\  ============================================================
 
0 [IF]

Object-oriented extensions from Neon

Original code from Andrew McKewan Version 1.0, 4 Feb 1997
mckewan@austin.finnigan.com
 

Additional code from Doug Hoffman Version 1.1, 6 Feb 2006
dhoffman@talkamerica.net

Added further extensions:

- private and public methods
- alternate late binding syntax as in Mops:    object message: **
- cascaded messages
- added EXM for executing ticked messages.
- modified mfa_offset to be faster as per Charles Melice
- added hooks for sending implicitly late-bound messages to VALUEs and LOCALS
  An example of how to do it in Carbon MacForth is provided.  See ?isValue and ?isLocal

\ Tested on Carbon MacForth, gForth, and SwiftForth

[THEN]

DECIMAL



\ =====================================================================
\ Misc words. You may already have some of these.

: HAVE  ( -- f )  BL WORD FIND NIP ;

have macforth.picture [IF]
\ just for Carbon MacForth, we must remove this name. Otherwise it will conflict
\ with the addr: selector
s" ADDR:"  FORTH-WORDLIST  >>FINDNAME
FORTH-WORDLIST >>REMOVEENTRY drop


\ Carbon MacForth version
: ?isLocal ( token -- t/f )
  info-find-token @infotype 13 = ;

\ Carbon MacForth version
 : ?isValue ( token -- t/f )
  info-find-token  @infotype 4 = ;
[ELSE]

: ?isLocal ( token -- t/f )
  drop false ;
  
 : ?isValue ( token -- t/f )
  drop false ;

[THEN]

1 CELLS CONSTANT CELL

: C+!  ( char addr -- )   DUP C@  ROT +  SWAP C! ;

: RESERVE  ( n -- )  HERE OVER ERASE ALLOT ;

\ Build link to list head at addr
: LINK  ( addr -- )   HERE  OVER @ ,  SWAP ! ;

: NOOP ;

HAVE PARSE 0= [IF]
: PARSE  ( -- addr len )  BL WORD COUNT ;
[THEN]

: move$ ( src$ptr\dest$ptr --) \ copy src to dest, dest must be long enough
 1 pick c@ 1+ cmove ;


\ =====================================================================
\ Class Structure.

0 VALUE ^Class  \ pointer to class being defined
0 VALUE newObject \ object being created
0 VALUE (ClassInit:) \ selector for ClassInit: message

\ The following offsets refer to the ^Class, or Pfa of the class.
\ The first 8 cells are the hashed method dictionary.

: IFA  8 CELLS + ; \ ivar dict Latest field
: DFA  9 CELLS + ; \ datalen of named ivars

\ IF XFA=0, then no class pointer is included in an ivar list
: XFA 10 CELLS + ; \ width of indexed area, <= 0 if not indexed
: SFA 11 CELLS + ; \ superclass ptr field
: TAG   12 CELLS + ; \ class tag field

13 CELLS CONSTANT classSize  \ size of class pfa

CREATE classTag \ contents of tag field for valid class


: ?isClass  ( pfa -- f )  \ is this a valid class?
 TAG @ classTag = ;

: ?isObj  ( pfa -- )  \ is this a valid object?
 @ DUP IF ?isClass THEN ;

: classAllot  ( n -- )  \ Allot space in the current class
 ^Class DFA +! ;

: classAlign  ( -- )  \ Align class data size (optional)
 ^Class DFA @  ALIGNED  ^Class DFA ! ;

: @width  ( ^class -- elWidth )  \ return the indexed element width for a class
 XFA @  0 MAX ;

\ Error if not compiling a new class definition
: ?Class   ^Class   0= ABORT" Not inside a class definition" ;

\ =====================================================================

\ Public and private methods, as in Mops

0 VALUE pub/priv

: public   -1 to pub/priv  ; \ following methods will be public = default
: private  0 to pub/priv  ; \ following methods will be private


\ =====================================================================
\ Objects have a pointer to their class stored in the first cell of
\ their pfa. When they execute, they return the address of the cell
\ following the class pointer, which is location of the first named
\ instance variable.
\
\ Object structure: | ^class | named ivars |
\
\ If an object is indexed, an indexed header appears after the data area.
\ This header consists of a cell containing the number of elements.
\ The indexed data follows this header.
\
\ Indexed object: | ^class | named ivars | #elems | indexed ivars |
\ =====================================================================

: (Obj)  ( -- )  CREATE  DOES> CELL+ ;


: >obj  ( xt -- ^obj )  \ get the object address from the execution token
 >BODY CELL+ ;

: >class  ( ^obj -- ^class )  \ get the class of an object
 CELL - @ ;


\ =====================================================================
\ Methods are stored in an 8-way linked-list from the MFA field.
\ Each method is identified by a 32-bit selector which is the parameter
\ field of the selector. Offsets are in cells.
\
\ Method Structure:
\ 0   link to next method
\ 1   selector
\ 2   method execution token (called mcfa below)
\ 3   flag  for pulic/private  -1 = public = default ,  0 = private
\ 4   ^class for which method was defined, used for private methods check
\ =====================================================================

\ Find the top of the method link for a given selector.
\ The "2/ 2/" below is to get a better distribution if the selectors
\ are aligned values.

\ : MFA  ( SelID ^Class -- SelID MFA ) \ original version
\ OVER 2/ 2/ 7 AND CELLS + ;

\ Faster version ala Charles Melice. It gives better distribution.
 : MFA  ( SelID ^Class -- SelID MFA )
  over dup >r
  2/ 2/
  r@ xor
  2/ 2/
  r> xor
  7 and
  cells + ;

0 [IF]
\ optional tool from Charles Melice
: CheckDistribution  ( xt -- )  \ usage:   ' var ( xt) CheckDistribution
    >BODY 0 LOCALS| n start |
    8 0 DO
        0 to n
        start I cells +
        begin @ ?dup while
            1 +to n
        repeat
        cr i . ." ->" n .
    LOOP ;
[THEN]

\ Search through a linked-list of (hashed) ivars for the given ivar.
\ This is called from vFind
: ((FINDIVAR))  ( ivarID IFA -- ivara true | false )
 BEGIN @ DUP
 WHILE 2DUP CELL+ @ = IF  2 CELLS + ( mfca )  NIP TRUE  EXIT THEN
 REPEAT NIP ;


0 value ^objClass

: checkClass ( addr -- t/f ) \ true if compiling in same class or a subclass of private method
 4 cells + @ >r \ store ^class of method on return stack
 ^Class \ start with ^class on stack for compares
 BEGIN
  dup 0= IF r> 2drop  false exit THEN  \ not compiling a class
  dup r@ = IF r> 2drop true exit THEN \ current class is either the same or a subclass of class where method in question was defined
  dup ^objClass = IF r> 2drop false exit THEN  \ superclass is object, end of superclass chain
  sfa @ ( get next ^superclass to the stack )
 AGAIN ;


 \  private/public methods version of ((FINDM))
: ((FINDM))  ( SelID MFA -- mcfa true | false ) \ abort on private message mis=use
 BEGIN @ DUP
 WHILE 2DUP CELL+ @ =
  IF \ we found the method, now check to see if it is public
       dup 3 CELLS + @ ( flag )
       IF
         \ It is a public method, so go ahead and use it.
         2 CELLS + ( mfca )  NIP TRUE  EXIT
       ELSE
         \ It is a private method, must check further.
           \ It is OK to use the private method IF we are defining a method in
           \ the same class as OR a subclass of the class in which the original
           \ private method was defined
           ( addr ) checkClass
          IF
           2 CELLS + ( mfca ) TRUE  EXIT
          ELSE
            true abort" private method!"
          THEN
       THEN
  THEN
 REPEAT NIP ;

: (FINDM)   ( SelID ^Class -- xt ) \ find selector in a class
 MFA ((FINDM)) IF  @ EXIT  THEN
 TRUE ABORT" Message not understood by class" ;

: FIND-METHOD   ( SelID ^obj -- ^obj xt )   \ find method in object's class
 TUCK >class (FINDM) ;


\ clFindM  returns the addr of a method's mcfa field, given a selID
\  and a class address.

: clFindM   ( selID ^class -- offs ^mcfa )
 MFA  ((findM)) IF  EXIT  THEN
 TRUE ABORT" Message not understood by class" ;


\ =====================================================================
\ Method execution. The current object address is store in the value ^base.
\ The object is only valid inside of a method definition. When we call a
\ method, we save the old object pointer and set it to the current object.
\ When the method returns, we restore the object pointer.

0 VALUE ^base  \ pointer to current object

: EXECUTE-METHOD  ( ^obj xt -- ) \ execute method, saving object pointer
 ^base >R   SWAP to ^base   EXECUTE   R> to ^base ;

: EXECUTE-IVAR  ( xt offset -- ) \ execute ivar method at given offset
 ^base >R  ^base + to ^base  EXECUTE  R> to ^base ;


\ Wrap catch so that it preserves the current object
HAVE CATCH [IF]
: CATCH  ( -- n )  ^base >R   CATCH   R> to ^base ;
[THEN]


\ =====================================================================
\ For late-bound method calls, we compile code to look up the method
\ at runtime and execute it. The syntax is:
\
\ Selector: [ object ]
\
\ The code between [ and ] must return an object reference. If the method
\ is not found in the class of the object, a runtime error occurs.
\ Because we use PARSE, [ and ] must be on the save source line.
\ =====================================================================

: (Defer)  ( ^obj SelId -- )  OVER >class (FINDM) EXECUTE-METHOD ;
: Defer,   ( SelId -- )  POSTPONE LITERAL  POSTPONE (Defer) ;

: Deferred  ( SelId -- )  \ Compile or execute a Deferred message send
 >R  [CHAR] ] PARSE  EVALUATE  R>
 STATE @ IF  Defer,  ELSE  (Defer)  THEN ;

\ True if string is "[" to start Deferred message send
: ?isLeftBracket  ( str -- f )   1+ C@ [CHAR] [ = ;


\ =====================================================================
\ Hash function for instance variable names. The "32 OR" is for
\ case-insensitive names. The compiler will warn you if you have
\ a hash collision.

\ original hash
\ : HASH  ( addr len -- n )
\  TUCK BOUNDS ?DO  5 LSHIFT  I c@ 32 OR XOR  LOOP
\  DUP 0< IF EXIT THEN  INVERT ;

 \ needed by Melice-style hash
 : LROT  ( x1 u -- x2 )
        2DUP LSHIFT >R  32 SWAP - RSHIFT R> OR ;

\ from Charles Melice
: HASH  ( addr u -- uh )
    TUCK OVER CHARS + SWAP  \ len addr+len addr 
    ?DO 5 LROT I C@ BL OR BL XOR BL 1- - XOR LOOP
    DUP 0< IF EXIT  INVERT THEN ;


: hash>  ( -- n )  BL WORD COUNT HASH ;


\ =====================================================================
\ Instance variable consists of five cell-sized fields. The fifth field
\ is used for indexed ivars only. Offsets are in cells.
\ 
\     Offset   Name      Description
\     ------   ----      ---------------------------------------
\        0     link      points to link of next ivar in chain
\        1     name      32-bit hash value of name
\        2     class     pointer to class
\        3     offset    offset in object to start of ivar data
\        4     #elem     number of elements (indexed ivars only)
\ 
\ In the stack diagrams, "ivar" refers to the starting address of this
\ structure.  The IFA field of a class points to the first ivar.
\ =====================================================================

: iclass     ( ivar -- 'class )   2 CELLS +   ;
: @IvarOffs  ( ivar -- offset )   3 CELLS + @ ;
: @IvarElems ( ivar -- #elems )   4 CELLS + @ ;


\ =====================================================================
\ Build SUPER and SELF pseudo ivars. These are always the last
\ two ivars in a class. When we define a class, we will patch the
\ class fields to the appropriate class and superclass.

CREATE ^Self
 0 ,  \ link
 hash> self , \ name
 0 ,  \ class
 0 ,    \ offset

CREATE ^Super
 ^Self ,  \ link
 hash> super , \ name
 0 ,  \ class
 0 ,    \ offset

\ Create a dummy class that "object" inherits from.

CREATE Meta classSize RESERVE

^Super   Meta IFA ! \ latest ivar
classTag Meta TAG ! \ class tag


\ support for cascaded messages
0 value doingCascade

\ we store the last $ptr following a selector here
\ a $ptr is a counted string or Pascal type str255
create lastWord 33 allot

: handle_| ( $ptr1 -- $ptr2 )
 locals| $ptr1 |
 false to doingCascade
 $ptr1 count ( addr len ) 1 = IF c@ [char] | = IF true to doingCascade THEN
        ELSE ( addr ) drop
        THEN
 doingCascade
 IF lastWord ( $ptr2 )
 ELSE $ptr1 lastWord move$ $ptr1
 THEN ;

\ =====================================================================
\ Determine if next word is an instance var.
\ Return pointer to class field in ivar structure if it is.

: vFind  ( str -- str false | ^iclass true )
 handle_| 
 ^Class
 IF DUP COUNT HASH ^Class IFA ((FINDIVAR))
  DUP IF  ROT DROP  THEN
 ELSE FALSE
 THEN ;

\ send ClassInit: message to ivar on stack
: InitIvar  ( ivar offset -- )
 OVER @IvarOffs + newObject +   ( ivar addr )
 (ClassInit:) ROT iclass @ (FINDM) EXECUTE-METHOD ;

: ClassInit  ( -- )  \ send ClassInit: to newObject
 newObject (ClassInit:)
 newObject >class (FINDM) EXECUTE-METHOD ;


\ =====================================================================
\ ITRAV traverses the tree of nested ivar definitions in a class,
\ building necessary class pointers and indexed area headers.

: ITRAV   ( ivar offset -- )  >R  ( ivar -- )
 BEGIN DUP ^Super <>
 WHILE DUP iclass @ IFA @
  OVER @IvarOffs R@ + RECURSE  ( initialize ivar's ivars )

  DUP iclass @  ( get ivar class )
  DUP XFA @  ( needs class pointer? )
  IF OVER @IvarOffs R@ + newObject +
   ( ivar ^Class ivarAddr -- )
   2DUP CELL - !   \ store class pointer
   OVER @width  ( indexed? )
   IF SWAP DFA @ +  \ addr of indexed area
    OVER @IvarElems  \ #elems
    SWAP !   \ upper array limit
   ELSE 2DROP
   THEN
  ELSE DROP
  THEN
  DUP R@ InitIvar  \ send ClassInit:
  @   \ next ivar in chain
 REPEAT R> 2DROP ;


\ =====================================================================
\ Compile an instance variable dictionary entry.
: <Var ( #elems ^Class | ^Class -- )
 BL WORD vFind ABORT" Duplicate instance variable"
 COUNT HASH     \ get hash value of name
 ALIGN ^Class IFA LINK ,   \ link & name
 DUP ,     \ class
 DUP XFA @  
 IF  CELL classAllot  THEN \ if indexed, allow for class ptr
 ^Class DFA @ ,    \ offset to ivar data
 DUP @width DUP
 IF  ROT DUP ,  * CELL+  THEN  \ #elems, cell for idx-hdr
 SWAP DFA @ + classAllot ;  \ Account for named ivar lengths


\ =====================================================================
\ Build an instance of a class. If we are inside a class definition, 
\ build an instance variable. Otherwise build a global object.

\ Compile the indexed data header into an object
: IDX-HDR   ( #elems ^Class | ^Class -- indlen )
 @width DUP IF  OVER , ( limit )  *  THEN ;

: (Build)   ( #elems ^Class | ^Class -- )
 ^Class
 IF <Var \ build an ivar if we are inside a class
 ELSE
  (Obj)   \ create object
  DUP >R ,  \ store class pointer
  HERE to newObject \ remember current object
  R@ DFA @ RESERVE \ allot space for ivars
  R@ IDX-HDR RESERVE \ allot space for indexed data
  R> IFA @ 0 ITRAV \ init instance variables
  ClassInit  \ send CLASSINIT: message
 THEN ;


\ =====================================================================
\ Build a class header with its superclass pointer.

: :Class  ( -- )
 CREATE 0 to ^Class
 DOES>  (Build) ;

: <Super  ( -- )
 HERE to ^Class   \ save current class
 classSize ALLOT   \ reserve rest of class data
 ' >BODY    \ pfa of superclass
 DUP ^Class classSize MOVE \ copy class data
 DUP ^Class SFA !  \ store pointer to superclass
 ^Super iclass !   \ store superclass in SUPER
 ^Class ^Self iclass !  \ store my class in SELF
 public \ set default method flag
  ;

: ;Class  ( -- )
 classAlign  \ align class data size (optional)
 0 ^Super iclass ! \ clear out super and self class pointers
 0 ^Self  iclass !
 0 to ^Class  \ clear class compiling flag
 ;


\ =====================================================================
\ Object Compiler. We rely on being able to classify the type of
\ object from it's execution token.
\
\
\ Key to instantiation actions
\ 1 = objType  defined as an object
\ 2 = classType  as a class
\ 3 = valueType  as a value
\ 4 = localType  as a local
\ 5 = leftBracketType  open bracket for defer group
\
\ =====================================================================

\ support for cascaded messages
0 value lastToken
0 value lastTokenID

: saveLast ( xt tokenID -- xt tokenID )
 2dup to lastTokenID  to lastToken ;

: ?isCascade ( $ptr -- t/f )
 count ( addr len ) 1 =
 IF c@ [char] | =
 ELSE drop false
 THEN ;

: ?is** ( *ptr -- t/f )
 count 2 = IF dup c@ [char] * = swap
      1+ c@ [char] * = and
     ELSE drop false
     THEN ;



\ ( str -- xt tokenID | tokenID )  Determine type of token referenced by string.
\ types 1-4 return an xt under the ID, types 5 and 6 return just the ID
: refToken
 DUP ?isCascade  IF drop lastToken lastTokenID EXIT THEN
 DUP ?isLeftBracket IF drop 5 EXIT   THEN
 DUP ?is**   IF drop 6 EXIT      THEN
 FIND 0= ABORT" undefined object"
 DUP ?isLocal   IF  4 saveLast EXIT  THEN
 DUP ?isValue   IF  3 saveLast EXIT  THEN
 DUP >BODY ?isObj IF  1 saveLast EXIT   THEN
 DUP >BODY ?isClass IF  2 saveLast EXIT   THEN
 TRUE ABORT" Invalid object type" ;

: (ivarRef)   ( xt offset -- )  \ compile ivar reference
 SWAP  POSTPONE LITERAL  POSTPONE LITERAL  POSTPONE EXECUTE-IVAR ;

: ivarRef   ( selID ^iclass -- ) \ compile ivar reference
 CELL+ FIND-METHOD  SWAP @  ( xt offset )  ?DUP
 IF (ivarRef)
 ELSE COMPILE,  ( optimize for offset zero )
 THEN ;

: callMethod  ( xt -- )   \ compile method call
 POSTPONE LITERAL  POSTPONE EXECUTE-METHOD ;

: (objRef)   ( SelID objCfa -- ) \ compile object reference
 >obj FIND-METHOD SWAP ( xt ^obj )
 POSTPONE LITERAL  callMethod ;

\ ( selID $str -- )  Build a reference to an object.
: compileRef
 refToken CASE
   1 ( object ) OF  (objRef)   ENDOF
   2 ( class  ) OF  >BODY (FINDM) callMethod ENDOF
   3 ( value  ) OF  COMPILE, Defer, ENDOF
   4 ( local  ) OF  COMPILE, Defer, ENDOF
   5 ( leftBracket ) OF   Deferred   ENDOF
   6 ( ** )   OF  ( ^obj selID )  defer,  ENDOF
 ENDCASE ;

\ ( selID $str -- )  Run time behavior.
: runRef
 refToken CASE
   1 ( object ) OF  >obj FIND-METHOD  ENDOF
   2 ( class  ) OF  >BODY (FINDM)  ENDOF
   3 ( value ) OF  EXECUTE  find-method  ENDOF
   5 ( leftBracket ) OF   Deferred EXIT  ENDOF
   6 ( ** )    OF  ( ^obj selID ) swap find-method ENDOF
 ENDCASE  EXECUTE-METHOD ;


\ =====================================================================
\ Selectors are immediate words that send a message to the object
\ that follows. The Yerk requirement that selectors end in ":" is
\ enforced here but not otherwise required by the implementation.

\ This is the message handler invoked by using a selector.
: handleMessage   ( SelID -- )
 BL WORD  STATE @
 IF vFind \ instance variable?
  IF    ivarRef \ compile ivar reference
  ELSE   compileRef \ compile object reference
  THEN
 ELSE runRef  \ run state - execute object ref
 THEN ;

: ?isSel  ( str -- flag )  \ true if word at addr is a selector xxx:
 DUP DUP C@ CHARS +  C@ [CHAR] :  =  SWAP C@ 1 > AND ;

: ?Selector ( -- )  \ Verify that following word is valid selector
 >IN @  BL WORD ?isSel 0= ABORT" Not a selector"  >IN ! ;

\ Create a selector that sends a message when executed.
: Selector  ( n -- )  ?Selector   \ selector needs a stack item ??? dbh
 CREATE IMMEDIATE  DOES> handleMessage ;

\ If the selector already exists, just return the existing selector,
\ otherwise create a new selector.
: getSelect  ( -- n )
 >IN @  BL WORD FIND
 IF >BODY NIP  ( already defined )
 ELSE DROP >IN ! Selector
  HERE
 THEN ;

Selector ClassInit:    getSelect ClassInit: to (ClassInit:)

\ =====================================================================
\ Build a methods dictionary entry. :M starts a method definition
\ by adding to the class method list and starting the compiler with
\ :NONAME. ;M ends a method and saves the method xt.

: :M ( -- )
 ?Class
 getSelect
 DUP ^Class MFA ((FINDM))  \ is method already defined?
 IF

  ^Class U> ABORT" Method redefined in same class"
 THEN
 ALIGN   \ align method
 ^Class MFA LINK  \ link into method chain
 ( SelID ) ,   \ name is selector's hashed value
 HERE 0 ,  \ save location for method xt
 pub/priv , \ store method type
 ^Class ,   \ store ^class
 :NONAME  ;  \ compile nameless definition

: ;M ( -- )  \ end a method definition
 ?Class  POSTPONE ;  ( here address from above should be on stack) SWAP ! ( save xt )  ; IMMEDIATE


\ =====================================================================
\ Build a new object on the heap for class. Use: Heap> className
\ gets heap, and returns ptr. Throws an error if not enough memory


: ?MEMERR  ( ior -- )  ABORT" Memory allocation error" ;

Selector Release: \ sent to an object before it is freed

: allocObj  ( size class -- )  \ allocate object and store in newObject
 OVER CELL+  \ allow for class ptr
 ALLOCATE ?MEMERR \ ( size class addr -- )
 DUP CELL+ to newObject  \ object address
 !   \ create the class ptr
 newObject SWAP ERASE ; \ clear to zero

: (heapObj)  ( #elems class | class -- ^obj )
 >R  ( save class on return stack )
 R@ DFA @  ( ivar data size )
 R@ @width ?DUP
 IF \ indexed object, add size of indexed area
  \ ( #elems size width -- )
  2 PICK * + ( indexed data )  CELL+ ( idxHdr )
  R@ allocObj
  newObject R@ DFA @ + !  ( store #elems in idxHdr )
 ELSE
  R@ allocObj \ non-indexed object
 THEN
 R> IFA @ 0 ITRAV \ initialize instance variables
 ClassInit  \ send ClassInit: message to new object
 newObject  ;  \ return object address

: HEAP>   ( -- addr )
 ' >BODY  DUP ?isClass 0= ABORT" Not a class"
 STATE @
 IF POSTPONE LITERAL  POSTPONE (heapObj)
 ELSE (heapObj)
 THEN ; IMMEDIATE

: <RELEASE  ( ^obj -- )  \ free heap object
 Release: [ DUP ] \ send Release: message to object
 CELL - FREE ?MEMERR ; \ deallocate it

\ Usage is   ' classname >heap or ['] classname >heap in a def.
: >heap ( ^class -- ^obj ) \ ^obj is in the heap
 >body (heapObj) ;


\ =====================================================================
\ Support for indexed instance variables. When object of these classes
\ are defined, the number of elements should be on the stack.

\ Set a class and its subclasses to indexed

: <Indexed  ( width -- )  ?Class  ^Class XFA !  ;

\ For some classes, we always want the class pointer to be stored whenever
\ the object is an instance variable, so that the object can receive late-
\ bound messages. This is already the case for indexed classes. Here we
\ fake it out by storing -1 into the indexed field. This is the reason
\ for the "0 MAX" in the definition of @width.

: <General  ( -- )  -1 <Indexed ;

\ self returns the same address as ^base, but is used for late bound calls
\ to the ivar, i.e.   Selector: [ self ]

: self  ( -- ) ?Class
 POSTPONE ^base ; IMMEDIATE

\ =====================================================================
\ Indexed primitives. These should be in code for best performance.

: idxBase  ( -- addr )  \ get base of idx data area
 ^base  DUP >class DFA @ +  CELL+ ;

: limit    ( -- n )  \ get idx limit (#elems)
 ^base  DUP >class DFA @ +  @ ;

: width    ( -- n )  \ width of an idx element
 ^base >class XFA @ ;

: ^elem   ( index -- addr ) \ get addr of idx element
 width *  idxBase + ;

\ Fast access to byte and cell arrays.
: At1  ( index -- char )   idxBase + C@ ;
: At4  ( index -- cell )   CELLS idxBase + @ ;

: To1  ( char index -- )   idxBase + C! ;
: To4  ( cell index -- )   CELLS idxBase + ! ;

: ++1  ( char index -- )   idxBase + C+! ;
: ++4  ( cell index -- )   CELLS idxBase + +! ;

\ Compute total length of object.
\ The length does not include class pointer.
: objlen  ( -- objlen )
 ^base >class DUP DFA @  ( non-indexed data )
 SWAP @width ?DUP
 IF  idxBase CELL - @ ( #elems ) * +  CELL+  THEN ;

\ =====================================================================
\ Runtime indexed range checking. Use +range and -range to turn range
\ checking on and off.

: ?range  ( index -- index )  \ range check
 DUP idxBase CELL - @ ( #elems )  U< IF EXIT THEN
 TRUE ABORT" Index out of range" ;

0 VALUE (?idx)  \ execution vector for range checking

: ?idx   (?idx) EXECUTE ;

: +range  ['] ?range to (?idx) ;  +range
: -range  ['] NOOP   to (?idx) ;

\ =====================================================================
\ Primatives for cell-sized objects.

: M@  ( -- n )  POSTPONE ^base  POSTPONE @ ; IMMEDIATE
: M!  ( n -- )  POSTPONE ^base  POSTPONE ! ; IMMEDIATE

\ =====================================================================
\ Define base class "object" from which all other classes inherit
\ Some of the common indexed methods are defined here.


:Class Object  <Super Meta

 :M Class:   ^base >class  ;M \ non-IX - leave class ptr
 :M Addr:    ^base         ;M \ get object address

 \ ( -- len )  Return total length of object
 :M  Length:    objlen      ;M

HAVE DUMP [IF]
 :M  Dump:     ^base objlen DUMP  ;M
[THEN]
 :M  Print:     ." Object@" ^base U.  ;M

 :M ClassInit:   ;M \ null method for object init
 :M Release:     ;M \ null method for object release

;Class

' object >body to ^objClass \ 12/18/05 dbh  for public/private methods

\ Bytes is used as the allocation primitive for basic classes
\ It creates an object of class Object that is n bytes long.
\ You can get the address by sending it an Addr: message.

: bytes  ( n -- )
 ?Class  ['] Object >BODY <Var  classAllot ;
 
0 value method-token
0 value obj-token

: putMparms ( method-token obj-token -- )
 to obj-token to method-token ;

: getMparms ( -- method-token obj-token )
 method-token obj-token ;
 
: exm ( method-token obj-token -- )
 swap >BODY swap
 >obj FIND-METHOD EXECUTE-METHOD ;

\ support for 2-dimensional arrays
\ Usage:   10 20 dimension <2array> theArray
\ Where:   <2array> is the class name of the array
\          theArray is the instance name of the array

    0 value ROWDIM
    0 value COLDIM

: DIMENSION
 to colDim   to rowDim
 colDim  rowDim  *  ;


\ =====================================================================
\ Load classes and test routines.


S" VAR11.F"    INCLUDED
S" ARRAY11.F"  INCLUDED
S" ORDERED-COL.F" INCLUDED
S" $X.F" INCLUDED
S" 2ARRAY.F"  INCLUDED
S" 2ARRAYGEN.F" INCLUDED
CR .( Classes loaded )

S" TEST11.F"   INCLUDED


