
+DOCGEN

( *! CListClass                                                         )
( **                                                                    )
( *T CList - lists for CIAO                                             )
( **                                                                    )
( *P A LIFO CList class that allows resizing and joining                )
( ** the first cell of the list pointed to by memptr holds the number   )
( ** of items in the list                                               )

class CList

Private:

      int memptr              \ holds pointer to allocated memory
      int size                \ size in cells

public:
      
      meth: init
      meth: Erase
      meth: Free
      meth: Add
      meth: Remove
      meth: items
      meth: FetchItem
      meth: Empty
      meth: ReSize
      meth: Join
      meth: Plist
      meth: GetSize
end-class

( **                                                                    )
( *H The CList class members                                            )
( **                                                                    )

: CList::Free           \ -- 
( *G frees the memory used by the list    
      memptr 0> 
      If
       memptr ::Free drop
       0 to size                               \ reset size
       0 to memptr                             \ clear memory referance
      Then
;

: CList::init           \ size -- f
( *G initialize the list prior to use                                   )

      Free                                      \ discard if allready inited
      to size                                   \ remember size
      size CELLS CELL+ Allocate >r to memptr    \ get memory
      0 memptr !                                \ set item count, first cell in list
      r>
      ;

: CList::Erase
      memptr size CELLS ::Erase
;
      

: CList::PList
( *G Print the list contents - diagnostic                               )
      Items 0 
      ?Do
        i FetchItem cr .
      Loop
;




: CList::Items          \ -- n
( *G return the number of items in the list                             )
      memptr @                                \ from first cell 
;

: CList::Add            \ item -- 
( *G add an item to the list                                            )
      Items 1+ size >
      If
       ." Failed to add item to List too many items"
      else
      memptr @ 1+ memptr !          \ bump item count 
      memptr @ CELLS memptr + !     \ add item
      Then

;            

: CList::Remove         \ -- item 
( *G remove an item from the list                                       )
      items 0= 
      If 
       ." No items in list"
       Else
        memptr Dup @ CELLS + @      \ get item
        memptr @ 1- memptr !        \ reduce item count
      Then
;



: CList::FetchItem            \ n -- item
( *G using a supplied index return a copy of that item                  )
      1+ CELLS memptr + @
;

: CList::Empty          \ -- 
( *G empty the list                                                     )
      0 memptr !                   \ set item count to zero
;

: CList::GetSize
( *G return the size of the list in items                               )
      Size
;

: CList::ReSize         \ n -- 
( *G resize the list, current contents are preserved                    )
      to size
      memptr
      If                                   \ list has been init'ed
       size CELLS CELL+ Allocate drop >r
       memptr r@ memptr @ CELLS CELL+ Move
       memptr ::Free Drop
       r> to memptr
       Else
        size CELLS CELL+ Allocate drop to memptr
      Then  
;

      
: CList::Join     \ *List --
( *G join will stick lists together, the size of the resulting list     )
( ** will be items1 + items2. to add any additional items you will need )
( ** to resize                                                          )

      { *CList -- }
      *CList (CList).Items ?Dup           \ size of list
      If
      Items +                             \ plus the size of the list to joint to
      Resize                              \ resize the list
      *CList (CList).Items 0
      Do
       i *CList (CList).FetchItem Add
      Loop
      Then
;


-DOCGEN

