\ DictStats.fth - word and dictionary statistics

((
Copyright (c) 1993, 2008, 2011, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20120715 SFP002 Replaced WID>TAB with WID-THREADS.
20120512 MPE001 Replaced "0 GetModuleHandle" with "FirstName".
))

only forth definitions
decimal

\ ===========
\ *! filename
\ *T Chapter Title
\ ===========

variable headerspace	\ -- addr
\ Holds accumulated size of headers.

(( del VFX build 3359
: wid>tab	\ wid -- addr len
\ Return thread table
  dup >threads swap >#threads @ cells
;
))

: (#words)	\ wid -- n
\ count number of words in vocabulary
\ wid threads point to control field.
\ link fields point to control field
  0  swap wid-threads bounds ?do	\ -- n ; start with count of 0 ; SFP002
    i				\ address of top nfa
    begin
      @ dup			\ get addr of next
     while			\ -- n ^ctl while there are more
      swap 1+ swap		\ increment counter
      ctrl>nfa			\ -- n nfa
      dup name> dup >link -	\ calculate length of header
      headerspace +!		\ add to headerspace total
      n>link			\ point at link to next
    repeat
    drop			\ ignore addr
  cell +loop
;

: #words	\ -- ; display number of words in vocabulary
  context @ dup cr .voc space 		\ say which vocabulary
  (#words)
  . ." words"				\ say how many
;

: #all-words	\ --
\ display statistics of words in vocabularies
  0 headerspace !
  0					\ initial word count total
  voc-link				\ for each vocabulary
  begin
    @ dup
   while				\ -- wordcount voccfa
    cr dup >name .name space 		\ show voc name
    dup (#words)			\ count words
    dup . ." words"			\ say how many
    rot + swap				\ accumulate total word count
    >voc-link				\ to next vocabulary
  repeat
  drop  cr . ." words in vocabularies"
;

: #all-wids	\ --
\ display statistics of word count and dictionary usage
  0 headerspace !
  0					\ initial word count total
  wid-link				\ for each wordlist
  begin
    @ dup
   while				\ -- wordcount wid
    cr  dup .voc	 		\ show wid/voc name
    dup (#words)			\ count words
    dup . ." words"			\ say how many
    rot + swap				\ accumulate total word count
    >wid-link				\ to next wid
  repeat
  drop  cr . ." words in total"
\  cr ." HERE = " here .
  cr ." Image size =        " here firstName - dup . >r
  cr ." Total header size = " headerspace @ .
  cr ."  equivalent to "
  headerspace @ #100 r>
  */  2 .r ." % of the image"
;

: .xx.yy	\ #thread #words --
  dup if
    #10000 swap */
    0 <# # # [char] . hold #s #> type
  else
    2drop ." Inf"
  endif
;

: #InThread	\ thread -- u ; num word in thread
  @ 0 swap begin			\ -- u link
    dup
   while
    swap 1+ swap  cell - @
  repeat
  drop
;

: .Balance	{ | #Min #Max NumThreads -- }
  0 -> #Max  $7FFF:FFFF -> #Min
  context @ >#threads @ -> NumThreads
  cr ." Number of threads = " NumThreads .
  0  context @ wid-threads bounds	\ SFP002
  ?do  i #InThread +  cell +loop
  cr ." Number of words = " dup .
  NumThreads 0 ?do			\ -- #words
    cr ." Thread # " i 3 .r space
    context @ >threads
    i cells + #InThread			\ -- #words #inThread
    dup #Min min -> #Min
    dup #Max max -> #Max
    dup 5 .r space space  over .xx.yy  ." %"
  loop
  cr ." Words in wordlist       = " .
  cr ." Max words in any thread = " #Max .
  cr ." Min words in any thread = " #Min .
  cr ." Ratio = " #Max #100 #Min 1 max */
  0 <# # # [char] . hold #s #> type
;


\ ********************
\ Hash test candidates
\ ********************

0 [if]
\ CRC-16 by calculation

((
: +CRC16	\ crc b -- crc'
\ *G Update CRC16 for one byte
  8 0 do
    2dup xor  1 and if
      swap 1 rshift $A001 xor swap	\ XOR in polynomial to crc
    Else
      swap 1 rshift swap		\ just shift crc
    Endif
    1 rshift				\ shift b
  Loop
  drop					\ discard b
;
))


\ CRC-16 using lookup table
\ User must set initial CRC to 0

create CRCTBL  #256 2* ALLOT

MARKER DISPOSE

: INITTABLE  ( -- )
  256 0 DO
    I  8 0 DO
      DUP 1 AND  >R
      1 RSHIFT  R>
      IF  $A001 XOR  THEN
    LOOP
    I 2*  CRCTBL + W!
  LOOP ;

INITTABLE  DISPOSE

: CRC-16  ( crc byt -- crc' )
  OVER  XOR  255 AND  2* CRCTBL + W@  SWAP  8 RSHIFT  XOR
;

: genCRC	\ crc addr len -- crc'
\ *G Generate a CRC for a block of memory, given an initial CRC
\ ** value.
  bounds				\ -- crc addr+len addr
\  ?do  i c@ +crc16  loop
  ?do  i c@ crc-16  loop
  $FFFF and
;


\ crc-16

defer CurrHash	\ wid nfa -- hash
\ Generate hash from counted string at nfa

: h16-1		\ wid nfa -- hash
  nip  0 swap  count genCRC
;

: h16-2		\ wid nfa -- hash
  nip  0 swap  count genCRC  2 rshift
;

: h16-3		\ wid nfa -- hash ; best so far
  nip  dup c@ 3 lshift swap  count genCRC
;

: h16-4		\ wid nfa -- hash
  nip  dup c@ 2 lshift swap  count genCRC
;

: h16-5		\ wid nfa -- hash
  nip  dup c@ 6 lshift swap  count genCRC
;

: h16-6		\ wid nfa -- hash
  nip  0 swap  dup c@ 1+ genCRC
;

: h16-7		\ wid nfa -- hash
  nip  dup c@ ( 1 lshift ) swap  count genCRC
;
assign h16-7 to-do currhash


\ *********************
\ Hash table processing
\ *********************

#1024 constant /Entries	\ -- n
\ Number of entries in hash table
/Entries cells buffer: HashTable

0 value /CurrThreads	\ -- n
\ number of threads
0 value CurrWidMask	\ -- mask
\ mask for number of threads

: DummyHash	\ wid nfa -- 0
\ dummy hash for testing
  2drop 0
;

: GenHash	\ wid nfa --
\ Run selected hash and update hash table
  CurrHash CurrWidMask and cells HashTable + incr
;

: RunHash	\ wid --
\ Generate hash table for a given thread.
\ wid threads point to control field.
\ link fields point to control field
  HashTable /Entries cells erase
  #512 dup to /CurrThreads  1- to CurrWidMask
((
  dup >#threads @
    dup to /CurrThreads
    1- to CurrWidMask
))
  dup wid-threads bounds ?do		\ -- wid ; SFP002
    i				\ address of top nfa
    begin
      @ dup			\ get addr of next
     while			\ -- wid ^ctl while there are more
      ctrl>nfa			\ -- wid nfa
      2dup GenHash
      n>link			\ point at link to next
    repeat
    drop			\ ignore addr
  cell +loop
  drop
;

: .HashTab	{ | #Min #Max -- }
  0 -> #Max  $7FFF:FFFF -> #Min
  cr ." Number of threads = " /CurrThreads .
  0  HashTable /CurrThreads cells bounds
  ?do  i @ +  cell +loop
  cr ." Number of words = " dup .
  /CurrThreads 0 ?do			\ -- #words
    cr ." Thread # " i 3 .r space
    HashTable i cells + @		\ -- #words #inThread
    dup #Min min -> #Min
    dup #Max max -> #Max
    dup 5 .r space space  over .xx.yy  ." %"
  loop
  cr ." Words in wordlist       = " .
  cr ." Max words in any thread = " #Max .
  cr ." Min words in any thread = " #Min .
  cr ." Ratio = " #Max #100 #Min 1 max */
  0 <# # # [char] . hold #s #> type
;

: .Balanced	\ --
  context @ RunHash  .HashTab
;

[then]


\ ======
\ *> ###
\ ======

decimal
