\ graphics.fth - Graphics for GTK in the Borland style.

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)

"Gdk-pixbuf currently supports PNG, XPM, JPEG, TIFF, PNM, RAS,
BMP, and even the patented GIF format"

To do
=====

Change history
==============
20110914 SFP003 First release.
20110909 RML003 Fixed shared graphics context problem.
20110909 SFP002 Fixed redraw problem with more callbacks.
20110821 RML001 First delivery
))

only forth definitions
decimal

\ ===========
\ *> gtkbindings
\ *S Graphics in the Borland style
\ ===========


\ **************
\ *N Global Data
\ **************

struct /gwindow	\ -- len
\ *G Structure to control a graphics window.
   int >winlink		\ link field
   int >GTKwin          \ window ID
   int >gc              \ graphics context - holds drawing color, line thickness etc
   int >pixmap          \ the pixmap where the drawing is done before display
   int >pixw		\ pixmap width
   int >pixh		\ pixmap height
\ -- start of GtkAllocation structure
   int >winx		\ window x position
   int >winy		\ window y position
   int >winw            \ window width
   int >winh            \ window height
\ -- end of GtkAllocation structure
   int >innerwindow     \ inner window - GTK refers to the client area inside the window (as we understand it) as a "window",
                        \ and we use it as our Gtk_Drawable which allows the pixmap to be displayed to the screen.
   int >dirty?          \ flag that causes the window to update itself
   int >filled?        \ internal flag to make the next command filled instead of stroked
   int >penx           \ internal x coordinate
   int >peny           \ internal y coordinate
end-struct

variable windows	\ -- addr
\ *G Anchors chain for keeping track of all created windows.
   windows off

CELL +USER CANVAS	\ -- addr
\ *G The current drawing window.

\ **********************
\ *N Internal operations
\ **********************

: window>	\ -- window
\ *G Get the current drawing window.
  CANVAS @ ;

: penx		\ -- addr
\ *G The current window's X coordinate for subsequent drawing
\ ** commands.
  window> >penx ;

: peny		\ -- addr
\ *G The current window's  Y coordinate for subsequent drawing
\ ** commands.
  window> >peny ;

: filled?	\ -- addr
\ *G The current window's internal flag affecting drawing commands;
\ ** see *\fo{filled}.
   window> >filled? ;

CREATE mycolor   0xffff0000 , 0xffff w, 0 w, 0 w,
\ *G Initial foreground color for new windows

: window-dims	\ window -- w h
\ *G Get the dimensions of a window from the window structure.
  dup >winw @ swap >winh @  ;

: load-pixbuf	\ zaddr -- pixbuf
\ *G Load an image into a Gtk_Pixbuf, which can be drawn to the
\ ** current window using *\fo{PUT}.
  gdk_pixbuf_new_from_file ;

\ : ->fggc 528 + @ ;
\ Hacky - fetches the foreground graphics context from a Gtk_Style.
\ The object that comes out is the point that lets us controlling the
\ current drawing color.

: redraw	{ window -- }
\ *G Make changes to a window's graphics visible.
  window dup >innerwindow @  swap >gc @
  window >pixmap @  0 0 0 0 window window-dims gdk_draw_drawable
;

: ?redraw   { window -- }
\ *G Internal - redraw a window only if it is "dirty" (affected
\ ** by any drawing commands). Immediately resets the window's
\ ** dirty flag, making it "clean" again
  window >dirty? @ if
    window >dirty? off   window redraw
  then
;

: ChainEach \ ... xt anchor -- ...
\ *G Execute xt on the contents of chain with the following structure:
\ *C   link | ...
\ *P The given xt must have the stack effect
\ *C   ... link -- ...
\ *P Where link is the address of the link field in the structure.
\ *P You can pass your own values to each link, just remember to clean up afterwards.
  begin
    @ dup
   while				\ -- xt link
    2dup 2>r swap execute 2r>
  repeat
  2drop
;

\ Callbacks

1 1 CCBproc: timeout_event_cb	\ 0 -- true ; -- entry
\ *G Callback run from a timer to redraw all "dirty" windows.
  drop
  ['] ?redraw windows ChainEach
  1
;

0 value gtimer	\ -- timer

: +gtimer	\ --
\ *G Start the graphics event timer.
  gtimer 0= if
    30 timeout_event_cb 0 g_timeout_add to gtimer
  endif
;

: -gtimer	\ --
\ *G Stop the graphics event timer
  gtimer if
    gtimer g_source_remove drop  0 to gtimer
  endif
;

: winResized	\ window --
\ *G Perform this when the window has been resized.
  { win | size[ 4 cells ] -- }
  win >GTKwin @ size[ gtk_widget_get_allocation	\ get window pos/size
  size[ 2 cells + @  win >winw !	\ width
  size[ 3 cells + @  win >winh !	\ height
  win >dirty? on			\ needs to be redrawn
;

3 0 CCBProc: GframeCallback	\ window event data -- ; -- entry
\ *G Callback to force window size to be updated and the window
\ ** redrawn.
\ cr dup cr .dword
  nip nip winresized
;

3 0 CCBProc: GExposeCallback	\ window event data -- ; -- entry
\ *G Callback to force window to be redrawn.
\ cr dup cr .dword
  nip nip  >dirty? on
;

((
\ Temporary - make these structs if possible
10 constant /gtkcolor
9 cells constant /gobject
))

: u+		\ x1 y x2 -- x1+x2 y
\ "Under-plus", add the number on the stack to the number 2 items down.
   rot + swap ;

: filled>	\ -- flag
\ *G Fetch the filled flag
  filled? @ filled? OFF  ;

: drawdest>	\ -- pixmap gc
\ *G Fetch the 2 objects from the current window that are passed
\ ** to all GDK graphics functions.
  window> >pixmap @  window> >gc @  ;

: dirty	\ --
\ *G Mark the current window as dirty, which signals the GUI's
\ ** internal timer to make changes to that window visible. Note
\ ** that the flag is reset by *\fo{redraw}.
  WINDOW> >DIRTY? ON  ;

: remove-window	\ window --
  windows DelLink ;

: add-window	\ window --
  windows AddLink  ;


\ ********************
\ *N Application words
\ ********************

: COLOR:	\ -- ; --
\ *G Builds a new GTK color. When the color is executed, the
\ ** foreground color is set. Use in the form:
\ *C  COLOR: red      0xffff0000 , 0xffff w, 0x0000 w, 0x0000 w,
  CREATE DOES> CANVAS @ >gc @ swap
  gdk_gc_set_rgb_fg_color ;
\  gdk_gc_set_foreground ;

\ *P The following colours are predefined:
\ *C  red     green   blue    yellow  orange  magenta
\ *C  cyan    white   black   ltgrey  grey
\ *C  dkred   dkgreen dkblue  dkyellow brown
\ *C  violet  dkcyan  dkgrey
\               argb-8888    r-16      g-16      b-16
COLOR: red      0xffff0000 , 0xffff w, 0x0000 w, 0x0000 w,
COLOR: green    0xff00ff00 , 0x0000 w, 0xffff w, 0x0000 w,
COLOR: blue     0xff0000ff , 0x0000 w, 0x0000 w, 0xffff w,
COLOR: yellow   0xffffff00 , 0xffff w, 0xffff w, 0x0000 w,
COLOR: orange   0xffff8800 , 0xffff w, 0x8888 w, 0x0000 w,
COLOR: magenta  0xffff00ff , 0xffff w, 0x0000 w, 0xffff w,
COLOR: cyan     0xff00ffff , 0x0000 w, 0xffff w, 0xffff w,
COLOR: white    0xffffffff , 0xffff w, 0xffff w, 0xffff w,
COLOR: black    0xff000000 , 0x0000 w, 0x0000 w, 0x0000 w,
COLOR: ltgrey   0xffcccccc , 0xcccc w, 0xcccc w, 0xcccc w,
COLOR: grey     0xff888888 , 0x8888 w, 0x8888 w, 0x8888 w,

COLOR: dkred    0xff880000 , 0x8888 w, 0x0000 w, 0x0000 w,
COLOR: dkgreen  0xff008800 , 0x0000 w, 0x8888 w, 0x0000 w,
COLOR: dkblue   0xff000088 , 0x0000 w, 0x0000 w, 0x8888 w,
COLOR: dkyellow 0xff888800 , 0x8888 w, 0x8888 w, 0x0000 w,
COLOR: brown    0xff884400 , 0x8888 w, 0x4444 w, 0x0000 w,
COLOR: violet   0xff880088 , 0x8888 w, 0x0000 w, 0x8888 w,
COLOR: dkcyan   0xff008888 , 0x0000 w, 0x8888 w, 0x8888 w,
COLOR: dkgrey   0xff444444 , 0x4444 w, 0x4444 w, 0x4444 w,

: onto		\ window --
\ *G Set the current target window structure for graphics commands.
  CANVAS !  ;

: pen		\ -- x y
\ *G Get the current drawing coordinates.
  penx @ peny @ ;

: at		\ x y --
\ *G Set the current drawing coordinates.
  peny ! penx ! ;

: filled	\ --
\ *G Makes the next command, such as rectangle or circle, filled
\ ** instead of stroked.
  filled? ON ;

: line		{ destx desty -- }
\ *G Draw a line from the pen to (destx,desty).
  drawdest>  pen  destx desty gdk_draw_line
  dirty
;

: lineto	\ destx desty --
\ *G Draw a line from the pen to (destx,desty) and set the pen to
\ ** (destx,desty).
  2dup line at ;

: linerel	\ dx dy --
\ *G Draw a line relative to the pen.
  pen u+ u+ line ;

: linerelto	\ dx dy --
\ *G Draw a line relative to the pen and move the pen to the end
\ ** of the line
  pen u+ u+ lineto ;

: ellipse	{ width height -- }
\ *G Draw an ellipse defined by width, height.
\ ** The ellipse is positioned such that the pen points to the
\ ** top left corner of an imaginary rectangle around the ellipse.
  drawdest>  filled>  pen  width height  0 360 64 *  gdk_draw_arc
  dirty
;

: circle	\ diameter --
\ *G Draw a circle. *\br{}
\ ** The circle is positioned such that the pen points to the top
\ ** left corner of an imaginary square around the circle.
  dup ellipse ;

: rectangle	{ width height -- }
\ *G Draw a rectangle.
  drawdest>  filled>   pen  width height   gdk_draw_rectangle
  dirty
;

: putpixel	\ --
\ *G Plot a single pixel.
  1 1 rectangle ;

: cleardevice	\ --
\ *G Clear the current drawing window using the current color.
   0 0 at window> window-dims filled rectangle ;

: put		{ pixbuf -- }
\ *G Draw a Gtk_Pixbuf to the current window at the current pen position.
  drawdest> pixbuf 0 0 pen pixbuf gdk_pixbuf_get_width
  pixbuf gdk_pixbuf_get_height 0 0 0  gdk_draw_pixbuf
  dirty
;

: gwin:		\ <name> -- ; -- window
\ *G Declare a named graphics window. The returned window is the
\ ** address of a *\fo{/gwindow} structure.
\ *C   gwin: MyWin    \ -- window
   /gwindow buffer: ;

: setupGwin	{ w h window -- }
\ *G Initialize a window control structure. This word is used to
\ ** create a new window. *\fo{SetupGwin} cannot be used with windows
\ ** defined in Glade.
  0 gtk_window_new window >GTKwin !	\ create new GtkWidget
  w window >winw ! h window >winh !
  window >GTKwin @ GTK_CONTAINER 10
                       gtk_container_set_border_width
  window >GTKwin @ GTK_OBJECT delete_event_name delete_event 0
                       gtk_signal_connect
  window >GTKwin @ GTK_OBJECT destroy_event_name destroy_event 0
                       gtk_signal_connect
  window >GTKwin @ w h gtk_window_resize
  window >GTKwin @     gtk_widget_show
;

: initGladeGwin	\ z$name builder window --
\ *G Use the Glade widget name (usually a drawing area) in the
\ ** Glade builder to set up the given *\fo{/gWindow} structure.
  {: zname builder win | size[ 4 cells ] -- :}
  builder zname gtk_builder_get_object
  win >GTKwin !				\ save widget
  win winResized
;

: initGwin	\ *widget gwindow --
\ *G Use the widget to set up the given *\fo{/gWindow} structure.
  tuck >GTKwin !  winResized
;

: addEvent	\ cbentry zname event window --
\ *G Add a callback to handle the name and event for a window
\ ** structure.
  {: cbentry zname event win -- :}
  win >GTKwin @ event gtk_widget_add_events
  win >GTKwin @ zname cbentry win g_signal_connect drop
;

: enable-graphics { window -- }
\ *G Enable the window for graphics operations and set the initial
\ ** state.
  window >gtkwin @     gtk_widget_get_window         window >innerwindow !
  window >innerwindow @ gdk_gc_new                   window >gc !
  window >gc @ mycolor gdk_gc_set_rgb_fg_color
  window >gc @ mycolor gdk_gc_set_foreground
  window >innerwindow @ window window-dims true
                       gdk_pixmap_new window         >pixmap !
  window add-window

  GframeCallback z" configure-event" GDK_CONFIGURE window addEvent
  GExposeCallback z" expose-event" GDK_EXPOSE window addEvent
;


\ ======
\ *> ###
\ ======

