\ gtkbindings.fth - bind to libgtk2 for GTK-style widgets

((
Copyright (c) 2008, 2011, 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20141006 SFP004 Updated documentation.
20110831 SFP003 Converted to MPE house style.
		Removed some O/S dependencies.
20110817 RML002 added "c" to bindings
20080325 VRW001 Initial release
))

\ ======
\ *! gtkbindings
\ *T Using libgtk2 widgets from Forth
\ ======
\ *P The source code for his chapter can be found in the directory
\ ** *\i{Examples/GTK}.
\ ** Additional directories provide tutorial code, including a
\ ** small editor.

\ *P GTK provides a GUI interface and graphical toolkit for many
\ ** operating systems. It also includes a GUI designer program
\ ** called *\i{Glade} whose output can be read and used by GTK
\ ** applications. This chapter is not a tutorial on GTK programming.

\ *P Several GTK+ tutorials exist, including
\ *C   http://developer.gnome.org/gtk-tutorial/

\ *P The book we used during development is
\ ** *\i{Foundations of GTK+ Development}, by Andrew Krause, published
\ ** by APress, ISBN 978-1-59059-793-4.

\ *P The file *\i{gtkbindings.fth} links to the libgtk library
\ ** and allows GTK calls to be made from Forth. This provides
\ ** VFX Forth with a powerful widget library that works on
\ ** Windows, Linux and Mac OS X.

\ *P The file *\i{gtktools.fth} provides basic functions including
\ ** the main GTK message loop.

\ *P The file *\i{graphics.fth} provides a graphics interface in
\ ** the style of the old Borland BGI interface. This is still
\ ** very useful for many applications.

\ *P The file *\i{GraphicsDemo.bld} loads all needed files
\ ** and provides a number of demonstrations, of which *\fo{ShowGraphics3}
\ ** is the most useful.

\ *P The file *\i{TextEd/editor.fth} contains a simple text editor
\ ** based on a tutorial by Micah Carrick.


\ ********************
\ *S External Linkages
\ ********************
\ *P The code loads required libraries and defines necessary API
\ ** calls. There are two points to be noted when using this code:
\ *(
\ *B The DLLs/libraries must be in the library search order when
\ ** VFX Forth runs or compiles this code.
\ *B Under Windows, we use the MinGW version of GTK+. Like
\ ** much Open Source software built with gcc, it uses a C calling
\ ** convention both for external functions and for callbacks.
\ *)

[undefined] CCB: [if]
[defined] Target_386_Windows [if]
: CCB:		\ #in #out "<name>" -- ; -- entry
\ *G Behaves as *\fo{CallBack:} but enforces a C parameter-passing
\ ** convention.
  fromC callback:  ;
: CCBproc:	\ #in #out "<name>" -- ; -- entry
\ *G Behaves as *\fo{CallProc:} but enforces a C parameter-passing
\ ** convention.
  fromC CallProc:  ;
[else]
: CCB:		\ #in #out "<name>" -- ; -- a-addr
  callback:  ;
: CCBproc:	\ #in #out "<name>" -- ; -- entry
  CallProc:  ;
[then]
[then]

\ =========
\ Libraries
\ =========

[defined] Target_386_Windows [if]
library: libgtk-win32-2.0-0.dll
library: libgobject-2.0-0.dll
library: libglib-2.0-0.dll
Library: libgdk-win32-2.0-0.dll
Library: libgdk_pixbuf-2.0-0.dll
Library: libpango-1.0-0.dll
\ Library: libpangocairo-1.0-0.dll
\ Library: libpangowin32-1.0-0.dll
[then]

[defined] Target_386_Linux [if]
isRTLD_Now_Global library: libgtk-x11-2.0.so
library: libgobject-2.0.so
library: libglib-2.0.so
Library: libgdk-x11-2.0.so.0
Library: libgdk_pixbuf-2.0.so.0
Library: libpango-1.0.so.0
also  types definitions
: uint  unsigned int  ;
previous definitions
[then]

[defined] Target_ARM_Linux [if]
isRTLD_Now_Global  library: libgtk-x11-2.0.so.0
library: libgobject-2.0.so.0
library: libglib-2.0.so.0
Library: libgdk-x11-2.0.so.0
Library: libgdk_pixbuf-2.0.so.0
Library: libpango-1.0.so.0
also  types definitions
: uint  unsigned int  ;
previous definitions
[then]

((
\ Generic OSX
[defined] Target_386_OSX [if]
library: libgtk-quartz-2.0.dylib
\ library: libgtk-x11-2.0.dylib
library: libgobject-2.0.dylib
library: libglib-2.0.dylib
Library: libgdk-quartz-2.0.dylib
\ Library: libgdk-x11-2.0.dylib
Library: libgdk_pixbuf-2.0.dylib
Library: libpango-1.0.dylib
[then]
))

\ MPE development OSX
[defined] Target_386_OSX [if]
library: /Users/gtkosx/gtk/inst/lib/libgtk-quartz-2.0.dylib
library: /Users/gtkosx/gtk/inst/lib/libgobject-2.0.dylib
library: /Users/gtkosx/gtk/inst/lib/libglib-2.0.dylib
Library: /Users/gtkosx/gtk/inst/lib/libgdk-quartz-2.0.dylib
Library: /Users/gtkosx/gtk/inst/lib/libgdk_pixbuf-2.0.dylib
Library: /Users/gtkosx/gtk/inst/lib/libpango-1.0.dylib
also  types definitions
: uint  unsigned int  ;
previous definitions
[then]


\ ============
\ Extend TYPES
\ ============

also types definitions
: GtkBuilder  void  ;
: GtkWidget   void  ;
: GtkWindow   void  ;
: GdkWindow   void  ;
: GtkTextBuffer void  ;
: GtkTextIter void  ;
: GObject     void  ;
: GError      void  ;
: gchar       char  ;
: gboolean    bool4  ;
: gsize       unsigned int  ;
: guint       unsigned int  ;
: gint        int  ;
: gpointer    void *  ;
previous definitions


\ =================
\ Primary GTK tools
\ =================

extern: void "c" gtk_init( int *, char *);
extern: void "c" gtk_main ( void );
extern: void "c" gtk_main_quit ( void );
extern: gboolean "C" gtk_events_pending( void );
extern: gboolean "C" gtk_main_iteration( void );
extern: gboolean "C" gtk_main_iteration_do( gboolean blocking );

extern: void "c" gtk_widget_show( void * widget );
extern: void "c" gtk_widget_show_all( void * widget );
extern: void "c" gtk_widget_hide( void * widget );
extern: void "c" gtk_widget_hide_all( void * widget );
extern: void "c" gtk_widget_destroy( int * );
extern: void * "c" gtk_widget_get_toplevel( void * );
extern: int * "c" gtk_button_new_with_label( char * );
extern: void "c" gtk_container_add( int * , int * );

extern: int "c" gtk_object_get_type( void );
extern: int "c" gtk_container_get_type( void );
extern: int "c" gtk_box_get_type( void );

extern: int "c" gtk_widget_get_type( void );
extern: void "C" gtk_widget_get_allocation(
  void * widget, void * allocation
);
extern: void "c" gtk_container_set_border_width(
  void * container, int border_width
);
\ ))
extern: void "C" gtk_widget_add_events( void * widget, int events );

extern: int * "c" gtk_widget_get_window( int *widget );
extern: int * "c" gtk_widget_get_style( int *widget );

extern: void "c" gtk_drawing_area_get_type( void );
extern: int * "c" gtk_drawing_area_new( void );
extern: void "c" gtk_drawing_area_size( void * drawingarea, int w, int h );
extern: void "C" gtk_widget_set_parent_window(
  GtkWidget * widget, GdkWindow * parent_window
);
Extern: void "C" gtk_widget_set_sensitive(
  GtkWidget * widget, gboolean sensitive
);


\ =======
\ Windows
\ =======

extern: int * "c" gtk_window_new( int );
extern: void "c" gtk_window_set_title(
  void * window, const char * title
);
extern: void "C" gtk_window_set_position(
  void * window, int position
);
extern: void "C" gtk_window_resize(
  void * window, int width, int height
);
extern: void "C" gtk_window_get_size(
  void * window, int *width, int *height
);
extern: void "C" gtk_window_set_default_icon_name( const gchar *name );
extern: int "c" gtk_window_get_type( void );
extern: void "c" gtk_window_resize( int * win, int width, int height );


\ =============
\ Memory Slices
\ =============

extern: void * "C" g_slice_alloc( gsize block_size );
extern: void "C" g_slice_free1( gsize block_size, gpointer mem_block );


\ =======
\ Dialogs
\ =======

\ Flags

1 constant GTK_DIALOG_MODAL
2 constant GTK_DIALOG_DESTROY_WITH_PARENT
4 constant GTK_DIALOG_NO_SEPARATOR

\ Message type

0 constant GTK_MESSAGE_INFO
1 constant GTK_MESSAGE_WARNING
2 constant GTK_MESSAGE_QUESTION
3 constant GTK_MESSAGE_ERROR
4 constant GTK_MESSAGE_OTHER

\ Buttons

0 constant GTK_BUTTONS_NONE
1 constant GTK_BUTTONS_OK
2 constant GTK_BUTTONS_CLOSE
3 constant GTK_BUTTONS_CANCEL
4 constant GTK_BUTTONS_YES_NO
5 constant GTK_BUTTONS_OK_CANCEL

\ Response codes

-1 constant GTK_RESPONSE_NONE

/* GTK won't return these unless you pass them in
 * as the response for an action widget. They are
 * for your convenience.
 */
-2 constant GTK_RESPONSE_REJECT
-3 constant GTK_RESPONSE_ACCEPT

/* If the dialog is deleted. */
-4 constant GTK_RESPONSE_DELETE_EVENT

/* These are returned from GTK dialogs, and you can also use them
 * yourself if you like.
 */
-5 constant GTK_RESPONSE_OK
-6 constant GTK_RESPONSE_CANCEL
-7 constant GTK_RESPONSE_CLOSE
-8 constant GTK_RESPONSE_YES
-9 constant GTK_RESPONSE_NO
-10 constant GTK_RESPONSE_APPLY
-11 constant GTK_RESPONSE_HELP

extern: void * "C" gtk_dialog_new( void );
extern: int "C" gtk_dialog_run( void * dialog );
extern: void "C" gtk_dialog_response(
  void * dialog, int response_id
);


\ ===========
\ Status Bars
\ ===========

Extern: GtkWidget * "C" gtk_statusbar_new( void );
extern: guint "C" gtk_statusbar_get_context_id(
  void * statusbar, const gchar * context_description
);
extern: guint "C" gtk_statusbar_push(
  void * statusbar, guint context_id, const gchar * text
);
extern: void "C" gtk_statusbar_pop(
  void * statusbar, guint context_id
);
(( Not on Linux
extern: void "C" gtk_statusbar_remove_all(
  void * statusbar, guint context_id
);
))


\ =====
\ Fonts
\ =====

extern: void "C" gtk_widget_modify_font(
  GtkWidget *widget, void * font_desc
);
extern: void * "C" pango_font_description_from_string( const char *str );
extern: void "C" pango_font_description_free( void *desc );


\ =====================
\ TextViews and Buffers
\ =====================

extern: gboolean "C" gtk_text_buffer_get_modified( void * buffer );
Extern: void "C" gtk_text_buffer_set_modified(
  GtkTextBuffer * buffer, gboolean setting
);
Extern: gchar * "C" gtk_text_buffer_get_text(
  GtkTextBuffer * buffer, GtkTextIter * start, GtkTextIter * end,
  gboolean include_hidden_chars
);
Extern: void "C" gtk_text_buffer_set_text(
  GtkTextBuffer * buffer, const gchar * text, gint len
);
extern: void * "C" gtk_text_view_get_buffer( void * text_view );

14 cells constant /GtkTextIter

Extern: void "C" gtk_text_buffer_get_start_iter(
  GtkTextBuffer * buffer, GtkTextIter * iter
);
Extern: void "C" gtk_text_buffer_get_end_iter(
  GtkTextBuffer * buffer, GtkTextIter * iter
);
Extern: void "C" gtk_text_buffer_get_bounds(
  GtkTextBuffer * buffer, GtkTextIter * start, GtkTextIter * end
);


\ =====
\ Files
\ =====

Extern: gboolean "C" g_file_set_contents(
  const gchar * filename, const gchar * contents, gsize length, GError ** error
);
Extern: gboolean "C" g_file_get_contents(
  const gchar * filename, gchar ** contents, gsize * length, GError ** error
);


\ =========
\ Clipboard
\ =========

Extern: void * "C" gtk_clipboard_get( guint selection );
Extern: void "C" gtk_text_buffer_cut_clipboard(
  void * buffer, void * clipboard, gboolean default_editable
);
Extern: void "C" gtk_text_buffer_copy_clipboard(
  void * buffer, void * clipboard
);
Extern: void "C" gtk_text_buffer_paste_clipboard(
  void * buffer, void * clipboard, void * override_location, gboolean default_editable
);
Extern: gboolean "C" gtk_text_buffer_delete_selection(
  void * buffer, gboolean interactive, gboolean default_editable
);


\ ===
\ GDK
\ ===

EXTERN: int * "c" gdk_pixmap_new(
  void *drawable, int width, int height, int depth
);
EXTERN: int * "c" gdk_pixbuf_get_from_drawable(
  void * dest, void * src, void * cmap,
  int src_x, int src_y, int dest_x, int dest_y,
  int width, int height
);

extern: int * "c" gtk_image_new_from_pixmap(
  void * pixmap, void * mask
);

EXTERN: void "c" gdk_draw_arc(
  void *drawable, void *gc, int filled,
  int x, int y, int width, int height,
  int angle1, int angle2
);

EXTERN: void "c" gdk_draw_drawable(
  int *drawable, int *gc, int *src,
  int xsrc, int ysrc, int xdest, int ydest,
  int width, int height
);

Extern: void * "C" gdk_gc_new( void * drawable );
EXTERN: void "c" gdk_gc_set_rgb_fg_color( int *gc, int *color );
EXTERN: void "c" gdk_gc_set_foreground( int *gc, int *color );

EXTERN: void "c" gdk_draw_line(
  int *drawable, int *gc,
  int x1_, int y1_, int x2_, int y2_
);

EXTERN: void "c" gdk_draw_rectangle(
  int *drawable, int *gc, int filled,
  int x, int y, int width, int height
);

extern: void * "c" gdk_pixbuf_new_from_file( const char *filename );
extern: int "c" gdk_pixbuf_get_width( void * pixbuf );
extern: int "c" gdk_pixbuf_get_height( void * pixbuf );
\ extern: void "c" gdk_pixmap_unref( void * pixmap );
extern: void "c" gdk_draw_pixbuf(
  void * drawable, void * gc, const void * pixbuf,
  int src_x, int src_y, int dest_x, int dest_y,
  int width, int height, int dither, int x_dither, int y_dither
);

extern: int "c" gdk_pixbuf_save(
  void * pixbuf, void * filename, void * type, void ** error,
  int zparam, int zvalue, int NULL
);

extern: void "C" gdk_threads_init( void );
extern: void "C" gdk_threads_enter( void );
extern: void "C" gdk_threads_leave( void );


\ ====
\ Misc
\ ====

#13 constant GDK_CONFIGURE
2 constant GDK_EXPOSE
((
typedef enum
{
  GDK_NOTHING		= -1,
  GDK_DELETE		= 0,
  GDK_DESTROY		= 1,
  GDK_MOTION_NOTIFY	= 3,
  GDK_BUTTON_PRESS	= 4,
  GDK_2BUTTON_PRESS	= 5,
  GDK_3BUTTON_PRESS	= 6,
  GDK_BUTTON_RELEASE	= 7,
  GDK_KEY_PRESS		= 8,
  GDK_KEY_RELEASE	= 9,
  GDK_ENTER_NOTIFY	= 10,
  GDK_LEAVE_NOTIFY	= 11,
  GDK_FOCUS_CHANGE	= 12,
  GDK_CONFIGURE		= 13,
  GDK_MAP		= 14,
  GDK_UNMAP		= 15,
  GDK_PROPERTY_NOTIFY	= 16,
  GDK_SELECTION_CLEAR	= 17,
  GDK_SELECTION_REQUEST = 18,
  GDK_SELECTION_NOTIFY	= 19,
  GDK_PROXIMITY_IN	= 20,
  GDK_PROXIMITY_OUT	= 21,
  GDK_DRAG_ENTER        = 22,
  GDK_DRAG_LEAVE        = 23,
  GDK_DRAG_MOTION       = 24,
  GDK_DRAG_STATUS       = 25,
  GDK_DROP_START        = 26,
  GDK_DROP_FINISHED     = 27,
  GDK_CLIENT_EVENT	= 28,
  GDK_VISIBILITY_NOTIFY = 29,
  GDK_NO_EXPOSE		= 30,
  GDK_SCROLL            = 31,
  GDK_WINDOW_STATE      = 32,
  GDK_SETTING           = 33,
  GDK_OWNER_CHANGE      = 34,
  GDK_GRAB_BROKEN       = 35,
  GDK_DAMAGE            = 36,
  GDK_EVENT_LAST        /* helper variable for decls */
} GdkEventType;
))

extern: void * "c" g_type_check_instance_cast(
  void * instance, int type
);

extern: int "c" g_timeout_add(
  int interval, void * function, void * data
);
extern: int "C" g_source_remove( uint tag );

\ extern: void * "c" g_main_context_new ();
extern: void * "c" g_main_context_default( void );
\ extern: void * "c" g_main_loop_new ( void * context, int );
\ extern: void   "c" g_main_loop_run ( void * loop );
extern: int "c" g_main_context_iteration(
  void * context, int may_block
);
extern: void "C" g_object_unref( void * object );
extern: void "C" g_clear_error( int ** err );
extern: void "C" g_error_free( int * err );
extern: void "C" g_free( gpointer mem );

extern: void * "C" gtk_builder_new( void );
extern: uint "C" gtk_builder_add_from_file(
  void * builder, char * filename, uint ** error
);
extern: void "C" gtk_builder_connect_signals_full(
  void * builder, void * func, void * user_data
);
extern: void * "C" gtk_builder_get_object(
  void * builder, const char * name
);

\ ********************
\ *S Signal Connection
\ ********************

\ ==============
\ *N Recommended
\ ==============

\ *[
extern: long "C" g_signal_connect_object(
  void * instance, char * detailed_signal, void * c_handler,
  void * gobject, uint connect_flags
);
extern: long "C" g_signal_connect_data(
  void * instance, char * detailed_signal, void * c_handler, void * data,
  int destroy_data, uint connect_flags
);

: g_signal_connect	\ instance detailed_signal c_handler data -- ulong
  0 0 g_signal_connect_data
;
\ *]

((
#define g_signal_connect_after(instance, detailed_signal, c_handler, data) \
    g_signal_connect_data ((instance), (detailed_signal), (c_handler), (data), NULL, G_CONNECT_AFTER)

#define g_signal_connect_swapped(instance, detailed_signal, c_handler, data) \
    g_signal_connect_data ((instance), (detailed_signal), (c_handler), (data), NULL, G_CONNECT_SWAPPED)
))


\ =============
\ *N Macros
\ =============
\ *P The GTK function *\b{gtk_signal_connect_full()} is the library
\ ** signal connection function; the C header file defines the other
\ ** connect-style functions as macros. These words create the
\ ** interface as it is more usually used.

\ *[
extern: unsigned int "c" gtk_signal_connect_full(
  int * object,
  char * name,
  int * func,
  int * unsupported,
  int * data,
  void * destroy_func,
  int object_signal,
  int after
);
\ *]
extern: unsigned int "c" gtk_signal_emit_by_name( void * object, void * name );

: gtk_signal_connect \ object name func data --
\ *G Connect an event handler to a signal.
  0 swap  0 0 0 gtk_signal_connect_full drop
;

: gtk_signal_connect_after \ object name func data --
\ *G Connect an event handler to a signal, but at the end of the
\ ** event chain.
  0 swap  0 0 1 gtk_signal_connect_full drop
;

: gtk_signal_connect_object   \ object name func slot_object
\ *G Connect an event handler which takes a *\i{GtkObject} pointer
\ ** as its argument to a signal.
  0 swap 0 1 0 gtk_signal_connect_full drop
;

: gtk_signal_connect_object_after   \ object name func slot_object
\ *G Connect an event handler which takes a *\i{GtkObject} pointer
\ ** as its argument to a signal, at the end of the event chain.
  0 swap 0 1 1 gtk_signal_connect_full drop
;


\ *************
\ *S GTK Types
\ *************
\ *P GTK operates on a number of different object types - and these are
\ ** frequently checked on call (with assertion failures if the wrong type is
\ ** used). These words should cause the object to be cast correctly.
\ *E \ object -- cast_object
\ ** GTK_OBJECT
\ ** GTK_CONTAINER
\ ** GTK_BOX
\ ** GTK_WIDGET
\ ** GTK_WINDOW

: GTK_OBJECT   \ object -- cast_object
  gtk_object_get_type g_type_check_instance_cast
;

: GTK_CONTAINER   \ object -- cast_container
  gtk_container_get_type g_type_check_instance_cast
;

: GTK_BOX   \ object -- cast_box
  gtk_box_get_type g_type_check_instance_cast
;

: GTK_WIDGET   \ object -- cast_widget
  gtk_widget_get_type g_type_check_instance_cast
;

: GTK_WINDOW   \ object -- cast_window
  gtk_window_get_type g_type_check_instance_cast
;

struct /GError	\ -- len
\ *G Definition of a *\b{GError} structure.
  int ge.domain				\ 32 bit
  int ge.code				\ C int
  int ge.*message			\ pointer to text
end-struct


\ *******************
\ *S GTK message pump
\ *******************

: gtk-step	\ -- flag
\ *G Iterate GTK's inner loop, process any events and exit immediately.
\ ** *\i{flag} is returned true if a GTK event was processed.
  g_main_context_default 0 g_main_context_iteration  ;

: gtk-step-blocking	\ -- flag
\ *G Iterate gtk's inner loop, process any events, wait until there are
\ **  events if none are waiting when the function was called.
  g_main_context_default -1 g_main_context_iteration  ;

: gtkPump	\ --
\ *G Empty the GTK event queue.
  begin  gtk_events_pending while  gtk_main_iteration drop  repeat
;

\ ==========
\ *N Windows
\ ==========

[defined] Target_386_Windows [if]
: gui-Idle	\ --
\ *G Wait for message or GTK events and process it/them.
  (Idle)  gtk-step-blocking DROP
;

: gui-BusyIdle	\ --
\ *G If a message or GTK events are available, process them.
  (BusyIdle)  gtk-step drop
;

: gui-EmptyIdle \ --
\ *G While messages and GTK events are available, process them.
  (EmptyIdle)  BEGIN gtk-step 0= UNTIL
;

: gui-AppIdle	\ --
\ *G Process messages until the *\b{GetMessage} API call returns
\ ** zero in response to a WM_QUIT message. The VFX Forth variable
\ ** *\fo{ExitCode} is set to the *\b{wParam} value of this message.
\ ** -- This version includes the GTK inner loop. --
  BEGIN (BusyIdle) gtk-step DROP AGAIN
;

: installGTKhooks	\ --
\ *G Install the GTK+ versions of the message pumps.
  ['] gui-Idle is Idle
  ['] gui-AppIdle is AppIdle
  ['] gui-BusyIdle is BusyIdle
  ['] gui-EmptyIdle is EmptyIdle
;
[then]


\ ========
\ *N Linux
\ ========

[defined] Target_386_Linux
[defined] Target_ARM_Linux or
[if]
: gui-WaitIdle	\ --
\ *G Wait for message or GTK events and process it/them.
  gtk-step-blocking DROP
;

: gui-BusyIdle	\ --
\ *G If a message or GTK events are available, process them.
  gtk-step drop
;

: gui-EmptyIdle \ --
\ *G While messages and GTK events are available, process them.
  BEGIN gtk-step 0= UNTIL
;

: installGTKhooks	\ --
\ *G Install the GTK+ versions of the message pumps.
  ['] gui-WaitIdle is WaitIdle
  ['] gui-BusyIdle is BusyIdle
  ['] gui-EmptyIdle is EmptyIdle
;
[then]


\ ===========
\ *N Mac OS X
\ ===========

[defined] Target_386_OSX [if]
: gui-WaitIdle	\ --
\ *G Wait for message or GTK events and process it/them.
  gtk-step-blocking DROP
;

: gui-BusyIdle	\ --
\ *G If a message or GTK events are available, process them.
  gtk-step drop
;

: gui-EmptyIdle \ --
\ *G While messages and GTK events are available, process them.
  BEGIN gtk-step 0= UNTIL
;

: installGTKhooks	\ --
\ *G Install the GTK+ versions of the message pumps.
  ['] gui-WaitIdle is WaitIdle
  ['] gui-BusyIdle is BusyIdle
  ['] gui-EmptyIdle is EmptyIdle
;
[then]


\ ********************************
\ *S Operating System Dependencies
\ ********************************
\ ========
\ *N Linux
\ ========
\ *P If GTK+ and Glade are not already installed, install them
\ ** using the system's package manager.

\ ==========
\ *N Windows
\ ==========
\ *P Under Windows, we use the MinGW version of GTK+. Like
\ ** much Open Source software built with gcc, it uses a C calling
\ ** convention both for external functions and for callbacks.

\ -------
\ *H GTK+
\ -------
\ *P We get our GTK + binaries from:
\ *C   http://www.gtk.org/download/win32.php

\ *P The Windows version of GTK+ used by MPE is supplied as a
\ ** directory tree that you place on your PC. During development,
\ ** the simplest way to work is to add the *\i{bin} folder to
\ ** your PC's system path. Depending on your version of Windows,
\ ** this is done from Control Panel -> System -> Advanced ->
\ ** Environment Variables. You may have to reboot after changing
\ ** the path.

\ *P If your application is being distributed into environments
\ ** controlled by an IT department, it may not be possible to
\ ** alter the system path. In this case make sure that the GTK
\ ** libraries and executables are in the same folder as your
\ ** application.

\ --------
\ *H Glade
\ --------
\ *P Glade is the GU designer for GTK+. At first, we tried to
\ ** install Glade to use the version of GTK+ used for application
\ ** run-time. This is a mistake! Installing a separate matched
\ ** Glade/GTK+ pair permits much more reliable development. MPE
\ ** gets Glade and GTK+ as one package from:
\ *C   http://ftp.gnome.org/pub/GNOME/binaries/win32/glade3/
\ *P If you are feeling adventurous, go to:
\ *C   http://glade.gnome.org/


\ ===========
\ *N Mac OS X
\ ===========
\ *P We use the GTK-OSX package. You can either use the package
\ ** built by MPE and supplied as *\i{gtkbins.tgz} or you can
\ ** work from the source package.

\ ** At application run time, it is usually sufficient to use
\ ** the prebuilt binaries if you use absolute paths to the
\ ** GTK shared libraries. For development, install from the
\ ** source package. One day, we'll find a suitable incantation
\ ** to run *\i{Glade} from the binary package.

\ -------------------------
\ *H Installing gtkbins.tgz
\ -------------------------
\ *P Launch a terminal, select your home folder and unpack
\ ** the package:
\ *E   tar -zxvf gtkbins.tgz
\ *P Executables will then be in *\i{~/gtkbins/bin} and shared
\ ** libraries in *\i{~/gtkbins/lib}. Edit *\i{gtkbindings.fth}
\ ** and other files such as *\i{gladetest.fth} so that library
\ ** references are correct.

\ -------------------------
\ *H Installing from source
\ -------------------------
\ *P During devlopment, it is convenient to use the source
\ ** distribution, although it is tedious to install.
\ *P If you want to use the *\i{Glade} designer and you already have
\ ** MacPorts or Fink installed, you will need to create a new user
\ ** and install GTK+ for that user in order to avoid conflicts.

\ *P Instructions for installing it are at:
\ *C  http://sourceforge.net/apps/trac/gtk-osx/wiki/Build

\ *P Read these instructions fully and carefully before you start.
\ ** In particular read the *\i{Snow Leopard} section and apply
\ ** the changes for a 32 bit build *\b{before} running *\i{jhbuild}.
\ ** Then perform the three main GTK+ builds:
\ *E $ jhbuild bootstrap
\ ** $ jhbuild build meta-gtk-osx-bootstrap
\ ** $ jhbuild build meta-gtk-osx-core

\ *P Then build the *\i{libglade} library needed at runtime and
\ ** the *\i{Glade} designer needed for development.
\ *E jhbuild build libglade
\ ** jhbuild build glade3

\ *P Executables will then be in *\i{~/gtk/inst/bin} and shared
\ ** libraries in *\i{~/gtk/inst/lib}. Edit *\i{gtkbindings.fth}
\ ** and other files such as *\i{gladetest.fth} so that library
\ ** references are correct.

\ *P In order to run *\i{Glade} properly, it should be launched by
\ *E $ jhbuild shell
\ ** $ glade-3 &
\ *P This incantation is required to persuade *\i{Glade} to find
\ ** its icons.


\ ======
\ *> ###
\ ======

decimal
