\ gtktools.fth - miscellaneous tools for GTK/GDK

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *> gtkbindings
\ ===========

\ ****************************
\ *S Loading GTK Builder files
\ ****************************

: zFindCallback	\ zCbName -- entry true | 0
\ *G Search the default search order for the given callback name.
\ ** On success, the entry point is returned.
  zcount search-context	if
    execute true
  else
    false
  endif
;

7 0 CCBproc: BuilderConnect_cb	\ *Builder *object zSignal zHandler *ConnObj flags *User --
\ *G The main Glade signal connection callback. This is called for
\ ** each signal handler specified in the Glade file, i.e. handlers
\ ** named by the GUI designer. Handler names are looked up in the
\ ** current context. The words found are linked to the appropriate
\ ** object by way of the GTK signal connection bindings.
  { *Builder *object zSignal zHandler *ConnObj flags *User | cback -- }
  zHandler zFindCallback if
    -> cback  *ConnObj if
      *object zSignal cback *ConnObj flags
      g_signal_connect_object
    else
      *object zSignal cback *User 0 flags
      g_signal_connect_data
    endif
  else
    ." Cannot find callback " zHandler .z$ ."  for " zSignal .z$ cr
  endif
;

variable CurrBuilder	\ -- addr
\ *G Holds the current builder handle.

: loadBuilderXML	\ z$ -- builder|0
\ *G Load a Glade GtkBuilder XML file and return the pointer or
\ ** zero on error. On success, the connections are made and the
\ ** variable *\fo{CurrBuilder} is set.
  {: | be[ cell ] -- :}			\ cell for error variable
  gtk_builder_new dup if		\ -- zname builder
    be[ off				\ needed on some OSes
    dup rot be[ gtk_builder_add_from_file if \ -- builder
      dup BuilderConnect_cb 0 gtk_builder_connect_signals_full
    else
      be[ @ 2 cells + @ .z$		\ "meaningful" error string
      be[ @ g_error_free
      g_object_unref  0
    endif
  else
    nip
  endif
  dup CurrBuilder !
;

: BuilderObject	\ z$ -- *widget
\ *G Obtain the widget pointer for the given named widget from the
\ ** current builder object.
  CurrBuilder @ swap  gtk_builder_get_object GTK_WIDGET
;

: freeBuilder	\ --
\ *G Free the current builder.
  CurrBuilder @ g_object_unref  0 CurrBuilder !
;


\ **********
\ *S Dialogs
\ **********

: runDlg	\ dialog -- response
\ *G Given a *\b{*GtkDialog}, run the dialog as a modal dialog
\ ** and return the result.
  dup >r gtk_widget_show_all
  r@ gtk_dialog_run
  r> gtk_widget_hide_all
;

AliasedExtern: NewErrDlg GtkWidget * "C" gtk_message_dialog_new(
  GtkWindow * parent, uint flags, uint type, uint buttons, const gchar * message
);

: ErrorBox	\ zMessage zTitle parent --
\ *G Displays an error box with the given message, title and
\ ** parent window.
  { zMessage zTitle parent | dlg -- }
  parent
  GTK_DIALOG_MODAL GTK_DIALOG_DESTROY_WITH_PARENT or
  GTK_MESSAGE_ERROR
  GTK_BUTTONS_OK
  zMessage
  NewErrDlg to dlg
  dlg zTitle gtk_window_set_title
  dlg gtk_dialog_run drop
  dlg gtk_widget_destroy
;

0 constant GTK_FILE_CHOOSER_ACTION_OPEN
1 constant GTK_FILE_CHOOSER_ACTION_SAVE
2 constant GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER
3 constant GTK_FILE_CHOOSER_ACTION_CREATE_FOLDER

extern: gchar * "C" gtk_file_chooser_get_filename( void * chooser );
extern: gboolean "C" gtk_file_chooser_set_filename(
  void * chooser, const char * filename
);
AliasedExtern: NewFileNameDlg GtkWidget * "C" gtk_file_chooser_dialog_new(
  gchar * title, GtkWindow * parent, uint action,
  gchar * button1_text, guint resp1,
  gchar * button2_text, guint resp2,
  int foo=0
);

: NewSaveDlg	\ parent -- dlg
\ Create a new file chooser dialogue for saving a file.
  z" Save File ..." swap GTK_FILE_CHOOSER_ACTION_SAVE
  z" gtk-cancel" GTK_RESPONSE_CANCEL
  z" gtk-save" GTK_RESPONSE_OK
  0 NewFileNameDlg
;

: NewOpenDlg	\ parent -- dlg
\ Create a new file chooser dialogue for opening a file.
  z" Open File ..." swap GTK_FILE_CHOOSER_ACTION_OPEN
  z" gtk-cancel" GTK_RESPONSE_CANCEL
  z" gtk-open" GTK_RESPONSE_OK
  0 NewFileNameDlg
;

: RunFileChooser	\ zaddr len dlg -- flag
  dup >r gtk_dialog_run
  GTK_RESPONSE_OK = if			\ -- zaddr len ; R: -- dlg
    r@ gtk_file_chooser_get_filename
    dup >r
    zcount rot 1- min  rot zplace  -1
    r> g_free
  else
    2drop 0
  endif
  r> gtk_widget_destroy
;

: AskSaveFileNameBox	\ zaddr len parent -- flag
\ *G Given a buffer and parent window, ask for a file name. On
\ ** success, the filename is returned (clipped as required) in
\ ** the buffer as a zero terminated string. If no file name is
\ ** given the buffer is left unchanged.
  NewSaveDlg >r				\ -- zaddr len ; R: -- dlg
  over c@ if				\ if an existing file
    over r@ swap gtk_file_chooser_set_filename drop
  endif
  r> RunFileChooser
;

: AskOpenFileNameBox	\ zaddr len parent -- flag
\ *G Given a buffer and parent window, ask for an existing file name. On
\ ** success, the filename is returned (clipped as required) in
\ ** the buffer as a zero terminated string. If no file name is
\ ** given the buffer is left unchanged.
  NewOpenDlg RunFileChooser
;


\ ******************
\ *S Event Callbacks
\ ******************
\ *P The GTK library calls back into Forth on various events. Here
\ ** are some sample definitions. You may need to modify them or
\ ** add new ones according to the complexity of your UI. You can
\ ** also use this code to test your GTK installation.

\ *P Compile the file *\i{gtkbindings.fth}. Then run
\ *C  .libs  .badExterns
\ *P This will display the load addresses of the libraries and
\ ** show you any unresolved external functions. A library value
\ ** of zero indicates that the library has not been found.
\ ** If everything is good at this point, run the test window:
\ *C   gtktest
\ *P which should display a "hello world" window.

: delete_event_fn \ *widget *event *data -- 0/1
\ *G Handles delete events. It returns zero so that the widget
\ ** is destroyed.
\  3drop 1           \ non-zero means don't destroy
  3drop 0
;
3 1 CCB: delete_event \ -- addr ; *widget *event *data -- 0/1
\ *G Callback for the delete event. Return 0 to perform the
\ ** default handling or return 1 to indicate that the callback
\ ** has done everything necessary.
  ' delete_event_fn to-callback delete_event

create delete_event_name
   z", delete_event"

: destroy_fn         \ *widget data --
\ *G Handles the destroy event for the application. It calls
\ ** *\b{gtk_main_quit()}.
  2drop gtk_main_quit
;
2 0 CCB: destroy_event
\ *G Callback for the destroy event.
  ' destroy_fn to-callback destroy_event

create destroy_event_name z", destroy"

create clicked_event_name z", clicked"

: wd_destroy	\ addr --
\ *G Destroy the widget
  gtk_widget_destroy
;
1 0 CCB: widget_destroy_cb	\ -- entry
\ *G Callback to handle the destroy event.
  ' wd_destroy to-callback widget_destroy_cb


\ ***************************
\ *S GTK startup and shutdown
\ ***************************

0 value GTKstarted?	\ -- x
\ *G Non-zero when GTK has been started.
0 value AppFinished?	\ -- x
\ *G Non-zero when app must quit.
0 value gtk_main?	\ -- x
\ *G Non-zero if *\fo{gtk_main} is in use.

: noVfxGtk	\ --
\ *G Mark the VFX Forth to GTK interface as unused.
  0 to GTKstarted?  0 to AppFinished?  0 to gtk_main?
;  ' noVfxGtk AtCold

: do_gtk_init	{ | temp[ cell ] -- }
\ *G Run *\b{gtk_init}.
  0 temp[ !  temp[ 0 gtk_init
;

: do_gtk_main	\ --
\ *G The tools version of *\fo{gtk_main}.
  -1 to gtk_main?  gtk_main
;

: initGTK	\ --
\ *G Call this to initialise the GTK system. Always call this
\ ** word to start the GTK system.
  GTKstarted? 0= if
    0 to AppFinished?  0 to gtk_main?
    do_gtk_init  installGTKhooks
    -1 to GTKstarted?
  endif
;

: GtkAppQuit	\ --
\ *G The GTK+ app should finish.
  -1 to AppFinished?  gtk_main?
  if  gtk_main_quit  endif		\ comment out if gtk_main not used
;


\ ****************
\ *S GTK test code
\ ****************

0 [if]

: hello_world_window	\ --
\ *G Launch the "hello world" window.
  0 gtk_window_new			\ -- window
  dup GTK_CONTAINER 10 gtk_container_set_border_width
  dup GTK_OBJECT delete_event_name delete_event 0
                     gtk_signal_connect \ -- window
  dup GTK_OBJECT destroy_event_name destroy_event 0
                     gtk_signal_connect \ -- window

  z" Hello, world" gtk_button_new_with_label \ -- window button

  2dup GTK_OBJECT swap GTK_OBJECT clicked_event_name widget_destroy_cb
  rot gtk_signal_connect_object

  2dup gtk_container_add            \ -- window button

  gtk_widget_show ( button )
  gtk_widget_show ( window )
;

: gtktest	\ --
\ *G Start GTK+, launch the hello world window, and wait until it
\ ** closes.
  do_gtk_init  hello_world_window  gtk_main
;

: hw		\ --
\ *G As *\fo{gtktest} but does not initialise GTK+ again.
  hello_world_window
  gtk_main
;

[then]


\ ======
\ *> ###
\ ======

decimal
