( *T I2C Base driver code 				)
( ** This file is always required as it defines the base protocol)

((
Copyright (c) 1999

MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com
))


\ ------------------------------------------------------------------------------
  ( *S Start Condition )
\ ( ** Bring SDA high-low while SCL is high )
\ ------------------------------------------------------------------------------

: start_i2c   	\ --
  sda_high		\		; set SDA HIGH
  scl_high		\		; set SCL HIGH
  sda_low		\		; set SDA LOW
;

\ ------------------------------------------------------------------------------
  ( *S Stop Condition )
\ ( ** Bring SDA low-high while SCL is high )
\ ------------------------------------------------------------------------------

: stop_i2c  	\ --
  sda_low		\		; set SDA LOW
  scl_high		\	        ; set SCL HIGH
  sda_high		\	        ; set SDA HIGH
;

\ ------------------------------------------------------------------------------
  ( *S Acknowledge )
\ ( ** Read acknowledgement bit, by storing in a byte and masking off LSB )
\ ------------------------------------------------------------------------------

: acknowledge_i2c  	\ -- ; check acknowledge bit, abort if 1
  sda_high      	\ 		; set SDA HIGH
  scl_high      	\  	        ; set SCL HIGH
  sda@	 	        \ -- sda  	; read the current state of sda (ack_bit) and leaves on stack
  if
    stop_i2c		\		; make I2C lines safe
    true abort" I2C acknowledge error"
  endif
  scl_low      		\ -- sda   	; set SCL LOW
;

\ ------------------------------------------------------------------------------
\ ( *S read a byte )
  ( ** MSB first )
\ ------------------------------------------------------------------------------

: read_byte_i2c		\ -- in_byte
 scl_low		 	\		; set scl LOW
  0	        		\ -- 0 		; put blank byte on the stack - will be used to hold the data byte
  8 0 do			\ -- 0 		; loop for 8 times to read in each bit of byte
    scl_high			\ -- 0 		; set SCL HIGH
    1 lshift sda@ OR	 	\ -- in_bits	; shift stack left 1, read SDA, OR both
    scl_low 			\ -- in_byte	; set SCL LOW
  loop				\        	; end the loop
;

\ ------------------------------------------------------------------------------
\ ( *S write a byte )
  ( ** Shift byte out with MSB first )
\ ------------------------------------------------------------------------------

: write_byte_i2c	\ byte --
  8 0 do			\ 	        	; loop 8 times
   scl_low			\			; set scl LOW
    dup $080 AND		\ -- byte 	    	; duplictae stack and AND with MSB
    0 = if                      \ -- byte 		; if MSB = 0
      sda_low  		        \ -- byte 		; set SDA LOW
    else  		        \ -- byte 		; if MSB = 1
      sda_high  		\ -- byte 		; set SDA HIGH
    endif			\ -- byte 	        ; end if statement
    scl_high  			\ -- byte 		; set SCL high
    1 lshift			\ -- byte<<1 	        ; shift byte left by 1
  loop				\ -- byte<<8 		; end loop
  drop				\ 		        ; drop the byte from stack
  scl_low		        \ 		        ; set SCL LOW
  acknowledge_i2c		\			; read acknowledge bit
 ;

 \ ------------------------------------------------------------------------------