\ OnGosub.fth - executes from list

: xtList,	\ wid --
\ *G Creates a table of xts. The wid is of a wordlist containing
\ ** words which are executed rather than compiled. These words
\ ** return true if the list is complete, or zero if there's
\ ** more to come. N.B. Immediate words are not executed but
\ ** are compiled.
  >r  begin
    bl get-word count
    2dup r@ search-wordlist if		\ -- caddr len xt ; in special voc
      nip nip  execute			\ -- flag
    else				\ -- caddr len
      search-context ?undef ,  0	\ must be in search order
    endif
  until
  r> drop
;

vocabulary ongosub.voc	\ --
\ *G Contains special words for *\fo{ONGOSUB}. Words in this
\ ** vocabulary return zero to continue compilation of the
\ ** *\fo{ONGOSUB ... END-ONGOSUB} structure or non-zero to
\ ** finish compilation of the structure.

also ongosub.voc definitions
\ Do not put comments after these words!
: \		postpone \  0  ;
: (		postpone (  0  ;
: ((		postpone ((  0  ;
: end-ongosub	-1  ;
previous definitions	\ )) ) for syntax highlighters

: (ongosub)	\ n --
\ *G Execute nth (zero-based) word in the following table.
  r> dup @ >r				\ adjust return address
  swap cells + cell +			\ index + base address
  @ execute
;

: ongosub	\ C: -- ; I: n --
\ *G Compiles a table of xts. At run-time the nth item is executed.
  ?comp discard-sinline			\ we will lay an address literal
  postpone (ongosub)  here 0 ,		\ run-time and exit address
  ['] ongosub.voc voc>wid xtList,	\ build list
  here swap !				\ patch exit address
; immediate

\ ONGOSUB test code
: tog
  ongosub
    vocs
    words
    .sources
( -1 ) \ test me
    false true
  end-ongosub
;



vocabulary atomlist.voc                                                     \ --
\ *G Contains special words for *\fo{ATOMLIST}. Words in this
\ ** vocabulary return zero to continue compilation of the
\ ** *\fo{ATOMLIST ... ATOM-LIST-END} structure or non-zero to
\ ** finish compilation of the structure.

also atomlist.voc definitions
: \ postpone \ #0 ;
: ((  postpone ((  #0  ;
: ( postpone (  #0  ;
: ATOM-LIST-END -1  ;
previous definitions	\ ) ))

-999 constant CountAtomsSwitch

: (atom-list)                                                             \ n --
\ *G Execute nth (zero-based) word in the following table.
  dup 0>
  IF
    r> dup @ >r                                          \ adjust return address
    swap cells +                                          \ index + base address
    @ execute
  ELSE
    CountAtomsSwitch =
    IF
        r> dup @ dup >r                                  \ adjust return address
        swap - cell - cell u/
    ELSE
        DROP true abort" Illegal atom number"
    THEN
  THEN
  ;

: ATOM-LIST                                                    \ C: -- ; I: n --
\ *G Compiles a table of xts. At run-time the nth item is executed.
  ?comp discard-sinline                         \ we will lay an address literal
  postpone (atom-list)  here 0 ,                     \ run-time and exit address
  ['] atomlist.voc voc>wid xtList,                                  \ build list
  here swap !                                               \ patch exit address
  ; immediate

: findAtomRun	\ addr -- addr'|0
\ *G Given the address/xt of a word containg an ATOM-LIST, return
\ ** the address just after the call to (ATOM-LIST). This is followed
\ ** by the end address and a number of cells containing xts.
  #1024 bounds do
    i c@ $E8 = if
      i getCallDest ['] (atom-list) = if
        i 5 +  unloop exit
      endif
    endif
  loop
  0
;

: atomAction	\ n xt -- xt'
\ *G Given an index and the xt of a word containing an ATOM_LIST
\ ** structure, return the xt of the word executed for that index.
  findAtomRun dup if
    swap cells + @
  else
    nip
  endif
;

((
\ ATOM_LIST test code
: tal
  500 0 do noop noop loop                                                                  ( --- )
  ATOM-LIST
    vocs
    words
    .sources
( -1 ) \ test me
    false true
  ATOM-LIST-END
  ;
))

\ *************************
\ *S Disassembler extension
\ *************************

also disassembler definitions

: disSubItem	\ --
\ Disassebly of item/xt in list.
  disinit @long disclose
  s"   " >disline
  >name count >DisLine  .DisLine
;

: disSubList	\ dest --
\ Disassemble a list.
  begin
    dup DisAddr <>
   while
    disSubItem
  repeat
  drop
;

: disListStart	\ caddr len -- dest
\ Disassemble the list, returning the end address.
  disinit @long -rot disclose		\ -- dest
  >disLine .disLine
;

: disOnGosub	\ --
\ Disassemble an *\fo{ONGOSUB} structure.
  disclose .disline			\ the call to (ONGOSUB)
  s" ONGOSUB" disListStart
  disSubList
  disinit s" END-ONGOSUB" >disline
;

: disAtomList	\ --
\ Disassemble an *\fo{ATOM-LIST} structure.
  disclose .disline			\ the call to (ATOM-LIST)
  s" ATOM-LIST" disListStart
  disSubList
  disinit s" END-ATOM-LIST" >disline
;

[+switch dasm-switch
  ' (ongosub)   runs disOngosub
  ' (atom-list) runs disAtomList
switch]

previous definitions

