\ STATEMACHINE.FTH - finite state machine code - PC host version

((
Original scheme
===============
Reference: J. Basile
The Journal of Forth Application and Research Vol 1 Number 2

Modified by
===========
Kurt Heinz, Synics AG

Modified again
==============
for MPE VFX Forth

5 fsm: game
  1 state:
    char f 2 action: .mes1
    char x 1 action: .grr
           1 default: .mes5
  2 state:
    char o 3 action: .mes2
           1 default: .mes5
  3 state:
    char r 4 action: .mes3
           1 default: .mes5  .s
  4 state:
    char t 1 action: .mes4
    char T 3 action: .mes2
    char q 5 action: drop
           1 default: .mes5

The phrase n FSM: <name> creates a new finite state machine called <name>,
with n states, numbered 1..n. An FSM consists of a structure:
  cell		current state (pointer into this structure)
    followed by n state entries, consisting of
  cell          pointer to state data

Each state in turn consists of the following structure
  cell		number of valid inputs
    followed by a structure for each valid input
  cell		input value
  cell		next state
  cell		xt of action for this input
    followed by a default
  cell		dummy
  cell		next state
  cell		xt of default action
))

struct /state
  cell field s.select	\ selection value
  cell field s.ops	\ output state
  cell field s.xt	\ xt of action
end-struct
  
: set-state	\ state# fsm -- ; make this state# current
  dup rot cells + swap !
;

: get-state	\ fsm -- state# ; get the current state#
  dup @  swap - cell /
;

: cycle		\ n1 fsm -- ; n1=input, fsm=FSM address
  tuck @ @				\ -- fsm n1 state
  dup @ swap cell+ swap			\ -- fsm n1 state' #states
  0 ?do             			\ -- fsm n1 state'
    2dup s.select @ = ?leave		\ at right place if inputs match
    /state +				\ -- fsm n1 state''
  loop
  rot over s.ops @ swap set-state	\ -- n1 state ; set new next state
  s.xt @ execute			\ -- ; perform output
;

variable >fsm<	\ -- addr ; pointer used during definition to the
                \ address of parameter field of a FSM


\ *******************************
\ Defining a Finite State Machine
\ *******************************

: FSM:		\ n -- <name> ; n FSM: <name>
  create
    here >fsm< !
    1+ cells allot			\ ptrs for current and n states
;

: state:	\ n -- ; begin state n, starting at 1
  cells >fsm< @ +  dup >fsm< @ !	\ make new state the current state
  here swap !				\ set state pointer into FSM
  0 ,					\ as yet unknown number of inputs
;

: action:	\ input nextState -- <action> ; define an input for the state
  swap , ,  ' ,				\ lay input, next state, xt
  >fsm< @ @ @ incr			\ update number of inputs
;

: default:	\ nextState -- <action> ; default value
  0 ,					\ dummy input
  ,					\ next state
  ' ,					\ must be the last entry of a state
;


\ ****************************
\ finite state machine example
\ ****************************

[undefined] Beeper [if]
: Beeper	\ --
;
[then]

: .mes1		( input -- )	drop ." level 1" cr  ;
: .mes2		( input -- )	drop ." level 2" cr  ;
: .mes3		( input -- )	drop ." level 3" cr  ;
: .mes4		( input -- )	drop ." you win" beeper cr  ;
: .mes5		( input -- )	drop ." sorry, try again" cr  ;
: .grr		( input -- )	drop ." Grrrrr" cr  ;

5 fsm: game	\ -- addr ; returns FSM structure address
  1 state:
    char f 2 action: .mes1
    char x 1 action: .grr
           1 default: .mes5
  2 state:
    char o 3 action: .mes2
           1 default: .mes5
  3 state:
    char r 4 action: .mes3
           1 default: .mes5
  4 state:
    char t 1 action: .mes4
    char T 3 action: .mes2
    char q 5 action: drop
           1 default: .mes5
  5 state:
           5 default: drop

: gamble	\ -- ; run the game
  1 game set-state
  begin
    key game cycle
    game get-state 5 =
  until
;

