\ Sudoku2.fth - Sudoku Solver in Forth.
\  No special extensions were used.
\  Tested on in win32forth, VFX and Swift (evaluation).
\  No locals were harmed during this experiment.
\
\  Version: 1900 01092005 - Robert Spykerman
\  email: robspyke@iprimus.com.au

anew sudoku

+idata

include %vfxpath%\lib\timers


\ ---------------------
\ Variables
\ ---------------------

81 buffer: sudokugrid 

9 cells buffer: sudoku_row

9 cells buffer: sudoku_col

9 cells buffer: sudoku_box


\  ---------------------
\  Logic
\  ---------------------
\  Basically :   
\     Grid is parsed. All numbers are put into sets, which are
\     implemented as bitmaps (sudoku_row, sudoku_col, sudoku_box)
\     which represent sets of numbers in each row, column, box.
\     only one specific instance of a number can exist in a
\     particular set.
\
\     SOLVER is recursively called
\     SOLVER looks for the next best guess using FINDNEXTSPACE
\     tries this trail down... if fails, backtracks... and tries
\     again.
\
    

\ Grid Related

: xy 9 * + ;   \  x y -- offset ;
: getrow 9 / ;
: getcol 9 mod ;
: getbox dup getrow 3 / 3 * swap getcol 3 / + ;

\ Puts and gets numbers from/to grid only
: setnumber sudokugrid + c! ;  \ n position --
: getnumber sudokugrid swap + c@ ;

: cleargrid sudokugrid 81 0 do dup i + 0 swap c! loop drop ;

\ --------------
\ Set related: sets are sudoku_row, sudoku_col, sudoku_box

\ ie x y --   ;  adds x into bitmap y
: addbits_row 1 rot lshift swap cells sudoku_row + dup @ rot or swap ! ;
: addbits_col 1 rot lshift swap cells sudoku_col + dup @ rot or swap ! ;
: addbits_box 1 rot lshift swap cells sudoku_box + dup @ rot or swap ! ;

\ ie x y --  ; remove number x from bitmap y
: removebits_row 1 rot lshift swap cells sudoku_row + dup @ rot invert and swap ! ; 
: removebits_col 1 rot lshift swap cells sudoku_col + dup @ rot invert and swap ! ;
: removebits_box 1 rot lshift swap cells sudoku_box + dup @ rot invert and swap ! ; 

\ clears all bitsmaps to 0
: clearbitmaps 9 0 do i cells 
                     0 over sudoku_row + !
                     0 over sudoku_col + !
                     0 swap sudoku_box + !
           loop ;

\ Adds number to grid and sets 
: addnumber                   \ number position --
    2dup setnumber 
    2dup getrow addbits_row 
    2dup getcol addbits_col 
         getbox addbits_box
;

\ Remove number from grid, and sets
: removenumber                \ position --
    dup getnumber swap    
    2dup getrow removebits_row 
    2dup getcol removebits_col 
    2dup getbox removebits_box 
    nip 0 swap setnumber
;

\ gets bitmap at position, ie
\ position -- bitmap

: getrow_bits getrow cells sudoku_row + @ ;  
: getcol_bits getcol cells sudoku_col + @ ;  
: getbox_bits getbox cells sudoku_box + @ ;  

\ position -- composite bitmap  (or'ed)
: getbits 
    dup getrow_bits 
    over getcol_bits 
    rot getbox_bits or or 
;

\ algorithm from c.l.f circa 1995 ? Will Baden
: countbits    ( number -- bits )
        [ HEX ] DUP  55555555 AND  SWAP  1 RSHIFT  55555555 AND  +
                DUP  33333333 AND  SWAP  2 RSHIFT  33333333 AND  +
                DUP  0F0F0F0F AND  SWAP  4 RSHIFT  0F0F0F0F AND  +
        [ DECIMAL ] 255 MOD
;

\ Try tests a number in a said position of grid
\ Returns true if it's possible, else false.
: try  \ number position -- true/false
    over 1 swap lshift 
    over getbits and 0= -ROT 2drop 
; 

\ --------------
: parsegrid  \ Parses Grid to fill sets.. Run before solver.
   sudokugrid \ to ensure all numbers are parsed into sets/bitmaps
   81 0 do 
     dup i + c@                             
       dup if                              
         dup i try if                    
           i addnumber                           
         else
           unloop drop drop FALSE exit      
         then  
       else
         drop 
       then 
   loop 
   drop
   TRUE
; 

\ Morespaces? manually checks for spaces ...
\ Obviously this can be optimised to a count var, done initially
\ Any additions/subtractions made to the grid could decrement
\ a 'spaces' variable.

: morespaces?
    0 81 0 do sudokugrid i + c@  0= if 1+ then loop ;


: findnextmove         \  -- n ; n = index next item, if -1 finished.
  -1  10                \  index  prev_possibilities  --
                         \  err... yeah... local variables, kind of...
  81 0 do 
    i sudokugrid + c@ 0= IF
      i getbits countbits 9 swap - 
      \ get bitmap and see how many possibilities
      \ stack diagram:
      \ index prev_possibilities  new_possiblities -- 
      2dup > if           
        \ if new_possibilities < prev_possibilities... 
        nip nip i swap  
        \ new_index new_possibilies --
      else \ else prev_possibilities < new possibilities, so:
        drop  \ new_index new_possibilies --         
      then                
    THEN
  loop
  drop
;

\ findnextmove returns index of best next guess OR returns -1
\ if no more guesses. You then have to check to see if there are
\ spaces left on the board unoccupied. If this is the case, you
\ need to back up the recursion and try again.

: solver
  findnextmove dup 0< if
    morespaces? if
      drop false exit
    else
      drop true exit
    then
  then
  
  10 1 do 
    i over try if          
      i over addnumber
      recurse  if
        drop unloop TRUE EXIT
      else
        dup removenumber
      then
    then
  loop

  drop FALSE
;

\ SOLVER

: startsolving         
   clearbitmaps  \ reparse bitmaps and reparse grid
   parsegrid     \ just in case..
   solver
   AND
;

\  ---------------------
\  Display Grid
\  ---------------------
\
\ Prints grid nicely

: .item		\ n --
  dup 0=
  if  drop ." . "  else  .  endif
;

: .sudokugrid 
  CR CR
  sudokugrid 
  81 0 do 
    dup i + c@ .item 
    i 1+ 
      dup 3 mod 0= if
         dup 9 mod 0= if
            CR
            dup 27 mod 0= if 
              dup 81 < if ." ------+-------+------" CR then
            then
         else
           ." ! "
         then      
      then
    drop
  loop
  drop 
  CR
;


\  ---------------------
\  Higher Level Words
\  ---------------------

: checkifoccupied  \ offset -- t/f
    sudokugrid + c@
;
    
: add                 \ n x y -- 
    xy 2dup
      dup checkifoccupied if
        dup removenumber
      then
    try if
      addnumber
      .sudokugrid
    else
      CR ." Not a valid move. " CR
      2drop
    then
;

: rm
    xy removenumber
    .sudokugrid
;

: clearit
    cleargrid
    clearbitmaps
    .sudokugrid
;

: solveit	\ --
  CR
  counter				\ start
  startsolving				\ gives flag
  counter rot - swap			\ duration
  if
    ." Solution Found " CR .sudokugrid
  else
    ." No Solution Found " CR
  then
  cr ." in " . ." ms"
;


: showit .sudokugrid ;


\ =====
\ Tools
\ =====

[undefined] get-word [if]
: get-word	\ char -- c-addr
\ *G A version of *\fo{WORD} that works across multiple lines. If a word
\ ** cannot be obtained, the input stream is *\fo{REFILL}ed.
  begin					\ -- char 
    dup word dup c@ 0=			\ scan input, no token?
  while					\ -- char addr
    drop  refill 0= #-411 ?throw	\ discard null, reload i/p
  repeat				\ -- char
  nip
;
[then]

: getNum	\ -- n
\ Get number from input stream
  0 0 bl get-word count >number 2drop drop
;


: Next3		\ caddr -- caddr'
\ Get a number and save it as three positions of the grid. Use
\ zero for undefined positions.
  >r  getNum
  10 /mod  10 /mod  r>			\ -- l m h caddr
  tuck c! 1+  tuck c! 1+   tuck c! 1+	\ -- addr'
;

: setgrid	\ caddr --
\ Fill the specified grid from 27 three-digit numbers
\ in the input strea.
  27 0
  do  Next3  loop
  drop  
;

: Grid:		\ -- ; Grid: <name> ; follow by 27 sets of 3
 \ Creates a named grid and uses SETGRID to fill it
  create
    here 81 allot setGrid
;

: >problem	\ caddr
\ Copy stated grid to problem area
  SudokuGrid 81 cmove
;

: Solve		\ caddr --
\ Solve the grid at caddr.
  >problem
  CR
  clearbitmaps
  parsegrid if
    CR ." Grid in source valid. " 
    .SudokuGrid
    solveit
  else
    CR ." Warning: Grid in source invalid. "
  then
;


\ =======
\ Puzzles
\ =======

\ Rob's test grid
Grid: Rob's
  090 004 007
  000 007 900
  800 000 000

  405 800 000
  300 000 002
  000 009 706

  000 000 004
  003 500 000
  200 600 080

Rob's >problem

\ Guardian Classic No. 101 Hard
Grid: Guardian
  200 005 100
  030 602 700
  000 000 089

  010 300 027
  000 000 000
  420 009 030

  840 000 000
  006 703 090
  003 800 006


((
"Gerry" <gerry@jackson9000.fsnet.co.uk> writes Re: Sudoku3

[..]

> Anton's #1 and #4 didn't need step 3, the rest of his did. #3 does
> appear to be the hardest and is the worst I've seen. However it found
> the 'minimal' puzzle rather easy as it didn't need to go to step 3.
> That seems strange as it took quite a long time in your results.

Obviously, what is difficult for a human being does not necessarily lead
to long computation times for a computer.

> Where were the other puzzles from, I'll gladly try them if you want.

I don't recall exactly where they are from. Most are from the Sudoku 
CLF thread started by Robert Spykerman. The minimal one could be from
a Sudoku website.

-marcel

-- ----------------

: rg 9 0 DO  BL WORD COUNT >FLOAT DROP F>S C, LOOP ;

CREATE grid0 
	rg      0 9 0  0 0 4  0 0 7
	rg      0 0 0  0 0 7  9 0 0
	rg      8 0 0  0 0 0  0 0 0

	rg      4 0 5  8 0 0  0 0 0
	rg      3 0 0  0 0 0  0 0 2
	rg      0 0 0  0 0 9  7 0 6

	rg      0 0 0  0 0 0  0 0 4
	rg      0 0 3  5 0 0  0 0 0
	rg      2 0 0  6 0 0  0 8 0
	," originally 4.36 ms for the computer" 

CREATE grid1 
	rg      0 0 6  0 5 0  0 0 0
	rg      0 7 0  0 3 9  1 0 0
	rg      0 8 0  0 0 0  0 3 0

	rg      0 0 0  0 0 2  5 1 8
	rg      0 0 0  0 0 0  0 0 0
	rg      7 5 9  8 0 0  0 0 0

	rg      0 6 0  0 0 0  0 7 0
	rg      0 0 2  5 9 0  0 4 0
	rg      0 0 0  0 6 0  3 0 0
	," 45 minutes human"

CREATE grid2 
	rg      9 2 0  0 0 0  0 0 8
	rg      0 8 0  0 0 0  0 5 1
	rg      0 0 1  5 0 0  3 0 0

	rg      0 0 0  9 0 7  8 0 0
	rg      0 0 0  0 0 0  0 0 0
	rg      0 0 3  6 0 2  0 0 0

	rg      0 0 6  0 0 4  7 0 0
	rg      5 7 0  0 0 0  0 8 0
	rg      8 0 0  0 0 0  0 9 3
	," 2 hours human"

CREATE grid3 
	rg      0 0 0  0 0 3  5 0 0
	rg      0 0 0  7 0 0  4 0 0
	rg      7 0 3  0 0 6  0 1 0

	rg      0 1 0  0 0 5  0 0 6
	rg      8 0 0  0 0 0  0 0 2
	rg      4 0 0  3 0 0  0 8 0

	rg      0 5 0  4 0 0  1 0 8
	rg      0 0 8  0 0 9  0 0 0
	rg      0 0 9  8 0 0  0 0 0
	," 2 hours for a human, maybe impossible"

CREATE grid4 
	rg      2 0 0  0 3 0  7 0 6
	rg      0 0 8  4 0 0  0 0 0
	rg      6 0 0  0 9 0  0 3 0

	rg      0 0 0  0 7 0  0 5 0
	rg      8 0 9  3 0 5  1 0 2
	rg      0 4 0  0 6 0  0 0 0

	rg      0 6 0  0 2 0  0 0 8
	rg      0 0 0  0 0 7  4 0 0
	rg      3 0 2  0 5 0  0 0 9
	," unknown source"

CREATE grid5
	rg      9 0 0  0 1 0  6 0 0
	rg      0 0 0  8 0 0  0 0 7
	rg      6 3 0  0 0 7  2 0 0

	rg      0 0 0  0 0 0  5 7 4
	rg      0 0 0  2 4 3  0 0 0
	rg      0 4 1  0 0 0  0 0 0

	rg      0 0 9  6 0 4  8 3 1
	rg      4 0 0  0 0 8  0 0 0
	rg      0 0 8  0 2 0  0 0 9
	 ," Paul Hsieh's example #1" 

CREATE grid6
	rg      0 0 9  6 0 0  1 0 0
	rg      0 0 0  0 0 0  0 2 4
	rg      6 0 0  0 0 2  0 9 0

	rg      3 0 0  4 2 0  0 0 0
	rg      0 8 0  3 1 0  0 0 7
	rg      2 0 4  0 0 0  0 1 8

	rg      0 0 0  7 0 4  8 0 2
	rg      5 4 0  0 3 0  0 0 1
	rg      7 9 0  5 0 0  0 0 0
	," Paul Hsieh's example #2"

CREATE grid7 
	rg      0 5 0  0 0 8  6 0 0
	rg      7 0 0  5 4 0  0 0 9
	rg      0 1 0  0 6 0  0 0 3

	rg      6 0 0  0 0 0  0 0 0
	rg      0 3 0  0 0 0  0 8 0
	rg      0 0 0  0 0 0  0 0 5

	rg      9 0 0  0 3 0  0 1 0
	rg      4 0 0  0 7 6  0 0 8
	rg      0 0 1  8 0 0  0 7 0
	," Paul Hsieh's example #3"

CREATE grid8
	rg	0 9 8  0 0 0  0 0 0
	rg	0 0 0  0 7 0  0 0 0
	rg	0 0 0  0 1 5  0 0 0

	rg	1 0 0  0 0 0  0 0 0
	rg	0 0 0  2 0 0  0 0 9
	rg	0 0 0  9 0 6  0 8 2

	rg	0 0 0  0 0 0  0 3 0
	rg	5 0 1  0 0 0  0 0 0
	rg	0 0 0  4 0 0  0 2 0
	," A `minimal' Sudoku (thought impossible for humans)"


CREATE grid9
	rg	0 0 1   0 0 0   8 0 0 
	rg	0 7 0   3 1 0   0 9 0 
	rg	3 0 0   0 4 5   0 0 7

	rg	0 9 0   7 0 0   5 0 0 
	rg	0 4 2   0 5 0   1 3 0 
	rg	0 0 3   0 0 9   0 4 0 

	rg	2 0 0   5 7 0   0 0 4
	rg	0 3 0   0 9 1   0 6 0 
	rg	0 0 4   0 0 0   3 0 0 
	," Ertl #1"

CREATE grid10
	rg	0 6 5   0 0 0   0 0 8
	rg	7 0 0   8 6 0   4 0 0 
	rg	0 0 0   0 2 0   0 0 9

	rg	0 4 0   0 0 1   0 0 2
	rg	0 0 0   2 0 7   0 0 0 
	rg	3 0 0   5 0 0   0 7 0 

	rg	4 0 0   0 5 0   0 0 0 
	rg	0 0 1   0 7 9   0 0 3
	rg	9 0 0   0 0 0   2 6 0 
	," Ertl #2"

CREATE grid11
	rg	0 0 0   0 7 0   9 4 0 
	rg	0 0 0   0 9 0   0 0 5
	rg	3 0 0   0 0 5   0 7 0 

	rg	0 0 7   4 0 0   1 0 0 
	rg	4 6 3   0 0 0   0 0 0 
	rg	0 0 0   0 0 7   0 8 0 

	rg	8 0 0   0 0 0   0 0 0 
	rg	7 0 0   0 0 0   0 2 8
	rg	0 5 0   2 6 0   0 0 0 
	," Ertl #3"

CREATE grid12
	rg	0 4 6   0 8 0   1 3 0 
	rg	3 9 0   5 0 6   2 0 7
	rg	0 5 0   1 3 7   4 0 0 

	rg	0 0 0   0 9 1   7 0 0 
	rg	1 0 0   0 0 0   6 4 0 
	rg	0 0 0   8 0 0   9 0 2

	rg	0 6 8   0 7 0   0 2 0 
	rg	0 0 5   0 0 3   8 7 0 
	rg	2 0 7   9 0 0   0 6 0 
	," Ertl #4"

CREATE grid13
	rg	7 0 8   0 3 0   0 0 0 
	rg	0 9 0   0 2 7   0 0 0 
	rg	0 2 1   8 0 0   9 7 0 

	rg	0 0 0   0 0 4   5 8 0 
	rg	0 0 7   0 0 0   2 0 0 
	rg	0 5 6   7 0 0   0 0 0 

	rg	0 1 5   0 0 3   6 2 0 
	rg	0 0 0   2 6 0   0 3 0 
	rg	0 0 0   0 5 0   7 0 9
	," Ertl #5"

CREATE grid14
	rg	6 0 8   9 0 2   0 0 7
	rg	0 0 0   0 7 0   9 0 0 
	rg	7 9 0   0 0 4   0 0 0 

	rg	5 0 0   0 0 7   3 0 0 
	rg	4 8 0   0 0 0   0 7 5 
	rg	0 0 7   6 0 0   0 0 4

	rg	0 0 0   2 0 0   0 1 6
	rg	0 0 1   0 3 0   0 0 0 
	rg	2 0 0   4 0 1   5 0 3
	," Ertl #6"

CREATE grid15
	rg	0 0 0   0 0 0   9 0 0 
	rg	1 7 0   0 0 5   0 0 2
	rg	0 8 0   9 2 1   0 0 7

	rg	0 1 0   0 9 0   5 0 0 
	rg	0 9 0   4 0 2   0 3 0  
	rg	0 0 4   0 7 0   0 2 0 

	rg	9 0 0   2 6 7   0 8 0 
	rg	6 0 0   8 0 0   0 7 1
	rg	0 0 8   0 0 0   0 0 0 
	," Ertl #7"

CREATE grid16
	rg	0 9 0  0 0 5  3 0 0 
	rg	0 0 0  0 2 0  8 0 5
	rg	5 0 8  0 0 6  0 7 0 

	rg	0 0 1  4 0 0  0 0 0 
	rg	0 8 2  7 0 1  5 3 0  
	rg	0 0 0  0 0 3  6 0 0 

	rg	0 2 0  6 0 0  4 0 8
	rg	4 0 9  0 3 0  0 0 0 
	rg	0 0 5  1 0 0  0 9 0 
	," Ertl #8"

))


\ ==========
\ User stuff
\ ==========

\ Print help menu
: help 
  CR
  ." Type clearit      ; to clear grid " CR
  ."      1-9 x y add  ; to add 1-9 to grid at x y (0 based) " CR
  ."      x y rm       ; to remove number at x y " CR
  ."      showit       ; redisplay grid " CR
  ."      solveit      ; to solve " CR
  ."      xxx solve    ; to solve grid xxx
  ."      Grid: <name> ; to define a named grid
  ."      help         ; for help " CR
  CR
;

: godoit
    CR
    clearbitmaps
    parsegrid if
      CR ." Grid in source valid. " 
    else
      CR ." Warning: Grid in source invalid. "
    then
    .sudokugrid
    help
;

godoit

