\ WinTimeChecks.fth - timestamps and checksums for PE files

((
Copyright (c) 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *! wintimechecks
\ *T Windows Timestamps and Checksums
\ ===========

cellbits #32 <> [if]
  1 abort" This code is for 32 bit Windows"
[then]


\ ***************
\ *S PE timestamp
\ ***************
\ *P This section generates a 32 bit number of seconds since
\ ** 1 Jan 1970.

struct /FILETIME	\ -- len
  4 field loDateTime		\ in 100ns units
  4 field hiDateTime
end-struct	

systemtime buffer: aSystemTime	\ -- addr
/FILETIME  buffer: aFileTime
variable ElapsedSeconds
	
Extern: BOOL SystemTimeToFileTime(
  void* lpSystemTime,
  void* lpFileTime
);

: mds/rem	\  d n -- rem dquot
\ *S Signed 64 bit divided by 32 bit divisor, returning a single
\ ** remainder and a double quotient. No protection is
\ ** provided. Symmetric division. A signed version of *\fo{MU/MOD}.
  over >r				\ sign of remainder=sgn(dividend)
  2dup xor >r				\ sign of quotient
  abs -rot dabs rot mu/mod		\ unsigned divide
  r> 0<
  if  dnegate  endif
  r> 0<
  if  rot negate -rot  endif
;

: mds/	\  d n -- dquot
\ *S Signed 64 bit divided by 32 bit divisor, returning a double
\ ** quotient. No protection is provided.
  mds/rem rot drop
;

10000000 constant 100ns/sec
1164447,3600 2constant EpochSecs

: D@		\ addr -- d
\ *G Fetch a double in native form. For the x86 ISA, the most
\ ** significant word is at the high address.
  dup @ swap cell + @			\ lo before high
;

: D!		\ d addr --
\ *G store a double in native form. The most significant word is
\ ** at the high address.
  tuck  2swap !  cell + !		\ lo before hi
;

: SecsSince1970	\ -- secs
\ *G Return the number of seconds since 1/1/1970.
  aSystemTime GetSystemTime
  aSystemTime aFileTime SystemTimeToFileTime drop
  aFileTime d@ 100ns/sec mds/		\ -- secs64
  EpochSecs d-
  drop
;


\ *******************
\ *S PE file checksum
\ *******************

Library: Imagehlp.dll

Extern: HANDLE CreateFileMappingA(
  HANDLE                hFile,
  LPSECURITY_ATTRIBUTES lpFileMappingAttributes,
  DWORD                 flProtect,
  DWORD                 dwMaximumSizeHigh,
  DWORD                 dwMaximumSizeLow,
  LPCSTR                lpName
);

Extern: DWORD IMAGEAPI MapFileAndCheckSumA(
  void*  Filename,
  DWORD* HeaderSum,
  DWORD* CheckSum
);

: OpenPEfile	\ caddr len -- pehandle
  r/w open-file abort" cannot open PE file"
;
  
: mapPEfile	\ pehandle -- maphandle
  NULL PAGE_READONLY 0 0 NULL CreateFileMappingA
  dup 0= abort" Cannot map file"
;

: PEfileSize	\ pehandle -- size
  file-size  abort" can't get size"  abort" file too big"
;

variable OldSum	\ -- addr ; holds previous checksum
variable NewSum	\ -- addr ; holds newly computed sum

: ChecksumPE	\ caddr len -- ior ; nz on error
  {: | name[ 256 ] -- :}
  name[ zplace
  name[ OldSum NewSum MapFileAndCheckSumA
;


\ ======
\ *> ###
\ ======

decimal
