(*
 * LANGUAGE    : ANS Forth with extensions
 * PROJECT     : Forth Environments
 * DESCRIPTION : Fannkuch shootout benchmark
 * CATEGORY    : Benchmarks 
 * AUTHOR      : Marcel Hendrix 
 * LAST CHANGE : June 5, 2006, Marcel Hendrix 
 *)

decimal

[undefined] timer-reset [if]
0 VALUE _time_
: timer-reset gettickcount TO _time_ ;
: .elapsed    gettickcount _time_ - . ." ms elapsed."  ;
[then]

0 [IF]
(*
    N  Full CPU Time s 
   10      16.04

 The Computer Language Shootout
 http://shootout.alioth.debian.org

 Contributed by Ian Osgood

       PROGRAM OUTPUT
       ==============
	12345678910
	21345678910
	23145678910
	32145678910
	31245678910
	13245678910
	23415678910
	32415678910
	34215678910
	43215678910
	42315678910
	24315678910
	34125678910
	43125678910
	41325678910
	14325678910
	13425678910
	31425678910
	41235678910
	14235678910
	12435678910
	21435678910
	24135678910
	42135678910
	23451678910
	32451678910
	34251678910
	43251678910
	42351678910
	24351678910
	Pfannkuchen(10) = 38

  FORTH> main ( iForth 2.0 on 3 GHz PIV )
  Pfannkuchen(10) = 38
  0.801 seconds elapsed. ok
\ Takes 0.594 with VFX Forth build 2297 on a 2.8 GHz PIV.
\ Takes 0.687 with VFX Forth build 2040 on a 2.8 GHz PIV.


  Results on 2 GHz PIV ...

	Ratio Program & Logs 	   Full CPU Time s 
	 Eiffel SmartEiffel 		 0.45
	 C gcc 				 0.45
	 D Digital Mars 		 0.49 
	 Clean 				 0.56
	 C++ g++ 			 0.70
	 OCaml 				 0.70
	 Oberon-2 OO2C 			 0.72
	 Ada 95 GNAT 			 0.94
	 Pascal Free Pascal 		 0.98 	
	 SML MLton 			 1.40
	 Haskell GHC #2 		 1.44
	 BASIC FreeBASIC 		 1.49
	 Java JDK -server 		 1.49
	 Scala #2 			 1.68
	 C# Mono			 1.68
	 Nice 				 1.83
	 Fortran G95 			 1.90
	 Haskell GHC 			 4.12
	 Smalltalk VisualWorks 		 6.94
	 Forth GForth 			16.04
	 Haskell GHC #3 		29.17

Sources for gcc

/*
 * The Computer Lannguage Shootout
 * http://shootout.alioth.debian.org/
 * Contributed by Heiner Marxen
 *
 * "fannkuch"	for C gcc
 *
 * $Id: fannkuch-gcc.code,v 1.33 2006/02/25 16:38:58 igouy-guest Exp $
 */

#include <stdio.h>
#include <stdlib.h>

#define Int	int
#define Aint	int

    static long
fannkuch( int n )
{
    Aint*	perm;
    Aint*	perm1;
    Aint*	count;
    long	flips;
    long	flipsMax;
    Int		r;
    Int		i;
    Int		k;
    Int		didpr;
    const Int	n1	= n - 1;

    if( n < 1 ) return 0;

    perm  = calloc(n, sizeof(*perm ));
    perm1 = calloc(n, sizeof(*perm1));
    count = calloc(n, sizeof(*count));

    for( i=0 ; i<n ; ++i ) perm1[i] = i;	/* initial (trivial) permu */

    r = n; didpr = 0; flipsMax = 0;
    for(;;) {
	if( didpr < 30 ) {
	    for( i=0 ; i<n ; ++i ) printf("%d", (int)(1+perm1[i]));
	    printf("\n");
	    ++didpr;
	}
	for( ; r!=1 ; --r ) {
	    count[r-1] = r;
	}

#define XCH(x,y)	{ Aint t_mp; t_mp=(x); (x)=(y); (y)=t_mp; }

	if( ! (perm1[0]==0 || perm1[n1]==n1) ) {
	    flips = 0;
	    for( i=1 ; i<n ; ++i ) {	/* perm = perm1 */
		perm[i] = perm1[i];
	    }
	    k = perm1[0];		/* cache perm[0] in k */
	    do {			/* k!=0 ==> k>0 */
		Int	j;
		for( i=1, j=k-1 ; i<j ; ++i, --j ) {
		    XCH(perm[i], perm[j])
		}
		++flips;
		/*
		 * Now exchange k (caching perm[0]) and perm[k]... with care!
		 * XCH(k, perm[k]) does NOT work!
		 */
		j=perm[k]; perm[k]=k ; k=j;
	    }while( k );
	    if( flipsMax < flips ) {
		flipsMax = flips;
	    }
	}

	for(;;) {
	    if( r == n ) {
		return flipsMax;
	    }
	    /* rotate down perm[0..r] by one */
	    {
		Int	perm0 = perm1[0];
		i = 0;
		while( i < r ) {
		    k = i+1;
		    perm1[i] = perm1[k];
		    i = k;
		}
		perm1[r] = perm0;
	    }
	    if( (count[r] -= 1) > 0 ) {
		break;
	    }
	    ++r;
	}
    }
}

    int
main( int argc, char* argv[] )
{
    int		n = (argc>1) ? atoi(argv[1]) : 0;

    printf("Pfannkuchen(%d) = %ld\n", n, fannkuch(n));
    return 0;
}

*)
[THEN]


[defined] VfxForth [if]
  +idata
  #0 loopalignment
[then]

#10 constant NUM
0 VALUE #perms

[undefined] buffer: [if]
create a        NUM allot
create pan      NUM allot
create counts   NUM allot
[else]
NUM buffer: a
NUM buffer: pan
NUM buffer: counts
[then]
  a NUM + 1- constant a-last

: .pan ( a -- ) 
	CR NUM 0 DO  DUP C@ 0 .R 1+  LOOP DROP ;

: flip ( n -- )
	DUP pan +  
	SWAP 2/ pan +  pan DO
          1-  I C@  OVER C@  I C! OVER C!
        LOOP 
	DROP ;

: flips ( -- n )
  	0 
	BEGIN  pan C@ 1 > 
	WHILE  1+  pan C@ flip 
	REPEAT ;

: fannkuch ( -- max-flips )
	NUM 0 DO  I 1+  a I + C!  LOOP
	0 NUM ( max-flips r )
	BEGIN  				
    	  #perms #30 < IF  a .pan  #perms 1+ TO #perms  ENDIF	\ print first 30 permutations
    	  a C@ 1 > IF  	a-last C@ NUM 				\ eliminate bad pans
	  		< IF  a pan NUM CMOVE			\ fill the pan and start flipping
      			      SWAP  flips MAX  SWAP
		       ENDIF
    		ENDIF
   	  BEGIN  DUP 1 > 					\ next permutation
	  WHILE  DUP 1- TUCK counts + C!  
	  REPEAT

    	  BEGIN  DUP >R  a C@  
	  	   a 1+ a R@ CMOVE
		 a R> + C! 				 	\ roll
          	 DUP counts +  DUP C@ 1-  DUP ROT C!  0=
    	  WHILE  1+ DUP NUM = IF  DROP EXIT  ENDIF
          REPEAT

  	AGAIN ;

: main ( -- )
	0 TO #perms 
	CR TIMER-RESET
	  fannkuch CR ." Pfannkuchen(" NUM 0 .R ." ) = " .
	CR .ELAPSED ;

: .ABOUT CR ." Try: main -- standard shootout test" ;

		.ABOUT 

                              (* End of Source *)
