\ bin2c.fth - convert binary image to C source

((
Copyright (c) 2004
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
))

decimal

\ ======
\ *! bin2c
\ *T Binary image to C source
\ ======
\ *P This code was developed to allow binary images to be embedded
\ ** into C source as an array. All output is done through
\ ** *\fo{EMIT}, *\fo{TYPE} and *\fo{CR}.

\ *P Note that these tools are designed for use from the keyboard
\ ** and that some global variables are used.

include %vfxpath%/Lib/Genio/FileBuff


[-sin


\ *************
\ *S File tools
\ *************
\ *P This section of the code is derived from Wil Baden's ToolBox.

: FILE-CHECK      ( n -- )  ABORT" File Access Error " ;

: MEMORY-CHECK    ( n -- )  ABORT" Memory Allocation Error " ;

: rewind-file	\ file-id -- ior
  0 0 rot reposition-file
;

0 value pData	\ -- addr ; pointer to data block
0 value /Data	\ -- n ; size of data block
0 value hData	\ -- handle ; handle of data file

: InitReadFile	\ handle -- size
\ *G Reset the file to the start and return its size.
  dup rewind-file file-check
  file-size file-check  drop
;

: OpenMouth	\ caddr len --
\ *G Open the file for read only.
  r/o open-file file-check  dup to hData
  InitReadFile  to /Data
;

: guzzle	\ file-id -- addr length
\ *G Reads file from disc to HERE without ALLOTing space.
\ ** The file is left open.
  dup InitReadFile			\ -- handle len
  here swap rot read-file file-check	\ -- #read
  here swap				\ -- addr #read
;

: slurp		\ file-id -- addr length
\ *G Reads the contents of a file into ALLOCATEd memory
\ ** and returns the address and length. Release the
\ ** memory using BURP.
  dup InitReadFile			\ -- handle len
  dup allocate memory-check		\ -- handle len addr
  dup to pData  dup >r  swap rot	\ -- addr len handle ; R: -- addr
  read-file file-check			\ -- #read
  r> swap				\ -- addr #read
;

: Hiccup	\ --
\ *G Close the file opened by *\fo{OpenMouth}.
  hData close-file file-check
;

: BURP		\ --
\ *G Release memory ALLOCATEd by SLURP.
  pData free memory-check
;


\ ************************
\ *S Generating the C code
\ ************************

MAX_PATH 1+ buffer: ImageName
  s" rsrc.img" ImageName place

MAX_PATH 1+ buffer: CFileName
  s" image.c" CFileName place

#64 buffer: BlockName
  s" ImageBlock" BlockName place

#64 buffer: SizeName
  s" ImageSize" SizeName place

: .x32		\ x --
\ *G Output x as a 32 bit hex C constant.
  base @ >r  hex
  ." 0x"  0 <#  #8 0 do  #  loop  #> type
  r> base !
;

: .Csize	\ len --
  cr ." int " SizeName count type ."  = " .x32 ." ;"
;

: .Chead	\ --
  cr ." // Generated by MPE bin2c utility"
  cr ." // From: " ImageName count type
     ."  to: " CFileName count type
;

: .Cstart	\ --
  cr ." unsigned int " BlockName count type ." [] = {"
;

: .Cend		\ --
  cr ." };" cr
;

: .line32	\ a-addr
  cr  3 spaces
  4 cells bounds
  do  i @ .x32 ." , "  cell +loop
;

: >Cimage	\ addr len --
  .Chead
  cr dup .Csize  cr .Cstart
  bounds
  ?do  i .line32  4 cells +loop
  cr ."    0x5555AAAA"			\ end marker
  .Cend
  cr
;


\ ******************
\ Writing the C file
\ ******************

FileDev: ofile	\ -- sid

: Copen		\ addr len --
  r/w create-file file-check ofile gen-handle !
  ofile op-handle !
;

: Cclose	\ --
  ofile close-gio file-check
;

: Image>C	\ --
  ImageName count OpenMouth
  hData slurp  hiccup
  ip-handle @ >r  op-handle @ >r
  CFileName count Copen  >CImage  Cclose
  r> op-handle !  r> ip-handle !
  burp
;

sin]


\ ======
\ *> ###
\ ======

decimal

