(*
 * LANGUAGE    : Forth
 * PROJECT     : Working on MPE and iForth
 * DESCRIPTION : VFX Forth OS dependent functions for the jack sound interface
 * CATEGORY    : Realtime Audio tool
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-06-03 01:56:28
 *)

JACKAUDIO DEFINITIONS
WINDOWS? [IF]
: START-BACK-THREAD	( xt parameters *threadid -- )
	>R >R >R
	0	\ security attributes
	$20000	\ stack size
	R>	\ thread function
	R>	\ arguments
	$0	\ creation flags
	R>	\ pointer to receive id
	CreateThread DROP ;
[THEN]


LINUX? [IF]
\ The stack oprder when calling this is like this ( xt arg 'thread -- )
: START-BACK-THREAD	( xt parameters threadid -- )
	-ROT 0 -ROT pthread_create DROP ;
[THEN]

DARWIN? [IF]
\ The stack oprder when calling this is like this ( xt arg 'thread -- )
: START-BACK-THREAD	( xt parameters threadid -- )
	-ROT 0 -ROT pthread_create DROP ;
[THEN]


Alias: S16B@	w@s
Alias: 16B!	w!
Alias: PLACE+	APPEND
Alias: 32B!	!	( n addr -- )
Alias: 32B@	@	( addr -- n )
Alias: S32B@	@	( addr -- n )
Alias: H.	.LWORD
Alias: ?MS	ticks ( -- n )
Alias: PI	%pi
Alias: $CR eol$	( -- addr )
: SF+!+					DUP SF@ F+ SF!+ ;
: (#.R)		( u r -- addr n )	BASE @ >R DECIMAL (U.R) R> BASE ! ;
: ($.R)		( u r -- addr n )	BASE @ >R HEX (U.R) R> BASE ! ;
: ??CR		( -- )			OUT @ IF CR THEN ;
: @+		( a -- a' n )		DUP @ SWAP CELL+ SWAP ;
: 0<=		( n -- flag )		0> 0= ;
: 32B!+		( n addr -- addr' )	TUCK ! 1 CELLS + ;
: 32B@+		( addr -- addr' n )	DUP @ SWAP CELL+ SWAP ;
: C@+		( addr -- addr' c )	DUP C@ SWAP CHAR+ SWAP ;
: FSQR		( F1: f1 -- f2 )	FDUP F* ;
: DEC.		( n -- )		BASE @ DECIMAL SWAP . BASE ! ;
: 2+!		( d a -- )		DUP >R 2@ D+ R> 2! ;
: REDUCE.2PI	( F n1 -- n2 )		FDUP %pi %pi F+ F> IF %pi F- %pi F- THEN ;
: FWITHIN	( F f1 f2 f3 -- flag )	2 FPICK F> F>= AND ;
: HTAB		( n -- )		OUT @ - 1 MAX SPACES ;
: ?FILE					THROW ;
: ?ALLOCATE				THROW ;
: 64BIT?	( -- f )		1 CELLS 8 = ;
: @LATEST-NAME	( -- addr )		LATEST CTRL>NFA ;
: .NAMEHEADER	( addr -- )		.NAME ;


[UNDEFINED] fromc [IF] : fromc ; [THEN]

:NONAME	( -- )	\ reads the current filesystem name to $FJACKSYSFILE;
	CurrSourceName $FJACKSYSFILE PLACE  $FJACKSYSFILE DUP C@ S" V_osdeps.frt" NIP - SWAP C!
	S" sysfiles/" $FJACKSYSFILE PLACE+
	WINDOWS?
	IF	$FJACKSYSFILE COUNT BOUNDS ?DO I C@ [CHAR] / = IF [CHAR] \ I C! THEN LOOP
	ELSE	$FJACKSYSFILE COUNT BOUNDS ?DO I C@ [CHAR] \ = IF [CHAR] / I C! THEN LOOP
	THEN ; DUP EXECUTE AtCold

64BIT?	[IF]	CR .( 64bit for vfx not implemented yet) ABORT \ INCLUDE fjack/opti64.frt
	[ELSE]	INCLUDE fjack/optv32.frt
	[THEN]

ALSO TYPES DEFINITIONS
	Alias: jack_client_t		INT
	Alias: jack_nframes_t		INT
	Alias: jack_port_t		INT
	Alias: jack_transport_state_t	INT
	Alias: jack_position_t		INT
	Alias: jack_midi_event_t	INT
	Alias: jack_midi_data_t		INT
	Alias: fftw_plan		INT
	Alias: jack_latency_range_t	INT
	Alias: jack_session_event_t	INT
	Alias: fftw_complex		double
	[UNDEFINED] size_t [IF] Alias: size_t	INT [THEN]
PREVIOUS DEFINITIONS
