(*
 * LANGUAGE    : Forth; iForth and MPE Forth working
 * PROJECT     : Forth Tool
 * DESCRIPTION : Config file for fJACK
 * CATEGORY    : Realtime Audio tools
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-05-04 01:23:15; Hanno Schwalm hanno@schwalm-bremen.de
		2011-08-08 added cCONSTANT optioning way
 *)
FORTH DEFINITIONS
\ *S What can be configured in fJACK?
\ *P Following are some system wide parameters defined in *\bold{fjack/config.frt} that could be changed by you!
\ ** All these configure options are defined by a cCONSTANT definer; it's used like a *\forth{CONSTANT} but it looks
\ ** for a definition having the same name. If such a definition exits, it won't define a new constant but will
\ ** use your predefined constant.
\ *P So - let's say you want to use fJACK on stage with a powerful computer, you
\ **.have a twelve channel audio hardware and use echoes for ~3seconds.
\ ** You will have to define *\forth{JACK-CHANNELS} and *\forth{#HISTORY-SIZE} *\bold{before} INCLUDE fjack.frt
\ *C 16 CONSTANT JACK-CHANNELS
\ *C $20000 CONSTANT #HISTORY-SIZE
\ *P This way you may configure your fJACK application without changing the fJACK source itself.

: cCONSTANT	( n -- )
	>IN @ BL WORD FIND NIP SWAP >IN !	\ <>0 found
	IF	CR ." fJACK config option: " BL WORD COUNT 2DUP TYPE ."  -- " EVALUATE . ." instead of default " .
	ELSE	CONSTANT
	THEN ;

#4 cCONSTANT JACK-CHANNELS
\ *G It can be set to any number >= 1 you like.
\ ** The default is 4, when you have a powerful computer and wish to use more audio channels,
\ ** just set this to 8 or whatever you need. Each channel has one input and one output port defined.
\ ** These ports are used for the audio processing plugins described later.
\ ** There is also a programmers interface to be used by you to define your own sound processing tools.
\ ** Also think about using other jack plugins available in the net to be used by your forth application.
\ ** Just one hint - more channels than needed will use a lot of memory and slow down the system a bit.

#32 cCONSTANT JACK-IO-SLOTS
\ *G Defines the maximum number of processing plugins that can be placed in the Input-Port->Output-Port
\ ** processing chain for a channel. Also see *\forth{#PLUGINS}

#32 cCONSTANT #PLUGINS
\ *G The maximum number of plugins that can be used at the same time in the input->output processing chain.

FALSE cCONSTANT JACK-AUTOSTARTING?
\ *G Should the jack server be autostarted with the client? This requires a moder jack server installed correctly.

[DEFINED] KICKSTART 0= [IF] : KICKSTART ; [THEN]
\ *G Sometimes you might want to start the jack server in a user specified way not using any of
\ ** the control applications or the autostart feature.
\ *P In this case you can define a word KICKSTART before loading fjack that does so.
\ *C : KICKSTART S" start /d C:\jack\jackdmp_0.71\bin\windows /i jackdmp -S -d portaudio -p1024" SYSTEM ;

[UNDEFINED] JACK-LOGIN-WAVFILE [IF] : JACK-LOGIN-WAVFILE S" fjack/snippets/start.wav" ; [THEN]
\ *G This is the wavfile played when starting the fjack client. You may choose any other wav filename and define
\ ** *\forth{JACK-LOGIN-WAVFILE} before loading fjack.frt

[UNDEFINED] JACK-CLIENT-NAME [IF] : JACK-CLIENT-NAME S" fJACK" ; [THEN]
\ *G Normally the fjack client is called fJACK. To support session management you may define another client name
\ ** for your target application. To further support jack session management it is assumed here that you will
\ ** have a shell script available in your iforth directory with exactly the name of the client name, this will be started
\ ** when reading your session. To use another client name you'll have to define \*forth{JACK-CLIENT-NAME} before loading fjack.

TRUE cCONSTANT JACK-MODULES?
\ *G Load the plugin modules? Without these modules no filtering is possible!

FALSE cCONSTANT JACK-VERBOSE?
\ *G Some extra information in the log

FALSE cCONSTANT JACK-MIDI?
\ *G Might switch on midi services.

#2048 cCONSTANT #FIR_BUFF
\ *G This defines the maximum number of taps in adaptive and fir filters; every defined FIR filter object uses 5*#FIR_BUFF DFLOATS
\ ** dynamic memory. Must be a power of 2 and at least 4. This is quite a good default as normal fir filters won't need larger buffers.
\ ** When using special effect plugins using the generic fir filter routines you might need larger buffers.

#16 cCONSTANT #jack-fifos	\ MUST be a power of 2
\ *G Gapless wav playing and writing is controlled via a FIFO with *\forth{#jack-fifos} elements. If
\ ** an applications writes small chunks of wav data at very high speed you could increase this.
\ ** More information is available in the wav handling chapter.

$800 cCONSTANT #max-frame-size	\ MUST be a power of 2
\ *G The maximum frame size suported. Only with very slow hardware you will want larger buffers as
\ ** the latency gets higher. #1024 is the default for the current jack versions so it must be at least
\ ** $400.

$10000 cCONSTANT #HISTORY-SIZE	\ must be a power of 2 and at least 4 * #max-frame-size
\ *G fJACK may want access to data played some time before for delaying or data analysis so data
\ ** are saved in this circular buffer. This history is available in every plugin.
\ ** The mixer and delay plugins both need the history size to be large enough.

FALSE cCONSTANT AUTO-HISTORY-CLEAR?
\ *G Whenever a plugin object is put into the processing chain the history might be cleared automatically.
\ ** I don't think it's worth to set this to TRUE but it's your decision.

1 cCONSTANT DEFAULT-FIR-WINDOW
\ *G The coefficients of FIR filters are calculated by a windowed inverse FFT, this windowing function can 0,1 or 2.
\ ** See *\forth{SET-FIR-WINDOWING} for more information.

#64 cCONSTANT JACK-BOARDJOBS
\ *G The number of wavboard jobs to be used at the same time. Must be at least one

TRUE cCONSTANT OPTIMIZED-FILTER-CODE?
\ *G Developers might wish to disable the sse2 optimized filter primitives.
