(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Audio processing
 * DESCRIPTION : Generic Effect for the jack plugins
 * CATEGORY    : audio processing tool
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-12-09 21:08:06
		: 2009-12-29 17:26:37 Tuning the effects object is like
 *)

JACKAUDIO DEFINITIONS
\ *S The generic Effect Filter
\ *P Allows the user defined definitions of effects.
\ ** The generic effect plugin is just a special type of FIR filter, the dsp algorithm is the same but the coefficients
\ ** are not calculated from a defined frequency response. So the *\forth{SET-OBJECT-FREQUENCY} tool is undefined, instead
\ ** it uses a callback function defining the coefficient for each tap number.

HERE ," Effect-Filter" CONSTANT #effect-id
#PRIVATE-DSP
	DSP-VALUE effect_data
	DSP-VALUE effect_muls
	CONSTANT #effect_struct

: calc-effect-val	\ ( -- ) (F val -- val' )
	effect_muls
	FDUP	#FJACK-SAMPLE NEGATE			#HISTORY-SIZE 1- AND
	DFLOATS fir_data + DUP DUP DF! FDUP		#HISTORY-SIZE DFLOATS + DF!
	FDUP	#FJACK-SAMPLE NEGATE 1-			#HISTORY-SIZE 1- AND #HISTORY-SIZE 2* +
	DFLOATS fir_data + DUP DF!			#HISTORY-SIZE DFLOATS + DF!
	DUP $08 AND IF #HISTORY-SIZE 2* 1- DFLOATS + THEN
	plug_taps (calc-fir-val) EXECUTE ;

: (init-effect-filter)
 	plug_dynmem #16 #B-ALIGNED TO effect_data
 	effect_data #HISTORY-SIZE 4 * DFLOATS + TO effect_muls
	1 TO plug_taps 1e0 effect_muls DF! ;

FORTH DEFINITIONS
: DEFINE-EFFECT-FILTER	( calc-xt|0 object -- )
\ *G calculates the effect filter coefficients for a given plugin object with the calculating function xt.
\ ** The calculating function must take two integer parameters *\forth{( number-of-filtertaps index -- )}
\ ** and must return the float *\forth{(F -- coefficient)} value for the index tap.
\ ** When *\forth{DEFINE-EFFECT-FILTER} is executed, it calls the function for every tap from 0 to {#HISTORY-SIZE},
\ ** so the default setting will allow processing in the ~3 seconds range for CD sampling rate.
\ ** When the xt is *\forth{FALSE} the effects filter is 'just-do-nothing'.
	-->DSP plug_filtype EFFECTS-FIR <> ABORT" Only effect filters allowed"
	DUP 0= IF DROP 1e0 effect_muls DF! 1 TO plug_taps EXIT THEN
	#HISTORY-SIZE 0
	DO	DUP #HISTORY-SIZE I ROT EXECUTE
		FDUP 0e0 F<> IF I 1+ TO plug_taps THEN
		effect_muls I DFLOATS + DF!
	LOOP DROP ;
: MAKE-EFFECT-FILTER	( -- object )
\ *G Constructs an effects plugin object leaves it's address. Calculation of the filter coefficients is done
\ ** with *\forth{DEFINE-EFFECT-FILTER}.
\ *P This word is meant for user extensions like hall and echo effects, combs ... let's see what we will make out of it.
	['] (init-effect-filter)
	#HISTORY-SIZE DFLOATS 5 * $20 + #effect_struct MAKE&ALLOCATE-PLUGIN-OBJECT
	EFFECTS-FIR TO plug_filtype
	#effect-id TO plug_type
	['] calc-effect-val TO plug_xt
	@DSP
	0 OVER DEFINE-EFFECT-FILTER ;
