(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Forth Environments
 * DESCRIPTION : Decoder plugins for digital audio modes
 * CATEGORY    : Sound processing tool
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-06-02 21:50:43; turnkey ready
 *)

\ *S The Decoder plugin
\ *P This plugin watches the audio data for a <=0 to >0 level changes, does some housekeeping,
\ ** calculates the frequency of the current audio signal and saves these data in a FIFO buffer.
\ ** The FIFO size is 32k, so when you are analysing at 2000/sec data are buffered for the last ~15 seconds.
\ ** This plugin is used for audio analysis with frequency-modulated signals like CW, WFAX, RTTY or alike.
\ *P The frequency of the anylser is set as usually by *\forth{SET-OBJ-FREQUENCY}.

DECIMAL ALSO JACKAUDIO ALSO DEFINITIONS

HERE ," Frequency modulation Decoder" CONSTANT #dec-id

#PRIVATE-DSP
	DSP-VALUE *buffer
	DSP-VALUE *moving4
	DSP-VALUE buff+
	DSP-VALUE buff-
	DSP-VALUE decstate
	DSP-FVALUE ticker
	DSP-FVALUE lastrisen
	DSP-FVALUE timediff
	DSP-FVALUE cumulator
	DSP-FVALUE last-val
	CONSTANT #decstruct
$7FFF CONSTANT #decmask

: decode-val
	FDUP #FJACK-SAMPLE 3 MOD SFLOATS *moving4 + SF!
	*moving4 SF@+ SF@+ F+ SF@ F+ 3e0 F/ FDUP	\ val acc acc
	decstate
	IF	-0.02e0 F< IF 0 TO decstate THEN
	ELSE	0.02e0 F>
		IF	1 TO decstate  last-val FNEGATE  FOVER last-val F-  F/ ticker F+
			FDUP lastrisen F- TO timediff TO lastrisen
		THEN
	THEN
	plug_frequency JACK-SAMPLE-RATE S>F F/ 	cumulator F+  FDUP TO cumulator  1e0 F>
	IF	-1e0 +TO cumulator JACK-SAMPLE-RATE S>F timediff F/
		*buffer  buff+ #decmask AND  SFLOATS + SF!
		buff+ buff- - #decmask <= IF 1 +TO buff+ THEN
	THEN
	TO last-val 1E0 +TO ticker ;
: CHK-DEC
	#dec-id plug_type <> ABORT" Invalid DECODER object" ;
: (init-decoder)
	plug_dynmem ALIGNED TO *moving4
	*moving4 4 SFLOATS + TO *buffer
	0 TO buff+  0 TO buff-  0 TO decstate
	0e0 TO cumulator  0e0 TO ticker 0e0 TO lastrisen 0e0 TO last-val ;
: (set-decoder-frequency)
	TO plug_frequency ;

FORTH DEFINITIONS
: MAKE-DECODER	( -- object )
\ *G Defines a decoder plugin
	['] (init-decoder)
	#decmask 1+ SFLOATS $80 + #decstruct MAKE&ALLOCATE-PLUGIN-OBJECT
	DECODER TO plug_filtype  #dec-id TO plug_type  10e0 TO plug_frequency
	['] (set-decoder-frequency) TO plug_xt_tune
	['] decode-val TO plug_xt @DSP ;
: DECODER:	\ ( -- ) name
\ *G defines a named decoder plugin
	MAKE-DECODER CONSTANT ;

: GET-DECODER-DATA	\ ( decoder -- true (F -- frequ ) | false (F -- ) )
\ *G Gets decoder data from the decoders fifo; typically this is used in a loop
\ *C *\forth{BEGIN GET-DECODER-DATA WHILE ... DF! ... REPEAT}
	-->DSP CHK-DEC
	buff+ buff- <> DUP
	IF 	*buffer buff- #decmask AND SFLOATS + SF@  1 +TO buff- THEN ;

PREVIOUS PREVIOUS DEFINITIONS
