(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Audio processing
 * DESCRIPTION : Implementing a delay filter for the jack plugins
 * CATEGORY    : audio processing tool
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-11-27 22:16:51 Hanno Schwalm
 *)


\ *S The Delayer Object
\ *P This can add an echo to the current signal, the level of the echo and the original signal can be set.

DECIMAL ALSO JACKAUDIO ALSO DEFINITIONS

HERE ," Audio Delayer" CONSTANT #delay-id
: CHK-DELAY	#delay-id plug_type <> ABORT" Invalid DELAY object" ;

#PRIVATE-DSP
	DSP-FVALUE del_level
	DSP-FVALUE del_time
	CONSTANT #del_struct

: calc-del-val	\ ( -- ) (F val -- val' )
	plug_level F*
	plug_input del_time @INTERPOLATED-HISTORY-DATA del_level F*
	F+ ;
: (init-delay-object)
	0e0 TO del_time  1e0 TO plug_level  0e0 TO del_level ;

FORTH DEFINITIONS

: MAKE-DELAYER	( -- object )
\ *G Construct a delayer object and leave it's address.
	['] (init-delay-object)
	0 #del_struct MAKE&ALLOCATE-PLUGIN-OBJECT
	DELAYER TO plug_filtype  #delay-id TO plug_type
	['] calc-del-val TO plug_xt @DSP ;
: DELAYER:	( -- ) \ name
\ *G Define a named delayer object
	 MAKE-DELAYER CONSTANT ;
: SET-DELAY-TIME	( useconds object -- )
\ *G Sets the dapay time for the delayer object. Time in usec.
	-->DSP CHK-DELAY S>F 1e6 F/  TO del_time ;
: SET-DELAY-LEVELS	\ ( object -- ) (F delayed-level now-level -- )
\ *G Sets the signal levels for a delayer object; both level must be floats.
	-->DSP CHK-DELAY TO plug_level TO del_level ;

PREVIOUS PREVIOUS DEFINITIONS
