(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Forth Environments
 * DESCRIPTION : a bank of notch iir filters
 * CATEGORY    :
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2008-07-17 08:17:35, Hanno Schwalm
		: 2009-01-16 21:26:40; multi channel supported
		2009-05-09 02:02:14 fixed slot handling as objects contain more information
 *)
\ *S A Bank of notch filters
\ *P Implements a bank of notch IIR butterworth filters, the slots 15-22 in a channel are used


DECIMAL ALSO JACKAUDIO ALSO DEFINITIONS
8 CONSTANT #NOTCHES
\ *G The number of notch filters available
CREATE NOTCH-TAB  #NOTCHES CELLS ALLOT
: MAKE-NOTCH-FILTERS
	#NOTCHES 0 DO BAND-STOP MAKE-IIR-FILTER  NOTCH-TAB I CELLS + ! LOOP ;
	MAKE-NOTCH-FILTERS
FORTH DEFINITIONS
: GET-FREE-NOTCH	( -- object flag )
\ *G Try to get a notch filter object from a bunch of 8, if all notches are already in use
\ ** flag is *\forth{FALSE}.
	0 0   #NOTCHES 0
	DO	I CELLS NOTCH-TAB + @ -->DSP plug_#used -1 =
		IF 2DROP @DSP TRUE LEAVE THEN
	LOOP ;

: NOTCH-THERE?		( frequ channel -- object flag )
\ *G checks for an already used notch filter with frequency - or very close to.
\ ** If such a notch is plugged in, the object and a *\forth{TRUE} flag are returned.
	LOCALS| channel frequ |
	channel 0 JACK-CHANNELS WITHIN 0= ABORT" channel not available for notch filter"
	0 0   #NOTCHES 0
	DO	I CELLS NOTCH-TAB + @ -->DSP
		plug_#used -1 >
		plug_channel channel = AND
		plug_frequency F>S frequ - ABS JACK-SAMPLE-RATE #1000 / < AND
		IF  2DROP @DSP TRUE LEAVE THEN
	LOOP ;

: SET-NOTCH-FILTER	\ ( frequ channel -- )
\ *G Sets a notch filter from a bunch of 8. If there is a notch already plugged in it is switched off.
	LOCALS| channel frequ |
	frequ channel NOTCH-THERE? IF REMOVE-JACK-PLUGIN EXIT THEN DROP
	GET-FREE-NOTCH 0= IF DROP EXIT THEN	( object )

	frequ S>F JACK-SAMPLE-RATE S>F F/ 1500E0 F*  F>S  OVER SET-OBJ-QUALITY
	frequ OVER SET-OBJ-FREQUENCY		( object )
	#23 #15
	DO	I channel *rg_PLUG @ 0=
		IF DUP I channel SET-JACK-PLUGIN LEAVE THEN
	LOOP DROP ;
: CLEAR-ALL-NOTCHES	( -- )
\ *G removes all notches from the processing queue.
	#NOTCHES 0 DO I CELLS NOTCH-TAB + @ REMOVE-JACK-PLUGIN LOOP ;

PREVIOUS PREVIOUS DEFINITIONS
