(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Audio processing
 * DESCRIPTION : Implementing a phaser for the jack plugins
 * CATEGORY    : audio processing tool
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-12-30 12:40:44
 *)


\ *S The Phaser Object
\ *P The phaser plugin emulates a phaser/flanger. You can control the timing of the shifted signal,
\ ** the timing is also modulated by an oszillator.

DECIMAL ALSO JACKAUDIO ALSO DEFINITIONS

HERE ," phaser" CONSTANT #phase-id
: CHK-PHASER	#phase-id plug_type <> ABORT" Invalid PHASER object" ;

#PRIVATE-DSP
	DSP-FVALUE phase_shift
	DSP-FVALUE phase_sinval
	DSP-FVALUE phase_speed
	DSP-FVALUE phase_magnitude
	CONSTANT #phase_struct

: calc-phase-val	\ ( -- ) (F val -- val' )
	phase_sinval phase_speed F+ REDUCE.2PI FDUP TO phase_sinval FSIN phase_magnitude F*
	phase_shift F+ plug_input @INTERPOLATED-HISTORY-DATA
	F+ plug_level F* ;
: (init-phase-object)
	0e0 TO phase_sinval ;

FORTH DEFINITIONS

: MAKE-PHASER	( -- object )
\ *G Construct a phaser object and leave it's address. The default settings are a timeshift of 10msec,
\ ** a strength of 0.2 and an oszillating speed of 5Hz.
	['] (init-phase-object)
	0 #phase_struct MAKE&ALLOCATE-PLUGIN-OBJECT
	PHASER-PLUGIN TO plug_filtype  #phase-id TO plug_type
	0.01e0 TO phase_shift
	phase_shift 0.2e0 F* TO phase_magnitude
	5e0 PI F* F2* JACK-SAMPLE-RATE S>F F/ TO phase_speed
	0.01e0 TO phase_shift
	['] calc-phase-val TO plug_xt @DSP ;
: PHASER:	( -- ) \ name
\ *G Define a named delayer object
	 MAKE-PHASER CONSTANT ;

: SET-PHASER-PARAMETERS	\ ( object -- ) (F frequ strength timeshift -- )
\ *G Sets parameters for the phaser object.
\ *P The strength may be choosen from 0 to 1, this controls the oszillating amount, the default of 0.2 is already pretty strong.
\ *P Timeshift in milliseconds, the maximum depends on the history size, probably up to 100 msecs are usefull.
\ *P frequency controls the shift oszillator in Hz.
	-->DSP CHK-PHASER
	1e3 F/ 0e0 FMAX 1e0 FMIN TO phase_shift
	0e0 FMAX 1e0 FMIN phase_shift F* TO phase_magnitude
	PI F* F2* JACK-SAMPLE-RATE S>F F/ TO phase_speed ;

PREVIOUS PREVIOUS DEFINITIONS
