(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Audio processing
 * DESCRIPTION : FIR filters for the jack plugins
 * CATEGORY    : audio processing tool
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-08-01 11:11:40, Hanno Schwalm
		: 2009-08-05 20:28:24 optimized routines for 64bit machines
		: 2009-08-22 01:13:47 all filters sizes are allowed and supported by the filter kernel
		: 2009-12-09 23:59:47; corrected coeff calculating with setting filter size.
		: 2009-12-11 00:25:45; added an equalizer FIR filter; bindings to the showfft module
 *)

JACKAUDIO DEFINITIONS
\ *S The Group of Finite Impulse Response Filters
HERE ," FIR-Filter" CONSTANT #fir-id
#PRIVATE-DSP
	DSP-VALUE fir_data
	DSP-VALUE fir_muls
	DSP-VALUE fir-window
	DSP-VALUE fir-tune-xt
	DSP-VALUE fir-coeff-xt
	DSP-VALUE fir-linear-equal
	CONSTANT #fir_struct
: CHK-FIR	#fir-id plug_type <> ABORT" Invalid FIR filter object" ;

$4000 CONSTANT #fir-plan-size
#fir-plan-size INVERSE-FFT-STRUCT CONSTANT filter-fft-plan
: (calc-lp	plug_frequency F< IF 1.0e0 ELSE 0e0 THEN ;
: (calc-hp	plug_frequency F< IF 0e0 ELSE 1e0 THEN ;
: (calc-bp	plug_frequency plug_frequency2 FWITHIN IF 1.0e0 ELSE 0e0 THEN ;
: (calc-bs	plug_frequency plug_frequency2 FWITHIN IF 0e0 ELSE 1e0 THEN ;
: (cl-fir-data)	fir_data #FIR_BUFF 5 * DFLOATS ERASE ;
: *equal-level	( i -- addr ) SFLOATS plug_static + ;
: frequ->equal	\ ( -- idx ) ( F frequency -- )
	JACK-SAMPLE-RATE 2/ S>F F/ 256e0 F* 0.5e0 F+ F>S 0 MAX #255 MIN ;
: calc-equalizer	\ (F level frequency -- your-level )
	FSWAP FDROP frequ->equal *equal-level SF@ ;

: CALC-FIR-FILTER-OLD
	plug_filtype DUP HIGH-PASS = SWAP LOW-PASS = OR 0=
	ABORT" Only HIGH- and LOW-PASS FIR filters can be tuned when the fftw3 lib is not installed"
	plug_frequency 2e0 F* JACK-SAMPLE-RATE S>F F/  PI F*
	plug_filtype HIGH-PASS = IF PI F- THEN		\ (F omega_g -- )
	plug_taps 1- 2/ LOCALS| centre |
	centre 0
	DO	FDUP I centre - S>F F*  FSIN   I centre - S>F PI F* F/
		plug_filtype HIGH-PASS =  I 1 AND 1 =  AND IF FNEGATE THEN
		I centre - S>F plug_taps 1- S>F F/ PI F* F2* FCOS 0.46e0 F*  0.54e0 F+
		F* FDUP fir_muls I DFLOATS + DF!  fir_muls plug_taps 1- I - DFLOATS + DF!
	LOOP
	PI F/ fir_muls centre DFLOATS + DF! ;

: CALC-FIR-FILTER
	plug_taps 2/ TO plug_timeshift
	FFTW3-LIB? 0= IF CALC-FIR-FILTER-OLD EXIT THEN
	plug_filtype USER-FIR <>
	IF	#fir-plan-size 2/ 0
		DO	0e0	\ complex part
			JACK-SAMPLE-RATE S>F I S>F F*  #fir-plan-size S>F F/
			plug_filtype
			CASE	LOW-PASS	OF	(calc-lp	ENDOF
				HIGH-PASS	OF	(calc-hp	ENDOF
				BAND-PASS	OF	(calc-bp	ENDOF
				BAND-STOP	OF	(calc-bs	ENDOF
				EQUALIZER	OF	(calc-bp	ENDOF
			ENDCASE
			fir-tune-xt ?DUP
			IF JACK-SAMPLE-RATE S>F I S>F F*  #fir-plan-size S>F F/ ( F: level frequency ) EXECUTE THEN
			filter-fft-plan I >FFT-COMPLEXDATA DF!+ DF!
		LOOP
		filter-fft-plan CALCULATE-INVERSE-FFT
		plug_taps 0
		DO	fir-window
			CASE	0 OF	I S>F plug_taps 1- S>F F/ PI F* F2* FCOS -0.5e0 F*
					I S>F plug_taps 1- S>F F/ PI F* 4e0 F* FCOS 0.08e0 F* F+  0.42e0 F+
				ENDOF
				1 OF	I S>F plug_taps 1- S>F F/  PI F* F2* FCOS -0.46e0 F* 0.54e0 F+
				ENDOF
				2 OF	1e0
				ENDOF
			ENDCASE
			filter-fft-plan  I plug_taps 1- 2/ -  #fir-plan-size 1- AND >FFT-DATA DF@ F*
			#fir-plan-size S>F F/
			fir_muls I DFLOATS + DF!
		LOOP
		plug_filtype EQUALIZER =
		IF	#256 0
			DO	I *equal-level SF@ FABS 0.001e0 F>
				IF I JACK-SAMPLE-RATE 2/ #256 */ S>F TO plug_frequency LEAVE THEN
			LOOP
			#256 0
			DO	#255 I - *equal-level SF@ FABS 0.001e0 F>
				IF #255 I - JACK-SAMPLE-RATE 2/ #256 */ S>F TO plug_frequency2 LEAVE THEN
			LOOP
		THEN
	ELSE	#FIR_BUFF 0
		DO	#FIR_BUFF I fir-coeff-xt ?DUP
			IF	( size index ) EXECUTE
			ELSE	NIP 0= IF 1e0 ELSE 0e0 THEN
			THEN
			FDUP 0e0 F<> IF I 1+ TO plug_taps THEN
			fir_muls I DFLOATS + DF!
		LOOP
	THEN ;

: calc-fir-val	\ ( -- ) (F val -- val' )
	fir_muls
	FDUP	#FJACK-SAMPLE NEGATE #FIR_BUFF 1- AND DFLOATS fir_data + DUP DUP DF! FDUP #FIR_BUFF DFLOATS + DF!	( addr )
	FDUP	#FJACK-SAMPLE NEGATE 1- #FIR_BUFF 1- AND #FIR_BUFF 2* + DFLOATS fir_data + DUP DF! #FIR_BUFF DFLOATS + DF!
	DUP $08 AND IF #FIR_BUFF 2* 1- DFLOATS + THEN
	plug_taps (calc-fir-val) EXECUTE ;

: (set-fir-filter-frequency)	( F: ???frequency2 frequency1 -- )
	plug_filtype BAND-STOP = plug_filtype BAND-PASS = OR IF	S>F FSWAP TO plug_frequency2 THEN
	TO plug_frequency CALC-FIR-FILTER ;
: (set-fir-filter-quality)	( taps -- )
	#FIR_BUFF 1- MIN 2 MAX 1 OR TO plug_taps CALC-FIR-FILTER ;
: (init-fir-filter)
	plug_dynmem #16 #B-ALIGNED TO fir_data
	fir_data #FIR_BUFF 4 * DFLOATS + TO fir_muls
	(cl-fir-data) ;
: -->EQUALIZER	( object -- )
	-->DSP CHK-FIR EQUALIZER plug_filtype <> ABORT" Not an equalizer FIR object" ;
: dB->level	( dB -- level )
	20e0 F/ 10e0 FSWAP F** ;

FORTH DEFINITIONS
: MAKE-FIR-FILTER	( filtype size -- object )
\ *G Constructs a linear-phase FIR filter object and leave it's address. 'type' can be *\forth{LOW-PASS HIGH-PASS BAND-PASS BAND-STOP USER-FIR}, the 'size'
\ ** is the number of taps corrected to odd.
\ ** FIR filters add a signal delay by half the number of taps.
\ *P You may select a different number of taps later via *\forth{SET-OBJ-QUALITY ( taps object -- )}, also you might
\ ** change the frequency response of this filter by *\forth{SET-FIR-TUNING}.
\ *P *\forth{BAND-PASS BAND-STOP USER-FIR} filters and the *\forth{SET-FIR-TUNING SET-FIR-WINDOWING SET-FIR-COEFF} are only possible
\ ** when the fftw3 library is installed.
	['] (init-fir-filter)
	#FIR_BUFF DFLOATS 5 * $80 +
	#fir_struct MAKE&ALLOCATE-PLUGIN-OBJECT
	#FIR_BUFF 1- MIN 2 MAX 1 OR TO plug_taps
	TO plug_filtype  #fir-id TO plug_type DEFAULT-FIR-WINDOW TO fir-window 0 TO fir-tune-xt 0 TO fir-coeff-xt
	0 TO fir-linear-equal
	CALC-FIR-FILTER
	['] (set-fir-filter-quality) TO plug_xt_quality
	['] calc-fir-val TO plug_xt
	['] (set-fir-filter-frequency) TO plug_xt_tune
	@DSP ;
: FIR-FILTER:	( type size -- ) \ name
\ *G Defines a named *\forth{MAKE-FIR-FILTER} plugin object, calling 'name' returns the object address.
	MAKE-FIR-FILTER CONSTANT ;

: MAKE-EQUALIZER	( size -- object )
\ *G Define an linear phase equalizer object with size taps corrected to odd, it's a special type of the generic FIR filter.
\ ** It offers 256 frequency bands in the
\ ** 0-JACK-SAMPLE-RATE/2 range, each band has it's own defined level. You can set these levels by
\ ** *\forth{SET-EQUALIZER-BAND} and *\forth{SET-EQUALIZER-RANGE}.
\ ** You may also select a different number of taps later via *\forth{SET-OBJ-QUALITY ( taps object -- )}.
\ *P For iForth users only: the showfft module now can directly control a bound equalizer filter,
\ ** see *\forth{SET-FFT-EQUALIZER} in the gui/showfft module.
	ALIGN HERE SWAP #256 SFLOATS ALLOT		\ static size
	BAND-PASS SWAP MAKE-FIR-FILTER -->DSP TO plug_static
	0 TO plug_xt_tune
	['] calc-equalizer TO fir-tune-xt
	#256 0 DO 1e0 I *equal-level SF! LOOP
	EQUALIZER TO plug_filtype CALC-FIR-FILTER
	@DSP ;
: EQUALIZER:	( size -- )
\ *G Defines a named *\forth{MAKE-EQUALIZER} plugin object, calling 'name' returns the object address.
	MAKE-EQUALIZER CONSTANT ;
: SET-EQUALIZER-BAND	\ ( object -- ) (F dBlevel frequency -- )
\ *G Sets the level for the specified frequency in the equalizer object and calculates the filter coefficients.
	-->EQUALIZER frequ->equal *equal-level dB->level SF!
	CALC-FIR-FILTER ;

: SET-EQUALIZER-RANGE	\ ( object -- )	(F dBlevel1 frequency1 dBlevel2 frequency2 -- )
\ *G Sets a linear level curve between the level/frequency pairs in for the equalizer objects levels
\ ** and calculates the filter coefficients. At least one point is set.
	-->EQUALIZER frequ->equal FSWAP frequ->equal SWAP	\ idx1 idx2		level1 level2
	2DUP >	IF SWAP FSWAP THEN
	2DUP = IF 1+ THEN DUP #255 = IF 1+ THEN #256 MIN SWAP	\ last first
	2DUP - OVER 2SWAP					\ diff first  last first
	?DO	FOVER FOVER FOVER F-
		I OVER - S>F OVER S>F F/			\ diff first		level1 level2 level1 diff fract
		fir-linear-equal 0= IF PI F* PI F+ FCOS 1e0 F+ 2e0 F/ THEN
		F*
		F+ I *equal-level dB->level SF!
	LOOP
	FDROP FDROP 2DROP CALC-FIR-FILTER ;
: SET-EQUALIZER-MODE	( mode object -- )
\ *G Sets the approximation mode for *\forth{SET-EQUALIZER-RANGE}, it can be either 0 for linear or 1 for sinoid mode.
	-->EQUALIZER TO fir-linear-equal ;

: EXPORT-EQUALIZER	( name len object -- )
\ *G Exports the equalizer setings to a file
	-->EQUALIZER
	$100 ALLOCATE ?ALLOCATE 0 LOCALS| fd buff |
	$FJACKSYSFILE COUNT buff PLACE  buff PLACE+
	buff COUNT R/W CREATE-FILE SWAP TO fd 0=
	IF plug_static #256 SFLOATS fd WRITE-FILE ?FILE fd CLOSE-FILE ?FILE ELSE S" Can't export the equalizer data because file is not writable" !info THEN
	buff FREE DROP ;
: IMPORT-EQUALIZER	( name len object -- )
\ *G Imports the equalizer setings from a file
	-->EQUALIZER
	$100 ALLOCATE ?ALLOCATE 0 LOCALS| fd buff |
	$FJACKSYSFILE COUNT buff PLACE  buff PLACE+
	buff COUNT R/O OPEN-FILE SWAP TO fd 0=
	IF plug_static #256 SFLOATS fd READ-FILE ?FILE DROP fd CLOSE-FILE ?FILE CALC-FIR-FILTER THEN
	buff FREE DROP ;


: SET-FIR-WINDOWING	( windowmode object -- )
\ *G Calculating the filter coefficents is done with a IFFT algorithm using a windowing function. This windowing
\ ** function can be set for each filter with *\forth{SET-FIR-WINDOWING}.
\ ** See the literature for descriptions of these windowing functions.
\ ** The default is set by *\forth{DEFAULT-FIR-WINDOW}.
\ *P The blackmann window *\forth{0} gives the deepest out-of-band suppression with a modest steepness.
\ *P The hamming window *\forth{1} has a steeper flank with less out of band suppression
\ *P The rectangle window *\forth{2} has the steepest flank, highest ripple and lowest out-of-band suppression.
	-->DSP CHK-FIR TO fir-window CALC-FIR-FILTER ;

: SET-FIR-TUNING	( 0|tuning-xt object -- )
\ *G The *\forth{MAKE-FIR-FILTER}s frequency response curve of each filter can be tuned to your needs by
\ ** adding a tuning-function-xt describing the desired freuency response. The function must have this stack effect:
\ *C (F level frequency -- your-level )
\ *P It must be able to return values for frequencies from 0 to sample-rate/2.
	-->DSP CHK-FIR TO fir-tune-xt CALC-FIR-FILTER ;

: SET-FIR-COEFF		( 0|set-coeff-xt object -- )
\ *G The *\forth{USER-FIR MAKE-FIR-FILTER} is a very special type of FIR filter, the dsp algorithm is the same but
\ ** there are many situations where you don't want to specify a frequency response of the filter but want to add effects
\ ** like hall, echo, delay, averagers and more.
\ *P To use this sort of filter you will need to think about the way the filter coefficients are calculated.
\ ** Whenever the filter is tuned, resized or defined there is the default function that sets x[0] to 1.0e0, all others
\ ** are zero. To define a filter according to your needs you will have to tell the coefficient-calculating-function execution token xt.
\ ** The calculating function must take two integer parameters *\forth{( number-of-filtertaps index)} and must return the float *\forth{(F -- coefficient)}
\ ** for that tap. This automatically set the filter size, so don't use *\forth{SET-OBJ-QUALITY} after calculating.
\ ** The maximum number of taps is *\forth{#FIR_BUFF} so when you need a larger range think of the generic effects plugin.
\ *P The graphics shows the signal flow, the number of taps is the number of multipliers.
	-->DSP CHK-FIR TO fir-coeff-xt CALC-FIR-FILTER ;
\ *P *\image{"The generic FIR filter signal flow" "firkern" "png" 50 50}
