(*
 * LANGUAGE    | iFORTH, ANS Forth
 * PROJECT     | Audio processing
 * DESCRIPTION | Implementing iir filters for jack
 * CATEGORY    | signal processing tool
 * AUTHOR      | Hanno Schwalm
 * LAST CHANGE | 2009-04-28 07:55:04, Hanno Schwalm
 *)

JACKAUDIO DEFINITIONS
\ *S The Infinite Impulse Response Filter

HERE ," IIR Filter" CONSTANT #iir-id

#PRIVATE-DSP
	DSP-VALUE iir_parameters
	DSP-VALUE iir_Q
	CONSTANT #iir_struct

0	IIR-FVALUE A0
	IIR-FVALUE A1
	IIR-FVALUE A2
	IIR-FVALUE B1
	IIR-FVALUE B2
	IIR-FVALUE Z0
	IIR-FVALUE Z1
	CONSTANT #FILPAR

CREATE butterworth-table \ BUTTERWORTH filter table for cascaded 10pole
	1.9754E0 DF,
	1.7820E0 DF,
	1.4142E0 DF,
	0.9080E0 DF,
	0.3129E0 DF,

0e0 FVALUE _a0val
0e0 FVALUE _a1val
0e0 FVALUE _a2val
0e0 FVALUE _b1val
0e0 FVALUE _b2val
0e0 FVALUE _divval
0e0 FVALUE _Wval^2
0e0 FVALUE _Wval
: CLEAR-IIR-VALUES
	0e0 TO _a0val	0e0 TO _a1val	0e0 TO _a2val
	0e0 TO _b1val	0e0 TO _b2val	0e0 TO _divval ;
: CALC-IIR-FILTER
	20 TO plug_timeshift	\ a very rough approximation
	JACK-SAMPLE-RATE S>F plug_frequency F/ PI FSWAP  F/ FTAN 1/F
	FDUP TO _Wval FDUP F* TO _Wval^2  CLEAR-IIR-VALUES
	plug_filtype
	CASE	LOW-PASS  OF 1e0 TO _a0val  0e0 TO _a1val  0e0 TO _a2val  1 TO iir_Q	ENDOF
		HIGH-PASS OF 0e0 TO _a0val  0e0 TO _a1val  1e0 TO _a2val  1 TO iir_Q	ENDOF
		BAND-PASS OF 0e0 TO _a0val  1e0 TO _a1val  0e0 TO _a2val  5 TO iir_Q	ENDOF
		BAND-STOP OF 1e0 TO _a0val  0e0 TO _a1val  1e0 TO _a2val  		ENDOF
		TRUE ABORT" Not a valid filtertype for IIR filters"
	ENDCASE
	_a1val iir_Q S>F F/ TO _a1val
 	5 0 DO	butterworth-table I DFLOATS +  DF@ iir_Q S>F F/ TO _b1val  1e0 TO _b2val
		iir_parameters I #FILPAR * +
		1e0	_b1val _Wval F* F+	_b2val _Wval^2 F* F+		TO _divval
		_a0val	_a1val _Wval F* F+	_a2val _Wval^2 F* F+		_divval F/		TO A0
		2e0	_a0val			_a2val _Wval^2 F* F- F*		_divval F/		TO A1
		_a0val	_a1val _Wval F* F-	_a2val _Wval^2 F* F+		_divval F/		TO A2
		2e0	1e0			_b2val _Wval^2 F* F- F*		_divval F/ FNEGATE 	TO B1
		1e0	_b1val _Wval F* F-	_b2val _Wval^2 F* F+		_divval F/ FNEGATE	TO B2
		0e0 TO Z0
		0e0 TO Z1
		DROP
	LOOP ;

: 10pole	\ ( -- ) 	(F val 'val )
	iir_parameters
  	FDUP A0 F* Z0 F+ FSWAP FOVER FOVER A1 F* FSWAP B1 F* F+ Z1 F+ TO Z0 A2 F* FOVER B2 F* F+ TO Z1 #FILPAR +
  	FDUP A0 F* Z0 F+ FSWAP FOVER FOVER A1 F* FSWAP B1 F* F+ Z1 F+ TO Z0 A2 F* FOVER B2 F* F+ TO Z1 #FILPAR +
  	FDUP A0 F* Z0 F+ FSWAP FOVER FOVER A1 F* FSWAP B1 F* F+ Z1 F+ TO Z0 A2 F* FOVER B2 F* F+ TO Z1 #FILPAR +
  	FDUP A0 F* Z0 F+ FSWAP FOVER FOVER A1 F* FSWAP B1 F* F+ Z1 F+ TO Z0 A2 F* FOVER B2 F* F+ TO Z1 #FILPAR +
 	FDUP A0 F* Z0 F+ FSWAP FOVER FOVER A1 F* FSWAP B1 F* F+ Z1 F+ TO Z0 A2 F* FOVER B2 F* F+ TO Z1
	DROP ;

: (set-iir-frequency)
	TO plug_frequency CALC-IIR-FILTER ;
: (set-iir-quality)
	TO iir_Q CALC-IIR-FILTER ;
: CHK-IIR
 	plug_type #iir-id <> ABORT" invalid IIR-FILTER object" ;
: (init-iir-filter)
	plug_dynmem #16 #B-ALIGNED TO iir_parameters ;

FORTH DEFINITIONS

: MAKE-IIR-FILTER	( type -- filter)
\ *G Constructs a 10pole IIR-FILTER (Butterworth) plugin object and leaves the address of the object.
\ ** The freqency of this filter can be set as usual with *\forth{SET-OBJ-FREQUENCY}
\ ** the quality (or sharpness) of the filter with *\forth{SET-OBJ-QUALITY}
	['] (init-iir-filter)
	#FILPAR 6 * #16 + #iir_struct MAKE&ALLOCATE-PLUGIN-OBJECT
	TO plug_filtype #iir-id TO plug_type  #10 TO iir_Q  10e0 TO plug_frequency
	CALC-IIR-FILTER
	['] 10pole TO plug_xt
	['] (set-iir-frequency) TO plug_xt_tune
	['] (set-iir-quality) TO plug_xt_quality
	@DSP ;
: IIR-FILTER:		( type -- )	\ name
\ *G Defines a named iir-filter plugin object, calling 'name' returns the object address.
	MAKE-IIR-FILTER CONSTANT ;
