(*
 * LANGUAGE    : Forth
 * PROJECT     : Working on MPE and iForth
 * DESCRIPTION : Interface for the libjack
 * CATEGORY    : Realtime Audio tools
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE  : 2009-05-04 01:24:36
		: 2011-07-24 23:10:21 added functions for session and latency support
 *)

JACKAUDIO DEFINITIONS
WINDOWS? [IF]
	Library: libjack.dll
	Library: libfftw3-3.dll
\	Library: msvcrt.dll
[THEN]

LINUX? [IF]
	Library: libfftw3.so.3
	Library: libjack.so.0
[THEN]

DARWIN? [IF]
	Library: libjack.0.dylib
	Library: libfftw3.3.dylib
[THEN]

\ enum JackPortFlags
$001 pCONSTANT JackPortIsInput
$002 pCONSTANT JackPortIsOutput
$004 pCONSTANT JackPortIsPhysical
$008 pCONSTANT JackPortCanMonitor
$010 pCONSTANT JackPortIsTerminal

\ enum JackOptions
$000 pCONSTANT JackNullOption
$001 pCONSTANT JackNoStartServer
$002 pCONSTANT JackUseExactName
$004 pCONSTANT JackServerName
$008 pCONSTANT JackLoadName
$010 pCONSTANT JackLoadInit
$020 pCONSTANT JackSessionID

\ enum JackStatus
 $001 pCONSTANT JackFailure
 $002 pCONSTANT JackInvalidOption
 $004 pCONSTANT JackNameNotUnique
 $008 pCONSTANT JackServerStarted
 $010 pCONSTANT JackServerFailed
 $020 pCONSTANT JackServerError
 $040 pCONSTANT JackNoSuchClient
 $080 pCONSTANT JackLoadFailure
 $100 pCONSTANT JackInitFailure
 $200 pCONSTANT JackShmFailure
 $400 pCONSTANT JackVersionError
 $800 pCONSTANT JackBackendError
$1000 pCONSTANT JackClientZombie

$001 pCONSTANT JackSessionSave
$002 pCONSTANT JackSessionSaveAndQuit
$003 pCONSTANT JackSessionSaveTemplate
$001 pCONSTANT JackSessionSaveError
$002 pCONSTANT JackSessionNeedTerminal

0 pCONSTANT JackCaptureLatency
1 pCONSTANT JackPlaybackLatency

\ enum jack_transport_state_t
$000 pCONSTANT JackTransportStopped
$001 pCONSTANT JackTransportRolling
$002 pCONSTANT JackTransportLooping
$003 pCONSTANT JackTransportStarting
$004 pCONSTANT JackNetStarting
#-2  pCONSTANT JackZombie

Extern: int "C" jack_get_version_string			( void );
Extern: int "C" jack_port_name_size			( void );
Extern: jack_client_t "C" jack_client_open		( const char * client_name, int options, void * status);
Extern: int "C" jack_client_close			( jack_client_t * client);
Extern: int "C" jack_activate				( jack_client_t * client);
Extern: int "C" jack_deactivate				( jack_client_t * client);
Extern: int "C" jack_get_client_name			( jack_client_t * client);
Extern: int "C" jack_is_realtime			( jack_client_t * client);
Extern: jack_nframes_t "C" jack_get_sample_rate		( jack_client_t * client);
Extern: float "C" jack_cpu_load				( jack_client_t * client);
Extern: jack_nframes_t "C" jack_get_buffer_size		( jack_client_t * client);

Extern: void "C" jack_port_get_latency_range		( jack_port_t * port, int mode, jack_latency_range_t * range);
Extern: void "C" jack_port_set_latency_range		( jack_port_t * port, int mode, jack_latency_range_t * range);
Extern: int "C" jack_recompute_total_latencies		( jack_client_t * client );

Extern: int "C" jack_set_session_callback		( jack_client_t * client, void * session_callback, void * arg);
Extern: int "C" jack_session_reply			( jack_client_t * client, jack_session_event_t * event);
Extern: void "C" jack_session_event_free		( jack_session_event_t * event);

\ Extern: int "C" jack_client_get_uuid			( jack_client_t * client);
Extern: int "C" jack_get_uuid_for_client_name		( jack_client_t * client, const char * client_name);
Extern: int "C" jack_reserve_client_name		( jack_client_t * client, const char * name, const char * uuid);

Extern: jack_port_t "C" jack_port_register		( jack_client_t * client, const char * port_name, const char * port_type, int flags, int buffer_size);
Extern: int "C" jack_port_unregister			( jack_client_t * client, jack_port_t * id);
Extern: int "C" jack_port_flags				( jack_port_t * port);
Extern: int "C" jack_port_get_buffer			( jack_port_t * port, jack_nframes_t size );
Extern: int "C" jack_port_connected			( jack_port_t * port);
Extern: int "C" jack_port_name				( jack_port_t * port);
Extern: int "C" jack_port_disconnect			( jack_client_t * client, jack_port_t * port);
Extern: int "C" jack_connect				( jack_client_t * client, const char * source_port, const char * destination_port);
Extern: int "C" jack_disconnect				( jack_client_t * client, const char * source_port, const char * destination_port);
Extern: int "C" jack_port_get_connections		( jack_port_t * port);
Extern: int "C" jack_port_by_name			( jack_client_t * client, jack_port_t * port);

Extern: void "C" jack_on_shutdown			( jack_client_t * client, void * Shutdowncallback, void * arg );
Extern: int "C" jack_set_process_callback		( jack_client_t * client, void * process_callback, void * arg);
Extern: int "C" jack_set_sample_rate_callback		( jack_client_t * client, void * srate_callback, void * arg);
Extern: int "C" jack_set_buffer_size_callback		( jack_client_t * client, void * bufsize_callback, void * arg);
\ Extern: int "C" jack_set_client_registration_callback	( jack_client_t * client, void * registration_callback, void * arg);
\ Extern: int "C" jack_set_port_registration_callback	( jack_client_t * client, void * registration_callback, void * arg);
Extern: int "C" jack_set_port_connect_callback		( jack_client_t * client, void * connect_callback, void * arg);
Extern: int "C" jack_set_graph_order_callback		( jack_client_t * client, void * graph_callback, void * arg);
Extern: int "C" jack_set_xrun_callback			( jack_client_t * client, void * xrun_callback, void * arg);
Extern: int "C" jack_set_sync_callback			( jack_client_t * client, void * sync_callback, void * arg);
Extern: int "C" jack_set_latency_callback		( jack_client_t * client, void * latency_callback, void * arg-fjack);

Extern: void "C" jack_set_error_function		( void * writeerror);
Extern: void "C" jack_set_info_function			( void * function);

Extern: jack_transport_state_t "C" jack_transport_query	( jack_client_t * client, jack_position_t * pos);
Extern: void "C" jack_transport_start			( jack_client_t * client);
Extern: void "C" jack_transport_stop			( jack_client_t * client);

Extern: void "C" jack_free				( void * ptr );
