(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Audio processing
 * DESCRIPTION : Implementing a mixer plugin
 * CATEGORY    : audio processing tool
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-12-03 18:37:43
 *)


[UNDEFINED] JACKAUDIO [IF] CR .( fjack/examples/mixproc needs fjack ... ) ABORT [THEN]

\ *S The Channel-Mixer (and processing) plugin object.
\ *P Allows mixing and processing of audio data.
DECIMAL JACKAUDIO DEFINITIONS

HERE ," Channel mixer" CONSTANT #mixer-id
: CHK-MIXER	#mixer-id plug_type <> ABORT" Invalid Mixer object" ;

#PRIVATE-DSP
	DSP-VALUE mix-levels
	DSP-VALUE mix-delays
	DSP-VALUE mix-vals
	DSP-VALUE mix-action
	CONSTANT #mix_struct

: @mix-level	( i -- ) ( F -- level )	SFLOATS mix-levels + SF@ ;
: @mix-delay	( i -- ) ( F -- time )	SFLOATS mix-delays + SF@ ;
: calc-mix-val	\ ( -- ) (F val -- val' )
	plug_level F* mix-vals SF! 1	( validcounter )
	JACK-CHANNELS 0
	DO	JACK-IO-SLOTS 0
		DO	I JACK-CHANNELS * J + @mix-level 0e0 F>
			IF	I 1- J SLOT/CHANNEL-HISTORY ?DUP
  				IF	I JACK-CHANNELS * J + SWAP OVER @mix-delay @INTERPOLATED-HISTORY-DATA
					@mix-level F* DUP SFLOATS mix-vals + SF! 1+
  				THEN
			THEN
		LOOP
	LOOP
	mix-vals SWAP mix-action EXECUTE ;
: mixer-adding	( array count -- ) ( F -- sum )
	0e0 0 DO SF@+ F+ LOOP DROP ;

: (init-mixer-object)
	plug_dynmem DUP TO mix-levels
	JACK-IO-SLOTS JACK-CHANNELS * SFLOATS + DUP TO mix-delays
	JACK-IO-SLOTS JACK-CHANNELS * SFLOATS + TO mix-vals
	mix-levels JACK-IO-SLOTS JACK-CHANNELS * 3 * SFLOATS ERASE ;

FORTH DEFINITIONS
: MAKE-MIXER	( -- object )
\ *G Construct a mixer plugin object and leave it's address.
\ ** The value in the processing stream is calculated by the presented value plus
\ ** all other slots/channels that have been added to the mixing by *\forth{SET-MIXER-LEVEL}.
\ ** *\forth{SET-OBJ-LEVEL} sets the level of the original source of this plugin.
\ *P The standard way to mix the audio data is just adding all values. There is a special feature here
\ ** to do some more tricks on the data, you may replace the 'adding' function by some other word with
\ ** *\forth{SET-MIXER-ACTION}.
	['] (init-mixer-object)
	1 CELLS 3 SFLOATS + JACK-IO-SLOTS * JACK-CHANNELS * $20 + #mix_struct MAKE&ALLOCATE-PLUGIN-OBJECT
	MIXER-PLUGIN TO plug_filtype   #mixer-id TO plug_type
	['] mixer-adding TO mix-action
	['] calc-mix-val TO plug_xt @DSP ;

: MIXER:
\ *G Creates a named mixer plugin.
	MAKE-MIXER CONSTANT ;

: SET-MIXER-INPUT	\ ( slot channel mixer -- ) (F level delay -- )
\ *G Sets the level and delaytime (seconds) in the mixer.
\ ** A slot-number <0 means the input port. You can only mix data from smaller slots
\ ** than the mixer objects slot. The graph shows the mixer inputs
	-->DSP CHK-MIXER
	DUP 0 JACK-CHANNELS WITHIN 0= ABORT" Illegal channel number for mixer"
	OVER -1 plug_slot WITHIN 0= ABORT" Illegal slot number for mixer"
	SWAP 1+ JACK-CHANNELS * + SFLOATS DUP
	mix-delays + SF! mix-levels + SF! ;

\ *P *\image{"Mixer sources" "mixer" "png" 50 50}

: SET-MIXER-ACTION	( action-xt object -- )
\ *G This replaces the default 'adding' function of a mixer plugin object by another one.
\ *P The stack effect of the new function must be
\ *C : my-function ( array cnt -- ) (F -- value )
\ *P The array holds cnt sfloats of input data read from all the activated mixer channels,
\ ** you can do whatever you want to do with these data but you *\bold{MUST} take the parameters and
\ ** *\bold{MUST} leave the floating result. All data have already been changed to the desired level,
\ ** the first position in the array holds the original input stream to this slot.
	-->DSP CHK-MIXER
	TO  mix-action ;
: .MIXER-INFO	( object -- )
\ *G Show the status of the mixer object; level is absolute; delay measured in seconds.
	-->DSP CHK-MIXER
	JACK-CHANNELS 0
	DO	JACK-IO-SLOTS 0
		DO	I JACK-CHANNELS * J + @mix-level 0e0 F>
			IF	I 1- J SLOT/CHANNEL-HISTORY
  				IF	CR ."  S:" J .idx ."  C:" I .idx
					."   Level:" I JACK-CHANNELS * J + @mix-level F.
					."   Delay:" I JACK-CHANNELS * J + @mix-delay F.
				THEN
			THEN
		LOOP
	LOOP ;
