(*
 * LANGUAGE    : Forth; iForth and MPE Forth working
 * PROJECT     : Forth Tool
 * DESCRIPTION : Audio processing client using JACK
 * CATEGORY    : Realtime Audio tools
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-04-28 07:55:40 Hanno Schwalm hanno@schwalm-bremen.de
 *		2009-05-09 00:05:17; COMP-FCHANNEL, added; bug when SFLOATS were NOT cells removed
			also slightly faster code as not swapping in COMP-xx, must be optimised
		: 2009-05-22 18:10:53 added DSP-DVALUE

 *)
\ *S Datatypes used in the client
\ *P Datatype implementation details might change in future releases.
\ ** Currently the object-makers all return object-pointers or are defined
\ ** as *\forth{CONSTANT}s. Please read the fjack/types.frt to understand how it works.
JACKAUDIO DEFINITIONS
: COMP-FVALUE	( n -- )
		CASE	0 OF	POSTPONE DF@			ENDOF
			1 OF	POSTPONE DF!			ENDOF
			-1 OF	POSTPONE DF+!			ENDOF
			TRUE ABORT" Wrong method for FVALUE"
		ENDCASE ;
: COMP-VALUE	( n -- )
		CASE	0 OF	POSTPONE @			ENDOF
			1 OF	POSTPONE !			ENDOF
			-1 OF	POSTPONE +!			ENDOF
			TRUE ABORT" Wrong method for VALUE"
		ENDCASE  ;
: COMP-DVALUE	( n -- )
		CASE	0 OF	POSTPONE 2@			ENDOF
			1 OF	POSTPONE 2!			ENDOF
			-1 OF	POSTPONE 2+!			ENDOF
			TRUE ABORT" Wrong method for 2VALUE"
		ENDCASE  ;


IS-IFORTH?	[IF]	USER DSP-OBJECT
			USER FIFO-OBJECT
			USER WBD-OBJECT
		[THEN]
IS-VFX?		[IF]	1 CELLS +USER DSP-OBJECT
			1 CELLS +USER FIFO-OBJECT
			1 CELLS +USER WBD-OBJECT
		[THEN]

: -->WBD	( object -- )	WBD-OBJECT ! ;
: -->DSP	( object -- )	DSP-OBJECT ! ;
: @DSP		( -- object )	DSP-OBJECT @ ;
: -->FIFO	( object -- )	FIFO-OBJECT ! ;
: @FIFO		( -- object )	FIFO-OBJECT @ ;

: COMP-CHANNEL,		POSTPONE CELLS @ POSTPONE LITERAL POSTPONE + ;
: COMP-FCHANNEL,	POSTPONE DFLOATS @ POSTPONE LITERAL POSTPONE + ;
: COMP-DSP-OBJ,		POSTPONE DSP-OBJECT POSTPONE @  @ POSTPONE LITERAL  POSTPONE + ;
: COMP-IIR-PAR,		POSTPONE DUP @ POSTPONE LITERAL POSTPONE + ;
: COMP-FIFO-OBJ,	POSTPONE FIFO-OBJECT POSTPONE @  @ POSTPONE LITERAL  POSTPONE + ;
: COMP-EFF-OBJ,		POSTPONE WBD-OBJECT POSTPONE @  @ POSTPONE LITERAL  POSTPONE + ;

IS-IFORTH?
[IF]
: CHAN-VALUE	( addr -- addr')	\ name
	CREATE	IMMEDIATE COMPILE-ONLY ALIGNED DUP , JACK-CHANNELS CELLS +
	DOES>	COMP-CHANNEL, %VAR @ %VAR OFF COMP-VALUE ;

: CHAN-FVALUE	( addr -- addr')	\ name
	CREATE	IMMEDIATE COMPILE-ONLY DFALIGNED DUP , JACK-CHANNELS DFLOATS +
	DOES>	COMP-FCHANNEL, %VAR @ %VAR OFF COMP-FVALUE ;

: DSP-VALUE	( n -- n' )		\ name
	CREATE	IMMEDIATE COMPILE-ONLY ALIGNED DUP , 1 CELLS +
	DOES>	COMP-DSP-OBJ, %VAR @ %VAR OFF COMP-VALUE ;

: DSP-DVALUE	( n -- n' )		\ name
	CREATE	IMMEDIATE COMPILE-ONLY ALIGNED DUP , 2 CELLS +
	DOES>	COMP-DSP-OBJ, %VAR @ %VAR OFF COMP-DVALUE ;

: DSP-FVALUE	( n -- n' )	\ name
	CREATE	IMMEDIATE COMPILE-ONLY DFALIGNED DUP , 1 DFLOATS +
	DOES>	COMP-DSP-OBJ, %VAR @ %VAR OFF COMP-FVALUE ;

: FIFO-VALUE	( n -- n' )
	CREATE	IMMEDIATE COMPILE-ONLY ALIGNED DUP , 1 CELLS +
	DOES>	COMP-FIFO-OBJ, %VAR @ %VAR OFF COMP-VALUE ;
: WBD-VALUE	( n -- n' )
	CREATE	IMMEDIATE COMPILE-ONLY ALIGNED DUP , 1 CELLS +
	DOES>	COMP-EFF-OBJ, %VAR @ %VAR OFF COMP-VALUE ;

: IIR-FVALUE	( n -- n' )	\ name
	CREATE	IMMEDIATE COMPILE-ONLY DFALIGNED DUP , 1 DFLOATS +
	DOES>	COMP-IIR-PAR, %VAR @ %VAR OFF COMP-FVALUE ;
[THEN]


IS-VFX?
[IF]
: @iFORTH-OPS
	OperatorType @ OperatorType OFF DUP 5 = IF DROP -1 THEN ;

: CHAN-VAL,	 \ xt
	>BODY COMP-CHANNEL, @iFORTH-OPS COMP-VALUE ;
: CHAN-VALUE
	CREATE	ALIGNED DUP , JACK-CHANNELS CELLS +
		['] CHAN-VAL, SET-COMPILER
	INTERP>	NoInterp ;

: CHAN-FVAL,	\ xt
	>BODY COMP-FCHANNEL, @iFORTH-OPS COMP-FVALUE ;
: CHAN-FVALUE
	CREATE	DFALIGNED DUP , JACK-CHANNELS DFLOATS +
		['] CHAN-FVAL, SET-COMPILER
	INTERP>	NoInterp ;

: DSP-VAL,	\ xt --
	>BODY COMP-DSP-OBJ, @iFORTH-OPS COMP-VALUE ;
: DSP-VALUE
	CREATE	ALIGNED DUP , 1 CELLS +
		['] DSP-VAL, SET-COMPILER
	INTERP> NoInterp ;

: DSP-DVAL,	\ xt --
	>BODY COMP-DSP-OBJ, @iFORTH-OPS COMP-DVALUE ;
: DSP-DVALUE
	CREATE	ALIGNED DUP , 2 CELLS +
		['] DSP-DVAL, SET-COMPILER
	INTERP> NoInterp ;

: DSP-FVAL,	\ xt
	>BODY COMP-DSP-OBJ, @iFORTH-OPS COMP-FVALUE ;
: DSP-FVALUE
	CREATE	DFALIGNED DUP , 1 DFLOATS +
		['] DSP-FVAL, SET-COMPILER
	INTERP> NoInterp ;

: FIFO-VAL,
	>BODY COMP-FIFO-OBJ, @iFORTH-OPS COMP-VALUE ;
: FIFO-VALUE
	CREATE	ALIGNED DUP , 1 CELLS +
		['] FIFO-VAL, SET-COMPILER
	INTERP> NoInterp ;

: EFF-VAL,
	>BODY COMP-EFF-OBJ, @iFORTH-OPS COMP-VALUE ;
: WBD-VALUE
	CREATE	ALIGNED DUP , 1 CELLS +
		['] EFF-VAL, SET-COMPILER
	INTERP> NoInterp ;

: IIR-VAL,	\ xt
	>BODY COMP-IIR-PAR, @iFORTH-OPS COMP-FVALUE ;
: IIR-FVALUE	( n -- n' )
	CREATE	DUP , 1 DFLOATS +
		['] IIR-VAL, SET-COMPILER
	INTERP> NoInterp ;

[THEN]
