\ HANOI.FTH - Towers of hanoi

((
A Famous Forth implementation of a classic problem. The original
version was written by Peter Midnight. This version was
converted to PowerForth/386 by MPE
from FORTH DIMENSIONS, Vol II, No. 2, page 32.
and then to VFX Forth.

This is also a good demonstration of the use of RECURSE.

To adjust the display to the speed of your CPU use:
  n SET-SPEED
where 500000 is a good start in 2011.

To run the demo, use
  n TOWERS
where n is the number of rings, e.g. 5.

To stop the display, press any key.
))

: hanoi  ;			\ marker

[defined] target_386_windows [if]
include %vfxpath%\Lib\Win32\Genio\Terminal
terminal: HanoiTerm
  RGB_BLACK HanoiTerm term-forecolor !	\ foreground colour
  RGB_WHITE HanoiTerm term-backcolor !	\ background colour

2variable oldIO	\ holds previous I/O device
: OpenHanoi	\ --
  op-handle @ ip-handle @ oldIO 2!
  80 25 10 HanoiTerm open-gio abort" can't open terminal"
  dup op-handle !  ip-handle !
;
: CloseHanoi	\ --
  #1000 ms
  oldIO 2@ ip-handle ! op-handle !
  HanoiTerm close-gio drop
;
[else]
: OpenHanoi  ;
: CloseHanoi  ;
[then]

decimal

12 constant nmax		\ max number of rings
variable <n>			\ holder
  nmax <n> !
variable <speed>		\ delay counter
  10000 <speed> !

create ring  nmax 2+ allot	\ array [1...n] of bytes
  ring nmax 2+ erase
char | constant v-bar		\ vertical bar character (pole)
char = constant base-char	\ base character
char o constant ring-char	\ ring character

: set-speed	\ n --
  <speed> !  ;

: speed		\ -- n
  <speed> @  ;

: delay		\ n --
  0 ?do				\ outer loop
    speed 0
    ?do  pause  loop		\ and waste time
  loop
;

: h-emit	\ char -- ; print and wait
  emit  1 delay  ;

[undefined] 4dup [if]
: 4dup		\ n1 n2 n3 n4 -- n1 n2 n3 n4 n1 n2 n3 n4
  3 pick  3 pick  3 pick  3 pick
;
[then]

: n		\ -- n ; get <n>
  <n> @  ;

: pos		\ location pos -- coordinate
  n 2* 1+ *  n +  ;

: halfdisplay	\ color size --
  0
  ?do  dup h-emit  loop
  drop
;

: <display>	\ line color size --
  2dup halfdisplay rot 3 <
  if  bl  else  v-bar  then
  h-emit  halfdisplay
;

: display	\ size pos line color --
  swap >r
  -rot over - r@		\ -- color size pos line
  at-xy
  r>				\ -- color size line
  -rot <display>
;

: presence	\ tower ring presence -- boolean
  ring + c@ =
;

: line		\ tower line -- display-line-of-top
  4 swap  n 0
  do  dup i presence 0=		\ now add boolean (yuk)
      abs rot + swap
      loop
  drop
;

: raise		\ size tower --
  dup pos swap line 2 swap
  do  2dup i bl display 2dup
      i 1- ring-char display
  -1 +loop
  2drop
;

: lower		\ size tower --
  dup pos  swap line  1+ 2
  do  2dup i 1- bl display
      2dup i ring-char display
  loop
  2drop
;

: moveleft	\ size source.tower destiny.tower --
  pos   swap pos 1-
  do  dup i 1+ 1 bl display
      dup i 1 ring-char display
      -1
  +loop
  drop
;

: moveright	\ size source.tower destiny.tower --
  pos 1+  swap pos 1+
  do  dup i 1- 1 bl display
      dup i 1 ring-char display
  loop
  drop
;

: traverse	\ size source.tower destiny.tower --
  2dup >
  if  moveleft  else  moveright  then
;

: tmove		\ size source.tower destiny.tower --
  key? if
    0 n 4 + at-xy
    ( cursor-on )  ( abort )  200 throw
  then
  rot rot 2dup raise
  >r 2dup r> rot traverse
  2dup ring + 1- c! swap lower
  100 delay
;

: multimov	\ size source destiny spare --
  3 pick 1 =
  if    drop tmove
  else  >r >r swap 1- swap r> r>
        4dup swap recurse
        4dup drop  rot 1+ -rot  tmove
        -rot swap recurse
  then
;

: maketower	\ tower --
  pos  4 n +  3
  do  dup i at-xy  v-bar emit loop
  drop
;
4 <n> !
: makebase	\ --
  0 n 4 + at-xy  n 6 * 3 + 0
  do  base-char emit  loop
;

: makering	\ tower size --
  2dup ring + 1- c! swap lower
;

: setup		\ --
  cls  n 1+ 0
  do  1 ring i + c!  loop
  3 0
  do  i maketower  loop
  makebase  1 n
  do  0 i makering  -1 +loop
;

: (towers)	\ quantity --
  1 max  nmax min  <n> !
  ( cursor-off ) setup  n 2 0 1
  begin
    over pos n 4 + at-xy  n 0
    do   7 emit  25 delay   loop
    rot 4dup multimov
  again
;

: towers	\ quantity --
  OpenHanoi
  ['] (towers) catch
  if  drop  then
  CloseHanoi
  cr
;
