\ Coos Haak provides REMEMBER using an idea from Michael Gassanenko

\ Derived by MPE from code provided by Coos Haak on comp.lang.forth

\ REMEMBER is a version of MARKER, used in the form:
\   REMEMBER <name>
\ When <name> is executed it removes everything after <name>, leaving
\ <name> intact. This implementation depends on the use of ANEW.

\ The really useful parts of this code are the definitions of 
\ DELIMITER and $. The definition of $ may be included in the
\ VFX Forth kernel at a later date.

: delimiter	\ char -- char'
\ *G Given a starting delimiter characterfor a string, return the ending
\ ** delimiter character. If the ending delimiter cannot be found, return
\ ** the starting delimiter character.
  s" (){}[]<>" bounds do
    dup i c@ =
    if  drop  i 1 chars + c@  leave  endif
  2 chars +loop
;

#256 chars buffer: $-BUFFER
\ *G Temporary buffer used by $ whose contents are valid until the
\ ** next use of $.

: $		\ "<dlm1>text<dlm2>" -- c-addr u
\ *G Parse a string between the delimiters above.
\ ** When the string is between quotes (" or ') use them
\ ** at the beginning and the end.
  source dup >r				\ preserve original length
  >in @ /string  bl skip		\ skip whitespace
  r> swap - 1+ >in !			\ step input stream past first char
  c@ delimiter parse			\ -- addr len ; extract string
  state @ if
    postpone sliteral
  else
    tuck $-buffer swap move
    $-buffer swap			\ -- addr len
  endif
; immediate

: REMEMBER          ( "name" -- )
\ *G REMEMBER is a version of MARKER, used in the form:
\ ** "REMEMBER <name>".
\ ** When <name> is executed it removes everything after <name>, leaving
\ ** <name> intact. This implementation depends on the use of ANEW.
  >in @
  BL PARSE-WORD
  $ [S" : ] pad PLACE			\ :
  2DUP pad APPEND			\ : name
  $ [ $ { ANEW --] pad APPEND   	\ : name s" anew --
  pad APPEND              		\ : name s" anew --name
  $ [--} EVALUATE ; " EVALUATE ]
  pad APPEND
  pad COUNT				\ : name s" anew --name--" evaluate ;
  EVALUATE
  >in !
;

