\ screen to text file conversion utility

((
This code released to convert screen files to text files.

(c) MicroProcessor Engineering Ltd., 1991, 2019
133 Hill Lane
Southampton
SO1 5AF

Phone: +44 (0)23 8063 1441

This code to be compiled on VFX Forth v4.81 or higher.

20190419 SFP002 Converted to VFX
20190418 SFP001 Updated page handling
))

only forth definitions
decimal

[undefined] filedev: [if]
include /home/stephen/Products/VfxCommunity/Sources/Lib/Genio/file.fth
[then]

64 buffer: source-buff			\ source buffer
256 buffer: screen-name			\ screen file name
256 buffer: dest-name			\ destination file name
filedev: screen-file			\ source screen file
filedev: dest-file			\ destination file

variable fth-page#			\ page number (starts at zero)
variable fth-line#			\ line number (starts at zero)

1 value first-screen			\ first screen is numbered 0 or 1
1 value DOS?				\ if nz, generate CR/LF otherwise LF

variable #blank-lines			\ send a blank line?

: bye-bye	\ -- ; quit application
  key? drop  bye			\ flush output
;

: close-files	\ -- ; close all files
  screen-file close-gio drop
  dest-file close-gio drop
;

: (write-dest)	\ caddr len -- ; write to o/p file ; SFP001
  dest-file write-gio if		\ write code
    cr ." Error writing destination file: aborting"	\ say so
    close-files					\ close files
    bye-bye					\ quit
  endif
;

: write-dest           \ len -- ; write to o/p file
  source-buff swap (write-dest)
;

create crlf$	\ -- caddr
  2 c,  13 c,  10 c,		\ DOS
create lf$	\ -- caddr
  1 c,  10 c,			\ Unix

: write-crlf	\ --
\ *G Write EOL in DOS or Unix format.
  DOS?
  if  crlf$  else  lf$  then
  count (write-dest)
;

: read-source	\ -- t/f ; 0=success
\ *G Read next line in block from the source file.
  source-buff 64 screen-file readex-gio if	\ get one line
    drop
    cr ." End of source file: closing"		\ say so
    close-files					\ close files
    bye-bye					\ quit
  endif
  64 <>						\ 0 = read 64 bytes
;

: open-source	\ -- ; try to open source file
  screen-name count r/o screen-file open-gio nip if
    cr ." Can't open source file: aborting"
    cr ." Use SCR2FTH -? for help"
    bye-bye
  endif
;

: open-dest	\ -- ; try to open source file
  dest-name count r/w dest-file open-gio nip if
    cr ." Can't open destination file: aborting"
    cr ." Use SCR2FTH -? for help"
    bye-bye
  endif
;

: open-files	\ -- ; open source and dest files
  open-source open-dest
;

: send-line?	\ len -- len t/f ; write if first blank line
  dup 0= if			\ is length zero?
    1 #blank-lines +!		\ increment count
    #blank-lines @
    1 =
    if -1  else 0  endif	\ only write if #1
  else
    #blank-lines off		\ write if non-blank
    -1
  endif
;

: write-line		\ -- ; write a line to dest. file
  source-buff 64 -trailing nip		\ actual number of chars
  send-line? if				\ should we send or not
    write-dest  write-crlf		\ write text
  else
    drop				\ else tidy up
  endif
;

: page-break		\ -- ; write a page break
  12 source-buff c!		\ make FF
  1 write-dest			\ write FF
;

: ?page-top		\ -- ; header if line=0 ; SFP001
  fth-line# @ 0= if
    s" \ Page " (write-dest)
    fth-page# @ first-screen + s>d <# #s #> (write-dest)
    write-crlf write-crlf
  endif
;

: tidy-file		\ -- ; manage counters, and inform user
  1 fth-line# +!			\ increment line#
  fth-line# @ 16 = if			\ if end of page
    1 fth-page# +!			\ increment page number
    fth-line# off			\ reset line counter
    page-break
    13 emit ." Screen/page " 		\  say 'next page'
    fth-page# @ first-screen + .
  endif
;

: convert		\ -- ; do the actual conversion
  open-files				\ try to open source
  fth-page# off fth-line# off		\ reset counters
  begin
    read-source				\ get some source
    0=					\ while successful
  while
    ?page-top				\ SFP001
    write-line				\ write to destination
    tidy-file				\ do page breaks, etc
  repeat				\ and the next
  close-files				\ close the files
;

: signon	\ -- ; say who we are
  ." Screen File to text file converter, v3.00"
  cr ." (c) MicroProcessor Engineering Ltd, 1990, 2019"
  cr cr
;

: .help		\ -- ; show help screen
     ." Usage: SCR2FTH <options> -i <source> -o <dest>
  cr
  cr ." <Source> is the name of a screen file to be converted to a text file."
  cr ." <Dest> is the name of the destination file."
  cr ." File extensions must be provided."
  cr
  cr ." Options:"
  cr ." -dos   use CR/LF for line endings"
  cr ." -unix  use LF for line endings"
  cr ." -0     first screen is zero"
  cr ." -1     first screen is one"
  cr ." -?     help"
;

: -?		\ addr -- ; do they want help?
  .help	 bye-bye
;

: -i		\ -- ; get the name of the source file
  getPathSpec dup 0= if
    2drop  cr ." No source file specified"  -?
  then
  screen-name place
;

: -o		\ -- ; get the name of the source file
  getPathSpec dup 0= if
    cr ." No destination file supplied: using CONVERT.FTH"
    2drop  s" convert.fth"			\ set default
  endif
  dest-name place
;

: -dos		( -- )   1 to DOS?  ;
: -unix		( -- )   0 to DOS?  ;
: -0		( -- )   0 to first-screen  ;
: -1		( -- )   1 to first-screen  ;

: info			\ -- ;  show what is about to happen
  ." Converting " screen-name count type	\ first message
  ."  to " dest-name count type
  cr cr					\ tidy screen
;

 : it                    \ -- ; the whole thing
  signon	 			\ what it is
  CommandLine to-source
  -1 to source-id  >in off  'SourceFile off
  getPathSpec 2drop			\ skip program name
  source >in @ /string evaluate		\ gather input
  info					\ show what is about to happen
  convert		 		\ do it
  bye-bye				\ and quit
;

: stripProgramName    \ c-addr u -- 'c-addr 'u
\ *G Strip a leading program name from a command line, returning
\ **  the command line tail.
  dup 0= if  exit  then
  bl skip               \ c-addr u --
  dup if
    over c@ [char] " = if
      1 /string
      [char] " scan
    else
      bl scan
    then
    1 /string bl skip
  then
;


\ *********************
\ *S Windows executable
\ *********************

[defined] Target_386_Windows [if]

Extern: BOOL PASCAL AllocConsole( void );
Extern: HANDLE PASCAL GetStdHandle( DWORD nStdHandle );
Extern: BOOL PASCAL SetConsoleMode( HANDLE hCons, DWORD mode );
Extern: BOOL PASCAL GetConsoleMode( HANDLE hCons, DWORD * mode );

also system
: MakeConsole	\ --
\ Create a Windows console for the Forth system.
  AllocConsole if
    STD_INPUT_HANDLE GetStdHandle stdin !
    STD_OUTPUT_HANDLE GetStdHandle stdout !
    stdin @ ENABLE_PROCESSED_INPUT SetConsoleMode drop
  endif
;
previous

[-short-branches
: WinMain	\ hinst hpinst lpcommand nshow -- res
\ *G Windows console mode application entry.
  {: hInstance hPInstance lpszCommand nShow | icc[ 2 cells ] -- :}

  Win32Catch				\ faults directed at THROWN

  NULL to hWndMain

  init-io				\ set up default console
  op-handle @ dup to task0-io		\ default i/o setup
  dup ip-default !  op-default !	\ in case of callbacks

  MakeConsole				\ in case Windows did not
  signon
  lpszCommand zcount stripProgramName	\ -- caddr len ; command line tail
  dup if
    ['] evaluate catch if  2drop  then
  else
    2drop
  then
  info					\ show what is about to happen
  convert		 		\ do it
  cr  key? drop				\ KEY? forces o/p flush
  ExitCode @
;
short-branches]

assign WinMain to-do EntryPoint
saveconsole scr2fth			\ save to disc
bye					\ ready to try it

[then]


\ *******************
\ *S Linux executable
\ *******************

[defined] Target_386_Linux [if]

: LinMain	\ hinst hpinst lpcommand nshow -- res
\ *G Linux console mode application entry.
  {: hInstance hPInstance lpszCommand nShow -- :}

  init-io				\ set up default console
  op-handle @ dup ip-default !  op-default !	\ in case of callbacks

  signon
  lpszCommand zcount stripProgramName	\ -- caddr len ; command line tail
  dup if
    ['] evaluate catch if  2drop  then
  else
    2drop
  then
  info					\ show what is about to happen
  convert		 		\ do it
  cr  key? drop				\ KEY? forces o/p flush
  ExitCode @
;

assign LinMain to-do EntryPoint
save scr2fth.elf			\ save to disc
bye					\ ready to try it

[then]
