\ CompileOnly.fth - Compile only Forth

0 [IF]
***********************************************************************
The original version of this code came from Ward MacFarland
in comp.lang.forth on 24 Jan 2006.

Compiling Interpreter

Intrigued by recent discussions of "interpreter-free" Forths, I came up
with something to experiment with that should run on most ANS Forths.

To enable this mode, simply use the new word: >>
The rest of the line is compiled as a temporary definition, then
executed.
This allows most compile-time words to run just fine, including 
RECURSE LOCALS| DO..LOOP IF..THEN

This only non-ANS requirement is the ability to change the Dataspace
pointer (and the Codespace pointer, for Forths with separate code and
data areas) programatically.  I think that most Forths allow changing
those pointers. 

MacForth uses DP to contain the Dataspace address ( HERE ) and CP to
contain the Codespace pointer.  Since MacForth also checks for code and
data space overruns, it also has the variables DATAEND and CODEEND which
need to be changed if DP and CP and changed. 

How >> does its "magic":
 - set a MARKER (for later cleanup)
 - move codespace and dataspace pointers to point to scratch areas
 - read in the rest of the line
 - call :NONAME to activate the compiler
 - EVALUATE the line we just read in
 - execute ; to terminate the scratch definition
 - EXECUTE the token we got from :NONAME
 - execute the MARKER word to restore everything


So, to show the printable ASCII chars,
>> 8 1  DO  cr  16 0 DO i j 16 * + emit LOOP  LOOP

Or to calculate the 20th Fibonacci number,
20 >>  dup 1 > if dup 1- recurse swap 2 - recurse + then 
.
6765 

even locals work,
-1 999 >> locals| X Y | cr x . y .
999 -1 


>> does NOT work with creating words, like:
>> CREATE ABC
>> 10 CONSTANT XYZ


Also, words that allot space will do so from the scratch area and not at
real HERE so this does not produce the expected result:
>> 100 ALLOT
************************************************************************
* 
[THEN]


(( Ward's original code
ONLY FORTH DEFINITIONS
DECIMAL

MARKER --Inline_compiler--

13 CONSTANT EOLChar  \ change this as needed.  Macs use CR for EOL
                     \ some Forths allow 0 to make WORD terminate at EOL
1000 CONSTANT ScratchSize
CREATE ScratchDP    ScratchSize allot
CREATE ScratchCP    ScratchSize allot


: SetDestination ( -- )        \ Change Codespace & Dataspace pointers
     \ change as appropriate for your Forth
    ScratchDP dp !
    ScratchCP cp !                   \ MF has separate code & data areas
    ScratchDP ScratchSize + DataEnd !  \ MF keeps track of end data
    ScratchCP ScratchSize + CodeEnd !  \ MF keeps track of end code
;


: >> ( -- )    \ compile rest of line as temporary word, then execute it
    s" marker xxx" evaluate      \ preserve current compilation info
    s" ' xxx" evaluate >R        \ save the token for the marker
\   LastCreatedToken @ >R        \ simpler MF-ism for above line
    EOLChar word >R              \ grab rest of line
    SetDestination               \ change DP etc
    :noname                      \ force compilation mode
    R> count ['] evaluate catch  \ compile the rest of the line
    IF
        R> execute        \ execute the marker to restore DP etc.
        -2 throw          \ ANS
\       Dflt(ERROR$)      \ MF - force error to be displayed and located
    ELSE
        postpone ;        \ end the compilation,  noname token is on TOS
        execute           \ execute the noname token
        R> execute        \ execute the marker to restore DP etc.
    THEN
;
))

ONLY FORTH DEFINITIONS
DECIMAL

MARKER --Inline_compiler--

0 CONSTANT EOLChar   \ change this as needed.  Macs use CR for EOL
                     \ some Forths allow 0 to make WORD terminate at EOL
1000 CONSTANT ScratchSize
CREATE ScratchDP    ScratchSize allot
variable OldDP

: SetDestination	\ --
\ Switch Codespace & Dataspace pointers
\ Change as appropriate for your Forth
  dp @ OldDP !  ScratchDP dp !
;

: OldDestination	\ --
\ Restore Codespace & Dataspace pointers, and
\ remove words from ScratchDP
\ Change as appropriate for your Forth
  OldDP @ dp !
  ScratchDP dup ScratchSize + trim-dictionary
;

: >>		\ i*x -- j*x
\ Compile rest of line as a temporary word, then execute it
\ Change if the compilation stack is not the data stack.
  EOLChar parse 2>r		\ get rest of line
  SetDestination
  :noname			\ start compilation
  2r> ['] evaluate catch	\ compile rest of line
  dup >r IF
    2drop drop  postpone ;	\ discard caddr len xt colonsys
  ELSE
    postpone ;  execute		\ end compilation and execute
  THEN
  OldDestination
  r> throw
;

0 [IF]
\ examples follow

>>  .s cr ." hello" .s

    ( 0 ) 
hello
    ( 0 ) 


\ loops work fine
>> 4 1 DO  cr  32 0 DO i j 32 *  + emit LOOP  LOOP

 !"#$%&'()*+,-./0123456789:;<=>?
@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_
`abcdefghijklmnopqrstuvwxyz{|}~


\ so do locals
-1 999 >> locals| X Y | cr x . y .

999 -1 


\ also recursion and conditionals
20 >> dup 1 > if dup 1- recurse swap 2 - recurse + then 
.s
    ( 1 ) \ 6765 


\ this should fail:

>> CREATE xyz  ," hello" .s

[THEN]
