\ Repeated Fractions

((
Copyright (c) 1999, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From: Anil Rodrigues on comp.lang.forth
=======================================
There is a systematic way to obtain the rational fractions that
best ( I think) approximate any no. if you have the no. accurately
already.

By repeated subtraction and reciprocals,

PI = 3 + 1
       --------
         7 + 1
            --------
             15 + 1
                 ---------
                  1 +  1
                      ----------
                       292 + 1
                            ---------
                             1 + 1
                                -----
                                  1.....
Stopping at 7, you get 22/7.
Stopping at 15 gives 333/106,  but the no. is very near 16, actually,
which is equivalent to stopping at the next 1: giving the well known
355/113.

From: Joseph Power on comp.arch.embedded
========================================
The following BASIC program was posted on comp.arch.embedded, referencing
Byte magazine, Programming Insights, May 1985, pp 429-430. This program
generates a ratio of two integers, given an input value A.

10 INPUT A: B=0: C=1: D=ABS( A-INT(A) )
20 IF D=0 THEN 40
30 E=1/D: F=C: C=INT(E) * C + B: B=F: D=E-INT(E); IF A*C<>INT(A*C) THEN 30
40 PRINT A*C;"/";C: GOTO 10

From: MPE
=========
The word .RATIOS ( float -- ) displays the pairs of integers needed to
represent the floating point value. To test it, use:
  %PI .RATIOS
and you will see a succession of the well known ratios. After each step,
press <space> to continue, or any other key to stop. When the denominator
or numerator overflow the 32 bit range, the displayed error will increase.

Change Notes
============
20120501 SFP001 Changed F>S to FR>S as per SFP021/022 in NDP387.FTH.
))

fvariable floatval      \ initial value         A
fvariable diff          \ difference            D
variable X              \ 1/difference

variable denominator
variable numerator      \ numerator             C
variable B              \ previous numerator    B

: .ratio        \ --
  cr
\  ." 1/diff = " x @ u.
  denominator @ u. ." / " numerator @ u.
  #20 >pos
  ."  Error = " floatval f@  denominator @ s>f  numerator @ s>f f/ f- f.
\  ."  New Diff = " diff f@ f.
;

: start         \ float --
  fdup floatval f!                      \ initial value
  fdup fint fr>s denominator !          \ initial denominator ; SFP001
\  fdup fint f>s denominator !           \ initial denominator
  fdup fint f- diff f!                  \ initial difference
  1 numerator !                         \ initial numerator
  0 b !                                 \ initial previous numerator
  .ratio
;

: step          \ --
  %1 diff f@ f/                         \ E=1/diff
  fdup fint fr>s x !                    \ X=int(1/diff) ; SFP001
\  fdup fint f>s x !                     \ X=int(1/diff)
  fdup fint f- diff f!                  \ next difference
  numerator @ >r
  r@ x @ *  b @ +  numerator !          \ num=num*diff+oldnum
  floatval f@ numerator @ s>f f*        \ denom=int(val*numerator)
  ( fint) fr>s denominator !            \ this is an approximation ; SFP001
\  ( fint) f>s denominator !             \ this is an approximation,
                                        \ so let rounding happen
  r> b !                                \ save old numerator
  .ratio                                \ display results
;

: .ratios       \ float --
  start
  begin
    step
    key bl <>
  until
;

