\ xorshift.fth

((
From Wikipedia, the free encyclopedia

Xorshift random number generators form a class of
pseudorandom number generators that was discovered by George
Marsaglia.[1] They generate the next number in their
sequence by repeatedly taking the exclusive or of a number
with a bit shifted version of itself. This makes them
extremely fast on modern computer architectures. The xor
shift primitive is invertible.[2]They are a subclass of
linear feedback shift registers, but their simple
implementation typically makes them faster and use less
space.[3] However, the parameters have to be chosen very
carefully in order to achieve a long period.[4]

Example implementation

A C version[note 1] of one xorshift algorithm [1] is:

#include <stdint.h>
uint32_t xor128(void) {
  static uint32_t x = 123456789;
  static uint32_t y = 362436069;
  static uint32_t z = 521288629;
  static uint32_t w = 88675123;
  uint32_t t;

  t = x ^ (x << 11);
  x = y; y = z; z = w;
  return w = w ^ (w >> 19) ^ (t ^ (t >> 8) );
}

This algorithm has a period of 2^128 - 1 and it passes the diehard tests.

Notes

    ^ In C, the "^" caret represents the Bitwise XOR, and "
    << " the bit shift.

References

    1. a b Marsaglia, George (July 2003). "Xorshift RNGs".
    Journal of Statistical Software. Vol. 8 (Issue  14). ^
    2. Rivest, R. (2010). "The invertibility of the XOR of
    rotations of a binary word". International Journal of
    Computer Mathematics: 10.
    doi:10.1080/00207161003596708.
    3. Brent, Richard P. (August 2004). "Note on
    Marsaglias Xorshift Random Number Generators". Journal
    of Statistical Software. Vol. 11 (Issue 5).
    4. Panneton, Franois (October 2005). "On the xorshift
    random number generators". ACM Transactions on Modeling
    and Computer Simulation (TOMACS). Vol. 15 (Issue 4).

External links

Press, WH; Teukolsky, SA; Vetterling, WT; Flannery, BP
(2007). "Section 7.1.2.A. 64-bit Xorshift Method". Numerical
Recipes: The Art of Scientific Computing (3rd ed.). New
York: Cambridge University Press. ISBN 978-0-521-88068-8.
))

decimal

123456789 value vx
362436069 value vy
521288629 value vz
88675123  value vw

: xor128	\ -- rand
  vx dup 11 lshift xor 			\ -- t
  vy to vx  vz to vy  vw to vz
  vw dup 19 rshift xor			\ -- t w'
  swap dup 8 rshift xor  xor
  dup to vw
;


\ Xorshift Random Number Generator, 32 Bits
\ http://de.wikipedia.org/wiki/Xorshift
\ from Matthias Koch

314159265 value seed32

: xor32		\ -- x
  seed32
  dup 13 lshift xor
  dup 17 rshift xor
  dup  5 lshift xor
  dup to seed32
;
