\ DASM386.FTH - VFX Forth high level kernel

((
Copyright (c) 2001, 2005, 2010, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

Derived with permission from code copyright (c) 1993
Construction Computer Software
Cape Town
South Africa

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


DocGen Tags
===========
T             title
S             Section sub-title
H             a general heading
N             Forced newline
P             begin paragraph (ends after any non * line)
D             Definition, next word is title, all following text
              and * lines are descriptions
C             Line in fixed font
E             Example

To do
=====
Fix 16/32 selections, especially n [EBP+EDI]

Change history
==============
20120214 SFP005 Added SHLD SHRD
20120203 SFP004 Added some 486 instructions.
20101202 SFP003 Fixed E8 CALL when forward.
20050905 SFP002 Added CMOVx
20010331 SFP001 Fixed IMUL disassembly for immediate operands
))

( *! dasm386                                                            )
( *T Intel 386+ Disassembler                                            )
( **                                                                    )
( *P The VFX Forth system includes a disassembler for debugging purposes.)
( ** Native code built by the system can be viewed at a machine code )
( ** level.                                                             )
( **                                                                    )

only forth definitions
vocabulary disassembler
  also disassembler definitions

\ words needed by PFW

: flip       \ n -- n' ; swap low word of n
  dup $FF and #8 lshift
  over $FF00 and #8 rshift      \ n hi lo --
  or swap $FFFF0000 and or
;

: split      \ n -- lo-word high-word
  dup $FFFF and swap #16 rshift $FFFF and
;

: unsplit    \ lo-word hi-word -- n
  MAKELONG
;

: scale         \ n cnt -- n' ; -CNT shift aright, +CNT shift aleft
  dup 0<
  if  abs rshift  else  lshift  then
;

: odd@          \ addr -- n
  @  ;

: boolean       \ n -- t/f
  0<>  ;

: <w            \ w -- n ; sign extend word
  dup $07FFF >
  if  $FFFF0000 or  endif
;

: either=       \ n a b -- t/f ; true if n=a or n=b
  2 pick = -rot                 \ n, b
  =  or                         \ n, a
;

create NextVal
  $C3 c,
here NextVal - constant NextWidth

\ --- WORDS NOT IN CCS KERNEL

: digits        \ d n -- d'
  0 ?do  #  loop
;

: DSTR0         ( d width -- adr cnt )
   >R <# R> digits #> ;

: STR0          ( n width -- adr cnt )
  >r  s>d
  <#  r> digits  #>
;

: <C            ( char --- n )   \ sign extend byte to 32bits
  DUP $7F >
  IF  $FFFFFF00 OR  THEN
;

: CS0
  1
;

: =cells        \ addr -- addr' ; adjust to boundary - PFW does not need this
  aligned
;

: BYTE-JUMPER  ( addr n -- addr' )
  0 2DUP +
  IF DO COUNT + ( =CELLS )  LOOP
  ELSE 2DROP
  THEN
;

: u<=           \ u1 u2 -- flag
  u> not
;

: u>=           \ u1 u2 -- flag
  u< not
;


\ --- DISASSEMBLER main portion

\ --- VARS

VARIABLE (DisAddr)              \ current disassembly address
VARIABLE ToDisAddr              \ to address in from to words
VARIABLE (DisSeg)               \ current disassembly segment
variable branch-target          \ highest branch target address
variable next-found             \ true when NEXT found

also forth definitions
: DisAddr  ( -- addr )
   \ current disassembly address
   (DisAddr) @ ;

: SET-DisAddr ( addr -- )
   \ set current disassembly addr to addr
   (DisAddr) ! ;
previous definitions

: ADD-DisAddr ( n -- )
   \ increment current disassembly address with n
   (DisAddr) +! ;

: DEC-DisAddr ( -- )
   \ increment current disassembly address with n
   DisAddr 1- SET-DisAddr ;

: SET-DisSeg ( n -- )
   (DisSeg) ! ;

: DisSeg ( -- n )
   (DisSeg) @ ;

: >branch-target        \ addr -- ; insert if higher than current value
  dup branch-target @ u>
  if  branch-target !  else  drop  endif
;


\ --- FLAGS FOR INSTRUCTION CONDITIONS
   VARIABLE (32bit?)    \ flag for current disassmbly width
   VARIABLE ((w)?)      \ operand length field query ))
   VARIABLE ($66?)      \ overide
   VARIABLE (16/32?)    \ only 16bit & 32bit registers
   VARIABLE (8/16?)     \ only  8bit & 16bit registers allowed
   VARIABLE (16bit?)    \ only 16bit regs and mem allowed
   VARIABLE (8bit?)     \ only  8bit registers and memory allowe

   \ lower bits of opcode
   VARIABLE Opcode1  \ first byte of instruction
   VARIABLE Opcode2  \ second byte of instruction
   \ lower bits of opcode
   VARIABLE (bit0)
   VARIABLE (bit1)
   VARIABLE (bit2)
   VARIABLE (bit3)


\ --- SET VARIABLES AND FLAGS

: SET-32bit?  ( n -- )
   \ set disassembly width to 32bit wide data
   (32bit?) ! ;

: ZERO-32bit? ( -- )
   \ reset 32bit data flag to zero
   (32bit?) OFF ;

: 32bit? ( -- flag )
   \ status of 32bit data flag
   (32bit?) @ ;

: SET-(w)? ( -- )
   \ set (w) field length operand in this instruction
   ((w)?) ON ;				\ ))

: ZERO-(w)? ( -- )
   \ zero (w) field length operand in this instruction
   ((w)?) OFF ;				\ ))

: (w)?  ( -- flag )
   \ status of (w) field flag
   ((w)?) @ ;				\ ))

: SET-$66?  ( -- )
   \ set 32bit data overide flag
  ($66?) ON ;

: ZERO-$66? ( -- )
   \ zero 32bit overide flag
   ($66?) OFF ;

: $66? ( -- flag )
   \ status of overide flag
   ($66?) @ ;

: SET-16/32? ( -- )
   \ set 16/32 bit regs only flag
   (16/32?) ON ;

: ZERO-16/32? ( -- )
   \ zero 16/32 bit regs only flag
   (16/32?) OFF ;

: 16/32? ( -- flag )
   (16/32?) @ ;

\ More flag handling

: SET-8/16? ( -- )
   \ set 8/16 bit regs only flag
   (8/16?) ON ;

: ZERO-8/16? ( -- )
   \ zero 8/16 bit regs only flag
   (8/16?) OFF ;

: 8/16? ( -- flag )
   (8/16?) @ ;

: SET-16bit? ( -- )
   \ set 16 bit regs only flag
   (16bit?) ON ;

: ZERO-16bit? ( -- )
   \ zero 16 bit regs only flag
   (16bit?) OFF ;

: 16bit? ( -- flag )
   (16bit?) @ ;

: SET-8bit? ( -- )
   \ set 8 bit regs only flag
   (8bit?) ON ;

: ZERO-8bit? ( -- )
   \ zero 16 bit regs only flag
   (8bit?) OFF ;

: 8bit? ( -- flag )
   (8bit?) @ ;

: SET-bit0 ( n -- )
   \ store n in bit0 flag
   (bit0) ! ;

: ZERO-bit0 ( -- )
   (bit0) OFF ;

: bit0 ( -- flag )
   (bit0) @ ;

: SET-bit1 ( n -- )
   \ store n in bit1 flag
   (bit1) ! ;

: ZERO-bit1 ( -- )
   (bit1) OFF ;

: bit1 ( -- flag )
   (bit1) @ ;

: SET-bit2 ( n -- )
   \ store n in bit2 flag
   (bit2) ! ;

: ZERO-bit2 ( -- )
   (bit2) OFF ;

: bit2 ( -- flag )
   (bit2) @ ;

: SET-bit3 ( n -- )
   \ store n in bit3 flag
   (bit3) ! ;

: ZERO-bit3 ( -- )
   (bit3) OFF ;

: bit3 ( -- flag )
   (bit3) @ ;

: INIT-VARS  ( -- )
   Opcode1 OFF
   Opcode2 OFF
   SET-(w)?           \ always use (w) operand length field
   ZERO-8bit?          \ untrue 8bit regs and mem
   ZERO-8/16?          \ untrue 8bit and 16bit flag only
   ZERO-16bit?         \ untrue 16bit regs and mem only
   ZERO-16/32? ;     \ untrue 16bit and 32bit flag only

\ mod/rm byte handling

   \ values of modrm byte
   VARIABLE (r/m)        \ register/memory
   VARIABLE (reg#)  \ register
   VARIABLE (modifier)  \ modifier

: r/m  ( -- n )
  (r/m) @ ;
: reg#  ( -- n )
  (reg#) @ ;
: modifier ( -- n )
  (modifier) @ ;
: SET-r/m  ( n -- )
   (r/m) ! ;
: SET-reg#  ( n -- )
  (reg#) ! ;
: SET-modifier ( n -- )
  (modifier) ! ;

   \ values of sib byte
   VARIABLE (regbase)  \ reg used as base register
   VARIABLE (indexreg)  \ reg used as index register
   VARIABLE (ss)         \ scale factor

: regbase ( -- n )
  (regbase)  @ ;
: indexreg ( -- n )
  (indexreg) @ ;
: ss    ( -- n )
   (ss)    @ ;
: SET-regbase ( n -- )
  (regbase)  ! ;
: SET-indexreg ( n -- )
  (indexreg) ! ;
: SET-ss ( n -- )
   (ss)    ! ;

\ --- DISASSEMBLY OUTPUT

14 CONSTANT DisDumpCol  \ starting column for dump
38 CONSTANT DisOpCol    \ starting column for opcode
48 CONSTANT DisParmCol  \ starting column for parameters

#128 constant LineSize	\ size of output line buffer
LineSize buffer: (DisLine$)	\ line buffer for disassembly
VARIABLE DisPos		\ output position for opcode info
VARIABLE DisDumpPos	\ output position for dump info

: DisLine$ ( -- addr cnt )
   (DisLine$) LineSize ;

: >DISLINE  ( adr cnt -- )   \ Add opcode information to disassembly line buffer.
   DisLine$ DisPos @ /STRING
   ROT MIN DUP
   DisPos  +!			\ increment print pos in dis buffer
   CMOVE ;

: >DISDUMP  ( adr cnt -- )   \ Add dump information to disassembly line buffer.
   DisLine$ DROP DisOpCol 1-    \ dump must not overwrite op
   DisDumpPos @ /STRING
   ROT MIN DUP
   DisDumpPos +!
   CMOVE ;

: BYTE>DUMP  ( byte -- )
   2 STR0 >DISDUMP ;

: WORD>DUMP  ( word -- )
   FLIP 4 STR0 >DISDUMP ;

: LONG>DUMP  ( long -- )
   SPLIT FLIP SWAP FLIP UNSPLIT
   8 STR0 >DISDUMP ;

: BYTE>DIS  ( byte -- )
   2 STR0 >DISLINE ;

: WORD>DIS  ( word -- )
   4 STR0 >DISLINE ;

: LONG>DIS  ( long -- )
   8 STR0 >DISLINE ;

variable char-holder

: CHAR>DIS  ( char -- )
   char-holder c!
   char-holder 1 >DISLINE  ;

: bl>dis	\ --
  bl char>dis  ;

: ">dis		\ --
  [char] " char>dis  ;

: [>DIS         ( --- )
   [char] [ CHAR>DIS ;

: ]>DIS         ( --- )
   [char] ] CHAR>DIS ;

: :>DIS         ( --- )
   [char] : CHAR>DIS ;

: +>DIS         ( --- )
   [char] + CHAR>DIS ;

: ->DIS         ( --- )
   [char] - CHAR>DIS ;

: $>dis		\ addr --
  count >disline  ;

: ,>DIS         ( --- )
   c" , " $>dis  ;

: #>DIS         ( --- )
   c" # " $>dis  ;

: DISADDR>DIS  ( -- )
   DisSeg CS0 <>
   IF DisSeg 4 STR0 >DISLINE :>DIS
   THEN
   DisAddr 8 STR0 >DISLINE ;

: DISINIT  ( -- )
\ Initialize disassembly output buffer to current
\ seg:address and output positions for dump and opcode.
  CS0 SET-DisSeg                        \ initialise segment
  DisLine$ BLANK                        \ initialise o/p buffer
  DisPos OFF
  DisDumpPos off

  c" ( " count >disline                 \ leading comment )
  DISADDR>DIS                           \ address of instruction

  DisDumpCol DisDumpPos !
  DisOpCol DisPos !
;

: disclose      \ -- ; add finishing text
  c"  )" count >disdump
;

\ --- BIT-TESTS

: @BYTE  ( -- byte )
   DisAddr C@
   DUP BYTE>DUMP
   1 ADD-DisAddr ;

: @WORD  ( -- word )
\    DisSeg DisAddr W@L
   DisAddr W@
   DUP WORD>DUMP
   2 ADD-DisAddr ;

: @LONG  ( -- long )
\    DisSeg DisAddr @L
   DisAddr ODD@
   DUP LONG>DUMP
   4 ADD-DisAddr ;

: @OPCODE  ( -- byte )
   @BYTE
   DUP 1 AND BOOLEAN SET-bit0
   DUP 2 AND BOOLEAN SET-bit1
   DUP 4 AND BOOLEAN SET-bit2
   DUP 8 AND BOOLEAN SET-bit3  ;

\  | bit7 | bit6 | bit5 | bit4 | bit3 | bit 2 | bit 1 | bit 0 |
\  |------|------|------|------|------|-------|-------|-------|
\  | mod  | mod  |opcode|opcode|opcode| r/m   | r/m   | r/m   |
\  |------|------|------|------|------|-------|-------|-------|

: SPLIT-BITS  ( byte -- bits76 bits543 bits012 )
   \ $38 = 00111000
   DUP -6 SCALE
   OVER $38 AND -3 SCALE
   ROT 7 AND ;

: @MODRM  ( -- )
   @BYTE SPLIT-BITS
   SET-r/m SET-reg# SET-modifier ;

: @SIB  ( -- )
   @BYTE SPLIT-BITS
   SET-regbase  SET-indexreg SET-ss ;

\ --- OPCODE LABELS

CREATE OpCodes
         \ Opcodes in order of apearance.
         ( 00) ", ADD"    ", OR"     ", ADC"    ", SBB"
         ( 04) ", AND"    ", SUB"    ", XOR"    ", CMP"
         ( 08) ", INC"    ", DEC"    ", PUSH"   ", POP"
         ( 0C) ", DAA"    ", DAS"    ", AAA"    ", AAS"
         ( 10) ", PUSHA"  ", POPA"   ", BOUND"  ", ARPL"
         ( 14) ", IMUL"   ", INS"    ", OUTS"   ", J"
         ( 18) ", TEST"   ", XCHG"   ", MOV"    ", LEA"
         ( 1C) ", NOP"    ", CBW"    ", CWD"    ", CALL"
         ( 20) ", FWAIT"  ", PUSHF"  ", POPF"   ", SAHF"
         ( 24) ", LAHF"   ", MOVS"   ", CMPS"   ", STOS"
         ( 28) ", LODS"   ", SCAS"   ", RET"    ", LES"
         ( 2C) ", LDS"    ", ENTER"  ", LEAVE"  ", INT"
         ( 30) ", INTO"   ", IRET"   ", AAM"    ", AAD"
         ( 34) ", XLAT"   ", LOOPNE" ", LOOPE"  ", LOOP"
         ( 38) ", JCXZ"   ", IN"     ", OUT"    ", JMP"
         ( 3C) ", LOCK"   ", REPNE"  ", REP"    ", ESC"
         ( 40) ", HLT"    ", CMC"    ", CLC"    ", STC"
         ( 44) ", CLI"    ", STI"    ", CLD"    ", STD"
         ( 48) ", LAR"    ", LSL"    ", CLTS"   ", SET"
         ( 4C) ", BT"     ", SHLD"   ", BTS"    ", SHRD"
         ( 50) ", LSS"    ", BTR"    ", LFS"    ", LGS"
         ( 54) ", MOVZX"  ", BTC"    ", BSF"    ", BSR"
         ( 58) ", MOVSX"  ", ROL"    ", ROR"    ", RCL"
         ( 5C) ", RCR"    ", SHL"    ", SHR"    ", SAR"
         ( 60) ", NOT"    ", NEG"    ", MUL"    ", DIV"
         ( 64) ", IDIV"   ", SLDT"   ", STR"    ", LLDT"
         ( 68) ", LTR"    ", VERR"   ", VERW"   ", SGDT"
         ( 6C) ", SIDT"   ", LGDT"   ", LIDT"   ", SMSW"
         ( 70) ", LMSW"   ", CDQ"    ", ???"    ", CWDE"
         ( 74) ", POPAD"

: OPCODE  ( n -- adr cnt )
  OpCodes SWAP BYTE-JUMPER COUNT ;

CREATE CondCodes
      \ Condition codes for J and SET.
    ( 00) ", O"   ", NO"  ", B/NAE "   ", NB/AE "
    ( 04) ", Z/E "   ", NZ/NE "  ", BE/NA "  ", NBE/A "
    ( 08) ", S"   ", NS"  ", P/PE "   ", NP/PO "
    ( 0C) ", L/NGE "   ", NL/GE "  ", LE/NG "  ", NLE/G "

: CONDCODE  ( n -- adr cnt )
   CondCodes SWAP BYTE-JUMPER COUNT ;

CREATE SegRegs	 \ -- addr
  ", ES"  ", CS"  ", SS"  ", DS"  ", FS"  ", GS"

: SEGREG  ( n -- adr cnt )
   SegRegs SWAP BYTE-JUMPER COUNT ;

CREATE Regs
    \   00       01       02       03
    ", AL"  ", CL"  ", DL"  ", BL"
    \   04       05       06       07
    ", AH"  ", CH"  ", DH"  ", BH"
    \   08       09       10       11
    ", AX"  ", CX"  ", DX"  ", BX"
    \   12       13       14       15
    ", SP"  ", BP"  ", SI"  ", DI"

: REG  ( n -- adr cnt )
   Regs SWAP BYTE-JUMPER COUNT ;

: r32>dis	\ n --
  [char] E CHAR>DIS
  $07 and 8 or reg >disline
;

CREATE R/M16-Table
         ", BX+SI"  ", BX+DI"  ", BP+SI"  ", BP+DI"
         ", SI"     ", DI"     ", BP"     ", BX"

: R/M16  ( n -- adr cnt )
   R/M16-Table SWAP BYTE-JUMPER COUNT ;

: PTR?  ( --- flag )
   modifier 0= r/m 4 = NOT AND ;


: PTR>DIS  ( -- )
   bit0  (w)?   AND
   \ if (w) field is in this instruction
   \ true if bit 0 is set or flase if not set
   \ 16bit or 32bit data
   16/32? OR
   IF  32bit?   8/16? NOT AND
      \ 32bit data and not 8/16bit data only
      IF [char] D CHAR>DIS THEN
      c" Word Ptr "
   THEN
   bit0 NOT  (w)?   AND
   \ if (w) field is in this instruction
   \ true if bit 0 is not set or flase if set 8bit data
   8bit? OR
   \ if only 8bit memory allowed in this instruction
   IF  c" Byte Ptr "  THEN
   count >DISLINE ;

\ --- PARAMETER DECODING
\
\     76543210        76543210
\          --- r/m         --- base
\       ---    ttt      ---    index
\     --       mod    --       ss
\
\  size = 16                                          mod = 3
\ r/m   mod = 0     mod = 1     mod = 2        w = 0    w = 1
\ ---   ----------  ----------  -------------  --------------
\ 0     [BX+SI]     [BX+SI+d8]  [BX+SI+d16]    AL       AX
\ 1     [BX+DI]     [BX+DI+d8]  [BX+DI+d16]    CL       CX
\ 2     [BP+SI]     [BP+SI+d8]  [BP+SI+d16]    DL       DX
\ 3     [BP+DI]     [BP+DI+d8]  [BP+DI+d16]    BL       BX
\ 4     [SI]        [SI+d8]     [SI+d16]       AH       SP
\ 5     [DI]        [DI+d8]     [DI+d16]       CH       BP
\ 6     d16         [BP+d8]     [BP+d16]       DH       SI
\ 7     [BX]        [BX+d8]     [BX+d16]       BH       DI
\
\ size = 32                                          mod = 3
\    r/m   mod = 0     mod = 1     mod = 2       w = 0    w = 1
\    ---   ----------  ----------  ------------- --------------
\    0     [EAX]       [EAX+d8]    [EAX+d32]     AL       EAX
\    1     [ECX]       [ECX+d8]    [ECX+d32]     CL       ECX
\    2     [EDX]       [EDX+d8]    [EDX+d32]     DL       EDX
\    3     [EBX]       [EBX+d8]    [EBX+d32]     BL       EBX
\    4     two-byte    two-byte    two-byte      AH       ESP
\    5     d32         [EBP+d8]    [EBP+d32]     CH       EBP
\    6     [ESI]       [ESI+d8]    [ESI+d32]     DH       ESI
\    7     [EDI]       [EDI+d8]    [EDI+d32]     BH       EDI
\
\  two-byte mode
\     base  mod = 0     mod = 1        mod = 2
\     ----  ----------- -------------- ---------------
\     0     [EAX+sidx]  [EAX+sidx+d8]  [EAX+sidx+d32]
\     1     [ECX+sidx]  [ECX+sidx+d8]  [ECX+sidx+d32]
\     2     [EDX+sidx]  [EDX+sidx+d8]  [EDX+sidx+d32]
\     3     [EBX+sidx]  [EBX+sidx+d8]  [EBX+sidx+d32]
\     4     [ESP+sidx]  [ESP+sidx+d8]  [ESP+sidx+d32]
\     5     [d32+sidx]  [d32+sidx+d8]  [d32+sidx+d32]
\     6     [ESI+sidx]  [ESI+sidx+d8]  [ESI+sidx+d32]
\     7     [EDI+sidx]  [EDI+sidx+d8]  [EDI+sidx+d32]
\
\
\  sidx
\     ss     scale factor      index   index register
\     0      *1                0       EAX
\     1      *2                1       ECX
\     2      *4                2       EDX
\     3      *8                3       EBX
\                              4       none
\                              5       EBP
\                              6       ESI
\                              7       EDI
\

: (OPCODE)>DIS  ( op# -- )
   OPCODE >DISLINE ;

: PARM-TAB  ( -- )
   DisParmCol DisPos ! ;

: OPCODE>DIS  ( op# -- )
   (OPCODE)>DIS PARM-TAB ;

: op$>dis	\ caddr len -- ; SFP004
  >disline  parm-tab
;

: @BYTE>DIS  ( -- )
   @BYTE BYTE>DIS ;

: @WORD>DIS  ( -- )
   @WORD WORD>DIS ;

: @LONG>DIS  ( -- )
   @LONG LONG>DIS ;

: SEXT>DIS  ( byte -- )
   <C DUP 0<
   IF [char] - CHAR>DIS THEN
   ( ELSE [char] + )
   ABS BYTE>DIS ;

: @DWORD  ( -- n )
   32bit?
   IF @LONG
   ELSE
      @WORD
   THEN ;

: DWORD>DIS  ( n -- )
   32bit?
   IF LONG>DIS
   ELSE
      WORD>DIS
   THEN ;

#256 buffer: StepBuff

: $stepOver	\ -- ; display and step over string at DISADR
  DisAddr StepBuff  over c@ 1+ 0 do	\ collect string
    over i + c@  over i + c!
  loop
  2drop
  bl>dis ">dis  StepBuff $>dis  ">dis	\ display string
  Disaddr count + aligned		\ step over string (aligned)
  (DisAddr) !
;

: ZStepOver	\ -- ; display and step over string at DISADDR
  bl>dis ">dis
  DisAddr zcount  2dup >disline  + 1+ (DisAddr) !
  ">dis
;

: w$StepOver	\ -- ; display and step over string at DISADDR
  bl>dis ">dis
  DisAddr wcount  2dup >disline  + 2+ aligned (DisAddr) !
  ">dis
;

: .dest		\ -- ; display inline data as branch target
  bl>dis  @long long>dis
;

also forth definitions
[switch dasm-switch drop
  ' (S")     runs $stepOver
  ' (C")     runs $stepOver
  ' (.")     runs $stepOver
  ' (abort") runs $stepOver
  ' (Z")     runs ZStepOver
  ' (W")     runs W$StepOver
  ' (do)     runs .dest
  ' (?do)    runs .dest
  ' (loop)   runs .dest
  ' (+loop)  runs .dest
  ' branch   runs .dest
  ' ?branch  runs .dest
  ' (of)     runs .dest
  ' lit      runs .dest
  also h-c  ' does_sin  previous
             runs .dest
  ' lv       run:  [>dis .dest ]>dis  ;
  ' lv@      run:  [>dis .dest ]>dis  ;
  ' lv!      run:  [>dis .dest ]>dis  ;
\  ' val@     run:  [>dis .dest ]>dis  ;
\  ' val!     run:  [>dis .dest ]>dis  ;
  ' makelvs  run:  c"   #args = " $>dis @word>dis
                   c" , #locals = " $>dis @word>dis  c"  cells" $>dis
             ;
switch]
previous definitions

: ShowBranchTarget      \ abs24 --
  dup inForth? 0=			\ get out if not in Forth
  if  drop  exit  endif

  dup >name name> over = if
    #8 Dispos @ #7 and - 0              \ to next tab stop
    ?do  bl char>dis  loop
    dup >name count >DISLINE
  then

  dasm-switch				\ process special cases
;

: abs>dis       \ addr -- ; deal with absolute branch target address
  dup >branch-target                    \ make branch target address
  dup dword>dis                         \ generate text
  ShowBranchTarget
;

: rel>dis       \ offset -- ; generate relative branch target
  DisAddr +                             \ form target address
  abs>dis                               \ treat as absolute target
;

: absCall>dis	\ addr -- ; SFP003
\ deal with absolute branch target address
  dup dword>dis ShowBranchTarget  ;

: relCall>dis	\ offset -- ; SFP003
\ generate relative branch target
  DisAddr + absCall>dis  ;

: @DWORD>DIS  ( -- )
   @DWORD DWORD>DIS ;

: @BYTE-OFFSET  ( -- )
   @BYTE <C rel>DIS ;

: @LONG-OFFSET  ( -- )
   @DWORD
   32bit?  NOT IF <W THEN
   rel>DIS ;

: (REG)>DIS  ( n -- )
   DUP 8 >=
   IF  32bit?   8/16? NOT AND    \ 32bit registers only
      16bit? NOT AND          \ 16bit regs only set ?
      IF [char] E CHAR>DIS THEN
   THEN
   REG >DISLINE ;

: REG>DIS  ( n -- )
   bit0  (w)?   AND
   \ if (w) field is in this instruction
   \ true if bit 0 is set or flase if not set
   \ 16bit or 32bit data
   16/32? OR
   16bit? OR
   IF \ 16bit and 32bit registers only
      \ instruction uses only modr/m byte
      \ for selection of registers
      8 +
   THEN
   (REG)>DIS ;

: REG32>DIS  ( n -- )
   $66? IF [char] E CHAR>DIS THEN
   8 + (REG)>DIS ;

: R/M16>DIS  ( n -- )
   R/M16 >DISLINE ;

: MEM(32bits)  ( -- )
  \ 32bit registers only for memory [reg]
  \ instruction uses only modr/m byte
  \ for selection of registers
  \ 32bit?= -1  mod= 0
  c" 0 " count >disline
  [>DIS
  [char] E CHAR>DIS
  r/m 8 + REG >DISLINE
  ]>DIS
;

: SIB>DIS
   indexreg 4 <>
   IF +>DIS
      indexreg REG32>DIS
      ss 0>
      IF [char] * CHAR>DIS
         1 ss SCALE [char] 0 + CHAR>DIS
      THEN
   THEN ;

: MODRM32-2BYTE  ( -- )
  @SIB
  [>DIS
  regbase 5 =  modifier 0=  AND
  IF  @LONG>DIS
  ELSE
      regbase REG32>DIS
  THEN
  SIB>DIS
  modifier 0 >
  IF  modifier 1 =
      IF  @BYTE ?DUP
          IF  +>DIS  SEXT>DIS THEN
      ELSE
          @LONG ?DUP
          IF  +>DIS  LONG>DIS  THEN
      THEN
  THEN
  ]>DIS
;

: R/M>DIS  ( -- )
   \ select r/m bits to registers including overide switch
   r/m $66? IF [char] E CHAR>DIS REG>DIS ELSE R/M16>DIS THEN ;

: MODRM16  ( -- )
   \ mod reg r/m
   \ 00  000 000 modr/m byte        mod r/m
   r/m 4 =  modifier 3 < AND  $66? AND
   \ 00 100  and overide then sib byte
   IF MODRM32-2BYTE EXIT THEN      \ process as 32bit data
   \ mod=0 and r/m not 6 | r/m=6 16bit displacement
   \ 32bit data overide then memory
   modifier 0= r/m 6 = NOT AND
   $66? AND
   IF MEM(32bits) EXIT THEN
   modifier 3 < IF [>DIS THEN
   \ mod = 00 or 10  then memory [  ]
   modifier
   CASE
      0 OF r/m 6 =         \ r/m = 110 displacement of 16bits
            IF @WORD>DIS
            ELSE R/M>DIS
            THEN
      ENDOF
      1 OF  R/M>DIS +>DIS @BYTE SEXT>DIS  ENDOF
      2 OF  R/M>DIS +>DIS @WORD>DIS  ENDOF
      3 OF  r/m REG>DIS                    ENDOF
   ENDCASE
   modifier 3 < IF ]>DIS THEN ;

: MODRM32  ( -- )
  \ 32bit?= -1  32bit data
  \ mod= 1 or 2 and r/m= 4 then sib byte present
  \ modr/m = 00 reg# 100
  r/m 4 =  modifier 3 < AND
  IF  MODRM32-2BYTE  EXIT  THEN
  \ 00 101 mod= 0 & r/m= 5 data 32
  modifier 0= r/m 5 = NOT AND
  IF  MEM(32bits)  EXIT  THEN
  modifier 3 <
  IF  [>DIS  THEN
  modifier 0= r/m 5 = AND
  IF  @LONG>DIS
  ELSE
      modifier 3 <
      IF    r/m REG32>DIS
      ELSE
            r/m REG>DIS
      THEN
  THEN
  modifier 1 =  modifier 2 =  OR
  IF  modifier 1 =
      IF  @BYTE ?DUP
          IF  +>DIS SEXT>DIS  THEN
      ELSE
          @LONG ?DUP
          IF  +>DIS LONG>DIS  THEN
      THEN
  THEN
  modifier 3 <
  IF  ]>DIS  THEN
;

: R/M+DISP>DIS  ( -- )
   32bit? IF
     MODRM32
   ELSE
     $66?
     IF  MODRM32  ELSE  MODRM16  THEN
   THEN ;

: OPS>DIS  ( -- )
   bit1
   IF reg# REG>DIS ,>DIS R/M+DISP>DIS
   ELSE
      R/M+DISP>DIS ,>DIS reg# REG>DIS
   THEN ;

: AC>DIS  ( --- )
   bit0
   IF  32bit?
      IF [char] E CHAR>DIS THEN
      c" AX"
   ELSE
      c" AL"
   THEN
  count >DISLINE ;

: IMM>DIS  ( -- )
   ,>DIS bit0
   IF @DWORD>DIS
   ELSE
      @BYTE>DIS
   THEN ;

: 2-OP-INST  ( opcode# -- )
   OPCODE>DIS
   bit2 NOT
   IF @MODRM OPS>DIS
   ELSE
      AC>DIS IMM>DIS
   THEN ;

: SEGREG>DIS  ( n -- )
   SEGREG >DISLINE ;

: SEGREG-OP  ( byte1 -- )
   DUP 1 AND $0A + OPCODE>DIS   \ PUSH/POP
   $18 AND -3 SCALE SEGREG>DIS ;

: M/R,SEG>DIS  ( -- )
   SET-bit0    \ there is no w field in this instruction
   bit1
   IF reg# SEGREG>DIS ,>DIS R/M+DISP>DIS
   ELSE
      R/M+DISP>DIS ,>DIS reg# SEGREG>DIS
   THEN ;

: REG,R/M*IMM  ( -- )
   @MODRM
   reg# REG>DIS  ,>DIS  R/M+DISP>DIS
   ,>dis #>dis @DWORD>DIS ;

: B/W/D>DIS ( -- )
   bit0
   IF  32bit?
      IF [char] D ELSE [char] W THEN
   ELSE [char] B THEN
   CHAR>DIS ;

: DADJ-INST         ( byte#1 --- )
   $0C bit3 IF 1+ THEN OPCODE>DIS ; \ DAS/DAA

: ADJ-INST         ( byte#1 --- )
   $0E bit3 IF 1+ THEN OPCODE>DIS ; \ AAS/AAA

: SEG:         ( byte#1 --- )
   \ 3F MNEM>BUF ( SEG:)
   $18 AND -3 SCALE SEGREG>DIS :>DIS ;

: ILLEGAL  ( -- )
   $72 OPCODE>DIS ;

VARIABLE (FORWARD)

: (DISASM-NEXT) ( -- )
    \ Continue disassembling next instruction on current line
    \ starting at DisAddr in current mode.
    \ Forward reference used by prefices, e.g. $66 (operand size)
    (FORWARD) @ EXECUTE ;

\ --- OPCODES GROUPS DETERMINED BY MODRM

: GROUP01  ( -- )
   \ reg    opcode   op#
   \                               mod reg r/m
   \  00    ADD      00   000000dw 00  000 000  regs to regs
   \                      0000000w 00  000 000  regs to mem
   \                      0000001w 00  000 000  mem to regs
   \                      100000sw 00      000  imm to mem/regs
   \                      0000010w              imm to ac
   \  01    OR       01
   \  02    ADC      02
   \  03    SBB      03
   \  04    AND      04
   \  05    SUB      05
   \  06    XOR      06
   \  07    CMP      07
   \ s - immediate data field must be extended
   \ w - byte or full size
   \ ws = 00 byte - 01 full - 10 illegal - 11 byte xtend full
   \ all                   imm to reg/mem
   @MODRM
   \ fetch opcode $ instruction and store in DisLine$
   reg# OPCODE>DIS
   PTR?  IF PTR>DIS THEN
   R/M+DISP>DIS ,>DIS
   bit0 bit1 NOT AND
   IF @DWORD>DIS
   ELSE
      @BYTE SEXT>DIS
   THEN ;

: GROUP02  ( byte -- )
   \ reg    opcode   op#
   \  00    ROL      59
   \  01    ROR      5A
   \  02    RCL      5B
   \  03    RCR      5C
   \  04    SHL      5D
   \  05    SHR      5E
   \  06    ???
   \  07    SAR      5F
   \ w - byte or full size
   \ byte=Cx               reg/mem by imm count
   \ byte=Dx bit1=0        reg/mem by 1
   \ byte=Dx bit1=1        reg/mem by cl
   @MODRM
   reg# 6 =
   IF DEC-DisAddr    \ @MODRM inc DisAddr
      ILLEGAL DROP
      EXIT
   THEN
   reg# DUP 7 = IF 1- THEN
   $59 + OPCODE>DIS
   PTR?  IF PTR>DIS THEN
   R/M+DISP>DIS ,>DIS
   $F0 AND $C0 =
   IF @BYTE>DIS
   ELSE
      bit1
      IF 1 (REG)>DIS
      ELSE
         [char] 1 CHAR>DIS
      THEN
   THEN ;

: GROUP06  ( -- )
   \ reg    opcode   op#
   \  00    SLDT     65 00001111 00000000 mod 000 r/m
   \        SLDT     store local descriptor table
   \        SLDT     reg16 | mem16
   \  01    STR      66 00001111 00000000 mod 001 r/m
   \        STR      store task register
   \        STR      reg16  | mem16
   \  01    LTR      68 00001111 00000000 mod 001 r/m
   \        LTR      load task register
   \        LTR      reg16 | mem16
   \  02    LLDT     67 00001111 00000000 mod 010 r/m
   \        LLDT     reg16 | mem16
   \  04    VERR     69 00001111 00000000 mod 100 r/m
   \        VERR     reg16 | mem16
   \  05    VERW     6A 00001111 00000000 mod 101 r/m
   \        VERW     reg16 | mem16
   \  06-07 ???
   @MODRM
   ZERO-(w)?        \ no (w) operand length field in instruction
   SET-16bit?       \ 16bit regs & mem allowed
   reg#
   CASE
    00 OF $65 >R  ENDOF
    01 OF $66 >R  ENDOF
    02 OF $4C >R  ENDOF
    03 OF $4E >R  ENDOF
    04 OF $51 >R  ENDOF
    05 OF $55 >R  ENDOF
    06 OF $72 >R  ENDOF
    07 OF $72 >R  ENDOF
   ENDCASE
   R> DUP OPCODE>DIS
   $72 =
   IF   DEC-DisAddr
   ELSE R/M+DISP>DIS
   THEN ;

: GROUP07  ( -- )
   \ reg    opcode   op#
   \  00    SGDT     6B  00001111 00000001 mod 000 r/m
   \        SGDT     mem
   \  01    SIDT     6C  00001111 00000001 mod 001 r/m
   \        SIDT     mem
   \  02    LGDT     6D  00001111 00000001 mod 010 r/m
   \  03    LIDT     6E  00001111 00000001 mod 011 r/m
   \  04    SMSW     6F  00001111 00000001 mod 100 r/m
   \  05    ???
   \  06    LMSW     70  00001111 00000001 mod 110 r/m
   \  07    ???
   @MODRM
   ZERO-(w)?        \ no (w) operand length field in instruction
   reg#
   CASE
    00 OF $6B >R  ENDOF
    01 OF $6C >R  ENDOF
    02 OF $6D >R  ENDOF
    03 OF $6E >R  ENDOF
    04 OF $6F >R  ENDOF
    05 OF $72 >R  ENDOF
    06 OF $70 >R  ENDOF
    07 OF $72 >R  ENDOF
   ENDCASE
   R> DUP OPCODE>DIS
   $72 =
   IF   DEC-DisAddr
   ELSE R/M+DISP>DIS
   THEN ;


: GROUP08  ( -- )
   \ reg    opcode   op#
   \  00-03 ???
   \  04    BT       4C  00001111 10111010 mod 100 r/m imm8
   \  05    BTS      4E  00001111 10111010 mod 101 r/m imm8
   \  06    BTR      51  00001111 10111010 mod 110 r/m imm8
   \  07    BTC      55  00001111 10111010 mod 111 r/m imm8
   @MODRM reg# 4 <
   IF DEC-DisAddr    \ @MODRM inc DisAddr
      ILLEGAL
   ELSE
      SET-16/32?    \ BT BTS BTR BTC 16bit and 32bit regs only
      ZERO-(w)?     \ no (w) operand length field in instructio
      reg#
      CASE
         04 OF $4C >R  ENDOF
         05 OF $4E >R  ENDOF
         06 OF $51 >R  ENDOF
         07 OF $55 >R  ENDOF
      ENDCASE
      R> OPCODE>DIS
      PTR?  IF PTR>DIS THEN
      R/M+DISP>DIS ,>DIS @BYTE>DIS
   THEN ;

\ --- TWO-BYTE OPCODES

: O2??  ( byte -- )
   DROP ILLEGAL ;

: O200  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    GR06
   \  01    GR07
   \  02    LAR      48
   \  03    LSL      49
   \  04-05 ???
   \  06    CLTS     50
   \  07-0F ???
   $0F AND
   DUP 0=
   IF DROP GROUP06
      EXIT
   THEN
   DUP 1 =
   IF DROP GROUP07
      EXIT
   THEN
   DUP 2 3 EITHER=
   IF 2 - $48 + OPCODE>DIS
      @MODRM OPS>DIS
      EXIT
   THEN
   dup 6 = IF				\ SFP004
     drop  $4A OPCODE>DIS exit
   ENDIF
   dup 8 =
   if  drop  s" INVD" op$>dis exit  endif
   dup 9 =
   if  drop  s" WBINVD" op$>dis exit  endif
   DROP  ILLEGAL
;

: O220  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    MOV      1A
   \  01    MOV      1A
   \  02    MOV      1A
   \  03    MOV      1A
   \  04    MOV      1A
   \  05    ???
   \  06    MOV      1A
   \  07-0F ???
   DROP
   ILLEGAL ;

: o240		\ byte -- ; SFP002
  s" CMOV" >disline
  $0F and condCode >disline
  parm-tab

  ZERO-(w)?
  Set-16/32?		\ only 16bit & 32bit regs allowed
  @MODRM
  reg# REG>DIS ,>DIS
  PTR? IF PTR>DIS THEN
  R/M+DISP>DIS
;

: O280  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    JO       17
   \  01    JNO
   \  02    JB
   \  03    JNB
   \  04    JZ
   \  05    JNZ
   \  06    JBE
   \  07    JNBE
   \  08    JS
   \  09    JNS
   \  0A    JP
   \  0B    JNP
   \  0C    JL
   \  0D    JNL
   \  0E    JLE
   \  0F    JNLE
   $17 (OPCODE)>DIS
   $0F AND CONDCODE >DISLINE
   PARM-TAB
   @LONG-OFFSET ;

: O290  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    SETO     4B
   \  01    SETNO
   \  02    SETB
   \  03    SETNB
   \  04    SETZ
   \  05    SETNZ
   \  06    SETBE
   \  07    SETNBE
   \  08    SETS
   \  09    SETNS
   \  0A    SETP
   \  0B    SETNP
   \  0C    SETL
   \  0D    SETNL
   \  0E    SETLE
   \  0F    SETNLE
   $4B (OPCODE)>DIS
   $0F AND CONDCODE >DISLINE
    ZERO-(w)?     \ no (w) operand length field in instruction
    SET-8bit?      \ set 8bit regs and memory only
   47 DisPos !
   @MODRM
   PTR?  IF PTR>DIS THEN
   R/M+DISP>DIS ;

: O2A0  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    PUSH FS  0A
   \  01    POP FS   0B
   \  02    ???
   \  03    BT       4C  00001111 10100011 mod reg r/m
   \  04    SHLD     4D
   \  05    SHLD     4D
   \  06-07 ???
   \  08    PUSH GS  0A
   \  09    POP GS   0B
   \  0A    ???
   \  0B    BTS      4E  00001111 10101011 mod reg r/m
   \  0C    SHRD     4F
   \  0D    SHRD     4F
   \  0E    ???
   \  0F    IMUL     14  1111011w mod 101 r/m
   $0F AND   \ 00001111 and : use only 4bits
   DUP 2 <
   IF DROP
      $0A bit0 IF 1+ THEN OPCODE>DIS
      c" FS" count >DISLINE
      EXIT
   THEN
   DUP 3 =
   IF \ BT = 3
      \ 00001111 10100011 mod reg r/m
      \ reg16,reg16 | reg32,reg32 | mem16,reg16 | mem32,reg32
      DROP $4C OPCODE>DIS
      @MODRM
      \ allow only 16bit & 32bit regs and mem
       SET-16/32?
      R/M+DISP>DIS ,>DIS reg# REG>DIS
      EXIT
   THEN
\ SFP005...
  dup $04 = if		\ SHLD r/m, reg, imm8
    drop  s" SHLD" op$>dis
    @modrm set-16/32? R/M+DISP>DIS ,>DIS reg# REG>DIS
    ,>dis #>dis @byte>dis
    exit
  endif
  dup $05 = if		\ SHLD r/m, reg, cl
    drop  s" SHLD" op$>dis
    @modrm set-16/32? R/M+DISP>DIS ,>DIS reg# REG>DIS
    s" , CL" >DISLINE
    exit
  endif
\ ...SFP005
   DUP 08 09 EITHER=  \ PUSH & POP GS regs
   IF DROP
      $0A bit0 IF 1+ THEN OPCODE>DIS
      c" GS"  count >DISLINE
      EXIT
   THEN
   DUP $0B =
   IF \ BTS = 0B
      \ 00001111 10101011 mod reg r/m
      \ reg16,reg16 | reg32,reg32 | mem16,reg16 | mem32,reg32
      DROP $4E OPCODE>DIS
      @MODRM
      \ allow only 16bit & 32bit regs and mem
       SET-16/32?
      R/M+DISP>DIS ,>DIS reg# REG>DIS
      EXIT
   THEN
\ SFP005...
  dup $0C = if		\ SHRD r/m, reg, imm8
    drop  s" SHRD" op$>dis
    @modrm set-16/32? R/M+DISP>DIS ,>DIS reg# REG>DIS
    ,>dis #>dis @byte>dis
    exit
  endif
  dup $0D = if		\ SHRD r/m, reg, cl
    drop  s" SHRD" op$>dis
    @modrm set-16/32? R/M+DISP>DIS ,>DIS reg# REG>DIS
    s" , CL" >DISLINE
    exit
  endif
\ ...SFP005
   DUP $0F =
   IF \ IMUL = 0F
      \ 1111011w mod 101 r/m
      \ reg16,reg16 | reg32,reg32 | reg16,mem16 | reg32,mem32
      DROP $14 OPCODE>DIS
      @MODRM
      \ allow only 16bit & 32bit regs and mem
       SET-16/32?
      reg# REG>DIS ,>DIS R/M+DISP>DIS
      EXIT
   THEN
   DROP ILLEGAL ;

: O2B0  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-01 ???
   \  02    LSS      50
   \  03    BTR      51  00001111 10110011 mod reg r/m
   \  04    LFS      52
   \  05    LGS      53
   \  06-07 MOVZX    54   00001111 1011011w mod reg r/m
   \  08-09 ???
   \  0A    GR08
   \  0B    BTC      55   00001111 10111011 mod reg r/m
   \  0C    BSF      56   00001111 10111100 mod reg r/m
   \  0D    BSR      57   00001111 10111101 mod reg r/m
   \  0E-0F MOVSX    58   00001111 1011111w mod reg r/m
   $0F AND   \ 00001111 and : use only 4bits
   dup 0 1 either= if			\ SFP004
     drop  s" CMPXCHG" op$>dis
     @MODRM
     R/M+DISP>DIS ,>DIS reg# REG>DIS
     exit
   endif
   DUP 2 =
   IF \ 02 = LSS
      \ 00001111 10110011 mod reg r/m
      2- $50 + OPCODE>DIS
      @MODRM
      \ allow only 16bit & 32bit regs and mem
       SET-16/32?
      reg# REG32>DIS ,>DIS R/M+DISP>DIS
      EXIT
   THEN
   DUP 3 =
   IF \ 03 = BTR
      \ 00001111 10110011 mod reg r/m
      DROP $51 OPCODE>DIS
      @MODRM
      \ allow only 16bit & 32bit regs and mem
       SET-16/32?
       R/M+DISP>DIS ,>DIS reg# REG>DIS
      EXIT
   THEN
   DUP 4 5 EITHER=
   IF \ LGS = 5 & LFS = 4
      \ LFS = 00001111 1011100 mod reg r/m
      \ LGS = 00001111 1011101 mod reg r/m
      2- $50 + OPCODE>DIS
      \ no (w) operand length field in this instruction
       ZERO-(w)?
       SET-16/32?    \ only 16bit & 32bit regs allowed
      @MODRM
      reg# REG>DIS ,>DIS R/M+DISP>DIS
      EXIT
   THEN
   DUP 6 7 EITHER=              \ MOVZX = 6 or 7
   OVER $0E $0F EITHER= OR    \ MOVSX = 0E or 0F
   IF 6 7 EITHER=               \ MOVZX = 6 or 7
      IF $54 ELSE $58 THEN
      OPCODE>DIS
      \ MOVZX = $54
      \ MOVZX = 6 or 7
      \ 00001111 | 1011011w | mod reg r/m
      \ MOVSX = $58
      \ MOVSX = 0E or 0F
      \ 00001111 | 1011111w | mod reg r/m
      SET-16/32?    \ only 16bit & 32bit regs allowed
      @MODRM
      reg# REG>DIS ,>DIS
      ZERO-16/32?   \ source can 8bit or 16bit
      SET-8/16?      \ set 8bit & 16bit regs only
      PTR?  IF PTR>DIS THEN
      R/M+DISP>DIS
      EXIT
   THEN
   DUP $0A =
   IF \ 0A = GROUP08
      DROP GROUP08
      EXIT
   THEN
   DUP $0B =
   IF \ 0B = BTC
      \ 00001111 10111011 mod reg r/m
      \ reg16,reg16 | reg32,reg32 | mem16,reg16 | mem32,reg32
      $0B - $55 + OPCODE>DIS
      @MODRM
       SET-16/32?
      \ allow only 16bit & 32bit registers and memory
      R/M+DISP>DIS ,>DIS reg# REG>DIS
      EXIT
   THEN
   DUP $0C $0D EITHER=  \  0C = BSF , 0D = BSR
   IF $0B - $55 + OPCODE>DIS
      @MODRM
       SET-16/32?
      \ allow only 16bit & 32bit registers and memory
      reg# REG>DIS ,>DIS R/M+DISP>DIS
      EXIT
   THEN
   DROP ILLEGAL ;

: O2C0		\ byte -- ; SFP004
  $0F and
  dup 8 >= if				\ BSWAP
    s" BSWAP" op$>dis  r32>dis
    exit
  endif
  drop illegal
;

CREATE OP2map	\ -- addr
  ' O200 , ' O2?? , ' O220 , ' O2?? ,
  ' o240 , ' O2?? , ' O2?? , ' O2?? ,	\ SFP002
  ' O280 , ' O290 , ' O2A0 , ' O2B0 ,
  ' O2C0 , ' O2?? , ' O2?? , ' O2?? ,

: OP2  ( byte2 -- )
   DUP $F0 AND $10 / 4* OP2map + @ EXECUTE ;

: OP0F  ( -- )
   @OPCODE DUP Opcode2 !   \ save second byte opcode
   OP2 ;

\ --- ONE BYTE OPCODES

: OP00  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-05 ADD      00
   \  ADD 000000dw mod reg r/m  regs > regs
   \      0000000w mod reg r/m  regs > mem
   \      0000001w mod reg r/m   mem > regs
   \  06    PUSH ES  0A
   \  07    POP ES   0B
   \  08-0D OR       01
   \  0E    PUSH CS  0A
   \  0F    OP0F
   DUP $0F =
   IF DROP OP0F
   ELSE
      DUP 6 AND 6 =
      IF SEGREG-OP
      ELSE
         $38 AND -3 SCALE 2-OP-INST
      THEN
   THEN ;

: OP10  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-05 ADC      02
   \  06    PUSH SS  0A
   \  07    POP SS   0B
   \  08-0D SBB      03
   \  0E    PUSH DS  0A
   \  0F    POP DS   0B
   DUP 6 AND 6 =
   IF SEGREG-OP
   ELSE
      $38 AND -3 SCALE 2-OP-INST
   THEN ;

: OP20  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-05 AND      04
   \  06    SEG ES
   \  07    DAA      0C
   \  08-0D SUB      05
   \  0E    SEG CS
   \  0F    DAS      0D
   DUP $18 OR $3E =
   IF SEG:
   ELSE
      DUP 7 AND 7 =
      IF DADJ-INST DROP
      ELSE
         $38 AND -3 SCALE 2-OP-INST
      THEN
   THEN  ;

: OP30  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-05 XOR      06
   \  06    SEG SS
   \  07    AAA      0E
   \  08-0D CMP      07
   \  0E    SEG CS
   \  0F    AAS      0F
   DUP $18 OR $3E =
   IF SEG:
   ELSE
      DUP 7 AND 7 =
      IF ADJ-INST DROP
      ELSE
         $38 AND -3 SCALE 2-OP-INST
      THEN
   THEN  ;

: OP40  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-07 INC      08
   \  08-0F DEC      09
   DUP $48 <
   IF 08 ELSE 09 THEN OPCODE>DIS
   7 AND REG32>DIS ;

: OP50  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-07 PUSH     0A
   \  08-0F POP      0B
   DUP $58 <
   IF $0A ELSE $0B THEN OPCODE>DIS
   7 AND REG32>DIS ;

: OP60  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    PUSHA    10
   \  01    POPA     11
   \  02    BOUND    12    reg,mem
   \  03    ARPL     13    reg16,reg16/mem16
   \  04    SEG FS
   \  05    SEG GS
   \  06    op size  66    32bit to 16bit data overide
   \  07    adr size 67
   \  08    PUSH     0A    imm full 011010s0 imm data
   \  09    IMUL     14
   \ IMUL reg,reg/mem*imm sext 00001111 10101111 mod reg r/m
   \  0A    PUSH     0A    imm byte 011010s0 imm data
   \  0B    IMUL     14
   \ IMUL reg,reg/mem*imm byte 011010s1 mod reg r/m imm
   \  0C    INSB     15
   \  0D    INSW/D   15
   \  0E    OUTSB    16
   \  0F    OUTSW/D  16
   $0F AND
   DUP 4 <
   IF DUP $10 + OPCODE>DIS
      CASE
         02 OF @MODRM reg# REG32>DIS
               ,>DIS [>DIS @DWORD>DIS ]>DIS
            ENDOF
         03 OF  ZERO-32bit?
               @MODRM MODRM16 ,>DIS reg# REG>DIS
            ENDOF
      ENDCASE
      EXIT
   THEN
   \ - FS: GS:
   DUP 4 5 EITHER=
   IF SEGREG>DIS :>DIS
      EXIT
   THEN
   \ -- operand size override
   DUP 6 =
   IF DROP
      32bit?  -1 XOR SET-32bit?
       SET-$66?
      (DISASM-NEXT)
      EXIT
   THEN
   \ -- address size override
   DUP 7 =
   IF DROP
      32bit?  -1 XOR SET-32bit?
      (DISASM-NEXT)
      EXIT
   THEN
   \ - PUSH imm
   DUP 8 $0A EITHER=
   IF DROP $0A OPCODE>DIS
      bit1
      IF @BYTE>DIS
      ELSE
         @DWORD>DIS
      THEN
      EXIT
   THEN
   \ - IMUL reg,reg/mem*imm
   DUP 09 =
   IF DROP
       ZERO-(w)?
       SET-16/32?    \ only 16bit & 32bit regs allowed
      $14 OPCODE>DIS
      REG,R/M*IMM
      EXIT
   THEN
   DUP $0B =
   IF \ IMUL = 0B
      \ reg16,reg/mem16,imm8 | reg32,reg/mem32,imm8
      DROP  $14 OPCODE>DIS
      ZERO-(w)?
      \ no (w) operand length field in this instruction
      SET-16/32?    \ only 16bit & 32bit regs allowed
      @MODRM
      reg# REG>DIS  ,>DIS  R/M+DISP>DIS
      ,>dis #>dis @BYTE SEXT>DIS
      EXIT
   THEN
   \ - INS/OUTS
   $0C - 2/ $15 + OPCODE >DISLINE
   bit0
   IF  32bit?
      IF [char] D ELSE [char] W THEN
   ELSE  [char] B THEN
   CHAR>DIS ;

: OP70  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    JO       17
   \  01    JNO
   \  02    JB
   \  03    JNB
   \  04    JZ
   \  05    JNZ
   \  06    JBE
   \  07    JNBE
   \  08    JS
   \  09    JNS
   \  0A    JP
   \  0B    JNP
   \  0C    JL
   \  0D    JNL
   \  0E    JLE
   \  0F    JNLE
   $17 (OPCODE)>DIS
   $0F AND CONDCODE >DISLINE
   PARM-TAB
   @BYTE-OFFSET ;

: OP80  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-01 GR01
   \  02    ???
   \  03    GR01
   \  04-05 TEST     18 1000010w mod reg r/m
   \ TEST = 1111011w mod 000 r/m
   \  06-07 XCHG     19
   \  08-0B MOV      1A
   \  0C    MOV ..   1A
   \  0D    LEA      1B  10001101 mod reg r/m
   \  0E    MOV      1A
   \  0F    POP      0B
   $0F AND
   DUP 4 <
   IF DROP  GROUP01
      EXIT
   THEN
   DUP 4 5 EITHER=
   IF \ TEST = 04-05
      \ TEST opcode = 18
      \ TEST
      4 - 2/ $18 + OPCODE>DIS
      @MODRM
      R/M+DISP>DIS ,>DIS reg# REG>DIS
      EXIT
   THEN
   DUP 6 7 EITHER=
   IF \ XCHG = 06-07
      \ XCHG 1000011w mod reg r/m
      4 - 2/ $18 + OPCODE>DIS
      @MODRM
      reg# REG>DIS ,>DIS R/M+DISP>DIS
      EXIT
   THEN
   DUP 8 $0A WITHIN?
   IF DROP
      $1A OPCODE>DIS
      @MODRM OPS>DIS
      EXIT
   THEN
   DUP $0B =
   IF \ MOV = 0B
      \ reg16,reg16 | reg16,mem16 | reg32,reg32 | reg32,mem32
      \ 1000101w mod reg r/m
       SET-16/32?    \ only 16bit & 32bit regs allowed
      DROP $1A OPCODE>DIS
      @MODRM
      reg# REG>DIS ,>DIS R/M+DISP>DIS
      EXIT
   THEN
   DUP $0C =
   IF \ MOV = 0C
      \ reg16,sreg | mem16,sreg
      \ 1000101w mod reg r/m
      SET-16bit?   \ only 16bit regs allowed
      DROP $1A OPCODE>DIS
      @MODRM
      r/m REG>DIS ,>DIS reg# SEGREG>DIS
      EXIT
   THEN
   DUP $0E =
   IF \ MOV = 0E
      \ sreg,reg16 | sreg,mem16
      \ 10001100 mod sreg r/m
      ZERO-(w)?
      \ no (w) operand length field in this instruction
      SET-16bit?   \ only 16bit regs allowed
      $1A OPCODE>DIS
      @MODRM M/R,SEG>DIS
      EXIT
   THEN
   $0D =
   IF \ LEA = 0D
      \ 10001101 mod reg r/m
      ZERO-(w)?
      \ no (w) operand length field in this instruction
      SET-16/32?    \ only 16bit & 32bit regs allowed
      $1B OPCODE>DIS
      @MODRM
      reg# REG>DIS ,>DIS R/M+DISP>DIS
   ELSE
      \ POP = 0F
      \ POP 10001111 mod 000 r/m
      \ mem16 : mem32
      ZERO-(w)?
      \ no (w) operand length field in this instruction
      SET-16/32?    \ only 16bit & 32bit regs allowed
      $0B OPCODE>DIS
      @MODRM
      PTR>DIS
      R/M+DISP>DIS
   THEN ;

: OP90  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    NOP      1C
   \  01-07 XCHG     19
   \  08    CBW      1D
   \  09    CWD      1E
   \  0A    CALL     1F
   \  0B    WAIT     20
   \  0C    PUSHF    21
   \  0D    POPF     22
   \  0E    SAHF     23
   \  0F    LAHF     24
   $0F AND
   DUP 0=
   IF DROP $1C OPCODE>DIS
      EXIT
   THEN
   DUP 8 <
   IF $19 OPCODE>DIS
      7 AND 0 REG32>DIS ,>DIS REG32>DIS
      EXIT
   THEN
   DUP $0A =
   IF DROP $1F OPCODE>DIS
      @DWORD @WORD>DIS :>DIS DWORD>DIS
   THEN
   DUP $0B =  \ WAIT = 0B
   IF DROP $20 OPCODE>DIS EXIT THEN
   DUP $0C =  \ PUSHF = 0C
   IF DROP $21 OPCODE>DIS EXIT THEN
   DUP $0D =  \ POPF = 0D
   IF DROP $22 OPCODE>DIS EXIT THEN
   DUP $0E =  \ SAHF = 0E
   IF DROP $23 OPCODE>DIS EXIT THEN
   DUP $0F =  \ LAHF = 0F
   IF DROP $24 OPCODE>DIS EXIT THEN
   7 AND
   32bit?
   IF   IF $71 ELSE $73 THEN
   ELSE $1D +
   THEN
   OPCODE>DIS ;

: OPA0  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-03 MOV      1A
   \  04    MOVSB    25
   \  05    MOVSW/D  25
   \  06    CMPSB    26
   \  07    CMPSW/D  26
   \  08-09 TEST     18
   \  0A    STOSB    27
   \  0B    STOSW/D  27
   \  0C    LODSB    28
   \  0D    LODSW/D  28
   \  0E    SCASB    29
   \  0F    SCASW/D  29
   $0F AND
   DUP 4 <
   IF DROP $1A OPCODE>DIS ( MOV)
      bit1
      IF [>DIS @DWORD>DIS ]>DIS ,>DIS AC>DIS
      ELSE
         AC>DIS ,>DIS [>DIS @DWORD>DIS ]>DIS
      THEN
      EXIT
   THEN
   DUP 4 7 WITHIN?
   IF 4 - 2/ $25 + (OPCODE)>DIS ( MOVS/CMPS)
      B/W/D>DIS
      EXIT
   THEN
   DUP 8 9 EITHER=
   IF DROP $18 OPCODE>DIS
      AC>DIS ,>DIS bit0 ( TEST)
      IF @DWORD>DIS
      ELSE
         @BYTE>DIS
      THEN
      EXIT
   THEN
   $0A - 2/ $27 + (OPCODE)>DIS ( STOS/LODS/SCAS)
   B/W/D>DIS ;

: OPB0  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-07 MOV byte 1A
   \  08-0F MOV w/dw 1A
   $1A OPCODE>DIS
   7 AND bit3
   IF 8 + THEN
   (REG)>DIS ,>DIS
   bit3
   IF @DWORD>DIS
   ELSE
      @BYTE>DIS
   THEN ;

: OPC0  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-01 GR02
   \  02-03 RET near 2A
   \  04    LES      2B   11000100 mod reg r/m
   \  05    LDS      2C   11000101 mod reg r/m
   \  06-07 MOV      1A
   \  08    ENTER    2D
   \  09    LEAVE    2E
   \  0A-0B RET far  2A
   \  0C    INT 3    2F
   \  0D    INT n    2F
   \  0E    INTO     30
   \  0F    IRET     31
   DUP $C2 <
   IF GROUP02
      EXIT
   THEN
   $0F AND
   bit2 NOT bit1 AND
   IF DROP $2A (OPCODE)>DIS bit3 ( RET)
      IF [char] F CHAR>DIS THEN
      bit0 NOT
      IF PARM-TAB @WORD>DIS THEN
      EXIT
   THEN
   DUP 4 5 EITHER=
   IF \ LDS = 5 & LES = 4
      \ LES = 11000100 mod reg r/m
      \ LDS = 11000101 mod reg r/m
      DROP bit0
      IF $2C ELSE $2B THEN
       ZERO-(w)?
      \ no (w) operand length field in this instruction
       SET-16/32?    \ only 16bit & 32bit regs allowed
      OPCODE>DIS ( LDS/LES)
      @MODRM
      reg# REG>DIS ,>DIS R/M+DISP>DIS
      EXIT
   THEN
   DUP 6 7 EITHER=
   IF DROP $1A OPCODE>DIS  ( MOV )
      @MODRM PTR>DIS R/M+DISP>DIS IMM>DIS
      EXIT
   THEN
   DUP 8 9 EITHER=
   IF 8 - $2D + OPCODE>DIS
      bit0 NOT
      IF @WORD>DIS ,>DIS @BYTE>DIS
      THEN
      EXIT
   THEN
   DUP $0C $0D EITHER=
   IF DROP $2F OPCODE>DIS bit0 ( INT)
      IF @BYTE>DIS ELSE [char] 3 CHAR>DIS THEN
      EXIT
   THEN
   DROP bit0
   IF $31 ELSE $30 THEN OPCODE>DIS ; ( IRET/INTO)

Defer NDPins?	\ -- flag ; NDP instruction and return true, otherwise false
  assign false to-do NDPins?

: OPD0  ( byte -- )
   \ bit0-3 opcode   op#
   \  00-03 GR02
   \  04    AAM      32
   \  05    AAD      33
   \  06    ???
   \  07    XLAT     34
   \  08-0F ESC coprocessor
   DUP $D4 <
   IF GROUP02
      EXIT
   THEN
   $0F AND
   DUP 4 5 EITHER=
   IF 4 - $32 + OPCODE>DIS
      EXIT
   THEN
   DUP 6 =
   IF ILLEGAL DROP
      EXIT
   THEN
   DUP 7 =
   IF DROP $34 OPCODE>DIS ( XLAT)
      EXIT
   THEN
   DROP
   NDPins? 0= if
     $3F OPCODE>DIS ( ESC)
     @MODRM R/M+DISP>DIS
   endif
;


: OPE0  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    LOOPNE   35
   \  01    LOOPE    36
   \  02    LOOP     37
   \  03    JCXZ     38
   \  04-05 IN       39
   \  06-07 OUT      3A
   \  08    CALL     1F
   \  09-0B JMP      3F
   \  0C-0D IN       39
   \  0E-0F OUT      3A
   $0F AND
   DUP 4 <
   IF 3 AND $35 + OPCODE>DIS @BYTE-OFFSET
      EXIT
   THEN
   bit2 ( IN and OUT )
   IF DROP bit1 ( OUT or IN )
      IF $3A ELSE $39 THEN OPCODE>DIS
      bit1 NOT IF AC>DIS ,>DIS THEN
      bit3 ( port or DX )
      IF c" DX" count >DISLINE
      ELSE
         @BYTE>DIS
      THEN
      bit1 IF ,>DIS AC>DIS THEN
      EXIT
   THEN
   DUP 8 =
   IF DROP $1F OPCODE>DIS ( CALL )
      @DWORD relCall>dis		\ SFP003
      EXIT
   THEN
   DUP 8 >
   IF $3B OPCODE>DIS ( JMP )
      3 AND
      CASE
         1 OF @DWORD rel>DIS                    ENDOF
         2 OF @DWORD @WORD>DIS :>DIS DWORD>DIS  ENDOF
         3 OF @BYTE-OFFSET                      ENDOF
      ENDCASE
      EXIT
   THEN ;

0 value CallAbsInd?	\ -- flag
\ Set for CALL [] addr

: OPF0  ( byte -- )
   \ bit0-3 opcode   op#
   \  00    LOCK     3C
   \  01    ???
   \  02    REPNE    3D
   \  03    REP      3E
   \  04    HLT      40
   \  05    CMC      41
   \  06-07 GR03
   \  08    CLC      42
   \  09    STC      43
   \  0A    CLI      44
   \  0B    STI      45
   \  0C    CLD      46
   \  0D    STD      47
   \  0E    GR04
   \  0F    GR05
   $0F AND
   DUP 1 =
   IF ILLEGAL DROP
      EXIT
   THEN
   DUP 6 7 EITHER=
   IF DROP @MODRM
      reg#  \ reg field select instruction
      CASE
         0 OF $18 >R ENDOF   \ TEST
         1 OF $72 >R ENDOF   \ ???
         2 OF $60 >R ENDOF   \ NOT
         3 OF $61 >R ENDOF   \ NEG
         4 OF $62 >R ENDOF   \ MUL
         5 OF $14 >R ENDOF   \ IMUL
         6 OF $63 >R ENDOF   \ DIV    11111011w mod 110 r/m
         7 OF $64 >R ENDOF   \ IDIV   11111011w mod 111 r/m
      ENDCASE
      R> DUP OPCODE>DIS
      PTR?  IF PTR>DIS THEN
      $72 <>
      IF R/M+DISP>DIS reg# 0=
         IF IMM>DIS THEN
      THEN
      EXIT
   THEN
   DUP $0E $0F EITHER=
   IF @MODRM
      DUP  $0E = reg# 1 > AND
      SWAP $0F = reg# 7 = AND OR
      IF  ILLEGAL  EXIT  THEN
      0 -> CallAbsInd?			\ assume not CALL [] addr
      reg#
      CASE
         0 OF $08 OPCODE>DIS  ( INC)
              PTR?  IF PTR>DIS THEN
           ENDOF
         1 OF $09 OPCODE>DIS  ( DEC)
              PTR?  IF PTR>DIS THEN
           ENDOF
         2 OF $1F OPCODE>DIS  -1 -> CallAbsInd?  ENDOF ( CALL [] addr )
         3 OF $1F OPCODE>DIS ENDOF ( CALL)
         4 OF $3B OPCODE>DIS ENDOF ( JMP)
         5 OF $3B OPCODE>DIS ENDOF ( JMP)
         6 OF $0A OPCODE>DIS ENDOF ( PUSH)
      ENDCASE
      R/M+DISP>DIS
      CallAbsInd? if
        DisAddr 4 - @ body> ShowBranchTarget
      endif
      EXIT
   THEN
   DUP 1 > IF 1- THEN
   DUP 2 > IF 1+ THEN
   DUP 6 > IF 2- THEN
   $3C + OPCODE>DIS
;


\ primary despatcher

CREATE OP1map   ' OP00 , ' OP10 , ' OP20 , ' OP30 , ' OP40 , ' OP50 , ' OP60 , ' OP70 ,
                ' OP80 , ' OP90 , ' OPA0 , ' OPB0 , ' OPC0 , ' OPD0 , ' OPE0 , ' OPF0 ,

: OP1  ( byte -- )
   DUP $F0 AND $10 / 4* OP1map + @ EXECUTE ;

: DISASM-NEXT  ( -- )
   INIT-VARS
   @OPCODE DUP Opcode1 !	\ save opcode first byte in var
   OP1 ;

' DISASM-NEXT (FORWARD) !

: (DISASM)  ( -- )
   BASE @ >R HEX
   -1 SET-32bit?
   ZERO-$66?			\ reset operand size overide flag
   DISINIT
   DISASM-NEXT
   disclose
   R> BASE ! ;

: .DISLINE  ( -- )
   CR DisLine$ -TRAILING TYPE ;

: DASHES  ( n -- )
   0 2DUP +
   IF DO [char] - EMIT LOOP ELSE 2DROP THEN ;

: ?LABEL  ( addr -- nfa true / false )
   >NAME DUP -1 <> DUP 0= IF NIP THEN ;

: LABEL>DIS  ( nfa -- )
   CR 72 DASHES
   CR 34 SPACES
\   NAME>  ID.
   .name
   CR 72 DASHES ;

: ?NEXT         \ addr -- flag ; true if string is NEXT
  dup @ 0= if
    cr ." In the weeds ..."
    next-found on
  endif
  NextVal NextWidth s= dup
  if  next-found on  endif
;

: NEXT>DIS  ( -- )
  BASE @ HEX
  DISINIT
  c" NEXT," count >DISLINE
  NextWidth 0
  DO @BYTE DROP LOOP
  BASE !
  disclose
;

also forth definitions
: al-init-dis   \ addr len -- ; initialise disassembly
\ *G Initialise the disassembly range before using (DASM).
  over +  ToDisAddr !           \ set end of disassembly
  SET-DisAddr                   \ set start of disassembler
  next-found off                \ reset next found flag
  branch-target off             \ reset max branch target
;

: ft-init-dis   \ from to -- ; initialise disassembly
\ *G Initialise the disassembly range before using (DASM).
  ToDisAddr !                   \ set end of disassembly
  SET-DisAddr                   \ set start of disassembler
  next-found off                \ reset next found flag
  branch-target off             \ reset max branch target
;
previous definitions

: dis-cont?     \ -- flag ; true if in disassembly range
  DisAddr ToDisAddr @ u<        \ in range
  halt? not  and                \ and halt key not pressed
;

: dasm-done?    \ -- t/f ; true if dasm finished
  next-found @                  \ next found
  if  branch-target @
      disaddr u<                \ and no branch beyond this position
      if  -1
      else  next-found off      \ beyond, so reset next-found flag
            branch-target @     \ return true if gap
            disaddr $100 + u>   \ is greater than 256 bytes (arbitrary)
      endif
  else  0
  endif
;

: get-cfa       \ -- cfa flag ; get CFA of word name in input stream
  bl word find dup ?undef
;

also forth definitions
: 1DISASM	\ --
\ *G Disassemble the next instruction. The range has already been set.
   DisAddr ?NEXT
   IF  NEXT>DIS
   ELSE
       (DISASM)
   THEN
   .DISLINE
;

: (dasm)        \ --
\ *G Disassemble a block of code whose range has already been set.
  disaddr  0
  BEGIN
      Dis-cont?  dasm-done? not  and
    while
      1DISASM  1+
  repeat
  cr  ." ( " disaddr rot - u. ." bytes, " u. ." instructions )"
;
previous definitions

( **                                                                    )
( *S Low-Level Disassembly Words                                        )
( **                                                                    )

also forth definitions

: disasm/f      \ addr --
( *G Disassemble memory starting at ADDR.                               )
  1 op-line# !
  dup $800 + ft-init-dis (dasm)
  CR
;

: disasm/ft     \ from to --
( *G Disassemble memory between the memory addresses FROM and TO.       )
  1 op-line# !  ft-init-dis
  BEGIN
      Dis-cont?
    WHILE
      1DISASM
  REPEAT
  CR
;

: DISASM/al     \ addr len --
( *G Disassemble LEN bytes of code starting at memory address ADDR.     )
  1 op-line# !  al-init-dis
  BEGIN
      Dis-cont?
    WHILE
      (DISASM) .DISLINE
  REPEAT
  cr
;

: dasm          \ -- ; DASM <word>
( *G Disassemble a given definition                                     )
  1 op-line# !
  get-cfa 0>                    \ find CFA or word, immediate?
  if  ." (is IMMEDIATE)"  endif
  dup $1000 +                   \ restrict range to 4k bytes
  ft-init-dis (dasm)            \ set up and disassemble
  CR
;

previous previous definitions

( *> ###                                                                )

