\ dis.fth - higher level disassembly support

((
Copyright (c) 1999, 2001, 2009
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20130816 SFP006 Changed DIS and friends to use 'SYN
20090429 SFP005 Minor documentation update.
SFP004 30 Mar 2001    Prevent blow up on JMP 0 etc
SFP003 20 July 1999   Updated to cope with words exported from a module
SFP002 19 July 1999   Re-refactored NEWLINE to improve display.
SFP001 15 July 1999   Added ABORT" and length display
                      refactored NEWLINE and tabbing to reduce size
                      but refactoring TAB was a waste because of the
                      header size
                      The use of GETPARENTCFA for defining words is
                      inappropriate for those defining words which
                      have been redefined in OPT386.FTH.
))


\ ==========
\ *> dasm386
\ *S Higher Level Disassembly
\ ==========
\ *P VFX Forth also contains some higher-level words to aid
\ ** disassembly. These words attempt to interpret the raw
\ ** assembler code as provided by *\fo{DASM} to identify items
\ ** such as inline strings and *\fo{USER} variables.

only forth  also disassembler definitions

: GetParentCFA  \ cfa -- parent-cfa | 0
  dup c@ $E8 = if
    1+
    dup @ + CELL+ [ also system ] ip>nfa name> [ previous ]
  else
    drop 0
  then
;

: GetRuntime    \ cfa -- addr
  dup c@ $E8 = if
    1+
    dup @ + 4+
  else
    drop 0
  then
;
interpreter
: GetRuntime    \ cfa -- addr
  dup c@ $E8 = if
    1+
    dup @ + 4+
  else
    drop 0
  then
;
target

variable tab
2 constant tab-step

\ ============================================================================

0 value valx

struct structx
  cell field fieldx
end-struct

structx buffer: instx


also forth definitions

: (dis)         \ xt -- ; hacked SFP001
  op-line# off  cr
  dup >name .name
  dup immediate?
  if ." (is IMMEDIATE) " then

  dup GetRuntime ?dup if		\ cfa 'cfa --
    case
      [ ' EMIT GetRuntime ] literal of
        ." is DEFERred. Action is "
        >body @ dup >name .name
        recurse
      endof
      [ ' EMIT-GEN GetRuntime ] literal of
        ." is a GENIO Output Function #"
        dup >body @ .lword
        GetParentCFA recurse
      endof
      [ ' KEY-GEN GetRuntime ] literal of
        ." is GENIO Input Function #"
        dup >body @ .lword
        GetParentCFA recurse
      endof
      [ ' TAB-STEP GetRuntime ] literal of
        ." is a CONSTANT = "
        >body @ .lword
      endof
      [ ' TAB GetRuntime ] literal of
        ." is a VARIABLE "
        drop
      endof
      [ ' VALX GetRuntime ] literal of
        ." is a VALUE = "
        >body @ .lword
      endof
      [ ' >IN GetRuntime ] literal of
        ." is USER VARIABLE, offset = "
        >body @ @ .lword
      endof
      [ ' INSTX GetRuntime ] literal of
        ." is a STRUCTure"
        drop
      endof
      [ ' STRUCTX GetRuntime ] literal of
        ." is a STRUCTure"
        drop
      endof
      [ ' FIELDX GetRuntime ] literal of
        ." is a STRUCTure FIELD, offset = "
        >body @ .lword
      endof
        dup 5 - GetRunTime ['] (;CODE) =
        if
          ." is a child of " ip>nfa dup .name
          cr ." Data area contains"
          swap >body #16 dump
          name> recurse
        else
          drop disasm/f
        endif
    end-case
  else
    dup c@ $E9 = if
      dup 1+ dup @ + 4+ inForth? if	\ prevents blow up with JMP 0 etc
        ." starts with a JMP instruction,"
        cr ." and is probably exported from a MODULE."
        dup 5 disasm/al
        cr  1+ dup @ + 4+ disasm/f
      else
        disasm/f
      endif
    else
      disasm/f
    endif
  then
;

\ SFP006...
: dis           \ -- ; DIS <name>
\ *G Disassemble a supplied target word using the higher level
\ ** interpreter to aid readability.
  'syn
  dup >name nfa>ctrl c@ $010 and if
    dup >name .name ." is a SYNONYM of "
    >body @
    dup >name .name cr
  then
  dup (dis)
  dup ndcs? over immediate? 0= and
  if  ." NDCS:" >code-len @ disasm/f  else  drop  then
;
\ ...SFP006

: see           \ -- ; SEE <name>
\ *G An alias for *\fo{DIS} supplied for compatibility with other
\ ** Forth systems.
  dis
;

previous previous definitions

\ ======
\ *> ###
\ ======

