; openlib.asm - builds an Mach-O file that calls libraries
; - no linker required
; Build: nasm -f bin -o openlib openlib.asm && chmod +x openlib

; Copyright (c) 2009, 2011
; MicroProcessor Engineering
; 133 Hill Lane
; Southampton SO15 5AF
; England
;
; tel: +44 (0)23 8063 1441
; fax: +44 (0)23 8033 9691
; net: mpe@mpeforth.com
;      tech-support@mpeforth.com
; web: www.mpeforth.com
;
; From North America, our telephone and fax numbers are:
;   011 44 23 8063 1441
;   011 44 23 8033 9691
;
;
; To do
; =====
;
; Change history
; ==============
; 20110802 SFP002 Lion changes
; 20090112 JOR001 first working version


; ********
; Prologue
; ********

   BITS 32

   TextSize equ 0x00800000  ; ask for 8 megabytes for my interpreter
   l_mhdr equ 28            ; 7 cells in mhdr

   org          0x1000

; Mach-O header

mhdr:
   dd 0xFEEDFACE  ; magic
   dd 7           ; cputype
   dd 3           ; cpusubtype
   dd 2           ; filetype, executable MH_EXECUTE
   dd 10          ; ncmds, number of commands that follow
   dd sizeofcmds  ; sizeofcmds, sans this header
   dd 0x85        ; flags

; Segments have to be aligned to page size (4096)


; *****************
; Page 0 protection
; *****************
; First page of our program space where
; we land by referencing a null pointer

pagezero:				; command 1
   dd 1              ; LC_SEGMENT
   dd _pagezero_     ; size
   db '__PAGEZERO'   ; segname
   times 6 db 0      ; pad to 16 chars
   dd 0              ; vmaddr
   dd 0x1000         ; vmsize
   dd 0              ; fileoff
   dd 0              ; filesize
   dd 0              ; maxprot
   dd 0              ; initprot
   dd 0              ; nsects
   dd 0              ; flags
_pagezero_ equ $-pagezero

; *******************
; Text (code) segment - Mem 0x1000..0x1FFF
; *******************

; Commands occupy part of the first
; segment which happens to be text in our case.
; Actual code (start) will go right after, although
; there's nothing preventing us from wasting the rest
; of the first text page and starting our code on the
; next one.

; This can be handy in the Forth cross-compiler
; since the headers will likely be generated after
; the kernel code. We can pre-allocate 3 pages
; at the beginning of the file and use an offset
; to make sure the kernel code goes after, starting
; on a new page boundary.

; Last but not least, segments and other load commands
; don't have to go in this order. The first segment
; can be a text segment.

; Data

data:					; command 2
   dd 1              ; LC_SEGMENT
   dd _data_         ; size
   db '__DATA'       ; segname
   times 10 db 0     ; pad to 16 chars
   dd 0x1000         ; vmaddr
   dd 0x1000         ; vmsize
   dd 0              ; fileoff
   dd 0x1000         ; filesize
   dd 7              ; maxprot
   dd 7              ; initprot
   dd 1              ; nsects
   dd 0              ; flags

data0:              ; section 0
   db '__const'      ; sectname
   times 9 db 0      ; pad to 16 chars
   db '__DATA'       ; segname
   times 10 db 0     ; pad to 16 chars
   dd 0x1000         ; addr
   dd datasize       ; size
   dd 0              ; file offset
   dd 0              ; align on 2^0
   dd 0              ; reloff
   dd 0              ; nreloc
   dd 0              ; flags
   dd 0              ; reserved1
   dd 0              ; reserved2
_data_ equ $-data

; Stubs, e.g. dlopen, dlclose, etc.

stubs:					; command 3
   dd 1              ; LC_SEGMENT
   dd _stubs_        ; size
   db '__IMPORT'     ; segname
   times 8 db 0      ; pad to 16 chars
   dd 0x2000         ; vmaddr
   dd 0x1000         ; vmsize
   dd 0x1000         ; fileoff
   dd 0x1000         ; filesize
   dd 7              ; maxprot
   dd 7              ; initprot
   dd 1              ; nsects
   dd 0              ; flags

stubs0:             ; section 0
   db '__jump_table' ; sectname
   times 4 db 0      ; pad to 16 chars
   db '__IMPORT'     ; segname
   times 8 db 0      ; pad to 16 chars
   dd 0x2000         ; addr
   dd 20             ; size, 4 5-byte stubs
   dd 0x1000         ; file offset
   dd 6              ; align on 2^6
   dd 0              ; reloff
   dd 0              ; nreloc
   dd 0x04000008     ; flags ; attribs S_ATTR_SELF_MODIFYING_CODE type S_SYMBOL_STUBS
   dd 0              ; reserved1
   dd 5              ; reserved2, stub size
_stubs_ equ $-stubs

text:					; command 4
   dd 1              ; LC_SEGMENT
   dd _text_         ; size
   db '__TEXT'       ; segname
   times 10 db 0     ; pad to 16 chars
   dd 0x4000         ; vmaddr
   dd TextSize       ; vmsize
   dd 0x3000         ; fileoff
   dd codesize       ; filesize
   dd 7              ; maxprot
   dd 7              ; initprot
   dd 1              ; nsects
   dd 0              ; flags

text0:              ; section 0
   db '__text'       ; sectname
   times 10 db 0     ; pad to 16 chars
   db '__TEXT'       ; segname
   times 10 db 0     ; pad to 16 chars
   dd _start         ; addr
   dd codesize       ; size
   dd _start-$$      ; offset
   dd 0              ; align on 2^0
   dd 0              ; reloff
   dd 0              ; nreloc
   dd 0x00000400     ; flags ; includes some instructions.
   dd 0              ; reserved1
   dd 0              ; reserved2
_text_ equ $-text

; Raw data used by the dynamic linker

linkage:				; command 5
   dd 1              ; LC_SEGMENT
   dd _linkage_      ; size
   db '__LINKEDIT'   ; link table
   times 6 db 0      ; pad
   dd 0x3000         ; vmaddr
   dd 0x1000         ; vmsize
   dd symbols-$$     ; fileoff
   dd _symbols_      ; filesize
   dd 7              ; maxprot
   dd 7              ; initprot
   dd 0              ; nsects
   dd 0              ; flags
_linkage_ equ $-linkage

; Symbol table

symtab:					; command 6
   dd 2              ; LC_SYMTAB
   dd _symtab_       ; size
   dd symbols-$$     ; symoff - file offset
   dd 4              ; nsyms
   dd strings-$$     ; stroff
   dd _strings_      ; strsize
_symtab_ equ $-symtab

; Additional symbol table info used by the dynamic linker

dysymtab:				; command 7
   dd 0x0b           ; LC_DYSYMTAB
   dd _dysymtab_     ; size
   dd 0              ; ilocalsym
   dd 0              ; nlocalsym
   dd 0              ; iextdefsym
   dd 0              ; nextdefsym
   dd 0              ; iundefsym
   dd 4              ; nundefsym
   dd 0              ; tocoff
   dd 0              ; ntoc
   dd 0              ; modtaboff
   dd 0              ; nmodtab
   dd 0              ; extrefsymoff
   dd 0              ; nextrefsyms
   dd indirect-$$    ; indirectsymoff
   dd 4              ; nindirectsyms
   dd 0              ; extreloff
   dd 0              ; nextrel
   dd 0              ; locreloff
   dd 0              ; nlocrel
_dysymtab_ equ $-dysymtab

; Our dynamic linker

dylinker:				; command 8
   dd 0x0e           ; LC_LOAD_DYLINKER
   dd _dylinker_     ; size
   dd 12             ; nameoff, of the string below
   db '/usr/lib/dyld', 0
   align 4, db 0
_dylinker_ equ $-dylinker

; Initial thread (register) state

thrstate:				; command 9
   dd 0x5            ; LC_UNIXTHREAD
   dd _thrstate_     ; size
   dd 0x01           ; i386_THREAD_STATE
   dd 0x10           ; i386_THREAD_STATE_COUNT
   times 10 dd 0x00  ; cpu thread state
   dd _start         ; eip
   times 05 dd 0x00  ;
_thrstate_ equ $-thrstate

; Dynamic libraries that we will be using

dylib:					; command 10
   dd 0x0c           ; LC_LOAD_DYLIB
   dd _dylib_        ; size
   dd 0x18           ; nameoff
   dd 0x02           ; timestamp
   dd 0x006F0103     ; currentver
   dd 0x00010000     ; compatver
   db '/usr/lib/libSystem.B.dylib', 0
   align 4, db 0
_dylib_ equ $-dylib

sizeofcmds equ $-pagezero

; Data

; Not aligned to page size so part of the same first
; virtual memory page as the load commands above.
; Second page if you count PAGEZERO.

signon:   ; sign on string issued by Mach-O system calls
   db      "OpenLib.asm example",0x0a,0
l_signon equ $-1-signon
; parameter strings and numbers for printf call to shared library
libname:
   db '/usr/lib/libSystem.B.dylib', 0
printfname:
   db      "printf",0
formatstring:
   db      "Hello from %s, my number is %d",0x0a,0
myname:
   db      "Stephen",0
mynum:
   dd      12345678

datasize equ $-$$

align 4096

; Stubs

dlopen:
   times 5 hlt

dlsym:
   times 5 hlt

dlclose:
   times 5 hlt

dlerror:
   times 5 hlt

align 4096

; Symbols

symsecnum equ 1
symbols:

   ; dlopen

   dd offdlopen   ; nstrx
   db 0x01        ; type N_EXT
   db symsecnum   ; sect
   dw 0x0101      ; desc
   dd dlopen      ; value

   ; dlsym

   dd offdlsym    ; nstrx
   db 0x01        ; type N_EXT
   db symsecnum   ; sect
   dw 0x0101      ; desc
   dd dlsym       ; value

   ; dlclose

   dd offdlclose  ; nstrx
   db 0x01        ; type N_EXT
   db symsecnum   ; sect
   dw 0x0101      ; desc
   dd dlclose     ; value

   ; dlerror

   dd offdlerror  ; nstrx
   db 0x01        ; type N_EXT
   db symsecnum   ; sect
   dw 0x0101      ; desc
   dd dlerror     ; value

indirect:        ; indirect symbol table

   dd 0, 1, 2, 3  ; dlopen, dlsym, dlclose, dlerror

; String table

strings:

   db 0x20, 0x00

offdlopen equ $-strings
   db "_dlopen", 0

offdlsym equ $-strings
   db "_dlsym", 0

offdlclose equ $-strings
   db "_dlclose", 0

offdlerror equ $-strings
   db "_dlerror", 0

_strings_ equ $-strings
_symbols_ equ $-symbols

align 4096

; Mac x86 ABI requires 16 byte stack alignments.
; This is required by SSE, apparently, and all
; Intel Macs in production come with SSE.

%macro clib_prolog 1
   mov ebx, esp        ; remember current esp
   and esp, 0xFFFFFFF0 ; align to next 16 byte boundary (could be zero offset!)
   sub esp, 12         ; skip ahead 12 so we can store original esp
   push ebx            ; store esp (16 bytes aligned again)
   sub esp, %1         ; pad for arguments (make conditional?)
%endmacro

; arg must match most recent call to clib_prolog

%macro clib_epilog 1
   add esp, %1         ; remove arg pad
   pop ebx             ; get original esp
   mov esp, ebx        ; restore
%endmacro

; ********************
; Start of application
; ********************
; By fixing the entry point we can make some parts of an interpreter
; simpler.

_start:

   jmp     _start

; storage. N.B. proves write permission in code segment
plib:     ; holds library handle
   dd      -1
pprintf:  ; holds address of printf
   dd      -1

; Extracts from the C header files
; void *dlopen(const char *filename, int flag);
; void *dlsym(void *handle, const char*symbol);
; int dlclose(void *handle);
; char *dlerror(void);

; #define RTLD_LAZY 0x00001 /* Lazy function call binding.  */
; #define RTLD_NOW  0x00002 /* Immediate function call binding.  */
; #define RTLD_BINDING_MASK   0x3 /* Mask of binding time value.  */
; #define RTLD_NOLOAD 0x00004 /* Do not load the object.  */
; #define RTLD_GLOBAL 0x00100

realstart:
; sign on
   mov     ecx, signon
   mov     edx, l_signon
   call    primtype

; Open the library
   clib_prolog 8
   push    1                     ; RTLD_LAZY
   push    libname               ; "libc.so.6"
   call    dlopen                ; open library
   clib_epilog 16
   mov     [plib], eax           ; EAX is library handle or error code
; mov     ebx, eax              ; return value in EAX
; call    showhex               ; display result
   mov     ebx, 1                ; return code
   cmp     eax, 0                ; test return value
   je      getout

; Find printf
   clib_prolog 8
   push    printfname            ; string "printf"
   push    dword [plib]          ; library handle
   call    dlsym                 ; find function address
   clib_epilog 16
   mov     [pprintf], eax        ; EAX = function address or error code
; mov     ebx, eax
; call    showhex               ; display result
   mov     ebx, 2                ; return code
   cmp     eax, 0                ; test return value
   je      getout

; call printf
   clib_prolog 4
   push    dword [mynum]         ; second paramter to format string
   push    myname                ; first parameter to format string
   push    formatstring          ; format string itself
   call    [pprintf]             ; print and expand format string
   clib_epilog 16
; mov     ebx, eax              ; number of characters displayed
; call    showhex               ; display result

; close library
   clib_prolog 12
   push    dword [plib]          ; library handle
   call    dlclose
   clib_epilog 16
; mov     ebx, eax
; call    showhex               ; display result
   mov     ebx, 3                ; error return code
   cmp     eax, 0                ; test result
   jne     getout

; back to OS X
   mov     ebx, 42               ; success return code
getout:
   mov     eax, 1                ; exit function number
   int     0x80                  ; OSX system call
   ret

; simple type routine, ECX=caddr EDX=len
primtype:
   mov     eax, 4                ; write call
   mov     ebx, 1                ; STDOUT=1
   int     0x80
   ret

; emit single character, EAX=char
primemit:
   push    eax
   mov     ecx, esp
   mov     edx, 1
   call    primtype
   pop     eax
   ret

; Display EBX as 8 digit hex number
showhex:
   push    eax
   push    ebx

   mov     ecx, 8
sh1:
   push    ecx
   push    ebx
   mov     eax, ebx
   shr     eax, 28
   cmp     eax, 9
   jle     sh2
   add     eax, 7
sh2:
   add     eax, 0x30
   call    primemit
   pop     ebx
   pop     ecx
   shl     ebx, 4
   sub     ecx, 1
   jne     sh1
   mov     eax, 0x20
   call    primemit

   pop     ebx
   pop     eax
   ret

codesize equ $-_start

