;
; Basic OS X calls - replicates tstmin.asm but as bin
;
; compile with:
; /usr/bin/nasm -f bin -o tstbin  tstbin.asm
;

; ********
; Prologue
; ********

   BITS 32

   TextSize equ 0x00800000  ; ask for 8 megabytes for my interpreter
   l_mhdr equ 28            ; 7 cells in mhdr

   org          0x1000

; *************
; Mach-O header
; *************

mhdr:
   dd 0xFEEDFACE  ; magic
   dd 7           ; cputype
   dd 3           ; cpusubtype
   dd 2           ; filetype, executable MH_EXECUTE
   dd 13          ; ncmds, number of commands that follow
   dd sizeofcmds  ; sizeofcmds, sans this header
   dd 0x85        ; flags

; Segments have to be aligned to page size (4096)


; *****************
; Page 0 protection
; *****************
; First page of our program space where
; we land by referencing a null pointer

pagezero:				; command 1
   dd 1              ; LC_SEGMENT
   dd _pagezero_     ; size
   db '__PAGEZERO'   ; segname
   times 6 db 0      ; pad to 16 chars
   dd 0              ; vmaddr
   dd 0x1000         ; vmsize
   dd 0              ; fileoff
   dd 0              ; filesize
   dd 0              ; maxprot
   dd 0              ; initprot
   dd 0              ; nsects
   dd 0              ; flags
_pagezero_ equ $-pagezero

; *******************
; Text (code) segment - Mem 0x1000..0x1FFF
; *******************

; Commands occupy part of the first
; segment which happens to be text in our case.
; Actual code (start) will go right after, although
; there's nothing preventing us from wasting the rest
; of the first text page and starting our code on the
; next one.

; This can be handy in the Forth cross-compiler
; since the headers will likely be generated after
; the kernel code. We can pre-allocate 3 pages
; at the beginning of the file and use an offset
; to make sure the kernel code goes after, starting
; on a new page boundary.

; Last but not least, segments and other load commands
; don't have to go in this order. The first segment
; can be a text segment.

text:					; command 2
   dd 1              ; LC_SEGMENT
   dd _text_         ; cmd size
   db '__TEXT'       ; segname
   times 10 db 0     ; pad to 16 chars
   dd 0x1000         ; vmaddr
   dd 0x1000         ; vmsize
   dd 0              ; fileoff
   dd 0x1000         ; filesize
   dd 7              ; maxprot
   dd 5              ; initprot
   dd 4              ; nsects
   dd 0              ; flags

text1:              ; section 1
   db '__text'      ; sectname
   times 10 db 0      ; pad to 16 chars
   db '__TEXT'       ; segname
   times 10 db 0     ; pad to 16 chars
   dd _main          ; vmaddr
   dd l_main         ; vmsize
   dd _main-$$       ; file offset
   dd 0              ; align on 2^0
   dd 0              ; reloff
   dd 0              ; nreloc
   dd 0x80000400     ; flags
   dd 0              ; reserved1
   dd 0              ; reserved2
_text_ equ $-text

text2:
   db '__symbol_stub' ; sectname
   times 3 db 0	     ; pad  to 16 chars
   db '__TEXT'       ; segname
   times 10 db 0     ; pad to 16 chars
   dd  symstubs      ; vmaddr
   dd  _symstubs_    ; vmsize
   dd  symstubs-$$   ; file offset
   dd  2             ; align on 2^2
   dd  0             ; reloff
   dd  0             ; nreloc
   dd  0x80000508    ; flags
   dd  0             ; reserved1 (index into indirect symbol table)
   dd  6             ; reserved2 (size of each stub)

text3:
   db '__stub_helper' ; sectname
   times 3 db 0      ; pad to 16 chars
   db '__TEXT'       ; segname
   times 10 db 0     ; pad to 16 chars
   dd  stubhelper    ; vmaddr - was 0x00001f84
   dd  _stubhelper_  ; vmsize
   dd  stubhelper-$$ ; file offset
   dd  2             ; align on 2^2
   dd  0             ; reloff
   dd  0             ; nreloc
   dd  0x80000500    ; flags
   dd  0             ; reserved1
   dd  0             ; reserved2


   Section
  sectname __stub_helper
   segname __TEXT
      addr 0x00001f84
      size 0x00000034
    offset 3972
     align 2^2 (4)
    reloff 0
    nreloc 0
     flags 0x80000500
 reserved1 0
 reserved2 0




; data goes here
segment .data

szlibname:
   db '/usr/lib/libSystem.B.dylib', 0
szprintf:
       db "printf",0
formatstring:
   db      "Hello from %s, my number is %d",0x0a,0
myname:
   db      "Stephen",0
mynum:
   dd      12345678

; storage. N.B. proves write permission in code segment
plib:     ; holds library handle
   dd      -1
pprintf:  ; holds address of printf
   dd      -1


; ******************************
; Code goes here - see Section 1 - should be address 0x1E60
; ******************************

_main:

; *******************************************
; This code appears to be added by the linker
; *******************************************
  push     0				; create a slot
  mov      ebp, esp			; pointed to by EBP
  and      esp, -16
  sub      esp, 16			; reserve four more slots
  mov      ebx, 4 [ebp]			; first stack parameter, argc
  mov      0 [esp], ebx			; into new slot 0
  lea      ecx, 8 [ebp]			; address of argv array
  mov      4 [esp], ecx			; into new slot 1
  add      ebx, 1			; argc + 1
  shl      ebx, 2			; argc + 1 cells
  add      ebx, ecx                     ; plus base of argv
  mov      8 [esp], ebx			; into new slot 2

entryloop1:				; walk argv array?
  mov      eax, 0 [ebx]			; ???
  add      ebx, 4
  test     eax, eax			; EAX = ???
  jnz      entryloop1

  mov      12 [ebp], ebx		; end into new slot 3
  call     entrypt2
  mov      0 [esp], eax			; return code
  call     0x01000F7C
  hlt
entrypt2:
   and     esp, 0xFFFFFFF0	; 16 byte align down
   push    0			; padding for alignment
   push    0
   push    1                     ; RTLD_LAZY
   push    szlibname             ; "libSystem.B.dylib"
   call    _dlopen               ; open library
   add     esp, 16               ; remove params and padding
   mov     [plib], eax           ; EAX is library handle or error code
   mov     ebx, 1                ; return code
   cmp     eax, 0                ; test return value
   je  near    failed

 ; Find printf
   push    0
   push    0
   push    szprintf              ; string "printf"
   push    dword [plib]          ; library handle
   call    _dlsym                ; find function address
   add     esp, 16               ; remove params and padding
   mov     [pprintf], eax        ; EAX = function address or error code
   mov     ebx, 2                ; return code
   cmp     eax, 0                ; test return value
   je      failed

; call printf
   push    0
   push    dword [mynum]         ; second paramter to format string
   push    myname                ; first parameter to format string
   push    formatstring          ; format string itself
   call    [pprintf]             ; print and expand format string
   add     esp, 16               ; remove params and padding

; close library
   push    0
   push    0
   push    0
   push    dword [plib]          ; library handle
   call    _dlclose
   add     esp, 16               ; remove params and padding
   mov     ebx, 3                ; error return code
   cmp     eax, 0                ; test result
   jne     failed

; back to OS X
   mov     ebx, 42               ; success return code
failed:
   mov     eax, 1                ; exit function number
   push    0                     ; padding
   int     0x80                  ; OSX system call
   pop     eax                   ; remove padding
   ret

l_main equ $-_main


; ****************************
; Symbol stubs - see Section 2
; ****************************

symstubs:			; was 0x0F6A

stub_dlclose:
  jmp     [a_dlclose]
stub_dlopen:
  jmp     [a_dlopen]
stub_dlsym:
  jmp     [a_dlsym]
stub_exit:
  jmp     [a_exit]		; 0F7C -> 2028

_symstubs_ equ $-symstubs

; *************
; Symbol helper - section 3
; *************

symhelper:

help_dlclose:			; was 0x0F84, len 0x34
   db       0x68                ; push long
   dd       0x0000
   db       0xE9                ; jmp
   dd       xdlclose-$-4        ; offset 0x001e
help_dlopen:
   db       0x68                ; push long
   dd       0x000f
   db       0xE9                ; jmp
   dd       xdlopen-$-4         ; offset 0x0014
help_dlsym:
   db       0x68                ; push long
   dd       0x001d
   db       0xE9		; jmp
   dd       xdlsym-$-4		; offset 0x000a
help_exit

