;
; Basic OS X calls through libSystem
;
; compile with:
; /user/bin/nasm -O1 -f macho tstmin.asm
; /usr/bin/gcc -g0 -mdynamic-no-pic -no_pie -arch i386 -o tstmin tstmin.o
; strip -u tstmin                           (removes unnecessary symbols)
; otool -h -l tstmin >tstmin.txt
;

extern _dlopen, _dlsym, _dlclose, _dlerror

; data goes here
segment .data

szlibname:
   db '/usr/lib/libSystem.B.dylib', 0
szprintf:
       db "printf",0
formatstring:
   db      "Hello from %s, my number is %d",0x0a,0
myname:
   db      "Stephen",0
mynum:
   dd      12345678

pprintf:  ; holds address of printf
   dd      -1

; ***************************
; storage later used by Forth
; ***************************
align 4, db 0
pmagic:		; identifies 32 bit system
   dd      0xfeedface			; magic number as identifier
   dd      _magiclen_			; length of this block from pmagic
plib:     ; holds library handle
   dd      -1				; handle for libSystem.B.dylib
   dd     _dlopen			; function address
   dd     _dlsym			; function address
   dd     _dlclose			; function address
   dd     _dlerror			; function address
_magiclen_ equ $-pmagic


; ********************
; MPE code starts here
; ********************

segment .text

global _main

_main:

; Open the library
   and     esp, 0xFFFFFFF0	; 16 byte align down
   push    0			; padding for alignment
   push    0
   push    1                     ; RTLD_LAZY
   push    szlibname             ; "libSystem.B.dylib"
   call    _dlopen               ; open library
   add     esp, 16               ; remove params and padding
   mov     [plib], eax           ; EAX is library handle or error code
   mov     ebx, 1                ; return code
   cmp     eax, 0                ; test return value
   je  near    failed

 ; Find printf
   push    0
   push    0
   push    szprintf              ; string "printf"
   push    dword [plib]          ; library handle
   call    _dlsym                ; find function address
   add     esp, 16               ; remove params and padding
   mov     [pprintf], eax        ; EAX = function address or error code
   mov     ebx, 2                ; return code
   cmp     eax, 0                ; test return value
   je      failed

; call printf
   push    0
   push    dword [mynum]         ; second paramter to format string
   push    myname                ; first parameter to format string
   push    formatstring          ; format string itself
   call    [pprintf]             ; print and expand format string
   add     esp, 16               ; remove params and padding

; close library
   push    0
   push    0
   push    0
   push    dword [plib]          ; library handle
   call    _dlclose
   add     esp, 16               ; remove params and padding
   mov     ebx, 3                ; error return code
   cmp     eax, 0                ; test result
   jne     failed

; back to OS X
   mov     ebx, 42               ; success return code
failed:
   mov     eax, 1                ; exit function number
   push    0                     ; padding
   int     0x80                  ; OSX system call
   pop     eax                   ; remove padding
   ret

