\ SaveOSX.fth - VFX Forth for OSX MachO image save

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20110119 SFP001 First Mach-o version.
))

decimal

\ ==========
\ *> turnkey
\ ==========
\ *S Saving to an Mach-o file
\ *P The following sequence performs this operation:
\ *E ok SAVE myfile.mo
\ ** ok BYE
\ ** poop> ./myfile.mo

only forth definitions

: Mb		\ n -- nMb ; nMb = n * 1048576
\ *G Given *\i{n}, returns n megabytes. Useful before *\fo{SET-SIZE}
\ ** or *\fo{ALLOCATE}.
  1048576 *
;

: Kb		\ n -- nKb ; nKb = n * 1024
\ *G Given *\i{n}, returns n kilobytes. Useful before *\fo{SET-SIZE}
\ ** or *\fo{ALLOCATE}.
  1024 *
;

: get-size	\ -- size
\ *G Return the amount of memory used by the Forth dictionary and
\ ** system headers.
  codesize @ pe-coderva +
;

: set-size	\ size --
\ *G Set the amount of memory to be used by the Forth dictionary and
\ ** system headers. This will not take effect until the system has
\ ** been *\fo{SAVE}d to form a new Mach-o file. The new dictionary
\ ** size will be used when the new Mach-o file is run.
  pe-codeRVA - newcodesize !
;

: get-stacks	\ -- size
\ *G Return the amount of memory available for Forth stacks and
\ ** user areas.
  StackSize @
;

: set-stacks	\ size --
\ *G Set the amount of memory to be used by the Forth stacks and
\ ** user area. This will not take effect until the system has
\ ** been *\fo{SAVE}d to form a new Mach-o file. The new size will
\ ** be used when the new Mach-o file is run.
  StackSize !
;

: BeforeSave	\ --
\ +G Prepare the system for saving.
  CheckDict
  ['] FirstName here removeSinInRange	\ remove all tokeniser data
  freeze
;

: AppImage	\ -- addr len
\ +G Return the base address and length of the application.
  loadbase @ PE-CODERVA +  here over -
;


\ The file starts at StartAddress, with the header region extending
\ up to _Entry.

StartAddress constant MoHdrs
_Entry StartAddress - constant /MoHdrs

\ The dictionary size is saved at
\   textseg_ts		segment size (in header)
\   codesize		actual (in dictionary)
\   newcodesize		requested (in dictionary)

textseg_ts StartAddress - equ offTSeg_/TS
textseg_fs StartAddress - equ offTSeg_/FS
textsec0_fs StartAddress - equ offTsec0_/FS

: prepMoHdr	\ addr --
\ +G Set up the headers for saving.
  StartAddress over /MoHdrs cmove	\ copy headers
  dup $1000 + $1000 erase		\ wipe stubs and command line
  dup $1000 + #20 $F4 fill		\ reset stubs
  newcodesize @ over offTSeg_/TS + !	\ set segment size
  newcodesize @ codesize !		\ reset in dictionary
  here _Entry - swap			\ -- size addr ; code size of dictionary
  2dup offTSeg_/FS + !
  offTsec0_/FS + !
;

: $SaveImage	\ caddr len -- len'
\ *G Save the Forth image to the given file. The image file size
\ ** is returned.
  r/w create-file #-412 ?throw >r	\ create o/p file
  /MoHdrs protAlloc			\ -- addr ; R: -- fid
  dup prepMoHdr				\ set up headers
  dup /MoHdrs r@ write-file #-415 ?throw \ Write headers
  protFree				\ Free buffer
  _Entry here over -			\ dictionary segment
  r@ write-file #-415 ?throw		\ Write rest of image
  r> close-file #-416 ?throw		\ close
  /MoHdrs here _Entry - +		\ size of file
;

defer OS_GetUmask	\ -- mode
defer OS_Chmod		\ caddr u mode -- ior

also system definitions

assign (OS_GetUmask)		to-do OS_GetUmask
assign (OS_Chmod)		to-do OS_Chmod

previous definitions

$1ff equ MODE_777	\ Octal 777 - rwx for all

: setExecPerms	\ caddr len --
\ *G Set execute permissions on the file.
  OS_GetUmask invert MODE_777 and	\ Generate desired permission bits
  OS_Chmod drop				\ Ignore chmod failures
;

: Save		\ "<name>" -- ; SAVE <name>
\ *G Save the application to the file given by the following file
\ ** name. No extension is applied to the name. The image is saved
\ ** as a Mach-o file.
  align BeforeSave
  GetPathSpec 2dup $SaveImage >r	\ -- caddr len ; R: -- /img
  2dup setExecPerms
  cr  r> . ." bytes saved to " type
;


\ ======
\ *> ###
\ ======

decimal

