//  VFX Forth tests of OSX

/* Original VFX test
#include <stdio.h>
// #include <stdlib.h>
#include <windows.h>
// #include <winsock2.h>

For 32 bit use, compile with
  gcc -m32 osxtest.c  -o osxtest

*/

#include <stdio.h>
#include <signal.h>
#include <sys/ucontext.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <semaphore.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/termios.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <mach-o/loader.h>
#include <dlfcn.h>


int main( int argc, char *argv[] )
{
   struct stat sbuf;
   struct sigaction act;
   siginfo_t info;
   ucontext_t context;
   int t1, t2, t3, t4, t5, sz1;
   sem_t sem;
   pthread_attr_t attr;
   pthread_mutex_t pmutex;
   struct ifreq myifreq;

   printf("\nGeneral data types\n");
   printf(" int size      = %d\n",(int)sizeof(int));
   printf(" short size    = %d\n",(int)sizeof(short));
   printf(" clock_t size  = %d\n",(int)sizeof(clock_t));
   printf(" time_t size   = %d\n",(int)sizeof(time_t));
   printf(" sizeof(off_t) = %d\n",(int)sizeof(off_t));
   printf(" sizeof addr   = %d\n",(int)sizeof(&sbuf));
   printf(" sizeof size_t = %d\n",(int)sizeof(size_t));

   printf("\ndlopen and friends\n");
   printf(" RTLD_LAZY = %d\n", (int)RTLD_LAZY );
   printf(" RTLD_NOW = %d\n", (int)RTLD_NOW );
   printf(" RTLD_GLOBAL = 0x%X\n", (int)RTLD_GLOBAL );
   printf(" RTLD_LOCAL = 0x%X\n", (int)RTLD_LOCAL );


   printf("\nMach-O File Format\n");
   printf("Types\n");
   printf(" S_REGULAR = %d\n", (int)S_REGULAR );
   printf(" S_ZEROFILL = %d\n", (int)S_ZEROFILL );
   printf(" S_SYMBOL_STUBS = %d\n", (int)S_SYMBOL_STUBS );
   printf("Attributes\n");
   printf(" S_ATTR_PURE_INSTRUCTIONS = 0x%X\n", (int)S_ATTR_PURE_INSTRUCTIONS );
   printf(" S_ATTR_SOME_INSTRUCTIONS = 0x%X\n", (int)S_ATTR_SOME_INSTRUCTIONS );
   printf(" S_ATTR_SELF_MODIFYING_CODE = 0x%X\n", (int)S_ATTR_SELF_MODIFYING_CODE );
   printf(" S_ATTR_NO_TOC = 0x%X\n", (int)S_ATTR_NO_TOC );
   printf(" S_ATTR_EXT_RELOC = 0x%X\n", (int)S_ATTR_EXT_RELOC );
   printf(" S_ATTR_LOC_RELOC = 0x%X\n", (int)S_ATTR_LOC_RELOC );
   printf(" S_ATTR_STRIP_STATIC_SYMS = 0x%X\n", (int)S_ATTR_STRIP_STATIC_SYMS );
   printf(" S_ATTR_NO_DEAD_STRIP = 0x%X\n", (int)S_ATTR_NO_DEAD_STRIP );
   printf(" S_ATTR_LIVE_SUPPORT = 0x%X\n", (int)S_ATTR_LIVE_SUPPORT );

   printf("\nOSX constants\n");
   printf(" PATH_MAX = %d\n", PATH_MAX);
   printf(" _SC_CLK_TCK = %d\n", _SC_CLK_TCK);		// get value with sysconf()
   printf(" CLOCKS_PER_SEC = %d\n", (int)CLOCKS_PER_SEC);
   printf(" FIONREAD = 0x%x\n", (int)FIONREAD);
   printf(" O_CREAT = 0x%x\n", (int)O_CREAT);
   printf(" NCCS = %d\n", (int)NCCS);
   printf(" TCSANOW = %d\n", (int)TCSANOW);
   printf(" ICANON = %d\n", (int)ICANON);
   printf(" ECHO = %d\n", (int)ECHO);

   printf("\nFile structures\n");
   t1 = (int) (void *) &sbuf;
   t2 = (int) (void *) &(sbuf.st_size);
   sz1 = t2-t1;
   printf("offset st_size = %d \n", sz1);

   printf("\nSignals\n");
   printf(" SA_SIGINFO = %d\n", SA_SIGINFO);
   printf(" SIGFPE = %d\n", SIGFPE);
   printf(" sigaction size = %d\n", (int)sizeof(act) );
   printf(" siginfo_t size = %d\n", (int)sizeof(info) );
   t1 = (int) (void *) &info;
   t2 = (int) (void *) &(info.si_addr);
   sz1 = t2-t1;
   printf(" offset si_addr = %d\n", sz1 );
//   printf(" offset si_addr = %d\n", (int)offsetof(siginfo_t, si_addr) );
   printf(" ucontext_t size = %d\n", (int)sizeof(context) );
   printf(" SIG_DFL = %d\n", (int)SIG_DFL);
   printf(" SIG_IGN = %d\n", (int)SIG_IGN);

   printf("\nSemaphores, Timers and Threads\n");
   printf(" sem_t size = %d\n", (int)sizeof(sem) );
   printf(" PTHREAD_STACK_MIN = %d\n", PTHREAD_STACK_MIN );
   printf(" pthread_attr_t size = %d\n", (int)sizeof(attr) );
   printf(" ITIMER_REAL = %d\n", ITIMER_REAL );
   printf(" ITIMER_VIRTUAL = %d\n", ITIMER_VIRTUAL );
   printf(" ITIMER_PROF = %d\n", ITIMER_PROF );
   printf(" pthread_mutex_t size = %d\n", (int)sizeof(pmutex) );


   printf("\nsa_flags\n");
   printf(" SA_NOCLDSTOP = %d, $%X\n", SA_NOCLDSTOP, SA_NOCLDSTOP);
//   printf(" SA_ONESHOT = %d, $%X\n", SA_ONESHOT, SA_ONESHOT);
   printf(" SA_ONSTACK = %d, $%X\n", SA_ONSTACK, SA_ONSTACK);
   printf(" SA_RESTART = %d, $%X\n", SA_RESTART, SA_RESTART);
   printf(" SA_NODEFER = %d, $%X\n", SA_NODEFER, SA_NODEFER);
   printf(" SA_SIGINFO = %d, $%X\n", SA_SIGINFO, SA_SIGINFO);
   printf(" SIGFPE = %d\n", SIGFPE);

   t1 = (int)( void *) &act;
   t2 = (int)(void *) &(act.sa_flags);
   t3 = (int)(void *) &(act.sa_sigaction);
   t4 = (int)(void *) &(act.sa_mask);
   printf("offset of sa_flags = %d\n", t2-t1);
   printf("offset of sa_sigaction = %d\n", t3-t1);
   printf("offset of sa_mask = %d\n", t4-t1);

   printf("\nSockets and Networking\n");
   printf(" IFNAMSIZ = %d, $%X\n", IFNAMSIZ, IFNAMSIZ);
//   printf(" IFHWADDRLEN = %d, $%X\n", IFHWADDRLEN, IFHWADDRLEN);
   printf(" ifreq size = %d\n", (int)sizeof(myifreq) );
   t1 = (int)( void *) &myifreq;
//   t2 = (int)(void *) &(myifreq.ifr_hwaddr.sa_data);
   t3 = (int)(void *) &(myifreq.ifr_name);
//   printf("offset of MAC address = %d\n", t2-t1);
   printf("offset of MAC name = %d\n", t3-t1);

   return 0;
}
