\ binder.fth - merge OSX stub and Forth image to Mach-O file

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20110812 SFP001 Tested on 10.6 and 10.7 with mods.
))

only forth definitions
decimal

\ ===========
\ *! binder
\ *T OS X stub and image binder
\ ===========
\ *P Cross compilation of the kernel produces a .img file with no
\ ** Mach-O headers. The headers are provided by *\i{stubosx}.
\ ** This binder code merges the stub and the image to produce
\ ** a Forth executable.

: StubName	\ -- caddr len ; Name of stub file
  s" 386OSX/stubosx"
;

: ImgName	\ -- caddr len ; Name of Forth image file
  s" VFXFORTH.img"
;

: AppName	\ -- caddr len ; Name of output
  s" vfxkern.mo"
;

0 value FileBase		\ -- addr ; base address of file buffer
0 value /Stub			\ -- len ; stub length
$0000:1000 value /ForthImg	\ -- len ; file size of Forth image (4k aligned)
8 MB value /ForthMem		\ -- len ; vmsize for Forth image (4k aligned)
0 value offDATAseg		\ -- addr ; offset of DATA segment in memory
0 value offLEseg		\ -- addr ; offset of LINKEDIT segment in memory

$1C constant /mohdr	\ -- len ; size of Mach-O header
$38 constant /seg	\ -- len ; size of LC_SEGMENT command
$44 constant /sect	\ -- len ; size of SECTION data

: InImg		\ offset -- addr
  FileBase +
;

: setImgData	\ --
  ImgName r/w open-file abort" open failed"
  dup file-size abort" size failed" drop
  $0FFF +  $-1000 and			\ page aligned size
  to /ForthImg
  close-file drop
;

: getBuff	\ --
  10 Mb allocate abort" allocation failed" to FileBase
;

: freeBuff	\ --
  FileBase free drop
;

: readStub	\ --
  StubName r/o open-file abort" open failed" >r
  r@ file-size abort" size failed" drop to /Stub
  FileBase /Stub r@ read-file abort" read failed" drop
  r> close-file drop
  /Stub $2000 < abort" Stub too short"
;

: CopySyms	\ --
\ Copy symbols to their new position.
  $2000 InImg  $2000 /ForthImg + InImg  /Stub $2000 - move
;

: AddForth	\ --
\ Add the Forth image into the buffer
  ImgName r/w open-file abort" open failed"
  $2000 InImg /ForthImg 2 pick read-file abort" read failed" drop
  close-file drop
;

: findCmd?	\ cmd -- offset true | cmd 0
\ Find the offset of the first command of the given type.
  >r
  FileBase dup /mohdr +	 swap $14 + @	\ first command, lencmds
  begin					\ -- caddr len ; R: -- cmd
    dup 0>
   while
    over @ r@ = if
      drop FileBase -  true  r> drop  exit
    endif
    over cell + @ /string
  repeat
  2drop  r> 0
;

: findCmd	\ cmd -- offset
\ Find the offset of the first command of the given type. THROW
\ if not found.
  findCmd? 0= if
    cr .dword space  -1 abort" Can't find command"
  endif
;

: findDATA	\ --
  FileBase /Stub s" __DATA" search 0= abort" no DATA"
  drop  8 - FileBase - to offDATAseg	\ start is 8 bytes before string
;

: findLE	\ --
  FileBase /Stub s" __LINKEDIT" search 0= abort" no LINKEDIT"
  drop  8 - FileBase - to offLEseg	\ start is 8 bytes before string
;

((
pagezero:				; command 1
   dd 1              ; LC_SEGMENT
   dd _pagezero_     ; size
   db '__PAGEZERO'   ; segname
   times 6 db 0      ; pad to 16 chars
   dd 0              ; vmaddr   ; $18
   dd 0x1000         ; vmsize   ; $1C
   dd 0              ; fileoff  ; $20
   dd 0              ; filesize ; $24
   dd 0              ; maxprot  ; $28
   dd 0              ; initprot ; $2C
   dd 0              ; nsects   ; $30
   dd 0              ; flags    ; $34
_pagezero_ equ $-pagezero
Section
  sectname __data
   segname __DATA
      addr 0x0000202c
      size 0x0000006e
    offset 4140
     align 2^0 (1)
    reloff 0
    nreloc 0
     flags 0x00000000
 reserved1 0
 reserved2 0
Section
  sectname __common
   segname __DATA
      addr 0x0000209c	; $20
      size 0x00000010   ; $24
    offset 0		; $28
     align 2^2 (4)	; $2C
    reloff 0            ; $30
    nreloc 0            ; $34
     flags 0x00000001   ; $38
 reserved1 0            ; $3C
 reserved2 0            ; $40
dysymtab:				; command 7
   dd 0x0b           ; LC_DYSYMTAB	$00
   dd _dysymtab_     ; size		$04
   dd 0              ; ilocalsym	$08
   dd 0              ; nlocalsym	$0C
   dd 0              ; iextdefsym	$10
   dd 0              ; nextdefsym	$14
   dd 0              ; iundefsym	$18
   dd 4              ; nundefsym	$1C
   dd 0              ; tocoff		$20
   dd 0              ; ntoc		$24
   dd 0              ; modtaboff	$28
   dd 0              ; nmodtab		$2C
   dd 0              ; extrefsymoff	$30
   dd 0              ; nextrefsyms	$34
   dd indirect-$$    ; indirectsymoff	$38
   dd 4              ; nindirectsyms	$3C
   dd 0              ; extreloff	$40
   dd 0              ; nextrel		$44
   dd 0              ; locreloff	$48
   dd 0              ; nlocrel		$4C
_dysymtab_ equ $-dysymtab		$50
))

$22 constant LC_DYLD_INFO
$8000:0022 constant LC_DYLD_INFO_ONLY
$02 constant LC_SYMTAB
$0B constant LC_DYSYMTAB
$26 constant LC_FUNCTION_STARTS

: PatchCmds	\ --
\ Patch the headers
  findData findLE
  offDATAseg >r
  /ForthMem r@ $1C + InImg +!		\ extend segment vmsize
  /ForthImg r@ $24 + InImg +!		\ extend segment filesize
  7 r@ $28 + InImg !			\ maxprot=rwx
  7 r@ $2C + InImg !			\ initprot=rwx
  /ForthImg r@ /seg +  3 /sect * +  $24 + InImg +!	\ __data section size
  case  r@ $30 + InImg @		\ number of sections
cr dup . ." sections"
    4 of  endof				\ built on 10.6 Snow Leopard
    5 of				\ built on 10.7 Lion
      /ForthMem
      r@ /seg +  4 /sect * +  $20 +	\ __common section address
      InImg +!
    endof
      -1 abort" Unknown DATA segment"
  endcase
  r> drop
  offLEseg >r
  /ForthMem r@ $18 + InImg +!		\ new vm address
  /ForthImg r> $20 + InImg +!		\ new file offset
  LC_DYLD_INFO_ONLY findCmd >r
  /ForthImg r@ $10 + InImg +!		\ bind offset
  /ForthImg r@ $20 + InImg +!		\ lazy bind offset
  /ForthImg r> $28 + InImg +!		\ export offset
  LC_SYMTAB findCmd >r
  /ForthImg r@ $08 + InImg +!		\ symbol offset
  /ForthImg r> $10 + InImg +!		\ string offset
  LC_DYSYMTAB findCmd >r
  /ForthImg r> $038 + InImg +!		\ indirectsymoff
  LC_FUNCTION_STARTS findCmd? if
    >r /ForthImg r> $08 + InImg +!	\ dataoff
  else
    drop
  endif
;

: writeApp	\ --
  AppName r/w create-file abort" create failed" >r
  FileBase				\ start address
  /Stub /ForthImg +			\ size
  r@ write-file abort" write failed"
  r> close-file drop
  cr /Stub /ForthImg + . ." bytes saved to " AppName type cr
;

: Bind		\ --
  setImgData getBuff
  readStub
  CopySyms AddForth PatchCmds
  writeApp
  freeBuff
;


\ ======
\ *> ###
\ ======

decimal
