\ startup.fth - Start up code for Mach-o OSX application

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20110115 SFP004 Converted to OSX from Linux ELF version.
20110108 SFP003 Corrected OSX/BSD system calls.
20110103 SFP002 Conversion to Forth
20090112 JOR001 first working version
))

decimal

interpreter
((
: bin-align	\ n --
 \ *G Force alignment to an N byte boundary where N is a
 \ ** power of two. The space stepped over is set to 0.
  chere
  over 1- +  swap negate and		\ -- dp'
  here over over - erase		\ wipe gap
  corg
;
))
: bin-align	\ u --
\ *G Align on a *\i{u} byte boundary, where *\i{u} is a power
\ ** of two.
  here  over 1- +  swap negate and  org  ;
: z16,"		\ -- ; z16," foo bar"
\ *G Lay a zero terminated string in 16 byte slot.
  chere #16 + >r  z",  r> corg  ;
target

$0001:0000 equ _file0		\ memory address of first byte of file
$0001:3000 equ _start		\ file entry point
$0080:0000 equ TextSize		\ ask for 8 megabytes for my interpreter

cdata
here _file0 <> [if]
  cr primary section at wrong address" abort
[then]
l: StartAddress				\ first location in cross-compiled code

StartAddress PE-LOADBASE - equ LOADBIAS	\ offset from code base to start of code
LOADBIAS PE-CODERVA <> [if]		\ sanity check
  cr ." LOADBIAS <> PE-CODERVA - whoops!" abort
[then]
StartAddress _file0 <> [if]		\ sanity check
  cr ." StartAddress <> _FILE0 - whoops!" abort
[then]


\ *************
\ Mach-O header
\ *************
\ The header and command list are in the first segment declared with
\ data.

l: mhdr
   $FEED:FACE ,		\ magic
   7 ,			\ cputype
   3 ,			\ cpusubtype
   2 ,			\ filetype, executable
   #10 ,		\ ncmds, number of commands that follow
l: ^sizeofcmds
   ( sizeofcmds ) 0 ,	\ sizeofcmds, sans this header
   $85 ,		\ flags

\ Segments have to be aligned to page size (4096)

\ First page of our program space where we land by referencing a
\ null pointer. Note that no space is actually allocated.
l: firstcmd		\ commands start here

l: pagezero
   1 ,			\ LC_SEGMENT - command 0
   0 ,			\ size
   z16," __PAGEZERO"	\ segname
   0 ,			\ vmaddr
   PE-LOADBASE ,	\ vmsize
   0 ,			\ fileoff
   0 ,			\ filesize
   0 ,			\ maxprot
   0 ,			\ initprot
   0 ,			\ nsects
   0 ,			\ flags
here pagezero - pagezero cell + !


\ Commands and the header occupy part of the first
\ segment which happens to be text in our case.
\ Actual code (start) will go right after, although
\ there's nothing preventing us from wasting the rest
\ of the first text page and starting our code on the
\ next one.

\ This can be handy in the Forth cross-compiler
\ since the headers will likely be generated after
\ the kernel code. We can pre-allocate 3 pages
\ at the beginning of the file and use an offset
\ to make sure the kernel code goes after, starting
\ on a new page boundary.

\ Last but not least, segments and other load commands
\ don't have to go in this order. The first segment
\ can be a text segment.

\ =============================
\ Data segment with one section
\ =============================

l: dataseg
  1 ,			\ LC_SEGMENT - command 1
  0 ,			\ size, filled in after section 0
  z16," __DATA"		\ segname
  PE-LOADBASE ,		\ vmaddr
  $1000 ,		\ vmsize
  0 ,			\ fileoff
  $1000 ,		\ filesize
  7 ,			\ maxprot
  3 ,			\ initprot
  1 ,			\ nsects
  0 ,			\ flags

l: datasec0		\ section 0
  z16," __const"	\ sectname
  z16," __DATA"		\ segname
  PE-LOADBASE ,		\ addr
l: datasec0_ds
  ( datasize ) 0 ,	\ section size, filled in later
  0 ,			\ offset
  0 ,			\ align on 2^0
  0 ,			\ reloff
  0 ,			\ nreloc
  0 ,			\ flags
  0 ,			\ reserved1
  0 ,			\ reserved2
here dataseg - dataseg cell + !


\ ================================
\ Stubs, e.g. dlopen, dlclose, etc. - segment with one section
\ ================================

l: stubseg
  1 ,			\ LC_SEGMENT - command 2
  ( _stubs_ ) 0 ,	\ size
  z16," __IMPORT"	\ segname
  PE-LOADBASE $1000 + ,	\ vmaddr
  $1000 ,		\ vmsize
  $1000 ,		\ fileoff
  $1000 ,		\ filesize
  7 ,			\ maxprot
  7 ,			\ initprot
  1 ,			\ nsects
  0 ,			\ flags

l: stubs0		\ section 0
  z16," __jump_table"	\ sectname
  z16," __IMPORT"	\ segname
  PE-LOADBASE $1000 + ,	\ vmaddr
  20 ,			\ size, 4 5-byte stubs
  $1000 ,		\ offset
  6 ,			\ align on 2^6
  0 ,			\ reloff
  0 ,			\ nreloc
  $0400:0008 ,		\ flags
  0 ,			\ reserved1
  5 ,			\ reserved2, stub size
here stubseg - stubseg cell + !
\ _stubs_ equ $-stubs


\ =======================================
\ Data segment used by the dynamic linker
\ =======================================

l: linkage
  1 ,			\ LC_SEGMENT - command 3
  ( _linkage_ ) 0 ,	\ size
  z16," __LINKEDIT"	\ link table
  PE-LOADBASE $2000 + ,	\ vmaddr
  $1000 ,		\ vmsize
l: linkage_symoff/len
  ( symbols-$$ ) 0 ,	\ fileoff
  ( _symbols_ ) 0 ,	\ filesize
  7 ,			\ maxprot
  1 ,			\ initprot
  0 ,			\ nsects
  0 ,			\ flags
here linkage -  linkage cell + !
\ _linkage_ equ $-linkage


\ ============
\ Symbol table
\ ============

l: symtab
  2 ,			\ LC_SYMTAB - command 4
  ( _symtab_ ) 0 ,	\ size
  ( symbols-$$ ) 0 ,	\ symoff
  4 ,			\ nsyms
  ( strings-$$ ) 0 ,	\ stroff    +16
  ( _strings_ ) 0 ,	\ strsize   +20
here symtab -  symtab cell + !
\ _symtab_ equ $-symtab

\ Additional symbol table info used by the dynamic linker

l: dysymtab
  $0b ,			\ LC_DYSYMTAB - command 5
  ( _dysymtab_ ) 0 ,	\ size
  0 ,			\ ilocalsym
  0 ,			\ nlocalsym
  0 ,			\ iextdefsym
  0 ,			\ nextdefsym
  0 ,			\ iundefsym
  4 ,			\ nundefsym
  0 ,			\ tocoff
  0 ,			\ ntoc
  0 ,			\ modtaboff
  0 ,			\ nmodtab
  0 ,			\ extrefsymoff
  0 ,			\ nextrefsyms
l: ^indirectsymoff
  ( indirect-$$ ) 0 ,	\ indirectsymoff
  4 ,			\ nindirectsyms
  0 ,			\ extreloff
  0 ,			\ nextrel
  0 ,			\ locreloff
  0 ,			\ nlocrel
here dysymtab -  dysymtab cell + !
\ _dysymtab_ equ $-dysymtab

\ Our dynamic linker

l: dylinker
  $0e ,			\ LC_LOAD_DYLINKER - command 6
  ( _dylinker_ ) 0 ,	\ size
  #12 ,			\ nameoff, of the string below
l: InterpName
  z", /usr/lib/dyld"
  align
here dylinker -  dylinker cell + !
\ _dylinker_ equ $-dylinker


\ ===============================
\ Initial thread (register) state
\ ===============================

l: thrstate
  5 ,			\ LC_UNIXTHREAD - command 7
  ( _thrstate_ ) 0 ,	\ size
  $01 ,			\ i386_THREAD_STATE
  $10 ,			\ i386_THREAD_STATE_COUNT - 16 items
  #10 cells allot&erase
  _start ,		\ 11th cell is EIP
  #5 cells allot&erase
here thrstate -  thrstate cell + !
\ _thrstate_ equ $-thrstate


\ =======================================
\ Dynamic libraries that we will be using
\ =======================================

l: dylib
  $0c ,			\ LC_LOAD_DYLIB - command 8
  ( _dylib_ ) 0 ,	\ size
  $18 ,			\ nameoff
  $02 ,			\ timestamp
  $006F:0103 ,		\ currentver
  $0001:0000 ,		\ compatver
  z", /usr/lib/libSystem.B.dylib"
  align
here dylib -  dylib cell + !
\ _dylib_ equ $-dylib


\ =============================
\ Main text segment and section
\ =============================

l: textseg
  1 ,			\ LC_SEGMENT - command 9
  ( _text_ ) 0 ,	\ size
  z16," __TEXT"		\ segname
  PE-LOADBASE $3000 + ,	\ vmaddr ; textseg+0x18
l: textseg_ts
  TextSize ,		\ vmsize ; textseg
  $3000 ,		\ fileoff
l: textseg_fs
  ( codesize ) 0 ,	\ filesize
  7 ,			\ maxprot
  7 ,			\ initprot
  1 ,			\ nsects
  0 ,			\ flags

l: textsec0		\ text segment section 0
  z16," __text"		\ sectname
  z16," __TEXT"		\ segname
  _start ,		\ addr
l: textsec0_fs
  ( codesize ) 0 ,	\ size
  _start _file0 - ,	\ offset
  0 ,			\ align on 2^0
  0 ,			\ reloff
  0 ,			\ nreloc
  $8000:0400 ,		\ flags
  0 ,			\ reserved1
  0 ,			\ reserved2
here textseg -  textseg cell + !
\ _text_ equ $-text

here firstcmd -  ^sizeofcmds !
\ sizeofcmds equ $-pagezero


\ ===============
\ Additional data
\ ===============
\ Not aligned to page size so part of the same first
\ virtual memory page as the load commands above.
\ Second page if you count PAGEZERO.

#256 bin-align
align l: signon	\ sign on string issued by OSX system calls
  z", openlibmo exampleX"		\ X replaced by LF
  $0a here 2- c!
here 1- signon - equ l_signon

\ parameter strings and numbers for printf call to shared library
align l: libname
  z", /usr/lib/libSystem.B.dylib"
align l: printfname
  z", printf"
align l: formatstring
  z", Hello from %s, my number is %dX"
  $0a here 2- c!
align l: myname
  z", Stephen"
align l: mynum
  12345678 ,

\ here signon -  datasec0_ds !		\ size of data in section
here _file0 -  datasec0_ds !		\ size of data in section
\ datasize equ $-signon

  #4096 bin-align			\ step to next 4k page

\ *****
\ Stubs - offset $0000:1000
\ *****

asmcode
l: dlopen
  hlt  hlt  hlt  hlt  hlt
l: dlsym
  hlt  hlt  hlt  hlt  hlt
l: dlclose
  hlt  hlt  hlt  hlt  hlt
l: dlerror
  hlt  hlt  hlt  hlt  hlt
end-code

\ Some buffer space for the command line
\ This is necessary because the Forth interpreter expects its
\ command line as one long string with all arguments in it;
\ This is necessary because some arguments only make sense
\ in conjunction with others, and it would complicate the rest of the code
\ dramatically to try to iterate over argv[] from within the interpreter
  #256 bin-align
l: CmdLineBuffer
  #4096 bin-align			\ step to next 4k page


\ *******
\ Symbols - offset $0000:2000
\ *******

l: symbols

l: sym_dlopen
  ( offdlopen ) 0 ,	\ nstrx
  $01 c,		\ type N_EXT
  $00 c,		\ sect
  $0101 w,		\ desc
  dlopen ,		\ value

l: sym_dlsym
  ( offdlsym ) 0 ,	\ nstrx
  $01 c,		\ type N_EXT
  $00 c,		\ sect
  $0101 w,		\ desc
  dlsym ,		\ value

l: sym_dlclose
  ( offdlclose ) 0 ,	\ nstrx
  $01 c,		\ type N_EXT
  $00 c,		\ sect
  $0101 w,		\ desc
  dlclose ,		\ value

l: sym_dlerror
  ( offdlerror ) 0 ,	\ nstrx
  $01 c,		\ type N_EXT
  $00 c,		\ sect
  $0101 w,		\ desc
  dlerror ,		\ value

l: indirect	\ indirect symbol table
  0 ,  1 ,  2 ,  3 ,	\ dlopen, dlsym, dlclose, dlerror

indirect _file0 -  ^indirectsymoff !


\ Symbol String table

l: strings
  $20 c,  0 c,

here strings -  sym_dlopen !	\ set offset
  z", _dlopen"

here strings -  sym_dlsym !	\ set offset
  z", _dlsym"

here strings -  sym_dlclose !	\ set offset
  z", _dlclose"

here strings -  sym_dlerror !	\ set offset
  z", _dlerror"

symbols _file0 -  linkage_symoff/len !		\ offset of symbols
here symbols -    linkage_symoff/len cell+ !	\ length of symbols

symbols _file0 -  symtab 2 cells + !		\ offset of symbols
strings _file0 -  symtab 4 cells + !
here strings -    symtab 5 cells + !

  #4096 bin-align		\ step to next page


\ *********************************************
\ The OSX file header transfers control to here - offset $3000
\ *********************************************

((	\ Only required for testing
[undefined] tib-len [if]
  $0800 equ tib-len
[then]

[undefined] rp-size  [if]
  $10000 equ rp-size
[then]

[undefined] fs-size  [if]
  $1000 equ fs-size
[then]

[undefined] sp-size  [if]
  $10000 equ sp-size
[then]
))

prefix intel

\ By fixing the entry point we can make some parts of an interpreter
\ simpler.
proc _entry
   jmp     entry2
end-code
_entry _start <> [if]
  cr ." _ENTRY <> _START" abort
[then]

\ storage. N.B. OSX proves write permission in code segment
\ This first set of data is common to all VFX Forth hosted
\ implementations
	align
l: loadbase	\ holds base address of system		$0000:4008
	PE-LOADBASE ,
l: codesize	\ holds code size set by load time	$0000:400C
	TextSize PE-CODERVA - ,
l: newcodesize	\ holds requested code size by user	$0000:4010
	TextSize PE-CODERVA - ,
l: numrelocs	\ number of relocations, 0=none		$0000:4014
	0 ,
l: relocbase	\ OFFSET from ENTRY to reloc table	$0000:4018
	0 ,	\ 0 = no table
l: pDllExports	\ OFFSET from ENTRY to DLL/SO chain	$0000:401C
	0 ,	\ 0 = no table

\ The second set is operating system specific to OSX implementations
l: StackSize	\ requested stack size			$0000:4020
	1 Mb ,
l: StackTop	\ top of allocated stack+1		$0000:4024
	0 ,
l: StackBase	\ base of Forth stack			$0000:4028
	0 ,
l: EntryESP	\ ESP given by the loader		$0000:402C
	0 ,	\ used to find argc, argv, envc ....

proc GetRA		\ return ESI=return address
	mov	esi, 0 [esp]
	ret
end-code

proc RelocSystem	\ returns EBX=runtime PE-LOADBASE, EDX=len
\ perform the relocation, leaving ESI=runtime PE-LOADBASE
\ N.B. this code must be position independent and not itself subject to relocation.
	call	GetRA				\ ESI=RelocSystem+5
	sub     esi, # 5			\ ESI=RelocSystem
	sub	esi, # RelocSystem StartAddress -	\ ESI=StartAddress
	sub	esi, # LOADBIAS			\ ESI=LoadBase

	mov	ecx, numrelocs PE-LOADBASE - [esi]	\ ECX := number of relocations
	mov	ebx, relocbase PE-LOADBASE - [esi]	\ EBX := offset from StartAddress to relocation table
	add	ebx, # LOADBIAS			\ rebias from loadbase
	add	ebx, esi			\ add loadbase, make absolute

	mov	eax, esi			\ new load address
	sub	eax, loadbase PE-LOADBASE - [esi]	\ - old = relocation
\ ESI=runtime loadbase
\ EBX=start of relocation table
\ ECX=#relocations
\ EAX=relocation
l: relocloop
	jecxz	relocx
	mov	edi, 0 [ebx]			\ get offset from relocation table
	add	ebx, # 4			\ update table address
	add	LOADBIAS [esi] [edi], eax	\ relocate entry
	dec	ecx				\ count relocations down
	jmp	relocloop
l: relocx
\ set up the system after relocation, ESI=runtime LOADBASE
	mov	eax, # 0
	mov	numrelocs PE-LOADBASE - [esi], eax	\ zap number of relocations
	mov	relocbase PE-LOADBASE - [esi], eax	\ zap offset to reloction table
\ *** the line below should reference the VMSIZE of the TEXT segment
        mov     eax, textseg $1C + PE-LOADBASE - [esi]	\ get TEXT size CELL
	mov	dword ptr codesize , eax	\ save dictionary size
	mov	dword ptr newcodesize , eax	\ save dictionary size
	mov	ebx, esi			\ set return values
	mov	edx, eax
\ touch every 4 k from the base
	mov	ecx, eax			\ length
	shr	ecx, # #12			\ # 4k blocks
	test	eax, # $FFF			\ round
	nz, if,  inc  ecx  endif,
l: touchloop
	mov	eax, 0 [esi]			\ read
	add	esi, # $01000
	dec	ecx
	jne	touchloop
	ret
end-code

\ hex
align
l: cld1		\ address of word to run, will be patched later
  0 ,
l: serial_no	\ serial number
  0 ,
l: copyright	\ DO NOT CHANGE THIS STRING - See STAMPEXE
  ", &(c) MicroProcessor Engineering Limited"

\ TIB - terminal input buffer
align l: init-tib
init-tib equ init-tib		\ make cross compiler happy
  tib-len allot


\ ****************************************
\ table of initial values of the user area
\ ****************************************

l: user-reset
  0 ,                   \ s0
  0 ,                   \ r0
  init-tib ,		\ 'TIB
here user-reset - equ user-reset-len
l: init-fence
  0 ,			\ FENCE, patched by cross compiler
l: init-dp
  0 ,			\ DP, patched by cross compiler
l: init-voc-link
  0 ,			\ VOC-LINK, patched by cross compiler
\ l: SavedFS0
\   0 ,


\ ***********************************************************************
\ COLD start entry code. The stack value is saved for use by warm starts.
\ The stacks are allocated as frames on the Windows stack.
\ ***********************************************************************
\ STACKS GROW DOWN
\   .. USER-SIZE bytes ..
\ ESI user area
\ TIB terminal input buffer in callbacks
\   /FSP-guard
\ FS0/FSP
\   .. FP-SIZE bytes ..
\ EBP/S0 data stack
\   ... min SP-SIZE
\ ESP/R0 return stack
\   .. RP-SIZE bytes ..

$040 equ /RSPguard	\ 16 cell guard space at top of return stack
/RSPguard negate equ /-RSPguard
$040 equ /PSPguard	\ 16 cell guard space at top of data stack
/PSPguard negate equ /-PSPguard
$040 equ /FSPguard	\ 4*16 byte guard space at top of float stack
/FSPguard negate equ /-FSPguard

\ The required space used on the stack is:
RP-SIZE USER-SIZE + TIB-LEN + FS-SIZE + SP-SIZE +
#4095 + #12 >>n		\ number of 4k pages
  equ #T0pages

align
proc entry2
  call	RelocSystem			\ relocate system as required
\ Next line needs to occur after relocation
  mov	[] EntryESP , esp		\ preserve loader's stack

l: ecld
((
\ Touch each 4k of stack space.
\ May be NECESSARY if not all stack space is committed.
  mov   esi, esp	\ ESI = top of stack
  mov   ecx, # #T0pages	\ ECX = number of pages
l: ecldTouchup
  mov   edi, #-4092 [esi]
\  mov   #-4092 [esi], eax
  sub   esi, # #4096
  dec   ecx
  jne   ecldTouchup
))

  mov	ebx, esp			\ current stack pointer

\ User area
  sub   ebx, # USER-SIZE		\ point at base of USER area
  mov	esi, ebx
  mov   4 [esi], esp			\ set R0
\ Float stack
  lea   eax, /-FSPguard [ebx]		\ float stack guard space
  mov   $50 [esi], eax			\ FS0
  mov   $54 [esi], eax			\ FSP
  sub   ebx, # FS-SIZE
\ Data stack
  lea	ebp, /-PSPguard [ebx]		\ reserve guard space
  lea   eax, -4 [ebp]			\ set S0
  mov   0 [esi], eax
  mov   ecx, # user-reset		\ point at initial user-var values
  mov   0 [ecx], ebp			\ INIT-S0
  sub	ebx, # SP-SIZE
\ Return stack
  mov	eax, ebx
  sub   eax, # /RSPguard		\ crash protection
  mov	esp, eax
  mov   4 [ecx], esp			\ INIT-R0

l: ewrm
  mov   ecx, # user-reset		\ point at initial user-var values
  mov   ebp, 0 [ecx]			\ INIT-S0
  add   ebp, # 4
  mov   esp, 4 [ecx]			\ INIT-R0

  mov	eax, [] cld1			\ initial word to run
  cld					\ force direction flag to increment
  call  eax				\ run it
  jmp   ewrm				\ whoops! - should never happen
end-code

((
\ INT $080 type routine, EBX=handle ECX=caddr EDX=len
proc primtype
  push    edx			\ len
  push    ecx			\ addr
  push    # 1			\ STDOUT
  sub     esp, # 4		\ extra space needed!
  mov     eax, # 4		\ write call
  int     $80
  add     esp, # #16
  ret
end-code

\ emit single character, EAX=char
proc primemit
   push    eax
   mov     ecx, esp
   mov     edx, # 1
   call    primtype
   pop     eax
   ret
 end-code

\ Display EBX as 8 digit hex number
proc showhex
   push    eax
   push    ebx

   mov     ecx, # 8
l: shx1
   push    ecx
   push    ebx
   mov     eax, ebx
   shr     eax, # #28
   cmp     eax, # 9
   jle     shx2
   add     eax, # 7
l: shx2
   add     eax, # $30
   call    primemit
   pop     ebx
   pop     ecx
   shl     ebx, # 4
   sub     ecx, # 1
   jne     shx1
   mov     eax, # $20
   call    primemit

   pop     ebx
   pop     eax
   ret
end-code
))

\ -----------------------------------------------------------------------------

decimal
external

here constant FirstName		\ -- addr ; marks boundary for ip>nfa etc
\ Must be above 64k for VFX code generator's code-len field to be good.

code LoadLibrary      \ z-addr -- handle|0
\ save Forth registers
  	push	edi
	push	esi
\ preserve EBX and ESP on extended data stack
	lea	ebp, -8 [ebp]		\ reserve two cells
	mov	0 [ebp], esp		\ old ESP
	mov	4 [ebp], ebx		\ old TOS
\ align return stack frame. Call has two parameters
	lea	eax, -8 [esp]		\ need two cells
	and	eax, # -16		\ drop to 16 byte boundary
	lea	esp, 0 [eax]		\ set ESP
\ set parms, call, get value
	mov	0 [esp], ebx		\ z-addr
	mov	4 [esp], # 1		\ RTLD_LAZY
	call	dlopen			\ make call
	mov     ebx, eax		\ return value in EAX, move to TOS
\ clean stacks
	mov	esp, 0 [ebp]		\ restore return stack
	lea	ebp, 8 [ebp]		\ restore data stack
\ restore regs
	pop	esi
	pop	edi
	next,
end-code

code FreeLibrary      \ addr -- status
\ save Forth registers
  	push	edi
	push	esi
\ preserve EBX and ESP on extended data stack
	lea	ebp, -8 [ebp]		\ reserve two cells
	mov	0 [ebp], esp		\ old ESP
	mov	4 [ebp], ebx		\ old TOS
\ align return stack frame. Call has one parameter
	lea	eax, -4 [esp]		\ need one cell
	and	eax, # -16		\ drop to 16 byte boundary
	lea	esp, 0 [eax]		\ set ESP
\ set parms, call, get value
	mov	0 [esp], ebx		\ addr
	call	dlclose			\ make call
	mov     ebx, eax		\ return value in EAX, move to TOS
\ clean stacks
	mov	esp, 0 [ebp]		\ restore return stack
	lea	ebp, 8 [ebp]		\ restore data stack
\ restore regs
	pop	esi
	pop	edi
	next,
end-code

code GetProcAddress   \ handle z-addr -- abs32|0
\ save Forth registers
  	push	edi
	push	esi
\ preserve EBX and ESP on extended data stack
	lea	ebp, -8 [ebp]		\ reserve two cells
	mov	0 [ebp], esp		\ old ESP
	mov	4 [ebp], ebx		\ old TOS
\ align return stack frame. Call has two parameters
	lea	eax, -8 [esp]		\ need two cells
	and	eax, # -16		\ drop to 16 byte boundary
	lea	esp, 0 [eax]		\ set ESP
\ set parms, call, get value
	mov	4 [esp], ebx		\ z-addr
	mov	eax, 8 [ebp]		\ handle
	mov	0 [esp], eax
	call	dlsym			\ make call
	mov     ebx, eax		\ return value in EAX, move to TOS
\ clean stacks
	mov	esp, 0 [ebp]		\ restore return stack
	lea	ebp, #12 [ebp]		\ restore data stack, 8 + 4 for Forth
\ restore regs
	pop	esi
	pop	edi
	next,
end-code

code RelocateSystem	\ --
\ Relocate the Forth system - can be called by shared library
\ load or initialisation routines.
	push	esi
	push	edi
	push	ebx
	call	RelocSystem
	pop	ebx
	pop	edi
	pop	esi
	next,
end-code

code nxcall      \ xn..x1 addr n -- rc
\ Call the external function at addr with n arguments, where x1
\ will be the top of the calling stack.
\ preserve Forth state
        push    esi                     \ ESI = UP
        push    edi			\ EDI = LP
\ make calling frame
	shl	ebx, # 2		\ cells to bytes
	sub	ebp, # 8		\ two more items on data stack
\ 12 [ebp  x1, first argument
\ 8 [ebp]  call address
\ 4 [ebp]  #bytes of args
\ 0 [ebp]  ESP value
	mov	4 [ebp], ebx		\ #bytes of args on data stack
	mov	0 [ebp], esp		\ current return stack to be restored
	sub	esp, ebx		\ make call frame
	and	esp, # -16		\ align stack
\ copy data from Forth EBP stack to calling ESP frame
	test	ebx, ebx
	nz, if,
	  lea     esi, #12 [ebp]	\ source address
	  lea	  edi, 0 [esp]		\ dest address
	  begin,
	    mov     eax, 0 [esi]	\ copy data, increment pointers
	    lea     esi, 4 [esi]
	    mov     0 [edi], eax
	    lea     edi, 4 [edi]
	    sub     ebx, # 4
	  z, until,
	endif,
\ call external function and restore stacks
        cld				\ force D bit just in case
	call	8 [ebp]
	cld
	mov	esp, 0 [ebp]
\ Clean arguments from Forth data stack
	mov	ebx, 4 [ebp]		\ number of bytes of args
	lea	ebp, #12 [ebp] [ebx]	\ remove transients, addr, args
\ restore Forth and return result
	mov	ebx, eax		\ return value
        pop     edi
        pop     esi
        next,
end-code

code nxcall64      \ xn..x1 addr n -- drc
\ Call the external function at addr with n arguments, where x1
\ will be the top of the calling stack. Return a double integer
\ in Forth order.

\ preserve Forth state
        push    esi                     \ ESI = UP
        push    edi			\ EDI = LP
\ make calling frame
	shl	ebx, # 2		\ cells to bytes
	sub	ebp, # 8		\ two more items on data stack
\ 12 [ebp  x1, first argument
\ 8 [ebp]  call address
\ 4 [ebp]  #bytes of args
\ 0 [ebp]  ESP value
	mov	4 [ebp], ebx		\ #bytes of args on data stack
	mov	0 [ebp], esp		\ current return stack to be restored
	sub	esp, ebx		\ make call frame
	and	esp, # -16		\ align stack
\ copy data from Forth EBP stack to calling ESP frame
	test	ebx, ebx
	nz, if,
	  lea     esi, #12 [ebp]	\ source address
	  lea	  edi, 0 [esp]		\ dest address
	  begin,
	    mov     eax, 0 [esi]	\ copy data, increment pointers
	    lea     esi, 4 [esi]
	    mov     0 [edi], eax
	    lea     edi, 4 [edi]
	    sub     ebx, # 4
	  z, until,
	endif,
\ call external function and restore stacks
        cld				\ force D bit just in case
	call	8 [ebp]
	cld
	mov	esp, 0 [ebp]
\ Clean arguments from Forth data stack
	mov	ebx, 4 [ebp]		\ number of bytes of args
	lea	ebp, #8 [ebp] [ebx]	\ remove transients, addr, args-1
\ restore Forth and return result
	mov	0 [ebp], eax		\ return low value
	mov	ebx, edx		\ return high value
        pop     edi
        pop     esi
        next,
end-code

code breakpoint
        int     3
        next,
end-code

: breakpoint    \ --
  $8B c, $C0 c, $CC c,                 \ lay MOV EAX, EAX ; INT 3
; immediate

cr ." This redefinition is OK!"

code int3	\ --
  mov  eax, eax				\ isolate from INT 3 padding bytes
  int  3
  next,
end-code

code reset-fpu
        finit
        next,
end-code

EntryESP constant EntryESP	\ -- addr
\ The address holds the value of ESP on entry. It is used for
\ argc and argv extraction.
CmdLineBuffer constant CmdLine	\ -- addr
\ Returns the address of a buffer in which the command line is
\ recreated from the ARGV list.


\ ===========
\ *! callback
\ *T Callback functions
\ ===========
\ *P The CALLBACK mechanism provides the facility to wrap Forth
\ ** definitions in code which is callable by Linux.


\ ----------------------------
\ *S Simple CALLBACK functions
\ ----------------------------

external
variable ip-default	\ -- addr
\ *G Holds the default value of *\fo{IP-HANDLE} that is set for
\ ** each CALLBACK entry.

variable op-default	\ -- addr
\ *G Holds the default value of *\fo{OP-HANDLE} that is set for
\ ** each CALLBACK entry.


\ SFP004...
\ *******************************************
\ Callback stack construction and destruction
\ *******************************************
\ Callbacks are preceded by the number of outputs and inputs
\  4   |  4  |  5		   | ...
\ #out | #in | call callback-entry | ... | call callback-exit | ret # #in*4 | ret for disassembler
\
\ STACKS GROW DOWN
\   .. USER-SIZE bytes ..
\ ESI user area
\ TIB terminal input buffer in callbacks
\   /FSP-guard
\ FS0/FSP
\   .. FP-SIZE bytes ..
\ EBP/S0 data stack
\   ... min SP-SIZE
\ ESP/R0 return stack
\   .. RP-SIZE bytes assumed
\ ESP is lowest pointer.

\ The required space used on the stack after the entry frame and
\ the offsets are here

USER-SIZE TIB-LEN + FS-SIZE + SP-SIZE + equ #CBdropped

TIB-LEN FS-SIZE + SP-SIZE + equ #esp_USER
FS-SIZE SP-SIZE + equ #esp_TIB
FS-SIZE SP-SIZE + equ #esp_FSPtop
SP-SIZE equ #esp_PSPtop


#CBdropped #4095 + #12 >>n equ #CBpages	\ number of 4k pages
#CBdropped #1024 + equ /AverageCB	\ guess memory used by callback
2 Mb  /AverageCB /  equ #MaxNestings	\ number of nestings


7 cells equ #CBsaved		\ number of items (inc CBRA) below parameters

#28 equ #CBF_pars
#24 equ #CBF_SYSRET
#20 equ #CBF_RA
#16 equ #CBF_ECX
#12 equ #CBF_EDI
  8 equ #CBF_EBP
  4 equ #CBF_EBX
  0 equ #CBF_ESI

intel prefix
code callback-entry
\ Preserve registers and required information
	                        \ Parameters 1..n
\ --- Start of Forth callback frame ---
  \ CALLBACK RETURN ADDRESS
  \ CALLBACK-ENTRY RETURN ADDRESS - MUST be left for CALLBACK-EXIT
  push  ecx			\ ecx
  push  edi			\ EDI
  push  ebp			\ EBP
  push  ebx			\ EBX
  push  esi			\ ESI
\ --- End of Forth callback frame ---

((
\ Normally commented out
\ Touch each 4k of stack space.
\ NECESSARY if not all stack space is committed.
  mov   esi, esp		\ ESI = top of stack before drop
  mov   ecx, # #CBpages		\ ECX = number of pages
l: touchup
  mov   edi, #-4092 [esi]	\ read
\  mov   #-4092 [esi], eax	\ write
  sub   esi, # #4096
  dec   ecx
  jne   touchup
))

  lea	esp, #CBdropped negate [esp]	\ drop stack
\ User area
  lea	esi, #esp_USER [esp]
\ TIB
  lea   edi, #esp_TIB [esp]		\ EDI = TIB
  mov   $2C [esi], edi			\ set 'TIB
\ Float stack, prevent crash into TIB
  lea   ecx, #esp_FSPtop /FSPguard - [esp]  \ float stack guard space
  mov   $50 [esi], ecx			\ FS0
  mov   $54 [esi], ecx			\ FSP
\ Data stack, below Float Stack
  lea	ebp, #esp_PSPtop /PSPguard - [esp] \ data stack guard space
  mov   0 [esi], ebp			\ set S0
\ Return stack, no protection as below Parameter/Data Stack
  mov   4 [esi], esp			\ set R0

\ Set USER variables
  mov   eax, # $0A
  mov   $18 [esi], eax		\ BASE is decimal
  xor   eax, eax
  mov	$60 [esi], eax		\ ThreadExit?
  mov	$64 [esi], eax		\ ThreadTCB
  mov	$68 [esi], eax		\ ThreadSync

  lea   eax, $FF0 [esi]		\ top of exception stack = UP+$0FF0
  mov   $FF4 [esi], eax		\ EXC0
  mov   $FF8 [esi], eax		\ EXCP

\ Set the default I/O pointers. Done this way so that it still
\ works with ATTACH/DETACH in a DLL before relocation.
  xchg	edx, esi		\ preserve UP
  call  getRA
l: cbe1				\ ESI points here
  xchg	edx, esi		\ EDX=cbe1, ESI=UP
  mov	eax, ip-default cbe1 - [edx]
  mov	$03C [esi], eax		\ IP-HANDLE
  mov	eax, op-default cbe1 - [edx]
  mov	$038 [esi], eax		\ OP-HANDLE

\ Get the return address and hence the callback address
  mov	eax, #CBdropped #CBF_RA + [esp]	\ return address from CALLBACK-ENTRY
  lea	ebx, #CBdropped #CBF_pars + [esp] \ points to lowest parameter
  push	eax			\ return from CALLBACK-ENTRY to following routine
\ Set up data stack, parameters in C order, EBX=^params, EBX=return
  mov   edx, -9 [eax]		\ EDX = NUM INPUT PARAMS
  cmp   edx, # 0
  je    pushparamx
  shl   edx, 2
  add   edx, ebx        	\ EDX points at last parameter, EBX at first
L: pushparam
  sub   ebp, 4
  mov   ecx, 0 [ebx]    	\ get input param
  add   ebx, 4
  mov   0 [ebp], ecx    	\ push param to data stack
  cmp   ebx, edx
  jne   pushparam
  mov   ebx, 0 [ebp]		\ pop TOS
  add   ebp, 4
l: pushparamx

  next,
end-code

code callback-exit
  pop   eax			\ save return address
  lea	esp, #CBdropped [esp]	\ restore ESP to before frame creation
  mov   esi, #CBF_RA [esp]	\ ESI = CALLBACK-ENTRY's return address
  mov   #CBF_RA [esp], eax	\ becomes new return address
  mov   esi, #-13 [esi]		\ number of output parameters

  mov   eax, ebx		\ 1st return parameter (EAX is not preserved
        			\ so this is valid even for void functions)
  cmp   esi, 2			\ check for 64 bit return value in EDX:EAX
  jne   onlyone
  mov	edx, ebx		\ EDX=most significant
  mov   eax, 0 [ebp]		\ EAX=least significant
l: onlyone

  \ restore saved regs
  pop   esi
  pop   ebx
  pop   ebp
  pop   edi
  pop   ecx
  ret				\ following RET # #in cells cleans stack
end-code
\ ...SFP004

\ ============================================================================

: callback-crash
  cr ." Undefined callback" bye
;

: set-callback	\ xt callback --
\ *G Make the xt be the action of the callback.
  SkipCall PatchCall
;

: callback,	\ #in #out -- address
\ *G Lay down a callback data structure. The first cell contains
\ ** the address of the entry point. The address of the data structure
\ ** is returned.
  here  dup >r  3 CELLS + ,	\ pointer to callback code
  ,  ,				\ #out #in ; SFP003
  ['] callback-entry doCall,	\ call to entry code
  ['] callback-crash doCall,	\ callback action word
  ['] callback-exit  doCall,	\ call to exit code
  $C3 c,			\ SFP003
  r>
;

target-only
: CALLBACK:	\ #in #out "<name>" -- ; -- a-addr
\ *G Create a callback function. *\i{#IN} and *\i{#OUT} refer to
\ ** the number of input and output parameters required for the
\ ** callback. When the definition *\fo{<name>}is executed it will
\ ** return the address of the callback function. For example
\ *C   2 1 CallBack: Foo
\ *P creates a callback named *\fo{Foo} with two inputs, and one output.
\ ** Executing *\fo{Foo} returns the entry point used by Linux. To use
\ ** it, pass *\fo{Foo} as the entry point required by Linux, e.g as the
\ ** address of a task action. *\fo{Foo} is built to use the "C" calling
\ ** convention.
\ *C   ' FooAction to-callback foo
\ *P Having defined an action for the callback, you can now use the
\ ** callback as if it was a C or assembler function called by the
\ ** operating system.
  create
    callback, drop
  does>
    @
;

: CallProc:	\ #in #out "<name>" -- ; -- entry
\ *G Create a callback function and start compilation of its
\ ** action. *\i{#IN} and *\i{#OUT} refer to
\ ** the number of input and output parameters required for the
\ ** callback. When the definition *\fo{<name>}is executed it will
\ ** return the entry point address of the callback function.
\ *E 4 1 CallProc: <name>  \ #in #out -- ; -- entry
\ ** \ Callback action ; x1 x2 x3 x4 -- op
\ **   ...
\ ** ;
\ ** <name>   \ returns entry point address
  create
    callback, >r :noname r> @ set-callback
  does>
    @
;
host&target

: to-callback   \ xt <"name"> --
\ *G Assign an XT as the action of a defined callback. This word is
\ ** state smart.
  ' execute			\ executing a callback returns its entry point
  state @ if
    [compile] literal  ['] set-callback doCall,
  else
    set-callback
  then
; immediate


\ ========================================
\ *S An example. Creating a signal handler
\ ========================================
\ *P A Linux signal handler has the prototype
\ *C   void sa_siginfo( int signum, siginfo_t * siginfo, ucontext_t * uc );
\ *P As far as Forth is concerned we need to execute a Forth word that
\ ** receives three parameters and returns none.
\ *C  (SigTrap)   \ signum *siginfo *ucontext --
\ *P The code fragment below achives this.
\ *E 3 0 callback: SigTrap  \ -- addr
\ ** \ executing SigTrap in Forth returns the C entry point.
\ **
\ ** : (SigTrap)  \ signum *siginfo *ucontext --
\ ** \ Action of SigTrap.
\ **   nip                                   \ discard siginfo
\ **   cr
\ **   cr ." Signal number " swap .sigName
\ **   uc.mcontext                           \ point at CPU context
\ **   cr ." at address " dup sc.EIP @ dup .dword
\ **      ." , probably in " ip>nfa .name
\ **   cr
\ **   ['] SigThrow swap sc.EIP !            \ force return to SigThrow
\ ** ;
\ ** assign (SigTrap) to-callback SigTrap

\ *P The callback entry code provides you with a default I/O device
\ ** and sets *\fo{BASE} to decimal. It does *\b{not} set up a
\ ** default *\fo{THROW} handler. If your callbacks must cope
\ ** with exceptions, you must provide a top-level *\fo{CATCH}
\ ** yourself.

\ ======
\ *> ###
\ ======

variable ExitCode	\ -- addr ; holds exit code returned to Windows
  0 ExitCode !

internal

: PRE-COLD
  0 ExitCode !
  cold
;
' pre-cold cld1 !	\ set up entry point code

external

defer (bye)	\ --


\ -----------------------------------------------------------------------------

decimal

