; VFX Forth stub loader.
; This file produces the Mach-O headers and commands used by VFX Forth
; for OSX. The file is extended with the Forth image by the file binder.fth
; after cross compilation of the kernel.
;
; compile with:
; /user/bin/nasm -O1 -f macho stubosx.asm
; /usr/bin/gcc -g0 -mdynamic-no-pic -no_pie -arch i386 -o stubosx stubosx.o
; strip -u stubosx                           (removes unnecessary symbols)
; otool -h -l stubosx >stubosx.txt
;

extern _dlopen, _dlsym, _dlclose, _dlerror

; data goes here
segment .data

szlibname:
   db '/usr/lib/libSystem.B.dylib', 0
szprintf:
       db "printf",0
formatstring:
   db      "Hello from %s, my number is %d",0x0a,0
myname:
   db      "Stephen",0
mynum:
   dd      12345678

pprintf:  ; holds address of printf
   dd      -1

; ***************************
; storage later used by Forth
; ***************************
align 4, db 0
pmagic:		; data here identifies 32 bit system
   dd      0xfeedface			; magic number as identifier
   dd      _magiclen_			; length of this block from pmagic
plib:     ; holds library handle
   dd      -1				; handle for libSystem.B.dylib
pdlopen:
   dd     _dlopen			; function address
pdlsym:
   dd     _dlsym			; function address
pdlclose:
   dd     _dlclose			; function address
pdlerror:
   dd     _dlerror			; function address
pMainESP:
   dd     -1				; ESP on entry to Main
pUnixESP:
   dd     -1                            ; ESP that points to argc/argv block
_magiclen_ equ $-pmagic


; ***********************
; Magic addesses in Forth
; ***********************

ForthBase equ 0x00003000		; Forth image start address
ForthIdent equ ForthBase		; holds counted string starting "VFX Forth"
ForthData  equ 0x00003100		; copy the data block at pmagic to here
					; good image contains 0xfeedface at ForthData
ForthEntry equ 0x00004000		; Forth entry point


; ********************
; MPE code starts here
; ********************
; Much of the code here is redundant, but is left here in case
; Apple change the rules again.

segment .text

global _main

_main:
   mov     eax, esp
   mov     [pMainESP], eax		; save current ESP
   add     eax, 4
   mov     [pUnixESP], eax		; where argc/argv block is (by inspection)
   mov     eax, [ForthData]		; get Forth ID
   cmp     eax, 0xfeedface		; check it
   je      IDgood
   mov     ebx, 43			; error code
   jmp     failed
IDgood:

; Open the system library
   and     esp, 0xFFFFFFF0	; 16 byte align down
   push    0			; padding for alignment
   push    0
   push    1                     ; RTLD_LAZY
   push    szlibname             ; "libSystem.B.dylib"
   call    [pdlopen]             ; open library
   add     esp, 16               ; remove params and padding
   mov     [plib], eax           ; EAX is library handle or error code
   mov     ebx, 1                ; return code
   cmp     eax, 0                ; test return value
   je  near    failed

; Copy the pmagic data block to the Forth image
   mov     ebx, pmagic
   mov     ecx, ForthData
   mov     edx, _magiclen_
copypm:
   mov     eax, [ebx]
   mov     [ecx], eax
   add     ebx, 4
   add     ecx, 4
   sub     edx, 4
   jne     copypm

; go run Forth
   mov     eax, ForthEntry
   jmp     eax

; Find printf
   push    0
   push    0
   push    szprintf              ; string "printf"
   push    dword [plib]          ; library handle
   call    _dlsym                ; find function address
   add     esp, 16               ; remove params and padding
   mov     [pprintf], eax        ; EAX = function address or error code
   mov     ebx, 2                ; return code
   cmp     eax, 0                ; test return value
   je      failed

; call printf
   push    0
   push    dword [mynum]         ; second paramter to format string
   push    myname                ; first parameter to format string
   push    formatstring          ; format string itself
   call    [pprintf]             ; print and expand format string
   add     esp, 16               ; remove params and padding

; close library
   push    0
   push    0
   push    0
   push    dword [plib]          ; library handle
   call    _dlclose
   add     esp, 16               ; remove params and padding
   mov     ebx, 3                ; error return code
   cmp     eax, 0                ; test result
   jne     failed

; back to OS X
   mov     ebx, 42               ; success return code
failed:                          ; exit code in EBX
   push    ebx                   ; push exit code
   mov     eax, 1                ; exit function number
   push    -1                    ; padding
   int     0x80                  ; OSX system call
   pop     eax                   ; remove padding
   ret
