\ syspatch.fth - OSX Operating system hooks for startup

((
Copyright (c) 2011, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20110613 SFP002 Fixed OS_SetEndOfFile
20110115 SFP001 OSX Conversion
))


also system definitions


\ **********************************
\ Resolving shared libraries and their functions
\ **********************************

\ SFP006...
create libsys$	\ -- caddr ; counted string
  256 allot&erase
  s" /usr/lib/libSystem.B.dylib" libsys$ place

: PleaseResolve	\ dll$ func$ -- xcall|0
\ *G Return address to call for a function in a DLL,
\ ** or 0 if not found.
  { | zname[ #256 ] -- }
  swap count zname[ zplace
  zname[ LoadLibrary          \ c-addr2 libptr --
  dup if
    swap count zname[ zplace
    zname[ GetProcAddress
  else
    nip
  then
;

: Resolve	\ dll$ func$ -- xcall
\ *G Return address to call for a function in a DLL.
\ ** A breakpoint occurs if the function fails.
  PleaseResolve dup 0=
  if  breakpoint  then
;

\ ...SFP006


\ ***************
\ VFX Support DLL or shared library
\ ***************

\ SFP007...

also Forth definitions
1 value InitSupport?	\ -- n
\ Returns true if the support DLL is to be loaded.
\ 0 value errno		\ -- addr ; SFP013
\ Returns the address of the libc variable errno.
previous definitions

\ internal

variable suppcall	\ -- addr ; holds address of FindConstant code
variable simcall	\ -- addr ; holds address of FindSimilar code
variable NumCall	\ -- addr ; holds address of NumConstants code
((
variable savecall	\ -- addr ; holds address of Save code
variable SaveExtCall	\ -- addr ; holds address of SaveExt call
variable SuppVerCall	\ -- addr ; holds address of VersionNum call
))
create SuppDLL$	\ -- addr
  ", vfxsupp.1.dylib"

: initVFXdll	\ --
  InitSupport? 0=
  if  suppcall off  exit  endif
  SuppDLL$ C" VFXSuppFindConstant"  PleaseResolve suppcall !
  SuppDLL$ C" VFXSuppFindSimilar"   PleaseResolve simcall !
  SuppDLL$ C" VFXSuppNumConstants"  PleaseResolve NumCall !
;

\ external
\ ...SFP007


\ ************
\ Standard I/O
\ ************

\ variable dll_SetConsoleMode
variable dll_WriteFile
variable dll_ReadFile
variable dll_GetSystemTime
variable dll_GetLocalTime
variable dll_time
variable dll_CreateFile
variable dll_CloseHandle
\ variable dll_GetFileSize
variable dll_SetFilePointer
variable dll_DeleteFile
variable dll_isatty
variable dll_poll
variable dll_ftruncate
variable dll_perror
variable dll_tcgetattr
variable dll_tcsetattr
variable dll_ioctl
variable dll_exit
variable dll_umask
variable dll_chmod
variable dll_syscall
variable dll_select
variable dll_nanosleep
variable dll_fsync	\ SFP010
variable dll_rename	\ SFP016

: (os_bin)      \ fam -- 'fam
;

: (os_r/o)      \ -- fam
  0
;

: (os_w/o)      \ -- fam
  1
;

: (os_r/w)	\ -- fam
  2
;

((
defer +instrument
  assign noop to-do +instrument
defer -instrument
  assign noop to-do -instrument
))

$200 equ O_CREAT
$400 equ O_TRUNC
$1b6 equ MODE_666	\ Octal 666
-1 equ INVALID_HANDLE_VALUE

: (OS_Create-File)      { c-addr u fam | zaddr[ #1024 ] -- fileid ior }
  c-addr u #1023 umin zaddr[ zplace	\ Convert name to 0 string
  MODE_666 fam O_CREAT or O_TRUNC or zaddr[	\ SFP002
  dll_CreateFile @ 3 nxcall
  dup INVALID_HANDLE_VALUE = if		\ set ior
    dup
  else
    0
  then
;

: (OS_Open-File)      { c-addr u fam | zaddr[ #1024 ] -- fileid ior }
\ +instrument
  c-addr u #1023 umin zaddr[ zplace                \ Convert name to 0 string
  fam zaddr[
  dll_CreateFile @ 2 nxcall
  dup INVALID_HANDLE_VALUE = if         \ set ior
    dup
  else
    0
  then
\ -instrument
;

: (OS_Close-File)       \ fileid -- ior
\ +instrument
  dll_CloseHandle @ 1 nxcall
\ -instrument
;

: (OS_Write-File)       { caddr u fileid -- ior }
  u caddr fileid
  dll_WriteFile @ 3 nxcall
  -1 =
;

: (OS_Read-File)	{ caddr u fileid -- u2 ior }
  u caddr fileid
  dll_ReadFile @ 3 nxcall
  dup -1 = if
    dup
  else
    0
  then
;

0 equ FILE_BEGIN	\ SEEK_SET in POSIX parlance
1 equ FILE_CURRENT	\ SEEK_CUR
2 equ FILE_END		\ SEEK_END

: (OS_file-position)    { fileid -- ud ior }
  FILE_CURRENT 0 0 fileid dll_SetFilePointer @ 4 nxcall64
  2dup -1 -1 d= if			\ fail
    2drop  0 0 -1
  else					\ success
    0
  then
;

: (OS_reposition-file)  { low high fileid -- ior }
  FILE_BEGIN high low fileid dll_SetFilePointer @ 4 nxcall64
  and 	\ -1 -1 means fail; combine the two cells
  -1 =
;

: (OS_file-end)	{ fileid -- ud ior }
  FILE_END 0 0 fileid dll_SetFilePointer @ 4 nxcall64
  2dup -1 -1 d= if			\ fail
    2drop  0 0 -1
  else					\ success
    0
  then
;

: (OS_file-size)	{ fileid | high low -- ud ior }
  fileid (OS_file-position) drop to high to low
  fileid (OS_file-end)
  low high fileid (OS_reposition-file) drop
;

: (OS_SetEndOfFile)	\ fileid -- ior ; SFP001 SFP002
  dup (OS_file-position) dup 0= if	\ -- fileid ud ior
    drop  swap rot dll_ftruncate @ 3 nxcall
  else
    nip nip nip				\ Remove fileid and dummy position from stack
  then
;

: (OS_delete-file)	{ c-addr u | zaddr[ #1024 ] -- ior }
  c-addr u zaddr[ zplace
  zaddr[ dll_DeleteFile @ 1 nxcall
;

: (OS_Rename-File)	\ ca-old lold ca-new lnew -- ior
  { | zold[ #1024 ] znew[ #1024 ] -- }
  #1024 2- min znew[ zplace
  #1024 2- min zold[ zplace
  znew[ zold[ dll_rename @ 2 nxcall
;

: (OS_GetUmask)	\ -- mode
  0 dll_umask @ 1 nxcall		\ Retrieve mask
  dup dll_umask @ 1 nxcall drop 	\ umask() is destructive; replace original mask
;

: (OS_Chmod)	{ c-addr u mode | zaddr[ #1024 ] -- ior }
  c-addr u zaddr[ zplace
  mode zaddr[ dll_chmod @ 2 nxcall
  0=
;

: (OS_Flush-File)	\ fileid -- ior ; SFP010
  dll_fsync @ 1 nxcall drop 0	\ always return ior=0
;


\ ************************
\ ALLOCATE FREE and RESIZE
\ ************************

  0 equ NewHeap?	\ -- n ; GlobalAlloc and friends

\ =======================
\ GlobalAlloc and friends
\ =======================

NewHeap? 0 = [if]

variable dll_GlobalAlloc
variable dll_GlobalFree
variable dll_GlobalReAlloc

: (os_allocate)	\ size -- a-addr ior
\ Returns a pointer to a fixed memory block, with zeroed contents
  1 dll_GlobalAlloc @ 2 nxcall
  ?dup if 0 else -1 -1 then
;

: (os_resize)	\ a-addr newsize -- 'a-addr ior
  swap dll_GlobalReAlloc @ 2 nxcall
  ?dup if 0 else -1 -1 then
;

: (os_free)	\ a-addr -- ior
  dup if
    dll_GlobalFree @ 1 nxcall drop 0
  then
;

: InitVfxHeap	\ -- ; initialise heap operations
  libsys$ C" calloc"             Resolve dll_GlobalAlloc !
  libsys$ C" realloc"            Resolve dll_GlobalReAlloc !
  libsys$ C" free"               Resolve dll_GlobalFree !
;

[then]


\ ***********************
\ Standard in and friends
\ ***********************

variable stdin	\ -- addr
\ *G Holds file handle used for UNIX-style input.
\ ** Leave as a *\fo{VARIABLE} for compatibility between
\ ** VFX Forth implementations.
variable stdout	\ -- addr
\ *G Holds file handle used for UNIX-style output.
\ ** Leave as a *\fo{VARIABLE} for compatibility between
\ ** VFX Forth implementations.
variable stderr	\ -- addr
\ *G Holds file handle used for UNIX-style error output.
\ ** Leave as a *\fo{VARIABLE} for compatibility between
\ ** VFX Forth implementations.
variable stddev	\ -- addr
\ *G Holds console type: 1=file, 2=char, 3=pipe, other=???

variable TermPrepped	\ -- addr
\ *G Set non-zero when the terminal functions have been set up.
variable StdinUnbuffered \ -- addr
\ *G Flags whether or not stdin has been explicitly unbuffered
variable OldLFlag	\ -- addr
\ *G Holds old valuf of c_lflag from termios structure if
\ ** stdin has been unbuffered

#20 equ NCCS
struct termios 	\ -- size
\ *S A structure to mimic the termios structure used by tcgetattr and tcsetattr
\ *[
  4 field termios.c_iflag	\ input mode flags
  4 field termios.c_oflag	\ output mode flags
  4 field termios.c_cflag	\ control mode flags
  4 field termios.c_lflag	\ local mode flags
  NCCS field termios.c_cc	\ control characters
  4 field termios.c_ispeed	\ input speed
  4 field termios.c_ospeed	\ output speed
end-struct
\ *]

$000 equ TCSANOW
$100 equ ICANON
$008 equ IECHO			\ ECHO in OSX terms
ICANON not equ ^ICANON
IECHO not equ ^IECHO

: (ReadTermios)	\ addr -- addr flag
  dup  stdin @ dll_tcgetattr @ 2 nxcall 0=
;

: (WriteTermios) \ addr --
  TCSANOW stdin @ dll_tcsetattr @ 3 nxcall drop
;

: (SetUnbufferedStdin) { | termios_struct[ termios ] -- }
  termios_struct[ (ReadTermios) if 	\ If we managed to get attributes...
    dup termios.c_lflag dup @ OldLFlag !
    ^ICANON ^IECHO and
    swap and! (WriteTermios)
    1 StdinUnbuffered !
  else	\ Don't know what the attributes are - don't try to change them
    drop
  then
;

: (RebufferStdin) { | termios_struct[ termios ] -- }
  StdinUnbuffered @ if
    termios_struct[ (ReadTermios) if
      dup termios.c_lflag
      OldLFlAG @ swap ! (WriteTermios)
      0 StdinUnbuffered !
    else
      drop
    then
  then
;

: prepTerminal	\ --
\ Set up the terminal.
  stdin @ dll_isatty @ 1 nxcall stddev !
  (SetUnbufferedStdin)
;

: ?PrepTerm	\ --
\ If not already done, set up the terminal.
  TermPrepped @ 0= if
    PrepTerminal  TermPrepped on
  then
;

: UnprepTerminal \ --
\ Restore the terminal.
  (RebufferStdin)
;

: ?UnprepTerm	\ --
\ Restore the terminal if set up.
  TermPrepped @ if
    UnprepTerminal  TermPrepped off
  then
;
' ?UnprepTerm AtExit		\ Cue up unbuffering action

: InitVfxConsole	\ -- ; SFP008
\ Initialise console I/O.
  libsys$ C" write"		Resolve dll_WriteFile !
  libsys$ C" read"		Resolve dll_ReadFile !
  libsys$ C" isatty"		Resolve dll_isatty !
  libsys$ C" poll"		Resolve dll_poll !
  libsys$ C" select"		Resolve dll_select !
  libsys$ C" tcgetattr"		Resolve dll_tcgetattr !
  libsys$ C" tcsetattr"		Resolve dll_tcsetattr !
  libsys$ C" ioctl"		Resolve dll_ioctl !
  libsys$ C" umask"		Resolve dll_umask !
  libsys$ C" chmod"		Resolve dll_chmod !
  \ The POSIX way of doing things...
  0 stdin !
  1 stdout !
  2 stderr !

  \ Default to "haven't touched it"
  0 TermPrepped !  0 StdinUnbuffered !  0 stddev !
\  stdin @ dll_isatty @ 1 nxcall stddev !
;

$4004:667f equ FIONREAD

: (OS_key?)	{ | nread[ cell ] -- flag }
  ?PrepTerm  nread[ off
  nread[ FIONREAD stdin @ dll_ioctl @ 3 nxcall -1 = if
    0					\ Error return from ioctl
  else
    nread[ @ 0<>
  then
;

: (OS_Key)	\ -- key ; SFP003
  { | iobuff[ cell ] -- char }
  ?PrepTerm  iobuff[ off
  1 iobuff[ stdin @ dll_ReadFile @ 3 nxcall drop
  iobuff[ c@
;

: (OS_Emit)     \ char -- ; SFP003
  { | iobuff[ cell ] -- }
  ?PrepTerm  iobuff[ off
  iobuff[ c!
  1 iobuff[ stdout @ dll_WriteFile @ 3 nxcall drop
;

: (OS_Type)     \ caddr len -- ; SFP015
  ?PrepTerm
  swap stdout @ dll_WriteFile @ 3 nxcall drop
;


\ --------------------------------------------------------------------------

#32 equ SPACE_CHAR

: (OS_CopyCommandLine)		\ --
\ *G Copy command line from argv[] into a single string buffer
  EntryESP @ dup @		\ -- *esp argc
  swap 4+ @			\ -- argc argv ; SFP002
  CmdLineBuffer			\ -- argc argv buff
  rot 0 do			\ -- argv buff
    over i 4 * + @ swap		\ -- argv argv[n] buff
    begin
      over c@			\ -- argv argv[n] buff char
      ?dup
    while
      case  dup
        $0A of  drop BL  endof
        $0D of  drop BL  endof
      endcase
      over c! 1+ swap 1+ swap
    repeat
    SPACE_CHAR over c! 1+ nip
  loop
  0 swap 1- c! drop
;

: (OS_GetLoadAddress)           \ -- a-addr
  StartAddress
;

: (OS_GetCommandLine)           \ -- z-addr
  CmdLineBuffer
;

: >tdw		\ addr -- seconds mins hours day month year dow
\ Extract data from a POSIX struct tm structure.
  dup @ swap			\ seconds
  dup 4 + @ swap		\ minutes
  dup 8 + @ swap		\ hours
  dup #12 + @ swap		\ day
  dup #16 + @ 1+ swap		\ month - 0-based for POSIX
  dup #20 + @ #1900 + swap	\ year - 1900-based for POSIX
  #24 + @			\ dow
;

: (OS_GetLocalTime)	\ -- seconds mins hours day month year dow
  { | st[ 12 cells ] timeval[ cell ] -- }
  timeval[ dll_time @ 1 nxcall drop
  st[ timeval[ dll_GetLocalTime @ 2 nxcall drop
  st[ >tdw
;

: (OS_GetSystemTime)	\ -- seconds mins hours day month year dow
  { | st[ 12 cells ] timeval[ cell ] -- }
  timeval[ dll_time @ 1 nxcall drop
  st[ timeval[ dll_GetSystemTime @ 2 nxcall drop
  st[ >tdw
;

: (OS_Exit)		\ -- ; primary exit routine
  ExitCode @ dll_exit @ 1 nxcall drop
;

struct timespec \ -- size
\ *S A structure to mimic the timespec structure for libc
\ *[
  4 field tv_sec
  4 field tv_nsec
end-struct
\ *]

: (OS_ms)	\ ms --
  { | tspec[ timespec ] -- }
  #1000 /mod				\ -- rem quotient
  tspec[ tv_sec !
  #1000000 * tspec[ tv_nsec !
  0 tspec[ dll_nanosleep @ 2 nxcall drop
;

\ --------------------------------------------------------------------------

: (syspatch)    \ --
\ *G Patch all xcall addresses for system hooks, configure console
\ ** ready for interpreter.
\ s" into (SysPatch)" systype .sysState
  initVFXdll				\ link to support DLL ; SFP007
  initVfxConsole			\ initialise base console ; SFP008
  InitVfxHeap
\ s" done initVFXheap" sysType SysCr
  libsys$ C" gmtime_r"           Resolve dll_GetSystemTime !
  libsys$ C" localtime_r"        Resolve dll_GetLocalTime !
  libsys$ C" time"               Resolve dll_time !
  libsys$ C" open"               Resolve dll_CreateFile !
  libsys$ C" close"              Resolve dll_CloseHandle !
\  libsys$ C" __fxstat"           Resolve dll_GetFileSize !
  libsys$ C" lseek"              Resolve dll_SetFilePointer !
  libsys$ C" ftruncate"          Resolve dll_ftruncate !
  libsys$ C" unlink"             Resolve dll_DeleteFile !
  libsys$ C" perror"             Resolve dll_perror !
  libsys$ C" exit"               Resolve dll_exit !
  libsys$ C" syscall"            Resolve dll_syscall !
  libsys$ C" nanosleep"          Resolve dll_nanosleep !
  libsys$ C" fsync"              Resolve dll_fsync !	\ SFP010
  libsys$ C" rename"             Resolve dll_rename !	\ SFP016
\ s" done Resolves" sysType SysCr

  (os_CopyCommandLine)
\ .sysState  s" out of (SysPatch)" systype
;
' (syspatch) AtCold

assign (OS_Key)                 to-do OS_Key
assign (OS_Key?)		to-do OS_Key?
assign (OS_Emit)                to-do OS_Emit
assign (OS_Type)                to-do OS_Type

assign (os_allocate)            to-do allocate
assign (os_resize)              to-do resize
assign (os_free)                to-do free

assign (OS_GetLoadAddress)      to-do OS_GetLoadAddress
assign (OS_GetCommandLine)      to-do OS_GetCommandLine
assign (OS_GetSystemTime)       to-do OS_GetSystemTime
assign (OS_GetLocalTime)        to-do OS_GetLocalTime

assign (os_bin)                 to-do os_bin
assign (os_r/o)                 to-do os_r/o
assign (os_w/o)                 to-do os_w/o
assign (os_r/w)			to-do os_r/w
assign (OS_Create-File)         to-do OS_Create-File
assign (OS_Open-File)           to-do OS_Open-File
assign (OS_Close-File)          to-do OS_Close-File
assign (OS_Write-File)          to-do OS_Write-File
assign (OS_Read-File)           to-do OS_Read-File
assign (OS_File-Size)           to-do OS_File-Size
assign (OS_File-Position)       to-do OS_File-Position
assign (OS_Reposition-file)     to-do OS_Reposition-file
assign (OS_SetEndOfFile)        to-do OS_SetEndOfFile		\ SFP001
assign (OS_Delete-File)         to-do OS_Delete-File
assign (OS_Flush-File)		to-do OS_FLush-File		\ SFP010
assign (OS_Rename-File)		to-do OS_Rename-File		\ SFP016

assign (OS_ms)			to-do ms
assign (OS_Exit)		to-do (bye)

((
\ These words were provided for testing various Unices
: SetInterp	\ caddr len --
  InterpName zplace
;

: SetLibSys	\ caddr len --
  libsys$ place
;
))

previous definitions

