\ GenLocals.fth - ANS, MPE and Forth200x Locals Support

((
Copyright (c) 1997-2008, 2009, 2010, 2011, 2012, 2014, 2018, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441Fcp

To do
=====

Change history
==============
20200621 SFP013 x86/x64/ARM conversion.
20180512 SFP012 Allowed for (LOCAL) name at HERE.
20141007 SFP011 Enhanced standards compliance by not putting
		LOCALVARS in the search order.
20120715 SFP010 Used WID-THREADS.
20120629 SFP009 Fixed set up for obnoxious CCSism.
20120131 SFP008 Made LOCALS| multi-line.
20111024 SFP007 Fixed spacing problem in LOCALS|.
20101014 SFP006 Fixed bug with :NONAME.
20101004 SFP005 Check for repeated locals definitions.
                Forth200x locals added.
20100329 SFP004 Added check for repeated bar indicator.
20090303 SFP003 Added floating point locals.
		Major rewrite.
20081110 SFP002 Restored ALIGNment in LBSIZE.
20080826 SFP001 Overhauled, previous change markers removed.
                GenLocals.fth is a merge of LOCALS.FTH and
                ExLocals.fth.
))


\ ===========
\ *! genlocals
\ *T Local variable support
\ ===========

only forth definitions  decimal

also environment definitions
 #16 constant #LOCALS		\ -- n ; max number of locals allowed
TRUE constant LOCALS		\ -- true ; wordset present
TRUE constant LOCALS-EXT	\ -- true ; extension wordset present
previous definitions


\ *************
\ +S Primitives
\ *************
\ +P These are unoptimised compilers. Optimised versions are
\ +* installed later. GenLocals.fth is compiled before the
\ +* optimiser. Code generators are therefore handled by DEFERred
\ +* words.

defer s_makelvs,	\ #iargs #frame -- patch-addr
\ Compiler for locals entry code
defer s_rellvs,		\ --
\ Compiler to release local arguments and temporaries
Defer s_lvarg,		\ offset --
\ Compiler for local arguments and temporaries
Defer s_lvarray,        \ offset --
\ Compiler for local arrays
defer PatchLocalSize	\ --
\ Set #locals to size defined by LVCOUNT.
Defer s_flvarg,		\ offset --
\ Compiler for FP local arguments and temporaries

\ vocabulary LOCALVARS	\ -- ; contains local variable names ; now in kernel
variable Patch-Address	\ -- addr ; holds address to patch later
((
variable LVCOUNT	\ -- addr ; offset of integer local arg/value
variable FLVCOUNT	\ -- addr ; offset of FP local arg/value

: hasLVs?	\ -- flag
\ Return true if current definition contains locals.
  lvcount @ flvcount @ or
;
))


: (s_rellvs,)	\ -- ; compile unoptimised release code
  compile rellvs
;
assign (s_rellvs,) to-do s_rellvs,

: (s_makelvs,)	\ #iargs #frame -- patch-addr
  compile makelvs
  here -rot                     \ needed when frame is adjusted at compile time
  swap 4/ w, 4/ w,              \ #iargs then #frame
;
assign (s_makelvs,) to-do s_makelvs,

: (s_lvarg,)    \ offset -- ; default behaviour of S_LVARG,
  OperatorType @off			\ any prefixed operator?
  case
    0 of  ['] lv@ doCall, ,  endof	\ lay function and offset
    1 of  ['] lv! doCall, ,  endof
    2 of  ['] lv  doCall, ,  endof
    3 of  ['] lv  doCall, ,   ['] incr doCall,  endof
    4 of  ['] lv  doCall, ,   ['] decr doCall,  endof
    5 of  ['] lv  doCall, ,   ['] +! doCall,  endof
    6 of  ['] lv  doCall, ,   ['] off doCall,  endof
    7 of  ['] negate doCall,  ['] lv  doCall, ,  ['] +! doCall,  endof
    8 of  drop  ['] cell doCall,  endof
    9 of  ['] lv  doCall, ,   ['] on doCall,  endof
                    bad-method
  endcase
;
assign (s_lvarg,) to-do s_lvarg,  \ until optimiser is installed

: (s_lvarray,)  \ offset -- ; compiler for local arrays
  ['] lv doCall, ,
;
assign (s_lvarray,) to-do s_lvarray,

: (PatchLocalSize)	\ -- ; set #locals to size defined by LVCOUNT
  lvcount @ negate cell- cell /		\ need #CELLS
  Patch-Address @ 2+ w!			\ LVCOUNT holds next free byte offset
;
assign (PatchLocalSize) to-do PatchLocalSize


\ ********
\ +S Tools
\ ********

also system definitions

CPU_i32 [if]
#12 constant /fpl	\ -- n ; allocated size of an FP local
[else]
#8 constant /fpl	\ -- n ; allocated size of an FP local
[then]

variable inTDP			\ true when in temp dictionary
variable TDPstart		\ what TDP was initialised to
variable TDPstartNow		\ what TDP was initialised to after extension
variable TDP                    \ temp dp for local use
variable OLDDP                  \ saved current dp
variable OLDLAST                \ saved last nfa
variable OLDCURRENT             \ saved current
variable LOCAL-OFFSET           \ offset used for local dictionary
  $8000 local-offset !		\   how far above dp to lay temp words
variable LOCAL-RESERVE          \ how much space for temp words
  $4000 local-reserve !
variable local-nests		\ local nesting level for quotations
  0 local-nests !
variable #iargs			\ number of single-cell integer arguments
variable #fargs			\ number of floating point arguments
variable temps?			\ true if local values required
variable localsDone?		\ true when all locals have been parsed
MAX_PATH buffer: localname	\ buffer for local arg/val name

create LOCAL-MARKER  ", $local$"

: START-LOCALS	\ -- ; SYSTEM
  dp @ TDPstart !			\ where temp dict starts
  dp @
   local-offset @ +			\ offset from primary dictionary
   local-nests @ local-reserve @ * +	\ allow for nesting
  dup tdp !  TDPstartNow !           			\ create new dp
\  also localvars
;

: FORGET-LOCALS	\ -- ; SYSTEM
  tdp off  lvcount off  flvcount off  TDPstart off
\  previous
  ['] localvars voc>wid	wid-threads erase	\ SFP010
;

: get-name	\ -- caddr len ; SYSTEM
\ Get next token, can go over a line.
  bl get-word count
;

also forth definitions

: >TEMP-DICT	\ -- ; FORTH
\ switch in temp dictionary
  dp @ olddp !                          \ first save current dp
  tdp @ dp !                            \ use temp dp
  last @ oldlast !                      \ save last
  current @ oldcurrent !                \ save vocabs
  ['] localvars voc>wid current !
  inTDP on
;

: >REAL-DICT	\ -- ; FORTH
\ switch out temp dictionary
  dp @ tdp !                            \ save temp dp
  olddp @ dp !                          \ use real current dp
  oldlast @ last !                      \ restore last
  oldcurrent @ current !                \ restore vocabs
  inTDP off
;

: inLocals?	\ addr -- flag
\ True if addr is in the temp dictionary
  TDPstart @  inTDP @
  if  dp @  else  tdp @  endif
  within
;

: DROP-LOCALS	\ -- ; FORTH
\ cleanup code for local stack vars
  lvcount @				\ only compile if we had lvars
  if  s_rellvs,  endif                  \ number of vars to lose
;

: CLEANUP-LOCALS	\ -- ; FORTH
\ cleanup code for semicolon
  lvcount @                             \ only compile if we had lvars
  if  forget-locals  endif              \ now lose local definitions
;

previous definitions


\ ==============
\ Creation tools
\ ==============

: $createLV	\ c-addr u -- ; SFP006
\ Do the CREATE portion of a local arg/value.
  <headerless> @ >r
  ($create) immediate
  r> <headerless> !
;

: LVAR		\ c-addr u --
\ build temp local var
\  2dup current @ search-wordlist if     \ see if name exists
\    ." Local variable "
\    CR LastNameFound .name ." is redefined "
\  then
  $createLV  lvcount @ ,  does> @ s_lvarg,  ;	\ SFP006

: LArray	\ caddr u --
  $createLV  lvcount @ ,  does> @ s_lvarray,  ;	\ SFP006

: #FLVAR	\ c-addr u offset --
\ build FP temp local var
  >r  $createLV  r> ,  does> @ s_flvarg,  ;	\ SFP006

: MakeLvar	\ c-addr u -- ; SYSTEM
\ build temp local. Used by LOCALS| and (LOCAL) and {.
\ Uses LVCOUNT
  >temp-dict  lvar  >real-dict  ;

: MakeLArray	\ caddr u -- ; SYSTEM
\ Build array local at given offset. Used by {.
  >temp-dict  Larray  >real-dict  ;

: Make#FLvar	\ c-addr u offset -- ; SYSTEM
\ Build temp FP local.  Needs offset because of different
\ mechanisms for args and temps.
\ cr ." FL " 2 pick 2 pick type space dup .
  >temp-dict  #flvar  >real-dict  ;


\ =============
\ Parsing tools
\ =============

: bar?          \ caddr len -- flag ; SYSTEM
\ Return true if the string is the bar character ($7C).
\ Only used by LOCALS|.
  s" |" str=  ;

: >LVend	\ --
\ Go to the end of the LV declaration.
  begin
    get-name  2dup S" }" str= -rot  s" :}" str= or
  until
;


: allocFLV	\ addr -- offset
\ Reserve space for an FP local using the given pointer.
  dup @ /fpl - dup rot !  ;

: clipLVC	\ --
\ At the end of the local declaration, if LVCOUNT is >0, force it to 0.
  lvcount @ 0>
  if  lvcount off  endif
;

vocabulary localargs	\ --
\ Holds words that are parsing elements when counting arguments.
vocabulary localgen	\ --
\ Holds words that are parsing elements when processing args and temps.

also localargs definitions
: F:	( -- )  parse-name 2drop  #fargs incr  ;
: 	( -- )	localsDone? on  ;
: 	( -- )	localsDone? on  ;
: |	( -- )	localsDone? on  ;
: ---	( -- )	localsDone? on  ;
: --	( -- )	localsDone? on  ;
: }	( -- )	localsDone? on  ;
: :}	( -- )	localsDone? on  ;
previous definitions

: setBar	\ -- ; SFP004
\ check for repeated bar indicator, set.
  temps? @ #-420 ?throw  temps? on  lvcount off  ;

also localgen definitions
: 	( -- )	setbar  ;		\ SFP004
: 	( -- )	setbar  ;		\ SFP004
: |	( -- )	setbar  ;		\ SFP004
: ---	( -- )	localsDone? on  clipLVC  >LVend  ;
: --	( -- )	localsDone? on  clipLVC  >LVend  ;
: }	( -- )	localsDone? on  clipLVC  ;
: :}	( -- )	localsDone? on  clipLVC  ;
: F:		\ --
  parse-name  temps? @
  if  lvcount else  flvcount  endif
  allocFLV Make#FLvar
;
previous definitions

: lbsize	\ -- u ; SYSTEM
\ Parse up to closing ']' and interpret string. ALIGN the result.
  state off
  [char] ] parse evaluate aligned
  state on
;

: CountArgs	\ --
\ Count the integer and FP arguments from '{' up to '|', '--' or '}'.
  save-input  localsDone? off
  begin
    get-name ['] localargs voc>wid search-wordlist
    if  execute  else  #iargs incr  endif
    localsDone? @
  until
  restore-input drop
;

: GenLV/LA	\ caddr len --
\ Create a named LVAR or array.
  2dup + 1- c@ [char] [ <> if		\ normal integer
    cell lvcount -!
\ cr ." LV " 2dup type space lvcount ?
    MakeLVAR  exit
  endif
  localname place localname count	\ -- caddr len ; avoids problems
  lbsize lvcount -!
\ cr ." LA " 2dup type space lvcount .
  MakeLArray
;

: GenLocals	\ --
\ Create the named arguments and locals. The structure is:
\   Fargs    12 bytes each
\   Iargs    4 bytes each
\   state    8 bytes
\   -----    offset=0, Locals Pointer (LP, EDI reg) points here
\   temps
  localsDone? off  temps? off
  #iargs @ 2+ cells dup lvcount !	\ TOS at +2 cells, NOS at +3 cells ...
  #fargs @ /fpl * + flvcount !
  begin
    get-name 2dup ['] localgen voc>wid search-wordlist
    if  nip nip execute  else  GenLV/LA  endif
    localsDone? @
  until
;

CPU_ARM [if]
: flushLocalsCache	\ --
\ Local names are kept outside the normal dictionary space.
\ This has to be flushed before use.
\ ." Start/End" TDPstartNow @ .dword space TDP @ .dword cr
  FlushCache  TDPstartNow @ TDP @ (flushcache)
;
[then]

previous definitions


\ *P For programming a hosted Forth with a GUI interface and for
\ ** other significant styles of programming, the ANS Forth
\ ** specification of local variables is inadequate. VFX Forth
\ ** and other modern Forth systems provide an alternative notation
\ ** with more functionality and better readability. A subset of
\ ** this notation became the basis of the Forth200x local
\ ** variables proposal. The ANS locals mechanism is supported
\ ** in VFX Forth for backwards compatibility.

\ ===========================
\ *S Extended locals notation
\ ===========================
( *P The MPE extended local syntax provides a number of significant     )
( ** benefits to the ANS standard.                                      )
\ *(
( *B Named inputs are in stack comment order rather than reverse to make)
( ** source more readable.                                              )
( *B The definition line can declare a number of true local             )
( ** variables for tempory data storage.                                )
( *B Ability to declare local arrays/buffers for structure definitions  )
( ** etc.                                                               )
\ *)

\ *P In this implementation, locals are allocated as a frame on
\ ** the return stack. Note that the word's return address is no
\ ** longer available.

( *P The following example shows a code extract from a WINPROC,         )
( ** there are the traditional 4 inputs, a local array storing a        )
( ** temporary structure and one output.                                )
( *E : WndProc    {: hWnd uMsg wParam lParam | clientrect[ RECT ] -- res :} )
( **   uMessage WM_SIZE =                                               )
( **   if                                                               )
( **     hWnd clientrect[ GetClientRect drop    \ Get client rect       )
( **     hWndChild @                            \ useto resize child    )
( **     #0                                                             )
( **     #0                                                             )
( **     clientrect[ RECT.right @                                       )
( **     clientrect[ RECT.bottom @                                      )
( **     TRUE MoveWindow drop                                           )
( **     0 exit                                                         )
( **   then                                                             )
( **                                                                    )
( **   ...... Other Messages ....                                       )
( **                                                                    )
( **   hWnd uMessage wParam lParam DefWindowProc      \ Msg default.    )
( ** ;                                                                  )

( *P The following syntax for named inputs and local variables is used. )
( *P The sequence:                                                      )
( *E {: ni1 ni2 ... | lv1 lv2 ... -- o1 o2 :}                           )
( *P defines named inputs, local variables, and outputs. The named      )
( ** inputs are automatically copied from the data stack on entry.      )
( ** Named inputs and local variables can be referenced by name within  )
( ** the word during compilation. The output names are dummies to allow )
( ** a complete stack comment to be generated.                          )
\ *(
\ *B The items between {: and | are named inputs.
\ *B The items between | and -- are local variables.
\ *B The items between -- and :} are outputs.
\ *)
\ *P For compatibility with previous implementations, { is
\ ** accepted in place of {: and } in place of :}. The change
\ ** to {: ... :} took place as a result of the Forth200x
\ ** standard.

\ *P Named inputs and locals return their values when referenced,
\ ** and must be preceded by *\fo{->} or *\fo{TO} to perform a
\ ** store, or by *\fo{ADDR} to return the address.

( *P Arrays may be defined in the form:                                 )
( *E arr[ n ]                                                           )
( *P Any name ending in the '[' character will be treated as an array,  )
( ** the expression up to the terminating ']' will be interpreted to    )
( ** provide the size of the array. Arrays only return their base       )
( ** address, all operators are ignored.                                )

( *P In the example below, a and b are named inputs, a+b and a*b are    )
( ** local variables, and arr[ is a 10 byte array.                      )
( *E : foo    {: a b | a+b a*b arr[ 10 ] -- :}                          )
( **   a b + -> a+b                                                     )
( **   a b * -> a*b                                                     )
( **   cr a+b .  a*b .                                                  )
( ** ;                                                                  )

\ *P Floating point arguments (inputs) and temporaries are
\ ** declared by placing *\fo{F:} before the name, but not for
\ ** arrays of floats, which should be declared as above.
\ ** Floating point locals use the CPU's native FP (80x87) stack,
\ ** and so are most suitable for use with the *\i{%lib%/x86/ndp387.fth}
\ ** floating point package. Floating point locals are stored in
\ ** the extended 80 bit (10 byte) format. This is the default
\ ** for the *\i{%lib%/x86/ndp387.fth} code. The default action of
\ ** an FP local is to return its value. The following operators
\ ** can be applied:
\ *(
\ *B none - return the value,
\ *B *\fo{TO} or *\fo{->} - store to the local,
\ *B *\fo{ADDR} - return the address of the data,
\ *B *\fo{ADD} or *\fo{+TO} - add to the value,
\ *B *\fo{SUB} or *\fo{-TO} - subtract from the value.
\ *)

\ *E : foo2 {: a f: f1 b f: f2 | f: f3 f: f4 c d e -- :}
\ **   ...
\ ** ;
\ *P The arguments *\i{a} and *\i{b} above are integer arguments
\ ** taken from the Forth data stack. The arguments *\i{f1} and
\ ** *\i{f2} are FP arguments taken from the floating point unit.
\ ** Local values *\i{f3} and *\i{f4} are FP locals and the others
\ ** are integer locals. An example of using FP locals follows:
\ *E : foo3	{: f: f1 | f: f2 f: f3 -- :}
\ **   0e0 -> f2  10e0 -> f3  ( noop )
\ **   f1 add f2  f1 sub f3  ( noop )
\ **   f2 f.  f3 f.
\ ** ;

also system
: {		\ --
\ *G The start of the traditional brace notation { ... }.
  lvcount @ #-421 ?throw		\ check for repeated definitions ; SFP005
  discard-sinline  <HeaderLess> @ >r
  0 #iargs !  0 #fargs !  FALSE temps? !  localsDone? off
  CountArgs  start-locals
\ lay the local entry code, we already have the number of arguments
  #iargs @ cells 0 s_makelvs, Patch-Address !
\ Now create the named args and temps
  GenLocals
  cell lvcount -!			\ because of FRADJUST - need next free
\ cr ." LVC " lvcount ?
  PatchLocalSize
  r> <HeaderLess> !
[ CPU_ARM ] [if]  FlushLocalsCache  [then]
; immediate
previous

: {:		\ --
\ *G The Forth200x name to start the extended local variable
\ ** notation. Use in the form:
\ *C {: ni1 ni2 ... | lv1 lv2 ... -- o1 o2 :}
  postpone {  ;  immediate


\ ************************
\ *S ANS local definitions
\ ************************
\ *P The ANS locals definitions are provided for use with ANS standard
\ ** compliant code. The ANS locals system offers limited functionality.

also system definitions
 #256 buffer: locname
previous definitions

also system

: (LOCAL)	\ Comp: c-addr u -- ; Exec: -- x
\ *G When executed during compilation, defines a local variable
\ ** whose name is given by c-addr/u. If u is zero, c-addr is ignored
\ ** and compilation of local variables is assumed to finish.
\ ** When the word containing the local variable executes, the local
\ ** variable is initialised from the stack. When the
\ ** local variable executes, its value is returned. The local variable
\ ** may be written to by preceding its name with *\fo{TO}.
\ ** The word *\fo{(LOCAL)} is intended for the construction of
\ ** user-defined local variable notations. It is only provided
\ ** for ANS compatibility.
  lvcount @ 0= if
    locname place			\ in case name is at HERE ; SFP012
    0 0 (s_makelvs,) Patch-Address !
    start-locals                        \ initialise locals
    2 cells lvcount !                   \ 2 cells used for frame management
    locname count			\ SFP012
  endif
  ?dup if                               \ OK not last
    makelvar                            \ build temp local variable
    cell lvcount +!                     \ bump for next
  else                                  \ is last local
    drop Patch-Address @
    lvcount @ 2 cells -                 \ discount frame management in patchup of MAKELVS
    cell u/ swap w!                     \ fix up two *words* after MAKELVS
  endif
;

-short-branches
: LOCALS|	\ "<name1> ... <namen> |" --
\ *G Create named local variables *\fo{<name1>} to *\fo{<namen>}.
\ ** At run time the stack effect is *\fo{( xn..x1 -- )}, such
\ ** that *\fo{<name1>} is initialised to *\fo{x1} and *\fo{<namen>}
\ ** is initialised to *\fo{xn}. Note that this means that the
\ ** order of declaration is the reverse of the order used in stack
\ ** comments! When referenced, a local variable returns its value.
\ ** To write to a local, precede its name with *\fo{TO}. All locals
\ ** created by *\fo{LOCALS|} are single-cell integers.
\ ** In the example below, a and b are named inputs.
\ *E : foo      \ a b --
\ **   locals| b a |
\ **   a b +  cr .
\ **   a b *  cr .
\ ** ;
  discard-sinline  <HeaderLess> @ >r
  lvcount @ if
    forget-locals  state off
    TRUE abort" previous LOCALS| not finished"
  then
  \ scan to count LVs
  save-input 				\ multi-line permitted ; SFP008
  0 begin				\ no arguments yet
    get-name bar? 0= 			\ check for final '|' ; SFP007 SFP008
   while
    cell+			        \ #bytes of arguments
  repeat
  >r  restore-input abort" LOCALS| ... | scan error"	\ SFP008
  \ create frame
  r> dup if				\ check for none defined ; SFP008
    0 s_makelvs, Patch-Address !	\ initialise the locals frame
    start-locals                        \ initialise locals
    2 cells lvcount !			\ 2 cells used for frame management, args start at +8
    begin				\ --
      get-name 2dup bar? 0=		\ check for '|' ; SFP007 SFP008
    while				\ -- c-addr len
      makelvar  cell lvcount +!		\ build local variable
    repeat				\ --
    2drop
  else					\ SFP008
    drop
  endif
  r> <Headerless> !
[ CPU_ARM ] [if]  FlushLocalsCache  [then]
;  IMMEDIATE
+short-branches
previous


\ ************************************
\ *S Local variable construction tools
\ ************************************

(( for DocGen
variable LVCOUNT	\ -- addr
\ *G Holds the offset in the frame for the next local integer variable.
))

also system

: ?init-locals	\ --
  lvcount @ 0= if
    0 #iargs !  0 #fargs !		\ SFP009
    start-locals
    cell negate lvcount !
    0 0 s_makelvs, Patch-Address !
  then
;

: FRADJUST      \ size -- offset
\ *G Adjust the size of the current local values frame. Used
\ ** by words that create additional local variables outside a
\ ** *\fo{LOCALS| ... |} or { ... } notation.
  negate lvcount +!			\ LVCOUNT points to first free cell
  PatchLocalSize
  lvcount @ cell +			\ return offset of new block
;

previous


\ ======
\ *> ###
\ ======

decimal

((
: foo1 { a b | c d e -- } noop a noop b + -> c noop ;
: foo2 { a f: f1 b f: f2 | f: f3 f: f4 c d e -- } noop noop ;
: foo3	{ f: f1 | f: f2 f: f3 -- }
  0e0 -> f2  0e0 -> f3  ( noop )
  f1 add f2  f1 sub f3  ( noop )
  f2 f.  f3 f.
;
: foo4 { f: f1 f: f2 } f1 f. f2 f. ;
: foo5	{	aaa	bbb	ccc | 	ddd	eee	fff	}
  aaa bbb ccc + + -> eee
  eee
;
))
