\ VfxForth.ctl - kernel build file

((
Copyright (c) 2001, 2002, 2003, 2004, 2006, 2008, 2011, 2012, 2013, 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20190613 SFP011 64 bit updates.
20190227 GFW010 Corrected filename case.
20120403 SFP009 ARM Linux preparation.
20110115 SFP008 OSX changes
20080205 VRW007 Linux changes.
20071128 SFP006 Added HOST_xxx and CPU_xxx equates.
20041130 SFP005 Added successful WDOSX target. Although a little
		restricted in some ways, WDosX is really easy to
		use.
20041129 SFP005 Added unsuccessful DOS32A target. This may be
		resurrected when we better understand the LX
		file format.
20011206 SFP004 Added 386WDM and SAMTWK targets - untested!
))


\ ============================
\ Check the target description
\ ============================

[undefined] Target_x64_Windows  [IF] FALSE equ Target_x64_Windows    [THEN]
[undefined] Target_x64_Linux    [IF] FALSE equ Target_x64_Linux      [THEN]
[undefined] Target_x64_OSX      [IF] FALSE equ Target_x64_OSX        [THEN]
[undefined] Target_aa64_Linux   [IF] FALSE equ Target_aa64_Linux     [THEN]

[undefined] Target_386_Windows  [IF] FALSE equ Target_386_Windows       [THEN]
[undefined] Target_386_Linux    [IF] FALSE equ Target_386_Linux         [THEN]
[undefined] Target_ARM_Linux    [IF] FALSE equ Target_ARM_Linux         [THEN]
[undefined] Target_386_OSX      [IF] FALSE equ Target_386_OSX           [THEN]
[undefined] Target_386_WDosX    [IF] FALSE equ Target_386_WDosX         [THEN]
[undefined] Target_386_Dos32a   [IF] FALSE equ Target_386_Dos32a        [THEN]
[undefined] Target_386_WDM      [IF] FALSE equ Target_386_WDM           [THEN]
[undefined] Target_SArm_Twk     [IF] FALSE equ Target_SARM_Twk          [THEN]

Target_x64_Windows
Target_x64_Linux OR
Target_aa64_Linux OR
Target_x64_OSX OR
Target_386_Windows OR
Target_386_Linux OR
Target_ARM_Linux OR
Target_386_OSX OR
Target_386_Dos32a OR
Target_386_WDosX OR
Target_386_WDM OR
Target_SArm_Twk OR
NOT [IF]
cr
cr .( ******************************************************* )
cr .( No valid target processor/host selected in VFXFORTH.CTL )
cr .( ******************************************************* )
cr
abort
[THEN]


[undefined] Host_Windows  [IF] FALSE equ Host_Windows [THEN]
[undefined] Host_DOS      [IF] FALSE equ Host_DOS [THEN]
[undefined] Host_Linux    [IF] FALSE equ Host_Linux [THEN]
[undefined] Host_OSX      [IF] FALSE equ Host_OSX [THEN]
[undefined] Host_Bare     [IF] FALSE equ Host_Bare [THEN]

Host_Windows  Host_DOS or  Host_Linux or  Host_OSX or  Host_bare or
0= [if]
cr
cr .( ****************************************** )
cr .( No valid host O/S selected in VFXFORTH.CTL )
cr .( ****************************************** )
cr
abort
[then]

[undefined] CPU_i32        [IF] FALSE equ CPU_i32 [THEN]
[undefined] CPU_x64        [IF] FALSE equ CPU_x64 [THEN]
[undefined] CPU_ARM        [IF] FALSE equ CPU_ARM [THEN]
[undefined] CPU_AA64       [IF] FALSE equ CPU_AA64 [THEN]
[undefined] CPU_ArmCortex  [IF] FALSE equ CPU_ArmCortex [THEN]

CPU_i32  CPU_x64 or  CPU_ARM or  CPU_AA64 or  CPU_ArmCortex or
0= [if]
cr
cr .( ************************************* )
cr .( No valid CPU selected in VFXFORTH.CTL )
cr .( ************************************* )
cr
abort
[then]


\ ========================

[undefined] User-size [if]
$2000     equ User-size                 \ size of each task user area and exception stack
$0800     equ Tib-len                   \ terminal i/p buffer length
$1000     equ FS-size                   \ size of float stack
\ large stacks
$1:0000 equ SP-size			\ size of each task data stack
$1:0000 equ RP-size			\ size of each task return stack
(( \ small stacks
$0:1000 equ SP-size			\ size of each task data stack
$0:1000 equ RP-size			\ size of each task return stack
))
[then]

\ Windows:
\ Increasing RP-SIZE too far may break the Windows stack allocation, which by
\ default is set to 2Mb in MAKE-EXE.C and VFXSUPP.C. If you get "early outs"
\ or crashes when RP-SIZE is increased, try upping the default values set by
\   ioh->SizeOfStackReserve         = 0x0200000;
\   ioh->SizeOfStackCommit          = 0x0200000;
\ in
\   void BuildDefaultImageOptionalHeader( void )

\ STACKS GROW DOWN
\   .. USER-SIZE bytes ..
\ ESI user area
\ TIB terminal input buffer in callbacks
\   /FSP-guard
\ FS0/FSP
\   .. FP-SIZE bytes ..
\ EBP/S0 data stack
\   ... min SP-SIZE
\ ESP/R0 return stack
\   .. RP-SIZE bytes assumed
\ ESP is lowest pointer.
\ USER-SIZE TIB-LEN + FS-SIZE + SP-SIZE + equ #CBdropped	\ in VfxForth.ctl

\ Is this necessary?
\ User-size equ per-task                  \ memory required per task


\ -------------------------
\ The bootstrap entry point
\ -------------------------

L: EntryPoint

only forth definitions
[forth]

INTERNAL

Target_x64_OSX [IF]
here equ beg-startup
  include x64Osx/startup64.fth	\ Startup code
here equ end-startup
here equ beg-reqdcode
  include x64Com/reqdcode.fth		\ x64 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_x64_Linux [IF]
here equ beg-startup
  include x64Lin/startup64.fth		\ Startup code
here equ end-startup
here equ beg-reqdcode
  include x64Com/reqdcode.fth		\ x64 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_x64_Windows [IF]
here equ beg-startup
  include x64Win/startup64.fth		\ x64 Windows Startup code
here equ end-startup
here equ beg-reqdcode
  include x64Com/reqdcode.fth		\ x64 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_386_Windows [IF]
hasP4
here equ beg-startup
  include 386win\startup.fth		\ Startup code
here equ end-startup
here equ beg-reqdcode
  include 386com\reqdcode.fth		\ 386 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_386_Linux [IF]
for386? [if] has386 [else] hasP4 [then]
here equ beg-startup
  include 386Lin/startup.fth		\ Startup code
here equ end-startup
here equ beg-reqdcode
  include 386Com/reqdcode.fth		\ 386 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_ARM_Linux [IF]
here equ beg-startup
  include ArmLin/startup.fth		\ Startup code
here equ end-startup
here equ beg-reqdcode
  include ArmCom/reqdcode.fth		\ Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_386_OSX [IF]
hasP4
here equ beg-startup
  include 386OSX/startup.Fth		\ Startup code
here equ end-startup
here equ beg-reqdcode
  include 386Com/REQDCODE.FTH		\ 386 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_386_Dos32a [IF]
here equ beg-startup
  include Dos32a\startup.fth		\ Startup code
here equ end-startup
here equ beg-reqdcode
  include 386com\reqdcode.fth		\ 386 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_386_WDosX [IF]
has386
here equ beg-startup
  include WDosX\startup.fth		\ Startup code
here equ end-startup
here equ beg-reqdcode
  include 386com\reqdcode.fth		\ 386 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_386_WDM [IF]
here equ beg-startup
  include 386wdm\startup.fth		\ Start Code
here equ end-startup
here equ beg-reqdcode
  include 386com\reqdcode.fth		\ 386 Kernel Code Prims
here equ end-reqdcode
[THEN]

Target_SARM_Twk [IF]
here equ beg-startup
  include samtwk\startup.fth		\ Start Code
here equ end-startup
here equ beg-reqdcode
  include samcom\reqdcode.fth
here equ end-reqdcode
[THEN]

EXTERNAL

Target_x64_OSX  Target_x64_Linux or  Target_x64_Windows or  [IF]
here equ beg-kernel
  include Common/kernel64.fth		\ Kernel
here equ end-kernel
[else]
here equ beg-kernel
  include Common/kernel.fth		\ Kernel
here equ end-kernel
[then]

Target_x64_OSX [IF]
  beg-startup end-startup sfile: x64Osx/startup64.fth %VFXPATH%/Kernel/x64Osx/startup64.fth"
  beg-reqdcode end-reqdcode sfile: x64Com/reqdcode.fth %VFXPATH%/Kernel/x64Com/reqdcode.fth"
[THEN]
Target_x64_Linux [IF]
  beg-startup end-startup sfile: x64Lin/startup64.fth %VFXPATH%/Kernel/x64Lin/startup64.fth"
  beg-reqdcode end-reqdcode sfile: x64Com/reqdcode.fth %VFXPATH%/Kernel/x64Com/reqdcode.fth"
[THEN]
Target_x64_Windows [IF]
  beg-startup end-startup sfile: x64Win\startup.fth %VFXPATH%\Kernel\x64Win\STARTUP.FTH"
  beg-reqdcode end-reqdcode sfile: x64Com\reqdcode.fth %VFXPATH%\Kernel\x64Com\reqdcode.fth"
[THEN]
Target_386_Windows [IF]
  beg-startup end-startup sfile: 386win\startup.fth %VFXPATH%\Kernel\386WIN\STARTUP.FTH"
  beg-reqdcode end-reqdcode sfile: 386com\reqdcode.fth %VFXPATH%\Kernel\386COM\REQDCODE.FTH"
[THEN]
Target_386_Linux [IF]
  beg-startup end-startup sfile: 386Lin/startup.fth %VFXPATH%/Kernel/386Lin/startup.fth"
  beg-reqdcode end-reqdcode sfile: 386Com/reqdcode.fth %VFXPATH%/Kernel/386Com/reqdcode.fth"
[THEN]
Target_ARM_Linux [IF]
  beg-startup end-startup sfile: ARMlin/startup.fth %VFXPATH%/Kernel/ARMlin/startup.fth"
  beg-reqdcode end-reqdcode sfile: ArmCom/REQDCODE.FTH %VFXPATH%/Kernel/ArmCom/REQDCODE.FTH"
[THEN]
Target_386_OSX [IF]
  beg-startup end-startup sfile: 386OSX/startup.fth %VFXPATH%/Kernel/386OSX/startup.fth"
  beg-reqdcode end-reqdcode sfile: 386Com/REQDCODE.FTH %VFXPATH%/Kernel/386Com/REQDCODE.FTH"
[THEN]
Target_386_Dos32a [IF]
  beg-startup end-startup sfile: Dos32a\startup.fth %VFXPATH%\Kernel\DOS32A\STARTUP.FTH"
  beg-reqdcode end-reqdcode sfile: 386com\reqdcode.fth %VFXPATH%\Kernel\386COM\REQDCODE.FTH"
[THEN]
Target_386_WDosX [IF]
  beg-startup end-startup sfile: WDosX\startup.fth %VFXPATH%\Kernel\WDOSX\STARTUP.FTH"
  beg-reqdcode end-reqdcode sfile: 386com\reqdcode.fth %VFXPATH%\Kernel\386COM\REQDCODE.FTH"
[THEN]
Target_386_WDM [IF]
  beg-startup end-startup sfile: 386wdm\startup.fth %VFXPATH%\Kernel\386WSM\STARTUP.FTH"
  beg-reqdcode end-reqdcode sfile: 386com\reqdcode.fth %VFXPATH%\Kernel\386COM\REQDCODE.FTH"
[THEN]
Target_SARM_Twk [IF]
  beg-startup end-startup sfile: SamTwk\startup.fth %VFXPATH%\Kernel\SAMTWK\STARTUP.FTH"
  beg-reqdcode end-reqdcode sfile: SamCom\reqdcode.fth %VFXPATH%\Kernel\SAMCOM\REQDCODE.FTH"
[THEN]


Target_x64_OSX  Target_x64_Linux or  Target_x64_Windows or [IF]
beg-kernel end-kernel sfile: Common/kernel64.fth %VFXPATH%/Kernel/Common/kernel64.fth"
[ELSE]
beg-kernel end-kernel sfile: Common/kernel.fth %VFXPATH%/Kernel/Common/kernel.fth"
[THEN]

\ here
\  include Common/METHODS.FTH		\ Methods
\ here sfile: Common/METHODS.FTH %VFXPATH%/Kernel/Common/METHODS.FTH"
\ here
\  include Common/VALUE.FTH		\ VALUE using Methods
\ here sfile: Common/VALUE.FTH %VFXPATH%/Kernel/Common/VALUE.FTH"
here
  include Common/GenLocals		\ Combined Locals
here sfile: Common/LOCAL.FTH %VFXPATH%/Kernel/Common/GenLocals.fth"

here
  include Common/environ.fth		\ Environment
here sfile: Common/environ.fth %VFXPATH%/Kernel/Common/environ.fth"
here
  include Common/generic.fth		\ Generic IO
here sfile: Common/generic.fth %VFXPATH%/Kernel/Common/generic.fth"
here
  include Common/build.fth		\ Build Numbering
here sfile: Common/build.fth %VFXPATH%/Kernel/Common/build.fth"
here
  include Common/version.fth		\ Version OP
here sfile: Common/version.fth %VFXPATH%/Kernel/Common/version.fth"

Target_386_Windows [IF]
here
  include 386win\syspatch.fth		\ Patch Kernel
here sfile: 386win\syspatch.fth %VFXPATH%\Kernel\386WIN\SYSPATCH.FTH"
here
  include 386win\pfwsupp.fth		\ Windows Constants
here sfile: 386win\pfwsupp.fth %VFXPATH%\Kernel\386WIN\PFWSUPP.FTH"
here
  include 386win\save-exe.fth		\ Binary Save
here sfile: 386win\save-exe.fth %VFXPATH%\Kernel\386WIN\SAVE-EXE.FTH"
[THEN]

Target_386_Linux [IF]
here
  include 386Lin/syspatch.fth		\ Patch Kernel
here sfile: 386Lin/syspatch.fth %VFXPATH%/Kernel/386Lin/syspatch.fth"
here
  include 386Lin/VfxSupp.fth		\ Linux Constants
here sfile: 386Lin/VfxSupp.fth %VFXPATH%/Kernel/386LIN/VfxSupp.fth"
here
  include 386Lin/Save-Elf.fth		\ SAVE and friends
here sfile: 386Lin/Save-Elf.fth %VFXPATH%/Kernel/386Lin/Save-Elf.fth"
[THEN]

Target_ARM_Linux [IF]
here
  include ArmLin/syspatch.fth		\ Patch Kernel
here sfile: ArmLin/syspatch.fth %VFXPATH%/Kernel/ArmLin/syspatch.fth"
here
  include ArmLin/VfxSupp.fth		\ Linux Constants
here sfile: ArmLin/VfxSupp.fth %VFXPATH%/Kernel/ArmLin/VfxSupp.fth"
here
  include ArmLin/Save-Elf.fth		\ SAVE and friends
here sfile: ArmLin/Save-Elf.fth %VFXPATH%/Kernel/ArmLin/Save-Elf.fth"
[THEN]

Target_386_OSX [IF]
here
  include 386OSX/syspatch.fth		\ Patch Kernel
here sfile: 386OSX/SYSPATCH.FTH %VFXPATH%/Kernel/386OSX/syspatch.fth"
here
  include 386OSX/VfxSupp.fth		\ OSX Constants
here sfile: 386OSX/VfxSupp.fth %VFXPATH%/Kernel/386OSX/VfxSupp.fth"
here
  include 386OSX/SaveOSX.fth		\ SAVE and friends
here sfile: 386OSX/SaveOSX.fth %VFXPATH%/Kernel/386OSX/SaveOSX.fth"
[THEN]

Target_386_Dos32a [IF]
here
  include Dos32a\syspatch.fth		\ Patch Kernel
here sfile: Dos32a\syspatch.fth %VFXPATH%\Kernel\DOS32A\SYSPATCH.FTH"
\ here
\   include Dos32a\save-exe.fth		\ SAVE and friends
\ here sfile: Dos32a\save-exe.fth %VFXPATH%\Kernel\Dos32a\SAVE-EXE.FTH"
[THEN]

Target_386_WDosX [IF]
here
  include WDosX\syspatch.fth		\ Patch Kernel
here sfile: WDosX\syspatch.fth %VFXPATH%\Kernel\WDosX\SYSPATCH.FTH"
here
  include WDosX\save-exe.fth		\ SAVE and friends
here sfile: WDosX\save-exe.fth %VFXPATH%\Kernel\WDosX\SAVE-EXE.FTH"
[THEN]

Target_386_WDM [IF]
here
  include 386wdm\syspatch.fth		\ Patch Kernel
here sfile: 386wdm\syspatch.fth %VFXPATH%\Kernel\386wdm\SYSPATCH.FTH"
here
  include 386wdm\save-exe.fth		\ SAVE and friends
here sfile: 386wdm\save-exe.fth %VFXPATH%\Kernel\386wdm\SAVE-EXE.FTH"
[THEN]

Target_x64_OSX [IF]
here
  include x64OSX/syspatch64.fth	\ Patch Kernel
here sfile: x64OSX/SYSPATCH64.FTH %VFXPATH%/Kernel/x64OSX/syspatch64.fth"
here
  include x64OSX/VfxSupp64.fth	\ OSX Constants
here sfile: x64OSX/VfxSupp64.fth %VFXPATH%/Kernel/x64OSX/VfxSupp64.fth"
here
  include x64OSX/SaveOSX64.fth	\ SAVE and friends
here sfile: x64OSX/SaveOSX64.fth %VFXPATH%/Kernel/x64OSX/SaveOSX64.fth"
[THEN]

Target_x64_Linux [IF]
here
  include x64Lin/syspatch64.fth	\ Patch Kernel
here sfile: x64Lin/SYSPATCH64.FTH %VFXPATH%/Kernel/x64Lin/syspatch64.fth"
here
  include x64Lin/VfxSupp64.fth	\ Linux Constants
here sfile: x64Lin/VfxSupp64.fth %VFXPATH%/Kernel/x64Lin/VfxSupp64.fth"
here
  include x64Lin/SaveElf64.fth	\ SAVE and friends
here sfile: x64Lin/SaveElf64.fth %VFXPATH%/Kernel/x64Lin/SaveElf64.fth"
[THEN]

Target_x64_Windows [IF]
here
  include x64Win/syspatch64.fth	\ Patch Kernel
here sfile: x64Win/syspatch64.fth %VFXPATH%/Kernel/x64Win/syspatch64.fth"
here
  include x64Win/VfxSupp64.fth	\ Windows Constants
here sfile: x64Win/VfxSupp64.fth %VFXPATH%/Kernel/x64Win/VfxSupp64.fth"
here
  include x64Win/SaveExe64.fth	\ SAVE and friends
here sfile: x64win/SaveExe64.fth %VFXPATH%/Kernel/x64Win/SaveExe64.fth"
[THEN]

Target_Sarm_Twk [IF]
here
  include samtwk\syspatch.fth		\ Patch Kernel
here sfile: samtwk\syspatch.fth %VFXPATH%\Kernel\SAMTWK\SYSPATCH.FTH"
here
  include samtwk\save-exe.fth		\ SAVE and friends
here sfile: samtwk\save-exe.fth %VFXPATH%\Kernel\SAMTWK\SAVE-EXE.FTH"
[THEN]

here
  include Common/console.fth		\ Default Console - may depend on SYSPATCH.FTH
here sfile: Common/console.fth %VFXPATH%/Kernel/Common/console.fth"
\  include Common/console.fth		\ Default Console - may depend on SYSPATCH.FTH
\ here sfile: Common/consolenew.fth %VFXPATH%/Kernel/Common/consolenew.fth"

Target_386_Windows	\ All Intel32 targets
Target_386_Linux or
Target_386_OSX or
Target_386_Dos32a or
Target_386_WDosX or
Target_386_WDM or
[IF]
here
  include 386Com/asm8687		\ Assembler
here sfile: 386Com/asm8687 %VFXPATH%/Kernel/386Com/asm8687.fth"
here here sfile: 386Com/OPTIMISE/BUILD %VFXPATH%/Kernel/386Com/OPTIMISE/BUILD.FTH"
  include 386Com/optimise/build		\ Optimiser
here end-sfile 386Com/optimise/build
here
  include 386Com/switches		\ Extensible switch chain
here sfile: 386Com/switches %VFXPATH%/Kernel/386Com/switches.fth"
here
  include 386Com/dasm386		\ 386 disassembler
here sfile: 386Com/dasm386 %VFXPATH%/Kernel/386Com/dasm386.fth"
here
  include 386Com/dasm387		\ 387 disassembler
here sfile: 386Com/dasm387 %VFXPATH%/Kernel/386Com/dasm387.fth"
here
  include 386Com/dis			\ decompiler
here sfile: 386Com/dis %VFXPATH%/Kernel/386Com/dis.fth"
[THEN]

Target_x64_Windows	\ All Intel64 targets
Target_x64_Linux or
Target_x64_OSX or
[IF]
here
  include x64Com/PreAsm
here sfile: x64/PreAsm %VFXPATH%/Kernel/x64Com/PreAsm.fth"
here
  include x64Com/hasmx64		\ Assembler
here sfile: x64/hasmx64 %VFXPATH%/Kernel/x64Com/hasmx64.fth"
here
  include x64Com/PostAsm
here sfile: x64Com/PostAsm %VFXPATH%/Kernel/x64Com/PostAsm.fth"
here here sfile: x64Com/OPTIMISE/BUILD %VFXPATH%/Kernel/x64Com/OPTIMISE/BUILD.FTH"
  include x64Com/optimise/build		\ Optimiser
here end-sfile x64Com/optimise/build
here
  include x64Com/switches		\ Extensible switch chain
here sfile: x64Com/switches %VFXPATH%/Kernel/x64Com/switches.fth"
here
  include x64Com/hdasmx64		\ 386 disassembler
here sfile: x64Com/hdasmx64 %VFXPATH%/Kernel/x64Com/hdasmx64.fth"
here
  include x64Com/hdasm387		\ 387 disassembler
here sfile: x64Com/hdasm387 %VFXPATH%/Kernel/x64Com/hdasm387.fth"
here
  include x64Com/hdasmSSE		\ 387 disassembler
here sfile: x64Com/hdasmSSE %VFXPATH%/Kernel/x64Com/hdasmSSE.fth"
here
  include x64Com/dis			\ decompiler
here sfile: x64Com/dis %VFXPATH%/Kernel/x64Com/dis.fth"
[THEN]


Target_ARM_Linux [if]	\ All ARM targets
here
  include ArmCom/PreAsm		\ Assembler
here sfile: ArmCom/PreAsm %VFXPATH%/Kernel/ArmCom/PreAsm.fth"
here
  include ArmCom/HAsmCortex		\ Assembler
here sfile: ArmCom/HAsmCortex %VFXPATH%/Kernel/ArmCom/HAsmCortex.fth"

here here sfile: ArmCom/optimise/build %VFXPATH%/Kernel/optimise/build.fth"
  include ArmCom/optimise/build		\ Optimiser
here end-sfile ArmCom/optimise/build

here
  include ArmCom/switches		\ Extensible switch chain
here sfile: ArmCom/switches %VFXPATH%/Kernel/ArmCom/switches.fth"

here
  include ArmCom/HDasmCortex		\ disassembler
here sfile: ArmCom/HDasmCortex %VFXPATH%/Kernel/ArmCom/HDasmCortex.fth"
[THEN]

Target_386_WDosX [IF]
here
  include wdosx\DosFuncs
here sfile: WDosX\DosFuncs.fth %VFXPATH%\Kernel\WDosX\DOSFUNCS.FTH"
[THEN]

Target_Sarm_Twk		\ All StrongARM targets
[IF]
[THEN]


\ -------------
\ end of kernel
\ -------------

        create copyright-date   ", 1998-2021 " align

Target_x64_Windows
Target_x64_Linux or
Target_x64_OSX or
Target_aa64_Linux or
[IF]
  create product-title    ", VFX Forth 64" align
[ELSE]
  create product-title    ", VFX Forth" align
[THEN]

Target_386_Windows [IF]
        create cpu_target       ", x86"
        create host_target      ", Windows"
        true constant Target_386_Windows
[THEN]

Target_386_Linux [IF]
        create cpu_target       ", x86"
        create host_target      ", Linux"
        true constant Target_386_Linux
	for386? constant for386?
[THEN]

Target_ARM_Linux [IF]
        create cpu_target       ", ARM"
        create host_target      ", Linux"
        true constant Target_ARM_Linux
[THEN]

Target_386_OSX [IF]
        create cpu_target       ", x86"
        create host_target      ", Mac OS X"
        true constant Target_386_OSX
[THEN]

Target_x64_OSX [IF]
        create cpu_target       ", x64"
        create host_target      ", Mac OS X"
        true constant Target_x64_OSX
[THEN]

Target_x64_Linux [IF]
        create cpu_target       ", x64"
        create host_target      ", Linux"
        true constant Target_x64_Linux
[THEN]

Target_x64_Windows [IF]
        create cpu_target       ", x64"
        create host_target      ", Windows"
        true constant Target_x64_Linux
[THEN]

Target_386_Dos32a [IF]
        create cpu_target       ", x86"
        create host_target      ", DOS"
        true constant Target_386_Dos32A
[THEN]

Target_386_WDosX [IF]
        create cpu_target       ", x86"
        create host_target      ", DOS"
        true constant Target_386_WDosX
[THEN]

Target_386_WDM [IF]
        create cpu_target       ", x86"
        create host_target      ", Win32 Driver Model"
        true constant Target_386_WDM
[THEN]

Target_SARM_Twk [IF]
        create cpu_target       ", StrongArm"
        create host_target      ", Embedded - TWK SA1100 module"
        true constant Target_SArm_Twk
[THEN]


: .CPU  \ --
  ." MPE ANS VFX Forth " cpu_target $.
;

