\ BUILD.FTH - build numbering system

((
Copyright (c) 2004
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20080506 SFP003 Added BUILD$, and time and date facilities.
20080502 SFP002 Added BUIDFILE for cross compiler compatibilty.
20040616 SFP001 Fixed MAKE-$, removed redundant $LENGTH.
))

decimal

\ ========
\ *> tools
\ *S Automatic build numbering
\ ========
\ *P The build numbering system allows you to generate a string
\ ** in the system which can be used for displaying version
\ ** information.

\ *P The system relies on a file (normally called *\i{BUILD.NO})
\ ** which holds the complete build version string. The string
\ ** can consist of any characters, e.g "Version 1.00.0034". The
\ ** contents of the file can be placed as a counted string in
\ ** the dictionary by *\fo{MAKE-BUILD}. After successful
\ ** compilation of your application, *\fo{UPDATE-BUILD} will
\ ** update the build number file by treating *\b{all} the digits
\ ** in the build string as a single number to be incremented.			)


\ =================
\ *N Internal words
\ =================
\ *P The following words are private to the build system, and can
\ ** be found in the *\fo{System} vocabulary.

( *D ^BUILD-BUFFER      Pointer to the application version buffer.      )
( *D BUILD.NO-SIZE      # bytes in the build number string.             )
\ ( *D $length            Length of counted string for buffer number      )
( *D CARRY              Carry to be added to next digit                 )
( *D TMP-BUILD-BUFFER   A tempory buffer used to build an updated no.   )
( *D SZBUILDFILE        String defining the build number file name      )
( *D *BUILDFILE         File pointer for currently in use build file    )

also system definitions

variable ^build-buffer
variable build.no-size
\ variable $length
variable carry

create tmp-build-buffer 100 allot
       tmp-build-buffer 100 erase

create szBuildFile  ", build.no"  #260 9 - allot
variable *buildfile


: Create-Build  \ --
\ *G Create a new build number file and store the file handle
\ ** in *\fo{*BUILDFILE}.
  szBuildFile count r/w Create-File swap *buildfile !
  abort" Can't create buld file"
;

: Open-Build    \ --
\ *G Open an existing build number file and store the file handle
\ ** in *\fo{*BUILDFILE}.
  szBuildFile count r/w Open-File swap *buildfile !
  abort" Can't open build file"
;

: Close-Build   \ --
\ *G Close the open build file.
  *buildfile @ Close-File drop
;

: Write-Build   \ Buffer len --
\ *G Write a buffer into the open build file.
  *buildfile @ write-file
  abort" Can't write build file"
;

: Read-Build    \ Buffer -- #bytes
\ *G Read contents of the open build file into a memory buffer.
  4096 *buildfile @ Read-File
  abort" Can't read build file"
;

: Build-Version \ addr --
\ *G Open a build file, read its contents and close it.
  open-build                    \ open file
  read-build                    \ read build no. string into buffer
  build.no-size !               \ set size of file
  close-build                   \ close file
;

: Digit?        \ char -- t/f
\ *G Check that a character is a decimal digit.
  [char] 0 [char] 9 within?
;

: ?Increment    \ addr --
\ *G Increment digit at ADDR as appropriate and store a carry flag.
  dup c@ carry @ +              \ add carry
  dup [char] 9 >                \ step to next column?
  If                            \ y:
    drop [char] 0 swap c!       \   wrap around to ascii 0 and set digit
    1 carry !                   \   set carry for next digit
  Else                          \ n:
    swap c!                     \   set this new ascii digit
    0 carry !                   \   no carry for next digit
  Endif
;

: ?Incr-Char    \ addr --
\ *G Increment char at ADDR if appropriate.
  dup c@ digit?                 \ is char a digit
  If                            \ y:
    ?increment                  \   may need to increment
  Else                          \ n:
    drop                        \   throw away addr
  Endif
;

: Build+        \ buffer --
\ *G Increment build no. stored at BUFFER.
  1 Carry !                     \ bump build.no by 1
  build.no-size @               \ for all chars in $:
  Begin                         \
    dup 0>                      \ while: still a char
  While                         \
    1-                          \   index into string
    2dup + ?incr-char           \   increment char if necessary
  Repeat
  2drop
;

: White-Space?  \ char -- t/f
\ *G TRUE if char is a white space character.
  >r                            \ check for -
  r@ bl =  r@ 9 =  OR           \   space or a tab
  r@ 13 =  r> 10 =  OR          \   cr or lf
  OR
;

: Make-$        \ addr --
\ *G Make counted string within buffer.
  0					\ length to use
  1 build.no-size @ 1 max ?DO
    over i + c@ white-space? not
    IF    drop i ( 1+ )  LEAVE  ENDIF	\ replace 0 with length
  -1 +LOOP
  swap c!
;

\ SFP003...
create Months	\ -- addr
  ", JanFebMarAprMayJunJulAugSepOctNovDec"

: Mon$		\ n -- caddr len ; n=1..12
  ( 1 max #12 min ) 1- 3 * Months + 1+ 3
;

: (date)	\ -- dow dd mm yyyy
  os_getlocaltime			\ -- secs min hrs day mon year dow
  swap >r swap >r swap >r
  nip nip nip
  r> r> r>
;

: (time)				\ -- hh mm ss
  os_getlocaltime 4drop			\ -- secs min hrs day mon year dow
  -rot swap
;

: GetDate$	\ buff --
\ *G Get the current date as a counted string in a buffer.
  base @ >r  decimal
  0 over c!  >r				\ set empty string
  (date)				\ -- dow dd mm yyyy ; R: -- base buff
  rot (u.) r@ append			\ dd
  bl r@ addchar
  swap Mon$ r@ append			\ mm
  bl r@ addchar
  (u.) r> append			\ yyyy
  drop
  r> base !
;

: (2.r)		\ u -- caddr len
  0 <# # # #>
;

: GetTime$	\ buff --
\ *G Get time as counted string in buffer.
  base @ >r  decimal
  0 over c!  >r				\ set empty string
  (time)				\ -- hh mm ss ; R: -- base buff
  rot (2.r) r@ append			\ hh
  [char] : r@ addchar
  swap (2.r) r@ append			\ mm
  [char] : r@ addchar
  (2.r) r> append			\ ss
  r> base !
;
\ ...SFP003

previous definitions


\ ===============
\ *N Public words
\ ===============
\ *P These words are the application words used to provide
\ ** automatic build numbering.

also system

: Make-Build    \ buffer --
\ *G Read the contents of the build number file and place as a counted
\ ** string in the application defined buffer for later use.
  dup ^build-buffer !           \ set pointer to application buffer
  dup 1+ build-version          \ read in version string
  make-$                        \ make into counted string
;

\ SFP003...
: Build$,	\ --
\ *G Read the contents of the build number file and place as a
\ ** counted string at *\fo{HERE}. *\fo{ALLOT} the required space.
  here dup Make-Build  c@ 1+ allot
;

: Date$,	\ --
\ *G Compile date as counted string.
  pad GetDate$  pad count $,
;

: Time$,	\ --
\ *G Compile time as counted string
  pad GetTime$  pad count $,
;

: DateTime$,	\ --
\ *G Compile date and time as counted string
  { | temp[ #256 ] -- }
  0 temp[ c!
  pad GetDate$  pad count temp[ append
  s" , " temp[ append
  pad GetTime$  pad count temp[ append
  temp[ count $,
;
\ ...SFP003

: Update-Build  \ --
\ *G Update the contents of the build number file ready for the next
\ ** build.
  ^build-buffer @               \ pointer to version string
  tmp-build-buffer              \ temporary buffer
  build.no-size @ 1+            \ size of version string (plus count byte)
  cmove                         \ copy version string
  tmp-build-buffer 1+           \ use temporary buffer to update string
  dup build+                    \ increment version string
  create-build                  \ make new build no. file
  build.no-size @ write-build   \ write new version string
  close-build                   \ close build no. file
;

: Set-BuildFile	\ c-addr u --
\ *G Set the build number file.
  #256 min  szBuildFile place
;

: BuildFile	\ -- ; Buildfile <filename>
\ *G Use GetPathSpec to parse a filename from the input stream,
\ ** and make it the current build number file.
  GetPathSpec Set-BuildFile
;

previous

\ *P The following example, defines which file to use, loads
\ ** the text into a buffer, and finally updates the build text.
\ ** By placing *\fo{Update-Build} last in your load file, your
\ ** build number file will only be updated for each successful
\ ** build.
\ *E s" MyBuild.no" Set-Buildfile    \ set file to use
\ ** #256 buffer: MyVersion$  \ -- caddr
\ **   MyVersion$ make-build         \ load version string
\ ** ...
\ ** update-build                    \ put this last in load file

\ ======
\ *> ###
\ ======

