\ console.fth - character I/O for TTY devices

((
Copyright (c) 1997-2004, 2005, 2007, 2010, 2011, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20151124 SFP008 Made KEY sensitive to the multitasker.
20110214 SFP007 Increased output buffer to 4kb (configurable).
20100305 SFP006 Buffered output.
20071126 SFP005 Added Unix/DOS sensitivity.
20071102 SFP004 Changed KEY again to speed message handling.
20070806 SFP003 Added 2 ms time delay in KEY.
20070624 SFP002 Added Multitasker hook for Windows version
20050701 SFP001 Added multitasker hook for DOS version.
		Changed ACCEPT to use current SID for KEY.
))

4 kb equ /DefConBuff		\ SFP007

Target_386_Linux  Target_ARM_Linux or  Target_386_OSX or
Target_x64_Linux or  Target_x64_OSX or [if]
  #10 equ EOLchar_con		\ character marking end of line
  #13 equ IGNchar_con		\ character to ignore
[else]
  #13 equ EOLchar_con		\ character marking end of line
  #10 equ IGNchar_con		\ character to ignore
[then]

also system definitions

struct console-sid      \ -- len ; defines the SID of a console device
  gen-sid +                             \ reuse field names of GEN-SID
  int cs.hIn				\ input handle
  int cs.hOut				\ output handle
  int cs.hErr				\ error handle
  int cs./Obuff				\ size of output buffer
  int cs.#oBuff				\ number of chars in output buffer
  int cs.oBuff				\ address of output buffer
end-struct

: flushOP-console \ sid -- ior
  dup cs.#oBuff 2@ dup if		\ -- sid caddr len
    OS_TYPE  cs.#oBuff off  0
  else
    nip nip
  endif
;

: open-console    \ addr len attribs sid -- handle/sid ior
  nip nip nip
  $40 c/line !
  0 over cs.#oBuff !  0 over gen-handle !
  0
;

: close-console   \ sid -- ior
  dup gen-handle on  cs.#oBuff off  0  ;

: read-console    \ addr len sid -- ior
  2drop drop  0  ;

: write-console   \ addr len sid -- ior
  2drop drop  0  ;

: key?-console	\ sid -- 0
  emptyidle flushOP-console drop OS_KEY?  ;

Target_386_WDosX [if]
: key-console   \ sid -- key ; SFP001, SFP008
  begin					\ -- sid
    emptyidle  dup flushOP-console drop
    OS_KEY? 0=
   while
    pause
  repeat
  drop  OS_KEY
;
[else]
: key-console   \ sid -- key ; SFP001, SFP008
  begin					\ -- sid
    emptyidle  dup flushOP-console drop
    OS_KEY? 0=
   while
    2 ms				\ SFP003, SFP004
  repeat
  drop  OS_KEY
;
[then]
((
Target_386_WDosX [if]
: key-console   \ sid -- key ; SFP001
  emptyidle flushOP-console drop
  begin OS_KEY? 0= while pause repeat
  OS_KEY
;
[else]
: key-console   \ sid -- key ; SFP001
  emptyidle flushOP-console drop
  begin OS_KEY? 0= while  emptyidle 2 ms  repeat	\ SFP003, SFP004
  OS_KEY
;
[then]
))

: ekey?-console   \ sid -- flag ; SFP001
  key?-console  ;

: ekey-console    \ sid -- echar ; SFP001
  key-console  ;

: emit-console    \ char sid --
  dup cs.#oBuff @ over cs./oBuff @ >= if
    dup flushOP-console drop
  endif
  tuck cs.#oBuff 2@ + c!  cs.#oBuff incr
  1 out +!
;

: emit?-console   \ sid -- flag
  drop  true  ;

: type-console    \ caddr len sid --
  -rot  bounds
  ?do  i c@ over emit-console  loop
  drop
;

: cr-console      \ sid --
  eol$ count rot type-console  out off  op-line# incr  ;

: lf-console      \ sid --
  cr-console  ;

: ff-console      \ sid -- ; page/cls on display devices
  cr-console  ;

: bs-console      \ sid -- ; destructive on display devices
  #8 swap emit-console  ;

: accept-console	\ address maxlen sid -- len
  {: address maxlen sid | alen -- len' :}
  0 -> alen

  address maxlen bounds ?do
    case  sid key-console

\ [ Target_386_Linux ] [IF] dup #127 = if drop #8 then [THEN]
      dup #127 = if  drop #8  then	\ treat DEL as BS

      #8 of                             \ <BACKSPACE>
        alen 0<> if                     \ if not at left edge
          alen 1- -> alen               \ decrement length
          #8 emit  #32 emit  #8 emit
          -1                            \ +LOOP modification -1
        else
          0                             \ Don't increment Loop
        then
      endof

      #9 of                             \ <TAB>
        bl dup emit  i c!               \ treat as space
        alen 1+ -> alen
        1
      endof

      EOLchar_con of                    \ <CR> for DOS/Win, <LF> for Unices ; SFP005
        bl dup emit                     \ output a <SPACE>
        i c!                            \ store on end of string
        leave                           \ and leave
      endof

      IGNchar_con of                    \ <LF> for DOS/Win, <CR> for Unices ; SFP005
        0				\ ignore
      endof

      dup emit                          \ DEFAULT
      i c!                              \ store char in buffer
      alen 1+ -> alen                   \ increment string length
      1                                 \ modify address by 1 (+LOOP)

    end-case
  +loop

  alen
;
((
: accept-console	\ address maxlen sid -- len
  {: address maxlen sid | len -- len' :}
  0 -> len

  address maxlen bounds ?do
    case  sid key-console

\ [ Target_386_Linux ] [IF] dup #127 = if drop #8 then [THEN]
      dup #127 = if  drop #8  then	\ treat DEL as BS

      #8 of                             \ <BACKSPACE>
        len 0<> if                      \ if not at left edge
          len 1- -> len                 \ decrement length
          #8 emit  #32 emit  #8 emit
          -1                            \ +LOOP modification -1
        else
          0                             \ Don't increment Loop
        then
      endof

      #9 of                             \ <TAB>
        bl dup emit  i c!               \ treat as space
        len 1+ -> len
        1
      endof

      EOLchar_con of                    \ <CR> for DOS/Win, <LF> for Unices ; SFP005
        bl dup emit                     \ output a <SPACE>
        i c!                            \ store on end of string
        leave                           \ and leave
      endof

      IGNchar_con of                    \ <LF> for DOS/Win, <CR> for Unices ; SFP005
        0				\ ignore
      endof

      dup emit                          \ DEFAULT
      i c!                              \ store char in buffer
      len 1+ -> len                     \ increment string length
      1                                 \ modify address by 1 (+LOOP)

    end-case
  +loop

  len
;
))

: bell-console    \ sid -- ; audible beeper
  drop  ;

: setpos-console  \ x y mode sid -- ior
  2drop 2drop 0  ;

: getpos-console  \ mode sid -- x y ior
  2drop  0 0 0  ;

: ioctl-console   \ addr len fn sid -- ior
  2drop 2drop  0  ;

: RFU-console     \ --
  1 abort" RFU not supported"  ;

: init-console    \ addr len sid -- ior
  drop 2drop  0  ;

: term-console    \ sid -- ior
  drop  0  ;

: config-console  \ sid -- ior ; produces a dialog
  drop  0  ;

create console-vectors
  ' Open-console ,
  ' Close-console ,
  ' Read-console ,
  ' Write-console ,
  ' Key-console ,
  ' Key?-console ,
  ' Ekey-console ,
  ' EKey?-console ,
  ' Accept-console ,
  ' Emit-console ,
  ' Emit?-console ,
  ' Type-console ,
  ' CR-console ,
  ' LF-console ,
  ' FF-console ,
  ' BS-console ,
  ' Bell-console ,
  ' setpos-console ,
  ' Getpos-console ,
  ' Ioctl-console ,
  ' FlushOP-console  ,
  ' RFU-console ,
  ' Init-console ,
  ' Term-console ,
  ' config-console ,

previous definitions

also system

target-only
: console:      \ -- ; -- addr
  create
    -1 ,			\ handle
    console-vectors ,		\ vector
    0 , 0 , 0 ,			\ stdin stdout stderr
    /DefConBuff ,		\ buffer size
    0 ,  here cell+ ,
    /DefConBuff allot
;
host&target
interpreter
: console:      \ -- ; -- addr
  create
    -1 ,			\ handle
    console-vectors ,		\ vector
    0 , 0 , 0 ,			\ stdin stdout stderr
    /DefConBuff ,		\ buffer size
    0 ,  here cell + ,		\ used, address
    /DefConBuff allot
;
target


\ ================================
\ Install VFX Forth default device
\ ================================

console: default-console

: init-io	\ --
  default-console dup ip-handle ! dup op-handle !  to PauseConsole
  NULL NULL NULL open-gen 2drop
;
' init-io AtCold

previous

\ =============================================================================
