\ Quotations.fth - nested definitions for VFX Forth

((
Copyright (c) 2021
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *> kerneldf
\ *S Nested definitions
\ ===========
\ *P Quotations provide nested colon definitions, in which the
\ ** inner definition(s) are nameless. The expression:
\ *E   : foo ... [: some words ;] ... ;
\ *P is equivalent to:
\ *E :noname some words ; Constant #temp#
\ ** : foo ... #temp# ... ;

\ *P A simple quotation is an anonymous colon definition that is
\ ** defined inside a colon definition or another quotation. It
\ ** has no access to locals of the enclosing definitions.
\ ** Quotations can use local variables and *\fo{RECURSE}.

\ *P A good example use of quotations is to provide a solution
\ ** to the use of *\fo{CATCH} in a form like the *\fo{TRY ... EXCEPT}
\ ** blocks of other languages.
\ *E : foo     \ i*x -- j*x
\ **   setup
\ **   [: fee fi fo fum ;] catch
\ **   if ... then
\ **   teardown
\ **   ( throw again )
\ ** ;

' localvars voc>wid wid-threads	2constant localThreads \ -- addr len

also system

: saveLThreads	\ -- ... addr
\ preserve compilation and locals state
\ This will leak memory if the compiler aborts
  <headerless> @  last @  last-colon-xt @
  tdp @  lvcount @  flvcount @  TDPstart @  current @
  localThreads dup ProtAlloc swap 	\ -- threads addr len
  over >r  move  r>			\ -- addr
  localThreads erase			\ wipe locals voc
  local-nests incr			\ increase locals nesting
  lvcount off  flvcount off		\ necessary resets
;

: restoreLThreads \ addr ... --
\ restore previous compilation and locals state
  dup localThreads move  ProtFree
  current !  TDPstart !  flvcount !  lvcount !  tdp !
  last-colon-xt !  last !  <headerless> !
  local-nests decr
;

previous

: [:		\ comp: -- i*x orig colon-sys
\ *G Compilation: suspends compiling to the current definition,
\ ** starts a new nested definition, and compilation continues
\ ** with this nested definition. Outer locals are not visible
\ ** in the nested definition. Locals may be defined in the
\ ** nested definition. Inside the nested definition *\fo{RECURSE}
\ ** applies to the nested definition.
  false  postpone [  :noname  ;
ndcs:
  saveLThreads  POSTPONE AHEAD  true
  postpone [  :noname
;

: ;]		\ comp: i*x orig colon-sys -- ; run-time: -- xt
\ *G Compilation: Ends the current nested definition, and resumes
\ ** compilation to the previous current definition. At run-time
\ ** the xt of the nameless definition is returned.
  discard-sinline  POSTPONE ;  >r   IF
    ]  POSTPONE THEN  r> POSTPONE Literal
    restoreLThreads
    discard-sinline
  ELSE
    r>
  THEN ( xt )
; immediate


\ ======
\ *> ###
\ ======

decimal


\ *********
\ Test code
\ *********

0 [if]

: t1		\ a b -- c
  {: a b -- :}
  [:
    {: | a b -- :}
    100 to a  200 to b
    cr ." Mul = " * .
    ."  a*b = " a b * .
  ;] >r
  a b r@ execute
  ."  a b " a . b .
  r> drop
;

[then]
