\ VFX Forth implementatioon of Secure Hash Algorithm (SHA-1)
\ Spec in FIPS PUB 180-1 -- www.itl.nist.gov/fipspubs/fip180-1.htm
\ SHA-1 required for the Digital Signature Algorithm (DSA) and the
\ Digital Signature Standard (DSS) and all secure federal hash apps.
\ Assumes Little Endian, byte addressable CPU, e.g. Pentium class.
\ DEPENDENCIES: CORE EXT WORDSET ; COMMON USAGE 3DROP ?DO CELL-
\ Use of this code is free subject to acknowledgment of copyright.
\ Copyright (c) 2001 Jabari Zakiya -- jzakiya@mail.com  10/31/2001

\ N.B. This code requires VFX Forth for Windows build 703 or later.
\ If you are using an earlier build, redefine EVALUATE as follows:
\ : EVALUATE  SAVE-INPUT N>R  EVALUATE  R> RESTORE-INPUT DROP ;

\ ===============================================================
\ MACRO wordset from Wil Baden's Tool Belt series in
\ Forth Dimensions (FD) Vol. 19, No. 2, July/August 1997
\ Original code has been modified to make more efficient
\ MACRO allows insertion of parameters following the macro
\ "\" represents place where parameter is inserted
\ Example:  MACRO  ??  " IF  \  THEN "
\ : FOO .. ?? EXIT .... ;  ?? compiles to -- IF EXIT THEN

\ PLACE and STRING for system if needed
\ Not needed for SwiftForth v 2.00.3, needed for Win32Forth V 4.10
\ : PLACE  ( caddr n addr -)  2DUP  C!  CHAR+  SWAP  CHARS  MOVE ;
: SSTRING ( char "ccc" -) WORD COUNT HERE OVER 1+ CHARS ALLOT PLACE ;

\ Versions of /STRING and ANEW if system doesn't have them
\ : /STRING ( a n k - a+k n-k) ( OVER MIN) TUCK - >R CHARS + R> ;
\ : ANEW  >IN @ BL WORD FIND IF EXECUTE ELSE DROP THEN >IN ! MARKER ;

: split-at-char  ( a  n  char  -  a  k  a+k  n-k)
  >R  2DUP  BEGIN  DUP  WHILE  OVER  C@  R@  -
            WHILE  1 /STRING  REPEAT  THEN
            R> DROP  TUCK  2>R  -  2R>
;

: DOES>MACRO  \ Compile the macro, including external parameters
  DOES> COUNT  BEGIN [CHAR]  \ split-at-char  2>R  EVALUATE  R@
               WHILE BL WORD COUNT EVALUATE 2R>  1 /STRING REPEAT
               R> DROP   R> DROP
;

\ Macro creation word which allows parameter insertion
: MACRO  CREATE  IMMEDIATE  CHAR  SSTRING  DOES>MACRO  ;

\ =====================  Start SHA-1 Code =====================
  DECIMAL
  32 CONSTANT CELLSIZE    \ CPU bitsize

  VARIABLE  SHAlen              \ Holds byte length of string
  CREATE SHAval  5 CELLS ALLOT  \ Holds hash after each block
  CREATE SHAsh  85 CELLS ALLOT  \ Fully extended hash array
  CREATE W      16 CELLS ALLOT  \ Holds message block
  1 W !                         \ For endian testing

  HEX
    5a827999  CONSTANT  K1   \ Constant for rounds  0 - 19
    6ed9eba1  CONSTANT  K2   \ Constant for rounds 20 - 39
    8f1bbcdc  CONSTANT  K3   \ Constant for rounds 40 - 59
    ca62c1d6  CONSTANT  K4   \ Constant for rounds 60 - 79

  DECIMAL

MACRO ]L   " ] LITERAL "

  0 VALUE H[E]  \ Pointer to addr of hash value E for each round

MACRO H[D] " H[E]  [ 1 CELLS ]L  +"  \ Return D adr
MACRO H[C] " H[E]  [ 2 CELLS ]L  +"  \ Return C adr
MACRO H[B] " H[E]  [ 3 CELLS ]L  +"  \ Return B adr
\ MACRO rol\ " DUP [ CELLSIZE \ TUCK - ]L RSHIFT SWAP LITERAL LSHIFT OR "

: SHAinit ( -)  \ Load initial hash values H0 - H4
  [ HEX ] 67452301 ( H0)  efcdab89 ( H1)  98badcfe ( H2)
          10325476 ( H3)  c3d2e1f0 ( H4)  [ DECIMAL ]
  SHAsh  4 0 DO  TUCK  !  CELL+  LOOP  !   \ Put initial hash in SHAsh array
  SHAsh  SHAval  5 CELLS  CMOVE            \ Put copy in SHAval array
  SHAsh  TO  H[E]                          \ Init pointer to last hash value
;

: UpDateHash ( -)  \ Update hash values and load arrays with new values
  SHAsh  SHAval  H[E]                                  \ Place array addresses on stack
  5 0 DO  DUP >R   @   SWAP  DUP >R  @  +  DUP         \ Compute updated hash subvalue
          R@  !  OVER  !  CELL+  R>  CELL+  R>  CELL+  \ Store updated hash subvalue
  LOOP  3DROP                                          \ Clear stack when done
  SHAsh  TO  H[E]                                      \ Init pointer to last subvalue
;

\ Put two copies of original Wi on stack, keep its address
MACRO Wi@  " DUP  @  TUCK"  ( [Wi] - wi [Wi] wi)

\ Create 2 copies of new Wi' from Wi on stack  ( ..Wi -..Wi' Wi')
MACRO Wi  " 15 PICK  14 PICK  XOR  8 PICK  XOR  3 PICK  XOR "

\ Drop 80 Wi cells from stack ( W0..W79 - )
MACRO WiDROP  " 5 0 DO  2DROP 2DROP 2DROP 2DROP 2DROP 2DROP 2DROP 2DROP  LOOP " 

\ ( - n )  n = B XOR C XOR D
\ MACRO F2  " H[D] DUP >R  CELL+  2@ ( B C) XOR  R> @ XOR "
MACRO F2  " H[C] 2@  ( B C) XOR  H[D] @  XOR "

\ ( - n )  n = (B AND C) OR (~B AND D)
\ MACRO F1  " H[C] 2@  OVER  AND  SWAP  INVERT  H[D]  @  AND  OR"
MACRO F1  " H[D]  DUP >R  CELL+  2@  OVER  AND  SWAP  INVERT  R>  @  AND  OR"

\ ( - n )  n = (B AND C) OR (B AND D) OR (C AND D)
MACRO F3  " H[D]  DUP >R  CELL+  2@  OVER  AND  SWAP  R@ @  AND OR  R> 2@ AND  OR "


: HashAdjust ( A' -)  \ Adjust hash array for next round
  H[B]  DUP  @  30 rol  SWAP  !  \ A'     ;B = S(B,30) = rol(B,30)
  H[D]  DUP  TO  H[E]            \ A' [E] ;H[E] points to former H[D]
  [ 4 CELLS ]L  +  !             \ --     ;new A to new top of hash array
;

: bytes>< ( m -- w )  \ Reverse cell bytes: 12345678 <-> 78563412
  [ HEX ]  DUP >R  18 LSHIFT  R@  FF00 AND  8 LSHIFT  OR
  R@  FF0000 AND  8 RSHIFT  OR  R>  18 RSHIFT  OR  [ DECIMAL ]
;

: cellsreverse  ( adr n -- )  \ Reverse bytes of n cells in array
  0 DO  DUP  @  bytes><  OVER !  CELL+  LOOP  DROP
;

: rnds1  ( [W0] -  W0..W19) \ Do hash algorithm for rounds  0 - 19
  16 0 DO  Wi@                          \ ..Wi [Wi] Wi
           H[E]  DUP  >R                \ ..Wi [Wi] Wi (E)
           [ 4 CELLS ]L  +  @  5 rol  + \ ..Wi [Wi] Wi+S(A,5)
           R>  @  +  K1  +  F1  +       \ ..Wi [Wi] Wi+S(A,5)+E+K1+F1
           HashAdjust                   \ . Wi [Wi]
           CELL+                        \ ..Wi [Wi]'
  LOOP     DROP                         \ W0..W15   ;original block
   4 0 DO  Wi  1 rol  DUP               \ ..Wi Wi
           H[E]  DUP  >R                \ ..Wi Wi (E)
           [ 4 CELLS ]L  +  @  5 rol  + \ ..Wi Wi+S(A,5)
           R>  @  +  K1  +  F1  +       \ ..Wi Wi+S(A,5)+E+K1+F1
           HashAdjust                   \ ..Wi
  LOOP                                  \ ..W19
;

: rnds2  ( ..W19 - ..W39)   \ Do hash algorithm for rounds 20 - 39
  20 0 DO  Wi  1 rol  DUP               \ ..Wi Wi
           H[E]  DUP  >R                \ ..Wi Wi (E)
           [ 4 CELLS ]L  +  @  5 rol  + \ ..Wi Wi+S(A,5)
           R>  @  +  K2  +  F2  +       \ ..Wi Wi+S(A,5)+E+K2+F2
           HashAdjust                   \ ..Wi
  LOOP                                  \ ..W39
;

: rnds3  ( ..W39 - ..W59)   \ Do hash algorithm for rounds 40 - 59
  20 0 DO  Wi  1 rol  DUP               \ ..Wi Wi
           H[E]  DUP  >R                \ ..Wi Wi (E)
           [ 4 CELLS ]L  +  @  5 rol  + \ ..Wi Wi+S(A,5)
           R>  @  +  K3  +  F3  +       \ ..Wi Wi+S(A,5)+E+K3+F3
           HashAdjust                   \ ..Wi
  LOOP                                  \ ..W59
;

: rnds4  ( ..W59 - ..W79)   \ Do hash algorithm for rounds 60 - 79
  20 0 DO  Wi  1 rol  DUP               \ ..Wi Wi
           H[E]  DUP  >R                \ ..Wi Wi (E)
           [ 4 CELLS ]L  +  @  5 rol  + \ ..Wi Wi+S(A,5)
           R>  @  +  K4  +  F2  +       \ ..Wi Wi+S(A,5)+E+K4+F2
           HashAdjust                   \ ..Wi
  LOOP                                  \ ..W79
;

: SHA1  ( adr -)  \ Compute 160-bit SHA-1 hash of 512-bit message block
  rnds1  rnds2  rnds3  rnds4  WiDROP    \ Do SHA-1, create H[A]-H[E]
  UpDateHash                            \ Compute|update hash values
;

: setlen  ( -- )  \ Store count into last two cells
  SHAlen @  DUP  [ CELLSIZE 3 - ]L  RSHIFT  [ W 56 CHARS + ]L !  ( hi)
  3 LSHIFT  [ W 60 CHARS + ]L !  ( lo)
;

W C@ [IF]  \ if little ENDIAN, e.g. PCs
      MACRO endian16 "  DUP  16  cellsreverse "  ( adr -- adr )
      MACRO endian14 "  DUP  14  cellsreverse "  ( adr -- adr )
[ELSE]      \ if big ENDIAN, e.g. Macs
      MACRO endian16 "  " ( adr -- adr )  \ Do nothing
      MACRO endian14 "  " ( adr -- adr )  \ Do nothing
[THEN]


\ Do all 64 byte blocks leaving remainder block
: dofullblocks ( adr1 count1 --  adr2 count2 )
  DUP  -64  AND  ( count 63 > - ?)
  IF  DUP >R  6 RSHIFT  ( count/64)
      0 DO  DUP  endian16  SHA1  64 +  LOOP
      R> 63 AND
  THEN
;

: dofinal ( addr count -- )  \ Hash partial|last block
  DUP >R  W  DUP >R  SWAP  CMOVE
  R> R@ +  128  OVER  C! CHAR+  55 R@ -  R> 55 >
  IF  8 + 0 FILL  W  endian16  SHA1  W  56  THEN
  0 FILL  setlen  W  endian14  SHA1
;

\ Compute SHA from a counted buffer of text
: SHAbuffer ( addr count -- )
  SHAinit  DUP  SHAlen !  dofullblocks  dofinal
;

\ ===============  Hash string display wordset  ===============
  DECIMAL

\ Array of digits 0123456789abcdef
: digit$  ( -- adr )  S" 0123456789abcdef"  DROP  ;

: intdigits ( -- )  0 PAD  ! ;
: savedigit ( n -- )  PAD  C@  1+  DUP  PAD  C!  PAD  +  C!  ;
: bytedigits ( n1 -- )
  DUP 4 RSHIFT digit$ + C@ savedigit 15 AND digit$ + C@ savedigit
;

  W C@ [IF] \ little ENDIAN
: celldigits ( a1 -- )  DUP 3 + DO I C@ bytedigits  -1 +LOOP ;
  [ELSE]    \ big ENDIAN
: celldigits ( a1 -- )  DUP 4 + SWAP DO I C@ bytedigits LOOP ;
  [THEN]

: SHAstring ( -- adr count )  \ Return counted SHA string array
  intdigits  [ SHAval 4 CELLS + ]L
  5 0 DO  DUP  celldigits  CELL-  LOOP  DROP  PAD  COUNT
;

\ Display SHA-1 hash value in hex ( A B C D E )
: HASH. CR  SHAstring  TYPE  SPACE  ;

: QuoteString ( adr cnt --)  [CHAR] " EMIT  TYPE  [CHAR] " EMIT ;

\ ====================  File hash wordset  ====================
  VARIABLE  rfileid     \ Holds fileid number of input file

: InputFileName  ( -- ior)
  CR  CR  ." Filename: "  PAD  DUP  80  ACCEPT ( adr #)
  R/O  OPEN-FILE  SWAP  rfileid !  ( ior)
;

: TryAgain?  ( -- ?)
  CR  CR ." Invalid iput file, try again? (Y/N)"
  KEY  DUP  EMIT  DUP [CHAR] N =  SWAP [CHAR] n = OR
;

\ Read n bytes from input file, store at addr array
: bytes@  ( adr n - )  rfileid @  READ-FILE  2DROP ;

: storelen  ( lo hi - )  \ Store count into last two cells
  D2*  D2*  D2*  [ W 56 CHARS + ]L !  [ W 60 CHARS + ]L !
;

: getpartial ( cnt  -- W'  cnt2 ?)
  W  2DUP  SWAP  DUP >R  bytes@        ( cnt1 adr1  )
  + 128 OVER C! CHAR+ 55 R@ - R> 55 >  ( adr2 cnt2 ?)
;

MACRO block@  " W  64  bytes@ "

: SHAfile ( -- )
  BEGIN  InputFileName  ( ior)           \ Enter filename
  WHILE  TryAgain? IF  EXIT  THEN        \ Not valid, try (not) again
  REPEAT SHAinit                         \ Valid file, init transform
  rfileid @  FILE-SIZE  DROP  ( ud )     \ Get bytesize of input file
  2 0  D-                                \ Dec cnt by 2 for CR|LF EOF
  CR ." Bytesize: " 2DUP  D.             \ Display filesize to screen
  2DUP  2>R                              \ Save message cnt on RETURN
  64  UM/MOD  ( rembytes nblocks )       \ Compute nblocks & rembytes
  0 ?DO  block@  W endian16  SHA1  Loop  \ Do n full blocks
  ( rembytes)  getpartial ( adr cnt ?)   \ Read remaining bytes
  IF 8 + 0 FILL  W endian16  SHA1  W 56 THEN  \ Do if rembytes > 55
  0 FILL  2R> storelen  W endian14  SHA1 \ Do last block
  CR  ." SHA Hash: "  SHAstring TYPE CR  \ Show SHA-1 hash for file
  rfileid @  CLOSE-FILE  DROP            \ Close the input file
;

\ ====================  SHA-1 Test Suite  =====================
  DECIMAL

\ Load W array with data on stack
: WLoad ( d0..d15 -- )  [ W 15 CELLS + ]L
  16 0 DO  TUCK  !  CELL-  LOOP  DROP
;

\ -------------------------------------------------------------
\ EXAMPLE 1: from FIPS PUB 180-1
\ Message: ASCII string 'abc'
\ Hash = A9993E36  4706816A  BA3E2571  7850C26C  9CD0D89D

\ Compute and display hash for ASCII string 'abc'
: EX1  S" abc"  SHAbuffer  HASH.  ;

\ -------------------------------------------------------------
\ EXAMPLE 2: from FIPS PUB 180-1
\ Message:'abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq'
\ Hash = 84983E44  1C3BD26E  BAAE4AA1  F95129E5  E54670F1

: EX2a  S" abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq" ;
: EX2  EX2a  SHAbuffer  HASH.  ;

\ -------------------------------------------------------------
\ EXAMPLE 3: from FIPS PUB 180-1
\ Message: 1 million copies of 'a' (61h), (8 million bits)
\ Hash = 34AA973C  D4C4DAA4  F61EEB2B  DBAD2731  6534016F

\ Load block of all 'a's (61h), must hash 15,625 times
: EX3a  W  64  [CHAR] a  FILL  ;

\ Last message block: 1st bit a '1', bit-count = 8 million
: EX3b  [ HEX ] 80000000 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        [ DECIMAL ] 8000000 WLoad
;
\ Do hash for message of 1 million copies of ASCII 'a' (61h)
: EX3  SHAinit  EX3a  15625 0 DO W SHA1 LOOP  EX3b  W SHA1 HASH. ;

\ -------------------------------------------------------------
\ Message: 10,000 SPACES 'BL' (20h), (80,000 bits)
\ Hash = 37C7DDC3  EFEC3BDC  4ED4CD05  12447278  DFF38E2E

\ Load block of all "BL' (20h), hash 156 full blocks + 16 bytes
: EX4a  W  64  BL  FILL  ;

\ Last message block: 16 SPACES, bit-count = 80,000
: EX4b  [ HEX ] 20202020 20202020 20202020 20202020 80000000
                0 0 0 0 0 0 0 0 0 0 [ DECIMAL ] 80000 WLoad
;
\ Do hash for message of 10,000 SPACES 'BL' (20h)
: EX4  SHAinit  EX4a  156 0 DO  W SHA1  LOOP  EX4b  W SHA1 HASH. ;

\ -------------------------------------------------------------
\ Message: blank string ''
\ Hash =  DA39A3EE  5E6B4B0D  3255BFEF  95601890  AFD80709

: SHATest ( -- )
  CR ." SHA-1 FIPS PUB 180-1 test suite + others:"
  S" "  SHAbuffer  HASH.  S" "  QuoteString
  EX1  S" abc"  QuoteString
  EX2  EX2a     QuoteString
  EX3  S" 1 million copies of ASCII 'a' (61h)" TYPE
  EX4  S" 10,000 copies of ASCII BL (20h)" TYPE  CR
;

\ ===========  VFX Forth specific performance test ===========

  [undefined] GetTickCount
  [IF] extern: DWORD PASCAL GetTickCount( void ) [THEN]

  VARIABLE  start-ms

: TIMER-RESET  ( -- )  GetTickCount  start-ms  !  ;

: MS?  ( -- u )  GetTickCount  start-ms  @  - ( abs )  ;

  DECIMAL 1000000 VALUE N#

: .### ( -) BASE @ >R DECIMAL MS? 0 <# # # # [CHAR] . HOLD #S #> R> BASE ! TYPE ;

  DECIMAL 100 VALUE N#

: [EX1]  S" abc"  SHAbuffer  ;
: [EX2]  EX2a     SHAbuffer  ;
: [EX3]  SHAinit  EX3a  15625 0 DO W SHA1 LOOP EX3b W SHA1 ;

: test1  [ DECIMAL ]
  cr ." SHA-1 test for EX1 for " N# . ." loops in milliseconds is "
  TIMER-RESET  N# 0 DO  [EX1]  LOOP  MS?  U.
;

: test2  [ DECIMAL ]
  cr ." SHA-1 test for EX2 for " N# . ." loops in milliseconds is "
  TIMER-RESET  N# 0 DO  [EX2]  LOOP  MS?  U.
;

: test3  [ DECIMAL ]
  cr ." SHA-1 test for EX3 for " N# . ." loops in milliseconds is "
  TIMER-RESET  N# 0 DO  [EX3]  LOOP  MS?  U.
;
