\ Docgen4.fth - fourth version with Markdown and HTML5.
\               and two letter tags.

((
Copyright (c) 2000-2005, 2006, 2008, 2009, 2011, 2015, 2017, 2018, 2019, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  901 313 4312 (North American access number to UK office)

Please note that MPE files have hard tabs set to 8. Hence, tabs
are at columns 1, 9, 17, 25 ...

To do
=====
Reduce size of code by using S\" and friends.


Change history
==============
20201019 GEW028 Added DG_Personality? ( xt -- flag ).
20200213 SFP027 Put DELETE-FILE before some CREATE-FILEs.
20180824 SFP026 Updated Texinfo not to use the hyperref package.
20171009 SFP025 Expansion to optional two-letter tags, e.g. *PA
20171001 JED024 Markdown paragraph fix.
                Markdowntype now escapes pipe '|', important for tables.
                Fixed markdown macros.
20170926 JED023 Markdown cleanup.
                All remaining HTML removed from Markdown personality.
                Markdown tables work.
20170804 JED022 Added initial commit of the Markdown personality.
20170618 JED021 *! accepts separate title.
                *^ creates blank file, no head node included.
		Tables are HTML5 compliant.
		Updated example code.
20170609 JED020 Cleaned html layout code.
                Added Title tag.
		Added *^ and *# tags.
		Added Bootstrap version number.
20170524 JED019 Added initial commit of the HTML 5 personality.
20150612 SFP018 Added cache flushes for ARM Linux.
20110412 SFP017 PARSED and DOCONLY set the INCDIR macro.
20091220 SFP016 Enhanced HTML strictness.
20090825 SFP015 Added image macro for TeXinfo.
20080416 SFP014 Added table support to Texinfo and HTML.
20060207 SFP013 Changed error handling with CREATE-FILE.
		Added BR macro for line breaks.
		Permitted nested lists using *( and *) tags and
		a choice of bulleted, numbered or lettered lists.
20060124 SFP012 Removed BGCOLOR for INDEX.HTML.
		Added HTMLBACK ( caddr len -- ) to set the HTML
		background colour.
		Permitted use of "\ *x" without trailing space.
20051130 SFP011 Tuned Texinfo for MikTex v2.4.
20050106 SFP010 Texinfo ignores index entries containing '{' or '}'.
20041025 SFP009 Added table of contents and index files for HTML.
20040608 SFP008 Added *[ *] and *L tags.
		Added *\command{text} formatting.
		Major overhaul.
20030824 SFP007 Fixed font switches for HTML
20020722 SFP006 Restored the SuperDocState check.
20010420 SFP005 Refactored DOCGEN hooks so that source inliner
		does not close the output file.
		Cleaned some string storage code.
20010404 SFP004 Made -DOCGEN close the active file.
		Made error handling more robust.
20010325 SFP003 Commented out test code.
		Moved control words to KERNEL.FTH
20010314 SFP002 Allowed user definition of comment position
20001121 SFP001 Finished derivation from DOCGEN 1
))

[defined] -docgen [if] -docgen [then]

[undefined] data-file [if]
: data-file	\ -- size ; DATA-FILE <filename>
\ Loads a file to memory at *\fo{HERE} and *\fo{ALLOT}s memory.
\ The size of the file is returned.
  GetPathSpec r/o open-file #-413 ?throw >r	\ -- handle
  r@ file-size #-414 ?throw drop		\ -- size ; R: -- handle
  here over r@ read-file #-414 ?throw drop	\ -- size ; R: -- handle
  dup allot
  r> close-file #-416 ?throw		\ close
;
[then]

[undefined] csplit [if]
: csplit	\ addr len char -- raddr rlen laddr llen
\ Extract a substring at the start of addr/len, returning
\ the string raddr/rlen which includes char (if found) and
\ the string laddr/llen which contains the text to left of char.
\ If the string does not contain the character, raddr is
\ addr+len and rlen=0.
  >r 2dup r> scan                       \ -- addr len addr' len'
  2swap 2 pick -                        \ -- addr' len' addr len-len'
;
[then]

[undefined] split$LR [if]
: split$LR	\ addr len char -- laddr llen raddr rlen
\ Extract a substring at the start of *\i{addr/len}, returning
\ *\i{raddr/rlen} the string remaining after char and the substring
\ *\i{laddr/llen} which includes all characters before but not
\ including *\i{char}. If the string does not contain the
\ character, *\i{raddr} is *\i{addr+len} and *\i{rlen}=0.
  rot dup 2swap scan			\ -- addr addr' len'
  dup 0= if				\ can't split
    -rot tuck over -			\ -- 0 addr' addr len
    2swap swap				\ -- addr len addr' 0
  else
    over swap				\ -- addr addr' addr' len'
    1 /string 2swap			\ -- addr'' len'' addr addr'
    over -				\ -- addr'' len'' addr sslen
    2swap				\ -- addr sslen addr'' len''
  endif
;
[then]

\ =========
\ *! docgen
\ *T DocGen Documentation Generator
\ =========
\ *Q "Documentation is like sex: when it is good, it is very, very
\ ** good; and when it is bad, it is better than nothing." - Dick Brandon

\ *******************
\ *S What DocGen does
\ *******************
\ *P DocGen is a simple form of literate programming that enables
\ ** you to generate software manuals and Forth glossaries directly
\ ** from the source code of the software. The documentation is
\ ** produced from formal comments within the source code. The
\ ** manual for VFX Forth is itself produced this way.

\ *P Unlike some other forms of literate programming, all editing
\ ** is performed directly in the source code itself with the same
\ ** editor you use for editing the source code. This makes it
\ ** easy to ensure that your documentation is accurate and up
\ **to date.

\ *P In order to provide for both on-screen and printed
\ ** documentation, DocGen can generate output in several forms,
\ ** referred to as personalities. DocGen is supplied with
\ ** several personalities.
\ *(
\ *B HTML personality. From June 2017, the output is HTML5 using
\ ** a CSS stylesheet, *\i{bootstrap.3.3.7.min.css}.
\ *B Markdown personality. Markdown is a simple alternative to HTML
\ ** that has gained acceptance in many communities for text markup
\ ** and software documentation.
\ *B Latex and Texinfo personality. This is for generating PDFs
\ ** and printed documentation. It generates TeX output for
\ ** use with the Texinfo package. From LaTeX you can generate
\ ** many other output formats.
\ *B LaTeX2e personality This generates plain Tex output for
\ ** LaTeX2e. From LaTeX you can generate many other output
\ ** formats.
\ *)

\ *P For all operating systems, there is a range of LaTex
\ ** implementations. For Windows we use *\i{Miktek} from
\ *C   http://www.miktex.org
\ *P or TexLive from
\ *C   https://www.tug.org/texlive/
\ *P For Linux and Macs we usually use the Texlive distribution.

\ *P If you wish to generate output in other forms, instructions
\ ** for writing additional personalities are provided later in
\ ** this chapter.

\ *P DocGen supports the use of formatting commands (macros)
\ ** within the DocGen formal comments.

\ *P The DocGen HTML personality can generate separate output files
\ ** for a table of contents and an index of documented words. The
\ ** Texinfo personality can also index documented words using the
\ ** facilities of Texinfo.

\ *P The source code for DocGen is in the file *\i{Lib/DocGen4.fth}.
\ ** The earlier code that generates HTML v2 is in *\i{Lib/DocGen3.fth}.

\ ***************
\ *S Using DocGen
\ ***************
\ *P In normal use DocGen is enbled by *\fo{+DOCGEN} and disabled
\ ** by *\fo{-DOCGEN} which causes DocGen comments to be generated
\ ** when source code is compiled.

\ *P For use with embedded systems code that cannot be compiled
\ ** by VFX Forth replace the use of *\fo{INCLUDE} by
\ ** *\fo{DOCONLY} or replace *\fo{INCLUDED} by *\fo{PARSED}.
\ ** These can also be used to generate manuals separately from the
\ ** compilation process, which is useful when the ordering of the
\ ** manual is not the same as the compilation order. Note that the
\ ** '*>' and '*!' tags described in the section "Marking up your text"
\ ** can provide fine grain control of ordering, especially for TeX and
\ ** PDF manuals.

\ *P An example of using DocGen for a software project is the
\ ** manual for the ForthEd2 text editor in *\i{Examples\ForthEd2}.
\ ** *\i{Examples\ForthEd2\Manual\DocFiles} contains everything
\ ** needed to generate an indexed manual in both HTML and PDF
\ ** formats. Feel free to use these as templates for your
\ ** projects.

\ *P DocGen is controlled by the following words.


\ *****************
\ The DOCGEN engine
\ *****************

MODULE DOCGEN

\ --------------
\ Error messages
\ --------------

decimal
-500 #ErrDef dgn_splat		"DOCGEN: Internal Error. Generic I/O code not supported"
-501 #ErrDef dgn_close		"DOCGEN: Cannot close output file"
-502 #Errdef dgn_open		"DOCGEN: Cannot open parse file"
-503 #ErrDef dgn_illegal	"DOCGEN: Illegal Dochandler invoked"
-504 #ErrDef dgn_NOTDEF		"DOCGEN: Only valid between [DOCGEN and DOCGEN]"
-505 #ErrDef dgn_EXTTOOLONG	"DOCGEN: File extension too long"
-506 #ErrDef dgn_command	"DOCGEN: Bad command, use: *\cmd{text}"
-507 #ErrDef dgn_create		"DOCGEN: Cannot create output/index/toc file"
-508 #ErrDef dgn_number		"DOCGEN: Invalid number format"

variable prev-op	\ -- addr ; hold previous i/o handles

: DG-throw	\ n -- ; use previous O/P handle
  prev-op @ ?dup
  if  op-handle !  prev-op off  endif
  throw
;

: ?dg-throw	\ flag n --
  swap if dg-throw else drop endif
;


\ ------------
\ Debug output
\ ------------

((
: [con		\ -- ; R: -- ip-handle op-handle ; Force I/O to console
  r>  ip-handle @ >r  op-handle @ >r  >r
  xconsole setIo
;

: con]		\ -- ; R: ip-handle op-handle -- ; Restore I/O from console
  r>  r> op-handle !  r> ip-handle !  >r
;
))

((
: [con
  postpone [io  postpone consoledev  postpone setio
; immediate

: con]
  postpone io]
; immediate
))

\ ----------------
\ Auxiliary stacks
\ ----------------
\ An auxiliary stack is global. Its is controlled by:

: AuxStack:	\ u -- ; -- addr
\ Creates an auxiliary stack of u cells.
  create
    0 ,  cells dup ,  allot		\ index, size, stack buffer
;

: >A		\ x stack --
\ Push an item onto an auxiliary stack.
\ THROWs if the stack will overflow.
  dup 2@ = abort" Auxiliary stack overflow"
  tuck  dup @ + 2 cells + !
  cell swap +!
;

: A>		\ stack -- x
\ Pop an item from an auxiliary stack.
\ THROWs if the stack will underflow.
  dup @ 0= abort" Pop from empty auxiliary stack"
  cell negate over +!
  dup @ + 2 cells + @
;

: A@		\ stack -- x
\ Copy the top item from an auxiliary stack.
\ THROWs if the stack will underflow.
  dup @ 0= abort" empty auxiliary stack"
  dup @ + cell + @
;

: Adepth	\ stack -- n
\ Return the depth of the stack.
  @ cell /
;


\ ----------------------------------
\ Generic I/O device for output file
\ ----------------------------------

((
The generic I/O device does not use the standard *\fo{FILEDEV:} type because
it may need to be modified for other file formats which do not use the
standard CR/LF terminator.
))

: docgen-type	\ addr len sid --
  over out +!
  gen-handle @ write-file drop
;

: docgen-emit   \ char sid --
  { | b[ cell ] -- }
  swap b[ c!
  b[ 1 rot gen-handle @ write-file drop
  1 out +!
;

: docgen-cr     \ sid --
  #13 over docgen-emit
  #10 swap docgen-emit
  out off
;

: splat         \ --
  dgn_splat DG-throw
;

create docgen-vectors
  ' splat ,			\ 0 open
  ' splat ,			\ 1 close
  ' splat ,			\ 2 write
  ' splat ,			\ 3 read
  ' splat ,			\ 4 key
  ' splat ,			\ 5 key?
  ' splat ,			\ 6 ekey
  ' splat ,			\ 7 ekey?
  ' splat ,			\ 8 accept
  ' docgen-Emit ,		\ 9 emit
  ' splat ,			\ 10 emit?
  ' docgen-Type ,		\ 11 type
  ' docgen-CR ,			\ 12 cr
  ' splat ,			\ 13 lf
  ' splat ,			\ 14 ff
  ' splat ,			\ 15 bs
  ' splat ,			\ 16 bell
  ' splat ,			\ 17 setpos
  ' splat ,			\ 18 getpos
  ' splat ,			\ 19 ioctl
  ' splat ,			\ 20 flushop
  ' splat ,			\ 21 RFU
  ' splat ,			\ 22 Init
  ' splat ,			\ 23 Term
  ' splat ,			\ 24 Config

[undefined] INVALID_HANDLE [if]
  -1 constant INVALID_HANDLE
[then]

: docgen:	\ "name" -- ; Exec: -- sid
  create
    INVALID_HANDLE ,  docgen-vectors ,  0 ,	\ handle, vector, RFU
;


\ -------------------------------
\ DOCGEN structures and variables
\ -------------------------------

0 value DG_Personality	\ -- addr ; holds the current personality - output type
variable LastDocState	\ -- addr ; holds previous tag state
variable CurrDocState	\ -- addr ; holds current tag state
variable NextDocState	\ -- addr ; holds next tag state during exit action
variable SuperDocState	\ -- addr ; holds true if internal documentation is required
  SuperDocState off

4 constant /tocs	\ -- n ; max number of TOC levels
/tocs value TocLevel	\ -- n ; number of levels in table of contents
/tocs cells buffer: Toc#s	\ -- addr ; array of TOC numbers
0 value TocStarted?	\ -- n
\ True after first reference, used to generate header

variable Index#		\ -- addr ; holds index number
0 value IdxStarted?	\ -- n
\ True after first reference, used to generate header

Defer OtherLines	\ -- ; what is done to lines other than formal comments
  assign noop to-do Otherlines

docgen: doc-gio	\ -- addr ; sid for O/P file
docgen: toc-gio	\ -- addr ; sid for table of contents file
docgen: idx-gio	\ -- addr ; sid for index file
MAX_PATH buffer: CurrOpFile$	\ name of current output file

7 constant MAX_EXT_LEN                  \ Must be binary-1

struct /Personality	\ -- len
  MAX_EXT_LEN chars  field p.$Ext	\ o/p file extension for this personality
\  MAX_PATH    chars  field p.$ActiveDoc	\ path name of active document
  aligned
  cell               field p.FileId	\ handle of active file
  FILETIBSZ 1+ chars field p.$LastLine	\ last line
  aligned
  cell               field p.DocState	\ DOCGEN state
  cell               field p.tagChain	\ anchor for tags
  cell               field p.wid	\ command wordlist
  cell               field p.type       \ xt of TYPE word
end-struct

struct /TAG	\ -- len ; must match DG_TAG:
  cell field tag.link			\ link to previous tag or 0
  cell field tag.ascii-id		\ character(s) for tag
  cell field tag.controlcode		\ control code for state
  cell field tag.XTEnterState		\ executed on entry
  cell field tag.XTExitState		\ executed on exit
  cell field tag.XTContinueState	\ executed on continue
end-struct

0 value defining?	\ -- addr ; nz if defining a new personality
0 value preprobuff	\ -- addr ; points to preprocessed char buffer

#16 AuxStack: BultStk	\ -- addr ; bullet nesting stack


\ -----------------------------------------------------------------------

: LastDefinitionLine    \ -- c-addr u
  DG_Personality p.$LastLine count
;

: >LastDefinitionLine   \ c-addr u -- ; store string in buffer
\  DG_Personality p.$LastLine 2dup c! 1 chars + swap move
  #255 min
  DG_Personality p.$LastLine place
;

\ -----------------------------------------------------------------------

: TagChar>TAG   \ char ^Personality -- tag|0 ; find tag
  p.tagchain @
  begin
    dup
  while					\ -- char ^tag
    2dup tag.ascii-id @ =
    if  nip  exit  then
    tag.link @
  repeat
  nip
;

: valid?	\ char -- flag ; true if valid tag
  dup [char] * =			\ main tag indicator
  swap [char] + =  SuperDocState @ and	\ internal documentation
  or
\  [char] * =
;

defer DocGen-Spacing	\ c-addr u -- c-addr' u'
\ *G If your house rules require DocGen comments to start other than
\ ** in the first column, assign an action to this *\fo{DEFER}red word. The
\ ** action should remove characters before the start of the comment,
\ ** returning the modified string as c-addr' u'.
  assign noop to-do DocGen-Spacing

[-sin
: DGcomment?	\ caddr len -- caddr len flag
\ Return true if the line start is that of a comment.
  over c@ dup				\ 1st char MUST be '(' or '\'
  [char] \ =  swap [char] ( =  or 	\ correct comment start?
  2 pick 1 + c@ bl =			\ 2nd Char MUST be space
  and
;

: DGtagStart?		\ caddr len -- caddr len flag
\ Return true if the third character is '*' or '+'.
  over 2 + c@ valid?
;

: prepTagLine	\ caddr len tagdata n -- caddr' len' tag | -- 0
\ In the given string, find the given tag and remove the comment
\ and tag. N is the size of the comment/tag start, 5 for
\ "\ *x " and 6 for "\ *xy ".
  >r
  DG_Personality TagChar>TAG dup if	\ -- caddr u tag | caddr u 0
    -rot				\ -- tag caddr u
    over c@ [char] ( = if		\ strip ( ... ) comments
      dup r@ min /string dup if		\ step over "( *x "
        2dup + 1- c@ [char] ) =		\ strip trailing ')' char
        if  1-  then
      then
    else
      dup r@ min /string		\ step over "\ *x "
    endif
    -trailing rot			\ -- caddr u tag
  else
    nip nip				\ -- 0
  then
  r> drop
;
sin]

: DG2charComment?	\ c-addr u -- 'c-addr 'u TAG | -- 0
\ Return true if the text matches a '*x' comment line. If true,
\ also return the remainder of the line after the tag.
  dup 4 < if 2drop 0 exit then          \ Not Long Enough "( *x "
  DGcomment? 0= if  2drop 0 exit  then	\ proper DocGen comment
  DGtagStart? 0= if  2drop 0 exit  then	\ 3rd char MUST be '+' or '*'
  dup 4 > if				\ if string is longer than 4 chars
    over 4 + c@ bl <>			\ 5th Char MUST be space
    if  2drop 0 exit  then
  endif
  over 3 + c@ 5 prepTagLine		\ "\ *x "
;

: DG3charComment?	\ c-addr u -- 'c-addr 'u TAG | -- 0
\ Return true if the text matches a '*xy' comment line. If true,
\ also return the remainder of the line after the tag.
  dup 5 < if 2drop 0 exit then          \ Not Long Enough "( *xy "
  DGcomment? 0= if  2drop 0 exit  then	\ proper DocGen comment
  DGtagStart? 0= if  2drop 0 exit  then	\ 3rd char MUST be '+' or '*'
  dup 5 > if				\ if string is longer than 5 chars
    over 5 + c@ bl <>			\ 6th Char MUST be space
    if  2drop 0 exit  then
  endif
  over 3 + w@ 6 prepTagLine		\ "\ *xy " ; w@ assumes little-endian
;

: IsFormalComment?      \ c-addr u -- 'c-addr 'u TAG | -- 0
  DocGen-Spacing			\ optional preprocess ; SFP002
  2dup 2>r DG2charComment? ?dup 0= if
    2r> DG3charcomment?
  else
    2r> 2drop
  then
;

\ Any controlcode other than 0 will cause the DOCSTATE's exit handler
\ to be called.
\ ..then..
\ If code action is -1 perform entryXT and reset state to 0
\ if code action is 1 change DOCSTATE and perform entry code
\ if code action is 0 perform DOCSTATE handler

[-short-branches
: HandleFormal  \ c-addr u ^tag --
  { c-addr u ^tag -- }

  \ Perform exitXT of current DOCSTATE tag if new control code is non-zero.
  \ This performs the "close" operation for those other than CONTINUE
  ^tag tag.controlcode @ if
    DG_Personality p.DocState @ ?dup if
      ^tag NextDocState !			\ so exit code knows what comes next
      tag.XTExitState @ execute
      0 NextDocState !				\ exit code is done
    then
  then

  CurrDocState @ LastDocState !			\ save previous tag state

  ^tag tag.controlcode @ ?dup if		\ run entry code?
    ^tag CurrDocState !				\ mark current/new tag state
    0< if
      c-addr u ^tag tag.XTEnterState @ execute
      0 DG_Personality p.DocState !		\ no continuation possible
    else
      ^tag DG_Personality p.DocState !
      c-addr u ^tag tag.XTEnterState @ execute
    then
  else						\ or carry on same state
    DG_Personality p.DocState @ ?dup if
      c-addr u rot tag.XTContinueState @ execute
    then
  then
;
short-branches]

((
: TagsSame?	\ -- flag ; true if current and previous tags are the same
  CurrDocState @ LastDocState @ =
;

: TagsDiff?	\ -- flag ; true if current and previous tags are different
  CurrDocState @ LastDocState @ <>
;

: NextTagsSame?	\ -- flag ; true if current and next tags are the same
  CurrDocState @ NextDocState @ =
;

: NextTagsDiff?	\ -- flag ; true if current and next tags are different
  CurrDocState @ NextDocState @ <>
;
))

\ Handling Forth expressions and conditional DG tags

: Tag		\ -- char1/char2
\ Like CHAR but handles 1 or 2 character tags
  bl word count  case
    1 of  c@  endof
    2 of  w@  endof
        1 abort" invalid tag"
  endcase
;

: [tag]		\ -- char1/char2
\ Compile time version of TAG.
  Tag  postpone literal
; immediate

0 value ForthVal	\ -- x ; value returned from executing Forth
0 value CondLevel	\ -- u ; nesting level of conditionals

: >ev		\ caddr len --
\ Process EV tag.
  evaluate
;

: >if		\ caddr len --
\ process IF tag. CONDLEVEL was zero, tags were being processed
\ normally. If the result of interpreting the line are zero,
\ apply 1 conditional level, so stopping tag processing.
  evaluate 0= if  1 to CondLevel  then
;

: >else		\ caddr len --
\ Tags were being processed, so stop them.
  2drop  1 to CondLevel
;

: >then		\ caddr len --
\ Tags were being processed, so this is a NOOP.
  2drop
;

: doCondline	\ caddr len ^tag --
\ CONDLEVEL is non-zero, so tags are NOT being processed.
  nip nip tag.ascii-id @
  case
    [tag] IF of  1 add CondLevel  endof
    [tag] EL of  CondLevel
                 1- dup if 1+ then
                 to CondLevel
              endof
    [tag] TH of  1 sub CondLevel  endof
  endcase
;

: doDGline	\ caddr len ^tag --
  DG_Personality p.fileid @ doc-gio gen-handle !
  ip-handle @ op-handle @  dup prev-op !  2>R
  doc-gio op-handle !			\ switch o/p to file
  ['] HandleFormal catch		\ process DOCGEN comment
  2R> op-handle ! ip-handle !		\ restore o/p
  prev-op off
  THROW
  postpone \				\ skip line
;

: dPreRefill    \ -- ; hook in REFILL called before REFILL action
  [defined] FlushCache [if] FlushCache [then]
  source >LastDefinitionLine
  [defined] FlushCache [if] FlushCache [then]
;

[-short-branches
: dRefill       \ -- ; hook in REFILL called after REFILL action
\ ." DG:" source type cr
  source IsFormalComment? ?dup if	\ -- caddr u ^tag | 0
    CondLevel
    if  doCondLine  else  doDGline  then
  else
    OtherLines
  then
;
short-branches]

: sidOpen?	\ sid -- flag
  gen-handle @ INVALID_HANDLE <>
;

[-sin

: closeSid	\ sid --
\ Close file for this sid.
  dup gen-handle @
  INVALID_HANDLE rot gen-handle !
  close-file dgn_close ?dg-throw
;

: CloseActive   \ -- ; close active output file
  doc-gio sidOpen? if
\   cr ." closing"
    doc-gio closeSid
    INVALID_HANDLE DG_Personality p.FileId !
  endif
  0 LastDocState !			\ stop tag comparison
;

: AllocPrepro	\ -- ; allocate another PREPRO buffer
  #1024 allocate #-406 ?throw -> preprobuff
  preprobuff #1024 erase
;

: FreePrepro	\ -- ; release PREPRO buffer
  preprobuff ?dup if
    0 -> preprobuff
    free #-407 ?throw
  endif
;
sin]


\ =======================================
\ +S Tools for constructing personalities
\ =======================================

: Toc?		\ -- flag ; true if making an index
  toc-gio sidOpen?
;

: Index?	\ -- flag ; true if making an index
  idx-gio sidOpen?
;

[-sin

: SetDGname	\ caddr u buffer --
\ Make counted string of name+ext. c-addr/u is the base name,
\ buffer holds the result.
  >r -white r@ place			\ counted string in buffer
  [char] . r@ addchar			\ add '.'
  DG_Personality p.$Ext r> $+		\ add extension
;

: MakeDocPath   \ c-addr u buffer --
\ Make counted string of opfile+ext. c-addr/u is the base name,
\ buffer holds the result.
  dup >r SetDGname r> CurrOpFile$ $move	\ set output file name
;

: NewDocFile    \ c-addr u --
  { | path[ #260 ] -- }

\ cr ." in NewDocFile"
  CloseActive

  path[ #260 erase
  path[ MakeDocPath
\ [con  cr ." NDF: "  path[ count type  4 spaces  path[ count expand type  key? drop  con]
  path[ count delete-file drop
  path[ count r/w bin Create-File dgn_create ?throw	\ SFP013
  dup DG_Personality p.FileId !
  doc-gio gen-handle !
;

: SwitchDocFile \ c-addr u --
  { c-addr u | path[ #260 ] -- }
\ cr ." in SwitchDocFile"
  CloseActive

  path[ #260 erase
  c-addr u path[ MakeDocPath
\ [con  cr ." opening " path[ count type cr  con]
  path[ count r/w bin open-file if
\ [con  ." New" cr  con]
    drop  c-addr u NewDocFile
  else
\ [con  ." old" cr  con]
    dup DG_Personality p.fileid !
    dup doc-gio gen-handle !
    dup file-size drop
    rot reposition-file drop
  then
\ [con  key drop  con]
;

: MakeFile	\ addr len -- handle
  { | file$[ max_path ] -- }
  file$[ setDGname
\ [con  cr ." MF: "  file$[ count type  4 spaces  file$[ count expand type  key? drop  con]
  file$[ count delete-file drop
  file$[ count r/w create-file
  dgn_create ?throw
;

: MakeTocFile	\ --
  s" contents" MakeFile toc-gio gen-handle !
;

: MakeIndexFile	\ --
  s" indices" MakeFile idx-gio gen-handle !
;

: +Toc$		\ n -- caddr u
\ Update table of contents level n (1..n), and return the string
\ for it.
  1- cells Toc#s +			\ array slot
  dup incr				\ update this one
  cell+					\ wipe following numbers
  Toc#s /tocs cells + over - erase
  <#					\ build string "TOC_a_b_c_d"
  0 /tocs 1- cells do			\ for each level
    Toc#s i + @ 0 #s 2drop		\ number
    [char] _ hold			\ !
  cell negate +loop
  [char] C hold  [char] O hold  [char] T hold
  0 0 #>
;

: +Idx$		\ -- caddr u
\ Update index number, and return the string for it.
  Index# dup @ 1+ dup rot !		\ get/update Index number
  0 <#					\ build string "IDX_n"
    #s
    [char] _ hold			\ !
    [char] X hold  [char] D hold  [char] I hold
  #>
;

: [DGop		\ sid1 -- sid2
\ switch output to new sid.
  op-handle @  swap op-handle !
;

: DGop]		\ sid --
\ switch back to previous sid.
  op-handle !
;

: [toc		\ -- sid
\ Switch output to table of contents file.
  toc-gio [DGop
;

: [idx		\ -- sid
\ Switch output to index file.
  idx-gio [DGop
;

: +ppchar	\ char -- ; add character to buffer
  preprobuff zcount + tuck c!  0 swap 1+ c!
;

: +ppbuff	\ c-addr u -- ; add string to buffer
  preprobuff zAppend
;

: leftstring	\ addr len char -- addr' len' ssaddr sslen
 \ *G Extract a substring at the start of addr/len, returning
 \ ** the string remaining after char and the substring ssaddr/sslen
 \ ** which does not include char. A *\fo{-1 THROW} is performed if
 \ ** the delimiter is not found.
  rot dup 2swap scan dup 0= throw	\ -- addr addr' len'
  over swap				\ -- addr addr' addr' len'
  1 /string 2swap			\ -- addr'' len'' addr addr'
  over -				\ -- addr'' len'' addr sslen
;

((
: .DGbox	\ caddr u --
  { | buff[ #256 ] -- }
  s" -" buff[ zplace
  buff[ zAppend
  s" -" buff[ zAppend
  0 buff[ z" DOCGEN" MB_OK MessageBox drop
;

: .DG#box	\ n --
  base @ >r  hex
  0 <# #s #> .DGbox
  r> base !
;
))

: (preprocmd)	\ addr len -- addr' len'
 \ *G Process a command sequence starting at the '*' character.
  1 /string dup 0= throw		\ check for * at end of line
  over c@ [char] \ <> throw		\ must be '*\'
  1 /string				\ step over '\'
  [char] { leftstring 			\ find command and look it up
  DG_Personality p.wid @
  search-wordlist 0= throw 		\ -- addr' len' xt
  preprobuff zcount			\ flush preprobuff to the output
  DG_Personality p.type @ execute
  0 preprobuff c!
  >r [char] } leftstring r> execute	\ extract field and execute xt
  0 preprobuff c!
;

: preprocmd	\ addr len -- addr' len'
  2dup ['] (preprocmd) catch if
    2drop  [char] * +ppchar  1 /string
  else
    2swap 2drop
  endif
;

: preprotab	\ addr len -- addr' len'
  8 preprobuff zStrlen #7 and - 0
  ?do  bl +ppchar  loop
;

: preprochar	\ addr len -- addr' len'
  case  over c@
    #9       of  preprotab  1 /string  endof
    #30      of  S" (" +ppbuff  1 /string  endof	\ )
    #31      of  S" )" +ppbuff  1 /string  endof
    [char] * of  preprocmd  endof
        +ppchar  1 /string
  end-case
;

: prepro        \ c-addr u -- 'c-addr 'u
  [defined] FlushCache [if] FlushCache [then]
  0 preprobuff c!
  begin
    dup
   while
    preprochar
  repeat
  2drop
  preprobuff zcount
;

: isANumber?	\ $ -- flag
\ Checks to see if a counted string is a number, returning
\ true if so.
  case  number?
    0 of  0  endof
    1 of  drop  -1  endof
    2 of  2drop  -1  endof
      dgn_number dg-throw
  endcase
;

create OpsList	\ -- addr
\ A counted string list of stack operations.
  ", + - * / 2* 2/ 4* 4/ DUP OVER SWAP CELL CELLS CELLS+ NEGATE D+ D-"
[defined] FlushCache [if] FlushCache [then]

: IsOp?		\ $ -- flag
\ Checks to see if a counted string is a stack operation,
\ returning true if so.
  count  OpsList count search nip nip
;

#256 buffer: CurrName	\ -- $
\ Buffer to hold the current Forth word name.

: getName	\ buffer --
\ Get the next word in the input and place it in the buffer.
\  bl word over $move  count upper
  parse-name rot uplace
;

: Num/Op?	\ $ -- flag
\ Return true if the string is a number or a stack operation.
  dup isANumber?
  if  drop true  exit  endif
  isOp?
;

: ParseGloName	\ -- caddr u
\ Find the glossary entry name and return it. Glossary entries
\ are assumed to be of the following form:
\ [ <number> | <stack> ]   - may be several
\ <definer>                - must exist
\ <name>                   - must exist
  save-input n>r
  LastDefinitionLine prepro to-source
  -1 to source-id  >in off
  0
  begin					\ skip expression
    drop >in @
    CurrName getName
    CurrName Num/Op? 0=
  until
  >in !
  CurrName getName			\ skip definer
  >in @ >r  bl word c@ 0= if		\ if there is no next word
    r> drop				\ use previous as result
  else
    r> >in !
    CurrName getName			\ get name
  endif
  CurrName count
  nr> restore-input drop
;

: illegal	\ -- ; illegal handler
  -503 ?dg-throw	\ true abort" Illegal Dochandler invoked"
;

: (NewList)	\ caddr len -- char|0
\ See NEWLIST below.
  -white 0=
  if  drop 0  exit  endif
  c@
  dup [char] 1 = if  exit  endif
  dup [char] A = if  exit  endif
  dup [char] a = if  exit  endif
  drop 0
;

: NewList	\ caddr len -- char|0
\ Inspect the first character of the text and try to use it as
\ a list type. Null text indicates an undered list with bullets.
\ Ordered lists are indicated by '1', 'A' or 'a'. The return value
\ is zero for a bulleted list which will use the current
\ personality's default, or one of the above characters.
  (NewList) dup BultStk >A
;

: EndList	\ caddr len --
\ Restore the previous list state. If the bullet stack is empty
\ assume a bulleted stack.
  Adepth 0> if
    BultStk A> drop
  else
    0 BultStk >A
  endif
;

: newfExpand    \ caddr u -- title_caddr u file_caddr u
\ Check if we have a separate title. If no title, use
\ the filename as title.
  bl split$LR ?DUP if
       2swap
  else drop 2dup        \ use filename as title
  endif
;


sin]


\ ************
\ Using DocGen
\ ************

[-sin

: (+DOCGEN)     \ -- ; enable DOCGEN
  assign dPreRefill to-do DOCGEN_PREREFILL
  assign dRefill    to-do DOCGEN_REFILL
;

: (-DOCGEN)     \ -- ; disable DOCGEN
  assign noop to-do DOCGEN_PREREFILL
  assign noop to-do DOCGEN_REFILL
;
assign (+docgen) to-do +docgen_hook
assign (-docgen) to-do -docgen_hook

(( for DOCGEN
0 value DocGen?		\ -- flag ; true if DOCGEN enabled
( *G Returns true if DocGen is enabled, otherwise false. )
))

: +DOCGEN       \ -- ; enable DOCGEN
\ *G Enables DocGen generation of documentation. After using
\ ** *\fo{+DOCGEN} an output file must be selected using either the
\ ** '*!' (create) or '*>' (append) tags.
  true to DocGen?
  AllocPrepro				\ Get buffer
  toc#s /tocs cells erase		\ initialise table of contents
  0 -> TocStarted?
  Index# off				\ initialise indexing
  0 -> IdxStarted?
  (+docgen)
;

: -DOCGEN       \ -- ; disable DOCGEN
\ *G Disables DocGen generation of documentation. Note that the
\ ** active output file is closed, and a new output file must
\ ** be specified after *\fo{+DOCGEN} is used again.
  CloseActive (-docgen)			\ SFP004
  Toc? if  toc-gio CloseSid  endif
  Index? if  idx-gio CloseSid  endif
  false to DocGen?
  FreePrepro
;

: (parsed)	\ -- ; the inner DocGen loop
  0 to-source				\ set input buffer
  begin					\ process comments via REFILL hooks
    refill
  while
    postpone \
  repeat
;

also system
: PARSED        \ c-addr u --
\ *G Similar to *\fo{INCLUDED} but performs no actual compilation.
\ ** Macros are expanded, and file extensions are resolved.
\ ** *\fo{PARSED} allows formal comments to be processed from any
\ ** suitable source code, for example embedded systems code
\ ** for cross compilers.
\ ** *\fo{+DOCGEN} must be used before *\fo{PARSED} to enable the DocGen system,
\ ** and *\fo{-DOCGEN} should be used after the last file has been processed.
\ ** See also *\fo{DOCONLY} below.
  EmptyIdle
  ResolveIncludeFileName		\ allow for missing extensions
  2dup r/o bin Open-File dgn_open ?dg-throw	\ abort" Cannot open parse file"
  -rot  ++IncDir			\ set include directory macro
  save-input n>r
  0 line# !
  to source-id
  FILETIBSZ allocate #-406 ?throw	\ new TIB
  ['] (parsed) catch			\ for later THROW >-
  source drop free drop			\ free new TIB      |
  source-id close-file drop		\		    |
  nr> restore-input drop		\		    |
  ?dup if				\ <-----------------
    -docgen  DG-throw
\    prev-op @ ?dup
\    if  op-handle !  prev-op off  endif
\    throw
  endif
  --IncDir
  [defined] FlushCache [if] FlushCache [then]
;
previous

: DocOnly	\ "<text>" -- ; DOCONLY <filename>
\ *G Similar to *\fo{INCLUDE} but performs no actual compilation. This
\ ** allows formal comments to be parsed from any source code,
\ ** for example embedded systems code for the MPE cross
\ ** compilers. Use in the form *\fo{DOCONLY <filename>}.
\ ** *\fo{+DOCGEN} must be used before *\fo{DOCONLY} to enable the
\ ** DocGen system, and *\fo{-DOCGEN} should be used after the last
\ ** file has been processed.
  GetPathSpec parsed
;

: +InternalDocs	\ --
\ *G Permits generation of documentation for internal use. DocGen
\ ** behaves as normal, but also accepts tags of the form +X as well as
\ ** the normal *X form. This can be used to generate an additional
\ ** level of documentation above that prepared for normal use.
  SuperDocState on
;

: -InternalDocs	\ --
\ *G Turns off generation of internal use documentation. Only tags of
\ ** the form *X will be accepted.
  SuperDocState off
;
-InternalDocs

: +TOC		\ --
\ *G Turn on Table of Contents generation. The table of contents
\ ** file is called "contents.ext" where ext is defined for the
\ ** personality being used.
\ ** The table of contents file is closed by  *\fo{-DOCGEN}.
\ ** *\fo{+TOC} only affects HTML generation. Tex and Texinfo
\ ** can already handle table of contents generation.
  MakeTocFile
;

\ *P The table of contents file has four levels, corresponding to
\ ** the CH, SE, SS, and SH tags (T, S, N and H tags).

: +Index	\ --
\ *G Turn on Index generation. The HTML index file is called
\ ** "indices.html". The index file is closed by  *\fo{-DOCGEN}.
\ ** For Texinfo and Latex2e the index entries are placed in the
\ ** output and the index is generated by Texinfo and LaTex.
\ ** At present *\fo{+Index} only affects HTML and Texinfo generation.
  MakeIndexFile
;

\ *P The HTML index file can later be sorted using the
\ ** command line *\b{SORT} utility which is still present in
\ ** all versions of Windows. The command line should be:
\ *C sort /+51 indices.html /O sorted.html
\ *P DocGen produces *\i{indices.html} with word names starting
\ ** in column 51. The output of the *\b{SORT} utility is
\ ** the file *\i{sorted.html}.
\ *P The command line:
\ *C sort /?
\ *P will display the command syntax.

\ *P *\b{N.B.} The index file contains no header or footer text,
\ ** e.g. <HTML>, because the sorting process will move these
\ ** in the file. After sorting, the file should be edited to
\ ** add suitable headers and footers. For HTML output, examples
\ ** follow.
\ *E <HTML><BODY bgcolor="#C2C1B4">
\ ** <BR><BR><HR><BR><H1><font color="#ff0000">
\ ** Index
\ ** </font></H1><BR><HR><BR><BR>
\ **   ...
\ ** </HTML>

\ *P Texinfo and Latex indices are sorted by the *\i{makeindex}
\ ** program. You do not need to call this yourself as it can
\ ** be handled by the *\i{texify} program. See the examples
\ ** at the end of the chapter for more details.

sin]


\ ***********************
\ *S Marking up your text
\ ***********************
\ *P DocGen output is derived from formal comments within the source
\ ** code. The output format defaults to HTML 5.0.
\ ** The layout of the final document is controlled by DocGen tags
\ ** at the start of the comment and by formatting macros within the
\ ** body of the comments.

\ ===============
\ *N Comment tags
\ ===============
\ *P The DocGen comment takes one of the following two forms:
\ *E \ *xy blah blah
\ ** \ *x blah blah
\ *P where the backslash character '\' must be in the first column,
\ ** and XY and X are the operation codes or tags and "blah blah"
\ ** is the control text.
\ *E ( *XY blah blah )
\ ** ( *x blah blah )
\ *P where the open bracket must be in the first column, the closing
\ ** bracket must be the last character, and XY and X are operation
\ ** codes or tags and "blah blah" is the control text.

\ *P The use of two-character tags and backslash comments is
\ ** recommended. Every single character tag has a corresponding
\ ** two-character tag. Not all two-character tags have single
\ ** character equivalents, for example the conditional
\ ** documentation tags, *IF, *EL and *TH have no single-character
\ ** equivalents. All new DocGen tags will be two-character tags.

\ *P Please examine MPE supplied sources for a view of how DocGen
\ ** is used to build library and API documentation.

\ ----------------------
\ *SH Two character tags
\ ----------------------
\ *DF **  The following text is a continuation for the current style. This
\ ***     can only be used after a GL, EX, DF, IT, QU or PA tag.
\ *DF !!  Create and select a new output file. The control text is the
\ ***     filename without extension. The remaining text is the description
\ ***     which in HTML becomes the contents of the <title>text</title>
\ ***    tag.
\ *DF ^^  Create and select a new output file. The control text is the
\ ***     filename without extension. No header is added. This tag
\ ***     allows you to generate completely custom formats in HTML
\ ***     using W tags.
\ *DF >>  Select and append to an existing output file. The control text is
\ ***     the filename without extension. No header is added.
\ *DF ##  Finalize the current HTML page. Outputs a canned HTML footer
\ ***     which closes off the sections used by the canned header.
\ *DF CH  Following text is a chapter/section title.
\ *DF SE  Following text is a section sub-title.
\ *DF SS  Following text is a section sub-sub-title.
\ *DF SH  A simple heading.
\ *DF DF  Following text is a definition. The first space delimited token
\ ***     is the term, the remaining text the description.
\ *DF PA  Begins a new paragraph.
\ *DF EX  Begins a paragraph which is a code example.
\ *DF L(  Starts a bullet list. The first character of the text
\ ***     defines how the lists are marked. No text produces bullets,
\ ***     *\b{1} produces a numbered list and *\b{A} or *\b{a} produce
\ ***     lettered lists. Lists may be nested. *\b{N.B.} Previous
\ ***     markups that did not use the *\b{*(} and *\b{*)} tags still
\ ***     function as before, but the new layout is generally visually
\ ***     better. Note that the LaTeX personality only handles bulleted
\ ***     and numbered lists. See the change notes section of this
\ ***     chapter for more details.
\ *DF IT  Following text is a bulleted item.
\ *DF L)  Ends a bulleted list.
\ *DF GL  Following text is a glossary entry. The preceeding line is
\ ***     output in a fixed font code format.
\ *DF RH  Following text is output directly when using HTML output.
\ *DF RL  Following text is output directly when using LaTeX output.
\ *DF RM  Following text is output directly when using Markdown output.
\ *DF RT  Following text is output directly when Using Texinfo output.
\ *DF CD  A fixed font line, e.g. a code example
\ *DF QU  A quotation which may extend over several lines.
\ *DF C(  The following code will also be copied into the documentation.
\ ***     DocGen lines are handled as usual. You *\b{must} use the *] tag to
\ ***     stop copying - it is not halted by other tags. Do not use
\ ***     any other tags before copying is stopped.
\ *DF C)  Stops copying code into the documentation.
\ *DF HR  Inserts a line break, full width horizontal rule and a line break.
\ ***     Any text after the tag is ignored.
\ *DF IF  The text after the tag is *\fo{EVALUATE}d as Forth source.
\ ***     If the result is non-zero, DocGen carries on. If the result
\ ***     is zero, DocGen stops prodution until the next EL or TH tag.
\ *DF EL  If the previous IF result was non-zero, production stops until
\ ***     the next TH tag. If the previous IF result was zero, production
\ ***     restarts.
\ *DF TH  The end of a DocGen IF ... EL ... TH set of lines.
\ *DF EV  Evaluate the rest of the line as Forth source.
\ ***     Use this to define Forth flags that can later be tested
\ ***     by an IF tag. Out using *\fo{EMIT}, *\fo{TYPE} and
\ ***     friends goes to the output document.
\ *EX *EV 1 constant Windows?
\ *** *EV 0 constant Mac?
\ *** *EV 0 constant Embedded?


\ ------------------------
\ *H Single character tags
\ ------------------------
\ *D *  The following text is a continuation for the current style. This
\ **    can only be used after a G, E, D, B, Q or P tag.
\ *D !  Create and select a new output file. The control text is the
\ **    filename without extension. The remaining text is the description
\ **    which in HTML becomes the contents of the <title>text</title>
\ **    tag.
\ *D ^  Create and select a new output file. The control text is the
\ **    filename without extension. No header is added. This tag
\ **    allows you to generate completely custom formats in HTML
\ **    using W tags.
\ *D >  Select and append to an existing output file. The control text is
\ **    the filename without extension. No header is added.
\ *D #  Finalize the current HTML page. Outputs a canned HTML footer
\ **    which closes off the sections used by the canned header.
\ *D T  Following text is a section title.
\ *D S  Following text is a section sub-title.
\ *D N  Following text is a section sub-sub-title.
\ *D H  A simple heading.
\ *D D  Following text is a definition. The first space delimited token
\ **    is the term, the remaining text the description.
\ *D P  Begins a new paragraph.
\ *D E  Begins a paragraph which is a code example.
\ *D (  Starts a bullet list. The first character of the text
\ **    defines how the lists are marked. No text produces bullets,
\ **    *\b{1} produces a numbered list and *\b{A} or *\b{a} produce
\ **    lettered lists. Lists may be nested. *\b{N.B.} Previous
\ **    markups that did not use the *\b{*(} and *\b{*)} tags still
\ **    function as before, but the new layout is generally visually
\ **    better. Note that the LaTeX personality only handles bulleted
\ **    and numbered lists. See the change notes section of this
\ **    chapter for more details.
\ *D B  Following text is a bulleted entry.
\ *D )  Ends a bullet list.
\ *D G  Following text is a glossary entry. The preceeding line is
\ **    output in a fixed font code format.
\ *D R  Following text is output directly when using TeX output.
\ *D W  Following text is output directly when using HTML output.
\ *D C  A fixed font line, e.g. a code example
\ *D Q  A quotation which may extend over several lines.
\ *D [  The following code will also be copied into the documentation.
\ **    DocGen lines are handled as usual. You *\b{must} use the *] tag to
\ **    stop copying - it is not halted by other tags. Do not use
\ **    any other tags before copying is stopped.
\ *D ]  Stops copying code into the documentation.
\ *D L  Inserts a line break, full width horizontal line and a line break.
\ **    Any text after the tag is ignored.

\ ------------------------------
\ *H Tags reserved for DocGen/SC
\ ------------------------------
\ *P The following tags are reserved by MPE for use with DocGen/SC
\ ** for safety critical applications.
\ *C  *O *A *V *U *M *X *Y *Z
\ *P If you extend DocGen yourself, avoiding these tags will
\ ** help to ensure compatibility with future versions of DocGen.
\ ** DocGen/SC is described later in this chapter.

\ ====================
\ *N Formatting macros
\ ====================
\ *P From VFX Forth v3.70 onwards DocGen supports formatting commands that
\ ** can be included inside DocGen comments. Commands are of the form
\ *C   ... *\starbslash{}command{text} ...
\ *P The *\i{text} is processed and output. The following commands
\ ** are supported in all three default personalities.
\ *D bold     Displays text in *\b{bold}.
\ *D b        as bold
\ *D fixed    Displays text in a *\f{fixed (typewriter)} font.
\ *D f        as fixed
\ *D italic   Displays text in *\i{italic}.
\ *D i        as italic
\ *D forth    Displays text in *\fo{bold and fixed}.
\ *D fo       as forth
\ *D br       generates a line break. The text is ignored.
\ *D starbslash  Displays *\starbslash{}. The text is ignored.

\ *P These commands are used to change the formatting of small
\ ** pieces of text and make it easier to identify what is being
\ ** referred to. For example, a reference to a Forth word can be
\ ** produced by:
\ *C \ ** the word *\starbslash{}forth{DUP} is often used
\ *P which produces the result:
\ *C \ ** the word *\forth{DUP} is often used

\ ===============
\ *N Table macros
\ ===============
\ *P Tables are described using DocGen macros, which are *\b{not}
\ ** present in the LaTex personality.
\ *D table{list}  Starts a table. The *\i{list} contains an optional
\ **              caption delimited by '"' characters, followed by
\ **              space separated numbers. These are the percentage
\ **              widths of each column in the table, and so the
\ **              list also defines the number of columns in the
\ **              table. There may be up to 16 columns.
\ *D endtable{}	  Marks the end of the table.
\ *D hrow{}	  Marks the start of a header row. Header rows are
\ **              formatted differently from normal rows.
\ *D row{}	  Marks the start of a normal row.
\ *D col{}	  Marks the start of a new column in a row - you
\ **              do need this for the first column.

\ ===============
\ *N Image macros
\ ===============
\ *P An image (graphics file) is incorporated using the *\fo{image}
\ ** macro, which is *\b{not} present in the LaTex personality.
\ ** It is used in the form:
\ *E  *\image{"caption" "file" "ext" hmm w%}
\ *P where *\i{caption} is the text of the image caption;
\ ** *\i{file} is the basic file name with no extension;
\ ** *\i{ext} is the file extension; *\i{hmm} is the image height
\ ** in millimetres for Texinfo and LaTex; and *\i{w%} is the image
\ ** width in percent for HTML. All five parameters must be present.

\ *P Not all personalities will use all the parameters. For example,
\ ** when using *\i{pdftex} the file extension may be unused, and
\ ** *\i{pdf} images are always used. When using HTML output
\ ** the *\i{hmm} parameter is ignored, and the image size is
\ ** taken from the percentage width.

\ *P When preparing PDF images for use with *\i{pdftex}, note that
\ ** most tools generate PDF pages. Consequently, when preparing
\ ** an image for export as a PDF image, first select portrait or
\ ** landscape page format as appropriate, then scale the image
\ ** to fit the page, and then save it as a PDF file. OpenOffice
\ ** Draw is a suitable tool for converting most graphics files
\ ** into PDF format.


\ *****************************
\ *S Defining a new personality
\ *****************************
\ *P You can extend the DocGen system yourself. Many utility words
\ ** are available in the module *\fo{DOCGEN}. Because of the number
\ ** of words you will have to write, we recommend that new
\ ** personalities be defined in a *\fo{VOCABULARY} or a
\ ** *\fo{MODULE}. Use the DocGen source code in
\ ** *\i{Sources\Lib\Docgen4.fth} as a model. The previous version
\ ** is in *\i{Sources\Lib\Docgen3.fth}.

\ *P *\b{WARNING}: From build 1720 onwards DocGen specific utility
\ ** words defined in the *\fo{DOCGEN} module are no longer
\ ** *\fo{EXPORT}ed. You must now surround the code that uses
\ ** them with the following fragment:
\ *E expose-module docgen
\ **   ...
\ ** previous


\ ===================================
\ *N Personality description notation
\ ===================================
\ *P Defining a new personality consists of naming it, defining
\ ** the file extension that will be added to the output file
\ ** name, and then defining the tags that the personality will
\ ** react to. It is recommended that you provide actions for all
\ ** the tags provided as default, so that DocGen will still be
\ ** able to generate documentation using any of the predefined
\ ** personalities. The example below is for the default HTML
\ ** personality.

\ *E [DocGen Docgen_html
\ **   DG_FileExt: html
\ **
\ ** \ the continuation tag    entry   exit    continuation
\ **   DG_Tag: *       0       illegal illegal illegal
\ **
\ ** \ tags whose text must fit on one line
\ **   DG_Tag: !       -1      >newf   illegal illegal
\ **   DG_Tag: >       -1      >tofl   illegal illegal
\ **   DG_Tag: T       -1      >titl   illegal illegal
\ **   DG_Tag: S       -1      >sect   illegal illegal
\ **   DG_Tag: N       -1      >sstl   illegal illegal
\ **   DG_Tag: H       -1      >head   illegal illegal
\ **   DG_Tag: R       -1      2drop   illegal illegal
\ **   DG_Tag: W       -1      >rweb   illegal illegal
\ **   DG_Tag: C       -1      >code   illegal illegal
\ **
\ ** \ tags that can have continuation lines
\ **   DG_Tag: D       1       >defn   defn>   defn
\ **   DG_Tag: P       1       >para   para>   para
\ **   DG_Tag: E       1       >exam   exam>   exam
\ **   DG_Tag: B       1       >bult   bult>   bult
\ **   DG_Tag: G       1       >glos   glos>   glos
\ **   DG_Tag: Q       1       >quot   quot>   quot
\ **
\ ** DG_Type: HTMLtype
\ **
\ ** [DG_Macros
\ ** : bold             \ caddr u --
\ **   ." <B>"  prepro HTMLtype  ." </B>"
\ ** ;
\ **
\ ** : fixed            \ caddr u --
\ **   ." <TT>"  prepro HTMLtype  ." </TT>"
\ ** ;
\ **
\ ** : italic           \ caddr u --
\ **   ." <I>"  prepro HTMLtype  ." </I>"
\ ** ;
\ **
\ ** : forth            \ caddr u --
\ **   ." <B><TT>"  prepro HTMLtype  ." </TT></B>"
\ ** ;
\ **
\ ** : starbslash       \ caddr u --
\ **   2drop  S" *\" HTMLtype
\ ** ;
\ ** DG_Macros]
\ **
\ ** DocGen]

\ *P *\fo{[DOCGEN <name>} defines a new personality. When
\ ** *\fo{<name>} is executed, it becomes the current personality.

\ *P *\fo{DG_FileExt: <ext>} specifies the extension that will be
\ ** added to the output file names for this personality.

\ *P *\fo{DG_Tag:} specifies the character, control code, and
\ ** actions of a tag. The given character is used as the second
\ ** character of the '*x' pair. The second entry is the control
\ ** code and specifies the state conditions required, see later.
\ ** The next three entries are the names of the Forth words
\ ** executed on entry to the tag, on exit from that tag when
\ ** another tag is selected, and how that tag line should be
\ ** handled. See later for how the three action words should be
\ ** defined.

\ *P *\fo{DG_Type: <name>} specifies that *\fo{<name>} is the word
\ ** used for *\fo{TYPE}ing output.

\ *P The pair *\fo{[DG_Macros} and *\fo{DG_Macros]} delimit command
\ ** words defined in the personality's private wordlist. These
\ ** provide the actions of the command macros.

\ *P *\fo{DOCGEN]} marks the end of the personality description.


\ ======================
\ *N Using control codes
\ ======================
\ *P DocGen operates using three states - entry, exit and
\ ** continuation. When a new tag is encountered, the previous
\ ** tag's exit code is run if the control code is non-zero. The
\ ** exit code can for example add a *\f{</P>} paragraph end
\ ** marker to HTML code.

\ *P If the tag's control code is 1, the new tag becomes the
\ ** currently active tag so that the '**' tag knows what to do,
\ ** and the entry code for the new tag is run. This is the normal
\ ** condition for a tag that can have continuation lines.

\ *P If the tag's control code is -1, the entry code is run, and
\ ** the currently active tag state is set to none. A condition
\ ** code of -1 is used for tags whose following text must fit on
\ ** a single line, for example section titles.

\ *P If the tag's control code is 0, the continuation action of
\ ** the previous tag is run.

\ ===========================
\ *N Writing the action words
\ ===========================
\ *P The stack comments of the three action words are as follows. Text output
\ ** of words such as *\fo{EMIT} and *\fo{TYPE} has already been set to go to the last
\ ** defined output file.
\ *E : entry      \ caddr u -- ; consumes the text for the line
\ ** : continue   \ caddr u -- ; consumes the text for the line
\ ** : exit       \ -- ; close the tag

\ *P The utility word *\fo{PREPRO} is part of the primary definition and converts any
\ ** special characters in the DocGen comments. *\fo{PREPRO} has the stack effect:
\ *E caddr u -- caddr' u'
\ *P where caddr u defines the input text and caddr' u' defines the output text.
\ ** The MPE handlers use the following special characters apart
\ ** from those special to the output format itself.
\ *D 0x09   tab character, convert to spaces, tab width is set
\ **        to 8 characters
\ *D 0x1E   convert to open bracket character
\ *D 0x1F   convert to close bracket character

\ *P The following code is for the glossary *\f{*G} tag in the
\ ** HTML output. The word *\fo{LASTDEFINITIONLINE} returns the
\ ** text for the previous line.

\ *E : >glos         \ caddr u --
\ **   cr ." <PRE><CODE><B>" LastDefinitionLine prepro type
\ **   ." </B></CODE></PRE><P ALIGN=JUSTIFY>"
\ **   cr prepro type
\ ** ;
\ **
\ ** : glos          \ caddr u --
\ **   cr prepro type
\ ** ;
\ **
\ ** : glos>         \ --
\ **   cr ." </P>" cr
\ ** ;

\ *P The utility words *\fo{NewDocFile} and *\fo{SwitchDocFile} are provided to aid
\ ** construction of the *! and *> tags which create a new file or append to
\ ** an existing file. Both have the stack comment:
\ *E caddr u --
\ *P which is the file name without extension. NewDocFile starts a new
\ ** file, and you can add any required file entry code as required.
\ ** SwitchDocFile switches to the end of the named file.

\ *P The word *\fo{PREPRO} is available for processing text for
\ ** tags and formatting macros. It has the stack comment
\ *C   caddr len -- caddr' len'
\ *P The input text is processed and passed to an output buffer.
\ ** Special DocGen charactes (TAB, 0x1E, 0x1F) are converted and
\ ** format commands are processed.


\ ======================
\ *N Formatting commands
\ ======================
\ *P DocGen supports formatting commands that can be included
\ ** inside DocGen comments. Commands are of the form
\ *C   ... *\command{text} ...
\ *P Each command is handled by a Forth word defined in a private
\ ** *\fo{wordlist} for the personality. The *\b{command} word is
\ ** passed the string text, and has the stack comment
\ *C   caddr len --

\ *P Formatting commands use *\fo{."} and friends for output. The
\ ** text defined by caddr/len must be processed by *\fo{PREPRO}
\ ** before output. The string returned by *\fo{PREPRO} must be
\ ** output by a version of *\fo{TYPE} specific to the personality.
\ *C   caddr len --

\ *P The HTML version, called *\fo{HTMLtype} outputs the text
\ ** converting special characters to the form required by the
\ ** output format, e.g it must handle '<' and '>' for HTML and
\ ** '\' for Tex.


\ =============================
\ *N Personality words glossary
\ =============================

: ?defining	\ -- ; error if not defining
  defining? 0= dgn_NOTDEF ?Throw
;

: [DocGen	\ "<spaces>name" -- ; -- ; start new DOCGEN personality
\ *G Defines the start of a new personality for DocGen. See the example above
\ ** for details of the use of *\fo{[DOCGEN} and *\fo{DOCGEN]}.
  create
    here to defining?
    /Personality allot&erase
    #threads @  8 #threads !
    wordlist defining? p.wid !
    #threads !
    INVALID_HANDLE defining? p.fileid !
  does>
    to DG_Personality
;

: DocGen]	\ -- ; end personality
\ *G Marks the end of a new personality for DocGen. See the example above
\ ** for details of the use of *\fo{[DOCGEN} and *\fo{DOCGEN]}.
  ?defining  false to defining?
;

: DG_FileExt:   \ "<spaces>text" -- ; define document file extension
\ *G Defines the file extension for the DocGen personality being defined.
\ ** See the example above
\ ** for details of the use of *\fo{DG_FILEEXT:}.
  ?defining
\  bl word count
  parse-name
  dup MAX_EXT_LEN > dgn_EXTTOOLONG ?throw
  defining? p.$Ext place
;

: DG_Personality? \ xt -- ; check if xt is current personality
\ *G Check if xt is the active personality
\ *E ' docgen_html DG_Personality? [if] ... [then]
  >body DG_Personality =
;

((
: Tag		\ -- char1/char2
  bl word count  case
    1 of  c@  endof
    2 of  w@  endof
        1 abort" invalid tag"
  endcase
;
))

: DG_Tag:	\ "<char>" "<code>" "<enter>" "<exit>" "<continue>" --
\ *G Defines how a tag character is handled by DocGen.
\ ** See the example above
\ ** for details of the use of *\fo{DG_TAG:}
\ the data here MUST match the /TAG structure
  ?defining
  here defining? p.tagchain dup @ , !   \ link
  Tag ,					\ ascii-id
\  bl word count evaluate ,              \ control code
  parse-name evaluate ,                 \ control code
  ' ,  ' ,  ' ,                         \ enter, exit, continue XTs
;

: DG_Type:	\ "<word>" --
\ *G Specifies the word which performs the action of *\fo{TYPE}
\ ** ( addr len -- ) for this personality. Special characters
\ ** are translated, e.g. in HTML the '<' character is issued
\ ** as "&lt;". Use in the form:
\ *C  DG_Type: <name>
  ?defining  ' defining? p.type !
;

: [DG_Macros	\ -- oldcurrent
\ *G Marks the start of defining formatting macros.
  ?defining
  current @
  also defining? p.wid @ dup context ! current !
;

: DG_Macros]	\ oldcurrent --
\ *G Marks the end of defining formatting macros.
  ?defining
  previous  current !
;

: .DG_Macros	\ --
\ *G Show the available formatting macros in the current
\ ** personality.
  DG_Personality if
    context @
    DG_Personality p.wid @ context !
    words
    context !
  else
    cr ." No DOCGEN personality set"
  endif
;

: .DG_Tags	\ --
\ *G Show the available tags in the current personality.
  DG_Personality if
    DG_Personality p.tagchain @
    begin
      dup
     while
      dup tag.ascii-id @ emit
      tag.link @
    repeat
    drop
  else
    cr ." No DOCGEN personality set"
  endif
;

\ =======================
\ Table and Image support
\ =======================

\ SFP014...

0 value #Tables	\ --
\ Number of captioned tables encountered so far.
0 value #Figures	\ --
\ Number of captioned figures encountered so far.
0 value #cols	\ -- n
\ Number of columns in current table
#16 cells buffer: percentages	\ -- addr
\ Table of column widths in percentages of the table width.
#256 buffer: param1$	\ -- addr
\ Counted string containing the first text paparmeter.
#256 buffer: param2$	\ -- addr
\ Counted string containing the second text paparmeter.
#256 buffer: caption$	\ -- addr
\ Counted string containg the optional caption.
#256 buffer: imgfile$	\ -- addr
\ Counted string containing the image file name.
#256 buffer: alttext$	\ -- addr
\ Counted string used by the rawimage macro.
#16 buffer: imgext$	\ -- addr
\ Counted string containing the image file extension.
#16 buffer: H$	\ -- addr
\ Holds height text.
#16 buffer: W$	\ -- addr
\ Holds width text.
0 value Hmm	\ -- n
\ Image height in mm.
0 value W%	\ -- n
\ Image width in percent.

0 value col#	\ -- n
\ Current column number, 0..n-1.

0 value ColActive	\ -- flag
\ Set non zero after the first column has been defined in a row.
0 value RowActive	\ -- flag
\ Set non zero after the first row has been defined.
0 value hRow?	\ -- flag
\ Set non-zero for header rows.

: getTextParam	\ caddr len dest$ -- caddr' len'
  >r  [char] " ExtractText #255 min r> place
; doNotSin

: SetCaption	\ caddr len -- caddr' len'
  caption$ getTextParam
\  [char] " ExtractText #255 min caption$ place
;

: GetNumList	\ caddr len --  n1..nn n
  base @ >r  decimal
  depth 2- >r  evaluate  depth r> -	\ -- n1..nn n
  r> base !
;

: SetCols	\ caddr len --
\ Given the list text, generate the column information.
  GetNumList dup 1 < abort" No columns in table"
  dup #16 > abort" Too many columns"
  dup -> #cols  1- 0 swap do		\ fill table
    dup 0< abort" width too small"
    dup #100 > abort" width too great"
    percentages i cells + !
  -1 +loop
;

: .dec		\ n --
\ Display n in decimal without a trailing space.
  base @ >r  decimal  0 .r  r> base !
;

: ColWidth	\ -- percent
\ Return current column width as a percentage.
  percentages col# cells + @
;

: getImgFile	\ caddr len -- caddr' len'
  imgFile$ getTextParam
\  [char] " ExtractText #255 min imgfile$ place
;

: getImgExt	\ caddr len -- caddr' len'
  [char] " ExtractText #15 min imgext$ place
;

: getW$		\ caddr len -- caddr' len'
  [char] " ExtractText #15 min W$ place
;

: getH$		\ caddr len -- caddr' len'
  [char] " ExtractText #15 min H$ place
;

: getAltText$	\ caddr len -- caddr' len'
  alttext$ getTextParam
\  [char] " ExtractText #255 min alttext$ place
;

: SetImage	\ caddr len --
\ Process the image parameters.
  SetCaption  getImgFile getImgExt
  GetNumList 2 <> abort" Not enough image paramters"
  -> W%  -> Hmm
;

: SetRawImage	\ caddr len --
\ Process the rawimage parameters file w h alt ext
  getImgFile getW$ getH$ getAltText$ getImgExt 2drop
;
\ ...SFP014


\ ========
\ User API
\ ========

\ User words
  export +docgen
  export -docgen
  export +InternalDocs
  export -InternalDocs
  export DocOnly
  export parsed
  export [docgen
  export docgen]
  export DG_Personality?
  export DG_FileExt:
  export DG_Tag:
  export DG_Type:
  export DocGen:
  export DocGen-Spacing
  export [DG_Macros
  export DG_Macros]
  export .DG_Macros
  export .dg_tags
  export +TOC
  export +Index

\ used by DocGen personalities
((
  export dg-throw
  export ?dg-throw
  export OtherLines
  export doc-gio
  export prepro
  export NewDocFile
  export SwitchDocFile
  export LastDefinitionLine
  export DG_Personality
))

end-module


\ JED019...
\ ***************
\ *S HTML5 output
\ ***************
\ *P This personality generates HTML output. From June 2017
\ ** onwards, the output is HTML5 using a CSS stylesheet,
\ ** *\i{bootstrap.3.3.7.min.css}. The CSS file is "minified"
\ ** to reduce its size.

\ *P The '*!' tag generates a default HTML header. When this is
\ ** not needed, use the '*^' tag instead and no header is made.
\ ** The '*^' tag is often used whn a fully custom *\i{index.html}
\ ** is being used, for example to use *\i{iframe} tags.

\ *P Providing that the '*#' tag is used to provide the HTML5
\ ** footer, HTML5 files produced by DocGen will pass validation
\ ** testing.

expose-module DOCGEN
MODULE DOCGEN_HTML

CREATE BootstrapAddr	\ -- addr
  data-file %idir%/bootstrap.3.3.7.min.css
  constant BootstrapCnt

[-sin
: HTMLtype	\ c-addr u -- ; as TYPE, but processes HTML5 characters
  bounds ?do
    case  i c@
      [char] < of  ." &lt;"  endof
      [char] > of  ." &gt;"  endof
      [char] & of  ." &amp;"  endof
      [char] ' of  ." &#039;"  endof	\ HTML 4.0 compliant
      [char] " of  ." &quot;"  endof
        emit
    end-case
  loop
;

: HTMLcodetype	\ c-addr u -- ; as TYPE, but processes HTML5 characters
  bounds ?do
    case  i c@
      bl       of  ." &nbsp;"  endof
      [char] < of  ." &lt;"  endof
      [char] > of  ." &gt;"  endof
      [char] & of  ." &amp;"  endof
        emit
    end-case
  loop
;

: opText	\ caddr u -- ; output processed text
  prepro HTMLtype
;

: opLine	\ caddr u -- ; output CR and processed text
  cr prepro HTMLtype
;

\ SFP012...
: ShortFilename	\ caddr len -- caddr' len'
\ Return name after last directory separator.
  dup if
    2dup + 0 2swap
    over + 1- do		\ -- caddr' len'
      i c@ dirChar? ?leave
      swap 1- swap 1+
    -1 +loop
  endif
;

: -Extension	\ caddr len -- caddr len'
  dup if
    2dup extension?
    if  1+ -  endif
  endif
;

: IsIndex?	\ -- flag
\ Return true if the current file is INDEX.xxx
  CurrOpFile$ count ShortFilename -Extension
  s" index" icompare 0=
;

#32 buffer: HTMLback$	\ -- caddr
\ Contains string for the HTML background colour.

: .bgcolor	\ --
\ Output the standard background colour.
  ." <HTML>"  IsIndex? 0= if
    ." <BODY bgcolor=" HTMLback$ $.  ." >"
  endif
;
\ ...SFP012

\ SFP014...
: endCol	\ --
\ Performed at the end of a column item
  ColActive if
    4 spaces  hRow?
    if  ." </th>" cr  else  ." </td>" cr  endif
  endif
;

: endRow	\ --
\ Performed at the end of a row.
  RowActive if
    endCol  ."   </tr>" cr
  endif
;

: anyRow	\ --
\ Common code for row and hrow.
  endRow  ."   <tr>" cr
  0 -> ColActive  -1 -> RowActive
  -1 -> col#
;
\ ...SFP014

: .HTMLHeader   \ caddr u --
\ Layout the HTML5 head node and the start of the body node.
     .\" <!DOCTYPE html>"
  cr .\" <html lang=\"en\">"
  cr .\" <head>"
  cr
  cr .\" <meta charset=\"utf-8\">"
  cr .\" <meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">"
  cr .\" <meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">"
  cr .\"     <!-- The above 3 meta tags *must* come first in the head -->"
  cr .\" <title>" type .\" </title>"
  cr
  cr .\" <!-- Bootstrap core CSS -->"
  cr .\" <link href=\"bootstrap.3.3.7.min.css\" rel=\"stylesheet\">"
  cr
  cr .\" </head>"
  cr
  cr .\" <body>"
  cr
  cr .\" <div class=\"container\">"
  cr
  cr .\" <div class=\"starter-template col-md-8 col-md-offset-2\">"
;

: .HTMLFooter
\ Close the html5 document.
  cr
  cr .\" </div>"        \ close col-md-8
  cr
  cr .\" </div>"        \ close container
  cr .\" </body>"
  cr .\" </html>"
;

: MakeBootstrapFile     \ --
\ *G HTML5 documents need a CSS file to output correctly.
\ ** This word generates the DocGen default CSS file,
\ ** *\i{bootstrap.3.3.7.min.css} in the current directory.
\ ** The file is only created if it does not exist.
  s" bootstrap.3.3.7.min.css" 2dup fileexist? if
    2drop
  else
    r/w create-file throw >r
    BootstrapAddr BootstrapCnt r@ write-file throw
    r> close-file throw
  endif
;

: >newf         \ caddr u -- ; *!
\ New file with header
  newfExpand NewDocFile .HTMLHeader cr
  MakeBootstrapFile
;

: >tofl		\ caddr u -- ; *>
\ Append to end of file. If the file is created, it will be empty.
  SwitchDocFile
;

: >simf         \ caddr u -- ; *^
\ Creates a new empty HTML file.
\ This is used when a user wants to create their own look and feel.
\ Used in conjuction with "\ *W ".
  NewDocFile
  MakeBootstrapFile
;

: >endf         \ caddr u -- ; *#
  2drop .HTMLFooter cr
;

sin]
[-sin

: StartLink	\ $ --
\ Generate a reference in the TOC or INDEX file to the current
\ location in the current output file. The string defines the
\ name of the link in the form:
\  file#name
  ." <A HREF=" [char] " emit	\ Reference in index/TOC
  CurrOpfile$ count type  [char] # emit  count type
  [char] " emit ." >"
;

: EndLink	\ --
\ finish a link
  ." </A>"
;

: CRindent	\ n --
\ Start a new line indented by n CODE characters.
  cr  ." <BR>"  0			\ indent
  ?do  ." &nbsp;"  loop
;

: NameDef	\ $ --
\ Make a name definition in the current output file.
  \ ." <A NAME=" [char] " emit
  ." <A id=" [char] " emit
  count opText  [char] " emit
  ." ></A>"
;

: .H1		\ caddr u --
\ Write the string as a level 1 heading
  cr ." <H1>"
  cr opText
  cr ." </H1>" cr		\ SFP007
  cr
;

: .toc		\ caddr u n --
\ Output TOC entry in TOC and o/p files at level n
  { caddr u n | toc$[ max_path ] -- }
  n +Toc$ toc$[ place
  [toc
    TocStarted? 0= if
      s" Table of Contents" .HTMLHeader cr
      s" Table of Contents" .H1
      true -> TocStarted?
    endif
    n 1- 2* CRindent			\ indented new line
    toc$[ StartLink			\ link
    caddr u opText
    EndLink
  DGop]
  cr  toc$[ NameDef
;

: >titl		\ caddr u --
  toc? if  2dup 1 .toc  endif
  .H1
;

: >sect		\ caddr u --
  toc? if  2dup 2 .toc  endif
  cr ." <H2>"
  cr opText
  cr ." </H2>" cr		\ SFP007
;

: >sstl         \ caddr u --
  toc? if  2dup 3 .toc  endif
  cr ." <H3>"
  opLine
  cr ." </H3>" cr		\ SFP007
;

: >head         \ caddr u --
  toc? if  2dup 4 .toc  endif
  cr ." <H4>" opText ." </H4>"
;

: >defn		\ caddr u --
  cr .\" <TABLE STYLE=\qBORDER:0px; WIDTH:100%\q><TR><TD STYLE=\qWIDTH:10%; VALIGN:TOP\q><P><I>"
  2dup 2dup bl scan nip - opText
  cr ." </I></P></TD>"
  cr .\" <TD STYLE=\qWIDTH:5%\q></TD>"
  cr .\" <TD STYLE=\qWIDTH:90%\q><P STYLE=\qALIGN:JUSTIFY\q>"
  bl scan opText
;

: defn		\ caddr u --
  opText
;

: defn>		\ --
  cr ." </P></TD></TR></TABLE>" cr
;

: >para         \ caddr u --
  cr .\" <P class=\qtext-justify\q>" opText
;

: para		\ caddr u --
  opLine
;

: para>         \ --
  cr ." </P>" cr
;

: >exam         \ caddr u --
  cr .\" <PRE><CODE>"
  opLine
;

: exam          \ caddr u --
  opLine
;

: exam>         \ --
  cr ." </CODE></PRE>" cr
;

: >[list	\ caddr u --
  cr
  case  NewList
    [char] 1 of  s\" <ol start=1 type=\q1\q>"  endof
    [char] A of  s\" <ol start=1 type=\qA\q>"  endof
    [char] a of  s\" <ol start=1 type=\qa\q>"  endof
      drop  s" <ul>"
  end-case
  type
;

: >list]	\ caddr u --
  2drop  cr
  case  BultStk A>
    [char] 1 of  ." </ol>"  endof
    [char] A of  ." </ol>"  endof
    [char] a of  ." </ol>"  endof
      ." </ul>"
  endcase
  cr
;

: >bult         \ caddr u --
  cr ." <LI>" opText
;

: bult          \ caddr u --
  opLine
;

: bult>         \ --
  ." </LI>"
;
sin]

[-sin
[-short-branches
: .idx		\ --
\ Output index entry in index and o/p files.
  { | idx$[ max_path ] -- }
  +idx$ idx$[ place
  [idx
    IdxStarted? 0= if
\ The next two lines would upset sorting of the index file.
\ Extending the line with blanks is not enough.
\      .bgcolor cr
\      s" Index" .H1
      true -> IdxStarted?
    endif
    cr ." <BR>" idx$[ StartLink
    #50 >pos  ParseGloName HTMLtype
    EndLink
  DGop]
  cr  idx$[ NameDef
;
short-branches]
sin]

: >glos         \ caddr u --
  index? if  .idx  endif
  cr .\" <P>"
  cr ." <CODE><B>"
  LastDefinitionLine -white prepro HTMLcodetype
  ." </B></CODE>"
  cr ." <BR>" opLine
;

: glos          \ caddr u --
  opLine
;

: glos>         \ --
  cr ." </P>" cr
;

: >quot         \ caddr u --
  ." <P><I>" opLine
;

: quot          \ caddr u --
  opLine
;

: quot>         \ --
  ." </I></P>" cr
;

: >code         \ caddr u --
  ." <PRE>" opText ." </PRE>" cr
;

: >rweb         \ caddr u --
  prepro type cr
;

: .codeline	\ -- ; output a code line to the documentation
  op-handle @ >r			\ switch o/p to file
  doc-gio op-handle !
  source prepro -trailing HTMLtype cr
  r> op-handle !			\ restore o/p
;

: >+cpy		\ caddr u -- ; start output of code
  prepro HTMLtype ." <PRE>" cr
  assign .codeline to-do OtherLines
;

: >-cpy		\ caddr u -- ; end output of code
  2drop  ." </PRE>" cr
  assign noop to-do OtherLines
;

: >rule		\ caddr u -- ; horizontal line
  2drop  cr ." <BR><HR><BR>" cr
;

[DocGen Docgen_html    \ -- ; -- ; select HTML5 for DOCGEN
\ *G Makes the HTML5 personality the current personality for
\ ** DocGen. HTML5 will remain the current personality until
\ ** another personality is selected.
  DG_FileExt:  html

\ the continuation tag
  DG_Tag: *       0       illegal illegal illegal
\ tags whose text must fit on a single line
  DG_Tag: !       -1      >newf   illegal illegal
  DG_Tag: >       -1      >tofl   illegal illegal
  DG_Tag: ^       -1      >simf   illegal illegal
  DG_Tag: #       -1      >endf   illegal illegal
  DG_Tag: T       -1      >titl   illegal illegal
  DG_Tag: S       -1      >sect   illegal illegal
  DG_Tag: N       -1      >sstl   illegal illegal
  DG_Tag: H       -1      >head   illegal illegal
  DG_Tag: R       -1      2drop   illegal illegal
  DG_Tag: W       -1      >rweb   illegal illegal
  DG_Tag: C       -1      >code   illegal illegal
  DG_Tag: [       -1      >+cpy   illegal illegal
  DG_Tag: ]       -1      >-cpy   illegal illegal
  DG_Tag: L       -1      >rule   illegal illegal
  DG_Tag: (       -1      >[list  illegal illegal	\ )
  DG_Tag: )       -1      >list]  illegal illegal
\ tags which permit continuation lines
  DG_Tag: D       1       >defn   defn>   defn
  DG_Tag: P       1       >para   para>   para
  DG_Tag: E       1       >exam   exam>   exam
  DG_Tag: B       1       >bult   bult>   bult
  DG_Tag: G       1       >glos   glos>   glos
  DG_Tag: Q       1       >quot   quot>   quot

\ the continuation tag
  DG_Tag: **      0       illegal illegal illegal	\ continuation
\ tags whose text must fit on a single line
  DG_Tag: !!      -1      >newf   illegal illegal
  DG_Tag: >>      -1      >tofl   illegal illegal
  DG_Tag: ^^      -1      >simf   illegal illegal
  DG_Tag: ##      -1      >endf   illegal illegal
  DG_Tag: CH      -1      >titl   illegal illegal	\ chapter
  DG_Tag: SE      -1      >sect   illegal illegal	\ section
  DG_Tag: SS      -1      >sstl   illegal illegal	\ sub-section
  DG_Tag: SH      -1      >head   illegal illegal	\ sub-head
  DG_Tag: RH      -1      >rweb   illegal illegal	\ raw HTML
  DG_Tag: RL      -1      2drop   illegal illegal	\ raw LaTeX
  DG_Tag: RM      -1      2drop   illegal illegal	\ raw Markdown
  DG_Tag: RT      -1      2drop   illegal illegal	\ raw TexInfo
  DG_Tag: CD      -1      >code   illegal illegal	\ code
  DG_Tag: C(      -1      >+cpy   illegal illegal	\ copy start
  DG_Tag: C)      -1      >-cpy   illegal illegal	\ copy end
  DG_Tag: HR      -1      >rule   illegal illegal	\ horizontal rule
  DG_Tag: L(      -1      >[list  illegal illegal	\ list start )
  DG_Tag: L)      -1      >list]  illegal illegal	\ list end
  DG_Tag: EV      -1      >ev     illegal illegal	\ evaluate
  DG_Tag: IF      -1      >if     illegal illegal	\ IF text
  DG_Tag: EL      -1      >else   illegal illegal	\ ELSE
  DG_Tag: TH      -1      >then   illegal illegal	\ THEN
\ tags which permit continuation lines
  DG_Tag: DF      1       >defn   defn>   defn
  DG_Tag: PA      1       >para   para>   para
  DG_Tag: EX      1       >exam   exam>   exam
  DG_Tag: IT      1       >bult   bult>   bult
  DG_Tag: GL      1       >glos   glos>   glos
  DG_Tag: QU      1       >quot   quot>   quot

DG_Type: HTMLtype

\ SFP012...
: HTMLback	\ caddr len --
\ *G Sets the HTML background colour for the next output file.
\ ** The string is the HTML colour reference (limited to 31
\ ** characters), e.g.
\ *C   s" #00C1B4" HTMLback
  #31 min HTMLback$ place
;
  s" #D0D0D0" HTMLback
\ ...SFP012

\ ===============
\ *N HTML5 macros
\ ===============

[DG_Macros
: bold		\ caddr u --
\ *G Macro to display string in bold text.
\ *C   *\starbslash{}bold{text in bold}
  ." <B>" opText ." </B>"
;

: fixed		\ caddr u --
\ *G Macro to display string in a fixed-width font.
\ *C   *\starbslash{}fixed{text in fixed font}
  ." <CODE>" opText ." </CODE>"		\ no <TT> in HTML5
;

: italic	\ caddr u --
\ *G Macro to display string in an italic font.
\ *C   *\starbslash{}italic{text in italic font}
  ." <I>" opText ." </I>"
;

: forth		\ caddr u --
\ *G Macro to display text as Forth source - bold and fixed font.
\ *C   *\starbslash{}forth{text in Forth font}
  ." <B><CODE>" opText ." </CODE></B>"
;

: br		\ caddr u --
\ *G Macro to insert a line break. Any text is ignored.
\ *C   *\starbslash{}br{}
  2drop  ." <BR>" cr
;

: starbslash	\ caddr u --
\ *G Macro to insert *\'. Any text is ignored.
\ *C   *\starbslash{}starbslash{}
  2drop  S" *\" HTMLtype
;

\ SFP014...
: table		\ caddr len --
\ *G Table start macro. The text is a quoted caption followed
\ ** by a list of column widths in per-cent. For a table with
\ ** three columns, use (say)
\ *C   *\starbslash{}table{ "Table caption" 20 20 60 }
  SetCaption SetCols  cr ." <table border>" cr
  0 -> RowActive  0 -> ColActive
;

: endtable	\ caddr len --
\ *G Table end macro. Any text is discarded
\ *C   *\starbslash{}endtable{}
  2drop  endRow  ." </table>" cr cr
;

: row		\ caddr len --
\ *G Macro that marks the start of a normal row. Any text is
\ ** discarded.
\ *C   *\starbslash{}row{}
  2drop  anyRow  0 -> hRow?
;

: hrow		\ caddr len --
\ *G Macro that marks the start of a header row. Any text is
\ ** discarded.
\ *C   *\starbslash{}hrow{}
  2drop  anyRow  -1 -> hRow?
;

: col		\ caddr len --
\ *G Macro that marks the start of a column. Any text is
\ ** discarded.
\ *C   *\starbslash{}col{}
  2drop  endCol  4 spaces  hRow?
  if  ." <th "  else  ." <td "  endif
  Col# 1+ -> Col#
  .\" width=\q" ColWidth .dec .\" %\q>"
  -1 -> ColActive
;

: image		\ caddr len --
\ *G Macro to display an image from a file.
\ *C  *\starbslash{}image{"caption" "file" "ext" hmm w%}
\ *P where *\i{caption} is the text of the image caption;
\ ** *\i{file} is the basic file name with no extension;
\ ** *\i{ext} is the file extension; *\i{hmm} is ignored for HTML;
\ ** and *\i{w%} is the image
\ ** width in percent for HTML. All five parameters must be present.
  SetImage
  cr ." <p><center>"
  cr s\" <img src=\q" type
  imgfile$ count type  ." ."  imgext$ count type
  s\" \q width=\q" type  w% .dec  s\" %\q>" type
  caption$ c@ if
    cr ." <br><i>"  caption$ count type  ." </i>"
  endif
  ." </center></p>" cr
;
\ ...SFP014

: lname		\ caddr len --
\ *G HTML specific. Defines the current location as a point in the file that
\ ** can be jumped to later by a *\fo{link}. The text is the
\ ** name of the location.
\ *C   *\starbslash{}lname{ linkname }
  .\" <a id=\q" type .\" \q></a>"
;
: link		\ caddr len --
\ *G HTML specific. Create a link to a previous *\fo{lname} in the current file.
\ ** the link is of the form "#linkname" and so must be in the
\ ** current output file.
\ *C   *\starbslash{}link{linkname}
  .\" <a href=\q#" 2dup type .\" \q>"  type  ." </a>"
;

: htarget		\ caddr len --
\ *G Defines the current location as a point in the file that
\ ** can be jumped to later by a *\fo{href}. The caption text
\ ** is optional.
\ *C   *\starbslash{}htarget{ "targetname" "caption" }
  param1$ getTextParam  param2$ getTextParam  2drop
  .\" <a id=\q" param1$ count type .\" \q> param2$ count type </a>"
;
: href		\ caddr len --
\ *G Create a link to another file or object. For HTML, the URL
\ ** text is what is linked to and the display text is what is
\ ** shown in the link.
\ *C   *\starbslash{}href{"url" "display"}
  param1$ getTextParam  param2$ getTextParam  2drop
  .\" <a href=\q" param1$ count type .\" \q>"  param2$ count type  ." </a>"
;

\ Must be last because I redefines a standard Forth word.
: b   bold  ;
\ *G Convenient redefinition.
: f   fixed  ;
\ *G Convenient redefinition.
: i   italic  ;
\ *G Convenient redefinition.
: fo  forth  ;
\ *G Convenient redefinition.

DG_Macros]

DocGen]

EXPORT docgen_html
EXPORT HTMLback
EXPORT MakeBootstrapFile


END-MODULE
previous
\ ...JED019

DocGen_HTML	\ Make this the default.

\ JED022...
\ ******************
\ *S Markdown output
\ ******************
\ *P This personality generates Markdown output. It does
\ ** not yet support all the features of the other
\ ** personalities; in particular it does not provide a table of contents
\ ** or an index.

\ *P To view Markdown text, the most convenient solution is probably to
\ ** add the Markdown Viewer to Google Chrome. Remember to check the box
\ ** to enable access to file URLs.

expose-module DOCGEN
MODULE DOCGEN_MARKDOWN

[-sin
: MARKDOWNtype	\ c-addr u -- ; as TYPE, but processes markdown characters
  bounds ?do
    case  i c@
      [char] \ of ." \\" endof
      [char] ` of ." \`" endof
      [char] * of ." \*" endof
      [char] _ of ." \_" endof
      [char] { of ." \{" endof
      [char] } of ." \}" endof
      [char] [ of ." \[" endof
      [char] ] of ." \]" endof
      [char] ( of ." \(" endof
      [char] ) of ." \)" endof
      [char] # of ." \#" endof
      [char] + of ." \+" endof
      [char] - of ." \-" endof
      [char] . of ." \." endof
      [char] ! of ." \!" endof
      [char] | of ." &#124;" endof      \ JED024
        emit
    end-case
  loop
;

: opText	\ caddr u -- ; output processed text
  prepro MARKDOWNtype
;

: opLine	\ caddr u -- ; output CR and processed text
  cr prepro MARKDOWNtype
;

\ SFP012...
: ShortFilename	\ caddr len -- caddr' len'
\ Return name after last directory separator.
  dup if
    2dup + 0 2swap
    over + 1- do		\ -- caddr' len'
      i c@ dirChar? ?leave
      swap 1- swap 1+
    -1 +loop
  endif
;

: -Extension	\ caddr len -- caddr len'
  dup if
    2dup extension?
    if  1+ -  endif
  endif
;

: mdHdrSep. \ cols# ---
\ Print markdown table header seperator.
  dup if
    ." |"
    0 do
      ."  --- |"
    loop
  else drop
  endif
;

\ SFP014...
: endCol	\ --
\ Performed at the end of a column item
  ColActive
  if  4 spaces ." |"  endif
;

: endRow	\ --
\ Performed at the end of a row.
  RowActive
  if  endCol  endif
  hrow?
  if  cr #cols mdHdrSep.  endif
;

: anyRow	\ --
\ Common code for row and hrow.
  endRow cr
  0 -> ColActive  -1 -> RowActive
  -1 -> col#
;
\ ...SFP014

: >newf         \ caddr u -- ; *!
\ New file with header
  newfExpand NewDocFile 2drop cr
;

: >tofl		\ caddr u -- ; *>
\ Append to end of file. If the file is created, it will be empty.
  SwitchDocFile
;

: .H1		\ caddr u --
\ Write the string as a level 1 heading
  cr cr ." # " opText cr
;

: >titl		\ caddr u --
  \ toc? if  2dup 1 .toc  endif
  .H1
;

: >sect		\ caddr u --
  cr cr ." ## " opText cr
;

: >sstl         \ caddr u --
  cr cr ." ### "  opText cr
;

: >head         \ caddr u --
  cr cr ." #### " opText cr
;

: >defn		\ caddr u --
  cr
  ." | Term | Description |" cr
  ." | --- | --- |" cr
  ." | " 2dup 2dup bl scan nip - opText ."  | " bl scan opText ."  | " cr
;

: defn		\ caddr u --
  ." |   | " opText ."  | " cr
;

: defn>		\ --
  cr cr
;

: >para         \ caddr u --
  cr opText
;

: para		\ caddr u --
  ."  " opText          \ JED024
;

: para>         \ --
  cr
;

: >exam         \ caddr u --
  cr  ." ```" cr
  prepro type
;

: exam          \ caddr u --
  cr prepro type
;

: exam>         \ --
  cr ." ```" cr
;

: >[list	\ caddr u --
  cr Newlist drop
  \  ." BultStk adepth :" BultStk adepth .
;

: >list]	\ caddr u --
  2drop BultStk A> drop
  cr
;

[-short-branches
: >bult         \ caddr u --
  \ ." in >bult: BultStk adepth :" BultStk adepth .
  BultStk adepth dup if
    cr 1- 4 * spaces
    BultStk A@ [char] 1 = if     \ md only does numbered lists
      ." 1. " opText
    else
      ." - " opText              \ just do unordered
    endif
    \ ." in >bult: BultStk adepth :" BultStk adepth .
  else       			\ old bullet lists
    drop cr ." - " opText
    \ ." in >bult: BultStk adepth :" BultStk adepth .
  endif
;
short-branches]

: bult          \ caddr u --
  opLine
;

: bult>         \ --
;
sin]

: >glos         \ caddr u --
  cr ." --------------------------------------------------------------------------------" cr
  ." ```" cr
  LastDefinitionLine -white prepro type
  cr ." ```"
  opLine
;

: glos          \ caddr u --
  opLine
;

: glos>         \ --
  cr
;

: >quot         \ caddr u --
  cr ." _" opLine
;

: quot          \ caddr u --
  opLine
;

: quot>         \ --
  ." _" cr      \ ." </I></P>" cr
;

: >code         \ caddr u --
  cr ."   " prepro type  cr
;

: .codeline	\ -- ; output a code line to the documentation
  op-handle @ >r			\ switch o/p to file
  doc-gio op-handle !
  source prepro -trailing type cr
  r> op-handle !			\ restore o/p
;

: >+cpy		\ caddr u -- ; start output of code
  prepro type ." ```" cr
  assign .codeline to-do OtherLines
;

: >-cpy		\ caddr u -- ; end output of code
  2drop  ." ```" cr
  assign noop to-do OtherLines
;

: >rule		\ caddr u -- ; horizontal line
  2drop cr ." ---" cr
;

[DocGen Docgen_markdown    \ -- ; -- ; select Markdown for DOCGEN
\ *G Makes the Markdown personality the current personality for
\ ** DocGen. Markdown will remain the current personality until
\ ** another personality is selected.
  DG_FileExt:  md

\ the continuation tag
  DG_Tag: *       0       illegal illegal illegal
\ tags whose text must fit on a single line
  DG_Tag: !       -1      >newf   illegal illegal
  DG_Tag: >       -1      >tofl   illegal illegal
  DG_Tag: ^       -1      2drop   illegal illegal
  DG_Tag: #       -1      2drop   illegal illegal
  DG_Tag: T       -1      >titl   illegal illegal
  DG_Tag: S       -1      >sect   illegal illegal
  DG_Tag: N       -1      >sstl   illegal illegal
  DG_Tag: H       -1      >head   illegal illegal
  DG_Tag: R       -1      2drop   illegal illegal
  DG_Tag: W       -1      2drop   illegal illegal
  DG_Tag: C       -1      >code   illegal illegal
  DG_Tag: [       -1      >+cpy   illegal illegal
  DG_Tag: ]       -1      >-cpy   illegal illegal
  DG_Tag: L       -1      >rule   illegal illegal
  DG_Tag: (       -1      >[list  illegal illegal	\ )
  DG_Tag: )       -1      >list]  illegal illegal
\ tags which permit continuation lines
  DG_Tag: D       1       >defn   defn>   defn
  DG_Tag: P       1       >para   para>   para
  DG_Tag: E       1       >exam   exam>   exam
  DG_Tag: B       1       >bult   bult>   bult
  DG_Tag: G       1       >glos   glos>   glos
  DG_Tag: Q       1       >quot   quot>   quot

\ the continuation tag
  DG_Tag: **      0       illegal illegal illegal	\ continuation
\ tags whose text must fit on a single line
  DG_Tag: !!      -1      >newf   illegal illegal
  DG_Tag: >>      -1      >tofl   illegal illegal
  DG_Tag: ^^      -1      2drop   illegal illegal
  DG_Tag: ##      -1      2drop   illegal illegal
  DG_Tag: CH      -1      >titl   illegal illegal	\ chapter
  DG_Tag: SE      -1      >sect   illegal illegal	\ section
  DG_Tag: SS      -1      >sstl   illegal illegal	\ sub-section
  DG_Tag: SH      -1      >head   illegal illegal	\ sub-head
  DG_Tag: RH      -1      2drop   illegal illegal	\ raw HTML
  DG_Tag: RL      -1      2drop   illegal illegal	\ raw LaTeX
  DG_Tag: RM      -1      2drop   illegal illegal	\ raw Markdown
  DG_Tag: RT      -1      2drop   illegal illegal	\ raw TexInfo
  DG_Tag: CD      -1      >code   illegal illegal	\ code
  DG_Tag: C(      -1      >+cpy   illegal illegal	\ copy start
  DG_Tag: C)      -1      >-cpy   illegal illegal	\ copy end
  DG_Tag: HR      -1      >rule   illegal illegal	\ horizontal rule
  DG_Tag: L(      -1      >[list  illegal illegal	\ list start )
  DG_Tag: L)      -1      >list]  illegal illegal	\ list end
  DG_Tag: EV      -1      >ev     illegal illegal	\ evaluate
  DG_Tag: IF      -1      >if     illegal illegal	\ IF text
  DG_Tag: EL      -1      >else   illegal illegal	\ ELSE
  DG_Tag: TH      -1      >then   illegal illegal	\ THEN
\ tags which permit continuation lines
  DG_Tag: DF      1       >defn   defn>   defn
  DG_Tag: PA      1       >para   para>   para
  DG_Tag: EX      1       >exam   exam>   exam
  DG_Tag: IT      1       >bult   bult>   bult
  DG_Tag: GL      1       >glos   glos>   glos
  DG_Tag: QU      1       >quot   quot>   quot

DG_Type: MARKDOWNtype

\ ==================
\ *N Markdown macros
\ ==================

[DG_Macros
: bold		\ caddr u --
\ *G Macro to display string in bold text.
\ *C   *\starbslash{}bold{text in bold}
  ." **" prepro type ." **"     \ JED024
;

: fixed		\ caddr u --
\ *G Macro to display string in a fixed-width font.
\ *C   *\starbslash{}fixed{text in fixed font}
  ." `" prepro type ." `"       \ JED024
;

: italic	\ caddr u --
\ *G Macro to display string in an italic font.
\ *C   *\starbslash{}italic{text in italic font}
  ." _" prepro type ." _"       \ JED024
;

: forth		\ caddr u --
\ *G Macro to display text as Forth source - bold and fixed font.
\ *C   *\starbslash{}forth{text in Forth font}
  ." **`" prepro type ." `**"   \ JED024
;

: br		\ caddr u --
\ *G Macro to insert a line break. Any text is ignored.
\ *C   *\starbslash{}br{}
  2drop  ."   " cr
;

: starbslash	\ caddr u --
\ *G Macro to insert '*\'. Any text is ignored.
\ *C   *\starbslash{}starbslash{}
  2drop  S" *\" MARKDOWNtype
;

\ SFP014...
: table		\ caddr len --
\ *G Table start macro. The text is a quoted caption followed
\ ** by a list of column widths in per-cent. For a table with
\ ** three columns, use (say)
\ *C   *\starbslash{}table{ "Table caption" 20 20 60 }
  SetCaption SetCols  cr
  0 -> RowActive  0 -> ColActive
;

: endtable	\ caddr len --
\ *G Table end macro. Any text is discarded
\ *C   *\starbslash{}endtable{}
  2drop  endRow  cr cr
;

: row		\ caddr len --
\ *G Macro that marks the start of a normal row. Any text is
\ ** discarded.
\ *C   *\starbslash{}row{}
  2drop  anyRow  0 -> hRow?
  ." |"
;

: hrow		\ caddr len --
\ *G Macro that marks the start of a header row. Any text is
\ ** discarded.
\ *C   *\starbslash{}hrow{}
  2drop  anyRow  -1 -> hRow?
  ." |"
;

: col		\ caddr len --
\ *G Macro that marks the start of a column. Any text is
\ ** discarded.
\ *C   *\starbslash{}col{}
  2drop  endCol  4 spaces
  Col# 1+ -> Col#
  -1 -> ColActive
;

: image		\ caddr len --
\ *G Macro to display an image from a file.
\ *C  *\starbslash{}image{"caption" "file" "ext" hmm w%}
\ *P where *\i{caption} is the text of the image caption;
\ ** *\i{file} is the basic file name with no extension;
\ ** *\i{ext} is the file extension; *\i{hmm} is ignored for HTML;
\ ** and *\i{w%} is the image
\ ** width in percent for HTML. All five parameters must be present.
  SetImage
  cr
  cr ." !["
  caption$ c@ if
    caption$ count type
  endif
  ." ](" imgfile$ count type  ." ."  imgext$ count type ."  "
  caption$ c@ if
    .\" \"" caption$ count type .\" \""
  endif
  ." )"
  cr
;

((
  : htarget		\ caddr len --
  \ *G Defines the current location as a point in the file that
  \ ** can be jumped to later by a *\fo{href}. The caption text
  \ ** is optional.
  \ *C   *\starbslash{}htarget{ "targetname" "caption" }
    param1$ getTextParam  param2$ getTextParam  2drop
    .\" <a id=\q" param1$ count type .\" \q> param2$ count type </a>"
  ;
))
: href		\ caddr len --
\ *G Create a link to another file or object. For HTML, the URL
\ ** text is what is linked to and the display text is what is
\ ** shown in the link.
\ *C   *\starbslash{}href{"url" "display"}
  param1$ getTextParam  param2$ getTextParam  2drop
  ." [" param2$ count type ." ](" param1$ count type ." ) cr
;

\ Must be last because I redefines a standard Forth word.

: b   bold  ;
\ *G Convenient redefinition.
: f   fixed  ;
\ *G Convenient redefinition.
: i   italic  ;
\ *G Convenient redefinition.
: fo  forth  ;
\ *G Convenient redefinition.

DG_Macros]

DocGen]

EXPORT docgen_markdown

END-MODULE
previous
\ ...JED022


\ ******************************
\ *S TeX output with texinfo.tex
\ ******************************
\ *P This personality generates TeXinfo output, using the file
\ ** texinfo.tex supplied with many TeX distributions. From TeX
\ ** you can get to many other formats including PDF. Although
\ ** we originally recommended

\ *P TeXinfo is under-documented, especially in terms of
\ ** examples, and has quirks. Despite this, Texinfo can
\ ** generate good quality PDF files with bookmarks and
\ ** thumbnails, a table of contents and an index. The MikTex CD package
\ ** (see below) includes the Texinfo manual file
\ ** *\i{doc\texinfo\texinfo.dvi}
\ ** which can be viewed using *\i{bin\yap.exe}. Texinfo is a
\ ** real "techie" extension to LaTex, but the more we use it,
\ ** the more we appreciate it. It is worth the admittedly steep
\ ** learning curve. Texinfo code can be found in the
\ ** examples at the end of this chapter.

\ *P Glossary entries (the ones with *\fo{*G} tags are indexed.
\ ** Additional indexing macros will be added in a future release.

\ *P For Windows, the MikTeX package should be suitable for use
\ ** with the output of DocGen and the file *\i{texinfo.tex} is
\ ** required. In addition, a converter from the TeX DVI
\ ** output format to PDF will be required if PDF manuals are to
\ ** be generated and if you do not have *\i{pdftex} or *\i{pdflatex}. MikTeX
\ ** includes *\i{pdflatex} which is a version of LaTeX that produces
\ ** PDF files directly. Most distributions include *\i{pdftex} which
\ ** performs the same operation and may be more suitable
\ ** for TeXinfo than *\i{pdflatex}.

\ *P The MikTeX home page is at www.miktex.org. Note that the v2.4
\ ** small distribution is over 24Mb. For other distributions, which are
\ ** much larger, a CD is available from the home page for a small
\ ** charge. We encourage you to get the CD of the latest
\ ** release if you intend to use Texinfo.

\ *P If you are using the v2.4 distribution supplied by MPE,
\ ** install the MikTex package by running the installer.
\ ** Make sure that the *\i{MIKTEX\BIN} directory is in your search
\ ** path, and generate a master document file using *\i{MANUAL.TEX}
\ ** as a template. Run DocGen with the *\fo{DOCGEN_TEXINFO} personality
\ ** to produce your output files, and make sure that you have
\ ** included lines of the form *\i{@include file.tex} for each file
\ ** in the manual. Then run *\i{LATEX.EXE MANUAL} to produce a DVI
\ ** file or *\i{PDFLATEX.EXE MANUAL} to produce a PDF file.

\ *P Not all versions of Tex include the indexing package *\i{texindex.exe}
\ ** or the Texinfo package.
\ ** For windows you can get them from the GnuWin32 project:
\ *C   http://gnuwin32.sourceforge.net/packages/texinfo.htm
\ *P For texindex you probably need only the binaries and can probably
\ ** discard everything except the directory containing texindex.exe
\ ** which needs to be put in some directory in your PATH. You may
\ ** also want to keep the documentation, texindex.1.txt.

expose-module DOCGEN
MODULE DOCGEN_TEXINFO

[-sin

: textype	\ c-addr u -- ; as type, but processes TeX characters
  bounds ?do
    case  i c@
      [char] @ of  ." @@"  endof
      [char] { of  ." @{"  endof
      [char] } of  ." @}"  endof
        emit
    end-case
  loop
;

: opText	\ caddr u -- ; output processed text
  prepro textype
;

: opLine	\ caddr u -- ; output processed text with cr
  prepro textype cr
;

: lpc           \ char -- char'
\ *G Convert char to lower case if it was alphabetic.
  dup [char] A >= if
    dup [char] Z <=
    if  $20 or  then
  then
;

: lower		\ addr len --
\ Convert the ASCII string described to lower case.
\ Happens in place.
  bounds
  ?do  i dup c@ lpc swap c!  loop
;

((
create qchars	\ -- addr
  char @ c,
\  char | c,
  0 c,

: inCharList?	\ char list -- flag
  begin
    dup c@ 0=			\ check for end
    if  2drop false  exit  endif
    2dup c@ =			\ check same
    if  2drop true  exit  endif
    1 chars +
  again
;

: .q  		[char] " emit  ;

: idxtype	\ caddr u --
\ Like TYPE, but handles escapes for indexing.
  -white
  bounds ?do
    case  i c@
      [char] @ of  ." @@"  endof
      [char] { of  ." @{"  endof
      [char] } of  ." @}"  endof
        emit
    end-case
  loop
;

: countChars	\ caddr u char -- n
\ Count the number of characters in a string.
 0 swap 2swap				\ -- 0 char caddr u
 bounds ?do
   dup i c@ =
   if  swap 1+ swap  endif
 loop
 drop
;

: matched?	\ caddr u char1 char2 -- flag
\ Return true if the number of times char1 exists in the string
\ matches that of char2
  >r >r					\ -- caddr u ; R: -- char2 char1
  2dup r> countChars -rot		\ -- n1 caddr u ; R: -- char2
  r> countChars				\ -- n1 n2
  =
;
))

\ SFP014...
: .frac		\ n --
\ Display a percentage as two digit decimal ".xy" with no
\ trailing space.
   base @ >r decimal
   ." ."  0 <# # # #> type
   r> base !
;

: anyRow	\ --
\ Common code for row and hrow.
  -1 -> RowActive  0 -> ColActive
  -1 -> col#
;
\ ...SFP014

sin]

: >newf         \ caddr u --
  newfExpand NewDocFile 2drop
;

: >tofl         \ caddr u --
  SwitchDocFile
;

: >titl         \ caddr u --
  cr ." @chapter " -white opText cr
  cr
;

: .sectiontitle	\ caddr u --
  -white opLine
  cr
;

: >sect         \ caddr u --
  cr ." @section " .sectiontitle
  cr ." @vskip -6pt" cr
;

: >sstl		\ caddr u --
  cr ." @subsection " .sectiontitle
  cr ." @vskip -4pt" cr
;

: >head         \ caddr u --
  cr ." @subheading " .sectiontitle
  cr ." @vskip -4pt" cr
;

: >defn         \ caddr u --
  cr cr
  ." @table @code" cr
  ." @item " 2dup 2dup bl scan nip - -trailing opLine
  bl scan -white opLine
;

: defn          \ caddr u --
  space  -white opLine
;

: defn>         \ --
  cr ." @end table" cr cr
;

: >para         \ caddr u --
  ." @*" cr  -white opLine
;

: para          \ caddr u --
  -white opLine
;

: para>		\ --
  ." @vskip 0pt" cr
;

: >exam         \ caddr u --
  ( ." @*" ) cr
  ." @cartouche" cr
  ." @example" cr
  opText
;

: exam		\ caddr u --
  cr opText
;

: exam>         \ --
  ."  @end example" cr
  ." @end cartouche" cr
;

: >[list	\ caddr u --
  cr
  case  NewList
    [char] 1 of  ." @enumerate 1 "  endof
    [char] A of  ." @enumerate A "  endof
    [char] a of  ." @enumerate a "  endof
      ." @itemize @bullet "
  endcase
;

: >list]	\ caddr u --
  2drop
  case  BultStk A>
    [char] 1 of  ." @end enumerate"  endof
    [char] A of  ." @end enumerate"  endof
    [char] a of  ." @end enumerate"  endof
      ." @end itemize"
  endcase
  cr
;

: >bult         \ caddr u --
  BultStk adepth 0= if
    ." @*  @bullet{} "  opLine
  else
    cr ." @item " opline
  endif
;

: bult          \ caddr u --
  ."    "  opLine
;

: bult>         \ --
  BultStk adepth 0= if
    [ expose-module docgen ]
    NextDocState @ CurrDocState @ <>
    [ previous ]
    if  ." @*" cr  endif
\  else
\    cr
  endif
;

: IndexEntry	\ -- ; SFP014
  ParseGloname 2dup lower
  2dup [char] { scan nip	\ ignore names with '{' chars cos of Texinfo
  if  2drop  exit  endif
  2dup [char] } scan nip	\ ignore names with '}' chars cos of Texinfo
  if  2drop  exit  endif
  ." @findex " opLine
;

: >glos         \ caddr u --
  ." @vskip 6pt" cr
  index?
  if  IndexEntry  endif
  cr ." @example" cr
  LastDefinitionLine opText
  ."  @end example" cr cr
  ." @vskip -2pt" cr
  opLine
;

: glos          \ caddr u --
  opLine
;

: glos>         \ --
;

: >rtex         \ caddr u --
  prepro type cr
;

: >quot         \ caddr u --
  ." @cite{ "  prepro -trailing textype
;

: quot          \ caddr u --
  space  prepro -trailing textype
;

: quot>         \ --
  ." } @*" cr
;

: >code         \ caddr u --
  ." @example" cr
  opText
  ."  @end example" cr
\  ." @vskip 0pt" cr
;

: .codeline	\ -- ; output a code line to the documentation
  op-handle @ >r			\ switch o/p to file
  doc-gio op-handle !
  source -trailing prepro textype cr
  r> op-handle !			\ restore o/p
;

: >+cpy		\ caddr u -- ; start output of code
  opLine  ." @example" cr
  assign .codeline to-do OtherLines
;

: >-cpy		\ caddr u -- ; end output of code
  2drop
  ."  @end example" cr cr
  assign noop to-do OtherLines
;

: >rule		\ caddr u -- ; display one line rule
  2drop
  cr ." @tex"
  cr ." \vskip 4pt \hrule height 1pt width \hsize \vskip 4pt"
  cr ." @end tex" cr
;

[DocGen DocGen_TexInfo	\ -- ; -- ; select TeX personality
\ *G This word makes the TexInfo personality the current personality for
\ ** DocGen. TexInfo will remain the current personality until another
\ ** personality is selected.
  DG_FileExt: tex

\ continuation tag
  DG_Tag: *       0       illegal illegal illegal
\ tags that consist of one line only
  DG_Tag: !       -1      >newf   illegal illegal
  DG_Tag: >       -1      >tofl   illegal illegal
  DG_Tag: ^       -1      2drop   illegal illegal
  DG_Tag: #       -1      2drop   illegal illegal
  DG_Tag: T       -1      >titl   illegal illegal
  DG_Tag: S       -1      >sect   illegal illegal
  DG_Tag: N       -1      >sstl   illegal illegal
  DG_Tag: H       -1      >head   illegal illegal
  DG_Tag: R       -1      >rtex   illegal illegal
  DG_Tag: W       -1      2drop   illegal illegal
  DG_Tag: C       -1      >code   illegal illegal
  DG_Tag: [       -1      >+cpy   illegal illegal
  DG_Tag: ]       -1      >-cpy   illegal illegal
  DG_Tag: L       -1      >rule   illegal illegal
  DG_Tag: (       -1      >[list  illegal illegal	\ )
  DG_Tag: )       -1      >list]  illegal illegal
\ tags that can have continuation lines
  DG_Tag: D       1       >defn   defn>   defn
  DG_Tag: P       1       >para   para>   para
  DG_Tag: E       1       >exam   exam>   exam
  DG_Tag: B       1       >bult   bult>   bult
  DG_Tag: G       1       >glos   glos>   glos
  DG_Tag: Q       1       >quot   quot>   quot

\ the continuation tag
  DG_Tag: **      0       illegal illegal illegal	\ continuation
\ tags whose text must fit on a single line
  DG_Tag: !!      -1      >newf   illegal illegal
  DG_Tag: >>      -1      >tofl   illegal illegal
  DG_Tag: ^^      -1      2drop   illegal illegal
  DG_Tag: ##      -1      2drop   illegal illegal
  DG_Tag: CH      -1      >titl   illegal illegal	\ chapter
  DG_Tag: SE      -1      >sect   illegal illegal	\ section
  DG_Tag: SS      -1      >sstl   illegal illegal	\ sub-section
  DG_Tag: SH      -1      >head   illegal illegal	\ sub-head
  DG_Tag: RH      -1      2drop   illegal illegal	\ raw HTML
  DG_Tag: RL      -1      2drop   illegal illegal	\ raw LaTeX
  DG_Tag: RM      -1      2drop   illegal illegal	\ raw Markdown
  DG_Tag: RT      -1      >rtex   illegal illegal	\ raw TexInfo
  DG_Tag: CD      -1      >code   illegal illegal	\ code
  DG_Tag: C(      -1      >+cpy   illegal illegal	\ copy start
  DG_Tag: C)      -1      >-cpy   illegal illegal	\ copy end
  DG_Tag: HR      -1      >rule   illegal illegal	\ horizontal rule
  DG_Tag: L(      -1      >[list  illegal illegal	\ list start )
  DG_Tag: L)      -1      >list]  illegal illegal	\ list end
  DG_Tag: EV      -1      >ev     illegal illegal	\ evaluate
  DG_Tag: IF      -1      >if     illegal illegal	\ IF text
  DG_Tag: EL      -1      >else   illegal illegal	\ ELSE
  DG_Tag: TH      -1      >then   illegal illegal	\ THEN
\ tags which permit continuation lines
  DG_Tag: DF      1       >defn   defn>   defn
  DG_Tag: PA      1       >para   para>   para
  DG_Tag: EX      1       >exam   exam>   exam
  DG_Tag: IT      1       >bult   bult>   bult
  DG_Tag: GL      1       >glos   glos>   glos
  DG_Tag: QU      1       >quot   quot>   quot

DG_Type: textype

\ =================
\ *N Texinfo macros
\ =================

[DG_Macros
: bold		\ caddr u --
\ *G Macro to display string in bold text.
\ *C   *\starbslash{}bold{text in bold}
  ." @strong{" opText ." }"
;

: fixed		\ caddr u --
\ *G Macro to display string in a fixed-width font.
\ *C   *\starbslash{}fixed{text in fixed font}
  ." @code{" opText ." }"
;

: italic	\ caddr u --
\ *G Macro to display string in an italic font.
\ *C   *\starbslash{}italic{text in italic font}
  ." @emph{" opText ." }"
;

: forth		\ caddr u --
\ *G Macro to display text as Forth source - bold and fixed font.
\ *C   *\starbslash{}forth{text in Forth font}
  ." @strong{@code{" opText ." }}"
;

: br		\ caddr u --
\ *G Macro to insert a line break. Any text is ignored.
\ *C   *\starbslash{}br{}
  2drop  ." @*" cr
;

: starbslash	\ caddr u --
\ *G Macro to insert *\'. Any text is ignored.
\ *C   *\starbslash{}starbslash{}
  2drop  S" *\" opText
;

\ SFP014...
: table		\ caddr len --
\ *G Table start macro. The text is a quoted caption followed
\ ** by a list of column widths in per-cent. For a table with
\ ** three columns, use (say)
\ *C   *\starbslash{}table{ "Table caption" 20 20 60 }
  SetCaption SetCols
  cr ." @float " caption$ c@ if
    1 add #Tables
    ." Table,tab:ex" #Tables .dec
  endif
  cr ." @multitable @columnfractions "
  #cols 0 ?do
    percentages i cells + @ .frac space
  loop
  cr  0 -> RowActive  0 -> ColActive
;

: endtable	\ caddr len --
\ *G Table end macro. Any text is discarded
\ *C   *\starbslash{}endtable{}
  2drop
  cr ." @end multitable"
  caption$ c@ if
    cr ." @caption{" caption$ count opText ." }"
  endif
  cr ." @end float" cr
;

: row		\ caddr len --
\ *G Macro that marks the start of a normal row. Any text is
\ ** discarded.
\ *C   *\starbslash{}row{}
  2drop  anyRow  0 -> hRow?
;

: hrow		\ caddr len --
\ *G Macro that marks the start of a header row. Any text is
\ ** discarded.
\ *C   *\starbslash{}hrow{}
  2drop  anyRow  -1 -> hRow?
;

: col		\ caddr len --
\ *G Macro that marks the start of a header row. Any text is
\ ** discarded.
\ *C   *\starbslash{}hrow{}
  ColActive if
    ."  @tab "
  else
    hRow?
    if  ."  @headitem "  else  ."  @item "  endif
  endif
  2drop  Col# 1+ -> Col#  -1 -> ColActive
;

: image		\ caddr len --
\ *G Macro to display an image from a file.
\ *C  *\starbslash{}image{"caption" "file" "ext" hmm w%}
\ *P where *\i{caption} is the text of the image caption;
\ ** *\i{file} is the basic file name with no extension;
\ ** *\i{ext} is the file extension; *\i{hmm} is the height in mm
\ ** except for HTML; and *\i{w%} is the image width in percent
\ ** for HTML, ignored here. All five parameters must be present.
  SetImage
  cr ." @float " caption$ c@ if
    1 add #Figures
    ." Figure,fig:ex" #Figures .dec
  endif
  cr ." @center @image{" imgfile$ count type ." ,," Hmm .dec ." mm}"
  caption$ c@ if
    cr ." @caption{" caption$ count opText ." }"
  endif
  cr ." @end float" cr
;

: rawimage	\ caddr len --
\ *G Macro to display an image from a file.
\ *C   *\starbslash{}rawimage{ "file" "w" "h" "alttext" ".ext" }
\ *P where *\i{file} is the basic file name with no extension;
\ ** *\i{ext} is the file extension; *\i{w%} is the image
\ ** width in mm; and *\i{h} is the height in mm.
\ ** All five parameters must be present.
  setRawImage
  cr ." @image{"
  imgfile$ count type ." ,"
  W$ count -white tuck type if ." mm" endif ." ,"
  H$ count -white tuck type if ." mm" endif ." ,"
  alttext$ count type ." ,"
  imgext$ count type  ." }" cr
;

\ ...SFP014

: href		\ caddr len --
\ *G Create a link to another file or object. The URL
\ ** text is what is linked to and the display text is what is
\ ** shown in the link.
\ *C   *\starbslash{}href{"url" "display"}
  param1$ getTextParam  param2$ getTextParam  2drop
  cr ." @uref{" param1$ count type ." , " param2$ count type ." }"
  cr
;

: url		\ caddr len --
\ *G Create a link to another file or object. The URL
\ ** text is what is linked to and the display text is also the link.
\ *C   *\starbslash{}url{"url"}
  param1$ getTextParam  2drop
  cr ." @uref{"  param1$ count type  ." }"
  cr
;

\ Must be last because they redefine standard words.
: b   bold  ;
\ *G Convenient redefinition.
: f   fixed  ;
\ *G Convenient redefinition.
: i   italic  ;
\ *G Convenient redefinition.
: fo  forth  ;
\ *G Convenient redefinition.

DG_Macros]

DocGen]

  EXPORT Docgen_texinfo

END-MODULE
previous


\ *****************
\ LaTeX Personality
\ *****************

expose-module DOCGEN
MODULE Docgen_Latex

\ *****************
\ *S LaTeX2e output
\ *****************
\ *P This personality generates LaTeX2e output. From LaTeX2e you can go to many
\ ** other formats including PDF. See the previous TeX section for details of
\ ** the required associated tools and their installation.

\ ===============
\ *N Installation
\ ===============
\ *P Install the MikTex or Texlive package by running the installer.
\ ** Make sure that the MIKTEX\BIN directory is your search path, and generate
\ ** a master document file using MANUAL.TEX as a template. Run DocGen with the
\ ** DOCGEN_LATEX personality to produce your output files, and make sure that
\ ** you have included lines of the form "\include{file}" for each file in the
\ ** manual. Then run "LATEX.EXE MANUAL" to produce a DVI file, or "PDFLATEX.EXE
\ ** MANUAL" or "PDFTEX MANUAL" to produce a PDF file.

\ ==============
\ *N Basic usage
\ ==============
\ *P The output of DocGen is a set of LaTex2e files as defined by the tags
\ ** above. The manual is then generated by processing these files with
\ ** PDFLATEX.EXE. An example of MANUAL.TEX is given below. Each file that you
\ ** want to process should have a line of the form "\include{filename}" where
\ ** the .TEX extension will be assumed.
\ *E \documentclass[a4paper, 10pt]{book}
\ ** \parindent 0pt
\ ** \parskip 1ex plus .5ex
\ ** \begin{document}
\ ** \include{docgen}
\ ** \end{document}

\ *P The output of running PDFLATEX MANUAL will be file called MANUAL.PDF ready
\ ** for distribution. There will also be a large number of MANUAL.* temporary
\ ** files, all of which can be deleted as required. The only one of potential
\ ** interest is MANUAL.LOG, which contains the error report. In most cases,
\ ** there will be many reports of the form "Overfull \hbox" which can be
\ ** ignored.

\ *P You can configure the appearance of the manual by editing MANUAL.TEX as
\ ** you require. LaTeX is a very powerful document processing system, and you
\ ** can modify nearly everything, as well as add indices and so on. Several
\ ** books about LaTeX and TeX are available from Amazon, and the best source
\ ** of information about the system is at www.tex.ac.uk which will point you
\ ** at implementations for many machines as well as several tutorial packages.

\ ======================
\ *N Adding a title page
\ ======================
\ *P Adding a title page requires a few more lines, and an example is given
\ ** below.
\ *E \NeedsTeXFormat{LaTeX2e}
\ ** \documentclass[a4paper, 10pt]{book}
\ ** \parindent 0pt
\ ** \parskip 1ex plus .5ex
\ ** \begin{document}
\ ** \author{MicroProcessor Engineering Ltd}
\ ** \title{DOCGEN/SC}
\ ** \maketitle
\ ** \include{docgensc}
\ ** \end{document}

\ =============================
\ *N Adding a Table of Contents
\ =============================
\ *P Adding a table of contents requires only a few more lines, and an example
\ ** is given below.
\ *E \NeedsTeXFormat{LaTeX2e}
\ ** \documentclass[a4paper, 10pt]{book}
\ ** \parindent 0pt
\ ** \parskip 1ex plus .5ex
\ ** \begin{document}
\ ** \author{MicroProcessor Engineering Ltd}
\ ** \title{DOCGEN/SC}
\ ** \maketitle
\ ** \pagenumbering{roman}
\ ** \tableofcontents
\ ** \newpage
\ ** \pagenumbering{arabic}
\ ** \include{docgensc}
\ ** \end{document}

\ *P Note that in order to generate a table of contents, LATEX
\ ** needs to be run twice. On the first run, the table of
\ ** contents will be inaccurate as the table of contents file
\ ** will be the one from the previous run which is unlikely
\ ** to match the new first run. After the second run on the
\ ** same files, the table of contents will be accurate.

[-sin
: textype       \ c-addr u -- ; as type, but processes TeX characters
  bounds ?do
    case  i c@
      [char] @ of  ." @"   endof
      [char] _ of  ." \_{}"  endof
      [char] ^ of  ." \^{ }"  endof
      [char] ~ of  ." \~{ }"  endof
      [char] $ of  ." \${}"  endof
      [char] { of  ." {$\{$}"  endof
      [char] } of  ." {$\}$}"  endof
      [char] # of  ." \#"  endof
      [char] % of  ." \%"  endof
      [char] & of  ." \&"  endof
      [char] < of  ." {$<$}"  endof
      [char] > of  ." {$>$}"  endof
      [char] / of  ." {$/$}"  endof
      [char] \ of  ." {$\backslash$}"  endof
        emit
    end-case
  loop
;

: opText	\ caddr u -- ; output processed text
  prepro textype
;

: opLine	\ caddr u -- ; output processed text with leading CR
  cr prepro textype
;

: opStripText	\ caddr u -- ; output processed text with -TRAILING
  prepro -trailing textype
;
sin]

: >newf         \ caddr u --
  NewDocFile
;

: >tofl         \ caddr u --
  SwitchDocFile
;

: >titl         \ caddr u --
  cr ." \chapter{" opStripText ." }"
  cr
;

: >sect         \ caddr u --
  cr ." \section{" opStriptext ." }"
  cr
;

: >sstl		\ caddr u --
  cr ." \subsection{" opStripText ." }"
  cr
;

: >head         \ caddr u --
  cr ." \paragraph{" opStripText ." }"
  cr
;

: >defn         \ caddr u --
  cr ." \begin{description}"
  cr ." \item["
  2dup 2dup bl scan nip - opStripText
  ." ]"
  cr bl scan opStripText
;

: defn          \ caddr u --
  space  opStripText
;

: defn>         \ --
  cr ." \end{description}" cr
;

: >para         \ caddr u --
  cr opLine
;

: para          \ caddr u --
  cr  opStripText
;

: para>		\ --
  cr
;

: >exam         \ caddr u --
\  cr ." \rule{\linewidth}{.2pt}"
  cr ." \begin{verbatim} "
  cr  prepro -trailing type
;

: exam		\ caddr u --
  cr  prepro -trailing type
;

: exam>         \ --
  cr ." \end{verbatim} "
\  cr ." \rule{\linewidth}{.2pt}"
  cr
;

: >[list	\ caddr u --
  cr
  case  NewList
    [char] 1 of  ." \begin{enumerate} "  endof
      ." \begin{itemize} "
  endcase
;

: >list]	\ caddr u --
  2drop  cr
  case  BultStk A>
    [char] 1 of  ." \end{enumerate}"  endof
      ." \end{itemize}"
  endcase
  cr
;

: >bult         \ caddr u --
  BultStk adepth 0= if			\ old style
    cr ." \begin{itemize} \item " opText
  else
    cr ." \item " opText
  endif
;

: bult          \ caddr u --
  opLine
;

: bult>         \ --
  BultStk adepth 0= if			\ old style
    cr ." \end{itemize} "
  endif
;

: >glos         \ caddr u --
  cr ." \pagebreak[2] \texttt{"
  LastDefinitionLine opStripText
  ." }"
  cr ." \vspace{-2ex}"
  cr ." \begin{quote}"
  opLine
;

: glos          \ caddr u --
  opLine
;

: glos>         \ --
  cr ." \end{quote}"
  cr cr
;

: >rtex         \ caddr u --
  cr  prepro type
;

: >quot         \ caddr u --
  cr ." \cite{ " opStripText
;

: quot          \ caddr u --
  space  opStripText
;

: quot>         \ --
  ." }" cr
;

: >code         \ caddr u --
  cr cr ." \texttt{ "
  opLine
  ." }" cr
;

: .codeline	\ -- ; output a code line to the documentation
  op-handle @ >r			\ switch o/p to file
  doc-gio op-handle !
  source -trailing prepro type cr
  r> op-handle !			\ restore o/p
;

: >+cpy		\ caddr u -- ; start output of code
  cr cr opText
  cr ." \begin{verbatim} "
  assign .codeline to-do OtherLines
;

: >-cpy		\ caddr u -- ; end output of code
  2drop
  cr ." \end{verbatim} "
  assign noop to-do OtherLines
;

: >rule		\ -- ; display one line rule
  2drop
  cr ." \vspace{\medskip)"
  cr ." \rule{\linewidth}{.2pt} \newline"
;

[DocGen DocGen_LaTex	\ -- ; -- ; select LaTeX personality
( *G This word makes the LaTex personality the current personality for	)
( ** DocGen. LaTex will remain the current personality until another	)
( ** personality is selected. See the previous section about TeX for	)
( ** details of the system requirements.				)
  DG_FileExt: tex

\ continuation tag
  DG_Tag: *       0       illegal illegal illegal
\ tags that consist of one line only
  DG_Tag: !       -1      >newf   illegal illegal
  DG_Tag: >       -1      >tofl   illegal illegal
  DG_Tag: ^       -1      2drop   illegal illegal
  DG_Tag: #       -1      2drop   illegal illegal
  DG_Tag: T       -1      >titl   illegal illegal
  DG_Tag: S       -1      >sect   illegal illegal
  DG_Tag: N       -1      >sstl   illegal illegal
  DG_Tag: H       -1      >head   illegal illegal
  DG_Tag: R       -1      >rtex   illegal illegal
  DG_Tag: W       -1      2drop   illegal illegal
  DG_Tag: C       -1      >code   illegal illegal
  DG_Tag: [       -1      >+cpy   illegal illegal
  DG_Tag: ]       -1      >-cpy   illegal illegal
  DG_Tag: L       -1      >rule   illegal illegal
  DG_Tag: (       -1      >[list  illegal illegal	\ )
  DG_Tag: )       -1      >list]  illegal illegal
\ tags that can have continuation lines
  DG_Tag: D       1       >defn   defn>   defn
  DG_Tag: P       1       >para   para>   para
  DG_Tag: E       1       >exam   exam>   exam
  DG_Tag: B       1       >bult   bult>   bult
  DG_Tag: G       1       >glos   glos>   glos
  DG_Tag: Q       1       >quot   quot>   quot

\ the continuation tag
  DG_Tag: **      0       illegal illegal illegal	\ continuation
\ tags whose text must fit on a single line
  DG_Tag: !!      -1      >newf   illegal illegal
  DG_Tag: >>      -1      >tofl   illegal illegal
  DG_Tag: ^^      -1      2drop   illegal illegal
  DG_Tag: ##      -1      2drop   illegal illegal
  DG_Tag: CH      -1      >titl   illegal illegal	\ chapter
  DG_Tag: SE      -1      >sect   illegal illegal	\ section
  DG_Tag: SS      -1      >sstl   illegal illegal	\ sub-section
  DG_Tag: SH      -1      >head   illegal illegal	\ sub-head
  DG_Tag: RH      -1      2drop   illegal illegal	\ raw HTML
  DG_Tag: RL      -1      >rtex   illegal illegal	\ raw LaTeX
  DG_Tag: RM      -1      2drop   illegal illegal	\ raw Markdown
  DG_Tag: RT      -1      2drop   illegal illegal	\ raw TexInfo
  DG_Tag: CD      -1      >code   illegal illegal	\ code
  DG_Tag: C(      -1      >+cpy   illegal illegal	\ copy start
  DG_Tag: C)      -1      >-cpy   illegal illegal	\ copy end
  DG_Tag: HR      -1      >rule   illegal illegal	\ horizontal rule
  DG_Tag: L(      -1      >[list  illegal illegal	\ list start )
  DG_Tag: L)      -1      >list]  illegal illegal	\ list end
  DG_Tag: EV      -1      >ev     illegal illegal	\ evaluate
  DG_Tag: IF      -1      >if     illegal illegal	\ IF text
  DG_Tag: EL      -1      >else   illegal illegal	\ ELSE
  DG_Tag: TH      -1      >then   illegal illegal	\ THEN
\ tags which permit continuation lines
  DG_Tag: DF      1       >defn   defn>   defn
  DG_Tag: PA      1       >para   para>   para
  DG_Tag: EX      1       >exam   exam>   exam
  DG_Tag: IT      1       >bult   bult>   bult
  DG_Tag: GL      1       >glos   glos>   glos
  DG_Tag: QU      1       >quot   quot>   quot

DG_Type: textype

\ ==================
\ *N LaTeX macros
\ ==================

[DG_Macros
: bold		\ caddr u --
\ *G Macro to display string in bold text.
\ *C   *\starbslash{}bold{text in bold}
  ." \textbf{"  prepro textype  ." }"
;

: fixed		\ caddr u --
\ *G Macro to display string in a fixed-width font.
\ *C   *\starbslash{}fixed{text in fixed font}
  ." \texttt{"  prepro textype  ." }"
;

: italic	\ caddr u --
\ *G Macro to display string in an italic font.
\ *C   *\starbslash{}italic{text in italic font}
  ." \textit{"  prepro textype  ." }"
;

: forth		\ caddr u --
\ *G Macro to display text as Forth source - bold and fixed font.
\ *C   *\starbslash{}forth{text in Forth font}
  ." \texttt{\textbf{"  prepro textype  ." }}"
;

: br		\ caddr u --
\ *G Macro to insert a line break. Any text is ignored.
\ *C   *\starbslash{}br{}
  2drop  ." \\ "
;

: starbslash	\ caddr u --
\ *G Macro to insert *\'. Any text is ignored.
\ *C   *\starbslash{}startbslash{}
  2drop  S" *\" textype
;

: href		\ caddr len --
\ *G Create a link to another file or object. For HTML, the URL
\ ** text is what is linked to and the display text is what is
\ ** shown in the link.
\ *C   *\starbslash{}href{"url" "display"}
\ *P The hyperref package is needed. Include the line below
\ ** where other packages are loaded.
\ *C   \usepackage{hyperref}
  param1$ getTextParam  param2$ getTextParam  2drop
  cr ." \href{" param1$ count type ." }{" param2$ count type ." }" cr
;

: b   bold  ;
\ *G Convenient redefinition.
: f   fixed  ;
\ *G Convenient redefinition.
: i   italic  ;
\ *G Convenient redefinition.
: fo  forth  ;
\ *G Convenient redefinition.

DG_Macros]

DocGen]

  EXPORT Docgen_latex

END-MODULE
previous


\ **********************
\ *S DocGen kernel hooks
\ **********************
\ *P The VFX Forth Kernel contains *\fo{DEFER}ed hooks at strategic
\ ** points within the compiler/interpreter. These hooks are used by
\ ** DocGen to install and uninstall itself.
\ ** The DocGen hooks are:-

(( for DOCGEN
defer DOCGEN_PREREFILL	\ --
\ *G Called within *\fo{REFILL} before another line of text is read.

defer DOCGEN_REFILL	\ --
\ *G Called within *\fo{REFILL} after the next text line has been read.
\ ** at this point you may process the INPUT buffer (from *\fo{SOURCE}) but
\ ** you must *\b{not} under any circumstances change it.

defer -DOCGEN_HOOK	\ --
\ *G Disables DocGen without closing the active output file. This
\ ** is useful for conditional generation of documentation,
\ ** particularly if several versions of the software exist.
\ ** Note that files containing such phrases must be *\fo{INCLUDE}d,
\ ** as *\fo{DocOnly} turns off the Forth interpreter.
\ *E [undefined] <someword> [if]  -docgen_hook  [then]
\ ** \ ** This documentation is needed if <someword>
\ ** \ ** exists ...
\ ** +docgen_hook
defer +DOCGEN_HOOK	\ --
\ *G Enables DocGen output again. See the previous example
))


\ ===============================
\ *S Organising Manual generation
\ ===============================
\ *P You can make manual generation very much easier by
\ ** creating some auxiliary files which manage the
\ ** process.
\ *(
\ *B DocGen control file, *\fo{INCLUDE}d by VFX Forth to create
\ ** the HTML and/or Tex files. This file often also
\ ** *\fo{INCLUDE}s the file below. To distinguish it, we
\ ** usually give this file a ".DGS" extension.
\ *B List of source files to process. If you are creating both
\ ** HTML and PDF documentation, the process is simplified by
\ ** using the same list for both generation phases. This file
\ ** is usually called something like *\i{jobfiles.fth}.
\ *B A batch file to run the whole process.
\ *)

\ *P Examples are given of all of these files, somewhat
\ ** attenuated to remove obvious repetition. The examples
\ ** are taken from the documentation for the MPE ARM
\ ** USB Stamp on-board software. Both HTML and PDF files
\ ** are produced.

\ *P These examples contain several "MPEisms", in particular
\ ** the use of *\i{###} files. After DocGen has finished, we will
\ ** find files *\i{###.html} and *\i{###.tex}. If these files
\ ** have anything in them, some documentation has been missed.
\ ** When a source file ends, the following section adds further
\ ** DocGen output to the junk file:
\ *E \ ======
\ ** \ *> ###
\ ** \ ======

\ *P At the start of each source file, you must declare where
\ ** output goes using one of the '*!' or '*>' tags. Thus the
\ ** *\i{###} junk files collect any documentation that has been
\ ** written but not associated with an output file. Looking
\ ** at the junk files tells you if this has happened.


\ -----------------------------
\ *N Sample DocGen Control file
\ -----------------------------
\ *P There are two sections to this file, one for the
\ ** HTML documentation and one for the Texinfo code
\ ** that then creates the PDF documentation.

\ *E \ USBSTAMP.DGS - ARM USB Stamp DOCGEN control file
\ **
\ ** \ turn on DOCGEN and select personality
\ ** +docgen  docgen_html
\ **
\ ** cr ." Starting HTML manual generation" cr

\ *P The following section produces the main HTML file with
\ ** a left-hand chapter selection menu.
\ *E \ *^ index
\ ** \ *W <!DOCTYPE html>
\ ** \ *W <html lang="en">
\ ** \ *W   <head>
\ ** \ *W     <meta charset="utf-8">
\ ** \ *W     <meta http-equiv="X-UA-Compatible" content="IE=edge">
\ ** \ *W     <meta name="viewport" content="width=device-width, initial-scale=1">
\ ** \ *W     <title>MPE VFX Forth for Windows User Manual</title>
\ ** \ *W
\ ** \ *W     <!-- Bootstrap core CSS -->
\ ** \ *W     <link href="bootstrap.min.css" rel="stylesheet">
\ ** \ *W     <script src="script.js"></script>
\ ** \ *W     <style type="text/css">
\ ** \ *W        #left_frame {
\ ** \ *W             width: 30%;
\ ** \ *W 	         height: 100vh;
\ ** \ *W 	         border: 0px;
\ ** \ *W        }
\ ** \ *W        #right_frame {
\ ** \ *W             width: 70%;
\ ** \ *W 	         height: 100vh;
\ ** \ *W 	         border: 0px;
\ ** \ *W        }
\ ** \ *W        #left_frame,
\ ** \ *W        #right_frame { float: left; }
\ ** \ *W
\ ** \ *W     </style>
\ ** \ *W   </head>
\ ** \ *W
\ ** \ *W   <body>
\ ** \ *W
\ ** \ *W     <iframe id="left_frame" src="mainmenu.html" name="menu"></iframe>
\ ** \ *W     <iframe id="right_frame" src="titlepg.html" name="main"></iframe>
\ ** \ *W
\ ** \ *W   </body>
\ ** \ *W </html>
\ ** \ *> ###


\ *P Then we can create the chapter selection menu, which
\ ** references the HTML files produced by DocGen.
\ *E \ *! menu
\ ** \ *W <A target="main" HREF="stamptitle.html" >Home</A><BR>
\ ** \ *W <A target="main" HREF="intro.html"      >Introduction</A><BR>
\ ** \ *W <A target="main" HREF="codearm.html"    >Low level Kernel</A><BR>
\ **   ...
\ ** \ *W <A target="main" HREF="romforth.html"   >ROM FORTH extensions</A><BR>
\ ** \ *W <A target="main" HREF="xmodemtxrx.html" >XModem File Transfers</A><BR>
\ ** \ *> ###

\ *P We can create a title page.
\ *E \ *! stamptitle
\ ** \ *W <TABLE border=0 cellPadding=0 cellSpacing=0 width="100%">
\ ** \ *W   <TBODY>
\ ** \ *W   <TR><TD>
\ ** \ *W     <IMG src="mpelogo.gif">
\ ** \ *W   </TD></TR>
\ ** \ *W   </TBODY>
\ ** \ *W </TABLE>
\ **
\ ** \ *W <CENTER>
\ ** \ *W <H1>MPE ARM Stamp Software Reference Manual</H1>
\ ** \ *W <P>7 October 2004</P>
\ ** \ *W <P><I>Documentation derived from the source code by DOCGEN
\ ** \ *W       with VFX Forth for Windows
\ ** \ *W </I></P>
\ ** \ *W <BR><BR><BR>
\ ** \ *W <B>(C)opyright 2004 MicroProcessor Engineering Limited.</B>
\ ** \ *W </CENTER>
\ ** \ *> ###

\ *P Now we can generate all the other files from the second file.
\ ** Because DocGen automatically adds a colour to the start of
\ ** the *\i{INDEX.HTML} file, we include a comment to remove it
\ ** because it will be in the wrong place when frames are used.
\ *E include stampfiles
\ **
\ ** cr ." HTML Manual generation done" cr
\ **
\ ** cr
\ ** cr ." ***********************************"
\ ** cr ." Remove the BODY tag in the first line of INDEX.HTML"
\ ** cr ." ***********************************"
\ ** cr
\ **
\ ** -docgen

\ *P The procedure is essentially the same for the PDF
\ ** documentation, except that the table of contents is
\ ** produced by Texinfo.

\ *E \ turn on DOCGEN and select personality
\ ** +docgen  docgen_texinfo
\ **
\ ** cr ." Starting Texinfo manual generation" cr
\ **
\ **
\ ** \ ===============
\ ** \ Tex manual file
\ ** \ ===============
\ ** \ *! manual
\ ** \ *R \input texinfo
\ ** \ *R @setfilename           usbstamp.info
\ ** \ *R @setcontentsaftertitlepage
\ ** \ *R @afourpaper
\ ** \ *R @settitle               MPE USB Stamp
\ ** \ *R @setchapternewpage      odd
\ ** \ *R @paragraphindent        0
\ ** \ *R @exampleindent          0
\ ** \ *R @finalout
\ **
\ ** \ *R @include titlepg.tex
\ ** \ *R @include intro.tex
\ ** \ *R @include codearm.tex
\ **   ...
\ ** \ *R @include romforth.tex
\ ** \ *R @include xmodemtxrx.tex
\ **
\ ** \ *R @bye
\ ** \ *> ###
\ **
\ ** include stampfiles
\ **
\ ** cr ." Texinfo Manual generation done" cr
\ **
\ ** -docgen

\ *P If you are generating an index, add the following three
\ ** lines before the line containing "@bye"
\ *E \ *R @unnumbered{Index}
\ ** \ *R @*
\ ** \ *R @printindex fn


\ --------------------
\ *N Example file list
\ --------------------
\ *P The first section defines a set of text macros to
\ ** reduce typing and ease changes when the project directories
\ ** (folders) are moved.

\ *E \ STAMPFILES.FTH - DOCGEN include files
\ **
\ ** c" c:\buildkit.dev\software\rom\arm"
\ **   setmacro CpuDir
\ ** c" c:\buildkit.dev\software\rom\common"
\ **   setmacro CommonDir
\ ** c" c:\buildkit.dev\software\rom\examples"
\ **   setmacro ExampleDir
\ ** c" c:\buildkit.dev\software\rom\arm\hardware\LPC210x"
\ **   setmacro StampDir

\ *P Now comes the list of files. This file was started before
\ ** the word *\fo{DocOnly} was available. The .FTH extension
\ ** is not required as the smart file include system will try
\ ** a range of extensions. MPE habit is to give files that
\ ** only contain DocGen comments a .MAN extension.

\ *E s" intro.man" parsed
\ ** s" %CpuDir%\codearm.fth" parsed
\ ** s" %CommonDir%\kernel62.fth" parsed
\ **   ...
\ ** s" %CommonDir%\voctools.fth" parsed
\ ** s" %CommonDir%\xmodemtxrx.fth" parsed
\ ** DocOnly romforth.man
\ ** DocOnly examples.man
\ ** DocOnly titlepg.man


\ ---------------------
\ *N Example batch file
\ ---------------------
\ *P The batch file controls the whole operation
\ ** and removes the collection of temporary files produced
\ ** by PDFTEX or TEXIFY. The "start /w" command is used to make
\ ** the batch file wait until a GUI program has finished.

\ *P Note that if an index is being generated by Texinfo,
\ ** the line
\ *C pdftex manual
\ *P must be replaced by
\ *C texify -p manual.tex
\ *P where the extension must be supplied.

\ *E @echo off
\ ** rem B.BAT controls the operation
\ ** echo *****************************
\ ** echo Date changed in USBSTAMP.DGS?
\ ** echo *****************************
\ ** del *.html
\ ** del *.tex
\ ** start /w c:\products\pfwvfx\bin\pfwvfx include usbstamp.dgs
\ **
\ ** echo Error log will be in manual.log
\ ** del manual.log
\ ** echo starting pass 1 ..
\ ** pdftex manual
\ ** echo .. pass1 complete, starting pass 2 ..
\ ** pdftex manual
\ ** echo .. pass 2 completed
\ **
\ ** pause
\ ** move manual.pdf USBStampCode.pdf
\ ** echo Manual is in file USBStampCode.pdf
\ **
\ ** echo Deleting temporary files
\ ** move manual.log temp.log
\ ** del manual.*
\ ** move temp.log manual.log
\ **
\ **
\ ** echo ***********************************
\ ** echo Modify the first line of INDEX.HTML
\ ** echo ***********************************
\ **
\ ** echo Press Control-C not to issue manuals
\ ** pause
\ ** del ###.*
\ ** del ..\*.html
\ ** del *.tex
\ **
\ ** move *.html ..
\ ** move *.pdf ..
\ ** copy *.gif ..
\ **
\ ** echo Done
\ ** pause


\ -----------------------------
\ *N Example Texinfo title page
\ -----------------------------
\ *P This example only applies to the Texinfo personality.

\ *E \ ==========
\ ** \ *! titlepg
\ ** \ ==========
\ ** \ *R @titlepage
\ ** \ *R
\ ** \ *R @title                  MPE ARM USB Stamp
\ ** \ *R @subtitle               v1.0
\ ** \ *R @author                 Microprocessor Engineering Limited
\ ** \ *R @page
\ ** \ *R
\ ** \ *R @vskip 0pt plus 1filll
\ ** \ *R
\ ** \ *R Copyright @copyright{} 2004 Microprocessor Engineering Limited
\ ** \ *R
\ ** \ *R Published by Microprocessor Engineering
\ ** \ *R
\ ** \ *R
\ ** \ *R
\ ** \ *R @page
\ ** \ *R
\ ** \ *R MPE ARM USB Stamp                      @*
\ ** \ *R User manual                            @*
\ ** \ *R Manual revision 1.00                   @*
\ ** \ *R @today{}                               @*
\ ** \ *R                                        @*
\ ** \ *R                                        @*
\ ** \ *R Software                               @*
\ ** \ *R Software version 6.20                  @*
\ ** \ *R                                        @*
\ ** \ *R                                        @*
\ ** \ *R For technical support                  @*
\ ** \ *R Please contact your supplier           @*
\ ** \ *R                                        @*
\ ** \ *R                                        @*
\ ** \ *R For further information                @*
\ ** \ *R MicroProcessor Engineering Limited     @*
\ ** \ *R 133 Hill Lane                          @*
\ ** \ *R Southampton SO15 5AF                   @*
\ ** \ *R UK                                     @*
\ ** \ *R                                        @*
\ ** \ *R Tel:       +44 (0)23 8063 1441         @*
\ ** \ *R e-mail:    mpe@@mpeforth.com     @*
\ ** \ *R   tech-support@@mpeforth.com     @*
\ ** \ *R web:       www.mpeforth.com      @*
\ ** \ *R                                        @*
\ ** \ *R @page
\ ** \ *R @end titlepage
\ ** \ ======
\ ** \ *> ###
\ ** \ ======

\ ************
\ *S DocGen/SC
\ ************
\ *P DocGen/SC is an extension of DocGen for documenting safety
\ ** critical systems. DocGen/SC allows test code to be provided
\ ** after each word and to be extracted to separate test files, so
\ ** automating the production of regression tests.

\ *P DocGen/SC produces documentation that has been accepted by
\ ** organisations such as the US FDA for medical equipment. The
\ ** documentation format and test files are also suitable for
\ ** other authorities and application domains including avionics
\ ** and transport systems. Contact MPE for more details.

\ *P All the tags of DocGen work as they did before.
\ ** Some new tags have been defined to control the safety
\ ** critical documentation process.
\ *D O  followed by "initials" "name" "organisation". A list of
\ ** authors is kept and authors only need to be defined once.
\ *D A  followed by "initials" selects the current author. Once
\ ** selected, the author's name and organisation will be output
\ ** for each word definition.
\ *D V  followed by "version_text" sets the version information
\ ** produced in the header for each definition.
\ *D U  followed by "10" sets the width of hard tabs, ASCII code 9,
\ ** to the given integer value. This value defaults to 8, and is
\ ** used when expanding tab characters in the source code and
\ ** test code output. In general, we recommend that hard tabs
\ ** are always set to 8 characters as this is the default value
\ ** used by many applications.
\ *D M  starts the notes section of the output.
\ *D X  followed by "filename" defines the file used to contain
\ ** the test code. This file is closed after each source file is
\ ** *\fo{PARSED} and each source file should select a test file
\ ** into which the code between *\fo{[TEST} and *\fo{TEST]} is
\ ** to be placed.
\ *D Y  marks the start of the test code section.
\ *D Z  marks the end of the definition and triggers checks.


\ ======
\ *> ###
\ ======
