\ FileHacks.fth - quick and dirty file tools

((
Copyright (c) 2003
Wil Baden

Copyright (c) 2004, 2005
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20121026 MPE002 Zero terminated the file on read.
20120808 MPE001 Allowed repeated use of BURP.
))

only forth definitions  decimal

\ ============
\ *! filehacks
\ *T Quick and dirty file tools
\ ============
\ *P This code was developed for quick and dirty file handling
\ ** during development. Note that these tools are designed for use
\ ** from the keyboard and that global variables are used. The
\ ** code is written for safety, not for speed. All errors cause a
\ ** *\fo{THROW}.
\ ** This code is derived from Wil Baden's ToolBox.

\ *P The facetious word names are entirely deliberate to remind
\ ** you to remove these words from production code.

0 value pData	\ -- addr
\ *G Pointer to data block loaded from a file.
0 value /Data	\ -- n
\ *G Size of data block
0 value hData	\ -- handle
\ *G Handle of data file

: FILE-CHECK      ( n -- )  ABORT" File Access Error " ;
\ *G Tests a file ior.

: MEMORY-CHECK    ( n -- )  ABORT" Memory Allocation Error " ;
\ *G Tests a memory ior.

: rewind-file	\ file-id -- ior
\ *G Resets a file to the start.
  0 0 rot reposition-file
;

: InitReadFile	\ handle -- size
\ *G Reset the file to the start and return its size.
  dup rewind-file file-check
  file-size file-check  drop
;

: OpenMouth	\ caddr len --
\ *G Open the file for read only.
  r/o open-file file-check  dup to hData
  InitReadFile  to /Data
;

: guzzle	\ file-id -- addr length
\ *G Reads file from disc to *\fo{HERE} without *\fo{ALLOT}ing
\ ** space. The file is left open.
  dup InitReadFile			\ -- handle len
  here swap rot read-file file-check	\ -- #read
  here swap				\ -- addr #read
;

: slurp		\ file-id -- addr length
\ *G Reads the contents of a file into *\fo{ALLOCATE}d memory
\ ** and returns the address and length. Release the
\ ** memory using *\fo{BURP}. The returned length is the
\ ** file size. The *\fo{ALLOCATE}d length is one byte longer and
\ ** contains a trailing zero byte.
  dup InitReadFile			\ -- handle len
  dup 1+ allocate memory-check		\ -- handle len addr
  dup to pData  dup >r  swap rot	\ -- addr len handle ; R: -- addr
  read-file file-check			\ -- #read
  r> swap				\ -- addr #read
  2dup + 0 swap c!			\ zero terminate
;

: Hiccup	\ --
\ *G Close the file opened by *\fo{OpenMouth}.
  hData close-file file-check
;

: BURP		\ --
\ *G Release memory *\fo{ALLOCATE}d by *\fo{SLURP}.
  pData if
    pData free memory-check
    0 to pData
  then
;

: Inhale	\ fname flen -- caddr len
\ *G Given a file name, reads the file into *\fo{ALLOCATE}d memory,
\ ** returning its address and length. Use in the form:
\ *C   s" <filename>" INHALE ... BURP
  OpenMouth  hData Slurp  Hiccup
;

: Suck		\ "<filename>" -- caddr len
\ *G Reads the given file into *\fo{ALLOCATE}d memory, returning
\ ** its address and length. Use in the form:
\ *C   SUCK <filename> ... BURP
  GetPathSpec Inhale
;

: Exhale	\ caddr len name namelen --
\ *G Write the memory region defined by *\i{caddr/len} to the file
\ ** whose name is given by *\i{name/namelen}. A *\fo{THROW} occurs
\ ** on any error.
  2dup fileExist? if
    2dup delete-file drop
  then
  r/w create-file #-412 ?throw >r	\ create
  r@ write-file				\ -- ior ; write
  r> close-file #-416 ?throw		\ close
  #415 ?throw
;

: Vomits	\ caddr len "<filename>" --
\ *G Like *\fo{Exhale}, but takes the filename from the input stream.
\ *C   VOMITS <filename>
  GetPathSpec Exhale
;


\ ======
\ *> ###
\ ======
