\ BUFFER.FTH - memory buffer I/O device

((
Copyright (c) 2005, 2006
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20050823 MPE002 Added SizedTextBuff.
20050801 MPE001 Changed and documented SETPOS action.
))


\ ========
\ *> GENIO
\ ========
\ *N Memory Buffer Device
\ *P This Generic IO Device uses blocks of memory for input
\ ** and/or output. The source code is in *\i{LIB\GENIO\BUFFER.FTH}.

\ *P In order to create a device use the *\fo{TEXTBUFF:} definition given
\ ** later. *\fo{TEXTBUFF:} is compatible with ProForth 2.

\ *P When opening a memory device the parameters to *\fo{OPEN-GEN}
\ ** have the following meaning:
( *E    ADDR            Address of memory to use as buffer or ZERO if   )
( **                    dynamic allocation is required.                 )
( **    LEN             The maximum length of the memory image.         )
( **    ATTRIBS         When ZERO the ADDR parameter is ignored and LEN )
( **                    bytes of memory are allocated from the heap.    )

MODULE GENIO-BUFFER

struct textbuff-sid     \ -- len
\ *G Defines the length of a SID for a text buffer device
\ *[
  gen-sid +                             \ reuse field names of GEN-SID
  int textbuff-len                      \ length of buffer
  int textbuff-attribs                  \ attributes
  int textbuff-wr                       \ address to write to, set by SETPOS, WRITE
  int textbuff-rd                       \ address to read from set by SETPOS, READ
end-struct
\ *]

: reset-textbuff        \ buffer sid -- ; set pointers, clear buffer
  2dup gen-handle !                     \ handle=address of buffer
  2dup textbuff-wr !                    \ write pointer
  2dup textbuff-rd !                    \ read pointer
  textbuff-len @ erase                  \ wipe buffer
;

: textbuff-#read        \ sid -- #bytes
  dup textbuff-rd @  swap gen-handle @ - \ number of bytes read from buffer
;

: textbuff-#written     \ sid -- #bytes
  dup textbuff-wr @  swap gen-handle @ - \ number of bytes written to buffer
;

: open-textbuff \ addr len attribs sid -- buffer ior
  >r                                    \ save SID
  -1 r@ gen-handle !                    \ assume failure
  dup r@ textbuff-attribs !             \ set required attribs
  over r@ textbuff-len !                \ set required size
  case                                  \ -- addr len attribs ; look at attribs
    0 of  nip  allocate dup 0=          \ get required memory, check result
          if  over r@ reset-textbuff  endif
      endof
          2drop                         \ discard attribs and len
          dup r@ reset-textbuff         \ set buffer addresses
          0                             \ success
  end-case
  r> drop
;

: close-textbuff        \ sid -- ior
  dup gen-handle @ -1 =			\ check if already closed
  if  drop 0  exit  endif
  dup textbuff-attribs @ 0= if          \ if attributes are 0, free off buffer
    gen-handle @on free                 \ free off the buffer, handle := -1
  else
    -1 swap gen-handle !                \ mark unused
    0                                   \ success
  endif
;

: read-textbuff         \ addr len sid -- ior
  >r
  r@ textbuff-rd @ -rot                 \ -- src dest len ; start of read
  dup r> textbuff-rd +!                 \ new SEEK position = len+POS
  move                                  \ copy data
  0                                     \ success
;

: write-textbuff        \ addr len sid -- ior
  >r
  r@ textbuff-wr @ swap                 \ -- addr dest len ; start of write
  dup r@ textbuff-wr +!                 \ update write pointer
  move                                  \ copy data
  0 r> textbuff-wr @ c!                 \ write 0 terminator
  0
;

: key-textbuff  \ sid -- char
  dup textbuff-rd @ c@
  swap textbuff-rd incr  ;

: key?-textbuff \ sid -- flag
  dup textbuff-#read                    \ number of bytes in buffer
  swap textbuff-len @ u<  ;

: ekey-textbuff \ sid -- echar
  key-textbuff  ;

: ekey?-textbuff \ sid -- flag
  key?-textbuff  ;

: accept-textbuff \ addr len sid -- len'
  0 swap 2swap                          \ -- 0 sid addr len
  bounds
  ?do
    dup key?-textbuff 0= ?leave         \ get out if no character
    dup key-textbuff dup $0D =          \ if carriage return
    if                                  \ sid
      drop  over key-textbuff drop      \ discard CR and following LF
      leave                             \ leave
    endif
    i c!  swap 1+ swap                  \ save character and count it
  loop
  drop                                  \ discard SID
;

: (emit-textbuff)       \ char sid --
  >r                                    \ stash SID
  r@ textbuff-wr @ tuck c!              \ write character
  0 swap 1 + c!                         \ write terminator
  r> textbuff-wr incr                   \ increment pointer
;

: emit-textbuff \ char sid --
  >r
  case
    $08 of  out @ if                    \ destructive backspace
              r@ textbuff-wr decr
              0 r@ textbuff-wr @ c!
              out decr
            endif
        endof
    $0A of  $0A r@ (emit-textbuff)      \ LF
            op-line# incr
        endof
    $0D of  $0D r@ (emit-textbuff)      \ CR
            out off
        endof
          r@ (emit-textbuff)
          out incr
  end-case
  r> drop
;

: emit?-textbuff        \ sid -- flag
  dup textbuff-#read                    \ number of bytes in buffer
  swap textbuff-len @ u<  ;

: type-textbuff         \ addr len sid --
  -rot  bounds
  ?do  i c@ over emit-textbuff  loop
  drop  ;

: cr-textbuff   \ sid --
  $0D over emit-textbuff  $0A swap emit-textbuff  ;

: lf-textbuff   \ sid --
  $0A swap emit-textbuff  ;

: ff-textbuff   \ sid -- ; page/cls on display devices
\ *G This word is run by *\fo{PAGE}, *\fo{CLS} and *\fo{FF-GEN}. It resets
\ ** (empties) the buffer.
  dup gen-handle @ swap reset-textbuff  ;

: bs-textbuff   \ sid -- ; destructive on display devices
  $08 swap emit-textbuff  ;

: bell-textbuff \ sid -- ; audible beeper
  drop  ;

: xypos-textbuff        { x y sid -- addr ; return address in buffer of x,y }
  sid gen-handle @  sid textbuff-#written       \ buffer addr and length
  y 0                                   \ number of lines
  ?do
    $0D scan dup                        \ look for CR
    if  2 /string  endif                \ if found, step over
  loop
  drop  x +                             \ discard len, add x
;

: setpos-textbuff       \ x y mode sid -- ior
\ *G This word is run by *\fo{SETPOS-GEN}. Mode controls the x and y
\ ** input values as follows.
\ *D 0   x = #bytes written, y is ignored
\ *D -1  x=col, y=line for next character to be written
\ *D -2  x = #bytes read, y is ignored
\ *D -3  x=col, y=line for next character to be read
  >r
  case
     0 of  drop  r@ gen-handle @ + textbuff-wr !  0  endof
    -1 of  r@ xypos-textbuff r@ textbuff-wr !     0  endof
    -2 of  drop  r@ gen-handle @ + textbuff-rd !  0  endof
    -3 of  r@ xypos-textbuff r@ textbuff-rd !     0  endof
           drop  2drop -1               \ error
  end-case
  r> drop
;

[-sin
: xypos         \ buffer len -- x y
  0 0  2swap bounds ?do			\ -- x y
     i c@ #13 = if                      \ look for CR
       nip 0 swap                       \ reset X to 0
       1+                               \ increment Y
       i 1 + c@ #10 =  1 and 1+         \ check for LF, return 2 for LF, 1 for not
     else
       swap 1+ swap                     \ increment X
       1                                \ step over one character
     endif
   +loop
;
sin]

: getpos-textbuff       \ mode sid -- x y ior
\ *G This word is run by *\fo{GETPOS-GEN}. Mode controls the x and y
\ ** return values as follows.
\ *D 0   x = #bytes written, y = 0
\ *D -1  x,y for next character to be written
\ *D -2  x = #bytes read, y = 0
\ *D -3  x,y for next character to be read
\ *D -4  x = addr, y = len of unread data
  >r
  case
     0 of  r@ textbuff-#written 0  0  endof
    -1 of  r@ gen-handle @ r@ textbuff-#written xypos  0  endof
    -2 of  r@ textbuff-#read 0  0     endof
    -3 of  r@ gen-handle @ r@ textbuff-#read xypos     0  endof
    -4 of  r@ textbuff-rd @  r@ textbuff-wr @ over -   0  endof
           drop  0 0  -1
  end-case
  r> drop  ;

: ioctl-textbuff        \ addr len fn sid -- ior
  4drop  0  ;

: flushOP-textbuff      \ sid -- ior
  drop 0  ;

: RFU-textbuff  \ --
  1 abort" RFU not supported"  ;

: init-textbuff \ addr len sid -- ior
  3drop  0  ;

: term-textbuff \ sid -- ior
  drop  0  ;

: config-textbuff       \ sid -- ior ; produces a dialog
  drop  0  ;

create textbuff-vectors	\ -- addr
  ' Open-textbuff ,
  ' Close-textbuff ,
  ' Read-textbuff ,
  ' Write-textbuff ,
  ' Key-textbuff ,
  ' Key?-textbuff ,
  ' Ekey-textbuff ,
  ' EKey?-textbuff ,
  ' Accept-textbuff ,
  ' Emit-textbuff ,
  ' Emit?-textbuff ,
  ' Type-textbuff ,
  ' CR-textbuff ,
  ' LF-textbuff ,
  ' FF-textbuff ,
  ' BS-textbuff ,
  ' Bell-textbuff ,
  ' setpos-textbuff ,
  ' getpos-textbuff ,
  ' Ioctl-textbuff ,
  ' FlushOP-textbuff  ,
  ' RFU-textbuff ,
  ' Init-textbuff ,
  ' Term-textbuff ,
  ' Config-textbuff ,


\ ==================
\ *H Device Creation
\ ==================

: initTextBuffSid	\ addr --
\ *G Initialise a previously allocated SID for
\ ** a text buffer to the default values.
  dup textbuff-sid erase
  -1 over gen-handle !
  textbuff-vectors over gen-vector !
  $4000 swap textbuff-len !
; doNotSin

: textbuff:       \ "name" -- ; Exec: -- sid
\ *G Create a MEMORY-DEVICE called *\i{name}.
  create
    here  textbuff-sid allot  initTextBuffSid
\    -1 ,  textbuff-vectors ,  $04000 ,  \ handle, vector, default size
\    0 ,  0 ,  0 ,                       \ attribs, write pointer, read pointer
;

\ MPE002...
: SizedTextBuff	\ size -- sid|0
\ *G Allocates and opens a SID with a buffer of size *\i{size}
\ ** bytes, and returns the SID on success or 0 on failure.
  textbuff-sid allocate			\ -- size sid flag
  if  2drop 0  exit  endif
  dup >r initTextBuffSid		\ -- size ; R: -- sid
  0 swap 0 r@ open-textbuff nip		\ -- flag ; R: -- sid
  if  drop 0  r> free drop  exit  endif
  r>					\ -- sid
; doNotSin

: AllocTextBuff	\ -- sid|0
\ *G Allocates and opens a SID with a default 16kb text buffer
\ ** and returns the SID on success or 0 on failure.
  $4000 SizedTextBuff
; doNotSin
\ ...MPE002

: FreeTextBuff	\ sid --
\ *G Closes the SID and frees memory allocated by *\fo{AllocTextBuff}
\ ** or *\fo{SizedTextBuff}.
  dup close-textbuff drop  free drop
; doNotSin

export textbuff-sid
export initTextBuffSid
EXPORT textbuff:
export AllocTextBuff
export SizedTextBuff
export FreeTextBuff

END-MODULE


\ =======================================================================
( *> ###                                                                )
\ =======================================================================

