(* ====================================================================
application bundle specifics

copyright 2013 Roelf Toxopeus

VFX version.
define the Info.plist, pkginfo and apprun files here.
Last: 29 July 2020 at 12:30:10 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
The Info.plist, pkginfo and apprun file contents:

(TXTLINE) -- parse entire line and 'compile' it as a string without count
or terminating zero.
TXTLINE+ -- (TXTLINE) and add a trailing linefeed.
TXTLINE -- (TXTLINE) and add a trailing zero and align, like ,Z" but without
quotes!

INFO.PLIST$1 -- first part of default Info.plist file as a string.
Goes up to
 <key>CFBundleExecutable</key>
 <string>
Used to write out to new Info.plist file in app bundle. Executable is
added and second part of default Info.plist file is written out.
INFO.PLIST$2 -- second part of default Info.plist. Starts at
 </string>
 <key>CFBundleIconFile</key>
Goes up to
 <key>CFBundleName</key>
 <string>
Used to write out to new Info.plist file in app bundle. App name is
added and third part of default Info.plist file is written out.
INFO.PLIST3 -- third part of default Info.plist. Starts at
 </string>
 <key>CFBundlePackageType</key>
and goes to end.

PKGINFO$ -- the PkgInfo string. Addjust if necessary.

See turnkey2.f
-------------------------------------------------------------------- ))

/forth
decimal

: (TXTLINE) ( <text> -- )   0 parse bl skip -trailing  here over allot swap cmove ;

: TXTLINE+ ( <text> -- )  (txtline) 10 c, ;

: TXTLINE ( <text> -- )   (txtline) 0 c, align ;

create INFO.PLIST$1
txtline+ <?xml version="1.0" encoding="UTF-8"?>
txtline+ <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
txtline+ <plist version=\"1.0\">
txtline+ <dict>
txtline+ <key>CFBundleAllowMixedLocalizations</key>
txtline+ <string>true</string>
txtline+ <key>CFBundleDevelopmentRegion</key>
txtline+ <string>English</string>
txtline+ <key>CFBundleExecutable</key>
txtline <string>

create INFO.PLIST$2
txtline+ </string>
txtline+ <key>CFBundleIconFile</key>
txtline+ <string>app.icns</string>
txtline+ <key>CFBundleInfoDictionaryVersion</key>
txtline+ <string>6.0</string>
txtline+ <key>CFBundleIdentifier</key>
txtline+ <string>org.VFXForth.app</string>
txtline+ <key>CFBundleName</key>
txtline <string>

create INFO.PLIST$3
txtline+ </string>
txtline+ <key>CFBundlePackageType</key>
txtline+ <string>APPL</string>
txtline+ <key>CFBundleSignature</key>
txtline+ <string>mpe4</string>
txtline+ <key>CFBundleVersion</key>
txtline+ <string>1.0.5</string>
txtline+ </dict>
txtline </plist>

create PKGINFO$
txtline APPLmpe4

(( --------------------------------------------------------------------
  The Info.plist, pkginfo and apprun files:
  
  INFOPLIST-FILE -- create Info.plist file, write out the plist strings to this
  file. Insert the given app name in plist.

  PKGINFO-FILE -- create the PKGInfo file, write out the default pkginfo.
-------------------------------------------------------------------- ))

: ?WRITE-ERROR ( fid ior -- )
      ?dup if ." WRITE-FILE error !" swap close-file drop throw else drop then ;

: WRITE-PLIST ( a n fid -- )
   >r
   info.plist$1 zcount r@ write-file r@ swap ?write-error
   2dup r@ write-file r@ swap ?write-error                \ CFBundleExecutable
   info.plist$2 zcount r@ write-file r@ swap ?write-error
   r@ write-file r@ swap ?write-error                     \ CFBundleName
   info.plist$3 zcount r@ write-file r> swap ?write-error ;

: INFOPLIST-FILE ( a n -- )
   S" Info.plist" file dup >r write-plist r> close-file throw ;
   
: WRITE-PKGINFO ( fid -- )
   dup pkginfo$ zcount rot write-file ?write-error ;

: PKGINFO-FILE ( -- )
   s" Pkginfo" file dup >r write-pkginfo r> close-file throw ;
   
\\ ( eof )