(* ====================================================================
bundlecheck

copyright 2013 Roelf Toxopeus

VFX version
Check we're launched from an appbundle or not. Allows to find our
resources.
Last: 4 February 2019 at 14:06:26 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
BUNDLED? -- check we're bundled, return usefull path depending on flag
true  -> return appbundlepath/Contents/
Usefull when we have to search through our bundled resources, frameworks, plugins etc.
false -> return macpath.
NoTE: zstring path !!!!!!!!!!

A non GUI app does have a main bundle!
The bundle path is the path to the folder the app is launched from.
All the standard bundle folders are found there as well. Some like
the Resource folder are the same as the the bundle path.
A possible Info.plist file goes in the bundle path as well.

.BUNDLE-INFO -- prints paths of bundle contents. Uses CoreFoundation's
CFBundle functions. It even prints the path of folders not existing,
i.e. it prints the path where such a folder should be if it existed!
-------------------------------------------------------------------- ))

/forth
decimal

s" .app/Contents/" nip constant #KEEPCHARS

: BUNDLED? ( -- zstring f )
   pathpad dup >r dup 1024 erase
   dup GetExeNameZ zcount 2dup s" .app" search
               if  nip #keepchars - - r@ zplace true
         else  2drop 2drop macpath count r@ zplace false
   then  r> swap ;

CoreFoundation.framework

function: CFBundleGetMainBundle ( -- CFBundleRef )
function: CFBundleCopyBundleURL ( CFBundleRef -- CFURLRef )
function: CFBundleCopyExecutableURL ( CFBundleRef -- CFURLRef )
function: CFBundleCopyResourcesDirectoryURL ( CFBundleRef -- CFURLRef )
function: CFBundleCopySharedSupportURL ( CFBundleRef -- CFURLRef )

: .BUNDLE-INFO ( -- )
   CFBundleGetMainBundle
   dup CFBundleCopyBundleURL                cr ." Bundle path:         " cfurl> zcount type
   dup CFBundleCopyExecutableURL            cr ." Executable path:     " cfurl> zcount type
   dup CFBundleCopyResourcesDirectoryURL    cr ." Resource path:       " cfurl> zcount type
   dup CFBundleCopySharedSupportURL         cr ." Shared support path: " cfurl> zcount type
   cfrelease drop ;

cr .( bundlechecker loaded)

\\ ( eof )
