\ VfxExternObjC.fth - VFX Forth Extern: interface for Objective C (Cocoa) functions.

((
copyright 2008
Roelf Toxopeus
Utrecht
Netherlands

copyright 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
16 October 2014 23:15:36 CEST -- added Cocoa-Fpret:   -rt
))

only forth definitions
decimal

\ ===========
\ *! *> extern
\ *S Objective C and Cocoa
\ ===========

only forth definitions  decimal


\ return selector for methodname
: @selector ( 0string -- sel )
  dup c@ [char] @  = IF char+ THEN      \ ignore leading 'identifier'
  sel_getUid dup 0= abort" @selector failed"
;

((
\ current structure
struct /selector   \ -- len
  int >selector      \ selector
  int >#in      \ number of input parameters
  0 field selname   \ selector zname
  ...
  int link      \ list link - see mylist.fth
  int stuct      \ selector field above

\ proposed new structure
struct /selector   \ -- len
  int >sellink      \ links all selectors
  int >selector      \ selector
  0 field selname   \ selector zname
))

variable cocoaselectors   \ -- addr
\ anchors a list of selector structures.

: >cocoaselectors   \ struct -- ; add structure to list
  cocoaselectors node node, ;

: relinkselector   \ struct -- ; find selector id
  dup cell+ cell+ @selector swap ! ;

: /COCOASELECTORS   \ -- ; find all selector ids
  ['] relinkselector cocoaselectors allnodes ;
\ AtCold perhaps

variable #selectors

: .selectorname      \ struct -- ; display selector information
  cr  dup cell+ cell+ zcount type   \ name
  out @ 31 > if
    2 spaces
  else
    out @ #31 and #32 swap - spaces
  endif
  @ .dword            \ id
  #selectors incr
;

: .COCOASELECTORS   \ --
\ *G Display information for all declared selectors.
  0 #selectors !
  ['] .selectorname cocoaselectors allnodes
  cr #selectors @ . ." selectors"
;

: .Selectors   \ --
\ *G Display information for all declared selectors.
  .COCOASELECTORS
;

also extern.voc definitions

\ Code between [A ... A] is compiled into the current definition.

: CocoaPrologue,   \ seladdr --
\ Lay the call entry prologue. The run-time Forth stack order will be:
\   -- x1..xm
\ which becomes
\   -- x1..xm sel
  >r  [a
    push  esi            \ save Forth regs
    push  edi
\ preserve EBX, Selector and ESP
    lea   ebp, #-12 [ebp]      \ save EBX, Selector and ESP
    mov   8 [ebp], ebx         \ TOS=xm
    mov   eax, [] r>         \ save selector
    mov   4 [ebp], eax
    mov   0 [ebp], esp         \ after call, ESP will be restored
  a] [a]
  FrameAlignment if
    \ make a frame on ESP and align it to OS convention, 16 byte aligned less 4 bytes
    [a
      lea   eax, FrameDrop [esp]   \ compile time constant
      and   eax, # FrameMask      \ 16 byte align
      lea   esp, FrameBackup [eax]   \ allow for size of call
    a] [a]
  else
    FrameDrop if
    [a
      lea   esp, FrameDrop [esp]   \ compile time constant
    a] [a]
    endif
  endif
; doNotSin

: genCocoaCall   \ xtSender --
\ Lay the code to call the external function, and clean up.
  func-pointer >r  [a         \ indirect call
    cld
    call  r> []            \ indirect through pointer
    cld
    mov   esp, 0 [ebp]         \ prologue saved ESP, used for C.
  a] [a]
;

: genCocoaCode   \ struct xtsender --
\ Generate code for selector call using objc_msgSendxxx
\ Make the word that calls a normal imported function.
  {: selstruct xtsender -- :}
  align  selstruct ,         \ back pointer to structure
  ExternName count (:)         \ make header and start definition
  last-colon-xt @ >r         \ save xt we have just made
  selstruct CocoaPrologue,
  copyFrame  xtsender genCocoaCall  returnData
  epilogue,
  postpone ;            \ finish definition
  r> set-comp-extern         \ set its compiler.
; doNotSin

: checkCallConv   \ --
\ check the calling convention
\  L>R? 0<>  REQUIRE_NREV 0<> and  CalleeCleaned? 0= and
\  0= abort" ObjC requires full C calling convention"
  varargs? abort" Can't handle varargs for selectors"
;

: CalcCocoaFrame   \ --
\ Calculate the O/S stack offsets. The incoming order is
\   x1..xn obj sel    (TOS=sel)
\ which will be converted to (C ESP stack)
\   obj sel x1..xn    (RTOS=obj)
\ where obj is RTOS
  numArgs @ 2- ArgList[] CalcDest   \ object
  numArgs @ 1- ArgList[] CalcDest   \ selector
  numArgs @ 2- 0         \ x1..xn
  ?do  i ArgList[] CalcDest  loop
;

\ Note: for 'STRET' functions, add the stret pointer as the last rightmost input parameter
\ i.e. CocoaSelector @frame ( frame-rect -- ret )

: CocoaStructFrame   \ --
\ Calculate the O/S stack offsets. The incoming order (Forth EBP stack) is
\   x1..xn struct obj sel   (TOS=sel)
\ which will be converted to (C ESP stack)
\   struct obj sel x1..xn   (RTOS=struct)
\ where struct is RTOS
  numArgs @ 3 - ArgList[] CalcDest   \ struct
  numArgs @ 2 - ArgList[] CalcDest   \ object
  numArgs @ 1 - ArgList[] CalcDest   \ selector
  numArgs @ 3 - 0         \ x1..xn
  ?do  i ArgList[] CalcDest  loop
;

: SelectorStruct   \ #in -- struct
\ Make a selector structure using the number of parameters and
\ the stored import name.
  ImportName count + 0 swap c!      \ zero terminate import name
  ImportName 1+ @selector      \ -- #in selector
  here >r
  , , ImportName $>z, r@ >cocoaselectors  \ selector, #in, zname, link into selector chain
  r>
;

: MakeSelector   \ #hidden -- struct ;  <name> ( params -- ret )
\ Parse the parameters for a selector with the given number of
\ parameters. Build the structure and return its address.
\ For most selectors, the hidden parameters are the object
\ and the selector (passed to objc_msgSend). For selectors
\ that use a structure, the structure could be the third parameter.
\ However, in this notation, the address of the structure is the
\ last (rightmost) parameter and so is not hidden.
  >r               \ number of hidden items
  singleName            \ get the selector name
  usesFloats off  setDefExt      \ set default declaraction type
  parse-params over -rot      \ -- #in #in #out
  swap r> + swap setArgs      \ -- #in ; object and selector are hidden parameters
  checkCallConv
  SelectorStruct         \ -- struct
;

(*
NoTE: if the msg_send calls are defined with GLOBAL: we can pass them straight away.
Don't need ['] in those cases.
*)

\ *P The selector and parameter and return value list are defined
\ ** as
\ *C  Cocoa: <name> ( p1 p2 p3 -- rval )
\ *P The brackets form the Forth stack comment, where *\i{p1} is
\ ** the leftmost parameter in Objective C documentation. The object
\ ** to which the selector is applied is a hidden rightmost
\ ** parameter. The return value *\i{rval} can be 0, 1 or 2 items.
\ ** If you want your code to be portable to other Forth systems
\ ** always set *\i{rval} to one item, even if this causes you
\ ** extra *\fo{drop}s and fussing.

also forth definitions
: Cocoa:   \ Cocoa: <name> ( params -- return )
\ *G Create a selector called *\fo{<name>} whose parameters and
\ ** return value are defined by a list as above. You must
\ ** remember to add an object pointer as the right hand item.
\ ** Executing the selector causes it to be processed by
\ ** *\fo{objc_msgSend}.
  2 MakeSelector         \ -- struct
  initOffsets CalcForthOff CalcCocoaFrame
  ['] objc_msgSend genCocoaCode
;

: SuperCocoa:   \ SuperCocoa: <name> ( params -- return )
\ *G Create a selector called *\fo{<name>} whose parameters and
\ ** return value are defined by a list as above. You must
\ ** remember to add an object pointer as the right hand item.
\ ** Executing the selector causes it to be processed by
\ ** *\fo{objc_msgSendSuper}.
  2 MakeSelector         \ -- struct
  initOffsets CalcForthOff CalcCocoaFrame
  ['] objc_msgSendSuper genCocoaCode
;

: Cocoa-Stret:   \ Cocoa-Stret: <name> ( params stret -- return )
\ *G Create a selector called *\fo{<name>} whose parameters and
\ ** return value are defined by a list as above. This selector
\ ** returns a structure. You provide the address of the structure
\ ** as an additional right hand paramter. You must
\ ** remember to add an object pointer as the right hand item.
\ ** Executing the selector causes it to be processed by
\ ** *\fo{objc_msgSend_stret}.
  2 MakeSelector         \ -- struct
  initOffsets CalcForthOff CocoaStructFrame
  ['] objc_msgSend_stret genCocoaCode
;

: SuperCocoa-Stret:   \ SuperCocoa-Stret: <name> ( params stret -- return )
\ *G Create a selector called *\fo{<name>} whose parameters and
\ ** return value are defined by a list as above. This selector
\ ** returns a structure. You provide the address of the structure
\ ** as an additional right hand paramter. You must
\ ** remember to add an object pointer as the right hand item.
\ ** Executing the selector causes it to be processed by
\ ** *\fo{objc_msgSendSuper_stret}.
  2 MakeSelector         \ -- struct
  initOffsets CalcForthOff CocoaStructFrame
  ['] objc_msgSendSuper_stret genCocoaCode
;

: Cocoa-Fpret:   \ Cocoa: <name> ( params -- ) ( F: -- x )
\ *G Create a selector called *\fo{<name>} whose parameters and
\ ** return value are defined by a list as above. You must
\ ** remember to add an object pointer as the right hand item.
\ ** Executing the selector causes it to be processed by
\ ** *\fo{objc_msgSend_fpret}.
\ ** VFX uses the hardware/internal fp stack as its numeric stack.
\ ** The call will automaticly push the double result on this
\ ** stack, no further action is required. The data stackpicture
\ ** for these calls will show no return.
  2 MakeSelector         \ -- struct
  initOffsets CalcForthOff CalcCocoaFrame
  ['] objc_msgSend_fpret genCocoaCode
;

previous previous definitions


\ ======
\ *> ###
\ ======

decimal

