(* ====================================================================
NIB file handling.

copyright 2006 Roelf Toxopeus

Part of GUI stuff.
VFX version.
An example of loading cocoanib files and showing and controlling the objects.
Last: 10 June 2020 at 11:59:05 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
NIB files are actualy bundles/folders containing XML files. These XML
files describe the GUI widgets used by an application.
Ctrl-click on a NIB file, allows you to open the bundle for inspection.

/MYNIBS -- initialise my/your working nib path. Two versions: for distro
and my setup. Needs attention !

>NIBPATH -- append nib filename to our default nibpaths as provided by
BUNDLED?. Return full path as NSString instance. It points to the Resources
folder used for nibs icns, images etc. as used in application bundles.

RESOURCES -- the default nibs path. The Resources folder.
Path does not have a nib bundle appended!

@NIB -- will load a given nibbundle from the resources folder. Does not
instantiate/unpack it. Returns a reference to the loaded nib bundle,
an NSBundle instance. Case sensitive !
If the nib doesn't exist, a nil wil be returned.

*NIBOBJECTS -- will point to an NSArray with the toplevel objects
found in the most recent instantiated nib bundle. These objects and
the arry should be strong referenced to prevent deallocation!

/NIB -- unpack/instantiate the given nib bundle, uses *NIBOBJECTS.
A strong reference for the array and the objects is established.

OBJECTS? -- shows the available objects in the most recent instantiated
nib.

The combination @NIB and /NIB is enough to get the nibs loaded, instan-
tiated and on screen.

An alternative and much used way is:
NIBWINDOW.CONTROLLER -- return a NSWindowController instance build from
given NIB path. Send its @showWindow: message to display the widgets
described by the NIB. Take care: you can feed it a nonexistent nibpath
and it will still return an instance reference. SHOW.NIB has means to
detect false nibpaths.

SHOW.NIB -- display widgets described by given nibname.
A check is made on nibname, if it doesn't exist, a zero is returned.
When it exists, it's displayed and its NSWindowController instance is
returned.

-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ NIB paths

\ initialize the pathpad with the path
: /NIBPATH ( -- za )
   bundled? 0= if
      dup >r zcount ( +root)  r@ zplace r>
   then  dup s" resources/" rot zappend ;

: >NIBPATH ( za -- nsstring )    zcount /nibpath dup >r zappend r> >nsstring ;                         

cocoaclass NSBundle
cocoa: @initWithPath: ( nsstring:ref -- id )

: RESOURCES ( -- NSBundle:ref)   z" " >nibpath NSBundle @alloc @initWithPath: ;

cocoaclass NSNib
cocoa: @initWithNibNamed:bundle: ( nibname nsbundle:ref -- id )
cocoa: @instantiateNibWithOwner:topLevelObjects: ( owner address -- bool )  \ OSX 10.6 and OSX 10.7
cocoa: @instantiateWithOwner:topLevelObjects: ( owner address -- bool )     \ OSX 10.8 and upwards

: @NIB ( za -- NSNib:ref )   >nsstring resources NSNib @alloc @initWithNibNamed:bundle: ;

variable *NIBOBJECTS

: OBJECTS? ( -- )
   *nibobjects @ dup @count                    
   0 ?do
      i over @objectAtIndex: @class class_getName zcount  cr i . space type
   loop drop ;

: /NIB ( nib -- flag )
   NSApp @ *nibobjects rot 
   z" instantiateWithOwner:topLevelObjects:" @selector
   over @respondsToSelector:
   if @instantiateWithOwner:topLevelObjects:
   else @instantiateNibWithOwner:topLevelObjects:
   then
   *nibobjects @ @retain drop ;

\ --------------------------------------------------------------------
\ NIB controller

cocoaclass NSWindowController

\ NOTE: use  @ALLOC  first !!!!!!!
cocoa: @initWithWindowNibPath:owner: ( windowNibPath owner -- id ) \ NSString object, id -- NSWindowController object
cocoa: @showWindow: ( sender -- ret ) \ id

: NIBWINDOW.CONTROLLER  ( nstring:nibpath -- ref )   NSApp @ NSWindowController @alloc @initWithWindowNibPath:owner: ;

: SHOW.NIB ( za -- windowcontroller )
   >nibpath
   dup >4thstring file-status nip if @release drop 0 exit then     \ check its validity
   nibwindow.controller 0 over @showWindow: drop ;

\\ ( eof )