(* ====================================================================
Thread awareness

copyright 2009 Roelf Toxopeus

Part of adding Cocoa event handling to Forth.
VFX version.
Make sure Cocoa is aware of multi(posix)threading.
Last: 15 October 2014 09:46:12 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
MULTI.AWARE -- should be executed before any serious threads are added.
This protects the Cocoa framework.
Uses a much advised tric: create an NSThread and immediately end it.

After MULTI.AWARE is executed we can safely add the IMPOSTOR thread
for Forth to run on.
-------------------------------------------------------------------- ))

/forth
decimal

cocoaclass NSThread
cocoa: @start ( -- ret )
cocoa: @cancel ( -- ret )
cocoa: @isMultiThreaded ( -- flag ) \ BOOL

: MULTI.AWARE ( -- )
   NSThread @alloc @init
   dup @start drop @cancel drop
   NSThread @isMultiThreaded
   0= abort" Cocoa not aware of multi threading ! Aborting..."
;

\\ ( eof )

