(* ====================================================================
Cocoa runners

copyright 2008 Roelf Toxopeus

Part of the interface to ObjC 2 Runtime.
VFX version.
Running ObjC selectors as methods/functions.
forCocoa and forSuperCocoa check for # of output parameters during compilation.
This is the portable version.
Last: 16 October 2014 15:09:45 CEST   -rt
==================================================================== *)

/forth
decimal

(( --------------------------------------------------------------------
Before running the method, we need to change the Forth stackorder
to ObjC stackorder. An 'argument' stack is used to temporarely store
the arguments while re-shuffling the stack.

-ARGP -- clears argument stack.
PUSHES --  pushes n parameters from data stack to the argument stack.
POPS -- pops n parameters from argument stack to data stack.
REORDER -- puts all parameters in the required ObjC parameter order.
REORDER-STRET -- same, but deals with the STructure RET pointer as well.
-------------------------------------------------------------------- ))

32 constant MAXARGS
create ARGSTACK here maxargs cells dup allot erase   \ argument stack
argstack maxargs cells + constant ARGBP            \ arguments stack base pointer
argbp VALUE argp                                    \ arguments stack pointer, top arg stack

: >ARG ( n -- )                        \ push n to arguments stack predecrement  
   argp cell- dup to argp ! ;

: ARG> ( -- n )                        \ pop n from argument stack postincrement
   argp dup @ swap cell+ to argp ;

: -ARGP ( -- )  argbp to argp ;               \ reset argument stack pointer

: ARGDEPTH ( -- n ) argbp argp - 4 / ;

: ?ARGP ( -- )
   argdepth maxargs > if -argp true abort" Argument stack overflow !" then ;

: PUSHES ( x1 ... xn n -- ) 0 ?do >arg loop ;

: POPS ( n -- x1 ... xn ) 0 ?do arg> loop ;

0 value #ARGS

: REORDER ( x1 ... xn object selector n -- object selector x1 ... xn n )
   dup >r
   2 - to #args -argp   \ do not push object and selector to argument stack
\ clear stack
   2>r
   #args pushes
\ fill stack
   2r>   #args pops
   r>
;

: REORDER-STRET ( x1 ... xn stret object selector n+stret -- stret object selector x1 ... xn n+stret )
   dup >r
   3 - to #args -argp   \ do not push stret, object and selector to argument stack
\ clear stack
   >r 2>r
   #args pushes
\ fill stack
   2r> r>  #args pops
   r>
;

(( --------------------------------------------------------------------
Running the selector as method.

Compile the actual run parts (DOES>) from the ObjC interface.
Note the fp returning variant, leaves the returned fp on the fp-stack.

#RETURNS -- variable containing the number of output parameters from
the last defined method. Will be set by parse-params in COCOASELECTOR
in cocoafunc9-port.fth
-------------------------------------------------------------------- ))

variable #RETURNS

: (FORCOCOA) ( x1 ... xn object a -- ret )
      2@ swap reorder objc_msgSend ;

: FORCOCOA ( -- )
   #returns @
   case 0 of does> (forcocoa) drop  exit endof
        1 of does> (forcocoa)       exit endof
        2 of does> (forcocoa) 2ret> exit endof
        1 abort" illegal return count in COCOA: !"
   endcase ;

: (FORSUPERCOCOA) ( x1 ... xn object a -- ret )
   2@ swap reorder objc_msgSendSuper ;

: FORSUPERCOCOA ( -- )
   #returns @
   case 0 of does> (forsupercocoa) drop  exit endof
        1 of does> (forsupercocoa)       exit endof
        2 of does> (forsupercocoa) 2ret> exit endof
        1 abort" illegal return count in SUPERCOCOA: !"
   endcase ;

: (FORCOCOA-STRET) ( x1 ... xn stret object a -- ret )
   2@ swap reorder-stret objc_msgSend_stret ;

: FORCOCOA-STRET ( -- )   does> (forcocoa-stret) ;
   
: (FORSUPERCOCOA-STRET) ( x1 ... xn stret object a -- ret )
   2@ swap reorder-stret objc_msgSendSuper_stret ;

: FORSUPERCOCOA-STRET ( -- )   does> (forsupercocoa-stret) ;

: (FORCOCOA-FPRET) ( x1 ... xn object a --  ) ( F: -- r )
   2@ swap reorder objc_msgSend_fpret drop ( ST0> ) ;

: FORCOCOA-FPRET ( -- )   does> (forcocoa-fpret) ;

\\ ( eof )