(* ====================================================================
Cocoa functions

copyright 2008 Roelf Toxopeus

Part of the interface to ObjC 2 Runtime.
VFX version.
Declare message as if it is a function
Together with cocoacore7.f and cocoabridge7.f this file defines the
interface to ObjC message sending.
Last: 16 October 2014 15:28:24 CEST     -rt
==================================================================== *)

(( --------------------------------------------------------------------
Turnkey proof:
A linked list is created, which will hold the message selectors.
Upon booting coco-vfx, the list is traversed and all our defined Cocoa
messages are initialised.

@SELECTOR -- return selector for messagename. It tests the selector name
for a leading '@' character, and skips it in the name when found.
This leading '@' is used as a Cocoa identifier in coco-vfx, it's optional.
Likewise the underscore _ prepended to foreign function names in some Forth
systems. The '@' is borrowed from ObjC where it serves as compiler hint.
COCOASELECTORS -- holds head linked list for Cocoa selectors.
>COCOASELECTORS -- adds a selector to the list.
/COCOASELECTORS -- initialise the messages in the list to their proper selectors,
execute this somewhere in the startup from a turnkeyed coco-vfx.
.COCOASELECTORS -- show the selectors in the list by their name.

COCOASELECTOR -- create a named structure from parsed name containing the
selector, #input-parameters and the name string. Pass a pointer to this structure
to the herefore mentioned list.
NoTE the syntax is similar to the one used for FUNCTION:
The required receiver and selector for passing are implied. You don't have to
put them in the parameterlist. At runtime return pointer to structure.
For 'STRET' functions, add the stret pointer to the parameterlist.
COCOA-STRET: @frame ( frame-rect -- ret )

The following words use COCOASELECTOR to create the needed structure and at runtime
use the selector and #input-parameters to run the Cocoa message as a function.
The stackpictures at define time differs from the ones used at runtime.
At runtime add the receiver as right most parameter on top of all the others,
including th possible STRET pointer.
COCOA: -- run the defined word with this stackpicture ( n1 ... nn receiver -- x )
SUPERCOCOA: -- as COCOA: but pass it to the superclass of the receiver.
COCOA-STRET: -- add a structure pointer for return(ed) values  ( n1 ... nn stret receiver -- x )
SUPERCOCOA-STRET: -- id. but pass it to superclass of the receiver.
COCOA-FPRET: -- run the defined word with this stackpicture ( n1 ... nn receiver -- ret ) ( F: -- r )

NoTE: ObjC is a dynamic binding system. The implementation of the method for
a certain selector depends on the receiver/class/object. The lookup is done
at runtime, hence the passing along of the receiver and selector when sending
the message.
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ Relink selectors

: @SELECTOR ( 0string -- sel )
   dup c@ [char] @  = if  char+  then      \ ignore leading 'identifier'
   sel_getUid dup 0= abort" @SELECTOR failed !" ;

VARIABLE COCOASELECTORS

: >COCOASELECTORS ( a -- )  cocoaselectors node node, ;

: RELINKSELECTOR ( a -- )  dup cell+ cell+ @selector swap ! ;

: /COCOASELECTORS ( -- )   ['] relinkselector cocoaselectors allnodes ;

: .SELECTORNAME ( a -- )  cell+ cell+ zcount cr type ;

: .COCOASELECTORS ( -- )   ['] .selectorname cocoaselectors allnodes ;

\ --------------------------------------------------------------------
\ COCOA: family

\ debug info
: .METHOD-OUTPUT ( -- )
   pocket zcount cr type space ." method returns " #returns ? ;

: .OUTPUT-WARNING ( -- )
   #returns @ dup 2 = swap 0 = or if .method-output then  ;
   
: COCOASELECTOR ( <name-sel> <... parameters ...> -- )
   >in @ >r  parse-params #returns ! 2+               \ add obj and sel now
   r@ >in !  bl parse pocket zplace pocket @selector
   .output-warning
   r> >in !  create
   here -rot , , pocket zcount z, >cocoaselectors ;

: COCOA: ( <name-sel> <... parameters ...> -- )
   cocoaselector forcocoa ;

: SUPERCOCOA: ( <name-sel> <... parameters ...> -- )
   cocoaselector forsupercocoa ;

: COCOA-STRET: ( <name-sel> <... parameters ...> -- )
   cocoaselector forcocoa-stret ;

: SUPERCOCOA-STRET: ( <name-sel> <... parameters ...> -- )
   cocoaselector forsupercocoa-stret ;

: COCOA-FPRET: ( <name-sel> <... parameters ...> -- )
   cocoaselector forcocoa-fpret ;

\\ ( eof )