(* ====================================================================
Extra window utillities.

copyright 2002 Roelf Toxopeus

Part of GUI stuff.
VFX version.
Last: 4 February 2019 at 12:45:46 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
Extra window utils, as used in MacForth.
There is much more you can do, add them when needed.

@MAINSIZE -- return dimensions fullscreen in given rect structure.
FULL.FRAME -- return dimensions fullscreen, uses w.bounds field from
given window.
FULL.SCREEN -- sets given window to full screen.
CONTENT.FRAME -- return dimensions contentview from given window.
CONTENT.SIZE -- return width and height contentview from given window.
MAKE.CONTENTVIEW -- set contentview for window. Main thread execution.

Some background colour presets for window:
DEFAULT.BG BBE.BG WHITE.BG RED.BG BLACK.BG
-------------------------------------------------------------------- ))

/forth
decimal

cocoaclass NSScreen
cocoa: @mainScreen ( -- NSScreen:mainscreen )

: @MAINSIZE ( a -- )   NSScreen @mainScreen @frame drop ;

: FULL.FRAME ( wptr4 -- ) ( F: -- x y w h )   +w.bounds dup @mainsize cg4@ ;
   
: FULL.SCREEN ( wptr4 -- )   dup >r full.frame  r> fix.window ;

\ get frame rect contentview
cocoa: @contentView ( -- contentviewref )

: CONTENT.FRAME ( wptr4 -- ) ( F: -- x y w h )
   dup +w.bounds dup >r swap +w.ref @ @contentView @frame Drop r> cg4@ ;

: CONTENT.SIZE ( wptr4 -- ) ( F: -- width height )
   content.frame f2swap fdrop fdrop ;

\ set window contentview
cocoa: @setContentView: ( viewref -- ret )

: MAKE.CONTENTVIEW ( viewref wptr4 -- )
   +w.ref @ >r
   z" setContentView:" @selector swap yes r> formain drop ;

(*
: DIMENSIONS ( F: T L B R -- wide high )    ;

: TL.WINDOW ( wptr4 -- ) ( F: -- top left )   ;

: BL.WINDOW ( wptr4 -- ) ( F: -- bottom right )  ;
*)

\ Some colour presets
cocoa: @redColor ( -- red )               \ NSColor object
cocoa: @whiteColor ( -- white )           \ NSColor object
cocoa: @blackColor ( -- black )           \ NSColor object
cocoa: @windowBackgroundColor ( -- ref )  \ macOS default background window color

: DEFAULT.BG ( wptr4 -- )
   NSColor @windowBackgroundColor swap +w.ref @ @setBackgroundColor: drop ;

: BBECOLOR ( -- NSColorRef )   \ BBEdit background window color
   248 s>f 255e0 f/
   249 s>f 255e0 f/
   248 s>f 255e0 f/
   255 s>f 255e0 f/
   rgba>nscolor ;
   
: BBE.BG ( wptr4 -- )
   bbecolor swap +w.ref @ @setBackgroundColor: drop ;

: WHITE.BG ( wptr4 -- )
   NSColor @whiteColor swap +w.ref @ @setBackgroundColor: drop ;

: RED.BG ( wptr4 -- )
   NSColor @redColor   swap +w.ref @ @setBackgroundColor: drop ;

: BLACK.BG ( wptr4 -- )
   NSColor @blackColor swap +w.ref @ @setBackgroundColor: drop ;

\\ ( eof )