(* ====================================================================
WebKit browsing

copyright 2006 Roelf Toxopeus

VFX version and Snowleopard Mac OSX 10.6 and later
Playing with webkit, must be called from main thread!!!
Last: 21 Dec 2016 07:46:19 CET   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Implement a crude webbrowser.

WEBWIN -- window for web content.
NEW.WEBVIEW -- create new webview instance, fitting the contentview.
Use INIT, no need for a frame size, works on Panther and upwards.
MAKE-WEBVIEW -- creates new webview and connects to given window.
HELLO-WEB -- initialise and open a window for web browsing
BYE-WEB -- closes it.

LOAD -- will open window and load and display the given url.
<GO -- attemp to go to back
GO> -- attemp to go forward
RELOAD -- reload current url

NoTE: as of OS X 10.11 many NSWindow NSView and WebKit usage should be executed on the main thread.
Examples of strategic placing PASS ...

Following words are executed on Main thread:
SHOW-VIEW ( ADD.WINDOW)
NEW.WEBVIEW
MAKE-WEBVIEW ( SHOW-VIEW NEW.WEBVIEW and MAKE.CONTENTVIEW)
LOADREQUEST
GOBACK
GOFORWARD
RELOAD

This is what you see when URL loading and navigating is not executed on the main thread:

maps  ok
knmi  ok
<go 2016-10-16 10:43:30.125 coco-sf[661:17814] WebKit Threading Violation - -[WebHistoryItem initWithWebCoreHistoryItem:] called from secondary thread
2016-10-16 10:43:30.125 coco-sf[661:17814] Additional threading violations for this function will not be logged.
2016-10-16 10:43:30.126 coco-sf[661:17814] WebKit Threading Violation - DOMDocument *kit(WebCore::Document *) called from secondary thread
2016-10-16 10:43:30.127 coco-sf[661:17814] Additional threading violations for this function will not be logged.
2016-10-16 10:43:30.127 coco-sf[661:17814] WebKit Threading Violation - DOMNode *kit(WebCore::Node *) called from secondary thread
2016-10-16 10:43:30.127 coco-sf[661:17814] Additional threading violations for this function will not be logged.

and the webkit activities stalls.

-------------------------------------------------------------------- ))


/forth
decimal

\ --------------------------------------------------------------------
\ Window for webview:
new.window WEBWIN

: SHOW-VIEW ( wptr4 -- )
   dup add.window
   0" Web View" swap set.wtitle ;

\ --------------------------------------------------------------------
\ Create a webview:

frameworK WebKit.framework
WebKit.framework

cocoaclass WebView

cocoa: @mainFrame ( -- frame )               \ WebView

variable NATIVEVIEW      \ global usage
variable MAINWEBFRAME   \ global usage

\ : NEW.WEBVIEW ( -- WebViewRef )  WebView @alloc @init ;

\ Main thread redefinitionS
: (NEW.WEBVIEW) ( -- )  WebView @alloc @init pocket ! ;
: NEW.WEBVIEW ( -- WebViewRef )  0 dup ['] (new.webview) pass pocket @ ;

\ SHOW-VIEW NEW.WEBVIEW and MAKE.CONTENTVIEW are executed on main thread
: MAKE-WEBVIEW ( wptr4 -- )
   dup >r show-view
   new.webview dup nativeview !            \ make new webview instance and save
   dup r> make.contentview                  \ connect to nswindow
   @mainFrame mainwebframe !               \ retrieve the main webframe and save
;

: HELLO-WEB ( -- )   webwin dup +w.ref @ if drop exit then make-webview ;

: BYE-WEB ( -- )   webwin close.window ;

\ --------------------------------------------------------------------
\ Interface: load url in web view

cocoaclass NSURL
cocoaclass NSURLRequest

cocoa: @URLWithString:   ( NSStringRef -- NSURLRef )    \ NSURL
cocoa: @requestWithURL: ( NSURLRef -- NSURLRequesRef ) \ NSURLRequest
cocoa: @loadRequest: ( NSURLRequesRef -- ret )          \ WebFrame

: REQUEST ( 0string -- NSURLRequesRef )
   >nsstring nsurl @URLWithString:
   NSURLRequest @requestWithURL: ;
   
: LOADREQUEST ( 0string frame -- )  >r request r> @loadRequest: drop ;

: LOAD ( url -- )   hello-web mainwebframe @ 2 0 ['] loadrequest pass ;

\ --------------------------------------------------------------------
\ Interface: more controls

cocoa: @canGoBack ( -- flag )     \ BOOL
cocoa: @goBack ( -- flag )        \ BOOL
cocoa: @canGoForward ( -- flag )  \ BOOL
cocoa: @goForward ( -- flag )     \ BOOL 
cocoa: @reload: ( sender -- ret ) \ id

: GOBACK ( obj -- )   dup @canGoBack if  @goBack  then  drop  ;
: GOBACK ( obj -- )   1 0 ['] goback pass ;
: <GO ( -- )  nativeview @ goback ;

: GOFORWARD ( obj -- )   dup @canGoForward if  @goForward  then  drop ;
: GOFORWARD ( obj -- )   1 0 ['] goforward pass ;
: GO> ( -- )   nativeview @ goforward ;

: RELOAD ( -- )   mainwebframe @ nativeview @ @reload: ;
: RELOAD ( -- )   0 0 ['] reload pass ;

\ --------------------------------------------------------------------
\ Shortcuts:

: BMB   0" http://www.bmbcon.org" load ;

: KNMI   0" https://www.knmi.nl" load ;

: COCOA   0" http://developer.apple.com/library/mac/index.html" load ;

: INFO   0" https://www.ixquick.com/" load ;

: UBU   0" http://www.ubu.com/" load ;

: VFXOSX   0" https://www.mpeforth.com/vfxmac.htm" load ;

: MAPS ( -- )   0" https://www.openstreetmap.org/#map=12/51.9811/5.9055" load ;

\\ ( eof )