(* ====================================================================
Early binding Cocoa

copyright 2011 Roelf Toxopeus
varXcall copyright 2011-2016 MicroProcessor Engineering
 
VFX version.
Calling Cocoa methods directly, bypassing message sending.
Just testing, no relinking.
Last: 1 Nov 2015 14:50:41 CET  -rt
==================================================================== *)

(( --------------------------------------------------------------------
Further development of the method swizzling idea's found in swizzling-
test.f
Create FUNCTION: like words for early binding, i.e. no method lookup,
ObjC methods.

OBJC-EXTERN-CALL -- external ObjC call with var args. The given address
can be one of the message_send functions or the real implementation
of the method belonging to the obj and sel. The reshuffling of the args
conforming the ObjC ABI is taken care of. 
OBJC-EXTERN-CALL is a variation on VFX's varXcall.

INSTANCEMETHOD -- create structure: imp sel #parameters
for fast objc call: add obj and sel to parameters: 2+
FASTCOCOA -- will retrieve the imp sel and #parameters and passes
them to OBJC-EXTERN-CALL directly.
INSTANCEFUNC -- will create a fast or early binding instance method.

CLASSMETHOD similar to INSTANCEMETHOD, but for class methods.
CLASSFUNC similar to INSTANCEFUNC, but for class methods.

Note: method implementations follow the rules of their messages:
so take care of mainthread recuirements.
-------------------------------------------------------------------- ))

/forth
decimal

code OBJC-EXTERN-CALL      ( x1..xn obj sel n+2 addr -- rc )
\ preserve Forth state
      push   esi               \ ESI = UP
      push   edi               \ EDI = LP
\ make calling frame
      mov   ecx, 0 [ebp]      \ n=#args
      shl   ecx, # 2            \ cells to bytes
      sub   ebp, # 8            \ two more items on data stack

\ 12 [ebp]  xn, last argument
\  8 [ebp]  call address
\  4 [ebp]  #bytes of args
\  0 [ebp]  ESP value
\ shuffle arguments to be like NXCALL
      mov   8 [ebp], ebx      \ call address
      mov   4 [ebp], ecx      \ #bytes of args on data stack
      mov   0 [ebp], esp      \ current return stack to be restored
      sub   esp, ecx            \ make call frame
      and   esp, # -16         \ align stack

\ copy data from Forth EBP stack to calling ESP frame
      test   ecx, ecx            \ to be sure... must be at least 2
   nz,    if,
      lea   esi, #8 [ebp] [ecx]   \ source address
        lea   edi, 0 [esp]         \ dest address

\ first copy the obj and sel parameters to the 'bottom'
      mov    eax, 16 [ebp]      \ copy xn-1 = obj
      mov   0 [edi], eax
      mov   eax, 12 [ebp]      \ copy xn = sel
      mov   4 [edi], eax
      lea   edi, 8 [edi]      \ update pointer
      sub   ecx, # 8            \ 2 items less  

      test   ecx, ecx            \ any other args?
    nz,    if,

\ then copy the real arguments
      begin,
       mov   eax, 0 [esi]      \ copy data, update pointers
       lea   esi, -4 [esi]
       mov   0 [edi], eax
       lea   edi, 4 [edi]
       sub   ecx, # 4
      z, until,
    then,
   then,

\ call external function and restore stacks
      cld                     \ force D bit just in case
      call   ebx
      cld
      mov   esp, 0 [ebp]

\ Clean arguments from Forth data stack
      mov   ebx, 4 [ebp]      \ number of bytes of args
      lea   ebp, #12 [ebp] [ebx]   \ remove transients, addr, args

\ restore Forth and return result
      mov   ebx, eax            \ return value
      pop   edi
      pop   esi
      next,
end-code

: @IMP ( class 0string -- sel imp )
   @selector tuck
   class_getInstanceMethod dup 0= abort" No Method for Selector !"
   method_getImplementation dup 0= abort" No IMP for Method !" ;

\ struct: sel #in imp
: INSTANCEMETHOD ( class <name> -- )
   >in @ >r  parse-params drop 2+      \ add obj and sel now
   r@ >in ! bl word count pocket zplace 
   r> >in ! swap pocket @imp rot rot   ( -- imp #in sel )
   create , , , ;
   
: 3@ ( a -- n1 n2 n3 )
   dup @ swap cell+
   dup @ swap cell+
   @ ;

: FASTCOCOA ( n0 ... nx obj a -- ? )
   3@                                  ( -- n0 ... nx obj sel #in imp )
   objc-extern-call ;

: INSTANCEFUNC ( class <name-sel> -- )
   instancemethod
   does>  ( n0 ... nx obj a -- ? )
      fastcocoa ;

\ ---

: @CLASSIMP ( class 0string -- sel imp )
   @selector tuck
   class_getClassMethod DUP 0= ABORT" No Method for Selector !"
   method_getImplementation DUP 0= ABORT" No IMP for Method !" ;

: CLASSMETHOD ( class <name> -- )
   >in @ >r  parse-params drop 2+      \ add obj and sel now
   r@ >in ! bl word count pocket zplace 
   r> >in ! swap pocket @classimp rot rot   ( -- imp #in sel )
   create , , , ;

: CLASSFUNC ( class <name-sel> -- )
   classmethod
   does>  ( n0 ... nx obj a -- ? )
      fastcocoa ;

\\ ...

\ testing:

NSWINDOW INSTANCEFUNC @close ( -- ret )

NEW.WINDOW WIN  WIN ADD.WINDOW

: CLOSE.IT ( -- )  win +w.ref @ 1 0 ['] @close pass ;

\\ ( eof )
