(* ====================================================================
File and URL app launching.

copyright 2013 Roelf Toxopeus

VFX version.
Use of NSWorkspace to open files/locations in their default applications.
Last: 25 March 2018 at 14:54:36 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
LAUNCHAPP -- launches new app instance. Example of toll-free bridging
between CFURL and NSURL. Here CFURL is used.

LAUNCHURL -- launches default browser with given url. NSURL is used.
GOTO" -- browser and go to quoted url, as example:
VFXWEB -- go to MPE home web page.

LAUNCHFILE -- launches default opener with given file.
VFXMANUAL -- open VFX manual in default PDF viewer.

More examples of NSWorkspace usage:
\ launch new Terminal instance. Interesting...
\ as if it launched a copy of currently running Terminal.app
\ with coco-vfx and whatever else running !
: >TERMINAL ( -- )
  z" /Applications/Utilities/Terminal.app" launchapp 2drop ;

\ launch another coco-vfx and quit current
: HELLO ( -- )
   pad dup GetExeNameZ launchfile  bye ;

\ hide all applications from sight
: HH ( -- )
  NSWorkspace @sharedWorkspace @hideOtherApplications drop ;
-------------------------------------------------------------------- ))

/forth
decimal

cocoaclass NSWorkspace
cocoa: @sharedWorkspace ( -- ref )
cocoa: @openURL: ( nsurl -- boolean )
cocoa: @openFile: ( nsstring -- boolean )
cocoa: @launchApplication: ( string -- ret )
cocoa: @launchApplicationAtURL:options:configuration:error: ( url options dict *error -- 0|ref )
cocoa: @hideOtherApplications ( -- ret )

cocoaclass NSURL
cocoa: URLWithString: ( nsstring -- nsurl )

cocoaclass NSDictionary
cocoa: @dictionary ( -- ref )

$00080000 constant NSWorkspaceLaunchNewInstance

: LAUNCHAPP ( zstring -- ref error)
   >cfurl dup
   NSWorkspaceLaunchNewInstance
   NSDictionary @dictionary
   0 >r rp@
   NSWorkspace @sharedWorkspace
   @launchApplicationAtURL:options:configuration:error: r>
   rot CFRelease drop ;

: LAUNCHURL ( zstring -- )
   >nsstring dup
   NSURL URLWithString:
   NSWorkspace @sharedWorkspace @openURL: drop
   @release drop ;

\ : URL: ( "<spaces>name" -- )   bl word count drop LaunchURL ;
   
: GOTO" ( -- )
   postpone z"
   state @
         if postpone launchurl
      else launchurl
   then ; immediate

: VFXWEB ( -- )   GOTO" http://www.mpeforth.com" ;

: LAUNCHFILE ( zstring -- )   
   >nsstring dup
   NSWorkspace @sharedWorkspace @openFile: drop
   @release drop ;

: VFXMANUAL ( -- )
   ms" %vfxpath%" pad zplace
   s" /Doc/VfxOsx.pdf" pad zappend
   pad launchfile ;

cr .( VFXWEB     will bring you to MPE)
cr .( VFXMANUAL  will open the VFX manual)

\\ ( eof )