(* ====================================================================
Another GUI example

copyright 2011 Roelf Toxopeus

VFX version.
Load a nib file with buttons and assign words to them.
Last: 7 August 2020 at 15:54:47 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Cocoa from within VFX demonstration:

Using a grid of buttons in a NIB.
Add tags in IB for every button: 0 up to 11
Use @tag to retrieve tag, which will be used as index.
This allows for an indexed execution array and only one method is needed!
Could have tagged the buttons with the calculated indexes: e.g. 0 4 8 12 etc.

Biggest challenge is getting the nib right in XCode4
What a lousy interface! This is 2013 remember...

NoTE: the Forth running the actions, is NOT the Forth thread IMPOSTOR.
It's a callback running on another thread/GCD-queue.
This can be either the Main thread, or a thread/dispatch-queue spawned
by MAIN, or perhaps even an 'ordinary' callback. You don't know and
maybe it's completely irrelevant as long as you're aware it's not IMPOSTOR.
You can suspend IMPOSTOR when needed.

DEF.ACTION -- default actions for buttons:
BUTTONS -- execution array for buttons, fill it with the defualt actions
DO.BUTTON -- default actions for testing

MyButtonsClassTagged -- class taking care of button actions, using
one method:    button:

/BUTTONS -- init and run demo
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ Utillities, Actions, etc.

\ NSControl class
cocoa: @intValue ( -- N )      \ for the buttons, this will be 'on/off' toggling
cocoa: @tag ( -- NSInteger )   \ tag is set to whatever you want in IB, here indexes

: DEF.ACTION ( sender -- )   @intValue . ;

12 cells buffer: BUTTONS
' def.action buttons !
buttons buttons cell+ 11 cells cmove 

: DO.BUTTON ( sender -- )
   impostor's dup @tag dup 1+ . ." : "
   cells buttons + @execute ;

\ --------------------------------------------------------------------
\ the so called IBActions connected to the buttons

\ nice syntactic sugar
synonym IBACTION: forthcall:

ibaction: *button: ( rec sel sender -- n )  nip nip do.button 0 ;

: BUTTONMETHODTYPES ( -- addr )   0" v@:V" ;

\ --------------------------------------------------------------------
\ this will be the delegate class for the button actions in the nib.
\ the OS takes care of assigning it as such.

NSObject new.class MyButtonsClassTagged

\ assign callback ( ibaction ) as method to the class and register its name with the objc runtime
*button:  buttonmethodtypes 0" button:"   MyButtonsClassTagged add.method

\ add the class
MyButtonsClassTagged add.class

\ --------------------------------------------------------------------
\ The Nib stuff

\ : /BUTTONS ( -- )   z" MyButtonsTagged.nib" @nib /nib 0= abort" Can't initiate NIB !" ;
: /BUTTONS ( -- )   z" MyButtonsTagged.nib" show.nib 0= abort" Can't initiate NIB !" ;

cr .( /buttons   -- to add buttons window to coco-vfx)

\\ ( eof )