(* ====================================================================
ShellView Console

Copyright 2014 Roelf Toxopeus

VFX version.
Behaves (hopefully) as a Terminal.app console.
Last: 27 September 2020 at 16:33:13 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
/DEFAULT-CONSOLE -- sets default-console as the I/O sid. This offers
a simple console, so no special handling for esc stuff is needed.
See syspatch.fth  and  console.fth
In VFX ACCEPT the cr=13 is thrown away, it's ignored. But a TextView
sends a 13 when hitting cr. Rather than rewritting the accept routine
for the default-console, it's dealt with in the shellview keyhandler.
See sfshellview7.fth -> *keyDown:

/CONSOLE -- sets the console to the sid saved by /DEFAULT-CONSOLE.

PIPE1 PIPE2 -- variables containing the NSPipeRefs.
Two pipes are needed: shellviewIn  <- forthOut
                      shellviewOut -> forthIn
/PIPES -- create NSPipe instances for I/O. Store refs in PIPE1, PIPE2.
-PIPES -- release NSPipe instances in PIPE1, PIPE2.
/IO -- create pipes for I/O, set the appropriate fd's in Forth's
stdin and stdout, return the appropriate NSFileHandle refs.
These will be used to create a SVShellView instance.

/STDIO -- set stdin and stdout to default standard values.

MYSHELLWIN -- the window for the shellview

/VIEWIO -- init shellview in window and set Forth I/O accordingly.
This makes MYSHELLWIN the I/O console for Forth.

-VIEWIO -- restore to default Terminal console. Restore I/O sid and
release used NSPipe refs.

/TERMINAL -- -VIEWIO and restarts QUIT. Warm boot in Terminal console.

/COCOA-CONSOLE -- /VIEWIO and restarts QUIT. Warm boot in new console
window

---

When running coco-vfx in the cocoa console, you can disconnect from
the Terminal if you whish.
1. Bring the Terminal console to the front
2. type  CTRL-Z
3. type  disown -h; fg <ret>
4. Close the Terminal console window or just quit Terminal.app

-------------------------------------------------------------------- ))

/forth
decimal

also system

\ --------------------------------------------------------------------
\ simple console

variable IP-SAVED variable OP-SAVED variable PACO-SAVED

: /DEFAULT-CONSOLE	( -- )
   ip-handle @ ip-saved ! op-handle @ op-saved ! PauseConsole paco-saved !
   default-console dup ip-handle ! dup op-handle ! to PauseConsole ;

: /CONSOLE ( -- )
   ip-saved @ ip-handle ! op-saved @ op-handle ! paco-saved @ to PauseConsole ;

\ --------------------------------------------------------------------
\ the pipes for IPC between Forth and SFShellView

cocoaclass NSPipe
cocoa: @pipe ( -- NSPipeRef )    \ class method
cocoa: @fileHandleForReading ( -- NSFileHandle )
cocoa: @fileHandleForWriting ( -- NSFileHandle )

variable PIPE1 variable PIPE2

: /PIPES ( -- pipe1 pipe2 )
   NSPipe @pipe dup 0= abort" Could not create a NSPipe instance !"
   dup pipe1 !
   NSPipe @pipe dup 0= if
        pipe1 dup @ @release drop off ." Could not create a NSPipe instance !"
        abort
   then
   dup pipe2 ! ;

: -PIPES ( -- )
   pipe1 dup @ @release drop off
   pipe2 dup @ @release drop off ;

\ COCOACLASS NSFileHandle
cocoa: @fileDescriptor ( -- Posixfd )

: /IO ( -- fhin fhout )
   /pipes
   /default-console
   >r dup @fileHandleForReading swap @fileHandleForWriting        ( fhin1 fhout1 )
   @fileDescriptor stdout !   \ xconsole xs.hout !                ( fhin1 )
   r> dup @fileHandleForReading swap @fileHandleForWriting        ( fhin1 fhin2 fhout2 )
   swap @fileDescriptor stdin !   \ xconsole xs.hin !             ( fhin1 fhout2 )
;

: /STDIO ( -- )   0 stdin !  1 stdout ! ;  \ 0 xconsole xs.hIn ! 1 xconsole xs.hout ! ;

previous

\ --------------------------------------------------------------------
\ the window and the SFShellView contentview

new.window MYSHELLWIN
100e0 100e0 500e0 200e0 myshellwin w.bounds
s" Coco-VFX I/O View"   myshellwin w.title

variable MYSHELLVIEW

: /VIEWIO ( -- )
   /io                          \ do this first, if no pipe, don't bother to continue
   myshellwin add.window
   myshellwin windowforshell dup myshellview !
   z" Monaco" 14 rot viewfont
;

: -VIEWIO ( -- )
   -pipes
   myshellview dup @ @release drop off
   \ myshellwin close.window
   /stdio
   /console ;

\ --------------------------------------------------------------------
\ the topwords

: /TERMINAL ( -- )
   -viewio
   cr ." I/O redirected to and from Terminal.app"
   .prompt
   myshellwin 1 0 ['] close.window pass
   quit ;

: /COCOA-CONSOLE ( -- )
   /viewio
   .about
   cr cr ." hello again"  cr
   quit ;

CR .( shellview console loaded)
CR .( /COCOA-CONSOLE   to redirect I/O to shellview)
CR .( /TERMINAL  to restore I/O to Terminal console)

\\ ( eof )