(* ====================================================================
Control panel Nib

copyright 2011 Roelf Toxopeus

VFX version.
Implement a control panel with three buttons and a slider: 4 defered actions
Last: 7 August 2020 at 15:51:24 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Cocoa from within VFX demonstration:

Demonstrates the interface to the objc-runtime in action:
-- some Cocoa calls
-- creating a new class for cocoa

NoTE: in VFX set the I/O stuff right for displaying strings.
     Use IMPOSTOR'S and don't forget to set FP stack in callback

First add some deferred actions for the buttons and the slider:
BUTTON1   BUTTON2   BUTTON3   SLIDER
/ACTIONS -- sets these to default actions.
Plug in other actions at will.

Then create the callbacks *BUTTON1  *BUTTON2  *BUTTON3  *SLIDER
which are the implementations for the methods, IBActions, declared
later in the TestController class. All callbacks receive 3 parameters
   receiver selector sender
Only the sender is passed to the deferred word. Its consumption is
mandatory. All callbacks return a zero on top.
Here CALLBACK is aliased with IBACTION: for consistency with Apple's
terminology in Interface Builder.

Now create the TestController class and add the methods
doButton1:  doButton2:   doButton3:   doSlider:
using the implementations created earlier.
Add the class to the runtime.

Get the panel's NIB testcontroller.nib and create nib window
controller TCCONTROLLER and show it with DO.SHOW.

Press buttons and move slider to execute the actions.
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ Utillities, Actions, etc.

defer BUTTON1   defer BUTTON2   defer BUTTON3   defer SLIDER

variable SCALER 1 scaler !

: SCALER+ ( -- )   impostor's cr ." scale by " 1 scaler +! scaler ? ;
: SCALER- ( -- )   impostor's cr ." scale by " -1 scaler +! scaler ? ;
: /SCALER ( -- )   impostor's cr ." reset scaler"  1 scaler ! ;

cocoa: @intValue ( -- n )  \ NSControl method
: SCALE ( x -- )   impostor's cr @intValue scaler @ * . ;

: /ACTIONS ( -- )
   ['] scaler+ is button1 ['] scaler- is button2  ['] /scaler is button3
   ['] scale is slider ;

\ --------------------------------------------------------------------
\ the so called IBAction connected to the buttons and slider in the panel

synonym IBACTION: forthcall:

ibaction: *BUTTON1 ( rec sel sender -- n )   nip nip button1  0 ;

ibaction: *BUTTON2 ( rec sel sender -- n )   nip nip button2  0 ;

ibaction: *BUTTON3 ( rec sel sender -- n )   nip nip button3  0 ;

ibaction: *SLIDER ( rec sel sender -- n )    nip nip slider   0 ;

: CONTROLTYPES ( -- a )   0" v@:V" ;

\ --------------------------------------------------------------------
\ The ForthClass, create class

NSObject new.class TestController

\ Assign callbacks as methods to the class and register their names with the objc runtime
*button1 controltypes 0" doButton1:" TestController add.method

*button2 controltypes 0" doButton2:" TestController add.method

*button3 controltypes 0" doButton3:" TestController add.method

*slider  controltypes 0" doSlider:"  TestController add.method

\ Add the class to the runtime

TestController add.class

\ --------------------------------------------------------------------
\ The Nib stuff

\ Topword
\ : DO.SHOW ( -- )   /actions z" testcontroller.nib" @nib /nib 0= abort" Can't initiate NIB !" ;
: DO.SHOW ( -- )   /actions z" testcontroller.nib" show.nib 0= abort" Can't initiate NIB !" ;

cr .( control panel with buttons and slider loaded ...)
cr .( do.show   -- to start things)

\\ ( eof )