(* ====================================================================
vfx menu

copyright 2008 Roelf Toxopeus

VFX version.
Add an application menu to coco-vfx.
Last: 15 September 2020 at 11:08:40 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
The following methods will be defined here:
menuQuit: -- for quit/bye menu, terminate NSApp. Run @terminate.

menuOpenFile: -- for Open File... menu, to open file in editor just
execute EDIT-FILE. 

menuLoadFile: -- for Load File... menu, execute a special LOAD-FILE,
DO-LOAD-FILE. To be safe put IMPOSTOR to sleep while loading.
The @release actions in LOAD-FILE will give trouble when used from
the menu.
NoTE: Some things are very thread related, f.i. graphic contexts
Initialization of such things during including a file should be
reinitialized after including. /GWINDOW is such an initializer.
Using LOAD-FILE directly in stead of menuLoadFile: doesn't have
this problem.

menuLoadEdit: -- for Load from Editor menu item, a special version for
LOAD-EDIT, DO-LOAD-EDIT.

menuVerbose: -- for the verbose/silent items. VERBOSE and SILENT wrapped
in DO-VERBOSE.

menuAbort: -- executes the defered DO-ABORT. Plug in what's appropriate,
for instance an exceptionhandler.
It reacts to the <esc> key, when the coco-sf part is up front.
Starts up with PANIC as default. Which sends a SIGINT signal to the
current process. Default response to SIGINT is termination of process.
  
For the Help menu's:
menuHelp: -- open swiftforth-linux-osx.pdf in PDF viewer, HELPME.
menuHelp2: -- open the coco-sf readme.pdf in PDF viewer, HELPME2. 
menuHelp3: -- open dpans94.pdf in PDF viewer, HELPME3.
menuHelp4: -- open handbook.pdf in PDF viewer, HELPME4.
 
All added to the SFMenuClass. The ObjC runtime calls these when the
corresponding menu's are activated. No need to create class instance
or delegate, the OS takes care.

VFXMenuClass and its methods is connected to the corresponding menubar
in vfxmenu.nib.
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ Utillities, Actions, etc.

cocoa: @terminate: ( sender -- ret )

: DO-LOAD-FILE ( -- )
   impostor's /forth state off ( impostor sleep )
   load-file cr
   ( impostor wake )
;

: DO-LOAD-EDIT ( -- )
   impostor's forth state off ( impostor sleep )
   load-edit cr
   ( impostor wake )
;

cocoa: @mainMenu ( -- id ) \ NSMenu object
cocoa: @itemChanged: ( menuitem -- ret ) \ NSMenuItem object
variable ~VERBOSE

: DO-VERBOSE ( menuitem -- )
   ~verbose dup @ tuck invert swap !
      if silent z" Verbose"
      else verbose z" Silent"
   then >nsstring over @setTitle: drop
   NSApp @ @mainMenu @itemChanged: drop ;

DEFER DO-ABORT

FUNCTION: kill ( pid signal -- ret )
: PANIC ( -- )   getpid 2 kill  DROP ;

: HELPME ( -- )
   s\" open \"" pad zplace
   ms" %vfxpath%" pad zappend
   s\" /Doc/VfxOsx.pdf\"" pad zappend
   pad ssystem drop ;

: HELPME2 ( -- )
   s\" open \"" pad zplace
   ms" %cocoalib%"  pad zappend       ( NoTE: ends with slash )
   s\" doc/readme.pdf\"" pad zappend
   pad ssystem drop ;

: HELPME3 ( -- )
   s\" open \"" pad zplace
   ms" %vfxpath%" pad zappend
   s\" /Doc/AnsForth.htm/dpans.htm\"" pad zappend
   pad ssystem drop ;

: HELPME4 ( -- )
   s\" open \"" pad zplace
   ms" %vfxpath%" pad zappend
   s\" /Doc/ProgramForth.pdf\"" PAD ZAPPEND
   pad ssystem drop ;

\ --------------------------------------------------------------------
\ the so called IBActions connected to the menus in the main menubar

variable VFXMenuClassREF

forthcall: *awakeFromNib ( rec sel -- ret )   drop VFXMenuClassREF ! 0 ;

forthcall: *menuQuit: ( rec sel sender -- n )  nip nip NSApp @ @terminate: drop 0 ;

forthcall: *menuOpenFile: ( rec sel sender -- n )  finit 2drop drop edit-file 0 ;

forthcall: *menuLoadFile: ( rec sel sender -- n )  finit 2drop drop do-load-file 0 ;

forthcall: *menuLoadEdit: ( rec sel sender -- n )  finit 2drop drop do-load-edit 0 ;

forthcall: *menuVerbose: ( rec sel sender -- n )  nip nip do-verbose  0 ;

forthcall: *menuAbort: ( rec sel sender -- n )  2drop drop do-abort 0 ;

forthcall: *menuHelp: ( rec sel sender -- n )  2drop drop helpme 0 ;

forthcall: *menuHelp2: ( rec sel sender -- n )  2drop drop helpme2 0 ;

forthcall: *menuHelp3: ( rec sel sender -- n )  2drop drop helpme3 0 ;

forthcall: *menuHelp4: ( rec sel sender -- n )  2drop drop helpme4 0 ;

: menuMethodTypes ( -- addr )   0" v@:@" ;

\ --------------------------------------------------------------------
\ The ForthClass

NSObject new.class VFXMenuClass

\ --- assign callback as method to the class and register its name with the objc runtime

*awakeFromNib  menuMethodTypes 0" awakeFromNib"  VFXMenuClass add.method   \ internaly used by runtime at instantiating nib

*menuQuit:     menuMethodTypes 0" menuQuit:"     VFXMenuClass add.method

*menuOpenFile: menuMethodTypes 0" menuOpenFile:" VFXMenuClass add.method

*menuLoadFile: menuMethodTypes 0" menuLoadFile:" VFXMenuClass add.method

*menuLoadEdit: menuMethodTypes 0" menuLoadEdit:" VFXMenuClass add.method

*menuVerbose:  menuMethodTypes 0" menuVerbose:"  VFXMenuClass add.method

*menuAbort:    menuMethodTypes 0" menuAbort:"    VFXMenuClass add.method

*menuHelp:     menuMethodTypes 0" menuHelp:"     VFXMenuClass add.method

*menuHelp2:    menuMethodTypes 0" menuHelp2:"    VFXMenuClass add.method

*menuHelp3:    menuMethodTypes 0" menuHelp3:"    VFXMenuClass add.method

*menuHelp4:    menuMethodTypes 0" menuHelp4:"    VFXMenuClass add.method

VFXMenuClass add.class

(( --------------------------------------------------------------------
  The Nib stuff
  VFXMENU -- topword, initializes the main menu and draws it.
-------------------------------------------------------------------- ))

\ : VFXMENU ( --  )   ['] panic is do-abort  z" vfxmenu.nib" @nib /nib 0= abort" Can't initiate NIB !" ;
: VFXMENU ( --  )   ['] panic is do-abort  z" vfxmenu.nib" show.nib 0= abort" Can't initiate NIB !" ;

cr .( simple menu for coco-vfx loaded ...)
cr .( vfxmenu   -- to add main menu to coco-vfx)

\\ ( eof )