(* ====================================================================
CoreGraphicsContext

Copyright 2006 Roelf Toxopeus

VFX version.
Core Graphics basic setup for displaying images and pictures.
Last: 3 February 2020 at 15:24:04 CET   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Using Quartz in NSWindow's needs CGContexts. Here many CGContext related
functions are defined.
See the Quartz 2D Programming Guide at:
https://developer.apple.com/library/mac/documentation/GraphicsImaging/Conceptual/drawingwithquartz2d/Introduction/Introduction.html#//apple_ref/doc/uid/TP30001066

quartz.framework -- the Quartz 2D libraries.

CGContextFlush -- will execute only at the advised framerate of 60Hz.
This will speed up drawing considerable. 
Trying to flush faster than the framerate will slow down the drawing!
NoTE: uses some finetuning.

FLUSHED -- returnes last time _CGContextFlush got executed. Updated
by CGContextFlush.
 
CGCONTEXT -- will return the CoreGraphicContextRef, if any, for the given window.
It's also known as the graphics port. It uses the NSGraphicsContext instance
method CGContext (or @CGContext as it is declared here).

/GWINDOW -- will fetch the CGContext from the OS for the given window
and store it in its window record.
The CGContext is per thread! If /GWINDOW is executed during an include
from the LOAD-FILE menu, make sure you execute it again from IMPOSTOR
before you run anything which makes use of CGContext.
For now I don't know why other threads don't complain when they use the
the CGContext asked for by IMPOSTOR. Because they're IMPOSTOR children?
Best thing is to initialize on thread which will run the graphic stuff.

GPORT -- return the grahics port or CGContextRef from given window's
window record. It's assumed to be set by /GWINDOW

CGCOLOR and CGRECT are color and rect pads for temporary usage.

Might add the following to simulate MacForth:
\ --- window for Core Graphics output
#USER CELL +USER TASKWINDOW
     CELL +USER TASKCGCONTEXT
TO #USER

\ set output window for graphics
: WINDOW ( wptr4 -- )
  TASKCGCONTEXT @ ?DUP IF CGContextRelease DROP THEN
  DUP @CGCONTEXT TASKCGCONTEXT ! TASKWINDOW ! ;

\ get current output window for graphics
: GET.WINDOW ( -- wptr4 )   TASKWINDOW @ ;
-------------------------------------------------------------------- ))

/forth
decimal

lacking quartz.framework framework quartz.framework

quartz.framework

function: CGContextSaveGState ( cgcontext -- ret )
function: CGContextRestoreGState ( cgcontext -- ret )

function: CGContextRetain ( cgcontext -- cgcontext ) 
function: CGContextRelease ( cgcontext -- ret )

function: CGContextBeginPath ( cgcontext -- ret )      \ hardly used?
function: CGContextClosePath ( cgcontext -- ret )

ascall: _CGContextFlush CGContextFlush ( cgcontext -- ret )         \ Note: use for drawing
function: CGContextSynchronize   ( cgcontext -- ret )    \       use for images

cocoaclass NSGraphicsContext

cocoa: @graphicsContextWithWindow: ( nswindow -- NSGraphicsContextRef ) \ NSGraphicsContextRef
COCOA: @graphicsContext ( -- NSGraphicsContextRef )   \ NSWindow Deprecated
COCOA: @CGContext ( -- CGContextRef )                 \ NSGraphicsContext
COCOA: @graphicsPort ( -- CGContextRef )              \ NSGraphicsContext Deprecated, Snow Leopard only

\ --------------------------------------------------------------------

variable flushed

: CGContextFlush ( cgcontext -- ret )
   counter flushed  2dup @ - 33 u> if ! _CGContextFlush  else 2drop drop 0 then ;

\ Snow Leopard
\ : CGCONTEXT ( wptr4 -- CGContextRef )
\   +w.ref @ NSGraphicsContext @graphicsContextWithWindow: @graphicsPort ;

\ Yosemite and up
: CGCONTEXT ( wptr4 -- CGContextRef )
   +w.ref @ NSGraphicsContext @graphicsContextWithWindow: @CGContext ;

(*
\ Snow Leopard
\ : CGCONTEXT ( wptr4 -- CGContextRef ) +w.ref @ @graphicsContext @graphicsPort ;

\ Yosemite and up
: CGCONTEXT ( wptr4 -- CGContextRef )   +W.REF @ @graphicsContext @CGContext ;
*)

: /GWINDOW ( wptr4 -- )   dup cgcontext swap +w.gport ! ;

: GPORT ( wptr4 -- CGContextRef )   +w.gport @ ;

0E0 fdup fdup fdup cg4: CGCOLOR
0E0 fdup fdup fdup cg4: CGRECT

cr .( CGContext loaded )

\\ ( eof )
