(* ====================================================================
Core Graphics drawing methods

copyright 2006-2017 Roelf Tooxpeus

VFX version
sf can now, 2014, deal with multiple output parameters:
changed defining CGContextGetPathCurrentPoint to return x and y
removed 2RET> where appropriate
See CHANGES.TXT
Last: 14 Nov 2016 16:52:32 CET   -rt
==================================================================== *)

(( --------------------------------------------------------------------
NoTE: all supplied and returned coordinates are fp numbers and passed
via the fp stack.

All words deal with current cgcontext.
LINE.WIDTH -- set linewidth for cgcontext.
PENLOC -- return pen location for cgcontext.
MOVE.TO -- move pen to x/y in cgcontext.
LINE.TO -- add a line using current blendmode etc.
from current penloc to x/y. Doesn't draw it!
SYNC.DRAW -- stroke, flush and set current point in cgcontext
DRAW.TO -- draw a line added to cgcontext, uses SYNC.DRAW
RMOVE -- move pen relative to current pen location.
RDRAW -- draw a line, relative to current pen location.
LINE -- draw a line form x1/y1 to x2/y2, uses DRAW.TO
DOT -- draw a dot at x/y, uses SYNC.DRAW. Pixel size = 5E-1
WIPE or CLEAR.WINDOW.

The jury is still not out on this:
Using integers for coordinates!!
: MOVE.TO  ( x y cgcontext --)
   ROT S>SF ROT S>SF CGContextMoveToPoint DROP ;
  
: LINE.TO ( x y cgcontext --)
   ROT S>SF ROT S>SF CGContextAddLineToPoint DROP ;

: DRAW.TO ( x y cgcontext --)
   DUP >R LINE.TO R> CGContextStrokePath DROP ;

: LINE ( x1 y1 x2 y2 cgcontext --)
   >R 2SWAP R@ MOVE.TO R> DRAW.TO ;

: DOT ( x y cgcontext --)
   >R 2DUP R> LINE ;
-------------------------------------------------------------------- ))

/forth
decimal

quartz.framework

function: CGContextGetPathCurrentPoint ( cgcontext -- x y )
function: CGContextMoveToPoint ( cgcontext sf sf -- ret ) 
function: CGContextAddLineToPoint ( cgcontext sf sf -- ret ) 
function: CGContextAddRect ( cgcontext sf sf sf sf -- ret )

function: CGContextStrokePath ( cgcontext -- ret )
function: CGContextFillPath ( cgcontext -- ret )
function: CGContextFillRect ( cgcontext sf sf sf sf -- ret )
function: CGContextClearRect ( cgcontext sf sf sf sf -- ret )

function: CGContextSetLineWidth ( cgcontext sf -- ret )
function: CGContextSetAlpha ( cgcontext sf -- ret )
function: CGContextSetRGBStrokeColor ( cgcontext sf sf sf sf -- ret )
function: CGContextSetRGBFillColor ( cgcontext sf sf sf sf -- ret )
function: CGContextSetGrayFillColor ( cgcontext sf sf -- ret )

\ --------------------------------------------------------------------

: LINE.WIDTH ( context --) ( F: width --)
   fpushs CGContextSetLineWidth drop ;

: PENLOC ( cgcontext -- ) ( F: -- x y )
   CGContextGetPathCurrentPoint swap sf>f sf>f ;

: MOVE.TO ( cgcontext -- ) ( F: x y -- )
    fswap fpushs fpushs CGContextMoveToPoint drop ;

: LINE.TO ( cgcontext -- ) ( F: x y -- )
   fswap fpushs fpushs  CGContextAddLineToPoint drop ;

(*
: DRAW.TO ( cgcontext -- ) ( F: x y -- )
   f2dup
   dup line.to dup CGContextStrokePath drop
   dup CGContextFlush drop
   move.to ;
*)

: SYNC.DRAW ( cgcontext -- ) ( F: x y -- )
   counter flushed  2dup @ - 33 u> if                                \ time to work?
      !                                                               \ next flushtime
      dup CGContextStrokePath drop dup _CGContextFlush drop move.to  \ stroke, flush and set current point
      else 2drop drop f2drop then ;
   
: DRAW.TO ( cgcontext -- ) ( F: x y -- )
   f2dup dup line.to sync.draw ;

: RMOVE ( cgcontext -- ) ( F: relativex relativey -- )
   penloc point+ move.to ;

: RDRAW ( cgcontext -- ) ( F: relativex relativey -- )
   penloc point+ draw.to ;
   
: LINE ( cgcontext -- ) ( F: x1 y1 x2 y2 -- )
   dup f2swap move.to draw.to ;

(*
: DOT ( cgcontext -- ) ( F: x y -- )
   f2dup
   dup 1e0 fdup 4fpushs CGContextFillRect drop
   dup CGContextFlush drop
   move.to ;
*)

: DOT ( cgcontext -- ) ( F: x y -- )
   f2dup dup 5e-1 fdup 4fpushs CGContextAddRect drop sync.draw ;

: WIPE \ CLEAR.WINDOW ( wptr4 -- )
   dup cgcontext >r
   r@ CGContextSaveGState drop
   cgrect over +w.ref @ @contentView @frame drop
   r@ swap +w.background CGColor@ CGContextSetRGBFillColor drop
   r@ cgrect cgrect@ CGContextFillRect drop
   r@ CGContextRestoreGState drop
   r> CGContextFlush drop ;
   
cr .( Core Graphics basic drawing loaded )

\\ ( eof )
