(* ====================================================================
Schedule word

VFX version

copyright 2016 Roelf Toxopeus

GCD CAUSE version
Last: 4 February 2019 at 13:12:32 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
CAUSE can be used for a one shot timer! See the RECURSIVE examples at
the end reagrding scheduling yourself in again.
CAUSE originates at STEIM (c) Frank Balde 1986. Based on words written
by R.Kuivila and D.Anderson in Forthmacs.
In stead of creating an irq-server like in the original CAUSE implemen-
tations, GCD is used to cause an action in the future.
The GCD dispatch_after_f function is central to the planning.

FROM-NOW - create a 64bit dispatch time using the walltime plus given
nanoseconds in the future.
CAUSE - schedule a Forth word (xt) to a given time from now, using
milliseconds as default. Uses the DEFAULTQUEUE for dispatch

The dispatch routines use nanoseconds:
S>NANO converts seconds to nanoseconds
MS>NAN converts milliseconds to nanoseconds
US>NAN converts microseconds to nanoseconds
These nanoseconds are 2 cell values on 32b Forth systems!

Note:
Activating or dispatching a task starting with nanosleep is not what
CAUSE (or GCD) does.

0 TASK PIPO
: HAIL ( -- )   PIPO ACTIVATE 6000 MS IMPOSTOR'S ." ahoi ahoi" ;
HAIL

has the same result as

: HAIL ( -- )   IMPOSTOR'S ." ahoi ahoi" ;
0 6000 ' HAIL CAUSE
  
but is executed differently.
Activity Monitor shows for CAUSE, a workthread is spawned at the time
of action, not before and sleeping till moment of action as in the
TASK version.

Perhaps for heavy load and strict realtime requirements, a sleeping
task, woken at the right time and executing a defered word and go to
sleep again, is a better approach.
-------------------------------------------------------------------- ))

/forth
decimal

function: dispatch_after_f ( when:lo when:hi queue *context work -- ret )
function: dispatch_walltime ( *timespec delta:lo delta:hi -- dispatch_time_t:lo dispatch_time_t:hi )

: s>nano  ( n -- d )   1000000000 m* ;
: ms>nano ( n -- d )   1000000 m* ;
: us>nano ( n -- d )   1000 m* ;

: FROM-NOW ( interval:lo interval:hi -- dispatch-time:lo dispatch-time:hi )
   0 -rot  dispatch_walltime ;

defer (CAUSE)
forthcall: *CAUSE ( n -- )   (cause) ;

: CAUSE ( param when:ms xt -- )
   is (cause) swap >r ms>nano from-now defaultqueue r> *cause dispatch_after_f drop ;

\\ ( eof )

\ example
variable var
: bump ( -- )  recursive  var dup @ 10 < if  1 swap +! 0 6000 ['] bump cause  then ;
0 6000 ' bump cause
var ? 8  ok
var ? 10  ok                                     
var ? 10  ok                                     
var ? 10  ok  

