(* ====================================================================
CoreFoundation

copyright 2001 Roelf Toxopeus

VFX version.
Handy Corefoundation utillities from Ward and Roelf.
More stuff gets in and out...
Last: 4 February 2019 at 14:03:27 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
CoreFoundation.framework -- liason framework, C API.

>CFSTRING -- create CFStringRef from given 0string or cstring.
Use CFRelease when done with string.
CF" -- convert following string to a CFString, works like S"
Use CFRelease when done with string.
CFSTRING> -- convert CFString to 0string in pocket.
Take care of it immediate. Aborts if not a CFString !

>CFNUMBER -- create CFNumber from given integer.
Use CFRelease when done with number.
CFNUMBER> -- convert CFNumber to integer

CFNUMBER@ -- using key-id, fetch number from CFDictionary
CFSTRING@) -- using key-id, fetch zstring from CFDictionary

>CFURL -- create a CFURL from given zstring. Allows Posix pathstyle
with spaces it seems. Use CFRelease when done with url.
CFURL> -- return zstring representation from given CFURLRef. zstring is
in PATHPAD, so consume before it gets overwritten.

NoTE: return from CFAllocatorGetDefault changes with machines and OS,
-1598992704 on 10.6 and -1407864448 on 10.7
CFAllocatorGetDefault constant myCFAllocatorSystemDefault
won't survive running turnkeyed 10.6 coco-sf on 10.7 
better to use 0 were possible
or  CFAllocatorGetDefault
or   (the global) kCFAllocatorSystemDefault @

Tip: in general when using CoreFoundation API, anything that is
returned by an API with "Get" in the name does not need to be released,
and anything returned by an API with "Create" or "Copy" in the name
DOES need to be released.
CFGetRetainCount -- returns reference count from CFTypeRef, should go
to 0 if we created the CFTypeRef, to avoid memory leaks.
CFRelease to decrement and CFRetain to increment the reference count.
Reference count = 0 means memory will be deallocated by OS.
Quitting the app (Forth) has same effect.
-------------------------------------------------------------------- ))


/forth
decimal

framework CoreFoundation.framework

\ --------------------------------------------------------------------
\ used a lot

function: CFAllocatorGetDefault ( -- CFAllocatorRef )
function: CFRelease ( CFTypeRef -- ret )
function: CFRetain ( CFTypeRef -- CFTypeRef )
function: CFGetRetainCount ( CFTypeRef -- CFIndex )

0 constant kNilOptions

global: kCFAllocatorSystemDefault

\ --------------------------------------------------------------------
\ info

function: CFShow ( CFTypeRef -- ret )   \ show information about CFTypeRef in CONSOLE
function: CFGetTypeID ( CFTypeRef -- CFTypeID )

\ --------------------------------------------------------------------
\ CFArray

function: CFArrayCreate ( CFAllocatorRef **values numValues CFArrayCallBacks -- CFArrayRef )
function: CFArrayCreateMutable ( CFAllocatorRef capacity CFArrayCallBacks -- CFMutableArrayRef )

\ --------------------------------------------------------------------
\ CFString stuff

function: CFStringCreateWithCString ( CFAllocatorRef ^0Str CFStringEncoding -- CFStringRef )
function: CFStringCreateWithBytes ( CFAllocatorRef addr cnt CFStringEncoding externalFlag -- CFStringRef )
function: CFStringCreateWithPascalString ( CFAllocatorRef pStr CFStringEncoding -- CFStringRef )

function: CFStringGetIntValue    ( CFStringRef -- u )
function: CFStringGetDoubleValue ( CFStringRef -- ud )

function: CFDataGetLength ( CFDataRef -- u )
function: CFDataGetBytes ( CFDataRef startCFIndex endCFIndex ^buffer -- ret )
function: CFStringGetCString ( CFStringRef *buffer bufferSize CFStringEncoding -- Boolean )

0 constant kCFStringEncodingMacRoman

: >CFSTRING ( zaddr -- CFStringRef )   0 SWAP kCFStringEncodingMacRoman CFStringCreateWithCString ;

(*
: CF" ( <string>"  -- CFStringRef )
   [compile] 0"
   state @
   IF    postpone >CFString  ELSE  >CFString  THEN ; immediate
*)

: CF"            \ Comp: "ccc<quote>" -- ; Run: -- CFReference
\ *G A Version of *\fo{C"} which returns a CFReference of the string
 [char] " parse >SyspadZ >CFString
;
comp: ( xt -- )
 drop  ['] (z") compile,
 [char] " parse z$,
 discard-sinline
 postpone >CFString
;

: CFSTRING> ( CFStringRef -- zaddr )     pocket tuck 256 kCFStringEncodingMacRoman CFStringGetCString 0= abort" Not a CFString !" ;
   
\ --------------------------------------------------------------------
\ CFNumber utillities

function: CFNumberCreate ( CFAllocatorRef CFNumberType *valuePtr -- CFNumberRef )
function: CFNumberGetValue ( object kCFNumberLongType &number -- ret )

 9 constant kCFNumberIntType
10 constant kCFNumberLongType

: >CFNUMBER ( u -- CFNumberRef )  >r 0 kCFNumberIntType rp@ CFNumberCreate r> drop ;

: CFNUMBER> ( CFNumberRef -- u ) kCFNumberLongType 0 >r rp@ CFNumberGetValue drop r> ;

\ --------------------------------------------------------------------
\ CFDictionary things

function: CFDictionaryCreateMutable ( allocator capacity *keyCallBacks *valueCallBacks -- CFMutableDictionaryRef )
function: CFDictionarySetValue ( dictionary cfstring:key cfnumber:value -- ret )
function: CFDictionaryGetValue ( dictionary cfstring:key -- CFTypeRef:object )
function: CFDictionaryContainsKey ( dictionary string -- flag )
function: CFDictionaryGetCount ( dictionary -- CFIndex )
function: CFDictionaryGetKeysAndValues ( dictionary *keys *values -- ret )

global: kCFCopyStringDictionaryKeyCallBacks
global: kCFTypeDictionaryKeyCallBacks
global: kCFTypeDictionaryValueCallBacks

: CFNUMBER@ ( key dictionary -- n )
   swap >cfstring CFDictionaryGetValue dup 0= if drop $DEADBEEF exit then
   cfnumber> ;

: CFSTRING@ ( key dictionary -- zaddr )
   swap >cfstring CFDictionaryGetValue dup 0= if drop 0" " exit then
   cfstring> ;

\ --------------------------------------------------------------------
\ CFRunLoop

function: CFMachPortCreateWithPort ( CFAllocatorRef portnum CFMachPortCallBack CFMachPortContext Boolean -- CFMachPort )
function: CFRunLoopGetCurrent ( -- CFRunLoopRef )
function: CFRunLoopGetMain ( -- CFRunLoopRef )
function: CFMachPortCreateRunLoopSource ( CFAllocatorRef CFMachPortRef CFIndex -- CFRunLoopSourceRef )
function: CFRunLoopAddSource ( CFRunLoopRef CFRunLoopSourceRef CFStringRef:mode -- ret )
function: CFRunLoopContainsSource ( CFRunLoopRef CFRunLoopSourceRef CFStringRef:mode -- Boolean )

global:   kCFRunLoopDefaultMode

\ --------------------------------------------------------------------
\ CFURL stuff

(*
function: CFURLCreateWithString ( CFAllocatorRef CFStringRef:URLString CFURLRef:baseURL -- CFURLRef )

: >CFURL ( zaddr -- CFURLRef )
      ( kCFAllocatorSystemDefault @ ) 0 swap >CFSTRING 0 CFURLCreateWithString ;
*)

function: CFURLCreateWithFileSystemPath ( allocator CFString:filePath pathStyle isDirectory   -- CFURLRef )

: >CFURL ( za -- CFURLRef )
   >cfstring dup >r
   ( kCFAllocatorSystemDefault @ ) 0
   swap
   ( kCFURLPOSIXPathStyle ) 0
   ( = not a directory ) 0 CFURLCreateWithFileSystemPath
   r> CFRelease drop ;

function: CFURLGetFileSystemRepresentation ( CFURLRef Boolean:resolveAgainstBase *buffer maxBufLen -- bool )

: CFURL> ( CFURLRef -- za )
   1 pathpad max-path CFURLGetFileSystemRepresentation 0= abort" No Path for CFURL !"
   pathpad ;

cr .( Corefoundation words loaded)

\\ ( eof )