(* ====================================================================
OSX Frameworks

Copyright 2010 Roelf Toxopeus

VFX version.
MacForth Framework (like) words, the OSX way for access to foreign
function libraries.
Redefines VFX stock version.
Last: 20 September 2019 at 07:57:08 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
FRAMEWORK -- create word named after an existing framework. When executed
the framework will be set as the top or first searched  library of all
imported libraries. The difference with the stock VFX version is, when
the new framework already exists, it will be executed and not redefined.

FRAMEWORK Carbon.framework -> execute Carbon.framework when existing <- new
                              else add to dictionary and execute it  <- old
Carbon.framework           -> execute it                             <- old

So the following is not necessary anymore:

LACKING System.framework  FRAMEWORK System.framework
System.framework

Just executing  FRAMEWORK System.framework  will cover both the undefined 
and defined cases.

Note: FRAMEWORK (actualy dlopen) will search for frameworks in all the
default Frameworks directories:
/Library/Frameworks
/System/Library/Frameworks
~/Library/Frameworks

Note:
FUNCTION: will always work, raises no exception when not imported!

function: pipo ( n1 n2 -- n3 ) ok
1 2 pipo 
 CS=001B DS=0023 ES=0023 SS=0023 FS=0000 GS=000F
 EAX=0000:0001 EBX=0000:0002 ECX=BFFE:CA71 EDX=0001:5CE4
 ESI=BFFF:DBC0 EDI=0000:0000 EBP=BFFF:CB74 ESP=BFFE:CB4C
 EIP=0000:0000 EFLAGS=0001:0286
--- RS top ---
 $0008:6A88 PIPO 
 $0000:0001
 $0000:0002
 $0000:0000
 $0000:0000
 $BFFF:DBC0
 $0001:7950 PLACE 
 $0000:0000
Signal number SIGSEGV
at address 0000:0000, probably in * outside dictionary *  

Press E to exit, R to restart, other to continue: 
Err# -57005 ERR: OSX Exception
 -> 1 2 pipo
           ^
crashed during execution, use find-libfunction to check availability.
z" pipo" find-libfunction . 0 ok

Example FRAMEWORK: usage in vanilla vfxosx:
.libs 
libmpeparser.0.dylib            00A0:11D0
/usr/lib/libSystem.B.dylib      0094:1ED8 ok

function: function: NSRunAlertPanel ( ns" ns" ns" n n -- ret )  ok
z" NSRunAlertPanel" find-libfunction . 0  ok

framework Cocoa.framework  ok
.libs 
Cocoa.framework/Cocoa           00C0:0180
libmpeparser.0.dylib            00A0:11D0
/usr/lib/libSystem.B.dylib      0094:1ED8 ok

z" NSRunAlertPanel" find-libfunction . -1842191646  ok

.libs 
Cocoa.framework/Cocoa           00C0:0180
libmpeparser.0.dylib            00A0:11D0
/usr/lib/libSystem.B.dylib      0094:1ED8 ok

framework Cocoa.framework  ok

.libs 
Cocoa.framework/Cocoa           00C0:0180
libmpeparser.0.dylib            00A0:11D0
/usr/lib/libSystem.B.dylib      0094:1ED8 ok
------------------------------------------------------------------- ))

only forth also definitions
decimal

: (framework) ( spaces<name> -- )   framework ;

: FRAMEWORK ( spaces<name> -- )
   >in @ >r  bl word find
   if    r> drop execute
   else  r> >in ! drop (framework)
   then ;

\ optional !
\ these redefinitions will catch non availability during defining
: check-import ( spaces<name> -- )
   >in @  bl word count pad zplace pad find-libfunction
   0= abort"  not in current libraries !"
   >in ! ;

: FUNCTION: ( spaces<name> -- )   check-import function: ;

: GLOBAL: ( spaces<name> -- )   check-import global: ;

: check-import2 ( spaces<names> -- )
   >in @  bl word drop bl word count pad zplace pad find-libfunction
   0= abort"  not in current libraries"
   >in ! ;

: ASCALL: ( spaces<names> -- )   check-import2 ascall: ;

cr .( New frameworkstuff loaded)

\\ ( eof)