(* ====================================================================
SIGINT handler

copyright 2020 Roelf Toxopeus

VFX version.
Installing a handler to catch raised SIGINT exceptions like the CMD.
key chord. Default action in macOS is to terminate process. The new
handler allows to 'savely' interrupt rogue Forth words a la ABORT.
Last: 15 September 2020 at 10:46:51 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
*doAbort -- the handler called by the exception.
/SIGINT --  installing the handler for the SIGINT (2) exception.

REACTIVATE -- a variation on ACTIVATE. If task is an existing thread,
terminate it and relaunch it a la ACTIVATE.
FORTH.AGAIN -- REACTIVATE task install SIGINT handler and run QUIT.
Dictionary is not affected.

doAbort -- reactivate IMPOSTOR, as if abort being executed.
/MYSIG -- set doAbort as action for the SIGINT handler and install.

The 1000 ms is a rough estimate for the Mach scheduler to react.
This could be too long for a GUI callback version. Experiment.

Just send a 'SIGINT kill' to your application to trigger the handler.
But not from IMPOSTOR self!
FUNCTION: kill ( pid signal -- ret )
: PANIC ( -- )   getpid 2 kill  DROP ;

Example: vfxmenu5.fth
1. include vfxmenu5.fth -> vfxmenu
2. include mysigint.fth -> /mysig

Note: IMPOSTOR could have activated, spawned or forked other tasks
and threads. If necessary, clean up if possible after SIGINT and
relaunch coco-vfx. This is not a replacement for a controlled task
termination.
-------------------------------------------------------------------- ))

/forth
decimal

variable 'abort  :noname 2drop drop ; 'abort !
forthcall: *doAbort ( signum *siginfo *ucontext -- )   ( impostor's) ." User interrupt" 'abort @execute ;

: /SIGINT ( -- )    *doabort sigint setsignal ;

: REACTIVATE ( taskid -- )
   dup tcb.up ( cell+) @ if     \ tcb.up gets cleared when tasks terminates
      dup done 1000 ms
      r> swap initiate
   else drop
   then ;

: FORTH.AGAIN ( task -- )   reactivate  /sigint  o_abort 0 ;

: doAbort ( signum *siginfo *ucontext -- )   2drop drop impostor forth.again ;

: /MYSIG ( -- )   ['] doAbort 'abort ! /sigint ;

cr .( /mysig  -- install SIGINT handler)

\\

\ test clean termination and reinitializing
\ User interrupt  means cmd. is hit in Terminal
/mysig  ok
.tasks 
MAIN   IMPOSTOR    ok
impostor cell+ ? -1339834368  ok
User interrupt
.tasks 
MAIN   IMPOSTOR    ok
impostor cell+ ? -1339301888  ok
User interrupt
.tasks 
MAIN   IMPOSTOR    ok
impostor cell+ ? -1338769408  ok

\ rogue word
: tt begin random cr . 500 ms again ;  ok
tt 
-1175601567 
-1156211442 
-882430089 
1082671596 
-1574068611 
1823025658 
838008051 
-726087496 User interrupt
 ok
 ok

\ include vfxmenu5.fth (adapted)
\ User interrupt  means abort menu item was hit

simple menu for coco-vfx loaded ...
vfxmenu   -- to add main menu to coco-vfx ok
vfxmenu  ok
.tasks 
MAIN   IMPOSTOR    ok
impostor cell+ ? -1339256832  ok
tt 
971430359 
403403980 
-1141714979 
1268228900 User interrupt
impostor cell+ ? -1339789312  ok
.tasks 
MAIN   IMPOSTOR    ok
tt 
-1690696363 
-700625422 
2474934 User interrupt
.tasks 
MAIN   IMPOSTOR    ok
impostor cell+ ? -1338191872  ok

\ include vfxmenu5.fth
\ User interrupt  means abort menu item was hit

simple menu for coco-vfx loaded ...
vfxmenu   -- to add main menu to coco-vfx ok
vfxmenu  ok
.tasks 
MAIN   IMPOSTOR    ok
impostor cell+ ? -1339256832  ok
tt 
971430359 
403403980 
-1141714979 
1268228900 User interrupt
impostor cell+ ? -1339789312  ok
.tasks 
MAIN   IMPOSTOR    ok
tt 
-1690696363 
-700625422 
2474934 User interrupt
.tasks 
MAIN   IMPOSTOR    ok
impostor cell+ ? -1338191872  ok