(* ====================================================================
Task control

copyright 2001 Roelf Toxopeus

VFX version.
Controlling threads and tasks unconditionally, no cancel-point considered.
Last: 14 October 2014 14:22:02 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
ACTIVATE -- run the given task with the code following ACTIVATE till a
closing ;

The thread controllers are for brute force control and when dealing with
self created threads without tasks. Many Mach and POSIX calls are now part
of VFX: Lib/Osx32/MultiOsx32.fth
SUSPEND -- suspend given thread
RESUME -- wake given thread
TERMINATE -- terminates given thread
These are wrapped in HALT, RESTART and TERMINATE, which also sets VFX
specific variables used by PAUSE and others.

SLEEP -- set given task to sleep
STOP -- goto sleep yourself, another task or interrupt will WAKE you...
WAKE -- wakeup giventask
DONE -- end/stop given task now

GET and RELEASE, Ward McFarland's versions, are now part of VFX
-------------------------------------------------------------------- ))

/forth
decimal

\ SF allows extra dictionary size for task
\ : TASK ( n -- ) drop task ;

: ACTIVATE ( taskid -- )   r> swap initiate ;

\ --- task controllers

: SUSPEND ( pthread -- )   pthread_mach_thread_np thread_suspend drop ;

: RESUME ( pthread -- )   pthread_mach_thread_np thread_resume drop ;

\ TERMINATE is in VFX

\ --- synonyms

: SLEEP ( task -- )  halt ;

\ STOP is in VFX

: WAKE ( task -- )   restart ;

: DONE ( task -- )   terminate ;

cr .( task-control loaded)

\\ ( eof )
