(* ====================================================================
fp precission conversion

copyright 2002 Roelf Toxopeus

VFX version.
Comes from MacForth version, which uses PPC instructions.
Converting between single precision (SF) and native format precision (F) fp.
Many OS calls need and/or return single precision fp. These 32bit SF's
are passed via the data stack.
Also contains the integer <=> fp formats conversions.
Last: 14 April 2018 at 16:34:18 CEST  -rt
==================================================================== *)

/forth
decimal

(( --------------------------------------------------------------------
conversion between fp formats

Could do (perhaps) with faster assembler versions from the following:

SF>F converts 32b SFLOAT on data stack to native FLOAT on fp stack.
F>SF does the reverse.
NoTE: F is not necessarely a DF. In SwiftForth it is, but not in iForth.
-------------------------------------------------------------------- ))

: SF>F ( sfloat -- ) ( F: -- r ) >r rp@ sf@ r> drop ;

: F>SF ( -- sfloat ) ( F: r -- ) 0 >r rp@ sf! r> ;

\ Fast versions of S>F D>F F>S F>D exist in VFX

(( --------------------------------------------------------------------
integer <=> (s)fp conversions

Some words are here for 'old times' sake. They should be replaced in
due time for 'standard' versions:
INT -- convert native fp format on fp stack to integer on stack.
FLOAT -- convert integer on data stack to native fp format on fp stack.
SFLOAT -- convert integer on data stack to 32b fp format on data stack.
SINT -- convert 32b fp format on data stack to integer on stack.
NoTE: these are re-definitions of the 'standard' versions.

S>SF -- convert integer on data stack to 32b fp format on data stack.
SF>S -- convert 32b fp format on data stack to integer on stack.
-------------------------------------------------------------------- ))

: S>SF ( n -- sfloat )   s>f f>sf ;

: SF>S ( sfloat -- n )  sf>f f>s ;

(( --------------------------------------------------------------------
passing sfloat and dfloats to calls

Many could be just aliasses for existing words, but might involve special
handling wrt call ABI on other architectures: 32b SF and VFX - 64b iForth
FPUSHS 2FPUSHS 4FPUSHS pass fp numbers as sf's on data stack.
The names are Marcel Hendrix's from iForth, the original MacForth names
were less attractive.
The fast assembler 4FPUSHS learned from RickVanNorman in the Faq-o-Matic
OpenGL Spin Cube. 2FPUSHS and FPUSHS are derived from that version.

FPUSHD 2FPUSHD 4FPUSHD and DF>F are aliasses for George Kozlowski
fp stack passing words found in fp-passing.f
Each DF consumes 2 cells on the data stack, take care when defining
functions and methods!

Could do with faster assembler versions.
-------------------------------------------------------------------- ))

\ --------------------------------------------------------------------
\ fp to sfloat on data stack conversion

: FPUSHS ( -- sfloat ) ( F: r -- )   f>sf  ;

: 2FPUSHS ( -- sf1 sf2 ) ( F: r1 r2 -- )
   fswap fpushs fpushs ;
   
: 4FPUSHS ( -- sf1 sf2 sf3 sf4 ) ( F: r1 r2 r3 r4 -- )
   f2swap 2fpushs 2fpushs ;

\ --------------------------------------------------------------------
\ fp to dfloat on data stack conversion

(( Kozlowski words missing
: FPUSHD  ( -- d ; F: r -- )   FSTACK>STACK ;
: 2FPUSHD ( -- d d ; F: r r -- )   2FSTACK>STACK ;
: 4FPUSHD ( -- d d d d ; F: r r r r -- )   4FSTACK>STACK ;
: DF>F ( d -- ; F: -- r )   STACK>FSTACK ;
))

\ for now:
: FPUSHD  ( -- d ; F: r -- )   0 dup 2>r rp@ df! r> r> ;
: DF>F ( d -- ; F: -- r )   >r >r rp@ df@ 2r> 2drop ;
synonym F>DF FPUSHD

\ --- not necesarily a better name, but in use a lot in the demo's

: D>R ( F: x1 -- x2 )  deg>rad ;

cr .( fp precisison loaded)

\\ ( eof )

