(* ====================================================================
waypoints

copyright 2010 Roelf Toxopeus

VFX version.
Usefull words to set current working directory, works like setting
a context vocabulary. Also working directory utillities.
Addapted paths to preferred MPE setup ...
Last: 17 December 2019 at 09:59:47 CET  rt
==================================================================== *)

/forth
decimal

(* --------------------------------------------------------------------
MAX-PATH -- returned value inherited from PATH_MAX found in
usr/include/sys/syslimits.h
However, most words defined here and elsewhere assume path is 255 chars max.
PATHPAD -- buffer for creating and receiving path strings.
OLDPWD -- buffer for previous working directory.

!OLDPWD -- save current working directory before changing it.

CHWD -- changes working directory, takes care of saving cwd before changing it.
Like CD but takes path from the stack.

CWD -- get current working directory in PATHPAD (1024 max), zero terminated.
WHAMI -- returns current working directory as string pair.
NoTE: both using the PADPATH instead of PAD. Could be a factor of PWD,
which prints the string.

SAVEPATH -- saves current working directory in save-buffer
RESTOREPATH -- changes working directory from save-buffer
-------------------------------------------------------------------- *)

1024 constant MAX-PATH
max-path buffer: PATHPAD

: CHWD ( a n -- )   \ name is a joke of course, chew on it ;-)
   changedir ;

: CWD ( -- zstring )   workingdir pathpad dup >r zplace s" /" r@ zappend r>  ;

: WHAMI ( -- a n )   workingdir ;

max-path buffer: 'cwd

: SAVEPATH ( -- )   whami 'cwd zplace ;

: RESTOREPATH ( -- )  'cwd zcount chwd ;

\ sf compatible like, but not realy:
: pushpath  SAVEPATH ;
: poppath   RESTOREPATH ;

(( --------------------------------------------------------------------
WAYPOINT -- used to create named 'waypoints' which when executed will
change the working directory to the path given at initialisation.
Originally a navigational aid used at sea and in the air.
Example usage:  WHAMI  WAYPOINT  MyPivot

MAC is the pivotal directory for the Cocoa package. It uses MACPATH,
set during compilation of coco-vfx. It is assumed you keep the VFX and
mac folders in the same relation as during compilation of coco-vfx.
Otherwise MACPATH is invalid.
VFX, FROOT, HOME and DESKTOP allow for quick and easy access to these
default directories.
Note: for now all paths are expanded path macros. Take care, they could
point to something else for whatever reason. When things go wrong,
you can derive the paths from MACPATH, which is hard coded in the
current running coco-vfx.
      
From MacForth: MYFOLDER -- set directory from file being included as
the current working directory. Further including can commence from this
directory as parent. Only from within INCLUDED/LOADED file!!!
INCLUDING ought to return a zero string when nothing is INCLUDED? It
appears it returns a 1 length zero string, so check for anything bigger
than 1.
Avoid getting lost, use with care.
-------------------------------------------------------------------- ))

: WAYPOINT ( a n <name> -- )
   create string,
   does> count chwd ;

\ macpath count   waypoint MAC
ms" %cocoalib%" waypoint MAC
ms" %vfxpath%"  waypoint VFX
ms" %vfxpath%" pad place
pad stripfilename
pad count       waypoint FROOT
ms" %$home%"    waypoint HOME
ms" %$home%" pad place
s" /Desktop" pad append
pad count       waypoint DESKTOP

also system
: MYFOLDER ( -- )
   'SourceFile @ dup 1 < abort" Only use MYFOLDER while loading a file"
   SFSTRUCT.szPathName zcount pathpad place pathpad stripfilename
   pathpad count changedir ;
previous

cr .( Waypoints loaded)

\\ ( eof )
