\ xterm.fth - OSX character I/O for TTY devices

((
Copyright (c) 2007, 2008, 2009, 2011, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20120329 SFP006 Fixed paste and OUT issues.
20110119 SFP005 OSX conversion.
20090917 SFP004 Added terminal preparation checks.
20080121 SFP003 Added save of history to INI file.
20080109 SFP002 Reverted to key 127 as delete before.
20080103 SFP001 Added history.
))

only forth definitions  decimal

\ ========
\ *> genio
\ *N XTERM Device
\ ========
\ *P The XTERM Generic IO Device controls an xterm or equivalent device.
\ ** Facilities are provided for cursor positioning, setting
\ ** the foreground and background colours, line editing and
\ ** line history. Cursor positioning uses ANSI escape sequences.
\ ** Any terminal emulator which supports these sequences, e.g.
\ ** in ANSI or VT100 mode, should work with this code. A good
\ ** introduction to ANSI escape sequences may be found at
\ ** *\i{http://en.wikipedia.org/wiki/ANSI_escape_code}.

\ *P In order to create a device use the *\fo{XTERM:} definition
\ ** given later. When opening a device the parameters to
\ ** *\fo{OPEN-GEN} are unused. For compatibility with future
\ ** versions please set them to -1, e.g.
\ *C   -1 -1 -1 <sid> open-gio

\ *P The IOCTL function has the following action
\ *P You can set the text foreground and background colours:
\ *C   <fcolour> <bcolour> #10 <sid> IOCTL-GIO drop
\ *P where *\i{colour} is a colour in the XTERM format. If
\ ** a colour is set to -1 the existing colour is left unchanged.
\ ** For XTERMs and VT100/220 compatible terminals, the following colours
\ ** are standard.
\ *E 0 constant Black
\ ** 1 constant Red
\ ** 2 constant Green
\ ** 3 constant Yellow
\ ** 4 constant Blue
\ ** 5 constant Magenta
\ ** 6 constant Cyan
\ ** 7 constant White
\ ** : +bright  \ color -- color'
\ ** \ Convert a colour into its bright version.
\ **   #60 +
\ ** ;

\ *P Similarly, terminal positioning control uses ANSI (VT100 and VT220)
\ ** sequences. If you are connecting using Telnet or other remote
\ ** access techniques (or even a real terminal), set it to ANSI,
\ ** VT100 or VT220 compatibility mode.

\ *P Line editing is performed using the cursor keys, BS (<- or ^H) to
\ ** delete before the cursor, and the DELETE keys to delete after
\ ** the cursor. You can also use ^W and ^R for cursor movement.
\ ** You can recall lines using the up (previous) and down (next)
\ ** cursor keys. Lines can be edited after recall. You can also
\ ** use ^E and ^D instead of the up and down keys. Note that
\ ** Linux implementations are not consistent in the codes returned
\ ** by keys such as BS and DELETE.

\ *P Unlike other devices, an *\fo{XTERM} uses three handles that
\ ** correspond to *\b{stdin}, *\b{stdout} and *\b{stderr}. By
\ ** default these are handles 0, 1 and 2 respectively. If you
\ ** wish to use different handles, you are responsible for their
\ ** management. Use these handles by setting them into the
\ ** *\fo{/xterm-sid} structure below. By default, the open
\ ** operation uses the preset handles, which are not closed
\ ** by the close operation.

\ *P The source code can be found in the file
\ ** *\i{Lib/Osx32/Genio/xterm.fth}, which is compiled during the second
\ ** stage build. If you change this file, perform a second stage
\ ** build when you wish to commit to using the changed file.


struct /xterm-sid	\ -- len
\ *G Defines the SID of an xterm console device.
\ *[
  gen-sid +			\ reuse field names of GEN-SID
  int xs.hIn			\ input handle
  int xs.hOut			\ output handle
  int xs.hErr			\ error handle
  int xs.flags			\ control flags
				\ bit 0 - QUIT control
				\ bit 1 - rfu
				\ bit 2 - 1=maintain history, 0=none
				\ bit 3 - 1=history in system ini file
  int xs.hiBuff			\ address of 64k history buffer
  int xs.hiIndex		\ current history line#
  int xs.hiLowIndex		\ lowest history line#
  int xs.hiSection		\ pointer to INI file section name
end-struct
\ *]

also system definitions

\ *****************
\ History functions
\ *****************
\ The history buffer consists of a number of fixed size line buffers
\ containing counted strings. The three history fields in the /xterm
\ structure above contain:
\   xs.hiBuff		\ pointer to an ALLOCATEd buffer
\   xs.hiIndex		\ the current history line#
\   xs.hiLowIndex	\ the lowest valid history line#
\ where line numbers are in the range 0..#HISTORY-1. When the buffer is empty,
\ both line numbers point to the last line. When a line is added, the buffer
\ is scrolled by one line, and the lowest index is decremented (until it
\ becomes 0).

$100 constant #history			\ -- n ; number of command lines for history
$100 constant MAXSTRING			\ -- n ; line buffer size
#history MAXSTRING * constant /history	\ -- n ; size of history buffer
/history MAXSTRING - constant /hscroll	\ -- n ; size of history buffer less one line
/history MAXSTRING - constant /hcurr	\ -- n ; offset to current line

((
: >hLast	\ addr -- addr' ; step to last history line
  /history MAXSTRING - +
;
))

: hLine		\ sid line# -- addr
  MAXSTRING *  swap xs.hiBuff @ +
; doNotSin

[-short-branches
: .History	\ sid -- ; show command line history buffers
  dup xs.hiBuff @ if
    #history 0 do
      cr  i 3 u.r ." : "
      dup i hLine count dup
      if  type  else  2drop ." <null>"  then
    loop
    cr
  then
  drop
;
short-branches]

: restart-history	\ sid -- ; reinitialise current index (ACCEPT started)
  #history 1- swap xs.hiIndex !
;

: clear-history         \ sid -- ; erase all strings in history buffer
  #history 1- over xs.hiIndex !
  #history 1- over xs.hiLowIndex !
  xs.hiBuff @ ?dup
  if  /history erase  then
;

: add-history-line      \ c-addr u sid -- ; Add line
  over if				\ only add non-blank lines
    dup xs.hiBuff @ if			\ to a buffer that exists
      dup xs.hiLowIndex @
      1- 0 max				\ update range
      over xs.hiLowIndex !
      dup xs.hiBuff @		\ scroll buffer by one line
        dup MAXSTRING + swap /hscroll move
      xs.hiBuff @ /hcurr + place	\ place text in last line
      exit
    then
  then
  drop 2drop
;

: curr-history	\ sid -- c-addr u
  dup xs.hiBuff @ if
    dup xs.hiIndex @ hLine
  else
    drop cNull
  then
  count
;

: step-history  \ sid step --
  over xs.hiIndex @ +	\ update current line# and range check
  over xs.hiLowIndex @ max  #history 1- min
  swap xs.hiIndex !
;

: at-last-history?	\ sid -- flag ; true if at last history line
  xs.hiIndex @ #history 1- =
;

: open-history	\ sid --
  dup xs.flags @ 4 and 0=
  if  drop  exit  endif
  /history allocate if			\ -- sid addr
    drop
    dup xs.hiBuff off			\ mark buffer unused
    4 swap xs.flags bic!		\ clear history demand
    exit
  then
  over xs.hiBuff !			\ mark buffer
  clear-history
;

: close-history	\ sid --
  dup xs.flags @ 4 and if
    dup xs.hiBuff @ ?dup
    if  free drop  endif
  endif
  drop
;

: write-history	{ sid -- }
  s" CommandHistory" Ini.Section
  base @ >r  hex
  S" LVALID" sid xs.hiLowIndex @ Ini.WriteInt
  #history 0 do
    i s>d <# #S #>  sid i hLine count Ini.WriteStr
  loop
  r> base !
;

: writeXtermCfg	\ sid --
  dup xs.flags @ $000C and $000C = if	\ if INI file load needed
    IniFile$ count Ini.Open 0= if
      dup write-history
    endif
    Ini.Close
  endif
  drop
;

: read-history	{ sid -- }
  s" CommandHistory" Ini.Section
  base @ >r  hex
  S" LVALID" #history Ini.ReadInt sid xs.hiLowIndex !
  #history 1- sid xs.hiIndex !
  #history 0 do
    i s>d <# #S #>  sid i hLine MAXSTRING  S" " Ini.ReadStr
  loop
  r> base !
;

: readXtermCfg	\ sid --
  dup xs.flags @ $000C and $000C = if	\ if INI file load needed
    IniFile$ count Ini.Open 0= if
      dup read-history
    endif
    Ini.Close
  endif
  drop
;


\ *********************************
\ System dependent XTERM primitives
\ *********************************

#10 constant EOLchar_con	\ character marking end of line
#13 constant IGNchar_con	\ character to ignore

\ -----
\ Linux
\ -----

[-sin

: read-xterm	\ addr len sid -- ior
  ?PrepTerm				\ SFP004
  xs.hIn @ -rot read -1 =
  if errno @ else 0 endif
;

: write-xterm	\ addr len sid -- ior
  ?PrepTerm				\ SFP004
  xs.hOut @ -rot write -1 =
  if errno @ else 0 endif
;

: key?-xterm	\ sid -- flag
  { | nread[ cell ] -- }
  ?PrepTerm  nread[ off			\ SFP004
  xs.hIn @ FIONREAD nread[ ioctl -1 =
  if  0  else  nread[ @ 0<>  endif
;

: key-xterm	\ sid -- key
  { | temp[ cell ] -- }
  begin
    dup key?-xterm 0=
   while
    emptyidle 2 ms
  repeat
  temp[ 1 rot read-xterm drop
  temp[ c@
;

: emit-xterm	\ char sid --
  { | temp[ cell ] -- }
  ?PrepTerm				\ SFP004
  swap temp[ c!  1 out +!
  xs.hOut @ temp[ 1 write drop
;

: emit?-xterm	\ sid -- flag
\ Return true if we can send more characters
  drop  true  ;

: type-xterm	\ caddr len sid --
  ?PrepTerm				\ SFP004
  over 0> if
    over out +!
    xs.hOut @ -rot write drop
  else
    2drop drop
  then
;

: cr-xterm	\ sid --
  #10 swap emit-xterm
  out off  op-line# incr
;

: lf-xterm	\ sid --
  cr-xterm
;

: open-xterm	\ addr len attribs sid -- handle/sid ior
\  ?PrepTerm				\ SFP004
  nip nip nip
  dup open-history			\ NOOP if not required
  dup xs.hOut @ over gen-handle !	\ mark as open
  $40 c/line !
  0
;

: close-xterm	\ sid -- ior
  dup close-history
  -1 swap gen-handle !			\ mark closed
  0
;

: ekey?-xterm	\ sid -- flag
  key?-xterm
;

: ekey-xterm	\ sid -- echar
  key-xterm
;

: ff-xterm	\ sid -- ; page/cls on display devices ; SFP006
  s\" \x1b[2J\x1b[H" rot type-xterm
  out off
;

: bs-xterm	\ sid -- ; destructive on display devices
  out @ if
    out @ >r
    s\" \x08 \x08" rot type-xterm
    r> 1- out !
  else
    drop
  endif
;

sin]

: u+$		\ u caddr --
\ Convert u to decimal digits and add them to the end of the
\ given counted string.
  base @ >r  decimal
  swap (u.) rot append
  r> base !
;

: setCSI	\ caddr --
\ Place the sequence "<ESC>[" as a counted string at caddr.
  s\" \x1b[" rot place
;

: xleft		\ n sid --
\ Move cursor left n positions.
  { | temp[ 16 ] -- }
  temp[ setCSI  swap temp[ u+$  s" D" temp[ append
  temp[ count rot type-xterm
;

: xright	\ n sid --
\ Move cursor right n positions.
  { | temp[ 16 ] -- }
  temp[ setCSI  swap temp[ u+$  s" C" temp[ append
  temp[ count rot type-xterm
;

: xdelAfter	\ sid --
\ Delete character after cursor
  s\" \x1b[1P" rot type-xterm
;

: xdelBefore	\ sid --
\ Delete character before cursor
  1 over xleft  xdelAfter
;

: xSetPos	\ x y sid --
\ Go to cursor position x,y. Position 0,0 is top left.
  { | temp[ 16 ] -- }
  -rot  temp[ setCSI		\ CSI
  0 max 1+ temp[ u+$		\ y
  s" ;" temp[ append
  0 max 1+ temp[ u+$		\ x
  s" H" temp[ append
  temp[ count rot type-xterm
;

: getpn		\ sid -- u
\ Read characters up to a non-digit, which is discarded. The
\ value of the digits is returned.
  0 begin				\ -- sid acc
    over key-xterm #10 digit		\ -- sid acc n true | sid acc 0
   while
    swap #10 * +
  repeat
  nip
;

: xGetPos	\ sid -- x y ; return cursor position
  >r
  s\" \x1b[6n" r@ type-xterm		\ cursor request
  r@ key-xterm drop  r@ key-xterm drop	\ discard "ESC ["
  r@ getpn 1- r> getpn 1- swap		\ y then x
;

: xSetx		\ u sid -- ; set cursor at column u (0 based)
  dup >r xGetPos nip  r> xSetPos
;

struct /xedit	\ -- len ; holds edit structure
  int xe.len		\ length of text in buffer
  int xe.addr		\ base address of buffer
  int xe.maxlen		\ length of buffer
  int xe.backup		\ cursor offset from end of text
  int xe.spos		\ number of chars on line before ACCEPT
  int xe.sid		\ SID of device
end-struct

: xeLeft	\ xe -- ; step left
  dup xe.backup @ over xe.len @ < if
    1 over xe.sid @ xleft  dup xe.backup incr
  endif
  drop
;

: xeRight	\ xe -- ; step right
  dup xe.backup @ 0> if
    1 over xe.sid @ xright  dup xe.backup decr
  endif
  drop
;

: curBuffPos	\ xe -- addr ; get address of char after cursor
  dup xe.len 2@ +  swap xe.backup @ -
;

: xetrim	\ xe -- ; trim backup as required
  dup xe.backup @ over xe.len @  min  swap xe.backup !
;

: xeDelB	\ xe -- ; delete before cursor
  dup xe.len @ over xe.backup @ > if	\ has char before cursor?
    dup xe.sid @ xDelBefore		\ delete on display
    dup xe.backup @ if			\ need to shuffle?
      dup curBuffPos dup 1-		\ -- xe src dest
      2 pick xe.backup @ cmove
      dup xe.len decr			\ reduce length
      dup xetrim			\ check backup
    else
      dup xe.len decr
    endif
  endif
  drop
;

[-short-branches
: xeDelA	\ xe -- ; delete after cursor
  dup xe.len @ if		        \ if there's anything to delete
    dup xe.backup @ 0> if		\ if there's a tail
      dup xe.sid @ xDelAfter		\ delete on display
      dup xe.backup @ 1 > if		\ and its more than one character
        dup curBuffPos dup 1+ swap	\ cursor position = dest, src=cp+1
        2 pick xe.backup @ 1- cmove	\ copy tail down one character
      then
      dup xe.len decr			\ cut length of line
      dup xe.backup @ 1- 0 max over xe.backup !	\ tail is now one char shorter
      dup xe.backup @ if
        dup xetrim
      then
    then
  then
  drop
; doNotSin
short-branches]

: xe?~		\ xe -- flag ; return true if next char is ~
  xe.sid @ key-xterm [char] ~ =
;

: xeHome	\ xe -- ; process HOME key
  dup xe.len @ over xe.backup !		\ change offset
  dup xe.spos @ swap xe.sid @ xSetX	\ change cursor
;

: xeEnd		\ xe -- ; process END key ; SFP006
  dup xe.backup @ if			\ if not at end
    0 over xe.backup !			\ change offset
    dup xe.spos @ over xe.len @ +  over xe.sid @ xSetx
  endif
  drop
;
((
: xeEnd		\ xe -- ; process END key
  0 over xe.backup !			\ change offset
  dup xe.spos @ over xe.len @ +  swap xe.sid @ xSetx
;
))

: xeInsertChar	\ char xe -- ; handle character
  2dup xe.sid @ emit-xterm		\ to display
  dup xe.backup @ if			\ if not at end of line
    dup >r curBuffPos			\ cursor position
    dup dup 1+ r@ xe.backup @ cmove>	\ -- char src ; R: -- xe ; space for one char
    c!  r>				\ insert character
  else
    tuck xe.len 2@ + c!
  then
  xe.len incr
;

: xeSetOUT	\ xe -- ; set OUT
  dup xe.spos @ over xe.len @ +
  swap xe.backup @ - out !
;

: xeSetInsert	\ xe -- ; set insert mode
  s\" \e[4h" rot xe.sid @ type-xterm
;

: xeSetReplace	\ xe -- ; set replace mode
  s\" \e[4l" rot xe.sid @ type-xterm
;

: xeDelLine	\ xe -- ; delete current line
  dup xe.spos @ over xe.sid @ xSetx
  s\" \e[K" rot xe.sid @ type-xterm
;

: xePrevHist	\ xe -- ; get previous history line
  dup >r xeDelLine			\ delete current entry
  r@ xe.sid @ curr-history		\ -- caddr len ; R: -- xe ; get new line
  dup r@ xe.len !  0 r@ xe.backup !	\ reset length and cursor offset
  2dup r@ xe.addr @ swap move		\ copy to buffer
  r@ xe.sid @ type-xterm		\ to display
  r> xe.sid @ -1 step-history		\ step to previous line
;

[-sin
: xeNextHist	\ xe -- ; get next history line
  dup xeDelLine 			\ clear current line
  dup xe.sid @ at-last-history? if	\ leave blank if already at last
    dup xe.len off  xe.backup off
  else					\ otherwise get next
    dup >r xe.sid @ +1 step-history	\ step to next line
    r@ xe.sid @ curr-history		\ get text
    dup r@ xe.len !  r@ xe.backup off	\ reset length and cursor offset
    2dup r@ xe.addr @ swap move		\ copy to buffer
    r> xe.sid @ type-xterm		\ and replace the line
  endif
;
sin]

\ Keys	   Linux                DOS
\ F1       csi 1 1 ~, csi P     0 59
\ F2       csi 1 2 ~, csi Q	0 60
\ F3       csi 1 3 ~, csi R	0 61
\ F4       csi 1 4 ~, csi S	0 62
\ F5       csi 1 5 ~		0 63
\ F6       csi 1 7 ~		0 64
\ F7       csi 1 8 ~		0 65
\ F8       csi 1 9 ~		0 66
\ F9       csi 2 0 ~		0 67
\ F10      csi 2 1 ~		0 68
\ F11      csi 2 3 ~		0 133
\ F12      csi 2 4 ~		0 134
\ up       csi A                0 72
\ left     csi D                0 75
\ right    csi C                0 77
\ down     csi B                0 80
\ home     csi H                0 71
\ end      csi F		0 79
\ delete   csi 3 ~		0 83
\ insert   csi 2 ~		0 82
\ PgUp     csi 5 ~		0 73
\ PgDn     csi 6 ~		0 81
\ del-<    127			8

[-sin
[-short-branches
: xeEsc		\ xe -- ; process ESC key
  dup xe.sid @ key-xterm [char] [ = if	\ ESC [ check
    case  dup xe.sid @ key-xterm		\ -- xe char
      [char] D of  dup xeLeft  endof
      [char] C of  dup xeRight  endof
      [char] H of  dup xeHome  endof
      [char] F of  dup xeEnd  endof
      [char] 3 of  dup xe?~ if  dup xeDelA  endif  endof
      [char] A of  dup xePrevHist  endof
      [char] B of  dup xeNextHist  endof
    endcase
  endif
  drop
;

: accept-xterm	\ address maxlen sid -- len'
  { | xe[ /xedit ] -- len' }
  xe[ xe.sid !
  xe[ xe.maxlen !
  xe[ xe.addr !
  0 xe[ xe.len !			\ #chars read to buffer
  0 xe[ xe.backup !			\ cursor offset from END of line
  out @ xe[ xe.spos !			\ characters already on line
  xe[ xe.sid @ restart-history		\ history starts with last line used
  xe[ xeSetInsert

  begin
    xe[ xe.sid @ key-xterm
    dup EOLchar_con <>
    xe[ xe.len @  xe[ xe.maxlen @ <> and
   while
    dup #9 =				\ convert <TAB> to <SPACE>
    if  drop bl  then
    case
      IGNchar_con of  endof
      #23 of  ( <CTRL><W> - step left ) xe[ xeleft  endof
      #18 of  ( <CTRL><R> - step right ) xe[ xeright  endof
      #8 of  ( <BACKSPACE> - delete before ) xe[ xeDelB  endof
      #127 of  ( <<-> - delete before ) xe[ xeDelB  endof
      #5 of   ( <CTRL><E> previous history line ) xe[ xePrevHist  endof
      #4 of   ( <CTRL><D> next history line ) xe[ xeNextHist  endof
      #27 of  ( ESC lead-in ) xe[ xeEsc  endof
        xe[ xeInsertChar
    end-case
    xe[ xeSetOUT
  repeat
  drop

  xe[ xeEnd				\ go to end of line
  bl xe[ xe.len 2@ + c!			\ trailing space
  bl xe[ xe.sid @ emit-xterm
  xe[ xeSetReplace
  xe[ xe.len 2@ xe[ xe.sid @ add-history-line	\ update history
  xe[ xe.len @
;
short-branches]
sin]

: bell-xterm	\ sid -- ; audible beeper ; SFP006
  out @ >r
  7 swap emit-xterm
  r> out !
;

: setpos-xterm	\ x y mode sid -- ior ; SFP006
\ Go to cursor position x,y. Position 0,0 is top left. Mode is ignored.
  nip  2 pick >r
  xSetPos
  r> out !  0
;

: getpos-xterm	\ mode sid -- x y ior ; SFP006
  out @ >r
  nip xGetPos 0
  r> out !
;

: setFBC	\ fc bc sid -- ; SFP006
\ Set foreground and background colours. A value of -1 is used to
\ leave the current colour unchanged.
  {: | prevout temp[ 64 ] -- :}
  out @ -> prevout
  -rot  temp[ setCSI
  dup -1 <> if
    #40 + temp[ u+$
    dup -1 <>
    if  s" ;" temp[ append  endif
  else
    drop
  endif
  dup -1 <>
  if  #30 + temp[ u+$  else  drop  endif
  s" m" temp[ append
  temp[ count rot type-xterm
  prevout out !
; doNotSin

: ioctl-xterm	\ addr len fn sid -- ior
  case  swap
    #10 of  setFBC 0  endof
      4drop -1
  end-case
;

: flushOP-xterm	\ sid -- ior
  xs.hOut @ flush-file  ;

: readex-xterm	\ caddr len sid -- u ior
  xs.hOut @ read-file  ;

: init-xterm	\ addr len sid -- ior
  drop 2drop  0  ;

: term-xterm	\ sid -- ior
  drop  0  ;

: config-xterm	\ sid -- ior ; produces a dialog
  drop  0  ;

create xterm-vectors
  ' Open-xterm ,
  ' Close-xterm ,
  ' Read-xterm ,
  ' Write-xterm ,
  ' Key-xterm ,
  ' Key?-xterm ,
  ' Ekey-xterm ,
  ' EKey?-xterm ,
  ' Accept-xterm ,
  ' Emit-xterm ,
  ' Emit?-xterm ,
  ' Type-xterm ,
  ' CR-xterm ,
  ' LF-xterm ,
  ' FF-xterm ,
  ' BS-xterm ,
  ' Bell-xterm ,
  ' setpos-xterm ,
  ' Getpos-xterm ,
  ' Ioctl-xterm ,
  ' FlushOP-xterm  ,
  ' readex-xterm ,
  ' Init-xterm ,
  ' Term-xterm ,
  ' config-xterm ,

create defxterm	\ -- addr
\ Default xterm structure - must match /xterm
  0 , xterm-vectors , 0 , 1 , 2 ,  0 ,	\ last one is flags
  0 , 0 , 0 ,				\ history

previous definitions

also system

: initXtermSid	\ addr --
\ *G Initialise a *\fo{/xterm} structure at *\i{addr}.
  defxterm swap /xterm-sid cmove
;

: xterm:	\ -- ; -- sid ; XTERM: <name>
\ *G Create a new terminal device.
  create
    here /xterm-sid allot initXtermSid
;

xterm: xconsole	\ -- addr
\ *G VFX Forth console.

: init-xcon	\ --
\ *G Set up to use the *\fo{xconsole} device. Performed at start up
\ ** and compilation.
  xconsole  $000C over xs.flags or!	\ demand history buffering
  dup ip-handle !  dup op-handle !	\ make it the current device
  dup ip-default !  dup op-default !	\ and the default device
  to PauseConsole
  NULL NULL NULL open-gen 2drop
  xconsole readXtermCfg
;
' init-xcon AtCold  init-xcon

: term-xcon	\ --
\ *G Shut down the *\fo{xconsole} device. Performed at shut down.
  xconsole writeXtermCfg
;
' term-xcon AtExit

((
: ReadXcon	\ --
  xconsole read-history
; ' ReadXcon AtIniLoad

: WriteXcon
  xconsole write-history
; ' WriteXcon AtIniSave
))

previous


\ ======
\ *> ###
\ ======

decimal

