\ TermControl.fth - exploring XTERMs

((
Copyright (c) 2007
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *> genio
\ *H XTERM Control
\ ===========
\ *P This section documents the control sequences required by the
\ ** Bash shell. The source code is in *\i{Lib/Osx64/TermControl.fth}.

: .csi		\ --
\ *G Issue an "ESC [" sequence.
  s\" \x1b[" type
;

: .pu		\ u --
\ *G Issue u as a decimal text sequence.
  base @ >r  decimal (u.) type  r> base !
;

: (at-xy)	\ x y --
\ *G Go to cursor position x,y. Position 0,0 is top left.
  .csi  1+ .pu  ." ;" 1+ .pu ." H"
;

: (home)	\ --
\ *G Cursor to top left.
  .csi ." H"
;

: (cls)		\ --
\ *G Clear screen and put cursor at top left.
  .csi ." 2J" (home)
;

: getpu		\ -- u
\ *G Read characters up to a non-digit, which is discarded. The
\ ** value of the digits is returned.
  0 begin
    key #10 digit			\ -- acc n true | acc 0
   while
    swap #10 * +
  repeat
;

: (xy?)		\ -- x y
\ *G Return the current X,Y position. Top left is 0,0.
  .csi ." 6n"				\ cursor request
  key drop key drop			\ discard "ESC ["
  getpu getpu swap
;

\ *[
: bold		.csi ." 1m" ;
: bright	.csi ." 41;37m"  ;
: normal	.csi ." 0m"  ;
: inverse	.csi ." 7m"  ;

0 constant Black
1 constant Red
2 constant Green
3 constant Yellow
4 constant Blue
5 constant Magenta
6 constant Cyan
7 constant White
\ *]

: +bright	\ color -- color'
\ *G Convert a colour into its bright version.
  #60 +
;

: setFBC	\ fc bc --
\ *G Set foreground and background colours. A colour value of -1 leaves
\ ** the current colour unchanged.
  .csi  dup -1 <> if
    #40 + .pn  dup -1 <>
    if  ." ;"  endif
  else
    drop
  endif
  dup -1 <>
  if  #30 + .pn  else  drop  endif
  ." m"
;

\ Keys
\ up       csi A
\ left     csi D
\ right    csi C
\ down     csi B
\ home     csi H
\ end      csi F
\ delete   csi 3 ~
\ insert   csi 2 ~
\ PgUp     csi 5 ~
\ PgDn     csi 6 ~
\ del-<    127
\ F1       csi 1 1 ~, csi P
\ F2       csi 1 2 ~, csi Q
\ F3       csi 1 3 ~, csi R
\ F4       csi 1 4 ~, csi S
\ F5       csi 1 5 ~
\ F6       csi 1 7 ~
\ F7       csi 1 8 ~
\ F8       csi 1 9 ~
\ F9       csi 2 0 ~
\ F10      csi 2 1 ~
\ F11      csi 2 3 ~
\ F12      csi 2 4 ~

\ ======
\ *> ###
\ ======

decimal

