(* ====================================================================
File Manager

copyright 2014 Roelf Toxopeus

VFX version.
Implements useful file and directory utillites, using NSFileManager.
Last: 17 December 2019 at 10:08:01 CET  -rt
==================================================================== *)

(( --------------------------------------------------------------------
(COPY-ITEM) -- expects the source and destination paths as NSString
references. You should reference the item you're copying in the
destination string as well:
 @" ~/Documents/mydoc.txt" @" ~/Desktop/mydoc.txt" (COPY-ITEM)
You don't have to quote paths with spaces, they are recognized allright.
The given NSString refs are not released.

COPY-ITEM differs from (COPY-ITEM): expects zero terminated strings
for source and destination paths. The created NSString refs will be
released.

(MOVE-ITEM) and MOVE-ITEM similar to (COPY-ITEM) and COPY-ITEM respectively,
but moving instead of copying.

RENAME-ITEM -- rename item at path pointed to by zstring a1 with newname
pointed to by zstring a2. A version of MOVE-ITEM.

TRASH-ITEM -- move item at path to the trash. Return the resulting path,
in which the name may be changed because of duplicates in the trash.
Uses NSURL <-> CFURL tollfree bridging.
-------------------------------------------------------------------- ))

cocoaclass NSFileManager
cocoa: @defaultManager ( -- object )
cocoa: @copyItemAtPath:toPath:error: ( NSString:source NSString:destination **error -- boolean )
cocoa: @moveItemAtPath:toPath:error: ( NSString:source NSString:destination **error -- boolean )
cocoa: @localizedDescription ( -- NSStringRef )  \ NSError
cocoa: trashItemAtURL:resultingItemURL:error: ( NSURL:item NSURL:result **error -- boolean )

: ?FILE-ERROR ( error boolean -- )   0= if @localizedDescription >4thstring type else drop then ;
   
: (COPY-ITEM) ( NSString:source NSString:destination -- )
   0 >r rp@ NSFileManager @defaultManager @copyItemAtPath:toPath:error:
   r> swap ?file-error ;

: COPY-ITEM ( za1 za2 -- )
   >nsstring swap >nsstring swap
   2dup (copy-item)
   @release drop @release drop ;

: (MOVE-ITEM) ( NSString:source NSString:destination -- )
   0 >r rp@ NSFileManager @defaultManager @moveItemAtPath:toPath:error:
   r> swap ?file-error ;

: MOVE-ITEM ( za1 za2 -- )
   >nsstring swap >nsstring swap
   2dup (move-item)
   @release drop @release drop ;

: RENAME-ITEM  ( za1 za2 -- )
   over >r zcount
   r> zcount pad place pad stripfilename
   pad count pathpad zplace s" /" pathpad zappend
   pathpad zappend
   pathpad move-item ;

: TRASH-ITEM ( za1 -- za2 )
   >cfurl dup ( *) 0 >r rp@ 0 >r rp@
   NSFileManager @defaultManager trashItemAtURL:resultingItemURL:error:
   r> swap ?file-error
   ( *) CFRelease drop r> cfurl> ;

\\ ( eof )