(* ====================================================================
ObjC executes Forth

copyright 2008 Roelf Toxopeus

Part of the interface to ObjC 2 Runtime.
VFX version.
Generic form of executing Forth words on Main thread.
Last: 25 September 2019 at 12:21:40 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
The following words allow execution of Forth words in the main eventloop.
Mostly used to execute GUI specific actions on the main thread from within
the Forth thread.

PBLOCK -- memory block for int and fp parameters:
#ints  n0 ... nn    #floats   r0 ... rn
cell   #ints cells  cell      #floats floats
!PARAMETERS -- pop #ints from data stack and#floats from fp stack to PBLOCK.
@PARAMETERS -- push ints and floats from PBLOCK to data and fp stacks. 

ForthClass -- wrapper class to execute Forth words in ObjC.
@doForth: -- class method, which pops the needed int and fp parameters
from PBLOCK and executes FWORD. For savety reasons FWORD is set to NOP
after execution.

PASS -- deliver xt to main runloop via the generic ForthClass. Pushing
the needed integer and fp parameters to PBLOCK. It synchronizes and
PASS is wrapped inside GET/RELEASE.
-------------------------------------------------------------------- ))

/forth
decimal

: float- ( n -- n' )   FPCELL - ;

: @+ ( addr -- a+4 n )   dup cell+ swap @ ;

512 buffer: PBLOCK

\ these should be in assembler
: !PARAMETERS ( n1 ... nn #ints #floats -- ) ( F: r1 ... rn -- )
   2dup floats swap cells + [ 2 cells ] literal + \ needed bytes
   pblock +                        \ start at bottom
   over 0 ?do float- dup f! loop   \ fp's
   cell- tuck !                  \ #fp's
   swap dup >r
   0 ?do cell- tuck ! loop       \ int's
   r> swap cell- ! ;             \ #int's

: @PARAMETERS ( -- n1 ... nn ) ( F: -- r1 ... rn )
   pblock
   @+ 0 ?do @+ swap loop               \ integers
   @+ 0 ?do dup f@ float+ loop drop ;  \ floats
   
defer fword

: nop ( -- )   ;

NSObject new.class ForthClass

forthcall: *doForth ( rec sel param -- void )   finit  @parameters  fword  ['] nop is fword  0 ;

: doForthtypes   0" v@:" ;

*doForth doForthtypes 0" doForth:" ForthClass ADD.CLASSMETHOD

cocoa: @doForth: ( n -- ret )

\ ForthClass add.class      \ will be executed in INIT.COCO, can't doit twice!

variable 'pass
: PASS ( #ints #floats xt -- )
   'pass get
   is fword !parameters
   0" doForth:" @selector 0 yes ForthClass formain drop
   'pass release ;

\\ ( eof )

\ test:
variable var
: bump var +! ;                                                                                    
1 1 0 ' bump pass var cr ? 500 ms many

\ try nested cocoa versions
: (bump) ( n -- )  var +! ;
: bump ( n -- )  1 0 ['] (bump) pass ;                                                                               
1 1 0 ' bump pass var cr ? 500 ms many
