(* ====================================================================
Run coco-vfx

copyright 2009 Roelf Toxopeus

Part of adding Cocoa event handling to Forth.
VFX version.
Running a cocoa interface on the main thread
Last: 25 September 2019 at 12:27:16 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
The ObjC interface will not work properly when it isn't run on the main
thread. At least wrt GUI and graphic stuff. One way to solve this is
the following:
Cocoa can run on main thread and QUIT on another in SwiftForth. In this
case the main thread, OPERATOR, is lost for further direct interaction
and you lose the main thread's privileges. Use FORMAIN or PASS for those
situations.

/COCOALINKS -- relink all our cocoa and cocoa related words before
we can use them safely in turnkey.
INIT.COCOA -- initialise, launch the NSApp part of coco-vfx and make
NSApp multi threading aware. This should happen _before_ PRETENDING
so the Cocoa framework is protected. It will also init the ForthClass.

RUNCOCOA -- run the NSApplication run method.

COCO.RUN -- topword: init, launch and run coco-vfx. Cocoa on main thread
and Forth on another, side by side.

Note: main thread will save autoreleasepool in mainpool (see VFX>COCOA).
might be drained automatically by main thread, so saved pool is not valid
after a while.
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ initialise and launch the NSApp part of coco-vfx

: /COCOALINKS ( -- )
   /cocoaclasses /cocoaselectors
   /forth-cocoaclasses /forth-cocoaclassmethods /forth-cocoamethods /forth-cocoaivars ;
   
: INIT.COCOA ( -- )   /cocoalinks  vfx>cocoa  multi.aware  ForthClass add.class ;
   
\ --------------------------------------------------------------------
\ setup IMPOSTOR running Forth and OPERATOR runnning Cocoa.
   
cocoa: @run ( -- ret )
: RUNCOCOA ( -- )   NSApp @ @run drop ;

: COCO.RUN ( -- )   init.cocoa pretending runcocoa ;

\\ ( eof )
