(* ====================================================================
No AppNap

copyright 2013 Roelf Toxopeus

Specific OS handling.
VFX64 version.
Protect coco-vfx from being put to sleep by OS
Last: 4 August 2020 at 15:58:39 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Run the following in the Terminal window of coco-vfx:

: TT ( -- )   BEGIN  CR ." ONE " 500 MS ." TWO " 1000 MS ." TEST" KEY? UNTIL ;

Notice after a short while the output gets irregular. As if the app is
frozen. Bringing the NSApp part to the front, kicks coco-vfx in to life
again. But Terminal is in the background, not very convenient in an in-
ter active environment. It looks like the NSApp part needs some attention
once in a while.
NoTE: no such issues with the regular vfx and the experimental coco-vfx
with its own I/O window.

As we see it now (could change):
As of OSX 10.9 Mavericks, there's a new feature causing an application
to be put to sleep by the OS: App Nap. This happens under certain
circumstances as described here:
https://developer.apple.com/library/mac/releasenotes/MacOSX/WhatsNewInOSX/Articles/MacOSX10_9.html#//apple_ref/doc/uid/TP40013207-CH100

The NSApp part of coco-vfx, the main thread/task, is not doing anything
for us when not having any GUI related or main thread related things
going on. Appearently the OS sees this as the whole application idling,
and puts the app napping. With concequenses for Forth...

Using the info from above mentioned Apple doc we can try to convince
the OS to keep coco-vfx awake. Also see the NSProcessInfo Class Reference.
The resulting code below seems to work and Activity Monitor doesn't
show coco-vfx running wild. It's just as in Mountain Lion etc.

NoTE: coco-vfx still shows App Nap as -yes- in Activity Monitor.

Update: defaults write coco-vfx NSAppSleepDisabled -bool YES
will change App Nap as -no- in Activity Monitor.
You only have to do this once, check coco-vfx.plist in ~/Library/Preferences

NONAP -- first checks if the nap stuff is availlable (10.9 and up),
if so, disables App Nap.
-------------------------------------------------------------------- ))

/forth
decimal

cocoaclass NSProcessInfo
cocoa: @processInfo ( -- info )          \ returns NSProcessInfo object
cocoa: @beginActivityWithOptions:reason: ( option nsstring -- objecttoken )

$ff constant NSActivityBackground                    \ used option 0x000000FFULL

variable ACTIVITY-TOKEN                                   \ for now, might need it later

: NONAP ( -- )
   z" beginActivityWithOptions:reason:" @selector
   NSProcessInfo @processInfo
   dup >r @respondsToSelector:
   if   NSActivityBackground @" coco-vfx stays awake!"
        r> @beginActivityWithOptions:reason: activity-token !
   else r> drop then ;

\\ ( eof )
