(* ====================================================================
WKWebKit browsing

copyright 2006 Roelf Toxopeus

VFX64 version.
Playing with webkit as of Yosemite 10.10.
Last: 1 September 2020 at 15:44:43 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Implement a crude webbrowser. This version uses the preferred WKWebView!
Only for 64bit Forth.

WEBWIN -- window for web content.
NEW.WEBVIEW -- create new webview instance, fitting the contentview.
Use INIT, no need for a frame size, works on Panther and upwards.
MAKE-WEBVIEW -- creates new webview and connects to given window.
HELLO-WEB -- initialise and open a window for web browsing
BYE-WEB -- closes it.

LOAD -- will open window and load and display the given url.
<GO -- attemp to go to back
GO> -- attemp to go forward
RELOAD -- reload current url

NoTE: as of OS X 10.11 many NSWindow NSView and WebKit usage should be executed on the main thread.
Examples of strategic placing PASS ...

Following words are executed on Main thread:
SHOW-VIEW ( ADD.WINDOW)
NEW.WEBVIEW
MAKE-WEBVIEW ( SHOW-VIEW NEW.WEBVIEW and MAKE.CONTENTVIEW)
LOADREQUEST
GOBACK
GOFORWARD
RELOAD

-------------------------------------------------------------------- ))


/forth
decimal

\ --------------------------------------------------------------------
\ Window for webview:
new.window WEBWIN

: SHOW-VIEW ( wptr4 -- )
   dup add.window
   0" Web View" swap set.wtitle ;

\ --------------------------------------------------------------------
\ Create a webview:

frameworK WebKit.framework
WebKit.framework

cocoaclass WKWebView

variable NativeView      \ global usage

\ Main thread redefinitions
: (NEW.WEBVIEW) ( -- )  WKWebView @alloc @init nativeview ! ;
: NEW.WEBVIEW ( -- WKWebViewRef )  0 dup ['] (new.webview) pass nativeview @ ;

\ SHOW-VIEW NEW.WEBVIEW and MAKE.CONTENTVIEW are executed on main thread
: MAKE-WEBVIEW ( wptr4 -- )
   dup show-view
   new.webview                  \ make new webview instance
   swap make.contentview        \ connect to nswindow
;

: HELLO-WEB ( -- )   webwin dup +w.ref @ if drop exit then make-webview ;

: BYE-WEB ( -- )   nativeview @ @release drop  webwin close.window ;

\ --------------------------------------------------------------------
\ Interface: load url in web view

cocoaclass NSURL
cocoaclass NSURLRequest

cocoa: @URLWithString:   ( NSStringRef -- NSURLRef )    \ NSURL
cocoa: @requestWithURL: ( NSURLRef -- NSURLRequesRef )  \ NSURLRequest
cocoa: @loadRequest: ( NSURLRequesRef -- WKNavigation ) \ WKWebFrame

: REQUEST ( 0string -- NSURLRequesRef )
   >nsstring nsurl @URLWithString:
   NSURLRequest @requestWithURL: ;
   
: LOADREQUEST ( 0string WKWebViewRef -- )  >r request r> @loadRequest: drop ;

: LOAD ( url -- )   hello-web nativeview @ 2 0 ['] loadrequest pass ;

\ --------------------------------------------------------------------
\ Interface: more controls

cocoa: @canGoBack ( -- flag )          \ BOOL
cocoa: @goBack ( -- WKNavigation )     \ id
cocoa: @canGoForward ( -- flag )       \ BOOL
cocoa: @goForward ( -- WKNavigation )  \ id
cocoa: @reload ( -- WKNavigation )     \ id

: GOBACK ( obj -- )   dup @canGoBack if  @goBack  then  drop  ;
: GOBACK ( obj -- )   1 0 ['] goback pass ;
: <GO ( -- )  nativeview @ goback ;

: GOFORWARD ( obj -- )   dup @canGoForward if  @goForward  then  drop ;
: GOFORWARD ( obj -- )   1 0 ['] goforward pass ;
: GO> ( -- )   nativeview @ goforward ;

: RELOAD ( -- )   nativeview @ @reload ;
: RELOAD ( -- )   0 0 ['] reload pass ;

\ --------------------------------------------------------------------
\ Shortcuts:

: BMB ( -- )   0" http://www.bmbcon.org" LOAD ;

: KNMI ( -- )   0" https://www.knmi.nl" LOAD ;

: HAIKU ( -- )   0" https://forthsalon.appspot.com" LOAD ;

: COCOA ( -- )   0" https://developer.apple.com/library/mac/index.html" LOAD ;

: INFO ( -- )   0" https://www.ixquick.com/" LOAD ;

: VFXOSX ( -- )   0" https://www.mpeforth.com/vfxmac.htm" load ;

: MAPS ( -- )   0" https://www.openstreetmap.org/#map=12/51.9811/5.9055" load ;

\\ ( eof )