(* ====================================================================
Custom Control Panel

copyright 2012 Roelf Toxopeus

VFX64 version
Programmaticly adding controls, like buttons and sliders
Last: 28 January 2020 at 16:28:23 CET  -rt
==================================================================== *)

(( --------------------------------------------------------------------
Interesting part is adding a method to an existing class:
Objective-C allows you to add your own methods to existing classes
through categories and class extensions:
Customizing Existing Classes

Create a window MYCONTROLWIN for the control objects: NSButton and
NSSlider, both NSControl subclasses. The button will randomly set a
scaler value for the slider.
Declare some methods from NSWindow, NSControl, NSSlider and NSActionCell.
Create a method for what to do when the button is hit and add it to
the NSButton class. Similar for the slider.
/BUTTON and /SLIDER will create instances, set their actions and
draw them in MYCONTROLWIN.
/CONTROL ties it all together and displays MYCONTROLWIN with the
controls.
Use SLIDER.MAX and SLIDER.MAX! to inspect and change the maximum
slider values.
SLIDER! sets the slider value programmatically.

Notice @initWithFrame: which needs the special COCOA.DFIX approach
because of passing the 4 doubles from the rectangle.
@setMaxValue: and @setMinValue: don't need this.
-------------------------------------------------------------------- ))

/forth
decimal

\ window containing controls
new.window MYCONTROLWIN
474e0 569e0 243e0 121e0 mycontrolwin w.bounds
s" controls"            mycontrolwin w.title

\ NSControl subclasses
cocoaclass NSButton
cocoaclass NSSlider

\ some methods needed
COCOA: initWithFrame: ( dummy dummy dummy dummy x y w h -- ret )
: @initWithFrame: ( -- ret ) ( x y w h -- )    4 cocoa.dfix  initWithFrame: ;
cocoa: @addSubview: ( view -- ret )                \ NSView
cocoa: @intValue ( -- n )                          \ NSControl
cocoa: @setMaxValue: ( -- ret ) ( r -- )           \ NSSlider
cocoa: @setMinValue: ( -- ret ) ( r -- )           \ NSSlider
cocoa: @setAction: ( sel -- ret )                  \ NSActionCell
cocoa: @setTarget: ( id -- ret )                   \ NSActionCell

\ some needed words for all
variable SCALER 12 choose scaler !
: CONTROLTYPES ( -- a )  0" v@:@" ;

\ add a method to the NSButton class
\ what to do when our button is hit
forthcall: *klik ( rec sel sender -- void )
   impostor's
   2drop drop
   12 choose scaler !
   0 20 at-xy ."  scaler: " scaler ? ;
 
*klik controltypes 0" klik" NSButton add.method

\ draw button and set it's action
: /BUTTON ( -- )
   10e0 10e0 100e0 100e0                                      ( frame size and position )
   NSButton @alloc @initWithFrame:                            ( button instance )
   @" Click me!" over @setTitle: drop
   dup mycontrolwin +w.ref @ @contentView @addSubview: drop   ( add control as view to windowview )
   dup dup @setTarget: DROP                                   ( self is target )
   0" klik" @selector swap @setAction: drop                   ( apply new method as action )
;

\ add a method to the NSSlider class
\ what to do when slider is moved
forthcall: *schuif ( rec sel sender -- void )
   impostor's nip nip @intValue scaler @ *  20 dup at-xy 4 .r cr ;

*schuif controltypes 0" schuif" NSSlider add.method

\ draw slider and set the action
variable MYSLIDER
: /SLIDER ( -- )
   200e0 10e0 30e0 100e0                                      ( frame size and position )
   NSSlider @alloc @initWithFrame:                            ( slider instance )
   dup mycontrolwin +w.ref @ @contentView @addSubview: drop   ( add control as view to windowview )
   dup dup @setTarget: drop                                   ( self is target )
   0" schuif" @selector over @setAction: drop                 ( apply new method as action )
   100e0 dup @setMaxValue: drop                               ( set maximum value )
     0e0 dup @setMinValue: drop                               ( set minimum value )
   myslider !
;

\ just to test the FPRET call
\ Note: in a FUNCTION: or COCOA: declaration do not use a F: in the first stackpicture,
\ the F: will be seen as a parameter!
cocoa-fpret: @maxValue ( -- ) ( F: -- r )  \ NSSlider
: SLIDER.MAX ( F: -- r )   myslider @ @maxvalue ;

: SLIDER.MAX! ( F: r -- )  myslider @ @setmaxValue: drop ;

\ test setting the slider programmaticly
cocoa: @setIntValue: ( n -- ret )   \ NSControl

: SLIDER! ( n -- )   MYSLIDER @ @setIntValue: drop ;

\ build control window
: /CONTROL ( -- )   mycontrolwin add.window /button /slider ;

cr .( /CONTROL to build control)

\\ ( eof )
