(* ====================================================================
Control panel Nib

copyright 2011 Roelf Toxopeus

VFX version.
Implement a control panel with one slider and its outlet to control
the slider programmaticly.
Last: 2 August 2020 at 10:50:04 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Cocoa from within VFX demonstration:

Demonstrates the interface to the objc-runtime in action:
-- some Cocoa calls
-- creating a new class for cocoa

NoTE: in VFX set the I/O stuff right for displaying strings.
     Use IMPOSTOR'S and don't forget to set FP stack in callback

The @awakeFromNib method in myclass, is executed during nib instantiation.
The method implementation here will just save the myclass instance ref
for further use.

This controller has just one slider:
The  @doSLider:   method will be used when user works the slider.

The slider has an outlet connected to myclass: myOutlet
BTW this is done in IB, not programmaticly.
The runtime will initiate the outlet, actually an instance variable,
using the setter method @setmyOutlet:
You can ommit this method, a default initiation method will be used instead.
The outlet should point to the NSSlider instance when initiated.

The user can programmaticly set the slider using the @myOutlet getter
method.

SLIDER! will set the slider to given value.

SS will run all automaticly. Watch slider move.

Obviously you can change the slider by dragging the 'walker' left and
right.
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ the so called IBAction connected to the buttons and slider in the panel

synonym IBACTION: forthcall:

variable MYCLASSREF

ibaction: *awakeFromNib ( rec sel -- ret )   drop myclassref ! 0 ;

cocoa: @intValue ( -- n )  \ NSControl method

ibaction: *doSlider: ( rec sel sender -- n )   impostor's nip nip @intValue  cr .  0 ;

ibaction: *myOutlet ( rec sel -- n )   drop  z" myOutlet" swap @ivar ;

ibaction: *setmyOutlet: ( rec sel obj -- ret )   nip swap Z" myOutlet" swap !ivar 0 ;

: CONTROLTYPES ( -- a )   0" v@:V" ;

: @OUTLETTYPES ( -- a )    0" @@:" ;

: !OUTLETTYPES ( -- a )    0" v@:@" ;

: VARTYPE ( -- a )   0" @" ;

\ --------------------------------------------------------------------
\ The ForthClass, create class

NSObject new.class myclass

\ Assign callbacks as methods to the class and register their names with the objc runtime

*awakeFromNib controltypes 0" awakeFromNib"  myclass add.method   \ internaly used by runtime at instantiating nib

*doSlider:  controltypes 0" doSlider:"  myclass add.method        \ internaly used by controler

*myOutlet @outlettypes 0" myOutlet"  myclass add.method           \ outlet getter, publicly used by us !       

*setmyOutlet: !outlettypes 0" setmyOutlet:"   myclass add.method  \ outlet setter, internaly used by controler

\ Add instance variable, the outlet to the class

vartype 0" myOutlet" myclass add.ivar                             \ the IBOutlet !

\ Add the class to the runtime

myclass add.class

\ --------------------------------------------------------------------
\ testing

\ Topword
\ : MYNIB ( -- nib )   z" mycontroller.nib" @nib dup /nib 0= abort" Can't initiate NIB !" ;
: MYNIB ( -- windowcontroller )   z" mycontroller.nib" show.nib dup 0= abort" Can't initiate NIB !" ;

cocoa: @myOutlet ( -- NSSlider:ref )  \ myclass !!!
cocoa: @setIntValue: ( n -- ret )     \ NSSlider

: SLIDER! ( n -- )   myclassref @ @myOutlet @setIntValue: drop ;

: SS ( -- )  mynib drop  begin  100 choose slider!  500 ms  key? until ;

cr .( control panel with slider loaded ...)
cr .( ss  -- to start things)

\\ ( eof )