(* ====================================================================
Web browsing

copyright 2006 Roelf Toxopeus

VFX version
Running the webview stuff from cocoanib on mainthread
Needs cocoawebview6.fth
Last: 20 October 2020 at 14:19:44 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
All GUI stuff should be run on the main thread. The web browser created
in cocoawebview3.f runs in Forth, here not on the main thread.

Solution is to create a control panel, run on the main thread, which
executes cocoawebview3 words. A Heads Up Display is used as control
panel: resources/runweb-sf.nib

Creation follows the normal procedure: method implementations,
class creation, loading NIB and run it.
-------------------------------------------------------------------- ))

pushpath  mac include hotcoco/cocoawebview.fth  poppath

/forth
decimal

\ --------------------------------------------------------------------
\ The so called IBAction connected to the buttons in the window

forthcall: *cocoa: ( rec sel sender:button -- n )   2drop drop cocoa 0 ;

forthcall: *bmb: ( rec sel sender:button -- n )   2drop drop bmb  0 ;

forthcall: *vfx: ( rec sel sender:button -- n )   2drop drop vfxosx  0 ;

forthcall: *info: ( rec sel sender:button -- n )   2drop drop info  0 ;

forthcall: *<go: ( rec sel sender:button -- n )   2drop drop <go  0 ;

forthcall: *go>: ( rec sel sender:button -- n )   2drop drop go>  0 ;

cocoa: @stringValue ( -- string )         \ NSControl method, NSTextView inherits from NSControl: -- NSString object
forthcall: *runit: ( rec sel sender:textview -- n )  nip nip @stringValue >4thstring drop load  0 ;

forthcall: *reload: ( rec sel sender:button -- n )   2drop drop reload  0 ;

: RUNWEBTYPES ( -- addr )  0" v@:V" ;

\ --------------------------------------------------------------------
\ Create class

NSObject new.class runWebClass

\ assign callback as method to the class and register its name with the objc runtime
*cocoa:    runwebtypes  0" runWebCocoa:"     runWebClass add.method
*bmb:      runwebtypes  0" runWebBMB:"       runWebClass add.method
*vfx:      runwebtypes  0" runWebVfx:"       runWebClass add.method
*info:     runwebtypes  0" runWebInfo:"      runWebClass add.method
*<go:      runwebtypes  0" runWebBackward:"  runWebClass add.method
*go>:      runwebtypes  0" runWebForward:"   runWebClass add.method
*runit:    runwebtypes  0" runWebIt:"        runWebClass add.method
*reload:   runwebtypes  0" runWebReload:"    runWebClass add.method

\ add the class
runWebClass ADD.CLASS

\ --------------------------------------------------------------------
\ The Nib stuff, a full path including nib filename

\ Show it
\ : DO.SHOW ( -- )    z" runweb-vfx.nib" @nib /nib 0= abort" Can't initiate NIB !" ;
: DO.SHOW ( -- )    z" runweb-vfx.nib" show.nib 0= abort" Can't initiate NIB !" ;

cr .( do.show  -- to start things)
cr .( use the control panel to navigate the web)

\\ ( eof )
