(* ====================================================================
Window delegation example

Copyright 2010 Roelf Toxopeus

VFX version.
Shows how to set your eventhandler for a window.
Last: 8 June 2020 at 10:40:02 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
Example of a 'self' delegate: NSWindow instances.
Adding a delegate method to the NSWindow class is enough to work with
the NSWindow delegate protocol. Appearently NSWindow instances are their
own delegates by default.

Here the windowShouldClose: delegate method is implemented. The action
of the window close button depends on the method implementation.
CLOSE? -- deferred. Implementation of the windowShouldClose: method.
Expects the sender, here the close button and leaves a true/false flag.
-CLOSE and CLOSE -- actions for CLOSE?
/DELEGATE and -DELEGATE -- setting CLOSE?
-------------------------------------------------------------------- ))

/forth
decimal

defer CLOSE? ( sender -- flag )

: -CLOSE ( sender -- NO )   drop impostor's ." We don't close!" cr no ;

: CLOSE ( sender -- YES )   drop impostor's ." Closing window... " cr 500 ms yes ;

FORTHCALL: *windowShouldClose: ( rec sel sender:button -- bool )   -rot 2drop close? ; 

: DELEGATETYPES   0" c@:@" ;

*windowShouldClose: DELEGATETYPES 0" windowShouldClose:"  NSWindow add.method

: /DELEGATE ( -- )   ['] -close is close? ;

: -DELEGATE ( -- )   ['] close is close? ;

\ --- demo

new.window MYWIN
S" Delegate Window" mywin w.title

: /DEMO ( -- )
   mywin add.window
   10 ms
   /delegate ;

CR .( window delegate test loaded ...)
CR .( /DEMO       -- init demo)
CR .( try window's red close button)
CR .( -DELEGATE   -- default window behaviour)

\\ ( eof )