(* ====================================================================
Draw colored boxes.

copyright 2006 Roelf Toxopeus

VFX64 version.
Last: 20 August 2020 at 10:47:52 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
Draw colored boxes, some slightly rotated, scaled and translated.
Also colors and transparency are randomly changed.

QWIN /GWINDOW    when included from LOAD-FILE menu.
Uses a borderless window QWIN.
TRANSFORM -- chooses one of the three possible transformations to
be applied to the given CGContext.
SLEEPY -- sleep a bit.
DRAW.A.BOX -- draws abox with the randomly provided transformation
and color: a reddish, a blueish and a yellowish box.
BOXES -- draws boxes until a key hit.
Q.RUNS -- runs BOXES in a background task QUARTZRUNNER.
-------------------------------------------------------------------- ))

/forth
decimal

new.window QWIN
   402e0 280e0 547e0 438e0 qwin w.bounds
   nowidgets                qwin w.style
   s" "                    qwin w.title

: /QWIN ( -- )   qwin dup add.window /gwindow ;

/qwin

: CHOOSE.TRANSPARENCY ( F: -- r )
   11 choose ?dup if 1e0 s>f f/ else 0e0 then ;

: CHOOSE.SCALE ( F:   -- xscal yscale )
   25e-2 4 choose 1+ s>f f* fdup ;

: CHOOSE.TRANSLATION ( F: -- xtrans ytrans )
   340 choose s>f 180 choose s>f ;

: CHOOSE.ROTATION ( F: -- angle )
   90 choose 45 - s>f d>r ;

: SCALE-IT ( cgcontext -- )  CHOOSE.SCALE CGContextScaleCTM DROP ;

: TRANSLATE-IT ( cgcontext -- )  CHOOSE.TRANSLATION CGContextTranslateCTM DROP ;

: ROTATE-IT ( cgcontext -- )  CHOOSE.ROTATION CGContextRotateCTM DROP ; 

CREATE TRANSFORMS  ' scale-it ,  ' translate-it ,  ' rotate-it ,

: TRANSFORM ( cgcontext -- )
   toss if
      3 choose cells transforms + @ execute
   else dup CGContextRestoreGState droP CGContextSaveGState drop  \ restore _and_ save for nexttime
   then ;

: COLORRECT ( cgcontext -- )
   dup cgcolor cgcolor@ CGContextSetRGBFillColor drop
   cgrect cgrect@ CGContextFillRect drop ;

: SLEEPY ( -- )   50 200 alea ms ;

: DRAW.A.BOX ( cgcontext -- )
   dup CGContextSaveGState drop
   dup transform
   dup colorrect
   dup CGContextRestoreGState drop
   sleepy CGContextFlush drop ;

: A.RED.BOX ( cgcontext -- )
   toss if
      1e0 1e0 1e0 5e-1 cgcolor cg4!
   else
      1e0 0e0 0e0 choose.transparency cgcolor cg4!
   then
   0e0 0e0 200e0 100e0 cgrect cg4!
   draw.a.box ;

: A.BLUE.BOX ( cgcontext -- )
   toss if
      1e0 1e0 1e0 5e-1 cgcolor cg4!
   else
      0e0 0e0 1e0 choose.transparency cgcolor cg4!
   then
   0e0 0e0 100e0 200e0 cgrect cg4!
   draw.a.box ;

: A.YELLOW.BOX ( cgcontext -- )
   toss if
      1e0 1e0 1e0 5e-1 cgcolor cg4!
   else      
      95e-2 1e0 0e0 choose.transparency cgcolor cg4!
   then
   90e0 90e0 173e0 189e0 cgrect cg4!
   draw.a.box ;

: BOXES ( -- )
   qwin cgcontext
   begin
      dup a.red.box
      dup a.blue.box
      dup a.yellow.box
   ?pause until
   drop ;

task QUARTZRUNNER

: Q.RUNS ( -- )    quartzrunner activate  allocpool  boxes  releasepool drop ;
: Q.DONE ( -- )    quartzrunner done ;

cr .( QWIN /GWINDOW    when included from LOAD-FILE menu)
cr .( Q.RUNS  starts)
cr .( Q.DONE  stops)

\\ ( eof )