(* ====================================================================
Path drawing

Copyright 2006 Roelf Toxopeus

VFX version.
CoreGraphics draw path tests.
Last: 23 April 2018 at 11:34:12 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
Draws 5 boxes, normal, translated, scaled, rotated and wheel.
Uses Paths to draw.
BOXES -- draws a box. Waits for a keystroke to draw next box.
AA -- draws each box in animation.

Note: uses the unsynced _CGContextFlush call
-------------------------------------------------------------------- ))

/FORTH
DECIMAL

NEW.WINDOW PATHWIN
 100e0 386e0 599e0 392e0 PATHWIN W.BOUNDS
 S" Path Window"         PATHWIN W.TITLE

: /PATHWIN ( -- )   PATHWIN DUP ADD.WINDOW /GWINDOW ;

/PATHWIN

\ the rest uses relative coordinates
: BOX1 ( -- )
   ." box starting in 0,0"
   PATHWIN CGContext
   DUP CGContextBeginPath DROP
   2E0 DUP LINE.WIDTH
        0e0 FDUP          F2DUP DUP MOVE.TO
        0e0  100e0 POINT+ F2DUP DUP LINE.TO
      100e0    0e0 POINT+ F2DUP DUP LINE.TO
        0e0 -100e0 POINT+       DUP LINE.TO
   DUP CGContextClosePath DROP
   DUP CGContextStrokePath DROP
   _CGContextFlush DROP ;

\ translating
: BOX2 ( -- )
   ." translate x+100 y+0"
   PATHWIN CGContext
   DUP CGContextSaveGState DROP
   100e0 0e0 DUP CGTRANSLATE
   DUP CGContextBeginPath DROP
   2e0 DUP LINE.WIDTH
        0e0 FDUP          F2DUP DUP MOVE.TO
        0e0  100e0 POINT+ F2DUP DUP LINE.TO
      100e0    0e0 POINT+ F2DUP DUP LINE.TO
        0e0 -100e0 POINT+       DUP LINE.TO
   DUP CGContextClosePath DROP
   DUP CGContextStrokePath DROP
   DUP _CGContextFlush DROP
   CGContextRestoreGState DROP ;

\ scaling
: BOX3 ( -- )
   ." scales 2x 0.5y"
   PATHWIN CGContext
   DUP CGContextSaveGState DROP
   2e0 5e-1 DUP CGSCALE
   DUP CGContextBeginPath DROP
   2e0 DUP LINE.WIDTH
        0e0 FDUP          F2DUP DUP MOVE.TO
        0e0  100e0 POINT+ F2DUP DUP LINE.TO
      100e0    0e0 POINT+ F2DUP DUP LINE.TO
        0e0 -100e0 POINT+       DUP LINE.TO
   DUP CGContextClosePath DROP
   DUP CGContextStrokePath DROP
   DUP _CGContextFlush DROP
   CGContextRestoreGState DROP ;

\ rotation
: BOX4 ( -- )
   ." rotates 10 degrees around 0,0"
   PATHWIN CGContext
   DUP CGContextSaveGState DROP
   -10e0 DUP CGROTATE
   dup CGContextBeginPath DROP
   2e0 DUP LINE.WIDTH
        0e0 FDUP          F2DUP DUP MOVE.TO
        0e0  100e0 POINT+ F2DUP DUP LINE.TO
      100e0    0e0 POINT+ F2DUP DUP LINE.TO
        0e0 -100e0 POINT+       DUP LINE.TO
   DUP CGContextClosePath DROP
   DUP CGContextStrokePath DROP
   DUP _CGContextFlush DROP
   CGContextRestoreGState DROP ;

\ Pivots around point somewher middle screen
: BOX5 ( -- )
   CR ." pivots around point 300,200"
   PATHWIN CGContext
   360 0 DO
    DUP CGContextSaveGState DROP
    300e0 200e0 DUP CGTRANSLATE
    I NEGATE S>F DUP CGROTATE
    DUP CGContextBeginPath DROP
        0e0 FDUP          F2DUP DUP MOVE.TO
        0e0  100e0 POINT+ F2DUP DUP LINE.TO
      100e0    0e0 POINT+ F2DUP DUP LINE.TO
        0e0 -100e0 POINT+       DUP LINE.TO
    DUP CGContextClosePath DROP
    DUP CGContextStrokePath DROP
    DUP _CGContextFlush DROP
    DUP CGContextRestoreGState DROP
    \ KEY DROP
   45 +LOOP
   DROP ;

: BOXES ( -- )
   CR ." hit any key to continue, until ok prompt"
   CR BOX1 KEY DROP CR BOX2 KEY DROP
   CR BOX3 KEY DROP CR BOX4 KEY DROP
   CR BOX5 CR ." done" ;
   
: AA ( -- )
   PATHWIN WIPE CR BOX1
   700 MS PATHWIN WIPE CR BOX2
   700 MS PATHWIN WIPE CR BOX3
   700 MS PATHWIN WIPE CR BOX4
   700 MS PATHWIN WIPE CR BOX5
   CR ." done" ;
   
CR .( path tests loaded)
CR .( try  BOXES  any key continues)
CR .( try  AA  runs it self)

\\ ( eof )