(* ====================================================================
CoreGraphics Affine Transform Matrices

Copyright 2009 Roelf Toxopeus

VFX64 version.
Current transformation matrix
Use together with CGContextSaveGState and CGContextRestoreGState
Most of these transform the space, not the points!!!
Last: 9 August 2020 at 21:08:16 CEST  -rt
==================================================================== *)

(* --------------------------------------------------------------------
Affine Transform Matrices to transform the CGContexts

CGatf -- a pad for temporary storage of a ATM
CGATF@ -- passing the AffineTransformstructure contents for calls.
Similar to CGRECT@ etc.

For transforming the context space there is: CGTRANSLATE CGSCALE and
CGROTATE, which do what their name suggest.
-------------------------------------------------------------------- *)

/FORTH
DECIMAL

\ -- provide the CGatf structure for the return!!!
FUNCTION: CGContextGetCTM ( *cgatf cgcontext -- ret )
FUNCTION: CGContextGetUserSpaceToDeviceSpaceTransform ( *cgatf cgcontext -- ret )

\ -- cgatf isn't a reference but structure I think, so adapt in the call
\ FUNCTION: CGContextConcatCTM ( cgcontext -- ret ) ( F: a b c d tx ty -- )
\ -- but 6 members is too big for passing by value, so try passing by reference
FUNCTION: CGContextConcatCTM ( cgcontext *CGatf -- ret )

FUNCTION: CGContextTranslateCTM ( cgcontext -- ret ) ( F: x y -- )
FUNCTION: CGContextScaleCTM ( cgcontext -- ret ) ( F: x y -- )
FUNCTION: CGContextRotateCTM ( cgcontext -- ret ) ( F: radians -- ) \ use D>R !!

CREATE CGatf 6 DFLOATS /ALLOT

\ Passing the AffineTransformstructure:
: CGATF@ ( addr -- ) ( F: -- a b c d tx ty )
   DUP >R CG2@ R> [ 2 DFLOATS ] LITERAL + CG4@ ;

\ Transform the context space:   
: CGTRANSLATE ( cgcontext -- ) ( F: x y -- )  CGContextTranslateCTM DROP ;

: CGSCALE ( cgcontext -- ) ( F: x y -- )   CGContextScaleCTM DROP ;

: CGROTATE ( cgcontext -- ) ( F: degrees -- )  D>R CGContextRotateCTM DROP ;

CR .( Transform matrices 64bit loaded )

\\ ( eof )