(* ====================================================================
Core Graphics drawing methods

Copyright 2006 Roelf Toxopeus

VFX64 version

Last: 9 August 2020 at 21:07:07 CEST   -rt
==================================================================== *)

(* --------------------------------------------------------------------
Note: all supplied and returned coordinates are fp numbers and passed
via the fp stack.
Apples way of passing CGRect via the stack ?!?!?!?
Fill up to 6 parameters and place rect via DFIX

All words deal with current cgcontext.
LINE.WIDTH -- set linewidth for cgcontext.
PENLOC -- return pen location for cgcontext.
MOVE.TO -- move pen to x/y in cgcontext.
LINE.TO -- add a line using current blendmode etc.
from current penloc to x/y. Doesn't draw it!
SYNC.DRAW -- stroke, flush and set current point in cgcontext
DRAW.TO -- draw a line added to cgcontext, uses SYNC.DRAW
RMOVE -- move pen relative to current pen location.
RDRAW -- draw a line, relative to current pen location.
LINE -- draw a line form x1/y1 to x2/y2, uses DRAW.TO
DOT -- draw a dot at x/y, uses SYNC.DRAW. Pixel size = 5E-1
WIPE or CLEAR.WINDOW.
-------------------------------------------------------------------- *)

/FORTH
DECIMAL

Quartz.framework

FUNCTION: CGContextGetPathCurrentPoint ( cgcontext -- ) ( F: -- x y )
FUNCTION: CGContextMoveToPoint ( cgcontext -- ret ) ( F: x y -- )
FUNCTION: CGContextAddLineToPoint ( cgcontext -- ret ) ( F: x y -- )
FUNCTION: CGContextAddRect ( cgcontext dummy dummy dummy dummy dummy x y x2 y2 -- ret )
: CGContextAddRect ( cgcontext -- ret ) ( F: x y x2 y2 -- )  5 DFIX  CGContextAddRect ;

FUNCTION: CGContextStrokePath ( cgcontext -- ret )
FUNCTION: CGContextFillPath ( cgcontext -- ret )
FUNCTION: CGContextFillRect ( cgcontext dummy dummy dummy dummy dummy x y x2 y2 -- ret )
: CGContextFillRect ( cgcontext -- ret ) ( F: x y x2 y2 -- )  5 DFIX  CGContextFillRect ;
FUNCTION: CGContextClearRect  ( cgcontext dummy dummy dummy dummy dummy x y x2 y2 -- ret )
: CGContextClearRect ( cgcontext -- ret ) ( F: x y x2 y2 -- ) 5 DFIX CGContextClearRect ;

FUNCTION: CGContextSetLineWidth ( cgcontext -- ret ) ( F: r -- )
FUNCTION: CGContextSetAlpha ( cgcontext -- ret ) ( F: alpha -- )
FUNCTION: CGContextSetRGBStrokeColor ( cgcontext -- ret ) ( F: r g b a -- )
FUNCTION: CGContextSetRGBFillColor ( cgcontext -- ret ) ( F: r g b a -- )
FUNCTION: CGContextSetGrayFillColor ( cgcontext -- ret ) ( F: gray alpha -- )

\ ---

: LINE.WIDTH ( context --) ( F: width --)
   CGContextSetLineWidth DROP ;

: PENLOC ( cgcontext -- ) ( F: -- x y )
   CGContextGetPathCurrentPoint ;

: MOVE.TO ( cgcontext -- ) ( F: x y -- )
    CGContextMoveToPoint DROP ;

: LINE.TO ( cgcontext -- ) ( F: x y -- )
   CGContextAddLineToPoint DROP ;

(*
: DRAW.TO ( cgcontext -- ) ( F: x y -- )
   f2dup
   dup line.to dup CGContextStrokePath drop
   dup CGContextFlush drop
   move.to ;
*)

: SYNC.DRAW ( cgcontext -- ) ( F: x y -- )
   COUNTER FLUSHED  2DUP @ - 33 U> IF                                \ time to work?
      !                                                              \ next flushtime
      DUP CGContextStrokePath DROP DUP _CGContextFlush DROP MOVE.TO  \ stroke, flush and set current point
      ELSE 2DROP DROP F2DROP THEN ;
   
: DRAW.TO ( cgcontext -- ) ( F: x y -- )
   F2DUP DUP LINE.TO SYNC.DRAW ;

: RMOVE ( cgcontext -- ) ( F: relativex relativey -- )
   PENLOC POINT+ MOVE.TO ;

: RDRAW ( cgcontext -- ) ( F: relativex relativey -- )
   PENLOC POINT+ DRAW.TO ;
   
: LINE ( cgcontext -- ) ( F: x1 y1 x2 y2 -- )
   dup f2swap move.to draw.to ;

(*
: DOT ( cgcontext -- ) ( F: x y -- )
   F2DUP
   DUP 1e0 FDUP CGContextFillRect DROP
   DUP CGContextFlush DROP
   MOVE.To ;
*)

: DOT ( cgcontext -- ) ( F: x y -- )
   F2DUP DUP 5e-1 FDUP CGContextAddRect DROP SYNC.DRAW ;

: WIPE \ CLEAR.WINDOW ( wptr4 -- )
   DUP CGContext
   DUP CGContextSaveGState DROP
   OVER CONTENT.FRAME           ( F: -- x y w h )
   SWAP +W.BACKGROUND CG4@      ( F: x y w h -- x y w h r g b a )
   DUP CGContextSetRGBFillColor DROP
   DUP CGContextFillRect DROP
   DUP CGContextRestoreGState DROP
   CGContextFlush DROP ;

(*
: WIPE \ CLEAR.WINDOW ( wptr4 -- )
   DUP CGContext >R
   R@ CGContextSaveGState DROP
   CGRECT OVER +W.REF @ @contentView @frame DROP
   R@ SWAP +W.BACKGROUND CGColor@ CGContextSetRGBFillColor DROP
   R@ CGRECT CGRect@ CGContextFillRect DROP
   R@ CGContextRestoreGState DROP
   R> CGContextFlush DROP ;
*)

CR .( Core Graphics basic drawing 64bit loaded )

\\ ( eof )
