(* ====================================================================
Wil Baden Random Number Generator

Copyright 2017 by FORTH, Inc.
copyright 1986+ Joel Ryan, STEIM, Roelf Toxopeus

See rnd.f in swiftforth/lib/options/
Changed RND in CHOOSE and added the other usual random things.
Last: 14 June 2020 at 14:14:41 CEST  -rt
==================================================================== *)

/forth
decimal

(( --------------------------------------------------------------------
Over the years I have collected a lot of RNG's. I use them a lot.
All my RNG packages define the same set of words.

Wil Badens RAND is very much faster than RANDOM16 (Paul Mennens) and
R250 (Skip Carter).
Quoting Wil: " Random Number Generators should be written for speed ..."

/RND -- initialise RNG. Add this word to Forth startup sequence.
RANDOM -- return unsigned random number
CHOOSE -- return random number in range n, exclusive.
RANF -- generate a random value from 0.0 to 1.0
ALEA -- ( low high -- low <= n < high ) return random number in range.
TOSS -- like tossing a coin : get 0 or -1
%CUT -- returns true if toss puts you within 0..%, where % is passed n.
50 %cut is 50-50 chance of t/f, 30 %cut is 30% true.

Made to fit deferred VFX random.
-------------------------------------------------------------------- ))

variable bud

: /rnd ( -- )  counter bud ! ;

\ Random in VFX is deferred
: Baden-RANDOM ( -- u )  bud @  3141592621 *  1+  dup bud ! ;

: /RANDOM   /rnd ['] baden-random is random ;

\ CHOOSE is now part of VFX
\ : CHOOSE   ( n -- u )  random um* nip ;

\ --------------------------------------------------------------------
\ Useful additions:

hex 7FFFFFFFFFFFFFFF decimal s>f fconstant max64-as-fp

: RANF ( f: -- x )   random s>f max64-as-fp f/ ;

: ALEA  ( low high - low <= n < high )    over - choose + ;

: TOSS  ( --- flag )   random 0< ;            \ 1 AND doesn't work well with this rng
        
: %CUT ( %-- flag )   100 choose > ;

cr .( Baden Pseudo Random Gen loaded)

/random

\\ ( eof )