(* ====================================================================
ABI call extra's

Copyright 2010 Roelf Toxopeus

VFX64 version
Extra words wrt Mac OSX function call ABI
Last: 27 September 2020 at 16:14:58 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
A fix for Apple's CGRect/NSRect member passing through  C stack.
DFIX -- push dummy zero's on data stack, with the purpose to fill all
parameter registers: 6 of them. Result will be that the next parameters
will be passed via the C stack. Very convenient for the rectangle doubles,
who will be popped from the fp stack and pushed on the C stack.
COCOA.DFIX -- similar to DFIX but takes the hidden receiver in to account.

Example:
FUNCTION: CGContextAddRect ( cgcontext dummy dummy dummy dummy dummy x y x2 y2 -- ret )
: CGContextAddRect ( cgcontext -- ret ) ( F: x y x2 y2 -- )  5 DFIX  CGContextAddRect ;

---
Also need to pass a 64b value as the second return value. Used in method
implementations, ie callbacks.

>2RET -- Pass a second long long (64 bit) in the rdx register.

Example >2RET
FORTHCALL: *textView:willChangeSelectionFromCharacterRange:toCharacterRange: ( rec sel textview loc1 len1 loc2 len2 -- loc3 len3 )
a textView delegate method.

For passing fp's in callbacks see fp-in-callbacks.fth
-------------------------------------------------------------------- ))

/forth
decimal

: DFIX ( n -- n.dummies x y w h )  ( F: x y w h -- )
   0 ?do 0 loop 4fpushdf ;

: COCOA.DFIX ( hidden:receiver n -- n.dummies x y w h hidden:receiver ) ( F: x y w h -- )
   swap >r dfix r> ;

code >2RET ( n -- )
   mov  rdx , rbx
   mov  rbx , 0 [rbp]
   lea  rbp , 8 [rbp]
   next,
end-code

\\ ( eof )