(* ===============================================================================
fp in callbacks

copyright 2020 Roelf Toxopeus

VFX64 specific version.
Adds fp passing in and out handling for callbacks.
Last: 20 October 2020 at 14:27:14 CEST  -rt
=============================================================================== *)

0 [IF] ===========================================================================
Callbacks are set up for passed ints, fp's aren't included.
In 32b intel, that's not a problem. fp's are passed through the C-stack anyway,
so you just add them to count of #ins for the callback.
Similar to the way parameters are passed for externals.

But in 64b Intel, they are passed via the XMM registers
How to do fp in coco-vfx64 for now?
Derived from coco-i4-64 example:

\ Pass the fp's through the data stack
FORTHCALL: *testFP: ( rec sel x -- ret )   ( finit) >fs 2drop fdup f* fs> ;

NSObject NEW.CLASS Foo
: testFPTYPES   0" v@:d" ;
*testFP: testFPTYPES 0" testFP:" Foo ADD.METHOD
Foo ADD.CLASS

COCOA: testFP: ( x -- x2 )
: @testFP: ( F: x1 -- x2 ) fs> swap testFP: >fs ;   \ swap x1 with receiver

2e0 Foo @alloc @init @testFP: f.   \ expect 4e0

This works, but is a bit elaborate, this should be simplified.

---
Consider a callback of the form
FORTHCALL: *testFP: ( rec sel -- ) ( F: x1 -- x2 )   2drop fdup f* ;
As is this will only initiate the data stack, the fp hint isignored.
What happens on fp stack? This will probably crash.
Let's see

Use *testFP: as method implementation
NSObject NEW.CLASS Foo
: testFPTYPES   0" d@:d" ;
*testFP: testFPTYPES 0" testFP:" Foo ADD.METHOD
Foo ADD.CLASS

Now we can add testFP: to our method calling mechanism
COCOA: @testFP: ( x -- x2 ) ( F: x1 -- x2 )
The fp hint is not ignored and the code generator will use this.
When run, the fp on the fpstack is passed via xmm0 to the method.
It's also expected that the returned fp is in xmm0.

So when *testFP: is executed, the fp is in xmm0, it just doesn't know it yet.
We need to move it to xmm8 and tell fdepth

code XMM0> ( -- ) ( F: -- x )
   sub	r13, # FPCELL
   movsd	0 [r13], xmm8
   movsd xmm8, xmm0
   ret
end-code

We also need to pass the fp in xmm0 when done
code >XMM0 ( -- ) ( F: x -- )
   movsd xmm0, xmm8
   movsd	xmm8, 0 [r13]
   add	r13, # FPCELL
   ret
end-code

Now *testFP: can be rewritten to use fp
FORTHCALL: *testFP: ( rec sel -- ) ( F: x1 -- x2 )   XMM0>  2drop fdup f* >XMM0 ;

NSObject NEW.CLASS Foo
: testFPTYPES   0" d@:d" ;
*Log: testFPTYPES 0" testFP:" Foo ADD.METHOD
Foo ADD.CLASS

COCOA: @testFP: ( -- ) ( F: x1 -- x2 )

\ test
Foo @alloc @init value myFoo
2e0 myFoo @testFP: f.     \ expect 4e0

It's trivial to add words to access and pass more fp's in callbacks this way.
Could the FORTHCALL: mechanism be streamlined to compile these words automatically?
Probably, but I will not pursue this.
Better to have the vendors fixing this in their under the hood callback implementations.
Until then, this will suffice I hope.
=========================================================================== [THEN]

(( -------------------------------------------------------------------------------
fp's are passed via the appropriate XMM registers.
keeping the > for 'to' and 'from' notation, clearer for balancing if needed.
in coco-i4-64 FPOPD and FPUSHD are used.

the 2fp parameters are often size:width-height location:x-y etc.
we need 4fp words as well for color:r-g-b-a

rectangle structures when taken, are passed via the C-stack, and must be pushed
on the Forth data stack.
make sure the stackpicture for callback has the dummies like used in COCOA: and
FUNCTION: i.e. fill up to 6 to force
FORTHCALL: *myRect: ( rec sel n3 n4 n5 n6 x y w h -- )   deal-with-rect drop-others ;
has to be tested!

--

XMM>  -- one fp in XMM0 pushed on fp stack
>XMM  -- one fp popped from fp stack in XMM0

2XMM> -- two fp's in XMM0 and XMM1 pushed on fp stack
2>XMM -- two fp's popped from fp stack in XMM0 and XMM1

4XMM> -- four fp's in XMM0 XMM1 XMM2 and XMM3 pushed on fp stack
4>XMM -- four fp's popped from fp stack in XMM0 XMM1 XMM2 and XMM3
------------------------------------------------------------------------------- ))

/FORTH
DECIMAL

code XMM> ( -- ) ( F: -- x )
   sub	r13, # FPCELL
   movsd	0 [r13], xmm8
   movsd xmm8, xmm0
   ret
end-code

code >XMM ( -- ) ( F: x -- )
   movsd xmm0, xmm8
   movsd	xmm8, 0 [r13]
   add	r13, # FPCELL
   ret
end-code

code 2XMM> ( -- ) ( F: -- x1 x2 )
   sub	r13, # FPCELL 2 *
   movsd	8 [r13], xmm8
   movsd 0 [r13], xmm1
   movsd xmm8, xmm0
   ret
end-code

code 2>XMM ( -- ) ( F: x1 x2 -- )
   movsd xmm0, xmm8
   movsd	xmm1, 0 [r13]
   movsd xmm8, 8 [r13]
   add	r13, # FPCELL 2 *
   ret
end-code

code 4XMM> ( -- ) ( F: -- x1 x2 x3 x4 )
   sub	r13, # FPCELL 4 *
   movsd	#24 [r13], xmm8
   movsd #16 [r13], xmm3
   movsd 8 [r13], xmm2   
   movsd 0 [r13], xmm1
   movsd xmm8, xmm0
   ret
end-code

code 4>XMM ( -- ) ( F: x1 x2 x3 x4 -- )
   movsd xmm0, xmm8
   movsd	xmm1, 0 [r13]
   movsd xmm2, 8 [r13]
   movsd xmm3, #16 [r13]
   movsd xmm8, #24 [r13]
   add	r13, # FPCELL 4 *
   ret
end-code

\\

\ example
FORTHCALL: *testFP: ( rec sel -- ) ( F: x1 x2 -- x3 x4 )   2XMM>  2drop fswap 2>XMM ;
NSObject NEW.CLASS Foo
: testFPTYPES ( -- za )   0" dd@:dd" ;
*testFP: testFPTYPES z" testFP:" Foo ADD.METHOD
Foo ADD.CLASS

COCOA: @testFP: ( -- ) ( F: x1 x2 -- x3 x4 )

\ test
Foo @alloc @init value myFoo

1e0 2e0 f.s             \ expect  1.0000000 2.0000000 -- ok F:-2 
myFoo @testFP: f.s      \ expect  2.0000000 1.0000000 -- ok F:-2 

\ testing callback with 10 parameters

defer paint
:noname ; is paint

\ rect structure is past via c-stack, 0's are dummies
forthcall: *draw ( rec sel 0 0 0 0 x y w h -- ret )   paint 0 ;

cocoaclass NSView
NSView new.class DrawView
: drawtypes   0" i@:" ;
*draw drawtypes 0" drawRect:" DrawView add.method
DrawView add.class

cocoa: initWithFrame: ( dummy dummy dummy dummy x y w h -- ref )
: @initWithFrame: ( -- ret ( F: x y w h -- ) 4 cocoa.dfix initWithFrame: ;
cocoa: @display ( -- ret )

: /gport ( win -- )
   dup content.frame
   DrawView @alloc @initWithFrame:
   tuck swap make.contentview
   @display drop ;

new.window win
win add.window
win /gport

\ test the callback:
function: class_getName ( id -- name )
function: object_getClass ( object -- class )
function: sel_getName ( sel -- *name )

: .what ( rec sel dummy dummy dummy dummy x y w h -- )
   cr depth .
   space 2swap swap >fs f. >fs f. swap >fs f. >fs f.       \ the rectangle 
   2drop 2drop                                             \ get rid of dummies
   sel_getName space zcount type                           \ the selector
   object_getClass class_getName space zcount type ;       \ the receiver

' .what is paint

\ resize the window