(* ====================================================================
integer and fp precission conversion

Copyright 2002 Roelf Toxopeus

VFX64 version.
A zoo of conversion words.
Comes from MacForth version, which uses PPC instructions.
Converting between single precision (SF) and native format precision (F) fp.
Many OS calls need and/or return single precision fp. These 32bit SF's
are passed via the data stack.
Also contains the integer <=> fp formats conversions.
Last: 18 August 2020 at 20:19:34 CEST   -rt
==================================================================== *)

/FORTH
DECIMAL

(( --------------------------------------------------------------------
Only for compatability with coco-i4-64 current version.

Integer precission
On the way out:
For explicit 32b handling: H@ H! H,  ( half cell sized memory operators )
Buffer/memory allocation and indexing for 32b items: HCELL HCELLS HCELL+

In the kernel:
explicit 16b handling: W@ W! W,         ( "word" sized memory operators )

The agreed upon versions for 32b handling in the kernel:
L@ L! L,
Defined here:
Buffer/memory allocation and indexing for 32b items:
LCELL LCELLS LCELL+
-------------------------------------------------------------------- ))

(*
: H@ ( a -- n )   l@ ;

: H! ( n a -- )   l! ;

: H, ( x -- )     l, ;

CELL 2/ CONSTANT HCELL

: HCELLS ( n -- n*hcell)   2 LSHIFT ;

: HCELL+ ( n -- n+hcell)   HCELL + ;
*)

CELL 2/ CONSTANT LCELL

: LCELLS ( n -- n*hcell)   2 LSHIFT ;

: LCELL+ ( n -- n+hcell)   LCELL + ;

(* --------------------------------------------------------------------
Floating point precission
These are meant for passing to library/OS calls.
VFX64 passes fp via registers, not via stack. So the words for passing
a variety of fp via the stack as in 32bit Intel Forths, are not needed. 
Exception is Apple's dealing with passing rectangles: x y w h
These go via the stack, see DFIX

FPUSHDF -- returns 64bit precission fp on data stack. Unlike FPUSHD which
returns a 2 cell something: 64bitfp 0
4FPUSHDF -- pushes 4 fp's as 64bitfp's on data stack.

In case we run in to trouble, some aliasses.
Also handy to (temporarely) move a fp away: FPUSH and to retrieve a
(temporarely moved) fp from the data stack: FPOP.
F>SF -- synonym for FPUSHS
SF>F -- synonym for FPOPS
F>DF -- synonym for FPUSHDF
DF>F -- synonym for FPOPD
-------------------------------------------------------------------- *)

synonym FPUSHDF fs>   ( -- f64 ) ( F: f64 -- )
synonym FPOPDF >fs   ( f64 -- ) ( F: -- f64 )

\ : FPUSHDF ( -- u ) ( F: r -- )    0 sp@ df! ;

\ : 4FPUSHDF ( -- dfx dfy dfw dfh ) ( F: x y w h -- )  
\  fpushdf >r fpushdf >r fpushdf fpushdf swap r> r> ;

\ code defenition like fs> would be great
: 4FPUSHDF ( -- dfx dfy dfw dfh ) ( F: x y w h -- )
   0 0 0 0 sp@ dup df! cell+ dup df! cell+ dup df! cell+ df! ;

(*
: F>SF ( -- 32float ) ( F: r -- )  FPUSHS ;

: SF>F ( 32float -- ) ( F: -- r )  FPOPS ;

: F>DF ( -- 64float ) ( F: r -- )  FPUSHDF ;

: DF>F ( 64float -- ) ( F: -- r )  FPOPD ;
*)

\ --- not necesarily a better name, but in use a lot in the demo's

: D>R ( F: x1 -- x2 )  deg>rad ;

\\  ( eof )