(* ====================================================================
petwords for porting harness

copyright 1988 Roelf Toxopeus

VFX64 version.
Much used words acumulated over the years using several Forth systems.
Originates in Mach2 -> Power/Carbon MacForth -> iForth -> SwiftForth -> VFX
Some are favourites, some are just system specific often found in source
files. Some will be replaced by nicer versions found in new systems,
while porting. Finding new favourites continues...
Last: 8 August 2020 at 10:04:08 CEST   -rt
==================================================================== *)

0 [IF] --------------------------------------------------------------------
List of pets
NoTE: some are commented out, because they're already in used Forth.
Others are superceded by better ones, but left for portability.

ANEW -- If next word found, execute it, assuming it's a marker. Then
re-create the marker. Effectively it's an persisting marker. 

LACKING -- If next word in inputstream exists, ignore rest of line.
Else continue interpretting line. Kind of conditional compiling.

DEFAULT-ORDER -- Set default context and wordlist for current.

/FORTH -- Set default context and wordlist for current.

PRIOR.STREAM -- Stop interpreting a file. 

\\ -- Stop interpreting a file.

COMPILING -- Leaves a flag, true when compiling. In SwiftForth use
STATE @ !!!!

ASCII - State smart version of CHAR and [CHAR], *but* capable of
leaving a 4 chars string, used a lot as id on the Mac.

ALLOTERASE -- Allot and erase memory in one go.

/ALLOT -- Allot and erase memory in one go.

BUFFER: -- Create word pointing to allotted memory.

RECURSIVE -- Used to make current word under construction visible.
Allows for simple BEGIN AGAIN loops, assuming tail optimization. 
Also allows for getting the xt from current word under construction,
calling ['] on word etc. All in all superior over RECURSE.
Some systems know this word as REVEAL.
NoTE: Charles Moore removed the smudge stuff all together.

QUERY -- Query for input, use with INTERPRET, very handy while debugging.
Very much preferred over REFILL as name.

0" -- Return zero terminated string (no count field). In some systems
known as Z"

POCKET -- Transient area for string manupilation.

PARSE>POCKET -- Parse text and move as 0string or zstring in pocket.
Transient area, use immediately.

NEXTWORD -- Scan for next word over line brakes till end of input.
Unlike Baden's NEXT-WORD, it returns a counted string like WORD.

SKIPS -- Skip text over string pointed by addr n. Allows other
'skippers' than those found in currently used Forth. Not case sensitive!

{ is used by many Forth systems for locals, so 2 preferred replacements:
(* -- Skip text till *)  MacForth, iForth, VFX
(( -- Skip text till ))  VFX, MFonVFX

LCOUNT -- Used as COUNT but for counted strings with a CELL size
count field.

CLS -- Clear screen.

SIFTING -- Print words containing pattern from next parsed word.
SIFTING DUP, will print all words containing DUP

.. -- Print what ever on the data and fp stacks, clearing them in
the process.
NoTE: lifo order left to right, unlike .S

.B -- Temporary change base to binary and print number.
Easy for eyeballing masks.

.H -- Temporary change base to hex and print number.    
Prefer over often found H.

.D -- Temporary change base to decimal and print number.
Obviously can't use D.

.BASE -- Print base in decimal.

?PAUSE -- Like KEY?, but when a key is hit, it pauses waiting
for another key. All keys except spacebar leave a true flag. 

MANY -- Looping in interpreter until a key hit. Uses ?PAUSE.

SNAP -- Suspend program, showing stack. Waits for a keypres to
continue. Q aborts.
Alternative: : SNAP   .s  key 27 = if cr prompt quit then ;                            
Example: : tt    begin random random random snap drop drop drop again ;      

SNAP" -- SNAP with message.

G -- Jump to line in file which caused an abort during an INCLUDE

STARTER -- Startup word in turnkey

EMPTY -- Wipe dictionary from user added words down till FENCE
-------------------------------------------------------------------- [THEN]

only forth also definitions
decimal

\ --------------------------------------------------------------------
\ --- used during evaluating file contents:

(( --------------------------------------------------------------------
  -- left over CarbonMacForth
  -- no replacement, phase out
: ANEW ( spaces<name> -- )
   >in @ >r bl word find if execute else drop then r> >in ! marker ;
-------------------------------------------------------------------- ))

: LACKING ( spaces<name> -- )   bl word find nip if [compile] \ then ;

(( --------------------------------------------------------------------
  -- left over CarbonMacForth
  -- replace with /FORTH
: DEFAULT-ORDER    only forth also definitions ;
-------------------------------------------------------------------- ))

: /FORTH   ( -- )   only forth definitions ;

(( --------------------------------------------------------------------
  -- left over CarbonMacForth
  -- replace with \\
: PRIOR.STREAM      \\ ;
-------------------------------------------------------------------- ))

\ \\ exists in VFX

\ debug while file including
: VERBOSE ( -- )   debug1 on ;

: SILENT ( -- )   debug1 off ;

\ --------------------------------------------------------------------
\ --- simple redefinitions and oneliners:

: COUNTER ( -- ms )   ms@  ;

: TIMER   ( ms -- )   ms@ swap - . ;

\ VFX EXPIRED does a PAUSE
: EXPIRED ( ms -- flag )  timedout? ;

\ INTERPRET exists in VFX

\ COMPILING exists in VFX

: ASCII ( spaces<chars> -- n )
   bl parse ?dup
   if over c@ swap 1- 3 min 0 ?do
      8 lshift over i 1+ + c@ + loop
      nip
      state @ if postpone literal then
   else true abort" ASCII NO Characters!" then ; immediate


: /ALLOT   ( n -- )   allot&erase ;

(( --------------------------------------------------------------------
  -- left over CarbonMacForth
  -- replace with /ALLOT
: ALLOTERASE ( n -- )  /allot ;
-------------------------------------------------------------------- ))

: @EXECUTE ( xt -- ??? )   @ ?dup if  execute  then ;

\ BUFFER: exists in VFX

: RECURSIVE ( -- )  reveal ; immediate

\ QUERY exists in VFX

\ --------------------------------------------------------------------
\ --- string stuff:

\ using Synonym or Alias: is the preferred way because of possible
\ extra compilation additions which redefing with coclon won't grasp.
Synonym 0" z"

\ large pocket for VFX
1024 buffer: POCKET   \ -- addr

\ temporary 0string or zstring in pocket. transient area, use immediately.
: PARSE>POCKET ( char -- )
   parse 1023 min pocket zplace ;

: NEXTWORD ( -- c-addr)   bl get-word ;

\ : SKIPS ( a n -- )   ParseUntil ;

: SKIPS ( a n -- )
         begin  nextword count dup
      while  2over istr= if  2drop exit  then
   repeat 2drop 2drop ;   \ <= eof, drop pattern as well!

\ (* exists in VFX

\ (( exists in VFX

: LCOUNT ( a -- a2 n )    dup @ swap cell+ ;

\ --------------------------------------------------------------------
\ --- user interface:

\ CLS exists in VFX

: SIFTING ( <string> -- )   sim ;

: ..  ( whatever on the stacks )
   cr depth 0 ?do . loop
   cr fdepth 0 ?do f. loop ;

: .B ( n -- )   base @ swap binary u. base ! ;

: .H ( n -- )   base @ swap hex u. base ! ;

: .D ( n -- )   base @ swap decimal u. base ! ;

: .BASE ( -- )   ." Base (in decimal): " base @ .d ;

: ?PAUSE ( -- f )
   key? dup if                     \ any key action?
      drop key drop               \ drop flag and pressed key
      cr ." Spacebar to Continue" cr
      key bl <> then ;            \ wait for key action and flag it

: MANY   ( -- )   ?pause 0= if 0 >in ! then ;

: SNAP ( -- )  .s  cr ." any key continues, q aborts" key [char] q = abort" aborts ..." ;

: SNAP" ( <string"> -- )   postpone cr postpone ." postpone snap ; immediate

\ G not needed in VFX, LOCATE and errors use the editor directly

: STARTER ( <name> -- )   ' to-do EntryPoint ;

\ EMPTY exists in VFX

cr .( Pets and port harness loaded)

\\  ( eof )