(* ====================================================================
Multi line string

VFX64 version

copyright 2018 Roelf Toxopeus

Last: 10 August 2020 at 09:01:32 CEST
==================================================================== *)

(( --------------------------------------------------------------------
PARSE-LINES -- parses text using passed delimiter over multiple lines.
               Returns char delimited string pair in transient area.
            
LINES, -- compiles parsed delimited text as a zero terminated counted
          string. Returns string pair ( -- a n ) at runtime.
          
S` -- parses ` delimited string over multiple lines. Depending on STATE,
      compiles it or returns it as a string pair in transient syspad.

Easy to create other delimiter versions. Whatever is appropriate:
: S~ ( "ccc"<"> -- )   [char] ~  postpone lines, ; immediate

: S| ( "ccc"<"> -- )   [char] |  postpone lines, ; immediate

: S" ( "ccc"<"> -- )   [char] "  postpone lines, ; immediate

etc.

Useful for Applescript where linefeeds are crucial.

: "test" ( -- za )
   S` tell application "Terminal"
   activate
   do script with command "ls;sleep 4;exit"
   end tell`
   DROP ;

: TT ( -- )   "test" doscript ;

Mach2 and MacForth had S" C" and 0" implemented for multiple lines.
-------------------------------------------------------------------- ))

/forth
decimal

(*
: LINES, ( delimiter "ccc"<"> -- )
   state @ 0= abort" Compile Only !"
   postpone (s")                                      \ runtime for string
   here >r                                            \ construct counted string here
   s" " r@ place                                      \ start with empty string
   begin 
      dup dup parse                                   \ parse till delimiter or end of line
      2dup r@ append                                  \ whatever we got, append to string
      + c@ <>                                         \ if the delimiter is parsed we're done
   while
      eol$ count r@ append                            \ not a delimiter so a linefeed is added to string
      source-id 0= if cr then                         \ at the keyboard, do a cr
      refill 0= abort" Unexpected End Of String !"    \ refill next line or abort 
   repeat
   drop                                               \ drop delimiter
   r> count nip 1+ allot                              \ allot memory for string
   0 c,                                               \ zero terminated
   align ; immediate                                  \ VFX (s") expects aligned string
   
: S` ( "ccc"<"> -- )   [char] `  postpone lines, ; immediate
*)

: PARSE-LINES ( delimiter "ccc"<char> -- a n )
   here >r                                            \ construct counted string here
   s" " r@ place                                      \ start with empty string
   begin 
      dup dup parse                                   \ parse till delimiter or end of line
      2dup r@ append                                  \ whatever we got, append to string
      + c@ <>                                         \ if the delimiter is parsed we're done
   while
      eol$ count r@ append                            \ not a delimiter so a linefeed is added to string
      source-id 0= if cr then                         \ at the keyboard, do a cr
      refill 0= abort" Unexpected End Of String !"    \ refill next line or abort 
   repeat
   drop                                               \ drop delimiter
   r> count ;                                         \ return string pair

: LINES, ( delimiter "ccc"<char> -- )
   postpone (s")                                      \ runtime for string
   parse-lines                                        \ parse delimiter over lines
   nip 1+ allot                                       \ allot memory for string
   0 c,                                               \ zero terminated
   align ;  ( immediate)                              \ VFX (s") expects aligned string

: S` ( "ccc"<`> -- )
   [char] ` state @ if
      optimising? >r unoptimised                      \ VFX64 code generator fix
      ( postpone) lines,
      r> if optimised then                            \ VFX64 code generator fix
      exit then
   parse-lines >syspad ; immediate

\\

\ compiled:
: ll ( -- a n )
   s` this is a
string, which has
more lines` ;

ll type this is a
string, which has
more lines ok

\ compile test:
: tt ll ;                 \ crashes without 10 aug 2020 fix

\ interpreted:
s` this is                                       
   a formatted 
   string over 
   five lines                                    
ending here` cr type                             
this is
   a formatted
   string over
   five lines
ending here ok
