\ PDFhelp.fth - implements HELP <wordname> for VFX Forths

((
Copyright (c) 2008, 2011, 2012, 2013, 2018
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20180828 SFP004 Updated documentation.
20130501 SFP003 Exposed DEBUGHELP? for all O/S.
20121119 SFP002 Allowed for FULLPATHNAME.
20110126 SFP001 OSX changes.
))

only forth definitions
decimal

\ ========
\ *> tools
\ *S PDF help system
\ ========
\ *P MPE documentation is produced by using *\i{DocGen} to produce
\ ** an indexed PDF file. The PDF help system parses the index file
\ ** produced by *\i{pdftex} to display the relevant page of the PDF
\ ** manual. To display a particular line in a PDF file requires the
\ ** following incantation for Adobe Reader v7 and beyond:
\ *C   <reader> /A "page=n=OpenActions" "<pdffile>"
\ *P The page number is the PDF file page number, not the page number
\ ** in the document section. For example, to display page 10 on a
\ ** Windows PC, use:
\ *C   "C:\Program Files\Adobe\Reader 8.0\Reader\AcroRd32.exe"
\ *C   /A "page=10=OpenActions" "%h%.pdf"
\ *C   "C:\Products\VfxCommunity\Sources\Manual\PDFs\VfxWin.pdf"
\ *P For VFX Forth for Windows, you can use the menu item
\ ** *\i{Option -> Set PDF help ...} for the configuration.
\ ** For all versions, when setting the base of the PDF file
\ ** name, do *\b{not} add the file extension. This is because
\ ** the PDF file and the index file share the same base name.

\ *P The page number is extracted from the index file *\i{VfxWin.vix},
\ ** from which this example comes:
\ *E \initial {A}
\ ** \entry {\code {abell}}{11}
\ ** \entry {\code {abl}}{12}
\ ** \entry {\code {abort}}{15, 200}
\ ** \entry {\code {abort"}}{200}
\ *P The file is parsed for the entry containing the word name, the
\ ** page number is extracted, and the file page is displayed. The
\ ** index file is derived from the *\i{.fns} file produced by
\ ** *\i{pdftex}.

\ *P The source code is in *\i{Lib\PDFhelp.fth}.

-sin

TextMacro: p	\ -- $text
\ *G Defines the page number macro *\i{p}.
TextMacro: h	\ -- $text
\ *G Define the help file macro *\i{h}.

#256 constant /Help$	\ -- n
\ *G Size of the command and base string buffers.
/Help$ buffer: HelpCmd$	\ -- addr
\ *G Holds the pathname and command line of the PDF viewer as a
\ ** counted string. In the command line, the page number is
\ ** supplied by the text macro *\i{%p%}, and the base help file
\ ** path/name with no extension is supplied by the text macro
\ ** *\i{%h%}. This string may include other macros.

\ *P For Acrobat Reader under Windows, we must use the full reader
\ ** pathname; we cannot use an association. The default string
\ ** is
\ *C   "<reader>" /A \qpage=%p%=OpenActions\q %h%.pdf
\ *P where *\i{<reader>} is
\ *C   "C:\Program Files\Adobe\Reader 8.0\Reader\AcroRd32.exe"
\ *P If you change the settings, use *\fo{S\"} as you may need to
\ ** have double-quotes characters in the string around path names
\ ** that include spaces.

\ *P One alternative is the free PDF-XChange Viewer from
\ *C   http://www.tracker-software.com/product/pdf-xchange-viewer
\ *C  "<path>\PDFXCview.exe" /A "page=%p%" "%h%.pdf"

\ *P Another alternative is the free Foxit Reader from
\ *C   http://www.foxitsoftware.com/
\ *P The required command strings for *\i{Foxit Reader} are
\ *C "<path>\Foxit Reader.exe" /A "page=%p%" "%h%.pdf"
\ *P The rules for the *\i{Foxit Reader} command line need to be
\ ** checked with every new major release!

\ *P Users have also suggested Nitro, Qpdfview and Sumatra among
\ ** many others. MPE has no position on this - it's a matter of
\ ** personal preference. Just check the viewer's manual for the
\ ** command line incantation!

\ *P For some Linux systems, e.g. Ubuntu, *\i{xpdf} is installed
\ ** by default. The VFX defaults are
\ *C   s\" xpdf %h%.pdf %p% &" HelpCmd$ place
\ *C   s" ~/VfxLin<ver>/Doc/VfxLin" HelpBase$ place
\ *C   #17 HelpPage0 !
\ *P If you used the default installer script, replace <ver> by
\ ** by one of Eval, Standard or Pro, e.g. VfxLinStandard.
\ *P Some Linux distributions, e.g. Debian, require shared
\ ** documentation files to be compressed. Type:
\ *C   sudo find / -name "VfxLin.pdf"
\ *P To see where VfxLin.pdf has been installed, and to see what extension,
\ ** e.g. *\i{.pdf.gz} is in use.

\ *P For OS X, the default PDF viewer is *\i{Preview}. It can be
\ ** run from the command line using:
\ *C  open -a Preview filename.pdf
\ *P However, going to a page number is undocumented. The best
\ ** solution we have found is to install the *\i{Skim} package
\ ** from:
\ *C   http://skim-app.sourceforge.net/
\ *P *\i{Skim} can be run using the supplied executable script
\ ** *\i{Bin/skimpage.scpt}. This is run in the form:
\ *C   skimpage.scpt "<file>" <pageno>
\ *P After copying the script file to a suitable directory such
\ ** as *\i{/usr/bin} (done by the install script), a suitable setup is:
\ *C   s\" skimpage.scpt \q%h%.pdf\q %p%" HelpCmd$ place
\ *C   s" ~/VfxForth/Doc/VfxOsx" HelpBase$ place
\ *C   #14 HelpPage0 !


/Help$ buffer: HelpBase$	\ -- addr
\ *G Contains a full path to the directory containing the help and
\ ** index files, plus the base file name with no extension. This
\ ** string may include macros.
\ ** A counted string, e.g for Windows
\ *C   %LOAD_PATH%\..\doc\VfxMan
\ *P and for Linux
\ *C   %LOAD_PATH%/../doc/VfxLin
\ *P and for OS X
\ *C   %LOAD_PATH%/../doc/VfxOsx
variable HelpPage0	\ -- addr
\ *G Holds the offset to be added to the index page number to
\ ** convert it to a PDF page number. This value may change
\ ** according to the manual version, so it is extracted from
\ ** the index file.
  #20 HelpPage0 !

[defined] Target_386_Windows [if]
  s\" \qC:\\Program Files\\Adobe\\Reader 9.0\\Reader\\AcroRd32.exe\q /A \qpage=%p%=OpenActions\q %h%.pdf" HelpCmd$ place
  s" %LOAD_PATH%\..\doc\VfxMan" HelpBase$ place
  #20 HelpPage0 !
[then]
[defined] Target_386_Linux [if]
  s\" xpdf %h%.pdf %p% &" HelpCmd$ place
  s" %LOAD_PATH%/../doc/VfxLin" HelpBase$ place
  #18 HelpPage0 !
[then]
[defined] Target_ARM_Linux [if]
  s\" xpdf %h%.pdf %p% &" HelpCmd$ place
  s" %LOAD_PATH%/../doc/VfxLin" HelpBase$ place
  #18 HelpPage0 !
[then]
[defined] Target_386_OSX [if]
  s\" skimpage.scpt \q%h%.pdf\q %p%" HelpCmd$ place
  s" /Users/sfp/Products/VfxCommunity/Sources/Manual/PDFs/VfxOsx" HelpBase$ place
  #14 HelpPage0 !
[then]
[defined] Target_x64_OSX [if]
  s\" skimpage.scpt \q%h%.pdf\q %p%" HelpCmd$ place
  s" /Users/sfp/Products/VfxCommunity/Sources/Manual/PDFs/VfxOsx64" HelpBase$ place
  #14 HelpPage0 !
[then]
[defined] Target_x64_Linux [if]
  s\" xpdf %h%.pdf %p% &" HelpCmd$ place
  s" %LOAD_PATH%/../doc/VfxLin" HelpBase$ place
  #18 HelpPage0 !
[then]


also system definitions

#256 buffer: iWordName	\ -- addr
\ +G Holds the word name as a counted string. Depends on the VFX
\ +* Forth max name size.
variable iPage#		\ -- addr
\ +G Holds the page number if found.
-1 value hIndex	\ -- handle
\ +G Holds index file handle.

create IndexMarker	\ -- addr
\ +G Holds counted string of "\entry {\code {".
  ", \entry {\code {"

#512 2+ buffer: LineBuff	\ -- addr
\ +G Line buffer for the current line.
0 value /Line	\ -- n
\ +G Size of the line last read.

: OpenIndex	\ --
\ +G Open the index file.
  { | fn[ MAX_PATH ] -- }
  -1 to hIndex
  HelpBase$ count fn[ place  s" .vix" fn[ append
\ cr fn[ count type
  fn[ count r/o open-file if
    cr fn[ count type cr
    -1 abort" Cannot open HELP index file"
  then
  to hIndex
;

: CloseIndex	\ --
\ +G Close the index file.
  hIndex close-file drop
;

: NextLine	\ --
\ +G Find the next line that starts *\fo{\entry}. A *\fo{THROW}
\ +* occurs if there is an error or end of file.
\ : read-line     \ c-addr u1 fileid -- u2 flag ior
\ bad  = ior<>0
\ EOF  = u2=0, flag=0, ior=0
\ good = u2=0..z, flag=true, ior=0
  LineBuff #512 hIndex read-line throw 0= throw to /Line
;

: IndexLine?	\ -- n
\ +G Return true if the last line read is an index line.
  IndexMarker c@ /Line >
  if  0 exit  endif
  IndexMarker count LineBuff swap is=
;

: NextIndex	\ --
\ +G Step to the next index line.
  begin  NextLine IndexLine? until
;

[undefined] csplit [if]
: csplit	\ caddr len char -- raddr rlen laddr llen
\ +G Extract a substring at the start of caddr/len, returning
\ +* the string raddr/rlen which includes char (if found) and
\ +* the string laddr/llen which contains the text to left of char.
\ +* If the string does not contain the character, raddr is
\ +* addr+len and rlen=0.
  >r 2dup r> scan                       \ -- addr len addr' len'
  2swap 2 pick -                        \ -- addr' len' addr len-len'
;
[then]

: ExtractIndex	\ --
\ +G Extract the word name and page number from the current
\ +* index line.
  LineBuff /Line IndexMarker c@ /string	\ string after index marker
  [char] } csplit -leading -trailing	\ word name
  iWordName place			\ save as word name.
  [char] { csplit 2drop			\ discard up to next '{' }
  dup if  1 /string  endif		\ step over '{'
  -leading #10 extractnum iPage# !	\ extract page number
  2drop
;

: ExtractPage	\ --
\ +G The first line is either the entry for the first PDF page
\ +* or an \initial entry.
  NextLine IndexLine? if
    ExtractIndex  iWordName count s" !!firstpage!!" istr= if
      iPage# @ HelpPage0 !
    endif
  endif
;

: (SearchHelp)	\ caddr len --
\ +G Search the help file. A *\fo{THROW} occurs if there is an
\ +* error or end of file. On return the variable *\fo{iPage#}
\ +* holds the required page number.
  ExtractPage
  begin
    NextIndex ExtractIndex
    2dup iWordName count istr=
  until
  2drop
;

: SetHelpPage	\ --
\ +G Set the page number macro.
  { | n$[ #64 ] -- }
  base @ >r decimal
  iPage# @ HelpPage0 @ + 1 max	 	\ PDF page number
  (u.) n$[ place  n$[ c" p" $setmacro	\ set P macro
  r> base !
;

[defined] fullPathname [if]
: SetHelpFile	\ --
\ +G Set the help file macro.
  HelpBase$ count fullPathname s" h" replaces
;
[else]
: SetHelpFile	\ --
\ +G Set the help file macro.
  HelpBase$ $ExpandMacros c" h" $setmacro
;
[then]

also forth definitions
0 value DebugHelp?	\ -- flag
\ *G Set this non-zero if you are having trouble setting up the
\ ** help system. The command line will be displayed.
previous definitions

[defined] Target_386_Windows [if]
: LaunchHelp	\ -- ior
\ +G Launch the PDF viewer. Return non-zero if the launch fails.
  SetHelpPage SetHelpFile		\ set the page and file macros
  HelpCmd$ count expand	drop		\ -- zaddr ; expanded string
  DebugHelp?
  if  ." Help command line :" dup .z$ ." :" cr  endif
  SW_SHOW WinExec #32 u<
;
[then]

[defined] Target_386_Linux
[defined] Target_ARM_Linux or
[defined] Target_x64_Linux or
[if]
: LaunchHelp	\ -- ior
  SetHelpPage SetHelpFile		\ set the page and file macros
  HelpCmd$ count expand drop		\ set macros
  DebugHelp?
  if  ." Help command line :" dup .z$ ." :" cr  endif
  >system				\ run expanded string
;
[then]

[defined] Target_386_OSX
[defined] Target_x64_OSX or
[if]
: LaunchHelp	\ -- ior
  SetHelpPage SetHelpFile		\ set the page and file macros
  HelpCmd$ count expand drop		\ set macros
  DebugHelp?
  if  ." Help command line :" dup .z$ ." :" cr  endif
  >system				\ run expanded string
;
[then]

: SearchHelp	\ caddr len --
\ +G Search the open help file. A *\fo{THROW} occurs if there is
\ +* an error or end of file. On return the variable *\fo{iPage#}
\ +* holds the required page number.
  (SearchHelp) LaunchHelp drop
;

only forth definitions  also system

: $Help		\ caddr len -- ior ; 0=success
\ *G Run the help file system using *\i{caddr/len} as the search key.
  OpenIndex
  ['] SearchHelp catch dup
  if  nip nip  endif			\ clean up caddr/len
  CloseIndex
;

: Help		\ "<word>" -- ; e.g. HELP dup
\ *G Get help on the given word name, e.g.
\ *C   help locate
\ *P Unlike *\fo{LOCATE}, *\fo{HELP} does not require the word to
\ ** be present in the Forth dictionary and current search order,
\ ** it only requires that a word have an index entry in the PDF
\ ** manual.
  bl word count $help if
    cr ." Error finding word or finding files"
  endif
;

\ The Windows save is in Studio\XTB.FTH.
[defined] Target_386_Linux
[defined] Target_ARM_Linux or
[if]
: PDFLoadCfg	\ --
\ *G Load the PDF help configuration from the INI file.
\ ** Linux and OS X only.
  s" Options" Ini.Section
  s" HelpCmd" HelpCmd$ /Help$ s" xpdf %h%.pdf %p% &" Ini.ReadStr
  s" HelpBase" HelpBase$ /Help$ s" /usr/share/doc/VfxForth/Doc/VfxLin" Ini.ReadStr
  s" HelpPage0" #18 Ini.ReadInt HelpPage0 !
; ' PDFLoadCfg AtIniLoad

: PDFSaveCfg	\ --
\ *G Load the PDF help configuration from the INI file.
\ ** Linux and OS X only.
  s" Options" Ini.Section
  s" HelpCmd" HelpCmd$ count Ini.WriteStr
  s" HelpBase" HelpBase$ count Ini.WriteStr
  s" HelpPage0" HelpPage0 @ Ini.WriteInt
; ' PDFSaveCfg AtIniSave
[then]

[defined] Target_386_OSX
[defined] Target_x64_OSX or
[if]
: PDFLoadCfg	\ --
  s" Options" Ini.Section
  s" HelpCmd" HelpCmd$ /Help$ s\" skimpage.scpt \q%h%.pdf\q %p%" Ini.ReadStr
  s" HelpBase" HelpBase$ /Help$ s" ~/VfxForth/Doc/VfxOsx" Ini.ReadStr
  s" HelpPage0" #18 Ini.ReadInt HelpPage0 !
; ' PDFLoadCfg AtIniLoad

: PDFSaveCfg	\ --
  s" Options" Ini.Section
  s" HelpCmd" HelpCmd$ count Ini.WriteStr
  s" HelpBase" HelpBase$ count Ini.WriteStr
  s" HelpPage0" HelpPage0 @ Ini.WriteInt
; ' PDFSaveCfg AtIniSave
[then]

previous

+sin


\ ======
\ *> ###
\ ======

((
XP
DWORD WINAPI GetFullPathName(
  __in   LPCTSTR lpFileName,
  __in   DWORD nBufferLength,
  __out  LPTSTR lpBuffer,
  __out  LPTSTR *lpFilePart
);

2000 Pro
BOOL PathFindOnPath(
  __inout   LPTSTR pszFile,
  __in_opt  LPCTSTR *ppszOtherDirs
);
))

(( doesn't go to page
: tpdf		\ --
  NULL NULL
  z" C:\Products\VfxCommunity\Sources\Manual\PDFs\VfxWin.pdf"
  z\" /A \qpage=10=OpenActions\q"
  NULL SW_SHOWNORMAL ShellExecute
;
))
(( doesn't run
: tpdf		\ --
  NULL NULL
  z\" \qC:\\Products\\VfxCommunity\\Sources\\Manual\\PDFs\\VfxWin.pdf\q /A \qpage=10=OpenActions\q"
  NULL NULL SW_SHOWNORMAL ShellExecute
;
))


(( OSX
The following method works with Skim, an open-source replacement
for Preview.app. First, download Skim, then save the following
code on a text file and name it "gotopage.scpt":

#!/usr/bin/osascript
on run argv
    tell application "Skim"
        activate
        open (item 1 of argv)

        tell document 1
            set myPage to (item 2 of argv) as integer
            go to page myPage
        end tell

    end tell
end run
You can now tell Skim to open a certain PDF and go to page 99 by
writing this on the terminal:

osascript gotopage.scpt "/full/path/to/doc/mydoc.pdf" 99

You might want to wrap the above line into an sh script of your
own. Also note that the path to the PDF must be absolute,
otherwise Skim won't find the file.

Answered by Julio Gorg

evince -p 5 foo.pdf

http://www.macports.org
Download and Install macports
Then install evince
  sudo port install evince
The dependency checks take a very long time if you have not
used MacPorts much. Start it off and leave it overnight!
If the install fails (usually p5-locale-gettext) with a list
of uninstalled dependencies, install the dependency that failed
  sudo port install p5-loacale-gettext
and recurse as required. It's incredibly tedious.
))

decimal
