\ BLOCKS.FTH - ANS BLOCK and BLOCK EXTENSION word sets

((
Copyright (c) 2001, 2004
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8033 9691


This file provides the ANS BLOCK facility. By default, eight block
buffers are provided. Before use, the word INITBLOCKS initialises
the buffer system. The size of a block is set at 1024 bytes. The
buffers are linked in a singly linked list. When a new buffer is
requested, the list is searched to see if that block number has
already been loaded. If the block has already been loaded, its
buffer is reused. If the block has not been loaded, the last buffer
in the list is saved if changed, unlinked, and moved to the top
of the list for use by the newly requested block. Thus the end
of the list represents the least recently used (LRU) buffer.

The words LOAD and THRU are provided, but the error handling has
minimal integration with the VFX Forth kernel. If you wish to
enhance this, you will need to provide new versions of the error
message handlers DOABORTMESSAGE and DOERRORMESSAGE.

If you change this file, please copy the changed version to a different
folder so that your copy will not be overwritten when you upgrade
VFX Forth. Words in the SYSTEM vocabulary are subject to change.


To do
=====

Change history
==============
20090618 SFP003 Improved start up
20040326 SFP002 Fixed UPDATE and friends
20010714 SFP001 Original release
))

only forth definitions  decimal

also system definitions

-1 value blkHandle	\ -- handle ; contains block file handle
max_path 1+ buffer: blkFilename	\ -- addr ; holds zstring of screen file name
variable disk-error	\ -- addr ; set true for block read beyond EOF.
#1024 constant b/buf	\ -- 1024 ; size of a block
#64 constant b/line	\ -- 64 ; size of source line in a block
#16 constant lines/b	\ -- 16 number of lines in a block
8 constant #blocks	\ -- n ; number of cached blocks
0 value blockMem	\ -- addr ; holds address of block buffers
0 value blockLink	\ -- addr ; single linked list, most recently used first
0 value blockCurr	\ -- addr ; current block structure ; SFP002

struct /block	\ -- n
  cell field blkNext		\ points to previously used block, -1=end
  cell field blk#		\ -1 for unused
  cell field blkUpdated		\ -1 for updated, 0 = unchanged
  b/buf field blkData		\ data area
end-struct

syserrdef blk_nofile	"BLK: No screen/block file open"
syserrdef blk_blk0	"BLK: Cannot LOAD screen 0"

: initBlockData	\ -- ; cold start initialisation ; SFP003
  -1 to blkHandle  0 to blockMem  disk-error off
;
' initBlockData AtCold

: close-blkfile	\ -- ; close block file
  blkHandle -1 <> if
    blkHandle close-file drop
    -1 -> blkHandle
  endif
;

: initblocks	\ -- ; initialise block buffer system
  blockMem 0= if
    #blocks /block * allocate #-406 ?throw
    -> BlockMem
    0 -> BlockLink  0 -> BlockCurr	\ SFP002
    BlockMem #blocks /block * bounds ?do
      blockLink i blkNext !		\ chain blocks
      i -> BlockLink
      -1 i blk# !  0 i blkUpdated !	\ initialise
    /block +loop
  endif
;
' initblocks AtCold
initblocks

: termblocks	\ -- ; shutdown block buffer system
  blockMem if
    blockMem free drop
    0 -> blockMem  0 -> blockLink
    0 -> blockCurr			\ SFP002
    close-blkfile
  endif
;
' termblocks AtExit

: blk-seek	\ blk --
  blkHandle -1 = blk_nofile ?throw	\ -1 = no screen file
  b/buf um* blkHandle reposition-file	\ seek to position
  dup disk-error !  #-417 ?throw	\ save status
;

: blk-read	\ addr blk -- ; read block to buffer
  blk-seek  b/buf 2dup
  blkHandle read-file #-414 ?throw
  dup 0= disk-error !			\ EOF if nothing read
  /string blank				\ wipe unread space
;

: blk-write	\ addr blk -- ; write block from buffer
  blk-seek
  b/buf blkHandle write-file #-415 ?throw
;

: findBlock	\ blk -- struct|0 ; return 0 if not found
  blocklink
  begin					\ -- blk# struct
    dup
   while				\ -- blk# struct
    2dup blk# @ = if
      nip  exit				\ return struct
    endif
    blkNext @
  repeat
  nip					\ return 0
;

: writeStruct	\ struct -- ; write block to disk
  dup blk# @ -1 =			\ -1 means unused
  if  drop  exit  endif
  dup blkData over blk# @ blk-write
  blkUpdated off
;

: GrabOldest	\ -- struct ; find oldest block, save it, move to top
  blockLink dup				\ -- curr next
  begin
    nip					\ remove curr
    dup blkNext @ dup blkNext @ 0=	\ get new curr and next
  until
  swap blkNext off			\ previous block is now last one
  dup blkUpdated @ if			\ write back if changed
    dup writeStruct
  endif
  blockLink over blkNext !		\ link at top of chain
  dup -> blockLink
;

((
: .blocks	\ -- ; display block buffer data
  blockLink
  begin
    dup
   while
    cr
    ."   Addr: " dup .dword
    ."   Next: " dup blkNext @ .dword
    ."   Blk#: " dup blk# @ .dword
    ."   Upd: "  dup blkUpdated @ .dword
    blkNext @
  repeat
  drop
;
))


\ ***************
\ ANS BLOCK words
\ ***************

also forth definitions

: save-buffers	\ -- ; write out changed buffers
  blocklink
  begin
    dup
   while				\ -- struct
    dup blkUpdated @ if
      dup writeStruct
    endif
    blkNext @
  repeat
  drop
;

: empty-buffers	\ -- ; mark all buffers as free
  blocklink
  begin
    dup
   while				\ -- struct
    -1 over blk# !
    dup blkUpdated off
    blkNext @
  repeat
  drop
  0 -> blockCurr			\ SFP002
;

: block		\ blk -- addr ; return block data address
  dup findblock dup if			\ -- blk struct|0
    nip					\ remove block number
  else
    drop  GrabOldest			\ -- blk struct
    2dup blkData swap blk-read		\ read from file
    tuck blk# !				\ set block number
  endif
  dup -> blockCurr			\ SFP002
  blkData
;

: buffer	\ blk -- addr ; return block data address
  dup findblock dup if			\ -- blk struct|0
    nip					\ remove block number
  else
    drop  GrabOldest			\ -- blk struct
    tuck blk# !				\ set block number
    dup blkData b/buf erase		\ wipe data area
  endif
  dup -> blockCurr			\ SFP002
  blkData
;

: flush		\ -- ; save changed data, mark all buffers as empty
  save-buffers empty-buffers
;

: update	\ -- ; mark current block as changed
  blockCurr ?dup if
    blkUpdated on
  endif
;

: load		\ i*x blk -- j*x ; interpret block blk
  dup 0= blk_blk0 ?throw
  save-input n>r
  dup blk !  block b/buf to-source  0 >in !
  interpret
  nr> restore-input drop
;

: thru		\ i*x u1 u2 -- j*x ; load blocks u1..u2
  1+ swap
  ?do  i load  loop
;

: list		\ n -- ; display source block n
  0 swap				\ initial line #
  dup scr !
  cr ." Screen: " dup .
  block b/buf bounds do
    cr  dup 2 .r ."  | " i b/line type
    1+
  b/line +loop
  drop
;

[-sin		\ source inliner misbehaves when redefinition uses the original!
: \		\ -- ; redefinition to cope with blocks
  blk @ if
    >in @  b/line 1- +  b/line negate and  >in !
  else
    postpone \
  endif
; immediate

: refill	\ -- flag ; redefinition to cope with blocks
  blk @ if
    1 blk +!
    blk @ block b/buf to-source  0 >in !
    true
  else
    refill
  endif
;
sin]


\ ***************
\ Common practice
\ ***************

: -->		\ -- ; LOAD next screen
  blk @ 0= blk_blk0 ?throw
  refill drop
; immediate

: ;s		\ -- ; terminate loading of current screen
  source nip >in !
;


\ *******
\ MPEisms
\ *******

: n		\ -- ; LIST next screen
  scr @ 1+ list
;

: p		\ -- ; LIST previous screen
  scr @ 1- list
;

: l		\ -- ; LIST current screen
  scr @ list
;

: using-file	\ c-addr u -- ; select screen file
  flush  close-blkfile
  blkFileName max_path 1+ erase
  blkFileName swap max_path min dup >r cmove
  blkFileName r> r/w open-file #-413 ?throw
  -> blkHandle
;

: using		\ -- ; USING <filename> ; select screen file
\  bl word count using-file
  parse-name using-file
;

: index		\ u1 u2 -- ; index blocks u1..u2 by showing first lines only
  1+ swap ?do
    i block  cr i 8 .r space b/line type
    halt? disk-error @ or ?leave
  loop
;

: qx		\ -- ; index file
  0 -2 index
;

previous definitions

previous definitions

