\ VfxUtil.fth - auxiliary file for the Forth Scientific Library
\               For VFX Forth 3.7 onwards

((
20041104 MPE005 Arrays now return the head of the array.
20040701 SFP004 Performance hackery to DMARRAY and }}.
20030604 SFP003 Fixed UMD/MOD
20000720 SFP002 Changed i j }} to step by j rather than i as some FSL code
                assumes that rows are contiguous in memory. This also helps
		with caching by rows.
		Modified RESET_SEARCH_ORDER yet again.
))

\                       contains commonly needed definitions.
\ dxor, dor, dand       double xor, or, and
\ sd*                   single * double = double_product
\ v: defines use(       For defining and settting execution vectors
\ %                     Parse next token as a FLOAT
\ S>F  F>S              Conversion between (single) integer and float
\ F,                    Store FLOAT at (aligned) HERE
\ F=                    Test for floating point equality
\ -FROT                 Reverse the effect of FROT
\ F2*  F2/              Multiply and divide float by two
\ F2DUP                 FDUP two floats
\ F2DROP                FDROP two floats
\ INTEGER, DOUBLE, FLOAT   For setting up ARRAY types
\ ARRAY DARRAY              For declaring static and dynamic arrays
\ }                         For getting an ARRAY or DARRAY element address
\ & &!                      For storing ARRAY aliases in a DARRAY
\ PRINT-WIDTH               The number of elements per line for printing arrays
\ }FPRINT                   Print out a given array
\ Matrix                    For declaring a 2-D array
\ }}                        gets a Matrix element address
\ Public: Private: Reset_Search_Order   controls the visibility of words
\ frame unframe             sets up/removes a local variable frame
\ a b c d e f g h           local FVARIABLE values
\ &a &b &c &d &e &f &g &h   local FVARIABLE addresses


\ This code conforms with ANS requiring:
\     	1. The Floating-Point word set

\ This code is released to the public domain, MPE October 1999

CR
CR .( VFXUTIL.FTH v3.71  4 November 2004 )
CR .( This code will break/crash if you use & for function )
CR .( addresses with DEFINES. )
CR


\ **********************
\ *S Compilation control
\ **********************

\ for control of conditional compilation of test code
true VALUE TEST-CODE?	\ -- flag
\ *G Set non-zero to compile test code.

true VALUE ?TEST-CODE	\ -- flag
\ Obsolete, for backward compatibility.

TRUE CONSTANT HAS-MEMORY-WORDS?	\ -- flag
\ *G for control of conditional compilation of Dynamic memory

FALSE CONSTANT DEBUG-ARRAYS?	\ -- flag
\ *G For control of conditional compilation of dereferencing
\ ** unallocated array error


\ ****************
\ FSL NonANS words
\ ****************

: DEFINED	\ c-addr -- t/f ; returns definition status of a word
\ *G returns true if the word represented by the counted string
\ ** c-addr has been defined.
  FIND SWAP DROP
;

: ~DEFINED	\ c-addr -- t/f ; returns definition status of a word
\ *G returns true if the word represented by the counted string
\ ** c-addr has not been defined.
  DEFINED 0=
;

Vocabulary HIDDEN	\ --
variable HIDDEN_SET        HIDDEN_SET Off
variable Private_Used      Private_Used Off
variable OLD_CURRENT       OLD_CURRENT Off

: Public:	\ --
\ *G Words defined after Public: will be in the CURRENT wordlist,
\ ** with words defined after Private: also being in the search order.
  HIDDEN_SET @ IF
    HIDDEN_SET OFF
    PREVIOUS  ALSO HIDDEN
    OLD_CURRENT @ 0= NOT IF
      OLD_CURRENT @ SET-CURRENT
    THEN
  THEN
;

: Private:	\ --
\ *G Words defined after Private: are added to a wordlist which
\ ** is normally unavailable. See Public:.
  HIDDEN_SET @ 0= IF
    HIDDEN_SET ON
    GET-CURRENT OLD_CURRENT !
    Private_Used @ IF PREVIOUS THEN
    ALSO HIDDEN DEFINITIONS
    Private_Used On
  THEN
;

: Reset_Search_Order	\ --
\ *G Remove the wordlist used by Private: from the search order.
  Public:
  Private_Used @ IF PREVIOUS THEN
  Private_Used Off
  OLD_CURRENT Off
;


84 CHARS ( or more ) buffer: fsl-pad	\ -- addr
\ *G This word appears to be unused.

: sd*		\ n d -- d(n*d)
\ *G unsigned single * double = double
\ ** This word appears to be unused.
 2 PICK * >R  UM*  R> +
;

: d*		\ d1 d2 -- d3
\ double * double = double
  3 PICK * >R  TUCK * >R  UM*  R> +  R> +
;

: DU<           \ ud1 ud2 -- t/f
  ROT SWAP 2DUP U<                      \ ud1l ud2l ud1h ud2h f
  IF   2DROP 2DROP  -1
  ELSE  <>
        IF   2DROP  0  ELSE  U<  THEN
  THEN
;

: DU>           \ ud1 ud2 -- t/f
  ROT SWAP 2DUP U>                      \ ud1l ud2l ud1h ud2h f
  IF   2DROP 2DROP  -1
  ELSE  <>
        IF   2DROP  0  ELSE  U>  THEN
  THEN  ;

: umd* 		( multiplicand . multiplier . -- product . . . )
				( ul uh vl vh)
	2 PICK OVER UM* 2>R ( ul uh vl vh R: x3 .)
	3 PICK UM* 2>R 	( ul uh vl R: x3 . x2 .)
	TUCK UM* 2>R 	( ul vl R: x3 . x2 . x1)
	UM* 		( x0 . R: x3 . x2 . x1)
	0 2R> D+ 	( x0 x1 . R: x3 . x2 .)
	2R@ D+
	2DUP 2R> DU< IF 	( x0 x1 . R: x3 .)
		R> 1+ >R
	THEN
	0 2R> D+ 	( x0 x1 x3 .)
;

4 cells buffer: tsumh	\ -- addr ; holds quad word
tsumh 2 cells + constant tsuml		\ tsumh holds high half, tsuml holds low

: umd/mod	\ uq ud -- udrem udquot
\ Coding UMD/MOD can give an enormous increase in speed.
  2>r  tsumh 2! tsuml 2!  2r>		\ save dividend
  #64 0 do				\ 64 times ; SFP003
    \ quad precision shift, preserving carry
    tsumh @ >r				\ save carry from shift
    tsumh 2@ d2*  tsuml @ 0<		\ high portion of quad word
    if  1 0 d+  then			\ add carry from low half ; SFP003
    tsumh 2!
    tsuml 2@ d2* tsuml 2!		\ shift low pair ; SFP003
    \ compare and subtract
    2dup tsumh 2@ du> 0= if		\ if divisor less than top 32 bits
      tsumh 2@ 2over d- tsumh 2!	\ subtract divisor
      1 tsuml cell + +!			\ increment quotient
      r> drop
    else
      r> 0< if				\ test preserved carry
        tsumh 2@ 2over d- tsumh 2!	\ subtract divisor
        1 tsuml cell + +!		\ increment quotient
      endif
    endif
  loop
  2drop					\ discard divisor
  tsumh 2@  tsuml 2@			\ return remainder and quotient
;

: dxor       ( d1 d2 -- d )             \ double xor
	ROT XOR >R XOR R>
;

: dor       ( d1 d2 -- d )              \ double or
	ROT OR >R OR R>
;

: dand     ( d1 d2 -- d )               \ double and
	ROT AND >R AND R>
;

: d>
         2SWAP D<
;


\ size of a regular integer
1 CELLS CONSTANT INTEGER

\ size of a double integer
2 CELLS CONSTANT DOUBLE

\ size of a regular float
1 floats CONSTANT FLOAT

\ size of a pointer (for readability)
1 CELLS CONSTANT POINTER

: %		\ "text" -- ; F: -- f | 
  BL WORD COUNT >FLOAT 0= if
    ShowSourceOnError
    cr true ABORT" NAN"
  then
  STATE @ IF POSTPONE FLITERAL THEN ; IMMEDIATE
                  
: FTUCK		\ F: x y -- y x y
  FSWAP FOVER ;

: F**2   FDUP F*     ;

: ZDUP    FOVER  FOVER  ;

: E.   FS. ;

: F2DROP    FDROP FDROP ;

: FSQR  ( F: r -- r^2 ) FDUP F* ;

: -FROT    FROT FROT ;
\ : F2*   2.0e0 F*     ;
\ : F2/   2.0e0 F/     ;
: F2DUP     FOVER FOVER ;
\ : F2DROP    FDROP FDROP ;
\ : F,   HERE FALIGN 1 FLOATS ALLOT F! ;
\ : F+! ( addr -- ) ( F: r -- )  DUP F@ F+ F! ;

%pi FCONSTANT PI
%pi FCONSTANT F=PI
%pi/2 FCONSTANT PI/2
%1  FCONSTANT F1.0
%1  FCONSTANT F=1
%0  FCONSTANT F=0


\ *****************
\ Array definitions
\ *****************

\ ============
\ Control data
\ ============

0 VALUE TYPE-ID               \ for building structures
FALSE VALUE STRUCT-ARRAY?

\ for dynamically allocating a structure or array
TRUE  VALUE is-static?     \ TRUE for statically allocated structs and arrays

: dynamic ( -- )     FALSE TO is-static? ;


struct /DAhead	\ -- size ; controls arrays up to 3 dimensions
  int DA.^data				\ pointer to data
  int DA./item				\ size of item/element in array
  int DA.#cols				\ c, 0..c-1
  int DA.#rows				\ r, 0..r-1
  int DA.type				\ type info
end-struct


\ =======
\ Helpers
\ =======

[-sin

: InitHead	\ r c /item head --
\ Initialise the array head, but not the type or data pointer.
  dup /DAhead erase			\ -- r c /item head
  tuck DA./item !
  tuck DA.#cols !
  DA.#rows !
;

: USetHead	\ r c /item head --
\ Set up the head for an unstructured array.
  >r
  3dup * * ?dup 0= if
    r> InitHead
  else
    r@ InitHead
    r@ /DAhead + r> DA.^data !
  endif
;

: SSetHead	\ r c /item head --
\ Set up the head for a structured array.
  dup >r USetHead
  type-id r> DA.type !
;

: cBuild	\ r c /item -- addr
\ ALLOT head and data space
  here >r  * * /DAhead + dup allot  r>	\ get space
  dup rot erase				\ wipe it
;

: iBuild	\ r c /item -- addr
\ IRESERVE head and data space in IDATA space.
  * * /DAhead + dup ireserve		\ get space
  dup rot erase				\ wipe it
;

sin]


\ ====================
\ Array defining words
\ ====================

: doMMChild	\ head -- head |
\ The children of all xMMATRIX definitions are immediate.
\ Perform their compilation/interpretation action.
  state @ if
    postpone literal  [o/f]
  then
;

: cMMATRIX	\ -- #r #c /item ; -- head
\ Build 2D unstructured matrix with ALLOT
  create  immediate
    3dup cBuild USetHead
  does>
    doMMChild
;

: iMMATRIX	\ -- #r #c /item ; -- head
\ Build 2D unstructured matrix in IDATA space.
  3dup iBuild dup >r UsetHead r>
  create  immediate
    ,
  does>
    @ doMMChild
;

: MMATRIX	\ -- #r #c /item ; -- head
\ Build 2D unstructured matrix.
  idata?
  if  cMMATRIX  else  iMMATRIX  endif
;


: doSMChild	\ head -- type ^data
\ The children of all xSMATRIX definitions return the type
\ information and the data address (not the head).
  dup DA.type @  swap /DAhead +
;

: cSMATRIX	\ -- #r #c /item ; -- type ^data
\ Build 2D structured matrix with ALLOT
  create
    3dup cBuild SSetHead
  does>
    doSMChild
;

: iSMATRIX	\ -- #r #c /item ; -- type ^data
\ Build 2D structured matrix in IDATA space.
  3dup iBuild dup >r SSetHead r>
  create
    ,
  does>
    @ doSMChild
;

: SMATRIX	\ -- #r #c /item ; -- type ^head 
\ Build 2D structured matrix.
  idata?
  if  cSMATRIX  else  iSMATRIX  endif
;

: MATRIX	\ #r #c /item --
  STRUCT-ARRAY?
  IF  SMATRIX  FALSE TO STRUCT-ARRAY?  ELSE  MMATRIX  THEN
;


\ ******************
\ Useful short forms
\ ******************

: DMMATRIX	\ /item -- ; -- head
\ Build dynamically allocated unstructured 2D matrix template.
  0 0 rot MMATRIX
;

: DSMATRIX	\ -- /item ; -- id data
\ Build dynamically allocated typed 2D matrix template.
  0 0 rot SMATRIX
;

: DMATRIX	\ /item --
  STRUCT-ARRAY?
  IF  DSMATRIX  FALSE TO STRUCT-ARRAY?  ELSE  DMMATRIX  THEN
;

: ARRAY		\ n /item --
  1 -ROT  MATRIX
;

: DARRAY	\ /item --
  DMATRIX
;

: &!		\ heada xtb --
\ word for aliasing arrays,
\  typical usage:  a{ & b{ &!  sets b{ to point to a{'s data
  /DAhead cmove
;

: &		\ addr -- addr
\ used when aliasing arrays.
; IMMEDIATE


\ ************
\ Array access
\ ************

: }		\ head n -- addr[n] ; fetches 1-D array addresses
  over DA./item @ *			\ -- head offset
  swap DA.^data @ +			\ -- addr
;

: }}		\ head row col -- addr[row][col]
\ Calculate 2-D array address.
  -rot  over DA.#cols @ *  rot +	\ -- head row*#cols+col
  over DA./item @ *			\ -- head (row*#cols+col)*dsize
  swap DA.^data @ +
;

FPsystem 2 = [if]	\ check for NDP387

variable #n}}	\ -- addr
\ Holds number of "normal" compilations of }}
variable #o}}	\ -- addr
\ Holds number of "optimised" compilations of }}
variable #b}}	\ -- addr
\ Holds number of "0 0" compilations of }}

also h-c definitions

: c_}}		{ row-res -- } 
\ head row col -- addr + (row*#cols+col)*dsize ; on working stack
\ nos/row := row*#cols
  nos-res lit=0? 0= if
    a[ imul ]a  row-res l>a  ,>a  0 DA.#cols 3os-res +n[reg]  use-a
  endif
\ nos/row := row*#cols+col
  tos-res lit=0? 0= if
    a[ add ]a  row-res l>a  tos-res ,gl>a  use-a
  endif
\ nos/row := (row*#cols+col)*dsize
  a[ imul ]a  row-res l>a  ,>a  0 DA./item 3os-res +n[reg] use-a
\ nos/row := addr + (row*#cols+col)*dsize
  a[ add ]a  row-res l>a  ,>a  0 DA.^data 3os-res +n[reg]  use-a
;

oc: }}
  drop  3 min-depth(ws)
\ -- addr row col ; on working stack
  tos-res lit=0?  nos-res lit=0? and if
#b}} incr
    drop(ws) drop(ws)
    0 DA.^data c_lit  c_+  c_@
    reset-opt  exit
  endif

\ If addr is not a register or a literal, try to get it into
\ a register. Causes improvement of .4% in MM.FTH.
  3os-res reg/lit? 0= if
    GetNewReg ?dup if
      pop(ws) >r  pop(ws) >r
      pop(ws) over P0sdMove,  push(ws)
      r> push(ws)  r> push(ws)
    endif
  endif

  nos-res count=1?  3os-res reg? and if
#o}} incr
    nos-res reg? if
      nos-res c_}}
      drop(ws) nip(ws)  reset-opt
    else
\      GetNewReg ?dup if
\        nos-res over P1sdMove,
\	dup c_}}
\	drop(ws) drop(ws) drop(ws)
\	push(ws)
\      else
        nos-res rEAX sdmove,
        rEAX c_}}
        drop(ws) pop(ws) drop(ws)
        GetNewReg ?dup
        if  nip  endif
        rEAX over sdMove,  push(ws)
\      endif
    endif
  else
#n}} incr
    c_-rot
    c_over 0 DA.#cols c_lit c_+ c_@ c_*
    c_rot c_+		
    c_over 0 DA./item c_lit c_+ c_@ c_*
    c_swap 0 DA.^data c_lit c_+ c_@ c_+
  endif
;

previous definitions

[then]


\ ===================================
\ Dynamic arry allocation and release
\ ===================================

\ memory allocation status variable, 0 for OK
0 VALUE malloc-fail?

: }malloc	\ head n --
\ Allocate memory for a dynamic 1-D array. Typical usage:
\   & a{ #elements }malloc
  swap					\ -- n head
  1 over DA.#rows !  2dup DA.#cols !	\ -- n head
  tuck DA./item @ *			\ -- head size
  allocate to malloc-fail?		\ -- head data
  swap DA.^data !
;

: }}malloc	\ head #r #c --
\ Allocate memory for a dynamic 2-D array. Typical usage:
\   & a{{ #rows #cols }}malloc
  rot >r				\ -- #r #c ; R: -- head
  2dup  r@ DA.#cols !  r@ DA.#rows !
  *  r@ DA./item @ *
  allocate to malloc-fail?
  r> DA.^data !
;

: }}free	\ head --
\ Release 2D dynamic array memory, typical usage:
\   & a{{ }}free
  dup DA.^data @ free to malloc-fail?
  0 swap DA.^data !
;

: }free		\ head --
\ Release 1D dynamic array memory, typical usage:
\   & a{ }free
  }}free
;


\ ===============
\ Array utilities
\ ===============

VARIABLE print-width	\ -- addr
\ Holds the number of columns of data required for array printing.
  6 print-width !

: }fprint	\ n head --
\ Print n elements of a float array.
  SWAP 0 DO
    I print-width @ MOD 0= I AND
    IF  CR  THEN
    DUP I } F@ F.
  LOOP
  DROP
;

: }iprint	\ n head -- ; print n elements of an integer array
  SWAP 0 DO
    I print-width @ MOD 0= I AND
    IF  CR  THEN
    DUP I } @ .
  LOOP
  DROP
;

: }fcopy	\ shead dhead n --
\ Copy one array into another.
  0 DO
    OVER I } F@  DUP  I } F!
  LOOP
  2DROP
;

: }}fprint	\ n m head --
\ Print nXm elements of a float 2-D array
  ROT ROT SWAP 0 DO
    DUP 0 DO
      OVER J I  }} F@ F.
    LOOP
    CR
  LOOP
  2DROP
;

: }}fcopy	\ shead dhead n m --
\ Copy nXm elements of 2-D array src to dest.
  SWAP 0 DO
    DUP 0 DO
      2 PICK J I }} F@
      OVER J I }} F!
    LOOP
  LOOP
  DROP 2DROP
;


\ **************************
\ function vector definition
\ **************************

: v:		\ -- ; i*x -- j*x
\ Equivalent to DEFER
  defer
;

: defines	\ xt --
\ Equivalent to IS and TO-DO
  postpone is
;  IMMEDIATE

: use(		\ -- xt ; state smart '
\ Equivalent to the MPE word ASSIGN, which is a start-smart tick.
  postpone assign
; IMMEDIATE


\ ************************************************
\ Floating point locals - its a kluge but it works
\ ************************************************

((
  Code for local fvariables, loosely based upon Wil Baden's idea presented
  at FORML 1992.
  The idea is to have a fixed number of variables with fixed names.
  I believe the code shown here will work with any, case insensitive,
  ANS Forth.

  : test  2e 3e FRAME| a b |  a f. b f. |FRAME ;
  test <cr> 3.0000 2.0000 ok

  PS: Don't forget to use |FRAME before an EXIT .
))

8 CONSTANT /flocals

: (frame) ( n -- ) FLOATS ALLOT ;

: FRAME|	\ --
  0 >R
  BEGIN
    BL WORD COUNT 1 =
    SWAP C@ [CHAR] | =
    AND 0=
   WHILE
    POSTPONE F,  R> 1+ >R
  REPEAT
  /FLOCALS R> - DUP 0< ABORT" too many flocals"
  POSTPONE LITERAL  POSTPONE (frame)
; IMMEDIATE

: |FRAME	\ --
  [ /FLOCALS NEGATE ] LITERAL (FRAME)
;

: &h            HERE [ 1 FLOATS ] LITERAL - ;
: &g            HERE [ 2 FLOATS ] LITERAL - ;
: &f            HERE [ 3 FLOATS ] LITERAL - ;
: &e            HERE [ 4 FLOATS ] LITERAL - ;
: &d            HERE [ 5 FLOATS ] LITERAL - ;
: &c            HERE [ 6 FLOATS ] LITERAL - ;
: &b            HERE [ 7 FLOATS ] LITERAL - ;
: &a            HERE [ 8 FLOATS ] LITERAL - ;

: a             &a F@ ;
: b             &b F@ ;
: c             &c F@ ;
: d             &d F@ ;
: e             &e F@ ;
: f             &f F@ ;
: g             &g F@ ;
: h             &h F@ ;


\ ***********************
\ Random number generator
\ ***********************

HERE 1+ VALUE seed

: RANDOM	\ -- n
\ note: '$' prefix means HEX
  seed $107465 * $234567 + DUP TO seed
;

: CHOOSE 	\ n -- +n ; Paul Mennen, 1991 
  RANDOM UM* NIP
;


\ *************
\ Miscellaneous
\ *************

: tab	\ #tabs -- ; used in SUNDAY.SEQ
  out @ 7 + -8 and >pos
  1- 0 max 8 * spaces
;


\ *******
\ the end
\ *******

Reset_Search_Order

