\ Forth Scientific Library test load

\ This file loads all of the FSL routines and runs as many of the test
\ routines as possible.  The goal is to produce a consistent command
\ window image so that modifications can be validated quickly.
\
\ \ before a test means that it presently causes run time errors.
\ ( ??? ) means that there is presently no test case for these routines.
\ ( random ) means that the test case produces random output.
\ ( overflow ) means that the test takes more than 8 stack items.

mc" %vfxpath%/Lib" setmacro optiondir
mc" %vfxpath%/Lib/x86" setmacro floatdir

-short-branches

\ debugging tools for when something breaks
\ get %optiondir%\wview.fth
\ get %optiondir%\exexcept.fth
\ get %optiondir%\stackmon.fth

\ load the floating point pack and set it up for FSL use
\  include %floatdir%/Hfp387	\ load floating point pack
  include %floatdir%/Ndp387	\ load floating point pack
  8 set-precision		\ so the tests line up
  char . dp-char !		\ select ANS number conversion
  char . fp-char !
  finit				\ initialise FP unit and stack
  floored			\ set NDP to round to floor

include Vfx4Util		\ FSL harness for VFX Forth 4.0+

anew xx	include ExpInt.seq      expint_test
anew xx	include Elip.seq        elip_test
anew xx	include ExpInt.seq
        include Horner.seq      horner_test
anew xx	include Logistic.seq
        CR % 1.0 % 1.0 % 0.0 logistic f.  .( = 0.731059)
        CR % 3.2 % 1.5 % 0.2 logistic f.  .( = 0.993307)
        CR % 3.2 % 1.5 % 0.2 d_logistic f. .( = 0.00997209)
anew xx	include cube_rt.seq     ( ??? )
anew xx	include cubic.seq       ( ??? )
anew xx include RegFalsi.seq    ( ??? )
anew xx	include Hartley.seq     hartley-test1
                                hartley-test2
	include Hartley.tst
        CR SINE-TEST  CR SHOW-FFT  CR SHOW-POWER
        CR .SPEED  CR .SPEED-FHT
anew xx	include Logistic.seq
	include Aitken.seq      1.2E aitken_test1
                                2.0E aitken_test2
anew xx	include Logistic.seq
	include Hermite.seq     1.2E hermite_test1
                                2.0E hermite_test2
anew xx	include Logistic.seq
	include Lagrange.seq    1.2E lagrange_test1
                                2.0E lagrange_test2
anew xx	include Logistic.seq
        include DivDiffs.seq
	include Newton.seq      1.2E fnewt-test1
                                2.0E fnewt-test2
                                1.2E bnewt-test1
                                2.0E bnewt-test2
anew xx	include Factorl.seq     factorial-test
anew xx	include ShellSrt.seq    test_sort
anew xx	include ExpInt.seq
	include Horner.seq
	include Factorl.seq
	include SeriesPW.seq    series_tests
                                test_1/(1+x)
anew xx	include Logistic.seq
	include PolRat.seq      1.5E0 polrat_test1 err-test1
                                2E0 polrat_test2 err-test2
                                -2E0 polrat_test3 err-test3
anew xx	include ExpInt.seq
	include Horner.seq
	include Gamma.seq       gamma-test
anew xx	include AdaptInt.seq    ( ??? )
anew xx	include ExpInt.seq
	include Horner.seq
	include Gamma.seq
	include PCylFun.seq     pcf-test
        include Polys.seq       test-cheby
                                test-hermite
                                test-laguerre
                                test-legendre
                                test-glaguerre
                                test-bessel
anew xx	include Dates.seq       date-test
anew xx	include R250.seq        lcm_test
                                r250_test
anew xx	include Ran4.seq        ( ??? )
anew xx	include Hilbert.seq     2 hilbert-test .( = 0.0833333)
                                3 hilbert-test .( = 0.000462963)
                                4 hilbert-test .( = 1.65344e-7)
                                5 hilbert-test .( = 3.7493E-12)
anew xx	include Find.seq        3 test-find
anew xx	include Gauleg.seq      8 gauleg-test
anew xx	include ExpInt.seq
	include Horner.seq
	include Elip12.seq      EK_test
anew xx	include Runge4.seq      10 dvib_test_twice
                                10 dvib_test
                                10 lorenz_test
                                10 cap_test
                                200e-3 50e-3 1e-3 cap_test2
                                800e-3 50e-3 1e-3 dvib_test2
anew xx	include Kande.seq       kande_test
anew xx	include Factorl.seq
	include TelScop1.seq    test_telescope1
anew xx	include Factorl.seq
	include TelScop2.seq    test_telescope2
anew xx	include Structs.seq     test0 test1 test2 test3
anew xx	include Structs.seq
	include Hilbert.seq
	include LuFact.seq      lufact-test
                                lufact-test2
                                lufact-test3
abort
anew xx	include Structs
	include Hilbert
	include Lufact
	include Dets		4 det-test
anew xx	include Structs
	include Hilbert
	include Lufact
	include BackSub		backsub-test
anew xx	include Structs
	include Hilbert
	include Lufact
	include BackSub
	include InvM		4 invm-test
anew xx	include Complex
	include CMath		zmath-test
                                ( 10 zmath-table ( overflow )
anew xx	include Complex
	include CMath
	include DFourier		dfourier-test1
                                dfourier-test2
anew xx	include Complex
	include CMath
	include FFourier		ffourier-test1
				ffourier-test2
anew xx	include GaussJ		TEST-MAT ( random )
				3EQS SOLVE-IT
anew xx	include GaussJ
	include SVD			PROBLEM1
				PROBLEM2
                                TEST-SINGULAR ( random )
                                TEST-SVDCMP ( random )
              1e-32 svdtol F! use( fpoly[odd] )FIT-SINE
              8 TO #vr use( fpoly[odd] )FIT-SINE

anew xx	include Amoeba		SHOW-PROBLEM
                                1E-3 TEST-AMOEBA
(( Prompts user, and it is not FP
anew xx	include Sunday		QTEST
))
anew xx	include ExpInt
	include Horner
	include Gauss		gauss-test
anew xx	include SPH_BES		( ??? )
anew xx	include CRC			( self-testing )
anew xx	include Logistic
	include PolRat
	include Integral		TEST-PROGRAM	\ requires checked F**
anew xx	include Big			( ??? )
anew xx	include Walsh		walsh-test

