\ Forth Scientific Library test load

\ This file loads all of the FSL routines and runs as many of the test
\ routines as possible.  The goal is to produce a consistent command
\ window image so that modifications can be validated quickly.
\
\ \ before a test means that it presently causes run time errors.
\ ( ??? ) means that there is presently no test case for these routines.
\ ( random ) means that the test case produces random output.
\ ( overflow ) means that the test takes more than 8 stack items.

[defined] VfxForth [if]
C" C:\Products\VfxForth.dev\Sources\Lib" setmacro optiondir
C" C:\Products\VfxForth.dev\Sources\Lib\x86" setmacro floatdir

-short-branches

\ debugging tools for when something breaks
\ get %optiondir%\wview.fth
\ get %optiondir%\exexcept.fth
\ get %optiondir%\stackmon.fth

\ load the floating point pack and set it up for FSL use
\  include %floatdir%\hfp387	\ load floating point pack
  include %floatdir%\ndp387	\ load floating point pack
  8 set-precision		\ so the tests line up
  char . dp-char !		\ select ANS number conversion
  char . fp-char !
  finit				\ initialise FP unit and stack
  floored			\ set NDP to round to floor

include Vfx4Util		\ FSL harness for VFX Forth 4.0+
\ include Vfx3Util		\ FSL harness for VFX Forth 3.7+
[then]

[defined] ENTER-SWIFT [if]
EMPTY
\ REQUIRES FPMATH /INPUT
include C:\MyApps\Forth200x\SwiftForth300\SwiftForth\Lib\Options\Fpmath.f
/input
: finit  /ndp  ;
\ INCLUDE LIBRARY\FSL-UTIL.F      \ Required for all
include C:\MyApps\Forth200x\SwiftForth300\SwiftForth\Unsupported\Fslib\Library\Fsl-util.f
: DU<           \ ud1 ud2 -- t/f
  ROT SWAP 2DUP U<                      \ ud1l ud2l ud1h ud2h f
  IF   2DROP 2DROP  -1
  ELSE  <>
        IF   2DROP  0  ELSE  U<  THEN
  THEN
;

: DU>           \ ud1 ud2 -- t/f
  ROT SWAP 2DUP U>                      \ ud1l ud2l ud1h ud2h f
  IF   2DROP 2DROP  -1
  ELSE  <>
        IF   2DROP  0  ELSE  U>  THEN
  THEN  ;

: umd* 		( multiplicand . multiplier . -- product . . . )
				( ul uh vl vh)
	2 PICK OVER UM* 2>R ( ul uh vl vh R: x3 .)
	3 PICK UM* 2>R 	( ul uh vl R: x3 . x2 .)
	TUCK UM* 2>R 	( ul vl R: x3 . x2 . x1)
	UM* 		( x0 . R: x3 . x2 . x1)
	0 2R> D+ 	( x0 x1 . R: x3 . x2 .)
	2R@ D+
	2DUP 2R> DU< IF 	( x0 x1 . R: x3 .)
		R> 1+ >R
	THEN
	0 2R> D+ 	( x0 x1 x3 .)
;
4 cells buffer: tsumh	\ -- addr ; holds quad word
tsumh 2 cells + constant tsuml		\ tsumh holds high half, tsuml holds low

: umd/mod	\ uq ud -- udrem udquot
\ Coding UMD/MOD can give an enormous increase in speed.
  2>r  tsumh 2! tsuml 2!  2r>		\ save dividend
  #64 0 do				\ 64 times ; SFP003
    \ quad precision shift, preserving carry
    tsumh @ >r				\ save carry from shift
    tsumh 2@ d2*  tsuml @ 0<		\ high portion of quad word
    if  1 0 d+  then			\ add carry from low half ; SFP003
    tsumh 2!
    tsuml 2@ d2* tsuml 2!		\ shift low pair ; SFP003
    \ compare and subtract
    2dup tsumh 2@ du> 0= if		\ if divisor less than top 32 bits
      tsumh 2@ 2over d- tsumh 2!	\ subtract divisor
      1 tsuml cell + +!			\ increment quotient
      r> drop
    else
      r> 0< if				\ test preserved carry
        tsumh 2@ 2over d- tsumh 2!	\ subtract divisor
        1 tsuml cell + +!		\ increment quotient
      then
    then
  loop
  2drop					\ discard divisor
  tsumh 2@  tsuml 2@			\ return remainder and quotient
;

GILD
[then]

[defined] Win32Forth-menu-bar [if]
s" src\lib\fmacro" "fpath+

fpi fconstant pi
fpi FCONSTANT F=PI

TRUE VALUE is-static?     \ TRUE for statically allocated structs and arrays

: ftuck  fover fswap  ;
: F**2   FDUP F*     ;
: ZDUP    FOVER  FOVER  ;

needs fsl_util.f
[then]

[undefined] anew [if]
: anew  \ " name" -- ; executes previous marker and replaces it
  >in @                                 \ will reuse following name later
  bl word find                          \ look for previously defined MARKER
  if  execute  else  drop  then
  >in !  marker
;
[then]

anew xx	include ExpInt.seq      expint_test
anew xx	include Elip.seq        elip_test
anew xx	include ExpInt.seq
        include Horner.seq      horner_test
anew xx	include Logistic.seq
        CR % 1.0 % 1.0 % 0.0 logistic f.  .( = 0.731059)
        CR % 3.2 % 1.5 % 0.2 logistic f.  .( = 0.993307)
        CR % 3.2 % 1.5 % 0.2 d_logistic f. .( = 0.00997209)
anew xx	include cube_rt.seq     ( ??? )
anew xx	include cubic.seq       ( ??? )
anew xx include RegFalsi.seq    ( ??? )
anew xx	include Hartley.seq     hartley-test1
                                hartley-test2
	include Hartley.tst
        CR SINE-TEST  CR SHOW-FFT  CR SHOW-POWER
        CR .SPEED  CR .SPEED-FHT
anew xx	include Logistic.seq
	include Aitken.seq      1.2E aitken_test1
                                2.0E aitken_test2
anew xx	include Logistic.seq
	include Hermite.seq     1.2E hermite_test1
                                2.0E hermite_test2
anew xx	include Logistic.seq
	include Lagrange.seq    1.2E lagrange_test1
                                2.0E lagrange_test2
anew xx	include Logistic.seq
        include DivDiffs.seq
	include Newton.seq      1.2E fnewt-test1
                                2.0E fnewt-test2
                                1.2E bnewt-test1
                                2.0E bnewt-test2
anew xx	include Factorl.seq     factorial-test
anew xx	include ShellSrt.seq    test_sort
anew xx	include ExpInt.seq
	include Horner.seq
	include Factorl.seq
	include SeriesPW.seq    series_tests
                                test_1/(1+x)
anew xx	include Logistic.seq
	include PolRat.seq      1.5E0 polrat_test1 err-test1
                                2E0 polrat_test2 err-test2
                                -2E0 polrat_test3 err-test3
anew xx	include ExpInt.seq
	include Horner.seq
	include Gamma.seq       gamma-test
anew xx	include AdaptInt.seq    ( ??? )
anew xx	include ExpInt.seq
	include Horner.seq
	include Gamma.seq
	include PCylFun.seq     pcf-test
        include Polys.seq       test-cheby
                                test-hermite
                                test-laguerre
                                test-legendre
                                test-glaguerre
                                test-bessel
anew xx	include Dates.seq       date-test
anew xx	include R250.seq        lcm_test
                                r250_test
anew xx	include Ran4.seq        ( ??? )
anew xx	include Hilbert.seq     2 hilbert-test .( = 0.0833333)
                                3 hilbert-test .( = 0.000462963)
                                4 hilbert-test .( = 1.65344e-7)
                                5 hilbert-test .( = 3.7493E-12)
anew xx	include Find.seq        3 test-find
anew xx	include Gauleg.seq      8 gauleg-test
anew xx	include ExpInt.seq
	include Horner.seq
	include Elip12.seq      EK_test
anew xx	include Runge4.seq      10 dvib_test_twice
                                10 dvib_test
                                10 lorenz_test
                                10 cap_test
                                200e-3 50e-3 1e-3 cap_test2
                                800e-3 50e-3 1e-3 dvib_test2
anew xx	include Kande.seq       kande_test
anew xx	include Factorl.seq
	include TelScop1.seq    test_telescope1
anew xx	include Factorl.seq
	include TelScop2.seq    test_telescope2
anew xx	include Structs.seq     test0 test1 test2 test3
anew xx	include Structs.seq
	include Hilbert.seq
	include LuFact.seq      lufact-test
                                lufact-test2
                                lufact-test3
abort
anew xx	include Structs
	include Hilbert
	include Lufact
	include Dets		4 det-test
anew xx	include Structs
	include Hilbert
	include Lufact
	include BackSub		backsub-test
anew xx	include Structs
	include Hilbert
	include Lufact
	include BackSub
	include InvM		4 invm-test
anew xx	include Complex
	include CMath		zmath-test
                                ( 10 zmath-table ( overflow )
anew xx	include Complex
	include CMath
	include DFourier		dfourier-test1
                                dfourier-test2
anew xx	include Complex
	include CMath
	include FFourier		ffourier-test1
				ffourier-test2
anew xx	include GaussJ		TEST-MAT ( random )
				3EQS SOLVE-IT
anew xx	include GaussJ
	include SVD			PROBLEM1
				PROBLEM2
                                TEST-SINGULAR ( random )
                                TEST-SVDCMP ( random )
              1e-32 svdtol F! use( fpoly[odd] )FIT-SINE
              8 TO #vr use( fpoly[odd] )FIT-SINE

anew xx	include Amoeba		SHOW-PROBLEM
                                1E-3 TEST-AMOEBA
(( Prompts user, and it is not FP
anew xx	include Sunday		QTEST
))
anew xx	include ExpInt
	include Horner
	include Gauss		gauss-test
anew xx	include SPH_BES		( ??? )
anew xx	include CRC			( self-testing )
anew xx	include Logistic
	include PolRat
	include Integral		TEST-PROGRAM	\ requires checked F**
anew xx	include Big			( ??? )
anew xx	include Walsh		walsh-test

