\ AUTOVAR.FTH - AutoVar example for CIAO

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441

PLEASE NOTE OUR PHONE NUMBER CHANGE

To do
=====

Change history
==============
))

((
This example shows how to create and use a class called AUTOVAR.
This class contains one private data member and two public methods.
one method initialises the data store, the other will return the
contents of that store and post-increment it.
))

class AutoVar           \ begin a new class definition
private:
  cell: m_data			\ Where the count will be stored
public:
  meth: read++			\ The method to read and increment
  meth: set			\ The method to initialise the count
end-class			\ end definition

: AutoVar::read++	\ -- n ; Method
  1 m_data dup @ -rot +!	\ read and increment data store
;

: AutoVar::set		\ n -- ; Method
  m_data !			\ write to data store
;

((
Here are three test routines which each take an initial value
for an AutoVar type and then run the read++ method 10 times
writing the result. The first case uses a static instance of
AutoVar, the second case uses a local instance, and finally the
third case shows how to use a heap allocated instance and how
to typecast an object pointer.
))

AutoVar new Foo	\ create a static instance of Autovar

: test1		\ n -- ; Test static instance FOO
  foo.set			\ init with supplied index
  #10 0 do
    cr foo.read++ .		\ read and increment 10 times!
  loop
;

: test2		\ n -- ; Same thing, local class this time
  AutoVar new Foobar            \ create local instance of AUTOVAR
  foobar.set
  #10 0 do
    cr foobar.read++ .
  loop
;

: test3		\ n -- ; Third time, heap allocated instance
  AutoVar dnew                  \ create instance and store pointer
  tuck (AutoVar).set            \ use type cast on heap pointer
  #10 0 do
    cr dup (AutoVar).read++ .
  loop
  delete                        \ destroy heap instance
;


((
AUTOVAR2 is a class derived from AUTOVAR to extend its
functionality. AutoVar2 has no publically accessable methods but
uses operators to perform the read and initialise.
))

Class AutoVar2 : private Autovar	\ Create new class, inherit from
					\ Autovar with all methods in
					\ private scope.
     oper: read++			\ Default operator performs
					\ read++ method
  to oper: set				\ TO operator performs set method
end-class

AutoVar2 new Foo2	\ create a static instance

: test4		\ n -- ; Test static instance FOO
  to Foo2				\ init with supplied index
  #10 0 do
    cr foo2 .				\ read and increment 10 times!
  loop
;

: test5		\ n -- ; Same thing, local class this time
  AutoVar2 new Foobar			\ create local instance
  to FooBar
  #10 0 do
    cr FooBar .
  loop
;

