\ ClassVFX.fth - intelligent objects

((
Copyright (c) 1999, 2000, 2005, 2008, 2010, 2011
Construction Computer Software Pty Ltd.
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441

net: tech-support@mpeforth.com
web: www.mpeforth.com

To do
=====

Change History
==============
))

\ ============
\ *! ccsobject
\ *T CCS Object Model
\ ============

\ ***************
\ *S Introduction
\ ***************
\ *P The CCS object model has been in use for 25 years in an
\ ** application that now (2011) contains over one million lines of
\ ** Forth source code.

\ *P The programmers working on this system range from deep
\ ** systems programmers to quantity surveyors who have become
\ ** application programmers. Above all, this is a pragmatic and
\ ** usable system.

\ *P From the outset, this was not designed as the ultimate OOP
\ ** system. It was an evolution from intelligent structures, driven
\ ** by the requirements of increasing complexity in a real
\ ** application.

\ ===============
\ *N Why Objects?
\ ===============
\ *P OOP has gained successs as a means of handling software complexity through
\ ** hierarchical structures of carefully designed object types. Each class
\ ** encapsulates a set o reusuable functions and reuses functions provided by
\ ** its base class or classes. Reuse of OO code eliminates much of the bug
\ ** prone code repeated throughout complex software projects.

\ ==========================
\ *N Definition of an Object
\ ==========================
\ *P Objects combine data and behavior and separation of interface from
\ ** implementation. Objects are complete entities from a problem domain.

\ ================
\ *N Extensibility
\ ================
\ *P In a truly extensible system, components use other components. The network
\ ** of these components is continuously expanded at run time by the addition
\ ** of new or existing subclassed components. A system wide garbage collector
\ ** is required to deallocate components.

\ =================
\ *N Object Methods
\ =================
\ *P Objects can only be addressed through methods associated with the object,
\ ** inherited by the object from a previously defined class or type cast
\ ** element of complex class using the dot notation.

\ *P Methods are classified to operate on the class or the
\ ** instance of the class. Methods are defined using the Object
\ ** Definition Language word *\fo{OPERATOR:} followed by the
\ ** name of the method. In the CCS method naming convention the
\ ** name must start with the '<' character and end with
\ ** '>' character.

\ *P The word *\fo{OPERATOR:} is the same word used to define
\ ** *\fo{TO} and other operators accepted by children of
\ ** *\fo{VALUE}.

\ ================
\ *N Class methods
\ ================
\ *P Class methods uses the *\fo{TYPE:} definition structure and can not address
\ ** any data space. These methods will be excuted during compilation as oposed
\ ** to laying down code that will be executed during runtime.

\ ===================
\ *N Instance methods
\ ===================
\ *P Instance methods use the instance of a *\fo{TYPE:} structure and can access the
\ ** dataspace of the instance if required. If access to the instance type:
\ ** structure is required during runtime then the method can be made immediate
\ ** with the object definition word *\fo{STRUCTURE-METHOD}. The method will then be
\ ** executed at compile time and not compiled inline.

\ =======================
\ *N Super Object Methods
\ =======================
\ *E     Class Methods
\ **     CLASS
\ **       mruns <default> make-inst
\ **       mruns <sizeof>  sizeprovider
\ **       mruns <addr>    ?complit
\ **       mruns <self>    ?complit
\ **    Instance Methods
\ **    INST
\ **       mruns <offsetof>  offsetprovider   structure-method
\ **       mruns +offsetof +provider        structure-method
\ **       mruns <sizeof>  sizeprovider     structure-method
\ **       mruns <default> inst-addr        structure-method
\ **       mruns <ptr>     inst-addr        structure-method
\ **       mruns <&ptr>    inst-addr        structure-method
\ **       mruns <addr>    inst-addr        structure-method
\ **       mruns <copy>    inst-copy        structure-method
\ **       mruns <blank>   inst-blank       structure-method
\ **       mruns <erase>   inst-erase       structure-method
\ **       mruns <init>    inst-erase       structure-method
\ **       mruns <zero>    inst-erase       structure-method
\ **       mruns <count>   inst-count       structure-method
\ **       mruns <self>    inst-self        structure-method
\ **       mruns <#bpe>    inst-bpeprovider structure-method

\ *P Methods are considered to be Modifiers, Inspectors, Constructors or Destructors.
\ *D Inspectors return the properties of an object instance.
\ *D Modifiers changes the properties or data of and object instance.
\ *D Constructors builds an object instance so that Instance Inspector
\ ** methods can be used on the object.
\ *D Destructors destroy the object instance.

\ =====================
\ *N Object terminology
\ =====================
\ *D {type:}    A Forth word defined with *\fo{TYPE:}.
\ *D {instance} A Forth word defined with an instance of *\fo{TYPE:}.
\ *D ODL        Object Definition Language. Words that are only valid during the
\ **            definition of a Object. (compile time)


\ ************************
\ *S Creating Object Types
\ ************************
\ *P The CCS Object model has an Object Definition Language that is only valid
\ ** when defining object types.
\ *E Object definition Language
\ ** TYPE: [name:]
\ **   SUPERCLASS
\ **   INHERITS  ( alias for SUPERCLASS )
\ ** EXTEND-TYPE [name:]
\ **   MRUNS <method>
\ **   STRUCTURE-METHOD
\ **   CLASS
\ **   INST
\ **   n SKIPPED
\ ** END-TYPE

\ *P CCS types are defined with *\fo{TYPE:} followed by a name ending with a
\ ** colon.
\ *C  TYPE: <name:> ... END-TYPE
\ *P The colon implies that the new word can itself create new Forth words
\ ** that behave according to the type definition. This is an instance of a
\ ** type. An instance of a type can not create new Forth words. The type
\ ** definition describes any data space requirements and valid methods that
\ ** can be used to address the object.

\ ==============
\ *N Inheritance
\ ==============
\ *P All objects inherit basic methods from the super object. The CCS object
\ ** model allows a single inheritance oppurtunity during definition of the
\ ** type: using the object definition word *\fo{INHERITS} followed by the type: name.

\ ================
\ *N Unary Objects
\ ================
\ *P An object with a single data field or member is called a unary object.
\ ** Some Objects have multiple fields but if the Instance Modifier Methods
\ ** address the data space as a monolithic unit then the object can be
\ ** classified as unary.
\ *E Example of an Unary object
\ ** TYPE: C-VAR:
\ **   1 SKIPPED          \ reserve 1 byte of data space
\ **   MRUNS <to> C!      \ method to modify data space
\ ** END-TYPE

\ ==================
\ *N Complex Objects
\ ==================
\ *P *\fo{TYPE:} definition with multiple fields or members are called complex
\ ** objects.

\ ===============
\ *N Array Object
\ ===============
\ *P An array is a rectangular multi dimensional data structure that holds data
\ ** elements of the same type in subdivisions called cells. The bounds of an
\ ** array are the lowest and highest permissible subscript integers. The value
\ ** of a dimension's lower bound is called its base.


\ *********************
\ *S Instance of a Type
\ *********************
\ ==================
\ *N Static Instance
\ ==================
\ *P A static instance of an object is compiled in the Forth dictionary and
\ ** its instance structure and structure data space is inside the Forth
\ ** dictionary space. The structure data space is erased for numeric and zero
\ ** terminated string objects. Counted string object structure data space are
\ ** blanked.
\ ==================
\ *N Local Instances
\ ==================
\ *P A local instance of an object is compiled on the Forth locals stack and
\ ** only the structure data space requirement is allocated. Unary object data
\ ** space is not initialized. It is assumed that the developer will be
\ ** assigning values to these instances just now. Complex objects with a
\ ** sub classed <default> class method can lay down initialization code.
\ ** The default class method for complex objects is to erase the structure
\ ** data space.

\ *****************************
\ *S Object Service Definitions
\ *****************************
\ *P The client/server model is very good model to use when
\ ** implementing an object model. The 'server code' operates on
\ ** an object instance or creates an object instance from an
\ ** existing class.

\ *P Object Service Definitions provide a service to an object.
\ ** The idea of a specific object service definition is to avoid
\ ** creating a new language using the object methods. Use Forth
\ ** and apply some discipline to parameter stack inputs and
\ ** outputs. Document the service requirement and object effect.
\ ** Announce and guarantee the service. Using standard naming
\ ** conventions developers can use standard tools to get access
\ ** to the service definitions.

\ *P The stack input of the specific object service definition
\ ** can only be one of the following three formats to qualify as
\ ** an object service.
\ *( 1
\ *B object  --- flag or object
\ *B object> --- flag or void
\ *B vobject> --- flag or void
\ *)

\ =================
\ *N Service type 1
\ =================
\ *P The definition copies the object to its own data space,
\ ** performs the service by operating on that data space and
\ ** returns a result flag or copies modified data space to the
\ ** object.

\ =================
\ *N Service type 2
\ =================
\ *P The definition performs the service by operating on the
\ ** client data space. A result flag or void is returned. The
\ ** client can use the modified object data space.

\ =================
\ *N Service type 3
\ =================
\ *P The definition performs the service by operating on the client
\ ** virtual data space and/or object data space. A result flag or
\ ** nothing is returned. The client can use the modified object
\ ** data space. The service defintion does not destroy the client
\ ** virtual data space, the client must make and destroy the virtual
\ ** data space.

\ *******************
\ *S Coding Standards
\ *******************
\ *N Parameter stack comments
\ *P These are the basic CCS house rules.
\ *(
\ *B The stack comment must indicate the *\fo{TYPE:} instance.
\ *B Object Pointers must have a trailing ">".
\ *B Virtual object instances must have a leading "v" and a trailing ">".
\ *)

\ ****************************
\ *S Elements of Simple Design
\ ****************************
\ *( 1
\ *B Passes its tests
\ *B Minimizes duplication
\ *B Maximizes clarity
\ *B Has fewer elements
\ *B Keep it simple
\ *B If it ain't broke, don't fix it
\ *)

\ ======
\ *> ###
\ ======

