\ String.fth - CCS string functions

((
Copyright (c) 1999, 2000, 2005, 2008, 2010, 2011
Construction Computer Software Pty Ltd.
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441

net: tech-support@mpeforth.com
web: www.mpeforth.com

To do
=====

Change History
==============
))

\ ============
\ *! stringccs
\ *T CCS String functions
\ ============
\ *P The CCS application requires a large number of string functions
\ ** which are not present in most Forth systems. The file
\ ** *\i{StringCCS.fth} contains these words.

code scan	\ caddr u char -- caddr2 u2
\ *G This version does NOT handle tabs.
  mov	edx, ebx			\ EDX=char
  mov	ecx, 0 [ebp]			\ ECX=u=len
  mov   ebx, 4 [ebp]			\ EBX=caddr
l$1:
  jecxz l$2				\ exit if len=0
  mov	al, 0 [ebx]			\ get character
  cmp	al, dl				\ get out if char to scan for
  je	l$2
  dec	ecx				\ step count and address
  inc	ebx
  jmp	l$1
l$2:
  mov	4 [ebp], ebx			\ 'c-addr
  mov	ebx, ecx			\ 'u=len'
  lea	ebp, 4 [ebp]			\ restore stack
  next,
end-code

code SKIP	\ c-addr u char -- 'c-addr 'u                    MPE.0000
\ *G This version does NOT handle tabs.
  mov	edx, ebx			\ EDX=char, DL=char
  mov	ecx, 0 [ebp]			\ ECX=u=len
  mov   ebx, 4 [ebp]			\ EBX=caddr
l$1:
  jecxz l$2				\ exit if len=0
  mov	al, 0 [ebx]			\ get character
  cmp	al, dl				\ get out if not char to skip
  jne	l$2
  dec	ecx				\ step count and address
  inc	ebx
  jmp	l$1
l$2:
  mov	4 [ebp], ebx			\ 'c-addr
  mov	ebx, ecx			\ 'u=len'
  lea	ebp, 4 [ebp]			\ restore stack
  next,
end-code

  #1024 CONSTANT B/BUF
: -LEADING                                 ( c-addr1 u-cnt1 --- c-addr2 u-cnt2 )
  \ *G Modify a string address/length pair to ignore any leading space
   32 SKIP
;

: -TRAILING            \ c-addr u1 -- c-addr u2                      17.6.1.0170
   \ *G Modify a string address/length pair to ignore any trailing space
   \ or tab characters.
   dup 0
   ?do
     2dup + 1 - c@  bl <> ?leave
     1-
   loop
;

: $-TRIM                                     ( c-addr u-cnt --- c-addr' u-cnt' )
   \ *G Trim string at addr cnt and return revised addr and cnt.
   -LEADING -TRAILING
;

: LWC                                                         ( char --- char' )
   \ *G ASCII Lowercase char on stack
   DUP #65 #90 WITHIN? IF BL + THEN
;

  +SHORT-BRANCHES
: LOWER                                                     ( c-addr u-cnt --- )
  \ *G Lowercase string at address cnt. Operates only on character values
  \ ** ASCII A to ASCII Z
  { %addr %cnt --- }
    begin
        %cnt 0>
    while
        %addr c@ [CHAR] A [CHAR] Z WITHIN?
        IF %addr c@  $20 OR %addr C! THEN
        inc %addr
        dec %cnt
    repeat
    ;
  -SHORT-BRANCHES

: STR-END=   ( s_addr s_cnt t_addr t_cnt --- bool )
	 \ *G Compare sub string address count with the end sub string of the target string
	 \ ** for the sub string count.
	 #2 PICK OVER <=
	 IF
        DUP /STRING #2 PICK NEGATE /STRING STR=
   ELSE
        4DROP
        False
   THEN
;

: STR-START=    ( s_addr s_cnt t_addr t_cnt --- bool )
   \ *G Compare sub string address count with the start sub string of the target string
	 \ ** for the sub string count.
	 #2 PICK >=
	 IF
        OVER STR=
   ELSE
        3DROP
        False
   THEN
;

: WPLACE                                                 ( c-addr1 u c-addr2 --- )
  \ *G Copy the string described by c-addr1 u to a wcounted string at
  \ ** the memory address described by c-addr2.
  #0 OVER W! WAPPEND
;

[undefined] LCOUNT [if]
: LCOUNT                                                   ( addr --- addr+4 cnt )
   \ *G Given the address of a 32-bit long-counted string in memory *\fo{LCOUNT} will
   \ ** return the address of the first character and the length in
   \ ** characters of the string.
   DUP 4+ SWAP @
;
[then]

: LAPPEND	\ c-addr u $dest --
( *G Add string c-addr u to the end of 32-bit counted string $dest. )
  >R  TUCK  R@ LCOUNT + SWAP MOVEX      \ add source to end
  R@ @ + R> !                           \ add length to count
;

: LPLACE        \ c-addr1 u c-addr2 --
  \ *G Copy the string described by c-addr1 u to a 32-bit counted string at
  \ ** the memory address described by c-addr2.
  #0 OVER ! LAPPEND
;

  +SHORT-BRANCHES
: -BYTE=                                    ( c-addr u-cnt char --- addr|false )
   \ *G Find first occurence of a byte value equal to char. Searches for char
   \ ** from addr-1 do not xam char at addr. minus_byte_equal
  { %addr %cnt %char | %result --- }
    zero %result
    dec %addr                                                  \ start at addr-1
    begin
        %cnt 0>
    while
        %addr c@ %char =
        IF
            %addr to %result
            zero %cnt
        else
            dec %addr
            dec %cnt
        THEN
    repeat
    %result
    ;
  -SHORT-BRANCHES

  +SHORT-BRANCHES
: BYTE=                                     ( c-addr u-cnt char --- addr|false )
   \ *G Find first occurence of a byte value equal to char. byte_greater_than
  { %addr %cnt %char | %result --- }
    zero %result
    begin
        %cnt 0>
    while
        %addr c@ %char =
        IF
            %addr to %result
            zero %cnt
        else
            inc %addr
            dec %cnt
        THEN
    repeat
    %result
    ;
  -SHORT-BRANCHES

  +SHORT-BRANCHES
: POS-JMP                      ( addr cnt sub_addr sub_cnt step --- addr|false )
  \ *G Return the start addr of sub_addr sub_cnt in addr cnt advancing the match
  \ ** with step. Return false if not found. pos_jump
  { %addr %cnt %subaddr %subcnt %step | %result -- addr or false }
    ZERO %result
    \
    %subcnt >R
    %subcnt 0>                                  \ sub count greater than zero...
    %step   0> AND                                \ ...step greater than zero...
    IF
        begin
        %cnt R@ >=                 \ ...argument count greater than sub count...
        while
            %addr %subaddr R@ S=
            IF
                %addr to %result
                zero %cnt
            ELSE
                %addr %cnt %step /STRING to %cnt to %addr
            THEN
        repeat
   THEN
   R> drop
   %result
  ;
  -SHORT-BRANCHES

: REMOVE-XTRA-BLANKS                                   \ addr cnt --- addr' cnt'
   \ *G Remove repeating xtra blanks, leaving at least 1 blank, from single
   \ ** byte string at addr cnt return result addr cnt within the original
   \ ** buffer space
   2DUP
   OVER >R                                                 \ running source addr
   #0 -ROT
   BEGIN
        DUP #1 >                              \ any chars to process and done ??
   WHILE
        OVER W@ $2020 =
        IF
             #2 PICK ?DUP
             IF
                  R@ + #2 PICK <>
                  IF
                       R> DROP
                       OVER >R                               \ save running addr
                  THEN
             ELSE
                  R> DROP
                  OVER >R                                    \ save running addr
             THEN
             ROT 1+ -ROT
        ELSE
             #2 PICK
             IF
                  OVER W@
                  R@ W!
                  R> 1+ >R
             THEN
        THEN
        #1 /STRING
   REPEAT
   2DROP
   R> DROP
   -
;

: REPLACE-CHAR                                ( addr cnt new_char old_char --- )
   \ *G Replace occurrences of old char with new char in string.
   { %addr %cnt %new %old --- }
   BEGIN
        %cnt 0>
   WHILE
        %addr c@ %old =
        IF
             %new %addr C!
        THEN
        inc %addr
        dec %cnt
   REPEAT
;

: ISOLATE-STR RUNWORD     ( c-add u-cnt start_char end_char --- c-addr2 u-cnt2 )
   \ *G Return the address and count of the first sub string in string at
   \ ** address count.
   SWAP                                   \ c-addr u-cnt end_char start_char ---
   2SWAP                                  \ end_char start_char c-addr u-cnt ---
   ROT                                    \ end_char c-addr u-cnt start_char ---
   SCAN                                             \ end_char c-addr2 u-cnt ---
   DUP                                                \ chars left to process ??
   IF   \ yes! end_char c-addr2 u-cnt2 ---
        OVER SWAP                          \ end_char c-addr2 c-addr2 u-cnt2 ---
        #1 /STRING                   \ jump over first character incase the same
        \ end_char c-addr2 c-addr3 u-cnt3
        ROT                                \ end_char c-addr3 u-cnt3 c-addr2 ---
        ROT                                \ end_char u-cnt3 c-addr2 c-addr3 ---
        2SWAP                              \ c-addr2 c-addr3 end_char u-cnt3 ---
        SWAP                               \ c-addr2 c-addr3 u-cnt3 end_char ---
        SCAN
        IF 1+ OVER - ELSE NIP #0 THEN
   ELSE ROT DROP
   THEN
  ;

CREATE CR$
       1 C, #13 C,

CREATE LF$
       1 C, #10 C,

 [-SIN
: N>$                                                         ( n --- addr cnt )
   \ *G Convert n to decimal string
   BASE @ >R
   DUP ABS SWAP 0< >R
   DECIMAL #0  <# #S R> SIGN #>
   R> BASE !
 ;
 sin]


: ZSTRBUF$   ( zstr --- addr u )
   \ *G Return buffer start address and the size of the buffer in number of character units
   ZCOUNT
;

: ZSTRCHARS  ( zstr --- u )
   \ *G Return number of characters units stored at zstring
   ZSTRBUF$ NIP
;

: ZSTRCHAR#>ADDR   ( zstr index# --- addr true|false )
   \ *G Convert character index number to character address.
   \ ** Index# is an option base 1 character position.
   DUP 0> NOT IF 2DROP False EXIT THEN
   OVER ZSTRCHARS 2DUP <= \ index within the string ??
   IF
        SWAP 1- /STRING DROP True
   ELSE
        3DROP
        False
   THEN
;

: ZSTREND-  ( zstr --- )
   \ *G Remove last character from zero terminated string
   ZSTRBUF$ DUP IF 1- THEN + #0 SWAP C!
;

: ZSTR-REMOVE-EXTRA-BLANK                                       ( zstr -- bool )
	 \ *G Scan zero terminated string for first occurance of repaeting blanks.
	 \ ** Remove the second blank and return true. If no repeating blanks was found return false.
	 ZCOUNT DUP #2 <
   IF 2DROP False EXIT THEN
   BEGIN
        OVER W@ $2020 =
        IF
             #1 /STRING OVER >R
             #1 /STRING CHAR+
             R> SWAP MOVEX
             True EXIT
        ELSE
             #1 /STRING
             #2 OVER >
        THEN
   UNTIL
   2DROP False
  ;

: ZSTR-#BLANKS-LEADING                                          ( zstr --- u )
   \ *G Return number of leading blanks at zstr
   ZSTRBUF$ OVER -ROT #32 SKIP DROP SWAP -
;

\ *H Number of words in a zero terminated text string
: ZSTR-#WORDS                                               ( zstr --- #words )
   \ *G Return number of words at zstr.
   ?DUP
   IF
        ZCOUNT $-TRIM DUP
        IF
             #1 -ROT                                          \ prime word count
             BEGIN
                   #32 SCAN DUP
                   IF
                        ROT 1+ -ROT
                        #32 SKIP
                        False
                   ELSE 2DROP
                        True
                   THEN
             UNTIL
        ELSE 2DROP
             False
        THEN
   ELSE False
   THEN
;

  #CCS-ERROR ERR_NAME_BLANK
  #CCS-ERROR ERR_NOT_SET
  #CCS-ERROR ERR_NOT_FOUND

: ZSTR>-GET-MACRO                                   ( zmacro> zbuf> --- err|false )
   \ *G Get the current value of the text macro.
   \ ** A valid existing text macro could have a leading and trailing %
   \ ** character.
   { %zmacro> %zbuf> --- }
   %zbuf> IF #0 %zbuf> C! THEN                    \ init user buffer if specified
   %zmacro> C@ 0= IF ERR_NAME_BLANK EXIT THEN
   \ Remove any leading %
   %zmacro> ZCOUNT OVER C@ [CHAR] % =
   IF   #1 /STRING
        DUP 0= IF ERR_NAME_BLANK EXIT THEN
   THEN
   \ Remove any trailing %
   2DUP + 1- C@ [CHAR] % =
   IF   1-
        DUP 0= IF ERR_NAME_BLANK EXIT THEN
   THEN
   [ ALSO SYSTEM ]
   \ MacroExists? CARNAL knowledge of FORTH system !!!!!!!!!!!!!
   wid-subst search-wordlist
   [ PREVIOUS ]
   IF   EXECUTE COUNT DUP
        IF   %zbuf>                          \ user specified a return buffer ??
             IF   %zbuf> ZPLACE
             ELSE 2DROP
             THEN
             False
        ELSE 2DROP ERR_NOT_SET
        THEN
   ELSE ERR_NOT_FOUND
   THEN
;


\ =======
\ *> ####
\ =======

