\ SystemCCS.fth - System extensions for ClassVFX

((
Copyright (c) 2000-2011
Construction Computer Software Pty Ltd.
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441

net: tech-support@mpeforth.com
web: www.mpeforth.com

To do
=====

Change History
==============
20120715 SFP002 WID-THREADS now in kernel.
20110412 SFP001 Refactoring. Documentation.
))

\ ============
\ *! systemccs
\ *T CCS System Extensions
\ ============

\ *******************
\ *S System Wordlists
\ *******************
\ ========================
\ *N Definition management
\ ========================
\ *P One of the difficult tasks in software engineering is to manage
\ ** multiple desperate programers, each with their own style and
\ ** creative abilities. Naming conventions of definitions and the
\ ** application interface contract are very problematic in a team
\ ** enviroment. To limit the conflict between programmers
\ ** and the misinterpretation of interface contracts the following
\ ** rules apply:
\ *(
\ *B Defitions marked as '**' do not form part of the interface and
\ ** can be changed without notice by the author.
\ *B Definitions marked "***" form part of the 'system' interface
\ ** and must have a correct stack picture comment and glossary
\ ** comment. The author can make changes to the definition only with
\ ** full audit and notification procedures.
\ *)

\ *P Using predefined system wordlists allows for a clear
\ ** understanding from the user of the definition and gives the
\ ** user the ability to debug the code without knowing the
\ ** intimate module construction of the author.

\ *P The Hidden wordlists are applicable to definitions that are
\ ** potential system dependencies or extensions to the current system.
\ ** Definitions that are possible name conflicts must be placed
\ ** in a module. Interface defintion must be exposed via export
\ ** functions and must have names that do not conflict with the
\ ** underlying system. The export list must be published separately
\ ** and can be renamed by the implementor to avoid name conflicts.

ALIGN #20 ALLOT&ERASE		\ space for horrors to come
CREATE CCSExtensions

[undefined] Pocket [if]
#256 BUFFER: Pocket	\ -- addr
\ *G Scratch buffer.
[then]

[undefined] LATEST-NFA [if]
: LATEST-NFA	\ -- nfa
\ *G Point to the last defintion's Name Field
  LATEST CTRL>NFA  ;
[then]

WORDLIST CONSTANT wid**
\ *G Two star temporary wordlist during compilitaion
WORDLIST CONSTANT wid***
\ *G Three star temporary wordlist during compilitaion
WORDLIST CONSTANT widHidden
\ *G System Hidden word list not to be used without consent from the owner.

\ ===========================================
\ *N Hidden wordlists and search order
\ ===========================================

: EXPOSE-**	\ --
\ *G Expose the ** wordlist.
  GET-ORDER wid** SWAP 1+ SET-ORDER  ;	\ Add private wordlist to order

: EXPOSE-***	\ --
\ *G Expose the *** wordlist.
  GET-ORDER wid*** SWAP 1+ SET-ORDER  ;	\ Add private wordlist to order

: EXPOSE-HIDDEN	\ --
\ *G Expose the hidden wordlist.
  GET-ORDER widHidden SWAP 1+ SET-ORDER  ;	\ Add private wordlist to order

: [OLDFORTH]	\ --
\ *G Search order for running CCS system. [imm]
  [COMPILE] [FORTH] ; IMMEDIATE

: [FORTH]	\ --
\ *G Set the search order for compiling CCS. [imm]
  ONLY winconstants
  EXPOSE-**
  EXPOSE-***
  ALSO FORTH
  DEFINITIONS
; IMMEDIATE

: [PLUG]	\ --
\ *G Set the search order for a plug. [imm]
  [COMPILE] [FORTH]
  EXPOSE-HIDDEN                                 \ Add private wordlist to order
; IMMEDIATE

: [PRIVATE]	\ --
\ *G Set the search order to make definitions Private. [imm]
  ONLY FORTH
  ALSO winconstants
  ALSO FORTH
  EXPOSE-***
  EXPOSE-**
  DEFINITIONS
; IMMEDIATE

: [FUNCTIONS]	\ --
\ *G Set the search order for compiling Windows functions. [imm]
  ONLY FORTH
  EXPOSE-**
  EXPOSE-***
  ALSO WINCONSTANTS
  ALSO EXTERNALS
  DEFINITIONS
; IMMEDIATE

ALSO SYSTEM DEFINITIONS

#0 VALUE Modulename?

: (MODULE?)	\ thread# nfa --- flag
   { %thread %nfa --- }
   %nfa NAME> >BODY @ CONTEXT @ =
   IF
        %nfa to Modulename?
        False
   ELSE
        True
   THEN
 ;

PREVIOUS DEFINITIONS

[FORTH]

: MODULE?	\ --- flag
   [ ALSO SYSTEM ]
   #0 to Modulename?
   ASSIGN (MODULE?) WIDMODULES WALKWORDLIST
   Modulename? 0<>
   [ PREVIOUS ]
;

: **            ( --- )
\ *G Compile time directive to move last definition from current to
\ ** two_star private wordlist. When HIDE-** is executed all definitions in
\ ** the two_star word list will be moved to the CCS hidden wordlist. [imm]
  SINACTIVE? 0= IF
    MODULE? IF
      LATEST-NFA CONTEXT @ MoveNameToWid
      0= ABORT" Can not move name"
    ELSE
      GET-CURRENT FORTH-WORDLIST = IF
        LATEST-NFA wid** MoveNameToWid DROP
      THEN
    THEN
  THEN
; IMMEDIATE

: ***           ( --- )
\ *G Compile time directive to move last definition from current wordlist to
\ ** three_star private wordlist. When HIDE-*** is executed all definitions
\ ** in the three_star word list will be moved to the hidden wordlist.
\ ** [imm]
  SINACTIVE? 0= IF
    MODULE? IF
      LATEST-NFA CONTEXT @ MoveNameToWid DROP
    ELSE
      GET-CURRENT DUP FORTH-WORDLIST = SWAP wid** = OR IF
        LATEST-NFA wid*** MoveNameToWid DROP
      THEN
    THEN
  THEN
; IMMEDIATE

: HIDE-(**) **  \ thread# nfa --- flag
\ CallBack for WalkWordList, Called for each word in a wordlist
  { %thread# %nfa | -- }
  %nfa widHidden MoveNameToWid DROP
  %nfa NAME> >CODE-GEN OFF
  True                                   \ keep walking the words in wordlist
;

[UNDEFINED] WID-THREADS [IF]
: WID-THREADS   ( wid --- addr c_ucnt  )
  DUP >THREADS                                           \ step to thread table
  SWAP >#THREADS @ CELLS
;
[THEN]

: WORDLIST-EMPTY?	\ wid --- flag
   True SWAP
   WID-THREADS BOUNDS ?DO                                    \ size of the table
     I @
     IF  INVERT LEAVE  THEN
   Cell +LOOP
;

: HIDE-**	\ --
\ *G Compile time directive to hide all ** definitions.
   ?EXEC
   BEGIN
        ASSIGN HIDE-(**)
        wid** WALKWORDLIST
        wid** WORDLIST-EMPTY?
   UNTIL
   wid** WID-THREADS ERASE
;

: HIDE-***	\ --
\ *G Compile time directive to hide all *** definitions.
   ?EXEC
   BEGIN
        ASSIGN HIDE-(**)
        wid*** WALKWORDLIST
        wid*** WORDLIST-EMPTY?
   UNTIL
   wid*** WID-THREADS ERASE
;

: last>hidden	\ --
\ Move the last word found to the HIDDEN vocabulary.
  LastNameFound widHidden MoveNameToWid 0=
  ABORT" Move to HIDDEN Failed"
;

: HIDE-WORD	\ {definition} --
\ *G Place this word in the Hidden vocabulary.
  Pocket OFF
  GET-TOKEN COUNT Pocket PLACE
  Pocket COUNT FORTH-WORDLIST SEARCH-WORDLIST IF
    >CODE-GEN OFF  last>hidden		\ clr source inliner
  ELSE
    Pocket COUNT GET-CURRENT SEARCH-WORDLIST IF
      >CODE-GEN OFF  last>hidden	\ clr source inliner
    ELSE
      Pocket COUNT wid*** SEARCH-WORDLIST IF
        >CODE-GEN OFF  last>hidden	\ clr source inliner
      ELSE
        Pocket COUNT wid** SEARCH-WORDLIST IF
          >CODE-GEN OFF  last>hidden	\ clr source inliner
        ELSE
          Pocket COUNT widHidden SEARCH-WORDLIST ?undef
          >CODE-GEN OFF			\ clr source inliner
        THEN
      THEN
    THEN
  THEN
;

: name>forth	\ nfa --
\ *G Move the given word to the *\fo{Forth} vocabulary.
  FORTH-WORDLIST MoveNameToWid 0=
  ABORT" Move to FORTH Failed"
;

: last>forth	\ nfa --
\ *G Move the last word found to the *\fo{Forth} vocabulary.
  lastNameFound name>forth  ;

: RUNWORD	\ --
\ *G Used during compilation to unhide a previously hidden word.
  SINACTIVE? 0= IF
    GET-CURRENT FORTH-WORDLIST <> IF
      GET-CURRENT wid**  =
      GET-CURRENT wid*** = OR
      IF
        LATEST-NFA name>forth
      THEN
    THEN
    DISCARD-SINLINE
  THEN
; IMMEDIATE

: API-EXPORT	\ {definition} --
\ *G Export the next inline definition to FORTH vocabulary.
  Pocket OFF
  GET-TOKEN COUNT Pocket PLACE
  Pocket COUNT GET-CURRENT SEARCH-WORDLIST IF
    >CODE-GEN OFF  last>forth		\ clr source inliner
  ELSE
    Pocket COUNT wid*** SEARCH-WORDLIST IF
      >CODE-GEN OFF  last>forth		\ clr source inliner
    ELSE
      Pocket COUNT wid** SEARCH-WORDLIST IF
        >CODE-GEN OFF  last>forth	\ clr source inliner
      ELSE
        Pocket COUNT FORTH-WORDLIST SEARCH-WORDLIST ?undef
        DROP
      THEN
    THEN
  THEN
;

: last>***	\ --
\ *G Move the last word found to the *** vocabulary.
  LastNameFound wid*** MoveNameToWid 0=
  ABORT" Move to *** Failed"
;

: API-MODULE	\ {definition} --
\ *G Export the next inline definition to *** vocabulary.
  Pocket OFF
  GET-TOKEN COUNT Pocket PLACE
  Pocket COUNT GET-CURRENT SEARCH-WORDLIST IF
    >CODE-GEN OFF  last>***		\ clr source inliner
  ELSE
    Pocket COUNT wid** SEARCH-WORDLIST IF
      >CODE-GEN OFF  last>***		\ clr source inliner
    ELSE
      Pocket COUNT FORTH-WORDLIST SEARCH-WORDLIST ?undef
      >CODE-GEN OFF  last>***		\ clr source inliner
    THEN
  THEN
;

\ ===================
\ *N Wordlist Helpers
\ ===================

: WID-PRINT	\ thread nfa -- flag
\ *G Print definition while walking wordlist
   { %thread %nfa --- }
   CR
   %nfa .DWORD SPACE
   %nfa COUNT TYPE
   True
;

: WID-WORDS	\ wid --
\ *G Print words in wid wordlist
   { %wid --- }
   ASSIGN WID-PRINT %wid WALKWORDLIST
;

: **-(XT-FIND)	\ addr cnt --- xt flag|false
\ *G Find the xt of the definition at addr cnt in the hidden wordlist two-stars
   wid** SEARCH-WORDLIST
;

: ***-(XT-FIND)	\ addr cnt --- xt flag|false
   \ *G Find the xt of the definition at addr cnt in the hidden wordlist three-stars
   wid*** SEARCH-WORDLIST
;

: HIDDEN-(XT-FIND)	\ addr cnt --- xt flag|false
   \ *G Find the xt of the definition at addr cnt in the hidden wordlist
   widHidden SEARCH-WORDLIST
;

\ ***************************
\ *S Error and Abort Messages
\ ***************************

\ ==================
\ *N Message numbers
\ ==================
\ *E  4         3         2         1
\ ** --------- --------- --------- ---------
\ ** 32   28   24   20   16   12   8    4
\ ** 0000 0000 0000 0000 0000 0000 0000 0000

VARIABLE NEXT-ERR **	\ -- addr
\ *G Global variable with next error number
  NEXT-ERR OFF                                               \ errors start at 0

: #CCS-ERROR	( <name> --- n )
\ *G Define next error number. Error text can be accessed using
\ ** CCS-ERR-MESSAGE
   NEXT-ERR @
   $1FFF:FFFF AND APPLICATION_ERROR_MASK OR           \ mark as application error
   [COMPILE] CONSTANT                                   \ create a long variable
   1 NEXT-ERR +!
; IMMEDIATE

\ =================
\ *N Static Buffers
\ =================

255 CONSTANT Abort$Size

CREATE Abort$	\ -- addr
\ *G Buffer for abort message
       Abort$Size 1+ 1+        \ 1 byte for count, 1 byte for zero term string
       ALLOT&ERASE

CREATE AbortFilename$	\ -- addr
\ *G Buffer for file name where error occured
       Abort$Size 1+  1+       \ 1 byte for count, 1 byte for zero term string
       ALLOT&ERASE

CREATE AbortWordName$	\ -- addr
\ *G Buffer for FORTH word name where error occured
       Abort$Size 1+  1+       \ 1 byte for count, 1 byte for zero term string
       ALLOT&ERASE

CREATE AbortCaption$	\ -- addr
\ *G Buffer for caption bar when error is displayed
       Abort$Size 1+ 1+        \ 1 byte for count, 1 byte for zero term string
       ALLOT&ERASE

CREATE AbortId$	\ -- addr
\ *G Buffer for error number message
       Abort$Size 1+ 1+        \ 1 byte for count, 1 byte for zero term string
       ALLOT&ERASE

CREATE bufException$	\ -- addr
\ *G Buffer for exception error message
       Abort$Size 1+ 1+        \ 1 byte for count, 1 byte for zero term string
       ALLOT&ERASE

CREATE Exception$	\ -- addr
\ *G Buffer for exception string
       Abort$Size 1+ 1+        \ 1 byte for count, 1 byte for zero term string
       ALLOT&ERASE

CREATE bufUserError$	\ -- addr
\ *G Buffer for user error message
       Abort$Size 1+ 1+        \ 1 byte for count, 1 byte for zero term string
       ALLOT&ERASE

\ ======================
\ *N Abort Buffer Access
\ ======================

: CLR-ABORT-FNAME$	( --- )
\ *G Clear AbortFilename$ text buffer
  AbortFilename$ OFF  ;

: >ABORT-FNAME$	( adr cnt --- )
\ *G Assign addr cnt to AbortFilename$ text buffer
   Abort$Size MIN #0 MAX  AbortFilename$ PLACE
;

: +ABORT-FNAME$	( addr cnt --- )
\ *G Append string to AbortFilename$ text buffer
    Abort$Size AbortFilename$ C@ - MIN #0 MAX
    AbortFilename$  APPEND
;

\ ====================================
\ *N AbortWordName$ Text Buffer Access
\ ====================================

: CLR-ABORT-FUNC$	( --- )
\ *G Clear AbortWordName$ text buffer
   AbortWordName$ OFF
;

: >ABORT-FUNC$	( addr cnt --- )
\ *G Assign addr cnt to AbortWordName$
   Abort$Size MIN #0 MAX AbortWordName$ PLACE
;

: +ABORTFUNC$	( addr cnt --- )
\ *G Append string to AbortWordName$
   Abort$Size AbortWordName$ C@ - MIN #0 MAX
   AbortWordName$ APPEND
;

\ ============================
\ *N Abort$ Text Buffer Access
\ ============================

: CLR-ABORT$	( --- )
\ *G Clear Abort$ text buffer
   Abort$ OFF
;

: >ABORT$	( adr cnt -- )
\ *G Assign addr cnt to Abort$ text buffer
   Abort$Size MIN #0 MAX Abort$ PLACE
;

: +ABORT$       ( adr cnt -- )
\ *G Append string to Abort$ text buffer
   Abort$Size Abort$ C@ - MIN #0 MAX
   Abort$ APPEND
;

\ ===================================
\ *N AbortCaption$ Text Buffer Access
\ ===================================

: CLR-ABORT-CAPTION$	( --- )
\ *G Clear AbortCaption$ text buffer
   AbortCaption$ OFF
;

: >ABORTCAPTION$        ( adr cnt -- )
\ *G Assign addr cnt to AbortCaption$ text buffer
   Abort$Size MIN #0 MAX AbortCaption$ PLACE
;

: +ABORTCAPTION$        ( addr cnt -- )
   Abort$Size AbortCaption$ C@ - MIN #0 MAX
   AbortCaption$ APPEND
;

\ =========================
\ *N AbortId$ Buffer Access
\ =========================

: CLR-ABORT-ID$         ( --- )
\ *G Clear AbortId$
   AbortId$ OFF
;

: >ABORT-ID$	( adr cnt -- )
\ *G Assign addr cnt to AbortId$
   Abort$Size MIN 0 MAX AbortId$ PLACE
;

: +ABORT-ID$    ( addr cnt --- )
\ *G Add string at addr cnt to AbortId$ buffer.
   Abort$Size AbortId$ C@ - MIN 0 MAX
   AbortId$ APPEND
;

\ ===========================
\ *N Exception$ Buffer Access
\ ===========================

: CLR-EXCEPT$   ( --- )
\ *G Clear Exception$
   Exception$ OFF
;

: CLR-EXCEPT-BUF$	( --- )
\ *G Clear bufException$
   bufException$ OFF
;

\ ===================================
\ *N User Error Message Buffer Access
\ ===================================

: CLR-USER-BUF$	( --- )
\ *G Clear user error buffer
   bufUserError$ OFF
;

\ ========================================
\ *N General Abort and Error Buffer Access
\ ========================================

: CLR-ABORT-BUF	( --- )
\ *G Clear all abort message buffers
   CLR-ABORT$
   CLR-ABORT-FNAME$
   CLR-ABORT-FUNC$
   CLR-ABORT-CAPTION$
   CLR-ABORT-ID$
   CLR-EXCEPT$
   CLR-EXCEPT-BUF$
   CLR-USER-BUF$
  ;

DEFER DISP-ABORT	\ --
\ *G The error display vector
  ASSIGN NOOP TO-DO DISP-ABORT

: CONSOLE-DISP-ABORT	( --- )
\ *G Print the ERROR buffers to the current I/O console
   AbortCaption$ C@			\ abort context set ??
   IF CR AbortCaption$ COUNT TYPE THEN
   Abort$ C@                            \ abort message set ??
   IF CR Abort$ COUNT TYPE THEN
   AbortFilename$ C@                    \ abort filename set ??
   IF CR AbortFilename$ COUNT TYPE THEN
   AbortWordName$ C@                    \ abort function set ??
   IF CR AbortWordName$ COUNT TYPE THEN
   AbortId$ C@                          \ abort message id set ??
   IF CR AbortId$ COUNT TYPE THEN
   CLR-ABORT-BUF
;
ASSIGN CONSOLE-DISP-ABORT TO-DO DISP-ABORT


\ ****************************
\ *S Guard the Parameter stack
\ ****************************

[undefined] [GuardSP [if]
8 cells constant /guard	\ -- n
\ *G The amount by which the stack is guarded.

: [GuardSP	\ -- ; R: -- oldS0 oldSP
\ *G Reserves guard space on the data stack and resets the data
\ ** stack pointers as if the stack was empty.
  ?comp
  postpone S0 postpone @ postpone >R
  postpone SP@ postpone >R                     \ save current stack pointers
  postpone sp@ postpone /guard postpone -
  postpone dup postpone s0 postpone ! postpone sp!   \ drop stack *** CARNAL ***
; immediate

: GuardSP]	\ -- ; R: oldS0 oldSP --
\ *G Restore the stack condition saved by *\fo{[GuardSP}.
  postpone R> postpone SP!                            \ reset stack pointers
  postpone R> postpone S0 postpone !
; immediate
[then]


\ ****************
\ *S Literal Lists
\ ****************

: ([LIT-LIST)	\ -- ; R: -- depth
   R> DEPTH >R >R
;

: [LIT-LIST	\ --- hival
\ *G Open Literal list. Must use LIT-LIST] to close list. [imm]
   ?COMP
   $7FFFFFFF
   POSTPONE ([LIT-LIST)
; IMMEDIATE

: (LIT-LIST])	\ -- change ; R: depth --
   R> DEPTH R> - SWAP >R
;

: LIT-LIST]                                             \ hival --- | n0....nx x
\ *G Close literal list. [imm]
   ?COMP
   POSTPONE (LIT-LIST])
   $7FFFFFFF ?PAIRS
; IMMEDIATE


\ ****************
\ *S Miscellaneous
\ ****************

: (COMP-INLINE-STRUCTURE)	\ size -- addr ; -- addr cnt
\ *G Compile time helper definition for inline structures. Returns address
\ ** of the structure during compile time. Returns the address and count of
\ ** the structure in the image during run time.
  postpone (w")  dup w,			\ helper and size
  here >r  allot  0 w,  align  r>	\ double zero terminated, aligned
;

((
: INVALID-XT$                                                      ( --- addr$ )
   c" <<Invalid XT>>"
;

: XT>NFA                                                          ( xt --- nfa )
   \ *G Convert xt to name field address. If non FORTH xt then return pointer
   \ ** to INVALID-XT$.
   { %xt --- }
   %xt INFORTH?
   IF   %xt >NAME
   ELSE INVALID-XT$
   THEN
;
))

: XT>NFA	\ xt --- nfa
\ *G Convert xt to name field address. If non FORTH xt then return pointer
\ ** to INVALID-XT$.
  dup INFORTH? IF
    >NAME
  ELSE
    drop c" <<Invalid XT>>"
  THEN
;

: NUMERIC?                                                     ( char --- flag )
   \ *G Is ascii code a base 10 digit
   [CHAR] 0 [CHAR] 9 WITHIN?
;

: UPC-ALPHA?                                                   ( char --- flag )
   \ *G Is ascii code an uppercase alpha char
    [CHAR] A [CHAR] Z WITHIN?                              \ uppercase ascii key
;

: LWC-ALPHA?                                                   ( char --- flag )
   \ *G Is ascii code an lowercase alpha char
   [CHAR] a [CHAR] z WITHIN?
;

\ =======
\ *> ####
\ =======

HIDE-**

