\ TypeDef.fth - CCS standard types

((
Copyright (c) 1999, 2000, 2005, 2008, 2010, 2011
Construction Computer Software Pty Ltd.
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441

net: tech-support@mpeforth.com
web: www.mpeforth.com

To do
=====
Add SUPER and SELF.
Add late binding (ideal for message handling in winprocs)

Change History
==============
))

\ ==========
\ *! typedef
\ *T Standard Types
\ ==========
\ *P The file *\i{TypeDef.fth} contains a class library.

  [FORTH]
  DECIMAL

  1  CONSTANT type_str
  2  CONSTANT type_zstr
  3  CONSTANT type_char
  4  CONSTANT type_word
  5  CONSTANT type_sword
  6  CONSTANT type_int
  7  CONSTANT type_dint
  8  CONSTANT type_d.dpl
  9  CONSTANT type_s.dpl
  10 CONSTANT type_float
  11 CONSTANT type_char(*)
  12 CONSTANT type_word(*)
  13 CONSTANT type_int(*)
  14 CONSTANT type_dint(*)
  15 CONSTANT type_s.dpl(*)
  16 CONSTANT type_d.dpl(*)
  17 CONSTANT type_float(*)
  18 CONSTANT type_str(*)
  19 CONSTANT type_sfloat
  20 CONSTANT type_ivar
  21 CONSTANT type_avar
  22 CONSTANT type_jvar
  23 CONSTANT type_lval
  24 CONSTANT type_wval
  25 CONSTANT type_cval
  26 CONSTANT type_tvar
  27 CONSTANT type_qvar

  [PRIVATE]

\ \ *S Primitives
: LONG-MAKE-INST                                                         \ n ---
   \
   [ ALSO SYSTEM ]                      \ add vocabulary for struct subprog defs
   MAKE-INST
   last-inst @ INST-ADDR ASSIGN ! XT-COMP?
   [ PREVIOUS ]
   ;


: WORD-MAKE-INST                                                        \ uw ---
   \
   [ ALSO SYSTEM ]                      \ add vocabulary for struct subprog defs
   MAKE-INST
   last-inst @ INST-ADDR ASSIGN W! XT-COMP?
   [ PREVIOUS ]
   ;

: BYTE-MAKE-INST                                                        \ uc ---
   \
   [ ALSO SYSTEM ]                      \ add vocabulary for struct subprog defs
   MAKE-INST
   last-inst @ INST-ADDR ASSIGN C! XT-COMP?
   [ PREVIOUS ]
   ;

: LONG-<FETCH>                                                         \ obj ---
   \ 32bit signded constant value fetch
   [ ALSO SYSTEM ]                      \ add vocabulary for struct subprog defs
   >R
   CASE
        R@ provider @                                 \ inspect address provider
        ['] lv    OF  R@ COMP-LOCAL-PROVIDER                             \ local
                      POSTPONE @
                  ENDOF
        #0        OF  R@ /inst-head + @ ?complit                   \ static data
                  ENDOF
                  DROP
   END-CASE
   R> DROP
   [ PREVIOUS ]
   ;

: WORD-<FETCH>                                                         \ obj ---
   \
   [ ALSO SYSTEM ]                      \ add vocabulary for struct subprog defs
   >R
   CASE
        R@ provider @                                 \ inspect address provider
        ['] lv    OF  R@ COMP-LOCAL-PROVIDER                             \ local
                      POSTPONE W@
                  ENDOF
        #0        OF R@ /inst-head + W@ ?complit                   \ static data
                  ENDOF
                  DROP
   END-CASE
   R> DROP
   [ PREVIOUS ]
   ;

: BYTE-<FETCH>                                                         \ obj ---
   [ ALSO SYSTEM ]                      \ add vocabulary for struct subprog defs
   >R
   CASE
        R@ provider @                                 \ inspect address provider
        ['] lv    OF  R@ COMP-LOCAL-PROVIDER                             \ local
                      POSTPONE C@
                  ENDOF
        #0        OF R@ /inst-head + C@ ?complit                   \ static data
                  ENDOF
                  DROP
   END-CASE
   R> DROP
   [ PREVIOUS ]
   ;

\ ***************
\ *S VALUE: types
\ ***************

  TYPE: VALUE:
  \ *G 32bit signded constant value
  \ *P Usage define - n VALUE: {instance name}
  \ *P Usage runtime - {instance name}
        4 SKIPPED
        CLASS  \ class methods
        :M <default> LONG-MAKE-INST ;M
        INST   \ instance methods
        MRUNS <default> LONG-<FETCH> STRUCTURE-METHOD
        MRUNS <blank>   OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <erase>   OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <zero>    OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <init>    OBJ-CRASH    STRUCTURE-METHOD
  END-TYPE

  TYPE: WORD-VALUE:
  \ *G 16bit unsignded constant value
  \ *P Usage define - uw WORD-VALUE: {instance name}
  \ *P Usage runtime - {instance name}
        2 SKIPPED
        CLASS  \ class methods
        :M <default> WORD-MAKE-INST ;M
        INST   \ instance methods
        MRUNS <default> WORD-<FETCH> STRUCTURE-METHOD
        MRUNS <blank>   OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <erase>   OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <zero>    OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <init>    OBJ-CRASH    STRUCTURE-METHOD
  END-TYPE

  TYPE: BYTE-VALUE:
  \ *G 8bit unsignded constant value
  \ *P Usage define - uc BYTE-VALUE: {instance name}
  \ *P Usage runtime - {instance name}
        1 SKIPPED
        CLASS  \ class methods
        :M <default> BYTE-MAKE-INST ;M
        INST   \ instance methods
        MRUNS <default> BYTE-<FETCH> STRUCTURE-METHOD
        MRUNS <blank>   OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <erase>   OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <zero>    OBJ-CRASH    STRUCTURE-METHOD
        MRUNS <init>    OBJ-CRASH    STRUCTURE-METHOD
  END-TYPE

\ *S Common class code

: LOCAL-MAKE-INST         runword                                   ( type --- )
   \ *G ODL Compile erase of local object
   \ ** Object Class specific method - <default>
   [ ALSO SYSTEM ]                      \ add vocabulary for struct subprog defs
   MAKE-INST
   State @                                       \ compiling a local instance ??
   IF   \ yes!
        last-inst @ INST-ERASE
   THEN
   [ PREVIOUS ]
   ;

\ **************
\ *S Unary Types
\ **************

\ =============
\ *N Byte types
\ =============

: C@OFF                                                        ( addr --- char )
   DUP C@ SWAP COFF ;

: C-VAR-<TOGGLE>                                                     ( obj --- )
   \ Change state of variable
   DUP C@ IF COFF ELSE CON THEN ;

  TYPE: C-VAR:
  \ *G 8bit unsigned value.
        1 SKIPPED
        CLASS \ Class methods
        INST  \ instance methods
        mruns <inc>     cincr
        mruns <dec>     cdecr
        mruns <add>     c+!
        mruns <sub>     c-!
        mruns <set>     con
        mruns <zero>    coff
        mruns <to>      c!
        mruns <default> c@
        mruns <@off>    C@OFF
        mruns <toggle>  C-VAR-<TOGGLE>
  END-TYPE

 TYPE: <C-VAR: inherits C-VAR:
  \ *G 8bit signed value
        CLASS \ Class methods
        INST  \ instance methods
        mruns <default> C@S
  END-TYPE


\ ===============
\ *N 16 bit types
\ ===============
\ -------------
\ *H Primitives
\ -------------

: TWISTW                  runword                                   ( w --- w' )
  \ *G Toggle endian of 16bit value on the pstack.
  { %word --- }
    %word $FFFF AND to %word
    %word #8 RSHIFT %word $FF AND #8 LSHIFT OR
    ;

: TW-COUNT                runword                       ( addr --- addr' u-cnt )
   \ *G Twisted WCOUNT
   WCOUNT TWISTW
   ;

: TWIST-WVAR              runword                                   ( addr --- )
   \ *G Toggle endian of 16bit value on the pstack.
   DUP >R W@ TWISTW R> W!
   ;

: W@OFF                                                           ( addr --- w )
   DUP W@ SWAP WOFF
   ;

\ -------------------
\ *N Type Definitions
\ -------------------

  TYPE: W-VAR:
  \ *G 16bit unsigned value variable
  \ *[  *\fo{W-VAR: members}
        C-VAR: x0   \ LSB
        C-VAR: x1   \ MSB
  \ *] End Doc-gen include
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <twist>    TWIST-WVAR
        MRUNS <fetch>    W@
        MRUNS <inc>      WINCR
        MRUNS <dec>      WDECR
        MRUNS <set>      WON
        MRUNS <add>      W+!
        MRUNS <sub>      W-!
        MRUNS <zero>     WOFF
        MRUNS <to>       W!
        MRUNS <default>  W@
        MRUNS <@off>     W@OFF
  END-TYPE

  TYPE: W<VAR: inherits W-VAR:
  \ *G 16bit signed value
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <default> W@S
  END-TYPE

  TYPE: WORD:
  \ *G Windows API type WORD
        2 SKIPPED
        CLASS \ Class methods
        INST  \ Instance methods
        mruns <default> W@
        mruns <to>      W!
        mruns <inc>     WINCR
        mruns <dec>     WDECR
        mruns <add>     W+!
        mruns <sub>     W-!
        mruns <zero>    WOFF
        mruns <set>     WON
  END-TYPE

\ *S L-VAR: 32Bit Signed Value Unary Object
\ *N L-VAR: Primitives
: TWISTL                   runword                                  ( n --- n' )
  \ *G Toggle endian of 32bit value on the pstack.
    DUP
    TWISTW #16 LSHIFT
    SWAP
    #16 RSHIFT TWISTW
    OR
    ;

: TWIST-LVAR               runword                                  ( addr --- )
   \ *G Toggle endian of 32bit value at address.
   DUP >R @ TWISTL R> ! ;

\ *N L-VAR: Structure Definition
  TYPE: L-VAR:
  \ *G  Structure for 32Bit Signed Value
  \ *[  *\fo{L-VAR: members}
        W-VAR: x0  \ LSW
        W-VAR: x1  \ MSW
  \ *] End Doc-gen include
  \ *E  L-VAR: Methods
  \ **       Instance methods
  \ **       1.0 Modifiers
  \ **       1.1 Arithmatic
  \ **           <inc>                                                   ( --- )
  \ **           <dec>                                                   ( --- )
  \ **           <add>                                                 ( n --- )
  \ **           <sub>                                                 ( n --- )
  \ **       1.2 Assignment
  \ **           <set>                                                   ( --- )
  \ **           <zero>                                                  ( --- )
  \ **           <to>                                                  ( n --- )
  \ **           <toggle>                                                ( --- )
  \ **           <invert>                                                ( --- )
  \ **           <twist>                                                 ( --- )
  \ **           <erase>                                                 ( --- )
  \ **           <init>                                                  ( --- )
  \ **           <blank>                                                ( ---- )
  \ **       1.3 Sign
  \ **           <sign?>                                            ( --- flag )
  \ **           <abs>                                                   ( --- )
  \ **           <negate>                                                ( --- )
  \ **           <neg>                                                   ( --- )
  \ **       2.0 Inspectors
  \ **           <fetch>                                               ( --- n )
  \ **           <default>                                             ( --- n )
  \ **           <f@>                                                  ( --- r )
  \ **           <m!>                                                    ( --- )
  \ **           <dbkey>                                                 ( --- )
  \ **           <count>                                        ( --- addr cnt )
  \ **           <sizeof>                                              ( --- n )
  \ **           <addr>                                             ( --- addr )
  \ **           <ptr>                                              ( --- addr )
  \ **           <&ptr>                                             ( --- addr )
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <twist>      TWIST-LVAR
        MRUNS <fetch>      @
        MRUNS <inc>        INCR
        MRUNS <dec>        DECR
        MRUNS <set>        ON
        MRUNS <add>        +!
        MRUNS <sub>        -!
        MRUNS <zero>       OFF
        MRUNS <to>         !
        MRUNS <default>    @
        MRUNS <@off>       @OFF
  END-TYPE

  TYPE: INT:
  \ *G Windows API type INT
        CELL SKIPPED
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <default> @
        MRUNS <to>      !
        MRUNS <inc>     incr
        MRUNS <dec>     decr
        MRUNS <add>     +!
        MRUNS <sub>     -!
        MRUNS <zero>    off
        MRUNS <set>     on
        MRUNS <twist>   TWIST-LVAR
  END-TYPE

  TYPE: R-VAR: INHERITS L-VAR:
  \ *G  Structure for 32Bit Signed resetting Value
  \ *E  R-VAR: Methods
  \ **       See L-VAR: Methods
  \ **       <default>                                                 ( --- n )
  \ **          sub classed to zero data value
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <default> @OFF
  END-TYPE

\ *S 32bit signed value and 8bit signed dpl
\ *N Primitives

: S@  RUNWORD                                                 ( addr --- n dpl )
   \ *G Fetch 32bit signed value and 8bit sign dpl at address.
   DUP >R @ R> 4+ C@S
   ;

: S! RUNWORD                                                  ( n dpl addr --- )
   \ *G Store 32bit signed value and 8bit sign dpl at address.
   DUP >R 4+ C! R> !
   ;

  TYPE: S-VAR:                                                          ( --- )
  \ *G 32bit Value and signed dpl structure
  \ *[  *\fo{S-VAR: members}
        L-VAR:  val
        <C-VAR: dpl
  \ *] End Doc-gen include
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <default> S@
        :M <zero> INST-ERASE ;M structure-method
        :M <to> S! ;M
        MRUNS <fetch> S@
  END-TYPE

\ *S 64bit signed value
\ *N Primitives

: D@		runword	( addr --- n0 n1 )
   \ *G Fetch 64bit value at address
    2@ SWAP
;

: D!		runword	( n0 n1 addr --- )
   \ *G Store 64bit value at address
   >R SWAP R> 2!
;

: TWISTD	runword	( n0\n1 --- n1'\n0' )
   \ *G Toggle endian of 64bit value on the pstack.
   TWISTL SWAP TWISTL SWAP
   ;

: TWIST-DVAR	runword	( addr --- )
   \ *G Toggle endian of 64bit value at address.
   DUP >R D@ TWISTD R> D!
   ;

\ *N D-VAR: Structure Definition

  TYPE: D-VAR:
  \ *G Structure for 64bit signed value
        \ <neg>      D-<NEG>
  \ *[  *\fo{D-VAR: members}
        L-VAR: x0  \ LSL
        L-VAR: x1  \ MSL
  \ *] End Doc-gen include
        CLASS \ class methods
        INST  \ Instance methods
        MRUNS <twist>      TWIST-DVAR
        MRUNS <default>    D@
        MRUNS <to>         D!
        MRUNS <fetch>      D@
  END-TYPE

\ *S 64bit signed value and 8bit signed dpl
\ *N Primitives

: B!		runword	( d dpl addr --- )
   \ *G Store 64bit signed value and 8bit sign dpl at address.
   DUP >R 8 + C! R> D!
   ;

: B@		runword	( addr --- d dpl )
   \ *G Fetch 64bit signed value and 8bit sign dpl at address.
   DUP >R D@ R> 8 + C@S
   ;

: TWISTB        runword	( d dpl ---  d' dpl )
  \ *G Twist B-VAR on pstack.
  >R TWISTD R>
  ;

\ *N Structure Definitions
  TYPE: B-VAR:
  \ *G  Structure for 64bit signed value and 8bit signed dpl
  \ *[  *\fo{B-VAR: members}
        D-VAR:  val
        <C-VAR: dpl
  \ *] End Doc-gen include
  \ *E  B-VAR: Methods
  \ **       Instance methods
  \ **       1.0 Modifiers
  \ **       1.1 Arithmatic
  \ **           <dec>                                                   ( --- )
  \ **           <inc>                                                   ( --- )
  \ **           <add>                                                ( b1 --- )
  \ **           <sub>                                                ( b1 --- )
  \ **       1.2 Assignment
  \ **           <twist>                                                 ( --- )
  \ **           <to>                                                 ( b1 --- )
  \ **       1.3 Sign
  \ **           <abs>                                                   ( --- )
  \ **           <neg>                                                   ( --- )
  \ **           <negate>                                                ( --- )
  \ **           <sign?>                                            ( --- flag )
  \ **       2.0 Inspectors
  \ **           <default>                                            ( --- b1 )
  \ **           <fetch>                                              ( --- b1 )
  \ **           <f@>                                                 ( --- f1 )
  \ **           <count>                                        ( --- addr cnt )
  \ **           <sizeof>                                              ( --- n )
  \ **           <addr>                                             ( --- addr )
  \ **           <ptr>                                              ( --- addr )
  \ **           <&ptr>                                             ( --- addr )
        CLASS \ Class methods
        INST  \ Instance methods
             MRUNS <to>       B!
             MRUNS <default>  B@
             MRUNS <fetch>    B@
             MRUNS <twist>    TWIST-DVAR
  END-TYPE

  TYPE: bytes:		\ size specified by n BYTES BYTE
      :m default ( n addr -- b )  + c@  ;m
      :m to      ( b n addr -- )  + c!  ;m
  END-TYPE

  TYPE: cstring:        \ size specified by BYTES before CSTRIN
        inst
        mruns to      $move
  END-TYPE

  TYPE: zstring:                         \ size specified by BYTES before ZSTRIN
        mruns to      zmove
  END-TYPE

  TYPE: (FIELD):      \ byte array, only ADDR operator, size specified by #bytes
  END-TYPE                                          \ e.g. 10 BYTES FIELD: <foo>

: FIELD:           runword                                             ( n --- )
   \ *G ODL Untyped field in structure definition.
   #BYTES                                             \ set size of the instance
   postpone (FIELD):
   ; IMMEDIATE

\ *S Three way Database Intersection structure with Instances

: J!                runword                 \ u-w m_acc# s_acc# s2_acc# addr ---
   \ Database accesion# intersection store
   DUP >R 8 + ! R@ 4+ ! R@ ! R> 12 + !
   ;

: J@                runword                  \ addr --- u-w m_acc# s_acc# s_acc#
   \ Database accesion# intersection fetch
   DUP >R 12 + @                                            \ fetch order number
   R@ @                                         \ fetch primary accession number
   R@ 4+ @                                      \ fetch secondary acesion number
   R> 8 + @                                          \ fetch 3rd accesion number
  ;

: TWIST-JVAR          runword                                         \ addr ---
   \ Database accesion# twist for string sort
   DUP >R   TWIST-LVAR                           \ twist primary accesion number
   R@ 4+   TWIST-LVAR                         \ twist secondary accession number
   R@ 8  + TWIST-LVAR                         \ twist secondary accession number
   R> 12 + TWIST-LVAR                                       \ twist order number
  ;

  TYPE: J-VAR:
        \ *G Master and secondary accesion number intersection with many
        \ ** instances. One to many with instances.
        \ *[  *\fo{J-VAR: members}
        L-VAR: macc#                                               \ master acc#
        L-VAR: sacc#                                            \ secondary acc#
        L-VAR: sacc2#                                           \ secondary acc#
        L-VAR: order#
        \ *] End Doc-gen include
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <to>        J!
        MRUNS <default>   J@
        MRUNS <twist>     TWIST-JVAR
  END-TYPE

\ *S Database Intersection structure with Instances
: A!                runword                         \ u-w m_acc# s_acc# addr ---
   \ Database accesion# intersection store
   DUP >R 4+ ! R@ ! R> 4+ 4+ !
   ;

: A@                runword                         \ addr --- u-w m_acc# s_acc#
   \ Database accesion# intersection fetch
   DUP >R 4+ 4+ @                                           \ fetch order number
   R@ @                                         \ fetch primary accession number
   R> 4+ @                                      \ fetch secondary aceesion number
 ;

: TWIST-AVAR          runword                                         \ addr ---
   \ Database accesion# twist for string sort
   DUP >R 4+ 4+ TWIST-LVAR                                  \ twist order number
   R@ TWIST-LVAR                                 \ twist primary accesion number
   R> 4+ TWIST-LVAR                            \ twist secondary accession number
 ;

  TYPE: A-VAR:
        \ *G Master and secondary accesion number intersection with many
        \ ** instances. One to many with instances.
        \ *[  *\fo{A-VAR: members}
        L-VAR: macc#                                               \ master acc#
        L-VAR: sacc#                                            \ secondary acc#
        L-VAR: order#
        \ *] End Doc-gen include
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <to>        A!
        MRUNS <default>   A@
        MRUNS <twist>     TWIST-AVAR
  END-TYPE

\ *S Database Intersection structure
: IVAR-<@>         runword                              \ addr --- m_acc# s_acc#
   \ Single Accession number Intersection Database fetch
   DUP >R @                                     \ fetch primary accession number
   R> 4+ @                                     \ fetch secondary accesion number
 ;

: IVAR-<to>        runword                              \ m_acc# s_acc# addr ---
   \ Single Accession number Intersection Database store
   DUP >R 4+ ! R> !
   ;

: TWIST-IVAR       runword                                            \ addr ---
   \ Database accesion# twist for string sort
   DUP >R 4+ TWIST-LVAR                       \ twist secondary accession number
   R> TWIST-LVAR                                 \ twist primary accesion number
   ;

  TYPE: I-VAR:
        \ *G Master and secondary accesion number intersection. Intersection
        \ ** dbase key variable. One to many.
        \ *[  *\fo{I-VAR: members}
        L-VAR: macc#                                               \ master acc#
        L-VAR: sacc#                                            \ secondary acc#
        \ *] End Doc-gen include
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <to>        IVAR-<to>
        MRUNS <default>   IVAR-<@>
        MRUNS <twist>     TWIST-IVAR
  END-TYPE

\ ---------------
\ #4 POINTER TYPE
\ ---------------
\ *S Pointer Types

  PTR: (POINTER):
  \ *G  Structure for 32Bit Unsigned long pointer to an instance of an object
  \ *E  (POINTER): Methods
  \ **       Instance methods
  \ **       1.0 Modifiers
  \ **       1.2 Assignment
  \ **           <ptr!>                                             ( addr --- )
  \ **           <ptr0>                                                  ( --- )
  \ **       2.0 Inspectors
  \ **           <default>                                          ( --- addr )
  \ **           <ptr@>                                             ( --- addr )
  \ **           <&ptr>                                         ( --- instance )
        4 SKIPPED
  END-TYPE

: POINTER:                runword                ( [instance_name] {type:} --- )
   \ *G Makes a pointer object of the type
   \ *E   POINTER: {name} [type:]
   [ ALSO SYSTEM ]
   with-inst OFF
   postpone (POINTER):                                      \ make a PTR: <name>
   ' >body dup ?type                                                \ get <type>
   DUP type-size @ 0=
   IF
         IP>NFA COUNT >ABORT-FUNC$
         S" Can not make a pointer to this type def." >ABORT$
         ACTION-OF ABORT ASSIGN (ABORT) =
         IF
               CR ." Can not make a pointer to this type def - "
               AbortWordName$ COUNT TYPE
               LATEST-NFA .NAME
        THEN
        ABORT
   THEN
   last-inst @ type-magicid !               \ change the MAGICID field of <name>
   State @                                       \ compiling a local instance ??
   IF   \ yes!
        last-inst @ INST-ERASE
   THEN
   [ PREVIOUS ]
   ; immediate

  L-VAR: gblUId  runword                               \ system global unique id

: OBJID-<set>                                                         ( obj -- )
   \
   <inc> gblUId
   gblUId 0= IF <inc> gblUId THEN
   gblUId SWAP ! ;

\ *S OBJID: 32Bit Unsigned unique identification Value

  TYPE: OBJID:
  \ *G  Structure for 32Bit UnSigned unique identification value
  \ *E  OBJID: Methods
  \ **       Instance methods
  \ **       1.0 Modifiers
  \ **           <set>                                                 ( --- )
  \ **              Increment the global unique identification value and assign
  \ **              this value to the object
  \ **           <to>                                                  ( n --- )
  \ **       2.0 Inspectors
  \ **           <default>                                             ( --- n )
  \ **           <fetch>                                               ( --- n )
  \ **           <count>                                        ( --- addr cnt )
  \ **           <sizeof>                                              ( --- n )
  \ **           <addr>                                             ( --- addr )
  \ **           <ptr>                                              ( --- addr )
  \ **           <&ptr>                                             ( --- addr )
        4 SKIPPED
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <default> @
        MRUNS <set> OBJID-<set>
        MRUNS <fetch>      @
        MRUNS <to>         !
  END-TYPE

\ *S Buffer type definition

  TYPE: sADDRLEN:
        \ *G Address and count structure
  \ *[  *\fo{sADDRLEN: members}
        L-VAR: address
        L-VAR: length
  \ *] End Doc-gen include
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <to>      D!
        MRUNS <default> D@
        MRUNS <fetch>   D@
  END-TYPE

\ *S Windows Standard Data Types
\ *[
  SYNONYM uint:       int:
  SYNONYM lptstr:     int:
  SYNONYM hmenu:      int:
  SYNONYM hbitmap:    int:
  SYNONYM handle:     int:
  SYNONYM HWND:       int:
  SYNONYM WPARAM:     int:
  SYNONYM LPARAM:     int:
  SYNONYM dword:      int:
  SYNONYM long:       int:
  SYNONYM char:       c-var:
  SYNONYM byte:       c-var:
  SYNONYM r-int:      r-var:
  SYNONYM sword:      W<VAR:
\ *]


\ FILE TYPE REQUIREMENTS

  CHAR . BYTE-VALUE: DotChar           runword
  CHAR : BYTE-VALUE: ColonChar         runword
  CHAR \ BYTE-VALUE: BackSlashChar     runword
  CHAR * BYTE-VALUE: StarChar          runword
  CHAR ! BYTE-VALUE: ShriekChar        runword
  CHAR ? BYTE-VALUE: QueryChar         runword
  CHAR / BYTE-VALUE: ForwardSlashChar  runword

: FILE-DRIVE?               runword                          \ addr cnt --- flag
   \ Does the file name address count include a valid drive specifier.
   -LEADING                \ remove leading blanks to check on pos of colon char
   2 >=                                 \ file name at least 2 charcters long ??
   IF                \ yes! file name can be tested for colonchar in position 2
       DUP >R 1+ C@ ColonChar =                            \ test for colon char
       R> C@
       UPC [CHAR] A [CHAR] Z WITHIN? AND            \ ...and valid drive specifier
   ELSE
       DROP False                          \ no! drop addr, not possible to test
   THEN
   ;

: FILE-PATH?                 runword                         \ addr cnt --- flag
   \ Does the file name address count include a path
   BackSlashChar BYTE= 0<>
   ;

: FILE-WILD?                 runword                       ( addr cnt --- flag )
   \ Check for presence of wildcard characters in adr\cnt and return true if
   \ there are.
   2DUP StarChar  BYTE=                                \ check for global wildcard
   -ROT QueryChar BYTE=                                  \ check for char wildcard
   OR 0<>
   ;

: FILE-UNC?                  runword                       ( addr cnt --- flag )
   \ Check for valid UNC path at adr\cnt and return true or false
   \ unc path = \\computer name\share name\folder\filename
   { %addr %cnt --- }
   %cnt 5 >                                        \ count must be at least 5 ??
   IF   \ yes!
        $5C 8 LSHIFT $5C OR                         \ build whack whack on stack
        %addr W@ =                         \ file name start with whack whack ??
        IF   \ yes!
             %addr %cnt 2 /STRING <to> %cnt <to> %addr     \ advance whack whack
             %addr %cnt $5C BYTE= ?DUP         \ computer and share separator ??
             IF   \ yes!
                  %addr -                                           \ calc chars
                  %addr %cnt ROT /STRING <to> %cnt <to> %addr          \ advance
                  %cnt 0>                             \ more chars to examine ??
                  IF   \ yes!
                       %addr %cnt $5C BYTE=      \ share and folder separator ??
                       IF   \ yes!
                            True                        \ possible unc path name
                       ELSE \ no!
                            False
                       THEN
                  ELSE \ no! chars finished
                       False
                  THEN
             ELSE \ no! separator not present
                  False
             THEN
        ELSE \ no! must start with whack whack
             False
        THEN
   ELSE \ no! chars not enough to contain UNC name
        False
   THEN
   ;

: ?PATH-ZFILE!              runword                             ( addr cnt --- )
   \ *G Remove trailing backslash and blank chars from a zfilename path
   { %addr %cnt --- }
   BEGIN
        %cnt 0>
   WHILE
        <dec> %cnt
        %addr %cnt +
        CASE
            DUP C@ BackSlashChar = ?OF COFF ENDOF
            DUP C@ BL            = ?OF COFF ENDOF
            DROP
            <zero> %cnt
        END-CASE
   REPEAT
   ;

: OBJ-TO   runword    ( source_addr\source_cnt\obj_addr\obj_cnt --- addr` cnt` )
   \ Move source addres count into object of count size.
   \ Return object remainder start address and count.
   \ Protected against negative source and object counts
   0 MAX 2SWAP 0 MAX 2OVER
   ROT MIN DUP >R MOVEX
   R> /STRING
   ;


\ *S String Primitives

: $-<LEN>                      runword                ( c-addr u-cnt --- u-cnt' )
   \ *G Trim string at addr cnt and return revised cnt
   -TRAILING NIP
;

: $-LEN                       runword                ( c-addr u-cnt --- u-cnt' )
   \ *G Trim string at addr cnt and return revised cnt
   -TRAILING NIP
;

: $-NULL?                       runword                ( c-addr u-cnt --- flag )
   \ *G Scan string at addr cnt and return true if oly blank char are found or
   \ ** false if any non blank chars is found.
   32 SKIP NIP 0=
   ;

: $TO$         runword         ( source_addr source_cnt dest_addr dest_cnt --- )
   \ *G Assign dest_addr for dest_cnt to source_addr for source_cnt. The
   \ ** trailing postive difference between dest_cnt and source_cnt will be
   \ ** filled with blanks.
   OBJ-TO BLANK
   ;

: (SQUEEZE)            runword           ( c-addr u-cnt char --- c-addr u-cnt' )
   \ *G Move all characters of value char to end of string.
  { %addr %cnt %char | %taddr %findadrr %remcnt --- addr' cnt' }
    %cnt 0>
    IF
        %addr to %findadrr
        %cnt  to %remcnt
        \
        BEGIN
            %findadrr %remcnt %char scan
            TO %remcnt TO %taddr
            %taddr %addr - to %cnt
            %remcnt
        WHILE
            %taddr %remcnt %char skip
            TO %remcnt TO %findadrr
            %findadrr %taddr %remcnt movex
            %taddr %remcnt +
            %findadrr %taddr - %char fill
            %taddr to %findadrr
        REPEAT
    THEN
    %addr %cnt ;

  +SHORT-BRANCHES
: (PARAGRAPH)                                                   ( addr cnt --- )
   \
   BEGIN
        DUP 1 >
   WHILE
        OVER W@ $2020 =
        IF OVER $20FF SWAP W! THEN
        1 /STRING
   REPEAT
   2DROP
   ;
  -SHORT-BRANCHES

: PARAGRAPH                 runword                   ( addr cnt --- addr cnt' )
   \ *G remove multple blank chars in situ
   2DUP (PARAGRAPH) #255 (SQUEEZE)
   ;

: $-<PARAGRAPH>                  runword                        ( addr cnt --- )
   \ *G Remove extra blanks from text.
   2DUP PARAGRAPH 2SWAP $TO$
   ;

: SQUEEZE                runword              ( c-addr u-cnt --- c-addr u-cnt' )
   \ *G Move blanks to end of string.
   #32 (SQUEEZE)
   ;

[undefined] ISOLATE-STR [if]
: ISOLATE-STR RUNWORD     ( c-add u-cnt start_char end_char --- c-addr2 u-cnt2 )
   \ *G Return the address and count of the first sub string in string at
   \ ** address count.
   SWAP                                   \ c-addr u-cnt end_char start_char ---
   2SWAP                                  \ end_char start_char c-addr u-cnt ---
   ROT                                    \ end_char c-addr u-cnt start_char ---
   SCAN                                             \ end_char c-addr2 u-cnt ---
   DUP                                                \ chars left to process ??
   IF   \ yes! end_char c-addr2 u-cnt2 ---
        OVER SWAP                          \ end_char c-addr2 c-addr2 u-cnt2 ---
        #1 /STRING                   \ jump over first character incase the same
        \ end_char c-addr2 c-addr3 u-cnt3
        ROT                                \ end_char c-addr3 u-cnt3 c-addr2 ---
        ROT                                \ end_char u-cnt3 c-addr2 c-addr3 ---
        2SWAP                              \ c-addr2 c-addr3 end_char u-cnt3 ---
        SWAP                               \ c-addr2 c-addr3 u-cnt3 end_char ---
        SCAN
        IF 1+ OVER - ELSE NIP #0 THEN
   ELSE ROT DROP
   THEN
  ;
[then]

: Z$CHAR+ RUNWORD                                             ( char zaddr --- )
   \ *G add char to end of zstring.
   ZCOUNT + TUCK C! 1+ COFF
  ;

: Z$CHAR- RUNWORD                                             ( zaddr --- char )
   \ *G remove and return char at end of zstring
   ZCOUNT DUP
   IF 1- THEN
   + DUP C@ SWAP COFF
  ;

: $-EXT$                 runword               ( addr cnt --- addr' cnt'|false )
   \ Return the addr and cnt of the file extension of a file.
   { %obj %size --- }
   %obj %size FILE-PATH?
   IF
        %obj %size TUCK + SWAP BackSlashChar -BYTE= ?DUP
        IF
             %obj - 1+ %obj %size ROT /STRING <to> %size <to> %obj
        THEN
   THEN
   %obj %size TUCK + TUCK SWAP DotChar -BYTE= ?DUP
   IF   TUCK -
   ELSE DROP False
   THEN
   ;

: Z$-TO                    runword               ( saddr scnt c-addr u-cnt --- )
   \ *G Given a source address and count and a target zero terminated string
   \ ** object. The target object u-cnt includes space for the terminator
   \ ** char. The buffer will always be zero terminated!
   \ ** Zero terminated string object Service definition.
   1- 2DUP + COFF                            \ make sure of last zero terminator
   OBJ-TO ERASE
   ;

: Z$-LEN                   runword                   ( c-addr u-cnt --- u-cnt' )
   \ *G Return address count of string or failing return count
   2DUP 0 SCAN                                      \ scan for zero in string ??
   IF   NIP SWAP -                                         \ found! fetch string
   ELSE DROP NIP THEN                              \ not found! give long string
   ;

: Z$-COUNT                runword             ( c-addr u-cnt --- c-addr u-cnt' )
   \ *G Return address count of string or failing return count.
   2DUP 0 SCAN                                      \ scan for zero in string ??
   IF   NIP OVER -                                         \ found! fetch string
   ELSE DROP THEN                                  \ not found! give long string
   ;

: Z$-LFT                   runword                  ( u-cnt1 c-addr u-cnt2 --- )
   \ *G Make the zero terminated string c-addr u-cnt2 u-cnt1 long preserving
   \ ** the current values to the left of u-cnt1 or fill with blanks if
   \ ** u-cnt1 less than the current zero terminated length of the string.
   { %u %addr %cnt --- }
   %u #0 MAX %cnt 1- MIN <to> %u
   %addr %cnt Z$-LEN >R
   %u R@  >
   IF
        %addr %u R@ /STRING BOUNDS
        ?DO
             BL I C!
        LOOP
   THEN
   %addr %u + COFF
   R> DROP
   ;

: Z$-RHT                    runword                 ( u-cnt1 c-addr u-cnt2 --- )
   \ *G Make the zero terminated string c-addr u-cnt2 u-cnt1 long preserving
   \ ** the current values to the right of u-cnt1 or fill with blanks if
   \ ** u-cnt1 less than the current zero terminated length of the string.
   { %u %addr %cnt --- }
   %u #0 MAX %cnt 1- MIN <to> %u
   %addr %cnt Z$-LEN >R
   %u R@ <
   IF
        %addr R@ DUP %u - /STRING %addr %cnt Z$-TO
   ELSE
        %u R@ <>
        IF
             %addr R@ %addr %cnt %u R@ - /STRING Z$-TO
             %addr %u R@ - BLANK
        THEN
        %addr %u + COFF
   THEN
   R> DROP
   ;

: Z$-CONCAT-             runword               ( s_addr s_cnt c-addr u-cnt --- )
   \ *G Concat source addr and cnt to zero terminated string object
   2SWAP 0 MAX 2 PICK MIN
   ?DUP                                                   \ any chars to copy ??
   IF
       2SWAP                                              \ yes! get target near
       TUCK
       Z$-COUNT TUCK + >R                             \ calculate target address
       - DUP                                             \ calculate target size
       1 >                             \ any space in the object to copy data ??
       IF
          R> SWAP
          Z$-TO
       ELSE
           R> DROP
           3DROP
       THEN
   ELSE
       3DROP
   THEN
   ;

: Z$-<CONCAT>               runword              ( saddr scnt c-addr u-cnt --- )
   \ *G Concat source addr and cnt to zero terminated string object.
   { %addr %cnt %obj %objcnt | %objsize --- }
   %cnt 0 MAX %objcnt 1- MIN TO %cnt
   %cnt                                                   \ any chars to copy ??
   IF
       %obj %objcnt Z$-LEN TO %objsize                                    \ yes!
       %obj %objsize                                  \ source addr and count...
       %obj %objcnt %cnt /string                  \ ...target addr and count ...
       Z$-TO                                    \ ...move string within himself.
       %addr %obj %cnt MOVEX                                        \ add string
   THEN
   ;

: Z$-BLANK                                                  ( c-addr u-cnt --- )
   \ *G Blank zero terminated string object.
   \ ** Zero terminated string specific method definition.
   1- 2DUP + COFF                            \ make sure of last zero terminator
   BLANK
   ;

: Z$-SQUEEZE                runword                         ( c-addr u-cnt --- )
   \ *G Remove blanks from zero terminated string object.
   2DUP Z$-COUNT SQUEEZE 2SWAP Z$-TO
   ;

: Z$-LOWER                                                  ( c-addr u-cnt --- )
   \ *G Lowercase zero terminated string object.
   Z$-COUNT LOWER
   ;

: Z$-UPPER                                                  ( c-addr u-cnt --- )
   \ *G Uppercase zero terminated string object.
   Z$-COUNT UPPER
   ;

: Z$-TRIM                     runword                       ( c-addr u-cnt --- )
   \ *G Remove leading and trailing blanks from zero terminated string object.
   2DUP                                                   \ obj addr and obj cnt
   Z$-COUNT                                              \ obj data addr and cnt
   $-TRIM 2SWAP Z$-TO
   ;

: Z$-TRAILING                    runword                    ( c-addr u-cnt --- )
   \ *G Remove trailing blanks from zero terminated string object.
   2DUP                                                   \ obj addr and obj cnt
   Z$-COUNT                                              \ obj data addr and cnt
   -TRAILING 2SWAP Z$-TO
   ;

: Z$-LEADING                   runword                      ( c-addr u-cnt --- )
   \ *G Remove leading blanks from zero terminated string object.
   2DUP                                                   \ obj addr and obj cnt
   Z$-COUNT                                              \ obj data addr and cnt
   -LEADING 2SWAP Z$-TO
   ;

: Z$-PATH?                   runword                   ( c-addr u-cnt --- flag )
   \ *G Does the zero terminated string contain a file path.
   Z$-COUNT FILE-PATH?
   ;

: Z$-DRIVE?                   runword                  ( c-addr u-cnt --- flag )
   \ *G Does the zero terminated string contain a drive reference.
   Z$-COUNT FILE-DRIVE?
   ;

: Z$-UNC?                    runword                   ( c-addr u-cnt --- flag )
   Z$-COUNT FILE-UNC?
   ;

: Z$-<TRIM_COUNT>    runword                         ( c-addr --- c-addr u-cnt )
   ZCOUNT $-TRIM
   ;

: Z$-<LEN>                                                  ( c-addr --- u-cnt )
   \ *G Return count of zero terminated object.
   ZCOUNT NIP
   ;

: Z$-<NULL?>            runword                              ( c-addr --- flag )
   \ *G Return true if zero terminated object is null else false.
   C@ 0= ;

: Z$-<UNC?>               runword                            ( c-addr --- flag )
   \ *G Return true if zero terminated string qualifies as a valid UNC path
   ZCOUNT FILE-UNC?
   ;

: Z$-<PAD>             runword                       ( char cnt obj objcnt --- )
   \ Concat source addr and cnt to zero terminated string object addr cnt
   { %char %cnt %obj %objcnt | %objsize --- }
   %cnt 0 MAX %objcnt 1- MIN TO %cnt
   %cnt
   IF
       %obj %objcnt Z$-LEN TO %objsize                                    \ yes!
       %obj %objsize                                  \ source addr and count...
       %obj %objcnt %cnt /string                  \ ...target addr and count ...
       Z$-TO                                    \ ...move string within himself.
       %obj %cnt %char FILL
   THEN
   ;

: Z$-<PAD->            runword                       ( char cnt obj objcnt --- )
   \ Pad string object with char at end for count or remaining data space in
   \ string.
   { %char %cnt %obj %objcnt | %diff %taddr --- }
   %cnt 0 MAX %objcnt 1- MIN TO %cnt                     \ calc max chars to pad
   %cnt                                                    \ any chars to pad ??
   IF
        %obj %objcnt 2DUP Z$-LEN /STRING        \ yes! calc remaining data space
        TO %diff TO %taddr
        %diff 1 >                                  \ any data space in string ??
        IF
             %taddr %diff erase                      \ yes! null term the string
             %taddr %diff 1- %cnt min %char FILL               \ fill data space
        THEN
   THEN
   ;

: Z$-<-PATH>            runword                              ( obj obj_cnt --- )
   \ *G remove path from file name
   \ ** Zero terminated file name object method
   { %obj %cnt --- }
   %obj %cnt Z$-COUNT
   TUCK + SWAP BackSlashChar -BYTE= ?DUP
   IF 1+ %obj TUCK - BLANK THEN
   %obj %cnt Z$-COUNT
   TUCK + SWAP ColonChar -BYTE= ?DUP
   IF 1+ %obj TUCK - BLANK THEN
   %obj %cnt Z$-TRIM
   ;

: Z$-<-ext>                 runword                          ( obj obj_cnt --- )
   \ Remove file extension if any of zero terminated file object at data addr
   \ and cnt length
   { %obj %size --- }
   %obj %size Z$-COUNT FILE-PATH?
   IF
        %obj %size TUCK + SWAP BackSlashChar -BYTE= ?DUP
        IF
             %obj - 1+ %obj %size ROT /STRING <to> %size <to> %obj
        THEN
   THEN
   %obj %size TUCK + SWAP DotChar -BYTE= ?DUP
   IF COFF THEN
   ;

: Z$-<+EXT>                  runword                ( addr cnt obj obj_cnt --- )
 \ remove any leading dotchars from source extension
   { %addr %cnt %obj %size | %zext[ #256 ] --- }
   %addr %cnt DotChar SKIP %zext[ #256 Z$-TO
   %obj %size Z$-COUNT Z$-<-ext>
   <count> DotChar %obj %size Z$-CONCAT-
   %zext[ ZCOUNT %obj %size Z$-CONCAT-
   ;

: Z$-[COUNT]                                              \ obj_addr obj_cnt ---
   \ Execute or compile count on zero terminated string object
   State @
   IF
        POSTPONE Z$-COUNT
   ELSE
        Z$-COUNT
   THEN
   ;

: Z$-[CONCAT-]                                                      ( inst --- )
   \ Execute or compile <CONCAT-> on zero terminated string object
   State @
   IF
        postpone Z$-CONCAT-
   ELSE
        Z$-CONCAT-
   THEN
   ;

: Z$-[CONCAT]                                                       ( inst --- )
   \ Execute or compile <CONCAT> on zero terminated string object
   State @
   IF
        postpone Z$-<CONCAT>
   ELSE
        Z$-<CONCAT>
   THEN
   ;

: Z$-[PAD]                                                          ( inst --- )
   \ Execute or compile <PAD> on zero terminated string object
   State @
   IF
        postpone Z$-<PAD>
   ELSE
        Z$-<PAD>
   THEN
   ;

: Z$-[ADD]                                           ( addr cnt zaddr zcnt --- )
   \
   \ Specific method - <add>
   State @
   IF
        postpone Z$-CONCAT-
   ELSE
        Z$-CONCAT-
   THEN
   ;

: Z$-[TO]                                                           ( inst --- )
   \ Execute or compile
   State @
   IF
        postpone Z$-TO
   ELSE
        Z$-TO
   THEN
   ;

: Z$-[TRIM]                                                           \ inst ---
   \ Execute or compile <trim> on zero terminated string object
   State @
   IF
        POSTPONE Z$-TRIM
   ELSE
        Z$-trim
   THEN
   ;

: Z$-[TRIM-TRAILING]                                                  \ inst ---
   \ Execute or compile <trim-trailing> on zero terminated string object
   State @
   IF
        POSTPONE Z$-TRAILING
   ELSE
        Z$-TRAILING
   THEN
   ;

: Z$-[TRIM-LEADING]                                                   \ inst ---
   \ Execute or compile <trim-leading> on zero terminated string object
   State @
   IF
        POSTPONE Z$-LEADING
   ELSE
        Z$-LEADING
   THEN
   ;

: Z$-[COPY]                                                 ( obj obj_cnt  --- )
   \ Execute or compile <copy> on zero terminated string object
   State @
   IF
        postpone tuck postpone Z$-TO
   ELSE
        tuck Z$-TO
   THEN
   ;

: Z$-[LEN]                                                           ( obj --- )
   \ Forth State smart
   State @
   IF
        POSTPONE Z$-<LEN>
   ELSE
        Z$-<LEN>
   THEN
   ;

: Z$-[NULL?]                                                         ( obj --- )
   \ Forth State smart
   State @
   IF
        POSTPONE Z$-<NULL?>
   ELSE
        Z$-<NULL?>
   THEN
   ;

: Z$-[-EXT]                                                  ( obj obj_cnt --- )
   \
   Z$-[COUNT]
   State @
   IF
        POSTPONE Z$-<-EXT>
   ELSE
        Z$-<-EXT>
   THEN
   ;

: Z$-[EXT$]                                                  ( obj obj_cnt --- )
   \
   State @
   IF
        POSTPONE ZCOUNT
        POSTPONE $-EXT$
   ELSE
        ZCOUNT $-EXT$
   THEN
   ;

: Z$-[+EXT]                                                  ( obj obj_cnt --- )
   \
   State @
   IF
        POSTPONE Z$-<+EXT>
   ELSE
        Z$-<+EXT>
   THEN
   ;

: Z$-[-PATH]                                                 ( obj obj_cnt --- )
   \
   State @
   IF
        POSTPONE Z$-<-PATH>
   ELSE
        Z$-<-PATH>
   THEN
   ;

: Z$-[WILD?]                                                 ( obj obj_cnt --- )
   \
   Z$-[COUNT]
   State @
   IF
        POSTPONE FILE-WILD?
   ELSE
        FILE-WILD?
   THEN
   ;

: NEW-INSTANCE          runword                           ( bpe #bytes obj --- )
   \ *G ODL Create a new instance with #bytes and bpe
   \ ** Object Class specific method - <default>
   [ ALSO SYSTEM ]
   >R
   #BYTES                                             \ set size of the instance
   #BYTES-PER-ELM                                        \ set bytes per element
   R> MAKE-INST                                              \ make the instance
   [ PREVIOUS ]
   ;

: MAT-NEW-INSTANCE       runword                          ( bpe #bytes obj --- )
   \ *G ODL Create a new instance with #bytes and bpe
   \ ** Object Class specific method - <default>
   [ ALSO SYSTEM ]
   >R
   #BYTES                                             \ set size of the instance
   #BYTES-PER-ELM                                        \ set bytes per element
   R> LOCAL-MAKE-INST                                        \ make the instance
   [ PREVIOUS ]
   ;

: NEW-ITEM               runword                              ( #bytes obj --- )
   \ *G ODL Make a new item of obj with #bytes size
   \ ** Object Class specific method - <default>
   [ ALSO SYSTEM ]
   SWAP
   #BYTES                                             \ set size of the instance
   MAKE-INST
   [ PREVIOUS ]
   ;

: ZFILENAME-<PATH>        runword                                 ( c-addr --- )
   \ *G Remove file name from file object leaving only the file path.
   \ ** Zero terminated file name object specific method - <path>
   DUP >R
   ZCOUNT
   TUCK + SWAP BackSlashChar -BYTE= ?DUP              \ any path in file name ??
   IF   \ yes!
        COFF
        R> ZCOUNT ?PATH-ZFILE!                 \ remove trailing backslash chars
   ELSE \ no!
        R> COFF                                         \ zero complete filename
   THEN
   ;

: Z$-<UPPER>                                                         ( obj --- )
   ZCOUNT UPPER
   ;

: Z$-REMOVE-EXTRA-BLANKS                                           ( zstr> --- )
    ZCOUNT 2DUP REMOVE-XTRA-BLANKS 2SWAP OBJ-TO DROP COFF
  ;

: Z$-<PARAGRAPH>             runword                            ( addr cnt --- )
   \
   { %addr %cnt --- }
   %addr %cnt Z$-COUNT (PARAGRAPH)
   %addr %cnt Z$-COUNT #255 (SQUEEZE) %addr %cnt Z$-TO
   ;

: Z$-[PARAGRAPH]                                             ( obj obj_cnt --- )
   State @
   IF
        POSTPONE Z$-<paragraph>
   ELSE
        Z$-<paragraph>
   THEN
   ;

: Z$-[BLANK]                                                 \ inst_addr cnt ---
   State @
   IF
        POSTPONE Z$-BLANK
   ELSE
        Z$-BLANK
   THEN
   ;

  DEFER ZFILENAME-<+PATH>    runword                                  \ obj> ---
  ASSIGN DROP TO-DO ZFILENAME-<+PATH>

  DEFER ZFILENAME-<+CJJ>     runword                                  \ obj> ---
  ASSIGN DROP TO-DO ZFILENAME-<+CJJ>

  DEFER CHARLOWER                                                ( addr cnt --- )
  \ *G Locale sensitive lower case
  \ *[
  ASSIGN LOWER TO-DO CHARLOWER
  \ *]

: Z$-<ANSI_LOWER>                                                    ( obj --- )
   ZCOUNT CHARLOWER
   ;

   DEFER CHARUPPER                                               ( addr cnt --- )
   \ *G Locale sensitive upper case
   \ *[
   ASSIGN UPPER TO-DO CHARUPPER
   \ *]

: Z$-<ANSI_UPPER>                                                    ( obj --- )
   ZCOUNT CHARUPPER
   ;

\ *********************************
\ *S Zero terminated string classes
\ *********************************

  Type: CLASSZSTR:
  \ *G Class template for zero terminated string type: definitions. Use
  \ ** only when defining a new zero terminated string type:
  \ *E  CLASSZSTR: Methods
  \ **       Instance methods
  \ **       File Methods
  \ **         1.0 Modifiers
  \ **             <-ext>                                                ( --- )
  \ **             <+ext>                                       ( addr cnt --- )
  \ **             <squeeze>                                             ( --- )
  \ **         2.0 Inspectors
  \ **             <wild?>                                          ( --- flag )
  \ **             <path?>                                          ( --- flag )
  \ **             <drive?>                                         ( --- flag )
  \ **             <unc?>                                           ( --- flag )
  \ **       Standard String methods
        0 SKIPPED
        CLASS \ Class methods
\           MRUNS <m!> Z$-class-<m!>
        INST  \ Instance methods
        \ ------------
        \ File Methods
        \ ------------
        \ Invalid methods
        :M <wild?> OBJ-CRASH ;M structure-method
        \
        :M <path?>      inst-count postpone Z$-PATH?   ;M structure-method
        :M <drive?>     inst-count postpone Z$-DRIVE?  ;M structure-method
        MRUNS <unc?>    Z$-<UNC?>
        :M <-ext>       inst-count Z$-[-EXT]           ;M structure-method
        :m <ext$>       inst-addr  Z$-[EXT$]           ;M structure-method
        :m <+ext>       inst-count Z$-[+EXT]           ;M structure-method
        :m <-path>      inst-count Z$-[-PATH]          ;M structure-method
        mruns <path>    zfilename-<path>
        mruns <+path>   ZFILENAME-<+PATH>
        mruns <+cjj>    ZFILENAME-<+CJJ>
        :M <squeeze>    inst-count postpone Z$-SQUEEZE ;M structure-method
        \ -----------------------
        \ Standard String methods
        \ -----------------------
        :M <paragraph>  inst-count Z$-[PARAGRAPH] ;M structure-method
        MRUNS <upper>   Z$-<UPPER>
        MRUNS <ansi_upper> Z$-<ANSI_UPPER>
        MRUNS <ansi_lower> Z$-<ANSI_LOWER>
        :M <lower>      inst-count postpone Z$-LOWER   ;M structure-method
        :M <copy>       inst-count Z$-[COPY]           ;M structure-method
        :M <concat->    inst-count Z$-[CONCAT-]        ;M structure-method
        :M <concat>     inst-count Z$-[CONCAT]         ;M structure-method
        :M <trim>       inst-count Z$-[TRIM]           ;M structure-method
        :M <trim-trailing> inst-count Z$-[TRIM-TRAILING] ;M structure-method
        :M <trim-leading> inst-count Z$-[TRIM-LEADING] ;M structure-method
        :M <len>        inst-addr  Z$-[LEN]            ;M structure-method
        :M <null?>      inst-addr  Z$-[NULL?]          ;M structure-method
        MRUNS <count>   ZCOUNT
        MRUNS <trim_count> Z$-<TRIM_COUNT>
        :M <ptr@>       inst-addr ;M structure-method
        :M <data>       inst-addr ;M structure-method
         MRUNS <init> COFF
        :M <$lft> inst-count postpone Z$-LFT ;M structure-method
        :M <$min> inst-count postpone Z$-LFT ;M structure-method
        :M <$rht> inst-count postpone Z$-RHT ;M structure-method
        :M <$min-> inst-count postpone Z$-RHT ;M structure-method
        :M <add> inst-count Z$-[ADD] ;M structure-method
        :M <to>         inst-count Z$-[TO]             ;M structure-method
        :M <addr&size>  inst-count ;M structure-method
  END-TYPE

  TYPE: CLASSZFILE:
  \ *G Class template for zero terminated file name string type: definitions.
  \ ** Use only when defining a new zero terminated string type:
        0 SKIPPED
        CLASS \ class methods
        INST  \ Instance methods
        :M <wild?>      inst-count Z$-[WILD?]          ;M   structure-method
        :M <path?>      inst-count postpone Z$-PATH?   ;M   structure-method
        :M <drive?>     inst-count postpone Z$-DRIVE?  ;M   structure-method
        :M <unc?>       inst-count postpone Z$-UNC?    ;M   structure-method
        :M <-ext>       inst-count Z$-[-EXT]           ;M   structure-method
        :M <ext$>       inst-addr  Z$-[EXT$]           ;M   structure-method
        :m <+ext>       inst-count Z$-[+EXT]           ;M   structure-method
        :m <-path>      inst-count Z$-[-PATH]          ;M   structure-method
        MRUNS <path>    ZFILENAME-<path>
        MRUNS <+path>   ZFILENAME-<+PATH>
        MRUNS <+cjj>    ZFILENAME-<+CJJ>
        :M <squeeze>    inst-count postpone Z$-SQUEEZE ;M   structure-method
        :M <paragraph>  inst-count Z$-[PARAGRAPH] ;M        structure-method
        :M <copy>       inst-count Z$-[COPY]           ;M   structure-method
        :M <concat->    inst-count Z$-[CONCAT-]        ;M   structure-method
        :M <concat>     inst-count Z$-[CONCAT]         ;M   structure-method
        :M <trim>       inst-count Z$-[TRIM]           ;M   structure-method
        :M <trim-trailing> inst-count Z$-[TRIM-TRAILING] ;M structure-method
        :M <trim-leading> inst-count Z$-[TRIM-LEADING] ;M   structure-method
        :M <len>        inst-addr  Z$-[LEN]            ;M   structure-method
        :M <null?>      inst-addr  Z$-[NULL?]          ;M   structure-method
        :M <blank>      inst-count Z$-[BLANK]          ;M   structure-method
        MRUNS <count>   ZCOUNT
        MRUNS <trim_count> Z$-<TRIM_COUNT>
        :M <ptr@>       inst-addr ;M                       structure-method
        :M <data>       inst-addr ;M                       structure-method
        MRUNS <init> COFF
        :M <$lft> inst-count postpone Z$-LFT ;M            structure-method
        :M <$min> inst-count postpone Z$-LFT ;M            structure-method
        :M <$rht> inst-count postpone Z$-RHT ;M            structure-method
        :M <$min-> inst-count postpone Z$-RHT ;M           structure-method
        :M <add> inst-count Z$-[ADD] ;M                    structure-method
        :M <addr&size>  inst-count ;M                      structure-method
  END-TYPE

  TYPE: Z$-VAR: INHERITS CLASSZSTR:
  \ *G Zero terminated string structure.
  \ **   Z$-VAR: requires a size on the parameter stack and therefore can
  \ **    only be used as a static instance or a structure member instance.
  \ **    Class defintions that require a size parameter stack input during
  \ **    execution can not be used as local instances. To use as a local
  \ **    instance, create a new type defintion using the class as a member
  \ **    instance and then use the new type def as the local instance.
  \ *E  Z$-VAR: Members
  \ **          classzstr: zstrdata
  \ **  Z$-VAR: Usage
  \ **   Static instance - 8 Z$-var: zMyVar
  \ **   Member instance example
  \ **     TYPE: Mytype:
  \ **           24 z$-VAR: zMyvar
  \ **     END-TYPE
  \ **  Z$-VAR: Methods
  \ **       Instance methods
  \ **       1.0 Modifiers
  \ **       2.0 Inspectors
  \ **           <default>                                          ( --- addr )
  \ **              default method has not been over loaded. The default
  \ **              instance method of a class is to supply the data address
  \ **              of the instance.
  \ **       See CLASSZSTR: for more methods
  \ **       Class Methods
  \ **           <default>                                          ( --- addr )
        classzstr: zstrdata
        CLASS \ class methods
        MRUNS <default> new-item
        INST  \ Instance methods
  END-TYPE

  TYPE: zBuf: INHERITS CLASSZSTR:
         MAX_PATH Z$-VAR: zstrdata
         C-VAR: zbufchartype                            \ 0=oem,1=char,2=unicode
  END-TYPE

  TYPE: zBUFFER: INHERITS CLASSZSTR:
        \ *G MAX_PATH byte size zero terminated buffer.
  \ *[  *\fo{zBUFFER: members}
        MAX_PATH Z$-VAR: zstrdata
        \ *] End Doc-gen include
        \ *E   zBUFFER: Methods
        \ **   <count> method overloaded to use MAX_PATH
        \ **   <to>    method overloaded to use MAX_PATH
        CLASS \ class methods
        INST  \ Instance methods
  END-TYPE

: ZBUFFER-<to>                                              ( addr cnt obj --- )
   \
   <sizeof> zBUFFER: Z$-TO
   ;

: ZBUFFER-<count>                                           ( addr cnt obj --- )
   \
   <sizeof> zBUFFER: Z$-COUNT
   ;

  EXTEND-TYPE zBUFFER:
        CLASS \ Class methods
        INST  \ instance methods
        MRUNS <to> ZBUFFER-<to>
        MRUNS <count> ZBUFFER-<count>
  END-TYPE


\ **************************
\ *S 96 bit and 128 bit data
\ **************************

: T-VAR-<TO>                runword                              ( t1 addr --- )
   \ *G Triad variable store
   DUP >R 8 + !
   R@ 4+ !
   R> !
   ;

: T-VAR-<FETCH>             runword                              ( addr --- t1 )
   \ *G Triad variable fetch
   DUP >R @
   R@ 4+ @
   R> 8 + @
   ;

  TYPE: T-VAR:
  \ *G Structure for 96bit signed value
  \ *E  T-VAR: Methods
  \ **       Instance methods
  \ **       1.0 Modifiers
  \ **       1.1 Arithmatic
  \ **           <add>                                                ( t1 --- )
  \ **           <inc>                                                   ( --- )
  \ **           <sub>                                                   ( --- )
  \ **       1.2 Assignment
  \ **       1.3 Sign
  \ **           <negate>                                                ( --- )
  \ **           <sign?>                                                 ( --- )
  \ **       2.0 Inspectors
  \ **           <fetch>                                              ( --- t1 )
  \ **           <default>                                            ( --- t1 )
  \ **           <count>                                        ( --- addr cnt )
  \ **           <sizeof>                                              ( --- n )
  \ **           <addr>                                             ( --- addr )
  \ **           <ptr>                                              ( --- addr )
  \ **           <&ptr>                                             ( --- addr )
        L-VAR: X0                                 \ least significant 32bit long
        L-VAR: X1                                      \ middle value 32bit long
        L-VAR: X2                                  \ most significant 32bit long
        CLASS \ class methods
        INST  \ Instance methods
        MRUNS <to>      T-VAR-<TO>
        MRUNS <default> T-VAR-<FETCH>
        MRUNS <fetch>   T-VAR-<FETCH>
  END-TYPE

: Q@                 runword                            ( addr --- n0 n1 n2 n3 )
   \ *G Fetch 128bit value at addr. Store format [ LSB.MSB ]
   \ ** quad_fetch.
   >R
   R@ @
   R@ 4+ @
   R@ 8 + @
   R> 12 + @
   ;

: Q!                  runword                           ( n0 n1 n2 n3 addr --- )
   \ *G Store 128bit value at addr. Store format [ LSB.MSB ]
   \ ** quad_store
   >R
   4 NREV R@ !
   R@ 4+ !
   R@ 8 + !
   R> 12 + !
   ;

  TYPE: Q-VAR:
        \ *G 128bit signed value
        \ *[  *\fo{Q-VAR: members}
        L-VAR: x0
        L-VAR: x1
        L-VAR: x2
        L-VAR: x3
        \ *] End Doc-gen include
        CLASS \ class methods
        INST  \ Instance methods
        MRUNS <default> Q@
        MRUNS <to>      Q!
        MRUNS <fetch>   Q@
  END-TYPE

\ ***********************
\ *S Vector address types
\ ***********************

: VECTOR-INIT                                                         ( obj -- )
   \
   ASSIGN NOOP SWAP !
   ;

: VECTOR-EXECUTE                                                     ( obj --- )
   \
   @ ?DUP IF EXECUTE THEN
   ;

  TYPE: VECTOR:
  \ *G 32bit CFA
  \ *[  *\fo{VECTOR: members}
        L-VAR: addrcfa
  \ *] End Doc-gen include
        CLASS \ Class methods
        INST  \ Instance methods
        MRUNS <twist>    TWIST-LVAR
        MRUNS <fetch>    @
        MRUNS <default>  @
        MRUNS <inc>      OBJ-CRASH STRUCTURE-METHOD
        MRUNS <dec>      OBJ-CRASH STRUCTURE-METHOD
        MRUNS <set>      OBJ-CRASH STRUCTURE-METHOD
        MRUNS <add>      OBJ-CRASH STRUCTURE-METHOD
        MRUNS <sub>      OBJ-CRASH STRUCTURE-METHOD
        MRUNS <zero>     VECTOR-INIT
        MRUNS <init>     VECTOR-INIT
        MRUNS <erase>    VECTOR-INIT
        MRUNS <to>       !
        MRUNS <execute>  VECTOR-EXECUTE
  END-TYPE


\ **********************************
\ *S Arrays of 8-Bit Unsigned Values
\ **********************************

: CHAR(*)-[to]                                 ( addr cnt obj_addr obj_cnt --- )
   \ Execute or compile counted string to method
   State @
   IF
        POSTPONE $TO$
   ELSE
        $TO$
   THEN
   ;

  TYPE: CHAR(3):
  \ *G Char array 3 long
  \ *[  *\fo{CHAR(3): members}
        C-VAR: x0
        C-VAR: x1
        C-VAR: x2
        CLASS \ Class methods
        INST  \ instance methods
       :M <to> INST-COUNT CHAR(*)-[to] ;M  structure-method
  \ *] End Doc-gen include
  END-TYPE

  TYPE: CHAR(4):
  \ *G Char array 4 long
  \ *[  *\fo{CHAR(3): members}
        C-VAR: x0
        C-VAR: x1
        C-VAR: x2
        C-VAR: x3
        CLASS \ Class methods
        INST  \ instance methods
       :M <to> INST-COUNT CHAR(*)-[to] ;M  structure-method
  \ *] End Doc-gen include
  END-TYPE

\ *S Export List
\ *[ *\fo{Ini API }
  API-EXPORT (POINTER):
  API-EXPORT VALUE:
  API-EXPORT WORD-VALUE:
  API-EXPORT BYTE-VALUE:
  API-EXPORT C-VAR:
  API-EXPORT CHAR(3):
  API-EXPORT CHAR(4):
  API-EXPORT <C-VAR:
  API-EXPORT W-VAR:
  API-EXPORT W<VAR:
  API-EXPORT WORD:
  API-EXPORT L-VAR:
  API-EXPORT INT:
  API-EXPORT R-VAR:
  API-EXPORT S-VAR:
  API-EXPORT D-VAR:
  API-EXPORT B-VAR:
  API-EXPORT bytes:
  API-EXPORT cstring:
  API-EXPORT zstring:
  API-EXPORT (FIELD):
  API-EXPORT J-VAR:
  API-EXPORT A-VAR:
  API-EXPORT I-VAR:
  API-EXPORT sADDRLEN:
  API-EXPORT OBJID:
  API-EXPORT CLASSZSTR:
  API-EXPORT CLASSZFILE:
  API-EXPORT Z$-VAR:
  API-EXPORT T-VAR:
  API-EXPORT Q-VAR:
  API-EXPORT zBuf:
  API-EXPORT zBUFFER:
  API-EXPORT VECTOR:
  API-EXPORT uint:
  API-EXPORT lptstr:
  API-EXPORT hmenu:
  API-EXPORT hbitmap:
  API-EXPORT handle:
  API-EXPORT HWND:
  API-EXPORT WPARAM:
  API-EXPORT LPARAM:
  API-EXPORT dword:
  API-EXPORT long:
  API-EXPORT char:
  API-EXPORT byte:
  API-EXPORT r-int:
  API-EXPORT sword:
  API-EXPORT Z$-<ANSI_UPPER>                                         ( obj --- )
  API-EXPORT Z$-<ANSI_LOWER>                                         ( obj --- )
  API-EXPORT Z$-REMOVE-EXTRA-BLANKS                                ( zstr> --- )
  API-EXPORT CHARLOWER                                          ( addr cnt --- )
  API-EXPORT CHARUPPER                                          ( addr cnt --- )
\ *] End Doc-gen include

\ *> ####
  HIDE-**
  [FORTH]                                                     \ Reset Vocabulary
