\ Typeops.fth - Operators used with ClassVFX

((
Copyright (c) 2000, 2010, 2011
Construction Computer Software Pty Ltd.
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441

To do
=====

Change History
==============
))

\ ==========
\ *> typeops
\ *T Typeops.fth
\ ==========
\ *P Operators used with ClassVFX

\ ***************
\ *S Array Object
\ ***************

\ *[
OPERATOR: <elm$!>                                          \ addr cnt elm# ---
OPERATOR: <elm$+>                                               \ addr cnt ---
OPERATOR: <elm$@>                                         \  elm# --- addr cnt
OPERATOR: <elmaddr>                                      \ n0...n1 --- elmaddr
OPERATOR: <#bpe>                                                    \ --- #bpe
\ *]


\ ***************************
\ *S Class Property Operators
\ ***************************
\ ==========================
\ *N Class Inspector Methods
\ ==========================

op# sizeof OPERATOR <sizeof>                                       ( --- un )
\ *G Return the overall data space requirement of the class. Class
\ ** Inspector.
\ *E   <sizeof> {type:}
\ **   Stack effect : ( --- un )

OPERATOR: <addr&size>                                \ buffer address and size
OPERATOR: <self>                                            \ body of instance
OPERATOR: <cfa>                                     \ debug method for members
OPERATOR: <execute>                                  \ execute object insatnce

\ ============================
\ *N Instance Modifier Methods
\ ============================

OPERATOR: <copy>                                                    \ addr ---
\ *G Copy instance of same type: to an instance of type: Instance Modifier
\ *E   <ptr> {instance} <copy> {instance}
\ **   Stack effect : ( addr --- )
OPERATOR: <erase>
\ *G Fill the instance data space with zeroes. Instance Modifier
\ *E   <erase> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <blank>
\ *G Fill the instance data space with blanks. Instance Modifier.
\ *E   <blank> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <static>                                      \ static create action

\ ============================
\ *N Instance Property Methods
\ ============================

OPERATOR: <count>	( --- addr u-cnt )
\ *G Return the data address and size of the type instance. Instance
\ ** inspector.
\ *E   <count> {instance}
\ **   <count> {ptr>.instance}
\ **   <count> {complex_instance.instance}
\ **   Stack effect : ( --- addr u-cnt )

OPERATOR: <mem_count>   ( --- addr u-cnt )
\ *G Return the data address and available size of the memory object
\ ** Instance inspector.

op# ADDR OPERATOR ADDROF	( --- addr )
\ *G Return the data address of the type. Instance inspector.
op# ADDR OPERATOR <addrof>      ( --- addr )
\ *G Return the data address of the type. Instance inspector.
op# ADDR OPERATOR <addr>        ( --- addr )
\ *G Return the data address of the type. Instance inspector. Synonym
\ ** <addrof>

op# offsetof OPERATOR <offsetof>	( --- un )
\ *G Return the offset of the instance member of a complex type. Instance or
\ ** class inspector.
\ *E   <offsetof> {type:.instance}
\ **   <offsetof> {ptr>.instance}
\ **   <offsetof> {complex_instance.instance}
\ **   Stack effect : ( --- un )

\ ==================
\ *N Virtual Objects
\ ==================
\ *P Objects that allocate their data space in virtual memory and their type
\ ** instance structure in the FORTH dictionary or local return stack are
\ ** called Virtual Objects.

\ *P When using Virtual Objects one should be careful when passing the instance
\ ** address and the data address to use the correct method.

\ *P The <&ptr> method returns the instance address and the <ptr@> method
\ ** returns the data address.

\ *P Virtual Objects has a strict naming convection that must be adhered to.
\ ** This will prevent confusion and incorrect methods applied to instances.
\ ** Start the name of a Virtual Object with a lowercase "v".
\ *C Example: A local vzFILENAME: %vzfile

\ ==================
\ *N Pointer Methods
\ ==================

OPERATOR: <ptr>		( --- addr )
\ *G Return the data address of the first element of the type structure.
\ ** Instance inspector.
\ *E   <ptr> {instance}
\ **   Stack effect : ( --- addr )
OPERATOR: <&ptr>        ( --- addr )
\ *G Return the data address of an instance of pointer type. This is used
\ ** to pass an instance of a pointer. Instance inspector.
\ *E   <&ptr> {pointer_instance}
\ **   Stack effect : ( --- addr )
OPERATOR: <ptr!>        ( instance_addr --- )
\ *G Set the data address of an instance of a type in an instance of POINTER:
\ ** Type at address must match the type of POINTER: Instance Modifier.
\ *E   <ptr!> {pointer_instance}
\ **   Stack effect : ( addr --- )
OPERATOR: <ptr@>        ( --- instance_addr )
\ *G Return the data address of the first element of the type structure of a
\ ** virtual object. Instance inspector.
\ *E   <ptr@> {virtual_instance}
\ **   Stack effect : ( --- addr )
OPERATOR: <ptr0>
\ *G Zero the data of an instance of pointer. Instance Modifier.
\ *E   <ptr0> {pointer_instance}
\ **   Stack effect : ( --- )

\ *N INSTANCE STATUS OPERATORS
op# default  OPERATOR <default>
op# FETCH    OPERATOR <fetch>

\ *N INSTANCE MODIFIER OPERATORS
op# to  OPERATOR <to>
\ *G Primary store method.
op# TWIST    OPERATOR <twist>

op# ZERO OPERATOR <zero>	( --- )
\ *G Fill the data space of an instance with zeroes. Instance Modifier
\ ** Restrictions: Does not apply to counted string instances.
\ *E   <zero> {instance} ---
\ **   Stack effect : ( --- )

op# SET OPERATOR <set>
\ *G Assign the value held by the instance to highest possible value that
\ ** it could hold. ( all value bits to 1 ) for numeric value objects.
\ ** Instance Modifier.
\ **  Restrictions: Numeric simple instance only. Avoid complex types.
\ *E  <set> {instance}
\ **  Stack effect : ( --- )

OPERATOR: <@off>
\ *G Fetch the current value of an unary numeric type and the fill the
\ ** data space of an instance with zeroes. Instance Modifier.
\ ** Restrictions: Does not apply to counted string instances.
\ *E   <@off> {instance}
\ **   Stack effect : ( --- n )

OPERATOR: <toggle>	( --- )
\ *G Check th e current value of the object, if zero then assign the value
\ ** held by the instance to highest possible value that it could hold. (
\ ** all value bits to 1 ) for numeric value objects. If any value other than
\ ** zero, then zero all the bits in the object.
\ ** Instance Modifier.
\ **  Restrictions: Numeric simple instance only. Avoid complex types.
\ *E  <toggle> {instance}
\ **  Stack effect : ( --- )


\ *********************
\ *S Arithmetic Methods
\ *********************

OPERATOR: <mul>
\ *G Multiply operator for unary numeric types. The input type must be on
\ ** the correct numeric stack for the type. Instance Modifier.
\ *E   [type] <mul> {instance}
\ **   Stack effect : ( [type] --- )
OPERATOR: <div>
\ *G Divide operator for unary numeric types. The input type must be on
\ ** the correct numeric stack for the type. Instance Modifier.
\ *E Usage : [type] <div> {instance}
\ ** Stack effect : ( [type] --- )

op# SUB OPERATOR <sub>
\ *G Subtract operator for unary numeric types. The input type must be on
\ ** the correct numeric stack for the type. Instance Modifier.
\ *E   [type] <sub> {instance}
\ **   Stack effect : ( [type] --- )
op# ADD OPERATOR <add>
\ *G Addition operator for unary numeric types. The input type must be on
\ ** the correct numeric stack for the type. Instance Modifier.
\ *E   [type] <add> {instance}
\ **   Stack effect : ( [type] --- )
op# INC OPERATOR <inc>	( --- )
\ *G Increment numeric value with one. Operator for unary numeric types.
\ ** Instance Modifier.
\ *E   <inc> {instance}
\ **   Stack effect : ( --- )
op# DEC OPERATOR <dec>	( --- )
\ *G Decrement numeric value with one. Operator for unary numeric types.
\ ** Instance Modifier.
\ *E   <dec> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <sign?>
\ *G Make the numeric value positive. Return a flag indicating the sign.
\ ** Flag = true negative & flag = false positive. Operator for unary
\ ** numeric types. Instance Modifier.
\ *E   <sign?> {instance}
\ **   Stack effect : ( --- flag )

OPERATOR: <negate>
\ *G change sign of value.
OPERATOR: <neg>
\ *G change sign of value to negative
OPERATOR: <abs>
\ *G change sign of value to positive

OPERATOR: <compare>    ( obj_inst1> obj_inst2> --- flag )
\ *G Compare 2 instances of the same class and return a flag
\ **   Stack effect : ( obj_inst1> obj_inst2> --- flag )
\ ** Return  1 if instance at obj_inst1> is greater than instance at obj_inst2>
\ ** Return  0 if instance at obj_inst1> is equal to instance at obj_inst2>
\ ** Return -1 if instance at obj_inst1> is less than instance at obj_inst2>

OPERATOR: <create>                                                 ( --- )
\ *G Perform create of the class

OPERATOR: <tostr>                                ( instance> --- addr cnt )
\ *G Convert binary class to string representation of the class
\ **   Stack effect : ( instance> --- addr cnt )

OPERATOR: <fromstr>                                ( addr cnt instance> --- )
\ *G Convert string representation of class at addr cnt to binary class of
\ ** the class in the object instance. If conversion fails then the instance
\ ** will contain the <init> method of the class.
\ **   Stack effect : ( addr cnt instance> ---  )

OPERATOR: <invert>                                              \ invert value
OPERATOR: <sqrt>                                                 \ square root
OPERATOR: <int>                                                      \ integer
OPERATOR: <fract>                                                 \ fractional
OPERATOR: <inv>                                                      \ inverse

OPERATOR: <tan>
OPERATOR: <cos>
OPERATOR: <sin>
OPERATOR: <atan>
OPERATOR: <acos>
OPERATOR: <asin>
OPERATOR: <log>
OPERATOR: <alog>
OPERATOR: <ln>
OPERATOR: <exp>
OPERATOR: <rad>deg>
OPERATOR: <deg>rad>
OPERATOR: <fill>                                     \ fill object with a char


\ ****************
\ *S Float Methods
\ ****************

OPERATOR: <f@>
\ *G Float operator for unary numeric types. Fetch object to float stack.
\ ** Fetch from object output as f-var type
\ *E   <f@> {instance}
\ **   Stack effect : ( --- f )

OPERATOR: <f!>
\ *G Float operator for unary numeric types. Store float from stack in
\ ** object. Float will be converted. Float input store in object.
\ *E   <f!> {instance}
\ **   Stack effect : ( f --- )

OPERATOR: <f>d>
\ *G Float operator for unary numeric types. Fetch float from object
\ ** and convert integer portion to 64bit signed value
\ *E   <f>d> {instance}
\ **   Stack effect : ( --- d1 )

OPERATOR: <f>b>
\ *G Float operator for unary numeric types. Fetch float from object
\ ** and convert to 64bit signed value and dpl. Fetch from float object
\ ** output as b-var type
\ *E   <f>b> {instance}
\ **   Stack effect : ( --- b1 )

OPERATOR: <mod>

\ **********************************************
\ *S Instance Constructor And Destructor Methods
\ **********************************************

OPERATOR: <make>
\ *G Set the object structure members to the init state. Allocate any virtual
\ ** data space required to function and allow inspector methods to operate.
OPERATOR: <destroy>
\ *G Clear the object structure members. Deallocate any virtual data space
\ ** and invalidate all inspector methods.
OPERATOR: <init>
\ *G Set the object structure members to the init state.

OPERATOR: <#dim!>
OPERATOR: <#dim@>
OPERATOR: <valid?>
\ *G Check the current state of the object. Return True if object passes
\ ** the object validation test else false. Instance Inspector.
\ *E   <valid?> {instance}
\ **   Stack effect : ( --- flag )
OPERATOR: <data>
\ *G Return the data address of an instance
OPERATOR: <busy?>
\ *G Return the current busy value of the object. Instance Inspector.
\ *P Usage : <busy?> {instance}
\ *P Stack effect : ( --- busy# )
OPERATOR: <busy!>
\ *G Set the current busy value of the object. Instance modifier
\ *P Usage : busy# <busy!> {instance}
\ *P Stack effect : ( busy# --- )
OPERATOR: <mine?>
OPERATOR: <prime>
\ *G Used after init for first time structure setup. eg dim matrix within a
\ ** structure

\ ****************
\ *S String Object
\ ****************
\ ==============================
\ *N Instance Property Operators
\ ==============================

OPERATOR: <len>		( --- len )
\ *G Return -trailing length of counted string objects or trailing zero
\ ** character of zero terminated string objects. Instance Inspector.
\ *E   <len> {instance}
\ **   Stack effect : ( --- u )

OPERATOR: <spare>   ( --- u )
\ *G Return the trailing buffer space available for an Z$ object. Instance Inspector.
\ *E   <spare> {instance}
\ **   Stack effect : ( --- u )

op# <len> OPERATOR <chars>	( --- u )
\ *G Return -trailing length of counted string objects or trailing zero
\ ** character of zero terminated string objects. Instance Inspector.
\ *E   <len> {instance}
\ **   Stack effect : ( --- u )

OPERATOR: <null?>	( --- flag )
\ *G Return true if no chars in counted string objects or zero terminated
\ ** string objects else false if any chars. Instance Inspector.
\ *E   <null?> {instance}
\ **   Stack effect : ( --- flag )

OPERATOR: <trim_count>  ( --- addr u-cnt )
\ *G Return remaining addr and cnt after performing -trailing and -leading on
\ ** the data space of the object. Instance Inspector.
\ *E   <trim_count> {instance}
\ **   Stack effect : ( --- c-addr u-cnt )

OPERATOR: <search>                                 ( addr len --- addr'|false )
\ *G Search object for sub string at addr cnt return addr' if found or false if not
\ ** found within the object current counted or terminated data space or overall data space.

\ ============================
\ *N Instance Modifier methods
\ ============================

OPERATOR: <concat->     ( c-addr u-cnt --- )
\ *G Add counted string to -trailing end of string object. Instance
\ ** Modifier.
\ *E   <concat-> {instance}
\ **    Stack effect : ( c-addr u-cnt --- )

OPERATOR: <concat>      ( c-addr u-cnt --- )
\ *G Add counted string to start of string object. Instance Modifier.
\ *E   <concat> {instance}
\ **   Stack effect :  ( c-addr u-cnt --- )

OPERATOR: <trim>        ( --- )
\ *G Remove leading and trailing blank characters from string object.
\ ** Instance Modifier.
\ *E   <trim> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <trim-trailing>	( --- )
\ *G Remove trailing blank characters from string object.
\ ** Instance Modifier.
\ *E   <trim-trailing> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <trim-leading>	( --- )
\ *G Remove leading blank characters from string object.
\ ** Instance Modifier.
\ *E   <trim-leading> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <upper>
\ *G ASCII Upper case string object. Only char values greater or equal to
\ ** ASCII a and lower or equal to ASCII z will be effected.

OPERATOR: <lower>
\ *G ASCII lower case string object. Only char values greater or equal to
\ ** ASCII A and lower or equal to ASCII Z will be effected.

OPERATOR: <ansi_upper>
\ *G Upper case string object to current ANSI locale setting

OPERATOR: <ansi_lower>
\ *G Lower case string object to current ANSI locale setting

OPERATOR: <$xlat>	( addr cnt --- )
\ *G Replace next string token found in string buffer instance.
\ *C Usage : <$xlat> [instance] ( addr cnt --- )

OPERATOR: <xlat>        ( --- )
\ *G Replace all constant string tokens found in string buffer instance.

OPERATOR: <expandmacro> ( --- )
\ *G Expand any macros in string buffer instance.

OPERATOR: <char+>                                             ( chr obj --- )
\ *G Add char on stack at end of string object
OPERATOR: <wchar+>                                           ( wchar obj --- )
\ *G Add wchar on stack at end of string object
OPERATOR: <char->   ( obj --- char )
\ *G Remove and return char at end of object
OPERATOR: <hide>   ( obj --- )
\ *G Perform system RC4 cipher on object
OPERATOR: <unhide>   ( obj --- )
\ *G Perform system RC4 cipher on object. Same as <hide> named to help developer
OPERATOR: <hidden?>   ( obj --- bool )
\ *G Return true if object is Candy system RC4 encrypted.Object must have a check member

\ *S 8-bit counted buffer operators
  OPERATOR: <b$init>                                                    ( --- )
  \ *G Init object buffer for byte count string access.
  OPERATOR: <b$sizeof>                                                ( --- u )
  \ *G Return the size of the byte counted string buffer
   OPERATOR: <b$make>                                                ( cnt --- )
  \ *G Init word counted string buffer object with cnt size string
  OPERATOR: <b$place>                                           ( addr cnt --- )
  \ *G Place string byte counted string buffer object
  OPERATOR: <b$append>                                          ( addr cnt --- )
  \ *G Append string byte counted string buffer object
  OPERATOR: <b$count>                                           ( --- addr cnt )
  \ *G COUNT of string byte counted string buffer object
  OPERATOR: <b$addr>                                                ( --- addr )
  \ *G String byte COUNT address byte counted string buffer object
  OPERATOR: <b$data>                                                ( --- addr )
  \ *G Byte count string data address byte counted string buffer object
  OPERATOR: <b$len>                                                 ( --- ulen )
  \ *G String byte length byte counted string buffer object
  OPERATOR: <b$char+>                                               ( char --- )
  \ *G Add char to end of byte counted string
  OPERATOR: <b$min>                                               ( char --- )
  \ *G Chop byte counted string to N len, or if smaller than N pad with
  \ ** trailing blanks to N
  OPERATOR: <b$wchar+>                                             ( wchar --- )
  \ *G Add wchar to end of byte counted string
  OPERATOR: <b$spare>                                                 ( --- ulen )
  \ *G Byte count string trailing buffer space
  OPERATOR: <b$search>
  \ *G Search byte count object for sub string at addr cnt return addr' if found or false if not
  \ ** found within the object current counted or terminated data space or overall data space.

\ *S 16-bit counted buffer operators
  OPERATOR: <w$init>                                                    ( --- )
  \ *G Init object buffer for word count string access.
  OPERATOR: <w$sizeof>                                                ( --- u )
  \ *G Return the size of the word counted string buffer
  OPERATOR: <w$make>                                                ( cnt  --- )
  \ *G Place string at buf> with 16-bit count
  OPERATOR: <w$place>                                           ( addr cnt --- )
  \ *G Place string at buf> with 16-bit count
  OPERATOR: <w$append>                                          ( addr cnt --- )
  \ *G Append string at buf> with 16-bit count
  OPERATOR: <w$count>                                           ( --- addr cnt )
  \ *G 16-bit COUNT of string at buf>
  OPERATOR: <w$addr>                                                ( --- addr )
  \ *G String word COUNT address
  OPERATOR: <w$data>                                                ( --- addr )
  \ *G Word count string data address
  OPERATOR: <w$len>                                                 ( --- ulen )
  \ *G String 16-bit byte length
  OPERATOR: <w$spare>                                                 ( --- ulen )
  \ *G Word count string trailing buffer space
  OPERATOR: <w$char+>                                               ( char --- )
  \ *G Add char to end of word counted string
  OPERATOR: <w$wchar+>                                             ( wchar --- )
  \ *G Add wchar to end of word counted string
  OPERATOR: <w$search>
  \ *G Search word count object for sub string at addr cnt return addr' if found or false if not
  \ ** found within the object current counted or terminated data space or overall data space.

\ *S 32-bit buffer operators
  OPERATOR: <l$init>                                                    ( --- )
  \ *G Init object buffer for 32-bit count string access.
  OPERATOR: <l$sizeof>                                                ( --- u )
  \ *G Return the size of the long counted string buffer
  OPERATOR: <l$make>                                                ( cnt  --- )
  \ *G Place string at buf> with long count
  OPERATOR: <l$place>                                      ( addr cnt buf> --- )
  \ *G Place string at buf> with 32-bit count
  OPERATOR: <l$append>                                     ( addr cnt buf> --- )
  \ *G Append string at buf> with 32-bit count
  OPERATOR: <l$count>                                      ( buf> --- addr cnt )
  \ *G 32-bit COUNT of string at buf>
  OPERATOR: <l$addr>                                                ( --- addr )
  \ *G String long COUNT address
  OPERATOR: <l$data>                                                ( --- addr )
  \ *G 32-bit count string data address
  OPERATOR: <l$len>                                                 ( --- ulen )
  \ *G String 32-bit byte length
  OPERATOR: <l$char+>                                               ( char --- )
  \ *G Add char to end of long counted string
  OPERATOR: <l$spare>                                                 ( --- ulen )
  \ *G 32-bit count string trailing buffer space
  OPERATOR: <l$wchar+>                                               ( char --- )
  \ *G Add wchar to end of 32-bit counted string
  OPERATOR: <l$search>
  \ *G Search long count object for sub string at addr cnt return addr' if found or false if not
  \ ** found within the object current counted or terminated data space or overall data space.

\ *********************
\ *S Filename Operators
\ *********************
\ ===============================
\ *N Instance Inspector Operators
\ ===============================

OPERATOR: <drive?>      ( --- flag )
\ *G Does the file name object include a valid drive specifier. True or false
\ ** Valid drive specifier <drive_letter>: <drive_letter> must be in
\ ** character position 1 and ColonChar must be in character position 2.
\ ** Instance Inspector. FORTH - drive_query
\ *E   <drive?> {instance}
\ **   Stack effect : ( --- flag )

OPERATOR: <path?>       ( --- flag )
\ *G Does the file name object include any path or folder specifier. True
\ ** or false. Valid path specifier BackSlashchar. Instance Inspector.
\ ** FORTH - path_query
\ *E   <path?> {instance}
\ **   Stack effect : ( --- flag )

OPERATOR: <unc?>        ( --- flag )
\ *G Does the file name object include a valid UNC share name. True or
\ ** false. Instance Inspector. FORTH - U_N_C_query
\ *E   Valid UNC share name   \\<computer_name>\<share_name>
\ **   <\\> must be in character position 1.
\ **   first leave is the computername.
\ **   second leave the share name.
\ **   <unc?> {instance}
\ **   Stack effect : ( --- flag )

OPERATOR: <wild?>       ( --- flag )
\ *G Does the file name object include any valid wild filename characters.
\ ** True or false. Valid valid wild filename characters is * and ?.
\ ** Instance Inspector. FORTH - wild_query
\ *E   <wild?> {instance}
\ **   Stack effect : ( --- flag )

\ *************************************
\ *S Filename Instance Modifier Methods
\ *************************************
\ ===============
\ *N Path Methods
\ ===============

OPERATOR: <path>        ( --- )
\ *G Remove remove one leave from filename object. Instance Modifier.
\ *E   <path> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <+path>       ( --- )
\ *G Add path to object using the filename according to CCS virtual
\ ** directory rules. Instance Modifier. FORTH - plus_path
\ *E   <+path> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <-path>       ( --- )
\ *G Remove all leaves from filename object except last. Instance Modifier.
\ ** FORTH - minus_path
\ *E   <-path> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <share>       ( --- )
\ *G Remove all leaves from filename object except drive mapping or share
\ ** name. Instance Modifier.
\ *E   <share> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <-share>      ( --- )
\ *G Remove share or drive mapping from zfilename. Instance Modifier.
\ *E   <-share> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <+filename>   ( addr cnt --- )
\ *G Add address count to zFILENAME object as the last leaf of the folder
\ ** path. Instance Modifier.
\ *E   <-share> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <-ext>        ( --- )
\ *G Remove existing file extesion from filename object Instance Modifier.
\ ** Instance Modifier.
\ *C   <-ext> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <+ext>        ( addr u-cnt --- )
\ *G Remove existing file extesion from filename object and add extension at
\ ** address count. Instance Modifier.
\ *C   <+ext> {instance}  (c-addr u-cnt  --- )

OPERATOR: <ext$>        ( --- c-addr u-cnt|false )
\ *G Does the file name object include any file extension. Instance
\ ** Inspector.
\ *E   <ext$> {instance}
\ **   Stack effect : ( --- c-addr u-cnt|false )

OPERATOR: <ptrext$>        ( --- c-addr|false )
\ *G Point to the file name object extension. Instance
\ ** Inspector.
\ *E   <ptrext$> {instance}
\ **   Stack effect : ( --- c-addr|false )

OPERATOR: <+cjj>        ( --- )
\ *G Add current Coy$ and Job$ as file extension to filename object name.
\ ** Instance Modifier.
\ *C   <+cjj> {instance}  ( --- )

OPERATOR: <squeeze>     ( --- )
\ *G Remove all blanks within string object. Instance Modifier.
\ *E   <squeeze> {instance}
\ **   Stack effect : ( --- )

OPERATOR: <paragraph>   \ -- ; remove repeating blanks from string

OPERATOR: <$lft>	\ u ---
OPERATOR: <$rht>        \ u ---
OPERATOR: <$min>        \ u ---
OPERATOR: <$min->       \ u ---
OPERATOR: <$mid>        \ from_u-cnt to_u-cnt ---

\ ***************
\ *S Matrix Cells
\ ***************

\ ===================
\ *N <mat!> Operators
\ ===================
\ *P [class] = the defined class of the data represented in one cell. If the
\ ** class is a simple numeric type then the data will be on the appropriate
\ ** stack. If the class is a complex type then the class <copy> method will
\ ** be used to assign the value of the cell. ie the data address of the
\ ** instance.

OPERATOR: <mat!>
\ *G Store class type for row# in matching class matrix instance.
\ *C Usage :   <mat!> [instance] \ [class] row# ---

OPERATOR: <matx2!>
\ *G Store class type for row# and col# in matching class matrix table instance.
\ *C Usage :   <matx2!> [instance] \ [class] row# col# ---

OPERATOR: <matx3!>
\ *G Store class type for row# and col# and table# in matching class matrix table instance.
\ *C Usage :   <matx3!> [instance] \ [class] row# col# table# ---

OPERATOR: <matx4!>
\ *G Store class type for row# and col# and table# and group# in matching class matrix table instance.
\ *C Usage :   <matx4!> [instance] \ [class] row# col# table# group# ---

\ ======================
\ *N <&cell..> Operators
\ ======================

OPERATOR: <&cellx1>     ( row# --- addr )
\ *G Fetch the class instance address of the required matrix element at row#.

OPERATOR: <&cellx2>     ( row# col# --- addr )
\ *G Fetch the class instance address of the required matrix element at row# and col#.

OPERATOR: <&cellx3>     ( row# col# table# --- addr )
\ *G Fetch the class instance address of the required matrix element at row# and col# and table#.

OPERATOR: <&cellx4>     ( row# col# table# group# --- addr )
\ *G Fetch the class instance address of the required matrix element at row# and col#
\ ** and table# and group#.

\ ===================
\ *N <mat@> Operators
\ ===================
\ *P [class] = the defined class of the data represented in one
\ ** cell. If the class is a simple numeric type then the data
\ ** will be fetched on the appropriate stack. If the class is a
\ ** complex type then the class <default> method will be used to
\ ** point to the data of the cell.

OPERATOR: <mat@>        ( row# --- [class] )
\ *G Fetch class type for row# in matching class matrix instance.
\ *C Usage :   <mat@> [instance]                                 \ row# --- [class]
\ *P Can not be used with dot notation

OPERATOR: <matx2@>      ( row# col# --- [class] )
\ *G Fetch class type for row# and col# in matching class matrix table instance.
\ *C Usage :   <matx2@> [instance] \ row# col# --- [class]
\ *P Can not be used with dot notation

OPERATOR: <matx3@>      ( row# col# table# --- [class] )
\ *G Fetch class type for row# and col# and table# in matching class matrix table instance.
\ *C Usage :   <matx3@> [instance] \ row# col# table# --- [class]
\ *P Can not be used with dot notation

OPERATOR: <matx4@>      ( row# col# table# group# --- [class] )
\ *G Fetch class type for row# and col# and table# and group# in matching class matrix table instance.
\ *C Usage :   <matx4@> [instance] \ row# col# table# group# --- [class]
\ *P Can not be used with dot notation
OPERATOR: <debug>
\ *> ####
     HIDE-**


