((
$Date: 2008-05-10 11:52:33 +0200 (sáb 10 de may de 2008) $
$Revision: 520 $
$Author: astrorafael $
$HeadURL: file:///home/desa/repos/vfx/objects/trunk/src/struct.fs $

data structures (like C structs)
written by Anton Ertl 1996-2000
This file is in the public domain. NO WARRANTY.
))

\ ###########################
\ *! struct
\ *R @node{Support Structs Package}
\ *T Support Structs Package
\ ###########################

\ ***************
\ *S Introduction
\ ***************

\ *P This module is a porting of GForth's ANS structures. It is *\b{not}
\ ** intended as a replacement for VFX structures, but as the necesary support
\ ** for the object oriented package also ported from GForth.
\ ** In a future, it could be possible to rewrite the object oriented package
\ ** to use the native VFX structures and this module would dissapear.
\ ** Also, there is an upcoming structures standard in Forth 200x
\ ** that could be taken into account.

\ *P To use this package, it is necessary to previously include
\ ** a floating point package from those available in VFX
\ ** or else redefine words
\ ** *\fo{dfaligned}, *\fo{sfaligned}, *\fo{faligned}, *\fo{floats},
\ ** *\fo{sfloats}, *\fo{dfloats}.

\ *P The necessary words that specify field size when defining class
\ ** field members are:
\ *(
\ *B *\fo{cell%}
\ *B *\fo{double%}
\ *B *\fo{char%}
\ *B *\fo{float%}
\ *B *\fo{sfloat%}
\ *B *\fo{dfloat%}
\ *)

\ *P When defining an array of , let's say three floats, use the expression:
\ *C  float% 3 *

\ *P When accomodating an VFX structure named *\fo{/foo}, use:
\ *C  char% /foo *

\ *P In addition, when defining public field members, the word *\fo{field%}
\ ** must be used.

\ ***************
\ +S Design notes
\ ***************

\ +P GForth structures differ from VFX structures in some ways:
\ +(
\ +B They use the "name last" convention of naming structures
\ +B Field creation carry both the next alignment and size.
\ +B Throwing exceptions uses *\fo{?throw} instead of *\fo{throw}
\ +)

\ +P To avoid collisions with VFX structs, the main field creation word
\ +* *\fo{field} has been renamed to *\fo{field%}.

\ +P A couple of memory allocation words are defined. Some auxiliar words
\ +* in the original package but not being ised in *\f{'objects.fs'} have
\ +* been removed.


FpSystem 0= [if]
   \   include /usr/share/doc/VfxForth/Lib/Ndp387.fth
   cr .( ==> Please include your favourite floating point package )
   quit
[then]

\ ***********
\ *S Glossary
\ ***********

MODULE GForth-Struct

: naligned ( addr1 n -- addr2 )
\ +G *\i{addr2} is the aligned version of *\i{addr1} with respect to the
\ ** alignment size *\i{n}.
   1- tuck +  swap invert and
;


: dofield ( -- )
\ +G Genereic run time code to compute the field address.   
 does> ( name execution: addr1 -- addr2 )
  @ +
;

: dozerofield ( -- )
\ +G Optimization to compute the offset for the first field.
\ +* Does nothing at runtime.
   immediate
 does> ( name execution: -- )
   drop
;

: create-field ( align1 offset1 align size "name" --  align2 offset2 )
\ *G Creates an struct field given by *\i{"name"} with the desired
\ ** *\i{align} and *\i{size} specification and taking the past history of
\ ** *\i{align1} and *\i{offset1} to produce the new
\ ** *\i{align2} and *\i{offset2}.
   create
   swap					\ align1 offset1 size align
   rot					\ align1 size align offset1
   over					\ align1 size align offset1 align
   naligned				\ align1 size align offset3
   dup , 				\ align1 size align offset3
   rot + >r				\ align1 align ; R: offset2
   naligned r> 				\ align2 offset2
;

: field% ( align1 offset1 align size "name" --  align2 offset2 )
\ *G Defining word for fields, using *\fo{create-field} above and
\ ** the runtime code given by *\fo{dofield} and *\fo{dozerofield}
\ ** The defined chilkd word  has the following stack signature:
\ ** *\f{addr1 -- addr2}.
   2 pick >r \ this uglyness is just for optimizing with dozerofield
   create-field
   r> if \ offset<>0
      dofield
   else
      dozerofield
   then
;

: end-struct% ( align size "name" -- )
\ *G Finish the whole structure definition and and tag it with "name".   
   over naligned \ pad size to full alignment
   2constant
;


1 chars 0 end-struct% struct%		\ --  align1 0
\ *G Start a new structure definion. Returns the initial aligment of 1 chars
\ ** and an offset of 0.


EXPORT struct%
EXPORT end-struct%
EXPORT field%
EXPORT create-field

\ ========================
\ *N Field size specifiers
\ ========================

\ type descriptors, all ( -- align size )
1 aligned   1 cells   2constant cell%	\ -- align size
\ *G Define a cell field.

1 chars     1 chars   2constant char%	\ -- align size
\ *G Define a char field.

1 faligned  1 floats  2constant float%	\ -- align size
\ *G Define a float field.

1 dfaligned 1 dfloats 2constant dfloat%	\ -- align size
\ *G Define a double precision floating point field.

1 sfaligned 1 sfloats 2constant sfloat%	\ -- align size
\ *G Define a single precision floating point field.

cell% 2*              2constant double%	\ -- align size
\ *G Define a double cell number field.

EXPORT cell%
EXPORT char%
EXPORT float%
EXPORT dfloat%
EXPORT sfloat%
EXPORT double%


\ ==========================
\ *N Memory allocation words
\ ==========================

\ *P Memory allocation words for these structures
\ ** either in the dictionary or in the heap.

: %align   ( align size -- )
\ +G Align memory in the dictionary with the respect to *\i{align}.
   drop here swap naligned here - allot
;


: %allot   ( align size -- addr )
\ *G Returns an aligned memory region *\i{addr} of a given *\i{size}
\ ** in the dictionary with the respect to *\i{align} bytes.   
   tuck %align
   here swap allot
;

 
: %alloc   ( align size-- addr )
\ *G Returns an aligned memory region *\i{addr} of a given *\i{size}
\ ** in the heap with the respect to *\i{align} bytes. Throws
\ ** a *\f{#-59} exception code if allocation fails.    
   nip ProtAlloc
;


EXPORT %allot
EXPORT %alloc

END-MODULE
