\ Pentium specific code for VFX Forth

((
Copyright (c) 1999
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441

PLEASE NOTE THAT OUR PHONE NUMBER CHANGED IN APRIL 2000

To do
=====

Change history
==============
20011030 JOS001 Enhanced CPUID and returned information.
))

prefix intel

code rdtsc      \ -- d ; reads Pentium time stamp counter
  sub   ebp, 8                  \ make space
  rdtsc                         \ read time stamp counter into EDX:EAX
  mov   4 [ebp], ebx            \ save old TOS
  mov   0 [ebp], eax            \ low portion of counter
  mov   ebx, edx                \ high portion of counter
  next,
end-code

2variable startclocks		\ TSC at time of test
#50, 2constant tsc-overhead	\ CPU dependent overhead - change as required

: start-tsc	\ -- ; mark start of test
  rdtsc startclocks 2!
;

: elapsed-tsc	\ -- d ; return clocks since last START-TSC
  rdtsc startclocks 2@ d-		\ base value
  tsc-overhead d-			\ remove CPU dependent overhead
  dabs					\ allow for wraparound
;

: ttsc		\ -- ; test overhead calculation
  begin
    start-tsc elapsed-tsc d.
  key? until
  key drop
;


\ The following version of cpuid shows all the parameters.
\ The original code was posted on comp.lang.forth 30 October 2001
\ by Jos v.d. Ven

decimal

code cpuid	\ initial_EAX_value -- ebx edx ecx eax
		\ Returns Pentium CPU identification info
  mov   eax, ebx	\ put initial_EAX_value in eax ( allowed: 0, 1 or 2)
  cpuid
  sub   ebp, # #12
  mov   8 [ebp], ebx
  mov   4 [ebp], edx
  mov   0 [ebp], ecx
  mov   ebx, eax
  next,
end-code

\ vendor$: ebx esx ecx from cpuid when initial_EAX_value = 0
: vendor$	\ ebx edx ecx -- adr count
  pad 8 + !
  pad 4 + !
  pad     !
  pad 12
;

: id-vendor-cpu	\ -- adr count
  0 cpuid drop vendor$
;

: decode-cpu-version	\ -- Reserved Type Family Model Stepping
			\ Short: R T F M S
  1 cpuid >r 3drop
  r@ 0xFFFFC000 and 14 rshift
  r@ 0x3000 and     12 rshift
  r@ 0xf00 and       8 rshift
  r@ 0xf0 and        4 rshift
  r> 0xf and
;
\ I found the following values:
\   R T F M S CPU
\   0 0 5 1 7 P66
\   0 0 5 7 0 P100
\   0 0 6 5 1 PII 400

: cache		\ -- l2
  2 cpuid 2drop nip  255 and
;

hex

: .cache	\ --
  cr ." Cache description:"
  cr
  case  cache
    00 of ." Null descriptor. " endof
    01 of ." Instruction TLB: 4K-Byte Pages, 4-way set associative, 32 entries."  endof
    02 of ." Instruction TLB: 4M-Byte Pages, 4-way set associative, 2 entries."  endof
    03 of ." Data TLB: 4K-Byte Pages, 4-way set associative, 64 entries." endof
    04 of ." Data TLB: 4M-Byte Pages, 4-way set associative, 8 entries." endof
    06 of ." 1st-level instruction cache: 8K Bytes, 4-way set associative, 32 byte line size." endof
    08 of ." 1st-level instruction cache: 16K Bytes, 4-way set associative, 32 byte line size." endof
    0A of ." 1st-level data cache: 8K Bytes, 2-way set associative, 32 byte line size." endof
    0C of ." 1st-level data cache: 16K Bytes, 4-way set associative, 32 byte line size." endof
    22 of ." 3rd-level cache: 512K Bytes, 4-way set associative, 64 byte line size." endof
    23 of ." 3rd-level cache: 1M Bytes, 8-way set associative, 64 byte line size." endof
    25 of ." 3rd-level cache: 2M Bytes, 8-way set associative, 64 byte line size." endof
    29 of ." 3rd-level cache: 4M Bytes, 8-way set associative, 64 byte line size." endof
    40 of ." No 2nd-level cache or, if processor contains a valid 2nd-level cache, no 3rd-level cache." endof
    41 of ." 2nd-level cache: 128K Bytes, 4-way set associative, 32 byte line size." endof
    42 of ." 2nd-level cache: 256K Bytes, 4-way set associative, 32 byte line size." endof
    43 of ." 2nd-level cache: 512K Bytes, 4-way set associative, 32 byte line size." endof
    44 of ." 2nd-level cache: 1M Byte, 4-way set associative, 32 byte line size." endof
    45 of ." 2nd-level cache: 2M Byte, 4-way set associative, 32 byte line size." endof
    50 of ." Instruction TLB: 4-KByte and 2-MByte or 4-MByte pages, 64 entries." endof
    51 of ." Instruction TLB: 4-KByte and 2-MByte or 4-MByte pages, 128 entries." endof
    52 of ." Instruction TLB: 4-KByte and 2-MByte or 4-MByte pages, 256 entries." endof
    5B of ." Data TLB: 4-KByte and 4-MByte pages, 64 entries." endof
    5C of ." Data TLB: 4-KByte and 4-MByte pages,128 entries." endof
    5D of ." Data TLB: 4-KByte and 4-MByte pages,256 entries." endof
    66 of ." 1st-level data cache: 8KB, 4-way set associative, 64 byte line size." endof
    67 of ." 1st-level data cache: 16KB, 4-way set associative, 64 byte line size." endof
    68 of ." 1st-level data cache: 32KB, 4-way set associative, 64 byte line size." endof
    70 of ." Trace cache: 12K-op, 8-way set associative." endof
    71 of ." Trace cache: 16K-op, 8-way set associative." endof
    72 of ." Trace cache: 32K-op, 8-way set associative." endof
    79 of ." 2nd-level cache: 128KB, 8-way set associative, sectored, 64 byte line size." endof
    7A of ." 2nd-level cache: 256KB, 8-way set associative, sectored, 64 byte line size." endof
    7B of ." 2nd-level cache: 512KB, 8-way set associative, sectored, 64 byte line size." endof
    7C of ." 2nd-level cache: 1MB, 8-way set associative, sectored, 64 byte line size." endof
    82 of ." 2nd-level cache: 256K Bytes, 8-way set associative, 32 byte line size." endof
    84 of ." 2nd-level cache: 1M Byte, 8-way set associative, 32 byte line size." endof
    85 of ." 2nd-level cache: 2M Byte, 8-way set associative, 32 byte line size." endof
          ." Cache type is unknown."
  endcase
;

decimal

: .cpu
  cr ." Vendor id cpu: " id-vendor-cpu type
  cr decode-cpu-version
     ." Type:"  3 roll . rot ." Family:" . ." Model:" swap .
     ." Stepping:" . drop
  .cache
;

((
\ The code that follows is very old 16 bit mode code for detecting
\ CPU types. It was last changed in 1989 or 1992.

decimal  36 base !

create cpu-tab
  80386, , ,            \ 0
  80286, , ,            \ 4
  80188, , ,            \ 8
  80186, , ,            \ 0c
  NECV20, , ,           \ 010
  NECV30, , ,           \ 014
  8088, , ,             \ 018
  8086, , ,             \ 01c
  80486, , ,            \ 020
decimal

hex

code (cpuid)                    \ -- tab-offset
  bx push                       \ save old TOS
  bx, bx sub                    \ 80386 table offset
  sp push                       \ push SP onto the stack
  ax pop                        \ get old pointer into AX
  ax, sp cmp  z,                \ compare ax and sp
  if,  ax, # 0f000 mov          \ 80286 or 80386
   ax push  popf  pushf  ax pop \ are top bits in CC preserved
   ax, # 0f000 and  z,          \
   if, bx, # 4 mov  endif,      \ must be 80286
  else, bx, # 8 mov             \ not 80286 or 80386
     cl, # 021 mov              \ 8018x shifts modulus 32
     al, # 0ff mov              \ start with all bits set
     al, cl shl                 \ will only shift once if 8018x
\ (cpuid) NECx check
     z,                         \ not 8018x test for NECx
     if,  cx, # -1 mov          \ mov -1 to CX
          sti                   \ enable interrupts for timer
          si push               \ save SI
          rep  es: [si] lodsb   \ if restored from interrupt
          si pop                \    correctly CX will be zero
          l$1 jcxz
          bx, # 018 mov         \ must be 808x
          l$2 jmp
l$1:      bx, # 010 mov         \ if so then cpu is a NECx
l$2:
     endif,

\ (cpuid)  pre-fetch-instruction queue
     std                        \ store backwards
     al, #  041 mov             \ INC CX instruction
     here 016 + al mov          \ store at last 2 bytes of p@iq
     di push  es push           \ preserve DI & ES
     ds push  es pop            \ make ES = DS
     di, here 010 + # mov       \
     al, # 0fb mov              \ STI instruction
     cx, # 03 mov               \ count of 3
     cli                        \ disable interrupts
     rep  stosb                 \ over write instructions
     cld  nop  nop  sti         \ 4 byte pre fetch
     sti  sti                   \ 2 bytes of 6 byte pre fetch
     es pop  di pop             \ restore ES & DI

     l$3 jcxz
     bx, # 4 add                \ inc point by 4 for xxx6 cpu's
l$3:
  endif,

\ (cpuid)  do 386/486 check
  bx, bx or     \ =0 for 386/486, so test
  z, if,  066 c, 08B c, 0D4 c,          \ MOV EDX, ESP
          066 c, 083 c, 0E4 c, 0FC c,   \ AND ESP, NOT 3
       066 c,  09C c,  066 c,  058 c,   \ PUSHFD   POP EAX
       066 c,  08B c,  0C8 c,           \ MOV ECX,EAX
       066 c,  035 c,  000 c, 000 c, 004 c, 000 c,  \ XOR EAX,#
       066 c,  050 c,  066 c,  09D c,   \ PUSH EAX   POPFD
       066 c,  09C c,  066 c,  058 c,   \ PUSHFD   POP EAX
       066 c,  033 c,  0C1 c,           \ XOR EAX, ECX
       066 c,  0C1 c,  0E8 c,  012 c,   \ SHR EAX,#18
       066 c,  051 c,  066 c,  09D c,   \ PUSH ECX   POPFD
       066 c,  08B c,  0E2 c,           \ MOV ESP, EDX
       ax, # 1 and
       nz, if,  bx, # 020 mov  endif,   \ is 486
  endif,

  next,   end-code              \ answer in BX, which is TOS

decimal

: .cpu          \ -- ; determines and displays CPU type
  base @  36 base !             \ Coded as A..Z, 0..9
  (cpuid) cpu-tab + 2@
  cr ." Using " d. ." CPU"
  base !  ;

))
