\ explore vocabularies and their contents

((
Copyright (c) 2021
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: info@vfxforth.com
web: https://vfxforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441

))

include vt100.fth
include isearch.fth

: in-order? ( wid -- f )
    \ *G check if wordlist is in current search-order
    {: wid | found :}
    0 to found
    get-order 0 do
        wid = if
            -1 to found
        then
    loop found ;

variable #browse-file-lines
10 #browse-file-lines !

variable selected-voc
get-current selected-voc !
: draw-voc ( addr-voc -- )
    dup >vocname count
    rot selected-voc @ = if -1 negative-text then \ invert selected
    line-type
    0 negative-text
    line-space ;

: draw-vocs ( -- )
    1 1 cursor-position
    terminal-linebreaks off
    black back-color
    white bright-text-color
    bold" Order:"  \ wordlists in current search-order
    cyan bright-text-color
    get-order 0 do
        draw-voc
    loop
    white bright-text-color
    bold" Vocs:" \ remaining vocabularies
    cyan bright-text-color
    voc-link @
    begin
        dup in-order? 0= if
            dup draw-voc
        then
        >voc-link @ ?dup 0=
    until ;

: draw-key ( c-addr-key n-key c-addr-text n-text -- )
    2swap
    white bright-text-color
    black back-color
    line-space
    line-type

    black text-color
    cyan bright-back-color
    line-type
    line-space ;

255 constant /draw-filter
/draw-filter buffer: draw-filter

: draw-footer ( -- )
    1 terminal-h @ cursor-position
    terminal-linebreaks off
    s" ESC" s" Exit" draw-key
    s" ←→" s" Select Wordlist" draw-key
    s" ↑↓" s" Scroll Words" draw-key
    s" Filter:" draw-filter count draw-key
    black back-color
    bl line-fill
    ;

\ fast integer based log10
create log10-values
10 , 100 , 1000 , 10000 , 100000 , 1000000 , 10000000 , 100000000 , 1000000000 ,
10000000000 , 100000000000 , 1000000000000 , 10000000000000 , 100000000000000 ,
1000000000000000 , 10000000000000000 , 100000000000000000 , 1000000000000000000 ,
: log10 ( u1 -- u2 )
    \ *G integer based log10
    \ ** O(N): linear search through precomputed array
    \ ** Truncation: use this function to determine the length of a base-10 number
    \ *E : base10length ( u-num -- u-len ) log10 1+ ;
    >r
    log10-values
    begin
        dup @ r@ <=
    while
        cell+
    repeat
    r> drop
    log10-values - cell / ;
: base10length ( n-num -- u-len )
    \ *G number of digits required for displaying n-num in base10 (accounts for negative sign)
    dup 0< if abs 1 else 0 then swap \ add extra length for minus
    log10 1+
    + ; \ extra length

255 constant /line-buffer
/line-buffer buffer: line-buffer
: skip-file-lines ( fd n -- )
    \ *G read ahead n lines
    0 ?do
        dup line-buffer /line-buffer rot read-line throw
        nip \ ignore length
        0= if leave then
    loop drop ;

: draw-file-content ( c-addr u line #lines )
    \ *G draw portion of a file at current cursor position
    {: line #lines :}
    line #lines + base10length 1+ terminal-taboffset !
    r/o open-file throw

        dup line 1- skip-file-lines

        #lines 0 ?do
            dup line-buffer /line-buffer rot read-line throw
            0= if drop leave then
            cr black text-color cyan bright-back-color
            line i + terminal-taboffset @ 1- .r space
            terminal-taboffset @ terminal-x !
            reset-colors
            line-buffer swap line-type
            bl line-fill
        loop

    close-file throw ;

: draw-file ( c-addr u line #lines )
    \ TODO: check log of line + #lines and .r by that
    black text-color cyan bright-back-color
    terminal-x off
    2dup + base10length 1+ line-spaces
    2over line-type bl line-fill
    draw-file-content
    reset-colors ;

: draw-xt ( xt -- )
    (whereis) if
        cr
        16 draw-file
    then ;

: locate ( <parse> -- )
    ' draw-xt ;

: get-file-line ( c-addr u line -- c-addr-line n-line )
    \ *G read single file line
    -rot r/o open-file throw >r
        1- r@ swap skip-file-lines
        line-buffer /line-buffer r@ read-line throw drop
        line-buffer swap
    r> close-file throw ;

: draw-xt-line ( xt -- )
    \ *G print line containing the xt's name in bold
    \ ** if the file-line does not contain \i{name}, print \i{name} instead
    {: xt :}
    xt (whereis) if
        get-file-line
        xt >name count isearch-split if
            line-type
            bold-type
            line-type
        else \ not in line
            ." NOT LINE: "
            2drop bold-type
        then
    else \ no locate information - interactive definition
        xt >name count bold-type
    then ;

update-terminal-size

variable words-offset
variable current-words-offset
variable #drawn-words
variable locate-lines
variable first-word
: draw-word ( nt -- f )
    dup count
    draw-filter count search nip nip 0= if
        drop true exit
    then

    current-words-offset @ 0<> if
        -1 current-words-offset +!
        drop true exit
    then
    erase-line

    1 #drawn-words +!

    name>
    #drawn-words @ 1 = if
        dup first-word !
    then
    red bright-text-color
    dup Immediate? if
        [char] I line-emit
    else
        dup ndcs? if
            [char] D line-emit
        else
            bl line-emit
        then
    then
    reset-colors
    dup hex.  \ show address
    >name count bold-type
    line-cr
    terminal-y @ locate-lines @ + terminal-h @ < ;
: draw-words ( -- )
    #drawn-words off
    words-offset @ current-words-offset !
    default-color text-color
    default-color back-color
    terminal-h @ 1- 2 scroll-margins
    1 2 cursor-position
    ['] draw-word selected-voc @ traverse-wordlist

    terminal-h @ 2- locate-lines @ - #drawn-words @ - dup 0> if
        \ clear remaining screen
        0 do
            erase-line line-cr
        loop
    else
        drop
    then

    first-word @ (whereis) if
        locate-lines @ 1- draw-file
    then

    scroll-fullscreen ;

: browse ( [<parse-name>] -- )
    \ *G enter browse mode
    \ ** optionally parse word and use as filter
    words-offset off
    parse-name draw-filter uplace
    altbuf-enable
    cursor-hide
    update-terminal-size
    first-word off
    terminal-h 2- 2/ 10 min locate-lines !
    erase-display
    draw-footer
    draw-vocs

    begin
        draw-words

        terminal-ekey case
            esc of -1 endof
            k-left  of ." LINKS" 0 endof
            k-right of ." RIGHT" 0 endof
            k-down  of  1 words-offset +! 0 endof
            k-up    of -1 words-offset +! 0 endof
            k-next  of terminal-h @ 2 -        words-offset +! 0 endof
            k-prior of terminal-h @ 2 - negate words-offset +! 0 endof
            k-home  of words-offset off 0 endof
            k-end   of 0 endof \ ignore end (for now)
            k-enter of 0 endof \ ignore return
            k-back  of
                draw-filter c@ 1- 0 max draw-filter c! 0
                draw-footer
            endof
            \ other ? add to filter
            dup upc draw-filter addchar
            words-offset off
            draw-footer
            0 swap
        endcase
        words-offset @ 0< if words-offset off then
    until
    altbuf-disable
    cursor-show ;

\ browse
