\ vt100.fth - esc.ape sequences for terminal(-emulators)

((
Copyright (c) 2021
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: info@vfxforth.com
web: https://vfxforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


Sources
=======
https://www.ecma-international.org/wp-content/uploads/ECMA-48_5th_edition_june_1991.pdf
https://www.ecma-international.org/wp-content/uploads/ECMA-35_4th_edition_march_1985.pdf
https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html
https://invisible-island.net/xterm/ctlseqs/ctlseqs.html

))

[undefined] @(n) [if]
: @(n)          \ addr -- u32
\ *G Network order 32 bit fetch.
  dup c@ 8 lshift
  over 1 + c@ or 8 lshift
  over 2 + c@ or 8 lshift
  swap 3 + c@ or
;
[THEN]


: hex. ( x -- )
    base @ >r hex u. r> base ! ;

: esc ( -- c )
    \ *G escape character ($1b #27)
    $1b ;
: esc. ( -- )
    \ *G emit ESC.
    $1b emit ;
: bel. ( -- )
    \ *G emit BEL.
    $7 emit ;
: csi. ( -- )
    \ *G control system introducer, sends "esc. ["
    esc. [CHAR] [ emit ;
: dcs. ( -- )
    \ *G Device Control String (mapping of C1 $90)
    esc. [CHAR] P emit ;
: st. ( -- )
    \ *G String Terminator (mapping of C1 $96)
    esc. [CHAR] \ emit ;

: param. ( u -- )
    \ csi. parameter
    0 .r ;

: param.; ( n -- )
    \ *G send parameter followed by semicolon
    param.
    [char] ; emit ;

: cursor-up ( n -- )
    \ *G CUU cursor up n lines
    csi. param. [CHAR] A emit ;

: cursor-down ( n -- )
    \ *G CUD cursor down n lines
    csi. param. [CHAR] B emit ;

: cursor-forward ( n -- )
    \ *G CUF move cursor forward
    csi. param. [CHAR] C emit ;

: cursor-back ( n -- )
    \ *G CUB move cursor back
    csi. param. [CHAR] D emit ;

: cursor-line-down ( n -- )
    \ *G CNL move cursor to beginning of line n (relative up)
    csi. param. [CHAR] E emit ;

: cursor-line-up ( n -- )
    \ *G CPL move cursor to beginning of line n (relative down)
    csi. param. [CHAR] F emit ;

: cursor-line ( n -- )
    \ *G CHA move cursor to column n
    \ *H Does not work
    csi. param. [CHAR] G emit ;

variable terminal-x
variable terminal-y
: cursor-position ( n-col n-row -- )
    \ *G CUP move cursor to row n, column n. values are 1-based
    dup  terminal-y !
    over terminal-x !
    csi. param.; param. [CHAR] H emit ;

: erase-in-display ( n-mode -- )
    \ *G erase part of the screen (cursor position unchanged):
    \ ** 0: clear from cursor to end of screen
    \ ** 1: clear from cursor to beginning of screen
    \ ** 2: clear entire screen
    \ ** 3: like 2, but also clear scrollback buffer
    csi. param. [CHAR] J emit ;

: erase-display ( -- )
    \ *G erase entire screen
    2 erase-in-display ;

: erase-in-line ( n-mode -- )
    \ *G erase part of the current line (cursor position unchanged):
    \ ** 0: clear from cursor to end of line
    \ ** 1: clear from cursor to beginning of line
    \ ** 2: clear entire line
    csi. param. [CHAR] K emit ;

: erase-line ( -- )
    \ *G erase entire line
    2 erase-in-line ;

: scroll-up ( n -- )
    \ *G scroll page up by n lines, new lines added at the bottom
    csi. param. [CHAR] S emit ;

: scroll-down ( n -- )
    \ *G scroll page down by n lines, new lines added at the top
    csi. param. [CHAR] T emit ;

\ graphic renditions
: sgr ( n -- )
    \ *G select graphic rendition
    csi. param. [CHAR] m emit ;

\ colors
0 constant black    \ *G black   (3-bit)
1 constant red      \ *G red     (3-bit)
2 constant green    \ *G green   (3-bit)
3 constant yellow   \ *G yellow  (3-bit)
4 constant blue     \ *G blue    (3-bit)
5 constant magenta  \ *G magenta (3-bit)
6 constant cyan     \ *G cyan    (3-bit)
7 constant white    \ *G white   (3-bit)
9 constant default-color \ *G implementation specific default color (4-bit)

: bright-color ( n-3/4-bit-color-dark -- n-3/4-bit-color-bright )
    \ *G converts a 3/4-bit-color to it's bright counterpart
    60 + ;

: text-color ( n-3/4-bit-color -- )
    \ *G set text color
    30 + sgr ;
: back-color ( n-3/4-bit-color -- )
    \ *G set background color
    40 + sgr ;

: reset-colors ( -- )
    \ *G reset text- and back-color
    default-color text-color
    default-color back-color ;

: bright-text-color ( n-3/4-bit-color -- )
    \ *G shorthand for converting a 3/4-bit-color to bright and setting it as text color
    bright-color text-color ;
: bright-back-color ( n-3/4-bit-color -- )
    \ *G shorthand for converting a 3/4-bit-color to bright and setting it as back-color color
    bright-color back-color ;

: (rgb-color) ( x-color -- )
    \ emit 24 bit color
    dup 16 rshift $FF and param.;  \ red
    dup  8 rshift $FF and param.;  \ green
                  $FF and param. ; \ blue

: rgb-text-color ( x-color -- )
    csi. 38 param.; 2 param.; \ sequence
    (rgb-color) [CHAR] m emit ;
: rgb-back-color ( x-color -- )
    csi. 48 param.; 2 param.; \ sequence
    (rgb-color) [CHAR] m emit ;

: reset-text ( -- )
    \ *G reset text to normal - reset all SGR(Select Graphic Rendition) attributes
    0 sgr ;
: toggleable-sgr ( f-enable n-attribute -- )
    \ toggle attribute depening on flag
    swap 0= if 20 + then sgr ;

: bold-text ( f -- )
    \ *G bold or increased intensity
    \ ** note: false resets bold and faint
    if 1 else 22 then sgr ;
: faint-text ( f -- )
    \ *G faint or decreased intensity
    \ ** note: false resets bold and faint
    2 sgr toggleable-sgr ;
: italic-text ( f -- )
    \ *G italic text
    3 toggleable-sgr ;
: underlined-text ( -- )
    \ *G underlined text - mutually exclusive with doubly underlined text
    \ ** note: false resets underlined and doubly underlined
    4 toggleable-sgr ;
: blinking-text ( -- )
    \ *G blinking text
    5 toggleable-sgr ;
: negative-text ( f -- )
    \ *G inverted text
    7 toggleable-sgr ;
: concealed-text ( f -- )
    \ *G hidden text
    8 toggleable-sgr ;
: crossed-out-text ( f -- )
    \ *G stroke / corssed-out-text
    9 toggleable-sgr ;
: doubly-underlined-text ( f -- )
    \ *G doubly underlined text - mutually exclusive with underlined text
    \ ** note: false resets underlined and doubly-underlined
    if 21 else 24 then sgr ;

\ modes
1  constant GUARDED-AREA-TRANSFER-MODE
    \ *G GUARDED AREA TRANSFER MODE (GATM)
2  constant KEYBOARD-ACTION-MODE
    \ *G KEYBOARD ACTION MODE (KAM)
3  constant CONTROL-REPRESENTATION-MODE
    \ *G CONTROL REPRESENTATION MODE (CRM)
4  constant INSERTION-REPLACEMENT-MODE
    \ *G INSERTION REPLACEMENT MODE (IRM)
5  constant STATUS-REPORT-TRANSFER-MODE
    \ *G STATUS REPORT TRANSFER MODE (SRTM)
6  constant ERASURE-MODE
    \ *G ERASURE MODE (ERM)
7  constant LINE-EDITING-MODE
    \ *G LINE EDITING MODE (VEM)
8  constant BI-DIRECTIONAL-SUPPORT-MODE
    \ *G BI-DIRECTIONAL SUPPORT MODE (BDSM)
9  constant DEVICE-COMPONENT-SELECT-MODE
    \ *G DEVICE COMPONENT SELECT MODE (DCSM)
10 constant CHARACTER-EDITING-MODE
    \ *G CHARACTER EDITING MODE (HEM)
11 constant POSITIONING-UNIT-MODE
    \ *G POSITIONING UNIT MODE (PUM)
12 constant SEND/RECEIVE-MODE
    \ *G SEND/RECEIVE MODE (SRM)
13 constant FORMAT-EFFECTOR-ACTION-MODE
    \ *G FORMAT EFFECTOR ACTION MODE (FEAM)
14 constant FORMAT-EFFECTOR-TRANSFER-MODE
    \ *G FORMAT EFFECTOR TRANSFER MODE (FETM)
15 constant MULTIPLE-AREA-TRANSFER-MODE
    \ *G MULTIPLE AREA TRANSFER MODE (MATM)
16 constant TRANSFER-TERMINATION-MODE
    \ *G TRANSFER TERMINATION MODE (TTM)
17 constant SELECTED-AREA-TRANSFER-MODE
    \ *G SELECTED AREA TRANSFER MODE (SATM)
18 constant TABULATION-STOP-MODE
    \ *G TABULATION STOP MODE (TSM)
21 constant GRAPHIC-RENDITION-COMBINATION
    \ *G GRAPHIC RENDITION COMBINATION (GRCM)
22 constant ZERO-DEFAULT-MODE
    \ *G ZERO DEFAULT MODE (ZDM)

: mode-on ( n-mode -- )
    csi. param. [char] h emit ;

: mode-off ( n-mode -- )
    csi. param. [char] l emit ;

\ mode renaming
: keyboard-lock ( -- )
    \ *G lock keyboard
    KEYBOARD-ACTION-MODE mode-on ;
: keyboard-unlock ( -- )
    \ *G unlock keyboard
    KEYBOARD-ACTION-MODE mode-off ;

: insert-mode ( -- )
    \ *G insert characters, move existing characters
    INSERTION-REPLACEMENT-MODE mode-on  ;
: replace-mode ( -- )
    \ *G replace characters at cursor position
    INSERTION-REPLACEMENT-MODE mode-off ;

: echo-on ( -- )
    \ *G disable local echo
    SEND/RECEIVE-MODE mode-on ;
: echo-off ( -- )
    \ *G enable local echo
    SEND/RECEIVE-MODE mode-off ;

\ popular private sequences
: cursor-save ( -- )
    \ *G save current cursor position
    csi. [CHAR] s emit ;

: cursor-restore ( -- )
    \ *G restore cursor position
    csi. [CHAR] u emit ;

: cursor-show ( -- )
    \ *G show cursor (VT220)
    csi. ." ?25h" ;

: cursor-hide ( -- )
    \ *G hide cursor (VT220)
    csi. ." ?25l" ;

: altbuf-enable ( -- )
    \ *G enable alternate screen buffer
    \ ** useful for fullscreen applications without deleting the terminal
    csi. ." ?1049h" ;

: altbuf-disable ( -- )
    \ *G enable alternate screen buffer
    csi. ." ?1049l" ;

: reset-terminal ( -- )
    \ *G reset to initial state
    esc. [CHAR] c emit ;

: scroll-margins ( n-bottom n-top -- )
    \ *G set top and bottom margins (DECSTBM)
    csi. param.; param. [CHAR] r emit ;
: scroll-fullscreen ( -- )
    \ *G reset top and bottom margins (DECSTBM)
    csi. [char] r emit ;

\ os sequences
: window-title ( c-addrn -- )
    \ *G set terminal window title
    esc. ." ]0;"
    type
    bel. ;

\ queries
16 constant /vt100-read-buf
/vt100-read-buf buffer: vt100-read-buf
: read-unprefixed-sequence ( n-offset -- c-addr n )
    \ receive vt100-sequence using 'key' without leading ESC[
    begin
        key >r
        r@ over vt100-read-buf + c! \ add char to buffer
        r@ [char] 0 >=
        r> [char] ; <=
        and
    while
        1+
        dup /vt100-read-buf >= abort" vt100 receive buffer overflow"
    repeat vt100-read-buf swap 1+ ;

: read-sequence ( -- c-addr n )
    \ *G receive vt100-sequence using 'key'
    key esc <> abort" vt100 expected ESC"
    key [char] [ <> abort" vt100 expected ["
    0 read-unprefixed-sequence ;

: between ( n-value n-min n-max -- f )
    \ check if value is between min and max
    rot >r
    r@ >=
    swap r> <= and ;

: isdigit? ( c -- f )
    \ returns true if c is a digit
    [char] 0 [char] 9 between ;

variable vt100-parameters
: parse-parameters ( c-addr n -- i*n n c-addr n )
    \ *G parse received string into parameters and remainder
    0 vt100-parameters !
    begin
        over c@ isdigit?
    while
        1 vt100-parameters +!   \ increment parameter count
        0.0 2swap >number       \ convert number
        rot drop                \ keep on using remaining string, drop double

        over c@ [char] ; = if      \ skip separator (if any)
            1 /string
        then
    repeat
    vt100-parameters @ -rot ;

: parse-expected ( c-addr-sequence n-sequence c-command n-parameters -- i*n )
    \ *G parse expected sequence and n parameters, abort on error
    >r >r
    parse-parameters
    1     <> abort" vt100 expected 1 as sequence remainder"
    c@ r> <> abort" vt100 unexpected sequence received"
    r>    <> abort" vt100 expected different parameter count" ;

: read-expected ( c-command n-parameters -- i*n )
    \ *G read expected sequence and n parameters, abort on error
    2>r read-sequence
    2r> parse-expected ;

: cursor-position? ( -- n-cols n-rows )
    \ *G get current cursor position
    csi. 6 param. [char] n emit
    [char] R 2 read-expected swap ;

\ TODO: catch SIGWINCH?
: terminal-size? ( -- n-cols n-rows )
    \ *G get terminal size
    cursor-save
    999 999 cursor-position
    cursor-position?
    cursor-restore ;

variable terminal-w \ *G terminal width in characters
variable terminal-h \ *G terminal height in characters
: update-terminal-size ( -- )
    \ *G query terminal size and save in \fo{terminal-w} and \fo{terminal-h}
    terminal-size? terminal-h ! terminal-w ! ;

variable terminal-linebreaks \ *G when true line-emit and line-type allow line-break
terminal-linebreaks on
variable terminal-taboffset \ *G offset tabs by this value. useful for displaying file with prefix line-numbers
terminal-taboffset off
: line-emit ( c -- )
    \ *G restrict emits per line to \fo{terminal-w} (if \fo{terminal-linebreaks} is set )
    \ ** only works in after setting \fo{terminal-w} and \fo{terminal-x}.
    \ ** Hint: this can be done by running \fo{update-terminal-size}
    \ ** followed by \fo{cursor-position}
    terminal-x @ terminal-w @ <=
    terminal-linebreaks @ or if
        dup 9 = if
            drop
            8 terminal-x @ terminal-taboffset @ - 8 mod -
            dup spaces
        else
            emit 1
        then
        terminal-x +!
    else
        drop
    then ;

: line-type ( c-addr n -- )
    \ *G restricts output to current line if \fo{terminal-linebreaks} is set (see \fo{line-emit}).
    bounds ?do
        i c@ line-emit
    loop ;
: line" ( <parse> -- )
    \ *G compile string into current definition, type on execute
    postpone s" postpone line-type ; immediate	\ "

: line-fill ( c -- )
    \ *G fill remaining line with c
    terminal-w @ terminal-x @ - dup 0> if
        0 do
            dup emit
        loop drop
    else
        2drop
    then ;

: line-space ( -- )
    \ *G emit single space
    bl line-emit ;

: line-cr ( -- )
    terminal-y @ terminal-h @ < if
        1 terminal-y @ 1+ cursor-position
    then ;

: line-spaces ( n -- )
    \ *G emit n spaces
    0 ?do
        line-space
    loop ;

\ output helpers
: bold-type ( c-addr n -- )
    -1 bold-text line-type 0 bold-text ;
: bold" ( <parse"> -- )
    postpone s" ( ") postpone bold-type ; immediate

: cell-byte-mask ( n -- x )
    \ masks the first n bytes
    0 over 0 ?do   \ positive mask
        8 lshift
        $FF or
    loop
    swap \ -- mask n
    cell swap - 8 * lshift ;

: str>cell ( c-addr n -- x )
    \ get string in single cell, tailing chars zerod
    cell-byte-mask
    swap @(n) and ;

: k-key ( C: <create> <parse> -- ; R: -- x )
    \ *G parses a string and saves its numeric value.
    \ ** that value is returned during runtime
    create
        parse-name str>cell ,
    does> @ ;

k-key k-down  [B
k-key k-end   [F
k-key k-f1    [11~
k-key k-f10   [21~
k-key k-f11   [23~
k-key k-f12   [24~
k-key k-f2    [12~
k-key k-f3    [13~
k-key k-f4    [14~
k-key k-f5    [15~
k-key k-f6    [17~
k-key k-f7    [18~
k-key k-f8    [19~
k-key k-f9    [20~
k-key k-home  [H
k-key k-left  [D
k-key k-next  [6~
k-key k-prior [5~
k-key k-right [C
k-key k-up    [A
$a constant k-enter
$8 constant k-back

\ input helpers
: terminal-ekey ( -- x )
    \ *G returns key, if escape sequence is passed, contents are returned in cell
    \ ** returns 0 if escape sequence is longer than cell
    key case
        esc of key? if
                vt100-read-buf off \ disable first cell
                key dup vt100-read-buf c!  \ store as first character to avoid ambigious returns
                case \ check for escape sequence
                    [char] [ of
                        1 read-unprefixed-sequence
                        cell > if
                            drop 0 \ too long
                        else
                            @(n) \ get first <cell> bytes
                        then
                    endof
                    [char] O of \ SS3 Single Shift 3
                        key dup vt100-read-buf 1+ c! \ save single byte
                        case
                            \ fix F1-F4
                            [char] P of k-f1 endof
                            [char] Q of k-f2 endof
                            [char] R of k-f3 endof
                            [char] S of k-f4 endof
                            \ pass on
                            vt100-read-buf @(n) swap
                        endcase
                    endof
                    0 swap \ unknown escape sequence
                endcase
            else
                esc \ just plain escape
            then
        endof
        $7F of k-back endof \ some terminals send $7F instead of $8 for backspace
        dup
    endcase ;

: show-ekey ( -- )
\ *G catch e-key and display sequence
    terminal-ekey
    vt100-read-buf 4 type ;
