\ DebugExit.fth

((
Copyright (c) 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *! debugexit
\ *T Debugging at ; and EXIT
\ ===========
\ *P This file shows two ways in which a simple debugger can be
\ ** implemented by inserting code at each use of *\fo{EXIT} and
\ ** *\fo{;}. The first way inserts extra code just before the
\ ** *\fo{RET} instructions. However, this makes it difficult to
\ ** debug with local variables as these have already been removed.
\ ** The second version overcomes this problem at the expense of
\ ** some complexity.


\ *****************
\ *S Simple version
\ *****************
\ *P This version is normally commented out.

((
defer DebugHook	\ --
\ *G A call to this *\fo{DEFER}red word is compiled before the *\fo{RET}
\ ** instructions compiled by *\fo{EXIT} and *\fo{;}. Defaults to *\fo{NOOP}.
  ' noop is DebugHook

also h-c
: DebugExit,	\ -- ; compile code for EXIT
  not-inlined-exit
  c_shuffle  reset-opt
  postpone DebugHook  discard-sinline
  a[ ret ]a  use-a
;

: DebugSemi,	\ -- ; compile code for ;
  c_shuffle  reset-opt
  postpone DebugHook  discard-sinline
  a[ ret ]a  use-a
  end-inline
;

: [DebugExit	\ --
\ *G Add a debugging check at every use of *\fo{EXIT} and *\fo{;}.
  assign DebugExit, to-do doexit,
  assign DebugSemi, to-do dosemicolon,
;

: DebugExit]	\ --
\ *G Restore normal compilation of words.
  assign c_exit      to-do doexit,
  assign c_;         to-do dosemicolon,
;
previous
))

\ ***************
\ *S Full version
\ ***************

defer DebugHook	\ --
\ *G A call to this *\fo{DEFER}red word is compiled before the *\fo{RET}
\ ** instructions compiled by *\fo{EXIT} and *\fo{;}. Defaults to *\fo{NOOP}.
  ' noop is DebugHook

0 value DebugExit?	\ -- flag
\ *G Return true if debug code is required.

: [DebugExit	\ --
\ *G Add a debugging check at every use of *\fo{EXIT} and *\fo{;}.
  1 to DebugExit?
;

: DebugExit]	\ --
\ *G Restore normal compilation of words.
  0 to DebugExit?
;

also h-c
: ;             \ --
  DebugExit? 0=
  if  postpone ;  exit  then
  ?comp
  OperatorType @off #-402 ?throw	\ check for unconsumed operator
  <HeaderLess> @ 0=
  if  ?csp reveal  then
  discard-sinline			\ no tokenising
  postpone DebugHook			\ call to DEFERred word
  drop-locals                           \ compiles in stack cleanup
  cleanup-locals                        \ cleans up local dictionary etc.
  dosemicolon,                          \ close colon definition
  state off
  last-colon-xt off
  0 -> original-xt			\ out of any redefinition
  end-sinline
; immediate

: EXIT          \ -- ; R: next-sys --
\ *G Compile code into the current definition to cause a definition to
\ ** terminate. This is the Forth equivalent to inserting an RTS/RET
\ ** instruction in the middle of an assembler subroutine.
  NoInterp  ;
ndcs: ( -- )
  discard-sinline
  DebugExit?
  if  postpone DebugHook  then
  drop-locals
  doexit,
;
previous


\ *******************
\ *S Example debugger
\ *******************

: MyDebugger	\ --
\ *S Installed in the debug hook. This one tells you which word
\ ** is being left and displays the stack. Like all debuggers,
\ ** a little carnal knowledge is required.
  r@ ip>nfa cr ." Leaving " .name  .s key drop
;  ' MyDebugger is DebugHook

[debugexit
: a  1 2  ;
: b  a a 2drop . .  ;
debugexit]


\ ======
\ *> ###
\ ======

decimal

