\ M4th Compatibility Stuff
\ Last Revision: 20 Jan 2012  05:49:53  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


\ Restores a number of useful MacForth primitives


only Forth definitions
>M4TH


decimal



: >VFX
	only Forth definitions
;

: >VFX+
	only Forth definitions
	also m4th
;


: LookAhead ( -- c-addr u )	\ look ahead and grab next word in input stream
								\ without changing compilation
	save-input
	bl parse-word 2>R
	restore-input drop	\ ? error check ?
	2R>	
;


: COMPILING  (  -- flag |  set if compiling *)
	state @ ;


: NEEDED ( #items --  abort if too few )
	>r depth r>  <  abort" not enough items on stack"
;


: BOOLEAN ( n -- flag )
	0<>
;

: DropFalse ( n -- 0 )
	drop false
; 


: @@  ( addr -- n | fetch n from addr  indirect *)
	@ @ ;
	

: TOGGLE  ( ADDR\MASK--   | TOGGLE THE CONTENTS OF ADDR *)
	\	Complements (XOR) the eight-bit value in addr by the bit mask given.
	over c@ XOR
	swap c!
;


: TOGGLE32 ( addr -- | toggles addr contents )
	dup @ 0= swap ! ;



:  @POINT  ( ADDRESS -- X\Y )
	dup w@s swap 2 +  w@s
	SWAP ;


:  !POINT  ( X\Y\ADDRESS --  )
	dup >R   w! r> 2 + w!				\ INTEL
;


: +POINT  ( x1\y1\x2\y2 --   x3\y3 --- sum vector )
	{ x1 y1 x2 y2 }
	x1 x2 +  y1 y2 +
;


: -POINT  ( x1\y1\x2\y2 --   x3\y3 --- sum vector )
	{ x1 y1 x2 y2 }
	x1 x2 -  y1 y2 -
;


:  SortTwo  ( n1\n2 -- nLo\nHi )
\ 	If n2 is less than n1, it swaps them.
	2dup >
	IF
		swap
	THEN
;


: EXTEND ( 16 bit -- 32 bit - sign extended )
	sp@ w@s swap drop
;


: Lo2  ( n1 -- n2 ) \ Sets the upper 16 bits of n1 to 0, leaving the low 16 bits unchanged
	$FFFF and
;


: HI2  ( n1 -- n2 ) \ Sets the low 16 bits of n1 to 0, leaving the high 16 bits unchanged.
	$FFFF0000 and
;


: HI2LO  ( n1 -- n2 ) \ Returns the high 16 bits of n1 as the low 16 bits of n2.
					  \ The high 16 bits of n2 are all zeros.
					  \ 12345678 becomes 1234
	16 rshift
;


: SWAPW ( n -- n' )
	\ swap the high and low halves of TOS
	\ equivalently:  ( hi:lo -- lo:hi )	$12345678 becomes $56781234
	16 ROR
;


: EndianSwap32	( n -- n' )  \ $12345678 becomes $78563412
	dup  24 rshift
	over $00FF0000 and 8 rshift OR
	over $0000FF00 and 8 lshift OR
	swap 24 lshift OR
;


: POINT>XY  ( POINT--X/Y )
	dup Hi2Lo extend 
	swap Lo2 extend ;


: XY>POINT  ( X\Y -- POINT )
	swap 16 lshift swap Lo2 or ;


: !RECT  ( T\L\B\R\ADDR -- STORE RECT )
	tuck 6 + w!   tuck 4 + w!   tuck 2 + w!   w! ;


: @RECT  ( ADDR -- T\L\B\R    FETCH RECTANGLE )
	dup w@s  swap 2 + dup w@s   swap 2 + dup w@s  swap 2 + w@s ;

: RECT  ( t\l\b\r -- )
	create  here 8  allot  !rect  ;


: w@s-e ( addr -- n )  \ w@s  with 16 bit endian swap
	dup c@ 8 lshift swap 1+ c@  + extend
;

\ some old sources are in the old endian form - picture resources are one
: @RECT-E  ( ADDR -- T\L\B\R    FETCH RECTANGLE )		\ resource rects need endian shift
	dup w@s-e  swap 2 + dup w@s-e   swap 2 + dup w@s-e  swap 2 + w@s-e ;


: XY><TLBR  ( X1\Y1\Y1\Y2 --  T\L\B\R  OR VICE VERSA)
	>R >R swap R> R> swap ;


: EXECUTE
\ a little slower, but safer
	?DUP IF EXECUTE THEN
;


: @EXECUTE  ( addr -- | execute token at addr *)
	@ execute ;


: RANGE  ( n\min\max -- n\flag | flag=true if min<=n<=max *)
	{ n min max }
	n
	n min >=
	n max <=
	AND
;

: RANGE.OF  ( n\min\max -- n |or| [] )
	postpone range 
	postpone ?OF
	postpone drop
; immediate

((

: BB
	CASE
		1 OF ." 1" ENDOF
		2 4 range.of ." 2-4 " ENDOF
\		2 4 range ?OF drop ." 2-4 " ENDOF
\	dup	2 4 within ?OF drop ." 2-4 " ENDOF
		5 OF ." 5" ENDOF
		.s
	ENDCASE ;

))



: BURY   \ ...depth_items...  n  depth -- ...depth+1_items...
	cells
	{ tos  thedelta }
	thedelta
	IF
		0 					\ push new tos
		sp@  dup cell-  thedelta cell+  cmove
		sp@ thedelta + tos swap !
	ELSE
		tos
	THEN
;




: LACKING  ( "<spaces>name" -- ) \ execute rest of line if name is not found
	bl word find  
	nip 
	IF  Postpone \ THEN  ;  IMMEDIATE


: HAVING  ( "<spaces>name" -- ) \ execute rest of line if name is found
	bl word find
	nip
	0= IF  Postpone \ THEN  ;  IMMEDIATE


: POSSIBLY  ( "<spaces>name" -- ) \ if name found, execute the word
	?exec
	bl word find 
	IF
		execute
	ELSE
		drop
	THEN  ;  immediate



: 1ARRAY  ( # of elements\element width --  | compile time )
		  ( index -- addr | run time )
	create
	dup  w,					\ elem width
	over w,					\ #elems
	*  allot&erase			\ here swap dup allot erase
	does>					\ index\base
		2dup 				\ index\base
		2 + w@   
		>= abort" 1ARRAY - index out of range" 		\ index.error	
		dup w@				\ index\base\width
		rot *
		+ 4 +
;


: ASCII ( "name"  ;  -- N )
\	Used in the form ASCII P.  Similar to CHAR & [CHAR]
\	ASCII will accept up to four characters. For example: ASCII PICT
\	ASCII STR_.    (Underline characters in any other than first position
\	are converted to spaces.) Only the last four characters are converted.

	0 bl  parse-word dup 1 >
	IF   over + swap
         DO   8 lshift i c@ dup 95 =
			  IF drop bl THEN
			  or
		 LOOP
	ELSE drop c@ or
	THEN
	
	state @ 		\ are we compiling?
	IF
		Postpone literal
	THEN	
;  immediate


: (H.)  ( n -- addr\len )
	base @ swap hex 	\ 8 U.R 
	s>d  <# # # # # ascii : hold # # # # ascii $ hold #>
	rot base ! ;

: H.  ( n -- )
	(H.) type ;


: EXC-DEPTH  ( -- #bytes )		\ exception stack depth in BYTES
	 exc0 @  excp @ - ;




\ *******************************************************

create BL$ 		1 C, bl C,			\ BL in a string.
create CR$ 		1 C, 13 C,			\ CR in a string.
create CR$CR$ 	2 C, 13 C,  13 C,	\ 2 CR's in a string.
create zCR		$0d c, 0 ,			\ CR as zString

create LF$ 		1 C, 10 C,			\ LF in a string.
create LF$LF$ 	2 C, 10 C, 10 C,	\ 2 LF's in a string.
\ create CRLF$ 	2 C, 13 C, 10 C,	\ CRLF in a string.  - in KERNEL.FTH
create TAB$ 	1 C, 9 C,


create ""  		0 w,				\ null string and null zstring

\ *******************************************************
\ *******************************************************




