\ M4th Compatibility Stuff
\ Last Revision: 20 Jan 2012  05:49:53  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


((

Contains a number of "MacForth-isms"

PRIOR.STREAM
ALLOTERASE
COMPILING
NEEDED
BOOLEAN
NOT
<ROT
@@
TOGGLE
<W@  >W@<  >W!<
@POINT  !POINT  +POINT  -POINT
SortTwo
EXTEND
LO2  HI2  HI2LO
SWAPW
POINT>XY  XY>POINT
RECT !RECT  @RECT
XY><TLBR
@EXECUTE
RANGE
RANGE.OF
DropFalse
BURY
LACKING
HAVING
POSSIBLY
""
BL$
CR$
CR$CR$
zCR
LF$



A few new words:

EndianSwap32
@RECT-E  ( like @RECT, but endian swaps)


as well as warnings for some "naughty" words that should be replaced

LocalBuff
localBufSize
R>DROP
DUP>R
MASK.HANDLE
H@
H@@
DBP+
DBP-
DBP@
)@
)!
),
LastCreatedToken
Field
IN.HEAP
0MAX
0>=
LOCAL.VOCAB
FlushDataCache  FlushCaches  FlushCacheRange
BS

))

only Forth definitions
>M4TH


decimal



\ ***************************************************************************
\ these words are MacForth words that are replaced with VFX synonyms
\ they will compile file, but when you see the warningsm i is recommended that
\ you change your source so that the optimizer can work most effectively.


: ReplacedWith ( caddr\len -- )
	cr ." Obsolete MacForth word: " 
	LastNameFound ?dup IF .NAME ELSE ." ?? " THEN
	."  replaced with " 2dup type
	evaluate	
	compiling
	IF
		."  - compiled into "
		last-colon-xt @  >name .NAME
	ELSE
		."  interpreted at  "
		ShowErrorLine
	THEN
;


\ ForbidColorizing		- see Colorize Editor
: PRIOR.STREAM ( -- )
\ AllowColorizing
	s" STOPINCLUDING" ReplacedWith
	postpone StopIncluding
; immediate



: NOT ( n -- flag ) 		\ NOTE BENE !!!!
	s" FALSE=" ReplacedWith
;  immediate



: 0$Len
	s" zstrlen" ReplacedWith
; immediate



: R>DROP
	s" R> DROP" ReplacedWith
; immediate


: DUP>R
	s" DUP >R " ReplacedWith
; immediate


: <ROT  ( a\b\c -- c\a\b )
	s" -ROT " ReplacedWith
; immediate


: ALLOTERASE  ( n -- )
	s" ALLOT&ERASE" ReplacedWith
; immediate



: <W@ ( addr -- n )		\ sign extended w@
	s" w@S" ReplacedWith
; immediate



: H@@  ( addr -- n )
	s" @@" ReplacedWith
; immediate


: H@
	s" @" ReplacedWith
; immediate



: >W@<  ( ADDR -- N | addr may be odd )
	s" w@" ReplacedWith
; immediate


: >W!<  ( N\ADDR --  | addr may be odd)
	s" w!" ReplacedWith
; immediate


: )@
	s" @" ReplacedWith
; immediate


: )!
	s" !" ReplacedWith
; immediate


: ),
	s" ," ReplacedWith
; immediate


: RASHIFT ( x1 u -- x2 )
	s" arshift" ReplacedWith
; immediate


: 0max
	s" 0 max" ReplacedWith
; immediate


: 0>=
	s" 0 >=" ReplacedWith
; immediate



\ ***************************************************************************
\ A few words that are no longer in use and not needed

: AnnounceObsolete ( -- )
	cr ." Obsolete MacForth word IGNORED:  "
	LastNameFound .NAME
	compiling
	IF
		."  compiled into "
		last-colon-xt @  >name .NAME
	ELSE
		."  interpreted at  "
		ShowErrorLine
	THEN
;  


: FlushDataCache ( -- )
	AnnounceObsolete  ; immediate


: FlushCaches 
	AnnounceObsolete  ; immediate


: FlushCacheRange 
	AnnounceObsolete  ; immediate


: MASK.HANDLE  ( addr -- addr *)
	AnnounceObsolete  ; immediate


: IN.HEAP  
	AnnounceObsolete  ; immediate


: DBP+
	AnnounceObsolete
; immediate


: DBP-
	AnnounceObsolete
; immediate




\ ***************************************************************************
\ A few more warnings about some possible name conflicts


variable SyntaxWarnings   SyntaxWarnings ON


: AnnounceSyntaxIssue ( cstr -- )
	SyntaxWarnings @
	IF
		cr ." MacForth Syntax differs for "
		LastNameFound ?dup IF .NAME ELSE ." ?? " THEN
		compiling
		IF
			."  - compiled into "
			last-colon-xt @  >name .NAME
		ELSE
			."  interpreted at  "
			ShowErrorLine
		THEN
		cr ."   **CAREFUL** " $.
	ELSE
		drop
	THEN
;


: Field
	c" check VFX syntax or use mfFIELD"  AnnounceSyntaxIssue
	compiling
	IF
		postpone field
	ELSE
		field
	THEN
; immediate


: MOVE
	c" check VFX syntax or use qdMove" AnnounceSyntaxIssue
	compiling
	IF
		postpone move
	ELSE
		move
	THEN
; immediate


: BS
	c" BS is a VERB under VFX" AnnounceSyntaxIssue
	compiling
	IF
		postpone BS
	ELSE
		BS
	THEN
; immediate






\ ***************************************************************************
\ A couple that require you to manually replace with alternates

: AbortObsolete ( cstr -- )
	AnnounceObsolete
	cr $.
	true abort"  Replace It"
;


: "
	c\" use C\" instead" AbortObsolete
; immediate


: 0"
	c\" use Z\" instead" AbortObsolete
; immediate


: 0,"
	c\" use Z\", instead" AbortObsolete
; immediate


: LastCreatedToken
	c" use LATEST-XT (a value) instead" AbortObsolete
; immediate


: UNIQUE.MESSAGE
	c" replace UNIQUE.MESSAGE with +WARNINGS or -WARNINGS or <flag> to WARNINGS?"
	AbortObsolete
;  immediate


: LocalBuf
	c" replace LocalBuff with { .. | arr[ nn ] } locals"
	AbortObsolete
; immediate


: MakeProcPtr
	c" replace MakeProcPtr with CALLBACK: - see VFX manual"
	AbortObsolete
; immediate


: ExternFunc
	c" replace ExternFunc with EXTERN: - see VFX manual"
	AbortObsolete
; immediate


: ExternProc
	c" replace ExternProc with EXTERN: - see VFX manual"
	AbortObsolete
; immediate



\ ***************************************************************************
\ A couple that just need to be removed


: !Obsolete ( -- )
	true abort" Obsolete function"
;


: LOCAL.VOCAB
	AnnounceObsolete
	true abort" No longer appropriate"
; immediate


: localBufSize  ( localBufPtr -- size )
	AnnounceObsolete
	!Obsolete
; immediate


: DBP@
	AnnounceObsolete
	true abort" No longer needed"
; immediate


: :|
	true abort" Not supported"
; immediate
