\ MACFORTH CHAINS
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************



only Forth definitions
>M4TH

decimal



CREATE ChainOfMFChains   0 ,  0 ,


: MFChain  ( "<spaces>name" -- )		\ special chain
	create
	0 ,									\ anchor for offspring chain
	ChainOfMFChains link,				\ ChainOfMFChains link field
;

(( 
linked entries  created by LinkToken or LinkLastToken are of the form:
<linktoprior>  
XT
 ))




: ToEveryMFChain ( xt -- )	\ XT takes ( ...\chainelement -- ...\  )
\ by keeping link as a local, should be mostly immune 
\ to stack imbalances caused by midbehaving XT's
	ChainOfMFChains @
	locals| link XT |
	begin
		link
	while
		link XT execute
		link @ -> link
	repeat
;


: ShowItsName ( xt -- )
	cr >name ?dup IF .name ELSE drop ." ???" THEN
;

: ShowCoC ( -- )
	['] ShowItsName ToEveryMFChain
;


: ShowMFTokenChain	\ anchor --
\ Shows a token chain created by mfCHAIN * LinkLastToken
\ first cell has chain's link field
\ second cell has the linked token
  @
  begin
    dup
  while
    dup  cell + @  ShowItsName
    @
  repeat
  drop
;


(( 
: ShowMFChainLinks	\ anchor --
\ Shows chains created by mfCHAIN
\ first cell has chain's link field
\ second cell has CoC link
  @
  begin
    dup
  while
    dup  ShowItsName
    @
  repeat
  drop
;
 ))


: <ShowMFChain  ( ^LF -- )  \ Pointer to ChainOfMFChains link field	
	4-  		\ ChainOfMFChains link field  --> MFChain
	dup ShowItsName 	\ name of this MFChain
	dup . 2 spaces		\ show its xt
	ShowMFTokenChain	\ show elements in this chain
	cr 
;


: ShowMFChain  ( anchor -- )
	4 +  <ShowMFChain
;


: DetailCoC  ( -- )
	['] <ShowMFChain  ToEveryMFChain
;




also system
: PruneOneMFChain { start end anchor -- start end  }
	start anchor  >
	IF			\ if anchor is in range being pruned, don't trim its chain
				\ should not happen !!
		start end anchor 4 -  	\ need head of the "daughter chain"
		trim-links
	THEN
	start end
;


: PruneChainOfMFChains { start end -- start end }
	\ when any MARKER-based action occurs, clean up MFChains
	\ prune ChainOfMFChains first ( to remove anchors of chains in the range )
	start end ChainOfMFChains trim-links

	\ then prune any remaining offspring chains
	start end
	['] PruneOneMFChain
	ToEveryMFChain ( start end xt -- start end )
;

' PruneChainOfMFChains prunes

previous


((
: MFChainExecute   ( ...\headPtr -- ... )		\ LATER
	drop
;
))

MFChain ON.FORGET.LINKAGE

MFChain NULLEVENTER		\ executed during a null event
MFChain PRELUDES		\ executed at cold start - before snapshot/turnkey tokens
MFChain RESTORER		\ executed at warm start - after snapshot token - NOT in TUrnkeys
MFChain QUITTER			\ executed early in QUIT
MFChain EVENTER			\ executed early in DO.EVENTS to allow special casing
						\ ( event -- event' ) - may modify event, be careful of 
MFChain BeforeMenuSelect	\ executed before >MENUBAR
MFChain PRESNAPSHOT			\ executed near the start of snapshot
MFChain EXODUS			\ executed in BYE
\ SE.LINK				\ defined in Editor


: MFChainSearch  ( ...\token\headPtr -- [...\nodePtr]or[...\0] )
	\ Apply token to every item in a relative chain.  Token behaves as:
	\	( ...\nodePtr -- ...\stop? )
	\	nodePtr points to the current link.
	\	stop? is a flag, TRUE to stop the search, FALSE to continue.
	\ The nodePtr returned from ChainSearch is the node where the stop?
	\ flag was returned TRUE, or 0 if every node was visited and none
	\ signalled TRUE.  The ellipses (...) indicate any number of
	\ optional user-maintained parameters.
	\   This scan algorithm is slower than possible, so the user-token
	\ can ChainCut the nodePtr from the chain without killing the scan.
	dup @
	locals|  nextPtr nodePtr theToken  |
	BEGIN
		nextPtr dup
	WHILE						\ if last node, exit with 0
		dup to nodePtr			\ old nextPtr becomes new nodePtr
		dup @ to nextPtr		\ save new nextPtr
		theToken execute if		\ execute token, stop if flag is TRUE
			nodePtr exit		\ ..giving back the nodePtr
		then
	REPEAT  ;

(( POSSIBLY USEFUL
: MFChainApply  ( token\headPtr  -- )
	\ Apply token to every item in a relative chain.  Token behaves as:
	\	( nodePtr --  )
	\	nodePtr points to the current link.
	\	stop? is a flag, TRUE to stop the search, FALSE to continue.
	\ The nodePtr returned from ChainSearch is the node where the stop?
	\ flag was returned TRUE, or 0 if every node was visited and none
	\ signalled TRUE.  The ellipses (...) indicate any number of
	\ optional user-maintained parameters.
	\   This scan algorithm is slower than possible, so the user-token
	\ can ChainCut the nodePtr from the chain without killing the scan.
	dup @
	locals|  nextPtr nodePtr theToken  |
	BEGIN
		nextPtr dup
	WHILE						\ if last node, exit with 0
		dup to nodePtr			\ old nextPtr becomes new nodePtr
		dup @ to nextPtr		\ save new nextPtr
		theToken execute
	REPEAT  ;
))


: LinkToken  ( token\headPtr -- )
	link,  ,  ;  \ the token is the payload


: LinkLastToken  ( headPtr -- )
	latest-xt
	swap LinkToken ;


: ExecMFChain&Show	\ anchor --
\ adaptation from ExecChain 
\ *G Execute the contents of chain with the following structure:
\ *C   link | xt | ...
\ *P Each word that is run has the stack effect
\ *C   link -- link
\ *P Where link is the address of the link field in the structure.
\ ** Thus, data that follows the xt can easily be accessed.
  @
  begin
    dup
  while
    dup cell + @ 
    ShowItsName			\ cr dup >name  .name
    execute
    @
  repeat
  drop
;


: ExecMFChainFromOldest        \ anchor -- 
\ like ExecChain, but oldest->newest
\ derived from WalkColdChain

  $DEADBEEF >r
  @
  begin
    dup
  while
    dup cell+ @ >r
    @
  repeat
  drop

  begin
    r@ $DEADBEEF <>
  while
    r> execute
  repeat
  r> drop
;


: ExecMFChainFromOldest&Show        \ anchor -- 
  cr ." Executing Chain: " dup >name .name
  $DEADBEEF >r
  @
  begin
    dup
  while
    dup cell+ @ >r
    @
  repeat
  drop

  begin
    r@ $DEADBEEF <>
  while
    cr r@ 3 spaces >name .name
    r> execute
  repeat
  r> drop
;


\ This is unlike 
: ON.FORGET		\ --     "<spaces>name"
	'
	on.forget.linkage link,
	,			\ save XT of newly added ON.FORGET word
;

prune:
\ ON.FORGET PRUNER
{: start end | link -- start end :}
\ on any MARKER-based action, execute any ON.FORGET words in the range being forgotten
\ by putting the link in a local, we should be somewhat protected against 
\ ON.FORGET words with innappropriate stack behavior
  	on.forget.linkage @ -> link
  	BEGIN
    	link
  	WHILE
    	link start end within
    	IF
    		link cell + @
    		cr ." executing ON.FORGET: " dup >name  .name
    		execute
    	THEN
    	link @ -> link
	REPEAT
  
	start end 		\ be sure to restore range for next pruner
;


: BetterQuit ( -- )
	skip.events OFF
	quitter execchain
	(quit)
;

' BetterQuit is QUIT



