\ EVENT STUFF
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


((

This is the traditional MacForth event handling code

WaitNextEvent - based stuff may eventually disapper, 
	  but too much stuff relies on it for now

Besides, Apple's StandardEventHandler is basically doing the same things
in handling "standard events"  that our event loop has always done

In many ways, our event handling offers more programmer control

Carbon Events (defined in a later file) allow access to specialized event
handling where necessary and coesist with our stuff nicely

Carbon Events DO require either a WaitNextEvent loop or a StandardEventHandler
to run - they are NOT run as interrupt level callbacks !!

))


only Forth definitions
>M4TH
decimal



\	MODIFIER KETS ----------------------------------------------------

\ these work on events that have been read into EVENT.RECORD
\ use _GetKeys other times ( see OptionKey? or ShiftKey? )
: Modifier?  ( bitMask -- flag )  event.record 14 + w@  and boolean ;
: ?CmdKey ( -- flag )		 256 Modifier?  ;  \ true if cmd key pressed
: ?ShiftKey ( -- flag )	 	 512 Modifier?  ;  \ true if shift key pressed
: ?CapsLock ( -- flag )		1024 Modifier?  ;  \ true if CAPS LOCK key pressed
: ?OptionKey  ( -- flag )	2048 Modifier?  ;  \ true if option key pressed
\ : OPTION.KEY?  ( -- flag )	2048 Modifier?  ;  \ true if option key pressed
: ?CntrlKey ( -- flag )		4096 Modifier?  ;  \ true if Control key pressed

((   see Events.h
typedef UInt16                          EventModifiers;
enum {
                                        /* modifiers */
  activeFlagBit                 = 0,    /* activate? (activateEvt and mouseDown)*/
  btnStateBit                   = 7,    /* state of button?*/
  cmdKeyBit                     = 8,    /* command key down?*/
  shiftKeyBit                   = 9,    /* shift key down?*/
  alphaLockBit                  = 10,   /* alpha lock down?*/
  optionKeyBit                  = 11,   /* option key down?*/
  controlKeyBit                 = 12,   /* control key down?*/
  rightShiftKeyBit              = 13,   /* right shift key down? Not supported on Mac OS X.*/
  rightOptionKeyBit             = 14,   /* right Option key down? Not supported on Mac OS X.*/
  rightControlKeyBit            = 15    /* right Control key down? Not supported on Mac OS X.*/
};

enum {
  activeFlag                    = 1 << activeFlagBit,
  btnState                      = 1 << btnStateBit,
  cmdKey                        = 1 << cmdKeyBit,
  shiftKey                      = 1 << shiftKeyBit,
  alphaLock                     = 1 << alphaLockBit,
  optionKey                     = 1 << optionKeyBit,
  controlKey                    = 1 << controlKeyBit,
  rightShiftKey                 = 1 << rightShiftKeyBit, /* Not supported on Mac OS X.*/
  rightOptionKey                = 1 << rightOptionKeyBit, /* Not supported on Mac OS X.*/
  rightControlKey               = 1 << rightControlKeyBit /* Not supported on Mac OS X.*/
};

 ))


\	EVENT NAMES ----------------------------------------------------

  0 constant NULL.EVENT			1 constant MOUSE.DOWN
  2 constant MOUSE.UP			3 constant KEY.DOWN
  4 constant KEY.UP				5 constant AUTO.KEY
  6 constant UPDATE.EVENT		\ 7 constant DISK.EVENT 
  8 constant ACTIVATE.EVENT		9 constant ABORT.EVENT
\ 10 constant NETWORK.EVENT		\ 11 constant DRVR.EVENT

 14 constant MOUSE.MOVED		15 constant SWITCH.EVENT
 16 constant COMMAND.KEY		17 constant IN.DESKTOP
 18 constant IN.MENUBAR			\ 19 constant IN.SYS.WINDOW
 20 constant IN.LOWER.WINDOW	21 constant IN.DRAG.BOX
 22 constant IN.SIZE.BOX		23 constant HIGH.LEVEL.EVENT
 24 constant IN.ZOOM.IN			25 constant IN.ZOOM.OUT
 26 constant IN.CLOSE.BOX
 29 constant IN.PROXY.ICON
 31 constant IN.NibControl


\	 ----------------------------------------------------

variable CurrentEvents		\ holds event mask for current @NEXTEVENT
	CurrentEvents off		\ used by syntheventers, etc.

AliasedExtern: _GetNextEvent UInt32 flag  GetNextEvent  ( UInt32 eventMask, UInt32 * &event );
\ : _GetNextEvent GetNextEvent ;
Extern: UInt32 flag  EventAvail    ( UInt32 eventMask, UInt32 * &event );
Extern: UInt32 flag  CheckEventQueueForUserCancel ( void );
\ Returns true if a user cancel event is in the queue, false otherwise. 
\ Should work with ESC and CMD-.

\ 10.46 - universal Carbon Event Queue word - can extract an eventref or simply check for presence
\ This checks BOTH classic events and Carbon events, unlike EventAvail which apparently only
\	checks for classic events
extern: UInt32 OSStatus ReceiveNextEvent ( UInt32 inNumTypes, UInt32 * EventTypeSpecList, double EventTimeout,
											UInt32 inPullEvent, UInt32 * EventRef );


: @EVENT  ( RECORD\MASK -- EVTCoDE )
	GetPort >R					\ GrafPtr on return stack
	over _GetNextEvent drop w@s
	R> <SetPort ;				\ 5.2 restore port in case a bad patcher changed it


: WaitNextEvent ( mask\^eventrecord\sleep\mousergn -- eventcode | )
	GetPort >R			\ GrafPtr on return stack
	2 pick >R			\ save address of event record
	_WaitNextEvent ( eventMask\&event\sleep\mouseRgn -- flag )
	drop R> w@s
	R> <SetPort ;		\ restore port in case a bad patcher changed it


: (?EVENT)  ( RECORD\MASK -- F | check event queue for event )
	over EventAvail IF w@ ELSE drop 0 THEN ;


defer ?EVENT  ( RECORD\MASK -- flag | Check for pending event )
	' (?EVENT) is ?EVENT
\ re-vectored in Carbon Events.fth to include our private event queue


variable TheEvRef    TheEvRef on

\ 10.46 - variant of (?EVENT) using ReceiveNextEvent
: (?AnyEvent)  (  -- flag | check for any pending os event )
	0		\ all events
	0		\ all events
	0.e		\ don't wait
	0		\ don't pull from queue
	TheEvRef
	ReceiveNextEvent drop
	TheEvRef @ 0<>
;

defer ?AnyEvent (  -- flag )  \ 10.46 check for ANY pending event
	' (?AnyEvent) is ?AnyEvent
\ also re-vectored in Carbon Events.fth to include our private event queue

\	 ----------------------------------------------------


: !ON.ACTIVATE ( token\wptr -- )
	-wptr  +on.activate !  ;

: !ON.UPDATE  ( token\wptr -- )
	-wptr  +on.update !  ;


: ON.ACTIVATE  ( wptr4 -- )
	'
	compiling
    IF		clit,		\ Literal, 
    		Postpone swap 
    		Postpone !on.activate
    ELSE  swap  !on.activate
    THEN  ;  immediate


: ON.UPDATE ( wptr4 -- )
	'
	compiling
    IF		clit,		\ Literal, 
    		Postpone swap
    		Postpone !on.update
    ELSE  swap !on.update
    THEN  ;  immediate


: (DISCARD.UPDATES)  ( wptr4  --  discards updates for window )
	-wptr 
	@  dup  BeginUpdate EndUpdate  ;


: DISCARD.UPDATES  (   --  discards updates for window )
   get.window  (discard.updates)  ;


: DISCARD.ACTIVATES  ( -- | remove all activate events )
	{  |  buf[ 32 ]  }
	BEGIN 
		buf[ 256 @event 0=
	UNTIL  ;

\	----------------------------------------------------

create OLD.WINDOW.BEHAVIOR true ,

: EVENT.LOOP ( -- )
	old.window.behavior @
	IF
		BEGIN
			do.events drop
		AGAIN
	ELSE
		do.events drop
	THEN ;


create EVENT.LOOP.TOKEN  0 ,


: MainEventLoop ( -- )
	get.window 
	locals| OldWindow |
	BEGIN
		active.window @ ?wvisible			\ 6.044
		IF
			 get.window -> OldWindow	\ save current wptr4
			 active.window @			\ 6.044
			 dup window					\ window = active wptr4
			 +window.action @ 
			 ?dup
			 IF 						\ do repetitive window action
				 execute				\ only if there is one
			 ELSE
			 	 do.events drop
			 THEN
			 OldWindow ?wvisible			\ 6.044
			 IF
			 	OldWindow (window) 	\ restore wptr4
			 ELSE
			 	active.window @				\ 6.044
			 	dup ?wvisible				\ 6.044
			 	IF   (window)
			 	ELSE drop
			 	THEN
			 THEN
        ELSE
        	 get.window active.window !			\ 6.044
        	 events on				\ no active wptr4 -- Handle all by default
			 do.events drop
		THEN
		Event.Loop.Token @execute       
	AGAIN ;


\ NULL EVENTS ----------------------------------------------------



: >NULL.EVENT ( -- )
	pause
	NullEventer ExecChain
;



\	MOUSE SPECIFICS ----------------------------------------------------

: MOUSE.WAS..  (  -- point | global mouse coordinates )
	mouse.down.record 
	12  + @  ;


: @MOUSE.DN  ( --  local  mouse pt at down )
	mouse.was.. Global>Local ;


: >MOUSE.DOWN  ( records mouse down event,  )
	mouse.up.record off  event.record
	mouse.down.record 2 + 16 cmove   ;


: >MOUSE.UP   ( records mouse up event  )
	mouse.down.record off  event.record
	mouse.up.record 2 + 16 cmove   ;





\ UPDATE EVENTS ----------------------------------------------------


: (UPDATE.WINDOW)  ( wptr4 -- | branches to window's update entry.point )
	dup  ?wvisible		\ ?wptr				\ is it MF window ??
	over ?wptr AND
	IF  @window.pair  >r >r
		outfile @ >R						\ 
		dup outfile !						\ 
		(mfwindow)							\ Select Window
		skip.events @ >R					\ 
		SKIP.EVENTS ON						\ no other events during update !!!
\	follower @ follower 4 + !	\ 6.0b10
\	STATUS follower !
		get.window @  BeginUpdate          	( Vis+update>Vis)
		get.window  +on.update @ execute  	( User token)
		get.window  show.controls         	( update controls)
		get.window @  EndUpdate            	( Restore clip,)
\	follower 4 + @ follower !
		R> SKIP.EVENTS !					\ 
		\ SKIP.EVENTS OFF					\ 
		R>	outfile !						\ 
		r> r> !window.pair					\ Restore wptr
	ELSE
		drop 
		pause		\  fix for LaserWriter foolishness
	THEN
;


: UPDATE.WINDOW  ( -- | branches to window's update entry.point )
	Event.Record 2 + @
	?dup
	IF
		_GetWRefCon
		dup	Event.Record 2 + !
		dup ?wptr
		IF
			(update.window)
		ELSE
			drop
		THEN
	THEN
 ;



\	Activate Events ----------------------------------------------------


: ?ACTIVATED.WINDOW ( wptr4 --- flag | true if window is active.window )
	dup IF active.window @ =  THEN ;	


: ?SYS.WINDOW ( -- flag | true if Active.Window = Sys.Window )
	sys.window ?activated.window ;


: HILITE.ACTIVE.WINDOW ( flag -- )		\ Actually hilites GET.WINDOW
	dup
	false= 255 and h&vbars				\ prob ok
	get.window  +w.hilite @execute ;	\ user hilite token - default = DROP


 create ACTIVATE.HOOK	0 ,			\ TESTING ONLY
 create DEACTIVATE.HOOK	0 ,			\ TESTING ONLY


: >DEACTIVATE.WINDOW   (  deactivates window )
	deactivate.hook @execute			\ TESTING ONLY
	event.record 2 + @ 					\ 
	active.window @ =
	IF
		Active.Window OFF				\ 
	THEN ;


: >ACTIVATE.WINDOW   					( activates window, executes )
	get.window  Active.Window !				\ 
				\ this is called by (ACTIVATE.WINDOW) which can be invoked
				\ by EDITOR or by DfltSELECT.WINDOW
				\ so we probably should not be accessing EVENT.RECORD
	activate.hook @execute					\ TESTING ONLY
;


VARIABLE TailGunner		\ last chance to get control...
  TailGunner OFF			\ initially no action

create SWITCHED.OUT?	false ,	
create SWITCHED.WINDOW	0 ,	


: WAKE.WINDOW ( flag -- )
	events @ >r
	events off
		dup
		IF
			>activate.window
			 hilite.active.window
		ELSE
			 hilite.active.window
			 >deactivate.window
		THEN
	r> events ! ;



\ On OSX, collapsing a window into toolbar and re-expanding it
\    causes the new window ACTIVATE before the last window DEACTIVATE
\ so we clear ACTIVE.WINDOW in >DEACTIVATE.WINDOW only if 
\		ACTIVE.WINDOW === window in event record

: (ACTIVATE.WINDOW) ( flag -- )
	\ GET.WINDOW will return the correct window
	DUP								\ 
	old.window.behavior @			\ ?sys.window or
	AND								\ both old.window.behavior AND an Activate 
	IF	
		r0 @ rp! 
		>R 							\ preserve flag
		s0 @ sp!
		R> 
	 	exc0 @ excp !				\ 10.46
	THEN
((
	OS X - 
		minimize window to dock,
		switch out of macforth,
		click document icon in dock,
		we get a window activate BEFORE we get SWITCH EVENT
	So, we have to adjust switched.window here, else
		we get ACTIVE.WINDOW <> CURRENTWINDOW
))
	dup IF
		SWITCHED.OUT? @ 
		IF ( SwitchIn ) get.window switched.window !  THEN
	THEN

	dup wake.window
	old.window.behavior @
	IF
		get.window 	?DUP
		IF
			+on.activate  @execute 
\			+on.activate  @  CATCH ?dup IF s" ACTIVATE CATCH" s.type   .throw event.loop THEN 
		ELSE
			drop							\ drop flag !!
		THEN
		TailGunner @Execute					\ hook for last window closed
											\ also when falls through Activate Loop
		get.window Active.Window @ =
		Active.Window @ 0=
		OR
		IF
			event.loop
		ELSE
			Active.Window @ ?wptr
			IF
				Active.Window @ window
			THEN
		THEN
	ELSE
		get.window  ?dup
		IF
			+on.newactivate	@execute
		THEN
	THEN 
;



: ACTIVATE.WINDOW  ( branches to window's activate/deactivate)
                   ( entry points w/ activate/deactivate flag )
\ enters with event.record 2 + having windowref
\ changes it to wptr4
	@window.pair 	locals| OldXForm OldWindow |
	event.record 2 + @								\ windowref
	dup GetWindowKind 0>
	IF
		_GetWRefCon ?DUP
		IF
			dup Event.Record 2 + ! 					\ change event msg to 4th window
			dup ?wptr								\ is it a m4th window?
			IF   
				(mfwindow)							\ make it TheWindow
				event.record  14 + w@   1 and		\ true if Activate event
				(activate.window)					\ does all the work...
			ELSE
				drop
			THEN
		THEN
	ELSE										\ it's a DA
				\ this code should be obsolete !!!!!
		DROP
		event.record  14 + w@   1 and			\ true if Activate event
		IF	 activate.hook						\ activating a DA
		ELSE deactivate.hook					\ deactivating a DA
		THEN
		@execute								\ TESTING ONLY
		active.window off
		OLD.WINDOW.BEHAVIOR @
		IF 
			TAILGUNNER @EXECUTE
			EVENT.LOOP
		THEN
	THEN
	OldWindow OldXForm  !window.pair
;


: >CLOSE.WINDOW ( event  -- event | hide window if mouse down in close box ) 
	get.window
	event.record 2 + @
	dup front.window =
	locals| isFront? EventWptr wasWindow |
	isFront? false=
	IF
		EventWptr SELECT.WINDOW		\ 
		EventWptr +w.attributes w@	\ 
		256 AND false=				\ do false= allow CLOSE.BEHIND ?
		IF
			EventWptr bring.to.front
			EventWptr SELECT.WINDOW
			drop  0					\ clear pending CLOSE.BOX event
			exit
		THEN
	THEN

	EventWptr window			\ ???

	EventWptr @  mouse.was..  TrackGoAway ( track.go.away )
	IF    
		EventWptr +on.close @ ?dup
		IF				\ token: ( event -- event' )
			execute
			\ the +ON.CLOSE token can clear the IN.CLOSE.BOX event
			\	for example, an unsaved editor window
		THEN

		dup in.close.box =
		IF
			EventWptr  hide.window
			EventWptr +dlogref# @ 		\  handle MODELESS dialogs
			?dup 
			IF
				DisposeDialog
				EventWptr +dlogref#   off
				EventWptr +windowref# off
			THEN
		THEN
	ELSE
		  drop  0		\ clear pending IN.CLOSE.BOX event for user
	THEN
	isFront? false= IF drop  0 THEN	\ if false= front window, don't let
									\ the IN.CLOSE.BOX event propogate
	wasWindow ?wvisible IF wasWindow window THEN
;


: >SELECT.WINDOW ( -- )
	event.record 2 + @  @ SelectWindow ;


: >GROW.WINDOW ( --  resizes window  scroll/content region) 
	event.record 2 + @

	dup front.window <>		\ 
	IF
		dup SELECT.WINDOW	\ 
	THEN
	window
	get.window size.window	  				\ computes +wcbounds
	get.window 					\ wptr
	dup  @ 						\ wptr\WindowRef
	dup  rot					\ 
	mouse.was.. 				\ WindowRef\WindowRef\wptr\mousept
	swap +wgrow    				\ WindowRef\WindowRef\mousept\^growData
    dup 4 +     				\ WindowRef\WindowRef\mousept\^growData\^TL
    dup  @			   			\ WindowRef\WindowRef\mousept\^growData\>limits\oldlimit
	locals|  oldlimit >limits growrect |
	oldlimit 0<		 			\ if upper limit is -1, compute end of screen
	IF	GrayBounds 4 + @point	\ H\W
		get.window 				\ 
			GetWindowPortBounds 	\ @ ScratchBoundsRect _GetWindowPortBounds
			@ Local>Global
			point>xy 			\ L\T
		-point   		\ 8 8  -point   		\ adjust for titlebar
		>limits !point
	THEN

	growrect GrowWindow ( WindowRef\startPt\&sizeRect -- newHeightWidth )	\ now grow screen
	oldlimit >limits  !						\ restore prior contents
	point>xy   SWAP		\ endian
	-1 SizeWindow ( >size.window ( WindowRef\wide\high\updateFlag -- )   ( track mouse )
	get.window clean.controls
	get.window size.window
	get.window clean.controls
	@pen get.window +wcbounds 4 + w@ 3 - min MoveTo
;


: >ZOOM.BOX  ( [7] or [8] -- | zoom box events )
	event.record 2 + @
	dup front.window <>		\ 6.04998c
	IF
		dup SELECT.WINDOW	\ 6.04998c
	THEN
	
	window				\ make it the window
	locals| thePart | 
	get.window @  mouse.was.. thePart TrackBox
	IF	get.window size.window				\ ensure wcbounds
      	get.window clean.controls
		get.window @ ( CARBON ) thePart -1 ZoomWindow		\ always brings to front
        get.window size.window
      	get.window clean.controls
      	@pen get.window +wcbounds 4 + w@ 3 - min MoveTo

	ELSE						\ not released in zoom box !!
		 drop 0					\ switch event to null
	THEN  ;


: >ZOOM.IN  ( -- | IN.ZOOM.IN events )
	7 >zoom.box ;


: >ZOOM.OUT  ( -- | IN.ZOOM.OUT events )
   8 >zoom.box ;



\  SYSTEM EVENTS  ----------------------------------------------------


create MENU.SELECTED.TOKEN	0 ,	 ( -- )
create CMD.KEY.FILTER		0 ,  ( char -- char' )

defer >execute.menu ( item\menuid -- )
	' execute.menu is >execute.menu

: >MENUBAR ( -- )
	BeforeMenuSelect ExecChain
 	menu.selected.token @execute
	mouse.was..
	menu.select	 ( point -- item\menuid )
	>execute.menu	\ execute.menu
;


: >>MENU ( -- )		\ event 28 - posted via <CommandEventHandler
					\ handle Key-based menu events without CMD key
	event.record 
	execute.menu  ( item\menu id  -- )
;

: >CMD.KEY ( -- )
	#menus @
	IF
		BeforeMenuSelect ExecChain				\ 10.46
		menu.selected.token @execute
		event.record 2 + c@		\ ENDIAN Bytes 2-5 contain the event message
		cmd.key.filter @execute	 ( char -- char' )
		?dup
		IF
			event.record 5 + c!
			event.record menu.event
			?dup false=
			IF				\ use old method if no result
				drop
				event.record 2 + c@  menu.key		\ ENDIAN Bytes 2-5 contain the event message
			THEN
	\		BeforeMenuSelect ExecChain
			execute.menu 
		THEN
	THEN ;


\	KEYSTROKE EVENT MAINTENENCE --------------------------------------------

: COMMAND.KEY? ( -- flag | true if command key pressed )
	event.record 15 + C@  1 AND 			\ if cmd key down 	Bytes 14-15 contain the modifiers bits (kbd state)
	boolean  ;


: >KEY.STROKE   ( records keydown event  )
	\ " >KEY.STROKE" DebugStr
    event.record  key.stroke 2 + 16 cmove    ;


defer >AUTO.COMMAND.KEY
	' noop IS >AUTO.COMMAND.KEY


: >AUTO.KEY
	command.key?
	IF  
		>auto.command.key			\ forward reference for user
	ELSE
		key.stroke @ false=			\ key.stroke buffer empty
		IF >key.stroke  THEN
	THEN  ;


: >ABORT.EVENT  (  --  aborts current process )
     26 true swap r$error ;


: >DESKTOP		\ probably obsolete
	beep  ;


: >DRAG.WINDOW
	event.record 2 + @  @ 
	mouse.was..
	GrayBounds  DragWindow ;



\ MultiFinder events ----------------------------------------------------


create BEFORE.SUSPEND	0 ,
create BEFORE.RESUME	0 ,


: MuFi.Hilite ( resume flag -- | handle hiliting for switch out )
    switched.window @ ?wvisible
    IF  @window.pair >R >R 
	        switched.window @ (window)
	        (activate.window)               	 \ turns off active.window
		R> R> !window.pair
    ELSE
    	drop
    THEN ;


: >SWITCH  ( -- )  \ switcher event
    event.record 5 + c@					\ 10.45 another endian issue with event record
\    event.record 2 + c@
    1 =
    IF                                  \ SUSPEND/RESUME event
        event.record 2 + c@  1 and		\ 10.45 another endian issue with event 
\        event.record 5 + c@  1 and
        dup  false= switched.out? ! 
        dup  
        IF                              \ it's a RESUME event
            switched.window @ 
            dup ?ACTIVE.WPTR
            IF
            	active.window !   		\ restore active.window
            ELSE
            	drop
            THEN
            before.resume @execute
        	MuFi.Hilite
        ELSE                                    \ it's a SUSPEND event
            active.window @ switched.window !
            before.suspend @execute
        	MuFi.Hilite
            do.events drop              \ MultiFinder switches now
        THEN
    THEN  ;


: NO.AEHANDLER ( -- )	\ default action for HIGH.LEVEL.EVENT
	pause
	high.level.event.record off
;


: DoExecuteEvent ( -- )	\ execute token at +where in event.record
	event.record 10 + @  execute
;


\	EVENT DEFINITIONS ----------------------------------------------------


: BAD.EVENT# ( -- | notify if Event# >62 )
	true 30 r$abort ;



false Value CORE.AE.HANDLED?


\	EVENT TABLE ----------------------------------------------------

create EVENT.TABLE
	(  0 )	' >NULL.EVENT ,			\ NULL.EVENT
	(  1 )	' >MOUSE.DOWN ,			\ MOUSE.DOWN
	(  2 )	' >MOUSE.UP ,			\ MOUSE.UP
	(  3 )	' >KEY.STROKE ,			\ KEY.DOWN
	(  4 )	' pause ,				\ KEY.UP
	(  5 )	' >AUTO.KEY ,			\ AUTO.KEY
	(  6 )	' UPDATE.WINDOW ,		\ UPDATE.EVENT
	(  7 )	' pause ,				\ obsolete -- was DISK.EVENT
	(  8 )	' ACTIVATE.WINDOW ,		\ ACTIVATE.EVENT
	(  9 )	' >ABORT.EVENT ,		\ ABORT.EVENT
	( 10 )	' pause ,				\ obsolete -- was NETWORK.EVENT
	( 11 )	' pause ,				\ obsolete -- was DRVR.EVENT
	( 12 )	' DoExecuteEvent ,		\ PRIVATE EVENT - EXECUTE MESSAGE AS TOKEN
	( 13 )	' pause ,				\ not used
	( 14 )	' >NULL.EVENT ,			\ MOUSE.MOVED
	( 15 )	' >SWITCH ,				\ SWITCH.EVENT


\ CSI synthesized events
	( 16 )	' >CMD.KEY ,			\ COMMAND.KEY
	( 17 )	' >DESKTOP ,			\ IN.DESKTOP
	( 18 )	' >MENUBAR ,			\ IN.MENUBAR
	( 19 )	' pause ,				\ obsolete -- was IN.SYS.WINDOW
	( 20 )	' >SELECT.WINDOW ,		\ IN.LOWER.WINDOW
	( 21 )	' >DRAG.WINDOW ,		\ IN.DRAG.BOX
	( 22 )	' >GROW.WINDOW ,		\ IN.SIZE.BOX
	( 23 )	' NO.AEHANDLER ,		\ HIGH.LEVEL.EVENT --> becomes >HIGHLEVELEVENTS
	( 24 )	' >ZOOM.IN ,			\ IN.ZOOM.IN
	( 25 )	' >ZOOM.OUT ,			\ IN.ZOOM.OUT
	( 26 )	' >CLOSE.WINDOW ,		\ IN.CLOSE.BOX
	( 27 )	' pause ,				\ 
	( 28 )	' pause ,				\ Carbon Menu Event
	( 29 )	' pause ,				\ IN.PROXY.ICON   	NEW>DRAG.WINDOW
	( 30 )	' pause ,				\ 
	( 31 )	' pause ,				\ IN.NibControl

 
\ Custom Internal Events
\ NOT maskable with EVENTS or IO.EVENTS
	( 32 )	' pause ,
	( 33 )	' pause ,
	( 34 )	' pause ,
	( 35 )	' pause ,
	( 36 )	' pause ,
	( 37 )	' pause ,
	( 38 )	' pause ,
	( 39 )	' pause ,
	( 40 )	' pause ,
	( 41 )	' pause ,

	( 42 )	' pause ,					( 43 )	' pause , 
	( 44 )	' pause ,					( 45 )	' pause ,
	( 46 )	' pause ,					( 47 )	' pause ,
	( 48 )	' pause ,					( 49 )	' pause ,
	( 50 )	' pause ,					( 51 )	' pause ,
	( 52 )	' pause ,					( 53 )	' pause ,
	( 54 )	' pause ,					( 55 )	' pause , 
	( 56 )	' pause ,					( 57 )	' pause ,
	( 58 )	' pause ,					( 59 )	' pause ,
	( 60 )	' pause ,					( 61 )	' pause , 
	( 62 )	' pause ,					( 63 )	' BAD.EVENT# ,




\	EVENT REMAPPING ----------------------------------------------------


: WHICH.KEY.EVENT  ( 3 -- [3] or [16]  returns 16 on cmd key )
	command.key?
	65536 
	currentevents @	
	and boolean and					\ and cmd key events enabled
	IF	\  don't do cmd.key if char<bl or char=127 (del)
		\ this allows editor to pick cmd-arrow keys, etc.
\		event.record 5 + c@ dup bl >=
		event.record 2 + c@ dup bl >=			\ ENDIAN
		swap 127 - and IF 13 + THEN 
	THEN ;


: WHICH.SWITCH.EVENT ( 15 -- 15 [or] 14  )
	event.record 2 + c@ 250 ( hex# 0FA ) =	\ high byte = FA if mouse.moved event
	IF 1- THEN ;


((
 create THIS.CONTROL  0 ,   ( latest control handle )
 create THIS.PART     0 ,   ( latest control part )

     10 constant IN.BUTTON         11 constant IN.CHECKBOX
     20 constant UP.BUTTON         21 constant DOWN.BUTTON
     22 constant PAGE.UP           23 constant PAGE.DOWN
    129 constant IN.THUMB
))


: WHICH.MOUSE.EVENT ( 1 -- [1] OR [17-25]   choose mouse event)
	drop
	>mouse.down  mouse.was..  Find.Window ( point -- WindowRef\where code )
	\ consider using HIWindowFindAtLocation
	over _GetWRefCon

	locals| wptr4 WPART WindowRef |
	wptr4 event.record 2 + ! 	            ( save M4th window ptr )

	wpart 5 =								( in grow rgn )
	IF  wptr4 +w.attributes
		w@ size.box and
 		wptr4 +NewWAttributes
		@ 16 and						\ kWindowResizableAttribute
		OR false=						\ neither old or new style size box
		IF 3 to wpart THEN					( change to IN.CONTENT )
	THEN

\ dimmed controls return 0\0
	wpart 3 =								\ inContent
	wptr4 active.window @ = and  			( FRONT AND IN CONTENT)
	IF
		@mouse.dn wptr4 Find.Control ( point\wptr4 --  ControlRef\PartCode  )
	\	16 rshift		\ 10.45 - change FIND.CONTROL
		this.part ! 						( flag control )
		this.control !
		1 to wpart
	ELSE
		17 +to wpart			 	      	( else remap  17-25 )
	THEN
	wpart 23 = 
	IF in.close.box -> wpart THEN
					\ in.close.box changed since high.level.event = 23
	1 wpart lshift
	currentevents @	
	and			   					( enabled? )
	IF 
		wpart 
	ELSE
		1 
	THEN ;


HEX
FFFFFFC7 value -KEYBOARD				\ not (KEY.DOWN or KEY.UP or AUTO.KEY)
5C9DFFC7 value INACTIVE.WINDOW.EVENTS	\ 10.46
\ 0410FFC2 value INACTIVE.WINDOW.EVENTS	\ allow closing behind windows
00000200 value ABORT.MASK
\ 00001000 value MuFi.EVENTS		\ 10.46 obsolete
00800000 value HIGH.LEVEL.EVENTS			\ 10.46
\ 00000400 value HIGH.LEVEL.EVENTS
DECIMAL



\ ACTIVE.WINDOW		(variable)			has the last activated MF window
\ CurrentWindow		(User Variable)		has the last Wptr set by (WINDOW)

: EventsMask  ( eventMask1 -- eventMask2 )  \ add in other bits
	abort.mask or  							\ always enable keyboard aborts
	key.stroke @ IF  -keyboard  and  THEN	\ KEY WAITING
	high.level.event.record @ IF			\ buffer high level events
		high.level.events invert and  		\ like we do with key strokes
	THEN

	switched.out? @
	IF inactive.window.events  and THEN

	Active.Window @
	dup  CurrentWindow @ -						\ is current =/= active?
	swap ?wvisible and							\ AND is active visible?
	IF
		inactive.window.events and
	THEN
\	MuFi.Events OR 					\ 10.46 obsolete must always allow MultiFinder Events
;


\ Chain EVENTER  						\ MOVED  head of Event-filter chain

: RunEventers  ( eventCode1 -- eventCode2 )
	skip.events @ >R
	skip.events off
	eventer ExecChain
	R> skip.events !
;

: SynthEvents  ( eventCode1 -- eventCode2 )  \ synthesize if needed
	dup
	IF	dup
		CASE
			mouse.down		OF  which.mouse.event	ENDOF
			key.down   		OF  which.key.event		ENDOF
			switch.event	OF  which.switch.event	ENDOF
		ENDCASE
	ELSE								\ null event
		currentevents @  256 and
		event.record w@  8 = and	\ deactivate DA
		IF  8 +  THEN
	THEN  ;


: >Event  ( eventCode -- addr )  \ point to default event action token
	0 max
	63 min					\ all events > 63 vector to #63
	4 *  event.table +  ;


create    Use.WaitNextEvent	true ,
 0 Value  WNE.REGION
 1 Value  WNE.SleepTime
 1 Value  WNE.SleepTime.bkg		\ possibly longer sleep time while in background
create    CAN.SLEEP		true ,	\ one-shot to disable sleep time for DO.IO.EVENTS


defer GetNextPrivateEvent  ( mask --- eventcode )	\ 10.46
	' dropfalse is GetNextPrivateEvent
	\ normally reset to (GetNextPrivateEvent)


: (@NEXTEVENT) ( mask --- eventCode )
	dup CurrentEvents !
	dup GetNextPrivateEvent ?dup		\ 10.46
	IF
		nip								\ don't need mask
	ELSE
		Event.Record
		\ 10.46 always use WaitNextEvent
\		Use.WaitNextEvent @				\ use WNE if requested
\		IF
			switched.out? @ IF wne.sleepTime.bkg ELSE wne.sleepTime THEN
			can.sleep @ and				\ speeds up (DfltDO.IO.EVENTS)
			wne.region
			WaitNextEvent ( mask\^eventrecord\sleep\mousergn -- eventcode )
			can.sleep on				\ always restore one-shot
\		ELSE
\			swap  @Event  ( RECORD\MASK -- EVTCODE )
\		THEN
	THEN
	runeventers						\ EVENTER words even run on null events
	synthevents
;


(( 
: (@NEXTEVENT) ( mask --- eventCode )
	{ | buf[ 32 ] }
	dup  CurrentEvents !
	dup buf[ ?event  ( RECORD\MASK -- F | CHECK EVENT QUEUE FOR EVENT )
	IF
		Event.Record
		Use.WaitNextEvent @				\ use WNE if requested
		IF
			switched.out? @ IF wne.sleepTime.bkg ELSE wne.sleepTime THEN
			can.sleep @ and
			wne.region
			WaitNextEvent ( mask\^eventrecord\sleep\mousergn -- eventcode )
			can.sleep on				\ always restore one-shot
		ELSE
			swap ( _SystemTask ) @Event  ( RECORD\MASK -- EVTCODE )
		THEN
		runeventers						\ EVENTER words even run on null events
		synthevents
	ELSE
		drop							\ mask
		0
		runeventers
	THEN
;
 ))


defer @NEXTEVENT ( mask --- eventCode )	
	' (@NEXTEVENT)   IS  @NEXTEVENT


variable SpecialEventAction		SpecialEventAction off

: >>EVENTS ( eventCode --- eventCode )		\ 10.46 - was >EVENTS
	SpecialEventAction @ ?dup	\ for AEQuitApp on OS X
	IF
		SpecialEventAction off	\ make it a one-shot
		execute
	THEN
	event.filter @execute
	dup
	>event @execute			\ allow event handler to change the event
;


defer >EVENTS ( mask --- eventCode )			\ 10.46 - make deferred
	' >>EVENTS  is >EVENTS



: GetNextEvent ( --- eventCode | DOES NOT EXECUTE EVENT.TABLE )
\ Does remove event from event queue
\ Replaces use of EXTERN: _GetNextEvent
	skip.events @				\ 10.43 DANGEROUS FLAG !!!
	IF
		0
	ELSE
		events @ eventsmask
		@nextevent
	THEN ;


: >DO.EVENTS ( mask --- eventCode )
	skip.events @				\ DANGEROUS FLAG !!!
	IF
		drop 0
	ELSE
		@nextevent >events
	THEN
;


: (DfltDO.EVENTS)  ( -- eventCode )  \ executes via event table
	Events @ EventsMask
	>do.events  ;


: (DfltDO.IO.EVENTS) ( -- )
	can.sleep off					\ disable WNE.SLEEPTIME for next >DO.EVENTS
	io.events @ EventsMask
	>do.events drop ;



defer DfltDO.EVENTS
	' (DfltDO.EVENTS)	  is	 DfltDO.EVENTS

defer DfltDO.IO.EVENTS
	' (DfltDO.IO.EVENTS)	is   DfltDO.IO.EVENTS


\	##### FORWARD REFERENCE RESOLUTION #####

' DfltDO.EVENTS	    dflt.con 56 + !
' DfltDO.IO.EVENTS  dflt.con 60 + !



\ ************************************


: SYS.UPDATE ( -- )
	pause
	sys.window front.window =
	IF
		sys.window GetWindowPortBounds
		4 + @point								\ h\w
		sys.window +wline.height @  / c/cols !
		20 - ascii M  CharWidth 1+  / c/line !
	THEN
;
sys.window on.update SYS.UPDATE


: SYS.ACTIVATE ( flag -- )
	IF	 
		sys.update
		sys.window dup window
				   dup clip>content  
		 			   infile !
		outfile on
		init.cursor
  		QUIT
	THEN  ;
sys.window on.activate sys.activate


: NEW.SYS.ACTIVATE ( flag -- )
	IF	 
		sys.update
		sys.window dup clip>content  infile !
		init.cursor
		s0 @ sp! 
\		retry OFF
		r0 @ rp!
		quit
	ELSE
		s0 @ sp!
\		retry OFF
		r0 @ rp!
		maineventloop
	THEN  ;

' new.sys.activate  sys.window +on.newactivate !



\ **********************************************************************


(( **********************************************************************

Carbon MacForth is an event-driven program, calling DO.EVENTS repetetively in
almost all windows to allow speedy user interaction and frequent invocations of 
PAUSE.

While OSX is intrinsically multitasking, a single greedy process can grab a 
significant chunk of CPU time running in a tight loop.  On a Core Duo Mac, this
could mean one entire CPU could be stuck running the one process.

WaitNextEvent is the OS call at the heart of DO.EVENTS and one of its
arguments is a "sleep time" - the number of Mac ticks (1/60 sec) yielded back to
the OS to prevent such "CPU hogging".


If you execute
0 -> WNE.SLEEPTIME
and watch Activity Monitor while idling in an Eitor window or SYS.WINDOW, you will
see its CPU Usage nearly at 100%.  Timing DO.EVENTS with a 0 sleeptime on a
3.06 GHz Core 2 Duo iMac reveals it to execute in about 17 microseconds. 

So, if we set 
1 -> WNE.SLEEPTIME
to give some time back to the OS (1 tick per WNE), Activity Monitor will show our
CPU usage while idling is <1% - MUCH better.  Our penalty is that DO.EVENTS now
executes in 16.8 MILLIseconds (1 tick) - MUCH slower. 

In OS X on an Intel Mac, you should not need a sleeptime >1, and as showm, a
sleeptime of 0 may be excessively greedy of CPU time.

In an interactive window like Editor or SYS.WINDOW, a 1 tick delay is
imperceptible.

In MF high level IO words, like TYPE where DO.IO.EVENTS is called often, the issue
is not a problem, since DO.IO.EVENTS uses a 0 sleeptime.

But, if we want to execute DO.EVENTS to check for user activity while we are
running a time critical task, like drawing a graph or doing a page count on
a long document we are preparing to print, that 1 tick/DO.EVENTS can really slow
us down.

There are some simple tricks to improve performance, always keeping in mind that
we do not want to hog the CPU unnecessarily.

PaperEvents shows how PrintFile does it.  Instead of calling DO.EVENTS
repeatedly, we call ?EVENT first to see if an interesting event is pending and
only execute DO.EVENTS if there is.  ?EVENT executes in a few microseconds,
with no sleep time, so we can do it at the end of every line we image with
little performance penalty. 

CheckEventQueueForUserCancel is another fast Toolbox call that will return true
if a user cancel event is in the event queue, like an ESC or CMD-.  This can
allow you to eliminate DO.EVENTS in time critical loops, with the penalty that
ONLY cancel events can get processed.


Another method that can be useful is to use FAST.DO.EVENTS (below) when you
need to poll for any user event fairly often with less time penalty than
DO.EVENTS. 

You can modify the EventSpeedup parameter to balance execution time with CPU
load.

Here are some actual metrics of FAST.DO.EVENTS made on the iMac described above

standard DO.EVENTS
WNE.SLEEPTIME = 0
17 uSec   99+% CPU time

WNE.SLEEPTIME = 1
16.7 msec  <1% CPU time



WNE.SLEEPTIME = 1

EventSpeedup	Execution Time	   CPU Load
----------------------------------------------
 1,000,000			 12 uSec		 99+%  (almost never call _WaitNextEvent)
     1,000			 32 uSec		 49%
       500			 51 uSec		 35%
       250			 85 uSec		 23%
       100			187 uSec		 12%
        50			352 uSec		  7%
        10		   1.55 mSec		  2%
         1		   8.4  mSec		  1%
         0		  16.8  mSec		 <1%  (equivalent to standard DO.EVENTS)


If you should want to experiment with changing DO.EVENTS action globally instead
of on an as-needed basis , recall that DO.EVENTS is vectored through mfCONSOLE and
has the default action of DfltDO.EVENTS which is itself a DEFER'red word. 

So, to change it globally, you would need to execute the following
     ' FAST.DO.EVENTS  IS  DfltDO.EVENTS


************************************************************************ ))


variable EventSpeedup       10 EventSpeedup !	\ a "good" compromise (see above)
variable EventSpeedupTally   0 EventSpeedupTally !


: FAST.DO.EVENTS  ( -- eventCode )
	Event.Record -1	?event			\ check event queue for ANY non-null event
	EventSpeedup @ 
		EventSpeedupTally @  <=		\ execute WaitNextEvent occasionally
	or
	IF
		EventSpeedupTally off
		(dfltdo.events)
	ELSE
		1 EventSpeedupTally +!
		0 >events			\ this ensures NULL.EVENTER's and PAUSE
							\ get executed every time
	THEN ;



