\ EZ NAVIGATION
\ An extended interface to the Standard File package
\ original by Greg Guerin - heavily adapted to use Nav Services
\ Last Revision: 12 Feb 2012  08:24:57  ward@megawolf.com



\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


only Forth definitions
>M4TH

decimal


ascii TEXT CONSTANT "TEXT


((
lacking REPLY   : REPLY NRRecord ;	\ just an alias to inner kernel structure

NavReplyRecord constant ReplySize

: ReplyOK?		( -- OK? )			NavReplyOK? ;
\ : ReplyType	( -- fileType )		NRRecord 2 + @ ;
: ReplySpec		( -- &spec )		NavReplySpec ;
: ReplyVolume	( -- VolRefNum )	NavReplyVolume ;
: ReplyDir		( -- dirID )		NavReplyDir ;
: ReplyName		( -- namePtr )		NavReplyName ;


: ClearReply  ( -- )  reply ReplySize erase ;
))


\ Navigation Dialogs are moveable and at minimum should support Window Updates

defer DoNavEvents  ' 3drop is DoNavEvents

3 0 CALLBACK: aNavEventProcPtr

' DoNavEvents to-callback aNavEventProcPtr



: DoNavUpdateOnly ( NavEventCallbackMessage\callBackParms\&callBackUD -- )
	locals| &CallBackUD callBackParms NavEventCallbackMessage |
	NavEventCallbackMessage 0=		\ an event has occurred
	IF
		callBackParms 26 +		\ ^NavEventData
		@
		dup w@ 
		update.event =
		IF	 \ if it is an update event, copy to event record
			Event.Record 16 move
			Event.Record 2 + @
			?active.WindowRef
			IF
				Event.Record 2 + @
				_GetWRefCon (?wptr)
				IF
					GetPort >R			\ protect against nav window still active port
					update.window 
					R> <SetPort
				THEN
			THEN
		ELSE \ we don't need to handle anything other than Updates
			drop
		THEN
((
	ELSE	\ other messages can be sent back in NavEventCallbackMessage
			\ like notifying of when to add controls !!
\ http://developer.apple.com/documentation/Carbon/Reference/Navigation_Services_Ref/nav_serv_ref/constant_5.html#//apple_ref/c/econst/kNavCBCustomize
\ http://developer.apple.com/documentation/Carbon/Conceptual/Prog_Navigation_Services/index.html?http://developer.apple.com/documentation/Carbon/Conceptual/Prog_Navigation_Services/Concepts/Navigation_Services-22.html
\ http://developer.apple.com/documentation/Carbon/Reference/Navigation_Services_Ref/index.html?http://developer.apple.com/documentation/Carbon/Reference/Navigation_Services_Ref/nav_serv_ref/function_group_2.html
))
	THEN
;

' DoNavUpdateOnly is  DoNavEvents



: ((NavGetFile)) (( AEDesc *defaultLocation,
				NavReplyRecord *reply,
				NavDialogOptions *dialogOptions,
				NavEventUPP eventProc,
				NavPreviewUPP previewProc,
				NavObjectFilterUPP filterProc,
				NavTypeListHandle typeList,
				void *callBackUD -- OSErr ))
	GetPort dup >R
		    GetPortTextFont	\ get current font
		r@ GetPortTextSize		\ get current font size
		r@ GetPortTextMode		\ get current textmode
		r@ GetPortTextFace		\ get current textface (style)
	r>
	locals| wasPort wasFace wasMode wasSize wasFont |
	NavGetFile
	wasPort <SetPort
	wasFont TextFont
	wasSize _TextSize
	wasMode TextMode
	wasFace TextFace
;


create ^0	0 , 0 , 0 , 0 ,
DEFER BeforeAlert  ( -- )		' noop is BeforeAlert
DEFER BeforeDialog  ( -- )		' noop is BeforeDialog


: (NavGetFile)  (( fileFilter\numTypes\&typeList\&reply\...
 				...\ client$\title$\action$\cancel$\message\NavEventProcPtr -- OSErr ))
	10 needed
	BeforeDialog		\ perform any pre-dialog actions
	>R					\ NavEventProcPtr
	0 InitializeNDoptions ( client$\title$\action$\cancel$\message$\filename$ -- )
	NavOptionFlags NDoptions +NDdialogOptionFlags !

	>R					\ &reply
	swap

	dup 0< IF drop 0 THEN					\ if #types <0, all types
	MakeNavTypeListHandle ( &list\numtypes -- handle )
	R>					\ &reply
	R>					\ NavEventProcPtr
	0
	locals|  NavEventUPP  NavEventProcPtr theReply NavTypeListHandle fileFilter |

	theReply NavReplyRecord erase
	NavEventProcPtr  dup IF NewNavEventUPP THEN -> NavEventUPP

	NavDefaultLocation  					\ defaultLocation
	theReply			\ reply
	NDoptions			\ dialogOptions
	NavEventUPP			\ eventProc
	0  					\ previewProc
	0  					\ filterproc
	NavTypeListHandle	\ typeList
	^0  				\ *callBackUD
	((NavGetFile))  	\ use THIS instead of _NavGetFile
	?dup 0=
	IF  
		theReply +NRvalidRecord c@
		IF
			theReply +NRselection
			1
			typeFSS
			NRAEKeyword
			NRAEDesc
			AEGetNthDesc 0=	\ no error
			IF	\ should we check the actual NRAEKeyword == typeFSS ??
				\ copy fs spec
				NDFSSpec  FSSpecRecord erase
				NRAEDesc +AEDescHandle @@  NDFSSpec  FSSpecRecord  cmove
				0	\ return if no error
			ELSE
				-2		\ " _AEGetNthDesc error" $>DB
			THEN
		ELSE
			-1 \ invalid record
		THEN
		\ theReply _NavDisposeReply .s drop \ check for non-0 result
	THEN
	0 -> NavPrefKey				\ make this a one-shot
	0 -> NavDefaultLocation		\ make this a one-shot
	7 -> NavOptionFlags			\ make this a one-shot

	theReply +NRvalidRecord c@	\ 605431 do this instead
	IF
		\ DONT do _NavCompleteSave
		theReply NavDisposeReply   drop \ check for non-0 result
		-1 theReply +NRvalidRecord c!
	THEN

	NavTypeListHandle	?dup IF to.heap THEN
	NavEventUPP			?dup IF DisposeNavEventUPP THEN
;


: NavGetFile ( fileFilter\numTypes\&typeList\&reply\...
 				...\ client$\title$\action$\cancel$\message -- OSErr )
	aNavEventProcPtr (NavGetFile)
;


: NavPutFile ( callbackData\Type\Creator\&reply\...
 				...\ client$\title$\action$\cancel$\message\filename$ -- OSErr )
	aNavEventProcPtr (NavPutFile)
;



\ similar to old StandardPutFile
: SPF  ( &prompt\&defaultName\&reply\NavEventProcPtr -- OSerr )
	locals| theNavEventProc theReply theFilename thePrompt |
	0   filetype @   Creator @ theReply   0 0 0 0  thePrompt  theFilename
	theNavEventProc
	(NavPutFile)  (( callbackData\Type\Creator\&reply\...
 				...\ client$\title$\action$\cancel$\message\filename$\NavEventProcPtr -- OSErr ))
;


: StandardPutFile ( &prompt\&defaultName\&reply --  )
	BeforeDialog		\ perform any pre-dialog actions
	3 needed  
	aNavEventProcPtr
	SPF drop ;



: StandardGetFile ( fileFilter\numTypes\&typeList\&reply -- )
\ if #types = -1, all types
	BeforeDialog		\ perform any pre-dialog actions
	ClearReply
	4 needed
	locals|  reply typeList numTypes filter |
	filter numTypes typeList reply
	0 0 0 0 0 
	aNavEventProcPtr
	(NavGetFile)  ( fileFilter\numTypes\&typeList\&reply\...
\ 				...\ client$\title$\action$\cancel$\message\NavEventProcPtr -- OSErr )
	drop
;


: AnyGetFile  ( prompt$ -- )
\ simple version with a prompt string to select any file
	>R
	0 		\ filter
	-1	0	\ any file type
	NRRecord 
	0 0 0 0
	R> 
	NavGetFile drop
	discard.activates
;	


(( 
: PromptedGetFile ( #types\typeList\Title$\Select$\Cancel$\Message$ -- )
	BeforeDialog		\ perform any pre-dialog actions
	ClearReply
	6 needed
	0 6 bury			\ no file filter
	NRRecord 4 bury		\ &reply
	0 4 bury			\ no client$
	aNavEventProcPtr
	(NavGetFile)  ( fileFilter\numTypes\&typeList\&reply\...
\ 				...\ client$\title$\action$\cancel$\message\NavEventProcPtr -- OSErr )
	drop
	discard.activates
;
 ))


\ File "Easy SFP"
\ Simple extensions to Standard File Package functions
\	by Greg Guerin.  Use freely, but give credit when distributing source.


: EZGet  ( typeList\#types -- )  \ select file to open
\ note,  0 -1 EZGET  will allow you to select ANY file type
	0 -rot swap NRRecord StandardGetFile
	Discard.Activates  ;	\ results are in NRRecord


: EZPut2  ( prompt\origName -- )  \ select file to create
	ClearReply
	NRRecord StandardPutFile
	Discard.Activates  ;	\ results are in NRRecord


: EZPut  ( prompt -- ) 
	c" Untitled" EZPut2  ;	\ origName is "Untitled"



: UseReplyName  ( -- namePtr )  \ set default vol & dir, return namePtr
	NavReplyVolume NavReplyDir DirSet  NavReplyName  ;


: UseReplyNameOnce  ( -- namePtr )  \ set VolRef# & iodir, return namePtr
	NavReplyVolume volref# ! NavReplyDir iodir ! NavReplyName ;





: PosixOpenFromFSSpec ( Spec\fam -- fd\ior )
\ use with NDFSSpec after one of the above Nav Services GetFile routines
\ to get a Posix FD instead of an HFS fileID
	>R
	FSSpec->ZString ( ^FSSpec -- zStr )
	zcount R> open-file
;	



: EZOpenIt  ( -- fileID )  \ return 0 if fail or "Cancel" button
	IO-Result OFF   0			\ assume "Cancel" button
	navreplyok? 
	if
		drop					\ discard the 0 we assumed
		UseReplyName OPEN$		\ open the file
	then  ;


: EZ1
	0 -1 EZGET
	SetVolDirToNav NavReplyName OPEN$ ;



StopIncluding ************************************

: EZOpenIt  ( -- fileID )  \ return 0 if fail or "Cancel" button
	IO-Result OFF   0			\ assume "Cancel" button
	ReplyOK? if
		drop					\ discard the 0 we assumed
		UseReplyName OPEN$		\ open the file
	then  ;


: EZCreateKind  ( kind -- fileID )  \ create according to NRRecord info
	FileType @  swap FileType !
	IO-Result OFF   0			\ assume "Cancel" button or error
	ReplyOK? if
		drop						\ if "Save", discard flag
		0 UseReplyName NEW.FILE		\ try to create the file  (S: fileID)
		IO-Result @  -48 = if		\ already exists? delete & try again
			UseReplyName DELETE$		\ remove the existing file
			IO-Result @ 0= if			\ if removed OK, try recreating it
				drop					\ discard 0 from first NEW.FILE
				0 UseReplyName NEW.FILE	\ failing twice returns w/ error
			then
		then
	then
	\ S: fileID  (which might be zero)
	swap FileType !  ;	\ restore default kind, return fileID


: EZCreateIt  ( -- fileID )  FileType @  EZCreateKind  ;

\ For EZOpenIt & EZCreateKind:
\	if fileID != 0  and  IO-Result == 0,  then SUCCESS
\	if fileID == 0  and  IO-Result == 0,  then "Cancel"
\	if fileID == 0  and  IO-Result != 0,  failed for reason in IO-Result


: EZOpenKind  ( kind -- fileID )  \ select file to open
	\ if kind = 0, accept any file-type, else only one type.
	dup>R rp@						\ addr of KIND in stack
	SWAP	0= 2* 1+				\ #types = -1 or +1
	EZGet  R>drop  EZOpenIt  ;	\ run dialog, discard kind, then open file

: EZOpen  ( -- fileID )  FileType @  EZOpenKind  ;

: EZNewFile  ( -- fileID )  c" Save as:" EZPut  EZCreateIt  ;



StopIncluding ************************************
