\ BASIC CARBON EVENTS
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com
\ NOT CLEAN FOR FORGETTING 


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


(( 

Carbon Events require either a WaitNextEvent loop or a StandardEventHandler
to invoke their handlers - they are NOT run at interrupt level !!


Carbon Events invoke EventHandler's which we create with
  CALLBACK:
  
Callbacks are executed directly by the MacOS so their "environment" is not
the same as normally executed Forth words.  In specific, USER variables
are not set for MacForth calls and they are heavily used in our high end
Text and Graphics words.

Before attempting any of our high level text or graphids words from a
Callback, be sure to explictly set (at a minimum)
	WINDOW and OUTFILE
Also, because you do NOT want to inadvertently invoke MacForth Event Handlers
from inside a Callback, set
  CALLBACK.CONSOLE MFCONSOLE !
in your event handler
  
CALLBACK.CONSOLE has DO.EVENTS and DO.IO,EVENTS stubbed out.


In some cases, like out mousewheel handler and Nib activate & update
routines, it is easier to have our EventHandlers post events in a queue that
we can poll with out traditional DO.EVENTS words. 

SendExecuteEvent ( XT -- )  is a handy word for posting a pseudo-event
for DO.EVENTS to execute later

))



only Forth definitions
>M4TH
decimal


(* *************************************************************************
Creating and Manipulating Event Handlers
	InstallEventHandler	Installs an event handler on a specified event target. 
	InstallStandardEventHandler	Installs the standard event handler for the specified target
	RemoveEventHandler	Removes the specified event handler.
	AddEventTypesToHandler	Adds events to an installed handler.
	RemoveEventTypesFromHandler	Removes events from an installed event handler.
	CallNextEventHandler	Calls the next handler in the handler chain.
************************************************************************* *)
(*
Carbon Events are specified as EventTypeSpec's - 8 byte constructs of:
	long: eventClass
	long: eventKind
Or as lists of EventTypeSpec's
*)

\ variable CarbonEventLogging		CarbonEventLogging off


\ Event handlers return this to allow subsequent handlers to operate
-9874 constant EventNotHandledErr
\ returning 0 means your handler has finished with the event and
\ the OS standard hadnlers will be skipped


\ Each class of UI item can have many eventKind's
\ EventClasses include:

ascii mous constant	kEventClassMouse
ascii keyb constant	kEventClassKeyboard
ascii text constant	kEventClassTextInput
ascii appl constant	kEventClassApplication
ascii eppc constant	kEventClassAppleEvent
ascii menu constant	kEventClassMenu
ascii wind constant	kEventClassWindow
ascii cntl constant	kEventClassControl
ascii cmds constant	kEventClassCommand
ascii tblt constant	kEventClassTablet
ascii vol_ constant	kEventClassVolume
ascii appm constant	kEventClassAppearance
ascii serv constant	kEventClassService
ascii tbar constant	kEventClassToolbar
ascii tbit constant	kEventClassToolbarItem
ascii acce constant	kEventClassAccessibility


Extern: UInt32 OSStatus InstallEventHandler ( UInt32 EventTargetRef, UInt32 EventHandlerUPP,
					UInt32 NumTypes, UInt32 ^TypeList, UInt32 UserData, UInt32 * ^EventHandlerRef );

Extern: UInt32 OSStatus InstallStandardEventHandler ( UInt32 EventTargetRef  );

Extern: UInt32 OSStatus RemoveEventHandler ( UInt32 EventHandlerRef );

Extern: UInt32 OSStatus AddEventTypesToHandler ( UInt32 EventHandlerRef, UInt32 #types, UInt32 * ^typeList );

Extern: UInt32 OSStatus RemoveEventTypesFromHandler ( UInt32 EventHandlerRef, UInt32 numTypes, UInt32 * ^typeList );

Extern: UInt32 OSStatus CallNextEventHandler ( UInt32 EventHandlerCallRef, UInt32 EventRef );


(* *************************************************************************
Creating and Manipulating Event Timers
	InstallEventLoopTimer  Installs a timer.
	InstallEventLoopIdleTimer  Installs a timer that fires only when there is no user activity.
	RemoveEventLoopTimer  Removes the specified timer.
	SetEventLoopTimerNextFireTime  Sets the next time that the specified timer will fire.
************************************************************************* *)

\ Note: TimerIntervals are floating point values 
\ corrections from George Kozlowski

Extern: UInt32 OSStatus InstallEventLoopTimer ( UInt32 EventLoopRef, double FireDelay, double TimerInterval,
			UInt32 EventLoopTimerUPP, UInt32 * ^inTimerData, UInt32 * ^EventLoopTimerRef );

Extern: UInt32 OSStatus InstallEventLoopIdleTimer ( UInt32 EventLoopRef, double FireDelay, double TimerInterval,
			EventLoopTimerUPP, UInt32 ^inTimerData, UInt32 ^EventLoopTimerRef );

Extern: UInt32 OSStatus RemoveEventLoopTimer ( UInt32 EventLoopTimerRef );

Extern: UInt32 OSStatus SetEventLoopTimerNextFireTime ( UInt32 EventLoopTimerRef, double TimerInterval );


(* *************************************************************************
Creating and Manipulating Events
	GetEventClass	Returns the class of an event (window, mouse, keyboard, and so on).
	GetEventKind	Returns the event kind for a given event.
	GetEventParameter	Obtains a parameter from the specified event. 
 	SetEventParameter	Sets a parameter associated with a particular event. 
 	CreateEvent	Creates an event.
 	CopyEvent	Copies an event.
 	RetainEvent	Increments the reference count of an event.
	ReleaseEvent	Releases, and possibly disposes of, the specified event.
	GetEventRetainCount	Returns the reference count of an event.
	ConvertEventRefToEventRecord	Converts an event reference into an event record.
	IsEventInMask	Returns whether an event reference matches a WaitNextEvent-style event mask.
************************************************************************* *)

Extern: UInt32 EventClass GetEventClass ( UInt32 EventRef );

Extern: UInt32 EventKind GetEventKind ( UInt32 EventRef );

Extern: UInt32 OSStatus GetEventParameter ( UInt32 EventRef, UInt32 EventParamName, UInt32 EventParamType, 
					UInt32 * ^ActualEventType, UInt32 MaxBufferSize, UInt32 * ^ActualSize, UInt32 * ^outData );

\ special case one of the commonest variants of GetEventParameter
: GetLongEventParameter ( ParamName\EventRef -- value )
	0 -rot	 					\ room for result..
	sp@ 8 + >R
	swap
	ascii long			\ typeLongInteger
	0
	4
	0
	R>
	GetEventParameter
	IF				\ in case of error, be sure we return 0
		DropFalse
	THEN
;

\ Another very common use of GetEventParameter
: GetControlPart ( EventRef -- partcode )	\ see GetLongEventParameter
	ascii cprt 					\ kEventParamControlPart
	dup							\ kEventParamControlRef
	0
	4
	0
	0 >R
	RP@
	GetEventParameter ( EventRef\EventParamName\EventParamType\..
					..^ActualEventType\MaxBufferSize\^ActualSize\^outData -- OSStatus )
	drop
	R> 16 rshift
;


\ You can get the Modifier key statues from most Carbon Events
\ see 
: GetEventModifiers ( EventRef -- modifiers )
		ascii kmod  swap GetLongEventParameter
		\ kEventParamKeyModifiers       = 'kmod'
;

((
  btnStateBit                   = 7,    /* state of button?*/
  cmdKeyBit                     = 8,    /* command key down?*/
  shiftKeyBit                   = 9,    /* shift key down?*/
  alphaLockBit                  = 10,   /* alpha lock down?*/
  optionKeyBit                  = 11,   /* option key down?*/
  controlKeyBit                 = 12,   /* control key down?*/

  btnState                      = 1 << btnStateBit,
  cmdKey                        = 1 << cmdKeyBit,
  shiftKey                      = 1 << shiftKeyBit,
  alphaLock                     = 1 << alphaLockBit,
  optionKey                     = 1 << optionKeyBit,
  controlKey                    = 1 << controlKeyBit,
))



\ ***************************************************************************
\ You can get the Modifier key statues from most Carbon Events

: GetEventModifiers ( EventRef -- modifiers )	\ added 10.47
		ascii kmod  swap GetLongEventParameter
		\ kEventParamKeyModifiers       = 'kmod'
;

Extern: UInt32 OSStatus SetEventParameter ( UInt32 EventRef, UInt32 EventParamName, UInt32 EventParamType, UInt32 Size, UInt32 * ^Data );

Extern: UInt32 OSStatus CreateEvent ( UInt32 CFAllocatorRef, UInt32 inClassID, UInt32 kind, LongLong EventTime,
										UInt32 EventAttributes, UInt32 * ^EventRef );

Extern: UInt32 EventRef' CopyEvent ( UInt32 EventRef );

Extern: UInt32 EventRef' RetainEvent ( UInt32 EventRef );

Extern: void ReleaseEvent ( UInt32 EventRef );

Extern: UInt32 count GetEventRetainCount ( UInt32 EventRef );

Extern: UInt32 flag ConvertEventRefToEventRecord ( UInt32 EventRefIn, UInt32 ^EventRecordOut );
	\ returns a Boolean value indicating whether the conversion was successful or not

Extern: UInt32 flag IsEventInMask	( UInt32 EventRef, UInt32 EventMask );



(* *************************************************************************
Dispatching Events
	SendEventToEventTarget	Sends an event to the specified event target. 
 	SendEventToEventTargetWithOptions	Sends an event to the specified event target with propagation options. 
	GetControlEventTarget	Obtains the event target reference for the specified control.
	GetWindowEventTarget	Obtains the event target reference for a specified window. 
	GetMenuEventTarget	Obtains an event target reference for the specified menu. 
	GetApplicationEventTarget	Obtains the event target reference for the application
	GetEventDispatcherTarget	Obtains the event target reference for the standard toolbox dispatcher.
	GetUserFocusEventTarget	Obtains the event target reference for the user focus.
	ProcessHICommand	Sends a command to the command chain.
************************************************************************* *)

Extern: UInt32 OSStatus SendEventToEventTarget ( UInt32 EventRef, UInt32 EventTargetRef );
Extern: UInt32 OSStatus SendEventToEventTargetWithOptions ( UInt32 EventRef, UInt32 EventTargetRef, UInt32 OptionBits );

Extern: UInt32 EventTargetRef GetControlEventTarget ( UInt32 ControlRef );
Extern: UInt32 EventTargetRef GetWindowEventTarget ( UInt32 WindowRef );

Extern: UInt32 EventTargetRef GetMenuEventTarget ( UInt32 MenuRef );
Extern: UInt32 EventTargetRef GetApplicationEventTarget ( void );

Extern: UInt32 EventTargetRef GetEventDispatcherTarget ( void );
Extern: UInt32 EventTargetRef GetUserFocusEventTarget ( void );

Extern: UInt32 OSStatus ProcessHICommand ( UInt32 ^HICommand );



(* *************************************************************************
Implementing Modal Windows
	RunAppModalLoopForWindow  Puts the window in an application-modal state.
	QuitAppModalLoopForWindow Quits the application-modal state for a window.
	BeginAppModalStateForWindow Puts the window in an application-modal state, but does not process events.
	EndAppModalStateForWindow Ends the application-modal state entered using the function BeginAppModalStateForWindow.
	SetWindowDefaultButton Specifies a default button for a window.
	GetWindowDefaultButton Returns the current default button for a window.
	SetWindowCancelButton Specifies a cancel button for a window.
	GetWindowCancelButton Returns the current cancel button for a window.
************************************************************************* *)


(* *************************************************************************
Managing Event Queues
	GetCurrentEventQueue Obtains the current event queue.
	GetMainEventQueue  Obtains the main event queue.
	PostEventToQueue  Adds an event to the specified event queue.
	RemoveEventFromQueue  Removes an event from the event queue.
	IsEventInQueue  Determines whether an event is in a particular queue.
	FlushEventsMatchingListFromQueue Removes events from the event queue by kind and class.
	FindSpecificEventInQueue Finds a specific event in the event queue.
	FlushSpecificEventsFromQueue Removes specified events from the event queue.
	FlushEventQueue Removes all events from the event queue.
	GetNumEventsInQueue Returns the number of events in the event queue.

************************************************************************* *)


Extern: UInt32 EventQueueRef  GetMainEventQueue ( void );
Extern: UInt32 EventQueueRef  GetCurrentEventQueue ( void );
Extern: UInt32 OSStatus  PostEventToQueue ( UInt32 inQueue,  UInt32 EventRef, UInt32 EventTargetRef );



\ specialized functions, not suitable for a basic suite of routines


(* *************************************************************************
Managing the Event Loop
	RunApplicationEventLoop  Runs the application event loop.
	QuitApplicationEventLoop Terminates the application event loop.
	GetMainEventLoop  Obtains a reference to the main event loop.
	GetCurrentEventLoop  Obtains a reference to the current event loop.
	GetCFRunLoopFromEventLoop  Obtains a Core Foundation CFRunLoop from an Carbon event loop reference.
************************************************************************* *)

Extern: void RunApplicationEventLoop ( void );
Extern: void QuitApplicationEventLoop ( void );
Extern: UInt32 EventLoopRef GetMainEventLoop ( void );
Extern: UInt32 EventLoopRef GetCurrentEventLoop ( void );
Extern: UInt32 EventLoopRef GetCFRunLoopFromEventLoop  ( UInt32 CFTypeRef );


(* *************************************************************************
Manipulating Event Time
	GetCurrentEventTime  Returns the current time since last system startup, in seconds.
	GetEventTime  Returns the time a specific event occurred.
	SetEventTime  Sets the event time for a given event.
	GetLastUserEventTime  Returns the last time a user input event arrived in the main event queue of the application.
************************************************************************* *)

Extern: double EventTime GetCurrentEventTime ( void );
Extern: double EventTime GetEventTime ( UInt32 EventRef );


(* *************************************************************************
Tracking the Mouse
	TrackMouseLocation  Tracks the mouse, blocking your application when there is no activity.
	TrackMouseLocationWithOptions Tracks the mouse with additional options.
	TrackMouseRegion  Tracks the mouse within a region.
	CreateMouseTrackingRegion  Creates a mouse tracking region.
	RetainMouseTrackingRegion  Retains a mouse tracking region.
	ReleaseMouseTrackingRegion  Releases a mouse tracking region.
************************************************************************* *)

\ specialized mouse tracking functions, more suitable for multithreaded code



(* *************************************************************************
Callback-Related Functions
	NewEventHandlerUPP	Creates an event handler UPP.
	DisposeEventHandlerUPP	Disposes of an event handler UPP. 
	NewEventLoopTimerUPP	Creates an event loop timer UPP. 
	DisposeEventLoopTimerUPP	Disposes of an event loop timer. 
************************************************************************* *)
(* 
	An EventHandlerProcPtr gets these arguments:
	( EventHandlerCallRef\EventRef\^UserData -- OSStatus )
	EventHandlerCallRef is a reference to the current handler call chain.  This is
		passed to your handler so that you can call CallNextEventHandler if you need to.
	EventRef is the event that triggered this call.
	^UserData is the application-specific data you passed in to InstallEventHandler
	OSStatus returns 0 if your EventHandler handled the event. 
	Return eventNotHandledErr (-9874) if you did not handle the event and perhaps 
		other handlers in the calling chain should take action
*)

Extern: UInt32 EventHandlerUPP NewEventHandlerUPP ( UInt32 EventHandlerProcPtr );

Extern: void DisposeEventHandlerUPP ( UInt32 EventHandlerUPP );


(* 
	An EventLoopTimerProcPtr gets these arguments:
	( EventLoopTimerRef\^UserData --  )
	EventLoopTimerRef is the timer that triggered this call.
	^UserData is the application-specific data you passed in to InstallEventHandler
*)

Extern: UInt32 EventLoopTimerUPP NewEventLoopTimerUPP ( UInt32 EventLoopTimerProcPtr );

Extern: void DisposeEventLoopTimerUPP ( UInt32 EventLoopTimerUPP );


(* *************************************************************************
Miscellaneous
	CreateTypeStringWithOSType	Converts an OSType to a Core Foundation String. 
	IsMouseCoalescingEnabled	Indicates whether mouse coalescing is enabled. 
	SetMouseCoalescingEnabled	Turns mouse coalescing on or off.
	IsUserCancelEventRef	Returns whether the specified event indicates the user wishes to cancel an operation.
	SetUserFocusWindow	Designates a window to receive user focus.
	GetUserFocusWindow	Returns the current user focus window.
************************************************************************* *)


Extern: UInt32 CFStringRef CreateTypeStringWithOSType ( UInt32 OSType );

Extern: UInt32 flag IsMouseCoalescingEnabled ( void  );

Extern: UInt32 OSStatus SetMouseCoalescingEnabled ( UInt32 inNewState, UInt32 * ^outOldState );

Extern: UInt32 flag IsUserCancelEventRef ( UInt32 EventRef );

Extern: UInt32 OSStatus SetUserFocusWindow ( UInt32 WindowRef );
Extern: UInt32 WindowRef GetUserFocusWindow ( void  );


\ ***************************************************************************
\ ***************************************************************************
\ ***************************************************************************


\ PRIVATE EVENT QUEUE

(* *************************************************************************

Maintain a private event queue to handle converted Carbon Events especially
	Activate, Deactivate, and Update events for windows that have a Mac 
	Standard Event Handler installed (like many Nib windows).

This mechanism may also be used to send "special" events to MacForth's Event 
	engine in the future (like scroll wheel events)

************************************************************************* *)


only Forth definitions
>M4TH
decimal


(*
Each Queue Element is 32 bytes:

 +0  +Qlink		= addr of next element
 +4  +payload	= 16 bytes for an "old-style" EventRecord
+20  +spare		= 12 bytes of "dry powder" for possible use with Special events
*)

create EventQueueHead  0 ,		\ contains 0 or pointer to oldest queued event
create FreeEventHead   0 ,		\ contains 0 or pointer to first unused Q Element
create EventQSemaphore 0 ,


256 constant Max#QueueElems
 32 constant QueueElemSize

0 value EventQueue


: NtheQueueEntry ( n -- addr )
	QueueElemSize *
	EventQueue +
;


: AddToFreeEntryList ( newQElemAddr -- )
	FreeEventHead @
	over !				\ link  prior head to newQElem
	FreeEventHead !		\ newQElem becomes head
;


: GetNextFreeEntry ( -- addr |or| 0 )	\ dequeues next entry from FreeList
	FreeEventHead @ dup
	IF
		dup @ FreeEventHead !	\ remove first item from FreeEntry list
		dup off					\ clear its link field
	THEN
;

: EmptyEventQueue ( -- )
	EventQSemaphore GET				\ 
	EventQueueHead off
	FreeEventHead  off
	Max#QueueElems 0		\ all go to free list
	DO
		i NtheQueueEntry AddToFreeEntryList
	LOOP
	EventQSemaphore RELEASE			\ 
;


\ ***************************************************************************

: LastQueueEntry ( -- QElemAddr |or| 0 )	\ get QElemAddr of last entry in list
	EventQueueHead @ dup
	IF
		BEGIN
			dup		\ prior entry
			@		\ next entry
		WHILE
			@
		REPEAT
	THEN
;

: EnqueueEntry ( NewQElemAddr -- )
	0 over !				\ link of new entry = 0
	LastQueueEntry ?dup
	IF
		!					\ link of prior last entry = NewEntryAddr
	ELSE
		EventQueueHead !
	THEN
;


: RemoveFirstMatchingEntry ( eventmask -- QElemAddr |or| 0 )
\ Tries to find a matching entry in EventQueueHead list
\ If found, removes entry from EventQueueHead list
\ Mask is OR'd values of all desired events, where
\	each event is represented as (1 shifted left by event number)
	EventQueueHead
	locals| PriorLinkAddress TheMask |
	EventQueueHead  @ 				\ start with oldest
	IF
		PriorLinkAddress @			\ first entry
		BEGIN
			\ current entry
			dup 4 + w@				\ get event type
			  1 swap lshift			\ 1 << event#
			  TheMask and
			  IF					\ it's a match
			  	dup @				\ get next link address
			  	PriorLinkAddress !	\ link it to the prior one
			  	exit
			  THEN
			dup
			@		\ next entry
		WHILE
			@
		REPEAT
		drop 0		\ no match found
	ELSE
\		drop
		0
	THEN
;


: GetFirstMatchingElem ( eventmask -- QElemAddr |or| 0 )
\ Tries to find a matching entry in EventQueueHead list
\ If successful, copies to EVENT.RECORD and recycles QElemAddr
	RemoveFirstMatchingEntry dup
	IF
		dup 4 + event.record QueueElemSize cmove	\ we copy MORE than a standard Event Record
		dup AddToFreeEntryList
	THEN
;


: AppendNewQEntry ( -- QElemAddr )
\ Adds a new empty Q entry
	GetNextFreeEntry ( -- addr |or| 0 )	\ dequeues it from FreeList
	?dup
	IF
		dup
		LastQueueEntry  ?dup
		IF						\ ^NewEntry\^NewEntry\^LastEntry
			!					\ link NewEntry onto prior LastEntry
		ELSE					\ ^LastEntry = 0   so queue is empty
			EventQueueHead !	\ so just add this entry at the top
		THEN
		dup OFF					\ clear link field of ^NewEntry
	ELSE			\ no free entries - re-use the oldest from EventQueueHead
		EventQueueHead @ dup	\ ^oldestEntry\^oldestEntry
		@ EventQueueHead !		\ next one becomes new oldest
		dup EnqueueEntry		\ add this one to end of list
		\  ???? ERRRROOORRR WARNING ?????
		LOG[[
			cr ." ****** EventQueue overflowed ******"
			base @ hex
			.rs
			cr
			base !
		]]LOG
 
	\	CR->Log  s" ****** EventQueue overflowed ******" Type->Log
	THEN
	dup off						\ clear the old event type
;


\ ***************************************************************************
\ We need to embellish a couple of primitives to work with our private event queue


: NEW.FLUSH.EVENTS ( -- )
	emptyeventqueue
	(flush.events)
;


: InitializeEventQueue ( -- )
	QueueElemSize Max#QueueElems *  NewPtrClear to EventQueue
	EmptyEventQueue
	['] NEW.FLUSH.EVENTS is flush.events 	\ dont patch until eventqueu initialized
;	PRELUDES linklasttoken



: DISCARD.UPDATES ( -- )
	EventQSemaphore GET				\ 
	BEGIN
		1 update.event lshift			\ 1 << event#
		RemoveFirstMatchingEntry 
		0=
	UNTIL
	EventQSemaphore RELEASE			\ 
	discard.updates
;


: DISCARD.ACTIVATES ( -- )
	EventQSemaphore GET				\ 
	BEGIN
		1 activate.event lshift			\ 1 << event#
		RemoveFirstMatchingEntry
		0=
	UNTIL
	EventQSemaphore RELEASE			\ 
	discard.activates
;


: ?EVENT+  ( RECORD\MASK -- flag | CHECK EVENT QUEUE FOR EVENT )
	EventQueueHead  @
	IF						\ anything in our private EventQueue ?
		2drop true
	ELSE
		(?event)
	THEN
;
	' ?EVENT+ is ?EVENT

\ ***************************************************************************
\ Now, we add a mechanism for converting Carbon EventRef (typically for
\	window Activates and Updates) to old-style Event Records and add to our
\	private event queue


create SCRATCH.EVENT.RECORD  	QueueElemSize 4 - allot&erase

: EnqueuePrivateEvent ( DataSrc -- flag )
	AppendNewQEntry ( -- QElemAddr )
	4 + QueueElemSize 4 - cmove
	false
;

: PostPrivateEvent ( ^eventrecord -- )
	EnqueuePrivateEvent ( DataSrc -- flag )
	drop
((	not
	CarbonEventLogging @ AND
	IF
		logcr " enqueued an event: " log$ 
		scratch.event.record w@ SysLogNumber
	THEN
))
;

: PostCarbonAsPrivateEvent	( EventRef -- )
	\ attempt to convert Carbon EventRef to EventRecord and add to our private Q
	EventQSemaphore GET					\ 
	scratch.event.record ConvertEventRefToEventRecord
	IF
		scratch.event.record PostPrivateEvent
	ELSE
(*  CarbonEvents.h
	 *    Currently, the following events can be converted: 
	 *    
	 *    kEventClassMouse: kEventMouseDown, kEventMouseUp,
	 *    kEventMouseMoved, kEventMouseDragged. 
	 *    
	 *    kEventClassKeyboard: kEventRawKeyDown, kEventRawKeyUp,
	 *    kEventRawKeyRepeat. 
	 *    
	 *    kEventClassControl: kEventControlTrack (in Mac OS X 10.4 and
	 *    later; converts to a mouseDown event). 
	 *    
	 *    kEventClassWindow: kEventWindowUpdate, kEventWindowActivate,
	 *    kEventWindowDeactivate, kEventWindowCursorChange. 
	 *    
	 *    kEventClassApplication: kEventAppActivated, kEventAppDeactivated.
	 *    
	 *    kEventClassAppleEvent: kEventAppleEvent.
*)
	THEN
	EventQSemaphore RELEASE 			\ 
;


: MakePrivateEvent ( Modifiers\Where\Message\Event# -- )
	\ use SCRATCH.EVENT.RECORD - don't reuse until it is Enqueued
	scratch.event.record  	QueueElemSize 4 - erase
	( Event#    )   scratch.event.record w!
	( Message   )   scratch.event.record 2 + !		\ typically WindowRef
	tickcount       scratch.event.record 6 + !
	( Where     )	scratch.event.record 10 + !		\ typically mouse location
	( Modifiers )	scratch.event.record 14 + w!	\ 
	\ remaining 12 bytes can be used for other messages
;


\ ***************************************************************************

\ Use this to postpone UI actions that rely on User variables
\ from within a Callback
\ Or to post a delayed action from within an event handler
\ SHOULD be safe to invoke from VFX Tasks

: SendExecuteEvent ( XT -- )
	0 swap 0 12 
	EventQSemaphore GET					\ 
	MakePrivateEvent ( Modifiers\Where\Message\Event# -- )
	scratch.event.record PostPrivateEvent 
	EventQSemaphore RELEASE 			\ 
;


\ ***************************************************************************
\ Finally, we patch @NEXTEVENT to handle our private queue
\ Giving it higher priority than the Classic Event Queue


: RESTORE.@NEXTEVENT ( -- )
	SKIP.EVENTS ON
	['] (@nextevent) IS @nextevent
	SKIP.EVENTS OFF
;  on.forget restore.@nextevent


create PRIVATE.EVENT.RECORD QueueElemSize allot&erase

: (GetNextPrivateEvent) ( mask --- eventcode )		\ 10.46
	EventQSemaphore GET			\ 
	EventQueueHead @			\ any data ?
	IF
		GetFirstMatchingElem ( eventmask -- addr |or| 0 )
		?dup
		IF	\ look in our private Queue first
			4 + Event.Record 16 move
			Event.Record private.event.record 28 move	\ make a copy
			Event.Record w@
		ELSE
			null.event
		THEN
	ELSE
		drop null.event
	THEN
	EventQSemaphore RELEASE			\ 
;
	' (GetNextPrivateEvent) is GetNextPrivateEvent



\ *************************************************************************
\ *************************************************************************


\ Application-Level Event Handler

(* *************************************************************************

Install a Carbon Event Handler that receives events directed to the 
Application Level

Initially, we will use it to catch Mouse Wheel events

************************************************************************* *)


\ For flexibility, we will use a DEFER word for our default Application Event handler:

DEFER ApplicationEventHandler ( EventHandlerCallRef\EventRef\userData -- OSErr )
	' 2drop is ApplicationEventHandler

\ 0 3 ' ApplicationEventHandler MakeProcPtr  ApplicationEventHandlerProcPtr

3 1  CALLBACK: ApplicationEventHandlerProcPtr
	' ApplicationEventHandler to-callback ApplicationEventHandlerProcPtr


0 value  ApplicationEventHandlerUPP


variable TheApplicationHandlerRef  TheApplicationHandlerRef off


10 constant kEventMouseWheelMoved


create ApplicationEventList
	kEventClassMouse , kEventMouseWheelMoved ,		\ eventClass  eventKind
\ We can easily extend the list of events in this list at any time by using
\	AddEventTypesToHandler



((
: ForgetApplicationEventHandler ( -- )
	TheApplicationHandlerRef @ _RemoveEventHandler drop
	ApplicationEventHandlerUPP _DisposeEventHandlerUPP
; 
\ on.forget ForgetApplicationEventHandler
))

: InstallApplicationEventHandler ( -- )
	ApplicationEventHandlerProcPtr NewEventHandlerUPP  -> ApplicationEventHandlerUPP

	GetApplicationEventTarget	\ EventTargetRef
	ApplicationEventHandlerUPP	\ EventHandlerUPP
	1							\ NumTypes
	ApplicationEventList		\ ^TypeList
	0							\ UserData
	TheApplicationHandlerRef	\ ^EventHandlerRef
	InstallEventHandler 
	drop
	\ abort" InstallEventHandler Failed"
; PRELUDES linklasttoken




variable TheApplicationHandler    TheApplicationHandler off


: SimpleApplicationEventHandler ( EventHandlerCallRef\EventRef\userData -- OSStatus )
	locals| userdata eventref eventhandlerref |
	TheApplicationHandler @ ?DUP	\ handler expects ( EventRef -- result )
	IF
		eventref swap execute
	ELSE
		eventNotHandledErr
	THEN
;

' SimpleApplicationEventHandler is ApplicationEventHandler



\ *************************************************************************
\ *************************************************************************

\ Scroll Wheel Events


(* *************************************************************************

Our Application level Carbon Event handler catches Mouse Wheel events.

A Mouse Wheel event can be asked to return a number of useful parameters:
    kEventMouseWheelMoved
        -->     kEventParamMouseLocation    typeHIPoint
        -->     kEventParamKeyModifiers     typeUInt32
        -->     kEventParamMouseWheelAxis   typeMouseWheelAxis
        -->     kEventParamMouseWheelDelta  typeLongInteger

If the current front window has no standard scrollbars, we ignore 
	Mouse Wheel events.
If the current front window has appropriate scroll bars, we create a new
"traditional" MacForth event:
	Control event:
		0 evnt.code.w		IN.NibControl -- 31
		2 message			WindowRef
		6 when				Tickcount
		10 where.pt			ControlRef
		14 modifiers.w		part code

Then we put it on our private event queue.

The ControlRef field lets you identify whether it is HBar or VBar

The Part Code will contain
	20 for up(left) button
	21 for down(right) button
	22 for Page up(left) 
	23 for Page down(right)

See DO.MOUSEWHEEL in the Editor for an example of useage

************************************************************************* *)

only Forth definitions
>M4TH
decimal



\ 31 constant IN.NibControl


: Forget>NibControl
	['] pause  event.table IN.NibControl 4 * + !
	TheApplicationHandler OFF
; on.forget Forget>NibControl


: >NibControl ( 31 -- 31 )
	\ GetControlReference to get address of MacForth control
	event.record dup 
	10 + @	this.control !					\ ControlRef
	14 + w@ this.part !
;

' >NibControl  event.table IN.NibControl 4 * + !



: EnqueueNibControlEvent ( PartCode\ControlRef -- )
	get.window @
	IN.NibControl
	EventQSemaphore GET				\ 
	MakePrivateEvent ( Modifiers\Where\Message\Event# -- )
	scratch.event.record PostPrivateEvent ( ^eventrecord -- )
	EventQSemaphore RELEASE			\ 
;


: MouseWheelHandler ( EventRef -- result )
	\ When a Scroll Wheel Mouse Event occurs in a front window with a scroll bar, 
	\	synthesize a "traditional" (old-style) event with a new type: IN.NibControl
	\ The controlref and part code will be set appropriately
	\ The synthesized event is added to our private event queue
	locals| EventRef |

	get.window
	DUP						\ could be 0 if in an alert, for example
	IF
		+vbar 2@ or 			\ do we have either a HBar or a VBar ?
	THEN

	IF
		EventRef GetEventClass kEventClassMouse =
		EventRef GetEventKind  kEventMouseWheelMoved =
		AND
		IF
			ascii mwdl			\ kEventParamMouseWheelDelta
			eventref GetLongEventParameter
			\ logcr "   WheelDelta = " log$ dup SysLogNumber
\	from sign of MouseWheelDelta we determine the part code
\	if option key is down, we convert to a page up/down
\			  kControlUpButtonPart          = 20,   \ also Left Button
\			  kControlDownButtonPart        = 21,	\ also Right Button
\			  kControlPageUpPart            = 22,	\ also Page Left
\			  kControlPageDownPart          = 23,	\ also Page Right

			0> IF 20 ELSE 21 THEN

((
typedef UInt16                          EventModifiers;
enum {
                                        /* modifiers */
  activeFlagBit                 = 0,    /* activate? (activateEvt and mouseDown)*/
  btnStateBit                   = 7,    /* state of button?*/
  cmdKeyBit                     = 8,    /* command key down?*/
  shiftKeyBit                   = 9,    /* shift key down?*/
  alphaLockBit                  = 10,   /* alpha lock down?*/
  optionKeyBit                  = 11,   /* option key down?*/
  controlKeyBit                 = 12,   /* control key down?*/
  rightShiftKeyBit              = 13,   /* right shift key down? Not supported on Mac OS X.*/
  rightOptionKeyBit             = 14,   /* right Option key down? Not supported on Mac OS X.*/
  rightControlKeyBit            = 15    /* right Control key down? Not supported on Mac OS X.*/
};
))
			ascii kmod					\ kEventParamKeyModifiers
			eventref GetLongEventParameter
			2048 and					\ mask for option key
			IF 2 + THEN					\ option key does paging

			0 sp@ >R			\ room for result..
			eventref
			ascii mwax			\ kEventParamMouseWheelAxis
			ascii mwax			\ typeMouseWheelAxis
			0
			2					\ sizeof(typeMouseWheelAxis)
			0
			R>
			GetEventParameter 	 drop
			swapw
			16 rshift
			IF
				get.window +vbar @ ?dup
				IF
					EnqueueNibControlEvent ( PartCode\ControlRef -- )
					0
				ELSE
					drop
					eventNotHandledErr
				THEN
			ELSE
				get.window +hbar @ ?dup
				IF
					EnqueueNibControlEvent ( PartCode\ControlRef -- )
					0
				ELSE
					drop
					eventNotHandledErr
				THEN
			THEN
		ELSE
			eventNotHandledErr
		THEN
	
	ELSE	\ no scroll bars, so just ignore it
		eventNotHandledErr
	THEN
;


' MouseWheelHandler  TheApplicationHandler !


\ *************************************************************************
\ *************************************************************************


\ Catch Window Events


(* *************************************************************************

When taking advantage of Apple's Standard Event Handlers on a window
	(especially Nib windows), the Standard Handler "eats" activate and 
	deactivate events and prevents them from getting to MacForth's event
	engine. 

This file creates a Window Carbon Event Handler to catch Activates, 
	Deactivates, and Updates before the Standard Event Handler does its
	thing.

To use it, add
	InstallWindowEventCatcher ( wptr4 -- )
after instantiating a window with ADD.WINDOW.FROM.NIB


A window may use a window-specific handler assigned by placing
	a token in wptr4 +SpecialEventHandler
	its stack map is: ( EventRef\theWPTR4 -- handled? )
	
The more common case is no window-specific action, but invoking
	a global action assigned by one of two DEFER words:

	WindowEventHandler ( EventRef\Wptr4 -- ) tries to convert the event to an
		old-style Event Record and place it in a special event queue
	WeirdWindowEventHandler	( EventRef\Wptr4 -- ) is invoked if the window receives
		a non-window class event

************************************************************************* *)



1 constant kEventWindowUpdate
\ 2 constant kEventWindowDrawContent
5 constant kEventWindowActivated
6 constant kEventWindowDeactivated
\ 7 constant kEventWindowGetClickActivation



\ Handle Activates, DeActivates, Updates
create WindowActivateTypeList
	kEventClassWindow  ,  kEventWindowActivated ,
	kEventClassWindow  ,  kEventWindowDeactivated ,
	kEventClassWindow  ,  kEventWindowUpdate ,

\ kEventWindowUpdate is sufficient
\	kEventClassWindow  ,  kEventWindowDrawContent ,

\ kEventWindowGetClickActivation could be used, but it skips activates
\ 	due to switching applications using CMD-Tab or the Dock
\	kEventClassWindow  ,  kEventWindowGetClickActivation ,


\ these Two Window Handler actions take:  ( EventRef\theWPTR4 -- handled? )
defer WindowEventHandler	  ( EventRef\theWPTR4 -- handled? )
	' drop IS WindowEventHandler

defer WeirdWindowEventHandler ( EventRef\theWPTR4 -- handled? )
	' drop IS WeirdWindowEventHandler


: (WindowActivateHandler)  ( EventRef\theWPTR4 -- handled? )
	locals| theWPTR4 eventref  |
	EventRef GetEventClass kEventClassWindow =
	IF
		EventRef theWPTR4  WindowEventHandler
	ELSE
		\ can we even GET a non-kEventClassWindow here ??
		EventRef theWPTR4  WeirdWindowEventHandler
	THEN
;

: WindowActivateHandler ( EventHandlerCallRef\EventRef\userData --  handled?  )
	locals| theWPTR4 eventref eventhandlerref |		\ UserData = WPTR4

	theWPTR4 +SpecialEventHandler @ ?dup
	IF	\ if there is a Window-Specific event handler, execute it
		\	that handler could call through to WindowEventHandler if appropriate
		EventRef theWPTR4  rot execute	( EventRef\theWPTR4 -- handled? )
		IF			\ +SpecialEventHandler did Not handle it
			EventRef theWPTR4 (WindowActivateHandler)
		ELSE
			0
		THEN
	ELSE	\ no +SpecialEventHandler, so execute the generic handler
		EventRef theWPTR4 (WindowActivateHandler)
	THEN
;


\ 0 3 ' WindowActivateHandler MakeProcPtr  WindowActivateHandlerProcPtr
3 1  CALLBACK: WindowActivateHandlerProcPtr
	' WindowActivateHandler to-callback WindowActivateHandlerProcPtr


0 value WindowActivateHandlerUPP




: RestoreWindowActivateHandler ( -- )
	WindowActivateHandlerProcPtr NewEventHandlerUPP
		-> WindowActivateHandlerUPP
;  PRELUDES linklasttoken

\ RestoreWindowActivateHandler


: InstallWindowEventCatcher ( wptr4 -- )
	dup >R
	+windowref# @ GetWindowEventTarget
	WindowActivateHandlerUPP
	3							\	 number of events in list
	WindowActivateTypeList
	R>								\ UserData = WPTR4
	dup +WindowEventHandlerRef		\ Stash EventHandlerRef in window tail
	InstallEventHandler abort" InstallWindowEventCatcher Failed"
;


\ *************************************************************************


: SimpleWindowEventHandler ( EventRef\Wptr4 -- handled? )
	locals| theWPTR4 EventRef |
	EventRef PostCarbonAsPrivateEvent
	0
;

' SimpleWindowEventHandler is WindowEventHandler



: SimpleWeirdWindowEventHandler ( EventRef\Wptr4 -- handled? )
	locals| theWPTR4 EventRef |
	EventNotHandledErr
;

' SimpleWeirdWindowEventHandler is WeirdWindowEventHandler





StopIncluding ************************************


