\ New Binary Controls
\ Created:       04/13/05  01:23:04 PM  wjm
\ Last Revision: 04/17/05  06:07:19 AM  wjm
\ Last Revision: 13 Jan 2012  08:20:05  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************



(* *************************************************************************

Nib windows are dandy features of Carbon.  They mostly run the same on
OS 9 and OS X (some window features and some types of controls are X-only).

Nib windows are the "modern" replacemnt for Dialogs.
Nib windows live in special package files that can each hold one or more Nib
	windows and all attendant controls.
Nib windows can only be created/edited with Apple's Interface Builder tool.

Unlike Dialog items, almost all user interface elements in Nib windows are
	controls of one sort of another.  Regular push buttons, radio buttons,
	check boxes, scroll bars, but also icon buttons, popup menus, grouped
	radio buttons, editable text fields, and lots of new widgets are all
	control types.  The data they contain is accessed differently than
	just getting the control value, as you would for a binary control or
	scroll bar.
Unlike Dialog item numbers, controls in Nib windows do NOT automatically get
	identifying numbers assigned. *YOU* must assign the identifiers in Project 
	Builder - at minimum, a control id, better to add a signature (typically,
	a 4 character value).  We use that sig/id pair to get the CtrlRef of a
	Nib control (using _GetControlByID)
	
While we can use "traditional" MacForth event handling to work with Nib controls,
	we then have to handle control hiliting on the click, following the mouse,
	and unhiliting the control when the mouse is released.
Easier, and usually preferable, is to make sure "Standard Handler" is checked
	in Interface Builder for the Nib window.  (Or add it programmatically
	with _InstallStandardEventHandler).  The standard handlers take care of
	all highliting issues, and check if the mouse was released while still 
	in the control.
If you use standard handlers, then your control needs a Carbon Event handler.

It turns out that normally we do not need a separate control handler for
	every control in a nib.  Typically, we used to run an event loop checking
	for control activity, then either update a variable with its current value
	or take some specific action.  Since the control highliting is handled by
	the standard handler, we only need to take action if a control was really 
	hit (mouse still in the control when button released). The Carbon Event
	type for this is kEventClassControl/kEventControlHit.  So we can create
	a generic handler for all nib controls that responds to that Carbon
	Event and can execute a control-specific token, or save the control
	value at a specific address.


*** See ":EXAMPLES:NIB EXAMPLES:" for actual usages of these words ***

NewNibBinControl: ( ID\sig\wptr "<spaces>name" --  | runtime: -- addr )
	is the new word for creating a Nib control structure. 
	Since the first thing stored in a NibCtrl record is the CtrlRef, you
	can do any of the usual control words like:
		myNibCtrl @ _GetControlValue
	The 3 parameters are sufficient to specify a particular control.  Since
	Nib controls are already sized, positioned, named, etc, we do not need
	to set those values as we do with the "traditional" NEW.CONTROL

InstallNibControl ( CntrlAddr -- )
	This is similar to ADD.CONTROL for "traditional" MacForth controls.
	AFTER the window containing the Nib has been added, each control is
	"attached" to MacForth using this word for each control.

!NibControlVariable ( addr\nibctrl -- )
	Optionally, assign the address of a variable to be updated with the
	current value of a Nib control.

!NibControlAction ( token\nibctrl -- )
	Optionally, assign a MacForth token to be executed when a Nib control is
	hit.

!NibControlUPP ( ^UPP\nibctrl -- )
	Optionally, assign a custom event handler UPP to a Nib control.

AddNibControlEvent ( type\class -- )
	You can additional Carbon events to a Nib control with this word.
	For some controls, you may want to respond to other events.  You might
	want a push button to respond when it is clicked at all, not if the
	mouse remains in.  

************************************************************************* *)


Only Forth Definitions
Decimal

>m4th


1  constant kEventControlHit			\ when mouse is released
13 constant kEventControlClick			\ when mouse is clicked in control



variable DefaultBinaryNibUPP 	 		\ contains UPP of binary control handler
			0 DefaultBinaryNibUPP !


\ We can install the same EventHandlerUPP on multiple targets
\ The EventHandler is a standard Mac EventHandler and will be called inside
\	a Mac UPP inside a MacForth ProcPtr
\ We can do lots of things inside such a thing, but must NEVER NEVER exit
\	the handler except by completing the word. NEVER execute QUIT, ABORT (or any 
\	of its friends that end up in quit. You should not put up alerts, dialogs
\	or attempt to switch windows.
\ And any word that executes DO.EVENTS explicitly or implicitly will 
\	always receive a null event.

\ For flexibility, we will use a DEFER word for our default binary control
\	handler:

DEFER BinaryNibControlHandler ( EventHandlerCallRef\EventRef\userData -- OSErr )

\ There is really very little a binary control handler usually needs to do.
\ Usually it is just setting a value that another part of the program will
\ check periodically.

\ We will offer a facility for posting a "traditional" MacForth event for
\ 	selected controls

\ You can specify a control-specific token for transient controls, like
\	push buttons, whose value changes only briefly while being pushed

: SimpleBinaryNibHandler ( EventHandlerCallRef\EventRef\userData -- OSErr )
	\ When we do InstallNibControl we set the user data to point to
	\   the MacForth Nib control structure
	\ EventRef is the event itself
	\ EventHandlerCallRef is the next chained event handler
	locals| mfCTRL  EventRef PriorHandler |

	mfCTRL @ this.control !

\ 10.47 move this here so control variable is set correctly even with a custom action
	mfCTRL 28 + @ ?dup				\ check for control-specific Variable
	IF								\ if the value is non-0
		( DBP+ )						\ address to update
		mfCTRL @ GetControl32BitValue	\ get the value
		swap !						\ save it
	THEN

	mfCTRL 8 + @ ?dup		\ is there a control-specific action?
	IF
		>R
		EventRef mfCTRL r> execute   \ xt gets ( EventRef\mfCtrl -- result ) 
		\ normally, your control action token should return eventNotHandledErr
		\ to allow system to do the control hiliting, etc
	ELSE
		eventNotHandledErr		\ return this to allow the default handlers to work
								\ for highlighting, toggling, etc
	THEN
(( 

	mfCTRL 28 + @ ?dup				\ check for control-specific Variable
	IF								\ if the value is non-0
		( DBP+ )						\ address to update
		mfCTRL @ GetControl32BitValue	\ get the value
		swap !						\ save it
	THEN
 ))
;
' SimpleBinaryNibHandler  is BinaryNibControlHandler



\ 0 3 ' BinaryNibControlHandler MakeProcPtr BinaryNibControlHandlerProc ( "<spaces>name"    procInfo\#params\token -- )	

3 1  CALLBACK: BinaryNibControlHandlerProc
	' BinaryNibControlHandler to-callback BinaryNibControlHandlerProc




: MakeBinaryNibUPP ( -- )
	BinaryNibControlHandlerProc		\ ['] BinaryNibControlHandler
	  NewEventHandlerUPP ( EventHandlerProcPtr -- EventHandlerUPP )
	  	DefaultBinaryNibUPP !
; PRELUDES linklasttoken	\ do it on restarting snapshots

\ MakeBinaryNibUPP			\ also, do it now, so we can use it


\ ***************************************************************************

: NewNibBinControl: ( ID\sig\wptr "<spaces>name" --  | runtime: -- addr )
	3 needed
	create			\ 
	0 ,				\ +0  <-- controlref
	ascii nibC ,	\ +4  --> mfControlType
					\ distinguish among old-style (ascii old_) and smart (ascii smrt)
	0 ,				\ +8  --> action token - gets ( EventRef\mfCtrl -- result ) 
					\		  a 0 in this field means no special action taken
	,				\ +12 --> wptr
	,				\ +16 --> controlsig
	,				\ +20 --> controlID
	0 ,				\ +24 --> address containing UPP for the 
					\		  EventHandler - a 0 means use DefaultBinaryNibUPP
	0 ,				\ +28 --> address that should be updated to the current
					\		  control value -- a 0 in this field means no save
	0 ,				\ +32 <-- event handler ref
	1 ,				\ +36 --> NumTypes
	kEventClassControl , \ +40 --> typelist.eventClass
	kEventControlHit   , \ +44 --> typelist.eventKind
;


: !NibControlVariable ( addr\nibctrl -- )
	\ set an address to maintain the current value of this control
	>R ( DBP- ) R> 28 + !
;

: !NibControlAction ( token\nibctrl -- )
	8 + !			\ xt gets ( EventRef\mfCtrl -- result )
;

: !NibControlUPP ( ^UPP\nibctrl -- )
	24 + !
;

: AddNibControlEvent ( type\class -- )
\	LastCreatedToken @ 
	latest-xt   >body dup >R 4 + @ ascii nibC <>
	abort" Must be used on a Nib Control"
	\ ERROR IF NOT USED IMMEDIATELY AFTER DEFINING A NEWNIBCONTROL:
	, ,
	1 R> 36 + +!	\ increment NumTypes
;  immediate



: ForgetBinaryNibControls ( -- )
	DefaultBinaryNibUPP @ DisposeEventHandlerUPP
; on.forget ForgetBinaryNibControls


\ *************************************************************************

: InstallNibControl ( CntrlAddr -- )	\ NIB window must be available
	dup >R
	12 + @ +windowref# @   \ get wRef to attach to
	dup 0=	abort" Window Not Open in InstallNibControl"
	  R@ 16 +				\ point to ControlID = (controlsig\controlID)
		R@ GetControlByID abort" GetControlByID failed in InstallNibControl"
	R@ dup @ swap SetControlReference	\ Set RefCon = MF control definition

	R@ @ GetControlEventTarget		\ EventTargetRef
	R@ 24 + @ ?dup					\ check for control-specific EventHandlerUPP
	IF								\ if EventHandlerUPP non-0
		( DBP+ ) @						\ use control-specific UPP
	ELSE
		DefaultBinaryNibUPP @ 		\ else use DefaultBinaryNibUPP
	THEN

	R@ 28 + @ ?dup					\ check for control-specific Variable
	IF								\ if there is one
		( DBP+ )					\ address to update
		R@ @ GetControl32BitValue	\ get the initial value from Nib
		swap !						\ save it
	THEN

	R@  36 + @						\ NumTypes
	R@  40 +						\ ^TypeList
	R@  							\ UserData == CntrlAddr
	R>  32 +						\ ^EventHandlerRef
	InstallEventHandler ( EventTargetRef\EventHandlerUPP\...
				...\NumTypes\^TypeList\UserData\^EventHandlerRef -- OSStatus )
				 abort" InstallEventHandler failed in InstallNibControl"
;


stopincluding