\ Nib Basics
\ Created:       04/04/05  04:27:52 PM  wjm
\ Last Revision: 05/29/06  01:35:21 PM  wjm
\ Last Revision: 12 Jan 2012  15:50:41  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


(* *************************************************************************

See Code Below for adapting Nib Windows that use a Mac Standard Event Handler
to work properly with MacForth's window activate routines.

A Nib File is actually an OS X Package, and looks like a folder on OS 9

Nib Files are used to design User Interface elements, notably Windows, Controls,
and Menus.  Apple recommends using Nib Files to replace Alerts and Dialogs
as well as for any window that has multiple controls and widgets.

Each Nib File can contain a single named user interface object, or many.

Nib Files can only be created on OS X using Interface Builder, but most Nib
Files can be used on older OS's that have CarbonLib installed.

Thus far, it does not seem that Nib-based Menus offer a lot of advantage over
existing older MacForth menus.

However, Nib-based windows can offer some advantages, if only for the ease of
designing fancy user interfaces (See :EXAMPLES:Tabbed Nib Demo")

Conventional OS X applications usually place their Nib File(s) in their
application package.  Since MacForth's development environment is not packaged
by default, we do not do that.  However, we it is easy to use Nib Files that are
in any folder, by asking the MacOS to give use a reference for the folder as a
framework.

OPENLOCALBUNDLE ( 0name$ -- bundleRef |or| 0 ) can be used to open a folder
(or a "real" bundle) as a bundle, where the folder/bundle is within the same
folder (or in a subfolder) as MacForth itself.
Given a BundleRef and the 0-terminated name of a Nib file,
GETNIBREFERENCEFROMBUNDLE ( CFBundleRef\NibFile0$ -- NibReference |or| abort )
is used to "open" that Nib file


During development, you can either keep your Nib files loose in the MacForth
folder or place them in the "MacForth NIBbles" folder (a sort of Resource Bag
for Nib files.  For turnkeys, you will almost certainly want to Bundle your
application (see ":EXTRAS:MAKE A BUNDLE") and place your Nib in the "Resources"
folder of your application bundle.  A handy shortcut to use to access your
Nib files is:
GETNIBREFERENCE ( 0$ -- NibReference |or| 0 )
It uses our own "poor man's search order" to find the specified Nib File.
First,   we check our own application bundle (if we are bundled)
Then,    we check the "MacForth NIBbles" folder (for a snapshot)
Finally, we check the folder our application is in (for development)


IMPORTANT:
Each Nib in a Nib File is accessed by its name, and Nib names, Nib File names,
framework names are treated by MacOS as CasE sENsiTive !!
Also note that Nib file names terminate in ".nib" in Finder, but we must specify
the name withOUT the suffix.


For more about Nibs, see:
http://developer.apple.com/documentation/Cocoa/Conceptual/CarbonCocoaDoc/Articles/CarbonUIInCocoa.html
http://developer.apple.com/documentation/Carbon/Reference/Interface_Builder/index.html

Also, the OS X developer disk contains Interface Builder and some
documentation for it.

************************************************************************* *)


Only Forth Definitions
Decimal

\ anew --NibBasics--
>m4th


true constant AreWeBundled?


0  constant kCFURLPOSIXPathStyle			\ CFURL.h


extern: UInt32 CFAllocatorRef CFAllocatorGetDefault ( void );

extern: UInt32 CFBundleRef CFBundleCreate ( UInt32 CFAllocatorRef, UInt32 CFURLRef );
extern: UInt32 CFURLRef CFURLCreateWithFileSystemPath ( UInt32 CFAllocatorRef, UInt32 filePath,
														UInt32 pathStyle, UInt32 isDirectory );

: GET-BUNDLEURL ( CFStringRef -- bundleUrl )
	\ return URL for specified bundle, false if error
	CFAllocatorGetDefault swap kCFURLPOSIXPathStyle 1 CFURLCreateWithFileSystemPath ;


: CREATE-BUNDLE ( bundleUrl -- bundleRef )
	\ create a bundle from the given Url and return a reference, false if error
	CFAllocatorGetDefault swap CFBundleCreate ;


: GET-BUNDLE ( 0string -- bundleRef | 0 )
	0 0 locals| theCFString theURL |
	>CFString dup dup -> theCFString
	IF
		get-bundleURL dup dup -> theURL
		IF
			create-bundle
			theURL  ?dup IF CFRelease THEN
		THEN	
		theCFString ?dup IF CFRelease THEN
	THEN
;

: MakePrivateBundleURL ( 0name$ -- CFURLRef )	\ previously GetPrivateBundleURL		???? WORKS ????
	\ create a CFURLRef for the folder or bundle specified by 0name$
	\ inside /Contents/MacOS/ subfolder in our Bundle
	>CFString 	  dup 0= abort" >CFString failed"
	dup >r
	0						\ use default allocator
	GetCFURLofMyContents	\ (MyCFURL)
	dup >r	\ BaseCFURLRef
	rot
	false 					\	true		\ a bundle IS a directory
	CFURLCreateCopyAppendingPathComponent
	r>		CFRelease	\ dispose the BaseCFURLRef
	r>		CFRelease	\ dispose the cfstring
;


: GetLocalBundleURL ( 0name$ -- CFURLRef )	\ LOCAL TO THIS APP
	\ create a CFURLRef for the folder or bundle specified by 0name$
	\ relative to the copy of MacForth executing this code
	>CFString 	  dup 0= abort" >CFString failed"
	dup >r
	0				\ use default allocator
	GetCFURLofMyApp'sFolder						\ MyCFURL
	dup >r	\ BaseCFURLRef
	rot
	false \	true		\ a bundle IS a directory
	CFURLCreateCopyAppendingPathComponent
	r>		CFRelease	\ dispose the BaseCFURLRef
	r>		CFRelease	\ dispose the cfstring
;


: <OpenPrivateBundle ( 0name$ -- bundleRef |or| 0 )
	\ Attempt to open the bundle specified by 0name$
	\ INSIDE our own Bundle's /Contents/MacOS/  subfolder
	AreWeBundled?
	IF
		MakePrivateBundleURL  dup 0= abort" MakePrivateBundleURL failed"
		dup >r 
		create-bundle
		r>		CFRelease			\ dispose the BundleURL
	ELSE							\ we are not in a bundlde
		drop 0
	THEN
;


: <OpenLocalBundle ( 0name$ -- bundleRef |or| 0 )
	\ Attempt to open the bundle specified by 0name$
	\ relative to the copy of MacForth executing this code
	GetLocalBundleURL 	  dup 0= abort" GetLocalBundleURL failed"
	dup >r 
	create-bundle
	r>		CFRelease	\ dispose the BundleURL
;


: OpenLocalBundle ( 0name$ -- bundleRef |or| 0 )
	\ Attempt to open the bundle specified by 0name$
	\  - we first check inside our own application bundle
	\  - then we check in the folder where our application bundle lives
	dup <OpenPrivateBundle ?dup
	IF
		nip
	ELSE
		<OpenLocalBundle
	THEN
;




extern: UInt32 OSStatus CreateNibReferenceWithCFBundle ( UInt32 CFBundleRef, 
											UInt32 CFStringRef, UInt32 * ^InNibRef );

extern: UInt32 OSStatus CreateWindowFromNib ( UInt32 nibRef, UInt32 CFStringRef, UInt32 * &window );

extern: void  DisposeNibReference ( UInt32 NibRef );


: CreateNibReferenceWithCFBundle+ ( CFBundleRef\CFStringRef -- NibReference\OSErr ) \ 6.05436
	\ same as _CreateNibReferenceWithCFBundle but with NibReference kept local
	0 >r rp@
	CreateNibReferenceWithCFBundle ( CFBundleRef\CFStringRef\^InNibRef -- OSStatus)
	r>												\ OSerr\NibReference
	swap
;

: (GetNibReferenceFromBundle) ( CFBundleRef\NibFileName0$ -- NibReference\OSErr )
	>CFString dup >r
	CreateNibReferenceWithCFBundle+
	r>  CFRelease									\ dispose CFString
;

: GetNibReferenceFromBundle ( CFBundleRef\NibFileName0$ -- NibReference |or| abort )
	(GetNibReferenceFromBundle)
	abort" CreateNibReferenceWithCFBundle failed"
;



: GetNibReference ( 0$ -- NibReference |or| 0 ) \ new - 6.05436
\ Make our own "poor man's search order" to find a specific Nib File
\ First, we check our own application bundle (if we are bundled)
\ Then, we check "MacForth NIBbles" folder (for a snapshot)
\ Finally, we check the folder our application is in (for development)
\ ** Note that Nib file names terminate in ".nib" in Finder, but
\	we must specify the name without the suffix in these calls
	>CFString >R
	AreWeBundled? dup
	IF		\ if we are bundled, then first look inside our own bundle
			\ this should find nibs loose inside /Contents/Resources/
			\ note; supplied string should NOT use the .nib suffix
		drop
		CFBundleGetMainBundle		\ return OUR BundleRef
		R@ CreateNibReferenceWithCFBundle+
		IF drop 0 THEN
	THEN
	?dup 0=
	IF	\ failed to find it in our application's Bundle
		\ so, look in our private named Nib Folder, relative to MF
		ASCII STR_ 402 GetResource dup	\ name of private Nib repository
		IF								\ default is "MacForth NIBbles"
			dup @ count pocket zplace
			ReleaseResource
			pocket
		ELSE
			z" MacForth NIBbles"
		THEN
		OpenLocalBundle dup				\ try to open it
		IF  
			R@ CreateNibReferenceWithCFBundle+ ( CFBundleRef\CFStringRef -- NibReference\OSErr )
			IF drop 0 THEN
		THEN
		?dup 0=
		IF	\ we have failed so far
			\ final case, use the folder containing our application
			\ this is useful during develoment, as we can keep the NIB
			\ file loose in the folder we test in.  For deployment, the NIB
			\ should be copied inside the RESOURCES folder of the Turnkey's Bundle.

			GetCFURLofMyApp'sFolder		\ mycfurl
			dup create-bundle
			R@ CreateNibReferenceWithCFBundle+ ( CFBundleRef\CFStringRef -- NibReference\OSErr )
			IF drop 0 THEN							\ return 0 if error
			swap CFRelease							\ dispose GetCFURLofMyApp'sFolder
		THEN
	THEN
	r>  CFRelease									\ dispose CFString
;


: ADD.WINDOW.FROM.NIB ( nibRef\nibName0$\wptr4 -- )	\ see (CREATE.WINDOW)
\ Nib Names are case sensitive !
	swap >CFString
	locals| CFName$ wptr4 |
	wptr4 close.window
	CFName$
	wptr4 +windowref#
	CreateWindowFromNib
	CFName$ CFRelease							\ dispose CFString
	abort" CreateWindowFromNib failed"
	wptr4 +windowref# @  wptr4 SetWRefcon		\ save our wptr4 as the wrefcon
	ascii NibW wptr4 +M4th.WTYPE !				\ 605421
								\ this can be overridden if Standard Event Handlers
								\ are not used on this window
;


: HasStandardHandler? ( wptr4 -- flag )
	+M4th.WTYPE @ ascii NibW =
;


: MakeNibRef  ( NibDir0$\NibFileName0$\addr -- )
	\ saves NibRef at addr
	\ NibFileName's and NibDir's are case sensitive !
	>R >R   openlocalbundle  DUP 0= abort" OPENLOCALBUNDLE failed"
	R> >CFString   R>   CreateNibReferenceWithCFBundle+
	abort" CreateNibReferenceWithCFBundle failed" 
;   



\ *************** comments and/or test code below ***************
StopIncluding



new.window www

\ open a Nib from "MacForth NIbbles" folder
z" Calculator" GetNibReference
z" Calculator3"
www add.window.from.nib
www show.window


\ Open a Nib from loose in App's Folder
z" testwindow1" GetNibReference .s
z" A Test Window" 
www ADD.WINDOW.FROM.NIB
www show.window


\ Open a Nib from /Contents/Resources/
z" testwindow" GetNibReference .s
z" A Test Window1" 
www ADD.WINDOW.FROM.NIB
www show.window


z" Calculator" GetNibReference .s

DATA STACK
     top
       13976240 00D5:42B0

ok \ 1 


