\ 39 Nib Miscellany.fth
\ Created:       20 Jul 2012  09:54:49  ward@megawolf.com
\ Last Revision: 20 Jul 2012  09:54:49  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to registered MacForth customers to use this
\   source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


(* *************************************************************************

Some additional common useful Nib functions


Even non-binary controls can be defined using  NewNibBinControl:
and added using InstallNibControl

But the events they respond to and the actions taken may be different

This has words for helping with Focus, basic Edit Text Controls, User Panes


I will try and add some DataBrowser & Clock Date control words soon


************************************************************************* *)

Only Forth Definitions


>M4TH
Decimal


\ Focus is the control that OS X highlites to take the next keyboard action,
\	be it text for a text control, a Tab to switch among controls, etc

Extern: UInt32 OSErr GetKeyboardFocus  ( UInt32 WindowRef, UInt32 * ^CtrlRef );
Extern: UInt32 OSErr ClearKeyboardFocus ( UInt32 WindowRef );

Extern: UInt32 OSErr SetKeyboardFocus ( UInt32 WindowRef, UInt32 CtrlRef, UInt32 CtrlPart );
(* from ControlDefinitions.h
	/**/
	/*   Control Part Codes                                                     */
	/**/
	enum {
	  kControlLabelPart             = 1,
	  kControlMenuPart              = 2,
	  kControlTrianglePart          = 4,
	  kControlEditTextPart          = 5,    /* Appearance 1.0 and later*/
	  kControlPicturePart           = 6,    /* Appearance 1.0 and later*/
	  kControlIconPart              = 7,    /* Appearance 1.0 and later*/
	  kControlClockPart             = 8,    /* Appearance 1.0 and later*/
	  kControlListBoxPart           = 24,   /* Appearance 1.0 and later*/
	  kControlListBoxDoubleClickPart = 25,  /* Appearance 1.0 and later*/
	  kControlImageWellPart         = 26,   /* Appearance 1.0 and later*/
	  kControlRadioGroupPart        = 27,   /* Appearance 1.0.2 and later*/
	  kControlButtonPart            = 10,
	  kControlCheckBoxPart          = 11,
	  kControlRadioButtonPart       = 11,
	  kControlUpButtonPart          = kAppearancePartUpButton,
	  kControlDownButtonPart        = kAppearancePartDownButton,
	  kControlPageUpPart            = kAppearancePartPageUpArea,
	  kControlPageDownPart          = kAppearancePartPageDownArea,
	  kControlClockHourDayPart      = 9,    /* Appearance 1.1 and later*/
	  kControlClockMinuteMonthPart  = 10,   /* Appearance 1.1 and later*/
	  kControlClockSecondYearPart   = 11,   /* Appearance 1.1 and later*/
	  kControlClockAMPMPart         = 12,   /* Appearance 1.1 and later*/
	  kControlDataBrowserPart       = 24,   /* CarbonLib 1.0 and later*/
	  kControlDataBrowserDraggedPart = 25   /* CarbonLib 1.0 and later*/
	};
*)



\ ***************************************************************************
\ get the menu "handle" (MenuRef) from a popup control
\ use that to modify the menu

extern: UInt32 MenuRef  GetControlPopupMenuHandle ( UInt32 ControlRef );


\ ***************************************************************************

((
	User Panes are a handy control for a nib
	We can do anything we want in them - stuff that standard controls dont handle
	I like them for short simple lists that do not warrant the complexity of a 
		Data Browser control - see Scientific Calculator demo or Vocab Browser

	The actual drawing must be done in an Extern: that we define and use to create
		a UserPaneDrawUPP

	The drawing action is then assigned as a data item for the control, as we
		do in the Vocab Browser:

	DrawListProc   NewControlUserPaneDrawUPP    DrawListUPP !

	ListPanel @ 0
	 ascii draw 				\ kControlUserPaneDrawProcTag
	  4 DrawListUPP
	   SetControlData ( ControlRef\PartCode\ResType\dataSize\^data -- OSStatus )
	    drop	\ abort" SetControlData error"
 ))


extern: UInt32 UPP  NewControlUserPaneDrawUPP ( UInt32 procptr );
extern: void 		DisposeControlUserPaneDrawUPP ( UInt32 procptr );





\ ***************************************************************************
\ EditTExt controls are useful for getting user text as was done in Dialogs

ascii cfst	CONSTANT kControlEditTextCFStringTag

(( 
    kEventTextInputUpdateActiveInputArea    = 1,
    kEventTextInputUnicodeForKeyEvent       = 2,
    kEventTextInputOffsetToPos              = 3,
    kEventTextInputPosToOffset              = 4,
    kEventTextInputShowHideBottomWindow     = 5,
    kEventTextInputGetSelectedText          = 6,
    kEventTextInputUnicodeText              = 7,
    kEventTextInputFilterText               = 14,
    kEventTextInputIsMouseEventInInlineInputArea = 16


  kEventTextInputUnicodeForKeyEvent = 2

 ))


: (SetControlText) ( cfstr\ctrlref -- )	\ does not release cfstr !
	{  | scratch[ 4 ] }
	swap scratch[ !
	0 kControlEditTextCFStringTag 4 scratch[
	SetControlData
	drop
;


: SetControlText ( addr\len\ctrlref -- )
	>R
	CFAllocatorGetDefault -rot
	kCFStringEncodingMacRoman 0
	CFStringCreateWithBytes
	dup R@ (SetControlText)
	CFRelease
	R> Draw1Control				\ ?? needed
;


: GetTextFromControl ( ^str\maxlen\ctrlref -- flag )
\ returns text from Edit Text controls (as in Nibs) into the supplied string buffer
	{ ^str maxlen ctrlref }
	ctrlref
	0
	kControlEditTextCFStringTag
	4
	^str		\ this will contain a CFStringRef 
	0
	GetControlData ( ControlRef\PartCode\ResType\maxSize\^buffer\^ActualSize -- OSStatus )
	0=
	IF
		^str @ dup >R
		^str maxlen
		0							\ use default encoding
		CFStringGetPascalString		\ TRUE if successful
		R> CFRelease
	ELSE
		0
	THEN
;


\ We can also use SetControlData to set some attributes (data tag constants) of a text control
\ See Vocabulary Browser for an example

\ some common ones:

\  kControlEditTextLockedTag     = 'lock', /* Boolean. Locking disables editability.*/
\  kControlEditTextSingleLineTag = 'sglc   /* Boolean. indicates whether the control should always be single-line */
\  kControlEditTextSpellCheckingTag = 'spck'  /* Boolean. indicates whether the control wants to have spell checking support. Available in Leopard and later. */

\ Control data tag constants are passed in the inTagName parameter of the getter and setter
\  functions SetControlData and GetControlData. 
\ kControlEditTextPart          = 5,



\ ***************************************************************************



stopincluding *********************************************************
******************* comments and/or test code below *******************
***********************************************************************

