\ VFX-style CONSOLE for MF
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************



only Forth definitions
>M4TH

decimal


((
struct /xterm-sid	\ -- len
\ Defines the SID of an xterm console device.
  gen-sid +			\ reuse field names of GEN-SID
\    	cell  field   gen-handle
\  		cell  field   gen-vector
\  		0     field gen-private
  int xs.hIn			\ input handle
  int xs.hOut			\ output handle
  int xs.hErr			\ error handle
  int xs.flags			\ control flags
				\ bit 0 - QUIT control
				\ bit 1 - rfu
				\ bit 2 - 1=maintain history, 0=none
				\ bit 3 - 1=history in system ini file
  int xs.hiBuff			\ address of 64k history buffer
  int xs.hiIndex		\ current history line#
  int xs.hiLowIndex		\ lowest history line#
  int xs.hiSection		\ pointer to INI file section name
end-struct
))


((
: SignalSetup
	#SigNest @ 
	IF 
		\ currentWindow @
		sys.window window
		sys.window SIZE.WINDOW
		\ currentWindow window
	THEN
;
))


\ ********** SAFE WRITE BUFFER *************************************

\ In the event we are in a signal handler, we can not safely output text
\ to a window, so we do "pseudo-writes" to a safe buffer


4000 value SafeWriteBufferSize

variable SafeWriteBuffer	SafeWriteBuffer off
variable SWB>IN				SWB>IN off
variable wasSWB>IN			wasSWB>IN off


: RemoveSafeWriteBuffer ( -- )
	SafeWriteBuffer @ ?dup IF disposeptr THEN
	SafeWriteBuffer off
	SWB>IN off
	wasSWB>IN off
; on.forget RemoveSafeWriteBuffer
presnapshot linklasttoken


: AllocateSafeWriteBuffer ( -- )  \ explicitly done in cold start
	RemoveSafeWriteBuffer
	SafeWriteBufferSize 1+ NewPtrClear 
		dup SafeWriteBuffer !
		dup SWB>IN !
			wasSWB>IN !
; 

\ AllocateSafeWriteBuffer		\ only do it here for testing


: SWB#Avail ( -- nn )
	SafeWriteBufferSize 
	SWB>IN @ SafeWriteBuffer @ -
	-
;

0 value ?NotifiedOfFull

: Type>SWB ( addr\cnt -- )
	\ circular buffer could be done, but not really necessary for this task
	\ check for cnt > SafeWriteBufferSize  for disasters?
	{: addr cnt :}
	SafeWriteBuffer @ 0= IF  exit THEN
\	SafeWriteBuffer @ 0= IF 2drop exit THEN
	?NotifiedOfFull 0=
	IF
		cnt SWB#Avail <
		IF
			addr cnt SWB>IN @  zplace	\ 
			cnt SWB>IN +!
		ELSE
			c" SafeWriteBuffer is full " dup SysLogMsg1 ( c$ -- )
			cr->log
			\ still need to do these explicitly as we are not using mfConsole here
			cr->log  count type->log

			true -> ?NotifiedOfFull
			beep			
		THEN
	THEN
;


: EmptySWB ( -- addr\cnt )	\ data is valid to write - one time
	SafeWriteBuffer @
	SWB>IN @ over - 
	SWB>IN @ wasSWB>IN !	\ save old SWB>IN just in case we want to rewrite
	SafeWriteBuffer @  SWB>IN !	\ reset SWB
	0 -> ?NotifiedOfFull
;



\ ********** REVECTOR PRIMITIVES *************************************
\ Most of these redirect through mfConsole

: read-mfterm	\ addr len sid -- ior			\ read-xterm
	3drop  0
;

: write-mfterm	\ addr len sid -- ior			\ write-xterm
	3drop  0
;



: KEY-mfterm	\ sid -- key		\ sid -- key		\ Key-xterm ,		\ 4
	drop
	#SigNest @ IF beep 100 ms drop  0 exit THEN
	mfKEY
;

: KEY?-mfterm	\ sid -- flag		\ Key?-xterm ,		\ 5
	drop
	#SigNest @ IF beep 100 ms drop true exit THEN
	anskey?
;

: ACCEPT-mfterm		\ address maxlen sid -- len'	\ Accept-xterm ,	\ 8
	drop
	ansaccept
;

: EMIT-mfterm	\ char sid --		\ Emit-xterm ,		\ 9
	drop
	#SigNest @ 
	IF 
		1 out +!
		sp@  1 Type>SWB drop
	ELSE
		mfEMIT			\ goes through TYPE
	THEN
;


: TYPE-mfterm	\ caddr len sid --		\ Type-xterm ,		\ 11
	drop

	#SigNest @ 
	IF 
		dup  out +!		
		Type>SWB
	ELSE
		mfTYPE
	THEN
;


: CR-mfterm	\ sid --			\ CR-xterm ,		\ 12
	drop 
	out off  op-line# incr

	#SigNest @ 
	IF
		cr$ count Type>SWB
	ELSE
	(CR)
	THEN
;


: lf-mfterm	\ sid --
	cr-mfterm
;

: open-mfterm	\ addr len attribs sid -- handle/sid ior
  nip nip nip
  dup xs.hOut @ over gen-handle !	\ mark as open
  $40 c/line !
  0
;

: close-mfterm	\ sid -- ior
  -1 swap gen-handle !			\ mark closed
  0
;

: ekey?-mfterm	\ sid -- flag
  key?-mfterm
;

: ekey-mfterm	\ sid -- echar
  key-mfterm
;

: emit?-mfterm	\ sid -- flag
\ Return true if we can send more characters
  drop  true  ;


: ff-mfterm	\ sid -- ; page/cls on display devices
	drop
  	page		\ ???? 
;

: bs-mfterm	\ sid -- ; destructive on display devices
	drop
;


: bell-mfterm	\ sid -- ; audible beeper
  7 swap emit-mfterm  ;


: setpos-mfterm	\ x y mode sid -- ior
\ Go to cursor position x,y. Position 0,0 is top left. Mode is ignored.
	3drop
	-1
;

: getpos-mfterm	\ mode sid -- x y ior
	2drop
	0 0 -1
;

: ioctl-mfterm	\ addr len fn sid -- ior
	4drop -1
;

: flushOP-mfterm	\ sid -- ior
	drop -1 ;

: readex-mfterm	\ caddr len sid -- u ior
	drop 2drop 0 -1 ;

: init-mfterm	\ addr len sid -- ior
  drop 2drop  0  ;

: term-mfterm	\ sid -- ior
  drop  0  ;

: config-mfterm	\ sid -- ior ; produces a dialog
  drop  0  ;


also system

create mfterm-vectors
  ' Open-mfterm ,	\ 0  OpenFnid
  ' Close-mfterm ,	\ 1
  ' Read-mfterm ,	\ 2
  ' Write-mfterm ,		\ 3
  ' KEY-mfterm ,		\ 4
  ' KEY?-mfterm ,	\ 5
  ' Ekey-mfterm ,	\ 6
  ' EKey?-mfterm ,	\ 7
  ' ACCEPT-mfterm ,	\ 8
  ' EMIT-mfterm ,	\ 9
  ' Emit?-mfterm ,	\ 10
  ' TYPE-mfterm ,	\ 11
  ' CR-mfterm ,		\ 12
  ' LF-mfterm ,		\ 13
  ' FF-mfterm ,		\ 14
  ' BS-mfterm ,		\ 15
  ' Bell-mfterm ,	\ 16
  ' setpos-mfterm ,	\ 17
  ' Getpos-mfterm ,	\ 18
  ' Ioctl-mfterm ,	\ 19
  ' FlushOP-mfterm  ,	\ 20
  ' readex-mfterm ,	\ 21
  ' Init-mfterm ,	\ 22
  ' Term-mfterm ,	\ 23
  ' config-mfterm ,	\ 24

previous


create defMFterm	\ -- addr
\ Default xterm structure - must match /xterm
  0 , mfterm-vectors , 0 , 1 , 2 ,  0 ,	\ last one is flags
  0 , 0 , 0 ,				\ history


: initmftermSid	\ addr --
\ *G Initialise a *\fo{/xterm} structure at *\i{addr}.
  defMFterm swap /xterm-sid cmove
;

: mfterm:	\ -- ; -- sid ; XTERM: <name>
\ *G Create a new terminal device.
  create
    here /xterm-sid allot initmftermSid
;

mfterm: mfXconsole	\ -- addr
\ *G VFX Forth console.

: init-mfcon	\ --
\ *G Set up to use the *\fo{xconsole} device. Performed at start up
\ ** and compilation.
  mfXconsole
\  $000C over xs.flags or!				\ demand history buffering
  dup ip-handle !   dup op-handle !		\ make it the current device
  dup ip-default !  dup op-default !	\ and the default device
  DROP 0 -> to PauseConsole
  NULL NULL NULL open-gen 2drop
;
\ ' init-xcon AtCold  init-xcon  -- init-mfcon DONE IN M4PLSTARTUP

: term-mfcon	\ --
\ *G Shut down the *\fo{xconsole} device. Performed at shut down.
;
\ ' term-xcon AtExit



\ ********** SAFER EXCEPTION HANDLER *************************************

[-opt
unoptimised

variable WaitingToShowException		WaitingToShowException off


: DfltAfterExceptionAction

	restartData  dup @ up!
	drop  \	dup cell + @ rp!  2 cells + @ sp!
	ip-default @ ip-handle !  op-default @ op-handle !

	EmptySWB ( -- addr\cnt )
	cr  type
	0 -> ?NotifiedOfFull

	cr ." Last word Interpreted : "
	here count ?dup 0=
	IF 
		'tib @ >in @
	THEN  type

	SigPause?
	if
		BEGIN key? WHILE key drop REPEAT	\ empty keystroke buffer
		TRUE BUSY
		cr ." Press E to ExitToShell, R to restart, other to continue: "
		CASE
			key upc
			FALSE BUSY
			[char] E of  (bye)      endof
			[char] R of  doRestart  endof
	  		SigThrow 
		ENDCASE
	else
		SigThrow
	endif
;

defer DoExceptionMessage ( -- )
	' DfltAfterExceptionAction is DoExceptionMessage


: AfterExceptionAction
	\ delay output from .SIGTRAP until we have left the callback
	WaitingToShowException @ 1 >
	IF
		c" NESTED in AfterExceptionAction " dup SysLogMsg1
		\ still need to do these explicitly as we are not using mfConsole here
		CR->Log count type->log CR->Log
		QUIT
	ELSE
		WaitingToShowException decr
		DoExceptionMessage
		QUIT
	THEN
;



: BadWord->Pocket ( -- )	\ 10.47 moved from 44 Dialogs.txt
	turnkeytoken @ 
	IF
		"" count
	ELSE
		saved>in @ dup 'tib @ +		\ wasOffset\start
		>in @ rot -
		-white
	THEN
	pocket place
;


: EMERGENCY!
	EmptySWB 2drop
	\ most of doRestart
	restartData  dup @ up!  dup cell + @ rp!  2 cells + @ sp!
	ip-default @ ip-handle !  op-default @ op-handle !
	\ still need to do these explicitly as we are not using mfConsole here
	cr->log s" Emergency recovery attempted ! "  Type->Log cr->log

	{  | str$[ 256 ] }
	
	s" A MAJOR error has occurred!"  str$[ place
	cr$cr$  count str$[ append
	
	s" Last word interpreted: "  str$[ append	\ 10.47
	BadWord->Pocket  pocket count str$[ append
	cr$cr$  count str$[ append

	s" Save your work and restart" str$[ append
	cr$cr$  count str$[ append

	str$[
	LogFile#
	IF
		c" for crash details, check "
		cr$
		LogFile$
	ELSE
		"" "" ""
	THEN
	
	Big:-(OK/QUIT# dfltissuealert 
	1 <>
	IF
		exittoshell
	THEN
	
	\ dump what we can from SafeWriteBuffer
	SafeWriteBuffer @   wasSWB>IN @ over - type
	QUIT
;


: SigUnNested2	\ signum *siginfo *ucontext -- eip
\ Handle unnested signal/fault.  safer than SigUnNested with windowed IO

	c" SigUnNested2 " SysLogMsg1
  	#SigNest incr
	
	WaitingToShowException @ 1 <
	?NotifiedOfFull	0= 				\ SWB#Avail 0 >
	AND
	IF
		WaitingToShowException incr
		.sigTrap
		\ still need to do this explicitly as we are not using mfConsole here
		SafeWriteBuffer @ SWB>IN @ over -  Type->Log

	  	#SigNest decr
		['] AfterExceptionAction
	ELSE			\ repeating Exceptions mean we are really munged up
		c" Emergency recovery attempted" dup SysLogMsg1
 		\ still need to do these explicitly as we are not using mfConsole here
		CR->Log
		count Type->Log	
		CR->Log
	
		['] EMERGENCY!
		WaitingToShowException off
	  	#SigNest off

	THEN
;


: SigNested2	\ signum *siginfo *ucontext -- eip
\ Handle nested signal/fault.
	c" SigNested2 BAD EXCEPTION" SysLogMsg1
	nip nip  #SigNest off			\ clear nesting

	-NestedSigs? 
	IF			\ check for permission
		['] (bye)
	ELSE
		['] SigThrow
	ENDIF
;


3 0 callback: SigGenTrap2	\ signum *siginfo *ucontext --

: SigGenTrapAction			\ signum *siginfo *ucontext --
\ *G Action of *\fo{SigGenTrap}. Displays an error message. On
\ ** return to OS X, a Forth *\fo{THROW} will occur.
\	c" SigGenTrapAction " SysLogMsg1

	2 pick c" OS EXCEPTION " SysLogNumber ( n\c1$ -- )

	dup >r
	#SigNest @
	if  
		SigNested2  
	else 
		SigUnNested2  
	endif
	r> setEIP
;
assign SigGenTrapAction to-callback SigGenTrap2



: setSigTraps2	\ --
\ *G Install the *\fo{SigGenTrap} signal handler for signals SIGILL,
\ ** SIGFPE, SIGBUS and SIGSEGV. Performed at startup.
  #SigNest off
  SigGenTrap2 SIGILL  setSignal
  SigGenTrap2 SIGFPE  setSignal
  SigGenTrap2 SIGSEGV setSignal
  SigGenTrap2 SIGBUS  setSignal
; 	\ ' setSigTraps2 AtCold
PRELUDES linklasttoken

opt]


StopIncluding ************************************

