\ SCRAP ROUTINES
\ Last Revision: 12 Feb 2012  08:22:02  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************



only Forth definitions
>M4TH
decimal



variable TEScrap.Changed	\ flag for scrap words defined later
  TEScrap.Changed OFF

: InitTEScrap  ( -- )
	TEScrap.Changed OFF  ;
Preludes LinkLastToken


Extern: UInt32 handle TEScrapHandle     ( void );
Extern: UInt32 length TEGetScrapLength  ( void );
Extern: void TESetScrapLength  ( UInt32 len );



(*
  	Under a future version of Carbon, there may be multiple scraps.
  	We'll need ScrapRefs to tell them apart.

	typedef struct OpaqueScrapRef* 			ScrapRef;
*)


(*
  	ClearCurrentScrap clears the current scrap. Call this
  	first when the user requests a Copy or Cut operation --
  	even if you maintain a private scrap! You should not wait
  	until receiving a suspend event to call ClearCurrentScrap. Call
  	it immediately after the user requests a Copy or Cut operation.
  	You don't need to put any data on the scrap immediately (although
  	it's perfectly fine to do so). You DO need to call GetCurrentScrap
  	after ClearCurrentScrap so you'll have a valid ScrapRef to pass
  	to other functions.
	EXTERN_API( OSStatus )
	ClearCurrentScrap				(void););) ;
*)
Extern: UInt32 OSStatus ClearCurrentScrap ( void );

(*
  	GetCurrentScrap obtains a reference to the current scrap.
  	The ScrapRef obtained via GetCurrentScrap will become
  	invalid and unusable after the scrap is cleared.

	EXTERN_API( OSStatus )
	GetCurrentScrap					(ScrapRef *				scrap) ;
*)

AliasedExtern: _GetCurrentScrap UInt32 OSStatus GetCurrentScrap    ( UInt32 * ^scrapref );
\ Extern: UInt32 OSStatus GetCurrentScrap    ( UInt32 * ^scrapref );

\ : _GetCurrentScrap  GetCurrentScrap ;	\ disambiguate

: GetCurrentScrap ( -- scrapref )
	0 sp@ _GetCurrentScrap drop ( result )
;


(*
  	GetScrapFlavorFlags tells you [a] whether the scrap contains
  	data for a particular flavor and [b] some things about that
  	flavor if it exists. This call never blocks, and is useful
  	for deciding whether to enable the Paste item in your Edit
  	menu, among other things.

	OSStatus GetScrapFlavorFlags	(ScrapRef 				scrap,
									 ScrapFlavorType 		flavorType,
									 ScrapFlavorFlags *		flavorFlags) ;
	struct ScrapFlavorInfo {
		ScrapFlavorType 				flavorType;  ( long )
		ScrapFlavorFlags 				flavorFlags; ( long )
	};
	enum {
		kScrapFlavorTypePicture		= FOUR_CHAR_CODE('PICT'),		/* contents of a PicHandle*/
		kScrapFlavorTypeText		= FOUR_CHAR_CODE('TEXT'),		/* stream of characters*/
		kScrapFlavorTypeTextStyle	= FOUR_CHAR_CODE('styl'),		/* see TEGetStyleScrapHandle*/
		kScrapFlavorTypeMovie		= FOUR_CHAR_CODE('moov'),		/* reference to a movie*/
		kScrapFlavorTypeSound		= FOUR_CHAR_CODE('snd '),		/* see SndRecord and SndPlay*/
		kScrapFlavorTypeUnicode		= FOUR_CHAR_CODE('utxt'),		/* stream of UTF16 characters*/
		kScrapFlavorTypeUnicodeStyle = FOUR_CHAR_CODE('ustl')		/* ATSUI defines; Textension uses*/
	};
	enum {
		kScrapFlavorMaskNone		= 0x00000000,
		kScrapFlavorMaskSenderOnly	= 0x00000001,
		kScrapFlavorMaskTranslated	= 0x00000002
	};
*)
Extern: UInt32 OSStatus GetScrapFlavorFlags  ( UInt32 scrapref, UInt32 flavorType, UInt32 * ^flags );

(*
  	GetScrapFlavorSize gets the size of the data of the specified
  	flavor. This function blocks until the specified flavor
  	data is available. GetScrapFlavorSize is intended as a prelude
  	to allocating memory and calling GetScrapFlavorData.
	
	EXTERN_API( OSStatus )
	GetScrapFlavorSize				(ScrapRef 				scrap,
									 ScrapFlavorType 		flavorType,
									 Size *					byteCount) ;
*)
Extern: UInt32 OSStatus GetScrapFlavorSize  ( UInt32 scrapref, UInt32 flavorType, UInt32 * ^size );


(*
  	GetScrapFlavorData gets the data from the specified flavor in the
  	specified scrap. This function blocks until the specified flavor
  	data is available. Specify the maximum size your buffer can contain;
  	on output, this function produces the number of bytes that were
  	available (even if this is more than you requested).
	
	EXTERN_API( OSStatus )
	GetScrapFlavorData				(ScrapRef 				scrap,
									 ScrapFlavorType 		flavorType,
									 Size *					byteCount,
									 void *					destination) ;
*)
Extern: UInt32 OSStatus GetScrapFlavorData ( UInt32 scrapref, UInt32 flavorType, UInt32 * ^Size, UInt32 * ^dest );

(*
  	PutScrapFlavor is a lot like PutScrap, with two differences:
  	we added a ScrapRef parameter at the beginning and you can
  	"promise" various aspects of a flavor. If you pass a NIL
  	data pointer, this is a promise that in the future you
  	will provide data for this flavor. Provide the data
  	through a subsequent call to PutScrapFlavor, either later
  	in the same code flow or during a scrap promise keeper proc.
  	If you know how big the data is, you can pass the size as
  	well, and this may allow subsequent callers of GetScrapFlavorInfo
  	to avoid blocking. If you don't know the size, pass -1.
  	If you pass a 0 size, you are telling Scrap Manager not to
  	expect any data for this flavor. In this case, the flavor
  	data pointer is ignored. NOTE: the last time you can provide
  	scrap flavor data is when your scrap promise keeper gets
  	called. It is NOT possible to call PutScrapFlavor while
  	handling a suspend event; suspend events under Carbon
  	simply don't work the way they do under Mac OS 8.

	EXTERN_API( OSStatus )
	PutScrapFlavor					(ScrapRef 				scrap,
									 ScrapFlavorType 		flavorType,
									 ScrapFlavorFlags 		flavorFlags,
									 Size 					flavorSize,
									 const void *			flavorData) /* can be NULL */;
*)
Extern: UInt32 OSStatus PutScrapFlavor    ( UInt32 scrapref, UInt32 flavorType, UInt32 flavorFlags, UInt32 Size, UInt32 * ^data );

(*
  	ScrapPromiseKeeper is a function you write which is called by
  	Scrap Manager as needed to keep your earlier promise of a
  	particular scrap flavor. When your function is called, deliver
  	the requested data by calling PutScrapFlavor.

	typedef CALLBACK_API( OSStatus , ScrapPromiseKeeperProcPtr )(ScrapRef scrap, ScrapFlavorType flavorType, void *userData) ;
	typedef STACK_UPP_TYPE(ScrapPromiseKeeperProcPtr) 				ScrapPromiseKeeperUPP;
	EXTERN_API(ScrapPromiseKeeperUPP)
	NewScrapPromiseKeeperUPP	   (ScrapPromiseKeeperProcPtr userRoutine) ;

	EXTERN_API(void)
	DisposeScrapPromiseKeeperUPP    (ScrapPromiseKeeperUPP	userUPP) ;

	EXTERN_API(OSStatus)
	InvokeScrapPromiseKeeperUPP	   (ScrapRef				scrap,
									ScrapFlavorType			flavorType,
									void *					userData,
									ScrapPromiseKeeperUPP	userUPP) ;
*)
Extern: UInt32 ScrapPromiseKeeperUPP NewScrapPromiseKeeperUPP      ( UInt32 procptr );
Extern: void   DisposeScrapPromiseKeeperUPP  ( UInt32 ScrapPromiseKeeperUPP );
Extern: UInt32 OSStatus InvokeScrapPromiseKeeperUPP   ( UInt32 scrapref, UInt32 flavorType, UInt32 * ^data, UInt32 ScrapPromiseKeeperUPP );

(*
  	SetScrapPromiseKeeper associates a ScrapPromiseKeeper with a
  	scrap. You can remove a ScrapPromiseKeeper from a scrap by
  	passing a NIL ScrapPromiseKeeper to SetScrapPromiseKeeper.
  	Pass whatever you like in the last parameter; its value will
  	be passed to your ScrapPromiseKeeper, which can do whatever
  	it likes with the value. You might choose to pass a pointer
  	or handle to some private scrap data which the
  	ScrapPromiseKeeper could use in fabricating one or more
  	promised flavors.


	EXTERN_API( OSStatus )
	SetScrapPromiseKeeper			(ScrapRef 				scrap,
									 ScrapPromiseKeeperUPP 	upp,
									 const void *			userData) ;
*)
Extern: UInt32 OSStatus SetScrapPromiseKeeper   ( UInt32 scrapref, UInt32 ScrapPromiseKeeperUPP, UInt32 * ^data );

(*
	CallInScrapPromises forces all promises to be kept.  If your application promises at least one
	flavor AND it does NOT adopt the new event model, you should call this function when your
	application is about to quit.  If your app promises no flavors, there is no need to call this
	function, and if your app adopts the new event model, this function will be called automagically
	for you.  It doesn't hurt to call this function more than once, though promise keepers may be asked
	to keep promises they already tried and failed. 

	EXTERN_API( OSStatus )
	CallInScrapPromises				(void) ;
*)
Extern: UInt32 OSStatus CallInScrapPromises ( void );

(*
  	GetScrapFlavorCount produces the number of
  	items which can be obtained by GetScrapFlavorInfoList.
	
	EXTERN_API( OSStatus )
	GetScrapFlavorCount				(ScrapRef 				scrap,
									 UInt32 *				infoCount) ;
*)
Extern: UInt32 OSStatus GetScrapFlavorCount   ( UInt32 scrapref, UInt32 * ^InfoCount );

(*
  	GetScrapFlavorInfoList fills a list (array)
  	of items which each describe the corresponding
  	flavor in the scrap. It fills no more array
  	elements as are specified. On exit, it produces
  	the count of elements it filled (which may be
  	smaller than the count requested). Yes, YOU
  	must provide the memory for the array.

	EXTERN_API( OSStatus )
	GetScrapFlavorInfoList			(ScrapRef 				scrap,
									 UInt32 *				infoCount,
									 ScrapFlavorInfo 		info[]) ;
*)
Extern: UInt32 OSStatus GetScrapFlavorInfoList   ( UInt32 scrapref, UInt32 * ^infoCount, UInt32 * ^data );



create CurrentScrap
	0 ,		\ scrap ref
	0 ,		\ size
	0 ,


: PUT.SCRAP  (  ADDR\CNT\FLAVOR   -- IO-RESULT)
				( COPIES STRING  TO  DESK SCRAP )
	locals| Flavor cnt addr |
	ClearCurrentScrap 		       abort" ClearCurrentScrap ERROR"
	CurrentScrap _GetCurrentScrap  abort" _GetCurrentScrap ERROR"
	CurrentScrap @ Flavor 0 ( flags )  cnt addr
	PutScrapFlavor ;  ( scrapref\flavorType\flavorFlags\Size\^data -- result )



: GET.SCRAP  ( dest.handle\flavor -- size )	\ 
				(  FETCHES DESK SCRAP TO HANDLE )
\ A handle to the memory location where the GetScrap function should place the data from the
\ scrap.  If you specify NIL in this parameter, the GetScrap function does not read in the
\ data but does return the offset of the data in the scrap and the number of bytes of the
\ requested scrap data type if the requested type exists in the scrap. 
	locals| Flavor  handle |
	CurrentScrap _GetCurrentScrap
		?dup IF exit THEN
	CurrentScrap @ Flavor CurrentScrap 4 + GetScrapFlavorSize  ( scrapref\flavorType\^size -- result )
		?dup IF exit THEN
	handle
	IF
		handle CurrentScrap 4 + @  resize.handle
			?dup IF exit THEN
		handle lock.handle
		CurrentScrap @
		 Flavor
		  CurrentScrap 4 +
		   handle @
		    GetScrapFlavorData  ( scrapref\flavorType\^Size\^dest -- result )
		handle unlock.handle
		?dup FALSE= IF CurrentScrap 4 + @ THEN		\ return length if not error
	ELSE
		CurrentScrap 4 + @		\ return size of scrap
	THEN
;


: Scrap.Len  ( -- addr )  \ address of 32-bit TEXT scrap length
	CurrentScrap  4 + off
	CurrentScrap _GetCurrentScrap 
	FALSE=
	IF
		CurrentScrap @ ascii TEXT CurrentScrap 4 + GetScrapFlavorSize 
		IF CurrentScrap  4 + off THEN
	THEN
	CurrentScrap 4 +
;



\ These scrap words don't abort on errors, but return a value in IO-Result

: ScrapTE>Desk  ( -- )	\ copy TE scrap to desk-scrap
	IO-Result OFF
	TEGetScrapLength 0>
	IF
		ClearCurrentScrap  ( CARBON  _ZeroScrap ) dup 0 >=		\ S: resultCode\flag
		IF
			drop				\ discard Zero.Scrap's result-code
			TEScrapHandle		\ keep handle on stack during this..
			dup HLock
			dup @ 
			TEGetScrapLength ascii TEXT
			PUT.SCRAP  (  ADDR\CNT\RESOURCE.TYPE   -- IO-RESULT)
			swap HUnlock
			\ leave TE-scrap handle UNLOCKED, assuming it started that way
		THEN
		IO-Result !
	THEN  ;


: ScrapExport  ( -- )	\ copy TE scrap to desk-scrap if necessary
	IO-Result OFF
	TEScrap.Changed @
	IF
		ScrapTE>Desk			\ retain result-code in IO-Result
		TEScrap.Changed OFF
	THEN  ;


: ScrapImport  ( -- )  \ copy desk-scrap to TE, if any TEXT in it
	IO-Result OFF
	Scrap.Len @ 0>
	IF
		0 TESetScrapLength
		TEScrapHandle HUnlock
		TEScrapHandle 0 SetHandleSize
		0  ascii TEXT  Get.Scrap	\ result = data-size or error-code
		1 32767 RANGE				\ S: result\inRange?
		IF
			drop
			TEScrapHandle ascii TEXT  Get.Scrap
			dup 0 max TESetScrapLength		\ store TEScrap length
		then
		0 min IO-Result !	\ put negative error-code or 0 into IO-Result
		TEScrap.Changed OFF	\ only change if TEscrap was changed
	THEN  ;




\ These scrap words abort if an error in scrap copying occurs

: Put.TEScrap  ( -- ) \ copy TE scrap to desk scrap, abort if error
	ScrapTE>Desk  ?File.Error  TEScrap.Changed OFF  ;

: ?Put.TEScrap  ( -- )  \ TE scrap to Clipboard if necessary, abort if error
	ScrapExport  ?File.Error  ;

: Get.TEScrap  ( -- )  \ copy Clipboard to TE scrap if TEXT, abort if error
	ScrapImport
	IO-Result @ -102 <> if  ?File.Error  then		\ only abort if not -102
	TEScrap.Changed OFF  ;



Extern: UInt32 teH TENew  ( UInt32 * &destR, UInt32 * &viewR );
Extern: void TESetText  ( UInt32 textP, UInt32 len, UInt32 teH );
Extern: void TECalText  ( UInt32 teH );
Extern: void TEDispose  ( UInt32 teH );




StopIncluding ************************************
