\ MISCELLANY
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


((

	AllWindowsFront

	SEARCH-TEXT
	
	ASK.NUMBER

	SYS.WINDOW.ProxyIcon
	
	LaunchURL
	
	FileContents->Handle
	
	MacForth Date&Time words
	
))

only Forth definitions
>M4TH
decimal


\ *********************************************************

: AllWindowsFront ( -- )	
	\ to bring all our windows to front use SetFrontProcess with our PSN
	mypsn SetFrontProcess drop
;


\ *********************************************************
\ Based on Boyer-Moore algorithm as described in Sedgewick's _Algorithms_

defer ?UPCASE  ( n -- n' )		' noop is ?UPCASE

: LoCase  ( n -- n' )
	ascii A ascii Z range IF 32 + THEN ;

: UPCASE  ( char -- char' )
	upc ;


: InitSkipTable  ( patternAddr\patternCnt\skipTable -- )
	ascii a dup ?upcase <> 
	locals| notCaseSensitive skipTable patternCnt patternAddr |
\	{ patternAddr patternCnt skipTable notCaseSensitive }
	
	SkipTable 256 patternCnt  fill

	patternCnt 0
	?DO
		\ later occurences of same char replace earlier ones
		patternCnt 1- i - dup
		patternAddr i + C@ skipTable + c!
		\ store in lower-case position as well if notCaseSensitive
		notCaseSensitive
		IF
			patternAddr i + C@   LoCase skipTable + c! 
		ELSE
			drop 
		THEN
	LOOP ;


\ If the notCaseSens? flag is on, then the pattern is assumed
\ to be in upper case already.
: (SEARCH-TEXT) ( textAddr\textLen\patternAddr\patternLen\notCaseSens?\tableAddr ... )
			( ... -- [false] or [hitaddr\true])
	swap IF ['] upcase IS ?upcase ELSE ['] noop IS ?upcase THEN
	over 
	locals| patternIndex skipTable patternLen patternAddr textEnd textAddr |

	textAddr +to textEnd

	patternAddr patternLen skipTable  InitSkipTable

	patternIndex 1- +to textAddr
	-1 +to patternAddr					\ because patternIndex is 1-based
	BEGIN
		textAddr textEnd >= IF false EXIT THEN
		textAddr c@ ?upcase	  patternIndex patternAddr + c@ =
		IF 
			-1 +to patternIndex -1 +to textAddr
		ELSE
			patternLen patternIndex - 1+
			textAddr c@ skipTable + c@
			MAX	+to textAddr
			patternLen -> patternIndex
		THEN
		patternIndex 1 <
	UNTIL
	textAddr 1+ true ;


: SEARCH-TEXT ( textAddr\textLen\patternAddr\patternLen\notCaseSens? ... )
			( ... -- [false] or [hitaddr\true])
	256 NewHandle dup 0= MemError?
	dup >R @ (search-text) R> DisposeHandle ;



\ *********************************************************
\ Numeric Conversions

: CHECK.SIGN	( addr\cnt -- addr\cnt\sign  )
	over c@  dup 45  =
	dup >r
	swap 43  = or
	IF 1 /string  THEN
	r> ;

create DP-CHARS		46 W,	\ decimal point separator
					44 W,	\ thousands separator  (ignored)


: $>NUMBER  ( addr\cnt  -- n\t or f)
	2 needed  dpl off
	check.sign 0 -1
	locals| FLAG SUM SIGN #CHARS ADDR |
	addr echo 1+ C@ - boolean #chars AND	( >0 chars, first -cr)
	IF addr #chars + addr
		DO  i c@
			CASE  48 57	  RANGE.OF I C@ 48 -  sum 10 *  + to sum	ENDOF
				dp-chars w@		OF #chars addr + i -  1- DPL !	ENDOF
				dp-chars 2 + w@	OF  ( ignore , )				ENDOF
				echo 1+ c@		OF  ( ignore eol char )			ENDOF
				( all else )	0 to flag  ( invalid char found )
			ENDCASE
		LOOP
		sum	 sign 0<  IF  negate  THEN
		flag  dup 0= IF nip THEN
	ELSE  false
	THEN  ;


: ASK.NUMBER  ( width -- [n\true] or [false] )
	1 needed
	pad swap  iexpect		\ EXPECT	\ always input from keyboard
	pad span @	2dup + 1- c@  echo 1+ c@ =  \ adjust for cr delim
	IF 1- THEN
	$>number ;


\ *********************************************************
\ Proxy Icons ( in document window titles )

\ we need an Alias Handle to add the proper Proxy Icon

Extern: UInt32 OSErr  NewAliasMinimal ( UInt32 * ^toSpec, UInt32 * ^Handle );
\ note: one of these days, should convert these to the FS* versions

\ Extern: UInt32 OSStatus  SetWindowProxyAlias ( UInt32 WindowRef, UInt32 AliasHandle );



create theAliasH  0 ,					\ scratch

: FRef->Alias ( ref# -- AliasHandle or 0 )
	{: ref# |  theFSSpec[ FSSpecRecord ]  fname$[ 256 ]  :}
	theAliasH off
	theFSSpec[  FSSpecRecord  erase
	fname$[ ref# GetFCBInfo ( ^namePtr\fileRef# -- VolRefNum\DirID\flags ) 
	\ (GetPathInfo) ( fileID -- VRef#\folder# )
	drop		\ flags
	io-result @ 0=
	IF
		fname$[ theFSSpec[ FSMakeFSSpec 0=
		IF
			theFSSpec[ theAliasH NewAliasMinimal
			0=		\ non-zero means success
		ELSE
			0
		THEN
	ELSE
		2drop
		0
	THEN

	dup
	IF
		drop
		theAliasH @
	ELSE
		theAliasH @ to.heap
	THEN
;

: BringFinderFront ( -- )
	{:  | TheProcess[ 8 ] TheProcessInfoRecord[ 64 ]   :}
	TheProcess[ 8 erase
	TheProcess[ GetNextProcess drop
	BEGIN								\ search all processes for Finder
		TheProcessInfoRecord[  64 erase
		60 TheProcessInfoRecord[ !
		TheProcess[ TheProcessInfoRecord[ GetProcessInformation
		IF
			TheProcessInfoRecord[ 8 + off	\ so zap the processNumber
			false							\ exit the loop
		ELSE								\ error before we located FINDER
			TheProcessInfoRecord[ 16 + @		\ get the processType
			ascii FNDR <>
		THEN
\	WHILE
		TheProcess[ GetNextProcess	dup 
		0= IF								\ we checked them all and no FInder
			TheProcessInfoRecord[ 16 + off	\ zap the processNumber
		THEN
		OR
		
	UNTIL

	TheProcessInfoRecord[ 16 + @				\ assuming no error
	IF
		TheProcess[ SetFrontProcess drop	\ then we have the FInder
	THEN
;


\ -------------------------------------------------------------------
\ expand the function of our IN.DRAG.BOX event handler:


: NEW>DRAG.WINDOW ( -- )
	{:  |  PathMenuItem[ 4 ]  :}
	get.window @  event.record IsWindowPathSelectClick
	IF
		get.window @ 0 PathMenuItem[ WindowPathSelect
		0=
		PathMenuItem[ @ 0=		\ selecting first item doesn't count
		AND
		IF
			BringFinderFront	\ bring finder to the front
		THEN
	ELSE
		>DRAG.WINDOW
	THEN
;

\ replace Event Table entries
' NEW>DRAG.WINDOW  event.table in.drag.box   4 * + !
' NEW>DRAG.WINDOW  event.table in.proxy.icon 4 * + !



: ForgetPathSelect ( -- )	\ let's be compulsive
	['] >drag.window  event.table in.drag.box 4 * + !
	['] pause         event.table in.proxy.icon 4 * + !
;  on.forget ForgetPathSelect



\ -------------------------------------------------------------------
\ Optional, add a proxy icon to SYS.WINDOW itself:

variable UseAppForProxy		UseAppForProxy ON	\ 10.47

: SYS.WINDOW.ProxyIcon ( -- )
	sys.window ?wvisible	\ 6.04998
	IF
\		MyRWResFile# FRef->Alias ?dup
\ 10.47 -  optionally use application itself rather than the embedded MACH-O file
		UseAppForProxy @
		IF
			GetCFURLofMyApp 
			dup >R CFUrl->CString 
			R> CFRelease
		ELSE
			MyCFURL CFUrl->CString
		THEN
		count  Path->FSSpec ?dup
		IF
			theAliasH  NewAliasMinimal 0=
			IF
				get.window	>R			\ preserve grafport around this call
				sys.window @ theAliasH @ SetWindowProxyAlias 0=
				IF
					sys.window @ 0 SetWindowModified drop
				THEN
				theAliasH @ to.heap		\ alias handle no longer needed
				R> window				\ restore grafport
			THEN
		THEN
((	this method allows menu to display from the mach-o app inside the
	 bundle but also uses macho-o app icon
		MyRWResFile# FRef->Alias ?dup
		IF
			get.window	>R		\ preserve grafport around this call
			sys.window @ over SetWindowProxyAlias 0=
			IF
				sys.window @ 0 SetWindowModified drop
			THEN
			to.heap				\ alias handle no longer needed
			R> window			\ restore grafport
		THEN
 ))
	THEN
;

restorer linklasttoken		\ done after adding sys.window


\ *********************************************************

extern: UInt32 OSStatus ICStart ( UInt32 * ICInstance, UInt32 OSType );
extern: UInt32 OSStatus ICStop  ( UInt32 ICInstance );
extern: UInt32 OSStatus ICLaunchURL  ( UInt32 ICInstance, UInt32 * hint$
					UInt32 * data, UInt32 len, UInt32 * ^selstart, UInt32 * ^selend );


(*
	EXTERN_API( OSStatus )
	ICLaunchURL						(ICInstance 			inst,
									 ConstStr255Param 		hint,
									 const void *			data,
									 long 					len,
									 long *					selStart,
									 long *					selEnd);
	   * Parses a URL out of the specified text and feeds it off to the
	   * appropriate helper.
	   * hint indicates the default scheme for URLs of the form "name@address".
	   * If hint is the empty string then URLs of that form are not allowed.
	   * data points to the start of the text. It must not be nil.
	   * len indicates the length of the text. It must be non-negative.
	   * selStart and selEnd should be passed in as the current selection of
	   * the text. This selection is given in the same manner as TextEdit,
	   * ie if selStart = selEnd then there is no selection only an insertion
	   * point. Also selStart  selEnd and 0  selStart  len and 0  selEnd  len.
	   * selStart and selEnd are returned as the bounds of the URL. If the
	   * routine returns an error then these new boundaries may be
	   * invalid but they will be close.
	   * The URL is parsed out of the text and passed off to the appropriate
	   * helper using the GURL AppleEvent.
*)



create myICInstance	0 ,
create ICStartSel	0 ,
create ICEndSel		0 ,


: myCreator ( -- creator )		\ get processSignature from MyProcessInfo
	MyProcessInfo 20 + @		\ which was read at launch time
;


: (LaunchURL) ( addr\cnt -- result )
	myICInstance myCreator ICStart ?dup 0=
	IF
		myICInstance @ 
		"" 							\ no hint
		2swap						\ instance\hint\addr\cnt
		ICStartSel off
		dup ICEndSel !				\ use all the text
		ICStartSel ICEndSel
		ICLaunchURL					\ pass result code back

		myICInstance @ ICStop drop	\ an error ONLY here would be suprising
	ELSE
		nip nip
	THEN
;


: LaunchURL ( str -- result )	\ takes counted string
	count (LaunchURL) ;


: URL: ( "<spaces>name"   -- )
	bl word LaunchURL abort" LaunchURL Error" ;


: URL" ( -- OSStatus )
	[compile] C"
	compiling
	IF
		POSTPONE LaunchURL
	ELSE
		LaunchURL
	THEN ; IMMEDIATE



\ *********************************************************
\ Uses POSIX calls - need POSIX path descriptor
\ can obtain from words like 
\	FSSpec->ZString   FileInfo->ZString   CFUrl->ZString


: FileContents->Handle ( c$path -- handle |or| 0 )
\ uses Posix file operations
	{  | fileid size handle }
	count  EXPAND r/o Open-File 0=
	IF
		dup -> fileid
		    file-size 2drop
		dup -> size
		from.heap -> handle
		0. fileid Reposition-File drop
		handle @  size fileid Read-File 2drop
		fileid close-file drop
		handle
	ELSE
		drop
		0
	THEN
;


\ *********************************************************
\ MacForth Date&Time words

>m4th
decimal


       60 60 * CONSTANT 1Hour		\ seconds in 1 hour
    1Hour 12 * CONSTANT 12Hours		\ seconds in 12 hours
  12Hours  2 * CONSTANT 1Day		\ seconds in 1 day


: Days>YMD  ( #days -- yr\mo\dy )  \ #days since 1Jan1904
	\ Range of correct results:  01Mar1900 .. 28Feb2100
	\ Range of valid days: -1401 .. 71647
	1401 +  4 * 3 + 1461 /mod  1900 +
	swap 4 / 1 +
	5 *
	3 - 153 /mod         ( year\remainder\quotient )
	swap 5 / 1 + swap
	3 + dup 12 > if
		12 - rot 1 +  rot rot
	then
	swap  ;

(( 

Use VFX TIME&DATE ( -- seconds mins hours day month year ) for local time
or   SYSTIME&DATE ( -- seconds mins hours day month year ) for UTC/GMT time

: TIME&DATE  ( -- sec min hr da mo year )
    \ contributed by Wil Baden
	\ allows easy ISO standard date output: yyyy-mm-dd
	\ see: http://www.iso.org/iso/en/prods-services/popstds/datesandtime.html
	@Clock 0 1Day  um/mod >r  60 /mod  60 /mod  R> Days>YMD swap rot
;

 ))

: YMD>Days  ( yr\mo\dy -- #days )  \ #days since 1Jan1904
    \ Range of valid dates:  01Mar1900 .. 28Feb2100
    \ Range of valid days: -1401 .. 71647
    \ improved version contributed by Wil Baden
    >r  dup 2 > if
        3 - swap
    else
        9 + swap 1-
    then

    DUP 100 < IF
        time&date  nip nip  nip nip  nip dup  100 mod - +
    THEN

    1900 - 1461 4 */
    swap  153 * 2 + 5 / +
    r> +  1402 -  ;


: HMS>Secs  ( H\M\S -- #secs )  \ #secs since midnite
	rot   60 *		\ convert hours to minutes
	rot + 60 *		\ sum all minutes & convert to secs
	+ ;				\ sum seconds

: Secs>HMS  ( #secs  -- H\M\S )  \ #secs since midnite
	0 60 um/mod		\ S: seconds\quotient
	0 60 um/mod		\ S: second\minutes\hours
	swap rot  ;


: ?DAYS  ( ClockTime -- #days )  \ convert seconds to days
	\ ClockTime is #seconds since:  01Jan1904 midnite
	1 needed  0 1day um/mod  nip   ;


: ?SECONDS  ( ClockTime -- #seconds )  \ #seconds since midnight
	\ ClockTime is #seconds since:  01Jan1904 midnite
	1 needed  0 1day um/mod  drop  ;


: DAYS>  ( #days -- year\day\month )  \ #days since 1Jan1904
	Days>YMD  swap  ;



ascii /  value DateDelim
ascii :  value TimeDelim


: ?HOLD	( char -- |  hold char in pad  if non-0 )
	?dup IF hold THEN ;


\ U.S. format (taditional MacForth), sorry Stephen
\ MM/DD/YY
: FMT.DATE$  ( #DAYS\flag -- addr\cnt )
	BASE @ >R   DECIMAL
	>R   DAYS> 100 * + 100 * SWAP ( 1900 - ) 100 mod +	\ crb
	S>D <# # # R@ IF DateDelim ?hold THEN # #
			   R> IF DateDelim ?hold THEN # # #>
	R> BASE ! ;


: GET.DATE$  ( addr -- | date string is 8 chars long )
	@clock ?days  -1 fmt.date$  rot swap  cmove  ;


: .DATE$  ( --- )
	@clock  ?days   -1 fmt.date$  type  ;


: (:00)  ( d -- | set up for time output )
	#  6 base !  #  decimal  TimeDelim ?hold  ;


variable 12H?	
	true 12H? !  \ change this to FALSE to display in military (24 hour) form


: FMT.TIME$  ( #seconds -- addr\cnt )
	base @ >r  decimal
	s>d <#  
		d>s
		12h? @
		IF	ascii M hold
			dup  12hours >=
			IF   12hours -
				  ascii P  
			ELSE  ascii A
			THEN  
			hold bl hold
			dup 1hour < if 12hours + then
		THEN
		s>d (:00)   (:00)   # # #>
	R> base !  ;


: GET.TIME$  ( addr -- | time string is 8 or 11 chars long, depending on 12H? )
	@clock ?seconds fmt.time$  rot swap cmove  ;


: .TIME$  ( -- )
	@clock ?seconds  fmt.time$  type  ;





StopIncluding ************************************

