\ Roelf Toxopeus
\ Last Revision: 18 Jan 2012  15:14:43  ward@megawolf.com

\ Using some MF search facilities to LEAP through a file.
\ Tweaked some words from MacForth's EDITOR file:
\ Whatever.needed.SE.word   rewritten as   myWhatever.needed.SE.word

\ Implements a Jeff Raskin like LEAP action search incrementally through a file:
\ fast mouseless file navigation, for edit-, word reference-, examining- and other-purposes.
\ It's safe: it doesn't change the file!

\ Intermediate test vesion, current state (doesn't mean it all works flawless):

\ Features:
\ -- Incremental search through top se window, without dialog window:
\    highliting in se.window is your feedback.
\ -- Leap is NOT case sensitive.
\ -- Automatic wrapping, use CONTROL key to reverse direction.
\ -- Immediate notice of no match: the 'suggested' not matching char won't show!
\ -- Changed back- and fore-color indicates leap search.
\ -- Internally used WRAP.SEARCH and -WRAP.SEARCH can be used with normal SE searching as well!
\    Don't need the LEAP stuff to use them. They set the Editors FOUND? hook.
\ -- CONTROL+SPACEBAR toggles between normal EDIT mode and LEAP mode (sic).

\ Usage:
\ -- hit CONTROL+SPACEBAR --> current cursor- or selection-start-position is leap-start position.
\ -- type target char --> immediate leaps to target.
\ -- optional add char(s) to target --> leaps to new target.
\ -- when an added char creates an unfindable searchstring, no further action is taken.
\    the not matching char is not added, last target stays put, as if nothing happened.
\ -- optional hit BS to del rightmost char --> leaps back to last point a char was added succesfully,
\    eventually to the leap-start-position.
\ -- optional hit RETURN/ENTER or keep RETURN/ENTER pressed --> leaps to next occurance target,
\    keeps jumping through the file.
\ -- press/hit CONTROL + others, to leap backwards.
\ -- press CONTROL + hit SPACEBAR twice release CONTROL --> last found target is new leap-start-position
\    with a fresh(0) searchstring.
\ -- hit CONTROL+SPACEBAR once when done --> EDIT mode.
\ -- use ARROW keys to set insertion point or use modified ARROW keys to extend
\    selection.

\ Note:
\ -- after returning in EDIT mode, every next (non-arrow)key will overwrite the selection!
\ -- the 'LEAPed to' selection is also shown in the MF search dialog when you open it. So any further
\ -- action could be taken there. 

\ Customize:
\ -- colors in  LEAP.COLORS  SE.COLORS
\ -- 'hot' keys in  DO.DIRECTION  DO.LEAP.KEY  LEAP?

\ Caveats, wanted fixing:
\ -- difficult to find the right 'hot' keys: SHIFT would be nicest, but SHIFTed chars are searchable.
\    Same for OPTION chars. CONTROL will do for now.
\ -- possibly leap through folders, disks, internet, where not?
\ -- whatever comes up.

\ Added support for colorized editor: turned off while leaping.
\ Coloring turned on when done leaping.
\ Last Revision: 2009-Feb-17 19:39:39   -rt


only Forth definitions

anew --leap--

>M4TH
decimal

also editorvocab definitions



\ RGBMagenta @RGB RGBColor LeapForeColor
 RGBRed		 @RGB RGBColor LeapForeColor
 RGBLtGray   @RGB RGBColor LeapBackColor
 
\ ------ Wrapping -------------------------------------------

\ Automaticly wraps around.
\ Does stop searching when target is not found after a research from top.

variable LastSelStart

: SEARCH$>DEL ( -- )
	search$ c@ 1- 0 max search$ c! ;


: STAY.PUT ( -- )
	search$>del						\ remove added character causing distress
	lastselstart @ selstart !		\ stay put
	lastjumpedfrom @ selend !
	found.it		\ myfound.it
;

: COMPLEMENT.SEARCH ( -- )
	selstart @ LastSelStart !
	selend @  LastJumpedFrom !
	BackFind?
		IF from.end ELSE from.top THEN 	
	re-search ;
		
		
: myNOT.FOUND ( -- )
	selend @  wasselstart <> 
	IF    complement.search  
	ELSE  beep stay.put
	THEN ;


: (myFOUND?) ( flag -- )
	IF   found.it			\ myfound.it 
	ELSE mynot.found 
	THEN 
;


\ save current FOUND? behavior and restore it when done.

variable SAVED.FOUND?

: WRAP.SEARCH ( -- )
	action-of found?  
	  dup ['] (myFOUND?) =
	IF
		drop
	ELSE
		saved.found? !
		['] (myFOUND?) is found?
	THEN
;

: FOUND.SEARCH ( -- )
	saved.found? @ ?dup IF is FOUND? THEN ;



\ ------ Found Leap History --------------------------------------

256 ( search$-length ) cells 2* constant /found

create FOUND /found allot&erase

variable FOUND#


: FOUND+ ( i -- 'found )
	found# @ + 255 and dup found# !
	cells 2* found + ;

	
: BUMP.FOUND ( n1 n2 -- )
	1 found+ 2! ;


: PREV.FOUND ( -- n1 n2 )
	-1 found+ 2@ ;


\ ------ Leap engine ----------------------------------------

\ wjm : add>search$ ( char -- )   search$ count + c! 1 search$ c+! ;

: ADD>SEARCH$ ( char -- )
	search$ count + c!   1 search$ c@ + search$ c! ;


: myMAKE.SEARCH ( char -- found flag )
	add>search$ (re-search) ;


: DO.OTHERS	( char -- )							\ default action: leap
	selend @ swap
	mymake.search found?
	selend @ <>
	IF
		selstart @ selend @ bump.found
	THEN ;


: DO.BS  ( -- )
	search$ c@ 
	if
		search$>del
		prev.found selend ! selstart !
		found.it		\ myfound.it
	then ;


: DO.RET ( -- )					\ leap to next current selection
	search$ c@ if  (re-search) found? then ;


: DO.DIRECTION ( -- )		 	\ CONTROL key sets leap direction
	?CntrlKey ( if true else false then ) -> BackFind? ;


: DO.LEAP.KEY	( char -- )   				( myDO.SEARCH.KEY )
	do.direction
	dup 13 = 
	over 3 = or
	if drop do.ret exit then
	dup 8 = if drop do.bs exit then
\	dup bs = if drop do.bs exit then
	do.others ;


: LEAP? ( char -- flag )   
	32 = 
	?CntrlKey and invert ;			\ CONTROL+SPACEBAR exits LEAP mode


: BLINK.KEY ( -- key ) 				\ takes care of blinking cursor
	BEGIN
		do.se.events drop
		?keystroke
	UNTIL
;


: (LEAP) ( -- )
	WholeWordFlag >R		\ 10.47 Doug Hoffman fix
	false to WholeWordFlag
	BEGIN
		blink.key dup leap?
	WHILE
		do.leap.key
	REPEAT
	drop  
	R> to WholeWordFlag
;


\ ------ setup cosmetics-------------------------------------

create PreLeapForeColor  6 allot
create PreLeapBackColor  6 allot


: FRESH.START ( -- )
	search$ off										\ for now, start clean
	found# off
	hide.range selstart @ selend !
	selstart @ LastSelStart !
	selstart @ selend @ 0 found+ 2! ;


: LEAP.COLORS ( -- )
	PreLeapForeColor  GetForeColor	\  @rgb WasForecolor !rgb
	PreLeapBackColor  GetBackColor 	\  @rgb WasBackcolor !rgb
	LeapForeColor RGBForeColor
	LeapBackColor RGBBackColor
	resize.se ;


: SE.COLORS ( -- )
	PreLeapForeColor RGBForeColor
	PreLeapBackColor RGBBackColor
\	SE.ForeColor RGBForeColor
\	SE.Backcolor RGBBackColor
	resize.se ;


\ ------ setup, run and restore------------------------------


variable wascolorflag


: UNCOLORED ( -- )			\ force uncolored
	wascolorflag @
	IF
		false +miscinfo +SE.ColorFlag !
		coord.bounds _InvalRect
		resize.se
	THEN
	['] noop is ColorRectClicked
;
	
: RestoreColored ( -- )
	wascolorflag @ ?dup
	IF
		+miscinfo +SE.ColorFlag !
		coord.bounds _InvalRect
		resize.se
	THEN
	['] ToggleColorMode is ColorRectClicked ;


: HELLO.LEAP ( -- )
	leaping ON
	+miscinfo +SE.ColorFlag @ wascolorflag !
	uncolored
	leap.colors
	fresh.start
	wrap.search ;


: BYE.LEAP ( -- )
	leaping @
	IF
		found.search
		se.colors
		RestoreColored
		leaping OFF
	THEN
;


: LEAP ( -- )
	se.window front.window
	IF
		hello.leap
		(leap)
		bye.leap
	ELSE
		beep
	THEN
;


\ ------ prepare for an Editor key hook or routine ----------


: SE.LEAP  ( char -- 'char )		\ CONTROL+SPACEBAR means LEAP
	dup
	32 = ?CntrlKey and
	if
		drop leap 0 
	then ;


: (StopLeaping) ( -- )
	bye.leap
;

' (StopLeaping) is StopLeaping


: ADD.LEAP ( -- )
	['] se.leap is EditorKeyFilter
;


: REMOVE.LEAP ( -- )
	StopLeaping 
	['] NOOP is EditorKeyFilter
;



?building 0= [IF] add.leap [THEN]

>m4th



StopIncluding

\ test
leap
remove.leap
