\ Mandelbrot task demo
\ Last Revision: 08/10/04  02:43:36 PM  wjm
\ Last Revision: 19 Jan 2012  07:32:06  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


((
The guts of this is "8 Line Mandelbrot"
	url: http://herkamire.com/jason/8_line_mandelbrot

from Jason Woofenden' s Forth Wiki
	url: http://jasonwoof.org//8_line_mandelbrot 

The original was not intended to be readable or well-factored, but "was written
to show that forth can create a mandelbrot in 8 lines"

I reformatted it and modified it only very slightly to use a graphic rather than
text display and throw in a few PAUSEs.

This is the original (all 8 lines of it):

fvariable ci fvariable c fvariable zi fvariable z
: >2? z f@ fdup f* zi f@ fdup f* f+ 4.0e f> ;
: nextr z f@ fdup f* zi f@ fdup f* f- c f@ f+ ;
: nexti z f@ zi f@ f* 2.0e f* ci f@ f+ ;
: pixel c f! ci f! 0e z f! 0e zi f! 150 50 do nextr nexti zi f! z f! >2? if i unloop exit then loop 0 ;
: left->right -1.5e 400 0 do fover fover pixel emit 0.005e f+ loop fdrop ;
: top->bottom -1e 400 0 do left->right 0.005e f+ loop fdrop ;
.( P5) cr .( 400 400) cr .( 255) cr top->bottom bye
))

only Forth definitions

anew --mandelbrot--

>M4TH

decimal


0 value Xpos						\ location of next point to draw
0 value Ypos						\ location of next point to draw
defer DrawPixel ( ColorIndex -- )	\ draw pixel at (Xpos,Ypos), ColorIndex 0-255

\ ---------------------------------------------------------
\ This is a reformatted and very slightly modified "8 Line Mandelbrot" (see below)

fvariable ci    fvariable c    fvariable zi    fvariable z


: >2? ( -- flag )
	z  f@ fdup f* 
	zi f@ fdup f* 
	f+ 				\ Z^2 + Zi^2
	4.0e
	f> ;

: nextr ( -- | F: -- f )
	z  f@ fdup f* 
	zi f@ fdup f* 
	f- 
	c f@  f+ ; 		\ ( Z^2 - Zi^2)+c

: nexti ( -- | F: -- f )
	z  f@ 
	zi f@ 
	f*   2.0e f*
	ci f@  f+ ; 		\ (2*Z*Zi)+c


: FigurePixelColor ( -- ColorIndex | F: f1\f2 -- )
	c f!   ci f! 
	0e z f! 0e zi f! 
	150 50 do
		nextr 
		nexti
		zi f! 
		z  f!
		>2? if  i unloop exit  then 
	loop
	0 ;


: LEFT->RIGHT ( -- | F: f1 -- f1 )	\ draw one line of image
	-1.5e 
	pause				\ pause with 2 items on the floating stack
	400 0 do 			\ for each column
		i -> xpos
		fover fover 
		FigurePixelColor DrawPixel
		0.005e f+ 
	loop 
	fdrop
;


: TOP->BOTTOM ( -- )	\ draw the whole image
	-1e
	400 0 do 			\ for each row
		i -> ypos
		left->right
		0.005e f+
\		pause			\ pause with 1 item on the floating stack
	loop
	fdrop ;


\ ---------------------------------------------------------
\ I needed a bunch of indexed colors, so for this quick and dirty demo
\ I borrowed a 'pltt' resource with 255 entries from another application
\	and added it to our Resource Bag

\ ?Library:  /Developer/SDKs/MacOSX10.5.sdk/System/Library/Frameworks/ApplicationServices.framework/ApplicationServices

Extern: UInt32 palettehandle GetNewPalette ( UInt32 resid   );	\ from a 'pltt' resource
Extern: void GetEntryColor ( UInt32 palettehandle, UInt32 Index, UInt32 ^rgbcolor   );

256 6 1array CTABLE				\ array of 255 RGBcolor s


: FillCTable ( -- )		\ copy 'pltt' to our CTABLE array
	1000 GetNewPalette ?dup 0= abort" Missing 'pltt' 1000 resource"
	locals| thePaletteH |
	
	256 0
	DO
		thePaletteH   i   i CTABLE  GetEntryColor
	LOOP
	thePaletteH to.heap
;

FillCTable
forget FillCTable	\ we don't need FillCTable word any more



0 0 0 RGBColor RGBScratch

: SwapEntries ( i\j -- )
	CTABLE 	dup >R
	@rgb RGBScratch !rgb		\ scratch = entry(j)
	CTABLE
	dup @rgb  R> !rgb
	>r
	RGBScratch @rgb r> !rgb
;

: ScrambleCTable ( -- )	\ randomly rearrange colors in table
	256 0
	DO
		random abs		\ 6.05436 - avoid negatives
		255 mod
		i
		SwapEntries
	LOOP
;

0 value #pix

: (DrawPixel) ( colorindex -- )
	\ demo would likely be faster if we draw it offscreen and _CopyBits it back
	\ rather than doing it one pixel at a time	
	Xpos Ypos  rot CTable  SetCPixel    ( x\y\^color -- )
	1 +to #pix	
\	pause			\ enable this for better time sharing
					\ but slower mandelbrot drawing
;

' (DrawPixel) is DrawPixel



: Invert16 ( addr -- )
	dup w@s
	$ffff swap - 
	swap w!
;

: DrawOne ( -- )
	0 -> #pix
	(ticks)	\ _tickcount
	ScrambleCTable
	pause
	top->bottom
	(ticks) swap -		\ #ticks it took

	200 200 RGBScratch GetCPixel	\ center should be all one big "bulb"
	\ now roughly invert the color to show the text
	RGBScratch     Invert16
	RGBScratch 2 + Invert16
	RGBScratch 4 + Invert16
	RGBScratch	RGBForeColor
	get.window flushwindow
	156 206 move.to
\	156 196 move.to
		#pix . ." pixels in "
		3 .R		\ s>d <# # ascii . hold #s #> type
		."  millisec"
\	206 216 move.to
\		." click to repeat"
;

\ ---------------------------------------------------------
\ Display the result in its own window

new.window MandelbrotWindow
300 320 2dup 400 400 +point MandelbrotWindow w.bounds
kWindowCollapseBoxAttribute MandelbrotWindow NewWAttributes
close.box MandelbrotWindow w.attributes
c" Mandelbrot" MandelbrotWindow w.title


: UpdateMandelbrotWindow ( -- )
	MandelbrotWindow window
	MandelbrotWindow outfile !
	DrawOne
;
MandelbrotWindow on.update UpdateMandelbrotWindow

: ActivateMandelbrotWindow ( flag -- )
	IF
		MandelbrotWindow window
		MandelbrotWindow outfile !
		screen.bounds invalrect	\ force an update
		init.cursor
		BEGIN
			do.events drop
			screen.bounds invalrect		\ repeatedly force an update
((			mouse.down =
			key.stroke @ OR
			IF
				screen.bounds invalrect	\ force an update
				key.stroke off
			THEN
))		AGAIN
	THEN
;

MandelbrotWindow on.activate ActivateMandelbrotWindow


MandelbrotWindow add.window

