\ PI calculating program                              17sep87mdp
\ converted to f83 by mark penacho  - Genie mail address M.PENACHO
\ Anyone remember the Forth Roundtable on Genie??
\ Ported to MacForth  by Ward McFarland
\ Last Revision: 11/05/05  05:15:00 AM  wjm
\ Last Revision: 12 Dec 2011  15:54:00  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


\ Substantially reformatted by wjm for legibility
\ The original algorithm was not changed except for removing a few unnecessary 
\ temporary variables for holding borrows and carries and the stack used instead )

((

a couple of examples at the end of this file
or simply open a new editor window
type 
5000 PI
and hit ENTER

))

only Forth definitions
decimal

anew --pi--

variable PSIZE                \ contains the number of digits of precision

\ the next three words define byte arrays - each byte contains a decimal digit
\ this permits multi-digit precision

create POWER  10240 allot
create TERM   10240 allot
create RESULT 10240 allot


\ some temporary variables, pointers and flags...

variable PPOINT             \ Pointer to array being divided
variable DIVISOR
variable ZERO               \ Zero=0 only when all digits of PPoint =0
variable EXP
variable PSIGN


\ A few simple implementations of Multi-digit precision decimal arithmetic...

: DIV ( -- | divide array pointed to by PPOINT by DIVISOR )
    0 zero !
    0                                \ put prior remainder on the stack
    ppoint @ dup  psize @ + 1+  swap
    DO  10 *   i c@  +               \ add prior remainder to next digit
        divisor @   /mod
        2dup +   zero @ or   zero !  \ Zero=0 only when all digits of PPoint =0
        i c!                         \ save quotient, leave remainder on stack
    LOOP
    drop ;


: ADD ( -- | add TERM array to RESULT array --> RESULT )
    0                                    \ put prior carry on the stack
    0   psize @
    DO  result i + c@   term i + c@  +   \ add each pair of digits
        +                                \ add carry
        dup  10 <                        \ generate new carry?
        IF  0  ELSE -10 + 1 THEN
        swap result i + c!               \ save result, leave carry on stack
    -1 +LOOP
    drop ;


: SUB ( -- | subtract TERM array from RESULT array --> RESULT )
    0                                    \ put prior borrow on the stack
    0   psize @
    DO  result i + c@    term i + c@ -   \ subtract each pair of digits
        swap -                           \ subtract borrow
        dup  0<                          \ generate new borrow?
        IF 10 + 1 ELSE 0 THEN
        swap result i + c!               \ save result, leave borrow on stack
    -1 +LOOP
    drop ;


\ Now, the algorithm itself...

: INIT ( pass# -- | initialize before each pass through P-CALC )
    power psize @ 1+ erase             \ clear POWER array
    term  psize @ 1+ erase             \ clear TERM array
    3 over 2* - psign !                \ PSIGN = +1 on pass 1,  -1 on pass 2
    1 =
    IF   result psize @ 1+ erase       \ clear RESULT array on pass 1
         16   5                        \ divisor =   5   power(0) = 16
    ELSE  4 239                        \ divisor = 239   power(0) =  4
    THEN
    divisor !  power c!
    power ppoint !
    div                                \ divide POWER array by DIVISOR
    1 exp ! ;


: P-PRINT ( -- | print RESULT array )
	cr
    result c@ [char] 0 + emit              \ print first digit
    [char] . emit                          \ print decimal point
    result psize @ + 1+ result 1+
    DO   i c@   [char] 0 +   emit          \ print remainder of array
    	 out @ 80 >= IF cr THEN
    LOOP ;


: P-CALC ( n -- | compute PI to N digits - leave answer in RESULT array )
    psize !
    3 1                                     \ make two Passes over the array
    DO  i init                              \ initialize for each Pass
        BEGIN
            power term psize @ 1+ move       \ copy POWER array to TERM array
            term ppoint !   exp @ divisor !
            div                             \ divide TERM array by EXP
            psign @ 0>  				    \ add/subtract TERM and RESULT arrays
            IF add  ELSE sub  THEN
            psign @ negate psign !          \ toggle PSIGN
            2 exp +!                        \ increment EXP by 2
            i 1 =  IF 25 ELSE 239 THEN
            divisor !    power ppoint !
            div                             \ divide POWER array by 25 or 239
            i 2 = IF div THEN               \ do it twice on Pass 2
            zero @  0=                      \ continue until POWER all zeroes
        UNTIL
    LOOP  ;


: PI ( n -- | compute & print PI to N places )
	10240 min
    p-calc
	p-print ;


\ print reformatted for the Great Language Shootout
: p-print2
	result psize @ over + swap
	DO
		10 0
		DO
			j i + c@ [char] 0 + emit
		LOOP
		4 spaces [char] : emit i result - 10 + . cr
	10 +LOOP
;

: PI2 ( n -- | compute & print PI to N places )
	10240 min
    p-calc
	p-print2 ;


StopIncluding ************************************

iMac
ticks 5000 p-calc ticks swap - cr . ."  msec"

520  msec





time for the algorithm alone on creaky old Mac G4/450

time.for 5000 p-calc

9.51 sec
9.51 sec
9.45 sec
9.46 sec
9.47 sec





1000 pi2
3141592653    :10 
5897932384    :20 
6264338327    :30 
9502884197    :40 
1693993751    :50 
0582097494    :60 
4592307816    :70 
4062862089    :80 
9862803482    :90 
5342117067    :100 
9821480865    :110 
1328230664    :120 
7093844609    :130 
5505822317    :140 
2535940812    :150 
8481117450    :160 
2841027019    :170 
3852110555    :180 
9644622948    :190 
9549303819    :200 
6442881097    :210 
5665933446    :220 
1284756482    :230 
3378678316    :240 
5271201909    :250 
1456485669    :260 
2346034861    :270 
0454326648    :280 
2133936072    :290 
6024914127    :300 
3724587006    :310 
6063155881    :320 
7488152092    :330 
0962829254    :340 
0917153643    :350 
6789259036    :360 
0011330530    :370 
5488204665    :380 
2138414695    :390 
1941511609    :400 
4330572703    :410 
6575959195    :420 
3092186117    :430 
3819326117    :440 
9310511854    :450 
8074462379    :460 
9627495673    :470 
5188575272    :480 
4891227938    :490 
1830119491    :500 
2983367336    :510 
2440656643    :520 
0860213949    :530 
4639522473    :540 
7190702179    :550 
8609437027    :560 
7053921717    :570 
6293176752    :580 
3846748184    :590 
6766940513    :600 
2000568127    :610 
1452635608    :620 
2778577134    :630 
2757789609    :640 
1736371787    :650 
2146844090    :660 
1224953430    :670 
1465495853    :680 
7105079227    :690 
9689258923    :700 
5420199561    :710 
1212902196    :720 
0864034418    :730 
1598136297    :740 
7477130996    :750 
0518707211    :760 
3499999983    :770 
7297804995    :780 
1059731732    :790 
8160963185    :800 
9502445945    :810 
5346908302    :820 
6425223082    :830 
5334468503    :840 
5261931188    :850 
1710100031    :860 
3783875288    :870 
6587533208    :880 
3814206171    :890 
7766914730    :900 
3598253490    :910 
4287554687    :920 
3115956286    :930 
3882353787    :940 
5937519577    :950 
8185778053    :960 
2171226806    :970 
6130019278    :980 
7661119590    :990 
9216420199    :1000 

deviates from the results on 
http://shootout.alioth.debian.org/iofile.php?test=pidigits&lang=all&sort=fullcpu&file=output
in digit 1000 (9 instead of 8)
but if we run 1010 pi2, the last 2 lines are:
9216420198    :1000 
9380952573    :1010 

indicating this method coprrectly rounded the 100th digit



The original Genie posting:

\ PI calculating program                              17sep87mdp

converted to f83 by mark penacho  - Genie mail address M.PENACHO

adopted from a program written in integer basic on an apple ][
that appeared in some very old computer magazine/book
I'm not sure of the original author of the integer basic
program, or what book it appeared in, or even when it appeared.
But he deserves credit.

the program calculates pi to about as many digits as you will
wait for.  on my 10Mhz AT, 2000 digits were done overnight, and
a couple hundred digits seem to finish almost immediately

uses a taylor series expansion of some inverse trig equation,
but I've not really analyzed the algorithm very much

WRITTEN IN FORTH BECAUSE IT'S FORTH

variable names and procedure names were taken almost
verbatim from the integer basic counterpart, except in cases to
avoid conflict with forth standard words (copy, sign, etc.)

much optimization could still be done to increase the speed

to adapt it to other forths, f83 behaves in these ways

CODE                        PRODUCES
====                        ========
: 5 0 DO I . LOOP ;         0 1 2 3 4

: 0 5 DO I . -1 +LOOP ;     5 4 3 2 1 0

so if loops are different on another forth, make changes
accordingly


[THEN]