\ Run As Kiosk
\ Last Revision: 01/14/05  04:58:58 PM  wjm
\ Last Revision: 12 Jan 2012  07:03:55  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


\ Allow MacForth to run in "Kiosk Mode".
\ Many possible variations, from totally owning the OS X interface, to
\	allowing the user some control over various 
\ See Apple Tech Note 2062 and the header file docs below

\ To see some of the possibilites, execute:
\ WeOwn
\ WeMostlyOwn
\ WeAllowSome
\ JustUs
\ To reset things to normal, execute:
\ WeAllow

forth definitions
decimal

anew --SetSystemUIMode--

>m4th


variable ShowBackgroundPict	ShowBackgroundPict ON

extern: UInt32 OSStatus SetSystemUIMode ( UInt32 SystemUIMode, UInt32 SystemUIOptions );


new.window BigWindow		\ a window to cover up the Finder
13 BigWindow w.type
full.screen.bounds @rect BigWindow w.bounds		\ make it huge



: BigWindowUpdate ( -- )
\	RGBYellow RGBBackColor
	RGBLtGray RGBBackColor
	400 GetPicture
	ShowBackgroundPict @ over AND
	IF
		full.screen.bounds DrawPicture  ( pictHandle\rectPtr -- )
	ELSE
		drop
		full.screen.bounds EraseRect
	THEN
;
BigWindow on.update BigWindowUpdate



: SpecialSwitchEventHandler ( -- )	\ optional
	\ we get this when we receive a switch event, 
	\ prior to changing SWITCHED.OUT? state
	BigWindow @			\ has BigWindow been ADDed ?
	IF
		BigWindow
		switched.out? @ 0=
		IF
			hide.window
		ELSE
			show.window
			allwindowsfront
		THEN
	THEN
;


: OverrideSomeEvents ( eventCode1 -- eventCode2 )	\ was IgnoreBigWindow 
	\ we do not want BigWindow activated by mistake,
	\ and we do not want BigWindow visible when we are switched out
	dup 
	mouse.down =
	IF	\ ignore mouse clicks in BigWindow
		event.record 10 + @		\ mouse location
		find.window  ( point -- WindowRef\where code )
		drop
		dup				\ window ref
		BigWindow @ =
		AND				\ check for non-0
		IF				\ click was in BigWindow
			drop		\ drop the eventCode 
			0			\ replace with 0
		THEN
	ELSE
		dup switch.event =
		IF
			SpecialSwitchEventHandler
		THEN
	THEN
\	false				\ Eventer STOP? flag == 0
;
Eventer linklasttoken



: BigWindowActivate ( flag -- )
	IF
		BigWindow @ GetNextWindow		\ 6.052 - get next front window
		BringToFront 					\ 6.052 - force it front
		BigWindow @ 0 SendBehind		\ send BigWindow behind everything
		event.loop
	THEN
;
BigWindow on.activate BigWindowActivate



: WeOwn ( -- )		\ Pretty much "stuck" in MacForth
					\ menus hidden, but cmd-key equivalents still work
	AllWindowsFront
	3					\ kUIModeAllHidden
		8				\ kUIOptionDisableProcessSwitch
		16	+			\ kUIOptionDisableForceQuit
		SetSystemUIMode abort" _SetSystemUIMode error"		\ kUIModeAllHidden, kUIOptionDisableProcessSwitch
	BigWindow add.window
;


: WeMostlyOwn ( -- )	\ Like WEOWN but menubar becomes visible
						\ when mouse is moved over it. Apple Menu disabled
	AllWindowsFront
	3					\ kUIModeAllHidden
		8				\ kUIOptionDisableProcessSwitch
		1 + 			\ kUIOptionAutoShowMenuBar
		4 + 			\ kUIOptionDisableAppleMenu
	   64 +				\ kUIOptionDisableHide
		SetSystemUIMode abort" _SetSystemUIMode error"		
	BigWindow add.window
;


: WeAllowSome ( -- )	\ Allow CMD-TAB to switch apps
						\ Dock & menubar appear when mouse positioned over them
	AllWindowsFront
	4 					\ kUIModeAllSuppressed
		0 
		SetSystemUIMode abort" _SetSystemUIMode error"
	BigWindow add.window
; 

: JustUs ( -- )			\ Allow CMD-TAB to switch apps
						\ Dock & menubar appear when mouse positioned over them
	AllWindowsFront
	1 					\ kUIModeContentSuppressed
		0 
		SetSystemUIMode abort" _SetSystemUIMode error"
	BigWindow add.window
; 

: WeAllow ( -- )		\ remove all special modes
	0 0 SetSystemUIMode abort" _SetSystemUIMode error"		\ default
	BigWindow ?wvisible IF BigWindow close.window THEN
;  on.forget weallow


cr ." See instructions at the start of the file"


StopIncluding ----------------------------------------------------------------------

0 0 SetSystemUIMode .s			\ default

4 0 SetSystemUIMode .s			\ kUIModeAllSuppressed

3 0 SetSystemUIMode .s			\ kUIModeAllHidden

3 1 SetSystemUIMode .s			\ kUIModeAllHidden, kUIOptionAutoShowMenuBar

3 8 SetSystemUIMode .s			\ kUIModeAllHidden, kUIOptionDisableProcessSwitch






From "System/Library/Frameworks/Carbon.framework/Versions/A/Frameworks/HIToolbox.framework/Versions/A/Headers/MacApplication.h"

/*
 *  Summary:
 *    Controls the presentation of system-provided user interface
 *    elements.
 *  
 *  Discussion:
 *    Note that the system UI mode is a per-process state. Switching
 *    from a process that is suppressing system UI to another process
 *    that is not will cause system UI elements to become visible
 *    automatically. Switching back to the first process will suppress
 *    system UI elements again.
 */
enum {

  /*
   * In this mode, all standard system UI elements are visible.
   */
  kUIModeNormal                 = 0,

  /*
   * In this mode, system UI elements which cover the "content area" of
   * the screen (the area other than the menubar) are hidden. However,
   * these elements may automatically show themselves in response to
   * mouse movements or other user activity; specifically, the Dock
   * will still show itself automatically when the mouse moves into the
   * Dock's auto-show region.
   */
  kUIModeContentSuppressed      = 1,

  /*
   * In this mode, system UI elements which cover the "content area" of
   * the screen (the area other than the menubar) are hidden. Unlike
   * kUIModeContentSuppressed, most UI elements will not automatically
   * show themselves in this mode.
   */
  kUIModeContentHidden          = 2,

  /*
   * In this mode, all system UI elements, including the menubar, are
   * hidden. However, these elements may automatically show themselves
   * in response to mouse movements or other user activity;
   * specifically, the Dock and menubar will still show themselves
   * automatically when the mouse moves into the Dock's auto-show
   * region. Available in Mac OS X 10.3 and later.
   */
  kUIModeAllSuppressed          = 4,

  /*
   * In this mode, all system UI elements, including the menubar, are
   * hidden. Most system UI elements will not automatically show
   * themselves in this mode. The application may request that the
   * menubar automatically show itself while in this mode by passing
   * the kUIOptionAutoShowMenuBar flag to SetSystemUIMode.
   */
  kUIModeAllHidden              = 3
};
typedef UInt32                          SystemUIMode;


/*
 *  Summary:
 *    Controls optional behavior of system-provided user interface
 *    elements.
 */
enum {

  /*
   * Requests that the menubar automatically show itself when the user
   * moves the mouse into the screen area that would ordinarily be
   * occupied by the menubar. Only valid with kUIModeAllHidden.
   */
  kUIOptionAutoShowMenuBar      = 1 << 0,

  /*
   * Disables all items in the Apple menu. Valid for all modes.
   */
  kUIOptionDisableAppleMenu     = 1 << 2,

  /*
   * The active application may not be changed while this process is
   * active. Currently disables the Command-Tab and Command-Shift-Tab
   * key sequences to switch the active process, and the global window
   * rotation key sequence selected by the user in the Keyboard
   * preference pane. SetFrontProcess may still be used to explicitly
   * switch the active process. Only valid with modes other than
   * kUIModeNormal.
   */
  kUIOptionDisableProcessSwitch = 1 << 3,

  /*
   * The Force Quit window may not be displayed while this process is
   * active. Currently disables the Command-Option-Escape key sequence
   * to open the Force Quit window and the Force Quit menu item in the
   * Apple menu. Only valid with modes other than kUIModeNormal.
   */
  kUIOptionDisableForceQuit     = 1 << 4,

  /*
   * The current login session may not be terminated while this process
   * is active. Currently disables the Power key and the Restart, Shut
   * Down, and Log Out menu items in the Apple menu. Only valid with
   * modes other than kUIModeNormal.
   */
  kUIOptionDisableSessionTerminate = 1 << 5,

  /*
   * The Hide menu item in the Application menu is disabled. Note that
   * this option does not prevent this application from being hidden if
   * Hide Others is selected in some other application. Available in
   * Mac OS X 10.3 and later.
   */
  kUIOptionDisableHide          = 1 << 6
};
typedef OptionBits                      SystemUIOptions;


/*
 *  SetSystemUIMode()
 *  
 *  Summary:
 *    Sets the presentation mode for system-provided user interface
 *    elements.
 *  
 *  Discussion:
 *    The presentation mode of an application determines which
 *    system-provided user interface elements are visible on thes
 *    screen. When the frontmost application changes its presentation
 *    mode, a kEventAppSystemUIModeChanged Carbon event is sent to all
 *    applications that have registered for the event. This event is
 *    also sent when an application is activated; it contains the newly
 *    active application's presentation mode.
 *  
 *  Mac OS X threading:
 *    Not thread safe
 *  
 *  Parameters:
 *    
 *    inMode:
 *      The new mode.
 *    
 *    inOptions:
 *      Options controlling how the new mode behaves.
 *  
 *  Result:
 *    An operating system result code.
 *  
 *  Availability:
 *    Mac OS X:         in version 10.2 and later in Carbon.framework
 *    CarbonLib:        not available in CarbonLib 1.x, is available on Mac OS X version 10.2 and later
 *    Non-Carbon CFM:   not available
 */


