\ Simple Nib Control Demo
\ Created:       04/14/05  11:50:39 AM  wjm
\ Last Revision: 04/23/05  10:28:02 AM  wjm
\ Last Revision: 13 Jan 2012  12:06:20  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


(* *************************************************************************

Try Nib WITH Standard Event handler installed
But we will watch for Activate Events using InstallWindowEventCatcher

Because there IS a Standard Event Handler, Carbon events for mouse
	clicks in the window ARE handled by the Mac, and therefore we
	do not have to bother with highliting parts or tracking the mouse.

Demonstrate how "smart" Binary Nib controls can be and shows how easily we
	can add handlers for the binary controls using  
	NewNibBinControl:   InstallNibControl  and   !NibControlAction

************************************************************************* *)

Only Forth Definitions
Decimal

anew --Simple_Nib_Control_Demo--

>m4th

13 constant kEventControlClick


\ First, create a regular MacForth window template:
NEW.WINDOW aNibWindow		


\ Now, we name our Nib's controls, and provide identifying information
\	so we can connect to them later...

1				\ control ID from IB
0				\ control SIG from IB
aNibWindow		\ the MacForth window referencing the Nib Window
NewNibBinControl: Ctrl1 ( ID\sig\wptr -- )

\ add kEventControlClick to the standard kEventControlHit
kEventControlClick kEventClassControl  AddNibControlEvent ( type\class\nibctrl -- )


2				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl2 ( ID\sig\wptr -- )

\ for some reason, grouped buttons seem to want this:
kEventControlClick kEventClassControl  AddNibControlEvent ( type\class\nibctrl -- )


3				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl3 ( ID\sig\wptr -- )


4				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl4 ( ID\sig\wptr -- )


5				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl5 ( ID\sig\wptr -- )

\ Ctrl5 IS A STATIC TEXT FIELD and does not trigger a kEventControlHit
\ So we must add kEventControlClick to its list of control events handled..
kEventControlClick kEventClassControl  AddNibControlEvent ( type\class\nibctrl -- )


6				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl6 ( ID\sig\wptr -- )

7				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl7 ( ID\sig\wptr -- )

8				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl8 ( ID\sig\wptr -- )

9				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl9 ( ID\sig\wptr -- )

10				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl10 ( ID\sig\wptr -- )

11				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl11 ( ID\sig\wptr -- )

12				\ control ID from IB
0				\ control SIG from IB
aNibWindow
NewNibBinControl: Ctrl12 ( ID\sig\wptr -- )



\ *************************************************************************
\ Add standard MacForth activate and update routines...

\ standard event handler seems to handle this OK on X
\ but not on OS9
: UpdateNibWindow ( -- )
	aNibWindow @ DrawControls
;

aNibWindow on.update UpdateNibWindow


: ActivateNibWindow
\ standard event handler handles activates, we "synthesize" an activate, though
	IF	beep
		BEGIN
			do.events 
			mouse.down =
			IF 
				beep	\ snick
				?IN.CONTROL  IF beep THEN
			 THEN
		AGAIN
	ELSE
((		CarbonEventLogging @
		IF
			logcr " ** old-style window DE-activate" log$
		THEN
))	THEN
;

aNibWindow on.activate ActivateNibWindow


\ *************************************************************************
\ Finally, the word that connects all the dots...



: TestMe ( -- )
	\ Locate our Nib file     new method  - 6.05436
	\ GetNibReference uses 
	z" testwindow" GetNibReference ( 0$ -- NibReference |or| 0 )
		dup 0= abort" GetNibReference failed"

(*	\ this method also works fine for more control over where to look for the Nib
	z" MacForth NIBbles" OpenLocalBundle dup 0= abort" Bundle not found!"
	z" testwindow" GetNibReferenceFromBundle
*)

	\ find the specific Nib within the Nib file and attach to aNibWindow
	\ this Nib DOES have standard handler installed
	z" A Test Window1" aNibWindow add.window.from.nib

	\ If we want a compositing window to work correctly on OS 8/9, we need this:
\	aNibWindow UndoCompositingForOS9

	\ this is the magic dust for us to receive Activate events
	\	even when a Standard Event Handler is installed:
	aNibWindow InstallWindowEventCatcher

\ "attach" our controls, so we get their CtrlRef's
\		and install a BinaryNibControlHandler
	Ctrl1  InstallNibControl
	Ctrl2  InstallNibControl
	Ctrl3  InstallNibControl
	Ctrl4  InstallNibControl
	Ctrl5  InstallNibControl
	Ctrl6  InstallNibControl
	Ctrl7  InstallNibControl
	Ctrl8  InstallNibControl
	Ctrl9  InstallNibControl
	Ctrl10 InstallNibControl
	Ctrl11 InstallNibControl
	Ctrl12 InstallNibControl

	aNibWindow show.window
;


having test test testme
\ cr ." Note: CarbonEventLogging flag is enabled"
cr ." Type TESTME or use CMD-\ to run demo"


\ *************************************************************************
\ Assign actions to some of the controls..
\ Doesn't matter where we add these (I am doing it late to make the main
\	portion of the code appear as simple as it is...

:noname ( EventRef\mfCtrl -- result )
	drop GetEventKind kEventControlHit =
	IF
		c" Note we get TWO events in the log, since we added Control Click "
		c" to the event list for this control."
		cr$
		c" Try moving the mouse out of the control before releasing button."
		400 dfltissuealert 		\ normal ISSUEALERT's are blocked in ProcPtr's
		drop
	THEN
	eventNotHandledErr
;	
Ctrl1 !NibControlAction

:noname ( EventRef\mfCtrl -- result )
	nip @ GetControl32BitValue  2 =
	IF
		c" Button Groups are a special type of control, the value "
		c" reflects the button pressed."
		"" ""
		400 dfltissuealert 		\ normal issuealert's are blocked in procptr's
		drop
	THEN
	eventNotHandledErr
;	
Ctrl2 !NibControlAction


' beep  Ctrl3 !NibControlAction


:noname ( EventRef\mfCtrl -- result )
	c" Popup controls can be handy, their value "
	c" reflects the item selected."
	"" ""
	400 dfltissuealert 		\ normal issuealert's are blocked in procptr's
	drop
	eventNotHandledErr
;	
Ctrl4 !NibControlAction


:noname ( EventRef\mfCtrl -- result )
	2drop
	c" This is like an old Dialog static text field."
	cr$ cr$
	c" But you can get control clicks in it."
	400 dfltissuealert 		\ normal ISSUEALERT's are blocked in ProcPtr's
	drop
	eventNotHandledErr
;
Ctrl5 !NibControlAction


:noname ( EventRef\mfCtrl -- result )
	2drop
	c" This obviously is not a binary control, but might be treated as one."
	cr$ ""
	c" The trick is to get the control part in the action routine."
	400 dfltissuealert 		\ normal ISSUEALERT's are blocked in ProcPtr's
	drop
	eventNotHandledErr
;
Ctrl6 !NibControlAction


:noname ( EventRef\mfCtrl -- result )
	2drop
	c" This obviously is not a binary control, but could be treated as one."
	cr$ ""
	c" The trick is to get the control part in the action routine."
	400 dfltissuealert 		\ normal ISSUEALERT's are blocked in ProcPtr's
	drop
	eventNotHandledErr
;
Ctrl7 !NibControlAction


:noname ( EventRef\mfCtrl -- result )
	2drop
	c" This is an Edit Text control and should not be used with NewNibBinControl:"
	cr$
	c" except to show that all UI elements in a Nib are considered controls."
	""
	400 dfltissuealert 		\ normal ISSUEALERT's are blocked in ProcPtr's
	drop
	eventNotHandledErr
;
Ctrl8 !NibControlAction


\ *************** comments and/or test code below ***************
StopIncluding

