\ Playing Sound Resources
\ Last Revision: 12 Dec 2011  16:32:25  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************

((

samples at the end of this file

))

only FORTH Definitions

anew --sndstuff--

>m4th

decimal



extern: UInt32 ioresult SndNewChannel ( UInt32 SndChannelPtr, UInt32 synthesizer, UInt32 init, UInt32 SndCallBackUPP );
extern: UInt32 ioresult SndDisposeChannel ( UInt32 SndChannelPtr, UInt32 quietNowflag );
extern: UInt32 ioresult SndPlay ( UInt32 SndChannelPtr, UInt32 SndListHandle, UInt32 asyncflag );
extern: UInt32 ioresult SndChannelStatus ( UInt32 SndChannelPtr, UInt32 theLength, UInt32 * ^SCStatus );



: NEW.SOUND.CHANNEL ( ChanPtr\type -- result )
	0				\ Long: init
	0				\ Long: ProcPtr
	SndNewChannel  ;


: DISPOSE.SOUND.CHANNEL ( ^channel\quitnow -- result )
	SndDisposeChannel  ;


structure SCStatus
	long:	+scStartTime				\ starting time for play from disk
	long:	+scEndTime					\ ending time for play from disk
	long:	+scCurrentTime				\ current time for play from disk
	byte:	+scChannelBusy				\ TRUE if channel is processing cmds
	byte:	+scChannelDisposed			\ reserved
	byte:	+scChannelPaused			\ TRUE if channel is paused
	byte:	+scUnused					\ unused
	long:	+scChannelAttributes		\ attributes of this channel
	long:	+scCPULoad					\ CPU load for this channel
structure.end


create aSCStatus    here SCStatus dup allot erase


: SndChannelStatus ( ChanPtr\^SCStatus -- result )
	SCStatus swap
	SndChannelStatus ( SndChannelPtr chan, short theLength, SCStatusPtr theStatus)
;



: (WaitSoundDone) ( chan -- )		\ stall until current sound commands complete
	locals| theChannel |
	theChannel
	IF
		BEGIN
			do.io.events  	 							\ but allow user to cancel out
			theChannel aSCStatus SndChannelStatus 0<>	\ exit if error
			aSCStatus +scChannelBusy c@ 0=
			OR
		UNTIL
	THEN ;


\ ------------------------------------------------------------------------------------------


variable MySSChan		MySSChan off	\ sampledSynth sound type

: ALLOCATE.SS.SOUND.CHANNEL ( -- )
	\ create a sound channel using sampled-sound data
	\ this channel can queue up to 128 SndCommands
	MySSChan 5 new.sound.channel abort" ALLOCATE.SS.SOUND.CHANNEL error!" ;
\ restorer linklasttoken			\  see RestoreSoundStuff
\ allocate.ss.sound.channel 		\  do it later


: FORGET.SS.SOUND ( -- )						\ dispose channel when quitting
	MySSChan @ ?dup
	IF	\ dup (shush)
  		true  dispose.sound.channel drop
		MySSChan off
	THEN  ;
on.forget FORGET.SS.SOUND						\ dispose when MySQChan is forgotten
\ exodus 		linklasttoken
\ presnapshot linklasttoken



: WaitSSChanDone ( -- )		\ stall until current sound commands complete
	MySsChan @ (WaitSoundDone) ;



\ factor the restorer word out so it catches and ignores errors, 
\ rather than aborting during the restorer chain load

: RestoreSoundStuff ( -- )
	MySSChan off
	['] allocate.ss.sound.channel catch drop
;
restorer linklasttoken		\ ensure this is  done again after restarting a snapshot


RestoreSoundStuff  \ initialize them now


\ ------------------------------------------------------------------------------------------



: PlayResourceAsync ( resid -- ioresult )
	locals| theResID |
	WaitSSChanDone

	MySsChan @
	ascii snd_ theResID getresource ?dup
	IF
		true			\ do it ASYNC
		sndplay
	ELSE
		drop
		-1
	THEN
;


\ add some sound effects words to spice up applications

: JOY ( -- )
	 10400 PlayResourceAsync drop ;
: OhOh ( -- )
	 10800 PlayResourceAsync drop ;
: HOWL ( -- )
	 11235 PlayResourceAsync drop ;
: CRACKLE ( -- )
	 10000 PlayResourceAsync drop ;
: SNAP ( -- )
	 10001 PlayResourceAsync drop ;
: POP ( -- )
	 10002 PlayResourceAsync drop ;
: SNICK ( -- )  
	 10003 PlayResourceAsync drop ;
: JEEZ  ( -- ) 
	 10004 PlayResourceAsync drop ;





StopIncluding ************************************


\ show off with a SNAP, CRACKLE, POP 
\ even long sound resources can be played asynchronously
\ although we explicitly force all but the last one to be synchronous here

\ select from here  <-----------

WaitSSChanDone
SNAP 						WaitSSChanDone
CRACKLE 					WaitSSChanDone
POP 						WaitSSChanDone
JOY 						WaitSSChanDone
OHOH 						WaitSSChanDone
HOWL

WaitSSChanDone	\ if we do something now that moves memory or purges the heap, async sound
				\ play may start playing garbage, so we will stop and wait for it

\ --------> to here and press ENTER



\ after including this file, 
\ select from here  <-----------

>m4th

: PlayAll
	cr ." ResID  ResName" cr
	0 locals| index |
	BEGIN
		1 +to index
		ascii snd_ index GetIndResource
		?dup
	WHILE
		\ PlayResourceAsync drop
		dup pocket GetResInfo  ( handle\addr -- )
		pocket w@  5 .r  2 spaces  pocket 6 + count type
		cr
		MySsChan @ swap true sndplay drop
		WaitSSChanDone
		500 ms				\ a little delay
	REPEAT
;


playall

\ --------> to here and press ENTER

