\ Event Monitor Tool.fth
\ Created:       3 May 2012  15:47:59  ward@megawolf.com
\ Last Revision: 3 May 2012  15:47:59  ward@megawolf.com


\ Adds a background task to show recent Events in special window
\ Often useful to see the sequence of Activaye/Update events


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\    to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


(* *************************************************************************

Very simple display currently.  Scrolling would risk triggering
update events for the monitor window

Event Monitoring is disabled when the tool's window is brough to front

************************************************************************* *)

Only Forth Definitions

\ anew --Event_Monitor_Tool.fth--

>M4TH
also ToolVocab definitions


Decimal




0  value EventDisplayMask		\ no events


\ also ToolVocab definitions

\ an EVENT.FILTER is executed AFTER  SynthEvents
\ an EVENTER      is executed BEFORE SynthEvents




new.window EventWindow
	\ 50 700 700 900 		EventWindow w.bounds
	close.box size.box +
    CLOSE.BEHIND +								\ 6.05437
							EventWindow w.attributes
having kWindowCollapseBoxAttribute  kWindowCollapseBoxAttribute EventWindow NewWAttributes  \ 6.053
	sys.window				EventWindow w.behind

 12 value MonitorTop
 86 value 2ndColumn
100 value MonitorBottom
 12 value MonitorLineHeight
  9 value MonitorTextSize
  4 value MonitorTextFont

create xEVENT.RECORD  16 allot&erase


: (ShowWindowEventInfo) ( -- )
	{   | namebuffer[ 256 ] }  
	xevent.record 2 + @
	dup ?active.WindowRef
	IF
		dup 
		_GetWRefCon dup ?wptr
		IF
			namebuffer[ swap  Get.WTitle
			ascii ^ emit namebuffer[ count type
			drop
		ELSE
			drop
		THEN
	ELSE
		dup ?active.wptr
		IF
			namebuffer[ swap  Get.WTitle
			namebuffer[ count type
		ELSE		
			6 .R
		THEN
	THEN
;

: ShowWindowEventInfo ( -- )
	2ndColumn MonitorTop moveto
	(ShowWindowEventInfo)
;


: ShowMouseWas ( -- )
	2ndColumn MonitorTop moveto
	xevent.record
	10  + @  point>xy  swap 
	." x = "    3 .R  
	."    y = " 3 .R
;


: ShowKeyInfo ( -- )
	2ndColumn MonitorTop moveto
	base @ hex
	xevent.record 2 + c@ 			\ 5 + c@ 
	dup s>d <# # # ascii $ hold #> type
	32 127 range 
	IF
		2 spaces ascii ' emit emit ascii ' emit
	ELSE
		drop
	THEN
	base !
;

: ShowActivateEvent ( -- )
	xevent.record  14 + w@
	1 =
	IF
		." activate"
	ELSE
		." deactivate"
	THEN
	ShowWindowEventInfo
;

: ShowSwitchEvent ( -- )
	xevent.record  2 + c@  
\	xevent.record  5 + c@  
	1 =
	IF
		." switched in"
	ELSE
		." switched out"
	THEN
;

: Show4CharCode ( nn -- )
	ascii ' emit sp@ 4 type drop ascii ' emit 
;

: ShowHiLevelEvent ( -- )
	2ndColumn MonitorTop moveto
	xevent.record 2  + @  endianswap32 Show4CharCode
	space
	xevent.record 10 + @  endianswap32 Show4CharCode
;

: ShowDoExecuteEvent ( -- )
	2ndColumn MonitorTop moveto
	xevent.record 10 + @  >name .name
;


: EventWindowCR ( -- )
	MonitorLineHeight +to MonitorTop
	MonitorTop MonitorBottom  > 
	IF MonitorLineHeight -> MonitorTop THEN			\ roll me over
	4 MonitorTop  moveto  80 spaces				\ erase next line
	4 MonitorTop  moveto
;


: DisplayEvent ( event -- event )
	get.window
	outfile @
	skip.events @
	col @
	mfconsole @
	locals| wasConsole wasCol wasSkip wasOutfile wasWindow theEvent |
	EventWindow ?wvisible
	EventWindow front.window <>
	and
	IF
		theEvent 1 swap lshift    EventDisplayMask AND
		IF  
			EventWindow window
			EventWindow outfile !
		\	dflt.con 	mfconsole !
			CALLBACK.CONSOLE 	mfconsole !
			skip.events on
		
			event.record  xevent.record  16 move	\ copy event record
			
			MonitorTextSize _TextSize
			MonitorTextFont textfont
			srccopy 		textmode
	
			4 MonitorTop    moveto
			theEvent
			CASE
				null.event  	 OF ." null"  		 ( ShowTime )		ENDOF
				mouse.down  	 OF ." mouse down"	ShowMouseWas		ENDOF
				mouse.up  		 OF ." mouse.up"	ShowMouseWas		ENDOF
				key.down  		 OF ." key down"	ShowKeyInfo			ENDOF
				key.up  		 OF ." key up"		ShowKeyInfo			ENDOF
				auto.key  		 OF ." auto key"	ShowKeyInfo			ENDOF
				update.event     OF ." update"		ShowWindowEventInfo	ENDOF
				activate.event   OF					ShowActivateEvent 	ENDOF
	
				12 				 OF ." ExecuteEvent" ShowDoExecuteEvent	ENDOF
				mouse.moved		 OF ." mouse moved"	 					ENDOF
				switch.event	 OF					ShowSwitchEvent		ENDOF
				command.key		 OF ." command key"	ShowKeyInfo			ENDOF
				in.desktop		 OF ." in desktop"	ShowMouseWas		ENDOF
				in.menubar		 OF ." in menubar"	ShowMouseWas		ENDOF
\				in.sys.window	 OF ." in SysWindow" ShowMouseWas		ENDOF
				in.lower.window  OF ." in lowerW"	ShowWindowEventInfo	ENDOF
				in.drag.box		 OF ." drag box"	ShowWindowEventInfo	ENDOF
				in.size.box 	 OF ." size box"	ShowWindowEventInfo	ENDOF
				high.level.event OF ." hi level"	ShowHiLevelEvent	ENDOF
				in.zoom.in 		 OF ." zoom in"		ShowWindowEventInfo	ENDOF
				in.zoom.out 	 OF ." zoom out"	ShowWindowEventInfo	ENDOF
				in.close.box 	 OF ." close box"	ShowWindowEventInfo	ENDOF

				28				 OF ." Carbon Menu Event"				ENDOF
				IN.PROXY.ICON	 OF ." Proxy Icon"	ShowWindowEventInfo	ENDOF
				
				IN.NibControl	 OF ." Nib Control"						ENDOF

				dup 2 .R ."  ???????"
			ENDCASE
	
			EventWindowCR

			wasWindow ?wptr
			IF
				wasWindow window
			ELSE	\ we could be in the process of switching windows
				\ 4 MonitorTop    moveto
				."  ** INVALID WINDOW"
				wasWindow 6 .R
				EventWindowCR
				."  ** not restored"
				EventWindowCR
				\ 0 -> EventDisplayMask
				\ true abort" Invalid wasWindow"
			THEN

			wasOutfile	outfile !
			wasCol		col !
			wasConsole	mfconsole !
			wasSkip		skip.events !
		THEN
	THEN
	theEvent
;


: CorrectDisplaySizes ( -- )
	EventWindow GetWindowPortBounds ( wptr4 -- ^rect )
	4 + w@ 
	20 - 			\ bottom margin
	0 max
	MonitorLineHeight / MonitorLineHeight *		\ round down to integral # lines
	 -> MonitorBottom
;


: UpdateEventMonitor ( -- )
	\ possibly restore display from archived array of events
;
EventWindow on.update UpdateEventMonitor


: EventWindow.Action ( -- )
	BEGIN
		do.events
		in.close.box = IF event.filter off THEN
		?keystroke IF drop beep THEN
	AGAIN
;

' EventWindow.Action EventWindow +Window.Action !

: New.ActivateEventMonitor ( flag -- )
	CorrectDisplaySizes		\ correct for any changes in window size
	IF
		\ if we add a scroll bar and the ability to scroll back through an archive of events,
		\ we should disable further event handling when this window is in front
		\ then we could display events in text box, so cut & copy could be used
		false busy
		init.cursor
	ELSE
		\ if we disable event handling with an activate, we must restore it here
	THEN ;

' New.ActivateEventMonitor EventWindow +on.newactivate !


: ActivateEventMonitor ( flag -- )
	dup New.ActivateEventMonitor
	IF
		EventWindow.Action
	THEN ;

EventWindow on.activate ActivateEventMonitor


\ all events except null
\	2 base !  1111:1111:1111:1111:1111:1111:1111:1110 -> EventDisplayMask

\ no null, mouse up or key up
2 base !  1111:1111:1111:1111:1111:1111:1110:1010 -> EventDisplayMask
decimal


: StartEventMonitor ( -- )
	40								\ top
	 screen.bounds 6 + w@s	120 -	\ right bounds
	  dup 200 -						\ left
	   swap
	    400							\ bottom
	     swap			EventWindow W.Bounds

	C" Event Monitor"	EventWindow w.title
	EventWindow add.window
	CorrectDisplaySizes
	['] DisplayEvent event.filter !
;



\ Need to save and restore old value if other routines
\ install their own event filters

\ For now, we are the only one, so we can have our way with it

: Restore.event.filter ( -- )
	event.filter off ; 
on.forget Restore.event.filter



Create EventMon.ToolGroup -1 , \ holds ID of this group of tools


: ForgetEventMonTool ( -- )
	EventMon.ToolGroup @ dup -1 <>
	over ToolCount <= and
	If  PruneToolItem  
		-1 EventMon.ToolGroup !
	ELSE drop
	THEN
;
on.forget ForgetEventMonTool


>m4th
also ToolVocab


: InstallEventMonTool ( low\high -- | Install the DS Tools menu group.)
	ToolsMenu# menu.handle 0= IF Tools.Menu THEN
    1 NewToolGroup          						\ Enough room to install items ?
    IF
    	EventMon.ToolGroup !      						\ Save separator item number
        ['] StartEventMonitor	C" Event Monitor"	NewToolItem drop
    THEN ;


?building 0= [IF]  InstallEventMonTool  [THEN]




stopincluding *********************************************************
******************* comments and/or test code below *******************
***********************************************************************
