\ FIFO-based CONSOLE FOR REMOTE CONTROL OF VFX
\ Last Revision: 9 Aug 2012  07:53:50  ward@megawolf.com

((

VFX in Terminal can be a little awkward for serious development
requiring an external source file editor and lots of repetitive
typing of path names for includes (although its keyboard shortcuts
help quite a bit)

It is possible to harness the power of MacForth's interactive editor 
on VFX without requiring all the baggage of the MacForth extensions.

The trick is to use MacForth as a remote console to replace Terminal
using PIPEs to handle the inter-process communication.

This file (along with pFIFOs.fth) replaces the standard VFX Console
with a custom one that talks with MacForthOnVFX



 TO SETUP MACFORTH ON VFX,  (Remote Editor)

1. Launch MacForthOnVFX
2. Include (underFile Menu)  "../MF EXTRAS/MF Remote/Remote Editor.fth"



 TO SETUP VFX, 

First, be sure that VFX's reference directory is set correctly:

1. Open Apple's Terminal (found in /Applications/Utilites)

2. Set current directory to the folder that contains the MacForth Source 
	code folders - ../VfxOsxPro/MacForthOnVFX/
	
	If you are not proficient with Terminal, this is how to do it:
	a. type "cd "  in the Terminal (the trailing space is important as is
		lower case) but do not terminate the line
	b. In Finder, locate your MacForthOnVFX folder and DRAG its icon onto the
		Terminal window.  This should emplace the full path into
		the command line.
		In my case, the line now looks like:
		iMac:~ ward$ cd /Volumes/Development/MPE/VfxRoot/MacForthOnVFX
		(yours will depend on where you put the parent folders, but the
		last folder should be 'MacForthOnVFX'
	c. Now, hit return in Terminal
	d. If you did it correctly, the Terminal prompt should reflect
		the new current directory, e.g. in my case
		iMac:MacForthOnVFX ward$ 

3. Still, in Terminal, type "vfxosx  <returnkey>"


To setup VFX to run with MacForth as its Console, instead of Terminal:

execute  (use PASTE into the Terminal):
 s" MF EXTRAS/MF Remote/client console.fth" included

then execute
 START  
to enter the "slave mode" and attempt to connect to the 
MacForth Remote Editor

 ))


\ *****************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ *****************************************************************



only Forth definitions

have M4TH 
[IF]
	>M4TH
	lacking TryPipeReadAll  myfolder mfINCLUDE" pFIFOs.fth"
	only Forth definitions
	anew --clientconsole--
[ELSE]
	[undefined] remotewords
	[IF]
		s" MF EXTRAS/MF Remote/pFIFOs.fth"  included
	[THEN]
[THEN]


only Forth definitions
also remotewords definitions


decimal



variable PipesPrepped  PipesPrepped off

ip-default @ 			 value wasConsole
action-of DoAbortMessage value wasDoAbortMessage
action-of DoErrorMessage value wasDoErrorMessage
action-of .locate        value was.Locate
\ action-of .nolocate		 value was.nolocate

[undefined] lf$
[IF]
	create LF$ 		1 C, 10 C,			\ LF in a string.
[THEN]


: RestoreError&Locate ( -- )
	wasDoAbortMessage is DoAbortMessage
	wasDoErrorMessage is DoErrorMessage
	was.Locate		  is .locate
	['] (quit) is quit
\	was.nolocate	  is .nolocate
;
have forgettor [IF] forgettor linklasttoken [THEN]


: Re-INIT-XCON
	xconsole
	dup ip-handle !  dup op-handle !	\ make it the current device
	dup ip-default !  dup op-default !	\ and the default device
    dup to PauseConsole
    dup xs.hOut @  swap gen-handle !
;

have M4TH [IF] also M4TH [THEN]
have init-mfcon
[IF]
	: RestoreNormalConsole ( -- )
		UP@ >R
		restartData @  up!
		wasConsole xconsole =
		IF
			Re-INIT-XCON	\ 
	\		init-xcon			\ restore
		ELSE
			init-mfcon
		THEN
		0 to PauseConsole
		R> UP!
	;
[ELSE]
	: RestoreNormalConsole ( -- )
		UP@ >R
		restartData @  up!
		Re-INIT-XCON  \ 
	\	init-xcon				\ restore
		0 to PauseConsole
		R> UP!
	;
[THEN]
have M4TH [IF]  previous [THEN]



(( 
variable PipeExit?  PipeExit? off


: ForceQuitOnStop ( -- )
	PipeExit? @
	IF
		PipeExit? OFF
		true abort" Remote move exitted - XCONSOLE Restored"
	THEN
;

' ForceQuitOnStop is EmptyIdle

 ))


: (STOP) ( -- )
	CloseUpPipes
	RestoreError&Locate
	RestoreNormalConsole
;


: SafeAbort$ ( c-addr -- )	\ derived from doRestart
\ Executed to perform a restart by running QUIT.
	restartData  dup @ up!
	swap 'AbortText !
	dup cell + @ rp!  2 cells + @ sp!
	ip-default @ ip-handle !  
	op-default @ op-handle !
	-2 throw
;

: ClientSigPipeHandler		\ signum *siginfo *ucontext --
	nip nip nip
	(stop)
\	QUIT
\	PipeExit? ON
\	doRestart
	c" SIGABRT or Pipe Error - XCONSOLE Restored" SafeAbort$
;


: ?PrepPipes ( -- )
	PipesPrepped @ 0=
	IF
		false -> IsThisTheEditor?
		InitializePipes
		['] ClientSigPipeHandler is UserSIGPIPEhandler
	THEN	
	?OpenReadPipe  drop
	?OpenWritePipe drop
	PipesPrepped on
	10:000 -> WriteMaxTime		\ 10 seconds to time out writes
;


: PoliteWrite ( addr\cnt --  )
\ Repeatedly tries to write for WriteMaxTime
\ If it times out or other error
\ 	then we restore normal Terminal console
\ 	and post an error message
	KeepTryingWrite ( addr\cnt -- ioR )
	0<
	IF 
		(stop)
		cr type
		true abort" PIPE WRITE FAILURE"
	THEN
;

(( 
: ReadUntilEmpty
	BEGIN
		emptyidle
		2 ms
		DataAvailable?  
		dup 0>
		IF
			cr ." avail: "  . true
		ELSE
			drop false
		THEN
	WHILE
		TryPipeReadAll  drop	\  cr ." flushed " .
	REPEAT
;
 ))





(( 
: ReadUntilEmpty
 cr ." ReadUntilEmpty "  DataAvailable? .
	BEGIN
		emptyidle
		2 ms
		ReceivingPipeFD readbuf[ DataAvailable? read dup 0 <=
		IF
			cr ." fail read " drop
			ErrNo .
		\	true
		ELSE
			cr ." read" BigReadBuffer swap type
		\	false
		THEN
		DataAvailable? not
	UNTIL
;
 ))


: Flush+Quit
	ReadUntilEmpty		\ Purge ReceivingPipe
	(quit)
;

\ ********** CONSOLE PRIMITIVES *************************************

: read-RemoteClient	\ addr len sid -- ior			\ read-xterm
	drop
	?OpenReadPipe dup 0>
	IF
		-rot read -1 =
		if errno  else 0 endif
	ELSE
		drop
		-1
	THEN
;


: write-RemoteClient	\ addr len sid -- ior			\ write-xterm
	3DROP  0
;


: KEY?-RemoteClient	\ sid -- flag		\ Key?-xterm ,
  { | nread[ cell ]  }
	drop 						\  
	DataAvailable? 0>
(( 
	?OpenReadPipe dup 0>
	IF
		FIONREAD nread[ ioctl 	\ Get the number of bytes that are immediately available for read
  		-1 =
  		if  0  else  nread[ @ 0<>  endif
  	ELSE
  		drop 0
  	THEN
 ))
;


: KEY-RemoteClient	\ sid -- key		\ sid -- key		\ Key-xterm ,
	{ sid | temp[ cell ] -- }
	BEGIN
		sid KEY?-RemoteClient 0=
	WHILE
		emptyidle
		2 ms
	REPEAT
	temp[ 1 sid read-RemoteClient drop
	temp[ c@
;


: ACCEPT-RemoteClient		\ address maxlen sid -- len'	\ Accept-xterm ,
	{ address maxlen sid | len }
	0 -> len
  	address maxlen bounds
	?DO
    	sid KEY-RemoteClient

		dup #127 = if  drop #8  then	\ treat DEL as BS
		
		CASE
			#8 of                             \ <BACKSPACE>
				len 0<>
				if                     		 \ if not at left edge
					len 1- -> len                 \ decrement length
	\				#8 emit  #32 emit  #8 emit
					-1                            \ +LOOP modification -1
				else
					0                             \ Don't increment Loop
				then
			endof
			
			#9 of                             \ <TAB>
				bl                \ treat as space
		\		dup emit
				i c!
				len 1+ -> len
				1
			endof
			
			#10 of                 \ <CR> for DOS/Win, <LF> for Unices ; SFP005
	\		EOLchar_con of                 \ <CR> for DOS/Win, <LF> for Unices ; SFP005
				bl
	\			dup emit                \ output a <SPACE>
				i c!                       \ store on end of string
				leave                      \ and leave
			endof
			
			#13 of                    \ <LF> for DOS/Win, <CR> for Unices ; SFP005
	\		IGNchar_con of                    \ <LF> for DOS/Win, <CR> for Unices ; SFP005
				0						\ ignore
			endof
			
	\		dup drawchar
	\		dup emit                          \ DEFAULT
			i c!                              \ store char in buffer
			len 1+ -> len                     \ increment string length
			1                                 \ modify address by 1 (+LOOP)
			
		END-CASE			\ NOT ENDCASE !!!!!!!!
	+LOOP
	
	len
;


: EMIT-RemoteClient	\ char sid --		\ Emit-xterm ,		\ 9
	drop
	{ | temp[ cell ] -- }
	temp[ c! 
	1 out +!
	temp[ 1 	\ TryPipeWrite drop
	PoliteWrite
;


: TYPE-RemoteClient	\ caddr len sid --		\ Type-xterm ,		\ 11
	drop
	dup  out +!	
	PoliteWrite
;


: CR-RemoteClient	\ sid --			\ CR-xterm ,		\ 12
	#10 swap EMIT-RemoteClient
	out off  op-line# incr
;


: lf-RemoteClient	\ sid --
	cr-RemoteClient
;


: open-RemoteClient	\ addr len attribs sid -- handle/sid ior
	?PrepPipes
	nip nip nip
 	
 	OpenPipes
	$40 c/line !
	0
;


: close-RemoteClient	\ sid -- ior
  -1 swap gen-handle !			\ mark closed
  0
;

: ekey?-RemoteClient	\ sid -- flag
	key?-RemoteClient
;

: ekey-RemoteClient	\ sid -- echar
	key-RemoteClient
;

: emit?-RemoteClient	\ sid -- flag
\ Return true if we can send more characters
	drop  true  ;


: ff-RemoteClient	\ sid -- ; page/cls on display devices
\ maybe implement as multiple LFs ?
	drop
\  	page		\ ???? 
;

: bs-RemoteClient	\ sid -- ; destructive on display devices
\ maybe implement ?
	drop
;


: bell-RemoteClient	\ sid -- ; audible beeper
  7 swap emit-RemoteClient  ;


: setpos-RemoteClient	\ x y mode sid -- ior
\ Go to cursor position x,y. Position 0,0 is top left. Mode is ignored.
	3drop
	-1
;

: getpos-RemoteClient	\ mode sid -- x y ior
	2drop
	0 0 -1
;

: ioctl-RemoteClient	\ addr len fn sid -- ior
\ could implement
	4drop -1
;

: flushOP-RemoteClient	\ sid -- ior
\ could implement ?
	drop 0 ;


: readex-RemoteClient	\ caddr len sid -- u ior
\ could implement
	drop 2drop 0 -1 ;


: init-RemoteClient	\ addr len sid -- ior
\ could implement
	drop 2drop  0  ;


: term-RemoteClient	\ sid -- ior
\ could implement
	drop  0  ;


: config-RemoteClient	\ sid -- ior ; produces a dialog
	drop  0  ;


also system

create RemoteClient-vectors
  ' Open-RemoteClient ,		\ 0  OpenFnid
  ' Close-RemoteClient ,	\ 1
  ' Read-RemoteClient ,		\ 2
  ' Write-RemoteClient ,	\ 3
  ' KEY-RemoteClient ,		\ 4
  ' KEY?-RemoteClient ,		\ 5
  ' Ekey-RemoteClient ,		\ 6
  ' EKey?-RemoteClient ,	\ 7
  ' ACCEPT-RemoteClient ,	\ 8
  ' EMIT-RemoteClient ,		\ 9
  ' Emit?-RemoteClient ,	\ 10
  ' TYPE-RemoteClient ,		\ 11
  ' CR-RemoteClient ,		\ 12
  ' LF-RemoteClient ,		\ 13
  ' FF-RemoteClient ,		\ 14
  ' BS-RemoteClient ,		\ 15
  ' Bell-RemoteClient ,		\ 16
  ' setpos-RemoteClient ,	\ 17
  ' Getpos-RemoteClient ,	\ 18
  ' Ioctl-RemoteClient ,	\ 19
  ' FlushOP-RemoteClient  ,	\ 20
  ' readex-RemoteClient ,	\ 21
  ' Init-RemoteClient ,		\ 22
  ' Term-RemoteClient ,		\ 23
  ' config-RemoteClient ,	\ 24

previous


\ base on /xterm-sid even though we will skip several elements

create defRemoteClient	\ -- addr
\ Default xterm structure - must match /xterm
  0 , 						\ gen-handle
  RemoteClient-vectors , 	\ gen-vector
  							\ gen-private
  0 , 						\ xs.hIn
  0 , 	\ 1 ,				\ xs.hOut
  0 ,	\ 2 ,  				\ xs.hErr
  0 ,						\ xs.flags
  0 , 0 , 0 ,				\ history


: initRemoteClientSid	\ addr --
\ *G Initialise a *\fo{/xterm} structure at *\i{addr}.
  defRemoteClient swap /xterm-sid cmove
;


: RemoteClient:	\ -- ; -- sid ; XTERM: <name>
\ *G Create a new terminal device.
	create
	here /xterm-sid allot
	initRemoteClientSid
;


RemoteClient: RemoteClientConsole	\ -- addr
\ *G VFX Forth console.


: INIT-CLIENTCON ( -- )
\ *G Set up to use the *\fo{xconsole} device. Performed at start up
\ ** and compilation.
	?PrepPipes
	RemoteClientConsole
\ $000C over xs.flags or!				\ demand history buffering
	dup ip-handle !   dup op-handle !	\ make it the current device
	dup ip-default !  dup op-default !	\ and the default device
	DROP
	0 to PauseConsole
  NULL NULL NULL open-gen 2drop
;
\ ' init-clientcon AtCold  init-xcon  -- init-clientcon DONE IN M4PLSTARTUP


: terminate-RemoteClient ( -- )
\ *G Shut down the *\fo{xconsole} device. Performed at shut down.
	WaitWritePipeEmpty
	(stop)
	QUIT
;
\ ' terminate-RemoteClient AtExit




\ *****************************************************************
\ *****************************************************************

0 value BigWriteBuffer		\ 8000 byte scratch buffer from Heap


: SendBufferAsHostCommand ( C-addr -- )
	\ we need to send commands in a single WRITE
	\ so host can digest the whole thing in one READ
	\ Host Commands start with command header
	\ MUST wait for write pipe to be empty before sending a command sequence
	WaitWritePipeEmpty
	$FF01 BigWriteBuffer !
	count BigWriteBuffer zappend

	\ MUST wait for write pipe to be empty before sending a command sequence
	WaitWritePipeEmpty

	BigWriteBuffer  zcount PoliteWrite
	
	\ and wait for it to finish
	WaitWritePipeEmpty
;

(( 
: SendBufferAsHostCommand ( C-addr -- )
	{  | msgbuf[ cell ] }
	\ Host Commands start with command header
	$FF01 msgbuf[ !

	\ MUST wait for write pipe to be empty before sending a command sequence
	WaitWritePipeEmpty
	msgbuf[  zcount PoliteWrite
	
	\ now send rest of the command
	count PoliteWrite

	\ and wait for it to finish
	WaitWritePipeEmpty
;
 ))

\ *****************************************************************



also substitutions

: +BasicLocateInfo  ( ^buffer -- )
	{ theBuff }

	locate_line count		theBuff append
	s"  clLine# !"			theBuff append
	lf$ count				theBuff append
	
	s" clSource: "			theBuff append
	locate_path count		theBuff append
	lf$ count				theBuff append
;


: LatestNameIn  \  wid -- nfa
\ method borrowed from  WalkWordList
  {  wid | temp[ 512 1+ cells ] -- }
  \ Copy wordlist to temporary buffer
 	wid temp[ over >#threads @ 1+ cells move
 	temp[ (max-def) nip ctrl>nfa
;

: LatestSourceFile ( -- nfa )
	['] sourcefiles voc>wid LatestNameIn
;

: FindSourceFile ( -- addr\cnt )
\ sometimes CurrSourceName returns a bogus string
	CurrSourceName expand 2dup
	FileExist? 0= ( path$\cnt -- flag )		\ IsFileMissing?
	IF
		2drop
		LatestSourceFile count
		expand
	THEN
;



: +BasicErrorInfo ( err#\buff -- )
	locals| theBuff |

	\ add "<nnn> clErr# !"
	(.)					theBuff append
	s"  clErr# !"		theBuff append
	lf$ count			theBuff append
	  
	\ set line#
	line# @  (.)  locate_line place

	\ set source file name
	locate_path OFF
	'SourceFile @ 0 >	\ not console or cached
	IF
		FindSourceFile locate_path place
	THEN

	theBuff +BasicLocateInfo

	\ add offending word
	s" clTheWord: "		theBuff append
	saved>in @ dup 'tib @ +		\ wasOffset\start
	>in @ rot -  
	-white				theBuff append
	lf$ count			theBuff append
	
;
previous



: SendAbortMessage ( c-addr -- )
	{  | msgbuf[ 2048 ] }
	ReadUntilEmpty		\ Purge ReceivingPipe
	msgbuf[ off
	-2					msgbuf[ +BasicErrorInfo
	\ add abort$
	s" clAbort: "		msgbuf[ append
	count				msgbuf[ append
	lf$ count			msgbuf[ append
	s" DisplayAbort"	msgbuf[ append
\	lf$ count			msgbuf[ append

	msgbuf[ SendBufferAsHostCommand ( C-addr -- )
;


: SendErrorMessage ( N -- )
	{  | msgbuf[ 2048 ] }
	ReadUntilEmpty		\ Purge ReceivingPipe 

	msgbuf[ off
						msgbuf[  +BasicErrorInfo
	s" DisplayError"	msgbuf[ append
\	lf$ count			msgbuf[ append

	msgbuf[ SendBufferAsHostCommand ( C-addr -- )
;


\ *****************************************************************

also system

: SendLocateInfo ( -- )
	{   | msgbuf[ 2048 ] }
	msgbuf[ off
	
							msgbuf[  +BasicLocateInfo  ( ^buffer -- )

	s" clTheWord: "			msgbuf[ append
	lookfor @ >name count	msgbuf[ append
	lf$ count				msgbuf[ append

	s" DisplayLocation"		msgbuf[ append
\	lf$ count				msgbuf[ append

	msgbuf[ SendBufferAsHostCommand ( C-addr -- )
;
previous


\ *****************************************************************

: HostInit ( -- )	\ ask host to do necessary initialization
					\ for now, it is just updating c/line
	{   | msgbuf[ 100 ] }
	\ start with command header
	$FF01 					msgbuf[ !

	s" HostInit "			msgbuf[ zappend
	lf$ count				msgbuf[ zappend
	
	getpid (u.)				msgbuf[ zappend
	s"  ClientPID ! "		msgbuf[ zappend
	lf$ count				msgbuf[ zappend
	

	WaitWritePipeEmpty
	msgbuf[  zcount PoliteWrite
;



\ *****************************************************************

0 value theList			\ temporary allocation for buildidng Context list
0 value TotalSize


: AddOneFromWL  ( thread# nfa -- flag ) \ Return TRUE to continue 
	nip
	dup c@
	IF
		count
	ELSE
		drop s" ??"
	THEN
	dup 1+ +to TotalSize
				theList  zappend
	lf$  count	theList  zappend
	true
;


: AddWid	( wid -- )	\ dump all names in wid
	['] AddOneFromWL swap WalkWordList  \ xt wid --
;


: BuildContextList ( -- )
	0 -> TotalSize
	theList off

	context vsize bounds
	swap
	do 
    	i @ ?dup
    	IF
    		AddWid
    	THEN
	cell negate +loop
	
	theList  zcount + 16 0 fill 
;


: BuildOtherList ( -- )
	0 -> TotalSize
	theList off

	voc-link
	BEGIN
		@ dup
	WHILE
		dup ?order		\ InContext?
		0=
		IF
			\ dup cr >VOCNAME count type
			dup AddWid
		THEN
		>voc-link
	REPEAT
	drop	
	theList  zcount + 16 0 fill 
;

: BuildFullList ( -- )
	0 -> TotalSize
	theList off

	voc-link
	BEGIN
		@ dup
	WHILE
	\	dup ?order		\ InContext?
	\	0=
	\	IF
			\ dup cr >VOCNAME count type
			dup AddWid
	\	THEN
		>voc-link
	REPEAT
	drop
	theList  zcount + 16 0 fill 
;




\ *****************************************************************
\	PUBLIC WORDS FOLLOW
\ *****************************************************************

only forth definitions
also remotewords
also substitutions


: START	 ( -- )
\	devpath $>asciiz chdir drop
\	IniDir$  count expand s" INIDIR"  replaces
\	IniFile$ count expand s" INIFILE" replaces


	cr ." Entering Remote Control Mode"
	cr

	BigWriteBuffer ?dup IF free drop THEN
	8000 allocate drop -> BigWriteBuffer
	init-clientcon

	['] SendAbortMessage is DoAbortMessage
	['] SendErrorMessage is DoErrorMessage
	['] SendLocateInfo   is .Locate
\	['] SendNolocateInfo is .noLocate

	['] Flush+Quit is quit

	." Starting Remote Client - PID = "  getpid .
	cr
	(.cold)
	HostInit
\	['] remotewords >body @ >MIN-ORDER
	only forth definitions
	(Quit)
;


: ExitTerm ( -- )
	cr cr
	." REMOTE CLIENT IS CLOSING PIPES AND RESUMING INTERACTIVE MODE"
	cr cr cr

	WaitWritePipeEmpty
	BigWriteBuffer ?dup IF free drop THEN
	0 -> BigWriteBuffer
	(stop)
	cr ." STOP'ped - XCONSOLE Restored" 
	quit
;


previous


\ also remotewords definitions

\ : START start ;



: BYEBYE  ( -- )
	cr ." REMOTE CLIENT is about to go BYE-BYE"
	(stop)
	BYE
;


: ReadChunk&Interpret ( n -- )
	{ chunksize }
	bigreadbuffer  chunksize ReadBigChunk ( addr/cnt -- total#read )
	drop		\ ? compare to chunksize
	bigreadbuffer  chunksize includemem
;


: DumpContext  ( -- )
	{   | msgbuf[ 256 ] }
\ create a LF-terminated list of all words in CONTEXT
\ with a few 0's at the end
	theList ?dup IF free drop THEN
	1:000:000 allocate drop to theList		\ allocate a temporary list buffer
\	BuildContextList
	BuildFullList

	ReadUntilEmpty							\ Purge ReceivingPipe

\ then send command to host to retrieve that list
	TotalSize	16 +	(.)  	msgbuf[ place
	s"  ImportRemoteContext "	msgbuf[ append
	msgbuf[ SendBufferAsHostCommand ( C-addr -- )

\ then send the list as one big chunk
	theList TotalSize  16 +  WriteBigChunk drop

	theList free drop						\ recover our temprary memory
	0 -> thelist
;


only forth definitions

cr ." type START to start remote console"


\ *****************************************************************
\ *****************************************************************
\\