\ REMOTE EDITOR - SECOND DRAFT
\ Last Revision: 23 Apr 2012  06:48:13  ward@megawolf.com

((

VFX in Terminal can be a little awkward for serious development
requiring an external source file editor and lots of repetitive
typing of path names for includes (although its keyboard shortcuts
help quite a bit)

It is possible to harness the power of MacForth's interactive editor 
on VFX without requiring all the baggage of the MacForth extensions.

The trick is to use MacForth as a remote console to replace Terminal
using PIPEs to handle the inter-process communication.

This file (along with pFIFOs.fth) replaces the standard VFX Console
with a custom one that talks with MacForthOnVFX



 TO SETUP MACFORTH ON VFX,  (Remote Editor)

1. Launch MacForthOnVFX
2. Include (underFile Menu)  "../MF EXTRAS/MF Remote/Remote Editor.fth"



 TO SETUP VFX, 

First, be sure that VFX's reference directory is set correctly:

1. Open Apple's Terminal (found in /Applications/Utilites)

2. Set current directory to the folder that contains the MacForth Source 
	code folders - ../VfxOsxPro/MacForthOnVFX/
	
	If you are not proficient with Terminal, this is how to do it:
	a. type "cd "  in the Terminal (the trailing space is important as is
		lower case) but do not terminate the line
	b. In Finder, locate your MacForthOnVFX folder and DRAG its icon onto the
		Terminal window.  This should emplace the full path into
		the command line.
		In my case, the line now looks like:
		iMac:~ ward$ cd /Volumes/Development/MPE/VfxRoot/MacForthOnVFX
		(yours will depend on where you put the parent folders, but the
		last folder should be 'MacForthOnVFX'
	c. Now, hit return in Terminal
	d. If you did it correctly, the Terminal prompt should reflect
		the new current directory, e.g. in my case
		iMac:MacForthOnVFX ward$ 

3. Still, in Terminal, type "vfxosx  <returnkey>"


To setup VFX to run with MacForth as its Console, instead of Terminal:

execute  (use PASTE into the Terminal):
 s" MF EXTRAS/MF Remote/client console.fth" included

then execute
 START  
to enter the "slave mode" and attempt to connect to the 
MacForth Remote Editor

 ))



\ *****************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ *****************************************************************

>m4th




only forth definitions
>m4th

lacking remotewords  myfolder mfINCLUDE" pFIFOs.fth"


only forth definitions
anew --remoteeditor--
>m4th

lacking remotewords  myfolder mfINCLUDE" pFIFOs.fth"
>m4th
also remotewords
also editorvocab

(( 

Experimentally validated a problem using a task (pthread) to post an event to Main
The event handler is likely called in the pThread's domain, making thread unsafe calls to
	update the editor
Thus, using a task to monitor Receive Pipe is impractical.
POSSIBLY could be made to work with
 - Cocoa Notifications
 - PTHREAD_KILL from task thread sent to MAIN with custom signal and appropriate handler


Instead, for now use a NullEventer to monitor Read Pipe

For faster polling, use eventer instead of nulleventer?
))

\ *************************************************************************
-1 constant InsertAtEND
-2 constant InsertAtSelend
 0 constant InsertAtStart


InsertAtSelend  value TerminalInsertionMode
true			value One-ShotEscapeMode
true			value EchoEscapeChars
1				value #LFsAfterEnter		\ accepts 0, 1, 2
1				value #LFsAfterReply		\ accepts 0, 1, 2


\ *************************************************************************
\ *************************************************************************
\ *************************************************************************


13 value TerminalMenu#


50 700 700 500 NEW.SEWINDOW TerminalWindow
		true TerminalWindow se.reserved!



: Indata>Terminal ( addr\cnt -- )
 	TerminalWindow TerminalInsertionMode (insert.text>se)
	TerminalWindow 4 MakeSelendVisible ( SEWindow\offset -- )
;


: ConnectedToRemote? ( -- flag )
	TerminalWindow ?wvisible
	?OpenWritePipe 0>  and
	?OpenReadPipe 0>   and
;



\ *************************************************************************

create TermMsgRect 0 , 0 ,


: UpdateTermStatus ( -- )
	TerminalWindow @ TermMsgRect InvalWindowRect drop
;


: WriteToRemote ( addr\cnt -- )
	KeepTryingWrite dup 0<
	IF						\ error
		UpdateTermStatus
		CASE
			-2 OF true abort" Pipe Not Open"		ENDOF
			-3 OF true abort" Pipe Write time out"	ENDOF
			abort" Pipe Write Error" 
		ENDCASE
	ELSE
		drop
	THEN
;


defer HandlePipeInput ( addr\cnt -- )
		' 2drop is HandlePipeInput
		
\ *************************************************************************


$FA6C $FB3A $FFFF	rgbcolor Term.BACKCOLOR	


0 value LastDisplayedStatus
0 value EscMode?




: ShowTerminalStatus ( -- )
	mfconsole @  
	{  wasconsole | scrbackcolor[ 8 ] scrforecolor[ 8 ] }
	dflt.con mfconsole !
	scrbackcolor[ GetBackColor
	scrforecolor[ GetForeColor
	RGBLtGray rgbbackcolor
	TermMsgRect eraserect
\	TermMsgRect invalrect
	\ fontsize
	
	OpenPipes
	
	20 14 moveto
	ConnectedToRemote?
	IF
		EscMode?
		IF
			rgbred rgbforecolor
			." next key sent immediately"	
		ELSE
			rgbblack rgbforecolor
			." <ENTER> key to send data"	
		THEN
	ELSE
		rgbblue rgbforecolor
		." REMOTE CLIENT OFF-LINE?"	
	THEN

	255 14 moveto
	?OpenReadPipe 0>
	IF
		rgbblack rgbforecolor
		." Read Pipe open"
	ELSE
		rgbred rgbforecolor
		." Read Pipe NOT open"
	THEN

	420  14 moveto
	?OpenWritePipe 0>
	IF
		rgbblack rgbforecolor
		." Write Pipe open"
	ELSE
		rgbred rgbforecolor
		." Write Pipe NOT open"
	THEN

	ConnectedToRemote? -> LastDisplayedStatus
	scrbackcolor[ rgbbackcolor
	scrforecolor[ rgbforecolor
	wasconsole mfconsole ! 
;




: FlashTermStatus ( -- )
	get.window 
	{  waswindow | scrcolor[ 8 ] }
	skip.events ON
	terminalwindow window
	\ TerminalWindow no.clip
	scrcolor[ GetBackColor
	RGBRed rgbbackcolor
	TermMsgRect cliprect
	TermMsgRect eraserect
	TerminalWindow FlushWindow
	200 ms
	3 0
	DO
		TermMsgRect invertrect
		TerminalWindow FlushWindow
		200 ms
	LOOP
	scrcolor[ rgbbackcolor
	ShowTerminalStatus
	skip.events OFF
	waswindow window
;


: SE.TIB->REMOTE 
	se.tib @  se.tib GetHandleSize
	dup #8000 > 		abort" Too large a selection"
 
	dup  (u.)					WriteToRemote
	s"  ReadChunk&Interpret"	WriteToRemote
	lf$ count   			 	WriteToRemote

	
	WriteToRemote
;


: TerminalEnterKey	( wptr -- )		\ Export.ENTER function for TerminalWindow
									\ analogous to SE.INTERPRET
	EscMode?
	IF
		drop exit
	THEN

	?OpenWritePipe 0>
	IF  
		DisableUndo
		SE.TIB->REMOTE
		#LFsAfterEnter
		CASE
			0 OF bl$	ENDOF
			1 OF lf$	ENDOF
			lf$lf$ swap
		ENDCASE
		count Indata>Terminal

		EnableUndo
	ELSE
		flashTermStatus
\		se.interpret
	THEN
;


0 value wasAutoCompletion

defer InstallTerminalPopup
	' noop is InstallTerminalPopup


: TerminalActivate					\ myse.enter action for TerminalWindow
	\ executed at the end of ENTER.SE
	\ forbid Autocompletion in Terminal for now
	Autocompletion @ -> wasAutoCompletion
	Autocompletion off
	SetAutocompletion
	Autocompletion.ToolGroup @ 1 +
		false ToolsMenu# item.enable

	['] true is Editor-Cntrl-Click		\ disable popup 
	InstallTerminalPopup				\ optionally replace popup
;


: TerminalDeactivate				\ myse.exit action for TerminalWindow
	\ executed during EXIT.SE
	wasAutoCompletion  Autocompletion !
	SetAutocompletion
	Autocompletion.ToolGroup @ 1 +
		true ToolsMenu# item.enable

	['] ($Locate) is (Locate)
	
	InstallEditorPopup					\ normal popup behavior
;


: TweakSEMenu ( flag -- )
	11 false	FileMenu# item.enable		\ disable SAVE & INCLUDE
	DROP
;



: TerminalKeyFilter ( char -- char )
	{ theC	 | buff[ cell ] }

	theC 27 =			\ ESCAPE ?
	IF
		beep
		EscMode?
		IF
			false
		ELSE
			true
		THEN to EscMode?
		UpdateTermStatus
		0
	ELSE
		EscMode?
		IF	\ send next key directly
			theC 13 = IF 10 TO theC THEN	\ return -> LF
			theC  3 = IF 10 TO theC THEN	\ ENTER  -> LF
			theC buff[ c!
			buff[ 1 WriteToRemote

			EchoEscapeChars
			IF
				buff[ 1 Indata>Terminal
			THEN

			One-ShotEscapeMode
			IF
				false -> EscMode?	\ one shot
			THEN
			\ undo range ??
			\ selstart <-- selend ??
			
			UpdateTermStatus
			0
		ELSE
			theC
		THEN
	THEN
;


: InstallTermWindow
	TerminalWindow +wbounds w@
		SEditHigh  50 max
		+
		TerminalWindow +wbounds 4 + w!
	
	TerminalWindow +wbounds 2+ w@
		seditwide  50 max
		+
		TerminalWindow +wbounds 6 + w!

 	0 0 20 1000		TermMsgRect !rect
	20 				TerminalWindow +bordertop w!

	TerminalWindow add.se
	c" Remote Edit" TerminalWindow set.wtitle
	
 
	Term.BACKCOLOR @rgb		TerminalWindow +SE.wasTopBackColor !rgb

	default.EOLMODE			TerminalWindow +se.orginaleolmode !
	default.EOLMODE			TerminalWindow +SE.EOLMode !
	10 						TerminalWindow +se.eolchar !
	['] ShowTerminalStatus	TerminalWindow +export.update !
	['] terminalEnterKey	TerminalWindow +export.enter !
	['] terminalActivate	TerminalWindow +myse.enter !
	['] terminalDeactivate	TerminalWindow +myse.exit !
	['] TweakSEMenu			TerminalWindow +MySE.MenuTweaks !
	
	['] TerminalKeyFilter	TerminalWindow +export.key !

\ even a noop will block Editor-Cntrl-Click
\	['] noop				TerminalWindow +MySE.CtrlClick !


	TerminalWindow add.se
	c" Remote Edit" TerminalWindow set.wtitle

	get.window >R
	TerminalWindow window
	Term.BACKCOLOR  rgbbackcolor
\	sys.window window
\	Term.BACKCOLOR  rgbbackcolor
	R> window
;


: MySigPipeHandler		\ signum *siginfo *ucontext --
	3drop
	beep
	\ nonono not from callback   true abort" PipeError"
	Lf$ count Indata>Terminal
	Lf$ count Indata>Terminal
	s" WRITE PIPE NOT ACTIVE"  Indata>Terminal 
	Lf$ count Indata>Terminal

	['] flashTermStatus	SendExecuteEvent	\ <-- no graphics direct from callback !!
;


: SetupTerminalAsEditor
	InstallTermWindow
	true -> IsThisTheEditor?
	c" EDITOR" 1 SetMenuTitle drop

	['] MySigPipeHandler is UserSIGPIPEhandler
	InitializePipes
	OpenPipes
;


: SetupTerminalAsClient		\ used for testing
	InstallTermWindow
	false -> IsThisTheEditor?
	c" CLIENT" 1 SetMenuTitle drop

	['] MySigPipeHandler is UserSIGPIPEhandler
	InitializePipes
	OpenPipes
;



\ *************************************************************************
\ *************************************************************************
\ *************************************************************************

0 value #READ
 
: MonitorInPipe ( -- )
	pause
	DataAvailable?
	IF
		TerminalWindow ?wvisible
		IF
			TryPipeReadAll dup 0>
			IF
				dup +to #READ
				BigReadBuffer swap  			\ ?? use local buffer ??
				HandlePipeInput					\ SendTextToTerminal
			ELSE
				drop
			THEN
	
	((  \ this is a bit faster if tons of data coming in FAST, but won't refresh the Editor 
		\ until there is a break in the data flow
			BEGIN
				TryPipeReadAll
				dup 0>
			WHILE
				BigReadBuffer swap  dup +to #READ			\ ?? use local buffer ??
				HandlePipeInput					\ SendTextToTerminal
			REPEAT
			drop
	))
	
		ELSE
			InstallTermWindow
		THEN
	THEN
;

\ *************************************************************************
\ *************************************************************************

: ADD/SHOWTERMINAL ( -- )
	TerminalWindow ?wvisible
	IF
		TerminalWindow select.window
	ELSE
		InstallTermWindow
	THEN
;


: LaunchVFXInTerminal ( -- )
	\ needs compiled applescript app in /contents/Resources
	s" %MyResourcesPath%/Launch VFX in Terminal.app" expand
	MakeCFURLfromPath ( addr\len -- CFURL )
	OpenCFURLRef drop
;


: HostInit ( -- )	\ update remote c/l
\ Terminal Window should be in front
\
    c/line @ (u.)				WriteToRemote
	s"  c/line !  DumpContext" 	WriteToRemote
	lf$ count 					WriteToRemote

	
\	s" DumpContext "  WriteToRemote
\	lf$ count 		WriteToRemote
\ assign .SYSPROMPT to-do .PROMPT
;



defer AfterRemoteInclude
	' noop is AfterRemoteInclude


: RemoteInclude ( addr\cnt -- )
	{ | buf[ 1024 ] }
	s\" S\" " 	\ "  the escaped string confuses the colorizer
						buf[ place
						buf[ append
	s\" \" included"	buf[ append
	lf$ count  			buf[ append
	\ buf[ count type
	buf[ count 	WriteToRemote
	AfterRemoteInclude
;


: RemoteINCLUDE..  ( -- )
	['] RemoteInclude is <<included>>
	(include..)
;


: SAVE&RemoteINCLUDE  ( -- )		\ see SAVE&INCLUDE
	['] RemoteInclude is <<included>>
	TerminalWindow (save&include)
;


: STOP.REMOTE
	lf$ count		WriteToRemote
	s" ExitTerm" 	WriteToRemote
	lf$ count		WriteToRemote
;

: EXIT.REMOTE
	lf$ count	WriteToRemote
	s" BYEBYE" 	WriteToRemote
	lf$ count	WriteToRemote
;


0 variable ClientPID				\ client PID


: SendSIGABRT ( -- )
	{ | cmd[ 256 ] }
	ClientPID @ ?dup
	IF
		s" kill -s ABRT "	cmd[ zplace
		(U.)				cmd[ zappend
		cmd[ >Shell
	THEN
;


: SendSIGKILL ( -- )
	{ | cmd[ 256 ] }
	ClientPID @ ?dup
	IF
		s" kill -s KILL "	cmd[ zplace
		(U.)				cmd[ zappend
		cmd[ >Shell
	THEN
;


: DO.TerminalMenu ( item -- ) 
	0 hilitemenu
	CASE
		1 OF  add/showterminal		ENDOF
		2 OF  LaunchVFXInTerminal	ENDOF
		\ 3
		4 OF  HostInit				ENDOF
		\ 5  separator
		6 OF  RemoteINCLUDE..		ENDOF
		7 OF  SAVE&RemoteINCLUDE	ENDOF
		\ 8  separator
		9 OF  STOP.REMOTE			ENDOF
	   10 OF  EXIT.REMOTE			ENDOF
		\ 11  separator
	   12 OF  SendSIGABRT			ENDOF
	   13 OF  SendSIGKILL			ENDOF
	ENDCASE ;

(( 
A CDEFGHIJKLMNO QRS UVWX Z  .\[]
 B             P   T    Y
 ))

: TerminalMenu ( -- )
	TerminalMenu# get.next.id  C" Terminal" TerminalMenu# new.menu
	C" Open Terminal Window"				TerminalMenu# append.items
	C" Launch VFX in Terminal"				TerminalMenu# append.items
	C" Launch VFX faceless("				TerminalMenu# append.items

	C" xx"									TerminalMenu# append.items
	4 C" Update C/LINE" 					TerminalMenu#  set.item$

	C" -(;Remote Include File.../I"			TerminalMenu# append.items
			6 1 TerminalMenu# set.item.modifiers \ cmd-shift-I
	C" Save & Remote Include/K"				TerminalMenu# append.items
			7 1 TerminalMenu# set.item.modifiers \ cmd-shift-K
	
	C" -(;Remote Disconnect;Remote Quit"	TerminalMenu# append.items
	C" -(;Remote SIGABRT;Remote SIGKILL"	TerminalMenu# append.items
(( 
	C" Emit Space to Remote"				TerminalMenu# append.items
		\ set to CMD-SHIFT-ESCAPE
		TerminalMenu# GetMenuHandle 8 0 27 SetMenuItemCommandKey drop
			8 9 TerminalMenu# set.item.modifiers
\			8 1 TerminalMenu# set.item.modifiers
 ))
	
	DrawMenuBar
	TerminalMenu# MENU.SELECTION: DO.TerminalMenu 
;
(( 
27 is ESC
TerminalMenu# GetMenuHandle 1 0 27 SetMenuItemCommandKey .s

TerminalMenu# GetMenuHandle 1 0 20 SetMenuItemCommandKey .s
			1 1 TerminalMenu# set.item.modifiers

TerminalMenu# GetMenuHandle 1 -1 20 SetMenuItemCommandKey .s
			1 1 TerminalMenu# set.item.modifiers

 ))

 
: AdjustTerminalMenuItem ( flag\item -- )
	swap TerminalMenu#  item.enable ( ITEM\ENABLE.FLAG\MENU.ID -- )
;


: AdjustTerminalMenu ( -- )
	TerminalMenu# GetMenuHandle 0= IF exit THEN

\ enabling some menu options depends on a remote connection
\ being available
	ConnectedToRemote?
	dup  6 AdjustTerminalMenuItem
	dup  9 AdjustTerminalMenuItem
	dup 10 AdjustTerminalMenuItem
	dup 12 AdjustTerminalMenuItem
	    13 AdjustTerminalMenuItem

	ConnectedToRemote?
	IF		\ pipes open &  TerminalWindow open
		TerminalWindow front.window =
		IF
			4 true  TerminalMenu#   item.enable
			7 false TerminalMenu#   item.enable
		ELSE
			4 false TerminalMenu#   item.enable		
			7 true  TerminalMenu#   item.enable
		THEN
	ELSE 	\ either no window or not both pipes
		4 false TerminalMenu#   item.enable		
		7 false TerminalMenu#   item.enable
	THEN
;  BeforeMenuSelect LinkLastToken	 \ do it when user clicks in menubar



\ *************************************************************************

: StartRemoteEditor
	SetupTerminalAsEditor
\	['] MonitorInPipe NULLEVENTER LinkToken  ( token\headPtr -- )
	['] MonitorInPipe EVENTER LinkToken  ( token\headPtr -- )
	TerminalMenu
	?OpenWritePipe 0>
	IF
		HostInit
	THEN
;	 
\ RESTORER linklasttoken	<--- Done in DEFAULT.EXTENSIONS

\ StartRemoteEditor


: StartClient		\ for testing only with 2 copies of MF
	SetupTerminalAsClient
	['] MonitorInPipe NULLEVENTER LinkToken  ( token\headPtr -- )
;

\ *************************************************************************

\ more advanced HandlePipeInput handler


0 variable clErr#					\ client error
0 variable clLine#					\ client line#
create clTheWord$  256 allot&erase	\ 
create clSource$  1024 allot&erase	\ 


: clAbort:
	aeol word						\ rest of line is error message
	count upcasepocket place
	\ cr pocket count type			\ <---- TEST ONLY
;

: clTheWord:
	aeol word						\ rest of line is error message
	count clTheWord$ place
	\ cr clTheWord$ count type		\ <---- TEST ONLY
;

: clSource:
	aeol word						\ rest of line is source file path
	count clSource$ place
;


(( 
: FlushIncoming ( -- )	\ slight stall, then throw away received data
	100 ms 
	TryPipeReadAll ?dup
	IF 
		DROP
	\	BigReadBuffer swap s.s s.dump
	THEN
;
 ))

: FlushIncomingLF ( -- )
	\ After the client error handler finishes, it loops around in ((QUIT))
	\ and sends a LF prior to next refill
	\ We want to retain our highliting of the offending word, so
	\ we must purge that LF
	ticks 1000 + { endtime }
	BEGIN
		TryPipeReadAll 0>		\ actually purge anything
		ticks endtime >			\ we don't want to get hung up here, though
		OR
		2 ms
	UNTIL
;


: DisplayAbort ( -- )		\ see ToEditorOnAbort"
	clSource$ c@
	IF
		clTheWord$ clLine# @ upcasepocket  AbortAlert ( badword$ line# c-addr -- )
		clSource$ count se.open.from.path
		IF
			clTheWord$ clLine# @ 2 - FindFrom
		ELSE
			c" Failed to open File:" cr$ "" clSource$
			Big:-(OK# IssueWindowCenteredAlert drop
			clSource$ 40 DUMP
		THEN
	ELSE
		FlushIncomingLF

	\	TermIncludeStartOffset -> IncludeStartOffset
		clLine# @ line# !
		clTheWord$ count pocket place
		upcasepocket clErr# @
		se.interpret.error?		\ handle.se.interpret.error ( abort$\err# -- )
	THEN
;


: DisplayError ( -- )		\ see ToEditorOnError
	clSource$ c@
	IF
		clTheWord$ clLine# @ clErr# @ ErrorAlert ( badword$ line# err# -- )
		clSource$ count se.open.from.path
		IF
			clTheWord$ clLine# @ 2 - FindFrom
		ELSE
			c" Failed to open File:" cr$ "" clSource$
			Big:-(OK# IssueWindowCenteredAlert drop
			clSource$ 40 DUMP
		THEN
	ELSE
		FlushIncomingLF	
	\	TermIncludeStartOffset -> IncludeStartOffset
		clLine# @ line# !
		clTheWord$ count pocket place
		0 clErr# @
		se.interpret.error?		\ handle.se.interpret.error ( abort$\err# -- )
	THEN
;



: DisplayLocation	\	((locate))   -- where is expand done?
	clTheWord$  count  search$ place		\ target name -> search$
	clSource$ count
	clLine# @
	se.locate ( c-addr u line# -- )
;


\ remote commands come in prefixed with a $FF01


variable ShowRemoteCommands ShowRemoteCommands Off

: RemoteCommandHandler ( addr\cnt -- )
	{ addr cnt }	
	ShowRemoteCommands @
	IF
		lf$ count Indata>Terminal
		lf$ count Indata>Terminal
		s" * COMMAND SEQUENCE *" Indata>Terminal
		lf$ count Indata>Terminal
		addr cnt  Indata>Terminal
		lf$ count Indata>Terminal
	THEN

	clErr#     off			clLine#      off
	clTheWord$ off			clSource$    off
	pocket     off			upcasepocket off

	addr cnt  IncludeMem
;

: ParsePipeInput ( addr\cnt -- )
	{ addr cnt }
	
	addr w@
	CASE
		\ prefix with $FF01 means executable commands from remote follow
		$FF01 OF addr 2 +  cnt 2 - RemoteCommandHandler	ENDOF
		\ header is optional for normal data
		$FF00 OF addr 2 +  cnt 2 - Indata>Terminal		ENDOF	\ <FF00> header == data
	\   ,,,,
	
		\ header is optional for normal data
		addr cnt Indata>Terminal
	ENDCASE
	
	clErr# @ 0=
	IF
		#LFsAfterReply
		CASE
			0 OF ""		ENDOF
			1 OF lf$	ENDOF
			lf$lf$ swap
		ENDCASE
		count Indata>Terminal
	THEN
	clErr# off

;

' ParsePipeInput is HandlePipeInput

\ *****************************************************************
\ *****************************************************************
also actels

0 value RemoteContextWL
0 value RemoteFullWL

: InitRemoteWLs ( -- )		\ avoid bad handles from prior snapshot
	0 -> RemoteFullWL
	0 -> RemoteContextWL
; preludes linklasttoken


: EmptyRemoteContextWL ( -- )	\ requires VOCB resource named "Empty Word List"
	RemoteContextWL ?dup 
	IF
		>>Discard
	THEN
	LoadNewWL -> RemoteContextWL
;

: EmptyRemoteFullWL ( -- )	\ requires VOCB resource named "Empty Word List"
	RemoteFullWL ?dup 
	IF
		>>Discard
	THEN
	LoadNewWL -> RemoteFullWL
;




: ScanAllWithXT ( addr cnt char xt -- )		\ xt gets addr\cnt
	{ addr cnt theChar theToken }
	BEGIN
		cnt 0>
	WHILE
		addr 
		dup c@ 0= IF drop exit THEN
		dup cnt theChar scan
		1-  to cnt

		2dup				\  wasAddr  newAddr wasAddr newAddr
		1+ -> addr
		-

		theToken execute	\ cr type
	REPEAT
;




: AddToRemoteContextWL ( addr\cnt -- )
	0 -rot
	RemoteContextWL
	>>AddEntry drop
;




: ImportRemoteContext ( n -- )
\ ticks swap
	dup 16 +  newptrclear
	{ size buff }
	buff size ReadBigChunk drop
\	buff size s.s  2drop
\	....
\ ticks over - s.s  drop

	EmptyRemoteContextWL
	buff size  10 ['] AddToRemoteContextWL ScanAllWithXT
	
	buff DisposePtr
	ReadUntilEmpty		\ Purge ReceivingPipe
\ ticks swap - s.s  drop
;


: FindInRemoteContext ( c$addr -- flag )
	{ | temp[ #256 ] -- }
	RemoteContextWL
	IF
		count temp[ uplace
		temp[ count RemoteContextWL >>FindName
	ELSE
		drop 0
	THEN
;

previous



\ *****************************************************************

\  Set Preferences  

\ ShowRemoteCommands on
ShowRemoteCommands off

\ false -> EchoEscapeChars
true  -> EchoEscapeChars

true  -> One-ShotEscapeMode
\ false  -> One-ShotEscapeMode


1	-> #LFsAfterEnter		\ accepts 0, 1, 2
1	-> #LFsAfterReply		\ accepts 0, 1, 2


\ InsertAtEND -> TerminalInsertionMode
InsertAtSelend -> TerminalInsertionMode



StartRemoteEditor




\ *****************************************************************
\ *****************************************************************
\ TESTING TESTING TESTING TESTING TESTING TESTING TESTING TESTING 
\ TESTING TESTING TESTING TESTING TESTING TESTING TESTING TESTING 
\ TESTING TESTING TESTING TESTING TESTING TESTING TESTING TESTING 
\ TESTING TESTING TESTING TESTING TESTING TESTING TESTING TESTING 
\ TESTING TESTING TESTING TESTING TESTING TESTING TESTING TESTING 
\ TESTING TESTING TESTING TESTING TESTING TESTING TESTING TESTING 
\ TESTING TESTING TESTING TESTING TESTING TESTING TESTING TESTING 
\ *****************************************************************
\ *****************************************************************

create Big1$ ," Now is the time for all good men to come to the aid of their country"
create Big2$ ," The quick brown fox jumped over the lazy dog"


: BigTEst1
	ticks
	{  |  str[ 1000 ] }
	100 0
	DO
\		do.events drop
		
		i 0 <# bl hold # # #s #> str[ place
		Big1$ count str[ append
		str[	count   Indata>Terminal
		LF$		count   Indata>Terminal
		Big2$   count   Indata>Terminal
		LF$		count   Indata>Terminal
	LOOP
	ticks swap - . ."  msec"
;




\ *****************************************************************
\ *****************************************************************
\ *****************************************************************
