\ Using PIPEs as FIFO's
\ Last Revision: 3 Apr 2012  16:35:50  ward@megawolf.com



\ *****************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved.
\ Permission is hereby granted to VFX Forth for Mac customers
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited.
\ *****************************************************************



only forth definitions
have M4TH
[IF]
	anew --mkfifo--
[THEN]

decimal

vocabulary remotewords
also remotewords definitions

((

FIFO buffers on OS X suppsedly auto-resize
Empirically, they are 8192 bytes and stay that way

Speed seems roughly 800-1000 bytes/millisec when read is in
a NullEventer Loop.

READ FIFO returns error if no data to receive
	UNLESS the read port is marked O_NONBLOCK

WRITE FIFO **hangs** if no pending read
	UNLESS the write port is marked O_NONBLOCK
	It still returns error -1 if no read pending


Some sources suggest using fflush to ensure writes are sent immediately.
I have not tried that since we are not opening pipe as stream.



see:
http://developers.sun.com/solaris/articles/named_pipes.html
http://www-cs.canisius.edu/PL_TUTORIALS/C/ADVANCED/ipc
http://linux.die.net/man/7/pipe

))


\ *****************************************************************
\ *****************************************************************


Extern: int mkfifo	( char *path, mode_t mode );
Extern: int __error (void );


: ErrNo ( -- n )
	__error @ ;

: ClearErrNo ( -- )
	0 __error ! ;



8 base !
0666 constant RWXAll
decimal


: (MakeRWPipe) ( 0$ -- result )
	RWXAll  mkfifo 	 			\ create with RW priv's for all
;


: MakeRWPipe ( 0$ -- result )
	ClearErrNo
	(MakeRWPipe)
	IF
		ErrNo 17 =		\ EEXIST /* File exists */
		IF				\ we don't consider that an error
			0
		ELSE
			ErrNo
		THEN
	ELSE
		0
	THEN
	ClearErrNo
;

\ *****************************************************************


true value IsThisTheEditor?


0 value SendingPipeFD
0 value ReceivingPipeFD
0 value WriteMonitorPipeFD		\ this is SendingPipeFD opened for Read


create toEditorName$	z", %IniDir%/toEditorPipe"
create fromEditorName$	z", %IniDir%/fromEditorPipe"


: zExpandName ( zname -- zname' )
	zcount expand  >SyspadZ
;

: CreatePipes ( -- )	\ only need to do this once
\ we NEED INIDIR macro set correctly
\ since out pipes will be made in  ../Library/Application Support/VfxForth/
	IniDir$  count expand s" INIDIR"  replaces
	IniFile$ count expand s" INIFILE" replaces

	toEditorName$   zExpandName MakeRWPipe abort" error creating toEditorPipe"
	fromEditorName$ zExpandName MakeRWPipe abort" error creating fromEditorPipe"
;


: CloseUpPipes ( -- )
	SendingPipeFD      0> IF SendingPipeFD      close drop THEN
	0 -> SendingPipeFD
	ReceivingPipeFD    0> IF ReceivingPipeFD    close drop THEN
	0 -> ReceivingPipeFD
	WriteMonitorPipeFD 0> IF WriteMonitorPipeFD close drop THEN
	0 -> WriteMonitorPipeFD
;

' CloseUpPipes AtExit


\ *****************************************************************
\ Pipe errors, especially trying to write to a Broken pipe
\ trigger a SIGPIPE Signal.  We better catch it or we will crash !
\ We also want to be able to catch a SIGABRT and exit gracefully

defer UserSIGPIPEhandler		\ signum *siginfo *ucontext --
	' 3drop is UserSIGPIPEhandler	\ assign specific action later


: SIGPIPEhandler		\ signum *siginfo *ucontext --
	\ ?? should we force an attempt to reopen after an error
	CloseUpPipes
\	EPIPE __error !		\ Broken pipe
	UserSIGPIPEhandler
;

3 0 callback: SIGPIPECallback		\ signum *siginfo *ucontext --

assign SIGPIPEhandler to-callback  SIGPIPECallback


\ *****************************************************************


: InitializePipes ( -- )
	CreatePipes
	SIGPIPECallback SIGPIPE setSignal	\ pipe errors
	SIGPIPECallback SIGABRT setSignal	\ SIGABRT
\	SIGPIPECallback SIGTERM setSignal	\ SIGTERM
;


: ReceivingPipeNameZ ( -- zaddr )
	IsThisTheEditor?
	IF
		toEditorName$
	ELSE
		fromEditorName$
	THEN
	zExpandName
;


: SendingPipeNameZ ( -- zaddr )
	IsThisTheEditor?
	IF
		fromEditorName$
	ELSE
		toEditorName$
	THEN
	zExpandName
;


4 constant NONBLOCK

: (OpenReadPipe)
	ReceivingPipeNameZ r/o nonblock or 0 open -> ReceivingPipeFD
;


: (OpenWritePipe)
	SendingPipeNameZ w/o nonblock or 0 open -> SendingPipeFD
;

: (OpenWriteMonitorPipe)	\ WRITE pipe opened for read ;-)  See Below
	SendingPipeNameZ R/O nonblock or 0 open -> WriteMonitorPipeFD
;



: ?OpenReadPipe ( -- FD )	\ multiple opens can give multiple FD's
	ReceivingPipeFD 0> 0=	\ so be sure we only do open if we NEED to
	IF
		(OpenReadPipe)
	THEN
	ReceivingPipeFD
;

: ?OpenWritePipe ( -- FD )	\ multiple opens can give multiple FD's
	SendingPipeFD 0> 0=		\ so be sure we only do open if we NEED to
	IF
		(OpenWritePipe)
	THEN
	SendingPipeFD
;


: ?OpenWriteMonitorPipe ( -- FD )	\ multiple opens can give multiple FD's
	WriteMonitorPipeFD 0> 0=		\ so be sure we only do open if we NEED to
	IF
		(OpenWriteMonitorPipe)
	THEN
	WriteMonitorPipeFD
;

\ *****************************************************************

0 value BigReadBuffer		\ OpenPipes will ALLOCATE this on setup

: InitPipes ( -- )			\ clear old vakues on startup
	0 -> SendingPipeFD
	0 -> ReceivingPipeFD
	0 -> WriteMonitorPipeFD
	0 -> BigReadBuffer
;
' InitPipes AtCold



: OpenPipes ( -- )
	?OpenReadPipe  drop
	?OpenWritePipe drop
\	?OpenWriteMonitorPipe drop	\ we do this ONLY when needed (see below)
	BigReadBuffer 0=
	IF 10:000 allocate drop -> BigReadBuffer THEN
	\ 8182 shold be more than enough, but hey, heap memory is cheap, right?
;


\ *****************************************************************

: DataAvailable? ( -- NN )	\ maybe return -1 if not open or error ?
	{ | nread[ cell ] }
	nread[ off  ?OpenReadPipe dup 0>
	IF
		fionread nread[ ioctl 	\ Get the number of bytes that are immediately available for read
  		-1 =
  		IF
  			0
  		ELSE
  			nread[ @
  		THEN
  	ELSE
  		drop 0
  	THEN
;


: (TryPipeRead) ( ^buff\maxCnt -- #read )	\ return -1 if error, -2 if cant open
	?OpenReadPipe dup 0>
	IF
		-rot  read
	ELSE
		2drop
		-2
	THEN
;


: TryPipeReadAll ( -- #read )		\ return -1 if receive error
	BigReadBuffer 10:000 (TryPipeRead)
((
	?OpenReadPipe dup 0>
	IF
		BigReadBuffer 10:000 read
	ELSE
		drop 0
	THEN
 ))
;


((
: TryPipeWrite ( addr\cnt -- #written )		\ return -1 if cannot open or error
	?OpenWritePipe ?dup
	IF
		-rot write
	ELSE
		2drop
		-2				\ cr ." No Receiver"
	THEN
;
 ))

: (TryPipeWrite) ( addr\cnt -- #written )	\ return -1 if error, -2 if cant open
	?OpenWritePipe dup 0>
	IF
		-rot  write
	ELSE
		2drop
		-2
	THEN
;



1000 value WriteMaxTime		\ ms to wait for write completion

: KeepTryingWrite ( addr\cnt -- result )	\ >=0 ==> #written,  neagtive = error
	ticks WriteMaxTime +
	-3								\ initalze Result with timeout
	{ addr cnt endticks RES }

	BEGIN
		-3 -> RES
		?OpenWritePipe dup 0>
		IF
			addr cnt write
			dup -> RES
			0<
		ELSE
			drop
			-2 -> RES		 		\ pipe open failure
			true
		THEN
		endticks ticks >
		AND
	WHILE
		emptyidle
		2 ms
	REPEAT
	RES
;


1000 value BigWriteWait


: WriteBigChunk ( addr/cnt -- total#written )
\ affected by receive speed
\ 130 ticks for 100,000 bytes is typical with dedicated receiver
	dup ticks BigWriteWait +
	{ addr cnt reqCnt endTime }
	BEGIN
		emptyidle
		ticks endTime <
		cnt 0>
		AND
	WHILE
		addr cnt 8000 min (TryPipeWrite)
		dup 0>
		IF
			dup    +to addr
			negate +to cnt
			ticks BigWriteWait + -> endTime
		ELSE
			drop
			2 ms
		THEN
	REPEAT
	reqCnt  cnt -
;


\ *****************************************************************

\ A "trick" using a read FD on the write pipe to see if it is empty
\ A side effect is that Write no longer blocks if there is no
\ other program with the pipe opened in read mode
\ So we normally keep WriteMonitorPipe closed

: #StillInWritePipe ( -- n )
	{  | nread[ cell ] }
	nread[ off  ?OpenWriteMonitorPipe dup 0>
	IF
		FIONREAD nread[ ioctl 	\ Get the number of bytes that are available
  		-1 =
  		IF  0  ELSE  nread[ @  THEN
	ELSE
		drop
		0
	THEN
;

: WaitWritePipeEmpty ( -- )
	ticks WriteMaxTime +
	{ endticks }
	BEGIN
		#StillInWritePipe 0>
		ticks endticks <
		AND
	WHILE
		1 ms
		emptyidle
	REPEAT
\ keep WriteMonitorPipe closed most of the time
	WriteMonitorPipeFD 0> IF WriteMonitorPipeFD close drop THEN
	0 -> WriteMonitorPipeFD
;

\ *****************************************************************

: ReadUntilEmpty ( -- )		\ flush any unread bytes in ReceivingPipe
	BEGIN
		emptyidle
		2 ms
		DataAvailable?  0>
	WHILE
		ReceivingPipeFD BigReadBuffer 8000 read drop
	REPEAT
;
\ *****************************************************************
((

: WriteBigChunk ( addr/cnt -- )
	{ addr cnt }
	BEGIN
		cnt 8000 >
	WHILE
		addr 8000 PoliteWrite
		8000  +to addr
		-8000 +to cnt
	REPEAT
	addr cnt PoliteWrite
;
 ))


1000 value BigReadWait

: ReadBigChunk ( addr/cnt -- total#read )
	dup ticks BigReadWait +
	{ addr cnt reqCnt endTime }
	BEGIN
		emptyidle
		endTime ticks >
		cnt 0>
		AND
	WHILE
		addr cnt 8000 min (TryPipeRead) ( ^buff\maxCnt -- #read )
		dup 0>
		IF
			dup    +to addr
			negate +to cnt
			ticks BigReadWait + -> endTime
		ELSE
			drop
			2 ms
		THEN
	REPEAT
	reqCnt  cnt -
;


\\ *****************************************************************


((

: WrPoll ( -- #desc\events )
	POLLWRBAND POLLOUT OR  SendingPipeFD
	FDPoll ( theEvents\thePipe-- #desc\events )
;

: RdPoll ( -- #desc\events )
	POLLIN  ReceivingPipeFD
	FDPoll ( theEvents\thePipe-- #desc\events )
;

 ))


\	#define	O_RDONLY	0x0000		/* open for reading only */
\	#define	O_WRONLY	0x0001		/* open for writing only */


((
#define	O_NONBLOCK	0x0004		/* no delay */
#define	O_APPEND	0x0008		/* set append mode */
#ifndef O_SYNC		/* allow simultaneous inclusion of <aio.h> */
#define	O_SYNC		0x0080		/* synch I/O file integrity */
#endif
#if !defined(_POSIX_C_SOURCE) || defined(_DARWIN_C_SOURCE)
#define	O_SHLOCK	0x0010		/* open with shared file lock */
#define	O_EXLOCK	0x0020		/* open with exclusive file lock */
#define	O_ASYNC		0x0040		/* signal pgrp when data ready */
#define	O_FSYNC		O_SYNC		/* source compatibility: do not use */
#define O_NOFOLLOW  0x0100      /* don't follow symlinks */
#endif /* (_POSIX_C_SOURCE && !_DARWIN_C_SOURCE) */
#define	O_CREAT		0x0200		/* create if nonexistant */
#define	O_TRUNC		0x0400		/* truncate to zero length */
#define	O_EXCL		0x0800		/* error if already exists */
#define	FMARK		0x1000		/* mark during gc() */
#define	FDEFER		0x2000		/* defer for next gc pass */
#define	FHASLOCK	0x4000		/* descriptor holds advisory lock */
#if !defined(_POSIX_C_SOURCE) || defined(_DARWIN_C_SOURCE)
#define	O_EVTONLY	0x8000		/* descriptor requested for event notifications only */
 ))

((

/*
 * Requestable events.  If poll(2) finds any of these set, they are
 * copied to revents on return.
 */
#define	POLLIN		0x0001		/* any readable data available */
#define	POLLPRI		0x0002		/* OOB/Urgent readable data */
#define	POLLOUT		0x0004		/* file descriptor is writeable */
#define	POLLRDNORM	0x0040		/* non-OOB/URG data available */
#define	POLLWRNORM	POLLOUT		/* no write type differentiation */
#define	POLLRDBAND	0x0080		/* OOB/Urgent readable data */
#define	POLLWRBAND	0x0100		/* OOB/Urgent data can be written */

/*
 * These events are set if they occur regardless of whether they were
 * requested.
 */
#define	POLLERR		0x0008		/* some poll error occurred */
#define	POLLHUP		0x0010		/* file descriptor was "hung up" */
#define	POLLNVAL	0x0020		/* requested events "invalid" */



structure pollfd
	long:    +pollfd-fd
	short:   +pollfd-events
	short:   +pollfd-revents
structure.end

extern int poll (struct pollfd *, nfds_t, int) __DARWIN_ALIAS_C(poll);

 ))

((
extern: UInt32 res poll ( UInt32 * pollfdArray, UInt32 #FDs, UInt32 timeout );


: FDPoll ( theEvents\thePipe-- #desc\events )
	{ | pollfd[ 8 ] }
	pollfd[  !				\ +pollfd-fd
	pollfd[  4 + w!			\ +pollfd-events

	pollfd[ 1 0 poll
	pollfd[  6 + w@			\ +pollfd-revents
;

 ))



((
http://linux.die.net/man/7/pipehttp://linux.die.net/man/7/pipe

POSIX.1-2001 says that write(2)s of less than PIPE_BUF bytes must be atomic: the
output data is written to the pipe as a contiguous sequence.  Writes of more
than PIPE_BUF bytes may be nonatomic: the kernel may interleave the data with
data written by other processes.  POSIX.1-2001 requires PIPE_BUF to be at least
512 bytes.  (On Linux, PIPE_BUF is 4096 bytes.) The precise semantics depend on
whether the file descriptor is nonblocking (O_NONBLOCK), whether there are
multiple writers to the pipe, and on n, the number of bytes to be written:

O_NONBLOCK disabled, n <= PIPE_BUF
All n bytes are written atomically; write(2) may block if there is not room for
n bytes to be written immediately

O_NONBLOCK enabled, n <= PIPE_BUF
If there is room to write n bytes to the pipe, then write(2) succeeds
immediately, writing all n bytes; otherwise write(2) fails, with errno set to
EAGAIN.

O_NONBLOCK disabled, n > PIPE_BUF
The write is nonatomic: the data given to write(2) may be interleaved with
write(2)s by other process; the write(2) blocks until n bytes have been written.

O_NONBLOCK enabled, n > PIPE_BUF
If the pipe is full, then write(2) fails, with errno set to EAGAIN.  Otherwise,
from 1 to n bytes may be written (i.e., a "partial write" may occur; the caller
should check the return value from write(2) to see how many bytes were actually
written), and these bytes may be interleaved with writes by other processes.


))

