\ Numbering&Time.fth
\ Created:       3 May 2012  08:49:42  ward@megawolf.com
\ Last Revision: 3 May 2012  08:49:42  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\    to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


(* *************************************************************************

Some  miscellaneous useful utility words
	#,###
	#,###.#

	TIMER-RESET  .ELAPSED
	TIME.FOR
	N.TIMES.FOR

************************************************************************* *)

Only Forth Definitions

\ anew --Numbering&Time.fth--

>M4TH
Decimal


\ Format numbers in 3-digit groups.
\ from Greg Guerin


\ DP-CHARS    w@ == radix point
\ DP-CHARS 2+ w@ == digit group (thousands) separator


: #,###  ( D -- addr\count )  	\ format double number in comma'ed format
	base @  dup dup * *  1-		\ compute value of 3-digit group for current base
	locals|  #999  |
	dup 0< >R
	dabs
	<#	BEGIN
			2dup  #999 s>d D>	\ actually 1000 in current BASE, minus 1
		WHILE
			# # # 
			dp-chars 2 + w@  hold	\ thousands separator
		REPEAT
		#s  R> sign
	#>  ;
(( 
MICROSECS #,### type
16,088,739,095
ok
 ))



: #,###.#  ( n -- addr\count )  \ format double number in comma'ed format
	base @  dup dup * *  1-		\ compute value of 3-digit group
	locals|  #999  |
	dup 0<  >R
	dabs
	<#	#  dp-chars w@ hold		\ radix point
		BEGIN
			2dup #999 s>d D>	\ actually 1000 in current BASE, minus 1
		WHILE
			# # #
			dp-chars 2 + w@  hold	\ thousands separator
		REPEAT
		#s  R> sign
	#>  ;

(( 
1599510049.6  #,###.# type
1,599,510,049.6
ok
 ))


\ *********************************************************
\ *********************************************************

\ Some Time Interval words

: MIL ( n -- n*1,000,000 )	\ handy shortcut 
	1:000:000 * ;


Lacking POWER10 : POWER10  ( n -- 10^n ) 0max 1 swap #DO 10 * LOOP ;

: CHOP-INSIGNIFICANTS  ( u1\n1 -- u2\n2 )
	power10 0 locals| zeros most |
	BEGIN
		dup most u< 0=
	WHILE
		dup most 10 * u< IF 5 + THEN	\ most sig zapped digit, so round up
		0 10 um/mod nip  1 +to zeros
	REPEAT
	zeros ;


: .MICROSECONDS  ( u -- )	\ display 3 significant digits of time
	3 chop-insignificants locals| zeros |
	zeros 0= IF . ." s"
	ELSE	\ we have three significant digits
		0 <# bl hold
			# zeros 3 mod 2 = IF ascii . hold THEN
			# zeros 3 mod 1 = IF ascii . hold THEN
			#
		#> type
		zeros 4 < 
		IF 
			." ms" 
		ELSE
			zeros 7 < 
			IF
				." s" 
			ELSE
				." ks"
			THEN
		THEN
	THEN ." ec" ;


: ShowMicroseconds ( d -- )		\ formatted dump of microseconds
								\ with attempt at prettyprinting
	base @ >R
	decimal
	2dup  999:000:000. du<	
	IF	\ less than 1000 sec
		d>s .microseconds
	ELSE
		500:000. d+				\ round
		1:000:000 mu/mod
		#,### type ."  sec"
		drop ( remainder )
	THEN
	R> base !  ;


: (ShowElapsed) ( d -- ) 		\ pass old value of microsecs on the stack
	microsecs 2swap D-  ShowMicroseconds  ;

: ShowElapsed ( d -- ) 		\ pass old value of microsecs on the stack
	microsecs 2swap D- cr ShowMicroseconds ;



create StartTime 0 , 0 ,

: TIMER-RESET ( -- )
	microsecs StartTime 2! ;

: .ELAPSED ( -- )
	StartTime 2@ (ShowElapsed) ;


: TIME.FOR  ( "rest-of-line" -- )
	10 parse 
	timer-reset
	evaluate
	.elapsed
;


((
\ try an empty line
time.for
1 sec
ok


time.for MakeWLFromContext
10.6 msec
ok

time.for  3456 7896 12 */ drop
26 sec
ok

))


\ CAREFUL - you should use this on stack neutral phrases
\ NOTE - cannot be cancelled unless one of the words you are testing
\		 does DO.EVENTS or DO.IO.EVENTS
: N.TIMES.FOR ( n --  ) \  time "rest-of-line"
	10 parse 		\ get rest of the line
	{ n addr cnt }
	timer-reset
	n 0
	?DO
		addr cnt evaluate
	LOOP
	.elapsed
;

((
For example, compare some multiply routine times

1 mil n.times.for 100 500 UM* 2drop
2.48 sec

1 mil n.times.for 100 500 m* 2drop
2.56 sec
ok

1 mil n.times.for 100 500 * drop
2.66 sec

8 set-precision
1 mil n.times.for 100.e 500.e f* f>d 2drop
3.64 sec

10 set-precision
1 mil n.times.for 100.e 500.e f* f>d 2drop
3.69 sec

18 set-precision
1 mil n.times.for 100.e 500.e f* f>d 2drop
3.74 sec
))




stopincluding *********************************************************
******************* comments and/or test code below *******************
***********************************************************************
