\ This file contains several extensions for manipulating
\ controls.
\
\ You can load this file and use it if you like, or just
\ refer to it when writing applications using controls.
\
\ In the examples used in the Level 2 manual, we will use these
\ extensions, so be sure to include the file before executing
\ those examples.
\
\ Last Revision: 25 Feb 2012  10:04:50  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************

Only Forth definitions
decimal

\ anew CONTROLS

>m4th

lacking shaperect  myfolder mfinclude" MacForth Graphics.fth"



: COMPUTE.CRECT  ( x\y\title$ -- ^rect)
    locals| CTITLE CY CX |
    get.textfont >r  get.textsize >r 
    0 textfont 13 textsize	\ Chicago 13 better size for OSX
    cx cy 2dup ctitle stringwidth 24 + 20 +point    \ l\t\r\b  6.049
    ShapeRect !OrderedRect  ShapeRect
    r> textsize  r> textfont  ;


: BINARY.CONTROL  ( wptr\x\y\title$\visible\type -- ControlRef )
    locals| CTYPE CVIS CTITLE CY CX CWPTR | 
    CX CY ctitle  compute.crect
    cwptr @ ( CARBON ) swap   ctitle  cvis  0  0  1  ctype  here
    ( wptr\rect.addr\title$\visible\min\val\max\button.type\refcon)
    NewControl ;


: CHECK.BOX ( wptr\x\y\title$\visible -- )
     1 BINARY.CONTROL VALUE ;

: RADIO.BUTTON ( wptr\x\y\title$\visible -- )
     2 BINARY.CONTROL VALUE ;


: PUSH.BUTTON ( wptr\x\y\title$\visible -- )
     0 BINARY.CONTROL VALUE ;


: LAST.PART  ( -- part code | returns value in THIS.PART )
    THIS.PART @  ;


: FOLLOW.MOUSE  ( -- flag | follows the mouse while the button )
        ( is down in the control area specified by the handle  )
        ( in THIS.CONTROL.  Flag true if button released in    )
        ( same control it was initially pressed in.            )
     LAST.CONTROL  @MOUSE  TRACK.CONTROL   ;


: IN.CONTROL?  ( -- flag | synonym for ?IN.CONTROL )
    ?IN.CONTROL ;


: @VPOS  ( -- n | returns the current value of the control for )
             ( the vertical scroll bar for the current window. )
    GET.WINDOW +VBAR @  GetControl32BitValue  ;


: @HPOS  ( -- n | returns the current value of the control for )
           ( the horizontal scroll bar for the current window. )
    GET.WINDOW +HBAR @ GetControl32BitValue ;


\ NOTE: These words will update the value of the control,
\ but WILL NOT udpate the appearance of the scroll bar.
\ Use UPDATE.CONTROL to update the scroll bar.

: !VPOS  ( n -- | stores n as the current value of the       )
         ( vertical scroll bar value for the current window. )
    GET.WINDOW  +VBAR @  SWAP ( OVER <CLIP> ) SetControl32BitValue  ;


: !HPOS  ( n -- | stores n as the current value of the         )
         ( horizontal scroll bar value for the current window. )
    GET.WINDOW  +HBAR @  SWAP (  OVER <CLIP> ) SetControl32BitValue  ;


: IN.SCROLL.BAR?  ( -- [0] or [scroll bar type\part code\-1] )
    @MOUSE.DN  GET.WINDOW  FIND.CONTROL DUP 
     IF
       >R  ( save part code on rstack )
       CASE  GET.WINDOW +HBAR @  ( horizontal scroll bar?)
                      OF   SCROLL.LEFT/RIGHT R> TRUE   ENDOF
             GET.WINDOW +VBAR @  ( vertical scroll bar? )
                      OF   SCROLL.UP/DOWN    R> TRUE   ENDOF
         	 R> DROP FALSE  SWAP
       ENDCASE
     ELSE
     	drop		\ 6.05437
     THEN  ;



\ Controls defining structure

structure <CFIELD>
       long:  +C.HANDLE		\ with carbon, this a ControlRef, NOT a handle !
    2 longs:  +C.BOUNDS   \ tlbr ; tl = br if C.POSITIONed
      short:  +C.VALUE
      short:  +C.VISIBLE
      short:  +C.MAX
      short:  +C.MIN
       long:  +C.TYPE
       long:  +C.REFCON
       long:  +C.TITLE    \ offset from DBP
structure.end

 0 CONSTANT A.PUSH.BUTTON
 1 CONSTANT A.CHECK.BOX
 2 CONSTANT A.RADIO.BUTTON
16 CONSTANT OLD.SCROLL.BAR			\ 6.05437
: A.SCROLL.BAR ScrollBarType ;		\ 6.05437

8 mfFIELD USEWFONT
\ Makes a control whose title is written in whatever font is
\ the current one for its window.  C.POSITION is unlikely to give a
\ satisfactory rect for such controls, as its computation is based on
\ Chicago 12.  Use C.BOUNDS to specify a better one.
\ Example
\    A.PUSH.BUTTON USEWFONT MY.CONTROL C.TYPE


create OK$	," Ok" 			\ default control name

: NEW.CONTROL  ( -- | create a new control array )
    CREATE
      0 ,        ( initial control handle     )
      0 , 0 ,    ( default bounds rect        )
      0 W,       ( default value              )
      TRUE W,    ( default visible flag       )
      1 W, 0 W,  ( default max, min           )
      0 ,        ( default type = push button )
      0 ,        ( default refcon             )
      OK$  ,  	 ( default string    )
;


\   Control Field Specifiers

: C.BOUNDS ( t\l\b\r\cfield --  )  5 NEEDED  +C.BOUNDS !RECT  ;

: C.POSITION  ( x\y\cfield -- )
    locals| CF CY CX |
    CX CY CF +C.BOUNDS 4 + !POINT
    CX CY CF +C.BOUNDS     !POINT  ;

: C.VISIBLE  ( flag\cfield -- ) 2 NEEDED   +C.VISIBLE W!  ;

: C.VALUE  ( value\cfield -- )  2 NEEDED   +C.VALUE   W!  ;

: C.RANGE  ( min\max\cfield -- | assign range to cfield )
    locals| CF CMAX CMIN |
    CMIN CF +C.MIN W!
    CMAX CF +C.MAX W!  ;


: C.TYPE  ( type\cfield -- )  2 NEEDED   +C.TYPE !    ;

: C.TITLE  ( $\cfield -- | assign $ as title )
    locals| CF CTITLE |
    CTITLE  CF +C.TITLE !  ;


\  Appending controls

: <DO.ADD.CONTROL>  ( wptr4\cfield -- control handle )
    locals| CF |  1 ( more ) NEEDED
    +windowref#	   @ ( CARBON )
    CF +C.BOUNDS      ( rect    )
    CF +C.TITLE    @  ( title )
    CF +C.VISIBLE W@  ( vis flag)
    CF +C.VALUE   W@  ( value   )
    CF +C.MIN     W@  ( min     )
    CF +C.MAX     W@  ( max     )
    CF +C.TYPE     @  ( type    )
    CF +C.REFCON   @  ( ptr     )
    NewControl  ;


: <DO.BIN.CONTROL>  ( wptr4\cfield -- control handle )
    locals| CF |  1 ( more ) NEEDED
    CF +C.BOUNDS 2@ =		\  was control defined with C.POSITION ?
    IF  cf +c.bounds @point  ( x\y )
        cf +c.title  @ 
        compute.crect
        @rect   
        cf +c.bounds !rect     \ compute the +crect
    THEN
    CF <DO.ADD.CONTROL>  ;


: APPEND.CONTROL  ( wptr4\cfield -- )
    locals| CF |   1 ( more ) NEEDED
    CF +C.TYPE @  3 <    ( add control based on type )
    IF    CF  <DO.BIN.CONTROL>
    ELSE  CF  <DO.ADD.CONTROL>
    THEN  CF  +C.HANDLE !    ;


: ON.CONTROL   ( ControlRef -- )
	1 SetControl32BitValue  ;

: OFF.CONTROL  ( ControlRef -- )
	0 SetControl32BitValue  ;


: TOGGLE.CONTROL  ( ControlRef -- toggle control )
	dup GetControl32BitValue 0= negate SetControl32BitValue  ;


: SET.CONTROL.RANGE  ( ControlRef\min\max -- )
	locals| CMAX CMIN ctrl |
	ctrl cmax SetControl32BitMaximum
	ctrl cmin SetControl32BitMinimum ;
