\ Carbon Menu Event Handler
\ Last Revision: 13 Apr 2012  11:22:57  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************

(( 

Classic MacForth detected CMD-key down during keystroke events in 
order to trigger a MenuSelect action.

Apple now allows alternate modifer keys for menu actions, like shift,
cpmmand, option, and combinations.

In order to support those, this file adds support for a Carbon Event
handler for Menu Events.

This should not impact anyone unless you were already doing some exotic things
with menu handling or command keys.
I am not sure why anyone would want to remap Shift-X to the menubar instead of
a capital letter, but who knows? It is allowed and we now support it. 


For now, I am still using >MENUBAR to track the mouse and >CMD.KEY to handle
command key events, However I have removed EXECUTE.MENU from >MENUBAR, so >MENUBAR,
only does the tracking.  I have replaced its menu item execution with
a kEventClassCommand Carbon Event handler which handles every type of menu
selection EXCEPT command keys for now. 
	
	
	See "/MF Examples/Menu Modifier Key test.fth" for example of usage

 ))



only forth definitions
\ anew --menuevent--
>m4th


DEFER CommandEventHandler ( EventHandlerCallRef\EventRef\userData -- OSErr )
	' 2drop is CommandEventHandler


3 1  CALLBACK: CommandEventHandlerProcPtr
	' CommandEventHandler to-callback CommandEventHandlerProcPtr


0 value  CommandEventHandlerUPP


variable TheCommandEventHandlerRef  TheCommandEventHandlerRef off

create CommandEventHandlerList
\ see discusssion below - can also be used to catch control events
	ascii cmds  ,   	\ kEventClassCommand
	1 ,					\ kEventCommandProcess


: CleanUpCommandEventHandler
	TheCommandEventHandlerRef @ ?dup IF RemoveEventHandler drop THEN
	CommandEventHandlerUPP      ?dup IF DisposeEventHandlerUPP  THEN
	['] execute.menu is >execute.menu
	['] >CMD.KEY   16 >event !
	['] pause      28 >event !
;  on.forget CleanUpCommandEventHandler



: InstallCommandEventHandler ( -- )
	CommandEventHandlerProcPtr NewEventHandlerUPP  -> CommandEventHandlerUPP

	GetApplicationEventTarget	\ EventTargetRef
	CommandEventHandlerUPP		\ EventHandlerUPP
	1							\ NumTypes
	CommandEventHandlerList		\ ^TypeList
	0							\ UserData
	TheApplicationHandlerRef	\ ^EventHandlerRef
	InstallEventHandler 
	drop
	\ abort" InstallEventHandler Failed"
;	PRELUDES linklasttoken



: <CommandEventHandler ( EventHandlerCallRef\EventRef\userData -- OSErr )
	{ EventHandlerCallRef  EventRef  userData  |  dirObj[ 16 ] }
	ascii mctx  EventRef GetLongEventParameter  \ kEventParamMenuContext      = 'mctx',
	\ get  Menu context flags indicating what type of menu was being tracked. 
	\ GetMenuType would be useless as it only works while the menu is open
	1 and 0= 		\ 1 is in menubar - other values for popup, menus embedded in controls, etc
	IF 				\ we only want this handler to work on menubar menus
		EventNotHandledErr
		exit 
	THEN

	EventRef GetEventModifiers
		\ see Modifier? for modifier key flags
	dup
	256 and 0= 			\ NOT a command key - could be another control key or direct menu selection
	over 4096 and OR	\ control key remaps keys to out of range that WHICH.KEY.EVENT remaps to >CMD.KEY
	IF
		EventRef  
		ascii ----  	\ EventParamName = kEventParamDirectObject = '----'
		ascii hcmd		\ EventParamType = kEventParamHICommand  = typeHICommand*/
		0				\ upcasepocket	\ -> ^ActualEventType
		16  			\ MaxBufferSize
		0				\ upcasepocket 8 +  @  --> 14 bytes == ^ActualSize
		dirObj[			\ UInt32 * ^outData
		GetEventParameter drop
		dirObj[ 8  + @			\ --> menuhandle
		\ modifiers\menuref		 on stack
			GetMenuID			\ where     = menu id
		0 max#menus range		\ 10.44 -> This handler can get events from the SYSTEM
								\ items under the Application Menu
								\ We DO NOT WANT THEM !!!!
		IF
			dirObj[ 12 + w@			\ message   = item#
			28 						\ private event 28 in event.table =  Carbon Menu Event
			EventQSemaphore GET				\ 
			MakePrivateEvent ( Modifiers\Where\Message\Event# -- )
			scratch.event.record PostPrivateEvent 
			EventQSemaphore RELEASE			\ 
			\ remove EXECUTE.MENU from >MENUBAR is we use this
	
			0						\ we handled the event
		ELSE	\ a weird menuid would mean this is the ystem part of application menu
			2drop
			EventNotHandledErr		\ pass the event along to next handler
		THEN
	ELSE						\ it was a command key, so let the old method handle it  >CMD.KEY
		drop
		EventNotHandledErr		\ pass the event along to next handler
	THEN
;
' <CommandEventHandler  is CommandEventHandler



: DoCarbonMenuSelect ( 28 -- 28 )
	\ new private event handles the menu execution
\	s.cr s" DoCarbonMenuSelect" s.type s.cr
	event.record 2+ w@			\ message = item#
	event.record 10 + @			\ where   = menu id
	execute.menu  ( item\menu id )
;
' DoCarbonMenuSelect  28 >event !




\\

\ kEventClassCommand            = 'cmds',
/*
 *  kEventClassCommand / kEventCommandProcess
 *  
 *  Summary:
 *    A command has been invoked and the application should handle it.
 *  
 *  Discussion:
 *    This event is sent when the user chooses a menu item or clicks a
 *    control. Any menu item selection, from either the menubar, a
 *    popup or contextual menu, or a popup or bevel button control,
 *    will cause this event to be sent; if the menu item does not have
 *    a command ID, the commandID field of the HICommand parameter will
 *    be zero, but the event will still contain a valid MenuRef and
 *    MenuItemIndex. Controls will send this event only if the control
 *    has a non-zero command ID set with SetControlCommandID. 
 *    
 *    Some senders of this event will also include the modifier keys
 *    that were pressed by the user when the command was invoked, but
 *    this parameter is optional and may not be present in all
 *    instances of this event. 
 *    
 *    When a command event is sent from a menu, a MenuContext parameter
 *    will be included on Mac OS X 10.2 and CarbonLib 1.6, indicating
 *    whether the command was sent from a menu in the menubar
 *    (kMenuContextMenuBar will be set) or from a popup menu
 *    (kMenuContextMenuBar will not be set). The MenuContext parameter
 *    also indicates whether the event was sent by a selection of a
 *    menu item with the mouse (kMenuContextMenuBarTracking or
 *    kMenuContextPopUpTracking will be set), or by a command key press
 *    (kMenuContextKeyMatching will be set). If you need to determine
 *    the source of a command event on a Carbon release prior to Mac OS
 *    X 10.2 or CarbonLib 1.6, we recommend installing a
 *    kEventMenuEnableItems handler on each menu and caching the
 *    MenuContext parameter in the EnableItems event in a menu
 *    property; when you receive the CommandProcess event, look up the
 *    cached MenuContext. 
 *    
 *    It is essential that your event handler for this event return
 *    eventNotHandledErr for any command events that you do not handle,
 *    especially for commands that are sent from menus attached to
 *    popup or bevel button controls; if, for example, you return noErr
 *    for command events sent in response to a menu selection from a
 *    popup button, the Menu Manager will return zero from
 *    PopUpMenuSelect, and the popup button control will not know that
 *    an item was selected from the popup menu and will not redraw with
 *    the new selection.
 *  
 *  Mac OS X threading:
 *    Not thread safe
 *  
 *  Parameters:
 *    
 *    --> kEventParamDirectObject (in, typeHICommand )
 *          The command to be handled.
   kEventParamDirectObject       = '----', /* type varies depending on event*/
*    
 *    --> kEventParamKeyModifiers (in, typeUInt32)
 *          The keyboard modifiers that were pressed when the command
 *          was invoked by the user. This parameter is optional and may
 *          not be present in all instances of this event.
   kEventParamKeyModifiers       = 'kmod', /* typeUInt32*/
*    
 *    --> kEventParamMenuContext (in, typeUInt32)
   kEventParamMenuContext        = 'mctx', /* typeUInt32*/
*          If the command was contained in a menu, information about
 *          the menu. This parameter is optional and may not be present
 *          in all instances of this event. Available in Mac OS X 10.2
 *          and CarbonLib 1.6, and later.
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in Carbon.framework
 *    CarbonLib:        in CarbonLib 1.1 and later
 */
