\ TOOLS MENU
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\    to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


only Forth definitions
>M4TH

decimal


09 value ToolsMenu#


lacking ToolVocab  vocabulary ToolVocab
also ToolVocab definitions


63 constant Max#Tools


create ToolTokens		\ Contains tokens for Tools menu selection  
	Max#Tools 1+  cells   allot&erase


: ^ToolToken  ( n -- addr | return address of word containing token n )
	cells ToolTokens + ;


: ToolCount  ( -- n | Return number of items in Tools menu )
	ToolsMenu# CountMenuItems ;


: ToolsSelection ( item# -- | Execute token for selected item )
	0 HiliteMenu
	1 ToolCount range
	if
		^ToolToken @execute
	then ;

>m4th
also toolvocab

: Tools.Menu  ( --- | Install the Tools menu )
	ToolsMenu# delete.menu
	0  c" Tools"  ToolsMenu#  new.menu
	false ToolsMenu# menu.enable
	DrawMenuBar
	ToolsMenu# menu.selection: ToolsSelection ;

definitions


: NewToolItem  ( token\item$ -- n | Install new tool in the menu. )
\		token will be executed when the item is selected.
\		item$ addresses a string containing the item, with any modifiers
\		n is the item's number within the menu
	locals| item$ theToken |
	ToolCount 1+ Max#Tools > abort" Too many tools for menu."
	ToolCount 0= IF true ToolsMenu# menu.enable THEN
	ToolCount 1+ true ToolsMenu# item.enable	\ Work around prune of disabled
	item$ ToolsMenu# append.items			\ Adjust the menu
	theToken ToolCount ^ToolToken !		\ Set up item's selection token
	ToolCount ;							\ Return item number


: NewToolGroup  ( #items -- n\TRUE  or  FALSE | Start a new group of tools.)
\	If room for #items, return n under TRUE, else return FALSE
\		n is item number of the separator line (0 for first group)
	ToolCount +								\ Old count plus new # items
	ToolCount if 1+ then					\ Plus 1 for separator
	Max#Tools > 0= dup
	IF
		ToolCount 
		IF
			0 C" (-" NewToolItem		\ Install separator, leaves #
		ELSE
			0							\ Separator's "item number"
		THEN
		swap							\ n\TRUE
	THEN ;


: PruneToolItem  ( n -- | Remove item n and all later items )
	0 ToolCount range  					\ n out of range would destroy heap
	IF
		ToolCount
		?DO
			i ToolsMenu# del.menu.item
		-1 +LOOP
		ToolCount 0= if false ToolsMenu# menu.enable then
	ELSE
		drop
	THEN ;



: TEST_DUMMY ( -- | dummy token for use when nothing else is available )	
	beep
	c" Use as:   'TEST <xxx>'" 
	cr$ cr$
	c" to make this menu item execute <xxx>"
	Big:-(OK# IssueWindowCenteredAlert drop ; 


create TEST.TOKEN  ' TEST_DUMMY ,			\ holds token to be executed

create TEST.ToolGroup -1 , 					\ holds ID of this tool group

0 value TestItem#							\ item# for use in Tools menu



: TEST.IT ( -- | execute contents of TEST.TOKEN )
	0 hilitemenu
	test.token @ ?dup
	IF execute ELSE beep THEN ;



ascii \ value TestItToolHotkey

: FIX.TEST.NAME ( -- | set menu item to match the name of the assigned token )
	ToolsMenu# menu.handle		\ is there a menu ??
	IF
		TestItem#
		test.token @ >name
		?dup  0= 
		IF
			c" Test 'IT'" 
		THEN		\ if it has no name, then use 'IT'
		ToolsMenu#  set.item$
		
		TestItToolHotkey
		IF
			ToolsMenu# -menu.handle TestItem#  0 TestItToolHotkey SetMenuItemCommandKey drop
		THEN
	THEN ;


prune:			\ RESTORE.TEST ( start end -- start end )
	\ restore default if test word forgotten
	2dup
	test.token @
	rot rot range
	IF  ['] test_dummy test.token !
		fix.test.name
	THEN
	drop ;


>m4th
also ToolVocab

((
A simple tool to add to the TOOLS menu to allow testing any word from a menu.
Particularly useful when you want to execute something WITHOUT typing;
e.g., when SYS.WINDOW is not in front, or you don't want to disturb a
graphics operation.  For example, I used it to test a print routine that
used COPY.BITS to print a portion of the front-most window.

The word to TEST must not require any stack parameters nor return any results
on the stack.

It allows you to change its tools menu entry to execute any word you want by 
simply using the following at any time:

	TEST <MY.WORD> 

If <MY.WORD> is forgotten, then the default action is restored by use of a
WhenForget word
))

: InstallTestTool ( -- | Install the TestIT Tools menu group )
	ToolsMenu# menu.handle 0= IF Tools.Menu THEN
	1 NewToolGroup          		\ Enough room to install items ?
	IF  Test.ToolGroup !      		\ Save separator item number
 	    ['] test.it
		test.token @  >name
		?dup 0= 
		IF
			c" Test IT"
		THEN
		NewToolItem to TestItem#
		fix.test.name
	THEN ;



: TEST ( ( <word> -- | install next word in input stream into Test Menu)
	'	\ exists? dup -found
	test.token ! 
	fix.test.name  ;


\ InstallTestTool   \ go ahead and install it

>m4th



StopIncluding ************************************


