\ Various Editor Utilies
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\    to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


((
	Contains:
		DATE/TIME Stamp  (Tools Menu)
		LOCATE  		 (Tools Menu)
		INCLUDE ERROR HANDLER 
		OPEN LAUNCH SCRIPT (Options Menu)
		EDITOR INDENT TOOL (Edit Menu)
		EDITOR COMMENT TOOL (Tools Menu)
		
		"MORE EDITOR TOOLS"
		HiliteEditorRange
		SELECT.SE
		CLEAR.SE
))

only Forth definitions
>M4TH


lacking ToolVocab  vocabulary ToolVocab



\ *******  DATE/TIME STAMP TOOL  ***************************************************

only Forth definitions
>M4TH
decimal

create USERNAME$  128 allot&erase

ALSO EDITORVOCAB
also ToolVocab definitions

ALSO SYSTEM


: (MAKE.STAMP$) ( str$ -- addr  )	\ prefix str$ - add date&time
	{:  | stamp$[ 256 ]	:}
									pocket place
	stamp$[ dup GetDate$	count	pocket append
	s"   "     						pocket append
	stamp$[ dup GetTime$	count	pocket append
	
	username$ c@
	IF
		s"   "     					pocket append
		username$			count	pocket append
	THEN
	se.eolchar @ sp@ 1				pocket append drop

	pocket ;

PREVIOUS			\ done with SYSTEM


: MAKE.REVISED.STAMP$ ( -- addr  )
	{:  | stamp$[ 256 ]	:}
	s" \ Last Revision: " 	(make.stamp$)
;


: MAKE.CREATED.STAMP$ ( -- addr  )		\ 10.44
	{:  | stamp$[ 256 ]	:}
	s" \ Created:       " 	(make.stamp$)
;



: STAMP.SE ( --| stamps the date into the current Editor window)
	Front.Window se.window  =		\ editor frontmost?
	se.window    get.window =		\ editor active window?
	AND
	IF
		make.revised.stamp$
		count
		se.window
	\	selend @ dup selstart !		\ insert at current end of selection instead of replacing
		delete.sel.range
		selend @
		(insert.text>se) ( addr\cnt\SEWindow\offset -- ) 
	ELSE
		beep
	THEN ;

PREVIOUS		\ done with EDITORVOCAB


Create DS.ToolGroup -1 , \ holds ID of this group of tools

>m4th
also ToolVocab


: InstallDSTool ( -- | Install the DS Tools menu group.)
	toolsmenu# menu.handle 0= IF Tools.Menu THEN

    1 NewToolGroup         			\ Enough room to install items ?
    IF  DS.ToolGroup !     			\ Save separator item number
        ['] stamp.se  c" Date Stamp"   NewToolItem drop
    THEN ;

\ should have  forgettor




\ *********  EDITOR LOCATE TOOL  ***************************************************

only Forth definitions
>M4TH
decimal
also ToolVocab 

also editorvocab definitions


also SUBSTITUTIONS

: SE.LOCATE ( c-addr u line# -- )
  	{  c-addr u line# | temp[ MAX_PATH ] }
\ depending on what name was passed to VFX INCLUDED we could see
\ 	SFSTRUCT.szPathName contain a full path name or
\	a partial path name assuming BasePath

\ first try LOCATE$ after ExpandMacro from info returned by (WHEREIS)
	c-addr u expand  locate_path place

	line# 2 - LineToStartLocate !

	locate_path count  Path->FSSpec 
	?dup
	IF
		OpenEditor&Locate
	ELSE
\ if that fails, we prepend BasePath and try again
		BasePath     count temp[ place
		
		s" /" temp[ append

		locate_path  count temp[ append

		temp[ count
		Path->FSSpec ?dup
		IF
			OpenEditor&Locate
		ELSE
			c" Failed to locate File:" cr$ cr$
			temp[
			Big:-(OK# IssueWindowCenteredAlert drop
		THEN
	THEN
;



also system

: ((LOCATE))	\ xt --
	\ adapted from VFX locate
	(whereis)					\ xt -- c-addr u line# TRUE | FALSE 
	IF							\ caddr u line# --
		lookfor @ >name count
			search$ place		\ target name -> search$
		se.locate
	ELSE
		c" No source file information available for:"
		cr$ cr$ lookfor @ >name
		!OK# IssueWindowCenteredAlert drop
	THEN
;

previous	\ don't need SYSTEM any more
previous	\ don't need SUBSTITUTIONS any more



: ($Locate) ( c-addr -- )
	FindAnywhere 0=
	IF
		cr$ cr$ c" Not found in any vocabulary"
		!OK# IssueWindowCenteredAlert drop
	ELSE
		((LOCATE))
	THEN
;

defer (Locate)		\  ( c-addr -- )
' ($Locate) is (Locate)


: <<Locate ( -- )	\ space"name"
	bl word (Locate)
;



: LOCATE.SELECTION  ( -- | LOCATE on selection range of front SE.WINDOW )
	selstart @ selend @ = 			\ just beep if nothing
	IF
		c" Select a word to locate in an Editor Window"
			"" "" ""  !OK# IssueWindowCenteredAlert drop
	ELSE
		range->search$					\ move selection range to Search$
		hide.range
		selend @  selstart !			\ protect selection from eprint
		search$ (Locate)
	THEN
;


>m4th
also EDITORVOCAB

: Locate
\	cr ." my locate"	\ i.e., not the default one
	<<Locate
;

' <<Locate is <LOCATE		\ needs restoration on forget



Create Locate.ToolGroup -1 , 		\ holds ID of this group of tools

also ToolVocab

: InstalLocateTool ( -- | Install the DS Tools menu group.)
	ToolsMenu# menu.handle 0= IF Tools.Menu THEN
	
	1 NewToolGroup          		\ Enough room to install items ?
	IF  Locate.ToolGroup !      	\ Save separator item number
		['] locate.selection c" Reference/R"	NewToolItem drop
	THEN ;


\ Possibly InstalLocateTool   \ go ahead and install it



\ needs forgettor




\ *********  INCLUDE ERROR HANDLER  ***********************************************

>m4th
also EditorVocab definitions


: LatestNameIn  \  wid -- nfa
\ method borrowed from  WalkWordList
  {  wid | temp[ 512 1+ cells ] -- }

  \ Copy wordlist to temporary buffer
 	wid temp[ over >#threads @ 1+ cells move
 	temp[ (max-def) nip ctrl>nfa
;

: LatestSourceFile ( -- nfa )
	['] sourcefiles voc>wid LatestNameIn
;

: FindSourceFile ( -- addr\cnt )
\ sometimes CurrSourceName returns a bogus string
	CurrSourceName expand 2dup
	IsFileMissing? ( path$\cnt -- flag )
	IF
		2drop
		LatestSourceFile count
		expand
	THEN
;

also SUBSTITUTIONS

: Alert&ShowError ( theabort$\err# -- )
	\ abort$ = 0  means was a regular error
	\ Last parsed word should be at HERE
	\ ERR# passed to us on stack
	\ CurrSourceName set by Include
	\ LINE# set by Interpret

	{: theabort$ err#  | PathBuf[ MAX_PATH ] :}

	state off
	(flush.type.buffer)		\ in case editor is in front
	
	\ stuff normally done in : ((QUIT))
		clean-opt  cleanup-locals	\ compiler state clean up
        OperatorType off			\ SFP055
        QuitHook					\ User defined clean up
\       init-quit
\       r0 @ rp!  
        s0 @ sp!
	
	\ copy offending word to POCKET
((
	\ Last parsed word should be at HERE
	\ but isn't always, for some reason

	here count ?dup 0=
	IF
		DROP
		'tib @ >in @
	THEN
))
	BadWord->Pocket				\ refactored 10.46
(( 
	saved>in @ dup 'tib @ +		\ wasOffset\start
	>in @ rot -
	-white
	pocket place
 ))


	se.interpreting @
	IF
		err# -> was.se.error
		theabort$ err# SE.INTERPRET.ERROR?	\ Better.SE.INTERPRET.ERROR?
	ELSE
		FindSourceFile
				PathBuf[ PLACE

		theabort$
		IF
			pocket line# @ theabort$ AbortAlert ( badword$ line# c-addr -- )
		ELSE
			pocket line# @ err# ErrorAlert
		THEN

		\ open editor and find the word
		'SourceFile @ 0 >	\ not console or cached
		EditOnError?  AND
		IF
			FindSourceFile	\ PathBuf[ count 
			se.open.from.path
			IF
				pocket line# @ 2 - FindFrom
			ELSE
				c" Failed to open File:" cr$ "" PathBuf[
				Big:-(OK# IssueWindowCenteredAlert drop
				
				PathBuf[ 40 DUMP
			THEN
		THEN
	THEN
;

previous	\ don't need SUBSTITUTIONS any more


: ToEditorOnAbort"	( abort$ -- )		\ see DoAbortMessage
	-2 Alert&ShowError
;


: ToEditorOnError	( err# -- )		\ see DoErrorMessage & (EDITONERROR)
	0 swap Alert&ShowError
;


0 value wasDoAbortMessage
0 value wasDoErrorMessage


: EnableEditorErrorRecovery
	action-of DoAbortMessage -> wasDoAbortMessage
	action-of DoErrorMessage -> wasDoErrorMessage

	['] ToEditorOnAbort" is DoAbortMessage
	['] ToEditorOnError  is DoErrorMessage
; restorer linklasttoken


: RestoreErrorRecovery
	wasDoAbortMessage  is DoAbortMessage
	wasDoErrorMessage  is DoErrorMessage
; on.forget RestoreErrorRecovery



\ *********  OPEN LAUNCH SCRIPT  ***************************************************

\ used in Options Menu

only Forth definitions
>M4TH
decimal

also EditorVocab
definitions


: (OpenLaunchScriptinEditor) ( -- )
	{ | aFSSpec[ FSSpecRecord ] }
	startupdir 2@  c" Default.Extensions" 
	aFSSpec[  FSMakeFSSpec
	0=
	IF
		aFSSpec[ OpenEditorFromFSSpec
	ELSE
		beep
	THEN
;


: OpenLaunchScriptinEditor ( -- )
	optionkey?  0=
	IF
		c" This allows you to edit the configuration file that is included late in initialization."
		cr$ cr$
		c" Changes will not have an effect until you quit and re-launch."
		ok# issuealert drop
	THEN
	(OpenLaunchScriptinEditor)
;

' OpenLaunchScriptinEditor is EditLaunchScript


also ToolVocab


: OpenBuildScript ( -- )
	optionkey?  0=
	IF
		c" This allows you to edit the script file used to add MacForth Extensions to VFX."
		cr$ cr$
		c" You will need to rebuild your working kernel for changes to have an effect."
		OK# issuealert drop
	THEN
	c" mfVersion" (Locate)
;
' OpenBuildScript is EditBuildScript


: (ViewLogFile)
	{ | aFSSpec[ FSSpecRecord ] scr[ 1024 ] }
\	GetLocalVolDir ( -- volref#\dirid )	
\	logfile$

	scr[ LocateLogFile ( namebuf -- vol\dir )
	scr[ aFSSpec[  FSMakeFSSpec
	0=
	IF
		aFSSpec[ OpenEditorFromFSSpec
		JumpEnd
	ELSE
		beep
	THEN
;

' (ViewLogFile) is ViewLogFile


\ *********  EDITOR INDENT TOOL  ***************************************************


(* This is an Editor add-on that allows an indent/outdent function.  

To use: Select a range of lines.  Press command-] to insert a tab in front of 
each line.  Or press command-[ to remove existing tabs/spaces in front of each 
line.  Note that the entire line must be selected before a tab will be inserted 
or deleted.  

Doug Hoffman
*)


only Forth definitions
>M4TH
decimal

also EditorVocab
definitions


: VisibleSelectionRange? ( -- flag)
	?ShiftKey
	IF
		true
	ELSE
		selstart @	 top.line @ - 0<
		screen.end @ selend   @ - 0< OR 0=
		dup 0=
		IF
			c" Use Shift Key with this function to select "
			c" a range longer than the visible screen."
			"" "" OK# IssueWindowCenteredAlert drop
		THEN
	THEN
;


: IndentProceed?	( -- [selstart\selend\true] or 0\0\[false])
	get.window VALID.SE?
	VisibleSelectionRange? AND
	IF
		StartOfSelectedLine		\ FindStartOfSelectedLine
		EndOfSelectedLine		\ FindEndOfSelectedLine
		true
	ELSE
		beep 
		0 0 false
	THEN
;


: DoIndent ( -- )
	IndentProceed? -rot
	0 locals| offset end start |

	IF	
		5 SaveRangeOp		\ SaveFormattingRange
		InhibitUndo @ >R
		InhibitUndo ON

		hide.range
		start -> offset
		start end COUNT.LINES
		
		1+  0
		DO	tab$ count get.window offset insert.text>se 
			1 +to offset	\ must do this because Place$At did an insert
			1 +to end
			offset get.line.length drop ( don't need remaining) -> offset
		LOOP
		start selstart !	end selend !  show.range
		true mark.change

		R> InhibitUndo !
		6 SaveRangeOp		\ CorrectFormattingEnd
		R> -> se.scrap
	THEN ;


: Tab/Space?	( char offset--flag)
	se.buffer @@ + c@ dup
	9 =	( char\flag)
	swap bl = OR ;


: DoUnindent ( -- )
	IndentProceed? -rot
	0 locals| offset end start |
	IF	
		5 SaveRangeOp		\ SaveFormattingRange
		InhibitUndo @ >R
		InhibitUndo ON

		hide.range
		start -> offset
		start end  COUNT.LINES 1+  0
		DO	offset  Tab/Space?
			IF	offset dup selstart ! 1+ selend ! (se.delete)
				-1 +to end
			THEN
			offset get.line.length drop ( don't need remaining) -> offset
		LOOP
		start selstart !	end selend !  show.range
		true mark.change

		R> InhibitUndo !
		6 SaveRangeOp		\ CorrectFormattingEnd
		R> -> se.scrap
	THEN ;


' DoIndent   is InDentRange
' DoUnindent is OutDentRange


: ForgetIndenter ( -- )
	['] beep dup  is InDentRange
				  is OutDentRange
; on.forget ForgetIndenter




\ *******  Commenting Tool  ***************************************************

only Forth definitions
>M4TH
also editorvocab definitions


create Comment.ToolGroup -1 , 	\ holds ID of this tool group


: SE.EOL$ ( -- cstr$ )
	SE.EOLMode @
	CASE
		CR=EOL   OF cr$			ENDOF
		LF=EOL   OF lf$			ENDOF
		CRLF=EOL OF LF->CRLF 	ENDOF
	ENDCASE
;

: InsertEOL ( -- )
	se.eol$ count se.window selend @
	(insert.text>se)
;

: ((-CommentRange-)) ( -- )
	selstart @ selend @
	{ wasSelStart wasSelEnd }
	se.window front.window =
	0= IF beep exit THEN
	?NoEditAllowed
\	IF				\ 10.46
	5 SaveRangeOp		\ SaveFormattingRange
	InhibitUndo @ >R
	InhibitUndo ON

	wasSelStart wasSelEnd =
	IF	\ no selection, so insert a blank line
		InsertEOL
	THEN
	S"  ))" se.window selend @
		(insert.text>se)
	InsertEOL
	selend @ -> wasSelEnd	\ wasSelEnd is end of new text selection
		
	wasSelStart selend !
	S" (( "  se.window selend @
		(insert.text>se)
	InsertEOL

	\ adjust  end of new text selection
	wasSelEnd   selend @ wasSelStart -  +
		dup selend ! selstart !
	show.selend
	show.range

	R> InhibitUndo !
	6 SaveRangeOp		\ CorrectFormattingEnd
	SEMenuBarAction	
\	THEN
;

>M4TH
also editorvocab

also ToolVocab


: RemoveCommentTool (  --  )
	Comment.ToolGroup @ dup -1 <> 		\ no room for group ?
	IF   PruneToolItem  
	ELSE drop
	THEN
;
on.forget RemoveCommentTool




: InstallCommentTool ( -- )
	ToolsMenu# menu.handle 0= IF Tools.Menu THEN
	1 NewToolGroup          		\ Enough room to install items ?
	IF
		Comment.ToolGroup !      	\ Save separator item number
	    ['] ((-CommentRange-))   c" X" NewToolItem		\ drop
		    ToolsMenu# menu.handle swap 
	    		c" (( Comment Range ))"  SetMenuItemText
	THEN 
;


>M4TH



\ *******  MORE EDITOR TOOLS  ***************************************************

only Forth definitions
>M4TH
also editorvocab definitions

\ this group appends text to end of editor...
\ WHILE PRESERVING CURRENT SELECTION
 
: EDITOR.WRITE ( addr\cnt\SEWindow -- actCount )
	over >R
	-1 insert.text>se
	R>
;

: EMIT>EDITOR ( char\SEWindow -- ) \ append char to SEWindow
	locals| TheWindow |
	sp@ 1 TheWindow -1 insert.text>se drop ;


: CR>EDITOR ( SEWindow -- | append CR to SEWindow )
	SE.EOLChar @  swap emit>editor ;


: BL>EDITOR ( SEWindow -- | append SPACES to SEWindow )
	bl swap emit>editor ;



\ this group appends text to end of editor...
\ CURRENT SELECTION BECOMES EOF

: (EDITOR.WRITE) ( addr\cnt\SEWindow -- actCount )
	over >R
	-1 (insert.text>se)
	R>
;

: (EMIT>EDITOR) ( char\SEWindow -- ) \ append char to SEWindow
	locals| TheWindow |
	sp@ 1 TheWindow -1 (insert.text>se) drop ;


: (CR>EDITOR) ( SEWindow -- | append CR to SEWindow )
	SE.EOLChar @  swap (emit>editor) ;





: HiliteEditorRange ( start\end\SEWindow -- )
	se.window  >R
	get.window >R
	dup window -> se.window
	hide.range
	selend ! selstart !
	RangeVis on
	invert.sel.range
	update.se
	R> window
	R> -> se.window 
;


: SELECT.SE ( SEWindow -- | select entire SEWindow )
	se.window >R
	-> se.window
	0  	  		 selstart !
	se.file.size @	 selend !
	R> -> se.window  ;



: CLEAR.SE ( SEWindow -- | select entire SEWindow and delete contents )
	se.window 
	get.window 
	locals| OldWind OldSE |
	dup select.se
	dup -> se.window 
		   window
    RangeVis off
	se.delete  
	oldwind window
	oldse -> se.window ;


>M4TH


StopIncluding ************************************
*******************************************************
