\ Recent Files Submenu
\ Last Revision: 28 Feb 2012  16:15:00  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\    to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


only Forth definitions
>M4TH

decimal


\ Easily changeable maximum # Recent Items - clipped to 200 maximum in code
40 value Max#RecentFiles
create RecentFileBag$ 	256 allot&erase
					s" Recent Files.bag" RecentFileBag$ place

1 value keepRecentFileBagLocal	\ inside bundle default
	\ if =  0    we will use our Application Support subfolder
	\ if = -1    we will keep it next to our app
	\	  useful if you use multiple app copies in different project folders
	\ if = other we will keep it inside our bundle/Contents/ folder
	\	  useful if you use multiple app copies in the same project folder

also editorvocab
definitions




0 value RecentFilesMenu#
4 constant 1stFileItem#
0 value RecentFilesHandle		\ in.heap

\ add a LF to make file easier to view in an editor
$5AA5A50A constant RFMarker	


\ RecentFilesHandle contains a concatenation of:
structure RecentFileElement
	long:	 +RecentFileMarker		\ identifier = $5A5AA50A
	short:	 +RecentFileFlags
	short:	 +RecentFileNameLen
	0 bytes: +RecentFileName
structure.end


\ +RecentFileFlags
\	$0001	\ locked
\	$1000	\ lost file


: ValidElement? ( addr -- addr\true |or| false )
	dup  +RecentFileMarker @ 
		 dup RFMarker =
		 swap  $FFFFFF00 and RFMarker $FFFFFF00 and =	\ old marker?
	OR

	IF
		true
	ELSE
		drop false
	THEN
;

: ElemLocked? ( ^element -- flag )
	+RecentFileFlags
	w@ $0001 and 0<>
;


: SetElemLocked ( flag\^element --  )
	+RecentFileFlags
	>r
	IF
		r@ 1 OR
	ELSE
		R@ w@ $fffe and
	THEN
	r> w!
;

: RemoveElement ( ^element -- )
	ValidElement?
	IF
		>R
		RecentFilesHandle
		dup @ r@ swap -		\ offset to element
		0					\ findstr
		r> +RecentFileNameLen w@ RecentFileElement +	\ findlen
		""					\ replcstr
		0					\ replacelen
		munger drop
\	ELSE
		\ error notification ?
	THEN
;

: AddToRecentFilesPathshandle ( addr\cnt -- )
\ latest entry goes first into list, so it auto-sorts newest->oldest
	{:  addr cnt | thebuff[ 1024 ] :}
	RFMarker thebuff[ +RecentFileMarker !
	0 		 thebuff[ +RecentFileFlags w!		\ flag field
	cnt 	 thebuff[ +RecentFileNameLen w!
	addr 	 thebuff[ +RecentFileName   cnt  move 
	RecentFilesHandle
	0							\ offset
	0							\ findstr
	0							\ findlen
	thebuff[					\ replcstr
	cnt RecentFileElement +		\ replacelen
	munger drop
;


: WalkRecentFilesPathshandle ( xt -- ^element |or| 0 )		\ xt ( ^element -- stop? ) \ false to continue
	\ newest -> oldest
	{:  theXT | theOffset theLimit ^element :}
	0 -> theOffset
	RecentFilesHandle handle.size -> theLimit
	
	BEGIN
		theOffset theLimit <
	WHILE
		RecentFilesHandle @ theOffset +			\ ^element
			dup -> ^element
			theXT execute IF ^element exit THEN
		^element +RecentFileNameLen w@ RecentFileElement + +to theOffset
	REPEAT
	0	\ ^element
;


: MatchRecentFile?  ( zstr\^element -- zstr\stop? )
	{ zstr element }
	zstr zcount  
	element +RecentFileName element +RecentFileNameLen w@
	str=
	zstr swap
;


: CheckForMatch ( zstr -- ^element |or| 0 )
	['] MatchRecentFile? WalkRecentFilesPathshandle
	nip
;


: PushAllRecentFiles ( -- marker\newestelement \ ... \ oldesteelement )
	$deadbeef
	['] false WalkRecentFilesPathshandle
	drop
;


: NE ( offset -- offset' )	\ next element
	dup
	RecentFilesHandle @ +
	+RecentFileNameLen w@ RecentFileElement +
	+
;


0 value RecentFileCount

: IncRecentCount
	drop
	1 +to RecentFileCount
	false
;


: CountRecentFiles ( -- n )
	0 -> RecentFileCount
	['] IncRecentCount  WalkRecentFilesPathshandle
	drop
	RecentFileCount
;


0 value #ToMatch

: (Nth?)  ( ^element -- stop? )
	drop
	-1 +to  #ToMatch
	#ToMatch 0=
;


: FindNthRecentFile ( N -- ^element )	\ 1-based
\ 1 is most recently added
\ returns 0 if not in range 1..#recentfiles
	dup
	IF
		dup CountRecentFiles >
		IF
			drop 0
		ELSE
			to #ToMatch
			['] (Nth?) WalkRecentFilesPathshandle ( xt -- ^element )
		THEN
	THEN
;


: FindOldestUnlockedFile ( -- ^element )	\ 0 - fail
	PushAllRecentFiles ( -- marker\newestelement \ ... \ oldesteelement )
	BEGIN
		dup  $deadbeef =					\ = last element, we're done
		IF drop 0 exit THEN
		dup +RecentFileFlags w@ 1 AND 0<>	\ continue if locked
	WHILE
		drop								\ next element
	REPEAT
	
	dup $deadbeef =
	IF
		drop 0
	ELSE
		>R
		BEGIN
			$deadbeef =
		UNTIL
		
		R>
	THEN
;


: PruneRecentFilesPathshandle ( max# -- )
	\ prune all "dead" links
;


: AddFileToRecentFilesPaths ( ref# -- )
	{ ref# | name$[ 1024 ] }
	
	\ remove oldest unlocked entry if list is too long
	CountRecentFiles Max#RecentFiles 200 min >=
	IF
		FindOldestUnlockedFile 
		?dup 
		IF 
			RemoveElement
		ELSE	\ no room for another
			exit
		THEN		
	THEN

	\ create full path name as Z$
	name$[ ref# GetFCBInfo	 ( ^namePtr\fileRef# -- VolRefNum\DirID\flags )
	drop
	name$[ FileInfo->ZString	\ get full path
	?DUP
	IF		\ 10.47 - rarely happens, but can occur
		\ remove older duplicates if any
		dup CheckForMatch ( zstr -- ^element |or| 0 )
		?dup 
		IF
			dup +RecentFileFlags w@ >R			\ preserve prior flags
			RemoveElement
		ELSE
			0 >R
		THEN
	
		\ add this one to the list
		zcount AddToRecentFilesPathshandle ( addr\cnt -- )
		\ restore locked flag state
		R>  $0001 and  							\ 10.46 only restore locked flag
		RecentFilesHandle @  +RecentFileFlags w!
	THEN
;



: ExtractFileName ( addr\cnt -- cfstr )
\ Get just the file name from full path
\ be sure to CFRelease cfstr when done
	CFAllocatorGetDefault rot rot
	0					\ roman
	CFURLCreateFromFileSystemRepresentation dup
	CFURLCopyLastPathComponent
	swap CFRelease
;



\ *******************************************************************************



0 value <itemcount>

: AppendOneRecentFile ( ^element -- stop? )
	{ ^element  |  item$[ 1024 ] }

	^element +RecentFileName
		^element  +RecentFileNameLen w@

\ shorten name
	ExtractFileName dup
	item$[ 1024 0 CFStringGetPascalString drop
	CFRelease
			
	item$[  FixLeadingDash		
	c"   " RecentFilesMenu# append.items			\ dummy string for item
	
	<itemcount> item$[ RecentFilesMenu# set.item$	\ correct the name

	^element +RecentFileName
		^element  +RecentFileNameLen w@
	IsFileMissing?
	IF
		^element  +RecentFileFlags dup w@ $1000 OR swap w!
		<itemcount> 2 RecentFilesMenu# item.style				\ italicize
	ELSE
		^element  +RecentFileFlags dup w@ $0FFF and swap w!		\ 10.46
	THEN

	^element +RecentFileFlags w@ $0001 AND
	IF
		<itemcount> $13 RecentFilesMenu# item.mark  ( item#\markchar\menuID -- )
	THEN


	1 +to <itemcount>
	false
;


: AddRecentFilesToMenu ( -- )
	1stFileItem# -> <itemcount>				\ 1st file item
	['] AppendOneRecentFile
	WalkRecentFilesPathshandle ( xt -- ^element )  drop
;

\ *******************************************************************************

0 value RecentFileBagFile#
0 value RecentBagFailure


: LocateRecentFileBag ( namebuf -- vol\dir )
	{ ^namebuf }
	keepRecentFileBagLocal ?dup
	IF
		-1 =
		IF
			RecentFileBag$ count ^namebuf place
			GetLocalVolDir
		ELSE
			RecentFileBag$ count ^namebuf place
\			s" :Resources:" ^namebuf place
\			RecentFileBag$ count ^namebuf append
			GetCFURLofMyResources dup >R
			CFURL->VolDir
			R> CFRelease
		THEN
	ELSE
		s" :VfxForth:"  ^namebuf place
		RecentFileBag$ count ^namebuf append
		GetMyAppSupportVolDir
	THEN
;

: (CreateRecentFileBag)  ( -- ioR )
	{  | scr[ 1024 ] }
	scr[  LocateRecentFileBag
	scr[  creator @  ascii RBAG  HCreate
	\ will return -48 (duplicate file) if exists
;


: (OpenRecentFileBag) ( -- ref# )
	{  | scr[ 1024 ] }
	scr[  LocateRecentFileBag  iodir ! volref# !
	scr[  rw  open$
;


: CloseRecentFileBag ( -- )
	RecentFileBagFile# ?dup IF fileclose 0 -> RecentFileBagFile# THEN
;


: OpenRecentFileBag ( -- ref# )		\ create if needed
	(OpenRecentFileBag) ?dup
	0=
	IF
		(CreateRecentFileBag) drop
		(OpenRecentFileBag)
	THEN
;

: LoadRecentFileBag ( -- )
	OpenRecentFileBag ?dup
	IF
		-> RecentFileBagFile#
		RecentFileBagFile# get.eof dup >R
\		RecentFilesHandle to.heap
\		from.heap to RecentFilesHandle
		RecentFilesHandle swap resize.handle drop
		RecentFilesHandle @
		R> RecentFileBagFile# FileREAD drop         ( addr\cnt\file# --  #read )
\		CloseRecentFileBag		\ keep it open !!!!
	ELSE
		io-result @ -> RecentBagFailure
\		?file.error  \ report error
	THEN
;


: SaveFileBag ( -- )
	OpenRecentFileBag ?dup
	IF
		-> RecentFileBagFile#
		RecentFilesHandle handle.size dup >R RecentFileBagFile# set.eof
		RecentFileBagFile# rewind
		RecentFilesHandle @
		R> RecentFileBagFile# FileWrite drop         ( addr\cnt\file# --  #written )
		RecentFileBagFile# flush.file
\		CloseRecentFileBag		\ keep it open !!!!
	ELSE
		\ report error ?
	THEN
;


defer UpdateFileBag		' SaveFileBag is UpdateFileBag		\ type RBAG

\ *******************************************************************************


: TrimExcessRecentFiles ( -- )
	CountRecentFiles   Max#RecentFiles 200 min   >=
	IF
		BEGIN
			FindOldestUnlockedFile 
			?dup 
			IF 
				RemoveElement
			ELSE			\ full of locked files
				exit
			THEN
			CountRecentFiles   Max#RecentFiles 200 min   <=
		UNTIL
	THEN
;


: UpdateRecentFilesMenu ( -- )
	RecentBagFailure IF exit THEN
	1stFileItem# RecentFilesMenu# PruneMenuTail		\ remove any items above item 4 from menu
	\ valid handle?
	CountRecentFiles 0>
	IF
		AddRecentFilesToMenu
	ELSE					\ if no recent files
		c" no recent files(" RecentFilesMenu# append.items
	THEN

	UpdateFileBag			\ 
;


: Item->Element ( item# -- ^element )
	1stFileItem# - 1+
	FindNthRecentFile ( N -- ^element )
;

: LockHierItem ( item# -- )
	Item->Element
	\ toggle flag in element
	dup  +RecentFileFlags w@
	$0001 XOR
	swap +RecentFileFlags w!
	\ redraw menu
	UpdateRecentFilesMenu
;

: DeleteOpenHierItem ( item# -- )
	Item->Element
	RemoveElement
	UpdateRecentFilesMenu
;

: OpenHierItem ( item# -- )
	Item->Element
	
	dup +RecentFileFlags w@ $1000 and
	IF
		c" File Not Found" "" "" ""
		!OK# IssueWindowCenteredAlert drop
	ELSE
		dup +RecentFileName
			swap  +RecentFileNameLen w@
	
		SE.OPEN.FROM.PATH ( path$\len -- sewptr |or| 0 )
		drop
	THEN
;


\ *******************************************************************************


: AddFileToRecent+UpdateMenu ( ref# -- )
	RecentBagFailure IF drop exit THEN
	AddFileToRecentFilesPaths
	UpdateRecentFilesMenu
;


: DoRecentFilesMenu  ( item# -- )		\ Hier menu action
	0 HiliteMenu
	dup
	CASE
\		1 OF drop se.open..		ENDOF	\ default action of File Menu's OPEN...
\		2 OF drop se.openany	ENDOF
	1stFileItem# 255 RANGE.OF
		ShiftKey?	\ ?ShiftKey	\ 						\ SHIFT KEY used to lock/unlock the selected item
		IF								\ SHIFT alone locks it - OPT-SHIFT unlocks it
		\	Optionkey? not swap LockHierItem ( flag\item# -- )
			LockHierItem ( item# -- )
			UpdateFileBag
		ELSE
			OptionKey?	\ ?OptionKey	\ OptionKey?					\ OPTION KEY used to delete the selected item from the list
			IF
				DeleteOpenHierItem
				UpdateFileBag
			ELSE						\ Neither SHIFT nor OPTION Opens the selected item
				OpenHierItem
				UpdateFileBag
			THEN
		THEN
								ENDOF
		nip
	ENDCASE
;


: AddRecentFilesMenu ( -- )
	RecentFilesMenu# ?dup IF delete.menu 0 -> RecentFilesMenu# THEN
	20 1stFreeMenu# -> RecentFilesMenu#
	-1 c" Open Recent" RecentFilesMenu# new.menu

	RecentBagFailure	\ RecentBagLocked?
	IF
		c" 'Recent Files Bag' could not be opened(" RecentFilesMenu# append.items
	ELSE
		c" Shift-Select toggles lock(;Opt-Select removes an entry(;-(" RecentFilesMenu# append.items
	
		c" recent1;recent2" RecentFilesMenu# append.items
		UpdateRecentFilesMenu
	THEN
	FileMenu# -menu.handle 4 RecentFilesMenu# SetMenuItemHierarchicalID drop
	drawmenubar
	
	RecentFilesMenu#  menu.selection:  DoRecentFilesMenu 
;

\ *******************************************************************************


0 value WasAFTER.SE.OPEN$ 	 ( ref# -- )
0 value wasBEFORE.SE.SAVE 	 ( ref# -- )
0 value wasBEFORE.SE.SAVE.AS ( ref# -- )

: AFTER.SE.OPEN$+RECENT ( ref# -- )
	dup >R
	AddFileToRecent+UpdateMenu
	R> WasAFTER.SE.OPEN$ ?dup IF execute ELSE drop THEN
;

: BEFORE.SE.SAVE+RECENT ( ref# -- )
	dup >R
	AddFileToRecent+UpdateMenu
	R> wasBEFORE.SE.SAVE.AS ?dup IF execute ELSE drop THEN
;

: BEFORE.SE.SAVE.AS+RECENT ( ref# -- )
	dup >R
	AddFileToRecent+UpdateMenu
	R> wasBEFORE.SE.SAVE.AS ?dup IF execute ELSE drop THEN
;


: CleanUpRecentFilesTool ( -- )
	CloseRecentFileBag
	WasAFTER.SE.OPEN$    -> AFTER.SE.OPEN$
	wasBEFORE.SE.SAVE	 -> BEFORE.SE.SAVE
	wasBEFORE.SE.SAVE.AS -> BEFORE.SE.SAVE.AS
	
	RecentFilesMenu#  ?dup IF delete.menu  0 -> RecentFilesMenu# THEN
	RecentFilesHandle ?dup IF to.heap THEN
	0 -> RecentFilesHandle
	0 -> RecentBagFailure
	false -> RecentFiles?
; on.forget CleanUpRecentFilesTool
presnapshot linklasttoken

>m4th
also editorvocab

: InstallRecentFilesTool ( -- )
	0 -> RecentBagFailure
	0 -> RecentFileBagFile#
	0 from.heap to  RecentFilesHandle
	LoadRecentFileBag

	AFTER.SE.OPEN$ 			  	 -> WasAFTER.SE.OPEN$		\ save old
	['] AFTER.SE.OPEN$+RECENT 	 -> AFTER.SE.OPEN$			\ assign new

	BEFORE.SE.SAVE.AS			 -> wasBEFORE.SE.SAVE.AS	\ save old
	['] BEFORE.SE.SAVE.AS+RECENT -> BEFORE.SE.SAVE.AS		\ assign new
	
	BEFORE.SE.SAVE				 -> wasBEFORE.SE.SAVE		\ save old
	['] BEFORE.SE.SAVE+RECENT 	 -> BEFORE.SE.SAVE.AS		\ assign new


	TrimExcessRecentFiles
	se.file.menu
	AddRecentFilesMenu
	true -> RecentFiles?
;  \ restorer linklasttoken		\ be sure it loads after editor
 

?building 0= [IF] InstallRecentFilesTool [THEN]




\ ***********************************************************************
\ mostly for debugging - can be removed


: ShowElement ( ^element -- stop? )
	cr
	dup 8 .R
	2 spaces
	dup +RecentFileFlags w@ 4 .R
	2 spaces
	dup  +RecentFileNameLen w@		\ count
	swap +RecentFileName swap type
	false
;


: ShowFileNameOnly ( ^element -- stop? )
	>R
	cr
	r@ 8 .R
	2 spaces
	r@ +RecentFileFlags w@ 4 .R
	2 spaces

	r@ +RecentFileName
	r> +RecentFileNameLen w@		\ count
	ExtractFileName ( addr\cnt -- cfstr )
	dup ShowCFString

	CFRelease
	false
;

>M4TH
also editorvocab


: SR ( -- )
	RecentFilesHandle @
	RecentFilesHandle handle.size
	dump
;


: ShowRecentFiles
	."   (newest first)"
	['] ShowElement
	WalkRecentFilesPathshandle ( xt -- ^element )  drop
;


: ShowRecentFiles2
	."   (newest first)"
	['] ShowFileNameOnly
	WalkRecentFilesPathshandle ( xt -- ^element )  drop
;

previous


