\ ACTEL & WORDLIST PRIMITIVES
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\    to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************

((
	Active Elements (ActEls) are heap based onjects that can be sent a variety of
	messages.  Used originally for Classic MacForth detached vocabularies that
	could use a variety of search algorithms, they can also be used for other object
	programming as described in the Manual.
	
	Very handy for manipulating name-value assoication lists
))

only Forth definitions

\ anew --vocprims--

>M4TH

vocabulary ACTELS

also actels definitions

decimal



: BadHandleAbort  ( ...\BadMessage# -- )
	\ receives the message# that caused the panic
	\ true abort" Bad message to Active Object !"  ;
	true 28 R$error ;


defer BadHandleAction	' beep is BadHandleAction		\ holds bad handle token
	' BadHandleAbort is BadHandleAction

: BadMessageAbort  ( ...\BadMessage# -- )
	\ receives the message# that caused the panic
	\ true abort" Bad message to Active Object !"  ;
	true 27 R$error ;

defer ElemPanic  ( ...\BadMessage# -- | user modifiable )
	\ receives the message# that caused the panic
	' BadMessageAbort   IS ElemPanic



Structure MinActelHeader	\ generic for all Active Elements (ActEls)
	long: +ActelAction		\ action token for this actel
	long: +ActelSpare1
	long: +ActelSpare2  	\ +ActelHeaderSize
	long: +ActelDataOffset
Structure.End



\ MinActelHeader constant MinElemHeaderSize

: ElemHeaderSize ( handle -- headersize )	\ was " 
	@ +ActelDataOffset @ ;


: >>ACT ( [...]\aHandle\message -- [???]  | send message to aHandle )
	over
	Handle?  ( handle -- [handle\TRUE]or[FALSE] )
	IF
		@ +ActelAction @
		execute
	ELSE
		BadHandleAction
	THEN
;


: Message ( message# -- )
	create ,
	does>
		@ >>ACT
;

: ElemAction!  ( token\handle -- )
	@ +ActelAction ! ;


: INHERITED, ( -- ** )
	\ ** emplaces a 00 in an action table
	0 ,   ;



\ ** MESSAGES that can be sent to any Actel:

\ First the generic messages for ACTELS:

   0 Message >>PanicAction	( ...\aHandle -- ... )
   1 Message >>New			( initialSpace\aHandle -- [newHandle] or [0] )
   2 Message >>Discard		( aHandle -- )
   3 Message >>Empty		( aHandle -- )
   4 Message >>Room			( newRoom\aHandle -- room\used )


\ Then the generic name-value assoication list messages:

  32 Message >>AddEntry			( value\NameAddr\nameLen\aHandle -- EntryTag |or| 0 )
  33 Message >>RemoveEntry		( EntryTag\aHandle -- flag | true if OK )
  34 Message >>(un)HideEntry	( flag\EntryTag\aHandle -- hide if true )

  35 Message >>FindName			(  NameAddr\nameLen\aHandle -- EntryTag or [0] )
  36 Message >>FindValue		(    value\aHandle -- EntryTag or [0] )
  37 Message >>Apply			( ..\token\aHandle -- [..\EntryTag] or [..\0] )

  38 Message >>GetValue			( EntryTag\aHandle -- value )
  39 Message >>GetName$			( EntryTag\aHandle -- addr\cnt)

  40 Message >>GetValue2		( EntryTag\aHandle -- value )
  41 Message >>SetValue2		( value\EntryTag\aHandle -- )

  42 Message >>Get#Entries		( aHandle -- #entries )


\ Finally, some specific ones for Vocabularies:

  64 Message >>GetFlags			( EntryTag\aHandle -- flags value )
  65 Message >>SetFlags			( flags value\EntryTag\aHandle -- )
  66 Message >>InterpretEntry	( state flag\EntryTag\aHandle -- )
  67 Message >>Prune			( low addr\high addr\aHandle -- )



: ActelMaker  ( <name selector-name> size -- actel handle )
\ 100 ActelMaker NewThing DictionaryAction
	create
	,			\ size offset 0
	'  ,		\ selector at offset 4
	does>
	0 locals| ah pfa |
	pfa @	MakeHandle -> ah
	pfa 4 + @  ah  elemaction!
	pfa @ ah @  12 + !
	ah >>empty		\ always pass >>empty to initialize actel
	ah ;



(*  STRUCTURE OF A SELECTOR TABLE:
	2 bytes:	starting msg#				\ default is 0
	2 bytes:	#entries					\ size of this table
	4 bytes:	token for BadMessage#		\ link to prior selector table or error notify
	4 bytes:	token for Action[0]
	4 bytes:	token for Action[1]
	...
	
	4 bytes:	token for Action[n-1]
*)

: DoSelector   ( ...\message#\PFA -- )	\ execute indexed token
	\ if message# not in range for this selector, pass it to bad message action
\	cr ." selector "  dup >name .name   ."   msg: "  over .
	{ msg# ^selector }
	msg# ^selector w@ -
	dup 0<
	IF								\ msg too small
		drop 
		msg# ^selector 4 + @		\ bad message token
	ELSE
		dup ^selector 2+ w@ >
		IF							\ msg too big
			drop 
			msg# ^selector 4 + @	\ bad message token
		ELSE						\ msg in range
			cells ^selector + 8 + @
		THEN
	THEN
	execute
;


: SELECTOR: ( name ( start msg# -- trace@\addr )
	create
	w,								\ to hold starting msg#
	0 w,							\ to hold limit
	here  
	!csp							\ stack mark for error-check
	does> DoSelector
;

: ;SELECTOR  ( trace@\addr -- )  		\ end of Selector: list
	\ ?comp
	?csp
	2 needed
	here  over -  4 /  1 -		\ sub 1 for default token
	swap 2 - w!					\ store the index limit
;
immediate



: ElemInherit  ( offspring Handle\parent Handle -- )
	\ copy action & instance variables from parent to offspring
	locals|  hParent  hOffspring  |
	hParent @					\ hParent is the source
	hOffspring @				\ hOffspring the destination
	hParent ElemHeaderSize		\ get Header Size
	cmove						\ copy Header  (instance stuff)
	hOffspring >>empty  		\ clear uninherited fields
;


: ElemNew  ( initialSpace\aHandle -- [newHandle] or [0] )
	\ create a new, empty ELEMENT based on aHandle
	swap							\ we will inherit from old A-List
	aligned		\ =cells							\ round up to even number
	over ElemHeaderSize  +
	4 +								\ allow a little for empty entry
	2 get.memory dup				\ dup 0<  272 r$error  ???
	IF	dup rot ElemInherit			\ set to initial default state
	THEN  ;



\ ** THE DEFAULT SET OF A-LIST PRIMITIVES **
\ Each primitive receives an aHandle that should have already been checked
\ for validity with Handle?.  This will normally be done by >>ACT


: EntryTag->Addr ( EntryTag\aHandle -- addr )
	@ + ;


((
First entry added at the END  of data area
Additional entries move towards the start
Search Names & Values is latest -> oldest
Duplicate names  are allowed, but only the most recent will be found with >>FindName
Duplicate values are allowed, but only the most recent will be found with >>FindValue

>>AddEntry preserves the string's case
>>FindName is cASe SeNSitiVe  -- 



For now, the name & value are invariant as set by >>AddEntry

each entry:
	+0	long:   <value1>	\ searchable with >>FindValue
	+4	long:   <value2>
	+8	long:   <linkTag>
	+12	byte:   <flags>
	+13	STRING: <name>		\ pascal format - 255 character max
		=halfcells
))

(*


The header of an old MacForth "vocabulary" looks like:

Structure VocHeader
	\ 1st 16 bytes are generic to Active Objects (Actors)
	 0			long: ^Voc.Action	\ action token of this object
	 4			long: spare
	 8			long: spare
	12			long: size of instance area ( offset to start data area )

	\ private instance area stuff follows:
	16			long: ^Voc.Latest		\ offset to latest entry
	20			long: ^Voc.Size			\ offset of "too full" limit
	24			long: ^Voc.Growth		\ growth factor for >>AddEntry
	Structure.End
*)


structure HVOCABHEADER
	MinActelHeader +
	long:	+vLATEST		\ EntryTag of most recently defined word
	long:	+vSIZE			\ handle.size 
	long:	+vFloor			\ lowest point that entries can reside (offset)
	long:	+vMapEntries	\ # entries in hash map
	long:	+vNumWords		\ # words
	long:   +vSpare
	216 +					\ bunch of extra room for future fun & games
		
structure.end

HVOCABHEADER mfFIELD +vHashTable


\ ---------- Basic Vocabulary Stuff ----------

((
First entry added at the END  of data area

For now, the name & value1 are invariant as set by >>AddEntry

each entry:
Structure
	+0	long:   <value1>	\ searchable with >>FindValue
	+4	long:   <value2>
	+8	long:   <linkTag>
	+12	byte:   <flags>
	+13	STRING: <name>		\ pascal format - 255 character max
		=halfcells
))

  32 constant LITTLEMAPSIZE		\ in longs
2048 constant BIGMAPSIZE		\ in longs

\ if vocab size is less than this size then no hash map is created
500 CONSTANT MinHashSize

: >hEnd  ( aHandle -- address pointing past the last byte in the vocab )
	@ dup +vSize @ + ;

: >hLatest  ( aHandle -- address of most recent entry )
	@ dup +vLatest @ + ;

: >hFloor  ( aHandle -- lowest address that an entry can occupy )
	@ dup +vFloor @ + ;

-5 mfFIELD NFA>VPA  ( addr -- addr' )  \ NFA to Vocab Ptr Address

: ET>NFA  ( EntryTag\aHandle -- NFA )
	@ + 13 + ;

: VP>NFA  ( vpa value\handle -- nfa )
	>hEnd swap - 5 + ;

: VPA>NEXT.ET  ( vpa -- eta )
	5 + count + =halfcells ;

: ET>VPA  ( entryTag\aHandle -- vpa )
	@ + 8 + ;


: >HEADVPA  ( vocab handle\addr\length -- ptr=addr of offset )
	rot @ 0 locals| INDEX VOCABADDR LEN ADDR |
	vocabAddr +vMapEntries @ BigMapSize = 
	IF
		len 
		BEGIN
			dup 
		WHILE
			1 - 
			dup addr + c@ 127 and 		\ char&$7F
			index 1 lshift + -> index 
		REPEAT 
		drop
		index dup 11 rshift +
		BigMapSize 1 - and 2 lshift
	ELSE
		len 3 and 5 lshift
		addr c@ 7 and 2 lshift +
	THEN 
	vocabAddr +vHashTable + ;




: hEmpty  ( aHandle -- )
\	cr ." hEmpty " .s
	@ locals| ADDR |
	addr +vSize @
	  addr +vLatest !
	addr hVocabHeader + 
	  addr +vMapEntries @ cells 
	    0  fill
	addr +vNumWords OFF
;


: hNew  ( initial space\aHandle -- newHandle or 0)
\	cr ." hNew " .s
	0 dup dup locals| totalSize Addr Child Parent requestedSize |
	requestedSize dup MinHashSize >=
	IF BigMapSize 4 * + ELSE LittleMapSize 4 * + THEN 
	hVocabHeader +
	\ =cells 
	
	dup -> totalSize
	NewHandleClear -> Child \ creates and erases
	Child 
	IF
		Child Parent ElemInherit
	
		child @ -> addr
		requestedSize MinHashSize >= IF BigMapSize ELSE LittleMapSize THEN
		dup addr +vMapEntries !
		4 * hVocabHeader + addr +vFloor !
		totalSize addr +vSize !
		Child hEmpty
	THEN
	Child ;


: hDiscard ( aHandle -- )	\ who would want to use this?
\	cr ." hDiscard " .s
	DisposeHandle
;

: hGetFlags  ( EntryTag\aHandle -- flags value )
\	cr ." hGetFlags " .s
	EntryTag->Addr 12 + c@ ;


: hSetFlags  ( flags value\EntryTag\aHandle -- )
\	cr ." hSetFlags " .s
	EntryTag->Addr 12 + c! ;


: hGetValue  ( EntryTag\aHandle -- value )
\	cr ." hGetValue " .s
	@ + @ ;


: hGetName$  ( EntryTag\aHandle -- addr\count )
\	cr ." hGetName$ " .s
	EntryTag->Addr 13 + count ;


: hGetEntry  ( EntryTag\aHandle -- value\flags )
\	cr ." hGetEntry " .s
	EntryTag->Addr dup @ swap 12 + c@ ;


: hApply  ( token\aHandle -- EntryTag or 0 )		\ >>Apply
\	cr ." hApply " .s
	0 0 0 locals| end nextTag tag aHandle token |
	aHandle HLock
	aHandle >hLatest to tag
	aHandle >hEnd to end
	BEGIN
		tag end <
	WHILE
		tag 13 + count + =halfcells to nextTag
		tag aHandle @ - 
		aHandle token execute		\ token takes ( EntryTag\aHandle -- flag ) \ true to exit hApply
		IF  aHandle HUnlock 
			tag aHandle @ - 
			EXIT 
		THEN
		nextTag to tag
	REPEAT
	aHandle HUnlock 0 ;

((
: hInterpretEntry  ( state\EntryTag\aHandle -- )
	hGetEntry
	bit-immediate and boolean rot >
	IF  Compile,
	ELSE execute THEN ;
))


\ ---------- Resizing Words ----------


: FreeRoom  ( aHandle -- free room )
	dup >hLatest swap >hFloor - ;

: UsedRoom  ( aHandle -- used room )
	dup >hEnd swap >hLatest - ;


: hRoom  ( newRoom\aHandle -- room\used )
\	cr ." hRoom " .s
	swap =halfcells swap
	0 locals| diff handle newRoom |
	newRoom IF
		newRoom handle FreeRoom - to diff
		diff 0>
		IF             						\ make it bigger
			diff ALIGNED -> diff
			handle HUnlock					\ 6.05436
			handle diff handle GetHandleSize + resize.handle
			0< 384 R$ERROR 					 \ " unable to resize vocab" 
			handle HLock	  				 \ while we move the contents
			handle >hLatest   dup diff +
			handle >hEnd 2 pick - BlockMove
			handle HUnlock
		ELSE									\ make it smaller
			diff 3 - ALIGNED -> diff
			handle HLock	  					\ while we move the contents
			handle >hLatest   dup diff +
			handle >hEnd 2 pick - BlockMove
			handle HUnlock
			handle diff handle GetHandleSize + SetHandleSize
		THEN
		diff handle @ +vLatest +! 
		diff handle @ +vSize +!
	THEN
	handle FreeRoom handle UsedRoom ;



\ ---------- APPEND Words ----------

: InsertEntry  ( head vpa\vpa\aHandle -- )
	locals| AHANDLE VPA HEADVPA |
	headVPA @ vpa !
	aHandle >hEnd vpa - headVPA ! ;


: hAddEntry  ( value\nameAddr\nameLen\aHandle -- EntryTag or 0 )
\	cr ." hAddEntry " .s
	0 locals| LATEST AHANDLE NAMELEN NAMEADDR thevalue |
	nameLen =halfcells 14 + aHandle FreeRoom >			\ auto resize actel if needed
	IF
		2000 aHandle hRoom drop 0= IF 0 EXIT THEN
	THEN
	nameLen =halfcells 14 + negate aHandle @ +vLatest +!
	aHandle >hLatest to latest
	0 latest 12 + c!						\ init flags to 0
	nameLen latest 13 + c!					\ move the name length
	nameAddr latest 14 + nameLen cmove		\ move the name body
	nameLen 1 and IF 0 latest 13 + count + c! THEN
	thevalue latest !
	0	  latest 4 + !
	aHandle latest 13 + count >headVPA
	latest 8 + aHandle InsertEntry
	
	aHandle @ +vNumWords incr	
	latest aHandle @ - ;


\ ---------- SMUDGE -----------

: h(un)HideEntry  ( flag\EntryTag\aHandle -- )
\ 	cr ." h(un)HideEntry " .s
	\ flag =  0  -- UNHIDE
	\ flag = -1  -- HIDE
	\ other flag -- toggles, like SMUDGE

	EntryTag->Addr 14 +		\ first character of name
	dup c@

	128  locals| TheMask TheValue TheAddr |
	CASE
		false OF TheValue TheMask invert and	TheAddr c! ENDOF
		true  OF TheValue TheMask or			TheAddr c! ENDOF
		TheAddr TheMask  toggle		( like SMUDGE )
	ENDCASE ;


\ ----- BEHEAD -----
\ probably the most complicated word of the lot
\ because of all the ptr adjustments necessary


: UNLINKEntry  ( EntryTag\aHandle -- )
	0 0 locals| offset offsetAddr aHandle EntryTag |
	aHandle EntryTag over et>nfa count >headVPA to offsetAddr
	aHandle >hEnd EntryTag aHandle et>vpa - to offset
	BEGIN
		offsetAddr @ offset <>
	WHILE
		aHandle >hEnd offsetAddr @ - to offsetAddr
	REPEAT
	EntryTag aHandle et>vpa @ offsetAddr ! ;


: SQUEEZEVOCAB  ( EntryTag\delta\aHandle -- )
	locals| aHandle DELTA EntryTag |
	aHandle >hLatest dup delta +
	EntryTag aHandle @ + aHandle >hLatest - 
	dup 100 > IF BlockMove ELSE cmove> THEN
	delta aHandle @ +vLatest +! ;


: ADJUSTVPTRS  ( delta\EntryTag\min vptr\aHandle -- )
	0 locals| entryAddr aHandle MINVPTR EntryTag DELTA |
	delta negate to delta
	aHandle @ +vHashTable
	aHandle @ +vMapEntries @
	0 DO			\ then scan it
		dup i 4 * + dup @ minVPtr > IF delta over +! THEN
		drop
	LOOP drop
	EntryTag aHandle entryTag->Addr to entryAddr
	aHandle >hLatest
	BEGIN
		dup entryAddr < 		\ over @ and
	WHILE
		8 + dup @ minVPtr > IF delta over +! THEN
		vpa>next.et
	REPEAT drop ;


: hRemoveEntry  ( EntryTag\aHandle -- ok? )
\ 	cr ." hRemoveEntry " .s
	0 locals| delta aHandle EntryTag |
	EntryTag 0= IF 0 EXIT THEN
	EntryTag aHandle UnlinkEntry
	EntryTag aHandle et>nfa c@ 14 + =halfcells to delta
	aHandle @ +vLatest @ EntryTag <>
	IF
		delta EntryTag aHandle @ +vSize @
			over 4 + - aHandle AdjustVPtrs
		aHandle @ +vNumWords decr	
	THEN
	EntryTag delta aHandle SqueezeVocab
	true ;



\ ---------- FIND Words ----------

: hFindName  ( name-addr\name-length\aHandle -- EntryTag or 0 )
\ 	cr ." hFindName " .s
	locals| aHandle nameLen nameAddr |
	nameLen 0= IF  							\ namelen=0 means return latest
		aHandle @ +vLatest @
		dup aHandle @ +vSize @ = IF drop 0 THEN
		EXIT
	THEN
	aHandle nameAddr nameLen >headVPA	\ vpa
	BEGIN	\ vpa
		dup @ 0= IF drop 0 EXIT THEN
		@ aHandle vp>nfa dup c@ nameLen = 
		IF	\ target length = namelen?
			dup count nameAddr \ -text	\ -TEXT   addr1  cnt  addr2 -- flag |  zero if the strings are equivalent
			swap s=
			\ 0=
			IF 13 - aHandle @ - EXIT THEN
		THEN
		nfa>vpa
	AGAIN  \ nfa of entry
;

: hFindValue  ( Value\aHandle -- EntryTag or 0 )
\ Searches Newest->Oldest entries - returns first match found
\ 	cr ." hFindValue " .s
	dup >hLatest over >hEnd
	locals| end curETA aHandle value |
	curETA 
	BEGIN
		dup  end <
		over  @ value <>
		AND
	WHILE
		13 + count + =halfcells
	REPEAT
	dup @ value = IF aHandle @ - ELSE drop 0 THEN
 ;


: hGetValue2  ( EntryTag\aHandle -- value )
\	cr ." hGetValue2 " .s
	@ + 4 + @ ;


: hSetValue2  ( value\EntryTag\aHandle -- )
\	cr ." hGetValue2 " .s
	@ + 4 + ! ;


: hGet#Entries ( aHandle -- #entries )
	@ +vNumWords @ ;




1 Selector: ElemAction  ( [...]\vMessage -- [...] )
	\ main dispatcher for all messages
	\ position corresponds to message value
	' ElemPanic ,			\ 0  >>PanicAction  ( aHandle -- )   *** first word is always the error-catcher
	' hNew ,				\ 1  >>New			( initial Space\aHandle -- newHandle |or| 0)
	' hDiscard ,			\ 2  >>Discard		( aHandle -- )
	' hEmpty ,				\ 3  >>Empty		( aHandle -- )
	' hRoom ,				\ 4  >>Room			( newRoom\aHandle -- room\used )
;SELECTOR



32 Selector: DictionaryAction
	' ElemAction ,			\ not in range 32-42, pass it to ElemAction
	' hAddEntry ,			\ 32  >>AddEntry 		( value\NameAddr\nameLen\aHandle -- flag |or| 0 )
	' hRemoveEntry ,		\ 33  >>RemoveEntry		( EntryTag\aHandle -- ok? )
	' h(un)HideEntry ,		\ 34  >>(un)HideEntry	( EntryTag\aHandle -- )

	' hFindName ,	 		\ 35  >>FindName	( NameAddr\nameLen\aHandle -- EntryTag |or| 0)
	' hFindValue , 			\ 36  >>FindValue	( value\aHandle -- EntryTag |or| 0)
	' hApply ,	 			\ 37  >>Apply		( token\aHandle -- EntryTag |or| 0)
	' hGetValue ,			\ 38  >>GetValue	( EntryTag\aHandle -- value )
	' hGetName$ , 			\ 39  >>GetName$	( EntryTag\aHandle -- addr\cnt )
	' hGetValue2 , 			\ 40  >>GetValue2	( EntryTag\aHandle -- value )
	' hSetValue2 , 			\ 41  >>SetValue2	( value\EntryTag\aHandle -- )
	' hGet#Entries ,		\ 42  >>Get#Entries ( aHandle -- #entries )
\ 	' hSetValue ,			\ ??  >>SetValue	( EntryTag\aHandle -- value )	\ ?? not useful
;SELECTOR



64 Selector: VocabAction
	' DictionaryAction ,	\ not in range 64-73, pass it to DictionaryAction
	' hGetFlags ,			\ 64  >>GetFlags	( EntryTag\aHandle -- flags value )
	' hSetFlags ,			\ 65  >>SetFlags	( flags value\EntryTag\aHandle -- )
	inherited,				\ 66  reserved		>>InterpretEntry
	inherited,				\ 67  reserved		>>Prune
	inherited,				\ 68  reserved
	inherited,				\ 69  reserved
	inherited,				\ 70  user available
	inherited,				\ 71  user available
	inherited,				\ 72  user available
	inherited,				\ 73  user available
;SELECTOR



\\
