\ Word List Extras
\ Last Revision: 25 Jul 2012  05:57:48  ward@megawolf.com


\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\    to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************



only forth definitions
>M4TH
also actels definitions


decimal



: LoadNamedWLResource ( cname$ -- handle )
\ be sure to open resource file, if not in RootResources.rsrc
	ascii VOCB swap GetNamedResource
;


: LoadNewWL ( -- handle )
	c" Empty Word List" LoadNamedWLResource ?dup
	IF
		dup DetachResource
		['] VocabAction over ElemAction! 	\ until VocabAction is locked in place
		dup >>empty
	ELSE
		true abort" LoadNamedWLResource failed"
	THEN
;



0 value TheWL		\ working WL to use with Partial and Fuzzy Matching

: InitTheWL ( -- )
	0 -> TheWL		\ avoid bad handles from prior snapshot
; preludes linklasttoken



: ClearTheWL ( -- )
	TheWL ?dup 
	IF
		handle.size
		  hVocabHeader >
		IF
			TheWL >>Discard
			0 -> TheWL
(( 
		ELSE
			BEEP
			CR->Log
			TheWL Num->Log
			s"  Bad value for TheWL in ClearTheWL"  Type->Log
 ))
		THEN
	THEN
	LoadNewWL -> TheWL	\ requires VOCB resource named "Empty Word List"
;



\ **** Partial Matches **************************************************

100 constant Max#Partials


create PartialMatchArray	\ 0-indexed array of <length, entrytag>'s
	Max#Partials 1+  2 cells * allot&erase
		$12345678 , $12345678 ,			\ guard space


0 value #Partials


: ClearPartials
	PartialMatchArray	 Max#Partials 1+  2 cells * erase
	Max#Partials 0
	DO
		$7fffffff i 2 cells * PartialMatchArray + !
	LOOP
;

: ^ShortestPartial ( -- addr )
	PartialMatchArray
;

: ^LongestPartial ( -- addr )
	Max#Partials  2 cells *  PartialMatchArray +
;

0 value TheLongestPartial

: InsertAPartial ( entrytag\matchvalue\N -- )
\   0 <= N < Max#Partials
	dup >r
	2 cells * PartialMatchArray +
	dup							\ insertion point
		dup 2 cells +			\ destination is one down
		Max#Partials r> - 2 cells *
		 cmove>					\ shift array down
	2!							\ save new entry
;


: WhichPartial? ( length -- N )		\ n negative means skip this
	{ thelength }
	thelength TheLongestPartial > IF thelength -> TheLongestPartial THEN

	thelength ^LongestPartial @ >
	^LongestPartial @ 0<> AND
	IF
		-1
	ELSE
		thelength ^ShortestPartial @ <=
		^ShortestPartial @  0<> AND
		IF
			0
		ELSE
			Max#Partials 0
			DO
				thelength
				i  2 cells * PartialMatchArray + @
				< IF i unloop exit THEN
			LOOP
			Max#Partials
		THEN
	THEN
;


: PossiblyUpdatePartials ( entrytag\length -- )
	dup WhichPartial? dup 0<
	IF
		3drop		\ not a best
	ELSE
		InsertAPartial
	THEN
;


: Partial? ( addr1\cnt1\UnkStr -- length |or| 0 )
	count
	{ addr1 cnt unkaddr unkcnt }
	unkcnt cnt >
	IF
		false
	ELSE
		addr1 unkaddr unkcnt is=	\ c-addr1 c-addr2 u -- flag
		IF cnt ELSE false THEN
	THEN
;


: CheckPartials ( cstr$\EntryTag\aHandle -- cstr$\flag ) \ true to exit hApply
	{ cstr$ EntryTag aHandle }
	EntryTag aHandle >>GetName$
	
	cstr$ Partial? ?dup
	IF
		1 +to #Partials
		EntryTag swap PossiblyUpdatePartials
	THEN
	cstr$ false
;


: (FindPartials) ( cstr -- )
	ClearPartials
	0 -> TheLongestPartial
	0 -> #Partials
	['] CheckPartials theWL >>apply
	2drop
;



\ **** FUZZY Matches **************************************************

\ Quick and Very dirty algotithm for ranking the match between strings
\ Seems to work OK on the type of words in a Forth vocabulary
\ Not so good as a general English language matcher

: FuzzyMatchStrings  ( addr1\cnt1\UnkStr -- matchRank )
	count
	0 0
	locals| thechar sum Unkcnt2 Unkaddr2 cnt1 addr1 |

	\ bonus for unknown string exactly matching first characters
	addr1 Unkaddr2 Unkcnt2    is=   ( addr1  addr2 cnt -- flag )
	IF
		Unkcnt2 dup * 10 * to sum
		\ penalty for mismatched string lengths..
		cnt1 Unkcnt2 - dup *  negate 2 * to sum
	ELSE
		\ penalty for mismatched string lengths..
		cnt1 Unkcnt2 - dup *  negate 10 * to sum
	THEN

	\ penalty for mismatched string lengths..
\	cnt1 Unkcnt2 - dup *  negate 10 * to sum		\ 6.0.4.2

	\ chart-by-char compare for #chars in shorter word
	Unkcnt2 cnt1 min	\ use shortest length of test and unknown strings
	0
	?DO
		addr1    i + C@ upc
		Unkaddr2 i + C@ upc

		=	\ MatchChar ?dup ( returns 0 or 8 now )
		IF
			\ highest credit for exact match
			50 +to sum
			\ 6 * +to sum
		ELSE
			\ less credit for being misaligned by one character
			i Unkcnt2 <
			IF
				Unkaddr2 1+ i + C@ upc
				addr1    i + C@ upc
				=
				IF
					30 +to sum
				THEN
			THEN
			i cnt1 <
			IF
				Unkaddr2 i + C@ upc
				addr1    1+ i + C@ upc
				=
				IF
					30 +to sum
				THEN
			THEN
		THEN
	LOOP
	sum 0 max
;


\ Maintain a sorted array of the "best" matches as we go
0 value BestFuzzyTag
0 value BestFuzzyValue
0 value #checked

30 constant Max#FuzzyMatches

create BestFuzzyMatches	\ 0-indexed array of (matchvalue, entrytag)'s
	Max#FuzzyMatches 1+  2 cells * allot&erase
		$12345678 , $12345678 ,			\ guard space

: ClearFuzzies
	BestFuzzyMatches	 Max#FuzzyMatches 1+  2 cells * erase
;

: ^BestFuzzy ( -- addr )
	BestFuzzyMatches
;

: ^LowestBest ( -- addr )
	Max#FuzzyMatches  2 cells *  BestFuzzyMatches +
;

: InsertAFuzzy ( entrytag\matchvalue\N -- )
\   0 <= N < Max#FuzzyMatches
	dup >r
	2 cells * BestFuzzyMatches +
	dup							\ insertion point
		dup 2 cells +			\ destination is one down
		Max#FuzzyMatches r> - 2 cells *
		 cmove>					\ shift array down
	2!							\ save new entry
;


: WhichBest? ( value == N )		\ n < 0 means it isnt one of the best
	{ thevalue }
	thevalue ^BestFuzzy @ >
	IF
		0
	ELSE
		thevalue ^LowestBest @ <
		IF
			-1
		ELSE
			Max#FuzzyMatches 0
			DO
				thevalue
				i  2 cells * BestFuzzyMatches + @
				> IF i unloop exit THEN
			LOOP
			Max#FuzzyMatches
		THEN
	THEN
;

: PossiblyUpdateBest ( entrytag\matchvalue -- )
	dup WhichBest? dup 0<
	IF
		3drop		\ not a best
	ELSE
		InsertAFuzzy
	THEN
;


: CheckFuzzyMatch ( cstr$\EntryTag\aHandle -- cstr$\flag ) \ true to exit hApply
	{ cstr$ EntryTag aHandle }
	EntryTag aHandle >>GetName$
	
	cstr$ FuzzyMatchStrings  ( addr1\cnt1\UnkStr -- matchRank )
\	dup IncCount
	EntryTag over	PossiblyUpdateBest
\	dup EntryTag aHandle >>SetValue2	\ save match value as value2
	dup BestFuzzyValue >
	IF
		-> BestFuzzyValue
		EntryTag -> BestFuzzyTag
	ELSE
		drop
	THEN
	
	1 +to #checked
	cstr$ false
;

: (FindBestFuzzies)  ( cstr -- )
	ClearFuzzies
	0 -> BestFuzzyTag
	0 -> BestFuzzyValue
	0 -> #checked
	['] CheckFuzzyMatch theWL >>apply
	2drop
;


\ **** VOCAB -> WORD LIST ***********************************************

0 value thisWID			\ 10.47

defer WLFilter  ( nfa -- nfa/true |or| false )		\ 10.47
	' true is WLFilter		\ default is pass all words
	

: AddOneToWL  ( thread# nfa -- flag ) \ Return TRUE to continue 
	nip
	WLFilter		\ 10.47
	IF
		dup name>	\ get XT
		swap count
		theWL >>AddEntry	( value\NameAddr\nameLen\aHandle -- EntryTag |or| 0 )
	\	drop
		thisWID swap theWL >>SetValue2			\ 10.47
	THEN
	true
;


: ((WID->WL))	( wid -- )	\ WAS ((V->WL)) -- add all names in wid to theWL
\ WalkWordList traverses the WID in reverse chronological order
\ i.e.,  Newset entry --> Oldest
	['] AddOneToWL swap WalkWordList  \ xt wid --
;




\ 10.47 - do not include duplicated vocabs

: InContextArray? ( voc *contextArray context  -- flag )
	cells bounds
	?do 
    	dup i  @ =
    	IF drop true unloop exit THEN
	cell +loop
	drop false
;


: MakeWLFromContext ( -- )
\ careful, there may be an active WLFilter 
	ClearTheWL

	\ copy context to context2[  eliminating duplicates		\ 10.47
	{   |  #context2   context2[ vsize ]   }
	0 -> #context2
	context2[ vsize erase
	get-order 		\ -- widn...wid1 n
	0 ?DO
		dup context2[  #context2  InContextArray?
		IF
			drop
		ELSE
			#context2 cells context2[ + !
			1 +to #context2
		THEN
	LOOP

	\ add Context Vocabs in "reverse" order
	\ so when we search, it is from "top" of Context list
\	context vsize bounds		\ 10.47
	context2[  vsize bounds
	swap cell-
	do 
    	i @ ?dup				\ get next WID
    	IF
    		dup -> thisWID		\ 10.47
    		((WID->WL))			\ add all words NEWEST to OLDEST
    	THEN
	cell negate +loop
;

 
(( 
: MakeWLFromContext ( -- )
	ClearTheWL
	\ add Context Vocabs in "reverse" order
	\ so when we search, it is from "top" of Context list
	context vsize bounds
	swap
	do 
    	i @ ?dup					\ get next WID
    	IF
    		dup -> thisWID			\ 10.47
    		((WID->WL))
    	THEN
	cell negate +loop
;
 ))




\ ***************************************************************************
\ ***************************************************************************
\ ***************************************************************************
\ ***************************************************************************


(* *************************************************************************

Do a tweaked Insertion sort of wordlist specified by SortableWL
Creates both an alphabetical sort and a sort by token

Roughly 80 mSec to sort all vocabularies of a standard MacForth build on
a 2010 iMac (about 9000 words).  Eliminating the token sort halves that

Since string compares are quite time-expensive, a "trick" is used to reduce
the # of compares needed, by using an index table based on the first letter
of each word

The Wordlist is of course, kept in a handle
One large array is needed, nut is ALLOCATE'd, so this code uses only about
4k of dictionary space

************************************************************************* *)

Only Forth Definitions

>M4TH
also actels
\ also toolvocab	
DEFINITIONS

Decimal


\ SortableArray is an array derived from SortableWL
\ Each element in SortableArray references an entry in SortableWL
\ SortableArray is built sequentially using >>APPLY on SortableWL
\ Insertion sort is performed as each entry is added, with the sort links saved in SortableArray


0 value SortableWL			\ will hold handle for wordlist

0 value SortableArray		\ will hold pointer to SortableArray ( ALLOCATE'd )
0 value #SortableEntries	\ #elements in SortableArray
0 value SortableOffset		\ scratch for adding elements


structure SWLEntry				\ structure of each eleemnt in SortableArray
	long: +SWLEntryTag
	long: +SWLEntryNameStart	\ pointer to start of name field in SortableWL
	long: +SWLEntryNextAlpha	\ sorting link - addr of next in alpha order
	long: +SWLEntryNextToken	\ sorting link - addr of next in token order - not implemented
structure.end



50 constant #Indices

create SearchIndexTable		#Indices cells allot&erase 
	\ array of pointers into SortableArray
	\ indexed by 1stLetter->Index
	\ each entry points to the first SortableArray item starting with first char
	\ speeds alpha sorting nearly 20x


\ calculate index into SearchIndexTable
: 1stLetter->Index ( char -- n )	\ targetted for vocabs, so assumed all Upper Case
									\ lower case will not fail, but sort will be
									\ much less efficient without increasing #Indices
	bl max
	ascii (  $FF  range
	IF
		ascii ' -
	ELSE						\ index 0 gets everything BELOW $28  '('
		drop 0					\ i.e. 0 - $27 (')
	THEN
	ascii  W ascii ' -  min		\ last "bin" is $30 == #48 for everything > 'V'
;



: ClearSortableWL ( -- )		\ Initialize SortableWL
	SortableWL    ?dup IF >>Discard   0 -> SortableWL    THEN
	SortableArray ?dup IF free drop   0 -> SortableArray THEN
	0 -> #SortableEntries
	0 -> SortableOffset
	SearchIndexTable  #Indices cells  erase
(( 
	#Indices 0
	DO
		-1 SearchIndexTable i cells + !
	LOOP
 ))
;


\ ***************************************************************************
\ Doing the sort

: Compare2 ( ^TheEntry1\^TheEntry2 -- true if 1<2 )		\ alpha compare
	    +SWLEntryNameStart @ count
	rot +SWLEntryNameStart @ count
	compare  -1 <>
;



: InsertAlphaLink ( ^TheEntry\^StartIndex -- )
	@ dup +SWLEntryNextAlpha @
	{ newEntry thisEntry nextToCheck }
	nextToCheck 0 >=		\ 0< 0=
	IF
		BEGIN
			nextToCheck  +SWLEntryNextAlpha @ 0<
			IF
				newEntry nextToCheck  +SWLEntryNextAlpha !
				exit
			THEN
			nextToCheck newEntry Compare2 
		WHILE
			nextToCheck -> thisEntry
			nextToCheck  +SWLEntryNextAlpha @ -> nextToCheck
		REPEAT
	THEN
	nextToCheck newEntry   +SWLEntryNextAlpha !
	newEntry    thisEntry  +SWLEntryNextAlpha !
;


\ 0 value #c				\ test
\ 0 value lastin			\ test


: DoAlphaLink	( ^TheEntry\^StartIndex -- )	\	over -> lastin	1 +to #c
	locals| ^StartIndex |
	^StartIndex @ 0<
	IF
		^StartIndex !
	ELSE
		dup ^StartIndex @  Compare2
		IF
			\ new entry is before start
			^StartIndex @ over +SWLEntryNextAlpha !
			^StartIndex !
		ELSE
			^StartIndex InsertAlphaLink
		THEN
	THEN
;


variable SearchIndexStart 	-1 SearchIndexStart !


: InsertTokenLink ( ^TheEntry\^StartIndex -- )
	@ dup +SWLEntryNextToken @
	{ newEntry thisEntry nextToCheck }
	nextToCheck  0 >=		\ 0< 0=
	IF
		BEGIN		\ walk through list
			nextToCheck  +SWLEntryNextToken @ 0<		\ last in sort order
			IF
				newEntry nextToCheck  +SWLEntryNextToken !
				exit
			THEN

			nextToCheck +SWLEntryNameStart @ 17 - @
			newEntry    +SWLEntryNameStart @ 17 - @
			>
		WHILE
			nextToCheck -> thisEntry
			nextToCheck  +SWLEntryNextToken @ -> nextToCheck
		REPEAT
	THEN
	nextToCheck newEntry   +SWLEntryNextToken !
	newEntry    thisEntry  +SWLEntryNextToken !
;


\ 0 value #c2				\ test
\ 0 value lastin2			\ test

: DoTokenLink	( ^TheEntry\^StartIndex -- )	\ over -> lastin2	1 +to #c2
	locals| ^StartIndex |
	^StartIndex @ 0<
	IF
		^StartIndex !
	ELSE
		dup    			+SWLEntryNameStart @ 17 - @		\ token for new entry
		^StartIndex @   +SWLEntryNameStart @ 17 - @		\ token for first sorted (lowest token so far) 
		>	\ <
		\ Compare2 ( ^TheEntry1\^TheEntry2 -- true if 1<2 ) 
		IF
			\ new entry is before prior lowest
			^StartIndex @ over +SWLEntryNextToken !
			^StartIndex !
		ELSE
			^StartIndex InsertTokenLink
		THEN
	THEN
;


: AddEntryToSortableArray ( EntryTag\aHandle -- flag )	\ tweaked insertion sort
	{ EntryTag aHandle | theArrayAddr }

	SortableArray  SortableOffset  + -> theArrayAddr

	EntryTag  							     theArrayAddr +SWLEntryTag  !		\ save entrytag
	EntryTag aHandle EntryTag->Addr 17 + DUP theArrayAddr +SWLEntryNameStart  !
	\ ^name still on the stack

	\  UpdateArray
	-1  theArrayAddr +SWLEntryNextAlpha !
	-1  theArrayAddr +SWLEntryNextToken !

	1+ c@   1stLetter->Index cells  SearchIndexTable +  theArrayAddr swap DoAlphaLink

\ possibly make the token search optional as it is a good deal slower than the alphabetic
	theArrayAddr SearchIndexStart   DoTokenLink

	SWLEntry +to SortableOffset
	0
;



: AlphaSortSortableWL ( -- )	\ this does it all except building SortableWL
	SortableWL >>Get#Entries
	dup -> #SortableEntries
	    SWLEntry * allocate drop  -> SortableArray
	
	-1 SearchIndexStart !

	#Indices 0				\ initialize SearchIndexTable
	DO
		-1 SearchIndexTable i cells + !
	LOOP

	0 -> SortableOffset
	['] AddEntryToSortableArray  SortableWL >>Apply drop
;



\ ***************************************************************************
\  Words to get nth entry from sorted wordlists

: GetNthUnSorted ( n -- ^WLNameField )		\ n is 1-based
	1 max
	#SortableEntries min  1-
	SWLEntry * SortableArray +  +SWLEntryNameStart @
;


: GetNthAlphaSorted ( n -- ^WLNameField )		\ n is 1-based
	1 max
	#SortableEntries min
	0
	{ desiredN currentN  }
	#Indices 0
	DO
		SearchIndexTable i cells + @ dup 0>
		IF
			BEGIN
				1 +to currentN
				currentN desiredN >= IF  +SWLEntryNameStart @  exit THEN
				+SWLEntryNextAlpha @ dup 0<
			UNTIL
			drop
		ELSE								\ no entries in this index
			drop
		THEN		
	LOOP
	-1
;


: GetNthTokenSorted ( n -- ^WLNameField )		\ n is 1-based
	1 max
	#SortableEntries min
	0
	{ desiredN currentN  }
	SearchIndexStart @ 
	BEGIN
		1 +to currentN
		currentN desiredN >= IF  +SWLEntryNameStart @  exit THEN
		+SWLEntryNextToken @ dup 0<
	UNTIL
	-1
;


\ ***************************************************************************
\  Words to build SortableWL

: (MakeSortableWL) ( XT -- )	\ XT should set context with desired vocabs
	TheWL						\ save for restoring
	0 -> TheWL
	{ theXT wasWL | temp[ vsize ] -- }

	context temp[ vsize move	\ save search order

	theXT  execute				\ XT should set context with desired vocabs

	MakeWLFromcontext
	TheWL -> SortableWL

	wasWL -> TheWL				\ restore TheWL
	temp[ context  vsize move	\ restore search order
;


: MakeSortableWL ( XT -- )		\ XT should set context with desired vocabs
	ClearSortableWL
	(MakeSortableWL)
;


: AllToContext ( -- )
\ put all vocabs on current search-order
	context VSIZE  erase
	voc-link
	BEGIN
		@ dup
	WHILE
		dup 
		ALSO  context !
		>voc-link
	REPEAT
	drop
;


: CurrentIscontext ( -- )
	get-current
	context VSIZE  erase
	context !
;


: AllToSortableWL ( -- )		\ all vocabs -> SortableWL
	['] AllToContext MakeSortableWL
;

: ContextToSortableWL ( -- )	\ only context -> SortableWL
	['] noop MakeSortableWL
;

: CurrentToSortableWL ( -- )	\ only current -> SortableWL
	['] CurrentIscontext MakeSortableWL
;




\\


MakeWLFromContext .s

DATA STACK
empty stack
theWL >>Get#Entries .
4119 
ok

