\ ExternPort.fth - porting harness for Extern.fth

\ Copyright (c) 2009
\ MicroProcessor Engineering
\ 133 Hill Lane
\ Southampton SO15 5AF
\ England
\
\ tel:   +44 (0)23 8063 1441
\ email: mpe@mpeforth.com
\        tech-support@mpeforth.com
\ web:   http://www.mpeforth.com
\ Skype: mpe_sfp
\
\ From North America, our telephone and fax numbers are:
\        011 44 23 8063 1441
\        901 313 4312 (North American access number to UK office)
\
\
\ To do
\ =====
\
\ Change history
\ ==============

only forth definitions
decimal

\ ===========
\ *! externport
\ *T Porting the EXTERN: interface
\ ===========
\ *P This file contains translations of VFX Forth words that may
\ ** not be present in other implementations. Because the EXTERN:
\ ** notation requires carnal knowledge of the host implementation
\ ** it is likely that the file *\i{Extern.fth} will itself be
\ ** modified. *\i{ExternPort.fth} is just a repository for words
\ ** that may not present in your Forth system.


\ **************
\ *S Assumptions
\ **************
\ *P We assume that a basic interface to Windows or a Unix style
\ ** library loader already exists.

((
: LoadLibrary		\ z-addr -- handle|0
\ *G Load a shared library who name is given as a zero-terminated string.
\ ** Return its address/handle or zero on error.

: GetProcAddress	\ handle z-addr -- addr|0
\ *G Given a library handle and the name of a function in the
\ ** library, return its execution address if found, or zero
\ ** if not found.

: FreeLibrary		\ handle -- status
\ *G Free the previously loaded library. Status is host specific,
\ ** and normally returns zero on error.

: AtCold	\ xt --
\ *G The word whose xt is given is run during start up and has the
\ ** stack effect
\ *C   --
))


\ ****************************
\ *S Generic portability layer
\ ****************************

variable lib-link	\ -- addr
\ *G Anchors the list of shared libraries.
 lib-link off

variable import-func-link
\ *G Anchors the list of imported functions.
  import-func-link off

: link,		\ var-addr --
\ *G Add a link to a chain anchored at address var-addr. The old contents of
\ ** var-addr are added to the dictionary as the new link, and the address
\ ** of the new link is placed at var-addr.
  here  over @ ,  swap !
;

: ParseUntil    \ c-addr u --
\ *G Parse the input stream for a white-space delimited string,
\ ** *\fo{REFILL}ing as necessary until the string is found or
\ ** input is exhausted. Mostly used for block comments. The
\ ** string compare is case insensitive.
  2>r
  begin
    begin
      bl word count dup
    while
      2r@ compare 0=
      if  2r> 2drop  exit  then
    repeat
    2drop  refill 0=
  until
  2r> 2drop
;

: ((            \ --
\ *G Block comment operator. Any source following this is ignored upto
\ ** and including the terminator, '))', which must be white space separated.
\ ))
  S" ))" ParseUntil
; immediate


variable NextErrCode
  1000 NextErrCode !

: SysErrDef	\ -- ; -- n
\ *G Creates an error/throw code and associates a message with it.
\ ** The message is delimited by '"' characters.
\ *C  SysErrDef <name> "<text>"
  NextErrCode @ constant  1 NextErrCode +!
  postpone \
;

: z$,		\ c-addr u --
\ *G Lay the given string in the dictionary as a zero terminated string.
  >r  here r@ move  r> allot  0 c,
;

: $>z,          \ addr --
\ *G Lay a zero terminated string in the dictionary, given a counted string.
  count z$,
;


\ *************
\ *S Structures
\ *************
\ *P Structures should really be written using the Forth200x
\ ** notation, but here are simple implementations of the VFX
\ ** structure notation.

: struct	\ -- addr 0 ; -- size
\ *G Begin definition of a new structure. Use in the form
\ ** *\fo{BEGIN-STRUCTURE <name>}. At run time *\fo{<name>}
\ ** returns the size of the structure.
  create
    here 0  0 ,                         \ mark stack, lay dummy
  does> @  ;                            \ -- rec-len

: end-struct	\ addr n --
\ *G Terminate definition of a structure.
  swap !  ;                             \ set len

: FIELD		\ n <"name"> -- ; Exec: addr -- 'addr
\ *G Create a new field within a structure definition of size n bytes.
  create
    over , +
  does>
    @ +
;

: int		\ <"name"> -- ; Exec: addr -- 'addr
\ *G Create a new field within a structure definition of size CELL bytes.
  cell field  ;

\ ======
\ *> ###
\ ======

decimal
