\ CommLine.fth - OSX command line operations

((
Copyright (c) 2006, 2009, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   htpp://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20110118 SFP003 OSX version.
20090129 SFP002 Linux specific version
20060127 MPE001 Major overhaul
))

only forth definitions
decimal


\ ===========
\ *> tools
\ *S Command Line parser
\ ===========
\ *P VFX Forth includes code for handling the OS X command line.
\ ** You can access the command line using Forth versions of
\ ** *\fo{ARGV[]} and *\fo{ARGC} as in C. At start up, VFX Forth
\ ** attempts to recreate the original command line using *\b{argc}
\ ** and the *\b{argv[]} strings so that the command line can be
\ ** treated as a sequence of Forth commands.

\ *P The attempt to recreate the command line is only partially
\ ** successful because the C parser uses double-quotes characters
\ ** to denote literal strings. These are passed to the application
\ ** as a single string with the double-quotes characters removed.
\ ** For example:
\ *C   vfxosx s" foo   bar" type
\ *P will give a command line error. You can escape the double-quotes
\ ** characters in the usual C manner, but this will not preserve the
\ ** spaces.
\ *C   vfxosx s\" foo   bar\" type
\ *P Most OS X system shells, including Bash, allow you to use
\ ** single-quotes characters for literal strings.
\ *C   vfxosx 's" foo   bar" type'


\ =================
\ +N INTERNAL Words
\ =================
\ +P These words are used by the internals of the command line mechanism.


\ ===============
\ +N Public words
\ ===============

(( for DocGen - Now in Kernel/386OSX/startup.fth
: argc	\ -- u
\ *G The number of defined arguments.
  EntryESP @ @
;

: argv[                 \ n -- pointer|0
\ *G Given an index of 0..argc-1 return a pointer to the command
\ ** line token's zero-terminated string. *\fo{0 argv[} returns the
\ ** executable's name. If the argument does not exist, the
\ ** pointer is zero.
  dup argc u>
  if  drop 0 exit  endif
  1+ cells EntryESP @ + @
; doNotSin

: CommandLine	\ -- c-addr len
\ *G Return the system command line. This is recreated from the
\ ** data passed to VFX Forth and may not be exact.
  CmdLine zcount
; doNotSin
))


\ =========
\ Test code
\ =========


\ ======
\ *> ###
\ ======

decimal
