\ SysConfig.fth - system state read/write in INI files

((
Copyright (c) 2008, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To Do
=====

Change History
==============
20160201 MPE003 Refactored editor configuration, default to nano.
20160127 MPE002 Changed default editor to emacs.
		Refactored to use VFXPATH if BASEPATH not set.
20080226 MPE001 First cut.
))

only forth definitions  decimal

\ ===========
\ *> lintools
\ *S System configuration
\ ===========

\ ----------------------------
\ *N Edit/Locate configuration
\ ----------------------------

\ ======================================
\ Grouped load/save - should be :NONAMEs
\ ======================================

also substitutions
: setLibMacro	\ caddr len --
\ Add /Lib to base string and set it as the LIB macro.
  { | temp[ 256 ] -- }
  temp[ place
  s" /Lib" temp[ append
  temp[ c" Lib" $setmacro
;

: loadEditCfg	\ --
\ load editor configuration.
  { | prog[ 256 ] -- }
  s" VISUAL" readEnv #255 umin dup if	\ allow for MacPorts
    prog[ place
  else
    2drop  s" EDITOR" readEnv #255 umin dup if
      prog[ place
    else
      2drop  s" nano" prog[ place
    endif
  endif
  s" Editor" Editor$ #255 prog[ count Ini.ReadStr
  s" Locate" Locate$ #255 s\" +%l% \q%f%\q &" Ini.ReadStr
  Locate$ c@ if
    assign .edlocate to-do .locate
  else
    assign .nolocate to-do .locate
  endif
  s" EditOnError" 1 Ini.ReadInt dup -> EditOnError?
  if   assign (EditOnError) to-do EditOnError  endif
;

: LoadSysCfg	\ --
\ *G Load the system configuration from the INI file.
  s" Options" Ini.Section
  loadEditCfg
  s" VfxPath" VfxPath #255 cNull count Ini.ReadStr
  s" DevPath" DevPath #255 cNull count Ini.ReadStr
  s" BasePath" BasePath #255 cNull count Ini.ReadStr
  VfxPath c@ if
    VfxPath count setLibMacro
  else
    BasePath c@ if
      BasePath count setLibMacro
    endif
  endif
; ' LoadSysCfg AtIniLoad

: SaveSysCfg	\ --
\ *G Save the system configuration to the INI file.
  s" Options" Ini.Section
  s" Editor" Editor$ count Ini.WriteStr
  s" Locate" Locate$ count Ini.WriteStr
  s" EditOnError" EditOnError? 1 and Ini.WriteInt
  s" VfxPath" VfxPath count Ini.WriteStr
  s" DevPath" DevPath count Ini.WriteStr
  s" BasePath" BasePath count Ini.WriteStr
; ' SaveSysCfg AtIniSave

: CheckSysPaths	\ --
\ *G Performed after the INI file has been loaded. Sets defaults
\ ** for the *\fo{VFXPATH} and *\fo{BASEPATH} macros when possible.
  VfxPath c@ 0= if
    s" VFXSRCDIR" readEnv dup if	\ present at installation
      255 umin VfxPath place
    else
      2drop
    then
  then
  BasePath c@ 0= if
    s" VFXROOTDIR" readEnv dup if	\ present at installation
      255 umin BasePath place
    else
      2drop
    then
  then
;  ' CheckSysPaths AtCold
previous


\ ======
\ *> ###
\ ======

