\ defmac.fth - default macros

((
Copyright (c) 2004, 2007, 2011, 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20190812 MPE005 Corrected MPE004.
20160203 MPE004 Removed default use of '&' in edit/locate commands.
		Replaced with checked version.
20110517 MPE003 Fixed edit on error action.
20110118 SFP002 OSX conversion.
20071211 SFP001 Made editor configurable.
))

decimal

\ ======
\ *> tmacro
\ *S MacOS specifics
\ ======
\ *P The code described here is specific to VFX Forth for MacOS.
\ ** Do not rely on any of the words documented here being
\ ** present in any other VFX Forth implementation.

\ create default macros

also substitutions definitions

(( now in KERNEL.FTH
#256 buffer: LOAD_PATH	\ -- addr
  0 LOAD_PATH c!
[undefined] LIBRARYDIR [if]
#256 buffer: LIBRARYDIR  0 LIBRARYDIR c!
[then]
#256 buffer: BasePath  0 BasePath c!

: lib LIBRARYDIR ;
: bin LOAD_PATH ;
#256 buffer: LOCATE_PATH  0 LOCATE_PATH c!
#256 buffer: LOCATE_LINE  0 LOCATE_LINE c!
: f locate_path ;
: l locate_line ;
))

previous definitions

also substitutions

: GetExeNameZ	\ zaddr --
\ *G Get the fully qualified name of the executable. It is saved
\ ** as a 0 terminated string at *\i{zaddr}. The buffer is assumed to
\ ** be at least 1024 characters long.
  { dest | psize[ cell ] buf[ MAX_PATH ] -- }
  1024 psize[ !
  buf[ psize[ _NSGetExecutablePath drop
  buf[ dest realpath drop
;

: InitMacros	\ --
\ *G Initialises the directory macros. Run at start up.
  { | buf[ MAX_PATH ] -- }
  buf[ 1+ GetExeNameZ
  buf[ 1+ zstrlen #255 umin buf[ c!
  buf[ stripFilename
  buf[ LOAD_PATH $move
  S" /../Lib" buf[ append
  buf[ LIBRARYDIR $move
  cDot to IncDir
;
' InitMacros AtCold
IncDir  InitMacros  to IncDir		\ this file is INCLUDED

previous


\ ****************************
\ *S Editor and LOCATE actions
\ ****************************

#256 buffer: editor$	\ -- addr
\ *G A buffer holding the path and name of the preferred editor
\ ** as a counted string, e.g.
\ *C   /bin/vi
  0 editor$ c!

0 value EditOnError?	\ -- flag
\ *G Set true to call the editor on an error.

: editor-is	\ "<editor-name>" --
\ *G Set your preferred editor, e.g.
\ *C   editor-is /bin/vi
\ *C   editor-is /Applications/UltraEdit.app/Contents/MacOS/UltraEdit
\ *C   editor-is edit
  GetPathSpec editor$ place
;

: .ed		\ --
\ *G Display the name of your preferred editor
  cr  editor$ c@ if
    ." Editor is: " editor$ count type
  else
    ." No editor set"
  endif
;

#256 buffer: locate$	\ -- addr
\ *G A buffer holding the macro expansion required to edit a
\ ** specific line of a file. This information is used by
\ ** *\fo{LOCATE}. In the example below the macros
\ ** *\i{%f%} and *\i{%l%} will be replaced by the file name and
\ ** line number.
\ *C   %f% -# %l%
\ *C   -- "%f%" --lc%l%:1 &
\ *C   +%l% "%f%" &
  s" %f% -# %l%" locate$ place

max_path buffer: EditCommand$	\ -- addr

: edit$		\ -- cstring
\ *G If the preferred editor has been set, return the program name,
\ ** otherwise return the default editor string for *\i{nano}.
  editor$ c@				\ select editor
  if  editor$  else  c" /usr/bin/nano"  endif
  count EditCommand$ ExpandMacro 2drop
  EditCommand$
;

0 value DebugLocate?	\ -- x
\ *G If non-zero, the string passed to the shell for *\fo{LOCATE}
\ ** is displayed before the system is called.

: $edit		\ cstring --
\ *G Edit the file provided as a counted string.
  edit$ swap $ExpandMacros
  DebugLocate?
  if  cr dup count type cr  endif
  $shell
;

: edit		\ "<filename>" --
\ *G Edit the file whose name follows in the input stream, e.g.
\ *C   EDIT release.txt
  { | temp[ #256 ] -- }
  GetPathSpec #255 min temp[ place  temp[ $edit
;

: .edlocate	\ --
  { | temp[ MAX_PATH ] }
  locate$ count temp[ ExpandMacro 2drop
  temp[ count + 1- c@ [char] & <>	\ last char not '&'
  if  s"  &" temp[ append  then		\ so add '&' to detach
  edit$ temp[ $shell
;
\ assign .edlocate to-do .locate

: (EditOnError)	\ -- ; run editor on error
\ *G Edit the file at an error, using the contents of the variables
\ ** *\fo{'SOURCEFILE} and *\fo{LINE#}.
  'SourceFile @ -1 =
  line# @ -1 = or
  EditOnError? 0= or
  if  exit  endif
  CurrSourceName line# @ LocateInfo	\ set locate macros
  .edlocate				\ go do it
;

: SetLocate	\ --
\ *G Tells VFX Forth how your editor can be called to go a
\ ** particular file and line. Use in the form
\ *C   SetLocate  <rest of line>
\ *P where the text after *\fo{SetLocate} is used to define
\ ** how parameters are passed to the editor, e.g. for Emacs,
\ ** use
\ *C   SetLocate +%l% "%f%" &
\ *P The rest of line following *\fo{SetLocate} is used as the
\ ** editor configuration string. Within the editor configuration
\ ** string 'f' will be replaced by the file name and 'l' will be
\ ** replaced by the line number. If you use file names with spaces,
\ ** you should put quotation marks around the *\f{%f%} text macro.
\ ** You must finish the line with " &" to run the editor
\ ** detached from VFX Forth - *\fo{LOCATE} adds this for you.
  0 word count locate$ place		\ Collect string
  assign .edlocate to-do .locate	\ locate through editor
  1 to EditOnError?
  assign (EditOnError) to-do EditOnError \ locate on error
;


\ ======
\ *> ###
\ ======

decimal

