\ VFXBase/Osx32/shell.fth - i32 OSX shell commands

((
Copyright (c) 2007, 2009, 2011, 2017
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20170623 SFP004 Added macro expansion.
		Added CWD words to replace CD words.
20120206 SFO003 Made $OSX a synonym for $SHELL.
20110201 SFP002 OSX conversion.
20090923 SFP001 Recoded to use the popen() call.
))

only forth definitions
decimal


\ ===========
\ *> osxtools
\ *S Shell operations
\ ===========
\ *P The VFX Forth console supports a number of command shell
\ ** operations.


\ *************
\ *N Primitives
\ *************
\ *P The words in this section are used to build the tools.

((
: PreSH		\ --
 \ *G Set up for calling a command shell.
  [ also system ] (RebufferStdIn) [ previous ]
;

: PostSH	\ --
 \ *G Restoration after calling a command shell.
  [ also system ] (SetUnbufferedStdIn) [ previous ]
;

: >Shell	\ z$ --
 \ *G Execute the given zero-terminated string as a shell command.
  PreSH  ssystem drop  PostSH
;
))

[undefined] csplit [if]
: csplit	\ caddr len char -- raddr rlen laddr llen
\ *G Extract a substring at the start of *\i{caddr/len}, returning
\ ** the string *\i{raddr/rlen} which includes *\i{char} (if found) and
\ ** the string *\i{laddr/llen} which contains the text to left of *\i{char}.
\ ** If the string does not contain the character, *\i{raddr} is
\ ** *\i{caddr+len} and *\i{rlen}=0.
  >r 2dup r> scan                       \ -- addr len addr' len'
  2swap 2 pick -                        \ -- addr' len' addr len-len'
;
[then]

: xtype		\ caddr len --
\ *G As *\fo{TYPE}, but LF characters cause a *\fo{CR}. This factor
\ ** copes with some user-written generic I/O devices that do not
\ ** implement *\fo{TYPE} correctly.
  begin
    0 max dup
   while				\ -- caddr len
    ALF csplit dup			\ split at LF
    if type else 2drop endif		\ type left string
    dup					\ check for CR
    if  cr 1 /string  endif
  repeat
  2drop
; doNotSin

0 value DebugShell?
: +DebugShell  1 to DebugShell?  ;
: -DebugShell  0 to DebugShell?  ;

: .ShellData	\ z$ -- z$
  DebugShell? if
    showmacros
    dup cr .z$
  endif
;

: >pShell	\ z$ -- ior
\ *G Execute the given zero-terminated string as a shell command,
\ ** write any output to the current output device,
\ ** and return the result code from the *\b{popen()} call.
\ ** This word provides consistent action regardless of whether
\ ** operation is running in a console or is detached.
\ ** This is the default action of *\fo{(>xShell)} below.
  { | lb[ 512 ] -- }
  .ShellData
  z" r" popen dup if
    >r
    begin
      lb[ 512 r@ fgets
     while
      lb[ zcount xtype
    repeat
    r> pclose
  endif
;

defer >xShell	\ z$ -- ior
\ *G Execute the given zero-terminated string as a shell command,
\ ** and return the result code from the relevant system call
\ ** such that zero=success. Most words that cause shell actions
\ ** use *\fo{>xShell} as a primitive.
\ ** To use a raw *\b{system} call instead as the action use:
\ *C   assign ssystem to-do >xShell
  assign >pShell to-do >xShell

also system
: (>Shell)	\ z$ -- ior
\ *G Execute the given zero-terminated string as a shell command,
\ ** and return the result code from *\fo{>xShell} above.
  TermPrepped @ if
    UnprepTerminal  >xShell  PrepTerminal
  else
    >xShell
  endif
; doNotSin

: >system	\ z$ -- ior
\ *G Execute the given zero-terminated string as a shell command
\ ** using the *\b{system()} API call, and return the result code.
  TermPrepped @ if
    UnprepTerminal  .ShellData ssystem  PrepTerminal
  else
    .ShellData ssystem
  endif
; doNotSin
previous

: >Shell	\ z$ --
\ *G Execute the given zero-terminated string as a shell command using
\ ** *\fo{(>Shell)} above.
\ ** Output from the command is written to the current output device.
\ ** Text macros are expanded before the operation.
  z$ExpandMacros (>Shell) drop
;

: ShellCmd	\ caddr len --
\ *G Execute the given *\i{caddr/len} string as a shell command.
  {: | cl[ 1024 chars ] -- :}
  1023 umin cl[ zplace				\ base command and params
  cl[ >Shell
;

: ShellLine	\ caddr len --
\ *G Execute the given counted string as a shell command. Before
\ ** execution, the remainder of the input line is added to the
\ ** given string.
  { | cl[ 1024 chars ] -- }
  cl[ zplace				\ base command and params
  0 word count dup if			\ get command tail
    s"  " cl[ zAppend			\ separator
    cl[ zAppend				\ command tail
  else
    2drop
  endif
  cl[ >Shell
;

: $shell	\ cmd$ tail$ --
\ *G Take the command and tail counted strings and execute them as
\ ** a shell command using *\b{system()}. Text macros are expanded.
  { | tbuff[ 512 ] -- }
  swap count tbuff[ zplace		\ start with the command
  s"  " tbuff[ zAppend			\ add separator
  count tbuff[ zAppend			\ add tail
  tbuff[ z$ExpandMacros ssystem drop
;

: $osx		\ cmd$ tail$ --
\ *G A synonym for *\fo{$shell}.
  $shell  ;


\ *********************
\ *N Command operations
\ *********************

: sh            \ -- ; "command"
\ *G Ask the host operating system to execute the supplied command line.
  cNull count ShellLine
;

: ls            \ -- ; "[spec]"
\ *G Display file information based on the supplied specification.
  s" ls" ShellLine
;

: dir		\ -- ; "[spec]"
\ *G Display file information based on the supplied specification.
\ ** As *\fo{LS} but with colouring.
  s" ls" ShellLine
;

: makedir         \ -- ; "name"
\ *G Create a new subdirectory from the current working one.
\ ** This word has been renamed to avoid a name conflict with
\ ** the system *\b{mkdir()} API.
  s" mkdir" ShellLine
;

: deldir         \ -- ; "name"
\ *G Remove a specified subdirectory. You can only remove an empty
\ ** directory. This word has been renamed to avoid a name conflict
\ ** with the system *\b{rmdir()} API.
  s" rmdir" ShellLine
;

: rm            \ -- ; "spec"
\ *G Delete a single file or group of files as described by the
\ ** given file specification. The wildcard '*' may also be used.
  s" rm" ShellLine
;

: cat           \ -- ; "spec"
( *G Perform an ASCII display of a file or group of files. No filtering )
( ** of the data is performed. This command should not be used to list  )
( ** binary files.                                                      )
  s" cat" ShellLine
;

: pwd           \ --
\ *G Display the currently active working directory using the shell
\ ** *\b{pwd} command.
  s" pwd" ShellLine
;

: cwd            \ -- ; ["name"]
\ *G Attempt to change current working directory either as an
\ ** offset from the current directory or as a complete path.
\ ** The wildcard '*' can be used to match the first directory.
\ ** If there is no tail, *\fo{CWD} displays the current directory.
\ ** No shell functions are used. Text macros are expanded.
  { | cl[ 1024 chars ] -- }
  0 word count dup if	\ -- caddr len
    cl[ 1023 substituteZ  cl[ chdir drop
\    cl[ zplace  cl[ chdir drop
  else
    2drop  0 cl[ c!  cl[ 1023 getcwd
    if  cl[ zcount type  endif
\    pwd
  endif
;

: cd             \ -- ; ["name"]
\ *G A synonym for *fo{CWD}. Use *fo{CWD} as *fo{CD}
\ ** will be removed in a future release to avoid conflicts
\ ** with hex numbers.
  cwd  ;


\ ======
\ *> ###
\ ======

decimal
