\ signal.fth - OS X signal handling

((
Copyright (c) 2007, 2009, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20110118 SFP002 OSX conversion.
20090416 SFP001 Added Exception frame dump.
))

only forth definitions
decimal


\ ===========
\ *> osxtools
\ *S OS X signal handling
\ ===========

\ *************
\ *N Structures
\ *************
\ *P Signal numbers
\ ** See *\i{/usr/include/sys/signal.h}.
\ *[
#define SIGHUP		 1
#define SIGINT		 2
#define SIGQUIT		 3
#define SIGILL		 4
#define SIGTRAP		 5
#define SIGABRT		 6
#define SIGIOT		 6
#define SIGEMT		 7
#define SIGFPE		 8
#define SIGKILL		 9
#define SIGBUS		10
#define SIGSEGV		11
#define SIGSYS		12
#define SIGPIPE		13
#define SIGALRM		14
#define SIGTERM		15
#define SIGURG		16
#define SIGSTOP		17
#define SIGSTP		18
#define SIGCONT		19
#define SIGCHLD		20
#define SIGTTIN		21
#define SIGTTOU		22
#define SIGIO		23
#define SIGXCPU		24
#define SIGXFSZ		25
#define SIGVTALRM	26
#define SIGPROF		27
#define SIGWINCH	28
#define SIGINFO		29
#define SIGUSR1		30
#define SIGUSR2		31
\ *]

struct /ts32	\ -- len
\ *G Regular thread state.
\ ** See *\i{/usr/include/mach/i386/_structs.h}
\ ** _STRUCT_X86_THREAD_STATE32.
  int ts.eax
  int ts.ebx
  int ts.ecx
  int ts.edx
  int ts.edi
  int ts.esi
  int ts.ebp
  int ts.esp
  int ts.ss
  int ts.eflags
  int ts.eip
  int ts.cs
  int ts.ds
  int ts.es
  int ts.fs
  int ts.gs
end-struct

struct /fs32	\ -- len
\ *G Regular FPU environment.
\ ** See *\i{/usr/include/mach/i386/_structs.h}
\ ** _STRUCT_X86_FLOAT_STATE32.
  8 +
  2 field fp.cw
  2 field fp.sw
  2 field fp.tag
  2 field fp.fop			\ opcode
  int fp.eip				\ PC
  int fp.cs				\ CS
  int fp.dp				\ data offset
  int fp.ds
  int fp.mxcsr
  int fp.mxcrrmask
  8 16 * field fp.stx/mmx		\ ST0..7 or MMX0..7
  8 16 * field fp.xmmx			\ XMM0..7
  14 16 * +
  4 +
end-struct

struct /excpt32	\ -- len
\ *G See *\i{/usr/include/mach/i386/_structs.h}
\ ** _STRUCT_X86_EXCEPTION_STATE32.
  int es.trapno
  int es.err
  int es.faultvaddr
end-struct

8 cells constant /dbg32	\ -- len ; dr0..dr7
\ *G See *\i{/usr/include/mach/i386/_structs.h}
\ ** _STRUCT_X86_DEBUG_STATE32.

4 constant /sigset_t	\ -- len
\ *G size of *\i{sigset_t}.

struct /sigaltstack	\ -- len
\ *G See *\i{/usr/include/sys/_structs.h}
  int sa.ss_sp
  int sa.ss_size
  int sa.ss_flags
end-struct

struct /mcontext	\ -- len
\ *G Machine dependent context structure.
\ ** See *\i{/usr/include/i386/_structs.h}.
  /excpt32 field mc.excpt32
  /ts32 field mc.ts32
  /fs32 field mc.fs32
end-struct

struct /ucontext	\ -- len
\ *G See *\i{/usr/include/sys/_structs.h}
  int uc.onstack
  int uc.sigmask
  /sigaltstack field uc.sigaltstack
  int uc.link				\ resuming context
  int uc.mcsize				\ size of machine context
  int uc.mcontext			\ pointer to mcontext
  /mcontext field uc.mcdata
end-struct

struct /siginfo	\ -- len
\ *G Siginfo structure in *\i{/usr/include/sys/signal.h}.
  int si.signo		/* signal number */
  int si.errno		/* errno association */
  int si.code		/* signal code */
  int si.pid		/* sending process */
  int si.uid		/* sender's ruid */
  int si.status		/* exit value */
  int si.addr		/* faulting instruction */
  int si.value		/* signal value */
  int si.band;		/* band event for SIGPOLL */
  7 4 * +		/* Reserved for Future Use */
end-struct

struct /sigaction	\ -- len
\ *G System *\b{sigaction} structure.
  int sa.action		\ action of signal, entrypoint
  int sa.mask		\ 32 bits
  int sa.flags		\ special flags
end-struct

Extern: int sigaction(int signum, void * act, void * oldact);
\ act can specify either
\   void sa_handler( int signum );
\ or
\   void sa_siginfo( int signum, siginfo_t * siginfo, ucontext_t * uc );
\ return 0 on success, -1 on error
Extern: int sigemptyset(void * set);
\ generate an empty signal set.

$0040 constant SA_SIGINFO
\ #define SA_ONSTACK	0x0001	/* take signal on signal stack */
\ #define SA_RESTART	0x0002	/* restart system on signal return */
\ #define SA_RESETHAND	0x0004	/* reset to SIG_DFL when taking signal */
\ #define SA_NOCLDSTOP	0x0008	/* do not generate SIGCHLD on child stop */
\ #define SA_NODEFER	0x0010	/* don't mask the signal we're delivering */
\ #define SA_NOCLDWAIT	0x0020	/* don't keep zombies around */
\ #define SA_SIGINFO	0x0040	/* signal handler with SA_SIGINFO args */


\ ******************
\ *N Signal handling
\ ******************
\ *P The following is the stack structure seen by the *\b{siginfo}
\ ** signal handler.
\ *E
\ **    |             |
\ **    +-------------+
\ **    | Return code |		8 bytes of code to return to kernel context at
\ **    |             |		end of the handler run via a special system call
\ **    |             |		(now a dummy left as a signature for debuggers
\ **    |             |		as many systems don't allow execution from the
\ **    |             |		stack)
\ **    +-------------+
\ **    | Floating    |		FPU state if FPU has been used by this process
\ **    | point state |
\ **    +-------------+
\ **    | User        |		OSX extension user context for the signal
\ **    | Context     |         handler. Includes pointer to register contents
\ ** +->|             |		which are modifiable by the signal handler
\ ** |  +-------------+
\ ** |  | Siginfo     |		Traditional signal information structure.
\ ** |  | structure   |<-+	Duplicates/simplifies some of User Context
\ ** |  +-------------+  |
\ ** +--| void *      |	 |	Pointer to OSX user context
\ **    +-------------+  |
\ **    | siginfo_t * |--+	Pointer to OSX Siginfo structure
\ **    +-------------+
\ **    | Signum      |		Actual signal number generated
\ **    +-------------+
\ **    | Return      |		Originally this pointed at the stack based
\ **    | address     |		return code above.  Now points directly at the
\ **    |             |		sigreturn syscall gate in the vDSO
\ **    +-------------+


create sigNames	\ -- addr
\ *G Holds the signal numbers and names as counted strings.
   0 , ", ???"	      1 , ", SIGHUP"    2 , ", SIGINT"     3 , ", SIGQUIT"
   4 , ", SIGILL"     5 , ", SIGTRAP"   6 , ", SIGIOT"     7 , ", SIGEMT"
   8 , ", SIGFPE"     9 , ", SIGKILL"  10 , ", SIGBUS"    11 , ", SIGSEGV"
  12 , ", SIGSYS"    13 , ", SIGPIPE"  14 , ", SIGALRM"   15 , ", SIGTERM"
  16 , ", SIGURG"    17 , ", SIGSTOP"  18 , ", SIGSTP"    19 , ", SIGCONT"
  20 , ", SIGCHLD"   21 , ", SIGTTIN"  22 , ", SIGTTOU"   23 , ", SIGIO"
  24 , ", SIGXCPU"   25 , ", SIGFSXZ"  26 , ", SIGVTALRM" 27 , ", SIGPROF"
  28 , ", SIGWINCH"  29 , ", SIGINFO"  30 , ", SIGUSR1"   31 , ", SIGUSR2"

: .sigName	\ n --
\ *G Given a signal number, display its name.
  dup #31 u>
  if  drop 0  endif
  sigNames swap 0 ?do
    cell +  count + aligned
  loop
  cell + count type
;

: .RSitem	\ x --
\ *G Display an item retrieved from the faulting return stack.
  cr ."  $" dup .dword			\ display stack item as number
  dup ['] firstname u>			\ above start of dictionary
  over here u< and if			\ below end of dictionary
    space  ip>nfa .name			\ show word name
  else
    drop
  endif
; doNotSin

: .mcontext	\ *mc --
\ *G Display data from the mcontext structure.
  mc.ts32				\ step to /ts32 structure
  cr
  ."  CS=" dup ts.cs w@ .word  ."  DS=" dup ts.ds w@ .word
  ."  ES=" dup ts.es w@ .word  ."  SS=" dup ts.ss w@ .word
  ."  FS=" dup ts.fs w@ .word  ."  GS=" dup ts.gs w@ .word
  cr
  ."  EAX=" dup ts.eax @ .dword  ."  EBX=" dup ts.ebx @ .dword
  ."  ECX=" dup ts.ecx @ .dword  ."  EDX=" dup ts.edx @ .dword
  cr
  ."  ESI=" dup ts.esi @ .dword  ."  EDI=" dup ts.edi @ .dword
  ."  EBP=" dup ts.ebp @ .dword  ."  ESP=" dup ts.esp @ .dword
  cr
  ."  EIP=" dup ts.EIP @ .dword  ."  EFLAGS=" dup ts.eflags @ .dword

  cr ." --- RS top ---"
  dup ts.esp @ 8 cells bounds
  do  i @ .RSitem  cell +loop
  drop
; doNotSin

: SigThrow	\ --
\ *G Runs the O/S THROW action.
  #-57005 throw
;

variable #SigNest	\ --
\ Counts Signal nesting level.

: .sigTrap	\ signum *siginfo *ucontext --
  nip					\ discard siginfo
  uc.mcontext @				\ find CPU context
  dup .mcontext				\ -- signum *mc ; display context
  cr ." Signal number " swap .sigName	\ -- *mc display signal number
  cr ." at address " ( dup ) mc.ts32 ts.EIP @ dup .dword
     ." , probably in " ip>nfa .name
  cr
;

3 0 callback: SigGenTrap	\ signum *siginfo *ucontext --
\ *G Generic trap handler that causes a *\fo{-57005 THROW} on return.
\ ** Callbacks are documented in a separate section of the manual.

1 value -NestedSigs?	\ -- x
\ *G Set non-zero to cause an exit if a nested signal exception
\ ** occurs in *\fo{(SigGenTrap)} below.
1 value SigPause?	\ -- x
\ *G Set non-zero to cause a pause when a signal is processed
\ ** in *\fo{(SigGenTrap)} below.

create restartData	\ -- addr
  up@ ,  r0 @ ,  s0 @ ,			\ UP, RSP, PSP

: setRestart	\ --
\ Set up the restart data area.
  restartData  up@ over !  r0 @ over cell + !  s0 @ swap 2 cells + !
;
' setRestart AtCold

: doRestart	\ --
\ Executed to perform a restart by running QUIT.
  restartData  dup @ up!  dup cell + @ rp!  2 cells + @ sp!
  ip-default @ ip-handle !  op-default @ op-handle !
  quit
;

: setEIP	\ xt *uc --
\ set address for signal return.
  uc.mcontext @ mc.ts32 ts.EIP !  ;

: SigNested	\ signum *siginfo *ucontext -- eip
\ Handle nested signal/fault.
  nip nip  #SigNest off			\ clear nesting
  -NestedSigs? if			\ check for permission
    cr ." Exiting on nested signal" cr
    ['] (bye)
  else
    ['] SigThrow
  endif
;

[-short-branches
: SigUnNested	\ signum *siginfo *ucontext -- eip
\ Handle unnested signal/fault.
  #SigNest incr  .sigTrap  #SigNest decr
  SigPause? if
    begin key? while key drop repeat
    cr ." Press E to exit, R to restart, other to continue: "
    case  key upc
      [char] E of  (bye)  ['] (bye)  endof
      [char] R of  ['] doRestart  endof
        ['] SigThrow swap
    endcase
  else
    ['] SigThrow
  endif
;
short-branches]

: (SigGenTrap)	\ signum *siginfo *ucontext --
\ *G Action of *\fo{SigGenTrap}. Displays an error message. On
\ ** return to OS X, a Forth *\fo{THROW} will occur.
  dup >r  #SigNest @
  if  SigNested  else  SigUnNested  endif
  r> setEIP
;
assign (SigGenTrap) to-callback SigGenTrap

: setSignal	\ callback signum --
\ *G Set a signal handler to execute the given callback. The
\ ** callback must have the stack effect *\fo{( signum *siginfo *ucontext -- )}
  { | act[ /sigaction ] -- }
  act[ sa.mask sigemptyset drop		\ clean masks
  SA_SIGINFO act[ sa.flags !		\ which type of handler
  swap act[ sa.action !			\ what to run
  act[ NULL sigaction drop		\ set it
;

: setSigTraps	\ --
\ *G Install the *\fo{SigGenTrap} signal handler for signals SIGILL,
\ ** SIGFPE, SIGBUS and SIGSEGV. Performed at startup.
  #SigNest off
  SigGenTrap SIGILL setSignal
  SigGenTrap SIGFPE setSignal
  SigGenTrap SIGSEGV setSignal
  SigGenTrap SIGBUS setSignal
; ' setSigTraps AtCold

\ Check for default I/O device
ip-default @ 0= [if]
  default-console  dup ip-default !  op-default !
[then]


\ ======
\ *> ###
\ ======

decimal

