\ TSEPOS 21 5 10 \

\ ========
\ *! clang
\ *T C Language Style Helpers
\ ========
\ *P VFX Forth supports a few "helper" definitions to aid the
\ ** parsing of "C" header files.

\ *P These are especially helpful in the parsing of Windows
\ ** resource scripts which are based on the Microsoft RC Language
\ ** for C, and for cutting and pasting from C header files.

: #define       \ <spaces"NAME"> <eol"value-def"> -- ; Exec: -- value
\ *G A simple version of C's #define preprocessor command.
\ ** Any text between the definition name and the end of the line is
\ ** *\fo{EVALUATE}d when *\fo{<NAME>} is invoked.
  create immediate
    0 word count -trailing		\ -- c-addr len
    here  over 1+ allot  place
  does>
    count evaluate
;

: //            \ --
\ *G An implementation of the C++ single line comment.
  postpone \
; immediate

: /*		\ --
\ *G A simple implementation of the C "/* ... */" comment.
  s" */" ParseUntil
; immediate

also system definitions
: next-enum	\ -- caddr len
\ +G A version of *\fo{PARSE-NAME} that works across multiple lines.
\ +* If a word cannot be obtained, the input stream is *\fo{REFILL}ed.
  begin
    [char] , parse dup 0=		\ scan input, no token?
  while					\ -- caddr len
    2drop  refill 0= #-411 ?throw	\ discard null, reload i/p
  repeat
;

: enum-term?	\ caddr len -- caddr len' flag
\ +G Return true if the last character is '}' or last two are '};'.
  2dup + 1- c@ [char] } =		\ trailing }?
  if  1- 1  exit  endif
  dup 2 >= if
   2dup + 2- s" };" s=
   if  2- 2 exit  endif
  endif
  0
;

: process-enum	\ u caddr len -- u'
\ +G Process one element of an enum.
  {: | temp[ 256 ] -- :}
  -white dup 2 >= if			\ check for C comments
    over s" //" s=
    if  2drop exit  endif
    over s" /*" s=
    if  2drop exit  endif
  endif
  dup 0= if				\ check if empty
    ( cr ." empty" ) 2drop  exit
  endif
  rot >r
  [char] = csplit 2swap dup if		\ string contains '='
    1 /string -white			\ step over '='
    isInteger? 1 <> #-13 ?throw		\ convert string
    r> drop >r				\ replace current value
  else
    2drop
  endif
\  s" EQU " temp[ place			\ build "EQU <name>"
  s" CONSTANT " temp[ place			\ build "EQU <name>"
  -white temp[ append
  r@ temp[ count evaluate		\ build constant
  r> 1+					\ -- u'
;

: (enum)	\ --
\ +G Process an enum once the name has been processed.
   0  begin
     next-enum enum-term? 0=		\ -- u caddr len ; terminator?
    while
     process-enum
   repeat
   process-enum  drop
;
previous definitions

also system
: enum		\ --
\ *G Process an enum of the form:
\ *C   enum <name> { a, b, c=10, d };
\ *P <name> is ignored. The elements appear as Forth constants.
\ ** The definition may extend over many lines. C comments may
\ ** occur after the ',' separator, e.g.
\ *C  JIM = 25, // comment about this line
  [char] { word drop  (enum)
;

: enum{		\ --
\ *G Process an enum of the form:
\ *C   enum{ a, b, c=10, d };
  (enum)
;
previous


\ ======
\ *> ###
\ ======

