\ replacements for ExternSysVx64.fth
\ starting at "Compatibility words" section.
\ Recoded: setReturn -> renamed: outMask
\ Recoded: parseArgs
\ Added: set2Ints  -> assumes existing lRANGEtype or whatever
\ Recoded: setReturn
\ Recoded: setArgs
\ test procedures at file end


\ **********************
\ *S Compatibility words
\ **********************
\ *P These words are mainly for users converting code from other
\ ** Forth systems. They provide a more "Forth-like" interface that
\ ** does not permit copy/paste from the operating system header files
\ ** or documentation.

: skipTo(	\ -- ; )
\ Step forward to a space-delimited '('. )
  begin
    parse-name dup 0= abort" missing ("	\ error at end of line ; )
    s" (" compare 0=			\ look for '('; )
  until
;

: param? \ caddr len -- flag
\ Should ignore S: F: N: R: stack indicators as parameters.
\ Leave a true flag when genuine parameter, false otherwise.
        2dup upper
        2dup s" S:" compare dup 0= if nip nip  exit then
   drop 2dup s" F:" compare dup 0= if nip nip  exit then
   drop 2dup s" N:" compare dup 0= if nip nip  exit then
   drop      s" R:" compare dup 0= if nip nip  exit then
   -1 or ;

: parse-params	\ -- #in #out
\ Parse a parameter list of the form:
\   ( a b c d -- e f )
  skipTo(				\ scan up to '('    )
  0 begin				\ -- u ; begin parsing
    parse-name  2dup s" --" compare	\ compare leaves zero flag when equal !
   while				\ -- u caddr len
    2dup param? if                      \ do we have a genuine param?
      s" ..." compare if		\ -- u
        dup 0 >=			\ if not varargs (* not ...  -rt *)
        if  1+  endif			\  count up
      else
        drop -1				\ #in=-1 for varargs  (* is ...  -rt *)
      endif
    else  2drop  then			\ ignore invalid param, try again
  repeat
   2drop                                \ get rid of dupped word+count
   0 begin                              \ begin parsing outputs
    parse-name s" )" compare
   while
     1+
  repeat
;

: doubles?  \ -- f
\ check for fp stackpicture
   >in @ parse-name rot >in ! s" (" compare 0= ;	\ ) for syntax highlighter

: outMask \ #double-rets #int-rets -- outmask
\ Create a mask from the given number of double and int returns.
   2dup 0 0 d= if 2drop 0 exit then          \ void
   2dup 0 1 d= if 2drop 1 exit then          \ int64
   2dup 0 2 d= if 2drop 2 exit then          \ 2int64s
   2dup 1 0 d= if 2drop 3 exit then          \ double
   2dup 2 0 d= if 2drop 6 exit then          \ 2doubles
   cr ." Unknown return possibility: " swap . .
   abort ;

: parseArgs  \ <text> -- #in-ints #in-doubles outmask
\ Parse a parameter list of the form:
\   ( a b c d -- e ) and an optional second ( r1 r2 r3 r4 - r5 r6 )
   parse-params           ( #ints #int-rets )
   doubles? 0= if 0 dup   ( #ints #int-rets 0 0 )
       else parse-params  ( #ints #int-rets #doubles #double-rets )
   then
   rot outmask ;

: setVoid	\ struct --
  0 over ai.Size !
  UINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: setInt64	\ struct --
  8 over ai.Size !
  SINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: set2Ints	\ struct --
  16 over ai.Size !
  lRANGEtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: setDouble	\ struct --
  8 over ai.Size !
  FLOATtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.FSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: set2Doubles	\ struct --
  16 over ai.Size !
  dPOINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.FSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: setReturn	\ outmask --
\ Set the return type as indicated by the mask
  case
    0 of  ReturnRes setVoid     endof
    1 of  ReturnRes setInt64    endof
    2 of  ReturnRes set2Ints    endof
    3 of  ReturnRes setDouble   endof
    6 of  ReturnRes set2Doubles endof
      -1 abort" Bad return data"          \ unknown mask
  endcase
;

-short-branches
: setArgs    \ #in-ints #in-doubles out-mask --
\ Generate the data handling structures. Varargs are indicated
\ by #in=-1.
   setReturn            \ set the return type
   >r ( *)                           \ dealing with the doubles later
   dup 0< if
      r> 2drop  1 to varargs?
   else
      0 to varargs?
      dup numArgs ! dup NumInts ! 0       \ set the int args first
      ?do  i ArgList[] setInt64  loop
      numArgs @   ( start index)
      r> ( *) dup numArgs +! NumFloats !  \ continue adding the doubles
      numArgs @ swap
      ?do  i ArgList[] setDouble  loop
   then
;
+short-branches

\ *P The following section provides shared library imports in the form:
\ *C   function: foo  ( a b c d -- x ) ( F: r1 r2 -- rx )
\ *P where the brackets *\b{must} be space delimited. The first argument
\ ** list represents the arguments seen by Forth as integers. The second
\ ** arguments list is optional, and is used when F.P. doubles are involved.
\ ** Imports use the default calling convention for the operating system.

also forth definitions

: FUNCTION:	\ "<name>" "<parameter list>" --
\ *G Generate a reference to an external function. The Forth name
\ ** is the same as the name of the external function.
\ ** The first parameter list is a Forth integers list plus an optional
\ ** second Forth doubles list. There should always be an integer list,
\ ** the first list, even when not used.
\ ** Use in the form:
\ *C  function: foo1 ( a b c d -- )
\ *C  function: foo2 ( a b c d -- e )
\ *C  function: foo3 ( a b c d -- e ) ( r1 r2 r3 -- )
\ *C  function: foo4 ( a b c d -- ) ( r1 r2 r3 -- r4 )
\ *C  function: foo5 ( a b c d -- ) ( r1 r2 r3 -- r4 r5 )
\ *C  function: foo6 ( -- ) ( r1 r2 r3 -- r4 )
\ *C  function: foo7 ( -- ) ( r1 r2 r3 -- r4 r5 )
\ *C  function: foo7 ( -- a ) ( r1 r2 r3 -- )
\ *P The inclusion of Forth stack indicators like S: F: N: R: is optional.
\ *P They will be ignored while parsing and are only informative on which
\ *P Forth stack you must place the parameters.
\ *P The returned value may be 0, 1 integer or 0, 1, 2 doubles
\ ** corresponding to void, int/long and 64 bit double
  initArgData  setFirstLine		\ SFP022
  singleName				\ name of Forth word and DLL func ; SFP020
  setDefExt				\ set default declaraction type
  parseArgs setArgs			\ the parameter parsing and setup for...
  GenerateImport
;

: ASCALL:	\ "<synonym-name>" "<name>" "<parameter list>" --
\ *G Generate a reference to an external function. The Forth name
\ ** is not the same as the name of the external function.
\ ** Use in the form:
\ *C  ascall: forthname funcname ( a b c d -- e )
  initArgData  setFirstLine		\ SFP022
  parse-name ExternName place		\ set name of Forth word
  parse-name ImportName place		\ set name of DLL function
  setDefExt				\ set default declaraction type
  parseArgs setArgs                     \ the parameter parsing and setup for...
  GenerateImport
;

: GLOBAL:	\ "<name>" --
\ *G Generate a reference to an external variable.
\ ** Use in the form:
\ *C  global: varname
  singleName genExternVar		\ generate variable ; SFP020
;

previous definitions

\ ================================

Test procedure:
Launch vfxosx64 = VfxForthOsx64/Bin/VfxForth_x64_mac.mo
Set vfxpath with   c" %load_path%" $ExpandMacros dup stripfilename setmacro vfxpath
and typepaste      include %vfxpath%/Lib/Osx64/Cocoa/new-coco64.bld
Launch coco-vfx64 in VfxForthOsx64/Bin

\ test with:
cocoa.framework
function: NSRangeFromString ( NSStringRef -- location length )
@" 1234 567" NSRangeFromString . .          \ expect: 567 1234
