\ PARSEL.FTH - long string parsing

((
Copyright (c) 2001, 2006
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20060809 SFP002 Added INITSTRINGBUFF to cold chain.
20010629 SFP001 Corrected (W$+) and replaced CMOVE with MOVEX
))

\ ========
\ *> tools
\ *S Long Strings
\ ========
( *P In order to extract very long strings from the source code, the	)
( ** word PARSE/L is provided. Support is also provided for counted	)
( ** strings with a 16 bit count. These words allow long strings such	)
( ** as those required for internationalisation to be generated without	)
( ** the restrictions of counted strings that use a character-sized	)
( ** count.								)

( *P The contents of this section are subject to change until the ANS	)
( ** Forth committee reaches a conclusion about internationalisation	)
( ** issues. An implementation of the					)
( ** system descibed in the paper on the MPE web site may be found	)
( ** in the file %LIB%\INTERNATIONAL.FTH.				)
\ =======================================================================

also system definitions

variable ^StringBuff	\ -- addr ; pointer to buffer on heap
variable #StringBuff	\ -- addr ; length used in buffer

: initStringBuff	\ --
  ^StringBuff off  #StringBuff off
;
' initStringBuff AtCold

: FreeStringBuff	\ -- ; release string buffer
  ^StringBuff @ ?dup
  if  free drop  initStringBuff  endif
;
' FreeStringBuff AtExit

: AllocStringBuff	\ -- ; ensure buffer is allocated
  FreeStringBuff
  #32768 allocate abort" Can't allocate string buffer"
  ^StringBuff !  #StringBuff off
;

: parse-more	\ char -- flag c-addr len ; flag is true if more needed
  >r                            \ character to scan for
  source >in @ /string          \ skip start of buffer ; -- addr(s) len(s)
  over swap r> scan >r          \ scan for terminator ; -- addr(s) addr(e)
  over -                        \ calculate string ; -- addr(s) len(s)
  dup r@ 0<> - >in +!           \ calculate length to update >IN
  r> 0= -rot                    \ return flag
;

previous definitions

also system
: parse/l       \ char -- c-addr len ; like PARSE over lines
( *G Parse the next token from the terminal input buffer using <char> as)
( ** the delimiter. The text up to the delimiter is returned as a       )
( ** c-addr u string. PARSE/L does not skip leading delimiters. In      )
( ** order to support long strings, PARSE/L can operate over multiple	)
( ** lines of input and line terminators are not included in the text.  )
( ** The string returned by PARSE/L remains in a single global buffer   )
( ** until the next invocation of PARSE/L. PARSE/L is designed for use  )
( ** at compile time and is not thread-safe or winproc-safe.		)
  AllocStringBuff
  begin					\ -- char
    dup parse-more			\ -- char flag c-addr len
    tuck ^StringBuff @ #StringBuff @ + swap move
    #StringBuff +!
   while
    refill 0= abort" Missing closing double quotes"
  repeat
  drop
  ^StringBuff @  #StringBuff @
;
previous

((  Moved to KERNEL.FTH
: wcount         \ addr1 -- addr2 len
( *G Given the address of a word-counted string in memory WCOUNT will   )
( ** return the address of the first character and the length in        )
( ** characters of the string.                                          )
  dup 2+ swap w@
;

-sin
: (W")		\ -- waddr u ; step over caller's in line string
( *G Returns the address and length of inline 16-bit word-counted string. )
( ** Steps over inline text. )
  r> wcount 2dup +  2+  aligned >r	\ step over count, string, 0, align
;

: ((W"))	\ -- waddr u ; dangerous factor!
\ *G A factor provided for the generation of long string actions that
\ ** have to step over an inline string. For example, to define W."
\ ** which uses a long string, you might compile (W.") and then use
\ ** W", to compile the inline string. The definition of (W.") then
\ ** might be:
\ *E : (W.")     \ --
\ **   ((W")) type
\ ** ;
  r>					\ caller's return address
  r> wcount 2dup +  2+ aligned >r
  rot >r
;
+sin
))

: wAppend	\ c-addr u $dest --
( *G Add string c-addr u to the end of word-counted string $dest. )
  >r  tuck  r@ wcount + swap movex      \ add source to end
  r@ w@ + r> w!                         \ add length to count
;

: (w$+)        \ c-addr u $dest -- ; SFP001
( *G Add string c-addr u to the end of word-counted string $dest. )
  >r  tuck  r@ wcount + swap movex      \ add source to end
  r@ w@ + r> w!                         \ add length to count
;

: w$+           \ $src $dest -- ; add $SRC to end of $DEST
( *G Add word-counted string $src to the end of word-counted string $dest. )
  >r wcount r> wAppend
;

: w$,		\ caddr len --
\ *G Lay a 16 bit string string into the dictionary at *\fo{HERE}, reserve
\ ** space for it and *\fo{ALIGN} the dictionary. The  inline string string
\ ** has a 16 bit count and 16 bit zero termination.
  tuck dup w,                           \ lay count word
  here swap move                        \ copy string
  allot                                 \ reserve space for string
  0 w,                                  \ trailing 0
  align                                 \ cell align it
;

: w",		\ -- ; compile a word counted string
\ *G multiline version of ",.
\ ** Interprets multiline text and lays down inline string string with 16 bit
\ ** count and 16 bit zero termination.
  [char] " parse/l  w$,
;

: ls"		\ c: -- ; i: -- caddr len
\ *G A version of *\fo{S"} that can extend over several lines.
\ ** Line separators are ignored.
  [char] " parse/l  ;			\ -- caddr len
ndcs: ( -- )  discard-sinline
  postpone (w")  w",  ;

: zls"		\ c: -- ; i: -- zaddr
\ *G A version of *\fo{Z"} that can extend over several lines.
\ ** Line separators are ignored. The returned address is that
\ ** of the start of the zero-terminated string.
  [char] " parse/l drop  ;			\ -- caddr
ndcs: ( -- )  discard-sinline
  postpone (w")  w",  postpone drop  ;


\ ========
( *> ### )
\ ========
