\ Random64.fth - random number generation.

((
Copyright (c) 2021
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20210429 SFP001 Rewrite.
))

only forth definitions
decimal

\ ========
\ *> tools
\ *S Random numbers
\ ========
\ *P The random number system in VFX Forth is based around
\ ** the *\fo{DEFER}red word *\fo{RANDOM} (see below). We have
\ ** found that a single random number generator (RNG) is not adequate
\ ** for all applications. Some applications need a particular
\ ** degree of randomness, others require more speed. If you do
\ ** not like the default RNG, you can install your own.

\ *P The implementation uses a seed in the variable *\fo{RandSeed),
\ ** which is set to some time value at start up.
\ ** The default implementation is in the *\fo{SYSTEM} vocabulary.

also system definitions

\ Fast Random Number Generator
\ algorithm by George Marsaglia "Xorshift RNGs"

variable RandSeed	\ -- addr
\ *G Seed value used by the RNG.
  88172645463325252 RandSeed !

: (RANDOM)	( -- x )
\ Default action of the RNG - it's good, fast, and by Marsaglia.
  RandSeed @
  dup 13 lshift xor
  dup 7 rshift xor
  dup 17 lshift xor
  dup RandSeed !
;
((
unsigned long long xor64(){
     static unsigned long long x=88172645463325252LL;
     x=(x<<13); x=(x>>7); return ( x=(x<<17) );
))
previous definitions

defer RANDOM	\ -- u
\ *G Generate a random number.
also system  assign (random) to-do RANDOM  previous

: CHOOSE 	\ n1 -- n2
\ *G Generate a random number *\i{n2) in the range 0..n1-1. The algorithm is
\ ** from Paul Mennen, 1991 .
  RANDOM UM* NIP
;

: +randDigits	\ buff$ -- ; 64 bit version
\ *G Add a 32 bit random number as 8 hex digits at the end
\ ** of the counted string in *\i{buff}. Used for generating
\ ** random names for things like semaphores.
  $1:0000:0000 choose swap		\ -- x buff
  8 0 do
    over #28 rshift $0F and		\ -- x buff nib ; top 4 bits
    dup 9 > if  7 +  then [char] 0 +	\ -- x buff char
    over addchar			\ -- x buff
    swap 4 lshift swap			\ -- x' buff
  loop
  0 over count + c!			\ zero terminate
  2drop
;


((
Andrew Haley posted on comp.lang.forth

variable rng

: random ( - n)
    rng @
    dup 0= or
    dup 6 lshift xor
    dup 21 rshift xor
    dup 7 lshift xor
    dup rng ! ;

This generates 1..FFFF:FFFF
))

\ ======
\ *> ###
\ ======

decimal

