\ RESOURCE.FTH - resource structure management

((
Copyright (c) 2001, 2002
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====
Convert styles to accept any Windows constant. This will allow us to
remove alarge number of string definitions.

Change history
==============
20020326 SFP001 Updated documentation.
))


\ ======================================================================
( *! win32rc                                                           )
( *T Windows Resource Compiling                                        )
\ *S Introduction
\ *P VFX Forth builds its GUI interfaces from Windows format resource
\ ** scripts with some extensions. These are commonly seens as RC files
\ ** produced by Windows resource editors. VFX Forth follows this
\ ** notation as far as is feasible. Resource compiling makes heavy use
\ ** of the BNF parser described elsewhere in this manual.

\ *P Because Windows resource editors include many C and C++ declarations
\ ** we have provided the tool BIN\STRPMSRC.EXE which produces an output
\ ** file for VFX Forth from Windows RC files. The syntax is:
\ *C   StrpMsRc <resource.h> <script.rc> <output>
\ *P This tool is available from the Tools menu of the Studio environment.

\ *P The header file <resource.h> is processed before the script file
\ ** <script.rc> and the resultant output is placed in <output>. The
\ ** output file <output> can then be imported into VFX Forth with only
\ ** minimal (if any) editing.
\ ======================================================================

               MODULE  RESOURCE

\ ======================================================================
( **                                                                   )
( *S Common Resource Structure Format                                  )
( **                                                                   )
( *P Each compiled RC is built using a data structure which is linked  )
( ** to a global list within Forth. During compilation of various      )
( ** resources enough information is built into the dictionary to a    )
( ** run-time routine to be able to create the resource.               )
( **                                                                   )
( *P The format of the internal data varies from resource to resource  )
( ** and is documented in each resource's separate source file.        )
( **                                                                   )
( *P There is however a 'common' structure for the internal resources  )
( ** which is always at the beginning and is described as:            )
( **                                                                   )
( *D CELL       Resource Handle. If the resource is created this holds )
( **            the windows handle for it. Otherwise it is NULL        )
( *D CELL       Previous Resource. A link to the previously defined    )
( **            resource.                                              )
( *D CELL       Creation Code. Pointer to code which will create       )
( **            this resource.                                         )
( *D CELL       Resource Type. One of the RESTYPE_xxx codes described  )
( **            later.                                                 )
( *D CELL       Resource ID. The Unique ID of the resource created.    )
( *D xxx CELLS  Resource Specific Data - Start of Custom Section -     )
( **                                                                   )
( *P For ease of navigation this 'header' is described by a structure  )
( ** as defined below:                                                 )
( **                                                                   )
( *C    struct  IRESOURCE                                              )
( **            CELL    field   IRESOURCE.handle                       )
( **            CELL    field   IRESOURCE.*prevlink                    )
( **            CELL    field   IRESOURCE.dwResType                    )
( **            CELL    field   IRESOURCE.dwResID                      )
( **            0       field   IRESOURCE.*CustomData                  )
( **    end-struct                                                     )
\ ======================================================================

STRUCT  IRESOURCE
        CELL    field   IRESOURCE.handle
        CELL    field   IRESOURCE.*prevlink
        CELL    field   IRESOURCE.dwResType
        CELL    field   IRESOURCE.dwResID
        CELL    field   IRESOURCE.*CustomData
END-STRUCT

$0010 constant RESTYPE_BITMAP
$0020 constant RESTYPE_CURSOR
$0030 constant RESTYPE_ICON
$0040 constant RESTYPE_TOOLBAR
$0050 constant RESTYPE_WINDOW
$0060 constant RESTYPE_MENU
$0070 constant RESTYPE_DIALOG
$0080 constant RESTYPE_COOLBAR

\ =======================================================================
( **                                                                    )
( *S Resource List Words                                                )
( **                                                                    )
( *P Useful words for use with the list of Resources. Covers searching  )
( ** modifying etc.                                                     )
( **                                                                    )
\ =======================================================================

: .ResType              \ type --
( +G Print human readable string for a RESTYPE_xxx constant. Must be 9  )
( +* characters long.                                                   )
  case
    RESTYPE_BITMAP of   ." Bitmap   "   endof
    RESTYPE_CURSOR of   ." Cursor   "   endof
    RESTYPE_ICON   of   ." Icon     "   endof
    RESTYPE_TOOLBAR of  ." Toolbar  "   endof
    RESTYPE_WINDOW of   ." Window   "   endof
    RESTYPE_MENU of     ." Menu     "   endof
    RESTYPE_DIALOG of   ." Dialog   "   endof
    RESTYPE_COOLBAR of  ." Coolbar  "   endof
    drop                ." <Unknown>"
  end-case
;

[-short-branches
: .Resources	\ --
( *G Display defined Resources in the list maintained by VFX Forth.	)

  cr
  cr ." Defined Windows Resources"
  cr
  cr ." Handle          Type            Id              *CustomData"
  cr ." -----------------------------------------------------------"
  0 >r
  res-link
  begin
    @ ?dup
  while
    r> 1+ >r
    >r
    cr
    r@ IRESOURCE.handle @    .lword   3 spaces [char] | emit 3 spaces
    r@ IRESOURCE.dwResType @ .ResType 3 spaces [char] | emit 3 spaces
    r@ IRESOURCE.dwResId @   .lword   3 spaces [char] | emit 3 spaces
    r@ IRESOURCE.*CustomData .lword
    r> IRESOURCE.*prevlink
  repeat
  cr ." -----------------------------------------------------------"
  cr r> u. ." resources defined."
  cr ." -----------------------------------------------------------"
  cr cr
;
short-branches]

: RESOURCE::LayHeader	\ id type --
( +G Lay the generic part of the Resource structure - IRESOURCE - and   )
( +* link it to the resource list.                                      )
        here
        NULL ,                  \ handle
        res-link @ ,            \ *prevlink
        res-link !
        ,                       \ dwResType
        ,                       \ dwResID
;

: RESOURCE::Find	\ id type -- *resource | 0
  2>r
  res-link
  begin
    @ ?dup
  while
    dup  IRESOURCE.dwResId @
    over IRESOURCE.dwResType @          \ *resource rid rtype --
    2r@                                 \ *resource rid rtype id type --
    rot =                               \ *resource rid id typeflag --
      -rot =                            \ *resource typeflag idflag --
    AND if
      2r> 2drop exit
    then
    IRESOURCE.*prevlink
  repeat
  2r> 2drop NULL
;

: RESOURCE::GetHandle   \ id type -- handle
  RESOURCE::Find ?dup if
    IRESOURCE.handle @
  else
    NULL
  then
;

\ ===============
\ *S Resource IDs
\ ===============

: NextID:       \ "<name>" -- ; NEXTID: <name>
( *G Produces a constant with the next available 16-bit ID number, and )
( ** increments the next available number. Used to produce unique IDs )
( ** for Windows resources. )
  <id> @ constant  <id> incr
;

\ =============================================================================

bl token <sp>
9 token <tab>
0 token <eol>
char , token <comma>
char " token <quote>
char | token <bar>

char 0 token '0'       char 1 token '1'
char 2 token '2'       char 3 token '3'
char 4 token '4'       char 5 token '5'
char 6 token '6'       char 7 token '7'
char 8 token '8'       char 9 token '9'

string  'NULL'                  NULL

string  'BEGIN'                 BEGIN
string  'END'                   END
string  'CAPTION'               CAPTION
string  'STYLE'                 STYLE
string  '+STYLE'                +STYLE
string  'EXSTYLE'               EXSTYLE
string  'ICON'                  ICON
string  'CURSOR'                CURSOR
string  'BRUSH'                 BRUSH
string  'MENU'                  MENU
string  'BUTTON'                BUTTON
string  'SEPARATOR'             SEPARATOR
string  'MENUITEM'              MENUITEM
string  'POPUP'                 POPUP
string  'CLASSSTYLE'            CLASSSTYLE
string  'WNDPROC'               WNDPROC
string  'TOOLBAR'               TOOLBAR
string  'FONT'                  FONT
string  'CLASS'                 CLASS
string  '+HOTLIST'              +HOTLIST

string  'DEFPUSHBUTTON'         DEFPUSHBUTTON
string  'PUSHBUTTON'            PUSHBUTTON
string  'LTEXT'                 LTEXT
string  'RTEXT'                 RTEXT
string  'CTEXT'                 CTEXT
string  'EDITTEXT'              EDITTEXT
string  'CONTROL'               CONTROL
string  'GROUPBOX'              GROUPBOX
string  'LISTBOX'               LISTBOX
string  'COMBOBOX'              COMBOBOX


\ ************
\ CLASS STYLEs
\ ************

string  'CS_BYTEALIGNCLIENT'    CS_BYTEALIGNCLIENT
string  'CS_BYTEALIGNWINDOW'    CS_BYTEALIGNWINDOW
string  'CS_CLASSDC'            CS_CLASSDC
string  'CS_DBLCLKS'            CS_DBLCLKS
string  'CS_GLOBALCLASS'        CS_GLOBALCLASS
string  'CS_HREDRAW'            CS_HREDRAW
string  'CS_NOCLOSE'            CS_NOCLOSE
string  'CS_OWNDC'              CS_OWNDC
string  'CS_PARENTDC'           CS_PARENTDC
string  'CS_SAVEBITS'           CS_SAVEBITS
string  'CS_VREDRAW'            CS_VREDRAW

\ ********
\ ExSTYLEs
\ ********

string  'WS_EX_CLIENTEDGE'      WS_EX_CLIENTEDGE
string  'WS_EX_DLGMODALFRAME'   WS_EX_DLGMODALFRAME
string  'WS_EX_NOPARENTNOTIFY'  WS_EX_NOPARENTNOTIFY
string  'WS_EX_TOPMOST'         WS_EX_TOPMOST
string  'WS_EX_ACCEPTFILES'     WS_EX_ACCEPTFILES
string  'WS_EX_TRANSPARENT'     WS_EX_TRANSPARENT
string  'WS_EX_MDICHILD'        WS_EX_MDICHILD
string  'WS_EX_TOOLWINDOW'      WS_EX_TOOLWINDOW
string  'WS_EX_WINDOWEDGE'      WS_EX_WINDOWEDGE
string  'WS_EX_CONTEXTHELP'     WS_EX_CONTEXTHELP
string  'WS_EX_RIGHT'           WS_EX_RIGHT
string  'WS_EX_LEFT'            WS_EX_LEFT
string  'WS_EX_RTLREADING'      WS_EX_RTLREADING
string  'WS_EX_LTRREADING'      WS_EX_LTRREADING
string  'WS_EX_LEFTSCROLLBAR'   WS_EX_LEFTSCROLLBAR
string  'WS_EX_RIGHTSCROLLBAR'  WS_EX_RIGHTSCROLLBAR
string  'WS_EX_CONTROLPARENT'   WS_EX_CONTROLPARENT
string  'WS_EX_STATICEDGE'      WS_EX_STATICEDGE
string  'WS_EX_APPWINDOW'       WS_EX_APPWINDOW
string  'WS_EX_OVERLAPPEDWINDOW' WS_EX_OVERLAPPEDWINDOW
string  'WS_EX_PALETTEWINDOW'   WS_EX_PALETTEWINDOW

\ *****
\ ICONs
\ *****

string  'IDI_APPLICATION'       IDI_APPLICATION
string  'IDI_HAND'              IDI_HAND
string  'IDI_QUESTION'          IDI_QUESTION
string  'IDI_EXCLAMATION'       IDI_EXCLAMATION
string  'IDI_ASTERISK'          IDI_ASTERISK
string  'IDI_WINLOGO'           IDI_WINLOGO
string  'IDI_WARNING'           IDI_WARNING
string  'IDI_ERROR'             IDI_ERROR
string  'IDI_INFORMATION'       IDI_INFORMATION

\ *******
\ CURSORs
\ *******

string  'IDC_ARROW'             IDC_ARROW
string  'IDC_IBEAM'             IDC_IBEAM
string  'IDC_WAIT'              IDC_WAIT
string  'IDC_CROSS'             IDC_CROSS
string  'IDC_UPARROW'           IDC_UPARROW
string  'IDC_SIZE'              IDC_SIZE
string  'IDC_ICON'              IDC_ICON
string  'IDC_SIZENWSE'          IDC_SIZENWSE
string  'IDC_SIZENESW'          IDC_SIZENESW
string  'IDC_SIZEWE'            IDC_SIZEWE
string  'IDC_SIZENS'            IDC_SIZENS
string  'IDC_SIZEALL'           IDC_SIZEALL
string  'IDC_NO'                IDC_NO
string  'IDC_APPSTARTING'       IDC_APPSTARTING
string  'IDC_HELP'              IDC_HELP

\ ************
\ STOCKOBJECTs
\ ************

string  'BLACK_BRUSH'           BLACK_BRUSH
string  'DKGRAY_BRUSH'          DKGRAY_BRUSH
string  'GRAY_BRUSH'            GRAY_BRUSH
string  'HOLLOW_BRUSH'          HOLLOW_BRUSH
string  'LTGRAY_BRUSH'          LTGRAY_BRUSH
string  'NULL_BRUSH'            NULL_BRUSH
string  'WHITE_BRUSH'           WHITE_BRUSH

\ ************
\ Brush COLORs
\ ************

string  'COLOR_ACTIVEBORDER'    COLOR_ACTIVEBORDER
string  'COLOR_ACTIVECAPTION'   COLOR_ACTIVECAPTION
string  'COLOR_APPWORKSPACE'    COLOR_APPWORKSPACE
string  'COLOR_BACKGROUND'      COLOR_BACKGROUND
string  'COLOR_BTNFACE'         COLOR_BTNFACE
string  'COLOR_BTNSHADOW'       COLOR_BTNSHADOW
string  'COLOR_BTNTEXT'         COLOR_BTNTEXT
string  'COLOR_CAPTIONTEXT'     COLOR_CAPTIONTEXT
string  'COLOR_GRAYTEXT'        COLOR_GRAYTEXT
string  'COLOR_HIGHLIGHT'       COLOR_HIGHLIGHT
string  'COLOR_HIGHLIGHTTEXT'   COLOR_HIGHLIGHTTEXT
string  'COLOR_INACTIVEBORDER'  COLOR_INACTIVEBORDER
string  'COLOR_INACTIVECAPTION' COLOR_INACTIVECAPTION
string  'COLOR_MENU'            COLOR_MENU
string  'COLOR_MENUTEXT'        COLOR_MENUTEXT
string  'COLOR_SCROLLBAR'       COLOR_SCROLLBAR
string  'COLOR_WINDOW'          COLOR_WINDOW
string  'COLOR_WINDOWFRAME'     COLOR_WINDOWFRAME
string  'COLOR_WINDOWTEXT'      COLOR_WINDOWTEXT

\ ************
\ Discardables
\ ************

string  'PRELOAD'               PRELOAD
string  'LOADONCALL'            LOADONCALL
string  'FIXED'                 FIXED
string  'MOVEABLE'              MOVEABLE
string  'DISCARDABLE'           DISCARDABLE
string  'PURE'                  PURE
string  'IMPURE'                IMPURE

\ **********
\ MENU Flags
\ **********

string  'CHECKED'               CHECKED
string  'GRAYED'                GRAYED

\ =============================================================================

: z$>U$,                \ *zaddr --
\ +G Given a 0 terminated string pointer lay it in the dictionary
\ +* as a unicode string.
  MAX_PATH bounds ?do
    i c@ dup c, 0 c,
    0= if unloop exit then
  loop
;

: FindWinConstant       \ c-addr u -- value
( +G Given an addr/len string, look it up in the WINCONSTANTS	)
( +* vocabulary	and return its value if found, or 0 if not found. )
  ['] WinConstants voc>wid search-wordlist
  if  execute  else  success off  0  then
;

::= <whitespacechar>    \ --
  <sp> | <tab> | <comma>
;;

::= <whitespacecharX>   \ --
  <sp> | <tab> | <comma> | <quote>
;;

::= <whitespace>	\ --
  [[ <whitespacechar> ]]
;;

((
::= <whitespaceOrquote>	\ --
  [[ <whitespacecharX> ]]
;;
))

::= <char>	\ --
  @token
  dup [char] , = if
     drop FALSE
  else
    #33 #255 within?
  then
  dup success !  +token
;;

::= <nchar>	\ --
  bnf-ignore-lines @ >r  bnf-ignore-lines off
  <char>
  r> bnf-ignore-lines !
;;

::= <cchar>	\ --
  @token
  false >R
  dup [char] 0 [char] 9 within? if r> drop true >r then
  dup [char] a [char] z within? if r> drop true >r then
  dup [char] A [char] Z within? if r> drop true >r then
  [char] _ = if r> drop true >r then
  r> dup success !  +token
;;

::= <ncchar>	\ --
  bnf-ignore-lines @ >r  bnf-ignore-lines off
  <cchar>
  r> bnf-ignore-lines !
;;

::= <word>	\ -- caddr len
        <whitespace>
        { source drop >in @ + 0 }
        <char> { 1+ }
        [[  <nchar> { 1+ }  ]]
;;

::= <cword>	\ -- caddr len
        <whitespace>
        { source drop >in @ + 0 }
        <cchar> { 1+ }
        [[  <ncchar> { 1+ }  ]]
;;

::= <digit>	\ --
          '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
;;

::= <ndigit>	\ n -- n'
  bnf-ignore-lines @ >r  bnf-ignore-lines off
  <digit>
  r> bnf-ignore-lines !
;;

::= <number>	\ -- n
          { 0 }                                 \ initial accumulator
          <whitespace>
\          <<  <digit> { #10 *  last-token @ [char] 0 - + }  >>
          <digit> { #10 *  last-token @ [char] 0 - + }
          [[  <ndigit> { #10 *  last-token @ [char] 0 - + }  ]]
;;

::= <discardable?>
        'PRELOAD'
        | 'LOADONCALL'
        | 'FIXED'
        | 'MOVEABLE'
        | 'DISCARDABLE'
        | 'PURE'
        | 'IMPURE'
;;

::= <any-discardable-tokens>
        [[
        <whitespace>
        <discardable?>
        ]]
;;

::= <_BEGIN_>
        <whitespace>
        'BEGIN'
;;

::= <_END_>
        <whitespace>
        'END'
;;


\ =-=-=-=-=-=-= EXSTYLE =-=-=-=-=-=-=-=

::= <exstyle-constant>
          'NULL'                  { NULL }
        | 'WS_EX_CLIENTEDGE'      { WS_EX_CLIENTEDGE }
        | 'WS_EX_DLGMODALFRAME'   { WS_EX_DLGMODALFRAME }
        | 'WS_EX_NOPARENTNOTIFY'  { WS_EX_NOPARENTNOTIFY }
        | 'WS_EX_TOPMOST'         { WS_EX_TOPMOST }
        | 'WS_EX_ACCEPTFILES'     { WS_EX_ACCEPTFILES }
        | 'WS_EX_TRANSPARENT'     { WS_EX_TRANSPARENT }
        | 'WS_EX_MDICHILD'        { WS_EX_MDICHILD }
        | 'WS_EX_TOOLWINDOW'      { WS_EX_TOOLWINDOW }
        | 'WS_EX_WINDOWEDGE'      { WS_EX_WINDOWEDGE }
        | 'WS_EX_CONTEXTHELP'     { WS_EX_CONTEXTHELP }
        | 'WS_EX_RIGHT'           { WS_EX_RIGHT }
        | 'WS_EX_LEFT'            { WS_EX_LEFT }
        | 'WS_EX_RTLREADING'      { WS_EX_RTLREADING }
        | 'WS_EX_LTRREADING'      { WS_EX_LTRREADING }
        | 'WS_EX_LEFTSCROLLBAR'   { WS_EX_LEFTSCROLLBAR }
        | 'WS_EX_RIGHTSCROLLBAR'  { WS_EX_RIGHTSCROLLBAR }
        | 'WS_EX_CONTROLPARENT'   { WS_EX_CONTROLPARENT }
        | 'WS_EX_STATICEDGE'      { WS_EX_STATICEDGE }
        | 'WS_EX_APPWINDOW'       { WS_EX_APPWINDOW }
        | 'WS_EX_OVERLAPPEDWINDOW' { WS_EX_OVERLAPPEDWINDOW }
        | 'WS_EX_PALETTEWINDOW'   { WS_EX_PALETTEWINDOW }
        | <cword> { FindWinConstant }
;;


::= <exstyle>
        <whitespace>
        <exstyle-constant>
        [[
                <whitespace>
                <bar>
                <whitespace>
                <exstyle-constant>
                { OR }
        ]]
;;

\ =-=-=-=-=-=-= STYLE =-=-=-=-=-=-=-=

::= <style-constant>
          'NULL'                { NULL }
        | <cword> { FindWinConstant }
;;

::= <style>
        <whitespace>
        <style-constant>
        [[
                <whitespace>
                <bar>
                <whitespace>
                <style-constant>
                { OR }
        ]]
;;

\ ======================================================================
( **                                                                   )
\ ======================================================================

S" Resource\BITMAP.FTH"   INCLUDED
S" Resource\ICON.FTH"     INCLUDED
S" Resource\CURSOR.FTH"   INCLUDED
S" Resource\MENU.FTH"     INCLUDED
S" Resource\TOOLBAR.FTH"  INCLUDED
S" Resource\DIALOG.FTH"   INCLUDED
S" Resource\WINDOW.FTH"   INCLUDED
S" Resource\COOLBAR.FTH"  INCLUDED

\ ======================================================================
( **                                                                   )
\ ======================================================================

EXPORT NextID:
EXPORT .Resources

EXPORT Bitmap
EXPORT Coolbar
EXPORT Cursor
EXPORT Dialog
EXPORT DialogEx
EXPORT Icon
EXPORT Menu
EXPORT Toolbar
EXPORT Window

EXPORT Resource::CreateBitmap
EXPORT Resource::CreateCoolbar
EXPORT Resource::CreateCursor
EXPORT Resource::GetDialogTemplate
EXPORT Resource::CreateIcon
EXPORT Resource::CreateMenu
EXPORT Resource::CreatePopupMenu
EXPORT Resource::CreateToolbar
EXPORT Resource::CreateParentWindow
EXPORT Resource::CreateChildWindow

EXPORT Resource::GetHandle

EXPORT RESTYPE_BITMAP
EXPORT RESTYPE_CURSOR
EXPORT RESTYPE_ICON
EXPORT RESTYPE_TOOLBAR
EXPORT RESTYPE_WINDOW
EXPORT RESTYPE_MENU
EXPORT RESTYPE_DIALOG
EXPORT RESTYPE_COOLBAR

export ChangeMenu

END-MODULE

NextID: IDC_STATIC

( *> ###                                                               )


