0 [IF]

Since I have for the most part, directly copied the Mops string class, I will
also simply copy the pertinent parts of the Mops Manual that describe how to
use these strings.

Mops strings are implemented as relocatable blocks of heap that can expand and
contract as their contents change.  A string object itself contains a handle
to the heap block that contains the string's data.  It also contains three
other ivars which we will describe below. 

Strings can be useful for a wide variety of programming needs.  They can serve
as file buffers, staging areas for text to be printed on the screen,
dictionaries, or vehicles for parsing user input.  You should consider using
strings for any run of bytes whose length and/or contents are likely to change
in the course of your program's execution.  Strings are not restricted to
ASCII text, although that will probably be their most common use.

Using strings is somewhat like using files, in that you must open the string
before you use it and close it when you're through.  This is done by sending a
New: message to each string before you use it, to allocate the string's heap
storage, and then sending a Release: message when you no longer need the
string.  Release: is actually inherited from String's superclass, Ptr.

Strings have a current size, which is the same as the length of the
relocatable block of heap containing the string's data.  Strings also have two
offets into the string data, called POS and LIM.  POS marks the current
position, and LIM the current end.  Most string operations operate on the
substring delimited by POS and LIM, which we call the active part of the
string, rather than the whole string.  We also keep the size of the string
(the whole size, that is) in an ivar, so that we can get it quickly without a
system call. 

ACCESSING 
pos:    ( -- n )    Returns the value of Pos 
>pos:   ( n -- )    Stores n in Pos 
lim:    ( -- n )    Returns the value of Lim 
>lim:   ( n -- )    Stores n in Lim 
len:    ( -- n )    Returns the value of Lim - Pos, i.e.  the length of the active
                    part
>len:   ( n -- )    Adds n and Pos, and stores the result in Lim 
skip:   ( n -- )    Adds n to Pos 
more:   ( n -- )    Adds n to Lim 
start:  ( --   )    Clears Pos, so that the active part now starts at the real
                    start of the string. 
begin:  ( -- )      Sets both Pos and Lim to 0.  Useful for setting up for an iterative
                    operation on the string
end:    ( -- )      Sets both Pos and Lim to the size (i.e.  the end) of the string. 
                    Useful for setting up for an iterative operation which has to go backwards
                    through the string
nolim:  ( -- )      Sets Lim to the end of the string 
reset:  ( -- )      Sets Pos to 0 and sets Lim to the end of the string.  The active
                    part will now be the whole string
step:   ( -- )      Steps forward in the string, setting Pos to Lim and then setting Lim to the end of the string 
<step:  ( -- )      Steps backward in the string, setting Lim to Pos and then clearing Pos 

MANIPULATION 
new: ( -- ) Creates a heap block for the string's data, and
    sets the handle.  The initial size is zero.  new: must be done
    before the string can be used
?new: ( -- ) Ensures a heap block is allocated, by calling new:
    if necessary (indicated by the handle being nilH).  If a block
    is already allocated, does nothing
size: ( -- n ) Returns the size of the (whole) string 
setSize: ( n -- ) Sets the size of the (whole) string to n,
    then does a reset:
clear: ( -- ) Ensures a heap block is allocated, calling new:
    if necessary, then sets its size to zero
get: ( -- addr len ) Returns the address and length of the
    active part of the string
all: ( -- addr len ) Returns the address and length of the
    entire string (not just the active part)
1st: ( -- c ) Returns the character at Pos 
^1st: ( -- addr ) Returns the address of the character at Pos 
uc: ( -- addr len ) Converts the active part to upper case and
    does a get:  ( see also >uc: )
put: ( addr len -- ) Ensures a heap block is allocated, calling
    new: if necessary, then replaces it with passed-in string, and
    does reset: as well
->: ( str -- ) Replaces the whole of this string (as in put:)
    with the active part of str, which must be a String
insert: ( addr len -- ) Ensures a heap block is allocated,
    calling new: if necessary, then inserts the string given by
    (addr len) at Pos.  Increments both Pos and Lim by len (thus
    the bytes at the Pos and Lim position will be the same as
    before, and the byte immediately preceding the Pos position
    will be the last of the inserted bytes)
$insert: ( str -- ) Inserts the active part of str, as for
    insert:
add: ( addr len ) Inserts (addr len) at the end of this string. 
    Pos and Lim are then set to the (updated) end position
$add: ( str -- ) Inserts the active part of str at the end of
    this string
+: ( c -- ) Appends the character c to the end of the string,
    and sets Pos and Lim to the (updated) end position
fill: ( c -- ) Overwrites each character in the active part of
    the string with the character c
search: ( addr len -- b ) Searches the active part of this
    string, starting from the left (i.e.  the Pos position), for
    the string (addr len).  If a match is found, Lim is set to
    indicate the first of the matching characters and true is
    returned.  If no match is found, Lim is unchanged and false is
    returned.  Note 1: Case control is handled via the value case?.
      Note 2: We use Lim rather than Pos, since
    it often happens after a search that some operation needs to be
    done on the part of the string preceding the matching
    substring.  If this isn't needed, step: is convenient for
    updating Pos to the matching substring position and preparing
    for another search
chsearch: ( c -- b ) Searches the active part of this string
    for the character c.  If it is found, Lim is set there and true
    is returned.  If it isn't found, Lim is unchanged and false is
    returned

OTHER

print:  Prints the active part of the string.
dump:   Prints a useful dump of the string.  Handy for debugging.    
    
[THEN]


have pad2 0= [if] create pad2 100 allot [then]

: NOPEN ." (not open)"  ;

:class  STRING  <super ptr <general
    var POS
    var LIM

:m release:
        release: super
        clear: pos  clear: lim ;m

:m $err:        \ called when we find something out of bounds.
    cr
    ." size: " get: size .
    ."   pos: "  get: pos .
    ."   lim: "  get: lim . cr
    abort
;m

:m $chk:
    get: pos 0<
    get: lim 0<             or
    get: pos  get: lim  u>  or
    get: lim  get: size u>  or  IF  $err: self  THEN
;m


:m POS:     \ ( -- pos )
    get: pos  ;m

:m >POS:    \ ( newpos -- )
    put: pos  ;m

:m LIM:     \ ( -- lim )
    get: lim  ;m

:m >LIM:    \ ( newlim -- )
     put: lim  ;m

\ Returns the value of Lim - Pos, i.e. the length of the active part
:m LEN:     \ ( -- length )
    get: lim  get: pos -  ;m

\ Adds n and Pos, and stores the result in Lim
:m >LEN:    \ ( newlength -- )
    get: pos  +  put: lim  ;m

:m SKIP:    \ ( n -- )  Increments POS by n.
    +: pos  ;m

:m MORE:    \ ( n -- )  Increments LIM by n.
    +: lim  ;m

\ Clears Pos, so that the active part now starts at the real start
\ of the string.
   \ Sets POS to 0 (the start of the string).

VFX [if]  \ was start:    VFX already uses start:
:m vstart: clear: pos ;m
[else]
:m start: clear: pos ;m
[then]

:m BEGIN:   \ Sets POS and LIM to 0, ready to begin some operation.
    clear: pos  clear: lim  ;m

\ Sets both Pos and Lim to the size (i.e. the end) of the string.
\ Useful for setting up for an iterative operation which has to go
\ backwards through the string
:m END:     \ Sets POS and LIM to the end of the string.
    get: size  dup  put: pos  put: lim  ;m

:m NOLIM:   \ Sets LIM to the end of the string.
    get: size put: lim  ;m

\ Clears Pos, and sets Lim to the end of the string.
\ The active part will now be the whole string
:m RESET:   \ Sets POS to 0, and LIM to the end.
    clear: pos  get: size  put: lim  ;m

:m STEP:    \ Steps down the string, by setting POS to LIM and
            \ then setting LIM to the end.
    get: lim put: pos get: size put: lim  ;m

:m <STEP:   \ Backward step.  Sets LIM to POS, then POS to 0.
    get: pos put: lim clear: pos  ;m


:m NEW:
    0 new: super \ allocate a ptr of (initially) zero size
    clear: size  clear: pos  clear: lim   ;m
    
:m ?NEW:
    Nil?: super IF new: self THEN ;m

\ Returns the size of the (whole) string
:m SIZE:    \ ( -- size )
    get: size  ;m

:m SETSIZE: \ ( newsize -- )
    ?new: self
    dup  resize: super  put: size  reset: self  ;m

:m CLEAR:
    ?new: self  0 resize: super  ;m

:m GET:     \ ( -- addr len ).  Gets the active part of the string.
   \ $chk: self
    get: super  get: pos  +  get: lim  get: pos  -  ;m

:m ALL:     \ ( -- addr len )   Gets all the string, ignoring POS and LIM.
    get: super  size: self  ;m

:m 1ST:     \ ( -- c )  Returns the char at POS.
    ^base @ get: pos + c@   ;m

:m ^1ST:    \ ( -- addr )  Returns the addr of the char at POS.
     ^base @ get: pos +   ;m

\ Converts the active part to upper case and does a get:
:m UC:      \ ( -- addr len )
    get: self  2dup  to-upper  ;m

:m >UC:     \ ( -- )    Converts active part of string to upper case
    get: self  to-upper  ;m

:m PUT: ( addr len -- )
        \ Replaces entire string with replacement string.  Does NEW:
        \ if not already done.
    ?new: self  clear: pos
     dup resize: super
     dup put: lim
     get: super swap move ;m

:m put$: ( str255 -- )
    count put: self ;m


\ Replaces the whole of this string (as in put:) with the
\ active part of the passed in string object.
:m ->:  ( strObj  -- )
    get:  **   put: self ;m

:m dump: 
    cr
    ." size: " get: size .
    ."   pos: "  get: pos .
    ."   lim: "  get: lim .
    cr  all: self type
    cr get: pos spaces [char] P emit
    cr get: lim spaces [char] L emit
    ;m


\ Ensures memory is allocated, calling new: if necessary, then inserts
\ the string given by (addr len) at Pos. Increments both Pos and Lim by len
\ (thus the bytes at the Pos and Lim position will be the same as before, and
\ the byte immediately preceding the Pos position will be the last of the
\ inserted bytes)
:m INSERT:  ( addr len -- )
    >r  \ stash len
    ?new: self
    size: self r@ + resize: super
    get: self drop ( src ) dup r@ + ( dest ) lim: self pos: self - ( len) move
    ( src ) get: super pos: self + ( dest ) r@ ( len ) move
    r@ +: pos
    r> +: lim  ;m


\ Inserts the char c at Pos. Pos and Lim are incremented by 1
:m  CHINSERT:   \ ( c -- )
    pad2 c!  pad2 1 insert: self   ;m

\ Inserts the active text from the given string object
:m $INSERT:  ( strObj  -- )
     get: **  insert: self ;m

\ Inserts ( addr len ) at the end of this string. Pos and Lim are then set to
\ the (updated) end position
:m ADD: ( addr len -- )
    end: self
    ( addr len )  insert: self  ;m

:m ADD$: ( str255 -- )
  count add: self ;m

\ does an ADD: with the active part of the passed in string object
:m $ADD:  ( strObj  -- )
    get: **  add: self ;m


:m +: \ ( char -- )  Appends a char to end of string
    pad2 c!  pad2 1 add: self  ;m


:m  OVWR:  ( addr len -- )

\ Overwrites the active part of SELF with the string ( addr len ).
\ Copying stops at the end of the active part, or when len characters
\ have been transferred.  POS is incremented by the number of chars
\ transferred.  This operation is faster than normal replacement, as the
\ length of SELF cannot change, so resize: is not called.
    >r  \ stash len
    ( addr )  get: self  r> min  dup >r  move
    r> +: pos   ;m


:m  CHOVWR: \ ( c -- )  Overwrites the first char of the active
        \ part of the string ( if any ) by the char c.
    get: self  IF  c!  1 skip: self  else  2drop  THEN   ;m


:m  $OVWR:  \ ( strObj -- )
    get: **  ovwr: self    ;m

:m PRINT:
    nil?: self
    IF  Nopen  ELSE  get: self  type  THEN  ;m

:m SEARCH:  \ ( addr len -- t | f )
    get: self 2swap (search) \ (search) is a Boyer-Moore routine
    IF ( offset ) pos: self + put: lim true
    ELSE  false
    THEN
    ;m

:m search$: ( str255 -- t | f )
    count search: self ;m


:m CHSEARCH:    \ ( c -- b )
    pad2 c!  pad2 1  search: self  ;m


\ do a backwards search for char starting end of
\ active part of string ( lim )
\ if found, then pos points to hit

VFX [if]

:m <CHSEARCH:    { c | strt ^1st addr -- b }
    $chk: self
    get: super dup to strt  get: pos +  to ^1st
    strt  get: lim +  to addr
    BEGIN
        addr 1 - to addr
        addr ^1st u<    IF  false  EXIT  THEN
        addr c@ c =     IF  addr strt -  put: pos  true  EXIT  THEN
    AGAIN
;m

[else]

:m <CHSEARCH:   \ { c | strt ^1st addr -- b }
    0 0 0 locals| strt ^1st addr c |
    to c
    $chk: self
    get: super dup to strt  get: pos +  to ^1st
    strt  get: lim +  to addr
    BEGIN
        addr 1 - to addr
        addr ^1st u<    IF  false  EXIT  THEN
        addr c@ c =     IF  addr strt -  put: pos  true  EXIT  THEN
    AGAIN
;m

[then]

:m LAST: \ ( -- char )  returns last char of active part of string
    get: self + 1- c@ ;m


VFX [if]

:m REPL:   { addr len | newsz sz dlen -- }

 len len: self  - dup to dlen size: super dup to sz + to newsz

 \ first make string bigger if necessary
 dlen 0> IF newsz resize: super THEN

 \ now shift the appropriate chars right or left
 get: self + ( src ) dup dlen + ( dest )
 sz lim: self - ( len ) cmove

 \ wait until here if making string smaller
 dlen 0< IF newsz resize: super THEN

 \ now overwrite the passed in addr len into the string object
 addr ( src ) get: self drop ( dest ) len cmove
 dlen +: lim
 ;m

[else]

:m REPL:  \ { addr len -- }
   0 0 0 locals| newsz sz dlen addr len |
 len len: self  - dup to dlen size: super dup to sz + to newsz

 \ first make string bigger if necessary
 dlen 0> IF newsz resize: super THEN

 \ now shift the appropriate chars right or left
 get: self + ( src ) dup dlen + ( dest )
 sz lim: self - ( len ) cmove

 \ wait until here if making string smaller
 dlen 0< IF newsz resize: super THEN

 \ now overwrite the passed in addr len into the string object
 addr ( src ) get: self drop ( dest ) len cmove
 dlen +: lim
 ;m

[then]

:m $REPL:  \ { strObj  -- }
    get: string  repl: self ;m

:m FILL:    \ ( c -- )
    get: self  rot  fill  ;m


:m DELETE:  \ Deletes the active part of the string.
        \ LIM is then set equal to POS.
    pad2 0  repl: self  ;m

\ From POS, deletes n characters or up to LIM,
\ whichever comes first.  LIM is reduced by the number
\ of characters deleted.

VFX [if]

:m DELETEN:  ( n -- )
    lim: self  0 { lim' n' }
    ( n ) lim' pos: self - min dup to n' \ n'= the actual number to delete
    pos: self + >lim: self
    delete: self \ delete just the active part
    lim' n' - >lim: self \ adjust lim
;m


[else]

:m DELETEN:  ( n -- )
    lim: self  0 locals| n' lim' |
    ( n ) lim' pos: self - min dup to n' \ n'= the actual number to delete
    pos: self + >lim: self
    delete: self \ delete just the active part
    lim' n' - >lim: self \ adjust lim
;m

[then]

;class


