((
$Date: 2008-02-11 00:27:54 +0100 (lun, 11 feb 2008) $
$Revision: 450 $
$Author: astrorafael $
$HeadURL$

2008 By Rafael Gonzalez Fuentetaja
This file is placed in the public domain. NO WARRANTY.
))

\ =======================================================================

\ *> examples

\ *S Simple TCP server example.
\ *P This example shows a multitasked server using the VFX preemptive
\ ** multitasker. A master server thread is created to listen to incoming
\ ** connections and spawns dynamically created slave threads.
\ ** Slave threads get the text from a client and print it in the Forth
\ ** console.
\
\ *P I/O through sockets is done using a GENIO *\fo{/SockDev} driver,
\ ** so words like *\fo{cr}, *\fo{type}, or *\fo{accept} are available.
\ ** Note that this example does not use event passing nor messages, so
\ ** there is no need to call *\fo{pause} and blocking sockets are used.
\ ** When remote clients close connections, slave tasks print exceptions
\ ** and die.
\ **
\ ** Upon starting the server, the Forth console is still available and
\ ** active tasks are shown using multitasker's *\fo{.tasks} word.
\
\ =======================================================================

\ *[

\ =========================================================================

include /usr/share/doc/VfxForth/Lib/Lin32/MultiLin32.fth
include ../src/sockets.fs
include ../src/Genio/sockets.fs


SOCKDEV_SVR Constant myflags

-1 Value master				\ master listening socket fd

7624 EndPoint: local any-ip
EndPoint: remote unknown-ip

256 Constant BufSize
BufSize +User RxBuffer			\ slave tasks reception buffers

\ ------------------------------------------------------------

variable  sock-sid			\ shared between tasks
semaphore sock-sem			\ to syncronize access

: sock-sid@	     \ -- sid
   sock-sem request sock-sid @ sock-sem signal ;

: sock-sid!	     \ sid --
   sock-sem request sock-sid ! sock-sem signal ;

\ ------------------------------------------------------------
\ COMMON WORDS
\ ------------------------------------------------------------

: default-io          \ The default Forth console I/O
   xconsole SetIO ;

: suicide             \ The way a task terminates itself cleanly
   pause termThread 0 pthread_exit ;

: is-action \ xt -- ; wrapper around real action code given by xt
   catch
   default-io dup errno-base <= if .errno-excp else drop then suicide ;

\ ------------------------------------------------------------
\ SLAVE TASKS
\ ------------------------------------------------------------

: new-task  \  -- tcb ; creates an unnamed, dynamically allocated task
   /TCB protAlloc dup initTCB ;


: .selfmsg	       \ #nread --
   dup if ." task " self . RxBuffer swap type cr else ." Read 0!" cr then ;


: (slave-task)         \ slave task action
   sock-sid@ setIO
   begin
      RxBuffer BufSize accept
      [io  default-io .selfmsg  io]
   again
;

: slave-task          \ slave task wrapper
   ['] (slave-task) is-action ;

\ ------------------------------------------------------------
\ MASTER TASK
\ ------------------------------------------------------------

: new-master		\ creates the master socket, in blocking mode
   TCP socket throw to master
   local master binds throw
   master listens throw
;

: new-sockdev	       \ -- sid ; dinamically allocated /sockdev
   /SockDev protAlloc dup initSockDev ;

: .connect
   cr ."  A client has connected " cr ;

: (server)	\ Main server action
   new-master
   begin
      new-sockdev dup >r sock-sid!
      remote master myflags r> open-gio throw
      .connect
      ['] slave-task new-task initiate
   again
;

: server
   ['] (server) is-action ;

: myshutdown
   sock-sem ShutSem ;

\ Main task creation and startup. Leaves a fully operating Forth console

sock-sem InitSem
task server-task
' server server-task initiate
' myshutdown AtExit
\ *]

\ ======
\ *> ###
\ ======
