((
$Date: 2008-02-11 00:27:54 +0100 (lun, 11 feb 2008) $
$Revision: 450 $
$Author: astrorafael $
$HeadURL$

2008 By Rafael Gonzalez Fuentetaja
This file is placed in the public domain. NO WARRANTY.

To do
=====

More comprenhensive accept entry point ?
Do we need ioctls tochange behaviour at run time ?
))

\ ######################
\ *! genio
\ *T GENIO Socket Device
\ ######################

\ *P This Generic IO Device operates on a TCP socket for input and output.
\ ** Flags specify whether to create a client or server socket and also
\ ** the blocking/non-blocking mode.

\ *************************
\ *S Socket Device Creation
\ *************************

\ *P In order to create a named socket device on the dictionary,
\ ** use the *\fo{SOCKDEV:} definition given later. These devices can be
\ ** created in the heap using *\fo{allocate}, and must be initialized using
\ ** *\fo{initSockDev}.

\ *P Under GENIO, sockets are opened by using either
\ ** *\fo{open-gen} or *\fo{open-gio}.
\ ** The following flags control  options at open time:
\ *(
\ *B *\fo{SOCKDEV_NONBLOCK}, to open the socket in non-blocking mode.
\ *B *\fo{SOCKDEV_SVR}, to open the socket in server mode using *\fo{accepts}.
\ ** Needs to pass the parent socket *\i{fd}.
\ *B *\fo{SOCKDEV_ECHO} to echo characters to remote peer socket when using
\ ** ANS *\fo{ACCEPT} word on this socket.
\ *)

\ *P When using this driver to program server sockets, it is up to the
\ ** developer to set up and configure the parent socket with
\ ** *\fo{socket}, *\fo{binds} and *\fo{listens} or *\fo{-blocking}.
\ ** An example:
\ *E
\ **   -1 Value master
\ **   7624 EndPoint: local any-ip
\ **   
\ ** : new-master   ( -- )	     
\ **    TCP socket throw to master
\ **    local master binds throw
\ **    master listens throw ;

   
\ *************
\ *S Exceptions
\ *************

\ *P When examining the stack signature of GENIO drivers, there were
\ ** no place to report I/O errors. Two options were available: either
\ ** ignore them or to throw exceptions. I have chosen the latter strategy
\ ** as a way to detect bugs and unexpected situations.

\ *P The following *\b{xxx-gio} words may throw exceptions when using the
\ ** underlying *\fo{sockets} vocabulary: *\br{}
\ ** *\fo{open-gio}, *\fo{close-gio}, *\fo{read-gio}, *\fo{readex-gio},
\ ** *\fo{write-gio}, *\fo{key?-gio}, *\fo{key-gio}, *\fo{accept-gio},
\ ** *\fo{emit-gio}, *\fo{type-gio}, *\fo{cr-gio}, *\fo{lf-gio},
\ ** *\fo{ff-gio} and *\fo{bs-gio}.

\ *P The following *\b{xxx-gio} words throw
\ ** *\fo{abort" operation not supported"} exceptions: *\br{}
\ ** *\fo{ekey-gio}, *\fo{ekey?-gio}, *\fo{bell-gio}, *\fo{setpos-gio},
\ ** *\fo{getpos-gio} and *\fo{ioctl-gio},

\ *P The following *\b{xxx-gio} words do nothing: *\br{}
\ ** *\fo{flushOP-gio}, *\fo{init-gio}, *\fo{term-gio} and *\fo{config-gio}.


\ ***********
\ *S Glossary
\ ***********


MODULE GENIO-SOCK
\ +G GENIO Module name for socket I/O.


struct /SockDev	\ -- len ; 
\ *G The socket SID structure, passed around as a handle by all *\b{xxx-gio}
\ ** words.
\ *[  
  gen-sid +				\ reuse field names of GEN-SID
   1 cells field sd.flags		\ Mode flags (see below).
end-struct
\ *]

1 Constant SOCKDEV_NONBLOCK
\ *G Non-Blocking bit for  *\i{fam} flag in  *\fo{open-gio}.
\ ** Also stored in *\fo{sd.flags}.

2 Constant SOCKDEV_SVR
\ *G Server bit for  *\i{fam} flag in  *\fo{open-gio}.
\ ** Also stored in *\fo{sd.flags}.

4 Constant SOCKDEV_ECHO
\ *G Echo bit in *\i{fam} flag in  *\fo{open-gio}.
\ ** Makes *\fo{ACCEPT} echo character back to the sender socket.
\ ** Option available for server sockets only.
\ ** Also stored in *\fo{sd.flags}.

\ ===========================================
\ +N Blocking/Non-Blocking socket API vectors
\ ===========================================

\ +P Handling blocking or non-blocking socket connections is deferred
\ +* until open time. It makes much easier life to code several entry points
\ +* with a slight overhead.

Defer socket-connect			\ addr fd -- ior
\ +G Vector for *\fo{connects} or  *\fo{connects-mt}.

Defer socket-accept			\ addr1 fd1 -- fd2 ior
\ +G Vector for *\fo{accepts} or  *\fo{acceptss-mt}.

Defer socket-send			\ c-addr1 +n1 fd -- +n2 ior
\ +G Vector for *\fo{sends} or  *\fo{sends-mt}.

Defer socket-recv			\ c-addr1 +n1 fd -- +n2 ior
\ +G Vector for *\fo{recvs} or  *\fo{recvs-mt}.

Defer socket-recv-all			\ c-addr1 +n1 fd -- +n2 ior
\ +G Vector for *\fo{recvs-all} or  *\fo{recvs-all-mt}.

Defer sd-key?				\ sid -- flag
\ +G Vector for *\fo{(sd-key?)} or *\fo{(sd-key?-mt)}.

\ ============================
\ +N Common, low level factors
\ ============================

: (sd-key?)          \ sid -- flag
\ +G Test for any character received by configuirng the socket as
\ +* non-blocking mode, do a peek and check for EAGAIN errno. Then, the socket
\ +* is configured as blocking again.   
   gen-handle @
   dup -blocking throw   dup peeks? throw   swap +blocking throw ;

: (sd-key?-mt)         \ sid -- flag
\ +G More efficient operation when socket is already open in non-blocking mode.
   gen-handle @ peeks? throw ;


: NotSockDev	\ --
\ +G Issue a SockDev error message.
  1 abort" operation not supported for SOCKET DEVICE"
;

: (sd-create) \ addr len sid -- addr fd sid 0 | sid ior
\ +G Creates the socket.
\ +* *\i{len} is unusued but *\i{addr} will be used later,   
   >r drop TCP socket dup if -rot 2drop then r> swap ;

: (sd-connect)	 \ addr fd sid -- fd sid 0 | sid ior
\ +G Connects or closes the socket if not succesfull.
\ +* *\i{addr} is a *\fo{/sockaddr_in},  *\fo{/enpoint} structure
\ +* or *\f{NULL}.   
   >r tuck socket-connect dup if swap closes drop then r> swap ;

: (sd-accepts)	 \ addr len sid  -- fd sid 0 | sid ior
\ +G Creates a new child socket from parent socket.
\ +* *\i{addr} is an *\fo{/enpoint} structure or *\f{NULL}.
\ +* *\i{len} is the parent spcket descriptor
   >r socket-accept dup if nip then r> swap ;

: (sd-gen!)			  \ fd sid -- sid 0
\ +G Sets the internal handle to be the file descriptor returned by socket
\ +* creation words.
   tuck gen-handle ! 0 ;

: open-client	\ addr len sid -- sid ior
\ +G Open socket device and connect to remote *\fo{/sockaddr_in} or
\ +* *\fo{/endpoint} *\i{addr} for client sockets.
\ +* *\i{len} is unusued. Leave it to -1.
   (sd-create)  ?dup if exit then	\ addr fd sid
   (sd-connect) ?dup if exit then	\ fd sid | sid ior
   (sd-gen!) ;

: open-server	\ addr len sid -- sid ior
\ +G Open socket device for children server sockets.
\ +* *\i{addr} is an *\fo{/enpoint} structure or *\f{NULL}.
\ +* *\i{len} is the parent socket descriptor returned
\ +* beforehand by *\fo{socket}.
   (sd-accepts)  ?dup if exit then	\ fd sid | sid ior
   (sd-gen!) ;
   
\ =============================
\ +N Socket driver Entry points
\ =============================

: sd-close         \ sid -- ior
\ +G Closes the socket and initializes the gen-handle back to invalid state.
   dup gen-handle @ >r			 
   INVALID_HANDLE_VALUE over gen-handle !
   sd.flags off
   r> closes
;

: sd-read          \ addr len sid -- ior
\ +G Read *\b{all} the bytes up to *\i{len} from the socket device.   
  gen-handle @ socket-recv-all nip
;

: sd-readex           \ addr len sid -- #read ior
\ +G Read *\b{all} the bytes up to *\i{len} from the socket device.
\ +* May return less bytes than requested.
    gen-handle @ socket-recv
;

: sd-write         \ addr len sid -- ior
\ +G 
  gen-handle @ socket-send nip 
;

: sd-key	\ sid -- char
\ +G    
  { | temp[ cell ] -- }
  temp[ 1 rot sd-read throw
  temp[ c@
;

: sd-ekey          \ sid -- echar
\ +G Not supported.    
  NotSockDev	\ 1 abort" EKEY not supported for SOCKDEV"
;

: sd-ekey?         \ sid -- flag
\ +G Not supported.    
  NotSockDev	\ 1 abort" EKEY? not supported for SOCKDEV"
;

: sd-emit          \ char sid --
\ +G    
  { | temp[ cell ] -- }
   swap temp[ c!		\ -- sid
   temp[ 1 rot sd-write throw
   1 out +!
;

: sd-emit?         \ sid -- flag
\ +G  ALways true. does not check for full Tx buffer.
  drop TRUE
;

: sd-type          \ addr len sid --
\ +G    
  sd-write throw
;

: sd-cr            \ sid --
\ +G  Send CR+LF characters   
   ACR over sd-emit
   ALF swap sd-emit			\ Should emit LF in Linux ?
   out off  op-line# incr
;

: sd-lf            \ sid --
\ +G Linefeed.   
   ALF swap sd-emit
   -1 out +!	
;

: sd-ff            \ sid -- ; page/cls on display devices
\ +G   
   #12 swap sd-emit
;

: sd-bs            \ sid -- ; destructive on display devices
\ +G   
   #8 swap sd-emit
;

: sd-bell          \ sid -- ; audible beeper
\ +G Not supported.             
   NotSockDev	
;

: sd-setpos        \ x y mode sid -- ior
\ +G Not supported.             
   NotSockDev	
;

: sd-getpos        \ mode sid -- x y ior
\ +G Not supported.             
   NotSockDev	
;

: sd-ioctl         \ addr len fn sid -- ior
\ +G Not supported. All is done at open time.   
   NotSockDev          
;

: sd-flushOP       \ sid -- ior
\ +G Not supported.          
   drop 0
;

: sd-init          \ addr len sid -- ior
\ +G Not supported.       
   3drop 0	
;

: sd-term          \ sid -- ior
\ +G Not supported.       
   drop 0	
;

: sd-config        \ sid -- ior ; produces a dialog
\ +G Not supported.       
   drop 0	
;

\ ---------------------------
\ +H Support for *\fo{ACCEPT}
\ ---------------------------

\ +P For the sd-accept entry points, we have deconstructed its loop
\ +* into pieces.

: wait-key		     \ sid -- sid
\ +G Wait until next character arrives. 
   begin dup sd-key? 0= while delayed repeat
;


: /loop				 \ addr1 len sid - cnt sid addr2 addr3
\ +G Initializes accept loop, converting (start, count) pair into
\ +* (starting , ending)  addresses *\i{addr2,addr3}.
   0 swap 2swap bounds
;

: cr?					  \ c -- c flag
\ +G Test for carriage return.
\ +* Returns true id *\f{CR} is detected.
   dup ACR =
;

: ?echo  \ cnt sid c -- cnt sid c
\ +G Conditionally echoes back the character, depending
\ +* on the *\fo{SOCKDEV_ECHO} flag.
   over sd.flags @ SOCKDEV_ECHO and 0= if exit then
   2dup swap sd-emit
;

: ?backspace   \ cnt1 sid c -- cnt2 sid index
\ +G Handle possible backspace character *\i{c} by going back
\ +* into the array and decreasing by 1 the count and loop index.
\ +* Normal operation increases by 1 the count and loop index.
   BSIN = if swap 1- swap -1 else swap 1+ swap  1 then
;

: discard-lf			   \ cnt sid c -- cnt sid
\ +G *\i{c} is a *\i{CR}, therefore, read *\i{LF} and discard it.
   drop dup sd-key drop ;

: sd-accept			   \ addr len sid -- #read
\ +G Accept chars from socket until *\i{cr} is detected.
\ +* Returns number of characters read. 
   /loop ?do				\ cnt1 sid
      wait-key dup sd-key		\ cnt1 sid c
      cr? if discard-lf leave endif	\ cnt1 sid c
      dup i c!				\ cnt1 sid c
      ?echo ?backspace			\ cnt2 sid index
   +loop
   drop
;

\ ========================
\ +N Open and vector table 
\ ========================

: +block-all
\ +G Set vectors for blocking mode.
   ['] connects    is socket-connect
   ['] accepts     is socket-accept
   ['] recvs       is socket-recv
   ['] recvs-all   is socket-recv-all
   ['] sends       is socket-send
   ['] (sd-key?)   is sd-key?

;

: -block-all
\ +G Set vectors for blocking mode.   
   ['] connects-mt    is socket-connect
   ['] accepts-mt     is socket-accept   
   ['] recvs-mt       is socket-recv
   ['] recvs-all-mt   is socket-recv-all
   ['] sends-mt       is socket-send
   ['] (sd-key?-mt)   is sd-key?   
;

: sd-open	\ addr len fam sid -- sid ior
\ +G *\i{addr} is an *\fo{/endpoint} structure.
\ +* *\i{fam} containst the two mode bits defined above.
\ +* *\i{len} is -1 for client socke or else the parent *\i{fd}.
\ +* when opening a child server socket.
\ +* *\i{sid} is the *\i{/SockDev} structure.
   2dup sd.flags !
   over  SOCKDEV_NONBLOCK and if -block-all  else +block-all  then
   swap  SOCKDEV_SVR      and if open-server else open-client then
;


create sd-vectors			\ -- ; Exec: -- addr
\ +G Table of GENIO execution tokens.
  ' sd-open ,
  ' sd-close ,
  ' sd-read ,
  ' sd-write ,
  ' sd-Key ,
  ' sd-Key? ,
  ' sd-Ekey ,
  ' sd-EKey? ,
  ' sd-Accept ,
  ' sd-Emit ,
  ' sd-Emit? ,
  ' sd-Type ,
  ' sd-CR ,
  ' sd-LF ,
  ' sd-FF ,
  ' sd-BS ,
  ' sd-Bell ,
  ' sd-setpos ,
  ' sd-Getpos ,
  ' sd-Ioctl ,
  ' sd-FlushOP  ,
  ' sd-readex ,
  ' sd-Init ,
  ' sd-Term ,
  ' sd-Config ,

\ ==================
\ *N Device Creation
\ ==================

: initSockDev	\ sid --
\ *G Initialise the sid for a socket device. Use it to initialize
\ ** the structure has been allocated from the heap.
  dup /sockdev erase
  sd-vectors INVALID_HANDLE_VALUE rot 2!
;

: SockDev:	\ "name" -- ; Exec: -- sid
\ *G Create a Socket based Generic IO device in the dictionary.
  create
    here  /sockdev allot  initSockDev
;

Export /sockdev
Export initSockDev
Export SockDev:
Export SOCKDEV_SVR
Export SOCKDEV_NONBLOCK
Export SOCKDEV_ECHO

END-MODULE


\ ======
\ *> ###
\ ======

