((
$Date: 2008-02-11 00:27:54 +0100 (lun, 11 feb 2008) $
$Revision: 450 $
$Author: astrorafael $
$HeadURL$

2008 By Rafael Gonzalez Fuentetaja
This file is placed in the public domain. NO WARRANTY.

To do
=====

Thorough test examples for non-blocking looping words with a large image file

))


\ ####################################
\ *> sockets
\ *T Linux TCP and UDP Socket Bindings
\ ####################################

\ *P Defines a socket wordset to be used
\ ** in a variety of situations. The main I/O words have been defined
\ ** with a similar stack effect as the *\i{file} wordset. The *\i{ior}
\ ** codes are throwable. 
\
\ *P Choosing names for this BDS style socket API has been
\ ** problematic because it uses common words such as *\i{bind} or
\ ** *\i{close} that could clash with other applications or wordlists.
\ ** I have opted to append an *\f{s} (for *\i{socket}) to most
\ ** of these words to keep them readable, concise and
\ ** avoid name clashes. So *\i{close} becomes *\i{closes} and
\ ** *\i{bind} becomes *\i{binds}.
\
\ *P Regarding portability, There has been no special effort to produce a
\ ** portable library, only to
\ ** make it work. It would probably be quite easy to port it to GForth.
\ ** Some asumptions taken:
\
\ *(
\ *B 1 char = 1 byte  =  8 bits.
\ *B 1 cell = 4 bytes = 32 bits.
\ *B There are words to read/write 16 bit values on unaligned addresses.
\ *B IP addresses are IPv4 32 bit addresses, fitting into 1 cell.
\ *B Structure defining words supplied by VFX Forth, whose *\fo{field} produces
\ ** unaligned fields. Apparently, that's what was needed to clone some
\ ** C wire structures.
\ *B There is a facility to access external shared library. I have used
\ ** the ones supplied bf VFX Forth.
\ *B Used VFX extended locals notation in some cases to interface
\ ** C system calls.
\ *B There are available some convenient but non ANS standard words like
\ ** *\fo{3drop 4drop 3dup 4dup}.
\ *B There is a word to print z-strings (actually, it is *\fo{.z$}).
\ *)

\ ****************************************
\ *S Non-blocking sockets and multitasking
\ ****************************************
\
\ *P A non-blocking I/O wordset has been included to use together with
\ ** cooperative multitaskers. Since *\i{VFX Forth for Linux} includes
\ ** a preemptive multitasker, blocking I/O can be used within each task and
\ ** the non-blocking counterparts have, in principle, little interest.
\ ** However, facilities like messages between tasks or task events are driven
\ ** using the traditional *\fo{pause} word. If you plan to use such
\ ** facilities, then you must use the non-blocking words to ensure that
\ ** *\fo{pause} is called frequently enough.
\
\ =======================================================================

\ ***********
\ *S Glossary
\ ***********

[undefined] libc.so.6 [if]
   Library: libc.so.6
[endif]

\ ===================================
\ *N Handling Linux *\fo{errno} codes
\ ===================================

\ *P Socket I/O uses a variety of system calls that may fail for a number
\ ** of reasons. The following words help keep track of *\i{errno} codes
\ ** thrown as exceptions.

\ *P As far as I know, VFX do not have yet embedded all the *\i{errno} codes
\ ** as part of its error handling wordset, nor do not offer an easy way to
\ ** automatically read the z-strings returned by *\i{strerror()}
\ ** and define them with *\fo{#AnonErr}.
\ ** All of VFX's error defining words are parsing words.
\ *P *\b{Note:} This section is likely to change as soon as possible.


-5000 Value errno-base
\ *G Base value for OS *\i{errno} throw codes.
\ ** Can be customized at runtime to coexist with other throw codes.

: errno>excp		     \ errno -- n1
\ *G Converts Linux *\i{errno} values in throw codes.
   negate errno-base + ;

: excp>errno		    \ n1 -- errno
\ *G Converts throw code back to genuine Linux *\i{errno} values.
   errno-base swap - ;

AliasedExtern: errno int * __errno_location(void);
\ *G *\fo{errno} is the well known *\i{errno} C global variable
\ ** used by libraries and system calls.
\ ** Actually, it is a thread local (aka *\fo{USER}) variable.
\ ** Can be read by *\fo{@} and written by *\fo{!}
\ This word will dissapear as soon as MPE include the proper verison of 'errno'.

: throwable		     \ flag -- -ior
\ *G Takes a boolean *\i{flag} and  makes it a throwable *\i{ior} with the
\ **\i{errno} code embedded.
\ ** *\fo{False} makes a zero *\i{ior}.   
   ( errno @ and dup if errno>excp then ;) \ this fails in prempt multithr.
   if errno @ errno>excp else 0 then ;   
   
Extern: char * strerror(int errnum);	\ -- z-addr
\ *G Get the human readable *\i{errno} string.
\ ** See *\i{man strerror} for details.

: .errno	      \ n1 --
\ *G Given the Linux *\i{n1 errno} number prints the associated z-string.
\ ** A factor for the word below.   
   strerror .z$ ;

: .errno-excp		\ n1 --
\ *G Prints the errno exception value thrown in *\i{n1}.
\ ** *\i{n1} must be < *\fo{errno-base}.
   excp>errno cr ." Exception ( errno = " dup . ." ): " .errno cr  ;

\ ========================
\ *N Handling IP addresses
\ ========================

\ *P The following words help using and declaring IP addresses.

\ ----------------
\ *H Reusable bits 
\ ----------------

struct /hostent				\ -- len
\ *G Mimics the Linux *\i{struct hostent}.
\ ** See *\i{man gethostbyname}.
   1 cells field h.name
   1 cells field h.aliases
   1 cells field h.addrtype
   1 cells field h.length
   1 cells field h.addr_list
end-struct

struct /sockaddr_in			\ -- len
\ *G Mimics the Linux *\i{struct sockaddr_in}.
\ ** See *\i{man ip}.
   2 chars field sin.family
   2 chars field sin.port
   1 cells field sin.addr
   8 chars field sin.padding
end-struct

2  Constant PF_INET
\ *G IP sockets family.

0  Constant INADDR_ANY
\ *G Used to listen to any IP address of a host.

-1 Constant INADDR_BROADCAST
\ *G The special broadcast address 255.255.255.255 encoded as binary.

\ *P The following constants are possible values left in *\i{h_errno}
\ ** C global variable. Comments are extracted form the *\f{<netdb.h>}
\ ** header file.

-1 Constant NETDB_INTERNAL
\ *G Internal error. See *\i{errno}.

0 Constant NETDB_SUCCESS
\ *G No problem. 

1 Constant HOST_NOT_FOUND
\ *G Authoritative Answer Host not found.

2 Constant TRY_AGAIN	
\ *G Non-Authoritative: Host not found, or SERVERFAIL.

3 Constant NO_RECOVERY
\ *G Non recoverable errors, FORMERR, REFUSED, NOTIMP.

4 Constant NO_DATA
\ *G Valid name, no data record of requested type.

4 Constant NO_ADDRESS
\ *G No address, look for MX record.
\ ** Currently, the same value as *\fo{NO_DATA}. 

AliasedExtern: h_errno int * __h_errno_location (void);
\ *G *\fo{h_errno} is the *\i{h_errno} C global variable.
\ ** Actually, it is a thread local (aka *\fo{USER}) variable.
\ ** Can be read by *\fo{@} and written by *\fo{!}

AliasedExtern: ux-gethostbyname void *  gethostbyname(const char * name);
\ +G See *\i{man gethostbyname}.

Extern: uint32  htonl(uint32 hostlong);
\ *G See *\i{man htonl}.

Extern: uint16  htons(uint16 hosthort);
\ *G See *\i{man htons}.

: c-string			 \ c-addr1 u -- c-addr2 
\ *G Converts a Forth string into an null terminated C string.
\ ** The z-string *\i{c-addr2}is stored in the *\fo{PAD}.
\ +* Taken from GForth.   
   tuck pad swap move pad + 0 swap c! pad ;
 
: gethostbyname				  \ c-addr1 u -- addr2 ior
\ *G Get a *\i{struct hostent} from a symbolic name or IP dotted notation
\ given by *\i{c-addr1 u} string.
\ ** Returns the *\fo{/hostent} structure base address *\i{addr2}
\ ** and an *\i{ior}.
\ ** Negative *\i{ior} values can be thrown as exceptions.
\ ** Positive *\i{ior} values denotes one of
\ ** *\fo{HOST_NOT_FOUND}, *\fo{TRY_AGAIN}, *\fo{NO_RECOVERY},
\ ** *\fo{NO_DATA NO_ADDRESS}
\ ** failed search results errors.
\ ** *\i{addr2} is undefined if *\i{ior} <> 0.
   c-string ux-gethostbyname dup 0<> if NETDB_SUCCESS exit then
   h_errno @ dup 0< if throwable exit then ;

   
: host>addr		    \ c-addr1 u -- u1 
\ *G Converts an Internet host name into a binary *\i{u1} IPv4 address.
\ ** The resulting address is in network byte order.
\ ** This word may throw *\i{errno} exceptions if internal errors are detected.
\ ** It may also call *\fo{abort"} with the texts shown in the constants
\ ** glossary above if the search results are one of the following 
\ ** *\fo{HOST_NOT_FOUND}, *\fo{TRY_AGAIN}, *\fo{NO_RECOVERY},
\ ** *\fo{NO_DATA NO_ADDRESS} values.
   gethostbyname dup 0< if throw then
   dup case
      HOST_NOT_FOUND of
	 abort" Authoritative Answer: Host not found."
      endof
      TRY_AGAIN of
	 abort" Non-Authoritative Answer: Host not found, or SERVERFAIL."
      endof
      NO_RECOVERY    of
	 abort" Non recoverable errors, FORMERR, REFUSED, NOTIMP."
      endof
      NO_DATA        of
	 abort" Valid name, no data record of requested type."
      endof
      NO_ADDRESS    of
	 abort" No address, look for MX record."
      endof
      drop >r h.addr_list @ @ @ r> \ the ok case
   endcase ;


: port!			  \ port addr1 --
\ *G Fills int the *\fo{/sockaddr_in} structure or *\fo{EndPoint:} given by
\ ** *\i{addr1} with the *\fo{PF_INET} protocol family and *\i{port}.
   swap htons over sin.port   w!
   PF_INET    swap sin.family w! ;

struct /endpoint			\ -- addr
\ *G Extends the *\fo{/sockaddr_in} structure with a *\fo{ep.rlen}
\ ** field for convenience. This field is a convenient placeholder
\ ** for storing returned length of a *\fo{/sockaddr_in} structure
\ ** in various Linux system calls.
\ ** Allocation of  *\fo{/endpoint} structures
\ ** can be made with *\fo{allocate} or *\fo{EndPoint:} (recommended).
   /sockaddr_in +
   1 cells field ep.rlen		\ addr1 -- addr2
end-struct
   
\ -------------
\ *H Main words
\ -------------

: EndPoint:			\ "<name>" -- addr1 
\ *G Defining word to create an IP end point.
\ ** The *\fo{/sockaddr_in} structure address *\i{addr1} is returned just to be
\ ** initialized by words
\ ** *\fo{known-ip}, *\fo{unknown-ip} *\fo{any-ip} or *\fo{broadcast-ip}.
\ ** *\br{}
\ ** Examples:   
\ *E
\ **  s" localhost" 7624 EndPoint: remote known-ip \ for clients
\ **  7624 EndPoint: local any-ip                  \ for servers 
\ **  EndPoint: receiver unknown-ip                \ to use in recvsfrom   
   Create here /endpoint allot&erase ; 

: known-ip				\ c-addr1 u1 port addr2 --
\ *G An initializer to a *\fo{/sockaddr_in} given by *\i{addr2} that declares
\ ** a well known *\i{port} plus *\i{c-addr1 u1} IP address, typically used
\ ** by clients to connect to remote peers or servers.
\ ** Host string can be either a symbolic name
\ ** or a IP dotted notation string. This word calls the resolver
\ ** *\fo{host>addr} word, so it may throw *\i{errno} exceptions.
   dup >r port!  host>addr r> sin.addr ! ;

: any-ip		   \ port addr1 --
\ *G An initializer to a *\fo{/sockaddr_in} given by *\i{addr1} that declares
\ ** an end point ready to listen to a given *\i{port} but any IP address.
\ ** Typically used by servers.
   dup >r port! INADDR_ANY r> sin.addr ! ;

: broadcast-ip			 \ port addr1 --
\ *G An iniializer to a *\fo{/sockaddr_in} given by *\i{addr1} that declares
\ ** an end point ready to send broadcast messages to a given *\i{port}.
\ ** Can be used by clients or in UDP based application protocols
\ ** to announce and discover.
   dup >r port! INADDR_BROADCAST r> sin.addr ! ;

: unknown-ip		   \ addr1 --
\ *G An initializer to a *\fo{/sockaddr_in} given by *\i{addr1} that declares
\ ** an uninitialized end point to be filled by *\fo{binds},
\ ** *\fo{accepts} or *\fo{recvsfrom} words. Syntactic sugar.
   drop ;


\ =================
\ *N TCP/IP Sockets
\ =================

\ ----------------
\ *H Reusable bits 
\ ----------------

\ *P Description of most words in this section is available by generating the
\ ** additional detailed level of documentation.

AliasedExtern: ux-socket int socket(int domain, int type, int protocol);
\ +G See *\i{man socket}.

AliasedExtern: ux-bind int bind(int fd, void * my_addr, int addrlen);
\ +G See *\i{man bind}.

AliasedExtern: ux-listen int listen(int fd, int backlog);
\ +G See *\i{man listen}.

AliasedExtern: ux-connect int connect(int fd, const int * serv_addr, int addr_len);
\ +G See *\i{man connect}.

AliasedExtern: ux-accept int accept(int fd, void * rem_addr, int * addr_len);
\ +G See *\i{man accept}.

AliasedExtern: ux-close int close(int fd);
\ +G See *\i{man close}.

AliasedExtern: ux-send int send(int s, const void * msg, int len, int flags);
\ +G See *\i{man send}.

AliasedExtern: ux-recv int recv(int s, void * msg, int len, int flags);
\ +G See *\i{man recv}.

Extern: int setsockopt(int fd, int level, int opt, void * val, int vallen);    
\ *G See *\i{man setsockopt}.

1 Constant SOL_SOCKET
\ *G Select socket level in *\i{get/setsockopt()}

2 Constant SO_REUSEADDR
\ *G Socket level option to reuse address.
\ ** Typically used in server sockets.

1 Constant SOCK_STREAM
\ *G Stream type socket.

2 Constant SOCK_DGRAM
\ *G Datagram type socket.

: reuse-IP		 \ fd -- ior 
\ *G Configure a socket *\i{fd} to reuse its IP address through
\ ** *\fo{setsockopt}. Usually done at servers.
\ ** Returns a throwable *\i{ior}.
   { | tmp[ cell ] -- }
   1 tmp[ !
   SOL_SOCKET SO_REUSEADDR tmp[ 1 cells setsockopt 0< throwable ;

\ -------------
\ *H Main words
\ -------------

#6  Constant TCP				
\ *G Used to create a TCP socket with *\fo{socket}.
\ ** Same value as Linux *\f{#define IPPROTO_TCP}.

#17 Constant UDP			
\ *G Used to create an UDP socket with *\fo{socket}.
\ ** Same value as Linux *\f{#define IPPROTO_UDP}.

: socket		  \ n1 -- fd ior
\ *G Creates a TCP or UDP socket given the constant *\fo{TCP} or *\fo{UDP}
\ ** passed in *\i{n1}. 
\ ** Returns a descriptor *\i{fd} and throwable *\i{ior}.
\ ** *\i{fd} is undefined when the *\i{ior} is not zero.
\ ** Socket I/O on the returned *\i{fd} is blocking by default.
\ ** See *\i{man socket}. 
   PF_INET over TCP = IF SOCK_STREAM  ELSE SOCK_DGRAM THEN rot ux-socket
   dup 0< throwable ;

: connects		     \ addr fd -- ior
\ *G Client connects to a remote socket server.
\ ** *\i{addr} is a remote *\fo{/sockaddr_in} or *\fo{/endpoint} structure.
\ ** Returns a throwable *\i{ior}.
   swap /sockaddr_in ux-connect 0< throwable ;

: binds			  \ addr fd -- ior
\ *G Binds the server socket to a port and address (ANY address usually).
\ ** Reuse port+IP address if possible.
\ ** *\i{addr} can be either a *\fo{/sockaddr_in} or
\ ** a *\fo{/endpoint} structure.
\ ** Returns a throwable *\i{ior}.
   swap over reuse-IP ?dup if -rot 2drop exit then
   /sockaddr_in ux-bind 0< throwable ;

#5 Value BackLog
\ *G Backlog of pending connections to *\fo{listens}.

: listens	       \ fd -- ior
\ *G Server socket *\i{fd} listens to incoming connections.
\ ** Returns a throwable *\i{ior}.
\ ** See *\i{man listen}.   
   BackLog ux-listen 0< throwable ;

: accepts			 \ addr1 fd1 -- fd2 ior
\ *G Accept an incoming socket connection. 
\ ** Returns a new socket descriptor *\i{fd2} and a throwable *\i{ior}.
\ ** *\i{addr1} *\b{must} be a defined an *\fo{/endpoint} struct
\ ** to be filled in with the incoming client IP address & port.
\ ** *\i{fd2} is undefined when *\i{ior} is non-zero.
\ ** Socket I/O on the returned *\i{fd2} is blocking by default.   
\ ** See *\i{man accept}.  
   swap dup ep.rlen ux-accept dup 0< throwable ;

: closes	   \ fd -- ior
\ *G Close the socket.
\ ** Returns a throwable *\i{ior}.
\ ** See *\i{man close}.
   ux-close 0< throwable ;

$4000 Constant MSG_NOSIGNAL
\ *G Transmission/reception flag to make *\fo{<recvs>} or *\fo{<sends>} not
\ ** to throw a *\i{SIGPIPE} signal when the other end closes a connection.
\ ** Only applicable to connection oriented sockets.
\ ** See  *\i{man send} or *\i{man recv}.

$02 Constant MSG_PEEK
\ *G Reception flag to make *\fo{<recvs>} or *\fo{<recvsfrom>} peek at
\ ** incoming data without dequeueing from system internal buffers.
\ ** See *\i{man recv}.

$100 Constant MSG_WAITALL
\ *G Reception flag to make *\fo{<recvs>} or *\fo{<recvsfrom>} block until
\ ** all requested bytes have been read.
\ ** See *\i{man recv}.

: <sends>				  \ c-addr1 +n1 fd n2 -- +n3 ior
\ *G Send *\i{n1} bytes starting from *\i{c-addr1} through socket
\ ** given by *\i{fd}.
\ ** Transmission flags must be or-ed in a single *\i{n2}.      
\ ** The actual amount sent is returned in *\i{n3}. In certain conditions,
\ ** *\i{n3} can be < *\i{n1}. 
\ ** Returns a throwable *\i{ior}. *\i{n3} is undefined if ior is non-zero.
\ ** See *\i{man send}.   
   >r -rot r> ux-send  dup 0< throwable ;

: sends				  \ c-addr1 +n1 fd -- +n2 ior
\ *G The common use case for transmission.
\ ** *\f{SIGPIPE} is disabled.
\ ** If the remote end closes connection, *\i{ior} will contain a
\ ** *\f{EPIPE} throwable errno code.
\ ** See *\i{man send}.      
\ ** Defined as :
\ *[    
   MSG_NOSIGNAL <sends> ;
\ *]


: <recvs>  \ c-addr1 +n1 fd n2 -- +n3 ior
\ *G Receive through socket given by *\i{fd} an amount of *\i{n1} bytes
\ ** and copy them into *\i{c-addr1}. 
\ ** Reception flags must be or-ed in a single *\i{n2}.   
\ ** Returns the actual amount received in *\i{n3} and a throwable *\i{ior}. 
\ ** If there is no data, the socket will block until something is available.
\ ** If the socket is non-blocking and there is no data, *\i{ior} will contain
\ ** a *\fo{EAGAIN} throwable exception.
\ ** See *\i{man recv}.   
   >r -rot r> ux-recv dup 0< throwable ;

#32 Constant EPIPE
\ *G *\i{errno} code returned when the other end closes a connection.

: epipe?				  \ #read ior1 -- #read ior2
\ *G Generate an artificial, throwable *\f{EPIPE} errno code
\ ** when the remote end closes the connection and signals are disabled.
\ ** This happens when *\fo{<recvs>} returns *\f{0 0}. 
    2dup 0= swap 0= and if drop EPIPE errno ! True throwable then ;

: recvs				  \ c-addr1 +n1 fd -- +n2 ior
\ *G The common use case for reception. 
\ ** If the remote side closes the connection, *\fo{recvs} generates
\ ** an artificial errno code through *\fo{epipe?}.
\ ** Defined as :
\ *[ 
   MSG_NOSIGNAL <recvs> epipe? ;
\ *]


: recvs-all			      \ c-addr1 +n1 fd -- +n2 ior
\ *G Blocks until *\b{all} *\i{n1} bytes are received.
\ ** If the remote side closes connection, \fo{recvs-all} generates
\ ** an artificial errno code through *\fo{epipe?}.
\ ** Defined as :
\ *[   
   [ MSG_WAITALL MSG_NOSIGNAL or ] literal <recvs> epipe? ;
\ *]

\ ==============
\ *N UDP Sockets
\ ==============

\ *P This section defines additional words to handle UDP sockets.

\ ----------------
\ *H Reusable bits 
\ ----------------

AliasedExtern: ux-sendto int sendto(int, const void *, int, int, const void *, int);
\ *G See *\i{man sendto}.

AliasedExtern: ux-recvfrom int recvfrom(int, void *, int, int, void *, int *);
\ *G See *\i{man recvfrom}.

\ -------------
\ *H Main words
\ -------------

: sendsto			     \ c-addr1 +n1 addr2 fd -- +n2 ior
\ *G Send *\i{n1} bytes starting from *\i{c-addr1} through socket 
\ ** given by *\i{fd} to an *\fo{/endpoint} or *\fo{/sockaddr_in} 
\ ** given by *\i{addr2}. *\i{n2} is the actual amount of data sent.
\ ** Returns a throwable *\i{ior}.
\ ** *\i{n2} is undefined if *\i{ior} is non-zero.
\ ** See *\i{man sendto}.   
   swap >r -rot 0 r> /sockaddr_in  ux-sendto dup 0< throwable ;

: <recvsfrom>		     { c-addr1 n1 addr2 fd n2 -- +n3 ior }
\ *G Receive data from a remote *\fo{/endpoint} *\i{addr2}.
\ ** *\i{n1} is the amount to read through socket *\i{fd} and *\i{c-addr1} 
\ ** is the receiving buffer. Recepcion flags in *\i{n2} must be
\ ** or-ed together.
\ ** Returns the actual amount received in *\i{n3} and a throwable *\i{ior}.
\ ** See *\i{man recvfrom}.
   fd c-addr1 n1 n2 addr2 dup ep.rlen ux-recvfrom dup 0< throwable ;


: recvsfrom			     \ c-addr1 +n1 addr2 fd -- +n2 ior
\ *G The common use case for reception.
\ ** Equivalent to *\fo{<recvsfrom>} when *\i{n2} = 0.  
\ ** See *\fo{<recvsfrom>}.
  0 <recvsfrom> ;

   
: recvsfrom-all			     \ c-addr1 +n1 addr2 fd -- +n2 ior
\ *G Blocks until all *\i{n1} bytes are received.
\ **  Equivalent to *\fo{<recvsfrom>} when *\i{n2} = *\fo{MSG_WAITALL} flag.
\ ** See *\fo{<recvsfrom>}.   
   MSG_WAITALL <recvsfrom> ;

\ ==============================================
\ *N Non-blocking connection-oriented socket I/O
\ ==============================================

\ *P The following words are intended for use with a cooperative
\ ** multitasker using *\fo{pause} and non-blocking sockets,
\ ** although they also work with blocking sockets without multitasker.
\ ** All *\fo{xxx-mt} words are the multitasked counterparts of
\ ** *\fo{xxx} words. 


\ +P NOTE 03Jan07: xxx-mt words have been refactored to avoid throwing
\ +* exceptions from the inside words. The internal ior codes have been
\ +* proagated to the outermost word instead. This has caused a lot of
\ +* headaches and have added significant complexity to xxx-mt words.

\ ----------------
\ *H Reusable bits 
\ ----------------

Extern: int fcntl(int fd, int cmd, LONG arg);
\ *G See *\i{man fcntl}.

3 Constant F_GETFL
\ *G File control 'get flags' command.
\ ** See *\i{man fcntl}.

4 Constant F_SETFL
\ *G File control 'set flags' command.
\ ** See *\i{man fcntl}.

$800 Constant O_NONBLOCK
\ *G Mark a file descriptor as non-blocking.
\ ** To perform a multitasked I/O, instead of the usual multiplexed I/O
\ ** using *\f{select()}.
\ ** See *\i{man open}.

: fcntl@		\ fd -- n1 ior
\ *G Get the socket file control flags
\ ** (only handles O_APPEND, O_ASYNC, O_NONBLOCK, O_DIRECT).
\ ** Returns a throwable *\i{ior}. *\i{n1} is undefined if *\i{ior} <> 0.
\ ** See *\i{man fcntl}.   
   F_GETFL 0 fcntl dup 0< throwable ;

: fcntl!		\ n1 fd -- ior
\ *G Set the socket file control flags
\ ** (only handles O_APPEND, O_ASYNC, O_NONBLOCK, O_DIRECT).
\ ** *\i{n1} is the new flag value.
\ ** Returns a throwable *\i{ior}.
\ ** See *\i{man fcntl}.      
   swap F_SETFL swap fcntl 0< throwable ;

11  Constant EAGAIN			\ Try Again
\ *G *\i{errno} code returned for non-blocking I/O  in
\ ** *\fo{accepts sends recvs sendsto recvsfrom}.

114 Constant EALREADY		        
\ *G *\i{errno} code returned for non-blocking socket when performing
\ ** *\fo{connects} when former *\fo{connects} request is still in progress.

115 Constant EINPROGRESS		\ Operation now in progress
\ *G *\i{errno} code returned for non-blocking socket when performing
\ ** *\fo{connects} request that would need to wait for completion.

: again?		    \ ior1 -- ior2 flag
\ *G Process *\i{ior1} to filter the *\fo{EAGAIN} value.
\ ** In this case, *\i{flag} is true and *\i{ior2} is zero.
\ ** Any other *\i{ior1} will produce the same value in *\i{ior2}
\ ** and a false *\i{flag}.
   dup excp>errno EAGAIN = if drop 0 True else False then ;

: in-progress?			  \ ior1 -- ior2 flag
\ *G Process *\i{ior1} to filter the *\fo{EINPROGRESS} or
\ ** *\fo{EALREADY} values when using  *\fo{connects}.
\ ** In this case, *\i{flag} is true and *\i{ior2} is zero.
\ ** Any other *\i{ior1} will produce the same value in *\i{ior2}
\ ** and a false *\i{flag}.
\ ** Please note that *\fo{connects} can return *\fo{EAGAIN} but this
\ ** denotes a different error situation (see *\i{man connect}.)
   dup excp>errno dup EINPROGRESS = swap EALREADY = or
   if drop 0 True else False then  ;

\ -------------
\ *H Main words
\ -------------

: -blocking		  \ fd -- ior
\ *G Configure the socket as non blocking.
\ ** Returns a throwable *\i{ior}. 
   dup fcntl@ ?dup if nip exit then
   O_NONBLOCK or swap fcntl! ;

: +blocking		  \ fd -- ior
\ *G Configure the socket as blocking.
\ ** Returns a throwable *\i{ior}. 
   dup fcntl@ ?dup if nip exit then
   O_NONBLOCK invert and swap fcntl! ;

: peeks?		 \ fd -- flag ior
\ *G Peek into the socket system buffer to check for data available.
\ ** Returns a true *\i{flag} if one or more bytes are waiting to be read
\ ** and a throwable *\i{ior}.
   { |  tmp[ 1 ] -- } 
   tmp[ 1 chars rot [ MSG_PEEK MSG_NOSIGNAL or ] literal <recvs>
   epipe? nip   again? 0= swap ;


1 Value #ms-delay
\ *G poll period in non-blocking *\fo{xxx-mt} I/O words below 
\ ** setting this value to 0 increases the CPU load. The default value
\ ** is enough to produce a negligible load but with a reduced performace.

: delayed	    \ --
\ *G By using this word, the caller is delayed *\fo{#ms-delay ms}.
\ ** and allows task switching if multitasking is enabled. 
   #ms-delay ms ;

: connects-mt			\ addr1 fd1 -- ior
\ *G Non-blocking I/O counterpart of *\fo{connects}.
\ ** This word loops over *\fo{connects} and uses internally *\fo{delayed}.
   Begin 2dup connects in-progress? While drop delayed Repeat -rot 2drop ;

: accepts-mt		       \ addr1 fd1 -- fd2 ior
\ *G Non-blocking I/O counterpart of *\fo{accepts}.   
\ ** This word loops over *\fo{accepts} and uses internally *\fo{delayed}.
   Begin 2dup accepts again? While 2drop delayed Repeat 2swap 2drop ;

: (sends-mt)		   \ c-addr1 +n1 fd ior1 -- c-addr2 n2 fd ior2
\ +G This word is an internal factor of *\fo{sends-mt} for non-blocking I/O
\ +* and should not be used.
\ +* Tries to send *\i{n1} bytes to socket given by *\i{fd}
\ +* and loops over until the socket says it would not block.
\ +* Returns the remaining buffer *\i{c-addr2 n2} to be sent.

   Begin
      drop 3dup			        \ c-addr1 +n1 fd c-addr1 +n1 fd 
      sends again?			\ c-addr1 +n1 fd n2 ior3 flag
   While				\ c-addr1 +n1 fd n2 ior3
	 drop delayed			\ c-addr1 +n1 fd n2
   Repeat
   rot					\ c-addr1 +n1 n2 ior3 fd
   2>r /string r> r> ;

: (recvs-mt)		   \ c-addr1 n1 fd ior1 -- c-addr2 +n2 fd ior2
\ +G This word is an internal factor of *\fo{recvs-mt} for non-blocking I/O
\ +* and should not be used.   
\ +* Tries to receive *\i{n1} bytes to socket given by *\i{fd}
\ +* and loops over until the socket says it would not block.
\ +* Returns the remaining buffer *\i{c-addr2 n2} to be received.
   Begin
      drop 3dup			        \ c-addr1 +n1 fd c-addr1 +n1 fd
      recvs again?		        \ c-addr1 +n1 fd n2 ior2 flag
   While				\ c-addr1 +n1 fd n2 ior
	 drop delayed			\ c-addr1 +n1 fd n2
   Repeat
   rot					\ c-addr1 +n1 n2 ior fd
   2>r /string r> r> ;

: #bytes-io			      \ c-addr1 n1 x1 ior n2 -- n3 ior
\ +G This word is a factor for some loop epilogs below and should not be used.
\ +* Returns the number of bytes processed in I/O in *\i{n3}
\ +* from the total requested *\i{n2} and the remaining amount *\i{n1}.
\ +* Reorders the result to place *\i{ior} on TOS.
\ +* *\i{x1} is a don't care cell left in the loop exist that gets
\ +* dropped in the way.
   rot drop  \ c-addr1 n1 ior n2
   rot -     \ c-addr1 ior n3
   rot drop  \ ior n3
   swap ;
   
: sends-mt			      \ c-addr1 n1 fd -- n2 ior
\ *G Non-blocking I/O counterpart of *\fo{sends}.
\ ** This word loops over *\fo{sends} and uses internally *\fo{delayed}.
   over >r 0 (sends-mt) r> #bytes-io ;


: recvs-mt			      \ c-addr1 +n1 fd -- +n2 ior
\ *G Non-blocking I/O counterpart of *\fo{recvs}.
\ ** This word loops over *\fo{recvs} and uses internally *\fo{delayed}.
   over >r 0 (recvs-mt) r> #bytes-io ;

: halt-loop?   \ n2 fd ior -- n2 ior fd flag
\ +G This word is an internal factor of  *\fo{recvs-all-mt}. Do not use.
\ +* This word factors the loop end detection while preserving stack ordering.
\ +* Returns true if *\i{ior} < 0 or *\i{n2} = 0, false otherwise.
   3dup nip 0< swap 0= or ;		


: recvs-all-mt			      \ c-addr1 +n1 fd -- +n2 ior
\ *G Non-blocking I/O counterpart of *\fo{recvs-all}.
\ ** This word loops over *\fo{recvs} and uses internally *\fo{delayed}.
   over >r 0 Begin (recvs-mt) halt-loop? until r> #bytes-io ;


\ =========================================
\ *N Non-blocking connectionless socket I/O
\ =========================================

\ *P The following words are intended for use with a cooperative
\ ** multitasker using *\fo{pause} and non-blocking sockets,
\ ** although they also work with blocking sockets without multitasker.
\ ** All *\fo{xxx-mt} words are the multitasked counterparts of
\ ** *\fo{xxx} words. 

\ ----------------
\ *H Reusable bits 
\ ----------------

: lover				  \ x1 x2 x3 -- x1 x2 x3 x1
\ *G "long" *\fo{over}, although the name actually suggest
\ **  something different :-)
   2 pick ;

: halt2-loop?		     \ n2 addr2 fd ior -- n2 addr2 ior fd flag
\ +G This word is an internal factor of *\fo{recvs-all-mt} and
\ +* *\fo{recvsfrom-all-mt}.  Do not use.   
\ +* This word factors the loop end detection while preserving stack ordering.
\ +* Returns true if *\i{ior} < 0 or *\i{n2} = 0, false otherwise.
   4dup -rot 2drop 0< swap 0= or ;

: (sendsto-mt)	\ c-addr1 +n1 addr2 fd ior -- c-addr3 +n2 addr2 fd ior
\ +G This word is an internal factor of *\fo{sendsto-mt} for non-blocking I/O
\ +* and should not be used.   
\ +* Tries to send *\i{n1} bytes to socket *\i{fd}
\ +* and loops over and over again until socket says it would not block.
\ +* Returns the remaining buffer *\i{c-addr3 n2} to be sent.
\ +* *\i{addr2} denotes the destination IP *\fo{/endpoint}.
\ +* *\i{ior} is propagated from *\fo{sendsto}.
   Begin
      drop 4dup				\ 2x (c-addr1 +n1 addr2 fd)
      sendsto again?			\ c-addr1 +n1 addr2 fd n2 ior flag
   While				\ c-addr1 +n1 addr2 fd n2 ior
	 drop delayed			\ c-addr1 +n1 addr2 fd n2
   Repeat
   rot					\ c-addr1 +n1 addr2 n2 ior fd
   2>r					\ c-addr1 +n1 addr2 n2
   swap >r				\ c-addr1 +n1 n2
   /string				\ c-addr2 +n2
   r> r> r> ;

: (recvsfrom-mt)   \ c-addr1 +n1 addr2 fd ior -- c-addr1 +n2 addr2 fd ior
\ +G This word is an internal factor of *\fo{recvsfrom-mt}
\ +* for non-blocking I/O and should not be used.   
\ +* Tries to receive *\i{+n1} bytes into buffer *\i{c-addr1}
\ +* from socket *\i{fd} and loops over and over again until socket says
\ +* it would not block.
\ +* Returns the remaining buffer *\i{c-addr3 n2} to be received.
\ +* *\i{addr2} denotes the destination IP *\fo{/endpoint}.
\ +* *\i{ior} is propagated from *\fo{recvsfrom}.
   Begin
      drop 4dup				\ 2x (c-addr1 +n1 addr2 fd)
      recvsfrom again?		    \ c-addr1 +n1 addr2 fd n2 ior flag
   While				\ c-addr1 +n1 addr2 fd n2 ior
	 drop delayed
   Repeat
   rot					\ c-addr1 +n1 addr2 n2 ior fd
   2>r					\ c-addr1 +n1 addr2 n2
   swap >r				\ c-addr1 +n1 n2
   /string				\ c-addr2 +n2
   r> r> r> ; 

\ -------------
\ *H Main words
\ -------------

: sendsto-mt			     \ c-addr1 +n1 addr2 fd -- +n2 ior
\ *G Non-blocking I/O counterpart of *\fo{sendsto}.
\ ** This word loops over *\fo{sendsto} and uses internally *\fo{delayed}.
   lover >r 0 (sendsto-mt) nip r> #bytes-io ;

: recvsfrom-mt			     \ c-addr1 +n1 addr2 fd -- +n2 ior
\ *G Non-blocking I/O counterpart of *\fo{recvsfrom}.
\ ** This word loops over *\fo{recvsfrom} and uses internally *\fo{delayed}.
   lover >r 0 (recvsfrom-mt) nip r> #bytes-io ;
   
: recvsfrom-all-mt		     \ c-addr1 +n1 addr2 fd -- +n2 ior
\ *G Non-blocking I/O counterpart of *\fo{recvsfrom-all}.
\ ** This word loops over *\fo{recvsfrom-all}
\ ** and uses internally *\fo{delayed}.
\ *P *\b{Warning:} Unpredictable results are obtained if several
\ ** simultaneous clients send their data to the same connectionless socket
\ ** in the receiving side. Data will be mixed in the
\ ** same buffer and a wrong byte count will be reported.
   lover >r 0 Begin (recvsfrom-mt) halt2-loop? until nip r> #bytes-io ;

\ ==================
\ *N Multiplexed I/O
\ ==================

\ *P An alternative to multitasked I/O when blocking sockets are used.
\ ** The *\i{poll()} system call is easier to implement than *\i{select()}. 

\ ----------------
\ *H Reusable bits 
\ ----------------

\ *P Internal documentation.

AliasedExtern: ux-poll int poll(void * ufds, unsigned int nfds, int timeout);
\ +G See *\i{man poll}.

\ -------------
\ *H Main words
\ -------------

$0001 Constant POLLIN
\ *G On *\fo{pollfd.events}: notify when data ready.
\ ** On *\fo{pollfd.revents}: data ready to read.

$0004 Constant POLLOUT
\ *G On *\fo{pollfd.events}: notify when write will not block.
\ ** On *\fo{pollfd.revents}: write will not block.

$0008 Constant POLLERR
\ *G On *\fo{pollfd.revents}: Error condition.

$0010 Constant POLLHUP
\ *G On *\fo{pollfd.revents}: Hangup. Remote socket closed.

$0020 Constant POLLNVAL
\ *G On *\fo{pollfd.revents}: Invalid request. File descriptor is not open.



struct /pollfd				\ -- len
\ *G Mimics the Linux *\i{struct pollfd}. See *\i{man poll}.         
\ *[
   1 cells field pollfd.fd		\ file descriptor
   2 chars field pollfd.events   	\ requested events
   2 chars field pollfd.revents   	\ returned events
end-struct
\ *]

: polls					  \ addr1 n1 n2 -- n3 ior
\ *G Poll the system for I/O events.
\ ** An array of *\fo{/pollfd} structures is passed starting at *\i{addr1}.
\ ** *\i{n1} is the array length in structure units (not bytes).
\ ** *\i{n2} is a timeout in milliseconds. *\i{n2} < 0 indicates
\ ** an infinite timeout.
\ ** *\i{n3} is the number of structures having *\fo{pollfd.revents} <> 0
\ ** If *\i{n3} = 0, a timeout has occured.
\ ** If *\i{n3} < 0, an internal error has occured and
\ ** *\i{ior} contains a throwable  *\i{errno} code.
\ ** See *\i{man poll}.
\ ** *\br{}
\ ** *\b{NOTE:} Tests made shows that when a remote side closes the socket
\ ** poll returns with an *\fo{POLLIN} event instead of *\fo{POLLHUP}.
\ ** Performing a *\fo{recvs} on that socket returns a zero length.
   ux-poll dup 0< throwable ;


\ ======
\ *> ###
\ ======

