\ Fuzzy logic Forth

((
by Rick van Norman
==================

The following code was presented with my paper FUZZY FORTH at the
1994 Rochester Forth Conference.  The code presented is hereby
released into the public domain for any use whatsoever, with all 
risk assumed by the user and none by the author.  Please give credit
where it is due.

It is my intention that this code be a beginning for a library of 
ANS compliant extensions for Forth.  Please hack at it and let me
know if I did anything to make it non-compliant.

This code will run on an ANS compliant system with the core extensions
word set.  It should be indifferent to word size etc. 

ANS requires a header for compliant source code.  Would someone be 
so kind as to provide one as an example?

SORRY IT IS SOOOOOOOOO LONG! 
))

\ Fuzzy Forth, (C) 1993, 1994 Rick VanNorman

[undefined] cell- [if]
: cell-		\ addr -- addr'
  cell -  ;
[then]

\ store two entries in the dictionary
: 2, ( n1 n2 -- )   SWAP , , ;

\ division by zero is allowed, and results in a zero returned
: DIV ( n1 n2 -- n3 )   dup if / else nip then ;

\ calculate the slope between x1=p2 and x2=p1, assuming ONE = 65536 and that
\ the y1=0 and y2=ONE.  if x1=x2, the slope is properly infinity, but
\ we only apply the calculated slope where we know there _is_ a slope.
\ FUZZIFY will trap this condition.

: SLOPE		( P2 P1 -- N )
  - 65536, ROT ?DUP
  IF  UM/MOD NIP ABS  ELSE  2DROP 0  THEN  ;


VARIABLE MLINK       \ TEMP VALUE FOR MEMBER LINKS

: LINK,              \ store a link in the current definition
   HERE MLINK @ !  HERE MLINK !  0 , ;


\ ----- 

: X? ( ADDR -- )
   @ 8 .R ;

: SEE ( ADDR -- )
   DUP X?  CELL+ BEGIN
      @ ?DUP 
   WHILE
      CR  
      DUP CELL - X?  \ SHOW FV
      DUP CELL + 6 CELLS BOUNDS 
      DO  I X?  CELL +LOOP 
   REPEAT ;
      
: SHOW ( ADDR -- )
   DUP X? CELL+
   BEGIN
      @ ?DUP 
   WHILE 
      DUP CELL - X?
   REPEAT ;      

\ input variable and membership function definition  
\ -----
\ 
\ a fuzzy input consists of 1) the crisp input variable and a range over which 
\ it is valid and 2) the set of membership functions defined on that range.  
\ a crisp input value is evaluated over each of the membership functions, 
\ resulting in the fuzzy value of the crisp in each membership function.
\ evaluation proceeds left-to-right, and the crisp is determined to be one of
\    a) left of the membership function [returns zero] 
\    b) between the leftmost point and the left top point [returns 0--256]
\    c) between the left top and the right top [returns 256]
\    d) between the right top and the rightmost point [returns 256--0]
\    e) right of the membership function [returns zero]
\               _____
\             /|     |\                  a typical membership function.
\            / |     | \                 note that points LT and RT can 
\           /  |     |  \                be equal, resulting in a triangle,
\       ___/   |     |   \___            and that LM:LT and RT:RM could be 
\        a | b |  c  | d | e             equal, resulting in a step function.
\          |   |     |   |                     
\         lm  lt     rt  rm                    LM <= LT <= RT <= RM
\
\ 0 100 INPUT ne
\     0  0 18  21 MEMBER ne_small
\    10 40 40  70 MEMBER ne_medium
\    50 90 90 120 MEMBER ne_big
\
\ Ŀ
\  ne  crisp  link  low (0) high (100)
\ 
\                  Ŀ
\                          
\    Ŀ
\     ne_small    fv  link  lm  (0) lt  (0) rt (18) rm  (21) ls  rs 
\    
\                          
\    Ŀ
\     ne_medium   fv  link  lm (10) lt (40) rt (40) rm  (70) ls  rs 
\    
\                          
\    Ŀ
\     ne_big      fv  0000  lm (50) lt (90) rt (90) rm (120) ls  rs 
\    
\

: INPUT ( low high "name" -- )
   CREATE
      0 ,                                    \ CRISP STORAGE
      HERE MLINK !  0 ,                      \ INITIAL LINK
      2, ;                                   \ the low and high ranges

: SET_SLOPE ( ilink -- )
   >R  R@ 1 CELLS + 2@ SLOPE  R@ 5 CELLS + !
       R@ 3 CELLS + 2@ SLOPE  R> 6 CELLS + ! ;

: MEMBER ( lm lt rt rm "name" -- )
   CREATE
      0 ,  LINK,                          \ fuzzy value and link to prev
      >R >R  2,  R> R> 2,   0 0 2,        \ store points, left to right
      MLINK @ SET_SLOPE                   \ calculate slopes of the edges
   DOES>  @ ;                             \ return the fuzzy value



\ output variable and de-fuzzification set definition
\ -----
\
\ a fuzzy output consists of a crisp output variable and a set of output 
\ singletons which accumulate the results of rule evaluations. 
\ before evaluation, all FV accumulators in the singletons are cleared.  
\ during evaluation, the results are applied to various FV accumulators.  
\ after evaluation, the accumulators are de-fuzzified and applied with
\ the singleton values to yield a crisp output value.
\ 
\ a typical output function, showing all singletons...
\
\     |   |        |
\ ____|___|________|______
\     9  10       12
\
\ OUTPUT ki
\     9 SINGLETON ki_small
\    10 SINGLETON ki_medium
\    12 SINGLETON ki_big
\ 
\ 
\     Ŀ
\      ki  crisp  link 
\     
\                      Ŀ
\                                  
\            Ŀ
\             ki_small    fv  link  singleton  (9)
\            
\                                  
\            Ŀ
\             ki_medium   fv  link  singleton (10)
\            
\                                  
\            Ŀ
\             ki_big      fv  0000  singleton (12)
\            
\ 

: OUTPUT
   CREATE
      0 ,                                    \ CRISP STORAGE
      HERE MLINK !  0 , ;                    \ INITIAL LINK

: SINGLETON ( value "name" -- )
   CREATE
      0 , LINK,                              \ fuzzy value and link to prev
      ( value) ,                             \ and the singleton value itself
   DOES>  DUP @ ROT MAX SWAP ! ;             \ accumulate rule evaluations

\ initialize an output set

: CLEAR_OUTPUT ( addr -- )
   CELL+
   BEGIN	                             \ 
      @ ?DUP WHILE  0 OVER CELL- !           \ following the links, write 0s
   REPEAT ;                                  \ to the fuzzy accumulators



\ RULE EVALUATION
\ -----
\ rules are of the form 
\     xyz abc & => ddd
\ which says "take the minimum of the membership functions as evaluated 
\ by the MEMBER words XYZ and ABC, leaving the result on the stack to be
\ absorbed by the SINGLETON definition DDD (a fuzzy output accumulator).
\ valid operations are AND (&) OR (|) and NOT (~), with each "logical"
\ value in the range 0--256


: &   MIN ;
: |   MAX ;
: ~   256 SWAP - ;

: => ; IMMEDIATE

\ conclusions after rule evaluation
\ ----- 
\ to generate a conclusion, we accumulate a "sum of area" and a "sum
\ of products" term for the set of singletons (which are taken to be
\ point sources whose weight was accumulated by the rule evaluations).  
\ dividing SOP by SOA results in the center of mass of the point sources, 
\ which we take to be the de-fuzzified value of the output variable

VARIABLE SOA      \ sum of area
VARIABLE SOP      \ sum of products

: CONCLUDE ( OVAR -- CRISP )
   0 SOA !  0 SOP !              \ CLEAR SUM OF AREA, SUM OF PRODUCT TERMS
   DUP >R                        \ save the address of the output var 
   CELL+
   BEGIN                         \ and pointing to the singleton's link
      @ ?DUP
    WHILE                        \ repeat while not at the end of the list
      DUP CELL- @  DUP SOA +!    \ SOA is the sum of all fuzzy values
      OVER CELL+ @               \ SOP is sum of all (fuzzy * weight) values
      * SOP +!                   \ 
   REPEAT                        \  
   SOP @ SOA @ DIV               \ divide SOP by SOA, giving the center of mass
   DUP R> ! ;                    \ which we interpret as the de-fuzzification


\ fuzzify the inputs
\ ----- 
\
\ we fuzzify the inputs by applying the crisp value to each of the 
\ membership functions and evaluating its position in the membership.  
\

\ fuzzyfiy the crisp value on the left edge of the membership (ie region B) 
\ via the slope of that line segment and its X position relative to the 
\ start of the line segment.  

: FUZLEFT ( CRISP MEMBER -- FUZZY )
   DUP >R  @ -  R> 4 CELLS + @ 256 */  ABS ;


\ fuzzyfiy the crisp value on the right edge of the membership (ie region D) 
\ via the slope of that line segment and its X position relative to the 
\ start of the line segment.  

: FUZRIGHT ( CRISP MEMBER -- FUZZY )
   DUP >R  3 CELLS + @ SWAP -  R> 5 CELLS + @ 256 */  ABS ;


\ given a crisp number, a membership function, and a cell offset, return
\ the crisp value and the range point for comparison

: IN ( crisp member n -- crisp member crisp range )
   >R 2DUP R> CELLS + @ ;


\ given a crisp value and a pointer to the membership points, figure out
\ where the crisp value falls (left to right) and assign a membership
\ value for it.  membership values are in the range (0--256)

: FUZZIFY ( CRISP MEMBER -- FUZZY )
   0 IN  <  IF ( BELOW LEFTMOST )  2DROP 0    EXIT THEN  ( is zero)
   1 IN  <  IF ( LM <= C <= LT  )  FUZLEFT    EXIT THEN  ( is calculated)
   2 IN  <= IF ( LT <= C <= RT  )  2DROP 256  EXIT THEN  ( is "one")
   3 IN  <= IF ( RT <= C <= RM  )  FUZRIGHT   EXIT THEN  ( is calculated)
               ( ABOVE RIGHTMOST)  2DROP 0    ;          ( is zero)

\ evaluate the crisp value in the specified membership function.  the 
\ result is stored in the first cell.  the address of the member passed
\ to this routine is the a pointer to its link field.

: APPLIES ( crisp member -- )
   TUCK  CELL+ FUZZIFY  SWAP CELL- ! ;


\ apply the crisp input to the specified fuzzy input variable

: APPLY ( crisp input -- )
   2DUP !  CELL+ BEGIN              \ keep the original crisp value
      @ ?DUP WHILE                  \ and scan through each of the memberships
      ( CRISP LINK)  2DUP APPLIES   \ applying the crisp to them
   REPEAT DROP ;


   
\ ---------------------- an example of use ----------------------------
\ a simple fuzzy application

0 100 INPUT ne
    0  0 18  21 MEMBER ne_small  
   10 40 40  70 MEMBER ne_medium 
   50 90 90 120 MEMBER ne_big    

0 80 INPUT oe
    0  0 18 21 MEMBER oe_small  
    0 30 30 60 MEMBER oe_medium 
   20 50 50 80 MEMBER oe_big    

OUTPUT kp
    8 SINGLETON kp_very_small 
   10 SINGLETON kp_small      
   12 SINGLETON kp_low_medium 
   16 SINGLETON kp_high_medium
   18 SINGLETON kp_big        
   24 SINGLETON kp_very_big   

OUTPUT ki
    9 SINGLETON ki_small    
   10 SINGLETON ki_medium   
   12 SINGLETON ki_big      

: RULES
   kp clear_output
   ki clear_output
   ne_small     oe_small     &    =>   kp_low_medium
   ne_small     oe_medium    &    =>   kp_big
   ne_small     oe_big       &    =>   kp_very_big
   ne_medium    oe_big       &    =>   kp_high_medium
   ne_medium                      =>   kp_small
   ne_big                         =>   kp_very_small
   ne_small                       =>   ki_medium
   ne_medium                      =>   ki_big
   ne_big                         =>   ki_small ;

\ -----

: TEST
   CR ." KP"
   91 0 DO CR
      I ne APPLY
      91 0 DO
         I oe APPLY
         RULES   kp CONCLUDE .
      5 +LOOP
      KEY? IF LEAVE THEN
   5 +LOOP
   CR ." KI"
   91 0 DO CR
      I ne APPLY
      91 0 DO
         I oe APPLY
         RULES   ki CONCLUDE .
      5 +LOOP
      KEY? IF LEAVE THEN
   5 +LOOP ;

\ all done

decimal


