\ Editor Colorizer version 3.0(beta)
\ Last Revision: 04/24/08  02:53:51 PM  wjm
\ Last Revision: 07/09/08  05:56:18 AM  wjm
\ Last Revision: 20 Dec 2011  06:47:51  ward@megawolf.com

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\    to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************


(* *************************************************************
   *************************************************************
   *************************************************************

 Syntax colorizing for the MacForth Editor 

 Original proof of concept and working version from Doug Hoffman
This is a major revision to allow multi-line color support for comments,
conditionals, etc

 Color can be enabled/disabled on a per-file basis by using the little palette
widget just to the left of the Horizontal scroll bar at the bottom of an editor
window.  This is saved as a preference when the file is saved, as is window position,
current selection, and jump marks.



Since this version must scan from the start of a file, long files (say, more than 10K
lines) can become a bit sluggish scrolling, so in those cases, reverting back to
mono-color may be preferable. 

The other time when mono-color is prefereable is when you want to programatically
dump a lot  of text in an editor window (say, running WORDS)

The original editor was highly tuned for rapid display on old slow 68K class Macs.
The drawing speed of these routines is considerably slower, but should be plenty
fast for most uses and on most modern Macs.

We welcome any suggestions for improvements in speed or behavior.


Two especially slow related operations to colorize in largish files are:
 drag-scrolling (selecting a section of text and dragging below or above the visible
area to force extending the selected text while scrolling.)
 continuous scrolling via clicking and holding in the vertical scrollbar

Both these actions default to being displayed as monocolor, colorizing when the mouse
is released.  Executing:
   TRUE TO COLOREDDRAGS
will force continuous uopdating of colorization during these scroll operations.



Colors, and optionally styles ("textface" in modern jargon) are controlled for
the following classes of phrases and words:

 Plain text
 Comments  (including multiline ones)
 "Meta comments" (normally only the stuff that follows a PRIOR.STREAM )
 Quoted strings
 Local definitions (but not words defined as local)
 "ticked" words  ( ' ['] POSTPONE [COMPILE] IS ) 
 Loop words
 Common creating words
 Conditionals (both immediate and compiled ones)
 [conditional]'s like [IF] [IFDEF]

 Vocabulary words
 "Mops-style" Selector: words

To see all the "special" words, execute:
	PARSERWID WORDSLIST


The default colors and texface can be modified either in this file, or after
compilation, in DEFAULT.EXTENSIONS

Textface can be the usual PLAIN BOLD ITALIC UNDERLINE (or combination), but many
monspaced fonts may change character width with BOLD or become hard to read in
ITALIC.  Be sure to test your changes for readability.

This is the syntax for (re)setting the various special class appearances:
	
    TEXTFACE	rgbCOLOR				STATE NAME
    ====================================================================
	plain		rgbBlack				PlainTextColor 		!EditorStyle
	plain		rgbBlue					CreatorColor 		!EditorStyle
	plain		rgbRed					ConditionalColor 	!EditorStyle
	plain		rgbTangerine			LoopColor 			!EditorStyle
	plain		rgbMidnight				TickColor 			!EditorStyle
	underline	rgbDkGray				QuoteColor 			!EditorStyle
	plain		rgbOrange				LocalsColor 		!EditorStyle
	plain		rgbVeryDkGrayColor		CommentColor 		!EditorStyle
	plain		rgbAqua					MetaCommentColor 	!EditorStyle
	italic		rgbAqua					SpecialCommentColor !EditorStyle
	plain 	  	rgbOlive				[Conditional]Color 	!EditorStyle
	plain 		rgbSalmon				VocabColor 			!EditorStyle
	plain 		rgbAqua					StructureColor 		!EditorStyle
	plain 		rgbBrown				SelectorColor 		!EditorStyle



These colors are available from "Colors"
	0000  0000  0000  RGBColor RGBBlack
	FC00  F37D  052F  RGBColor RGBYellow
	F2D7  0856  84EC  RGBColor RGBMagenta
	DD6B  08C2  06A2  RGBColor RGBRed
	0241  AB54  EAFF  RGBColor RGBCyan
	0000  64AF  11B0  RGBColor RGBGreen
	0000  0000  D400  RGBColor RGBBlue
	FFFF  FFFF  FFFF  RGBColor RGBWhite
	F000  F000  F000  RGBColor RGBLtGray
	A000  A000  A000  RGBColor RGBDkGray

These colors are added in this file:
	FFFF   8080   0000  RGBColor rgbTangerine 
	6666   CCCC   FFFF  RGBColor rgbSkyBlue
	6666   6666   6666  RGBColor rgbSteelColor
	0000   8080   FFFF  RGBColor rgbAqua 
	0000   0000   8080  RGBColor rgbMidnight
	4000   4000   4000  RGBColor rgbVeryDkGrayColor
	FFFF   6666   6666  RGBColor rgbSalmon
	D000   8000   0000  RGBColor rgbOrange
	8080   4040   0000  RGBColor rgbBrown
	4F6A   4040   0000  RGBColor rgbOlive

You can easily define additional colors
To make it easy, use the file  ":Snippets:Color Picking"

To show these colors as they look in the editor:
	ALtGrayWord
	ADkGrayWord
	ASteelColorWord
	AVeryDkGrayWord
	ABlackWord
	
	AYellowWord
	ASalmonWord
	ATangerineWord
	AnOrangeWord
	ARedWord
	
	ASkyBlueWord
	ACyanWord
	AnAquaWord
	ABlueWord
	AMidnightWord
	
	AMagentaWord
	AGreenWord
	AnOliveWord
	ABrownWord


\ ForbidColorizing   	- even commented, this turns colorizing off for whatever follows
\ AllowColorizing		- even commented, this turns colorizing back on


HOW IT WORKS:

A special "vocabulary" is set up to contain the words that trigger color changes
and the drawing routines use this vocabulary to parse the file contents using a
very similar mechanism to compiling a file in Forth.

The parsing is done in two phases -
1.  ScanFromStartToAddr does it from the start of the file up to the top visible
	line, without any drawing
2.  DrawOnlyVisibleLines does as its name suggests - draws the visible lines

#2 is pretty fast.
#1 can be slowish if the file is large and you are viewing near the end of the
	file. 

To speed up redisplay when text is being inserted from the keyboard (or via a paste
that does not trigger changing the top line), ScanFromStartToAddr saves the
color and ColorizerState for the top line.  If the top line has not changed,
RestoreTopLineState and DrawOnlyVisibleLines is sufficient and the redisplay is
much faster on a long file.



ENJOY !!


************************************************************* *)


only forth definitions

\ anew --ColorEditorDisplay--

>m4th
also editorvocab definitions
also forth also m4th


decimal

\ *************************************************************
\ *************************************************************

\ More Colors:

$FFFF  $8080  $0000  RGBColor rgbTangerine 
$6666  $CCCC  $FFFF  RGBColor rgbSkyBlue
$6666  $6666  $6666  RGBColor rgbSteelColor
$0000  $8080  $FFFF  RGBColor rgbAqua 
$0000  $0000  $8080  RGBColor rgbMidnight
$4000  $4000  $4000  RGBColor rgbVeryDkGrayColor
$FFFF  $6666  $6666  RGBColor rgbSalmon
$C000  $8000  $0000  RGBColor rgbOrange
$8080  $4040  $0000  RGBColor rgbBrown
$4F6A  $4040  $0000  RGBColor rgbOlive

\ *************************************************************
\ *************************************************************
\ *************************************************************


: EditorFontStyle: ( textface\r\g\b  "<spaces>name" -- ) (  -- addr )
	RGBColor
	w,
;

: !EditorStyle ( Textface\^rgbcolor\^Editorstyle -- )
	dup >r
	6 move
	r> 6 + w!
;

: ApplyEditorFontStyle ( ^EditorFontStyle -- )
	dup RGBForecolor
	6 + w@ TextFace
;

plain		rgbBlack				@rgb EditorFontStyle: PlainTextColor
plain		rgbBlue					@rgb EditorFontStyle: CreatorColor
plain		rgbRed					@rgb EditorFontStyle: ConditionalColor
plain		rgbTangerine			@rgb EditorFontStyle: LoopColor
plain		rgbMidnight				@rgb EditorFontStyle: TickColor
underline	rgbSkyBlue				@rgb EditorFontStyle: QuoteColor
\ underline	rgbDkGray				@rgb EditorFontStyle: QuoteColor
plain		rgbOrange				@rgb EditorFontStyle: LocalsColor

\ probably make the 2 flavors of comments the same color eventually
\ keep them distinct for testing
plain		rgbSteelColor			@rgb EditorFontStyle: CommentColor
\ plain		rgbVeryDkGrayColor		@rgb EditorFontStyle: CommentColor
plain		rgbSteelColor			@rgb EditorFontStyle: MetaCommentColor
\ plain		rgbCyan					@rgb EditorFontStyle: MetaCommentColor
italic		rgbCyan					@rgb EditorFontStyle: SpecialCommentColor

plain 	  	rgbOlive				@rgb EditorFontStyle: [Conditional]Color 
plain 		rgbSalmon				@rgb EditorFontStyle: VocabColor
plain 		rgbAqua					@rgb EditorFontStyle: StructureColor
plain 		rgbBrown				@rgb EditorFontStyle: SelectorColor
underline	rgbOrange				@rgb EditorFontStyle: HUH?Color

underline	rgbBlack				@rgb EditorFontStyle: InterpretedColor
underline	rgbRED					@rgb EditorFontStyle: CommandColor



\ *************************************************************

\ 13 value EOLChar	     \ in editor
\ 202 constant NBSpChar  \ in editor


\ internal deferred words
defer InitColor ( -- )						' noop is InitColor
defer AdjustOneShots ( -- )					' noop is AdjustOneShots
defer ParseEOL ( -- )						' noop is ParseEOL
defer ParseAWord ( addr\cnt -- addr\cnt )	' noop is ParseAWord
defer PreParser ( addr\cnt -- addr\cnt\continueparsing? )
\ the PreParser handles special color state terminators, ) or *)
	' true is PreParser
defer PostParser	  ( addr\cnt -- addr\cnt ) ' noop is PostParser
defer OtherPostParser ( addr\cnt -- addr\cnt ) ' noop is OtherPostParser
\ the OtherPostParser can handle other "specials", like numbers or selectors:


((
 Editor Defer Words that we modfiy:
	AfterCharInsert
	AfterRangeInsert
	UpdateAfterScroll
	UpdateAfterFlush
	UpdateAfterRedrawFrom
	BOTTOM.ONE
	TOP.ONE
	DRAW.ALL.LINES
))

\ *************************************************************

0 value OneShotColorFlag
0 value OneShotColorState
create TempSavedFontState 	8 allot&erase


0 value CurrentLine#		\ 0-based index of VISIBLE lines

: CR.MOVE2
	ParseEOL
	MoveToNewLine	\ cr.move
	1 +to CurrentLine#
	\ ? check if in visible range
;

: DrawABlank	( --heh-- )
	bl DrawChar
;


: DrawSeparator ( addr\char -- addr\MoreInLine? )	\ no parsing
	true swap
	CASE
		bl			OF	DrawABlank				ENDOF
		13			OF	MoveToNewLine	1 +to CurrentLine#	drop false	ENDOF
		10			OF	MoveToNewLine	1 +to CurrentLine#	drop false	ENDOF
		09			OF	tab.move				ENDOF
		BSOUT		OF	se.backspace.move		ENDOF
		NBSpChar	OF	Draw.NBspace			ENDOF
		255 DrawChar			\ missing char for other controls
	ENDCASE
;

: Parse&DrawSeparator ( addr\char -- addr\MoreInLine? )
	true swap
	CASE
		bl			OF	DrawABlank				ENDOF
		13			OF	cr.move2	drop false	ENDOF
		10			OF	cr.move2	drop false	ENDOF
		09			OF	tab.move				ENDOF
		BSOUT		OF	se.backspace.move		ENDOF
		NBSpChar	OF	Draw.NBspace			ENDOF
		MissingCharChar DrawChar		\ missing char for other controls
	ENDCASE
;


: (Parse&DrawWord) ( addr\cnt --  )
	ParseAWord			\ is WordParser
	0 swap	_DrawText  ( textBuf\byteOffset\byteCount -- )
	AdjustOneShots		\ is DecrementColorizerStateCount
;


0 value theCount		\ length of the word being parsed
0 value theAddr			\ start address of the word being parsed
0 value theChar			
0 value #parsed			\ can be removed -- keeps tally for timing analysis

: Parse&DrawWord ( --  )
	theCount 0>			\ most delimiters don't change color
	IF
		theAddr theCount  (Parse&DrawWord)
		theAddr theCount + to theAddr
		0 -> theCount
		1 +to #parsed		\ can be removed -- keeps tally for timing analysis
	THEN
;

: ParseWordNoDraw ( --  )
	theCount 0>			\ most delimiters don't change color
	IF
		theAddr theCount ParseAWord ( addr\cnt -- addr\cnt )
		( theAddr theCount ) + to theAddr
		0 -> theCount
		1 +to #parsed		\ can be removed -- keeps tally for timing analysis
		AdjustOneShots		\ is DecrementColorizerStateCount
	THEN
;

: DrawWordNoParse ( --  )
	theCount 0>			\ most delimiters don't change color
	IF
		theAddr 0 theCount 	_DrawText
		theAddr theCount + to theAddr
		0 -> theCount
	THEN
;

: ParseSeparator ( char -- notEOL? )
	SE.EOLChar @ = 
	dup 
	IF ParseEOL THEN
	0=
;


0 value EndingAddr

: DrawThisLineWhileParsing ( addr -- addr' )
	0 -> theCount
	dup -> theAddr					\ theAddr is start of this line

	BEGIN							\ running addr on stack
		dup EndingAddr <			\ not at end of range
	WHILE
		dup c@						\ addr\char
		dup -> theChar				\ ??? needed
		dup  bl >					\ not a control character
		swap NBSpChar <>	AND		\ AND theChar <>  NBSP
		IF
			1 +to theCount				\ increment count
			1+							\ increment running address
			true
		ELSE							 \ char = a separator
			Parse&DrawWord				 \ also updates theAddr
			theChar Parse&DrawSeparator	 \ this also detects EOL's
			swap						 \  MoreInLine?\addr
			1+							 \ increment running address
			0 -> theCount
			1 +to theAddr				 \ incr theAddr
			swap						 \ addr\MoreInLine?
		THEN
		\ addr'\flag = true if ok to continue
		0=								 \ no more in this line? 
		IF
			exit
		THEN
	REPEAT
	theCount 0>
	IF
		-1 to theChar
		Parse&DrawWord
	THEN
;


: ParseThisLineWithoutDraw ( addr -- addr' )
	0   -> theCount
	dup -> theAddr					\ theAddr is start of this line

	BEGIN							\ addr on stack
		dup EndingAddr <			\ not at end of range
	WHILE
		dup c@
		dup -> theChar
		dup  bl >					\ not a control character
		swap NBSpChar <>	AND		\ AND theChar <>  NBSP
		IF
			1 +to theCount			\ increment count
			1+						\ increment running address
			true
		ELSE
			ParseWordNoDraw
			theChar ParseSeparator
			swap	\ >R
			1+						\ increment running address
			0 -> theCount
			1 +to theAddr
			swap	\ R>
		THEN
		\ addr'\flag = true if ok to continue
		0= 
		IF
			theCount 0>
			IF
				-1 to theChar
				ParseWordNoDraw
			THEN
			exit
		THEN
	REPEAT
	theCount 0>
	IF
		-1 to theChar
		ParseWordNoDraw
	THEN
;

: DrawThisLineWithoutParsing ( addr -- addr' )
	0 -> theCount
	dup -> theAddr					\ theAddr is start of this line

	BEGIN							\ addr on stack
		dup EndingAddr <				\ not at end of range
	WHILE
		dup c@
		dup -> theChar				\ ??? needed
		dup  bl >
		swap NBSpChar <>	AND		\ AND theChar <>  NBSP
		IF
			1 +to theCount			\ increment count
			1+						\ increment running address
			true
		ELSE
			DrawWordNoParse
			theChar DrawSeparator
			swap	\ >R
			1+						\ increment running address
			0 -> theCount
			1 +to theAddr
			swap	\ R>
		THEN
		\ addr'\flag = true if ok to continue
		0=
		IF
			theCount 0>
			IF
				-1 to theChar
				DrawWordNoParse
			THEN
			exit
		THEN
	REPEAT
	theCount 0>
	IF
		-1 to theChar
		DrawWordNoParse
	THEN
;


: ScanFromStartToAddr ( EndingAddr -- )
\ this parses without drawing
\ use to calculate colorizer state at top line of visible area
	SE.DataRange +  dup to file.end
	min			\ ensure that addr is <= file.end 
	-> 	EndingAddr 
	se.handle  HLock
	InitColor
	SE.DataStart
	BEGIN					\ addr
		dup EndingAddr <		\
	WHILE
		ParseThisLineWithoutDraw ( addr -- addr' )
	REPEAT
	drop
	se.handle  HUnLock
;


\ *************************************************************
\ *************************************************************

\ ColorizerState is the basis of our somewhat involuted state machine

0 value ColorizerState
\ a few multiple states are used  like $0200:0000 + $0004:0000
\ 

\   $0000:0000  null state - no special color
\ .....................
\   $0000:00xx  short term effects
\	$0000:0001	affect this word only
\	$0000:0002	affect this word and next word
\	$0000:0003	affect this word and next 2 words
\ ...
\ .....................
\   $0000:xx00  one line effects
\	$0000:0100	  \  --\--  etc
\ ...
\ .....................
\   $0xxx:0000  multi-line effects
\	$0001:0000		in a quoted string
\	$0002:0000		in a ( .. )
\	$0004:0000		in a (*  *)
\	$0008:0000		in a ((  ))
\	$0010:0000		in a /*  */
\ ...
\	$0020:0000		in a [  ]
\	$0040:0000		in a locals|
\	$0080:0000		in a { } or {*  *}
\ ...
\	$0100:0000		in a |****  ****|
\	$0200:0000		in a [IF] ... clause
\ ...
\	$0800:0000		in an EXTERN: definition ( can terminate one of 2 ways )
\ .....................
\ ... 
\ ...
\	$8000:0000	    color changes disabled	- prior.stream  |********|  \\


: CanChangeColor? ( -- flag )
	ColorizerState 0 >=		\ EXcludes $8000:0000 -> $FFFF:FFFF
;


: ClearStateFlag ( nn -- )				\ Clears just the specified flag(s)
	-1 xor
	ColorizerState AND
	-> ColorizerState
;

: SetColorizerStateBits ( nn -- )		\ Sets just the specific flag(s)
	ColorizerState OR
	-> ColorizerState
;

: ColorizerStateClear? ( -- flag )
	ColorizerState 0= boolean
;


: InMultiLinePhrase? ( -- flag )
	ColorizerState $0FFF:0000 and boolean
; 


: AreWeInThisState? ( state -- flag | true if ok to change color )
	CanChangeColor?					\ be sure color not DISabled
	IF
		ColorizerState AND boolean	\ and we are in specified state
	ELSE
		drop
		false
	THEN
;

\ *************************************************************


0 value wasColorizerState
create	SavedTopForeColor 	6 allot&erase
create	SavedTopBackColor 	6 allot&erase
create	SavedStyle 			0 ,

create [IF]Depth  			0 ,  				\ handle nested [IF]



: SaveTopLineState ( -- ) 
	top.line @  		SE.oldtopline !
	ColorizerState      $FFFF:0000 and	SE.oldtoplineState !
	SE.wasTopForeColor	GetForeColor
	SE.wasTopBackColor	GetBackColor
	get.textstyle		SE.oldtopStyle !

	[IF]Depth @ 		SE.old[IF]Depth !
	SavedStyle @		SE.oldSavedStyle !
	SavedTopForeColor	SE.oldSavedTopForeColor 6 move
	SavedTopBackColor	SE.oldSavedTopBackColor 6 move
	wasColorizerState	SE.oldwasColorizerState !
;

: RestoreTopLineState ( -- ) 
	SE.oldtoplineState @ -> ColorizerState
	SE.wasTopForeColor 	 	RGBForecolor
	SE.wasTopBackColor 	 	RGBBackColor
	SE.oldtopStyle @ 	 	TextFace

	SE.old[IF]Depth @  		[IF]Depth !
	SE.oldSavedStyle @		SavedStyle !
	SE.oldSavedTopForeColor	SavedTopForeColor	 6 move
	SE.oldSavedTopBackColor	SavedTopBackColor	 6 move
	SE.oldwasColorizerState @ -> wasColorizerState
;



: DrawOnlyVisibleLines ( -- )		\ see <DRAW.ALL.LINES			\  ( EndingAddr -- )
	\ assume color and style and state setup correctly for top line
	\ and continue parsing the visible area
	SE.DataRange +  dup -> file.end  -> EndingAddr 
	0 -> CurrentLine#
	TextBounds ClipRect
	TextBounds EraseRect

	LEFT.EDGE+HBar ActiveBounds w@  se.lineheight + MoveTo

	se.handle  HLock

	SE.DataStart  top.line @ +
	BEGIN					\ addr
		dup EndingAddr <				\ not at end of se.buffer
		CurrentLine# #vislines @ <		\ AND still only visible lines
		AND
	WHILE
		ColorThisFile?
		IF
			DrawThisLineWhileParsing ( addr -- addr' )
		ELSE
			DrawThisLineWithoutParsing ( addr -- addr' )
		THEN
	REPEAT
	SE.DataStart  - screen.end !
	se.handle  HUnLock
;



variable DebugColor 		DebugColor off			\ enable color debugging
0 0 0 rgbcolor wasColor
0 value wasTextFace


: <DrawVisible ( -- ) 
	0 -> wasColorizerState
	ColorThisFile?
	IF
		[IF]Depth OFF
		wasColor GetForeColor
		get.textstyle -> wasTextFace
		top.line @  SE.oldtopline @ =
		IF
			RestoreTopLineState
			DrawOnlyVisibleLines
		ELSE
			DebugColor @ IF beep THEN				\ for testing - beep when we parse from the start
			SE.DataStart top.line @ +				\ Parse to top line
			ScanFromStartToAddr ( EndingAddr -- )
			SaveTopLineState
			DrawOnlyVisibleLines
		THEN
		wasColor RGBForecolor						\ restore for normal editor updates
		wasTextFace TextFace
	ELSE
		0 -> ColorizerState
		DrawOnlyVisibleLines
	THEN
;



: DrawVisible ( -- )
	skip.events @ >R
	skip.events  ON
	action-of pause >R
	['] noop is pause			\ disable tasks
	col @
	line# @
	se.window dup currentwindow ! outfile !
	line# !
	col !

	<DrawVisible

	col @
	line# @
	line# !
	col !
	
	R> is pause
	R>  skip.events !
;



\ *************************************************************
\ *************************************************************

: initColorizer ( -- )
	0 -> ColorizerState
	PlainTextColor RGBForecolor
	plain TextFace
\ can play with alternative fonts here ...
\	" Monaco"  GET.FONT.NUMBER _TextFont
\	" Courier"  GET.FONT.NUMBER _TextFont
\	" Courier New"  GET.FONT.NUMBER _TextFont
\	" Andale"  GET.FONT.NUMBER _TextFont
\	" VT100"  GET.FONT.NUMBER _TextFont
\	" Bitstream Vera Sans Mono" GET.FONT.NUMBER _TextFont
\	" Inconsolata" GET.FONT.NUMBER _TextFont
;

' initColorizer is InitColor

\ *************************************************************
\ *************************************************************

: SaveColorizerState ( -- )
	ColorizerState -> wasColorizerState
	SavedTopForeColor GetForeColor
	SavedTopBackColor GetBackColor
	get.textstyle	SavedStyle !
;

: ?SaveColorizerState ( -- )
	wasColorizerState 0=
	IF
		SaveColorizerState
	THEN
;

: RestoreColorizerState ( -- )
	wasColorizerState -> ColorizerState
	SavedTopForeColor 	RGBForecolor
	SavedTopBackColor 	RGBBackColor
	SavedStyle @ 	 	TextFace
;

: ?RestoreColorizerState ( -- )
	wasColorizerState
	IF
		RestoreColorizerState
		[IF]Depth @ 
		IF
			0 -> wasColorizerState		\ make it a one-shot
		THEN
	THEN
;
\ *************************************************************

\ create a special vocabulary for the words our parser can recognize
\ 1000 make.vocab  value ParserWID 		\ similar to WORDLIST

WORDLIST  value ParserWID 

: DecrementColorizerStateCount ( -- )
	CanChangeColor?
	\ InMultiLinePhrase?
	IF
		ColorizerState $ff and
		1- 0 max
		ColorizerState $ffff:ff00 AND
		OR
		dup -> ColorizerState
		0= 
		IF
			initColor 
		THEN

		OneShotColorFlag  \ restore state prior to one-shot before next word
		IF
			OneShotColorState -> ColorizerState
			TempSavedFontState  ApplyEditorFontStyle	\ restore face & color
			0 -> OneShotColorFlag						\ one-shot
		THEN

	THEN
;

' DecrementColorizerStateCount is AdjustOneShots


: HandleEOL ( -- )		\ possibly restore colorizing at end of line
	CanChangeColor?		\ ColorizerState 0>=
	InMultiLinePhrase? 0= AND
	IF
		initColor
	THEN
	[IF]Depth @ 								\ are we in an [IF] ?
	IF
		ColorizerState  $0000:0100 and			\ are we in a one-liner ?
		IF										\ then restore [IF] state & style
			?RestoreColorizerState
			0 -> wasColorizerState
		ELSE
			ColorizerState		$01FF:0000 and 0=	\ not currently in a multi-line
			[IF]Depth @ boolean
			and
			IF
				?RestoreColorizerState
\			ELSE
			THEN
		THEN
	ELSE
		ColorizerState  $0020:0000 and			\ are we in [  ..  ] ?
		IF
			?RestoreColorizerState
			0 -> wasColorizerState
		THEN
	THEN
;

' HandleEOL is ParseEOL ( -- )


: WordParser ( addr\cnt -- addr\cnt )	\ all the work is done in here
	PreParser
	IF
		2dup 
		ParserWID search-wordlist ?dup
		IF							\ found a 'special' word
			0>
			IF						\ check "immediate" flag - use carefully
									\ ignores ColorizerState
				execute 	( addr\cnt -- addr\cnt )
				PostParser	( addr\cnt -- addr\cnt )
			ELSE					\ not immediate
				CanChangeColor?
				IF
					execute		( addr\cnt -- addr\cnt )
					PostParser	( addr\cnt -- addr\cnt )
				ELSE
					drop			\ other color state overrides this
				THEN
			THEN
		ELSE						\ not in ParserWID
			ColorizerStateClear?	\ not in a special state already
			IF						\ handle words not in the special list
									\ possibly numbers
				OtherPostParser	( addr\cnt -- addr\cnt )
			THEN
		THEN
	THEN
;

' WordParser is ParseAWord ( addr\cnt -- addr\cnt )


\ *************************************************************
\ *************************************************************

\ some defining words for our parser vocabulary:

: CreatingWord:  ( "<spaces>name" -- )
\ can miscolor following word if inside a : .. ;
	create
	does>  
	drop
	ColorizerStateClear?
	IF
		2 SetColorizerStateBits			\ affect this word and the one following
		CreatorColor ApplyEditorFontStyle
	THEN
;

: StructureWord:  ( "<spaces>name" -- )
	create
	does>  drop
	ColorizerStateClear?
	IF
		2 SetColorizerStateBits			\ affect this word and the one following
		StructureColor ApplyEditorFontStyle
	THEN
;

: ConditionalWord:  ( "<spaces>name" -- )
	create
	does>  drop
	ColorizerStateClear?
	IF  1 SetColorizerStateBits		\ affect just this word
		ConditionalColor ApplyEditorFontStyle
	THEN
;

: MaybeWord:  ( "<spaces>name" -- )	( LACKING HAVING POSSIBLY )
	create
	does>  drop
	ColorizerStateClear?
	IF
		2 SetColorizerStateBits			\ affect this word and the one following
		[Conditional]Color ApplyEditorFontStyle
	THEN
;

: TickWord:  ( "<spaces>name" -- )		( ' ['] Postpone [COMPILE]  )
	create does>  drop
	ColorizerStateClear?
	IF
		2 SetColorizerStateBits			\ affect this word and the one following
		TickColor ApplyEditorFontStyle
	THEN
;

: VocabWord:  ( "<spaces>name" -- )
	create does>  drop
	ColorizerStateClear?
	IF
		1 SetColorizerStateBits			\ affect just this word
		VocabColor ApplyEditorFontStyle
	THEN
;

: LoopWord:  ( "<spaces>name" -- )
	create does>  drop
	ColorizerStateClear?
	IF  1 SetColorizerStateBits			\ affect just this word
		LoopColor ApplyEditorFontStyle
	THEN
;

: StartQuoteWord:  ( "<spaces>name" -- )
	create does>  drop
	ColorizerStateClear?
	IF 
		$0001:0000 SetColorizerStateBits
		QuoteColor ApplyEditorFontStyle
	THEN
;


: Start[Conditional]:  ( "<spaces>name" -- )
	create does>  drop
	ColorizerStateClear?
	IF
		$0200:0000 SetColorizerStateBits
		[Conditional]Color ApplyEditorFontStyle
	THEN
	ColorizerState    $0200:0000 =		\ are we in an [IF] ?
	IF
		1 [IF]Depth +!
	THEN
;


: OneColoredWord:  ( Textface\^RGB  "<spaces>name" --  )	\ set special color for next word
	create 
		here dup >R   8 allot
		6  move
		r> 6 + w! 
	does>
		ColorizerStateClear?
		IF
		\	RGBForecolor
			ApplyEditorFontStyle
			1 SetColorizerStateBits		\ affect just this word 
		ELSE
			drop
		THEN
;


: AlwaysColorMe: ( Textface\^rgbcolor "<spaces>name" -- )	\ one shot function to color this word
	\ Like OneColoredWord: except will also show its color inside comments etc
	create									
		here dup >R   8 allot
		6  move
		r> 6 + w! 
		IMMEDIATE							\ <------ why we need the one-shot
	does> 
		ColorizerState -> OneShotColorState
		true -> OneShotColorFlag				\ current state is flag
		TempSavedFontState dup GetForeColor
			get.textstyle swap 6 + w!
		ApplyEditorFontStyle
;


: SpecialCase[IF]Comment ( style\state --- )
	\ we want to retain comment colorizing in [IF] ... [THEN[ clauses
	\ we also do not want a commented out [THEN] to terminate the clause
	ColorizerState 	?dup
	IF								\ commented embedded in another state
		$0200:0000	=				\ in a [IF] clause
		[IF]Depth @ 0 > AND			\
		IF
			?SaveColorizerState		\ save current ColorizerState and style
			SetColorizerStateBits
			ApplyEditorFontStyle
		ELSE
			ColorizerState $0020:0000 =		\ in [  ... ]
			IF
				?SaveColorizerState		\ save current ColorizerState and style
				SetColorizerStateBits
				ApplyEditorFontStyle
			ELSE
				2drop
			THEN
		THEN
	ELSE							\ normal comment
		SetColorizerStateBits
		ApplyEditorFontStyle
	THEN
;


: StartOneLineComment ( -- )
	CommentColor  $0000:0100  SpecialCase[IF]Comment
;


: EndCompilableCodeArea ( -- )
	ColorizerState 0=
	IF
		$8000:0000	 SetColorizerStateBits
		MetaCommentColor ApplyEditorFontStyle 
	THEN
;

: StartLocals ( -- )
	ColorizerState 0=		\ CanChangeColor?
	IF
		$0040:0000 SetColorizerStateBits
		LocalsColor ApplyEditorFontStyle 
	THEN
;

: EndLocals ( -- )
	$0040:0000 AreWeInThisState?
	IF
		$0040:0000  ClearStateFlag
		1	SetColorizerStateBits
	THEN
;

: StartVFXLocals ( -- )
	ColorizerState 0=		\ CanChangeColor?
	IF
		$0080:0000 SetColorizerStateBits
		LocalsColor ApplyEditorFontStyle 
	THEN
;

: EndVFXLocals ( -- )
	$0080:0000 AreWeInThisState?
	IF
		$0080:0000  ClearStateFlag
		1	SetColorizerStateBits
	THEN
;

: handleEXTERN: 
	ColorizerStateClear?
	IF
		CreatorColor ApplyEditorFontStyle
		\ 1 SetColorizerStateBits		\ affect just this word 
		$0800:0001 -> ColorizerState
	THEN
;

\ *************************************************************
\ *************************************************************

\ Now we add some words to our special parser wordlist

only forth definitions
>m4th
also editorvocab
also forth also m4th
ParserWID set-current


 
: \
	StartOneLineComment
;

: // ( -- )
	StartOneLineComment
;

: --\-- 		\ one of ward's special metacomments - just like \
	\ StartOneLineComment
	SpecialCommentColor  $0000:0100  SpecialCase[IF]Comment
;


: ForbidColorizing	\ even commented, this turns colorizing off for whatever follows
	SaveColorizerState
	$8000:0000 -> ColorizerState
; immediate		\ <--- !!!!!!!!!!


\ ForbidColorizing  - THIS LINE MUST PRECEDE DEFINING MULTILINE PRIMITIVES !!!!
\					- or the colorizing of this file gets REALLY confused !!

: LOCALS|
	StartLocals ;

: |
	EndLocals ;

: {
	StartVFXLocals ;
	
: {:
	StartVFXLocals ;

: }
	EndVFXLocals ;

: :}
	EndVFXLocals ;


: EXTERN:
	handleEXTERN: ;

: EXTERN
	handleEXTERN: ;

: LOCALEXTERN:
	handleEXTERN: ;

: ALIASEDEXTERN:
	handleEXTERN: ;


: (	
	CommentColor $0002:0000  SpecialCase[IF]Comment
;

: );				\ end Extern:
	 $0800:0000 AreWeInThisState?
	 IF		\ continue as if  \ but in plain color
	 	CommentColor $0000:0100   -> ColorizerState ApplyEditorFontStyle
	 THEN
;

: )				\ normal comment terminator should be caught by preparser
	$0002:0000 AreWeInThisState?
	IF
		$0002:0000  ClearStateFlag 
	ELSE		\ external terminator
		 $0800:0000 AreWeInThisState?
		 IF		\ continue as if  \ but in plain color
	 		CommentColor $0000:0100  -> ColorizerState ApplyEditorFontStyle
		 THEN
	THEN
;
 
: [
	 InterpretedColor $0020:0000  SpecialCase[IF]Comment
;

: ]
	$0020:0000 AreWeInThisState?
	IF
		$0020:0000  ClearStateFlag
	THEN
;


: (*
	 MetaCommentColor $0004:0000  SpecialCase[IF]Comment
;

: *)				\ should be caught by preparser
	$0004:0000 AreWeInThisState?
	IF
		$0004:0000  ClearStateFlag
	THEN
;


: ((
	MetaCommentColor $0008:0000  SpecialCase[IF]Comment
;

: ))		 			\ should be caught by preparser 
	$0008:0000 AreWeInThisState?
	IF
		$0008:0000  ClearStateFlag
	THEN
;

: /*
	MetaCommentColor $0010:0000  SpecialCase[IF]Comment
;

: */
	$0010:0000 AreWeInThisState?
	IF
		$0010:0000  ClearStateFlag
	THEN
;

: |****			\ one of ward's special metacomments - like (*
	SpecialCommentColor $0100:0000  SpecialCase[IF]Comment
;

: ****|				\ one of ward's special metacomments - like *)
	$0100:0000 AreWeInThisState? 
	IF
		$0100:0000  ClearStateFlag
	THEN
;


: PRIOR.STREAM
	EndCompilableCodeArea
;

: |********|		\ one of ward's special metacomments - works like PRIOR.STREAM
	EndCompilableCodeArea
	SpecialCommentColor ApplyEditorFontStyle 
;

: \\
	EndCompilableCodeArea
;

: StopIncluding
	EndCompilableCodeArea
;

: "
	ColorizerState 0=
	IF
		$0001:0000 SetColorizerStateBits
		QuoteColor ApplyEditorFontStyle
	ELSE
		$0001:0000 AreWeInThisState?
		IF
			$0001:0000 ClearStateFlag
			1 SetColorizerStateBits		\ affect just this word
		THEN
	THEN
;

: CHAR
	ColorizerState 0=
	IF
		$0000:0002 SetColorizerStateBits
		QuoteColor ApplyEditorFontStyle
	THEN
;

\ restore normal colorizing - cancel ForbidColorizing

: AllowColorizing	\ even commented, this turns colorizing on for whatever follows
	0 -> ColorizerState
	?RestoreColorizerState
	0 -> wasColorizerState
	\ joy
; immediate		\ <--- !!!!!!!!!!

\ AllowColorizing	\ even commented, this turns colorizing back on


\ force ParserWID to be in Search Order temporarily
also 
ParserWID context !

: ASCII   char  ;
: [CHAR]  char  ;
\ : [CHARS] char  ;
previous


: [ELSE] 
	ColorizerStateClear?
	IF	\ cover the [ELSE] without [IF] condition
		HUH?Color ApplyEditorFontStyle
		$0000:0001  SetColorizerStateBits		\ affect just this word 
	THEN
;

: [THEN]
	$0200:0000  ColorizerState	= 			\ FALSE if in an [IF]
	IF
		[IF]Depth @ 1 - 0 max  dup [IF]Depth !	\ handle nesting
		0=
		IF \ beep
			$0200:0000  ClearStateFlag		\ leave [IF] state
			1	SetColorizerStateBits			\ apply to this word
		THEN
	THEN

	ColorizerStateClear?
	IF	\ cover the [THEN] without [IF] condition
		HUH?Color ApplyEditorFontStyle
		$0000:0001  SetColorizerStateBits		\ affect just this word 
	THEN
;


Start[Conditional]: [if]
Start[Conditional]: [ifdef]
Start[Conditional]: [ifndef]

\ QuoteWord: "
StartQuoteWord: ,"
StartQuoteWord: ."
StartQuoteWord: s"
StartQuoteWord: c"
StartQuoteWord: z"
StartQuoteWord: z",
StartQuoteWord: 0"
StartQuoteWord: 0,"
StartQuoteWord: abort"
StartQuoteWord: error"
StartQuoteWord: include"
StartQuoteWord: open"
StartQuoteWord: cdinclude"
StartQuoteWord: mfinclude"
\ StartQuoteWord: |"
\ StartQuoteWord: |0"

ConditionalWord: IF
ConditionalWord: ELSE
ConditionalWord: THEN
ConditionalWord: ENDIF
ConditionalWord: CASE
ConditionalWord: ENDCASE
ConditionalWord: BEGINCASE
ConditionalWord: NEXTCASE
ConditionalWord: NEXT-CASE
ConditionalWord: END-CASE
ConditionalWord: OF
ConditionalWord: ENDOF
ConditionalWord: RANGE.OF


\ give these a unique color?  
\ biggest issue is unpaired use, so maybe conditional color is adequate
ConditionalWord: >R
ConditionalWord: R>
ConditionalWord: R@
ConditionalWord: R>DROP
ConditionalWord: 2R>
ConditionalWord: 2>R
ConditionalWord: 2R@
ConditionalWord: RP@
\ ConditionalWord: DUP>R


MaybeWord: LACKING
MaybeWord: HAVING
MaybeWord: POSSIBLY

TickWord: [COMPILE]
TickWord: POSTPONE
TickWord: [']
TickWord: '
TickWord: IS
TickWord: EXISTS?
TickWord: [UNDEFINED]
TickWord: [DEFINED]
TickWord: HAVE


LoopWord: BEGIN
LoopWord: AGAIN
LoopWord: REPEAT
LoopWord: UNTIL
LoopWord: WHILE
LoopWord: #DO
LoopWord: DO
LoopWord: ?DO
LoopWord: LOOP
LoopWord: +LOOP
LoopWord: UNLOOP
LoopWord: I
LoopWord: J
LoopWord: EXIT
LoopWord: LEAVE

plain CommandColor OneColoredWord: :NONAME
\ CreatorColor OneColoredWord:  EXTERN:

underline CommandColor OneColoredWord: HEX
underline CommandColor OneColoredWord: DECIMAL
underline CommandColor OneColoredWord: BINARY
underline CommandColor OneColoredWord: IMMEDIATE


CreatingWord: :
CreatingWord: Code
CreatingWord: value
CreatingWord: FValue
CreatingWord: structure
CreatingWord: structure.end
CreatingWord: selector:
CreatingWord: ;selector
CreatingWord: defer
CreatingWord: variable
CreatingWord: 2variable
CreatingWord: Fvariable
CreatingWord: constant
CreatingWord: 2constant
CreatingWord: Fconstant

CreatingWord: create
CreatingWord: RGBColor
CreatingWord: New.Window
CreatingWord: NewTask
CreatingWord: user
CreatingWord: library:
having ?library:  CreatingWord: ?library:
having :MF		  CreatingWord: :MF

CreatingWord: XTERM:
CreatingWord: ENVMACRO:
CreatingWord: BUFFER:
\ CreatingWord: LOCALEXTERN:
\ CreatingWord: ALIASEDEXTERN:
CreatingWord: RUN:
CreatingWord: FIELD:
CreatingWord: CFIELD:
CreatingWord: MASM:
CreatingWord: LBL:
CreatingWord: CONSOLE:
CreatingWord: TEXTMACRO:
CreatingWord: OPERATOR:
CreatingWord: ALIAS:
CreatingWord: SYNONYM
CreatingWord: COMP:
CreatingWord: (:)
CreatingWord: REMEMBER:
CreatingWord: PRUNE:
CreatingWord: CALLPROC:
CreatingWord: CALLBACK:
CreatingWord: mfCHAIN
CreatingWord: marker
CreatingWord: anew


CreatingWord: CreatingWord:
CreatingWord: ConditionalWord:
CreatingWord: LoopWord:
CreatingWord: StartQuoteWord:
CreatingWord: Start[Conditional]:
CreatingWord: OneColoredWord:
CreatingWord: TickWord:
CreatingWord: MaybeWord:
CreatingWord: EditorFontStyle:
CreatingWord: StructureWord:
CreatingWord: VocabWord:
CreatingWord: +user
CreatingWord: mffield
CreatingWord: rect

StructureWord: byte:
StructureWord: bytes:
StructureWord: short:
StructureWord: shorts:
StructureWord: long:
StructureWord: longs:
StructureWord: string:
StructureWord: addr:
StructureWord: addrs:
StructureWord: point:
StructureWord: points:
StructureWord: ptr:
StructureWord: ptrs:
StructureWord: handle:
StructureWord: rect:
StructureWord: struct:
StructureWord: structs:
StructureWord: bytes.of:
StructureWord: one.of:
StructureWord: or.of:
StructureWord: xor.of:
StructureWord: float:
StructureWord: 2array:
StructureWord: 1array:

StructureWord: struct
StructureWord: end-struct
StructureWord: field
StructureWord: int
StructureWord: array-of
StructureWord: subrecord
StructureWord: end-subrecord
StructureWord: variant
StructureWord: end-variant
StructureWord: union
StructureWord: part
StructureWord: end-union
StructureWord: field-type
StructureWord: instance
StructureWord: OFFSET
StructureWord: BOFFSET
StructureWord: begin-structure
StructureWord: end-structure
StructureWord: +FIELD

VocabWord: previous
VocabWord: set-first-wordlist
VocabWord: also
VocabWord: get-order
VocabWord: set-order
VocabWord: minimum-order
VocabWord: only
VocabWord: default-order
VocabWord: get-current
VocabWord: set-current
VocabWord: get-first-wordlist
VocabWord: definitions
VocabWord: forth-wordlist
VocabWord: current
VocabWord: context
VocabWord: trunk
VocabWord: vocabulary
VocabWord: wordlist
VocabWord: ResetMinSearchOrder
VocabWord: >MIN-ORDER
VocabWord: -ORDER
VocabWord: +ORDER
VocabWord: >m4th
VocabWord: >vfx
VocabWord: >VFX+


underline RGBRed	AlwaysColorMe: VFX
underline RGBRed	AlwaysColorMe: MPE
underline RGBRed	AlwaysColorMe: Pelc
underline RGBBlue	AlwaysColorMe: MacForth
underline RGBBlue	AlwaysColorMe: MFonVFX
underline RGBBlue	AlwaysColorMe: MegaWolf
underline RGBBlue	AlwaysColorMe: McFarland

\ some for testing...
 
plain RGBYellow		AlwaysColorMe: AYellowWord
plain RGBRed		AlwaysColorMe: ARedWord 
plain RGBCyan		AlwaysColorMe: ACyanWord
plain RGBMagenta	AlwaysColorMe: AMagentaWord
plain RGBGreen		AlwaysColorMe: AGreenWord
plain rgbBrown		AlwaysColorMe: ABrownWord
plain rgbSkyBlue	AlwaysColorMe: ASkyBlueWord

plain RGBBlack		AlwaysColorMe: ABlackWord
plain RGBBlue		AlwaysColorMe: ABlueWord
plain RGBLtGray		AlwaysColorMe:	ALtGrayWord
plain RGBDkGray		AlwaysColorMe:	ADkGrayWord
plain rgbTangerine 	AlwaysColorMe: ATangerineWord
plain rgbSteelColor	AlwaysColorMe: ASteelColorWord
plain rgbAqua 		AlwaysColorMe: AnAquaWord
plain rgbMidnight	AlwaysColorMe: AMidnightWord
plain rgbVeryDkGrayColor	AlwaysColorMe: AVeryDkGrayWord
plain rgbSalmon		AlwaysColorMe: ASalmonWord
plain rgbOrange		AlwaysColorMe: AnOrangeWord
plain rgbOlive		AlwaysColorMe: AnOliveWord


\ *************************************************************
\ *************************************************************
\ Back to normal Forth
 

only forth definitions
>m4th
also editorvocab definitions
also forth also m4th		\ Forth searched first

((
: myPostParser	( addr\cnt -- addr\cnt )
	OneShotColorFlag
	IF	\ restore state prior to one-shot
		OneShotColorState -> ColorizerState
		TempSavedFontState  ApplyEditorFontStyle	\ restore face & color
		0 -> OneShotColorFlag						\ one-shot
	THEN
;

' myPostParser is PostParser
))

\ simple preparser to handle *) and )
\ maybe should be a chain?   slower that way, but would be easily extensible

: myPreParser ( addr\cnt -- addr\cnt\continueparsing? )
	0 locals| LastChar |
	\ if in quote,   check last character for "
	\ if in comment, check last character for terminator
	\ optional -- check for terminal : for MOPS

	OneShotColorFlag  \ restore state prior to one-shot before next word
	IF
		OneShotColorState -> ColorizerState
		TempSavedFontState  ApplyEditorFontStyle	\ restore face & color
		0 -> OneShotColorFlag						\ one-shot
	THEN

\ handle cases where no space before " or ) or *) 
	ColorizerState  
[	$0001:0000		\ in a quoted string
	$0002:0000 OR	\ in a ( .. )
	$0004:0000 OR	\ in a (*  *)
	$0008:0000 OR	\ in a (( ))
\	$0100:0000 +	\ in a |****  ****| - maybe - requiring preceding space is not egregious
]	literal
	AND
	IF
		2dup + 1- c@ -> LastChar 
		$0001:0000 AreWeInThisState?	\ in a " quoted string "
		IF
			LastChar ascii " =
			IF
				$0001:0000  ClearStateFlag
				1	SetColorizerStateBits
			THEN
		ELSE
			$0002:0000					\ in a ( comment )			
			AreWeInThisState?
			IF
				LastChar ascii ) =
				IF
					$0002:0000  ClearStateFlag
					?RestoreColorizerState			\ restored
					1	SetColorizerStateBits
				THEN
			ELSE
				$0004:0000				\ in a (* comment *)
				AreWeInThisState?
				IF
					2dup + 3 - 
					dup c@  bl =
					swap 1+ c@  ascii * =
					AND
					LastChar ascii ) =
					AND
					IF	
						$0004:0000  ClearStateFlag
						?RestoreColorizerState
						1 SetColorizerStateBits
					THEN
				ELSE
					$0008:0000				\ in a (( comment ))
					AreWeInThisState?
					IF
						2dup + 3 - 
						dup c@  bl =
						swap 1+ c@ ascii ) =
						AND
						LastChar ascii ) =
						AND
						IF	
							$0008:0000  ClearStateFlag
							?RestoreColorizerState
							1 SetColorizerStateBits
						THEN
					THEN
				THEN
			THEN
		THEN
	THEN
	true			\ for now, always continue parsing
;

' myPreParser is PreParser 


\ a simple OtherPostParser for colorizing words that end with a colon :
\ that do not fit other categories
: ?ObjectSelector ( addr\cnt -- addr\cnt )
	2dup + 1- c@ ascii : =
	IF
		SelectorColor ApplyEditorFontStyle 
		1 SetColorizerStateBits				\ affect just this word
	THEN
;

' ?ObjectSelector is OtherPostParser



(*
\ an experimental post parser for marking words that are not in any vocabulary
: testPostParser ( addr\cnt -- addr\cnt )
	over
	SE.DataStart  top.line @ +
	screen.end @ within
	IF	\ only do this for words on visible screen
		\ note: locals and numbers will be marked
		\		as well as misspelled and unknown words
		2dup
		2>R
		depth >R (allVocabs) depth R> -
		2R>
		#find-pair
		nip 0=
		IF
			HUH?Color ApplyEditorFontStyle	
		THEN
	THEN
;

' testPostParser is OtherPostParser
*)

\ *************************************************************
\ *************************************************************
\ *************************************************************

also editorvocab


: DrawLinesPossiblyColored ( -- )
	ColorThisFile?
	IF
		DrawVisible			\ 
	ELSE
		<DRAW.ALL.LINES
	THEN
;
\ this is the main editor display word that we must modify
 ' DrawLinesPossiblyColored  is DRAW.ALL.LINES



false value ColoredDrags		\ true makes things slower on big files

: <BOTTOM.ONE+ ( -- )
	<BOTTOM.ONE
	ColorThisFile?		\ +MISCINFO +SE.ColorFlag @
	ColoredDrags AND
	IF
		draw.screen
	THEN
;


: <TOP.ONE+ ( -- )
	<TOP.ONE
	ColorThisFile?
	ColoredDrags AND
	IF 
		draw.screen
	THEN
;

' <BOTTOM.ONE+ is BOTTOM.ONE
' <TOP.ONE+    is TOP.ONE



: ?RecolorScreen ( -- )
	ColorThisFile?
	IF
		TextBounds cliprect		\ ?? let editor set clip
		draw.screen
	
	ELSE						\ ?? move this to <DRAW.ALL.LINES					
		TextBounds ClipRect
		TextBounds EraseRect
		draw.screen
	THEN
;

' ?RecolorScreen  is UpdateAfterScroll
' ?RecolorScreen  is UpdateAfterRedrawFrom

: ?RecolorAfterChar ( char -- )
	bl <=
	ColorThisFile? AND
	IF
		TextBounds cliprect		\ whole screen
		draw.screen
	THEN
;
' ?RecolorAfterChar is AfterCharInsert


: ?RecolorAfterInsert ( addr\cnt -- )
	\ ? recalculate clip?
	2drop 
	?RecolorScreen
;

' ?RecolorAfterInsert  is  AfterRangeInsert

\ ' ?RecolorScreen  is AfterInsert				\ ?? needed

' ?RecolorScreen  is UpdateAfterFlush			\ ?? needed


\ *************************************************************
\ Housekeeping stuff...

: ForgetColorizer ( -- )
\	ParserWID to.heap
; on.forget ForgetColorizer



